(* Theorema 
    Copyright (C) 2010 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`Provers`ReductionRingProver`"]

Needs[ "Theorema`Provers`"]
Needs[ "Theorema`Common`"]
Needs[ "Theorema`Language`"]
Needs[ "Theorema`Provers`RewriteInteractiveProver`"]

Begin["`Private`"]

(* TODO:
? Enhance ordering rules s.t. "a>b || b>a || a==b" can be shown directly for total orderings.
? The "add abbreviation"-rule might be useful when, e.g., "(a+b)*c==c*(a+b)" shall be proved only using commutativity of +,
	which means that only domain-membership of "c" and "a+b" has to be checked. At the moment, both sides of the equality are fully
	expanded, and hence domain-membership of "a", "b" and "c" has to be checked.
	However, "a+b" could (temporarily) be abbreviated by some "x".
*)

(* ::Section:: *)
(* Auxiliary Functions *)

Options[ checkDomainMembership] = {"marker" -> "C"};
checkDomainMembership[ terms_List, dom_, rules_List, opts___?OptionQ] :=
	Module[ {allRefs = {}, allConds = And$TM[], tmp, f},
		Scan[ (
				tmp = followsImmediately[ f = DomainOperation$TM[ dom, Element$TM][ #], rules];
				If[ MatchQ[ tmp, {_List, {}}],
					allRefs = Join[ allRefs, First[ tmp]],
					AppendTo[ allConds, f]
				]
			)&,
			terms
		];
		allConds = If[ Length[ allConds] > 0,
						makeRRPFML[ formula -> allConds, "marker" -> ("marker" /. {opts} /. Options[ checkDomainMembership])],
						Null
					];
		{allRefs, allConds}
	]


(* ::Section:: *)
(* Orderings *)

inferenceRule[ RRPorderingGoal] =
this:PRFSIT$[ g_FML$, k_List, _, rest___] :> performProofStep[
	Module[ {rw = getOptionalComponent[ this, "rewriting"]},
		orderingGoal[ g, k, getRulesFromRW[ rw, "type" -> "Backward", "ground" -> $backwardGround],
			{Theorema`Knowledge`isPartIrreflOrder$TM, Theorema`Knowledge`isPartReflOrder$TM, Theorema`Knowledge`isTotalIrreflOrder$TM},
			RRPorderingGoal,
			rest
		]
	]	
]

orderingGoal[ g:FML$[ _, Not$TM[ (r:Except[ Equal$TM])[ x_, x_]], __], k_List, gRules_List, types_List, n_, rest___] :=
	Module[ {type, fml, dom, domRefs, domFml},
		(
			dom = (formula@fml)[[2]];
			{domRefs, domFml} = checkDomainMembership[ {x}, dom, gRules, "marker" -> "G"];
			If[ domFml === Null,
				makeTERMINALNODE[ makePRFINFO[ name -> n, used -> {g, fml}, generated -> {}, "data" -> {domRefs, "irreflNot"}],
									proved
				],
				
				makeANDNODE[ makePRFINFO[ name -> n, used -> {g, fml}, generated -> {domFml}, "data" -> {x, dom, "irreflNot"}],
					RRPToBeProved[ goal -> domFml, kb -> k, rest]
				]
			]
		) /; ({fml, type} = Catch[ Scan[ With[ {f = Cases[ k, FML$[ _, #[ r, _], __], {1}, 1]},
									If[ f =!= {},
										Throw[ Append[ f, #]]
									]]&,
									Intersection[ types, {Theorema`Knowledge`isPartIrreflOrder$TM, Theorema`Knowledge`isTotalIrreflOrder$TM}]
								];
								Throw[ {$Failed, $Failed}]
							];
			type =!= $Failed)
	]
orderingGoal[ g:FML$[ _, (Not$TM[ (r:Except[ Equal$TM])[ fun_[ a_, b_], fun_[ x_, y_]]])|((r:Except[ Equal$TM])[ fun_[ a_, b_], fun_[ x_, y_]]), __], k_List, gRules_List, _List, n_, rest___] :=
	Module[ {fml, dom, domRefs, domFml, newG},
		performProofStep[
			fml = First[ fml];
			dom = (formula@fml)[[3]];
			newG = Which[
				a === x,
				r[ b, y],
				
				b === y,
				r[ a, x],
				
				True,
				$Failed
			];
			If[ newG === $Failed,
				$Failed,
				
				newG = makeRRPFML[ formula -> Switch[ formula@g, _Not$TM, Not$TM[ newG], _, newG], "marker" -> "G"];
				{domRefs, domFml} = checkDomainMembership[ DeleteDuplicates[ {a, b, x, y}], dom, gRules];
				If[ domFml === Null,
					makeANDNODE[ makePRFINFO[ name -> n, used -> {g, fml}, generated -> {newG}, "data" -> {domRefs, "orderEmbedding"}],
						RRPToBeProved[ goal -> newG, kb -> k, rest]
					],
					
					makeANDNODE[ makePRFINFO[ name -> n, used -> {g, fml}, generated -> {newG, domFml}, "data" -> {dom, domRefs, "orderEmbedding"}],
						{RRPToBeProved[ goal -> domFml, kb -> k, rest], RRPToBeProved[ goal -> newG, kb -> addNewKB[ k, domFml], rest]}
					]
				]
			]
		] /; ((fml = Cases[ k, FML$[ _, Theorema`Knowledge`isOrderEmbedding$TM[ fun, r, _], __], {1}, 1]) =!= {})
	]
orderingGoal[ g:FML$[ _, Not$TM[ (r:Except[ Equal$TM])[ x_, y_]], __], k_List, gRules_List, types_List, n_, rest___] :=
	Module[ {type, fml, dom, domRefs, domFml, transRefs, transElems, newG},
		(* TODO: Maybe it is reasonable to replace !(x < y) by (y < x || x == y) in case of a total ordering. *)
		(
			dom = (formula@fml)[[2]];
			Switch[ type,
				Theorema`Knowledge`isPartIrreflOrder$TM|Theorema`Knowledge`isTotalIrreflOrder$TM,
				{domRefs, domFml} = checkDomainMembership[ Join[ {x, y}, transElems], dom, gRules];
				If[ domFml === Null,
					makeTERMINALNODE[ makePRFINFO[ name -> n, used -> {g, fml}, generated -> {},
										"data" -> {dom, r, y, x, transElems, transRefs, domRefs, "transAsymNot"}],
									proved
					],
					
					makeANDNODE[ makePRFINFO[ name -> n, used -> {g, fml}, generated -> {domFml},
										"data" -> {dom, r, y, x, transElems, transRefs, domRefs, "transAsymNot"}],
						RRPToBeProved[ goal -> domFml, kb -> k, rest]
					]
				],
				
				Theorema`Knowledge`isPartReflOrder$TM,
				newG = makeRRPFML[ formula -> Not$TM[ Equal$TM[ x, y]], "marker" -> "G"];
				{domRefs, domFml} = checkDomainMembership[ Join[ {x, y}, transElems], dom, gRules];
				If[ domFml === Null,
					makeANDNODE[ makePRFINFO[ name -> n, used -> {g, fml}, generated -> {newG},
										"data" -> {dom, r, y, x, transElems, transRefs, domRefs, "transAntisymNot"}],
						RRPToBeProved[ goal -> newG, kb -> k, rest]
					],
					
					makeANDNODE[ makePRFINFO[ name -> n, used -> {g, fml}, generated -> {newG, domFml},
										"data" -> {dom, r, y, x, transElems, transRefs, domRefs, "transAntisymNot"}],
						{RRPToBeProved[ goal -> domFml, kb -> k, rest], RRPToBeProved[ goal -> newG, kb -> addNewKB[ k, domFml], rest]}
					]
				]
			]
		) /; ({fml, type} = Catch[ Scan[ With[ {f = Cases[ k, FML$[ _, #[ r, _], __], {1}, 1]},
									If[ f =!= {},
										Throw[ Append[ f, #]]
									]]&,
									types
								];
								Throw[ {$Failed, $Failed}]
							];
			type =!= $Failed && ({transRefs, transElems} = findChain[ y, x, r, k, gRules]; transRefs =!= {}))
	]
orderingGoal[ g:FML$[ _, (r:Except[ Equal$TM])[ x_, x_], __], k_List, gRules_List, {___, Theorema`Knowledge`isPartReflOrder$TM, ___}, n_, rest___] :=
	Module[ {fml, dom, domRefs, domFml},
		(
			fml = First[ fml];
			dom = (formula@fml)[[2]];
			{domRefs, domFml} = checkDomainMembership[ {x}, dom, gRules, "marker" -> "G"];
			If[ domFml === Null,
				makeTERMINALNODE[ makePRFINFO[ name -> n, used -> {g, fml}, generated -> {}, "data" -> {domRefs, "refl"}],
									proved
				],
				
				makeANDNODE[ makePRFINFO[ name -> n, used -> {g, fml}, generated -> {domFml}, "data" -> {x, dom, "refl"}],
					RRPToBeProved[ goal -> domFml, kb -> k, rest]
				]
			]
		) /; ((fml = Cases[ k, FML$[ _, Theorema`Knowledge`isPartReflOrder$TM[ r, _], __], {1}, 1]) =!= {})
	]
orderingGoal[ g:FML$[ _, (r:Except[ Equal$TM])[ x_, y_], __], k_List, gRules_List, types_List, n_, rest___] :=
	Module[ {type, fml, dom, domRefs, domFml, transRefs, transElems},
		(
			{transRefs, transElems} = findChain[ x, y, r, k, gRules];
			If[ transRefs =!= {},
				dom = (formula@fml)[[2]];
				{domRefs, domFml} = If[ transElems === {},
										{{}, Null},
										checkDomainMembership[ Join[ {x, y}, transElems], dom, gRules, "marker" -> "G"]
									];
				If[ domFml === Null,
					makeTERMINALNODE[ makePRFINFO[ name -> n, used -> {g, fml}, generated -> {},
											"data" -> {dom, r, y, x, transElems, transRefs, domRefs, "trans"}],
										proved
					],
					
					makeANDNODE[ makePRFINFO[ name -> n, used -> {g, fml}, generated -> {domFml},
											"data" -> {dom, r, y, x, transElems, transRefs, domRefs, "trans"}],
						RRPToBeProved[ goal -> domFml, kb -> k, rest]
					]
				]
			]
		) /; ({fml, type} = Catch[ Scan[ With[ {f = Cases[ k, FML$[ _, #[ r, _], __], {1}, 1]},
										If[ f =!= {},
											Throw[ Append[ f, #]]
										]]&,
										types
									];
									Throw[ {$Failed, $Failed}]
								];
			type =!= $Failed)
	]
orderingGoal[ ___] := $Failed
	
findChain[ x_, y_, r_, k_List, rules_List] := findChain[ x, y, r, k, rules, True]
findChain[ x_, y_, r_, k_List, rules_List, expandRight_] :=
	Module[ {refs, elems},
		refs = followsImmediately[ r[ x, y], rules];
		If[ MatchQ[ refs, {_List, {}}],
			{First[ refs], {}},
			
			Catch[
				Scan[
					With[ {f = First[ #], z = Last[ #]},
						{refs, elems} = findChain[ z, y, r, k, rules, False];
						If[ refs =!= {},
							Throw[ {Prepend[ refs, f], Prepend[ elems, z]}]
						]
					]&,
					Cases[ k, f:(FML$[ _, r[ x, z_], __]) :> {f, z}]
				];
				If[ expandRight,
					Scan[
						With[ {f = First[ #], z = Last[ #]},
							{refs, elems} = findChain[ x, z, r, k, rules, True];
							If[ refs =!= {},
								Throw[ {Append[ refs, f], Append[ elems, z]}]
							]
						]&,
						Cases[ k, f:(FML$[ _, r[ z_, y], __]) :> {f, z}]
					]
				];
				Throw[ {{}, {}}]
			]
		]
	]
	
inferenceRule[ RRPorderingKB] =
this:PRFSIT$[ g_FML$, k_List, _, rest___] :> performProofStep[
	Module[ {rw = getOptionalComponent[ this, "rewriting"]},
		orderingKB[ g, k, getRulesFromRW[ rw, "type" -> "Backward", "ground" -> $backwardGround],
			{Theorema`Knowledge`isPartIrreflOrder$TM, Theorema`Knowledge`isPartReflOrder$TM, Theorema`Knowledge`isTotalIrreflOrder$TM},
			RRPorderingKB,
			rest
		]
	]	
]

orderingKB[ g_FML$, {pre___, form:FML$[ _?activatedQ, Not$TM[ (r:Except[ Equal$TM])[ x_, x_]], __], post___}, gRules_List, {___, Theorema`Knowledge`isPartReflOrder$TM, ___}, n_, rest___] :=
	Module[ {fml, dom, domRefs, domFml},
		(
			fml = First[ fml];
			dom = (formula@fml)[[2]];
			{domRefs, domFml} = checkDomainMembership[ {x}, dom, gRules, "marker" -> "G"];
			If[ domFml === Null,
				makeTERMINALNODE[ makePRFINFO[ name -> n, used -> {form, fml}, generated -> {}, "data" -> {domRefs, "reflNot"}], proved],
				
				makeANDNODE[ makePRFINFO[ name -> n, used -> {form, fml}, generated -> {domFml}, "data" -> "reflNot"],
					RRPToBeProved[ goal -> domFml, kb -> {pre, post}, rest]
				]
			]
		) /; ((fml = Cases[ {pre, post}, FML$[ _, Theorema`Knowledge`isPartReflOrder$TM[ r, _], __], {1}, 1]) =!= {})
	]
orderingKB[ g_FML$, {pre___, form:FML$[ _?activatedQ, (r:Except[ Equal$TM])[ x_, x_], __], post___}, gRules_List, types_List, n_, rest___] :=
	Module[ {type, fml, dom, domRefs, domFml, thinnedKB = {pre, post}},
		(
			dom = (formula@fml)[[2]];
			{domRefs, domFml} = checkDomainMembership[ {x}, dom, gRules, "marker" -> "G"];
			If[ domFml === Null,
				makeTERMINALNODE[ makePRFINFO[ name -> n, used -> {form, fml}, generated -> {}, "data" -> {domRefs, "irrefl"}], proved],
				
				makeANDNODE[ makePRFINFO[ name -> n, used -> {form, fml}, generated -> {domFml}, "data" -> "irrefl"],
					RRPToBeProved[ goal -> domFml, kb -> thinnedKB, rest]
				]
			]
		) /; ({fml, type} = Catch[ Scan[ With[ {f = Cases[ thinnedKB, FML$[ _, #[ r, _], __], {1}, 1]},
										If[ f =!= {},
											Throw[ Append[ f, #]]
										]]&,
										Intersection[ types, {Theorema`Knowledge`isPartIrreflOrder$TM, Theorema`Knowledge`isTotalIrreflOrder$TM}]
									];
									Throw[ {$Failed, $Failed}]
								];
			type =!= $Failed)
	]
orderingKB[ ___] := $Failed
	
inferenceRule[ RRPorderingEqualGoal] = 
this:PRFSIT$[ g:FML$[ _, Equal$TM[ x_, y_], __], k_List, _, rest___] :>
	Module[ {t = {Theorema`Knowledge`isPartReflOrder$TM, Theorema`Knowledge`isTotalIrreflOrder$TM},
				rw, gRules, tp, r, fml, dom, domRefs, domFml, xknown = False, newG},
		performProofStep[
			rw = getOptionalComponent[ this, "rewriting"];
			gRules = getRulesFromRW[ rw, "type" -> "Backward", "ground" -> $backwardGround];
			r = (formula@fml)[[1]];
			{t, domFml} = checkDomainMembership[ {If[ xknown, y, x]}, dom, gRules];
			domRefs = Join[ domRefs, t];
			newG = makeRRPFML[ formula -> Switch[ tp,
													Theorema`Knowledge`isPartReflOrder$TM,
													And$TM[ r[ x, y], r[ y, x]],
													_,
													And$TM[ Not$TM[ r[ x, y]], Not$TM[ r[ y, x]]]
											],
								"marker" -> "G"
					];
			If[ domFml === Null,
				makeANDNODE[ makePRFINFO[ name -> RRPorderingEqualGoal, used -> {g, fml, First[ domRefs]}, generated -> {newG},
											"data" -> {Rest[ domRefs], dom, tp, r}],
					RRPToBeProved[ goal -> newG, kb -> k, rest]
				],
				
				makeANDNODE[ makePRFINFO[ name -> RRPorderingEqualGoal, used -> {g, fml, First[ domRefs]}, generated -> {newG, domFml},
											"data" -> {dom, tp, r}],
					{RRPToBeProved[ goal -> domFml, kb -> k, rest], RRPToBeProved[ goal -> newG, kb -> addNewKB[ k, domFml], rest]}
				]
			]
		] /; (
				domRefs = Cases[ k, FML$[ _, DomainOperation$TM[ d_, Element$TM][ x] /; MemberQ[ k, FML$[ _, (Alternatives@@t)[ _, d], __]], __], {1}, 1];
				If[ domRefs =!= {},
					xknown = True,
					domRefs = Cases[ k, FML$[ _, DomainOperation$TM[ d_, Element$TM][ y] /; MemberQ[ k, FML$[ _, (Alternatives@@t)[ _, d], __]], __], {1}, 1]
				];
				If[ domRefs =!= {},
					dom = (formula@First[ domRefs])[[0, 1]];
					fml = {};
					Scan[ (fml = Join[ fml, Cases[ k, FML$[ _, #[ _, dom], __]]])&, t];
					Switch[ fml,
						{},	(* the domain is not ordered *)
						False,
						
						{_},	(* the domain is ordered by exactly 1 suitable order relation *)
						fml = First[ fml];
						tp = Head[ formula@fml];
						True,
						
						_,		(* the domain is ordered by several suitable order relations *)
						fml = Cases[ fml, FML$[ _?activatedQ, __]];
						If[ Length[ fml] === 1,
							fml = First[ fml];
							tp = Head[ formula@fml];
							True,
							False
						]
					]
				]
			)
	]
	
	
(* ::Section:: *)
(* DomainSets and DomainTuples *)

inferenceRule[ RRPmembershipDomainTuples] =
this:PRFSIT$[ g:FML$[ _, fml:((DomainOperation$TM[ dom_, Element$TM][ expr_])|(isTuple$TM[ expr_])|_And$TM), __], k_List, _, rest___] :> performProofStep[
	Module[ {refs = $Failed, conds, rules},
		Switch[ fml,
			_And$TM,
			If[ TrueQ[ $applyAllPossibleInferences],
				Block[ {$proposedRules = {{RRPforallGoal, "Repeated"}, {RRPmembershipDomainTuples}}},
					Replace[ this, {inferenceRule[ RRPandGoal], _ -> $Failed}]
				],
				$Failed
			],
			
			_isTuple$TM,
			refs = Cases[ k, FML$[ _, DomainOperation$TM[ Theorema`Knowledge`DomainTuples$TM[ _], Element$TM][ expr],__], {1}, 1];
			If[ refs === {},
				$Failed,
				makeTERMINALNODE[ makePRFINFO[ name -> RRPmembershipDomainTuples, used -> {{g}, refs}, generated -> {}, "domMembership" -> False], proved]
			],
			
			_,
			rules = getRulesFromRW[ getOptionalComponent[ this, "rewriting"], "type" -> "Backward", "ground" -> $backwardGround];
			
			(* General Part *)
			If[ MatchQ[ expr, Subscript$TM[ _, _]],
				refs = followsImmediately[ DomainOperation$TM[ Theorema`Knowledge`DomainTuples$TM[ dom], Element$TM][ First[ expr]], rules];
				{refs, conds} = If[ MatchQ[ refs, {_List, {}}],
							{First[ refs], rngToCondition[ RNG$[ STEPRNG$[ Last[ expr], 1, BracketingBar$TM[ First[ expr]], 1]]]},
							{$Failed, $Failed}
						]
			];
			If[ refs === $Failed,
				{refs, conds} = Catch[ 
					Scan[
						With[ {tmp = followsImmediately[ DomainOperation$TM[ Theorema`Knowledge`DomainTuples$TM[ dom], Element$TM][ Last[ #]], rules]}, 
							If[ MatchQ[ tmp, {_List, {}}], 
								Throw[ {Prepend[ First[ tmp], First[ #]], {}}]
							]
						]&,
						Cases[ k, f:(FML$[ _, elemTuple$TM[ expr, t_], __]) :> {f, t}]
					];
					Throw[ {$Failed, $Failed}]
				]
			];
			
			(* Special Part *)
			If[ refs === $Failed && MatchQ[ dom, Theorema`Knowledge`DomainTuples$TM[ _]],
				{refs, conds} = tuplesDM[ expr, dom, rules];
				If[ conds === {fml}, refs = $Failed]
			];
			
			If[ refs === $Failed,
				$Failed,
				
				refs = DeleteDuplicates[ refs];
				conds = DeleteDuplicates[ conds];
				If[ conds === {},
					makeTERMINALNODE[ makePRFINFO[ name -> RRPmembershipDomainTuples, used -> {{g}, refs}, generated -> {}], proved],
					
					conds = makeRRPFML[ formula -> makeConjunction[ conds, And$TM], "marker" -> "G"];
					makeANDNODE[ makePRFINFO[ name -> RRPmembershipDomainTuples, used -> {{g}, refs}, generated -> {conds}],
						RRPToBeProved[ goal -> conds, kb -> k, rest]
					]
				]
			]
		]
	]
]

inferenceRule[ RRPmembershipDomainSets] =
this:PRFSIT$[ g:FML$[ _, fml:((DomainOperation$TM[ dom_, Element$TM][ expr_])|(isSet$TM[ expr_])|_And$TM), __], k_List, _, rest___] :> performProofStep[
	Module[ {refs = $Failed, conds, rules},
		Switch[ fml,
			_And$TM,
			If[ TrueQ[ $applyAllPossibleInferences],
				Block[ {$proposedRules = {{RRPforallGoal, "Repeated"}, {RRPmembershipDomainSets}}},
					Replace[ this, {inferenceRule[ RRPandGoal], _ -> $Failed}]
				],
				$Failed
			],
			
			_isSet$TM,
			refs = Cases[ k, FML$[ _, DomainOperation$TM[ Theorema`Knowledge`DomainSets$TM[ _], Element$TM][ expr],__], {1}, 1];
			If[ refs === {},
				$Failed,
				makeTERMINALNODE[ makePRFINFO[ name -> RRPmembershipDomainSets, used -> {{g}, refs}, generated -> {}, "domMembership" -> False], proved]
			],
			
			_,
			rules = getRulesFromRW[ getOptionalComponent[ this, "rewriting"], "type" -> "Backward", "ground" -> $backwardGround];
			
			(* General Part *)
			{refs, conds} = Catch[
				Scan[
					With[ {tmp = followsImmediately[ DomainOperation$TM[ Theorema`Knowledge`DomainSets$TM[ dom], Element$TM][ Last[ #]], rules]}, 
						If[ MatchQ[ tmp, {_List, {}}], 
							Throw[ {Prepend[ First[ tmp], First[ #]], {}}]
						]
					]&,
					Cases[ k, f:(FML$[ _, Element$TM[ expr, s_], __]) :> {f, s}]
				];
				Throw[ {$Failed, $Failed}]
			];
			
			(* Special Part *)
			If[ refs === $Failed && MatchQ[ dom, Theorema`Knowledge`DomainSets$TM[ _]],
				{refs, conds} = setsDM[ expr, dom, rules];
				If[ conds === {fml}, refs = $Failed]
			];
			
			If[ refs === $Failed,
				$Failed,
				
				refs = DeleteDuplicates[ refs];
				conds = DeleteDuplicates[ conds];
				If[ conds === {},
					makeTERMINALNODE[ makePRFINFO[ name -> RRPmembershipDomainSets, used -> {{g}, refs}, generated -> {}], proved],
					
					conds = makeRRPFML[ formula -> makeConjunction[ conds, And$TM], "marker" -> "G"];
					makeANDNODE[ makePRFINFO[ name -> RRPmembershipDomainSets, used -> {{g}, refs}, generated -> {conds}],
						RRPToBeProved[ goal -> conds, kb -> k, rest]
					]
				]
			]
		]
	]
]


(* ::Subsection:: *)
(* Auxiliary Functions *)

tuplesDM[ expr_, dom_, rules_List] :=
	Module[ {out},
		out /; MatchQ[ out = followsImmediately[ DomainOperation$TM[ dom, Element$TM][ expr], rules], {_List, {}}]
	]
tuplesDM[ joinTuples$TM[ a___], dom_, rules_] :=
	Module[ {allRefs = {}, allConds = {}, ref, cond},
		Scan[ (
				{ref, cond} = tuplesDM[ #, dom, rules];
				allRefs = Join[ allRefs, ref];
				allConds = Join[ allConds, cond]
			)&,
			{a}
		];
		{allRefs, allConds}
	]
tuplesDM[ appendElem$TM[ t_, a_], dom:(_[ inner_]), rules_] :=
	Module[ {refs, conds},
		{refs, conds} = tuplesDM[ t, dom, rules];
		{refs, Append[ conds, DomainOperation$TM[ inner, Element$TM][ a]]}
	]
tuplesDM[ prependElem$TM[ a_, t_], dom:(_[ inner_]), rules_] :=
	Module[ {refs, conds},
		{refs, conds} = tuplesDM[ t, dom, rules];
		{refs, Append[ conds, DomainOperation$TM[ inner, Element$TM][ a]]}
	]
tuplesDM[ (Rest$TM|Most$TM)[ t_], dom_, rules_] :=
	Module[ {refs, conds},
		{refs, conds} = tuplesDM[ t, dom, rules];
		{refs, Append[ conds, Not$TM[ Equal$TM[ t, Tuple$TM[]]]]}
	]
tuplesDM[ Reverse$TM[ t_], dom_, rules_] :=
	tuplesDM[ t, dom, rules]
tuplesDM[ Tuple$TM[ a___], _[ inner_], _] :=
	{{}, Map[ DomainOperation$TM[ inner, Element$TM][ #]&, {a}]}
tuplesDM[ TupleOf$TM[ r:RNG$[ STEPRNG$[ i_, a_, b_, 1]], cond_, term_], dom:(_[ inner_]), rules_] :=
	Module[ {c2, t2, const},
		If[ const,
			{{}, {Element$TM[ a, IntegerInterval$TM[ DirectedInfinity[ -1], DirectedInfinity[ 1], False, False]],
					Element$TM[ b, IntegerInterval$TM[ DirectedInfinity[ -1], DirectedInfinity[ 1], False, False]],
					DomainOperation$TM[ inner, Element$TM][ term]}},
					
			tuplesDM[ TupleOf$TM[ r, c2, t2], dom, rules]
		] /;
		(
			{c2, t2} = substituteFree[ {cond, term}, {i -> "$rngVar$"}];
			If[ FreeQ[ {c2, t2}, "$rngVar$"],
				const = True;
				True,
				
				const = False;
				FreeQ[ {c2, t2} = ({c2, t2} /. Subtract$TM[ b, Subtract$TM[ "$rngVar$", a]] -> i), "$rngVar$"]
			]
		)
	]
tuplesDM[ TupleOf$TM[ RNG$[ STEPRNG$[ i_, 1, BracketingBar$TM[ T_], 1]], _, Subscript$TM[ T_, i_]], dom_, rules_] :=
	tuplesDM[ T, dom, rules]
tuplesDM[ TupleOf$TM[ r:RNG$[ STEPRNG$[ _, a_, b_, 1]], cond_, term_], _[ inner_], _] :=
	{{}, {Element$TM[ a, IntegerInterval$TM[ DirectedInfinity[ -1], DirectedInfinity[ 1], False, False]],
			Element$TM[ b, IntegerInterval$TM[ DirectedInfinity[ -1], DirectedInfinity[ 1], False, False]],
			Forall$TM[ r, cond, DomainOperation$TM[ inner, Element$TM][ term]]}}
tuplesDM[ expr_, _[ inner_], rules_List] :=
	Module[ {i = VAR$[ Unique[ Theorema`Knowledge`VAR$i$TM]], tmp},
		{First[ tmp], {isTuple$TM[ expr]}}
			/; MatchQ[ tmp = followsImmediately[
								Forall$TM[ RNG$[ STEPRNG$[ i, 1, BracketingBar$TM[ expr], 1]], True, DomainOperation$TM[ inner, Element$TM][ Subscript$TM[ expr, i]]],
								rules
							],
						{_List, {}}
				]
	]
tuplesDM[ expr_, dom_, _] := {{}, {DomainOperation$TM[ dom, Element$TM][ expr]}}

setsDM[ expr_, dom_, rules_List] :=
	Module[ {out},
		out /; MatchQ[ out = followsImmediately[ DomainOperation$TM[ dom, Element$TM][ expr], rules], {_List, {}}]
	]
setsDM[ (Union$TM|Intersection$TM)[ a___], dom_, rules_] :=
	Module[ {allRefs = {}, allConds = {}, ref, cond},
		Scan[ (
				{ref, cond} = setsDM[ #, dom, rules];
				allRefs = Join[ allRefs, ref];
				allConds = Join[ allConds, cond]
			)&,
			{a}
		];
		{allRefs, allConds}
	]
setsDM[ Set$TM[ a___], _[ inner_], _] :=
	{{}, Map[ DomainOperation$TM[ inner, Element$TM][ #]&, {a}]}
setsDM[ Theorema`Knowledge`TupleToSet$TM[ T_], _[ inner_], _] :=
	{{}, {DomainOperation$TM[ Theorema`Knowledge`DomainTuples$TM[ inner], Element$TM][ T]}}
setsDM[ SetOf$TM[ RNG$[ SETRNG$[ x_, A_]], _, x_], dom_, rules_] :=
	setsDM[ A, dom, rules]
setsDM[ SetOf$TM[ r:RNG$[ SETRNG$[ x_, A_]], cond_, term_], _[ inner_], _] :=
	{{}, {isSet$TM[ A], Forall$TM[ r, cond, DomainOperation$TM[ inner, Element$TM][ term]]}}
setsDM[ SetOf$TM[ r:RNG$[ PREDRNG$[ x_, DomainOperation$TM[ dom_, Element$TM]]], cond_, term_], _[ inner_], _] :=
	{{}, {Theorema`Knowledge`isDomain$TM[ dom], Forall$TM[ r, cond, DomainOperation$TM[ inner, Element$TM][ term]]}}
setsDM[ expr_, _[ inner_], rules_List] :=
	Module[ {x = VAR$[ Unique[ Theorema`Knowledge`VAR$x$TM]], tmp},
		{First[ tmp], {isSet$TM[ expr]}}
			/; MatchQ[ tmp = followsImmediately[
								Forall$TM[ RNG$[ SETRNG$[ x, expr]], True, DomainOperation$TM[ inner, Element$TM][ x]],
								rules
							],
						{_List, {}}
				]
	]
setsDM[ expr_, dom_, _] := {{}, {DomainOperation$TM[ dom, Element$TM][ expr]}}


(* ::Section:: *)
(* Commutative Rings with Unit *)
	
inferenceRule[ RRPmembershipCommRings1] =
this:PRFSIT$[ g:FML$[ _, (DomainOperation$TM[ _, Element$TM][ _])|_And$TM, __], k_List, _, rest___] :>
Module[ {refs = $Failed, conds, other, R, dref, rules, simp, newK, dp, pr},
	performProofStep[
		dref = First[ dref];
		R = First[ formula@dref];
		{conds, other, simp} = splitRingConditions[ conds, R];
		If[ TrueQ[ simp],
			If[ other === {},
				rules = getRulesFromRW[ getOptionalComponent[ this, "rewriting"], "type" -> "Backward", "ground" -> $backwardGround];
				{conds, refs, simp} = simplifyRingConditions[ conds, rules, R, False];
				If[ refs === $Failed || simp =!= True,
					$Failed,
					
					refs = DeleteDuplicates[ refs];
					dp = Position[ conds, {}, {1}, Heads -> False];
					pr = Delete[ {{{RRPmembershipDomainTuples}}, {{RRPmembershipIntegerInterval}}, {}}, dp];
					conds = Delete[ conds, dp];
					If[ conds === {},
						makeTERMINALNODE[ makePRFINFO[ name -> RRPmembershipCommRings1, used -> {{g, dref}, refs}, generated -> {}, "dom" -> R], proved],
				
						other = makeRRPFML[ formula -> makeConjunction[ Last[ conds], And$TM], "marker" -> "G"];
						conds = Append[ Map[ makeRRPFML[ formula -> makeConjunction[ #, And$TM], "marker" -> "C"]&, Most[ conds]], other];
						newK = k;
						makeANDNODE[ makePRFINFO[ name -> RRPmembershipCommRings1, used -> {{g, dref}, refs}, generated -> conds, "dom" -> R],
							MapThread[ Block[ {$proposedRules = #2}, With[ {tmp = RRPToBeProved[ goal -> #1, kb -> newK, rest]}, newK = addNewKB[ newK, #1]; tmp]]&, {conds, pr}]
						]
					]
				],
				
				conds = makeRRPFML[ formula -> makeConjunction[ conds, And$TM], "marker" -> "C"];
				other = makeRRPFML[ formula -> makeConjunction[ other, And$TM], "marker" -> "G"];
				makeANDNODE[ makePRFINFO[ name -> RRPsplitConjunction, used -> {g}, generated -> {conds, other}],
					{RRPToBeProved[ goal -> conds, kb -> k, "proposedRules" -> {{RRPmembershipCommRings1}}, rest],
					RRPToBeProved[ goal -> other, kb -> addNewKB[ k, conds], rest]}
				]
			],
			
			$Failed	
		]
	] /;
	(
		conds = formula@g;
		conds = Switch[ conds,
					_And$TM,
					List@@conds,
					_,
					{conds}
				];
		dref = Cases[ k, FML$[ _, (Theorema`Knowledge`isCommRing1$TM|Theorema`Knowledge`isReductionRing$TM|Theorema`Knowledge`isAlgoReductionRing$TM)[ dom_] /; !FreeQ[ conds, DomainOperation$TM[ dom, Element$TM][ __]], __], {1}, 1];
		dref =!= {}
	)
]

inferenceRule[ RRPcommRing1Equal] =
this:PRFSIT$[ g:FML$[ _, eq:Equal$TM[ _, __], __], k_List, _, rest___] :> performProofStep[
	Module[ {R, dref = Cases[ k, FML$[ _, (Theorema`Knowledge`isCommRing1$TM|Theorema`Knowledge`isReductionRing$TM|Theorema`Knowledge`isAlgoReductionRing$TM)[ dom_], __] /;
									MemberQ[ eq, DomainOperation$TM[ dom, 0|1]|
												(DomainOperation$TM[ dom, Plus$TM|Minus$TM|Subtract$TM|Times$TM][ __])|
												DomainOperation$TM[ dom, SumOf$TM][ _, _, _]
									],
							{1}, 1
						],
			ring, other, refs, rules, newK, pr, dp},
		If[ dref === {},
			$Failed,
			
			dref = First[ dref];
			R = First[ formula@dref];
			rules = getRulesFromRW[ getOptionalComponent[ this, "rewriting"], "type" -> "Backward", "ground" -> $backwardGround];
			{ring, refs} = simplifyCommRing1Equality[ eq, rules, R];
			If[ ring === $Failed,
				$Failed,
				
				dp = Position[ ring, {}, {1}, Heads -> False];
				pr = Delete[ {{{RRPmembershipDomainTuples}}, {{RRPmembershipIntegerInterval}}, {}, {}}, dp];
				ring = Delete[ ring, dp];
				If[ ring === {},
					makeTERMINALNODE[ makePRFINFO[ name -> RRPcommRing1Equal, used -> {{g, dref}, refs}, generated -> {}, "dom" -> R], proved],
					
					other = makeRRPFML[ formula -> makeConjunction[ Last[ ring], And$TM], "marker" -> "G"];
					ring = Append[ Map[ makeRRPFML[ formula -> makeConjunction[ #, And$TM], "marker" -> "C"]&, Most[ ring]], other];
					newK = k;
					makeANDNODE[ makePRFINFO[ name -> RRPcommRing1Equal, used -> {{g, dref}, refs}, generated -> ring, "dom" -> R],
						MapThread[ Block[ {$proposedRules = #2},
									With[ {tmp = RRPToBeProved[ goal -> #1, kb -> newK, rest]},
										newK = addNewKB[ newK, #1];
										tmp
									]
								]&,
								{ring, pr}
						]
					]
				]
			]
		]
	]
]

inferenceRule[ RRPcommRing1EqualKB] =
this:PRFSIT$[ g_, k_List, _, rest___] :> performProofStep[
	Module[ {R, dref = Cases[ k, fml:FML$[ _, (Theorema`Knowledge`isCommRing1$TM|Theorema`Knowledge`isReductionRing$TM|Theorema`Knowledge`isAlgoReductionRing$TM)[ dom_], __] :> {dom, fml}],
			allConds = {}, allRefs = {}, rules, newK, pr, dp, eqs, u = {}, gen = {}},
		If[ dref === {},
			$Failed,

			eqs = Cases[ k, fml:FML$[ _?activatedQ, e:Equal$TM[_, _], __] :>
								Module[ {dom},
									(
										dref = Cases[ dref, First[ dom]];
										{e, fml}
									) /; 
									(
										dom = Cases[dref, {d_, _} /; MemberQ[ e, DomainOperation$TM[ d, 0|1]|
																				(DomainOperation$TM[ d, Plus$TM|Minus$TM|Subtract$TM|Times$TM][ __])|
																				DomainOperation$TM[ d, SumOf$TM][ _, _, _]
																	],
													{1},
													1
												];
										dom =!= {}
									)
								]
					];
			If[ eqs === {},
				$Failed,
				
				dref = First[ dref];
				R = First[ dref];
				dref = Last[ dref];
   				rules = getRulesFromRW[ getOptionalComponent[ this, "rewriting"], "type" -> "Backward", "ground" -> $backwardGround];
				{allConds, dp, allRefs} = simplifyCommRing1Equality[ eqs[[All, 1]], rules, R];
				If[ allConds === $Failed || ((gen = Last[ allConds]) === {}),
					$Failed,

					gen = Map[ makeRRPFML[ formula -> #, "marker" -> "A"]&, gen];
					u = Delete[ eqs[[All, 2]], dp];
					allConds = Most[ allConds];
					dp = Position[ allConds, {}, {1}, Heads -> False];
					pr = Delete[ {{{RRPmembershipDomainTuples}}, {{RRPmembershipIntegerInterval}}, {}}, dp];
					allConds = Delete[ allConds, dp];
					If[allConds === {},
						makeANDNODE[ makePRFINFO[ name -> RRPcommRing1EqualKB, used -> {u, allRefs, {dref}}, generated -> gen, "dom" -> R],
									RRPToBeProved[ goal -> g, kb -> joinNewKB[ Select[k, (!MemberQ[ u, #])&], gen], rest]
						],

						allConds = Map[ makeRRPFML[ formula -> makeConjunction[ #, And$TM], "marker" -> "C"]&, allConds];
						newK = k;
						makeANDNODE[ makePRFINFO[ name -> RRPcommRing1EqualKB, used -> {u, allRefs, {dref}}, generated -> {gen, allConds}, "dom" -> R],
							Append[
								MapThread[ Block[ {$proposedRules = #2},
											With[ {tmp = RRPToBeProved[ goal -> #1, kb -> newK, rest]},
												newK = addNewKB[ newK, #1];
												tmp
											]
										]&,
										{allConds, pr}
								],
								RRPToBeProved[ goal -> g, kb -> joinNewKB[ Select[newK, (!MemberQ[ u, #])&], gen], rest]
							]
						]
					]
				]
			]
		]
	]
]


(* ::Subsection:: *)
(* Auxiliary Functions *)

(* 'splitRingConditions[ conds, R ]' splits the list 'conds' into the two sublists 'ring' and 'other',
	where 'ring' consists of all elements of 'conds' that are (universally quantified)
	'R'-membership, 'DomainTuples[R]'-membership or integer-interval membership conditions, and 'other' consists of
	all other elements.
	The function returns the triple '{ring, other, ringMembership}', where 'ringMembership' is a boolean value
	indicating whether any of the elements of 'ring' really is an 'R'-membership condition or not.
*)
splitRingConditions[ conds_List, R_] :=
	Module[ {thinned = DeleteDuplicates[ conds, alphaEquivalent], ring, rest, tmp, qe, rng, simp = False},
		{tmp, {ring, rest}} = Reap[ Scan[ Function[ c,
						Switch[ c,
							DomainOperation$TM[ R, Element$TM][ __],
							simp = True;
							Sow[ c, "ring"],
	
							DomainOperation$TM[ Theorema`Knowledge`DomainTuples$TM[ R], Element$TM][ _],
							Sow[ c, "ring"],
	
							Element$TM[ _, _IntegerInterval$TM],
							Sow[ c, "ring"],
							
							_Forall$TM,
							{qe, rng} = getQuantifiedExpression[ c];
							Switch[ qe,
								Element$TM[ _, _IntegerInterval$TM],
								Sow[ c, "ring"],
								
								DomainOperation$TM[ Theorema`Knowledge`DomainTuples$TM[ R], Element$TM][ _],
								Sow[ c, "ring"],
	
								DomainOperation$TM[ R, Element$TM][ __],
								simp = True;
								Sow[ c, "ring"],
								
								_,
								Sow[ c, "rest"]
							],
							
							_,
							Sow[ c, "rest"]
						]
					],
					thinned
				],
				{"ring", "rest"}
			];
		{Flatten[ ring, 1], rest = Flatten[ rest, 1], simp}
	]
	
(* 'simplifyRingConditions[ conds, R ]' removes redundant conditions from the list 'conds',
	i.e. formulas that already follow from other formulas in 'conds' in the commutative ring with unit R.
	Moreover, some (simple) inferences from tuples and integers are also used.
	The function returns a triple '{integers, tuples, ring}', where
	- 'integers' is the list containing all (quantified) IntegerInterval-membership conditions from 'conds',
	- 'tuples' is the list containing all (quantified) 'DomainTuples[R]'-membership conditions from 'conds',
	- 'ring' is the list containing the simplified (quantified) ring-membership conditions.
	
	'ring' is a list of pairs of the form '{{orig, origRanges}, {simp, simpRanges}}', where
	- 'orig' is the original formula, with all leading universal quantifiers stripped away,
	- 'origRanges' is a list containing, for each stripped quantifier, the range and the condition,
	- 'simp' is the simplified formula, i.e. where some subparts were replaced by '$checked',
	- 'simpRanges' are the respective ranges for 'simp', with redundant ranges removed.
	
	'simplifyRingConditions[ conds, rules, R, True ]' does essentially the same as 'simplifyRingConditions[ conds, R ]',
	only that in addition 'proveRingConditions' is called on the output, for proving (by backward rewriting w.r.t. 'rules')
	some of the domain-membership conditions. New 'integers' and 'tuples' conditions may be generated during this process.
*)
simplifyRingConditions[ conds_List, R_] := simplifyRingConditions[ conds, {}, R, False]
simplifyRingConditions[ conds_List, rules_List, R_, unknownAsRing_] :=
	Module[ {thinned = DeleteDuplicates[ conds, alphaEquivalent], integers, tuples, ringG, ringQ, rest, tmp, qe, rng, simp1 = False, simp2, refs},
		{tmp, {integers, tuples, ringG, ringQ, rest}} = Reap[ Scan[ Function[ c,
						Switch[ c,
							DomainOperation$TM[ R, Element$TM][ __],
							Scan[ Sow[ #, "ringG"]&, c],
	
							DomainOperation$TM[ Theorema`Knowledge`DomainTuples$TM[ R], Element$TM][ _],
							Sow[ c, "tuples"],
	
							Element$TM[ _, _IntegerInterval$TM],
							Sow[ c, "integers"],
							
							_Forall$TM,
							{qe, rng} = getQuantifiedExpression[ c];
							Switch[ qe,
								Element$TM[ _, _IntegerInterval$TM],
								Sow[ c, "integers"],
								
								DomainOperation$TM[ Theorema`Knowledge`DomainTuples$TM[ R], Element$TM][ _],
								Sow[ c, "tuples"],
	
								DomainOperation$TM[ R, Element$TM][ __],
								Scan[ Sow[ {#, rng}, "ringQ"]&, qe],
								
								_,
								Sow[ c, "rest"]
							],
							
							_,
							Sow[ c, If[ TrueQ[ unknownAsRing], "ringG", "rest"]]
						]
					],
					thinned
				],
				{"integers", "tuples", "ringG", "ringQ", "rest"}
			];
		integers = Flatten[ integers, 1];
		tuples = Flatten[ tuples, 1];
		ringG = Flatten[ ringG, 1];
		ringQ = Flatten[ ringQ, 1];
		If[ Flatten[ rest, 1] === {},
			ringG = Join[ Map[ Function[ c,
									{{c, {}}, simplifyRingConditions[ {c, {}}, integers, tuples, With[ {d = Depth[ c]}, Select[ ringG, Depth[ #] < d &]], ringQ, R]}
								],
								ringG
						  ],
						  Map[ Function[ c,
						  			{c, simplifyRingConditions[ c, integers, tuples, ringG, With[ {d = Depth[ First[ c]]}, Select[ ringQ, Depth[ First[ #]] < d &]], R]}
						  		],
						  		ringQ
						  ]
					];
			If[ MemberQ[ ringG, {_, {$checked, _List}}],
				simp1 = True;
				ringG = DeleteCases[ ringG, {_, {$checked, _List}}]
			];
			
			If[ ringG =!= {},
				{ringG, tuples, integers, refs, simp2} = proveRingConditions[ ringG, tuples, integers, rules, R],
				refs = {}
			];
			tuples = DeleteDuplicates[ tuples, alphaEquivalent];
			integers = DeleteDuplicates[ integers, alphaEquivalent];
			ringG = DeleteDuplicates[ ringG, alphaEquivalent];
			{{tuples, integers, ringG}, refs, simp1 || simp2},
			
			{{$Failed, $Failed, $Failed}, $Failed, False}
		]
	]
simplifyRingConditions[ {expr_, rng_List}, integers_, tuples_, ringG_, ringQ_, R_] :=
	Module[ {s, r},
		{s, r} = simplifyRingConditions[ {expr /. {_?(alphaMemberQ[ ringG, #]&) :> $checked, DomainOperation$TM[ R, 0|1] :> $checked}, rng}, integers, tuples, ringQ, R];
		{s //. DomainOperation$TM[ R, Plus$TM|Minus$TM|Subtract$TM|Times$TM][ $checked ...] -> $checked, r}
	]
simplifyRingConditions[ a:{_, Except[ {}, _List]}, _, _, {___, b_, ___}, _] /; rangeEquivalent[ a, b] :=
	{$checked, {}}
simplifyRingConditions[ expr:{e:(DomainOperation$TM[ R_, Plus$TM|Minus$TM|Subtract$TM|Times$TM][ ___]), ranges_List}, integers_, tuples_, ring_, R_] :=
	{
		Map[ Function[ c,
				With[ {r = thinnedRange[ c, ranges, Range]},
					If[ r =!= {} && MemberQ[ ring, _?(rangeEquivalent[ #, {c, r}]&)],
						$checked,
						First[ simplifyRingConditions[ {c, r}, integers, tuples, ring, R]]
					]
				]
			],
			e
		],
		ranges
	}
simplifyRingConditions[ expr:{e:Subscript$TM[ _, _], _List}, integers_, tuples_, ring_, R_] /;
		decomposeRingCondition[ expr, e, tuples, integers, R, False] === {{}, {}, {}} :=
	{$checked, {}}
simplifyRingConditions[ expr:{e:(DomainOperation$TM[ R_, SumOf$TM][ rng_RNG$, cond_, _]), _List}, integers_, tuples_, ring_, R_] :=
	Module[ {s, r, newTuples, newIntegers},
		{s, newTuples, newIntegers} = decomposeRingCondition[ expr, e, tuples, integers, R, False];
		Assert[ newTuples === {} && Length[s] <= 1];
		If[ s === {},
			s = $checked,
			{s, r} = simplifyRingConditions[ s[[1, 1]], integers, tuples, ring, R];
			s = s //. DomainOperation$TM[ R, Plus$TM|Minus$TM|Subtract$TM|Times$TM][ $checked ...] -> $checked
		];
		If[ s === $checked,
			If [newIntegers === {},
				{$checked, {}},
				{DomainOperation$TM[ R, SumOf$TM][rng, cond, $checked], {}}
			],
			{DomainOperation$TM[ R, SumOf$TM][ rng, cond, s], DeleteCases[ r, {rng, _}]}
		]
	]
simplifyRingConditions[ {$checked, _}, _, _, _, _] := {$checked, {}}
simplifyRingConditions[ expr_, _, _, _, _] := expr

proveRingConditions[ conds_List, tuples_List, integers_List, rules_List, R_] :=
	Module[ {known = conds[[All, 1]], aux = conds, newConds = {}, newTuples = tuples, newIntegers = integers, cur, cnd, tup, int, rf, refs = {}, simp = False},
		While[ aux =!= {},
			cur = First[ aux];
			{cnd, tup, int, rf} = proveRingConditions[ cur, known, newTuples, newIntegers, rules, R];
			If[ aux[[1, 1, -1]] === {},
				(* We proved domain membership of a ground term t.
					Hence, we can replace every occurrence of t in the remaining conditions by $checked. *)
				With[ {t = aux[[1, 1, 1]]},
					aux = Replace[ aux, {o_, {s_, r___}} :> {o, {s /. t -> $checked, r}}, {1}]
				],
				(* We proved domain membership of a quantified term t.
					Hence, we cannot immediately replace every occurrence of t in the remaining conditions, but rather add t to known (unless it is already there). *)
				If[ cnd =!= {} || rf =!= {} || tup =!= {} || int =!= {},
					known = Join[ known, aux[[1, 1]]]
				]
			];
			aux = Rest[ aux];
			If[ cnd === $Failed,
				AppendTo[ newConds, thinnedRange[ DomainOperation$TM[ R, Element$TM][ cur[[1, 1]]], cur[[1, 2]], Expression]],
				
				simp = True;
				aux = Join[ cnd, aux];
				newTuples = Join[ newTuples, tup];
				newIntegers = Join[ newIntegers, int];
				refs = Join[ refs, rf]
			]
		];
		{postprocessExpr[ newConds], newTuples, newIntegers, DeleteDuplicates[ refs], simp}
	]
proveRingConditions[ {{orig_, origRanges_List}, {simp_, _List}}, _List, tuples_List, integers_List, rules_List, R_] :=
	Module[ {expr = preprocessExpr[ thinnedRange[ DomainOperation$TM[ R, Element$TM][ orig], origRanges, Expression]], refs},
		refs = followsImmediately[ expr, rules, "RenameBound" -> False];
		If[ MatchQ[ refs, {_List, {}}],
			{{}, tuples, integers, First[ refs]},
			Append[ decomposeRingCondition[ getQuantifiedExpression[ expr], simp, tuples, integers, R], {}]
		]
	]
proveRingConditions[ {orig:{_, Except[ {}, _List]}, {_}}, {___, k_, ___}, _List, _List, _List, _] /; rangeEquivalent[ orig, k] :=
	{{}, {}, {}, {}}
proveRingConditions[ {orig:{origExpr_, origRanges_List}, {simp_}}, _List, tuples_List, integers_List, rules_List, R_] :=
	Module[ {expr = thinnedRange[ DomainOperation$TM[ R, Element$TM][ origExpr], origRanges, Expression], refs},
		refs = followsImmediately[ expr, rules, "RenameBound" -> False];
		If[ MatchQ[ refs, {_List, {}}],
			{{}, {}, {}, First[ refs]},
			Append[ decomposeRingCondition[ orig, simp, tuples, integers, R], {}]
		]
	]
	
(* 'decomposeRingCondition[ {orig, ranges}, simp, tuples, integers, R, nonemptyTupInt ]' decomposes the ring-membership-condition on
	'{orig, ranges}' (i.e. the expression obtained from 'orig' by prepending a universal quantifier for each element of 'ranges') by
	splitting additions, multiplications, subtractions and sums into their individual components.
	The function returns a triple of new ring, tuple, and integer conditions. *)
decomposeRingCondition[ expr:{e_, _List}, tuples_List, integers_List, R_] :=
	decomposeRingCondition[ expr, e, tuples, integers, R, True]
decomposeRingCondition[ orig_, simp_, tuples_List, integers_List, R_] :=
	decomposeRingCondition[ orig, simp, tuples, integers, R, True]
decomposeRingCondition[ {DomainOperation$TM[ R_, Element$TM][ expr_], ranges_List}, simp_, tuples_List, integers_List, R_, nonemptyTupInt_] :=
	decomposeRingCondition[ {expr, ranges}, simp, tuples, integers, R, nonemptyTupInt]
decomposeRingCondition[ {e:(DomainOperation$TM[ R_, Plus$TM|Minus$TM|Subtract$TM|Times$TM][ argsOrig___]), ranges_List}, s:(_[argsSimp___]), _List, _List, R_, _] /;
		Length[e] === Length[s] :=
	Module[ {r},
		{DeleteCases[ MapThread[ Function[ {c1, c2},
						If[ c2 === $checked,
							$checked,
							r = thinnedRange[ c1, ranges, Range];
							{{c1, r}, {c2}}
						]
					],
					{{argsOrig}, {argsSimp}}
				],
				$checked
			],
			{}, {}
		}
	]
decomposeRingCondition[ {Subscript$TM[ T_, var_], {pre___, {RNG$[ STEPRNG$[ var_, a_, b_, 1]], _}, post___}}, _, tuples_List, integers_List, R_, nonemptyTupInt_] /;
		FreeQ[ {T, a, b}, var] :=
	Module[ {e, newTuples, newIntegers = {}},
		e = thinnedRange[ DomainOperation$TM[ Theorema`Knowledge`DomainTuples$TM[ R], Element$TM][ T], {pre, post}, Expression];
		newTuples = If[ alphaMemberQ[ tuples, e], {}, {e}];

		If[ nonemptyTupInt || newTuples === {},
  			If[ !(TrueQ[ a >= 1] ||
  					alphaMemberQ[ integers, e = thinnedRange[ Element$TM[ a, IntegerInterval$TM[ 1, DirectedInfinity[ 1], True, False]], {pre, post}, Expression]]),
				AppendTo[ newIntegers, e]
			];
			If[ (nonemptyTupInt || newIntegers === {}) &&
				!(b === BracketingBar$TM[ T] ||
					alphaMemberQ[ integers, e = thinnedRange[ Element$TM[ b, IntegerInterval$TM[ DirectedInfinity[ -1], BracketingBar$TM[ T], False, True]], {pre, post}, Expression]]),
				AppendTo[ newIntegers, e]
			]
		];
		{{}, newTuples, newIntegers}
	]
decomposeRingCondition[ {DomainOperation$TM[ R_, SumOf$TM][ rng:RNG$[ STEPRNG$[ var_, a_, b_, 1]], cond_, summand_], ranges_List}, DomainOperation$TM[ R_, SumOf$TM][ _RNG$, _, summandSimp_], tuples_List, integers_List, R_, nonemptyTupInt_] :=
	Module[ {e, integerLimit, newRing, newIntegers = {}}, 
		newRing = If[ summandSimp === $checked,
					{},
					{{{summand, Append[ ranges, {rng, cond}]}, {summandSimp}}}
				  ];
		Map[ If[ nonemptyTupInt || newIntegers === {},
				integerLimit = Switch[ #,
									_Integer,
									True,
									BracketingBar$TM[ _], 
									alphaMemberQ[ tuples, thinnedRange[ DomainOperation$TM[ Theorema`Knowledge`DomainTuples$TM[ R], Element$TM][ First[ #]], ranges, Expression]],
									_,
									False
								];
				If[ integerLimit === False,
					If[ !alphaMemberQ[ integers, e = thinnedRange[ Element$TM[ #, IntegerInterval$TM[ DirectedInfinity[ -1], DirectedInfinity[ 1], False, False]], ranges, Expression]],
						AppendTo[ newIntegers, e]
					]
				]
			]&,
			{a, b}
		];
		{newRing, {}, newIntegers}
	]
decomposeRingCondition[ {_, _List}, _, _List, _List, _, _] := {$Failed, $Failed, $Failed}

(* 'simplifyCommRing1Equality[ a1 == a2 == ... == an, rules, R ]' tries to simplify the equalities 'a1 == a2, a2 == a3, ..., a(n-1) == an'
	by expanding each constituent (according to the axioms of commutative rings with unit) and cancelling terms that
	appear on both sides. Domain-membership conditions are generated for every term whose membership in 'R' is a necessary
	precondition in order for a rewrite to be performed, and all such conditions are immediately simplified/proved using
	'simplifyRingConditions' together with 'rules'. Note that the function can also deal with sums.
	The function returns a pair of the form '{{tuples, integers, ring, failed}, refs}', where
	- 'tuples' is the list of all unproved conditions on terms being domain-tuples of 'R',
	- 'integers' is the list of all unproved conditions on terms being elements of some integer interval,
	- 'ring' is the list of all unproved conditions on terms being elements of 'R',
	- 'failed' is the list of all (simplified) equalities that could not be proved,
	- 'refs' is the list of references needed for proving some of the conditions. *)
simplifyCommRing1Equality[ eq:Equal$TM[ first_, rest__], rules_List, R_] :=
	Module[ {l = first, p, conds, failed = {}, refs, eqs},
		eqs = Map[ (
						p = Equal$TM[ l, #];
						l = #;
						p
					)&,
					{rest}
				];
		{conds, failed, refs} = simplifyCommRing1Equality[ eqs, rules, R];
		If[ conds === $Failed,
			{$Failed, $Failed},
			
			{Append[ Most[ conds], Join[ Last[ conds], Extract[ eqs, failed]]], refs}
		]
	]
simplifyCommRing1Equality[ expr:{Equal$TM[ _, _] ...}, rules_List, R_] :=
	Module[ {tuples, integers, ring, failed = {}, simp = {}, refs, tmp},
		ring = Map[ Switch[ DomainOperation$TM[ R, 0],
						First[ #],
						simplifyCommRing1Equality[ Last[ #], R],
						Last[ #],
						simplifyCommRing1Equality[ First[ #], R],
						_,
						simplifyCommRing1Equality[ DomainOperation$TM[ R, Subtract$TM]@@#, R, False]
					]&,
					expr
				];
		If[ MatchQ[ ring, {$Failed ..}],
			{$Failed, $Failed, $Failed},

			failed = Position[ ring, $Failed, {1}, Heads -> False];
			If[ failed =!= {},
				ring = Delete[ ring, failed]
			];
			simp = Cases[ ring, {e_Equal$TM, _List} :> e];
			{{tuples, integers, ring}, refs, tmp} = simplifyRingConditions[ Flatten[ ring[[All, 2]], 1], rules, R, True];
			{{tuples, integers, ring, simp}, failed, refs}
		]
	]
simplifyCommRing1Equality[ exprOrig_, R_, origSubtract_:True] :=
	Module[ {pOrig, nOrig, p, n, expr = exprOrig, summands, exprAux, sign, pos1, pos2, conds1 = {}, conds2, conds3, simp1 = False, simp2},
		summands = collectSummands[ expr, R];
		(* 'summands' now is a list of pairs of the form '{{expression, positive}, position}'. *)
		
		While[ Length[ summands] > 1,
			{{exprAux, sign}, pos1} = First[ summands];
			summands = Rest[ summands];
			sign = !sign;
			pos2 = Cases[ summands, {{_?(alphaEquivalent[ exprAux, #]&), sign}, pos_} :> pos, {1}, 1];
			If[ pos2 =!= {},
				pos2 = First[ pos2];
				summands = DeleteCases[ summands, {{_, _}, Alternatives[ Append[ pos1, BlankSequence[]], Append[ pos2, BlankNullSequence[]]]}];
				expr = ReplacePart[ expr, {pos1, pos2} -> $dummy];
				conds1 = Join[ conds1, {exprAux}, collectOtherArguments[ expr, pos1], collectOtherArguments[ expr, pos2]]
			]
		];
		conds1 = conds1 /. $dummy -> DomainOperation$TM[ R, 0];
		expr = expr //. {
					DomainOperation$TM[ R, Plus$TM][ $dummy...] :> $dummy,
					e:(DomainOperation$TM[ R, Plus$TM][ ___, $dummy, ___]) :> DeleteCases[e, $dummy], 
					DomainOperation$TM[ R, Subtract$TM][ a_, $dummy] :> a,
					DomainOperation$TM[ R, Subtract$TM][ $dummy, a_] :> DomainOperation$TM[ R, Minus$TM][ a],
					DomainOperation$TM[ R, Minus$TM][ $dummy] -> $dummy
				};
		If[ expr === $dummy, expr = DomainOperation$TM[ R, 0]];
		simp1 = (conds1 =!= {});

		{pOrig, nOrig, conds2, simp2} = expandRing1Expression[ expr, R];
		{conds3, p, n} = termSetDifference[ pOrig, nOrig, R];
		If[ conds3 =!= {},
			simp2 = True;
			conds2 = Join[ conds2, Flatten[ conds3], Map[ termSetToTerm[ #, R, Inner]&, Join[ p, n]]]
		];
		If[ simp2,
			simp1 = True;
			conds1 = Join[ conds1, conds2];
			p = Equal$TM[ termSetToTerm[ p, R], termSetToTerm[ n, R]],

			p = Switch[ expr,
					DomainOperation$TM[ R, Subtract$TM][ _, _],
					simp1 = simp1 || origSubtract;
					conds1 = Join[ conds1, List@@expr];
					Equal$TM@@expr,
					
					DomainOperation$TM[ R, Minus$TM][ _],
					simp1 = True;
					AppendTo[ conds1, First[ expr]];
					Equal$TM[ First[ expr], DomainOperation$TM[ R, 0]],
					
					_,
					Equal$TM[ expr, DomainOperation$TM[ R, 0]]
				]
		];
		If[ First[ p] === Last[ p], p = True];
		If[ simp1,
			{p, conds1} //. DomainOperation$TM[ R, Plus$TM|Times$TM][ a_] :> a,
			$Failed
		]
	]
	
collectSummands[ expr_, R_] :=
	Sort[ Rest[ DeleteCases[ collectSummands[ R, True, {}][ expr], {{DomainOperation$TM[ R, 0], _}, _List}]], (Length[ Last[ #1]] <= Length[ Last[ #2]])&]
collectSummands[ R_, sign_, pos_List][ expr:(DomainOperation$TM[ R_, Plus$TM][ args___])] :=
	Prepend[ Flatten[ MapIndexed[ collectSummands[ R, sign, Join[ pos, #2]][ #1]&, {args}], 1], {{expr, sign}, pos}]
collectSummands[ R_, sign_, pos_List][ expr:(DomainOperation$TM[ R_, Subtract$TM][ a_, b_])] :=
	Prepend[ Join[ collectSummands[ R, sign, Append[ pos, 1]][ a], collectSummands[ R, !sign, Append[ pos, 2]][ b]], {{expr, sign}, pos}]
collectSummands[ R_, sign_, pos_List][ expr:(DomainOperation$TM[ R_, Minus$TM][ a_])] :=
	Prepend[ collectSummands[ R, !sign, Append[ pos, 1]][ a], {{expr, sign}, pos}]
collectSummands[ _, sign_, pos_List][ expr_] := {{{expr, sign}, pos}}

collectOtherArguments[ _, {}] := {}
collectOtherArguments[ expr_, {pre___, i_}] := 
	Join[ Delete[ List@@Part[ expr, pre], i], collectOtherArguments[ expr, {pre}]]
	
termSetToTerm[ {}, R_] := DomainOperation$TM[ R, 0]
termSetToTerm[ {a_}, R_] := termSetToTerm[ a, R, Inner]
termSetToTerm[ m:{_, __}, R_] := DomainOperation$TM[ R, Plus$TM]@@(Map[ termSetToTerm[ #, R, Inner]&, m])
termSetToTerm[ {DomainOperation$TM[ R_, SumOf$TM][ rng_RNG$, cond_, s_List]}, R_, Inner] := DomainOperation$TM[ R, SumOf$TM][ rng, cond, termSetToTerm[ s, R, Inner]]
termSetToTerm[ {a_}, R_, Inner] := a
termSetToTerm[ m:{_, __}, R_, Inner] := DomainOperation$TM[ R, Times$TM]@@m

termSetEqual[ a_, b_] := (Length[ a] === Length[ b] && termSetEqual[ a, b, {}])
termSetEqual[ {}, {}, _] := True
termSetEqual[ {a_, r___}, {pre___, b_, post___}, repl_List] /; termSetEqual[ a, b, repl] :=
	termSetEqual[ {r}, {pre, post}, repl]
termSetEqual[ _List, _List, _] := False
termSetEqual[ DomainOperation$TM[ D_, SumOf$TM][ RNG$[ STEPRNG$[ v1_VAR$, a1_, b1_, s1_]], c1_, sum1_List],
			  DomainOperation$TM[ D_, SumOf$TM][ RNG$[ STEPRNG$[ v2_VAR$, a2_, b2_, s2_]], c2_, sum2_List], repl_List] :=
	(
		Length[ sum1] === Length[ sum2] && 
		With[ {repl2 = Prepend[ repl, v1 -> v2]},
			alphaEquivalent[ {a1, b1, s1, c1}, {a2, b2, s2, c2}, repl2] && termSetEqual[ sum1, sum2, repl2]
		]
	)
termSetEqual[ a_, b_, repl_List] := alphaEquivalent[ a, b, repl]

termSetDifference[ a_List, b_List, R_] :=
	Module[ {out, anew, bnew},
		{out, {anew, bnew}} = Reap[ termSetDifference[ R][ a, b], {"left", "right"}];
		{out, Flatten[ anew, 2], Flatten[ bnew, 2]}
	]
termSetDifference[ _][ {}, b_List] := (Sow[ b, "right"]; {})
termSetDifference[ _][ a_List, {}] := (Sow[ a, "left"]; {})
termSetDifference[ R_][{ a_List, r___}, b_List] :=
	Module[ {i = Position[ b, _?(termSetEqual[ a, #]&), {1}, 1, Heads -> False]},
		If[ i === {},
			Sow[ {a}, "left"];
			termSetDifference[ R][ {r}, b],
			
			Prepend[ termSetDifference[ R][ {r}, Delete[ b, i[[1, 1]]]], Map[ termSetToTerm[ {#}, R, Inner]&, a]]
		]
	]
	
expandRing1Expression[ expr_, R_] :=
	With[ {SumR = DomainOperation$TM[ R, SumOf$TM]},
	Module[ {out, conds, conds2, simp = False, rules, sumMap, index = 0, sums},
		out = expr /. (e:(SumR[ RNG$[ _STEPRNG$], _, _])) :> (sumMap[ ++index] = e; $dummy[ index]);

		rules = {
			(* 0,1 *)
			DomainOperation$TM[ R, Minus$TM][ DomainOperation$TM[ R, 0]] :> (simp = True; DomainOperation$TM[ R, 0]),
			e:(DomainOperation$TM[ R, Plus$TM][ pre___, DomainOperation$TM[ R, 0], post___]) :>
				(simp = True; Sow[ {pre, post}, "cond"]; DeleteCases[e, DomainOperation$TM[ R, 0]]),
			DomainOperation$TM[ R, Plus$TM][] -> DomainOperation$TM[ R, 0],
			DomainOperation$TM[ R, Subtract$TM][ a_, DomainOperation$TM[ R, 0]] :> (simp = True; Sow[ {a}, "cond"]; a),
			DomainOperation$TM[ R, Subtract$TM][ DomainOperation$TM[ R, 0], a_] :>
				(simp = True; Sow[ {a}, "cond"]; DomainOperation$TM[ R, Minus$TM][ a]),
			DomainOperation$TM[ R, Times$TM][ pre___, DomainOperation$TM[ R, 0], post___] :>
				(simp = True; Sow[ {pre, post}, "cond"]; DomainOperation$TM[ R, 0]),
			e:(DomainOperation$TM[ R, Times$TM][ pre___, DomainOperation$TM[ R, 1], post___]) :>
				(simp = True; Sow[ {pre, post}, "cond"]; DeleteCases[ e, DomainOperation$TM[R, 1]]),
			DomainOperation$TM[ R, Times$TM][] -> DomainOperation$TM[ R, 1],
			DomainOperation$TM[ R, Plus$TM|Times$TM][ a_] :> a,
			
			(* - *)
			DomainOperation$TM[ R, Minus$TM][ DomainOperation$TM[ R, Minus$TM][ a_]] :> (simp = True; Sow[ {a}, "cond"]; a),
			DomainOperation$TM[ R, Subtract$TM][ a_, b_] :>
				If[ alphaEquivalent[ a, b],
					simp = True;
					Sow[ {a}, "cond"];
					DomainOperation$TM[ R, 0],
					Sow[ {b}, "cond"];
					DomainOperation$TM[ R, Plus$TM][ a, DomainOperation$TM[ R, Minus$TM][ b]]
				],
			DomainOperation$TM[ R, Minus$TM][ e:(DomainOperation$TM[ R, Plus$TM][ a__])] :>
				(Sow[ {a}, "cond"]; Map[ DomainOperation$TM[ R, Minus$TM], e]),
			DomainOperation$TM[ R, Plus$TM][ pre___, a_, mid___, DomainOperation$TM[ R, Minus$TM][ b_], post___] /; alphaEquivalent[ a, b] :>
				(simp = True; Sow[ {pre, a, mid, post}, "cond"]; DomainOperation$TM[ R, Plus$TM][ pre, mid, post]),
			DomainOperation$TM[ R, Plus$TM][ pre___, DomainOperation$TM[ R, Minus$TM][ a_], mid___, b_, post___] /; alphaEquivalent[ a, b] :>
				(simp = True; Sow[ {pre, a, mid, post}, "cond"]; DomainOperation$TM[ R, Plus$TM][ pre, mid, post]),
				
			(* associativity of + and * *)
			e:(DomainOperation$TM[ R, op:(Plus$TM|Times$TM)][ ___, DomainOperation$TM[ R, op_][__], ___]) :>
				With[ {new = Flatten[ e]}, Sow[ List@@new, "cond"]; new],
				
			(* distributivity *)
			DomainOperation$TM[ R, Times$TM][ pre___, e:(DomainOperation$TM[ R, Plus$TM][ a__]), post___] :>
				(Sow[ {pre, a, post}, "cond"]; Map[ DomainOperation$TM[ R, Times$TM][ pre, #, post]&, e]),
			DomainOperation$TM[ R, Times$TM][ pre___, DomainOperation$TM[ R, Minus$TM][ a_], post___] :>
				(Sow[ {pre, a, post}, "cond"]; DomainOperation$TM[ R, Minus$TM][ DomainOperation$TM[ R, Times$TM][ pre, a, post]])
				};
				
		{out, conds} = Reap[ out //. rules, "cond"] /. $dummy[ i_] :> sumMap[ i];
		out = Switch[ out,
				DomainOperation$TM[ R, Plus$TM][ __],
				List@@out,
				
				DomainOperation$TM[ R, 0],
				{},
				
				_,
				{out}
			];
		{out, conds2} = Reap[ Replace[
									{DeleteCases[ out, DomainOperation$TM[ R, Minus$TM][ _]],
									Cases[ out, DomainOperation$TM[ R, Minus$TM][ a_] :> a]},
								{DomainOperation$TM[R, Times$TM][ pre___, e:(SumR[ RNG$[ STEPRNG$[var:VAR$[ base_Symbol], a_, b_, 1]], cond_, summand_]), post___] :>
									Module[ {newVar, fv = Cases[ freeVariables[ {pre, post}], VAR$[ x_Symbol] :> x]},
										Sow[ Join[ makeSumConditions[ e], {pre, post}], "conds"];
										{If[ MemberQ[ fv, base],
											newVar = VAR$[ freshName[ base, Join[ fv, Cases[ summand, VAR$[ x_Symbol] :> x, {0, Infinity}, Heads -> True]]]];
											SumR[ RNG$[ STEPRNG$[ newVar, a, b, 1]],
												substituteFree[ cond, {var -> newVar}],
												expandRing1Expression[ DomainOperation$TM[ R, Times$TM][ pre, substituteFree[ summand, {var -> newVar}], post], R]
											],

											SumR[ RNG$[ STEPRNG$[ var, a, b, 1]], cond, expandRing1Expression[ DomainOperation$TM[ R, Times$TM][ pre, summand, post], R]]
										]}
									],
									e:(DomainOperation$TM[ R, Times$TM][ __]) :> List @@ e,
									SumR[ rng:RNG$[ STEPRNG$[ __, 1]], cond_, summand_] :> {SumR[ rng, cond, expandRing1Expression[ summand, R]]},
									a_ :> {a}
								}, {2}],
							"conds"
						];

		conds = Join[ conds, conds2];
		{out, sums} = {
						DeleteCases[ out, {SumR[ _RNG$, _, {_List, _List, _, _}]}, {2}],
						{Cases[ First[ out], {SumR[ _RNG$, _, {_List, _List, _, _}]}], Cases[ Last[ out], {SumR[ _RNG$, _, {_List, _List, _, _}]}]}
					  };
					  
		{sums, conds2} = Reap[ Replace[ sums, {sum:(SumR[ rng_RNG$, cond_, {pos_, neg_, c_, s_}])} :>
												(simp = simp || s;
												Sow[ makeSumConditions[ sum], "conds"];
												{Map[ {SumR[ rng, cond, #]}&, pos], Map[ {SumR[ rng, cond, #]}&, neg]}
												),
								{2}],
							"conds"
						];
		sums = {Join[ Flatten[ sums[[1, All, 1]], 1], Flatten[ sums[[2, All, 2]], 1]], Join[ Flatten[ sums[[1, All, 2]], 1], Flatten[ sums[[2, All, 1]], 1]]};
		Join[ Join[ out, sums, 2], {Flatten[ Join[ conds, conds2]], simp}]
	]
	]
	
makeSumConditions[ DomainOperation$TM[ R_, SumOf$TM][ rng_RNG$, cond_, {{_}, {}, c_, _}]] := Flatten[ Map[ makeUniversalCond[ rng, cond, R], c], 1]
makeSumConditions[ DomainOperation$TM[ R_, SumOf$TM][ rng:RNG$[ STEPRNG$[ var_, a_, b_, 1]], cond_, expr_]] :=
	Module[ {out, pos, neg, c},
		{pos, neg, c} = If[ MatchQ[ expr, {_List, _List, _List, _}],
							Most[ expr],
							{{expr}, {}, {}}
						];
		out = Flatten[ Join[ Map[ makeUniversalCond[ rng, cond, R], pos], Map[ makeUniversalCond[ rng, cond, R], neg], Map[ makeUniversalCond[ rng, cond, R], c]], 1];
		If[ !MemberQ[ out, Element$TM[ a, _IntegerInterval$TM]],
			AppendTo[ out, Element$TM[ a, IntegerInterval$TM[ DirectedInfinity[ -1], DirectedInfinity[ 1], False, False]]]
		];
		If[ !MemberQ[ out, Element$TM[ b, _IntegerInterval$TM]],
			AppendTo[ out, Element$TM[ b, IntegerInterval$TM[ DirectedInfinity[ -1], DirectedInfinity[ 1], False, False]]]
		];
		out
	]
	
makeUniversalCond[ rng_RNG$, cond_, R_][ {x_}] := makeUniversalCond[ rng, cond, R][ x]
makeUniversalCond[ rng_RNG$, cond_, R_][ x:{_, __}] := makeUniversalCond[ rng, cond, R][ DomainOperation$TM[ R, Times$TM]@@x]
makeUniversalCond[ rng:RNG$[ STEPRNG$[ var_, a_, b_, 1]], cond_, R_][ Subscript$TM[ T_, var_]] :=
	Module[ {out = {DomainOperation$TM[ Theorema`Knowledge`DomainTuples$TM[ R], Element$TM][ T]}},
		If[ a =!= BracketingBar$TM[ T],
			AppendTo[ out, Element$TM[ a, IntegerInterval$TM[ 1, DirectedInfinity[ 1], True, False]]]
		];
		If[ b =!= BracketingBar$TM[ T],
			AppendTo[ out, Element$TM[ b, IntegerInterval$TM[ DirectedInfinity[ -1], BracketingBar$TM[ T], False, True]]]
		];
		out
	]
makeUniversalCond[ rng:RNG$[STEPRNG$[ var_, __]], cond_, R_][ c_] :=
	With[ {cnew = c /. List -> DomainOperation$TM[ R, Times$TM]},	(* There might be sum-expressions with multiplication replaced by List. *)
		{Which[
			!isFree[ var, cnew],
			cnew,
	
			MatchQ[ cnew, _Forall$TM] || MatchQ[ cnew, DomainOperation$TM[ _, Element$TM][ _]],
			Forall$TM[ rng, cond, cnew],
			
			True,
			Forall$TM[ rng, cond, DomainOperation$TM[ R, Element$TM][ cnew]]
		]}
	]
	
rangeEquivalent[ A:{_, r1_List}, B:{_, r2_List}] :=
	Length[ r1] === Length[ r2] && rangeEquivalent[ A, B, {}]
rangeEquivalent[ {a_, {}}, {b_, {}}, repl_List] :=
	alphaEquivalent[ a, b, repl]
rangeEquivalent[ {a_, {prea___, {ra_RNG$, ca_}}}, {b_, {preb___, {rb_RNG$, cb_}}}, repl_List] :=
	Module[ {newRules},
		If[ Length[ ra] === Length[ rb], 
			newRules = MapThread[
							If[ varType[ #1] === varType[ #2],
								#1 -> #2,
								$Failed
							]&,
							{rngVariables[ ra], rngVariables[ rb]}
						];
			If[ MemberQ[ newRules, $Failed],
				False,
				With[ {rl = Join[ newRules, repl]},
					alphaEquivalent[ ca, cb, rl] &&
					rangeEquivalent[ {a, {prea}}, {b, {preb}}, rl]
				]
			],
			False
		]
	]

getQuantifiedExpression[ Forall$TM[ r_RNG$, cond_, inner_]] :=
	Module[ {innerRanges, out},
		{out, innerRanges} = getQuantifiedExpression[ inner];
		{out, Prepend[ innerRanges, {r, cond}]}
	]
getQuantifiedExpression[ expr_] := {expr, {}}

thinnedRange[ expr_, rng_List, return_:All] :=
	Module[ {fv, newExpr = expr, newRanges = {}},
		fv = freeVariables[ expr];
		Scan[ With[ {rv = rngVariables[ First[ #]]},
				If[ Intersection[ rv, fv] =!= {},
					fv = Complement[ Join[ fv, freeVariables[ #]], rv];
					Switch[ return,
						Expression,
						newExpr = Forall$TM@@Append[ #, newExpr],
						Range,
						PrependTo[ newRanges, #],
						_,
						newExpr = Forall$TM@@Append[ #, newExpr];
						PrependTo[ newRanges, #]
					]
				]
			]&,
			Reverse[ rng]
		];
		Switch[ return,
			Expression,
			newExpr,
			Range,
			newRanges,
			_,
			{newExpr, newRanges}
		]
	]


(* ::Section:: *)
(* Cancellative Commutative Monoids *)
	
inferenceRule[ RRPmembershipCCMonoids] =
this:PRFSIT$[ g:FML$[ _, f_, __], k_List, _, rest___] :>
Module[ {refs = $Failed, conds, other, M, dref, rules, simp, mult},
	performProofStep[
		dref = First[ dref];
		M = formula@dref;
		Switch[ M,
			_[ _, _, _],
			mult = False;
			M = List@@M,
			
			_,
			mult = True;
			M = First[ M];
			M = Prepend[ {DomainOperation$TM[ M, Times$TM], DomainOperation$TM[ M, 1]}, M]
		];
		{conds, other} = splitMonoidConditions[ conds, M];
		If[ conds === {},
			$Failed,
			
			If[ other === {},
				rules = getRulesFromRW[ getOptionalComponent[ this, "rewriting"], "type" -> "Backward", "ground" -> $backwardGround];
				{conds, refs, simp} = simplifyMonoidConditions[ conds, rules, M, False, True];
				If[ refs === $Failed || simp =!= True,
					$Failed,
					
					refs = DeleteDuplicates[ refs];
					If[ conds === {},
						makeTERMINALNODE[ makePRFINFO[ name -> RRPmembershipCCMonoids, used -> {{g, dref}, refs}, generated -> {}, "dom" -> If[ mult, {First[ M]}, M]], proved],
						
						conds = makeRRPFML[ formula -> makeConjunction[ conds, And$TM], "marker" -> "G"];
						makeANDNODE[ makePRFINFO[ name -> RRPmembershipCCMonoids, used -> {{g, dref}, refs}, generated -> {conds}, "dom" -> If[ mult, {First[ M]}, M]],
							RRPToBeProved[ goal -> conds, kb -> k, rest]
						]
					]
				],
				
				conds = makeRRPFML[ formula -> makeConjunction[ conds, And$TM], "marker" -> "C"];
				other = makeRRPFML[ formula -> makeConjunction[ other, And$TM], "marker" -> "G"];
				makeANDNODE[ makePRFINFO[ name -> RRPsplitConjunction, used -> {g}, generated -> {conds, other}],
					{RRPToBeProved[ goal -> conds, kb -> k, "proposedRules" -> {{RRPmembershipCCMonoids}}, rest],
					RRPToBeProved[ goal -> other, kb -> addNewKB[ k, conds], rest]}
				]
			]
		]
	] /;
	(
		conds = Switch[ f,
					_And$TM,
					List@@f,
					_,
					{f}
				];
		dref = Cases[ k, FML$[ _, (Theorema`Knowledge`isCCMonoid$TM|Theorema`Knowledge`isCommPPDomain$TM)[ dom_, ___] /; !FreeQ[ conds, DomainOperation$TM[ dom, Element$TM][ __]], __], {1}, 1];
		dref =!= {}
	)
]
	
inferenceRule[ RRPCCMonoidEqual] =
this:PRFSIT$[ g:FML$[ _, eq:Equal$TM[ _, __], __], k_List, _, rest___] :> performProofStep[
	Module[ {M, dref = Cases[ k, FML$[ _, (Theorema`Knowledge`isCCMonoid$TM|Theorema`Knowledge`isCommPPDomain$TM)[ dom_], __] /;
									MemberQ[ eq, DomainOperation$TM[ dom, 1]|(DomainOperation$TM[ dom, Times$TM][ __])],
							{1}, 1
						],
			monoid, eqs, refs, rules, mult},
		If[ dref === {},
			dref = Cases[ k, FML$[ _, Theorema`Knowledge`isCCMonoid$TM[ dom_, op_, n_], __] /;
									MemberQ[ eq, n|(op[ __])],
						{1}, 1
					]
		];
		If[ dref === {},
			$Failed,
			
			dref = First[ dref];
			M = formula@dref;
			Switch[ M,
				_[ _, _, _],
				mult = False;
				M = List@@M,
				
				_,
				mult = True;
				M = First[ M];
				M = Prepend[ {DomainOperation$TM[ M, Times$TM], DomainOperation$TM[ M, 1]}, M]
			];
			rules = getRulesFromRW[ getOptionalComponent[ this, "rewriting"], "type" -> "Backward", "ground" -> $backwardGround];
			{{monoid, eqs}, refs} = simplifyCCMonoidEquality[ eq, rules, M];
			If[ monoid === $Failed,
				$Failed,
				
				refs = DeleteDuplicates[ refs];
				If[ monoid =!= {},
					monoid = makeRRPFML[ formula -> makeConjunction[ monoid, And$TM], "marker" -> If[ eqs === {}, "G", "C"]]
				];
				If[ eqs =!= {},
					eqs = makeRRPFML[ formula -> makeConjunction[ eqs, And$TM], "marker" -> "G"]
				];
				M = If[ mult, {First[ M]}, M];
				If[ monoid === {},
					If[ eqs === {},
						makeTERMINALNODE[ makePRFINFO[ name -> RRPCCMonoidEqual, used -> {{g, dref}, refs}, generated -> {}, "dom" -> M], proved],
						makeANDNODE[ makePRFINFO[ name -> RRPCCMonoidEqual, used -> {{g, dref}, refs}, generated -> {eqs}, "dom" -> M],
							RRPToBeProved[ goal -> eqs, kb -> k, rest]
						]
					],
					
					If[ eqs === {},
						makeANDNODE[ makePRFINFO[ name -> RRPCCMonoidEqual, used -> {{g, dref}, refs}, generated -> {monoid}, "dom" -> M],
							RRPToBeProved[ goal -> monoid, kb -> k, rest]
						],
						makeANDNODE[ makePRFINFO[ name -> RRPCCMonoidEqual, used -> {{g, dref}, refs}, generated -> {monoid, eqs}, "dom" -> M],
							{RRPToBeProved[ goal -> monoid, kb -> k, rest], RRPToBeProved[ goal -> eqs, kb -> addNewKB[ k, monoid], rest]}
						]
					]
				]
			]
		]
	]
]


(* ::Subsection:: *)
(* Auxiliary Functions *)

splitMonoidConditions[ conds_List, {dom_, _, _}] :=
	Module[ {thinned = DeleteDuplicates[ conds, alphaEquivalent], monoid, rest, tmp, qe, rng},
		{tmp, {monoid, rest}} = Reap[ Scan[ Function[ c,
						Switch[ c,
							DomainOperation$TM[ dom, Element$TM][ __],
							Sow[ c, "monoid"],
							
							_Forall$TM,
							{qe, rng} = getQuantifiedExpression[ c];
							Sow[ c, If[ MatchQ[ qe, DomainOperation$TM[ dom, Element$TM][ __]], "monoid", "rest"]],
							
							_,
							Sow[ c, "rest"]
						]
					],
					thinned
				],
				{"monoid", "rest"}
			];
		{Flatten[ monoid, 1], rest = Flatten[ rest, 1]}
	]
	
(* 'simplifyMonoidConditions[ conds, M ]' removes redundant conditions from the list 'conds',
	i.e. formulas that already follow from other formulas in 'conds' in the cancellative commutative monoid 'M'.
	The function returns a list 'monoid' containing the simplified (quantified) monoid-membership conditions.
	'monoid' is a list of pairs of the form '{{orig, origRanges}, {simp, simpRanges}}', where
	- 'orig' is the original formula, with all leading universal quantifiers stripped away,
	- 'origRanges' is a list containing, for each stripped quantifier, the range and the condition,
	- 'simp' is the simplified formula, i.e.where some subparts were replaced by'$checked',
	- 'simpRanges' are the respective ranges for 'simp', with redundant ranges removed.
	
	'simplifyMonoidConditions[ conds, rules, M, True ]' does essentially the same as 'simplifyMonoidConditions[ conds, M ]',
	only that in addition 'proveMonoidConditions' is called on the output, for proving (by backward rewriting w.r.t. 'rules')
	some of the domain-membership conditions.
*)
simplifyMonoidConditions[ conds_List, M_List] := simplifyMonoidConditions[ conds, {}, M, False, True]
simplifyMonoidConditions[ conds_List, rules_List, M:{dom_, _, _}, unknownAsMonoid_, fs_] :=
	Module[ {thinned = DeleteDuplicates[ conds, alphaEquivalent], monoidG, monoidQ, rest, tmp, qe, rng, simp1 = False, simp2, refs},
		{tmp, {monoidG, monoidQ, rest}} = Reap[ Scan[ Function[ c,
						Switch[ c,
							DomainOperation$TM[ dom, Element$TM][ __],
							Scan[ Sow[ #, "monoidG"]&, c],
							
							_Forall$TM,
							{qe, rng} = getQuantifiedExpression[ c];
							If[ MatchQ[ qe, DomainOperation$TM[ dom, Element$TM][ __]],
								Scan[ Sow[ {#, rng}, "monoidQ"]&, qe],
								Sow[ c, "rest"]
							],
							
							_,
							Sow[ c, If[ TrueQ[ unknownAsMonoid], "monoidG", "rest"]]
						]
					],
					thinned
				],
				{"monoidG", "monoidQ", "rest"}
			];
		monoidG = Flatten[ monoidG, 1];
		monoidQ = Flatten[ monoidQ, 1];
		If[ Flatten[ rest, 1] === {},
			monoidG = Join[ Map[ Function[ c,
									{{c, {}}, simplifyMonoidConditions[ {c, {}}, With[ {d = Depth[ c]}, Select[ monoidG, Depth[ #] < d &]], monoidQ, M]}
								],
								monoidG
						  ],
						  Map[ Function[ c,
						  			{c, simplifyMonoidConditions[ c, monoidG, With[ {d = Depth[ First[ c]]}, Select[ monoidQ, Depth[ First[ #]] < d &]], M]}
						  		],
						  		monoidQ
						  ]
					];
			If[ MemberQ[ monoidG, {_, {$checked, _List}}],
				simp1 = True;
				monoidG = DeleteCases[ monoidG, {_, {$checked, _List}}]
			];
			
			If[ TrueQ[ fs] && monoidG =!= {},
				{monoidG, refs, simp2} = proveMonoidConditions[ monoidG, rules, M],
				refs = {}
			];
			{monoidG, refs, simp1 || simp2},
			
			{$Failed, $Failed, False}
		]
	]
simplifyMonoidConditions[ {expr_, rng_List}, monoidG_, monoidQ_, M:{_, op_, n_}] :=
	Module[ {s, r},
		{s, r} = simplifyMonoidConditions[ {expr /. {_?(alphaMemberQ[ monoidG, #]&) :> $checked, n :> $checked}, rng}, monoidQ, M];
		{s //. op[ $checked ...] -> $checked, r}
	]
simplifyMonoidConditions[ a:{_, Except[ {}, _List]}, {___, b_, ___}, _] /; rangeEquivalent[ a, b] :=
	{$checked, {}}
simplifyMonoidConditions[ {e:(op_[ ___]), ranges_List}, monoid_, M:{_, op_, n_}] :=
	{
		Map[ Function[ c,
				With[ {r = thinnedRange[ c, ranges, Range]},
					If[ r =!= {} && MemberQ[ monoid, _?(rangeEquivalent[ #, {c, r}]&)],
						$checked,
						First[ simplifyMonoidConditions[ {c, r}, monoid, M]]
					]
				]
			],
			e
		],
		ranges
	}
simplifyMonoidConditions[ {$checked, _}, _, _] := {$checked, {}}
simplifyMonoidConditions[ expr_, _, _] := expr

proveMonoidConditions[ conds_List, rules_List, M:{dom_, _, _}] :=
	Module[ {known = conds[[All, 1]], aux = conds, newConds = {}, cur, cnd, rf, refs = {}, simp = False},
		While[ aux =!= {},
			cur = First[ aux];
			{cnd, rf} = proveMonoidConditions[ cur, known, rules, M];
			If[ aux[[1, 1, -1]] === {},
				(* We proved domain membership of a ground term t.
					Hence, we can replace every occurrence of t in the remaining conditions by $checked. *)
				With[ {t = aux[[1, 1, 1]]},
					aux = Replace[ aux, {o_, {s_, r___}} :> {o, {s /. t -> $checked, r}}, {1}]
				],
				(* We proved domain membership of a quantified term t.
					Hence, we cannot immediately replace every occurrence of t in the remaining conditions, but rather add t to known (unless it is already there). *)
				If[ cnd =!= {} || rf =!= {},
					known = Join[ known, aux[[1, 1]]]
				]
			];
			aux = Rest[ aux];
			If[ cnd === $Failed,
				AppendTo[ newConds, thinnedRange[ DomainOperation$TM[ dom, Element$TM][ cur[[1, 1]]], cur[[1, 2]], Expression]],
				
				simp = True;
				aux = Join[ cnd, aux];
				refs = Join[ refs, rf]
			]
		];
		{postprocessExpr[ newConds], refs, simp}
	]
proveMonoidConditions[ {{orig_, origRanges_List}, {simp_, _List}}, _List, rules_List, M:{dom_, _, _}] :=
	Module[ {expr = preprocessExpr[ thinnedRange[ DomainOperation$TM[ dom, Element$TM][ orig], origRanges, Expression]], refs},
		refs = followsImmediately[ expr, rules, "RenameBound" -> False];
		If[ MatchQ[ refs, {_List, {}}],
			{{}, First[ refs]},
			Prepend[ {{}}, decomposeMonoidCondition[ getQuantifiedExpression[ expr], simp, M]]
		]
	]
proveMonoidConditions[ {orig:{_, Except[ {}, _List]}, {_}}, {___, k_, ___}, _List, _] /; rangeEquivalent[ orig, k] :=
	{{}, {}}
proveMonoidConditions[ {orig:{origExpr_, origRanges_List}, {simp_}}, _List, rules_List, M:{dom_, _, _}] :=
	Module[ {expr = thinnedRange[ DomainOperation$TM[ dom, Element$TM][ origExpr], origRanges, Expression], refs},
		refs = followsImmediately[ expr, rules, "RenameBound" -> False];
		If[ MatchQ[ refs, {_List, {}}],
			{{}, First[ refs]},
			Prepend[ {{}}, decomposeMonoidCondition[ orig, simp, M]]
		]
	]
	
decomposeMonoidCondition[ expr:{e_, _List}, M_] :=
	decomposeMonoidCondition[ expr, e, M]
decomposeMonoidCondition[ {DomainOperation$TM[ dom_, Element$TM][ expr_], ranges_List}, s_, M:{dom_, _, _}] :=
	decomposeMonoidCondition[ {expr, ranges}, s, M]
decomposeMonoidCondition[ {e:(op_[ argsOrig___]), ranges_List}, s:(_[argsSimp___]), {_, op_, _}] /; Length[e] === Length[s] :=
	Module[ {r},
		DeleteCases[ MapThread[ Function[ {c1, c2},
					If[ c2 === $checked,
						$checked,
						r = thinnedRange[ c1, ranges, Range];
						{{c1, r}, {c2}}
					]
				],
				{{argsOrig}, {argsSimp}}
			],
			$checked
		]
	]
decomposeMonoidCondition[ {_, _List}, _, _] := $Failed

(* 'simplifyCCMonoidEquality[ a1 == a2 == ... == an, rules, M ]' tries to simplify/prove the equalities 'a1 == a2, a2 == a3, ..., a(n-1) == an'
	in the theory of (multiplicatively written) cancellative commutative monoids.
	It essentially goes along the same lines as function 'simplifyCommRing1Equality', only that it is much simpler, since no sums have to be dealt with. *)
simplifyCCMonoidEquality[ eq:Equal$TM[ first_, rest__], rules_List, M_] :=
	Module[ {l = first, monoid, failed = {}, p, refs, simp}, 
		monoid = Map[ (
						p = simplifyCCMonoidEquality[ {l, #}, M];
						l = #;
						p
					)&,
					{rest}
				];
		If[ MatchQ[ monoid, {$Failed ..}],
			{{$Failed, $Failed}, $Failed},
			
			p = Position[ monoid, $Failed, {1}, Heads -> False];
			If[ p =!= {},
				monoid = Delete[ monoid, p];
				failed = Map[ With[ {i = First[ #]}, Equal$TM[ eq[[i]], eq[[i + 1]]]]&, p]
			];
			failed = Join[ failed, Cases[ monoid, {e_Equal$TM, _List} :> e]];
			{monoid, refs, simp} = simplifyMonoidConditions[ Flatten[ monoid[[All, 2]], 1], rules, M, True, True];
			{{monoid, failed}, refs}
		]
	];
simplifyCCMonoidEquality[ {origL_, origR_}, {_, op_, n_}] :=
	Module[ {l = origL, r = origR, factsL, factsR, exprAux, posL, posR, conds = {}},
		factsL = collectFactors[ origL, op];
		factsR = collectFactors[ origR, op];
		(* 'factsL' and 'factsR' now are lists of pairs of the form '{expression, position}'. *)
		
		While[ factsL =!= {},
			{exprAux, posL} = First[ factsL];
			factsL = Rest[ factsL];
			If[ exprAux === n,
				l = ReplacePart[ l, posL -> $dummy];
				conds = Join[ conds, collectOtherArguments[ l, posL]],
				
				posR = Cases[ factsR, {_?(alphaEquivalent[ exprAux, #]&), pos_} :> pos, {1}, 1];
				If[ posR =!= {},
					posR = First[ posR];
					factsL = DeleteCases[ factsL, {_, Append[ posL, BlankSequence[]]}];
					factsR = DeleteCases[ factsR, {_, Append[ posR, BlankNullSequence[]]}];
					l = If[ posL === {}, $dummy, ReplacePart[ l, posL -> $dummy]];
					r = If[ posR === {}, $dummy, ReplacePart[ r, posR -> $dummy]];
					conds = Join[ conds, {exprAux}, collectOtherArguments[ l, posL], collectOtherArguments[ r, posR]]
				]
			]
		];
		Scan[ (
				r = ReplacePart[ r, # -> $dummy];
				conds = Join[ conds, collectOtherArguments[ r, #]]
			)&,
			Cases[ factsR, {n, pos_} :> pos]
		];
		
		conds = conds /. $dummy -> n //. op[ a_] :> a;
		{l, r} = {l, r} //. {
								e:(op[___, $dummy, ___]) :> DeleteCases[e, $dummy],
								op[] -> $dummy,
								op[ a_] :> a
							};
		If[ l === $dummy, l = n];
		If[ r === $dummy, r = n];

		If[ conds === {},
			$Failed,
			{If[ l === r, True, Equal$TM[ l, r]], conds}
		]
	]
	
collectFactors[ expr_, op_] :=
	Sort[ collectFactors[ expr, op, {}], (Length[ Last[ #1]] <= Length[ Last[ #2]])&]
collectFactors[ expr:(op_[ args___]), op_, pos_List] :=
	Prepend[ Flatten[ MapIndexed[ collectFactors[ #1, op, Join[ pos, #2]]&, {args}], 1], {expr, pos}]
collectFactors[ expr_, _, pos_List] := {{expr, pos}}


(* ::Section:: *)
(* Integer Intervals *)

inferenceRule[ RRPmembershipIntegerInterval] =
this:PRFSIT$[ g:FML$[ _, f_, __], k_List, _, rest___] :> performProofStep[
	Module[ {refs = $Failed, conds, other, rules, simp},
		conds = Switch[ f,
					_And$TM,
					List@@f,
					_,
					{f}
				];
		{conds, other} = splitIntegerConditions[ List@@conds];
		If[ conds === {},
			$Failed,
			
			If[ other === {},
				rules = getRulesFromRW[ getOptionalComponent[ this, "rewriting"], "type" -> "Backward", "ground" -> $backwardGround];
				{conds, refs, simp} = simplifyIntegerConditions[ conds, rules, k];
				If[ refs === $Failed || simp =!= True,
					$Failed,
					
					refs = DeleteDuplicates[ refs];
					If[ conds === {},
						makeTERMINALNODE[ makePRFINFO[ name -> RRPmembershipIntegerInterval, used -> {{g}, refs}, generated -> {}], proved],
						
						conds = makeRRPFML[ formula -> makeConjunction[ conds, And$TM], "marker" -> "G"];
						makeANDNODE[ makePRFINFO[ name -> RRPmembershipIntegerInterval, used -> {{g}, refs}, generated -> {{conds}}],
							RRPToBeProved[ goal -> conds, kb -> k, rest]
						]
					]
				],
				
				conds = makeRRPFML[ formula -> makeConjunction[ conds, And$TM], "marker" -> "C"];
				other = makeRRPFML[ formula -> makeConjunction[ other, And$TM], "marker" -> "G"];
				makeANDNODE[ makePRFINFO[ name -> RRPsplitConjunction, used -> {g}, generated -> {conds, other}],
					{RRPToBeProved[ goal -> conds, kb -> k, "proposedRules" -> {{RRPmembershipIntegerInterval}}, rest],
					RRPToBeProved[ goal -> other, kb -> addNewKB[ k, conds], rest]}
				]
			]	
		]
	]
]


(* ::Subsection:: *)
(* Auxiliary Functions *)

(* 'splitIntegerConditions[ conds ]' splits the list 'conds' into two sublists 'int' and 'other':
	- 'int' consists of those elements of 'conds' that are (universally quantified) relations of the form 'Element[ x, IntegerInterval[ a, b, True, True]]',
	- 'other' consists of all other elements.
*)
splitIntegerConditions[ conds_List] :=
	Module[ {int, other, qe, rng, tmp},
		{tmp, {int, other}} =
			Reap[ Scan[ Function[ c,
						Switch[ c,
							Element$TM[ _, _IntegerInterval$TM?isClosedInterval],
							Sow[ c, "int"],
							
							_Forall$TM,
							{qe, rng} = getQuantifiedExpression[ c];
							Switch[ qe,
								Element$TM[ _, _IntegerInterval$TM?isClosedInterval],
								Sow[ c, "int"],
								
								_,
								Sow[ c, "other"]
							],
							
							_,
							Sow[ c, "other"]
						]
					],
					DeleteDuplicates[ conds, alphaEquivalent]
				],
				{"int", "other"}
			];
		{Flatten[ int, 1], Flatten[ other, 1]}
	]
	
(* 'simplifyIntegerConditions[ conds, rules, kb ]' tries to simplify the integer-interval conditions 'cond',
	using the backward rules 'rules' and the knowledge 'kb'. It essentially calls 'proveIntegerConditions', i.e.
	unlike in, e.g., 'simplifyRingConditions', no simplification of the condtions w.r.t. each other takes place.
*)
simplifyIntegerConditions[ conds_List, rules_List, kb_List] :=
	Module[ {int, qe, rng},
		int = Map[ Function[ c,
					Switch[ c,
						Element$TM[ _, _IntegerInterval$TM?isClosedInterval],
						c,
						
						_Forall$TM,
						{qe, rng} = getQuantifiedExpression[ c];
						Switch[ qe,
							Element$TM[ _, _IntegerInterval$TM?isClosedInterval],
							c,
							
							_,
							$Failed
						],
						
						_,
						$Failed
					]
				], conds
			];
		int = Flatten[ int, 1];
		If[ int === {} || MemberQ[ int, $Failed],
			{$Failed, $Failed, False},
			proveIntegerConditions[ int, rules, kb]
		]
	]
	
isClosedInterval[ _[ _, _, True, True]] = True
isClosedInterval[ _[ DirectedInfinity[ -1|1], _, _, True]] = True
isClosedInterval[ _[ _, DirectedInfinity[ -1|1], True, _]] = True
isClosedInterval[ _[ DirectedInfinity[ -1|1], DirectedInfinity[ -1|1], _, _]] = True
isClosedInterval[ _] = False

tripleToCondition[ {a_, l_, r_}] := Element$TM[ a, IntegerInterval$TM[ l, r, !MatchQ[ l, _DirectedInfinity], !MatchQ[ r, _DirectedInfinity]]]

proveIntegerConditions[ conds_List, rules_List, kb_List] :=
	Module[ {known = {}, aux, auxZ, newConds = {}, cur, cnd, rf, refs = {}, simp = False, add, qe, rng},
		aux = Map[ (
					{qe, rng} = getQuantifiedExpression[ preprocessExpr[ #]];
					{{First[ qe], qe[[2, 1]], qe[[2, 2]]}, rng}
				)&,
				conds
			];
		(* It is desirable to have "bigger" intervals at the end. *)
		auxZ = Cases[ aux, {{_, DirectedInfinity[ -1], DirectedInfinity[ 1]}, _}];
		aux = Join[ DeleteCases[ aux, {{_, DirectedInfinity[ -1], DirectedInfinity[ 1]}, _}], auxZ];
		While[ aux =!= {},
			cur = First[ aux];
			{cnd, rf, add} = proveIntegerConditions[ cur, known, rules, kb];
			If[ add, AppendTo[ known, cur]];
			aux = Rest[ aux];
			If[ cnd === $Failed,
				AppendTo[ newConds, postprocessExpr[ thinnedRange[ tripleToCondition[ First[ cur]], Last[ cur], Expression]]],
				
				simp = True;
				Scan[ If[ MatchQ[ #, {{_, DirectedInfinity[ -1], DirectedInfinity[ 1]}, _}], AppendTo[ aux, #], PrependTo[ aux, #]]&, cnd];
				refs = Join[ refs, rf]
			]
		];
		{newConds, refs, simp}
	]
proveIntegerConditions[ {{x_Integer, a_, b_}, _}, _, _, _] /; (x >= a && x <= b) := {{}, {}, False}
proveIntegerConditions[ expr_, {___, k_, ___}, _, _] /; rangeEquivalent[ expr, k] := {{}, {}, False}
proveIntegerConditions[ {t_, r_}, _, rules_List, _] :=
	Module[ {out},
		{{}, First[out], True} /;
		(
			out = followsImmediately[ thinnedRange[ tripleToCondition[ t], r, Expression], rules, "RenameBound" -> False];
			MatchQ[ out, {_List, {}}]
		)
	]
proveIntegerConditions[ {{Subtract$TM[ b_, Subtract$TM[ a_, 1]], 1, b_}, r_}, _, _, _] :=
	{{thinnedRange[ {a, 1, b}, r], thinnedRange[ {b, DirectedInfinity[ -1], DirectedInfinity[ 1]}, r]}, {}, True}
proveIntegerConditions[ {{a_, a_, a_}, r_}, _, _, _] := {{{{a, DirectedInfinity[ -1], DirectedInfinity[ 1]}, r}}, {}, True}
proveIntegerConditions[ {{a:Except[DirectedInfinity[ -1]], a_, b_}, r_}, _, _, _] := {{{{a, DirectedInfinity[ -1], b}, r}}, {}, True}
proveIntegerConditions[ {{b:Except[DirectedInfinity[ 1]], a_, b_}, r_}, _, _, _] := {{{{b, a, DirectedInfinity[ 1]}, r}}, {}, True}
proveIntegerConditions[ {{BracketingBar$TM[ T_], _?NonPositive, DirectedInfinity[ 1]}, r_}, _, rules_List, kb_List] :=
	Module[ {refs, qe, rng},
		{{}, refs, True} /;
		(
			refs = Cases[ kb, FML$[_, _?(
						(
							{qe, rng} = getQuantifiedExpression[#];
	            			MatchQ[ qe, (isTuple$TM|(DomainOperation$TM[ Theorema`Knowledge`DomainTuples$TM[ _], Element$TM]))[ _]] &&
	            				rangeEquivalent[ {T, r}, {First[ qe], rng}]
	            		)&), __]
	            	];
				If[ refs === {},
					refs = followsImmediately[ thinnedRange[ isTuple$TM[ T], r, Expression], rules, "RenameBound" -> False];
					If[ MatchQ[ refs, {_List, {}}],
						refs = First[ refs];
						True,
						False
					],
					True
				]
		)
	]
proveIntegerConditions[ {{z_, a_, b_}, {}}, known_, rules_, kb_] :=
	Module[ {knownIntervals, kbIntervals, refs1, refs2, tmp, add1, add2, success = False},
		{{}, refs1, add1} /;
		(
			knownIntervals = Cases[ known, {{z, a2_, b2_}, {}} :> {a2, b2}];
			kbIntervals = Cases[ kb, f:FML$[ _, Element$TM[ z, i_IntegerInterval$TM?isClosedInterval], __] :> {{i[[1]], i[[2]]}, f}];
			Catch[
				Scan[
					With[ {x = #[[1, 1]], y = #[[1, 2]]},
						success = False;
						If[ x === a || a === DirectedInfinity[ -1],
							refs1 = {};
							add1 = True;
          					success = True,
          						
          					{tmp, refs1, add1} = proveIntegerConditions[ {{x, a, DirectedInfinity[ 1]}, {}}, known, rules, kb];
							success = (tmp === {})
						];
						If[ success,
							If[ y === b || b === DirectedInfinity[ 1],
								refs2 = {};
								add2 = True,
									
								{tmp, refs2, add2} = proveIntegerConditions[ {{y, DirectedInfinity[ -1], b}, {}}, known, rules, kb];
								success = (tmp === {})
							]
						];
						If[ success,
							refs1 = Join[ {Last[ #]}, refs1, refs2];
							add1 = add1 && add2;
							Throw[Null]
						]
					]&,
					kbIntervals
				]
			];
			If[ success =!= True,
				Catch[
					Scan[
						With[ {x = First[ #], y = Last[ #]},
							success = False;
							If[ x === a || a === DirectedInfinity[ -1],
								refs1 = {};
								success = True,
								
									{tmp, refs1, add1} = proveIntegerConditions[ {{x, a, DirectedInfinity[ 1]}, {}}, known, rules, kb];
								success = (tmp === {})
							];
							If[ success,
								If[ y === b || b === DirectedInfinity[ 1],
									refs2 = {},
									
									{tmp, refs2, add2} = proveIntegerConditions[ {{y, DirectedInfinity[ -1], b}, {}}, known, rules, kb];
									success = (tmp === {})
								]
							];
							If[ success,
								refs1 = Join[ refs1, refs2];
								add1 = False;
								Throw[Null]
							]
						]&,
						knownIntervals
					]
				]
			];
			success
		)
	]
proveIntegerConditions[ expr_, _, _, _] := decomposeIntegerConditions[ expr]

decomposeIntegerConditions[ {{(Plus$TM|Times$TM)[ args__], a:((_?NonNegative)|(DirectedInfinity[ -1])), DirectedInfinity[ 1]}, r_}] :=
	{Map[ {{#, a, DirectedInfinity[ 1]}, thinnedRange[ {#, a}, r, Range]}&, {args}], {}, True}
decomposeIntegerConditions[ {{Subtract$TM[ args__], DirectedInfinity[ -1], DirectedInfinity[ 1]}, r_}] :=
	{Map[ {{#, DirectedInfinity[ -1], DirectedInfinity[ 1]}, thinnedRange[ #, r, Range]}&, {args}], {}, True}
decomposeIntegerConditions[ {{Minus$TM[ arg_], DirectedInfinity[ -1], DirectedInfinity[ 1]}, r_}] :=
	{{{{arg, DirectedInfinity[ -1], DirectedInfinity[ 1]}, r}}, {}, True}
decomposeIntegerConditions[ _] := {$Failed, $Failed, True}


(* ::Section:: *)
(* Rule composition *)

(*
	Rule priorities:
	 0 - 19: termination + high priority
	20 - 24: medium priority, rewriting
	25 - 44: medium priority, goal-oriented
	45 - 64: medium priority, whole prfsit
	65 - 84: medium priority, kb-oriented
	85 - 89: low priority, goal-oriented
	90 - 94: low priority, kb-oriented
	95 - 99: low priority, whole prfsit + rewriting
*)

propositionalLogicRules = {"Propositional Logic Rules",
	{RRPgoalInKB, True, True, 1, "term"},
	{falseInKB, True, True, 1, "term"},
	{trueGoal, True, True, 1, "term"},
	{contradictionKB, True, True, 1, "term"},
	{RRPnotEquGoal, True, True, 40},
	{RRPandGoal, True, True, 85},
	{RRPandKB, True, False, 4},
	{RRPorGoal, True, True, 16},
	{RRPorKB, True, True, 91},
	{RRPimplGoalDirect, True, True, 10},
	{equivGoal, True, True, 41},
	{RRPequivKB, False, True, 90},
	{RRPsimpImplication, True, True, 70},
	{RRPmodusPonensTollens, True, True, 71},
	{RRPcontradiction, False, True, 99}
	};
	
miscRules = {"Miscellaneous Rules",
	{RRPexpandNegationGoal, False, True, 35},
	{RRPexpandNegationKB, True, True, 80},
	{RRPexpandPiecewise, True, True, 95},
	{RIexpandSuch, True, True, 94}
	};
	
quantifierRules = {"Quantifier Rules",
	{RRPforallGoal, True, True, 19},
	{RRPsplitForallGoal, False, True, 88},
	{RRPexistsKB, True, True, 65}
	};
	
rewriteRules = {"Rules for Rewriting",
	{RRPgenerateRewriteRules, True, False, 2},
	{RRPgoalRewriting, True, True, 44},
	{RRPelementarySubstitution, True, True, 22},
	{RRPgroundSubstitution, True, True, 21},
	{RRPexpandDef, True, True, 97},
	{RRPimplicitDef, True, True, 97}
	};
	
interactiveRules = {"Interactive Rules",
	{RRPorGoalInteractive, False, True, 15},
	{RRPexchangeGoalKBInteractive, False, True, 100, "interactive"},
	{RRPsplitForallGoalInteractive, False, True, 100, "interactive"},
	{RRPinstantiateInteractive, True, True, 100, "interactive"},
	{RRPselectImplicationInteractive, True, True, 100, "interactive"},
	{RRPrewritePS, False, False, 100, "interactive"},
	{RRPintroduceFormula, True, True, 100, "interactive"},
	{RRPaddAbbreviation, True, True, 100, "interactive"}
	};
	
managementRules = {"Management Rules",
	{RRPaddKnowledge, False, False, 100, "interactive"},
	{RIremoveKnowledge, False, False, 100, "interactive"},
	{RIhideKnowledge, False, False, 100, "interactive"}
	};
	
orderingRules = {"Ordering Rules",
	{RRPorderingGoal, True, True, 41},
	{RRPorderingKB, True, True, 81},
	{RRPorderingEqualGoal, True, True, 41}
	};
	
collectionRules = {"Domain-Collection Rules",
	{RRPmembershipDomainTuples, True, True, 41},
	{RRPmembershipDomainSets, True, True, 41}
	};
	
ringRules = {"CommRing1 Rules",
	{RRPmembershipCommRings1, True, True, 42},
	{RRPcommRing1Equal, True, True, 42},
	{RRPcommRing1EqualKB, False, True, 84}
	};
	
monoidRules = {"CCMonoid Rules",
	{RRPmembershipCCMonoids, True, True, 42},
	{RRPCCMonoidEqual, True, True, 42}
	};
	
integerRules = {"Integer Rules",
	{RRPmembershipIntegerInterval, True, True, 42}
	};
	
(* All but "integerRules", "orderingRules", "collectionRules", "monoidRules" and "ringRules" are exactly the same as in the RewriteInteractiveProver. *)

registerRuleSet[ "Prover for the theory of Reduction Rings", reductionRingProver,
		{propositionalLogicRules, miscRules, quantifierRules, rewriteRules, interactiveRules, managementRules, integerRules, orderingRules, collectionRules, monoidRules, ringRules}
	]

End[]

EndPackage[]