(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 9.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[   2642655,      71160]
NotebookOptionsPosition[   2124793,      57230]
NotebookOutlinePosition[   2472808,      66091]
CellTagsIndexPosition[   2437087,      65203]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["Gr\[ODoubleDot]bner Rings", "Title",
 CellChangeTimes->{{3.590837152246185*^9, 3.590837155585586*^9}, 
   3.607771716512438*^9},
 TextAlignment->Center,
 CellID->1451885783],

Cell[CellGroupData[{

Cell["Preliminaries", "Section",
 CellChangeTimes->{{3.61338381272198*^9, 3.613383813722398*^9}, {
  3.613704212916864*^9, 3.613704214914052*^9}, {3.645519657680581*^9, 
  3.645519661426262*^9}},
 CellID->1605568989],

Cell[CellGroupData[{

Cell["Description", "Subsection",
 CellChangeTimes->{{3.660734560434437*^9, 3.6607345624780407`*^9}},
 CellID->499915303],

Cell[TextData[{
 "This notebook defines the functor \[OpenCurlyDoubleQuote]",
 Cell[BoxData[
  FormBox[
   RowBox[{"GroebnerRing", "[", "\[ScriptCapitalR]", "]"}], TraditionalForm]]],
 "\[CloseCurlyDoubleQuote] as an extension of the domain \[ScriptCapitalR] by \
an implementation of ",
 StyleBox["Buchberger\[CloseCurlyQuote]s algorithm",
  FontSlant->"Italic"],
 " for computing ",
 StyleBox["Gr\[ODoubleDot]bner bases",
  FontSlant->"Italic"],
 " of finite tuples in \[ScriptCapitalR]. The function that implements the \
algorithm is called GB and makes use of a range of auxiliary functions, most \
notably functions for (totally) reducing a given element of \[ScriptCapitalR] \
modulo a tuple of such elements. Furthermore, function GB takes the ",
 StyleBox["chain criterion",
  FontSlant->"Italic"],
 " for avoiding useless reductions into account."
}], "Text",
 CellChangeTimes->{{3.6607345921124926`*^9, 3.660734808173872*^9}},
 CellID->29975522],

Cell[TextData[{
 "Apart from several sample computations in section \
\[OpenCurlyDoubleQuote]Tests\[CloseCurlyDoubleQuote], the main part of this \
notebook is concerned with proving function GB correct, i.e. proving that, \
for any given tuple of elements of \[ScriptCapitalR], ",
 Cell[BoxData[
  FormBox[
   RowBox[{"GB", "[", "T", "]"}], TraditionalForm]]]
}], "Text",
 CellChangeTimes->{{3.6607348438823347`*^9, 3.6607349355480957`*^9}},
 CellID->57224452],

Cell[CellGroupData[{

Cell["\<\
is again a tuple of elements of \[ScriptCapitalR],\
\>", "Item",
 CellChangeTimes->{{3.660734916781263*^9, 3.660734924222476*^9}},
 CellID->45190291],

Cell[TextData[{
 "generates the same ideal as ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 " in \[ScriptCapitalR], and"
}], "Item",
 CellChangeTimes->{{3.660734916781263*^9, 3.660734944736512*^9}},
 CellID->90776693],

Cell["\<\
is a Gr\[ODoubleDot]bner basis according to the definition given in \
\[OpenCurlyDoubleQuote]ReductionRings.nb\[CloseCurlyDoubleQuote].\
\>", "Item",
 CellChangeTimes->{{3.660734916781263*^9, 3.6607349714593587`*^9}},
 CellID->8961680]
}, Open  ]],

Cell["\<\
Of course, GB can only be shown to satisfy these requirements if \
\[ScriptCapitalR] is known to possess some algebraic structure: It has to be \
an algorithmic reduction ring (again according to the definition from \
\[OpenCurlyDoubleQuote]ReductionRings.nb\[CloseCurlyDoubleQuote]). This is \
assumed throughout sections \[OpenCurlyDoubleQuote]Correctness of Auxiliary \
Functions\[CloseCurlyDoubleQuote] and \[OpenCurlyDoubleQuote]Correctness of \
Function GB\[CloseCurlyDoubleQuote].\
\>", "Text",
 CellChangeTimes->{{3.6607349815525765`*^9, 3.660735200212785*^9}, {
  3.660735240390083*^9, 3.6607352406951003`*^9}},
 CellID->71065762],

Cell["\<\
The correctness proof, of course, relies heavily on results proved in \
\[OpenCurlyDoubleQuote]ReductionRings.nb\[CloseCurlyDoubleQuote] about \
confluence properties of the underlying reduction relation. One of its key \
steps is to state and prove an induction rule for proving properties for all \
possible inputs of function GBAux (i.e. the main auxiliary function GB is \
defined in terms of; this function is defined recursively). This induction \
rule, in turn, relies on a Noetherian (i.e. well-founded) order relation on \
the set of possible inputs, that moreover can be used to prove termination of \
GBAux (see Subsection \[OpenCurlyDoubleQuote]Termination & Induction Schemas\
\[CloseCurlyDoubleQuote]).
Once having the induction rule all of the three verification obligations are \
discharged by finding sufficiently powerful statements about the arguments of \
GBAux and proving them invariant under recursive calls \[Dash] of course by \
induction.\
\>", "Text",
 CellChangeTimes->{{3.6607352428172216`*^9, 3.6607354389314384`*^9}, {
  3.6607354800787926`*^9, 3.660735506210287*^9}, {3.660735557634228*^9, 
  3.6607356684865685`*^9}, {3.660735712441083*^9, 3.6607357412037277`*^9}},
 CellID->10851318]
}, Closed]],

Cell[CellGroupData[{

Cell["Theory Packages", "Subsection",
 CellChangeTimes->{{3.655465027736725*^9, 3.65546503045113*^9}},
 CellID->16638123],

Cell[BoxData[
 RowBox[{
  RowBox[{"$PackageDirectory", "=", 
   RowBox[{"FileNameJoin", "[", 
    RowBox[{"{", 
     RowBox[{
     "$RootDirectory", ",", "\"\<scratch\>\"", ",", "\"\<amaletzk\>\"", ",", 
      "\"\<Theorema Environment Data\>\""}], "}"}], "]"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.666438783233775*^9, 3.666438821083324*^9}, 
   3.666693954745158*^9},
 CellID->498486718],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<LogicSets.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554650438047533`*^9, 3.6554651094340687`*^9}, {
   3.661601041781954*^9, 3.661601058863984*^9}, {3.6616010933868446`*^9, 
   3.6616011050088654`*^9}, {3.662099216256549*^9, 3.662099218081752*^9}, 
   3.6666939951209393`*^9},
 CellID->119228788],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<NatInt.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554650438047533`*^9, 3.6554651094340687`*^9}, {
   3.661601041781954*^9, 3.661601058863984*^9}, {3.6616010933868446`*^9, 
   3.6616011050088654`*^9}, 3.666693996159422*^9},
 CellID->297313380],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<Tuples.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554650438047533`*^9, 3.6554651094340687`*^9}, {
   3.661601041781954*^9, 3.661601075992814*^9}, {3.661601107099269*^9, 
   3.6616011073488693`*^9}, 3.666693997064526*^9},
 CellID->86850003],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<Functors.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554650438047533`*^9, 3.6554651094340687`*^9}, {
   3.661601041781954*^9, 3.661601109610873*^9}, 3.66669399791951*^9},
 CellID->131034841],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<ReductionRings.m\>\""}], "}"}], 
   "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.6554650438047533`*^9, 3.6554651094340687`*^9}, {
   3.661601041781954*^9, 3.6616011124344783`*^9}, 3.6666939987524223`*^9},
 CellID->477360655],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<GroebnerRings.m\>\""}], "}"}], 
   "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.655465034553937*^9, 3.655465038734744*^9}, {
  3.6666939757608967`*^9, 3.666693999824239*^9}},
 CellID->123628739],

Cell[BoxData[
 RowBox[{"MakeTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<GroebnerRings.m\>\""}], "}"}], 
   "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.6554651139424763`*^9, 3.655465120978089*^9}, 
   3.666939421630968*^9, 3.6669394522475357`*^9, {3.6669396576080227`*^9, 
   3.666939657840248*^9}},
 CellID->115117422],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DeleteCases", "[", 
  RowBox[{
   RowBox[{"Map", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Quiet", "[", 
       RowBox[{"Check", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"With", "[", 
           RowBox[{
            RowBox[{"{", 
             RowBox[{"s", "=", 
              RowBox[{"\"\<VAR$\>\"", "<>", "#", "<>", "\"\<$TM\>\""}]}], 
             "}"}], ",", 
            RowBox[{"Context", "[", "s", "]"}]}], "]"}], ";", "#"}], ",", 
         "Null"}], "]"}], "]"}], "&"}], ",", 
     RowBox[{"{", 
      RowBox[{
      "\"\<i\>\"", ",", "\"\<j\>\"", ",", "\"\<n\>\"", ",", "\"\<m\>\"", ",", 
       "\"\<P\>\"", ",", "\"\<x\>\"", ",", "\"\<y\>\"", ",", "\"\<z\>\"", 
       ",", "\"\<a\>\"", ",", "\"\<A\>\"", ",", "\"\<b\>\"", ",", "\"\<c\>\"",
        ",", "\"\<\[ScriptCapitalR]\>\"", ",", "\"\<\[ScriptCapitalL]\>\"", 
       ",", "\"\<\[ScriptCapitalG]\>\"", ",", "\"\<t\>\""}], "}"}]}], "]"}], 
   ",", "Null"}], "]"}]], "Input",
 CellChangeTimes->{{3.6592618519471025`*^9, 3.6592618631011224`*^9}, {
   3.659681175379635*^9, 3.659681185706853*^9}, 3.659683710620489*^9, 
   3.65968441725893*^9, 3.6597695015379353`*^9, {3.6597698034056654`*^9, 
   3.659769808569275*^9}, {3.6597702854721127`*^9, 3.659770290573322*^9}, {
   3.659848272310069*^9, 3.6598483467534*^9}, {3.6598515848824887`*^9, 
   3.6598516763142495`*^9}, {3.6598555136659927`*^9, 3.659855522090008*^9}, {
   3.6598559051642804`*^9, 3.659855919516306*^9}, {3.6602862205309095`*^9, 
   3.6602862222781124`*^9}, {3.660479643354128*^9, 3.660479646380533*^9}, {
   3.6616030717375193`*^9, 3.661603091253154*^9}},
 CellID->20666029],

Cell[BoxData[
 RowBox[{"{", "}"}]], "Output",
 CellChangeTimes->{{3.6616030688983145`*^9, 3.6616030927819567`*^9}, 
   3.6616612048340015`*^9, 3.661669205818514*^9, 3.6616695620763397`*^9, 
   3.6616697551920795`*^9, 3.662097318815398*^9, 3.6620984270485477`*^9, 
   3.6621118504805727`*^9, 3.662112088500394*^9, 3.662115545411705*^9, 
   3.666175061844001*^9, 3.6661752587057467`*^9},
 CellID->468046677]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Theory Analysis", "Subsection",
 CellChangeTimes->{{3.640668830437483*^9, 3.640668834318727*^9}},
 CellID->1386333768],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"First", "[", 
  RowBox[{"Timing", "[", 
   RowBox[{"AnalyzeProofs", "[", "\"\<Theory\>\"", "]"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6406688560104437`*^9, 3.640668867513785*^9}, 
   3.64067664060292*^9, 3.644222019629964*^9, {3.661669216676133*^9, 
   3.6616692424473786`*^9}},
 CellID->1039312110],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"/home/amaletzk/Theorema \
Notebooks/Reduction Ring Theory/Fields\\\" (\"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"8\"\>", 
   "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields\" (", 1, " of ", "8", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLSMWc0efaPX
jgAPWJl0
  "],
 CellID->880067753],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"/home/amaletzk/Theorema \
Notebooks/Reduction Ring Theory/GroebnerRings\\\" (\"\>", "\[InvisibleSpace]",
    "2", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"8\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings\" (", 2, " of ", "8", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPSnVTV79I1e
OwIAGTSavg==
  "],
 CellID->604617084],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"164\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "164"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPRFtto9+kav
HQETtpn6
  "],
 CellID->964310657],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"164\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "164"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLRL3cI9+kav
HQETbpoJ
  "],
 CellID->1623802069],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"164\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "164"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLSH64k9+kav
HQES6Zn7
  "],
 CellID->1392221509],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"164\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "164"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgDRDyrs9+kav
HQESZpn4
  "],
 CellID->1474649765],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"/home/amaletzk/Theorema \
Notebooks/Reduction Ring Theory/IntegerQuotientRings\\\" (\"\>", 
   "\[InvisibleSpace]", "3", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"8\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings\" (", 3, " of ", "8", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPQpg1979I1e
OwIAF5Kamg==
  "],
 CellID->1623284855],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"/home/amaletzk/Theorema \
Notebooks/Reduction Ring Theory/Integers\\\" (\"\>", "\[InvisibleSpace]", "4",
    "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"8\"\>", 
   "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers\" (", 4, " of ", "8", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLSB7989+kav
HQETn5og
  "],
 CellID->2059021584],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"/home/amaletzk/Theorema \
Notebooks/Reduction Ring Theory/Polynomials\\\" (\"\>", "\[InvisibleSpace]", 
   "5", "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"8\"\>",
    "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials\" (", 5, " of ", "8", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPSbkr979I1e
OwIAGpCbAw==
  "],
 CellID->468682694],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"325\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "325"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPSreX/36Bu9
dgQAG6abKw==
  "],
 CellID->768345119],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"325\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "325"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPSCIoG9+kav
HQESmZnJ
  "],
 CellID->203140207],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"325\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "325"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgHSeqsZefaPX
jgAPfpli
  "],
 CellID->839315771],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"325\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "325"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgHSPt91efaPX
jgAR/Jm8
  "],
 CellID->1909090729],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"325\"\>"}],
  SequenceForm["Reading proof ", 201, " of ", "325"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLS8YvBefaPX
jgAN7Jk6
  "],
 CellID->781859586],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"325\"\>"}],
  SequenceForm["Reading proof ", 251, " of ", "325"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPSNkKy9+kav
HQEVo5o9
  "],
 CellID->792515171],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"325\"\>"}],
  SequenceForm["Reading proof ", 301, " of ", "325"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPRFg7q9+kav
HQEU55om
  "],
 CellID->2077551358],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"/home/amaletzk/Theorema \
Notebooks/Reduction Ring Theory/PolyTuples\\\" (\"\>", "\[InvisibleSpace]", 
   "6", "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"8\"\>",
    "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples\" (", 6, " of ", "8", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgHROV8defaPX
jgAUcZol
  "],
 CellID->385817707],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"/home/amaletzk/Theorema \
Notebooks/Reduction Ring Theory/ReductionRings\\\" (\"\>", 
   "\[InvisibleSpace]", "7", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"8\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings\" (", 7, " of ", "8", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgHTbo+69+kav
HQEXu5qa
  "],
 CellID->1914681352],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"236\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "236"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPQzzp69+kav
HQEU0poi
  "],
 CellID->309156947],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"236\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "236"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgHTGmUV79Y1e
OwIAFruafQ==
  "],
 CellID->538743880],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"236\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "236"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLTkjX179Y1e
OwIAFT+aVg==
  "],
 CellID->1509225573],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"236\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "236"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgHSD7p29+kav
HQEUfpow
  "],
 CellID->599128823],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"236\"\>"}],
  SequenceForm["Reading proof ", 201, " of ", "236"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPT3Cb/26hu9
dgQAG5+bKA==
  "],
 CellID->448464536],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"/home/amaletzk/Theorema \
Notebooks/ElementaryTheories\\\" (\"\>", "\[InvisibleSpace]", "8", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"8\"\>", 
   "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"/home/amaletzk/Theorema Notebooks/ElementaryTheories\" \
(", 8, " of ", "8", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLTpDP59+kav
HQEQSpmE
  "],
 CellID->726826737],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"384\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "384"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLQ9j/A+faPX
jgAM3pkG
  "],
 CellID->519974804],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"384\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "384"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPQ+ZfV9+kav
HQER75mw
  "],
 CellID->1538620868],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"384\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "384"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLRYj8M+faPX
jgAQJJmK
  "],
 CellID->2116298062],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"384\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "384"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPSXN+H79I1e
OwIAGj6a3w==
  "],
 CellID->1975937998],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"384\"\>"}],
  SequenceForm["Reading proof ", 201, " of ", "384"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgDRTRuE+faPX
jgAPrpmD
  "],
 CellID->1696687010],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"384\"\>"}],
  SequenceForm["Reading proof ", 251, " of ", "384"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPSW9LZ9+kav
HQEVtZpJ
  "],
 CellID->1940201374],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"384\"\>"}],
  SequenceForm["Reading proof ", 301, " of ", "384"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLTbpHn79I1e
OwIAFAKaHg==
  "],
 CellID->1823397800],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"384\"\>"}],
  SequenceForm["Reading proof ", 351, " of ", "384"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLSp5NZ9+kav
HQEQtZmr
  "],
 CellID->446834775],

Cell[BoxData["\<\"Creating forward rules:\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgDQzx9F9+kav
HQEPDpl4
  "],
 CellID->1578517663],

Cell[BoxData["\<\"Preparing ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLTD3xP79I1e
OwIAF7uarQ==
  "],
 CellID->31752406],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 1, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgHSUkvQBfaPX
jgAOiplB
  "],
 CellID->1289398078],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 51, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLSVWuwBfaPX
jgAPMpln
  "],
 CellID->1707398801],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 101, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLRlz5ID+kav
HQETnpoT
  "],
 CellID->480147046],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 151, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPQyrXcH9I1e
OwIAFhSaaA==
  "],
 CellID->1571230922],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 201, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgHTQPaOD+kav
HQET/ZoN
  "],
 CellID->1160857634],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 251, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPSUg8UH9Y1e
OwIAFsiacw==
  "],
 CellID->1266278869],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 301, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLQY19aD+kav
HQEPY5mA
  "],
 CellID->1832033513],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 351, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLTRgW8H9Y1e
OwIAFsOakw==
  "],
 CellID->1459714591],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "401", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 401, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPQ9EfND+kav
HQES+pnV
  "],
 CellID->372738200],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "451", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 451, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPS5zvJD+kav
HQEXLZp6
  "],
 CellID->1119015511],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "501", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 501, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLT/pO2H9I1e
OwIAFPSaRA==
  "],
 CellID->266406868],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "551", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 551, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgHSvy89D+kav
HQEWTpp2
  "],
 CellID->1662811202],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "601", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 601, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPRqQ+vD+kav
HQESSpnE
  "],
 CellID->905324562],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "651", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 651, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLRfUd1hfaPX
jgASu5nr
  "],
 CellID->648137395],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "701", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 701, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgHTfx3OH9Y1e
OwIAGhSa+g==
  "],
 CellID->285997184],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "751", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 751, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLSbicQRfaPX
jgAOaplA
  "],
 CellID->959440624],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "801", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 801, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPQl49gj+kav
HQEUYZoQ
  "],
 CellID->1729084317],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "851", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 851, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPRemXlH9I1e
OwIAFJ6aJQ==
  "],
 CellID->318527011],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "901", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 901, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPTsq/eP6Bu9
dgQAGiOa/Q==
  "],
 CellID->1995813270],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "951", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 951, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLSVocpRfaPX
jgAOQpk+
  "],
 CellID->683642649],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1001", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 1001, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLRgcsZRfaPX
jgAP8JmL
  "],
 CellID->2093499806],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1051", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 1051, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPQRlpVH9Y1e
OwIAFHWaIA==
  "],
 CellID->1424876236],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 1101, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgHTP8xdH9Y1e
OwIAGmSbCg==
  "],
 CellID->187493898],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 1151, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPQGc81j+kav
HQESP5nA
  "],
 CellID->138253922],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 1201, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgHRYaOYxfaPX
jgARwZnE
  "],
 CellID->2019201381],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 1251, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPQn+5XH9I1e
OwIAF4eaig==
  "],
 CellID->170458589],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 1301, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgLS77stj+kav
HQETMZoN
  "],
 CellID->346169200],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 1351, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgHTCKY3j+kav
HQETw5oD
  "],
 CellID->859857300],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1401", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 1401, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPTZluzj+kav
HQEW05pt
  "],
 CellID->101944774],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1451", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "1474"}],
  SequenceForm["Creating rule ", 1451, " of ", 1474],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgPTM0NXH9Y1e
OwIAFWqaSg==
  "],
 CellID->313520785],

Cell[BoxData["\<\"Done\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAuKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/qeR8vwYkI7Z3Zm91eC14xJuu70PgHTA4ZPH9Y1e
OwIAFtiajQ==
  "],
 CellID->1157430104]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"First", "[", 
  RowBox[{"Timing", "[", 
   RowBox[{"AnalyzeProofs", "[", "]"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.6406688560104437`*^9, 3.640668867513785*^9}, 
   3.64067664060292*^9, 3.644222019629964*^9, 3.648883848335284*^9, {
   3.6616692230565443`*^9, 3.6616692330873623`*^9}},
 CellID->892463300],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\GroebnerRings\\\" (\"\>", 
   "\[InvisibleSpace]", "1", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"1\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\GroebnerRings\" (", 1, " of ", "1", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.6620966361169987`*^9},
 CellID->564482905],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"170\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "170"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.6620966362729993`*^9},
 CellID->306351101],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"170\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "170"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.662096642076209*^9},
 CellID->726477269],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"170\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "170"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.66209664825382*^9},
 CellID->4348995],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"170\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "170"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.66209665390103*^9},
 CellID->41231562],

Cell[BoxData["\<\"Creating forward rules:\"\>"], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.662096656365834*^9},
 CellID->360733829],

Cell[BoxData["\<\"Preparing ...\"\>"], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.6620966563814344`*^9},
 CellID->816780095],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "420"}],
  SequenceForm["Creating rule ", 1, " of ", 420],
  Editable->False]], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.6620966564594345`*^9},
 CellID->352919787],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "420"}],
  SequenceForm["Creating rule ", 51, " of ", 420],
  Editable->False]], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.662096656662235*^9},
 CellID->523041642],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "420"}],
  SequenceForm["Creating rule ", 101, " of ", 420],
  Editable->False]], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.662096656849435*^9},
 CellID->379549855],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "420"}],
  SequenceForm["Creating rule ", 151, " of ", 420],
  Editable->False]], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.6620966570522356`*^9},
 CellID->68832474],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "420"}],
  SequenceForm["Creating rule ", 201, " of ", 420],
  Editable->False]], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.6620966572550364`*^9},
 CellID->173122215],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "420"}],
  SequenceForm["Creating rule ", 251, " of ", 420],
  Editable->False]], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.662096657457836*^9},
 CellID->530215840],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "420"}],
  SequenceForm["Creating rule ", 301, " of ", 420],
  Editable->False]], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.6620966576450367`*^9},
 CellID->298553875],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "420"}],
  SequenceForm["Creating rule ", 351, " of ", 420],
  Editable->False]], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.662096657847837*^9},
 CellID->499074680],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "401", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "420"}],
  SequenceForm["Creating rule ", 401, " of ", 420],
  Editable->False]], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.6620966580662374`*^9},
 CellID->23924730],

Cell[BoxData["\<\"Creating thinned data ...\"\>"], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.6620966581442375`*^9},
 CellID->429839575],

Cell[BoxData["\<\"Creating theory data ...\"\>"], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.662096658175438*^9},
 CellID->115076950],

Cell[BoxData["\<\"Creating backward rules ...\"\>"], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.662096658175438*^9},
 CellID->11557326],

Cell[BoxData["\<\"Creating forward rules ...\"\>"], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.6620966581910377`*^9},
 CellID->386496720],

Cell[BoxData["\<\"Preparing ...\"\>"], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.6620966581910377`*^9},
 CellID->20784557],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "9"}],
  SequenceForm["Creating rule ", 1, " of ", 9],
  Editable->False]], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.6620966582066374`*^9},
 CellID->67136478],

Cell[BoxData["\<\"Done\"\>"], "Print",
 CellChangeTimes->{3.6616694570569553`*^9, 3.6620966582066374`*^9},
 CellID->27448579]
}, Open  ]],

Cell[BoxData["9.17285880000000020118022803217172622681`6.983104621640288"], \
"Output",
 CellChangeTimes->{3.6616694819545994`*^9, 3.6620966582222376`*^9},
 CellID->368709728]
}, Closed]],

Cell[BoxData[
 RowBox[{
  RowBox[{"ComputeFormulaPositions", "[", "]"}], ";"}]], "Input",
 CellChangeTimes->{{3.6406695270485573`*^9, 3.6406695415951242`*^9}, {
   3.64067664934828*^9, 3.640676663567005*^9}, 3.6412902845300865`*^9, 
   3.641290346822996*^9, {3.6442219079512787`*^9, 3.644221909273131*^9}},
 CellID->19428626],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CheckIntegrity", "[", 
  RowBox[{"\"\<CheckPositions\>\"", "\[Rule]", 
   RowBox[{"{", 
    RowBox[{"FileNameJoin", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"CurrentValue", "[", "\"\<NotebookDirectory\>\"", "]"}], ",", 
       RowBox[{"CurrentValue", "[", "\"\<NotebookFileName\>\"", "]"}]}], 
      "}"}], "]"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.64067003109768*^9, 3.6406700374311132`*^9}, {
   3.6406797903560762`*^9, 3.640679792950843*^9}, 3.6442220291527653`*^9},
 CellID->1909803754],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{
  3.640670040858255*^9, 3.640670099679185*^9, 3.640670397721546*^9, 
   3.640679793300535*^9, 3.6406798406436234`*^9, 3.640679891162312*^9, 
   3.640679971009297*^9, 3.64068009182163*^9, 3.640680173340745*^9, 
   3.641037395456129*^9, 3.641116403388618*^9, 3.641286881014108*^9, 
   3.642225466578094*^9, 3.642234672070362*^9, 3.642240151894621*^9, 
   3.6423063269273643`*^9, {3.642308832539488*^9, 3.642308857563932*^9}, 
   3.642321396697977*^9, 3.6423285030212173`*^9, 3.642493582768566*^9, 
   3.642763270724422*^9, 3.642834604308413*^9, 3.643428832901271*^9, 
   3.6434441292624083`*^9, 3.643532851309215*^9, 3.644918543299676*^9, 
   3.645518304732339*^9, 3.649416144914283*^9, 3.649504862206789*^9},
 CellID->797779344]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Cases", "[", 
   RowBox[{
    RowBox[{"NotebookGet", "[", "]"}], ",", 
    RowBox[{"Cell", "[", 
     RowBox[{"_", ",", "\"\<FormalTextInputFormula\>\"", ",", "___"}], "]"}], 
    ",", 
    RowBox[{"{", 
     RowBox[{"0", ",", "Infinity"}], "}"}]}], "]"}], "//", 
  "Length"}]], "Input",
 CellChangeTimes->{{3.64966454060702*^9, 3.649664569556057*^9}},
 CellID->18261744],

Cell[BoxData["225"], "Output",
 CellChangeTimes->{3.6496645707478724`*^9},
 CellID->772937485]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Auxiliary Functions", "Subsection",
 CellChangeTimes->{{3.640668579267561*^9, 3.640668583347287*^9}},
 CellID->1454182661],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "LoadPackage", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"LoadPackage", "[", "name_String", "]"}], ":=", 
   RowBox[{"LoadPackage", "[", 
    RowBox[{"{", "name", "}"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"LoadPackage", "[", 
    RowBox[{"l", ":", 
     RowBox[{"{", "___String", "}"}]}], "]"}], ":=", 
   RowBox[{"Scan", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Get", "[", 
       RowBox[{"FileNameJoin", "[", 
        RowBox[{"{", 
         RowBox[{"$HomeDirectory", ",", "\"\<Theorema Notebooks\>\"", ",", 
          RowBox[{"#", "<>", "\"\<.m\>\""}]}], "}"}], "]"}], "]"}], "&"}], 
     ",", "l"}], "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.640668603629292*^9, 3.640668727460284*^9}, 
   3.641286816086794*^9, 3.6434459961973133`*^9, {3.66160074884464*^9, 
   3.661600764101466*^9}, {3.666439821218811*^9, 3.6664398289628153`*^9}},
 CellID->2074904605],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "FindFormula", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"FindFormula", "[", "label_String", "]"}], ":=", 
   RowBox[{"NotebookLocate", "[", "label", "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.640677756847624*^9, 3.6406777896243896`*^9}, 
   3.640677822869519*^9, {3.640677854902379*^9, 3.6406778579321823`*^9}},
 CellID->1980565193],

Cell[BoxData[
 RowBox[{"LoadPackage", "[", "\"\<ProofPalette\>\"", "]"}]], "Input",
 CellChangeTimes->{{3.640668742006607*^9, 3.640668751417616*^9}},
 CellID->444074165],

Cell[BoxData[
 RowBox[{"LoadPackage", "[", "\"\<TheoryTools/TheoryTools\>\"", 
  "]"}]], "Input",
 CellChangeTimes->{{3.640668759133181*^9, 3.640668768089197*^9}, {
   3.6616007773458896`*^9, 3.661600781542297*^9}, 3.666439832071404*^9},
 CellID->798575911]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Auxiliary Notions (\[Checkmark])", "Section",
 CellChangeTimes->{{3.593495725784487*^9, 3.593495732147601*^9}, {
  3.59377076619969*^9, 3.59377076728689*^9}, {3.649414316899413*^9, 
  3.649414320239662*^9}},
 CellID->1452921687],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " Notions"
}], "Subsection",
 CellChangeTimes->{{3.594017382923796*^9, 3.594017387248188*^9}, {
  3.645768728746114*^9, 3.645768729755661*^9}},
 CellID->287260853],

Cell["", "OpenEnvironment",
 CellID->207462652],

Cell[CellGroupData[{

Cell[TextData[{
 "Definition: ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " Notions"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.594017427286558*^9, 3.594017434141912*^9}},
 CellID->608117277],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"x", "..."}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"print", "[", 
     RowBox[{"x", "..."}], "]"}], ":=", 
    RowBox[{"\[Wolf]Print", "[", 
     RowBox[{"x", "..."}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(print)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1079943823", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6314495144447527`*^9, 3.6314495221558323`*^9}, {
  3.6457686420533047`*^9, 3.6457686528983603`*^9}, {3.64829063522996*^9, 
  3.648290639639655*^9}},
 CellTags->{
  "ID:1079943823", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "print"},
 CellID->1079943823],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1723749034]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->712958051]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Theorema",
  FontSlant->"Italic"],
 " Notions (\[Checkmark])"
}], "Subsection",
 CellChangeTimes->{{3.593770761517057*^9, 3.593770763408195*^9}, {
  3.594017356748529*^9, 3.594017360442698*^9}, {3.645768732336473*^9, 
  3.6457687332692327`*^9}, {3.649414308885401*^9, 3.6494143124287167`*^9}},
 CellID->334464829],

Cell[CellGroupData[{

Cell["Definitions", "Subsubsection",
 CellChangeTimes->{{3.594017368357338*^9, 3.594017370074435*^9}},
 CellID->657395164],

Cell["", "OpenEnvironment",
 CellID->704307368],

Cell[CellGroupData[{

Cell["Definition: processed", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.645768446806375*^9, 3.645768448634256*^9}},
 CellID->2046755789],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"T", ",", "i", ",", "j"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"processed", "[", 
     RowBox[{"i", ",", "j", ",", "T"}], "]"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         SubscriptBox["T", "k"], "\[NotEqual]", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"i", ",", "j"}], "\[RightAngleBracket]"}]}], "\[And]", 
        RowBox[{
         SubscriptBox["T", "k"], "\[NotEqual]", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"j", ",", "i"}], "\[RightAngleBracket]"}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(processed)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1033951332", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645768470371194*^9, 3.645768535237401*^9}},
 CellTags->{
  "ID:1033951332", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed"},
 CellID->1033951332],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->547704634]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->95587551],

Cell["", "OpenEnvironment",
 CellID->546908141],

Cell[CellGroupData[{

Cell["Definition: allPairs, pairs", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.59081191489364*^9, 3.590811920306068*^9}, {
  3.593765262339091*^9, 3.593765263674924*^9}, {3.594638995456747*^9, 
  3.594638999682806*^9}},
 CellID->227588461],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"k", ",", "j"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1967331668}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590811924672281*^9, 3.590811931528107*^9}, {
   3.590811978386294*^9, 3.590811989901865*^9}, {3.594638033720185*^9, 
   3.594638061480287*^9}, {3.594638128033971*^9, 3.594638128256049*^9}, 
   3.645768490104237*^9},
 CellTags->{
  "ID:1967331668", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1967331668],

Cell[BoxData[
 RowBox[{
  RowBox[{"pairs", "[", "k", "]"}], ":=", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"i", ",", "k"}], "\[RightAngleBracket]"}], 
    UnderscriptBox["|", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"k", "-", "1"}]}]]}], 
   "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(pairs 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1698292943", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6482919948824797`*^9, 3.648292007739118*^9}, {
  3.648292073155397*^9, 3.648292092832309*^9}},
 CellTags->{
  "ID:1698292943", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs 1"},
 CellID->1698292943],

Cell[BoxData[
 RowBox[{
  RowBox[{"pairs", "[", 
   RowBox[{"j", ",", "k"}], "]"}], ":=", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Piecewise]", GridBox[{
      {
       RowBox[{
        RowBox[{"pairs", "[", "j", "]"}], 
        TagBox["\:22c8",
         Identity,
         SyntaxForm->"a*b"], 
        RowBox[{"pairs", "[", 
         RowBox[{
          RowBox[{"j", "+", "1"}], ",", "k"}], "]"}]}], "\[DoubleLeftArrow]", 
       
       RowBox[{"k", "\[GreaterEqual]", "j"}]},
      {
       RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
       "\[DoubleLeftArrow]", "otherwise"}
     }]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(pairs 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1961920283", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594638122733042*^9, 3.594638212997779*^9}, {
   3.594638324570799*^9, 3.594638349462316*^9}, 3.594638407007603*^9, 
   3.594639074585411*^9, {3.606126392866226*^9, 3.606126394083959*^9}, {
   3.60612684336094*^9, 3.6061268505053368`*^9}, {3.615713528680735*^9, 
   3.615713537022277*^9}, {3.648292208657324*^9, 3.648292210477517*^9}, {
   3.648292294000093*^9, 3.648292325539136*^9}},
 CellTags->{
  "ID:1961920283", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs 2"},
 CellID->1961920283],

Cell[BoxData[
 RowBox[{
  RowBox[{"allPairs", "[", "k", "]"}], ":=", 
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"i", ",", "i"}], "\[RightAngleBracket]"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "k"}]]}], 
    "\[RightAngleBracket]"}], 
   TagBox["\:22c8",
    Identity,
    SyntaxForm->"a*b"], 
   RowBox[{"pairs", "[", 
    RowBox[{"2", ",", "k"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(allPairs)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1516706954", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590811959940657*^9, 3.590812291742648*^9}, {
   3.590812344278364*^9, 3.590812350026736*^9}, {3.594638064095589*^9, 
   3.594638103770809*^9}, {3.594638144530171*^9, 3.594638158738885*^9}, {
   3.594638361669126*^9, 3.594638385425623*^9}, {3.606126388215915*^9, 
   3.606126389322834*^9}, {3.606126828325432*^9, 3.606126835575524*^9}, {
   3.615713502845813*^9, 3.615713515014458*^9}, 3.648292336559531*^9},
 CellTags->{
  "ID:1516706954", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "allPairs"},
 CellID->1516706954],

Cell[BoxData[
 RowBox[{
  RowBox[{"update", "[", 
   RowBox[{"k", ",", "j"}], "]"}], ":=", 
  RowBox[{"k", 
   TagBox["\:22c8",
    Identity,
    SyntaxForm->"a*b"], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"j", ",", "j"}], "\[RightAngleBracket]"}], 
     TagBox["\:293b",
      Identity,
      SyntaxForm->"a*b"], 
     RowBox[{"pairs", "[", "j", "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(update)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:329670188", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648294027814323*^9, 3.648294074051206*^9}, {
  3.649404533268659*^9, 3.6494045538060417`*^9}},
 CellTags->{
  "ID:329670188", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "update"},
 CellID->329670188],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1008147046]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->118409702],

Cell["\<\
Note that \[OpenCurlyDoubleQuote]allPairs\[CloseCurlyDoubleQuote], \
\[OpenCurlyDoubleQuote]pairs\[CloseCurlyDoubleQuote] and \
\[OpenCurlyDoubleQuote]update\[CloseCurlyDoubleQuote] only store indices, not \
elements!\
\>", "Text",
 CellChangeTimes->{{3.594637992659927*^9, 3.594638012984477*^9}, {
  3.594638478024145*^9, 3.59463848658783*^9}, {3.6482942555111017`*^9, 
  3.6482942611116056`*^9}},
 CellID->214949191],

Cell["", "OpenEnvironment",
 CellID->657408265],

Cell[CellGroupData[{

Cell["Definition: Domain of Pairs of Natural Numbers", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.64836891250559*^9, 3.648368927846455*^9}},
 CellID->1299046839],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"n", ",", "R"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1283917386}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648368933584914*^9, 3.648368937141016*^9}, 
   3.648372059427598*^9, {3.6483722365020933`*^9, 3.648372236724966*^9}},
 CellTags->{
  "ID:1283917386", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1283917386],

Cell[BoxData[
 RowBox[{
  RowBox[{"NatPairs", "[", 
   RowBox[{"n", ",", "R"}], "]"}], ":=", 
  UnderscriptBox["\[CapitalDelta]", "N"]}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 14652249}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6483689423907423`*^9, 3.6483689531231613`*^9}, 
   3.648372061271434*^9, {3.648372238900926*^9, 3.648372239050088*^9}},
 CellTags->{
  "ID:14652249", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->14652249],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "x"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", "N"], 
      RowBox[{"[", "x", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"isTuple", "[", "x", "]"}]},
         {
          RowBox[{
           RowBox[{"\[LeftBracketingBar]", "x", "\[RightBracketingBar]"}], 
           "\[Equal]", "2"}]},
         {
          RowBox[{
           SubscriptBox["x", "1"], "\[Element]", 
           SubscriptBox["\[DoubleStruckCapitalN]", 
            RowBox[{"1", ",", "\[Ellipsis]", ",", "n"}]]}]},
         {
          RowBox[{
           SubscriptBox["x", "2"], "\[Element]", 
           SubscriptBox["\[DoubleStruckCapitalN]", 
            RowBox[{"1", ",", "\[Ellipsis]", ",", "n"}]]}]},
         {
          RowBox[{"R", "[", 
           RowBox[{
            SubscriptBox["x", "1"], ",", 
            SubscriptBox["x", "2"]}], "]"}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(NatPairs \[Element])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1012271681", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648368966568204*^9, 3.648369032104726*^9}, {
  3.648372065886207*^9, 3.6483720697247677`*^9}, {3.6483722425538054`*^9, 
  3.6483722509279623`*^9}},
 CellTags->{
  "ID:1012271681", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element]"},
 CellID->1012271681],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->750257285]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->672258777]
}, Closed]],

Cell[CellGroupData[{

Cell["Tests", "Subsubsection",
 CellChangeTimes->{{3.593770775733779*^9, 3.593770776513757*^9}},
 CellID->200708094],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"processed", "[", 
  RowBox[{"1", ",", "3", ",", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", "1"}], "\[RightAngleBracket]"}], 
    "\[RightAngleBracket]"}]}], "]"}]], "Computation",
 CellChangeTimes->{{3.594639493017169*^9, 3.594639508388174*^9}, {
   3.594639539106227*^9, 3.594639539436631*^9}, {3.594639598789307*^9, 
   3.594639607583975*^9}, {3.619249627023704*^9, 3.619249630694136*^9}, 
   3.645768708261572*^9},
 CellID->1171053509],

Cell[BoxData[
 FormBox["False", TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.606126659672483*^9, 3.615713258616062*^9, 3.61571466834841*^9, {
   3.615714714228213*^9, 3.615714760495578*^9}, 3.615714844381404*^9, {
   3.619249617389901*^9, 3.619249631213377*^9}, 3.631446620451812*^9, {
   3.631448882131184*^9, 3.6314488852924147`*^9}, {3.631448920782597*^9, 
   3.631449018348303*^9}, 3.6319440132492228`*^9, 3.63194407845774*^9, 
   3.6319441843366127`*^9, 3.631953095964971*^9, 3.631953184693646*^9, 
   3.635060954076249*^9, 3.645768747926551*^9, 3.645776649054196*^9, 
   3.6482925466463833`*^9, 3.661669580484372*^9, 3.6616698253766026`*^9, 
   3.666694295118277*^9, 3.666694330946423*^9},
 CellID->2030739290]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1171053509"]]]]}, {None, None}},
 CellChangeTimes->{3.666694332375251*^9},
 CellID->238027729],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1171053509-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666694332433668*^9},
 CellID->572441673]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"processed", "[", 
  RowBox[{"1", ",", "3", ",", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "2"}], "\[RightAngleBracket]"}], 
    "\[RightAngleBracket]"}]}], "]"}]], "Computation",
 CellChangeTimes->{{3.631446680152231*^9, 3.63144670170191*^9}, 
   3.645768710618661*^9},
 CellID->1147098130],

Cell[BoxData[
 FormBox["True", TheoremaForm]], "ComputationResult",
 CellChangeTimes->{{3.6314466991259613`*^9, 3.6314467025665417`*^9}, {
   3.631448975935293*^9, 3.63144898255611*^9}, 3.6319531900987253`*^9, 
   3.635060958600396*^9, 3.645768750627264*^9, 3.64577665009691*^9, 
   3.648292548501771*^9, 3.6616695811239734`*^9, 3.661669827888207*^9, 
   3.666694336220992*^9},
 CellID->1543197942]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1147098130"]]]]}, {None, None}},
 CellChangeTimes->{3.666694337134251*^9},
 CellID->1024977621],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1147098130-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6666943371937304`*^9},
 CellID->1352090126]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"processed", "[", 
  RowBox[{"1", ",", "3", ",", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"1", ",", "2"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"3", ",", "2"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"2", ",", "4"}], "\[RightAngleBracket]"}]}], 
    "\[RightAngleBracket]"}]}], "]"}]], "Computation",
 CellChangeTimes->{{3.645768831260263*^9, 3.6457688457877197`*^9}},
 CellID->1987206771],

Cell[BoxData[
 FormBox["True", TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645768846598303*^9, 3.645776651149809*^9, 
  3.648292550682995*^9, 3.6616695819819746`*^9, 3.6616698308990126`*^9, 
  3.666694343787703*^9},
 CellID->1879346961]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1987206771"]]]]}, {None, None}},
 CellChangeTimes->{3.666694344752088*^9},
 CellID->578228279],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1987206771-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666694344813312*^9},
 CellID->441030501]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"allPairs", "[", "0", "]"}]], "Computation",
 CellChangeTimes->{{3.645768762756009*^9, 3.6457687659037247`*^9}},
 CellID->618819632],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.6457687716639347`*^9, 3.6457766521824493`*^9, 3.648292444177867*^9, 
   3.661669582605976*^9, 3.661669852770251*^9, {3.6666943525787983`*^9, 
   3.666694373507621*^9}},
 CellID->1355411954]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c618819632"]]]]}, {None, None}},
 CellChangeTimes->{3.666694374417115*^9},
 CellID->1346320737],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c618819632-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666694374487471*^9},
 CellID->1995525275]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"allPairs", "[", "1", "]"}]], "Computation",
 CellChangeTimes->{3.645768776553953*^9},
 CellID->1130747252],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"1", ",", "1"}], "\[RightAngleBracket]"}], 
   "\[RightAngleBracket]"}], TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645768777342539*^9, 3.6457766532128277`*^9, 
  3.648292447106061*^9, 3.6616695834327774`*^9, 3.6616698554534554`*^9, 
  3.666694377272262*^9},
 CellID->1139002741]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1130747252"]]]]}, {None, None}},
 CellChangeTimes->{3.666694378423242*^9},
 CellID->1111910764],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1130747252-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.66669437849358*^9},
 CellID->1023788914]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"allPairs", "[", "2", "]"}]], "Computation",
 CellChangeTimes->{3.645768782298009*^9},
 CellID->1098463539],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "1"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"2", ",", "2"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "2"}], "\[RightAngleBracket]"}]}], 
   "\[RightAngleBracket]"}], TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645768783014254*^9, 3.6457766543188753`*^9, 
  3.648292451420046*^9, 3.661669584056779*^9, 3.6616698583102603`*^9, 
  3.6666943804755583`*^9},
 CellID->539950968]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1098463539"]]]]}, {None, None}},
 CellChangeTimes->{3.666694381414575*^9},
 CellID->1782642311],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1098463539-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6666943814938183`*^9},
 CellID->219375137]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"allPairs", "[", "3", "]"}]], "Computation",
 CellChangeTimes->{{3.593765283987323*^9, 3.593765326826361*^9}, {
   3.593765798213798*^9, 3.593765798843862*^9}, 3.594639037196087*^9},
 CellID->1917426498],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "1"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"2", ",", "2"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", "3"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "2"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "3"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"2", ",", "3"}], "\[RightAngleBracket]"}]}], 
   "\[RightAngleBracket]"}], TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.60612680821001*^9, 3.606126871768988*^9, 
  3.615713493662694*^9, 3.615713543117505*^9, 3.631446772372219*^9, 
  3.6457687581008053`*^9, 3.645776655277975*^9, 3.648292457946517*^9, 
  3.6616695847743797`*^9, 3.661669861367866*^9, 3.666694383323574*^9},
 CellID->58665367]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1917426498"]]]]}, {None, None}},
 CellChangeTimes->{3.666694384312993*^9},
 CellID->1124327149],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1917426498-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666694384389708*^9},
 CellID->1789828913]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"allPairs", "[", "4", "]"}]], "Computation",
 CellChangeTimes->{{3.64829249329635*^9, 3.648292495800482*^9}},
 CellID->191290873],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "1"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"2", ",", "2"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", "3"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"4", ",", "4"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "2"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "3"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"2", ",", "3"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "4"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"2", ",", "4"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", "4"}], "\[RightAngleBracket]"}]}], 
   "\[RightAngleBracket]"}], TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.6482924961910973`*^9, 3.661669585445181*^9, 
  3.6616698657670736`*^9, 3.6666943864107122`*^9},
 CellID->935380317]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c191290873"]]]]}, {None, None}},
 CellChangeTimes->{3.666694387330267*^9},
 CellID->1954139160],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c191290873-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666694387411824*^9},
 CellID->1250936321]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"pairs", "[", "10", "]"}]], "Computation",
 CellChangeTimes->{{3.648292463817152*^9, 3.648292466869205*^9}},
 CellID->125498398],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "10"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"2", ",", "10"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", "10"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"4", ",", "10"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"5", ",", "10"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"6", ",", "10"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"7", ",", "10"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"8", ",", "10"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"9", ",", "10"}], "\[RightAngleBracket]"}]}], 
   "\[RightAngleBracket]"}], TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.648292467809958*^9, 3.6616695860847816`*^9, 
  3.6616698759850917`*^9, 3.666694389074728*^9},
 CellID->412770255]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c125498398"]]]]}, {None, None}},
 CellChangeTimes->{3.6666943899719057`*^9},
 CellID->1021847407],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c125498398-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666694390048152*^9},
 CellID->34256619]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"update", "[", 
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", "4"}], "\[RightAngleBracket]"}], 
    "\[RightAngleBracket]"}], ",", "5"}], "]"}]], "Computation",
 CellChangeTimes->{{3.648294166475052*^9, 3.648294172712922*^9}, {
  3.6482942302030277`*^9, 3.648294233983843*^9}},
 CellID->1562111062],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", "4"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"5", ",", "5"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", "5"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"2", ",", "5"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", "5"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"4", ",", "5"}], "\[RightAngleBracket]"}]}], 
   "\[RightAngleBracket]"}], TheoremaForm]], "ComputationResult",
 CellChangeTimes->{{3.6482942187218723`*^9, 3.648294234927169*^9}, 
   3.661669586708783*^9, 3.6616698877163124`*^9, {3.666694408850893*^9, 
   3.666694416012982*^9}},
 CellID->350255224]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1562111062"]]]]}, {None, None}},
 CellChangeTimes->{3.666694416996004*^9},
 CellID->965480556],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1562111062-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666694417070943*^9},
 CellID->230386765]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Lemmas (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.648369073780802*^9, 3.648369074641408*^9}, {
  3.649414300919784*^9, 3.649414304498095*^9}},
 CellID->2027758943],

Cell["", "OpenEnvironment",
 CellID->1347355957],

Cell[CellGroupData[{

Cell["Lemma: Simple Properties of NatPairs", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648369084026124*^9, 3.6483690968159857`*^9}, 
   3.648369249672823*^9, {3.6483707179070168`*^9, 3.6483707213687267`*^9}},
 CellID->1285769589],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"n", ",", "i", ",", "j", ",", "R"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 988862829}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648369279873826*^9, 3.648369293025572*^9}, {
  3.648370917636973*^9, 3.6483709179615993`*^9}},
 CellTags->{
  "ID:988862829", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->988862829],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"NatPairs", "[", 
      RowBox[{"n", ",", "R"}], "]"}]], 
    RowBox[{"[", "i", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"ProdDomain", "[", 
      RowBox[{
      "\[DoubleStruckCapitalZ]Dom", ",", "\[DoubleStruckCapitalZ]Dom"}], 
      "]"}]], 
    RowBox[{"[", "i", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(NatPairs \[Element] ProdDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:16377100", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6616612429458685`*^9, 3.6616612946603594`*^9}},
 CellTags->{
  "ID:16377100", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element] ProdDomain"},
 CellID->16377100],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{"n", ",", "R"}], "]"}], "]"}]], 
    RowBox[{"[", "i", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"ProdDomain", "[", 
       RowBox[{
       "\[DoubleStruckCapitalZ]Dom", ",", "\[DoubleStruckCapitalZ]Dom"}], 
       "]"}], "]"}]], 
    RowBox[{"[", "i", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(NatPairs \[Element] ProdDomain DomainTuples)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:136514135", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661661304379177*^9, 3.661661320573005*^9}},
 CellTags->{
  "ID:136514135", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element] ProdDomain DomainTuples"},
 CellID->136514135],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"NatPairs", "[", 
      RowBox[{"n", ",", "R"}], "]"}]], 
    RowBox[{"[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"i", ",", "j"}], "\[RightAngleBracket]"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"i", "\[Element]", 
         SubscriptBox["\[DoubleStruckCapitalN]", 
          RowBox[{"1", ",", "\[Ellipsis]", ",", "n"}]]}]},
       {
        RowBox[{"j", "\[Element]", 
         SubscriptBox["\[DoubleStruckCapitalN]", 
          RowBox[{"1", ",", "\[Ellipsis]", ",", "n"}]]}]},
       {
        RowBox[{"R", "[", 
         RowBox[{"i", ",", "j"}], "]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(NatPairs \[Element] alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1872634730", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648369128609074*^9, 3.6483691919316607`*^9}, {
  3.6483692428823423`*^9, 3.648369244475065*^9}, {3.6483692860951157`*^9, 
  3.648369296074065*^9}, {3.648525894053009*^9, 3.648525899050996*^9}},
 CellTags->{
  "ID:1872634730", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element] alternative", "IR"},
 CellID->1872634730],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"NatPairs", "[", 
      RowBox[{"n", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["<",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}], "]"}]], 
    RowBox[{"[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"i", ",", "j"}], "\[RightAngleBracket]"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"i", "\[Element]", 
         SubscriptBox["\[DoubleStruckCapitalN]", 
          RowBox[{"1", ",", "\[Ellipsis]", ",", "n"}]]}]},
       {
        RowBox[{"j", "\[Element]", 
         SubscriptBox["\[DoubleStruckCapitalN]", 
          RowBox[{"1", ",", "\[Ellipsis]", ",", "n"}]]}]},
       {
        RowBox[{"j", ">", "i"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(NatPairs< \[Element] alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1931599387", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6483693035578327`*^9, 3.648369319790317*^9}},
 CellTags->{
  "ID:1931599387", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs< \[Element] alternative", "IR"},
 CellID->1931599387],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"NatPairs", "[", 
      RowBox[{"n", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[LessEqual]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}], "]"}]], 
    RowBox[{"[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"i", ",", "j"}], "\[RightAngleBracket]"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"i", "\[Element]", 
         SubscriptBox["\[DoubleStruckCapitalN]", 
          RowBox[{"1", ",", "\[Ellipsis]", ",", "n"}]]}]},
       {
        RowBox[{"j", "\[Element]", 
         SubscriptBox["\[DoubleStruckCapitalN]", 
          RowBox[{"1", ",", "\[Ellipsis]", ",", "n"}]]}]},
       {
        RowBox[{"j", "\[GreaterEqual]", "i"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(NatPairs\[LessEqual] \[Element] alternative)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2124565744", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6483693292246523`*^9, 3.648369335042576*^9}},
 CellTags->{
  "ID:2124565744", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs\[LessEqual] \[Element] alternative", "IR"},
 CellID->2124565744],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"NatPairs", "[", 
      RowBox[{"n", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["<",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}], "]"}]], 
    RowBox[{"[", "x", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{"k", ",", 
     RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"l", ">", "k"}], "\[And]", 
     RowBox[{"x", "\[Equal]", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"k", ",", "l"}], "\[RightAngleBracket]"}]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(NatPairs< \[Element] alternative II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1917398594", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649135512572839*^9, 3.649135517253199*^9}},
 CellTags->{
  "ID:1917398594", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs< \[Element] alternative II"},
 CellID->1917398594],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"NatPairs", "[", 
      RowBox[{"n", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[LessEqual]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}], "]"}]], 
    RowBox[{"[", "x", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{"k", ",", 
     RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"l", "\[GreaterEqual]", "k"}], "\[And]", 
     RowBox[{"x", "\[Equal]", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"k", ",", "l"}], "\[RightAngleBracket]"}]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(NatPairs\[LessEqual] \[Element] alternative II)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:272474466", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648525916196514*^9, 3.648526000359509*^9}, {
  3.648526816816532*^9, 3.648526817629773*^9}},
 CellTags->{
  "ID:272474466", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs\[LessEqual] \[Element] alternative II"},
 CellID->272474466],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"NatPairs", "[", 
      RowBox[{"n", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["<",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}], "]"}]], 
    RowBox[{"[", "x", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"NatPairs", "[", 
     RowBox[{"n", ",", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       TagBox["\[LessEqual]",
        Identity], 
       TagBox[")",
        "AutoParentheses"]}]}], "]"}]], 
   RowBox[{"[", "x", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(NatPairs< \[LessEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2005363738", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649409214241397*^9, 3.649409225477697*^9}},
 CellTags->{
  "ID:2005363738", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs< \[LessEqual]"},
 CellID->2005363738],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{"n", ",", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         TagBox["<",
          Identity], 
         TagBox[")",
          "AutoParentheses"]}]}], "]"}], "]"}]], 
    RowBox[{"[", "x", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", 
     RowBox[{"NatPairs", "[", 
      RowBox[{"n", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[LessEqual]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}], "]"}], "]"}]], 
   RowBox[{"[", "x", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DomainTuples NatPairs< \[LessEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1661715038", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649409750922407*^9, 3.64940976851696*^9}},
 CellTags->{
  "ID:1661715038", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NatPairs< \[LessEqual]"},
 CellID->1661715038],

Cell[BoxData[
 RowBox[{
  RowBox[{"n", "\[Element]", 
   SubscriptBox["\[DoubleStruckCapitalN]", "0"]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1020080047}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649135771170228*^9, 3.649135776110366*^9}},
 CellTags->{
  "ID:1020080047", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1020080047],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"NatPairs", "[", 
      RowBox[{"n", ",", "R"}], "]"}]], 
    RowBox[{"[", "i", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"NatPairs", "[", 
      RowBox[{
       RowBox[{"n", "+", "1"}], ",", "R"}], "]"}]], 
    RowBox[{"[", "i", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(NatPairs +1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1327808237", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6483707594961987`*^9, 3.648370804571213*^9}, {
  3.648370863314588*^9, 3.648370863686028*^9}},
 CellTags->{
  "ID:1327808237", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs +1"},
 CellID->1327808237],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{"n", ",", "R"}], "]"}], "]"}]], 
    RowBox[{"[", "i", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{
        RowBox[{"n", "+", "1"}], ",", "R"}], "]"}], "]"}]], 
    RowBox[{"[", "i", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DomainTuples NatPairs +1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:798140928", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648370815334556*^9, 3.648370832972797*^9}, {
  3.648370866490199*^9, 3.648370866897826*^9}},
 CellTags->{
  "ID:798140928", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NatPairs +1"},
 CellID->798140928],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{"m", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]], 
  RowBox[{"n", "\[GreaterEqual]", "m"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 2111464127}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649404221987938*^9, 3.649404233613016*^9}},
 CellTags->{
  "ID:2111464127", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->2111464127],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"NatPairs", "[", 
      RowBox[{"m", ",", "R"}], "]"}]], 
    RowBox[{"[", "i", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"NatPairs", "[", 
      RowBox[{"n", ",", "R"}], "]"}]], 
    RowBox[{"[", "i", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(NatPairs \[GreaterEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1834895057", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6494042440151577`*^9, 3.64940425371987*^9}},
 CellTags->{
  "ID:1834895057", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[GreaterEqual]"},
 CellID->1834895057],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{"m", ",", "R"}], "]"}], "]"}]], 
    RowBox[{"[", "i", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{"n", ",", "R"}], "]"}], "]"}]], 
    RowBox[{"[", "i", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DomainTuples NatPairs \[GreaterEqual])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1703465728", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649404264038302*^9, 3.6494042692755537`*^9}},
 CellTags->{
  "ID:1703465728", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NatPairs \[GreaterEqual]"},
 CellID->1703465728],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2009530504]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->954284704],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(NatPairs \[Element] ProdDomain)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(NatPairs \[Element] ProdDomain)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["n", "ExpressionVariable"], ",", 
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["R", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"NatPairs", "[", 
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], ",", 
                  StyleBox["R", "ExpressionVariable"]}], "]"}]], "[", 
             StyleBox["i", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"ProdDomain", "[", 
                
                RowBox[{
                 "\[DoubleStruckCapitalZ]Dom", ",", 
                  "\[DoubleStruckCapitalZ]Dom"}], "]"}]], "[", 
             StyleBox["i", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["n", "ExpressionVariable"], ",", 
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["R", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"NatPairs", "[", 
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], ",", 
                  StyleBox["R", "ExpressionVariable"]}], "]"}]], "[", 
             StyleBox["i", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"ProdDomain", "[", 
                
                RowBox[{
                 "\[DoubleStruckCapitalZ]Dom", ",", 
                  "\[DoubleStruckCapitalZ]Dom"}], "]"}]], "[", 
             StyleBox["i", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:16377100"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p16377100-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p16377100-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:16377100", "Proof|ID:16377100-1"},
 CellID->22673357],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p16377100-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->196582360]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(NatPairs \[Element] ProdDomain DomainTuples)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(NatPairs \[Element] ProdDomain DomainTuples)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["n", "ExpressionVariable"], ",", 
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["R", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
                RowBox[{"NatPairs", "[", 
                  RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
             StyleBox["i", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
                RowBox[{"ProdDomain", "[", 
                  
                  RowBox[{
                   "\[DoubleStruckCapitalZ]Dom", ",", 
                    "\[DoubleStruckCapitalZ]Dom"}], "]"}], "]"}]], "[", 
             StyleBox["i", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["n", "ExpressionVariable"], ",", 
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["R", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
                RowBox[{"NatPairs", "[", 
                  RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
             StyleBox["i", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
                RowBox[{"ProdDomain", "[", 
                  
                  RowBox[{
                   "\[DoubleStruckCapitalZ]Dom", ",", 
                    "\[DoubleStruckCapitalZ]Dom"}], "]"}], "]"}]], "[", 
             StyleBox["i", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:136514135"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p136514135-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p136514135-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:136514135", "Proof|ID:136514135-1"},
 CellID->585962876],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p136514135-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->33667364]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(NatPairs \[Element] alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(NatPairs \[Element] alternative)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["n", "ExpressionVariable"], ",", 
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["j", "ExpressionVariable"], ",", 
            StyleBox["R", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"NatPairs", "[", 
                  RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "\[Element]", 
                 SubscriptBox["\[DoubleStruckCapitalN]", 
                  RowBox[{"1", ",", "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]]}], "\[And]", 
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "\[Element]", 
                 SubscriptBox["\[DoubleStruckCapitalN]", 
                  RowBox[{"1", ",", "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]]}], "\[And]", 
               RowBox[{
                 StyleBox["R", "ExpressionVariable"], "[", 
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["n", "ExpressionVariable"], ",", 
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["j", "ExpressionVariable"], ",", 
            StyleBox["R", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"NatPairs", "[", 
                  RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "\[Element]", 
                 SubscriptBox["\[DoubleStruckCapitalN]", 
                  RowBox[{"1", ",", "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]]}], "\[And]", 
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "\[Element]", 
                 SubscriptBox["\[DoubleStruckCapitalN]", 
                  RowBox[{"1", ",", "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]]}], "\[And]", 
               RowBox[{
                 StyleBox["R", "ExpressionVariable"], "[", 
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1872634730"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1872634730-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1872634730-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1872634730", "Proof|ID:1872634730-1"},
 CellID->1738547973],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1872634730-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1666042888]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(NatPairs< \[Element] alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(NatPairs< \[Element] alternative)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["n", "ExpressionVariable"], ",", 
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["j", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"NatPairs", "[", 
                  RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", "<"}], "]"}]], 
               "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "\[Element]", 
                 SubscriptBox["\[DoubleStruckCapitalN]", 
                  RowBox[{"1", ",", "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]]}], "\[And]", 
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "\[Element]", 
                 SubscriptBox["\[DoubleStruckCapitalN]", 
                  RowBox[{"1", ",", "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]]}], "\[And]", 
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], ">", 
                 StyleBox["i", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["n", "ExpressionVariable"], ",", 
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["j", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"NatPairs", "[", 
                  RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", "<"}], "]"}]], 
               "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "\[Element]", 
                 SubscriptBox["\[DoubleStruckCapitalN]", 
                  RowBox[{"1", ",", "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]]}], "\[And]", 
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "\[Element]", 
                 SubscriptBox["\[DoubleStruckCapitalN]", 
                  RowBox[{"1", ",", "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]]}], "\[And]", 
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], ">", 
                 StyleBox["i", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1931599387"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1931599387-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1931599387-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1931599387", "Proof|ID:1931599387-1"},
 CellID->60767287],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1931599387-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2095809430]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(NatPairs\[LessEqual] \[Element] alternative)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(NatPairs\[LessEqual] \[Element] alternative)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["n", "ExpressionVariable"], ",", 
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["j", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"NatPairs", "[", 
                  RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    "\[LessEqual]"}], "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "\[Element]", 
                 SubscriptBox["\[DoubleStruckCapitalN]", 
                  RowBox[{"1", ",", "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]]}], "\[And]", 
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "\[Element]", 
                 SubscriptBox["\[DoubleStruckCapitalN]", 
                  RowBox[{"1", ",", "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]]}], "\[And]", 
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                 StyleBox["i", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["n", "ExpressionVariable"], ",", 
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["j", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"NatPairs", "[", 
                  RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    "\[LessEqual]"}], "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "\[Element]", 
                 SubscriptBox["\[DoubleStruckCapitalN]", 
                  RowBox[{"1", ",", "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]]}], "\[And]", 
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "\[Element]", 
                 SubscriptBox["\[DoubleStruckCapitalN]", 
                  RowBox[{"1", ",", "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]]}], "\[And]", 
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                 StyleBox["i", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2124565744"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2124565744-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2124565744-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2124565744", "Proof|ID:2124565744-1"},
 CellID->2051885642],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2124565744-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1918391475]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(NatPairs< \[Element] alternative II)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(NatPairs< \[Element] alternative II)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["n", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"NatPairs", "[", 
                  RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", "<"}], "]"}]], 
               "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 RowBox[{
                   StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                 "\[Ellipsis]", ",", 
                 StyleBox["n", "ExpressionVariable"]}]}, {
               RowBox[{
                 RowBox[{
                   StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
                 "\[Ellipsis]", ",", 
                 StyleBox["n", "ExpressionVariable"]}]}}]], 
           RowBox[{
             RowBox[{
               StyleBox["l", "ExpressionVariable"], ">", 
               StyleBox["k", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["k", "ExpressionVariable"], ",", 
                   StyleBox["l", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["n", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"NatPairs", "[", 
                  RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", "<"}], "]"}]], 
               "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 RowBox[{
                   StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                 "\[Ellipsis]", ",", 
                 StyleBox["n", "ExpressionVariable"]}]}, {
               RowBox[{
                 RowBox[{
                   StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
                 "\[Ellipsis]", ",", 
                 StyleBox["n", "ExpressionVariable"]}]}}]], 
           RowBox[{
             RowBox[{
               StyleBox["l", "ExpressionVariable"], ">", 
               StyleBox["k", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["k", "ExpressionVariable"], ",", 
                   StyleBox["l", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1917398594"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1917398594-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1917398594-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1917398594", "Proof|ID:1917398594-1"},
 CellID->1425710701],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1917398594-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->98700243]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(NatPairs\[LessEqual] \[Element] alternative II)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(NatPairs\[LessEqual] \[Element] alternative II)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["n", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"NatPairs", "[", 
                  RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    "\[LessEqual]"}], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 RowBox[{
                   StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                 "\[Ellipsis]", ",", 
                 StyleBox["n", "ExpressionVariable"]}]}, {
               RowBox[{
                 RowBox[{
                   StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
                 "\[Ellipsis]", ",", 
                 StyleBox["n", "ExpressionVariable"]}]}}]], 
           RowBox[{
             RowBox[{
               StyleBox["l", "ExpressionVariable"], "\[GreaterEqual]", 
               StyleBox["k", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["k", "ExpressionVariable"], ",", 
                   StyleBox["l", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["n", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"NatPairs", "[", 
                  RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    "\[LessEqual]"}], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 RowBox[{
                   StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                 "\[Ellipsis]", ",", 
                 StyleBox["n", "ExpressionVariable"]}]}, {
               RowBox[{
                 RowBox[{
                   StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
                 "\[Ellipsis]", ",", 
                 StyleBox["n", "ExpressionVariable"]}]}}]], 
           RowBox[{
             RowBox[{
               StyleBox["l", "ExpressionVariable"], "\[GreaterEqual]", 
               StyleBox["k", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["k", "ExpressionVariable"], ",", 
                   StyleBox["l", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:272474466"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p272474466-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p272474466-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:272474466", "Proof|ID:272474466-1"},
 CellID->136931469],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p272474466-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1186718445]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(NatPairs< \[LessEqual])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(NatPairs< \[LessEqual])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["n", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"NatPairs", "[", 
                  RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", "<"}], "]"}]], 
               "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"NatPairs", "[", 
              RowBox[{
                StyleBox["n", "ExpressionVariable"], ",", "\[LessEqual]"}], 
              "]"}]], "[", 
           StyleBox["x", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["n", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"NatPairs", "[", 
                  RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", "<"}], "]"}]], 
               "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"NatPairs", "[", 
              RowBox[{
                StyleBox["n", "ExpressionVariable"], ",", "\[LessEqual]"}], 
              "]"}]], "[", 
           StyleBox["x", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2005363738"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2005363738-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2005363738-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2005363738", "Proof|ID:2005363738-1"},
 CellID->911690168],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2005363738-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->109415471]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DomainTuples NatPairs< \[LessEqual])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DomainTuples NatPairs< \[LessEqual])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["n", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", "<"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", 
              RowBox[{"NatPairs", "[", 
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], ",", "\[LessEqual]"}], 
                "]"}], "]"}]], "[", 
           StyleBox["x", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["n", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", "<"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", 
              RowBox[{"NatPairs", "[", 
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], ",", "\[LessEqual]"}], 
                "]"}], "]"}]], "[", 
           StyleBox["x", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1661715038"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1661715038-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1661715038-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1661715038", "Proof|ID:1661715038-1"},
 CellID->1036844371],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1661715038-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->123023466]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(NatPairs +1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(NatPairs +1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["n", "ExpressionVariable"], ",", 
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["R", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             StyleBox["n", "ExpressionVariable"], "\[Element]", 
             SubscriptBox["\[DoubleStruckCapitalN]", "0"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"NatPairs", "[", 
                  RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}]], "[", 
               StyleBox["i", "ExpressionVariable"], "]"}], "\[Implies]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"NatPairs", "[", 
                  RowBox[{
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "+", "1"}], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}]], "[", 
               StyleBox["i", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["n", "ExpressionVariable"], ",", 
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["R", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             StyleBox["n", "ExpressionVariable"], "\[Element]", 
             SubscriptBox["\[DoubleStruckCapitalN]", "0"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"NatPairs", "[", 
                  RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}]], "[", 
               StyleBox["i", "ExpressionVariable"], "]"}], "\[Implies]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"NatPairs", "[", 
                  RowBox[{
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "+", "1"}], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}]], "[", 
               StyleBox["i", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1327808237"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1327808237-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1327808237-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1327808237", "Proof|ID:1327808237-1"},
 CellID->2053606910],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1327808237-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->184747131]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DomainTuples NatPairs +1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DomainTuples NatPairs +1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["n", "ExpressionVariable"], ",", 
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["R", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             StyleBox["n", "ExpressionVariable"], "\[Element]", 
             SubscriptBox["\[DoubleStruckCapitalN]", "0"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
               StyleBox["i", "ExpressionVariable"], "]"}], "\[Implies]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "+", "1"}], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
               StyleBox["i", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["n", "ExpressionVariable"], ",", 
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["R", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             StyleBox["n", "ExpressionVariable"], "\[Element]", 
             SubscriptBox["\[DoubleStruckCapitalN]", "0"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
               StyleBox["i", "ExpressionVariable"], "]"}], "\[Implies]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "+", "1"}], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
               StyleBox["i", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:798140928"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p798140928-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p798140928-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:798140928", "Proof|ID:798140928-1"},
 CellID->974929944],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p798140928-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->963692269]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(NatPairs \[GreaterEqual])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(NatPairs \[GreaterEqual])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["n", "ExpressionVariable"], ",", 
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["R", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             StyleBox["n", "ExpressionVariable"], "\[Element]", 
             SubscriptBox["\[DoubleStruckCapitalN]", "0"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["m", "ExpressionVariable"], "\[Element]", 
                    SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
              RowBox[{
                StyleBox["n", "ExpressionVariable"], "\[GreaterEqual]", 
                StyleBox["m", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}]], "[", 
                 StyleBox["i", "ExpressionVariable"], "]"}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}]], "[", 
                 StyleBox["i", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["n", "ExpressionVariable"], ",", 
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["R", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             StyleBox["n", "ExpressionVariable"], "\[Element]", 
             SubscriptBox["\[DoubleStruckCapitalN]", "0"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["m", "ExpressionVariable"], "\[Element]", 
                    SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
              RowBox[{
                StyleBox["n", "ExpressionVariable"], "\[GreaterEqual]", 
                StyleBox["m", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}]], "[", 
                 StyleBox["i", "ExpressionVariable"], "]"}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}]], "[", 
                 StyleBox["i", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1834895057"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1834895057-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1834895057-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1834895057", "Proof|ID:1834895057-1"},
 CellID->1231124039],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1834895057-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2145280537]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DomainTuples NatPairs \[GreaterEqual])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DomainTuples NatPairs \[GreaterEqual])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["n", "ExpressionVariable"], ",", 
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["R", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             StyleBox["n", "ExpressionVariable"], "\[Element]", 
             SubscriptBox["\[DoubleStruckCapitalN]", "0"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["m", "ExpressionVariable"], "\[Element]", 
                    SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
              RowBox[{
                StyleBox["n", "ExpressionVariable"], "\[GreaterEqual]", 
                StyleBox["m", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
                 StyleBox["i", "ExpressionVariable"], "]"}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
                 StyleBox["i", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["n", "ExpressionVariable"], ",", 
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["R", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             StyleBox["n", "ExpressionVariable"], "\[Element]", 
             SubscriptBox["\[DoubleStruckCapitalN]", "0"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["m", "ExpressionVariable"], "\[Element]", 
                    SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
              RowBox[{
                StyleBox["n", "ExpressionVariable"], "\[GreaterEqual]", 
                StyleBox["m", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
                 StyleBox["i", "ExpressionVariable"], "]"}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
                 StyleBox["i", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1703465728"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1703465728-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1703465728-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1703465728", "Proof|ID:1703465728-1"},
 CellID->1999261425],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1703465728-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1185705224]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->2014150930],

Cell[CellGroupData[{

Cell["Theorem: Properties of Pairs and allPairs", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6483694211769753`*^9, 3.64836944656343*^9}, {
  3.648369945479884*^9, 3.64836994791982*^9}},
 CellID->643050674],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"j", ",", 
   RowBox[{"k", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 971135058}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648369470453247*^9, 3.648369474656508*^9}, {
   3.6483695364594183`*^9, 3.6483695406408033`*^9}, 3.6483696823331013`*^9, {
   3.648369761105887*^9, 3.648369761364896*^9}, 3.648369793910568*^9, {
   3.6483698469599867`*^9, 3.6483698473681183`*^9}},
 CellTags->{
  "ID:971135058", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->971135058],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   RowBox[{"DomainTuples", "[", 
    RowBox[{"NatPairs", "[", 
     RowBox[{"k", ",", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       TagBox["<",
        Identity], 
       TagBox[")",
        "AutoParentheses"]}]}], "]"}], "]"}]], 
  RowBox[{"[", 
   RowBox[{"pairs", "[", "k", "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership pairs unary)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1202991311", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648369561993596*^9, 3.648369589663659*^9}},
 CellTags->{
  "ID:1202991311", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership pairs unary", "IR"},
 CellID->1202991311],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"a", "=", "1"}], ",", "\[Ellipsis]", ",", "k"}]], 
   RowBox[{"k", ">", "a"}]], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", 
      RowBox[{"pairs", "[", "k", "]"}], "\[RightBracketingBar]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     SubscriptBox[
      RowBox[{"pairs", "[", "k", "]"}], "i"], "\[Equal]", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"a", ",", "k"}], "\[RightAngleBracket]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(pairs unary content')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1932739315", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648369631028431*^9, 3.648369713304759*^9}, {
  3.648369851537121*^9, 3.6483698551911373`*^9}, {3.64939085903158*^9, 
  3.649390861954669*^9}},
 CellTags->{
  "ID:1932739315", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs unary content'"},
 CellID->1932739315],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"a", "=", "1"}], ",", "\[Ellipsis]", ",", "k"}]], 
   RowBox[{"k", ">", "a"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"a", ",", "k"}], "\[RightAngleBracket]"}], 
    TagBox["\:22ff",
     Identity,
     SyntaxForm->"a+b"], 
    RowBox[{"pairs", "[", "k", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(pairs unary content)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1628109432", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649405145426482*^9, 3.649405158618739*^9}, {
  3.649405563606674*^9, 3.6494055640094957`*^9}},
 CellTags->{
  "ID:1628109432", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs unary content"},
 CellID->1628109432],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   RowBox[{"DomainTuples", "[", 
    RowBox[{"NatPairs", "[", 
     RowBox[{"k", ",", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       TagBox["<",
        Identity], 
       TagBox[")",
        "AutoParentheses"]}]}], "]"}], "]"}]], 
  RowBox[{"[", 
   RowBox[{"pairs", "[", 
    RowBox[{"j", ",", "k"}], "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership pairs binary)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:540671405", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648369809843079*^9, 3.648369812429143*^9}, {
  3.648369858324048*^9, 3.648369858733101*^9}},
 CellTags->{
  "ID:540671405", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership pairs binary", "IR"},
 CellID->540671405],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{"a", ",", 
     RowBox[{"b", "=", "1"}], ",", "\[Ellipsis]", ",", "k"}]], 
   RowBox[{
    RowBox[{"b", ">", "a"}], "\[And]", 
    RowBox[{"b", "\[GreaterEqual]", "j"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"a", ",", "b"}], "\[RightAngleBracket]"}], 
    TagBox["\:22ff",
     Identity,
     SyntaxForm->"a+b"], 
    RowBox[{"pairs", "[", 
     RowBox[{"j", ",", "k"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(pairs binary content')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:881225230", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648369829131374*^9, 3.648369882186467*^9}, {
  3.649390873378325*^9, 3.649390877001148*^9}, {3.649405017197448*^9, 
  3.649405057331389*^9}, {3.6494055662945337`*^9, 3.649405566514873*^9}},
 CellTags->{
  "ID:881225230", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs binary content'"},
 CellID->881225230],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{"a", ",", 
     RowBox[{"b", "=", "1"}], ",", "\[Ellipsis]", ",", "k"}]], 
   RowBox[{"b", ">", "a"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"a", ",", "b"}], "\[RightAngleBracket]"}], 
    TagBox["\:22ff",
     Identity,
     SyntaxForm->"a+b"], 
    RowBox[{"pairs", "[", 
     RowBox[{"2", ",", "k"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(pairs binary content)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:717573619", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64940739612747*^9, 3.649407406220566*^9}},
 CellTags->{
  "ID:717573619", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs binary content"},
 CellID->717573619],

Cell[BoxData[
 RowBox[{
  RowBox[{"allPairs", "[", "0", "]"}], "\[Equal]", 
  RowBox[{
  "\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(allPairs 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:794667329", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6485388844790697`*^9, 3.648538891043173*^9}},
 CellTags->{
  "ID:794667329", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "allPairs 0", "IR"},
 CellID->794667329],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   RowBox[{"DomainTuples", "[", 
    RowBox[{"NatPairs", "[", 
     RowBox[{"k", ",", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       TagBox["\[LessEqual]",
        Identity], 
       TagBox[")",
        "AutoParentheses"]}]}], "]"}], "]"}]], 
  RowBox[{"[", 
   RowBox[{"allPairs", "[", "k", "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership allPairs)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1385740688", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6483699027923393`*^9, 3.648369908396092*^9}},
 CellTags->{
  "ID:1385740688", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership allPairs", "IR"},
 CellID->1385740688],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{"a", ",", 
     RowBox[{"b", "=", "1"}], ",", "\[Ellipsis]", ",", "k"}]], 
   RowBox[{"b", "\[GreaterEqual]", "a"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"a", ",", "b"}], "\[RightAngleBracket]"}], 
    TagBox["\:22ff",
     Identity,
     SyntaxForm->"a+b"], 
    RowBox[{"allPairs", "[", "k", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(allPairs content)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1430860208", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649390217198257*^9, 3.6493902631885*^9}},
 CellTags->{
  "ID:1430860208", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "allPairs content"},
 CellID->1430860208],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->968800477]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->51649926],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership pairs unary)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership pairs unary)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", 
              RowBox[{"NatPairs", "[", 
                RowBox[{
                  StyleBox["k", "ExpressionVariable"], ",", "<"}], "]"}], 
              "]"}]], "[", 
           RowBox[{"pairs", "[", 
             StyleBox["k", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", 
              RowBox[{"NatPairs", "[", 
                RowBox[{
                  StyleBox["k", "ExpressionVariable"], ",", "<"}], "]"}], 
              "]"}]], "[", 
           RowBox[{"pairs", "[", 
             StyleBox["k", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1202991311"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1202991311-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1202991311-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1202991311", "Proof|ID:1202991311-1"},
 CellID->1926557973],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1202991311-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1260065281]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(pairs unary content')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(pairs unary content')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["k", "ExpressionVariable"], ">", 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{"pairs", "[", 
                    StyleBox["k", "ExpressionVariable"], "]"}], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               SubscriptBox[
                RowBox[{"pairs", "[", 
                  StyleBox["k", "ExpressionVariable"], "]"}], 
                StyleBox["i", "ExpressionVariable"]], "\[Equal]", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["k", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["k", "ExpressionVariable"], ">", 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{"pairs", "[", 
                    StyleBox["k", "ExpressionVariable"], "]"}], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               SubscriptBox[
                RowBox[{"pairs", "[", 
                  StyleBox["k", "ExpressionVariable"], "]"}], 
                StyleBox["i", "ExpressionVariable"]], "\[Equal]", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["k", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1932739315"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1932739315-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1932739315-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1932739315", "Proof|ID:1932739315-1"},
 CellID->1207788650],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1932739315-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->758612481]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(pairs unary content)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(pairs unary content)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["k", "ExpressionVariable"], ">", 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["k", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "\:22ff", 
             RowBox[{"pairs", "[", 
               StyleBox["k", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["k", "ExpressionVariable"], ">", 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["k", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "\:22ff", 
             RowBox[{"pairs", "[", 
               StyleBox["k", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1628109432"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1628109432-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1628109432-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1628109432", "Proof|ID:1628109432-1"},
 CellID->135860045],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1628109432-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1770197684]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership pairs binary)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership pairs binary)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["j", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", 
              RowBox[{"NatPairs", "[", 
                RowBox[{
                  StyleBox["k", "ExpressionVariable"], ",", "<"}], "]"}], 
              "]"}]], "[", 
           RowBox[{"pairs", "[", 
             RowBox[{
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["j", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", 
              RowBox[{"NatPairs", "[", 
                RowBox[{
                  StyleBox["k", "ExpressionVariable"], ",", "<"}], "]"}], 
              "]"}]], "[", 
           RowBox[{"pairs", "[", 
             RowBox[{
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:540671405"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p540671405-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p540671405-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:540671405", "Proof|ID:540671405-1"},
 CellID->1030769530],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p540671405-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{3.649406042942514*^9},
 CellID->1057299996]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(pairs binary content')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(pairs binary range')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["j", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ">", 
               StyleBox["a", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], "\[GreaterEqual]", 
               StyleBox["j", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["b", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "\:22ff", 
             RowBox[{"pairs", "[", 
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], ",", 
                 StyleBox["k", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["j", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ">", 
               StyleBox["a", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], "\[GreaterEqual]", 
               StyleBox["j", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["b", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "\:22ff", 
             RowBox[{"pairs", "[", 
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], ",", 
                 StyleBox["k", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:881225230"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p881225230-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p881225230-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:881225230", "Proof|ID:881225230-1"},
 CellID->960050214],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p881225230-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->347641866]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(pairs binary content)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(pairs binary content)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["b", "ExpressionVariable"], ">", 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["b", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "\:22ff", 
             RowBox[{"pairs", "[", 
               RowBox[{"2", ",", 
                 StyleBox["k", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["b", "ExpressionVariable"], ">", 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["b", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "\:22ff", 
             RowBox[{"pairs", "[", 
               RowBox[{"2", ",", 
                 StyleBox["k", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:717573619"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p717573619-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p717573619-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:717573619", "Proof|ID:717573619-1"},
 CellID->1036460169],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p717573619-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1623918929]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(allPairs 0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(allPairs 0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{"allPairs", "[", "0", "]"}], "\[Equal]", 
         RowBox[{
          "\[LeftAngleBracket]", 
           "\[RightAngleBracket]"}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{"allPairs", "[", "0", "]"}], "\[Equal]", 
         RowBox[{
          "\[LeftAngleBracket]", 
           "\[RightAngleBracket]"}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:794667329"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p794667329-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p794667329-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:794667329", "Proof|ID:794667329-1"},
 CellID->1738668792],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p794667329-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2133578009]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership allPairs)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership allPairs)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", 
              RowBox[{"NatPairs", "[", 
                RowBox[{
                  StyleBox["k", "ExpressionVariable"], ",", "\[LessEqual]"}], 
                "]"}], "]"}]], "[", 
           RowBox[{"allPairs", "[", 
             StyleBox["k", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", 
              RowBox[{"NatPairs", "[", 
                RowBox[{
                  StyleBox["k", "ExpressionVariable"], ",", "\[LessEqual]"}], 
                "]"}], "]"}]], "[", 
           RowBox[{"allPairs", "[", 
             StyleBox["k", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1385740688"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1385740688-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1385740688-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1385740688", "Proof|ID:1385740688-1"},
 CellID->1882316559],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1385740688-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->782413460]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(allPairs content)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(allPairs content)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["b", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["b", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "\:22ff", 
             RowBox[{"allPairs", "[", 
               StyleBox["k", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["b", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["a", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["b", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "\:22ff", 
             RowBox[{"allPairs", "[", 
               StyleBox["k", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1430860208"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1430860208-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1430860208-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1430860208", "Proof|ID:1430860208-1"},
 CellID->1876195014],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1430860208-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1624086186]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->2123044522],

Cell[CellGroupData[{

Cell["Theorem: Properties of update", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648369954666053*^9, 3.648369961700397*^9}},
 CellID->774910772],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"k", "\[Element]", 
   SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 963656485}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648369992644178*^9, 3.648370041412921*^9}, {
  3.648370072876596*^9, 3.648370073839692*^9}, {3.648370464178219*^9, 
  3.648370466151206*^9}},
 CellTags->{
  "ID:963656485", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->963656485],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", 
     RowBox[{"NatPairs", "[", 
      RowBox[{"k", ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[LessEqual]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}], "]"}], "]"}]], 
   RowBox[{"[", "P", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1750952092}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648369965591441*^9, 3.648369985363243*^9}, {
   3.6483700258560677`*^9, 3.648370034434209*^9}, 3.648370076790762*^9},
 CellTags->{
  "ID:1750952092", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1750952092],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"up", "=", 
   RowBox[{"update", "[", 
    RowBox[{"P", ",", 
     RowBox[{"k", "+", "1"}]}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1500640019}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648370051858952*^9, 3.6483700784502172`*^9}, {
  3.648370468661193*^9, 3.648370469106596*^9}},
 CellTags->{
  "ID:1500640019", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1500640019],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   RowBox[{"DomainTuples", "[", 
    RowBox[{"NatPairs", "[", 
     RowBox[{
      RowBox[{"k", "+", "1"}], ",", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       TagBox["\[LessEqual]",
        Identity], 
       TagBox[")",
        "AutoParentheses"]}]}], "]"}], "]"}]], 
  RowBox[{"[", "up", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership update)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:723173326", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648370084610054*^9, 3.648370090971936*^9}, {
  3.648370472979113*^9, 3.6483704734603853`*^9}},
 CellTags->{
  "ID:723173326", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership update", "IR"},
 CellID->723173326],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "a"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      TagBox["\:22ff",
       Identity,
       SyntaxForm->"a+b"], "P"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      TagBox["\:22ff",
       Identity,
       SyntaxForm->"a+b"], "up"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(update content I)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:392184336", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649131186671835*^9, 3.6491312062177258`*^9}},
 CellTags->{
  "ID:392184336", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "update content I"},
 CellID->392184336],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"a", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"k", "+", "1"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"a", ",", 
      RowBox[{"k", "+", "1"}]}], "\[RightAngleBracket]"}], 
    TagBox["\:22ff",
     Identity,
     SyntaxForm->"a+b"], "up"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(update content II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:341058317", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649151801334594*^9, 3.649151822133856*^9}},
 CellTags->{
  "ID:341058317", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "update content II"},
 CellID->341058317],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->522299237]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1129995189],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership update)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership update)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    "\[LessEqual]"}], "]"}], "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", 
              RowBox[{"NatPairs", "[", 
                RowBox[{
                  RowBox[{
                    StyleBox["k", "ExpressionVariable"], "+", "1"}], ",", 
                  "\[LessEqual]"}], "]"}], "]"}]], "[", 
           RowBox[{"update", "[", 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "+", "1"}]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    "\[LessEqual]"}], "]"}], "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", 
              RowBox[{"NatPairs", "[", 
                RowBox[{
                  RowBox[{
                    StyleBox["k", "ExpressionVariable"], "+", "1"}], ",", 
                  "\[LessEqual]"}], "]"}], "]"}]], "[", 
           RowBox[{"update", "[", 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "+", "1"}]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:723173326"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p723173326-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p723173326-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:723173326", "Proof|ID:723173326-1"},
 CellID->1574692164],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p723173326-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->310043985]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(update content I)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(update content I)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    "\[LessEqual]"}], "]"}], "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], "\:22ff", 
             StyleBox["P", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], "\:22ff", 
             RowBox[{"update", "[", 
               RowBox[{
                 StyleBox["P", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["k", "ExpressionVariable"], "+", "1"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    "\[LessEqual]"}], "]"}], "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], "\:22ff", 
             StyleBox["P", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], "\:22ff", 
             RowBox[{"update", "[", 
               RowBox[{
                 StyleBox["P", "ExpressionVariable"], ",", 
                 RowBox[{
                   StyleBox["k", "ExpressionVariable"], "+", "1"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:392184336"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p392184336-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p392184336-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:392184336", "Proof|ID:392184336-1"},
 CellID->393000099],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p392184336-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1623641860]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(update content II)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(update content II)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    "\[LessEqual]"}], "]"}], "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "+", "1"}]}]}}]], 
         RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "+", "1"}]}], 
             "\[RightAngleBracket]"}], "\:22ff", 
           RowBox[{"update", "[", 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "+", "1"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    "\[LessEqual]"}], "]"}], "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "+", "1"}]}]}}]], 
         RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "+", "1"}]}], 
             "\[RightAngleBracket]"}], "\:22ff", 
           RowBox[{"update", "[", 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "+", "1"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:341058317"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p341058317-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p341058317-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:341058317", "Proof|ID:341058317-1"},
 CellID->1522358468],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p341058317-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->975211483]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1143846827],

Cell[CellGroupData[{

Cell["Lemma: Properties of processed", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648888183013509*^9, 3.648888199621606*^9}, {
  3.648888351367065*^9, 3.6488883579238997`*^9}, {3.649041173345683*^9, 
  3.649041178763947*^9}, {3.6490413800767612`*^9, 3.649041381042499*^9}, {
  3.649131056316865*^9, 3.649131064348309*^9}},
 CellID->1184365251],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", "j", ",", 
    RowBox[{"isTuple", "[", "P", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"processed", "[", 
     RowBox[{"i", ",", "j", ",", "P"}], "]"}], "\[DoubleLongLeftRightArrow]", 
    
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[Not]", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"i", ",", "j"}], "\[RightAngleBracket]"}], 
         TagBox["\:22ff",
          Identity,
          SyntaxForm->"a+b"], "P"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[Not]", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"j", ",", "i"}], "\[RightAngleBracket]"}], 
         TagBox["\:22ff",
          Identity,
          SyntaxForm->"a+b"], "P"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(processed alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1597870023", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64913109631213*^9, 3.6491311260809097`*^9}, {
  3.649131558729898*^9, 3.6491315624444933`*^9}},
 CellTags->{
  "ID:1597870023", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed alternative", "IR"},
 CellID->1597870023],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", "j", ",", "P"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"processed", "[", 
     RowBox[{"i", ",", "j", ",", "P"}], "]"}], "\[Implies]", 
    RowBox[{"processed", "[", 
     RowBox[{"j", ",", "i", ",", "P"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(processed symmetric)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:40246499", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649056291066627*^9, 3.649056323980137*^9}},
 CellTags->{
  "ID:40246499", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed symmetric"},
 CellID->40246499],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"k", "\[Element]", 
   SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1131819573}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648889784464409*^9, 3.648889789436606*^9}},
 CellTags->{
  "ID:1131819573", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1131819573],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", "k"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{"processed", "[", 
     RowBox[{"i", ",", "j", ",", 
      RowBox[{"allPairs", "[", "k", "]"}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(processed allPairs)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1083804603", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648889805329378*^9, 3.6488898294085083`*^9}},
 CellTags->{
  "ID:1083804603", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed allPairs"},
 CellID->1083804603],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{"k", ",", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         TagBox["\[LessEqual]",
          Identity], 
         TagBox[")",
          "AutoParentheses"]}]}], "]"}], "]"}]], 
    RowBox[{"[", "P", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", 
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"k", "+", "1"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"processed", "[", 
      RowBox[{"i", ",", "j", ",", 
       RowBox[{"update", "[", 
        RowBox[{"P", ",", 
         RowBox[{"k", "+", "1"}]}], "]"}]}], "]"}], "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"\[And]", 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{"i", "\[Element]", 
            SubscriptBox["\[DoubleStruckCapitalN]", 
             RowBox[{"1", ",", "\[Ellipsis]", ",", "k"}]]}]},
          {
           RowBox[{"j", "\[Element]", 
            SubscriptBox["\[DoubleStruckCapitalN]", 
             RowBox[{"1", ",", "\[Ellipsis]", ",", "k"}]]}]},
          {
           RowBox[{"processed", "[", 
            RowBox[{"i", ",", "j", ",", "P"}], "]"}]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(processed update)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:316507172", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648888207969057*^9, 3.648888332846758*^9}, {
   3.6488884471910467`*^9, 3.6488884484109497`*^9}, 3.648889792316772*^9},
 CellTags->{
  "ID:316507172", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed update"},
 CellID->316507172],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", "j", ",", "i0", ",", "j0", ",", 
    RowBox[{"t", "..."}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"processed", "[", 
     RowBox[{"i", ",", "j", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"i0", ",", "j0"}], "\[RightAngleBracket]"}], ",", 
        RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}]}], "]"}], 
    "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           RowBox[{"i", "\[NotEqual]", "i0"}], "\[Or]", 
           RowBox[{"j", "\[NotEqual]", "j0"}]}]},
         {
          RowBox[{
           RowBox[{"i", "\[NotEqual]", "j0"}], "\[Or]", 
           RowBox[{"j", "\[NotEqual]", "i0"}]}]},
         {
          RowBox[{"processed", "[", 
           RowBox[{"i", ",", "j", ",", 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"t", "..."}], "\[RightAngleBracket]"}]}], "]"}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(processed prepend)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1179241697", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649041195648801*^9, 3.6490413187269573`*^9}, {
  3.649041535252778*^9, 3.6490415574365587`*^9}},
 CellTags->{
  "ID:1179241697", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed prepend"},
 CellID->1179241697],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1183261273]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->186310157],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(processed alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(processed alternative)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}, {
             RowBox[{"isTuple", "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"processed", "[", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["P", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\:22ff", 
                    StyleBox["P", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\:22ff", 
                    StyleBox["P", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}, {
             RowBox[{"isTuple", "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"processed", "[", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["P", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\:22ff", 
                    StyleBox["P", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\:22ff", 
                    StyleBox["P", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1597870023"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1597870023-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1597870023-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1597870023", "Proof|ID:1597870023-1"},
 CellID->932327505],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1597870023-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->559044054]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(processed symmetric)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(processed symmetric)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["j", "ExpressionVariable"], ",", 
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"processed", "[", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["P", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{"processed", "[", 
             RowBox[{
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["P", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["j", "ExpressionVariable"], ",", 
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"processed", "[", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["P", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{"processed", "[", 
             RowBox[{
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["P", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:40246499"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p40246499-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p40246499-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:40246499", "Proof|ID:40246499-1"},
 CellID->1343087848],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p40246499-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1360525692]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(processed allPairs)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(processed allPairs)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{"processed", "[", 
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"], ",", 
                 RowBox[{"allPairs", "[", 
                   StyleBox["k", "ExpressionVariable"], "]"}]}], "]"}], 
             TagBox[")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               StyleBox["k", "ExpressionVariable"]}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{"processed", "[", 
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"], ",", 
                 RowBox[{"allPairs", "[", 
                   StyleBox["k", "ExpressionVariable"], "]"}]}], "]"}], 
             TagBox[
             ")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1083804603"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1083804603-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1083804603-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1083804603", "Proof|ID:1083804603-1"},
 CellID->465417913],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1083804603-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->20057220]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(processed update)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(processed update)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    "\[LessEqual]"}], "]"}], "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "+", "1"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "+", "1"}]}]}}]], 
         RowBox[{
           RowBox[{"processed", "[", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               RowBox[{"update", "[", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "+", "1"}]}], 
                 "]"}]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", 
                RowBox[{"1", ",", "\[Ellipsis]", ",", 
                  StyleBox["k", "ExpressionVariable"]}]]}], "\[And]", 
             RowBox[{
               StyleBox["j", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", 
                RowBox[{"1", ",", "\[Ellipsis]", ",", 
                  StyleBox["k", "ExpressionVariable"]}]]}], "\[And]", 
             RowBox[{"processed", "[", 
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"], ",", 
                 StyleBox["P", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    "\[LessEqual]"}], "]"}], "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "+", "1"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "+", "1"}]}]}}]], 
         RowBox[{
           RowBox[{"processed", "[", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               RowBox[{"update", "[", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], ",", 
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], "+", "1"}]}], 
                 "]"}]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", 
                RowBox[{"1", ",", "\[Ellipsis]", ",", 
                  StyleBox["k", "ExpressionVariable"]}]]}], "\[And]", 
             RowBox[{
               StyleBox["j", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", 
                RowBox[{"1", ",", "\[Ellipsis]", ",", 
                  StyleBox["k", "ExpressionVariable"]}]]}], "\[And]", 
             RowBox[{"processed", "[", 
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"], ",", 
                 StyleBox["P", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:316507172"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p316507172-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p316507172-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:316507172", "Proof|ID:316507172-1"},
 CellID->660680371],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p316507172-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1967876509]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(processed prepend)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(processed prepend)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["j", "ExpressionVariable"], ",", 
            StyleBox["i0", "ExpressionVariable"], ",", 
            StyleBox["j0", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"t", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"processed", "[", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["i0", "ExpressionVariable"], ",", 
                    StyleBox["j0", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}]}], "]"}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["i0", "ExpressionVariable"]}], "\[Or]", 
                 RowBox[{
                   StyleBox["j", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["j0", "ExpressionVariable"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["j0", "ExpressionVariable"]}], "\[Or]", 
                 RowBox[{
                   StyleBox["j", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["i0", "ExpressionVariable"]}]}], "\[And]", 
               RowBox[{"processed", "[", 
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["i", "ExpressionVariable"], ",", 
            StyleBox["j", "ExpressionVariable"], ",", 
            StyleBox["i0", "ExpressionVariable"], ",", 
            StyleBox["j0", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"t", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"processed", "[", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["i0", "ExpressionVariable"], ",", 
                    StyleBox["j0", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}]}], "]"}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["i0", "ExpressionVariable"]}], "\[Or]", 
                 RowBox[{
                   StyleBox["j", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["j0", "ExpressionVariable"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["j0", "ExpressionVariable"]}], "\[Or]", 
                 RowBox[{
                   StyleBox["j", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["i0", "ExpressionVariable"]}]}], "\[And]", 
               RowBox[{"processed", "[", 
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1179241697"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1179241697-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1179241697-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1179241697", "Proof|ID:1179241697-1"},
 CellID->2041135352],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1179241697-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1629853247]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Definitions", "Section",
 CellChangeTimes->{{3.590814495331113*^9, 3.590814500335643*^9}, {
   3.590816307063519*^9, 3.590816308616468*^9}, {3.590834000388527*^9, 
   3.590834004560697*^9}, 3.590834107885453*^9, 3.60777172991225*^9, {
   3.645766703528338*^9, 3.645766705362876*^9}},
 CellID->509234429],

Cell[CellGroupData[{

Cell["Informal Explanation", "Subsection",
 CellChangeTimes->{{3.590830406688664*^9, 3.590830411996763*^9}},
 CellID->1820866479],

Cell[TextData[{
 "The functor \[OpenCurlyDoubleQuote]GroebnerRing\[CloseCurlyDoubleQuote] \
extends a domain ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " by adding several new operations, most notably operations for computing \
Groebner bases. The only operations that ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " has to provide are, besides the usual ring operations and a test for \
equality, the following:"
}], "Text",
 CellChangeTimes->{{3.590814535015566*^9, 3.590814683792137*^9}, {
  3.590815612389775*^9, 3.590815612854046*^9}, {3.590815900588747*^9, 
  3.590815912465344*^9}, {3.591943783426792*^9, 3.591943783744542*^9}},
 CellID->1002761981],

Cell[CellGroupData[{

Cell["\<\
An irreflexive partial order relation (\[OpenCurlyDoubleQuote]\[Succeeds]\
\[CloseCurlyDoubleQuote]) which should also be noetherian.\
\>", "Item",
 CellChangeTimes->{3.59204273196003*^9},
 CellID->1613933805],

Cell[TextData[{
 "A function that, for all ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 ", returns ",
 StyleBox["all",
  FontWeight->"Bold"],
 " minimal non-trivial common reducibles of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " in a tuple (\[OpenCurlyDoubleQuote]lcrd\[CloseCurlyDoubleQuote]). Note \
that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Equal]", "y"}], TraditionalForm]]],
 " is also possible!\nThis is different to [Buchberger2003]: There, only ",
 StyleBox["one",
  FontWeight->"Bold"],
 " minimal non-trivial common reducible is considered, and the case ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Equal]", "y"}], TraditionalForm]]],
 " is ruled out."
}], "Item",
 CellChangeTimes->{{3.592042742322896*^9, 3.592042750441021*^9}},
 CellID->307340730],

Cell[TextData[{
 "A function that returns a valid reduction multiplier for reducing ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " modulo ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " (\[OpenCurlyDoubleQuote]rdm\[CloseCurlyDoubleQuote]). We follow the same \
paradigm as in [Buchberger2003] by setting this reduction multiplier to 0 iff \
no reduction is possible (modulo ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 ").\nOf course, reduction has to be compatible with the ordering, i.e. ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " is only reducible modulo ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " if ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Succeeds]", 
    RowBox[{"x", "-", 
     RowBox[{
      RowBox[{"rdm", "[", 
       RowBox[{"x", ",", "y"}], "]"}], "*", "y"}]}]}], TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.592042742322896*^9, 3.592042759074751*^9}},
 CellID->1910860741],

Cell[TextData[{
 "A function that returns the critical pair of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " and minimial non-trivial common reducible ",
 Cell[BoxData[
  FormBox["z", TraditionalForm]]],
 " (\[OpenCurlyDoubleQuote]cp\[CloseCurlyDoubleQuote]). This function cannot \
be defined in the Groebner ring, because according to [Stifter1988] and \
[Stifter1989], the reductions that are used to obtain the critical pair have \
to be irrelative, and this can only be ensured in ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 " itself.\nThe reason why in [Buchberger2003] the function \
\[OpenCurlyDoubleQuote]cp\[CloseCurlyDoubleQuote] (resp. \
\[OpenCurlyDoubleQuote]cpd\[CloseCurlyDoubleQuote]) can be defined in the \
Groebner ring functor is that there only critical pairs ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[NotEqual]", "y"}], TraditionalForm]]],
 " are considered, and in this case reductions are always irrelative."
}], "Item",
 CellChangeTimes->{{3.592042742322896*^9, 3.592042763764432*^9}},
 CellID->1871827641]
}, Open  ]],

Cell[TextData[{
 "Building upon ",
 Cell[BoxData[
  FormBox["D", TraditionalForm]]],
 "\[CloseCurlyQuote]s operations, the \[OpenCurlyDoubleQuote]GroebnerRing\
\[CloseCurlyDoubleQuote] functor defines the following new functions:"
}], "Text",
 CellChangeTimes->{{3.590814535015566*^9, 3.590814683792137*^9}, {
  3.590815612389775*^9, 3.590815612854046*^9}, {3.590815900588747*^9, 
  3.590815912465344*^9}, {3.590816051328113*^9, 3.590816090998818*^9}},
 CellID->701994113],

Cell[CellGroupData[{

Cell[TextData[{
 "A function that reduces ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " modulo ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " in one step (\[OpenCurlyDoubleQuote]rd\[CloseCurlyDoubleQuote])."
}], "Item",
 CellChangeTimes->{{3.59204277479837*^9, 3.592042780711912*^9}},
 CellID->687289781],

Cell[TextData[{
 "A function that totally reduces ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " modulo a tuple ",
 Cell[BoxData[
  FormBox["Y", TraditionalForm]]],
 " (\[OpenCurlyDoubleQuote]trd\[CloseCurlyDoubleQuote])."
}], "Item",
 CellChangeTimes->{{3.59204277479837*^9, 3.592042788037416*^9}},
 CellID->1002810321],

Cell[TextData[{
 "A function that totally reduces ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " modulo a tuple ",
 Cell[BoxData[
  FormBox["Y", TraditionalForm]]],
 " and collects the cofactors as well (\[OpenCurlyDoubleQuote]tcrd\
\[CloseCurlyDoubleQuote])."
}], "Item",
 CellChangeTimes->{{3.59204277479837*^9, 3.592042796714147*^9}},
 CellID->101699731],

Cell[TextData[{
 "A function that computes the Groebner basis of a tuple ",
 Cell[BoxData[
  FormBox["X", TraditionalForm]]],
 " (\[OpenCurlyDoubleQuote]GB\[CloseCurlyDoubleQuote]).\nNote here that ",
 StyleBox["not",
  FontWeight->"Bold"],
 " the critical-pair difference is reduced, but that both constituents of the \
critical pair are reduced separately and only ",
 StyleBox["then",
  FontWeight->"Bold"],
 " the difference is formed. This has to be done due to [Stifter1985], since \
otherwise wrong results would be obtained. Example: \
\[LeftAngleBracket]2091,2337,2829\[RightAngleBracket] in \
\[DoubleStruckCapitalZ] gives -246, if critical pair difference is used, \
whereas the gcd is 123!"
}], "Item",
 CellChangeTimes->{{3.59204277479837*^9, 3.592042801795943*^9}, {
  3.59410312781665*^9, 3.594103130228677*^9}, {3.594103199787234*^9, 
  3.59410326794131*^9}},
 CellID->1471005405]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Gr\[ODoubleDot]bner Ring", "Subsection",
 CellChangeTimes->{{3.590830397583161*^9, 3.590830399414385*^9}, {
  3.6457667841793823`*^9, 3.645766794048307*^9}},
 CellID->1076888715],

Cell["", "OpenEnvironment",
 CellID->508559908],

Cell[CellGroupData[{

Cell["Definition: Groebner Ring", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.590810466487417*^9, 3.590810471212538*^9}, {
   3.590816314378241*^9, 3.590816315711609*^9}, 3.590834012881357*^9, {
   3.590834100627162*^9, 3.590834102777678*^9}, 3.610098654566184*^9, {
   3.613375973240609*^9, 3.613375973424726*^9}},
 CellID->388885171],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", "D"]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1573502139}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590810478223491*^9, 3.590810479577169*^9}},
 CellTags->{
  "ID:1573502139", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1573502139],

Cell[BoxData[
 RowBox[{
  RowBox[{"GroebnerRing", "[", "D", "]"}], ":=", 
  UnderscriptBox["\[CapitalDelta]", "N"]}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 377704681}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590810492586065*^9, 3.590810536568862*^9}, {
   3.607339752824867*^9, 3.607339753674482*^9}, 3.645766807285335*^9},
 CellTags->{
  "ID:377704681", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->377704681],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "x", ",", "y", ",", "i", ",", "j", ",", "l", ",", "mi", ",", "mj", ",", "C",
    ",", "P", ",", 
   RowBox[{"R", "..."}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1468120413}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590810519204274*^9, 3.590810526475304*^9}, {
   3.590811113815022*^9, 3.590811122583638*^9}, {3.590811532127638*^9, 
   3.590811538508059*^9}, {3.590811805934485*^9, 3.590811806708129*^9}, {
   3.590812522337135*^9, 3.59081253820046*^9}, {3.590813117501017*^9, 
   3.590813127460218*^9}, {3.592026774625217*^9, 3.592026780089743*^9}, {
   3.592039938629474*^9, 3.592039939648736*^9}, {3.592041417665219*^9, 
   3.592041420906844*^9}, {3.592041513004348*^9, 3.592041515723723*^9}, {
   3.592041594824149*^9, 3.592041611697621*^9}, {3.592041747760502*^9, 
   3.592041757145427*^9}, 3.593763634589205*^9, 3.59403587284446*^9, {
   3.594636351134745*^9, 3.594636351356311*^9}, {3.59463793350088*^9, 
   3.594637933714704*^9}, {3.594642919639365*^9, 3.594642920408186*^9}, {
   3.595057617978057*^9, 3.595057621271173*^9}, {3.595057658915725*^9, 
   3.595057662878155*^9}, {3.595245326259666*^9, 3.595245338079507*^9}, {
   3.595245370579881*^9, 3.595245372429706*^9}, {3.595245432171063*^9, 
   3.595245433828291*^9}, {3.595247918317629*^9, 3.595247920906298*^9}, {
   3.645767669519658*^9, 3.64576767143998*^9}, {3.645768368916648*^9, 
   3.645768369215679*^9}, {3.645768917178562*^9, 3.645768929122849*^9}, {
   3.6457734535528*^9, 3.6457734543809023`*^9}},
 CellTags->{
  "ID:1468120413", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1468120413],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Element]", "N"], 
   RowBox[{"[", "x", "]"}]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["\[Element]", "D"], 
   RowBox[{"[", "x", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing \[Element])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1774031579", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590810548699564*^9, 3.590810569770297*^9}, {
  3.592041630937578*^9, 3.592041631896009*^9}, {3.594035875652224*^9, 
  3.594035878563569*^9}, {3.594642923169162*^9, 3.594642924606098*^9}, {
  3.595057632050218*^9, 3.595057633007402*^9}, {3.60733975607305*^9, 
  3.607339756996495*^9}},
 CellTags->{
  "ID:1774031579", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing \[Element]"},
 CellID->1774031579],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["0", "N"], ":=", 
  UnderscriptBox["0", "D"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:182210603", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590810617984028*^9, 3.590810626819227*^9}, {
  3.607339758103916*^9, 3.607339758916852*^9}},
 CellTags->{
  "ID:182210603", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing 0"},
 CellID->182210603],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["1", "N"], ":=", 
  UnderscriptBox["1", "D"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:814933967", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590810632183932*^9, 3.590810640872741*^9}, {
  3.607339761094176*^9, 3.607339762091435*^9}},
 CellTags->{
  "ID:814933967", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing 1"},
 CellID->814933967],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["+", "N"], "y"}], ":=", 
  RowBox[{"x", 
   UnderscriptBox["+", "D"], "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing +)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1684568525", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590810648086526*^9, 3.590810660336159*^9}, {
  3.607339762904646*^9, 3.607339763643719*^9}},
 CellTags->{
  "ID:1684568525", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing +"},
 CellID->1684568525],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["-", "N"], "y"}], ":=", 
  RowBox[{"x", 
   UnderscriptBox["-", "D"], "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing -)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:584522550", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590810710973222*^9, 3.59081072100195*^9}, {
  3.607339764603978*^9, 3.607339765601353*^9}},
 CellTags->{
  "ID:584522550", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing -"},
 CellID->584522550],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["-", "N"], "x"}], ":=", 
  RowBox[{
   UnderscriptBox["-", "D"], "x"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing - unary)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1458964616", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590810730954042*^9, 3.590810739650624*^9}, {
  3.607339766598423*^9, 3.607339767411305*^9}},
 CellTags->{
  "ID:1458964616", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing - unary"},
 CellID->1458964616],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["*", "N"], "y"}], ":=", 
  RowBox[{"x", 
   UnderscriptBox["*", "D"], "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing *)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1496488880", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590810669910874*^9, 3.590810679909174*^9}, {
  3.607339768298328*^9, 3.607339769184829*^9}},
 CellTags->{
  "ID:1496488880", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing *"},
 CellID->1496488880],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", 
    UnderscriptBox["\[Succeeds]", "N"], "y"}], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", 
    UnderscriptBox["\[Succeeds]", "D"], "y"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing \[Succeeds])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:342394137", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590810754961668*^9, 3.590810767252134*^9}, {
  3.590915557134551*^9, 3.590915558531057*^9}, {3.591944082290723*^9, 
  3.591944089190963*^9}, {3.607339770200787*^9, 3.607339771123946*^9}, {
  3.645766834367797*^9, 3.6457668372908707`*^9}},
 CellTags->{
  "ID:342394137", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing \[Succeeds]"},
 CellID->342394137],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["multN", "N"], "[", "x", "]"}], ":=", 
  RowBox[{
   UnderscriptBox["multN", "D"], "[", "x", "]"}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing multN)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1566608311", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648368465750918*^9, 3.6483684794680367`*^9}},
 CellTags->{
  "ID:1566608311", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing multN"},
 CellID->1566608311],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "N"], "[", 
   RowBox[{"x", ",", "y"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   UnderscriptBox["mult", "D"], "[", 
   RowBox[{"x", ",", "y"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing mult 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:628841792", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649585064059825*^9, 3.649585075702774*^9}},
 CellTags->{
  "ID:628841792", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing mult 1"},
 CellID->628841792],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "N"], "[", 
   RowBox[{"x", ",", "y", ",", "i"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   UnderscriptBox["mult", "D"], "[", 
   RowBox[{"x", ",", "y", ",", "i"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing mult 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:378340035", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6495850388825207`*^9, 3.649585049489229*^9}},
 CellTags->{
  "ID:378340035", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing mult 2"},
 CellID->378340035],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lcrd", "N"], "[", 
   RowBox[{"x", ",", "y", ",", "i", ",", "j"}], "]"}], ":=", 
  RowBox[{
   UnderscriptBox["lcrd", "D"], "[", 
   RowBox[{"x", ",", "y", ",", "i", ",", "j"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing lcrd)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:822076309", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590810847711658*^9, 3.59081086548938*^9}, {
  3.607339783108974*^9, 3.607339784142542*^9}, {3.610098474024149*^9, 
  3.610098478557001*^9}},
 CellTags->{
  "ID:822076309", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lcrd"},
 CellID->822076309],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["rdm", "N"], "[", 
   RowBox[{"x", ",", "y"}], "]"}], ":=", 
  RowBox[{
   UnderscriptBox["rdm", "D"], "[", 
   RowBox[{"x", ",", "y"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing rdm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:849156542", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590810870534436*^9, 3.590810885810094*^9}, {
  3.607339789992259*^9, 3.607339791266072*^9}},
 CellTags->{
  "ID:849156542", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing rdm"},
 CellID->849156542],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lcrd", "N"], "[", "x", "]"}], ":=", 
  RowBox[{
   UnderscriptBox["lcrd", "D"], "[", "x", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing lcrd unary)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:809363439", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6457731429176397`*^9, 3.645773146894947*^9}},
 CellTags->{
  "ID:809363439", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lcrd unary"},
 CellID->809363439],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["cpm", "N"], "[", 
   RowBox[{"x", ",", "y", ",", "l", ",", "i", ",", "j"}], "]"}], ":=", 
  RowBox[{
   UnderscriptBox["cpm", "D"], "[", 
   RowBox[{"x", ",", "y", ",", "l", ",", "i", ",", "j"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing cpm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1671864047", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645773691708338*^9, 3.645773708147462*^9}},
 CellTags->{
  "ID:1671864047", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing cpm"},
 CellID->1671864047],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"rd", "[", 
    RowBox[{"x", ",", "y"}], "]"}], TraditionalForm]]],
 " reduces ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " modulo ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 ", if possible."
}], "EnvironmentText",
 CellChangeTimes->{{3.596192933098709*^9, 3.596192950757366*^9}, 
   3.596193814879557*^9},
 CellID->1362566519],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["rd", "N"], "[", 
   RowBox[{"x", ",", "y"}], "]"}], ":=", 
  RowBox[{"x", 
   UnderscriptBox["-", "N"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["rdm", "N"], "[", 
     RowBox[{"x", ",", "y"}], "]"}], 
    UnderscriptBox["*", "N"], "y"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing rd)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:558323540", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590811251050826*^9, 3.590811296933772*^9}, {
  3.607339792411164*^9, 3.607339793556132*^9}},
 CellTags->{
  "ID:558323540", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing rd"},
 CellID->558323540],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"rdT", "[", 
    RowBox[{"x", ",", "C", ",", "i"}], "]"}], TraditionalForm]]],
 " reduces ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " modulo the first possible element of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " starting from ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 ", if there is one."
}], "EnvironmentText",
 CellChangeTimes->{{3.649495133798478*^9, 3.649495160583733*^9}, {
  3.649495312881132*^9, 3.649495342398161*^9}},
 CellID->1551627756],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["rdT", "N"], "[", 
   RowBox[{"x", ",", "C", ",", "i"}], "]"}], ":=", 
  RowBox[{"\[Piecewise]", GridBox[{
     {"x", "\[DoubleLeftArrow]", 
      RowBox[{"i", ">", 
       RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]},
     {
      RowBox[{
       UnderscriptBox["let", 
        RowBox[{"x1", "=", 
         RowBox[{
          UnderscriptBox["rd", "N"], "[", 
          RowBox[{"x", ",", 
           SubscriptBox["C", "i"]}], "]"}]}]], 
       RowBox[{"\[Piecewise]", GridBox[{
          {"x1", "\[DoubleLeftArrow]", 
           RowBox[{"x", 
            UnderscriptBox["\[Succeeds]", "N"], "x1"}]},
          {
           RowBox[{
            UnderscriptBox["rdT", "N"], "[", 
            RowBox[{"x", ",", "C", ",", 
             RowBox[{"i", "+", "1"}]}], "]"}], "\[DoubleLeftArrow]", 
           "otherwise"}
         }]}]}], "\[DoubleLeftArrow]", "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing rdT)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1403530194", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649495029772912*^9, 3.649495097523718*^9}},
 CellTags->{
  "ID:1403530194", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing rdT"},
 CellID->1403530194],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"trd", "[", 
    RowBox[{"x", ",", "C"}], "]"}], TraditionalForm]]],
 " totally reduces ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " modulo tuple ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ", making use of auxiliary function ",
 Cell[BoxData[
  FormBox["rdT", TraditionalForm]]],
 "."
}], "EnvironmentText",
 CellChangeTimes->{{3.596192960332472*^9, 3.596192989969512*^9}, 
   3.596193813171517*^9, {3.645769030339172*^9, 3.645769030339282*^9}, {
   3.645769070910647*^9, 3.645769070910756*^9}, {3.6494952705329247`*^9, 
   3.649495271086669*^9}},
 CellID->2034063019],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["trd", "N"], "[", 
   RowBox[{"x", ",", "C"}], "]"}], ":=", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["let", 
     RowBox[{"x1", "=", 
      RowBox[{
       UnderscriptBox["rdT", "N"], "[", 
       RowBox[{"x", ",", "C", ",", "1"}], "]"}]}]], 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["trd", "N"], "[", 
         RowBox[{"x1", ",", "C"}], "]"}], "\[DoubleLeftArrow]", 
        RowBox[{"x", 
         UnderscriptBox["\[Succeeds]", "N"], "x1"}]},
       {"x", "\[DoubleLeftArrow]", "otherwise"}
      }]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing trd)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:689058884", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590811097547198*^9, 3.590811149248184*^9}, {
   3.592040971320777*^9, 3.592040971692672*^9}, {3.592041769940404*^9, 
   3.59204177151747*^9}, {3.607339796581358*^9, 3.607339797836971*^9}, 
   3.6457690202465143`*^9, 3.6457693267369347`*^9, 3.645773500084735*^9, {
   3.649487982157732*^9, 3.649487983272237*^9}, {3.649495207794087*^9, 
   3.649495259974362*^9}},
 CellTags->{
  "ID:689058884", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing trd"},
 CellID->689058884],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"lmTuple", "[", 
    RowBox[{"x", ",", "y"}], "]"}], TraditionalForm]]],
 " returns the tuple consisting of all lcrd-multiplier-triples of the two \
elements ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 ", i.e. the result is a tuple consisting of elements of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"l", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       SubscriptBox["m", "x"], ",", 
       SubscriptBox["m", "y"]}], "\[RightAngleBracket]"}]}], 
    "\[RightAngleBracket]"}], TraditionalForm]]],
 " where ",
 Cell[BoxData[
  FormBox["l", TraditionalForm]]],
 " is a least common non-trivial reducible of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "x"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "y"], TraditionalForm]]],
 " are the corresponding multipliers."
}], "EnvironmentText",
 CellChangeTimes->{{3.596193051074086*^9, 3.596193154175139*^9}, {
   3.596193533978049*^9, 3.596193612316742*^9}, 3.596193810299366*^9, {
   3.610099051513561*^9, 3.610099083468072*^9}, {3.610099148399598*^9, 
   3.610099151319463*^9}, 3.610099269048171*^9, {3.645769156101419*^9, 
   3.645769300896875*^9}, 3.645769403222836*^9, {3.645772450478397*^9, 
   3.645772450626514*^9}, {3.645772481617474*^9, 3.645772486680374*^9}, {
   3.645772676157433*^9, 3.6457726794385233`*^9}, {3.649504680399942*^9, 
   3.649504681292376*^9}},
 CellID->666716832],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lmTuple", "N"], "[", 
   RowBox[{"x", ",", "y"}], "]"}], ":=", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Piecewise]", GridBox[{
      {
       RowBox[{
        UnderscriptBox["lmTupleAux", "N"], "[", 
        RowBox[{"y", ",", 
         RowBox[{"pairs", "[", 
          RowBox[{"2", ",", 
           RowBox[{
            UnderscriptBox["multN", "D"], "[", "y", "]"}]}], "]"}]}], "]"}], 
       "\[DoubleLeftArrow]", 
       RowBox[{"x", "\[Equal]", "y"}]},
      {
       RowBox[{
        UnderscriptBox["cpmTuple", "N"], "[", 
        RowBox[{"x", ",", "y"}], "]"}], "\[DoubleLeftArrow]", "otherwise"}
     }]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing lmTuple)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1528624471", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.590813671989039*^9, 3.590813791132313*^9}, {
   3.59081386033703*^9, 3.590813882486047*^9}, {3.591943451907128*^9, 
   3.591943475999585*^9}, {3.592041014808201*^9, 3.592041018536057*^9}, {
   3.59463630581078*^9, 3.594636316559394*^9}, {3.594638521423177*^9, 
   3.594638572626171*^9}, {3.595244986502244*^9, 3.595245039272563*^9}, {
   3.595245269334146*^9, 3.595245278881878*^9}, {3.595247193990375*^9, 
   3.595247218740873*^9}, {3.595247686717673*^9, 3.595247714612775*^9}, {
   3.595247812414809*^9, 3.595247876888999*^9}, 3.595247926132646*^9, {
   3.595248369385284*^9, 3.595248413274778*^9}, {3.595248563144118*^9, 
   3.595248578662604*^9}, {3.607339866984721*^9, 3.607339896464107*^9}, {
   3.610098440823806*^9, 3.610098446866263*^9}, {3.610098567608838*^9, 
   3.610098607937981*^9}, {3.610098756078075*^9, 3.610098790106672*^9}, {
   3.610099100791485*^9, 3.610099119059428*^9}, {3.610099204808068*^9, 
   3.61009921064556*^9}, {3.610099328762665*^9, 3.610099367814251*^9}, {
   3.610099417147477*^9, 3.610099448502119*^9}, {3.610102494561526*^9, 
   3.610102494674246*^9}, {3.610102601635393*^9, 3.610102602085253*^9}, {
   3.61010335766321*^9, 3.610103360933293*^9}, {3.610103495761268*^9, 
   3.610103496101641*^9}, {3.6318556082555103`*^9, 3.6318556086625023`*^9}, 
   3.631856606970002*^9, {3.6457726040295563`*^9, 3.645772644136286*^9}, {
   3.6457726920173883`*^9, 3.6457727644949417`*^9}, {3.6457738193419857`*^9, 
   3.645773831637828*^9}, {3.649504502326264*^9, 3.649504509474811*^9}, 
   3.649564872316319*^9, 3.649565968546289*^9},
 CellTags->{
  "ID:1528624471", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lmTuple"},
 CellID->1528624471],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lmTupleAux", "N"], "[", 
   RowBox[{"x", ",", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], "]"}], ":=", 
  RowBox[{
  "\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing lmTupleAux 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1192872489", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610102742554179*^9, 3.610102758027676*^9}},
 CellTags->{
  "ID:1192872489", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lmTupleAux 1"},
 CellID->1192872489],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lmTupleAux", "N"], "[", 
   RowBox[{"x", ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"i", ",", "j"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"R", "..."}]}], "\[RightAngleBracket]"}]}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    UnderscriptBox["cpmTuple", "N"], "[", 
    RowBox[{"x", ",", "i", ",", "j"}], "]"}], 
   TagBox["\:22c8",
    Identity,
    SyntaxForm->"a*b"], 
   RowBox[{
    UnderscriptBox["lmTupleAux", "N"], "[", 
    RowBox[{"x", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"R", "..."}], "\[RightAngleBracket]"}]}], 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing lmTupleAux 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1479034574", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610102523388122*^9, 3.610102624094787*^9}, 
   3.610102654630619*^9, {3.610102717675195*^9, 3.610102734134739*^9}, {
   3.610102774107618*^9, 3.610102787674383*^9}, {3.610102828757215*^9, 
   3.610102847807423*^9}, {3.610103501700722*^9, 3.610103502434768*^9}, {
   3.610103896627203*^9, 3.610103897591004*^9}, {3.615801463971723*^9, 
   3.615801471689614*^9}, {3.645772814900178*^9, 3.645772825285109*^9}, {
   3.649062183845086*^9, 3.649062184323401*^9}, {3.6495045187400618`*^9, 
   3.649504526521378*^9}, {3.6495663975233297`*^9, 3.649566397675694*^9}},
 CellTags->{
  "ID:1479034574", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lmTupleAux 2"},
 CellID->1479034574],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"cpmTuple", "[", 
    RowBox[{"x", ",", "y"}], "]"}], TraditionalForm]]],
 " returns a tuple of pairs of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"l", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       SubscriptBox["m", "x"], ",", 
       SubscriptBox["m", "y"]}], "\[RightAngleBracket]"}]}], 
    "\[RightAngleBracket]"}], TraditionalForm]]],
 ", where ",
 Cell[BoxData[
  FormBox["l", TraditionalForm]]],
 " is a minimal non-trivial common reducible of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " w.r.t. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"1", ",", "1"}], ")"}], TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "x"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "y"], TraditionalForm]]],
 " are the corresponding critical-pair multipliers. Every minimal non-trivial \
common reducible of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " w.r.t. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"1", ",", "1"}], ")"}], TraditionalForm]]],
 " is contained in the tuple returned by the function. Note that we cannot \
use the function cpm, as for cpm the two indices have to be distinct; \
instead, we use rdm. If ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " are distinct, rdm always returns suitable critical-pair multipliers."
}], "EnvironmentText",
 CellChangeTimes->{{3.649504537187324*^9, 3.6495046681931477`*^9}, {
  3.649504964066737*^9, 3.649504965778262*^9}, {3.649566114325779*^9, 
  3.649566276530155*^9}},
 CellID->1669742674],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["cpmTuple", "N"], "[", 
   RowBox[{"x", ",", "y"}], "]"}], ":=", 
  RowBox[{
   UnderscriptBox["let", 
    RowBox[{"L", "=", 
     RowBox[{
      UnderscriptBox["lcrd", "D"], "[", 
      RowBox[{"x", ",", "y", ",", "1", ",", "1"}], "]"}]}]], 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       SubscriptBox["L", "k"], ",", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         RowBox[{
          UnderscriptBox["rdm", "D"], "[", 
          RowBox[{
           SubscriptBox["L", "k"], ",", "x"}], "]"}], ",", 
         RowBox[{
          UnderscriptBox["rdm", "D"], "[", 
          RowBox[{
           SubscriptBox["L", "k"], ",", "y"}], "]"}]}], 
        "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "L", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing cpmTuple 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1529751135", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6495658651738453`*^9, 3.649565911681417*^9}, 
   3.6495659939134817`*^9},
 CellTags->{
  "ID:1529751135", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing cpmTuple 1"},
 CellID->1529751135],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"cpmTuple", "[", 
    RowBox[{"x", ",", "i", ",", "j"}], "]"}], TraditionalForm]]],
 " returns a tuple of pairs of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"l", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       SubscriptBox["m", "1"], ",", 
       SubscriptBox["m", "2"]}], "\[RightAngleBracket]"}]}], 
    "\[RightAngleBracket]"}], TraditionalForm]]],
 ", where ",
 Cell[BoxData[
  FormBox["l", TraditionalForm]]],
 " is a minimal non-trivial common reducible of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " w.r.t. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"i", ",", "j"}], ")"}], TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "1"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "2"], TraditionalForm]]],
 " are the corresponding critical-pair multipliers. Every minimal non-trivial \
common reducible of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " w.r.t. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"i", ",", "j"}], ")"}], TraditionalForm]]],
 " is contained in the tuple returned by the function."
}], "EnvironmentText",
 CellChangeTimes->{{3.649504537187324*^9, 3.6495046681931477`*^9}, {
  3.649504964066737*^9, 3.649504965778262*^9}, {3.649566124592992*^9, 
  3.6495661607638283`*^9}, {3.649566429026786*^9, 3.649566444500855*^9}},
 CellID->203149419],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["cpmTuple", "N"], "[", 
   RowBox[{"x", ",", "i", ",", "j"}], "]"}], ":=", 
  RowBox[{
   UnderscriptBox["let", 
    RowBox[{"L", "=", 
     RowBox[{
      UnderscriptBox["lcrd", "D"], "[", 
      RowBox[{"x", ",", "x", ",", "i", ",", "j"}], "]"}]}]], 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       SubscriptBox["L", "k"], ",", 
       RowBox[{
        UnderscriptBox["cpm", "D"], "[", 
        RowBox[{"x", ",", "x", ",", 
         SubscriptBox["L", "k"], ",", "i", ",", "j"}], "]"}]}], 
      "\[RightAngleBracket]"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "L", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing cpmTuple 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:551185449", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649504413703558*^9, 3.6495044503530188`*^9}, 
   3.649565996820709*^9, {3.649566392184623*^9, 3.649566414062989*^9}, {
   3.649571408955715*^9, 3.649571409213032*^9}},
 CellTags->{
  "ID:551185449", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing cpmTuple 2"},
 CellID->551185449],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"GB", "[", "C", "]"}], TraditionalForm]]],
 " computes a (not necessarily reduced) Gr\[ODoubleDot]bner basis of tuple ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 "."
}], "EnvironmentText",
 CellChangeTimes->{{3.596193167495689*^9, 3.596193214107766*^9}, 
   3.596193808789274*^9, 3.607771744263031*^9, {3.6457693701370163`*^9, 
   3.6457693709481583`*^9}},
 CellID->1547004803],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["GB", "N"], "[", "C", "]"}], ":=", 
  RowBox[{
   UnderscriptBox["GBAux", "N"], "[", 
   RowBox[{"C", ",", 
    RowBox[{"allPairs", "[", 
     RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], "]"}], 
    ",", "1", ",", "1", ",", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing GB)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1533920470", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645767069467291*^9, 3.6457670738579082`*^9}, {
  3.6457688714983892`*^9, 3.645768871646961*^9}, {3.647690177507332*^9, 
  3.6476902037642193`*^9}, {3.648285834329937*^9, 3.648285834624675*^9}},
 CellTags->{
  "ID:1533920470", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GB"},
 CellID->1533920470],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["GBAux", "N"], "[", 
   RowBox[{"C", ",", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "x", ",", 
    "y", ",", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], "]"}], ":=", 
  "C"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing GBAux 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1976234692", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6457670813647203`*^9, 3.645767095157061*^9}, {
  3.6476901652011557`*^9, 3.6476901906432047`*^9}, {3.647690315152137*^9, 
  3.6476903154502907`*^9}},
 CellTags->{
  "ID:1976234692", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GBAux 1"},
 CellID->1976234692],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["GBAux", "N"], "[", 
   RowBox[{"C", ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"i", ",", "j"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"R", "..."}]}], "\[RightAngleBracket]"}], ",", "x", ",", "y", 
    ",", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], "]"}], ":=", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Piecewise]", GridBox[{
      {
       RowBox[{
        UnderscriptBox["GBAux", "N"], "[", 
        RowBox[{"C", ",", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"R", "..."}], "\[RightAngleBracket]"}], ",", "i", ",", "j", 
         ",", 
         RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], "]"}], 
       "\[DoubleLeftArrow]", 
       RowBox[{
        RowBox[{
         SubscriptBox["C", "i"], "\[Equal]", 
         UnderscriptBox["0", "N"]}], "\[Or]", 
        RowBox[{
         SubscriptBox["C", "j"], "\[Equal]", 
         UnderscriptBox["0", "N"]}]}]},
      {
       RowBox[{
        UnderscriptBox["GBAux", "N"], "[", 
        RowBox[{"C", ",", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"R", "..."}], "\[RightAngleBracket]"}], ",", "i", ",", "j", 
         ",", 
         RowBox[{
          UnderscriptBox["lmTuple", "N"], "[", 
          RowBox[{
           SubscriptBox["C", "i"], ",", 
           SubscriptBox["C", "j"]}], "]"}]}], "]"}], "\[DoubleLeftArrow]", 
       "otherwise"}
     }]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing GBAux 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:144716839", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645767102416054*^9, 3.645767199340426*^9}, {
  3.645767680981659*^9, 3.645767686589417*^9}, {3.645767960014741*^9, 
  3.645767962182386*^9}, {3.645768196393848*^9, 3.6457682095996227`*^9}, {
  3.645773464652727*^9, 3.6457734672292213`*^9}, {3.647690168162436*^9, 
  3.647690168871063*^9}, {3.6476902148893337`*^9, 3.6476902280545197`*^9}, {
  3.648529633677803*^9, 3.648529685239771*^9}, {3.648534050985529*^9, 
  3.648534055134179*^9}, {3.648982006944772*^9, 3.648982007241762*^9}, {
  3.6490675998206453`*^9, 3.649067601371084*^9}},
 CellTags->{
  "ID:144716839", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GBAux 2"},
 CellID->144716839],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["GBAux", "N"], "[", 
   RowBox[{"C", ",", "P", ",", "i", ",", "j", ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"l", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
       "\[RightAngleBracket]"}], ",", 
      RowBox[{"R", "..."}]}], "\[RightAngleBracket]"}]}], "]"}], ":=", 
  RowBox[{"\[Piecewise]", GridBox[{
     {
      RowBox[{
       UnderscriptBox["GBAux", "N"], "[", 
       RowBox[{"C", ",", "P", ",", "i", ",", "j", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"R", "..."}], "\[RightAngleBracket]"}]}], "]"}], 
      "\[DoubleLeftArrow]", 
      RowBox[{
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{
          SubscriptBox["C", "i"], "\[Equal]", 
          SubscriptBox["C", "j"]}], "\[And]", 
         RowBox[{"mi", "\[Equal]", "mj"}]}], 
        TagBox[")",
         "AutoParentheses"]}], "\[Or]", 
       RowBox[{
        UnderscriptBox["chainCrit", "N"], "[", 
        RowBox[{"i", ",", "j", ",", "l", ",", "C", ",", "P"}], "]"}]}]},
     {
      RowBox[{
       UnderscriptBox["let", 
        RowBox[{"h", "=", 
         RowBox[{
          RowBox[{
           UnderscriptBox["trd", "N"], "[", 
           RowBox[{
            RowBox[{"l", 
             UnderscriptBox["-", "N"], 
             RowBox[{"mi", 
              UnderscriptBox["*", "N"], 
              SubscriptBox["C", "i"]}]}], ",", "C"}], "]"}], 
          UnderscriptBox["-", "N"], 
          RowBox[{
           UnderscriptBox["trd", "N"], "[", 
           RowBox[{
            RowBox[{"l", 
             UnderscriptBox["-", "N"], 
             RowBox[{"mj", 
              UnderscriptBox["*", "N"], 
              SubscriptBox["C", "j"]}]}], ",", "C"}], "]"}]}]}]], 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{
            UnderscriptBox["GBAux", "N"], "[", 
            RowBox[{"C", ",", "P", ",", "i", ",", "j", ",", 
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"R", "..."}], "\[RightAngleBracket]"}]}], "]"}], 
           "\[DoubleLeftArrow]", 
           RowBox[{"h", "\[Equal]", 
            UnderscriptBox["0", "N"]}]},
          {
           RowBox[{
            UnderscriptBox["GBAux", "N"], "[", 
            RowBox[{
             RowBox[{"C", 
              TagBox["\:293a",
               Identity,
               SyntaxForm->"a*b"], "h"}], ",", 
             RowBox[{"update", "[", 
              RowBox[{"P", ",", 
               RowBox[{
                RowBox[{
                "\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], "+", 
                "1"}]}], "]"}], ",", "i", ",", "j", ",", 
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"R", "..."}], "\[RightAngleBracket]"}]}], "]"}], 
           "\[DoubleLeftArrow]", "otherwise"}
         }]}]}], "\[DoubleLeftArrow]", "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing GBAux 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:189044945", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6457672438008013`*^9, 3.645767251593217*^9}, {
   3.645767298923203*^9, 3.645767337085554*^9}, {3.645767383118289*^9, 
   3.64576752473984*^9}, {3.645767561504184*^9, 3.6457676085621157`*^9}, {
   3.645768216718992*^9, 3.645768256926362*^9}, {3.645772491177433*^9, 
   3.645772519042301*^9}, {3.645772682158334*^9, 3.6457726848158216`*^9}, {
   3.645772901399315*^9, 3.64577291514465*^9}, {3.64577415647261*^9, 
   3.645774162807465*^9}, {3.647690173744401*^9, 3.647690174265512*^9}, {
   3.6476903233429337`*^9, 3.6476903312785807`*^9}, {3.648290524903431*^9, 
   3.648290542026781*^9}, {3.648290725369234*^9, 3.648290729635234*^9}, {
   3.648292584949109*^9, 3.648292588487661*^9}, {3.648294117809107*^9, 
   3.648294134091977*^9}, 3.6483771988060627`*^9, 3.648443105943869*^9, {
   3.6485292549078093`*^9, 3.648529274813999*^9}, 3.6485296202906437`*^9},
 CellTags->{
  "ID:189044945", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GBAux 3"},
 CellID->189044945],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"chainCrit", "[", 
    RowBox[{"i", ",", "j", ",", "l", ",", "C", ",", "P"}], "]"}], 
   TraditionalForm]]],
 " checks whether the critical situation originating from pair ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{
     SubscriptBox["C", "i"], ",", 
     SubscriptBox["C", "j"]}], ")"}], TraditionalForm]]],
 " and least common reducible ",
 Cell[BoxData[
  FormBox["l", TraditionalForm]]],
 " really has to be considered (i.e. whether its critical pairs have to be \
reduced), using Buchberger\[CloseCurlyQuote]s chain criterion; ",
 Cell[BoxData[
  FormBox["P", TraditionalForm]]],
 " is the tuple of critical pairs that have not been considered yet. The \
predicate holds iff ",
 Cell[BoxData[
  FormBox["l", TraditionalForm]]],
 " can be ignored."
}], "EnvironmentText",
 CellChangeTimes->{{3.59619341190503*^9, 3.596193513888025*^9}, {
  3.596193702354539*^9, 3.596193800148364*^9}, {3.645772932362173*^9, 
  3.645772977043655*^9}, {3.6457735261209517`*^9, 3.645773527923483*^9}, {
  3.64829057840269*^9, 3.64829059189112*^9}},
 CellID->468041515],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["chainCrit", "N"], "[", 
   RowBox[{"i", ",", "j", ",", "l", ",", "C", ",", "P"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"i", "\[NotEqual]", "j"}]},
       {
        RowBox[{"processed", "[", 
         RowBox[{"i", ",", "i", ",", "P"}], "]"}]},
       {
        RowBox[{"processed", "[", 
         RowBox[{"j", ",", "j", ",", "P"}], "]"}]},
       {
        RowBox[{
         UnderscriptBox[
          UnderscriptBox["\[Exists]", 
           RowBox[{
            RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
            RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
          RowBox[{
           RowBox[{"k", "\[NotEqual]", "i"}], "\[And]", 
           RowBox[{"k", "\[NotEqual]", "j"}]}]], 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"\[And]", 
           RowBox[{"\[Piecewise]", GridBox[{
              {
               RowBox[{
                RowBox[{
                 UnderscriptBox["rdm", "N"], "[", 
                 RowBox[{"l", ",", 
                  SubscriptBox["C", "k"]}], "]"}], "\[NotEqual]", 
                UnderscriptBox["0", "N"]}]},
              {
               RowBox[{"processed", "[", 
                RowBox[{"k", ",", "k", ",", "P"}], "]"}]},
              {
               RowBox[{"processed", "[", 
                RowBox[{"i", ",", "k", ",", "P"}], "]"}]},
              {
               RowBox[{"processed", "[", 
                RowBox[{"j", ",", "k", ",", "P"}], "]"}]}
             }]}]}], 
          TagBox[")",
           "AutoParentheses"]}]}]}
      }]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing chainCrit)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1766960499", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594637258737032*^9, 3.59463734033244*^9}, {
  3.59463742967946*^9, 3.594637491222606*^9}, {3.594637647059857*^9, 
  3.594637721458136*^9}, {3.59463878445477*^9, 3.594638846659329*^9}, {
  3.594638883578569*^9, 3.594638905555725*^9}, {3.594639921280334*^9, 
  3.594639940298158*^9}, {3.594640580968024*^9, 3.594640591354008*^9}, {
  3.607340027987263*^9, 3.607340054239634*^9}, {3.610098386985024*^9, 
  3.610098408702971*^9}, {3.645767908954451*^9, 3.645767927962611*^9}, {
  3.645768126742022*^9, 3.645768156546657*^9}, {3.645768314139288*^9, 
  3.645768324311273*^9}, {3.645768374388782*^9, 3.6457683855595818`*^9}, {
  3.645768894440772*^9, 3.645768894921385*^9}, {3.648810151240979*^9, 
  3.648810166370591*^9}, {3.648810568492683*^9, 3.648810568786747*^9}, {
  3.6489555390339823`*^9, 3.648955592951199*^9}},
 CellTags->{
  "ID:1766960499", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing chainCrit"},
 CellID->1766960499],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"ard", "[", "C", "]"}], TraditionalForm]]],
 " returns the \[OpenCurlyDoubleQuote]auto-reduced\[CloseCurlyDoubleQuote] \
tuple of elements in ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ", i.e. the tuple consisting of those elements which cannot be reduced to 0 \
modulo the other elements.\n",
 StyleBox["Attention!",
  FontWeight->"Bold"],
 " Auto-reduction of a Gr\[ODoubleDot]bner basis does ",
 StyleBox["not",
  FontSlant->"Italic"],
 " yield a Gr\[ODoubleDot]bner basis in general; this is only the case in \
special reduction rings (e.g. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"K", "[", "X", "]"}], TraditionalForm]]],
 "), where we have ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"a", 
       SubsuperscriptBox["\[Rule]", "B", "*"], 
       RowBox[{
        RowBox[{"0", "\[And]", "b"}], 
        SubsuperscriptBox["\[Rule]", 
         RowBox[{"B", "\[Union]", 
          RowBox[{"{", "a", "}"}]}], "*"], "0"}]}], ")"}], "\[Implies]", 
     "b"}], 
    SubsuperscriptBox["\[Rule]", "B", "*"], "0"}], TraditionalForm]]],
 " for Gr\[ODoubleDot]bner bases ",
 Cell[BoxData[
  FormBox[
   RowBox[{"B", "\[Union]", 
    RowBox[{"{", "a", "}"}]}], TraditionalForm]]],
 "."
}], "EnvironmentText",
 CellChangeTimes->{{3.645775246133754*^9, 3.645775296561284*^9}, {
  3.64769103118467*^9, 3.6476910327282333`*^9}, {3.649580920821789*^9, 
  3.649581087055869*^9}, {3.6495811346319304`*^9, 3.649581145655768*^9}, {
  3.6495811908389463`*^9, 3.649581245117393*^9}, {3.649581936048156*^9, 
  3.6495819569584713`*^9}, {3.649582120513228*^9, 3.649582137054285*^9}},
 CellID->458982664],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["ard", "N"], "[", "C", "]"}], ":=", 
  RowBox[{
   UnderscriptBox["ardAux", "N"], "[", 
   RowBox[{"C", ",", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing ard)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1777604038", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6457747764512*^9, 3.645774783344411*^9}, {
   3.645774867560499*^9, 3.645774882665203*^9}, {3.645775351498755*^9, 
   3.645775351756864*^9}, 3.645776057148971*^9, {3.649505001733881*^9, 
   3.64950500288865*^9}},
 CellTags->{
  "ID:1777604038", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing ard"},
 CellID->1777604038],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["ardAux", "N"], "[", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "P"}], 
   "]"}], ":=", "P"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing ardAux 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:620529741", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645774892558385*^9, 3.645774898750112*^9}, {
  3.645775348959907*^9, 3.645775392185935*^9}, {3.645775435140626*^9, 
  3.6457754477134447`*^9}, {3.645775547564621*^9, 3.6457755525027647`*^9}, {
  3.645776062490337*^9, 3.645776067873632*^9}, {3.649505005498889*^9, 
  3.649505006423311*^9}},
 CellTags->{
  "ID:620529741", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing ardAux 1"},
 CellID->620529741],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["ardAux", "N"], "[", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"x", ",", 
      RowBox[{"R", "..."}]}], "\[RightAngleBracket]"}], ",", "P"}], "]"}], ":=", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["let", 
     RowBox[{"h", "=", 
      RowBox[{
       UnderscriptBox["trd", "N"], "[", 
       RowBox[{"x", ",", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"R", "..."}], "\[RightAngleBracket]"}], 
         TagBox["\:22c8",
          Identity,
          SyntaxForm->"a*b"], "P"}]}], "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Piecewise]", GridBox[{
        {
         RowBox[{
          UnderscriptBox["ardAux", "N"], "[", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"R", "..."}], "\[RightAngleBracket]"}], ",", "P"}], "]"}],
          "\[DoubleLeftArrow]", 
         RowBox[{"h", "\[Equal]", 
          UnderscriptBox["0", "N"]}]},
        {
         RowBox[{
          UnderscriptBox["ardAux", "N"], "[", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"R", "..."}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"P", 
            TagBox["\:293a",
             Identity,
             SyntaxForm->"a*b"], "x"}]}], "]"}], "\[DoubleLeftArrow]", 
         "otherwise"}
       }]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GroebnerRing ardAux 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:174904624", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645776053209857*^9, 3.645776140140369*^9}, {
  3.649505009004125*^9, 3.649505018981471*^9}},
 CellTags->{
  "ID:174904624", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing ardAux 2"},
 CellID->174904624],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1222673264]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1724139512]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Tests", "Section",
 CellChangeTimes->{{3.645766606382124*^9, 3.6457666072002087`*^9}},
 CellID->963895538],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<Fields.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554650438047533`*^9, 3.6554651094340687`*^9}, {
   3.661601041781954*^9, 3.6616011124344783`*^9}, 3.6666939987524223`*^9, {
   3.66669447260254*^9, 3.6666944735453367`*^9}},
 CellID->1816127517],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<Integers.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554650438047533`*^9, 3.6554651094340687`*^9}, {
   3.661601041781954*^9, 3.6616011124344783`*^9}, 3.6666939987524223`*^9, {
   3.66669447260254*^9, 3.666694479705483*^9}},
 CellID->1873029144],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<IntegerQuotientRings.m\>\""}], 
    "}"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.6554650438047533`*^9, 3.6554651094340687`*^9}, {
   3.661601041781954*^9, 3.6616011124344783`*^9}, 3.6666939987524223`*^9, {
   3.66669447260254*^9, 3.666694487098187*^9}},
 CellID->1110067096],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<PolyTuples.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554650438047533`*^9, 3.6554651094340687`*^9}, {
   3.661601041781954*^9, 3.6616011124344783`*^9}, 3.6666939987524223`*^9, {
   3.66669447260254*^9, 3.666694491993288*^9}},
 CellID->1197775081],

Cell[BoxData[
 RowBox[{"Get", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"NotebookDirectory", "[", "]"}], ",", "\"\<PolyIO.m\>\""}], 
    "}"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.666695109147286*^9, 3.666695121717403*^9}},
 CellID->851552697],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"GenerateComputationRules", "[", "\"\<active\>\"", "]"}]], "Input",
 CellChangeTimes->{{3.666694610616609*^9, 3.6666946168419743`*^9}},
 CellID->430172879],

Cell[BoxData[
 InterpretationBox[
  RowBox[{
  "125", "\[InvisibleSpace]", "\<\" formulas turned into computation rules.\"\
\>"}],
  SequenceForm[125, " formulas turned into computation rules."],
  Editable->False]], "Print",
 CellChangeTimes->{{3.666694617736566*^9, 3.666694647409486*^9}, {
   3.66669472010032*^9, 3.66669474367059*^9}, 3.6666947918285513`*^9, 
   3.666694891689855*^9, 3.6666952722586193`*^9, 3.666695462843245*^9, 
   3.6669355496230927`*^9, 3.6669364724942827`*^9, 3.666936599794191*^9, 
   3.66693697705588*^9, 3.666937063342227*^9, 3.6670436135774403`*^9},
 CellID->1649399640]
}, Open  ]],

Cell[CellGroupData[{

Cell["\[DoubleStruckCapitalQ]", "Subsection",
 CellChangeTimes->{{3.592043465736425*^9, 3.592043466408566*^9}, {
  3.594015855609246*^9, 3.594015856896769*^9}},
 CellID->553055740],

Cell["", "OpenEnvironment",
 CellID->572285523],

Cell[CellGroupData[{

Cell["\<\
Definition: Reduction Field and Gr\[ODoubleDot]bner Ring of \
\[DoubleStruckCapitalQ]\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.593493077646172*^9, 3.593493094508568*^9}, {
  3.593493266364826*^9, 3.593493267367732*^9}},
 CellID->2012283505],

Cell[BoxData[
 RowBox[{"\[ScriptCapitalG]\[ScriptCapitalQ]", ":=", 
  RowBox[{
  "GroebnerRing", "[", "\[ScriptCapitalQ]", "]"}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalG] \[ScriptCapitalQ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:336511876", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.593493324864282*^9, 3.593493359987526*^9}, {
  3.607340386629562*^9, 3.60734038758986*^9}, {3.645781230205714*^9, 
  3.645781234933929*^9}, {3.645781581001783*^9, 3.645781583111521*^9}},
 CellTags->{
  "ID:336511876", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalQ]"},
 CellID->336511876],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1595859076]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->914612762],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["rdm", "\[ScriptCapitalG]\[ScriptCapitalQ]"], "[", 
  RowBox[{
   FractionBox["7", "2"], ",", 
   RowBox[{"-", "4"}]}], "]"}]], "Computation",
 CellChangeTimes->{{3.5934934160813*^9, 3.593493439228454*^9}, {
   3.59349506414043*^9, 3.593495065470676*^9}, {3.593495632839658*^9, 
   3.593495641131288*^9}, {3.593757585815953*^9, 3.593757587065053*^9}, {
   3.593764969117268*^9, 3.593764980482681*^9}, {3.593765759702713*^9, 
   3.593765763068752*^9}, {3.6457736499585657`*^9, 3.6457736514740057`*^9}, 
   3.645781317332996*^9, 3.645781586496162*^9},
 CellID->460713032],

Cell[BoxData[
 FormBox[
  RowBox[{"-", 
   FractionBox["7", "8"]}], TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.607340445027412*^9, 3.610103443429014*^9, 3.61580161131121*^9, 
   3.615801760444767*^9, {3.631449351286912*^9, 3.631449357499194*^9}, {
   3.645773639815812*^9, 3.645773652265807*^9}, 3.6457737368285713`*^9, 
   3.645775086726378*^9, 3.645775605651773*^9, 3.645776480411322*^9, 
   3.645781270432805*^9, 3.6457813186500483`*^9, {3.666694623964028*^9, 
   3.666694650542091*^9}, 3.666935476980575*^9, {3.6669355170623007`*^9, 
   3.6669355540130177`*^9}},
 CellID->1219447368]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c460713032"]]]]}, {None, None}},
 CellChangeTimes->{3.666935555105481*^9},
 CellID->1022441481],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c460713032-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {106., 112.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666935555378112*^9},
 CellID->1436640939]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalG]\[ScriptCapitalQ]"], "[", 
  RowBox[{
   FractionBox["7", "2"], ",", 
   RowBox[{"-", "4"}], ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.593765425391834*^9, 3.593765449142532*^9}, {
   3.610103447744197*^9, 3.610103448071136*^9}, {3.645773654932273*^9, 
   3.64577365626258*^9}, 3.645781321301705*^9, 3.6457815883783712`*^9},
 CellID->1058478970],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", "1", "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.607340449408249*^9, 3.610103449569515*^9, 3.615801614128419*^9, 
   3.615801762367444*^9, 3.6314493627271013`*^9, {3.645773643683654*^9, 
   3.645773657020623*^9}, 3.645773738703012*^9, 3.645775088641365*^9, 
   3.645775608518175*^9, 3.645776482247532*^9, 3.645781272731164*^9, 
   3.645781322513425*^9, 3.666694654539253*^9, 3.666935714062541*^9},
 CellID->1917317166]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1058478970"]]]]}, {None, None}},
 CellChangeTimes->{3.66693571504642*^9},
 CellID->2127071249],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1058478970-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {106., 112.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6669357153126173`*^9},
 CellID->1965686465]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["cpm", "\[ScriptCapitalG]\[ScriptCapitalQ]"], "[", 
  RowBox[{
   FractionBox["7", "2"], ",", 
   RowBox[{"-", "4"}], ",", "1", ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.59446030295423*^9, 3.594460320836334*^9}, {
   3.6101034522213*^9, 3.610103452560434*^9}, {3.645773659578863*^9, 
   3.645773660940794*^9}, 3.645781325041388*^9, 3.645781589898673*^9},
 CellID->1725275638],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    FractionBox["2", "7"], ",", 
    RowBox[{"-", 
     FractionBox["1", "4"]}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.607340452582168*^9, 3.610103454727458*^9, 3.615801616809471*^9, 
   3.61580176421647*^9, 3.631449366421712*^9, 3.645773661624929*^9, {
   3.645773740842984*^9, 3.645773775118881*^9}, 3.645775090706311*^9, 
   3.645781274993313*^9, 3.645781326239513*^9, 3.666694658235365*^9, 
   3.666935717502914*^9},
 CellID->1471124563]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1725275638"]]]]}, {None, None}},
 CellChangeTimes->{3.666935718523868*^9},
 CellID->1462783354],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1725275638-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {106., 112.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666935718782543*^9},
 CellID->31606615]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lmTuple", "\[ScriptCapitalG]\[ScriptCapitalQ]"], "[", 
  RowBox[{"4", ",", 
   FractionBox["3", "2"]}], "]"}]], "Computation",
 CellChangeTimes->{{3.63144937241086*^9, 3.631449373047841*^9}, {
   3.645773789032194*^9, 3.645773790691318*^9}, 3.645781329277317*^9, 
   3.6457815917744513`*^9},
 CellID->681863368],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"1", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       FractionBox["1", "4"], ",", 
       FractionBox["2", "3"]}], "\[RightAngleBracket]"}]}], 
    "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{{3.631449373888969*^9, 3.6314493772315397`*^9}, 
   3.6457737992659492`*^9, 3.645773839491849*^9, 3.6457750927687187`*^9, 
   3.6457812778312073`*^9, 3.6457813305495567`*^9, 3.666694664061652*^9, 
   3.666935720405705*^9},
 CellID->2046403607]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c681863368"]]]]}, {None, None}},
 CellChangeTimes->{3.666935721392426*^9},
 CellID->1100457786],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c681863368-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {106., 112.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666935721656023*^9},
 CellID->1053359543]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalQ]"], "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    FractionBox["7", "2"], ",", 
    RowBox[{"-", "4"}], ",", 
    FractionBox["1", "17946"]}], "\[RightAngleBracket]"}], 
  "]"}]], "Computation",
 CellChangeTimes->{{3.593494911230733*^9, 3.593494956631068*^9}, 
   3.645781333183577*^9, 3.64578159321472*^9},
 CellID->380754896],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    FractionBox["7", "2"], ",", 
    RowBox[{"-", "4"}], ",", 
    FractionBox["1", "17946"]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{{3.610103578744948*^9, 3.610103603008846*^9}, 
   3.61010366757265*^9, 3.610103705314242*^9, 3.615801621890438*^9, 
   3.615801769904948*^9, 3.631449386211639*^9, 3.6457738506365843`*^9, 
   3.645775611594078*^9, 3.64578128027299*^9, 3.6457813344507227`*^9, 
   3.666694668842505*^9},
 CellID->695736833]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c380754896"]]]]}, {None, None}},
 CellChangeTimes->{3.666694669781911*^9},
 CellID->1874687287],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c380754896-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {106., 112.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6666946700327663`*^9},
 CellID->623992335]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["ard", "\[ScriptCapitalG]\[ScriptCapitalQ]"], "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    FractionBox["7", "2"], ",", 
    RowBox[{"-", "4"}], ",", 
    FractionBox["1", "17946"]}], "\[RightAngleBracket]"}], 
  "]"}]], "Computation",
 CellChangeTimes->{{3.645775039507814*^9, 3.645775052489315*^9}, 
   3.645781337097643*^9, 3.645781594799958*^9},
 CellID->1319269262],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   FractionBox["1", "17946"], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.64577509534718*^9, 3.64577562031009*^9, 
  3.645776485385104*^9, 3.645781282688081*^9, 3.645781338406003*^9, 
  3.666694672436157*^9, 3.666935723871339*^9},
 CellID->815793440]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1319269262"]]]]}, {None, None}},
 CellChangeTimes->{3.6669357248836117`*^9},
 CellID->537945533],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1319269262-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {106., 112.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666935725134572*^9},
 CellID->1252991721]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\[DoubleStruckCapitalZ]", "Subsection",
 CellChangeTimes->{{3.594023410029535*^9, 3.594023412370074*^9}},
 CellID->1783762617],

Cell["", "OpenEnvironment",
 CellID->413357741],

Cell[CellGroupData[{

Cell[TextData[{
 "Definition: Gr\[ODoubleDot]bner Ring of \[DoubleStruckCapitalZ] (or ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[ScriptCapitalR]", "\[DoubleStruckCapitalZ]"], 
   TraditionalForm]]],
 ")"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.59402342779568*^9, 3.594023449386866*^9}},
 CellID->2097093576],

Cell[BoxData[
 RowBox[{"\[ScriptCapitalG]\[ScriptCapitalZ]", ":=", 
  RowBox[{
  "GroebnerRing", "[", "\[ScriptCapitalZ]", "]"}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalG] \[ScriptCapitalZ])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:193404746", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594023456116466*^9, 3.594023479354317*^9}, {
  3.6457812912819157`*^9, 3.6457812960803556`*^9}, {3.6457816004436283`*^9, 
  3.6457816025110407`*^9}},
 CellTags->{
  "ID:193404746", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]"},
 CellID->193404746],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->953586710]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1137065047],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lmTuple", "\[ScriptCapitalG]\[ScriptCapitalZ]"], "[", 
  RowBox[{"6", ",", 
   RowBox[{"-", "10"}]}], "]"}]], "Computation",
 CellChangeTimes->{{3.645773979613538*^9, 3.6457739822561703`*^9}, 
   3.645781347778739*^9, 3.6457816071499662`*^9},
 CellID->1290937647],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"5", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"1", ",", 
       RowBox[{"-", "1"}]}], "\[RightAngleBracket]"}]}], 
    "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.6457739841028337`*^9, 3.6457751450670137`*^9, 3.645775486065033*^9, 
   3.645776525388197*^9, {3.6457794377430477`*^9, 3.645779450366433*^9}, {
   3.645781394988924*^9, 3.645781419244809*^9}, {3.666694724641418*^9, 
   3.666694745905456*^9}, 3.666694794449294*^9},
 CellID->1065167918]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1290937647"]]]]}, {None, None}},
 CellChangeTimes->{3.6666947955005493`*^9},
 CellID->745975735],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1290937647-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {106., 112.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666694795772977*^9},
 CellID->1868486125]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lmTuple", "\[ScriptCapitalG]\[ScriptCapitalZ]"], "[", 
  RowBox[{"6", ",", "4"}], "]"}]], "Computation",
 CellChangeTimes->{3.645774000288472*^9, 3.645781350243997*^9, 
  3.6457816087726383`*^9},
 CellID->34460033],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"3", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"1", ",", "1"}], "\[RightAngleBracket]"}]}], 
    "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645774000893763*^9, 3.645775147274693*^9, 
  3.645779452885704*^9, 3.645781422626532*^9, 3.666694798740707*^9},
 CellID->2084155435]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c34460033"]]]]}, {None, None}},
 CellChangeTimes->{3.666694799791038*^9},
 CellID->813692243],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c34460033-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {106., 112.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6666948000476227`*^9},
 CellID->2017922153]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lmTuple", "\[ScriptCapitalG]\[ScriptCapitalZ]"], "[", 
  RowBox[{"6", ",", 
   RowBox[{"-", "32"}]}], "]"}]], "Computation",
 CellChangeTimes->{{3.645774029231339*^9, 3.645774029749236*^9}, 
   3.645781351680764*^9, 3.645781610099621*^9},
 CellID->210527065],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"16", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"3", ",", 
       RowBox[{"-", "1"}]}], "\[RightAngleBracket]"}]}], 
    "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645774030206378*^9, 3.645775149338443*^9, 
  3.645779456034095*^9, 3.645781425829186*^9, 3.666694801811654*^9},
 CellID->233209813]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c210527065"]]]]}, {None, None}},
 CellChangeTimes->{3.666694803039237*^9},
 CellID->2009197914],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c210527065-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {106., 112.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666694803299492*^9},
 CellID->1673135784]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalZ]"], "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"6", ",", 
    RowBox[{"-", "10"}], ",", "4", ",", 
    RowBox[{"-", "32"}]}], "\[RightAngleBracket]"}], "]"}]], "Computation",
 CellChangeTimes->{{3.594023507196046*^9, 3.594023566236402*^9}, {
   3.594036519284828*^9, 3.594036519431344*^9}, {3.5941030535863*^9, 
   3.594103062197513*^9}, 3.645781369874148*^9, 3.64578161223849*^9},
 CellID->2004775239],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"6", ",", 
    RowBox[{"-", "10"}], ",", "4", ",", 
    RowBox[{"-", "32"}], ",", 
    RowBox[{"-", "2"}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.607340685033291*^9, 3.607340748715928*^9, 
  3.610103780129468*^9, 3.615801818336378*^9, 3.631449440331676*^9, 
  3.645774071968647*^9, 3.645775159467504*^9, 3.645781429114352*^9, 
  3.666694808059627*^9},
 CellID->2097566239]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c2004775239"]]]]}, {None, None}},
 CellChangeTimes->{3.6666948090812807`*^9},
 CellID->1339960540],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c2004775239-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {106., 112.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666694809337407*^9},
 CellID->136544494]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["ard", "\[ScriptCapitalG]\[ScriptCapitalZ]"], "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"6", ",", 
    RowBox[{"-", "10"}], ",", "4", ",", 
    RowBox[{"-", "32"}], ",", "2"}], "\[RightAngleBracket]"}], 
  "]"}]], "Computation",
 CellChangeTimes->{{3.645775176377749*^9, 3.64577518836729*^9}, 
   3.6457813714597683`*^9, 3.645781613897146*^9},
 CellID->314892606],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", "2", "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645775189715773*^9, 3.645775489950197*^9, 
  3.64577556561998*^9, 3.645776529359284*^9, 3.6457794662153883`*^9, 
  3.6457814324722633`*^9, 3.666694811238172*^9},
 CellID->1102276527]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c314892606"]]]]}, {None, None}},
 CellChangeTimes->{3.666694812462551*^9},
 CellID->825370302],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c314892606-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {106., 112.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6666948127214003`*^9},
 CellID->1666078827]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalZ]"], "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"2091", ",", "2337", ",", "2829"}], "\[RightAngleBracket]"}], 
  "]"}]], "Computation",
 CellChangeTimes->{3.645781372823606*^9, 3.6457816153701067`*^9},
 CellID->1654602455],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
   "2091", ",", "2337", ",", "2829", ",", "1845", ",", "1353", ",", "1107", 
    ",", "246", ",", "123"}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645774107376429*^9, 3.645774175693429*^9, 
  3.645781436438002*^9, 3.666694825634289*^9},
 CellID->1491910187]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1654602455"]]]]}, {None, None}},
 CellChangeTimes->{3.666694826661265*^9},
 CellID->1152043447],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1654602455-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {106., 112.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{{3.666694826923242*^9, 3.6666948307728777`*^9}},
 CellID->643984759]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["ard", "\[ScriptCapitalG]\[ScriptCapitalZ]"], "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
   "2091", ",", "2337", ",", "2829", ",", "1845", ",", "1599", ",", "1353", 
    ",", "1107", ",", "861", ",", "615", ",", "369", ",", "123"}], 
   "\[RightAngleBracket]"}], "]"}]], "Computation",
 CellChangeTimes->{{3.645774314433775*^9, 3.645774337117379*^9}, {
   3.645775196787771*^9, 3.6457752042998323`*^9}, 3.645781374411418*^9, 
   3.6457816165999403`*^9},
 CellID->123561440],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", "123", "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645774337494687*^9, 3.645775205690734*^9, 
  3.6457755690092583`*^9, 3.6457765318896923`*^9, 3.645779469304899*^9, 
  3.645781439361951*^9, 3.666694833889835*^9},
 CellID->355642102]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c123561440"]]]]}, {None, None}},
 CellChangeTimes->{3.666694834935936*^9},
 CellID->1426675444],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c123561440-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {106., 112.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6666948352027273`*^9},
 CellID->657498851]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 FormBox[
  SubscriptBox["\[DoubleStruckCapitalZ]", "24"], 
  TraditionalForm]]]], "Subsection",
 CellChangeTimes->{{3.594111051425592*^9, 3.594111054968721*^9}},
 CellID->1940243944],

Cell["", "OpenEnvironment",
 CellID->921750930],

Cell[CellGroupData[{

Cell[TextData[{
 "Definition: Gr\[ODoubleDot]bner Ring of ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalZ]", "24"], TraditionalForm]]],
 " (or ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[ScriptCapitalR]", "24"], TraditionalForm]]],
 ")"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.594111073532906*^9, 3.594111090932806*^9}},
 CellID->369221017],

Cell[BoxData[
 RowBox[{"\[ScriptCapitalG]\[ScriptCapitalZ]24", ":=", 
  RowBox[{
  "GroebnerRing", "[", "\[ScriptCapitalZ]24", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalG] \[ScriptCapitalZ]24)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1821700886", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594111094953551*^9, 3.594111110109059*^9}, {
  3.6457816253599367`*^9, 3.6457816333657103`*^9}},
 CellTags->{
  "ID:1821700886", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]24"},
 CellID->1821700886],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1831110154]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1312555615],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalZ]24"], "[", 
  RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
  "]"}]], "Computation",
 CellChangeTimes->{{3.594111133265843*^9, 3.594111142747825*^9}, 
   3.645781654841292*^9},
 CellID->210353575],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.607760809348652*^9, 3.608006279594525*^9, 
  3.610103826574468*^9, 3.615801873858963*^9, 3.6314496303681097`*^9, 
  3.6457766071959953`*^9, 3.645776678686323*^9, 3.645781686615196*^9, 
  3.666694896017622*^9},
 CellID->1123616661]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c210353575"]]]]}, {None, None}},
 CellChangeTimes->{3.666694897179661*^9},
 CellID->1279883406],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c210353575-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {113., 120.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6666948974633293`*^9},
 CellID->320835099]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalZ]24"], "[", 
  RowBox[{"\[LeftAngleBracket]", "6", "\[RightAngleBracket]"}], 
  "]"}]], "Computation",
 CellChangeTimes->{{3.607760813282549*^9, 3.607760824745377*^9}, {
   3.607762211021081*^9, 3.607762214230755*^9}, 3.645781656543898*^9},
 CellID->2041351713],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", "6", "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.607760825571718*^9, {3.607762211567405*^9, 3.607762214759021*^9}, 
   3.608006282006695*^9, 3.61010382895304*^9, 3.610103907638524*^9, 
   3.615801876018272*^9, 3.631449634053815*^9, 3.645776681991857*^9, 
   3.6457816893874187`*^9, {3.666694899903675*^9, 3.666694918411083*^9}},
 CellID->1960744693]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c2041351713"]]]]}, {None, None}},
 CellChangeTimes->{3.666694919593152*^9},
 CellID->955142166],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c2041351713-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {113., 120.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666694919882516*^9},
 CellID->1202136422]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalZ]24"], "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"6", ",", "8"}], "\[RightAngleBracket]"}], "]"}]], "Computation",
 CellChangeTimes->{{3.607760920541331*^9, 3.607760931863272*^9}, 
   3.607762188725419*^9, {3.607762460509647*^9, 3.607762460845677*^9}, 
   3.645781658019553*^9},
 CellID->1718024098],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"6", ",", "8", ",", "2"}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.60776093339226*^9, 3.60776193304691*^9, 
  3.607762189905469*^9, 3.607762462093354*^9, 3.608006284554496*^9, 
  3.61010391111485*^9, 3.615801879047486*^9, 3.631449638660005*^9, 
  3.645776690641706*^9, 3.645781692452422*^9, 3.666694922750743*^9},
 CellID->1268604381]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1718024098"]]]]}, {None, None}},
 CellChangeTimes->{3.666694923916999*^9},
 CellID->147506085],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1718024098-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {113., 120.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666694924196575*^9},
 CellID->795415407]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["ard", "\[ScriptCapitalG]\[ScriptCapitalZ]24"], "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"6", ",", "8", ",", "2"}], "\[RightAngleBracket]"}], 
  "]"}]], "Computation",
 CellChangeTimes->{{3.64577670013833*^9, 3.645776712601969*^9}, 
   3.645781659428747*^9},
 CellID->1660233602],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", "2", "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645776713659679*^9, 3.6457816953661747`*^9, 
  3.66669492581355*^9},
 CellID->1236673481]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1660233602"]]]]}, {None, None}},
 CellChangeTimes->{3.6666949270644493`*^9},
 CellID->1531881759],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1660233602-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {113., 120.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6666949273539343`*^9},
 CellID->1891696089]
}, Closed]],

Cell["\<\
Note that new elements might be greater than the old ones (w.r.t. the \
Noetherian ordering):\
\>", "Text",
 CellChangeTimes->{{3.607762486472984*^9, 3.607762506965752*^9}, {
  3.645781705219055*^9, 3.645781711706991*^9}},
 CellID->162878423],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalZ]24"], "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"8", ",", "6"}], "\[RightAngleBracket]"}], "]"}]], "Computation",
 CellChangeTimes->{{3.60776248059803*^9, 3.607762480860373*^9}, 
   3.645781661382963*^9},
 CellID->561089538],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"8", ",", "6", ",", "22"}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.607762481980799*^9, 3.608006288644174*^9, 
  3.61010391667905*^9, 3.615801882928908*^9, 3.631449642078402*^9, 
  3.6457767186798153`*^9, 3.645781699345311*^9, 3.666694931011435*^9},
 CellID->1487225701]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c561089538"]]]]}, {None, None}},
 CellChangeTimes->{3.6666949322570868`*^9},
 CellID->1246547808],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c561089538-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {113., 120.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666694932540453*^9},
 CellID->553142869]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["ard", "\[ScriptCapitalG]\[ScriptCapitalZ]24"], "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"8", ",", "6", ",", "22"}], "\[RightAngleBracket]"}], 
  "]"}]], "Computation",
 CellChangeTimes->{{3.645776732867635*^9, 3.645776735330373*^9}, 
   3.645781662637567*^9},
 CellID->372196881],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", "22", "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.6457767361350803`*^9, 3.64578170203273*^9, 
  3.6666949357625637`*^9},
 CellID->1782329282]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c372196881"]]]]}, {None, None}},
 CellChangeTimes->{3.6666949369502573`*^9},
 CellID->1465396252],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c372196881-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {113., 120.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666694937233564*^9},
 CellID->1202561127]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalZ]24"], "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"10", ",", "9"}], "\[RightAngleBracket]"}], "]"}]], "Computation",
 CellChangeTimes->{{3.607762048065322*^9, 3.607762082136002*^9}, {
   3.607762126073396*^9, 3.607762136632602*^9}, 3.645781664148416*^9},
 CellID->1596303038],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"10", ",", "9", ",", "1"}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{{3.607762048980089*^9, 3.60776205226617*^9}, 
   3.607762083002282*^9, {3.607762127443899*^9, 3.607762137932911*^9}, 
   3.6080062916398*^9, 3.610103920327193*^9, 3.615801886075713*^9, 
   3.6314496463415413`*^9, 3.645776757814369*^9, 3.6457817216392117`*^9, 
   3.666694939790051*^9},
 CellID->96031009]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1596303038"]]]]}, {None, None}},
 CellChangeTimes->{3.666694941030613*^9},
 CellID->1550093037],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1596303038-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {113., 120.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6666949413196163`*^9},
 CellID->1545764092]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["ard", "\[ScriptCapitalG]\[ScriptCapitalZ]24"], "[", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"10", ",", "9", ",", "1"}], "\[RightAngleBracket]"}], 
  "]"}]], "Computation",
 CellChangeTimes->{{3.645776773213105*^9, 3.645776774031245*^9}, 
   3.6457816654025993`*^9},
 CellID->364287853],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", "1", "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645776774958434*^9, 3.645781724473379*^9, 
  3.666694942917618*^9},
 CellID->289900299]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c364287853"]]]]}, {None, None}},
 CellChangeTimes->{3.666694944065542*^9},
 CellID->1727929368],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c364287853-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {113., 120.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6666949443479843`*^9},
 CellID->1122644056]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{"\[DoubleStruckCapitalQ]", "[", 
   RowBox[{"x", ",", "y", ",", "z"}], "]"}], TraditionalForm]]]], "Subsection",\

 CellChangeTimes->{{3.594466512409256*^9, 3.594466525503822*^9}, {
  3.645779785238605*^9, 3.645779786002248*^9}},
 CellID->438160971],

Cell["", "OpenEnvironment",
 CellID->1899836489],

Cell[CellGroupData[{

Cell[TextData[{
 "Definition: Gr\[ODoubleDot]bner Ring of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[DoubleStruckCapitalQ]", "[", 
    RowBox[{"x", ",", "y", ",", "z"}], "]"}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.59446653772576*^9, 3.594466560566009*^9}, {
  3.64577979753371*^9, 3.645779802209031*^9}},
 CellID->418711622],

Cell[BoxData[
 RowBox[{"\[ScriptCapitalG]\[ScriptCapitalQ]xyz", ":=", 
  RowBox[{
  "GroebnerRing", "[", "\[ScriptCapitalQ]xyz", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalG] \[ScriptCapitalQ]xy)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1158015694", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594466566856977*^9, 3.594466584620404*^9}, {
  3.607340983575106*^9, 3.607340984604111*^9}, {3.6457769137875347`*^9, 
  3.645776922133088*^9}, {3.645777105715609*^9, 3.645777112102729*^9}, {
  3.6457810435551157`*^9, 3.645781048969571*^9}, {3.645781127837743*^9, 
  3.645781129162697*^9}, {3.6457815516201143`*^9, 3.645781552702067*^9}},
 CellTags->{
  "ID:1158015694", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalQ]xy"},
 CellID->1158015694],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1667815660]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1792593212],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"poly1", "=", 
  RowBox[{
   UnderscriptBox["<<", "\[ScriptCapitalQ]xyz"], 
   RowBox[{"[", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"7", "x", " ", "y", " ", "z"}], "-", 
      RowBox[{
       FractionBox["2", "3"], 
       SuperscriptBox["x", "2"]}], "+", 
      RowBox[{
       FractionBox["3", "8"], 
       SuperscriptBox["z", "2"]}], "-", "y", "-", "4"}], 
     TagBox[")",
      "AutoParentheses"]}], "]"}]}]}]], "Computation",
 CellChangeTimes->{{3.645781086345557*^9, 3.6457810924293413`*^9}, 
   3.645781130896298*^9},
 CellID->749980045],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"7", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"1", ",", "1", ",", "1"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"-", 
       FractionBox["2", "3"]}], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"2", ",", "0", ",", "0"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      FractionBox["3", "8"], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"0", ",", "0", ",", "2"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"0", ",", "1", ",", "0"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"-", "4"}], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"0", ",", "0", ",", "0"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.666695280834413*^9, 3.666695471726309*^9, 
  3.666936478406272*^9},
 CellID->683271910]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c749980045"]]]]}, {None, None}},
 CellChangeTimes->{3.6669364801509857`*^9},
 CellID->588655274],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c749980045-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {211., 217.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6669364808037167`*^9},
 CellID->1614854082]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalQ]xyz"], 
  RowBox[{"[", "poly1", "]"}]}]], "Computation",
 CellChangeTimes->{{3.6666953916117773`*^9, 3.666695401245496*^9}},
 CellID->444957564],

Cell[BoxData[
 FormBox["True", TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.666695403719852*^9, {3.666695466567841*^9, 3.666695484175748*^9}, 
   3.666936483165592*^9},
 CellID->1389037508]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c444957564"]]]]}, {None, None}},
 CellChangeTimes->{3.666936484848777*^9},
 CellID->101779177],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c444957564-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {211., 217.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666936485494933*^9},
 CellID->882684762]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"poly2", "=", 
  RowBox[{
   UnderscriptBox["<<", "\[ScriptCapitalQ]xyz"], 
   RowBox[{"[", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"3", 
       SuperscriptBox["x", "5"], 
       SuperscriptBox["z", "3"]}], "-", 
      RowBox[{"x", " ", 
       SuperscriptBox["y", "3"], 
       SuperscriptBox["z", "3"]}], "+", 
      RowBox[{
       FractionBox["4", "5"], 
       SuperscriptBox["x", "3"], "y", " ", 
       SuperscriptBox["z", "2"]}], "-", 
      RowBox[{"3", "x", " ", "y", " ", "z"}], "+", 
      RowBox[{
       FractionBox["1", "2"], 
       SuperscriptBox["z", "2"]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "]"}]}]}]], "Computation",
 CellChangeTimes->{3.6457811411311893`*^9},
 CellID->84261983],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"5", ",", "0", ",", "3"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"1", ",", "3", ",", "3"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      FractionBox["4", "5"], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"3", ",", "1", ",", "2"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"-", "3"}], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"1", ",", "1", ",", "1"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      FractionBox["1", "2"], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"0", ",", "0", ",", "2"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645777017322402*^9, 3.6457772115455313`*^9, 
  3.64578114255899*^9, 3.666695495040433*^9, 3.6669365284312897`*^9},
 CellID->1405364557]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c84261983"]]]]}, {None, None}},
 CellChangeTimes->{3.666936530135832*^9},
 CellID->1568740038],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c84261983-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {211., 217.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666936530780168*^9},
 CellID->630430316]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"poly3", "=", 
  RowBox[{
   UnderscriptBox["<<", "\[ScriptCapitalQ]xyz"], 
   RowBox[{"[", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"3", 
       SuperscriptBox["x", "2"], "z"}], "-", "y"}], 
     TagBox[")",
      "AutoParentheses"]}], "]"}]}]}]], "Computation",
 CellChangeTimes->{3.645781145476644*^9},
 CellID->1496730429],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"2", ",", "0", ",", "1"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"0", ",", "1", ",", "0"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645777021652474*^9, 3.645777209245468*^9, 
  3.645781146882132*^9, 3.666936535808527*^9},
 CellID->1282426654]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1496730429"]]]]}, {None, None}},
 CellChangeTimes->{3.6669365376273518`*^9},
 CellID->2059153055],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1496730429-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {211., 217.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666936538281904*^9},
 CellID->1534939845]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalQ]xyz", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["lcrd", "\[ScriptCapitalG]\[ScriptCapitalQ]xyz"], "[", 
    RowBox[{"poly1", ",", "poly2", ",", "1", ",", "1"}], "]"}], 
   "]"}]}]], "Computation",
 CellChangeTimes->{{3.645777033969936*^9, 3.645777063270175*^9}, {
   3.64577711832686*^9, 3.645777120808913*^9}, {3.6457771866000357`*^9, 
   3.6457771912859697`*^9}, {3.6457811513137608`*^9, 3.645781176296383*^9}, 
   3.645781560723365*^9},
 CellID->1444448452],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    SuperscriptBox["x", "5"], " ", "y", " ", 
    SuperscriptBox["z", "3"]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.6457770648637943`*^9, {3.645777121693622*^9, 3.645777161320527*^9}, {
   3.6457771919891644`*^9, 3.645777206265705*^9}, 3.645781178304019*^9, 
   3.666936547643065*^9, {3.666936590742138*^9, 3.666936602794621*^9}},
 CellID->1031808536]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1444448452"]]]]}, {None, None}},
 CellChangeTimes->{3.666936605050161*^9},
 CellID->948688360],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1444448452-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {211., 217.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.66693660575799*^9},
 CellID->21100703]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalQ]xyz", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["cpm", "\[ScriptCapitalG]\[ScriptCapitalQ]xyz"], "[", 
    RowBox[{"poly1", ",", 
     UnderscriptBox["1", "\[ScriptCapitalQ]xyz"], ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"1", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"1", ",", "1", ",", "1"}], "\[RightAngleBracket]"}]}], 
       "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], ",", "1", ",", 
     "1"}], "]"}], "]"}]}]], "Computation",
 CellChangeTimes->{{3.645777235467147*^9, 3.645777271706874*^9}, {
   3.645781153008976*^9, 3.645781181233642*^9}, 3.645781562384179*^9},
 CellID->563077933],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    FractionBox["1", "7"], ",", 
    RowBox[{"x", " ", "y", " ", "z"}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{{3.64577724733821*^9, 3.645777272479014*^9}, 
   3.645781183095936*^9, 3.666936613247848*^9},
 CellID->1664994205]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c563077933"]]]]}, {None, None}},
 CellChangeTimes->{3.666936615270176*^9},
 CellID->1036358125],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c563077933-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {211., 217.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6669366159783287`*^9},
 CellID->1303560488]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", "\[ScriptCapitalQ]xyz"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["trd", "\[ScriptCapitalG]\[ScriptCapitalQ]xyz"], "[", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          FractionBox["2", "21"], 
          TagBox[")",
           "AutoParentheses"]}], ",", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"2", ",", "0", ",", "0"}], "\[RightAngleBracket]"}]}], 
        "\[RightAngleBracket]"}], ",", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"-", 
           FractionBox["3", "56"]}], 
          TagBox[")",
           "AutoParentheses"]}], ",", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"0", ",", "0", ",", "2"}], "\[RightAngleBracket]"}]}], 
        "\[RightAngleBracket]"}], ",", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          FractionBox["1", "7"], 
          TagBox[")",
           "AutoParentheses"]}], ",", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"0", ",", "1", ",", "0"}], "\[RightAngleBracket]"}]}], 
        "\[RightAngleBracket]"}], ",", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          FractionBox["4", "7"], 
          TagBox[")",
           "AutoParentheses"]}], ",", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"0", ",", "0", ",", "0"}], "\[RightAngleBracket]"}]}], 
        "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"poly1", ",", 
       UnderscriptBox["1", "\[ScriptCapitalQ]xyz"]}], 
      "\[RightAngleBracket]"}]}], "]"}], "]"}]}]], "Computation",
 CellChangeTimes->{{3.645777290526781*^9, 3.645777316745391*^9}, {
   3.6457811571402273`*^9, 3.645781186208552*^9}, 3.6457815647827044`*^9},
 CellID->707667009],

Cell[BoxData[
 FormBox["0", TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645777317996956*^9, 3.6457811897634974`*^9, 
  3.666936618476619*^9},
 CellID->1825061739]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c707667009"]]]]}, {None, None}},
 CellChangeTimes->{3.666936620594318*^9},
 CellID->1750862647],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c707667009-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {211., 217.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666936621303926*^9},
 CellID->1389165957]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["chainCrit", "\[ScriptCapitalG]\[ScriptCapitalQ]xyz"], "[", 
  RowBox[{"1", ",", "2", ",", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"5", ",", "1", ",", "3"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], ",", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"poly1", ",", "poly2", ",", "poly3", ",", 
     UnderscriptBox["0", "\[ScriptCapitalQ]xyz"]}], "\[RightAngleBracket]"}], 
   ",", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"1", ",", "4"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"2", ",", "4"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"3", ",", "4"}], "\[RightAngleBracket]"}]}], 
    "\[RightAngleBracket]"}]}], "]"}]], "Computation",
 CellChangeTimes->{{3.59463980651697*^9, 3.594639819402611*^9}, {
   3.594640034777827*^9, 3.59464004801119*^9}, {3.59464013292029*^9, 
   3.594640227941778*^9}, {3.607341089450816*^9, 3.60734110197543*^9}, {
   3.631449679570619*^9, 3.631449685306025*^9}, {3.645777336784148*^9, 
   3.645777358137576*^9}, {3.645777410383856*^9, 3.645777413440447*^9}, 
   3.645781161883234*^9, 3.645781193768668*^9, 3.645781567320857*^9},
 CellID->496046602],

Cell[BoxData[
 FormBox["True", TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.60734110280185*^9, 3.60802362306378*^9, 3.610104052709569*^9, 
   3.615801965017417*^9, {3.63144985654362*^9, 3.631449859764426*^9}, 
   3.64577741493536*^9, 3.6457811957140102`*^9, 3.6669366234045277`*^9},
 CellID->1815416274]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c496046602"]]]]}, {None, None}},
 CellChangeTimes->{3.6669366254080667`*^9},
 CellID->2055356184],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c496046602-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {211., 217.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666936626125886*^9},
 CellID->866446722]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lmTuple", "\[ScriptCapitalG]\[ScriptCapitalQ]xyz"], "[", 
  RowBox[{"poly2", ",", "poly3"}], "]"}]], "Computation",
 CellChangeTimes->{{3.631449871082885*^9, 3.631449872031981*^9}, {
   3.64577741833602*^9, 3.64577742047115*^9}, 3.6457775889271193`*^9, 
   3.645781198737863*^9, 3.6457815696009417`*^9},
 CellID->2143832469],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"1", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"5", ",", "0", ",", "3"}], "\[RightAngleBracket]"}]}], 
       "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          FractionBox["1", "3"], ",", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"0", ",", "0", ",", "0"}], "\[RightAngleBracket]"}]}], 
         "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], ",", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          FractionBox["1", "3"], ",", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"3", ",", "0", ",", "2"}], "\[RightAngleBracket]"}]}], 
         "\[RightAngleBracket]"}], "\[RightAngleBracket]"}]}], 
      "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
   "\[RightAngleBracket]"}], TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.631449872929653*^9, 3.645777591732449*^9, 
  3.6457812002339287`*^9, 3.6669366461566887`*^9},
 CellID->369046373]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c2143832469"]]]]}, {None, None}},
 CellChangeTimes->{3.6669366481820517`*^9},
 CellID->173456624],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c2143832469-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {211., 217.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666936648899621*^9},
 CellID->391417162]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalQ]xyz", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["ard", "\[ScriptCapitalG]\[ScriptCapitalQ]xyz"], "[", 
    RowBox[{
     UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalQ]xyz"], "[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"poly2", ",", "poly3"}], "\[RightAngleBracket]"}], "]"}], "]"}],
    "]"}]}]], "Computation",
 CellChangeTimes->{{3.645777610499056*^9, 3.645777615471879*^9}, {
  3.64577769965414*^9, 3.645777709432386*^9}, {3.6457812042627687`*^9, 
  3.64578121309529*^9}, {3.645781571380155*^9, 3.645781572412188*^9}},
 CellID->722194560],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{
     RowBox[{"-", "y"}], "+", 
     RowBox[{"3", " ", 
      SuperscriptBox["x", "2"], " ", "z"}]}], ",", 
    RowBox[{
     RowBox[{"x", " ", "y", " ", "z"}], "-", 
     RowBox[{
      FractionBox["1", "5"], " ", "x", " ", 
      SuperscriptBox["y", "2"], " ", "z"}], "-", 
     FractionBox[
      SuperscriptBox["z", "2"], "6"], "+", 
     RowBox[{
      FractionBox["1", "3"], " ", "x", " ", 
      SuperscriptBox["y", "3"], " ", 
      SuperscriptBox["z", "3"]}]}], ",", 
    RowBox[{
     SuperscriptBox["y", "2"], "-", 
     FractionBox[
      SuperscriptBox["y", "3"], "5"], "-", 
     FractionBox[
      RowBox[{"x", " ", 
       SuperscriptBox["z", "2"]}], "2"], "+", 
     FractionBox[
      RowBox[{
       SuperscriptBox["y", "4"], " ", 
       SuperscriptBox["z", "2"]}], "3"]}], ",", 
    RowBox[{
     RowBox[{"-", 
      FractionBox[
       RowBox[{"x", " ", 
        SuperscriptBox["y", "2"]}], "2"]}], "+", 
     RowBox[{"9", " ", 
      SuperscriptBox["x", "4"], " ", 
      SuperscriptBox["y", "2"]}], "-", 
     FractionBox[
      RowBox[{"9", " ", 
       SuperscriptBox["x", "4"], " ", 
       SuperscriptBox["y", "3"]}], "5"], "+", 
     FractionBox[
      SuperscriptBox["y", "6"], "3"]}], ",", 
    RowBox[{
     RowBox[{"3", " ", 
      SuperscriptBox["x", "2"], " ", 
      SuperscriptBox["y", "2"]}], "-", 
     FractionBox[
      RowBox[{"3", " ", 
       SuperscriptBox["x", "2"], " ", 
       SuperscriptBox["y", "3"]}], "5"], "-", 
     FractionBox[
      RowBox[{"x", " ", "y", " ", "z"}], "2"], "+", 
     FractionBox[
      RowBox[{
       SuperscriptBox["y", "5"], " ", "z"}], "3"]}]}], 
   "\[RightAngleBracket]"}], TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645777629224595*^9, 3.645777714305509*^9, 
  3.64578121824852*^9, 3.6669366868048058`*^9},
 CellID->778638775]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c722194560"]]]]}, {None, None}},
 CellChangeTimes->{3.666936688825539*^9},
 CellID->21965254],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c722194560-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {211., 217.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666936689551189*^9},
 CellID->709996602]
}, Closed]],

Cell[TextData[{
 "As can be seen, the Gr\[ODoubleDot]bner basis computed by function \
\[OpenCurlyDoubleQuote]GB\[CloseCurlyDoubleQuote] coincides with the one \
computed by ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 "\[CloseCurlyQuote]s built-in \[OpenCurlyDoubleQuote]GroebnerBasis\
\[CloseCurlyDoubleQuote] function (up to multiplication by constants):"
}], "Text",
 CellChangeTimes->{{3.6457779918084717`*^9, 3.645778033968535*^9}},
 CellID->153491769],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"GroebnerBasis", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"3", 
       SuperscriptBox["x", "5"], 
       SuperscriptBox["z", "3"]}], "-", 
      RowBox[{"x", " ", 
       SuperscriptBox["y", "3"], 
       SuperscriptBox["z", "3"]}], "+", 
      RowBox[{
       FractionBox["4", "5"], 
       SuperscriptBox["x", "3"], "y", " ", 
       SuperscriptBox["z", "2"]}], "-", 
      RowBox[{"3", "x", " ", "y", " ", "z"}], "+", 
      RowBox[{
       FractionBox["1", "2"], 
       SuperscriptBox["z", "2"]}]}], ",", 
     RowBox[{
      RowBox[{"3", 
       SuperscriptBox["x", "2"], "z"}], "-", "y"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", "y", ",", "z"}], "}"}], ",", 
   RowBox[{"MonomialOrder", "\[Rule]", "DegreeReverseLexicographic"}]}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.645777723496063*^9, 3.645777797254936*^9}},
 CellID->1101254799],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"-", "y"}], "+", 
    RowBox[{"3", " ", 
     SuperscriptBox["x", "2"], " ", "z"}]}], ",", 
   RowBox[{
    RowBox[{"30", " ", 
     SuperscriptBox["y", "2"]}], "-", 
    RowBox[{"6", " ", 
     SuperscriptBox["y", "3"]}], "-", 
    RowBox[{"15", " ", "x", " ", 
     SuperscriptBox["z", "2"]}], "+", 
    RowBox[{"10", " ", 
     SuperscriptBox["y", "4"], " ", 
     SuperscriptBox["z", "2"]}]}], ",", 
   RowBox[{
    RowBox[{"90", " ", 
     SuperscriptBox["x", "2"], " ", 
     SuperscriptBox["y", "2"]}], "-", 
    RowBox[{"18", " ", 
     SuperscriptBox["x", "2"], " ", 
     SuperscriptBox["y", "3"]}], "-", 
    RowBox[{"15", " ", "x", " ", "y", " ", "z"}], "+", 
    RowBox[{"10", " ", 
     SuperscriptBox["y", "5"], " ", "z"}]}], ",", 
   RowBox[{
    RowBox[{"30", " ", "x", " ", "y", " ", "z"}], "-", 
    RowBox[{"6", " ", "x", " ", 
     SuperscriptBox["y", "2"], " ", "z"}], "-", 
    RowBox[{"5", " ", 
     SuperscriptBox["z", "2"]}], "+", 
    RowBox[{"10", " ", "x", " ", 
     SuperscriptBox["y", "3"], " ", 
     SuperscriptBox["z", "3"]}]}], ",", 
   RowBox[{
    RowBox[{"15", " ", "x", " ", 
     SuperscriptBox["y", "2"]}], "-", 
    RowBox[{"270", " ", 
     SuperscriptBox["x", "4"], " ", 
     SuperscriptBox["y", "2"]}], "+", 
    RowBox[{"54", " ", 
     SuperscriptBox["x", "4"], " ", 
     SuperscriptBox["y", "3"]}], "-", 
    RowBox[{"10", " ", 
     SuperscriptBox["y", "6"]}]}]}], "}"}]], "Output",
 CellChangeTimes->{3.645777798426079*^9},
 CellID->1633409070]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{"\[DoubleStruckCapitalZ]", "[", 
   RowBox[{"x", ",", "y"}], "]"}], TraditionalForm]]]], "Subsection",
 CellChangeTimes->{{3.596191479804769*^9, 3.596191492717959*^9}, {
  3.596192461569363*^9, 3.596192461901982*^9}, {3.645779782524728*^9, 
  3.645779782806164*^9}},
 CellID->2093452402],

Cell["", "OpenEnvironment",
 CellID->1713907510],

Cell[CellGroupData[{

Cell[TextData[{
 "Definition: Gr\[ODoubleDot]bner Ring of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[DoubleStruckCapitalZ]", "[", 
    RowBox[{"x", ",", "y"}], "]"}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.596191516114268*^9, 3.596191532492729*^9}, {
  3.59619246361635*^9, 3.596192464319168*^9}, {3.645779791364962*^9, 
  3.6457797916593723`*^9}},
 CellID->810757233],

Cell[BoxData[
 RowBox[{"\[ScriptCapitalZ]xy", ":=", 
  RowBox[{"PolyTuples", "[", 
   RowBox[{"\[ScriptCapitalZ]", ",", "\[ScriptCapitalT]2"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalZ]xy)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1921709458", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.596191572717378*^9, 3.596191575934689*^9}, {
  3.596191628353664*^9, 3.596191636296058*^9}, {3.596192467106728*^9, 
  3.596192470824888*^9}, {3.6457780692642097`*^9, 3.645778072515918*^9}, {
  3.6457781343831673`*^9, 3.645778135752284*^9}, {3.6457783314795*^9, 
  3.6457783324707212`*^9}, {3.64578093825847*^9, 3.645780940624741*^9}, {
  3.645781001059987*^9, 3.6457810023871927`*^9}},
 CellTags->{
  "ID:1921709458", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalZ]xy"},
 CellID->1921709458],

Cell[BoxData[
 RowBox[{"\[ScriptCapitalG]\[ScriptCapitalZ]xy", ":=", 
  RowBox[{
  "GroebnerRing", "[", "\[ScriptCapitalZ]xy", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalG] \[ScriptCapitalZ]xy)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:281007089", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.596191541721945*^9, 3.596191590594404*^9}, {
  3.596192472669584*^9, 3.596192474216192*^9}, {3.607341173408438*^9, 
  3.607341173960876*^9}, {3.645778075018854*^9, 3.6457780783094177`*^9}, {
  3.645778141495479*^9, 3.6457781428276777`*^9}, {3.645780959174*^9, 
  3.645780966368587*^9}, {3.645781500633254*^9, 3.6457815017770967`*^9}},
 CellTags->{
  "ID:281007089", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]xy"},
 CellID->281007089],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->282724972]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1301683811],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"poly4", "=", 
  RowBox[{
   UnderscriptBox["<<", "\[ScriptCapitalZ]xy"], 
   RowBox[{"[", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"x", "+", 
      RowBox[{"4", "y"}], "+", "2"}], 
     TagBox[")",
      "AutoParentheses"]}], "]"}]}]}]], "Computation",
 CellChangeTimes->{{3.645778170345675*^9, 3.645778205603258*^9}, {
  3.645781510913827*^9, 3.645781512905799*^9}},
 CellID->928226933],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"1", ",", "0"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"4", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"0", ",", "1"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"2", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"0", ",", "0"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.645778303158366*^9, {3.6457783354257803`*^9, 3.6457783600073347`*^9}, 
   3.645778566534252*^9, 3.645778896154838*^9, 3.645781015745545*^9, {
   3.666937057636237*^9, 3.66693706582718*^9}, 3.6669376677060633`*^9},
 CellID->1031522555]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c928226933"]]]]}, {None, None}},
 CellChangeTimes->{3.666937669716654*^9},
 CellID->494727030],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c928226933-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {211., 217.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666937670430828*^9},
 CellID->1681718319]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalZ]xy"], 
  RowBox[{"[", "poly4", "]"}]}]], "Computation",
 CellChangeTimes->{{3.666937108484133*^9, 3.666937119683359*^9}},
 CellID->1154153169],

Cell[BoxData[
 FormBox["True", TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.66693712090314*^9},
 CellID->565150268]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1154153169"]]]]}, {None, None}},
 CellChangeTimes->{3.6669371228464193`*^9},
 CellID->250297112],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1154153169-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {211., 217.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6669371235507183`*^9},
 CellID->1398378781]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"poly5", "=", 
  RowBox[{
   UnderscriptBox["<<", "\[ScriptCapitalZ]xy"], 
   RowBox[{"[", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"x", " ", "y"}], "+", 
      RowBox[{"4", "x"}], "+", "3"}], 
     TagBox[")",
      "AutoParentheses"]}], "]"}]}]}]], "Computation",
 CellChangeTimes->{{3.645778191554359*^9, 3.645778218013579*^9}, 
   3.6457815165859203`*^9},
 CellID->1686609323],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"1", ",", "1"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"4", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"1", ",", "0"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"0", ",", "0"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645778586210778*^9, 3.6457788990553017`*^9, 
  3.645781018250785*^9, 3.6669374188796186`*^9, 3.666937672384571*^9},
 CellID->1452783109]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1686609323"]]]]}, {None, None}},
 CellChangeTimes->{3.666937674386504*^9},
 CellID->765097563],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1686609323-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {211., 217.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6669376751057253`*^9},
 CellID->806774256]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"poly6", "=", 
  RowBox[{
   UnderscriptBox["<<", "\[ScriptCapitalZ]xy"], 
   RowBox[{"[", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"x", " ", 
       SuperscriptBox["y", "2"]}], "+", 
      RowBox[{"4", "y"}], "+", "2"}], 
     TagBox[")",
      "AutoParentheses"]}], "]"}]}]}]], "Computation",
 CellChangeTimes->{{3.6457782228889923`*^9, 3.645778237045854*^9}, 
   3.645781517693183*^9},
 CellID->493960096],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"1", ",", "2"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"4", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"0", ",", "1"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"2", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"0", ",", "0"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645778595605673*^9, 3.6457789017055197`*^9, 
  3.645781021044552*^9, 3.666937423663443*^9, 3.666937676630508*^9},
 CellID->1934080482]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c493960096"]]]]}, {None, None}},
 CellChangeTimes->{3.666937678639336*^9},
 CellID->1616194477],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c493960096-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {211., 217.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666937679360806*^9},
 CellID->462334974]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"poly7", "=", 
  RowBox[{
   UnderscriptBox["<<", "\[ScriptCapitalZ]xy"], 
   RowBox[{"[", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       SuperscriptBox["x", "2"], "y"}], "+", 
      RowBox[{"4", "x"}], "+", "3"}], 
     TagBox[")",
      "AutoParentheses"]}], "]"}]}]}]], "Computation",
 CellChangeTimes->{{3.645778246598831*^9, 3.645778264810707*^9}, 
   3.645781518910885*^9},
 CellID->1132140614],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"1", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"2", ",", "1"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"4", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"1", ",", "0"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"0", ",", "0"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645778604305485*^9, 3.6457789045163383`*^9, 
  3.6457810232529573`*^9, 3.666937433024262*^9, 3.666937681441421*^9},
 CellID->1903936635]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1132140614"]]]]}, {None, None}},
 CellChangeTimes->{3.666937683737794*^9},
 CellID->2064683746],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1132140614-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {211., 217.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666937684464569*^9},
 CellID->2072598081]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalZ]xy", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["lcrd", "\[ScriptCapitalG]\[ScriptCapitalZ]xy"], "[", 
    RowBox[{"poly4", ",", "poly5", ",", "1", ",", "1"}], "]"}], 
   "]"}]}]], "Computation",
 CellChangeTimes->{{3.6457786176997623`*^9, 3.6457786405770187`*^9}, {
  3.645781523181505*^9, 3.6457815343457127`*^9}},
 CellID->737891318],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"-", "x"}], " ", "y"}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.6457786436251*^9, 3.6457789070557737`*^9, 
  3.645781026455947*^9, 3.666937458534416*^9},
 CellID->456067319]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c737891318"]]]]}, {None, None}},
 CellChangeTimes->{3.666937460828925*^9},
 CellID->1015696530],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c737891318-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {211., 217.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666937461544416*^9},
 CellID->1495241193]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalZ]xy", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["ard", "\[ScriptCapitalG]\[ScriptCapitalZ]xy"], "[", 
    RowBox[{
     UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalZ]xy"], "[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"poly4", ",", "poly5"}], "\[RightAngleBracket]"}], "]"}], "]"}],
    "]"}]}]], "Computation",
 CellChangeTimes->{{3.645778922755761*^9, 3.6457789536947947`*^9}, {
  3.6457815244730167`*^9, 3.645781538546232*^9}},
 CellID->1018642778],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"2", "+", "x", "+", 
     RowBox[{"4", " ", "y"}]}], ",", 
    RowBox[{"5", "+", 
     RowBox[{"18", " ", "y"}], "+", 
     RowBox[{"4", " ", 
      SuperscriptBox["y", "2"]}]}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{{3.645778937278345*^9, 3.645778954900433*^9}, 
   3.6457810296952543`*^9, 3.666937470503051*^9},
 CellID->81922570]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1018642778"]]]]}, {None, None}},
 CellChangeTimes->{3.6669374724935637`*^9},
 CellID->1625548729],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1018642778-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {211., 217.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666937473206943*^9},
 CellID->1981492116]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"GroebnerBasis", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"x", "+", 
      RowBox[{"4", "y"}], "+", "2"}], ",", 
     RowBox[{
      RowBox[{"x", " ", "y"}], "+", 
      RowBox[{"4", "x"}], "+", "3"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", "y"}], "}"}], ",", 
   RowBox[{"MonomialOrder", "\[Rule]", "DegreeReverseLexicographic"}], ",", 
   RowBox[{"CoefficientDomain", "\[Rule]", "Integers"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.607341304183858*^9, 3.607341353312032*^9}, {
  3.645779013759791*^9, 3.645779019320702*^9}},
 CellID->1620459251],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"2", "+", "x", "+", 
    RowBox[{"4", " ", "y"}]}], ",", 
   RowBox[{
    RowBox[{"-", "5"}], "-", 
    RowBox[{"18", " ", "y"}], "-", 
    RowBox[{"4", " ", 
     SuperscriptBox["y", "2"]}]}]}], "}"}]], "Output",
 CellChangeTimes->{3.645779020479648*^9},
 CellID->1167063876]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalZ]xy", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["ard", "\[ScriptCapitalG]\[ScriptCapitalZ]xy"], "[", 
    RowBox[{
     UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalZ]xy"], "[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"poly6", ",", "poly7"}], "\[RightAngleBracket]"}], "]"}], "]"}],
    "]"}]}]], "Computation",
 CellChangeTimes->{{3.631450115866868*^9, 3.631450116809414*^9}, {
  3.645779032948251*^9, 3.6457790368896227`*^9}, {3.645781526941573*^9, 
  3.645781541973913*^9}},
 CellID->1874818521],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"2", "+", 
     RowBox[{"4", " ", "y"}], "+", 
     RowBox[{"x", " ", 
      SuperscriptBox["y", "2"]}]}], ",", 
    RowBox[{"3", "+", 
     RowBox[{"4", " ", "x"}], "+", 
     RowBox[{
      SuperscriptBox["x", "2"], " ", "y"}]}], ",", 
    RowBox[{
     RowBox[{"2", " ", "x"}], "-", 
     RowBox[{"3", " ", "y"}]}], ",", 
    RowBox[{
     RowBox[{"-", "4"}], "-", 
     RowBox[{"8", " ", "y"}], "-", 
     RowBox[{"3", " ", 
      SuperscriptBox["y", "3"]}]}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.6314501239553347`*^9, 3.645779039611787*^9, 
  3.6457810343967133`*^9, 3.66693756053489*^9},
 CellID->1919959885]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1874818521"]]]]}, {None, None}},
 CellChangeTimes->{3.666937562508067*^9},
 CellID->580296865],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1874818521-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {211., 217.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6669375632229652`*^9},
 CellID->112711656]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"GroebnerBasis", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"x", " ", 
       SuperscriptBox["y", "2"]}], "+", 
      RowBox[{"4", "y"}], "+", "2"}], ",", 
     RowBox[{
      RowBox[{
       SuperscriptBox["x", "2"], "y"}], "+", 
      RowBox[{"4", "x"}], "+", "3"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", "y"}], "}"}], ",", 
   RowBox[{"MonomialOrder", "\[Rule]", "DegreeReverseLexicographic"}], ",", 
   RowBox[{"CoefficientDomain", "\[Rule]", "Integers"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.607341304183858*^9, 3.607341353312032*^9}},
 CellID->334307522],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"2", " ", "x"}], "-", 
    RowBox[{"3", " ", "y"}]}], ",", 
   RowBox[{"4", "+", 
    RowBox[{"8", " ", "y"}], "+", 
    RowBox[{"3", " ", 
     SuperscriptBox["y", "3"]}]}], ",", 
   RowBox[{"2", "+", 
    RowBox[{"4", " ", "y"}], "+", 
    RowBox[{"x", " ", 
     SuperscriptBox["y", "2"]}]}], ",", 
   RowBox[{"3", "+", 
    RowBox[{"6", " ", "y"}], "+", 
    RowBox[{
     SuperscriptBox["x", "2"], " ", "y"}]}]}], "}"}]], "Output",
 CellChangeTimes->{3.607341354146081*^9, 3.608024045123539*^9, 
  3.615802077976342*^9, 3.631450131100707*^9},
 CellID->1032387437]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{
   SubscriptBox["\[DoubleStruckCapitalZ]", "24"], "[", 
   RowBox[{"x", ",", "y"}], "]"}], TraditionalForm]]]], "Subsection",
 CellChangeTimes->{{3.594111051425592*^9, 3.594111054968721*^9}, {
  3.607762605967956*^9, 3.607762612409246*^9}, {3.64577977833939*^9, 
  3.645779778673761*^9}},
 CellID->484498399],

Cell["", "OpenEnvironment",
 CellID->438098684],

Cell[CellGroupData[{

Cell[TextData[{
 "Definition: Gr\[ODoubleDot]bner Ring of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[DoubleStruckCapitalZ]", "24"], "[", 
    RowBox[{"x", ",", "y"}], "]"}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.594111073532906*^9, 3.594111090932806*^9}, {
  3.607762593211863*^9, 3.607762602264591*^9}, {3.645779772197019*^9, 
  3.645779774771872*^9}},
 CellID->269683380],

Cell[BoxData[
 RowBox[{"\[ScriptCapitalZ]24xy", ":=", 
  RowBox[{"PolyTuples", "[", 
   RowBox[{"\[ScriptCapitalZ]24", ",", "\[ScriptCapitalT]2"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalZ]24xy)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1512392075", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.607762731433203*^9, 3.607762796175256*^9}, {
  3.6457793333544493`*^9, 3.6457794020490723`*^9}, {3.6457802432583838`*^9, 
  3.64578024963925*^9}},
 CellTags->{
  "ID:1512392075", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalZ]24xy"},
 CellID->1512392075],

Cell[BoxData[
 RowBox[{"\[ScriptCapitalG]\[ScriptCapitalZ]24xy", ":=", 
  RowBox[{
  "GroebnerRing", "[", "\[ScriptCapitalZ]24xy", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalG] \[ScriptCapitalZ]24xy)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:543229319", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.607762822147151*^9, {3.645779355638915*^9, 3.645779376811718*^9}, {
   3.645780255748831*^9, 3.645780257666185*^9}, {3.645781464030305*^9, 
   3.645781465351355*^9}},
 CellTags->{
  "ID:543229319", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]24xy"},
 CellID->543229319],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1764652756]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1776414057],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalZ]24xy", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalZ]24xy"], "[", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], "]"}], 
   "]"}]}]], "Computation",
 CellChangeTimes->{{3.594111133265843*^9, 3.594111142747825*^9}, {
   3.607762838558035*^9, 3.607762838706709*^9}, {3.608024116262919*^9, 
   3.608024124720802*^9}, 3.631450149143076*^9, {3.645780308003632*^9, 
   3.6457803369838247`*^9}, 3.645781476285914*^9},
 CellID->2076222122],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.607760809348652*^9, 3.607762839586637*^9, {3.607762954826059*^9, 
   3.607762961161926*^9}, 3.607765185343121*^9, 3.608006307250386*^9, 
   3.608006418055206*^9, 3.608024125270817*^9, 3.610104502418526*^9, {
   3.614928951321909*^9, 3.614928969840172*^9}, {3.614929144332658*^9, 
   3.614929152366965*^9}, 3.614929291708282*^9, {3.615802112571871*^9, 
   3.615802139936889*^9}, {3.6314502293685493`*^9, 3.6314502324581823`*^9}, 
   3.6349669068160257`*^9, 3.645780367307741*^9, {3.666936960053121*^9, 
   3.666936980146628*^9}, 3.6669375816682663`*^9, 3.666937721071192*^9},
 CellID->1961765370]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c2076222122"]]]]}, {None, None}},
 CellChangeTimes->{3.666937723077183*^9},
 CellID->231019000],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c2076222122-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {226., 232.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666937723806088*^9},
 CellID->1617801885]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalZ]24xy", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalZ]24xy"], "[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"poly4", ",", "poly5"}], "\[RightAngleBracket]"}], "]"}], 
   "]"}]}]], "Computation",
 CellChangeTimes->{{3.6457803831542873`*^9, 3.645780385304264*^9}, 
   3.645781478754155*^9},
 CellID->1987464113],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"2", "+", "x", "+", 
     RowBox[{"4", " ", "y"}]}], ",", 
    RowBox[{"3", "+", 
     RowBox[{"4", " ", "x"}], "+", 
     RowBox[{"x", " ", "y"}]}], ",", 
    RowBox[{"19", "+", 
     RowBox[{"6", " ", "y"}], "+", 
     RowBox[{"20", " ", 
      SuperscriptBox["y", "2"]}]}], ",", 
    RowBox[{"6", "+", 
     RowBox[{"12", " ", "y"}]}], ",", "12", ",", "21", ",", 
    RowBox[{"1", "+", 
     RowBox[{"23", " ", 
      SuperscriptBox["y", "2"]}]}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645780400296679*^9, 3.666936996040262*^9, 
  3.666937587425069*^9, 3.666937738841489*^9},
 CellID->1211232653]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1987464113"]]]]}, {None, None}},
 CellChangeTimes->{3.666937740961225*^9},
 CellID->1858134368],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1987464113-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {226., 232.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6669377416900473`*^9},
 CellID->881951091]
}, Closed]],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"2", "+", "x", "+", 
     RowBox[{"4", " ", "y"}]}], ",", 
    RowBox[{"3", "+", 
     RowBox[{"4", " ", "x"}], "+", 
     RowBox[{"x", " ", "y"}]}], ",", 
    RowBox[{"19", "+", 
     RowBox[{"6", " ", "y"}], "+", 
     RowBox[{"20", " ", 
      SuperscriptBox["y", "2"]}]}], ",", 
    RowBox[{"6", "+", 
     RowBox[{"12", " ", "y"}]}], ",", "12", ",", "18", ",", 
    RowBox[{"23", "+", 
     RowBox[{"22", " ", 
      SuperscriptBox["y", "2"]}]}], ",", "3", ",", 
    RowBox[{"22", "+", 
     RowBox[{"23", " ", 
      SuperscriptBox["y", "2"]}]}]}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "Computation",
 CellID->410409823],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalZ]24xy", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalZ]24xy"], "[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"poly4", ",", "poly6"}], "\[RightAngleBracket]"}], "]"}], 
   "]"}]}]], "Computation",
 CellChangeTimes->{3.64578050256633*^9, 3.645781480163782*^9},
 CellID->388032786],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"2", "+", "x", "+", 
     RowBox[{"4", " ", "y"}]}], ",", 
    RowBox[{"2", "+", 
     RowBox[{"4", " ", "y"}], "+", 
     RowBox[{"x", " ", 
      SuperscriptBox["y", "2"]}]}], ",", 
    RowBox[{"2", "+", 
     RowBox[{"4", " ", "y"}], "+", 
     RowBox[{"22", " ", 
      SuperscriptBox["y", "2"]}], "+", 
     RowBox[{"20", " ", 
      SuperscriptBox["y", "3"]}]}], ",", 
    RowBox[{"12", "+", 
     RowBox[{"12", " ", 
      SuperscriptBox["y", "2"]}]}], ",", 
    RowBox[{"18", "+", 
     RowBox[{"6", " ", 
      SuperscriptBox["y", "2"]}]}], ",", 
    RowBox[{"20", "+", 
     RowBox[{"22", " ", "y"}], "+", 
     RowBox[{"4", " ", 
      SuperscriptBox["y", "2"]}], "+", 
     RowBox[{"2", " ", 
      SuperscriptBox["y", "3"]}]}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645780524599128*^9, 3.666937610568469*^9, 
  3.666937799107333*^9},
 CellID->1924293860]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c388032786"]]]]}, {None, None}},
 CellChangeTimes->{3.666937801063414*^9},
 CellID->842604710],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c388032786-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {226., 232.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666937801814081*^9},
 CellID->1428414825]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalZ]24xy", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalZ]24xy"], "[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"poly4", ",", "poly7"}], "\[RightAngleBracket]"}], "]"}], 
   "]"}]}]], "Computation",
 CellChangeTimes->{3.645780544692596*^9, 3.64578148152826*^9},
 CellID->1393649830],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"2", "+", "x", "+", 
     RowBox[{"4", " ", "y"}]}], ",", 
    RowBox[{"3", "+", 
     RowBox[{"4", " ", "x"}], "+", 
     RowBox[{
      SuperscriptBox["x", "2"], " ", "y"}]}], ",", 
    RowBox[{"19", "+", 
     RowBox[{"12", " ", "y"}], "+", 
     RowBox[{"16", " ", 
      SuperscriptBox["y", "2"]}], "+", 
     RowBox[{"16", " ", 
      SuperscriptBox["y", "3"]}]}], ",", 
    RowBox[{"15", "+", 
     RowBox[{"12", " ", "y"}]}], ",", 
    RowBox[{"15", "+", 
     RowBox[{"18", " ", "y"}]}], ",", "18", ",", "21", ",", 
    RowBox[{"1", "+", 
     SuperscriptBox["y", "2"], "+", 
     SuperscriptBox["y", "3"]}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645780559428782*^9, 3.645780591601698*^9, 
  3.6669379266314983`*^9},
 CellID->1840469971]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1393649830"]]]]}, {None, None}},
 CellChangeTimes->{3.666937928821301*^9},
 CellID->334182051],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1393649830-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {226., 232.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666937929569324*^9},
 CellID->1317407945]
}, Closed]],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"2", "+", "x", "+", 
     RowBox[{"4", " ", "y"}]}], ",", 
    RowBox[{"3", "+", 
     RowBox[{"4", " ", "x"}], "+", 
     RowBox[{
      SuperscriptBox["x", "2"], " ", "y"}]}], ",", 
    RowBox[{"19", "+", 
     RowBox[{"12", " ", "y"}], "+", 
     RowBox[{"16", " ", 
      SuperscriptBox["y", "2"]}], "+", 
     RowBox[{"16", " ", 
      SuperscriptBox["y", "3"]}]}], ",", 
    RowBox[{"15", "+", 
     RowBox[{"12", " ", "y"}]}], ",", 
    RowBox[{"9", "+", 
     RowBox[{"6", " ", "y"}]}], ",", "18", ",", "21"}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "Computation",
 CellID->311652298],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalZ]24xy", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalZ]24xy"], "[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"poly5", ",", "poly6"}], "\[RightAngleBracket]"}], "]"}], 
   "]"}]}]], "Computation",
 CellChangeTimes->{{3.645780615497316*^9, 3.6457806173750353`*^9}, 
   3.6457814831129637`*^9},
 CellID->966365481],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"3", "+", 
     RowBox[{"4", " ", "x"}], "+", 
     RowBox[{"x", " ", "y"}]}], ",", 
    RowBox[{"2", "+", 
     RowBox[{"4", " ", "y"}], "+", 
     RowBox[{"x", " ", 
      SuperscriptBox["y", "2"]}]}], ",", 
    RowBox[{"14", "+", 
     RowBox[{"16", " ", "x"}], "+", "y"}], ",", 
    RowBox[{"6", "+", 
     RowBox[{"21", " ", "y"}]}], ",", 
    RowBox[{"4", "+", 
     RowBox[{"23", " ", 
      SuperscriptBox["y", "2"]}]}], ",", 
    RowBox[{"5", "+", 
     RowBox[{"22", " ", "x"}], "+", "y"}], ",", "12", ",", "6", ",", "21", 
    ",", 
    RowBox[{"2", "+", "x", "+", "y"}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645780628677948*^9, 3.666938025848509*^9},
 CellID->1237511167]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c966365481"]]]]}, {None, None}},
 CellChangeTimes->{3.666938027930944*^9},
 CellID->149562406],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c966365481-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {226., 232.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6669380286828213`*^9},
 CellID->1554428390]
}, Closed]],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"3", "+", 
     RowBox[{"4", " ", "x"}], "+", 
     RowBox[{"x", " ", "y"}]}], ",", 
    RowBox[{"2", "+", 
     RowBox[{"4", " ", "y"}], "+", 
     RowBox[{"x", " ", 
      SuperscriptBox["y", "2"]}]}], ",", 
    RowBox[{"14", "+", 
     RowBox[{"16", " ", "x"}], "+", "y"}], ",", 
    RowBox[{"6", "+", 
     RowBox[{"21", " ", "y"}]}], ",", 
    RowBox[{"19", "+", 
     RowBox[{"2", " ", "x"}], "+", 
     RowBox[{"23", " ", "y"}]}], ",", "12", ",", 
    RowBox[{"20", "+", 
     SuperscriptBox["y", "2"]}]}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "Computation",
 CellID->1424170364],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalZ]24xy", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalZ]24xy"], "[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"poly5", ",", "poly7"}], "\[RightAngleBracket]"}], "]"}], 
   "]"}]}]], "Computation",
 CellChangeTimes->{{3.6457806750246983`*^9, 3.645780676974934*^9}, 
   3.645781484293754*^9},
 CellID->248651614],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"3", "+", 
     RowBox[{"4", " ", "x"}], "+", 
     RowBox[{"x", " ", "y"}]}], ",", 
    RowBox[{"3", "+", 
     RowBox[{"4", " ", "x"}], "+", 
     RowBox[{
      SuperscriptBox["x", "2"], " ", "y"}]}], ",", 
    RowBox[{"3", "+", "x", "+", 
     RowBox[{"20", " ", 
      SuperscriptBox["x", "2"]}]}], ",", 
    RowBox[{"6", "+", 
     RowBox[{"18", " ", "x"}]}], ",", 
    RowBox[{"9", "+", 
     RowBox[{"15", " ", "y"}]}], ",", 
    RowBox[{"21", "+", "x", "+", 
     RowBox[{"2", " ", 
      SuperscriptBox["x", "2"]}]}], ",", 
    RowBox[{"3", "+", 
     RowBox[{"21", " ", "x"}]}], ",", 
    RowBox[{"x", "+", 
     RowBox[{"23", " ", 
      SuperscriptBox["x", "2"]}]}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645780691065043*^9, 3.666938080028421*^9},
 CellID->934060400]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c248651614"]]]]}, {None, None}},
 CellChangeTimes->{3.666938082132092*^9},
 CellID->425268034],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c248651614-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {226., 232.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6669380828801937`*^9},
 CellID->121844055]
}, Closed]],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"3", "+", 
     RowBox[{"4", " ", "x"}], "+", 
     RowBox[{"x", " ", "y"}]}], ",", 
    RowBox[{"3", "+", 
     RowBox[{"4", " ", "x"}], "+", 
     RowBox[{
      SuperscriptBox["x", "2"], " ", "y"}]}], ",", 
    RowBox[{"3", "+", "x", "+", 
     RowBox[{"20", " ", 
      SuperscriptBox["x", "2"]}]}], ",", 
    RowBox[{"6", "+", 
     RowBox[{"18", " ", "x"}]}], ",", 
    RowBox[{"18", "+", 
     RowBox[{"6", " ", "y"}]}], ",", 
    RowBox[{"3", "+", 
     RowBox[{"23", " ", "x"}], "+", 
     RowBox[{"22", " ", 
      SuperscriptBox["x", "2"]}]}], ",", 
    RowBox[{"21", "+", 
     RowBox[{"3", " ", "x"}]}], ",", 
    RowBox[{"21", "+", 
     RowBox[{"3", " ", "y"}]}], ",", 
    RowBox[{"3", "+", 
     RowBox[{"22", " ", "x"}], "+", 
     RowBox[{"23", " ", 
      SuperscriptBox["x", "2"]}]}]}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "Computation",
 CellID->166800873],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalZ]24xy", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalZ]24xy"], "[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"poly6", ",", "poly7"}], "\[RightAngleBracket]"}], "]"}], 
   "]"}]}]], "Computation",
 CellChangeTimes->{{3.645780720294505*^9, 3.645780723865295*^9}, 
   3.6457814860261908`*^9},
 CellID->1503343685],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"2", "+", 
     RowBox[{"4", " ", "y"}], "+", 
     RowBox[{"x", " ", 
      SuperscriptBox["y", "2"]}]}], ",", 
    RowBox[{"3", "+", 
     RowBox[{"4", " ", "x"}], "+", 
     RowBox[{
      SuperscriptBox["x", "2"], " ", "y"}]}], ",", 
    RowBox[{
     RowBox[{"22", " ", "x"}], "+", 
     RowBox[{"3", " ", "y"}]}], ",", 
    RowBox[{"12", " ", "y"}], ",", 
    RowBox[{"20", "+", 
     RowBox[{"4", " ", "y"}], "+", 
     RowBox[{"21", " ", 
      SuperscriptBox["y", "3"]}]}], ",", "12", ",", 
    RowBox[{"6", " ", 
     SuperscriptBox["y", "2"]}], ",", 
    RowBox[{"20", "+", 
     RowBox[{"4", " ", "y"}]}], ",", 
    RowBox[{"4", "+", 
     RowBox[{"2", " ", "y"}]}], ",", 
    RowBox[{"16", "+", 
     RowBox[{"23", " ", 
      SuperscriptBox["y", "2"]}]}], ",", "6", ",", 
    RowBox[{"2", "+", "y"}], ",", "3", ",", "x"}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645780739796832*^9, 3.666938199251328*^9},
 CellID->365732408]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1503343685"]]]]}, {None, None}},
 CellChangeTimes->{3.666938201255211*^9},
 CellID->312984509],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1503343685-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {226., 232.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.666938202024271*^9},
 CellID->1443280734]
}, Closed]],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"2", "+", 
     RowBox[{"4", " ", "y"}], "+", 
     RowBox[{"x", " ", 
      SuperscriptBox["y", "2"]}]}], ",", 
    RowBox[{"3", "+", 
     RowBox[{"4", " ", "x"}], "+", 
     RowBox[{
      SuperscriptBox["x", "2"], " ", "y"}]}], ",", 
    RowBox[{
     RowBox[{"22", " ", "x"}], "+", 
     RowBox[{"3", " ", "y"}]}], ",", 
    RowBox[{"12", " ", "y"}], ",", 
    RowBox[{"4", "+", 
     RowBox[{"20", " ", "y"}], "+", 
     RowBox[{"3", " ", 
      SuperscriptBox["y", "3"]}]}], ",", 
    RowBox[{"18", " ", 
     SuperscriptBox["y", "2"]}], ",", 
    RowBox[{"16", "+", 
     RowBox[{"20", " ", "y"}]}], ",", 
    RowBox[{"4", "+", 
     RowBox[{"23", " ", 
      SuperscriptBox["y", "3"]}]}], ",", 
    RowBox[{"16", "+", 
     RowBox[{"2", " ", 
      SuperscriptBox["y", "2"]}]}]}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "Computation",
 CellID->1183610157],

Cell["\<\
A one-element set is not necessarily a Gr\[ODoubleDot]bner basis:\
\>", "Text",
 CellChangeTimes->{{3.607766363431278*^9, 3.607766377697355*^9}},
 CellID->547291024],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"res", "=", 
  RowBox[{
   UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalZ]24xy"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     UnderscriptBox["<<", "\[ScriptCapitalZ]24xy"], 
     RowBox[{"[", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"16", "x", " ", "y"}], "+", "2"}], 
       TagBox[")",
        "AutoParentheses"]}], "]"}]}], "\[RightAngleBracket]"}], 
   "]"}]}]], "Computation",
 CellChangeTimes->{{3.645780775476667*^9, 3.6457808281766367`*^9}, 
   3.645781488788686*^9},
 CellID->2006674578],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"16", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"1", ",", "1"}], "\[RightAngleBracket]"}]}], 
       "\[RightAngleBracket]"}], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"2", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"0", ",", "0"}], "\[RightAngleBracket]"}]}], 
       "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"18", ",", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"0", ",", "0"}], "\[RightAngleBracket]"}]}], 
      "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"22", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"1", ",", "1"}], "\[RightAngleBracket]"}]}], 
       "\[RightAngleBracket]"}], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"2", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"0", ",", "0"}], "\[RightAngleBracket]"}]}], 
       "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}]}], 
   "\[RightAngleBracket]"}], TheoremaForm]], "ComputationResult",
 CellChangeTimes->{{3.6457808050891237`*^9, 3.6457808301484947`*^9}, 
   3.666938321979209*^9, 3.6670436211993923`*^9},
 CellID->658239665]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c2006674578"]]]]}, {None, None}},
 CellChangeTimes->{3.667043623043281*^9},
 CellID->133220999],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c2006674578-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {218., 225.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6670436237850018`*^9},
 CellID->93241346]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalZ]24xy", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], "res", 
    TagBox[")",
     "AutoParentheses"]}], "]"}]}]], "Computation",
 CellChangeTimes->{{3.645780834298642*^9, 3.645780877861499*^9}},
 CellID->183209066],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"2", "+", 
     RowBox[{"16", " ", "x", " ", "y"}]}], ",", "18", ",", 
    RowBox[{"2", "+", 
     RowBox[{"22", " ", "x", " ", "y"}]}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645780879754014*^9, 3.666938339187229*^9, 
  3.667043633103154*^9},
 CellID->1718893651]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c183209066"]]]]}, {None, None}},
 CellChangeTimes->{3.667043634957281*^9},
 CellID->173442579],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c183209066-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {218., 225.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.66704363568779*^9},
 CellID->1487544269]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"res0", "=", 
  RowBox[{
   UnderscriptBox["ard", "\[ScriptCapitalG]\[ScriptCapitalZ]24xy"], "[", 
   "res", "]"}]}]], "Computation",
 CellChangeTimes->{{3.645780893318697*^9, 3.6457809025489902`*^9}, 
   3.645781491770978*^9, {3.6670436429879293`*^9, 3.667043670691174*^9}},
 CellID->1367741268],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"18", ",", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"0", ",", "0"}], "\[RightAngleBracket]"}]}], 
      "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"22", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"1", ",", "1"}], "\[RightAngleBracket]"}]}], 
       "\[RightAngleBracket]"}], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"2", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"0", ",", "0"}], "\[RightAngleBracket]"}]}], 
       "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}]}], 
   "\[RightAngleBracket]"}], TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.64578090427862*^9, 3.666938344058926*^9, 
  3.6670436737727757`*^9},
 CellID->511099276]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1367741268"]]]]}, {None, None}},
 CellChangeTimes->{3.667043675661834*^9},
 CellID->309629685],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1367741268-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {218., 225.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6670436763877783`*^9},
 CellID->1668723514]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalZ]24xy", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], "res0", 
    TagBox[")",
     "AutoParentheses"]}], "]"}]}]], "Computation",
 CellChangeTimes->{{3.667043680203175*^9, 3.667043723063356*^9}},
 CellID->1250393302],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"18", ",", 
    RowBox[{"2", "+", 
     RowBox[{"22", " ", "x", " ", "y"}]}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.667043725751*^9},
 CellID->1927486347]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1250393302"]]]]}, {None, None}},
 CellChangeTimes->{3.6670437275962477`*^9},
 CellID->2036526983],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1250393302-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {218., 225.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.667043728323484*^9},
 CellID->1762191997]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalZ]24xy", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["GB", "\[ScriptCapitalG]\[ScriptCapitalZ]24xy"], "[", 
    "res0", "]"}], "]"}]}]], "Computation",
 CellChangeTimes->{{3.667043733929183*^9, 3.667043756126377*^9}},
 CellID->1733660062],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"18", ",", 
    RowBox[{"2", "+", 
     RowBox[{"22", " ", "x", " ", "y"}]}]}], "\[RightAngleBracket]"}], 
  TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.66704376190031*^9},
 CellID->924097278]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1733660062"]]]]}, {None, None}},
 CellChangeTimes->{3.667043763650392*^9},
 CellID->576174505],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1733660062-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {218., 225.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.667043764373796*^9},
 CellID->1929933941]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Theorem: Gr\[ODoubleDot]bner Rings of Reduction Rings are Reduction Rings (\
\[Checkmark])\
\>", "Section",
 CellChangeTimes->{{3.649136090102188*^9, 3.6491361105677547`*^9}, {
  3.649652589661607*^9, 3.6496525935382967`*^9}, {3.650185392688982*^9, 
  3.6501853937596493`*^9}, {3.6507933280391083`*^9, 3.650793330599265*^9}},
 CellID->1382014194],

Cell[TextData[{
 "Apparently, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"GroebnerRing", "[", "\[ScriptCapitalR]", "]"}], TraditionalForm]]],
 " is a conservative extension of every algorithmic reduction ring \
\[ScriptCapitalR] in the sense that it does not redefine any operations \
present in algorithmic reduction rings. Hence, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"GroebnerRing", "[", "\[ScriptCapitalR]", "]"}], TraditionalForm]]],
 " clearly is an algorithmic reduction ring as well."
}], "Text",
 CellChangeTimes->{{3.649583595451274*^9, 3.649583688932379*^9}},
 CellID->1670632347],

Cell[CellGroupData[{

Cell["\<\
Gr\[ODoubleDot]bner Rings of Unital Commutative Rings are Unital Commutative \
Rings (\[Checkmark])\
\>", "Subsection",
 CellChangeTimes->{{3.6491361304919653`*^9, 3.649136159708892*^9}, {
  3.649583558388838*^9, 3.64958356235918*^9}},
 CellID->279781653],

Cell["", "OpenEnvironment",
 CellID->1257814565],

Cell[CellGroupData[{

Cell["\<\
Theorem: \[ScriptCapitalG] is Commutative Ring with unit\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.649582222732123*^9, 3.649582240233585*^9}},
 CellID->915451344],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"isCommRing1", "[", "\[ScriptCapitalR]", "]"}]], 
  RowBox[{"isCommRing1", "[", 
   RowBox[{"GroebnerRing", "[", "\[ScriptCapitalR]", "]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalG] isCommRing1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:280174026", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6495822457862*^9, 3.649582264601447*^9}},
 CellTags->{
  "ID:280174026", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isCommRing1"},
 CellID->280174026],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->774494674]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1228140165],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[ScriptCapitalG] isCommRing1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[ScriptCapitalG] isCommRing1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isCommRing1", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isCommRing1", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isCommRing1", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:280174026"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p280174026-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p280174026-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:280174026", "Proof|ID:280174026-1"},
 CellID->79132233],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p280174026-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->273864868]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Gr\[ODoubleDot]bner Rings of Reduction Rings are Reduction Rings \
(\[Checkmark])\
\>", "Subsection",
 CellChangeTimes->{{3.6491361687449713`*^9, 3.649136178640733*^9}, {
  3.6496525822678547`*^9, 3.64965258579679*^9}, {3.6501853868341*^9, 
  3.650185387941751*^9}, {3.6507933216912813`*^9, 3.650793324330558*^9}},
 CellID->732003135],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "isReductionRing", "[", "\[ScriptCapitalR]", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1057748777}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649583745457139*^9, 3.649583749538539*^9}},
 CellTags->{
  "ID:1057748777", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1057748777],

Cell["", "Text",
 CellID->223648837],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"\[ScriptCapitalG]", "=", 
   RowBox[{
   "GroebnerRing", "[", "\[ScriptCapitalR]", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 464516108}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649583773886771*^9, 3.6495837835491056`*^9}},
 CellTags->{
  "ID:464516108", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->464516108],

Cell[CellGroupData[{

Cell["\<\
Lemma: Auxiliary Notions are Preserved (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.649584341935975*^9, 3.6495843511681232`*^9}, {
  3.6496494648725452`*^9, 3.649649473121643*^9}, {3.649652575349527*^9, 
  3.6496525789881907`*^9}, {3.6501853831781607`*^9, 3.6501853842548018`*^9}, {
  3.6507933137368107`*^9, 3.650793316406434*^9}},
 CellID->216509956],

Cell[TextData[{
 StyleBox["Remark",
  FontWeight->"Bold",
  FontSlant->"Italic"],
 ": All of the following formulas hold for any domain \[ScriptCapitalR], not \
only for reduction rings."
}], "Text",
 CellChangeTimes->{{3.649646989713027*^9, 3.649647025862999*^9}},
 CellID->754492038],

Cell["", "OpenEnvironment",
 CellID->1092859258],

Cell[CellGroupData[{

Cell["Lemma", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{3.649584371354794*^9},
 CellID->1393668874],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "S"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainSets", "[", "\[ScriptCapitalG]", "]"}]], 
      RowBox[{"[", "S", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainSets", "[", "\[ScriptCapitalR]", "]"}]], 
      RowBox[{"[", "S", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DomainSets \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1148013153", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649584374831284*^9, 3.649584404371516*^9}},
 CellTags->{
  "ID:1148013153", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainSets \[ScriptCapitalG]", "IR"},
 CellID->1148013153],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "T"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
      RowBox[{"[", "T", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "\[ScriptCapitalR]", "]"}]], 
      RowBox[{"[", "T", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DomainTuples \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1668771272", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649584411587224*^9, 3.649584426291765*^9}},
 CellTags->{
  "ID:1668771272", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples \[ScriptCapitalG]", "IR"},
 CellID->1668771272],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "f"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"n", "\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox[
       SubscriptBox["\[Sum]", "\[ScriptCapitalG]"], 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]], 
      RowBox[{"f", "[", "i", "]"}]}], "\[Equal]", 
     RowBox[{
      UnderscriptBox[
       SubscriptBox["\[Sum]", "\[ScriptCapitalR]"], 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]], 
      RowBox[{"f", "[", "i", "]"}]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Sum] \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1928426019", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6495857627646093`*^9, 3.6495858066056643`*^9}},
 CellTags->{
  "ID:1928426019", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Sum] \[ScriptCapitalG]", "IR"},
 CellID->1928426019],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"a", ",", "b", ",", "c", ",", "m"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 546598180}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6495868921444693`*^9, 3.64958689463776*^9}},
 CellTags->{
  "ID:546598180", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->546598180],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
    RowBox[{"[", "T", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"k", "=", "0"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "T", "\[RightBracketingBar]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["reduceM", "\[ScriptCapitalG]"], "[", 
      RowBox[{"a", ",", "c", ",", "T", ",", "k"}], "]"}], "\[Equal]", 
     RowBox[{
      UnderscriptBox["reduceM", "\[ScriptCapitalR]"], "[", 
      RowBox[{"a", ",", "c", ",", "T", ",", "k"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reduceM \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:581321955", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.64958803228433*^9, {3.6495880644790487`*^9, 3.649588125831882*^9}, {
   3.649588157210514*^9, 3.6495881593090153`*^9}},
 CellTags->{
  "ID:581321955", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "reduceM \[ScriptCapitalG]", "IR"},
 CellID->581321955],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
   RowBox[{"[", "c", "]"}]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"m", ",", "c"}]], "\[ScriptCapitalG]"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"m", ",", "c"}]], "\[ScriptCapitalR]"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[RightArrow] \[ScriptCapitalG] multiplier)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:405682622", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649586900431262*^9, 3.649586921879122*^9}, {
  3.6495870469280567`*^9, 3.649587061928833*^9}},
 CellTags->{
  "ID:405682622", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow] \[ScriptCapitalG] multiplier", "IR"},
 
 CellID->405682622],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
   RowBox[{"[", "c", "]"}]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"{", "c", "}"}]], "\[ScriptCapitalG]"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[Rule]", 
        RowBox[{"{", "c", "}"}]], "\[ScriptCapitalR]"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[RightArrow] \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1165811909", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649586978686574*^9, 3.649586981964794*^9}, {
  3.649587064572033*^9, 3.649587072985865*^9}, {3.66185181602229*^9, 
  3.6618518235571036`*^9}},
 CellTags->{
  "ID:1165811909", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow] \[ScriptCapitalG]", "IR"},
 CellID->1165811909],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
   RowBox[{"[", "c", "]"}]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[LeftRightArrow]", 
        RowBox[{"{", "c", "}"}]], "\[ScriptCapitalG]"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox[
       SubscriptBox["\[LeftRightArrow]", 
        RowBox[{"{", "c", "}"}]], "\[ScriptCapitalR]"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[LeftRightArrow] \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:306221528", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649586994985965*^9, 3.649587000818623*^9}, {
  3.6495876195604353`*^9, 3.649587629153784*^9}, {3.661851831731518*^9, 
  3.6618518381743293`*^9}},
 CellTags->{
  "ID:306221528", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[LeftRightArrow] \[ScriptCapitalG]", "IR"},
 CellID->306221528],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalG]"], "[", 
   RowBox[{"a", ",", "c", ",", "m"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
   RowBox[{"a", ",", "c", ",", "m"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReducible \[ScriptCapitalG] multiplier)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:113746156", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649586932020336*^9, 3.649586971587502*^9}},
 CellTags->{
  "ID:113746156", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isReducible \[ScriptCapitalG] multiplier", "IR"},
 CellID->113746156],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalG]"], "[", 
   RowBox[{"a", ",", "c"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalR]"], "[", 
   RowBox[{"a", ",", "c"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReducible \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1372373678", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649587162762032*^9, 3.649587165708991*^9}},
 CellTags->{
  "ID:1372373678", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isReducible \[ScriptCapitalG]", "IR"},
 CellID->1372373678],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["symred", "\[ScriptCapitalG]"], "[", 
   RowBox[{"a", ",", "b", ",", "c", ",", "m"}], "]"}], 
  "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["symred", "\[ScriptCapitalR]"], "[", 
   RowBox[{"a", ",", "b", ",", "c", ",", "m"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(symred \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:12812135", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649645931002143*^9, 3.6496459479034452`*^9}},
 CellTags->{
  "ID:12812135", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "symred \[ScriptCapitalG]", "IR"},
 CellID->12812135],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["areCorrelative", "\[ScriptCapitalG]"], "[", 
   RowBox[{"a", ",", "b", ",", "c"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["areCorrelative", "\[ScriptCapitalR]"], "[", 
   RowBox[{"a", ",", "b", ",", "c"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(areCorrelative \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:168109679", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649645997096574*^9, 3.649646027617305*^9}},
 CellTags->{
  "ID:168109679", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "areCorrelative \[ScriptCapitalG]", "IR"},
 CellID->168109679],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", "j"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isNTCR", "\[ScriptCapitalG]"], "[", 
     RowBox[{"a", ",", "b", ",", "c", ",", "i", ",", "j"}], "]"}], 
    "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
     RowBox[{"a", ",", "b", ",", "c", ",", "i", ",", "j"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isNTCR \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1427340407", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6496469243629313`*^9, 3.649646957720841*^9}},
 CellTags->{
  "ID:1427340407", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isNTCR \[ScriptCapitalG]", "IR"},
 CellID->1427340407],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", "j"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isMNTCR", "\[ScriptCapitalG]"], "[", 
     RowBox[{"a", ",", "b", ",", "c", ",", "i", ",", "j"}], "]"}], 
    "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
     RowBox[{"a", ",", "b", ",", "c", ",", "i", ",", "j"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMNTCR \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:587977465", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649647497821816*^9, 3.649647502860858*^9}},
 CellTags->{
  "ID:587977465", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isMNTCR \[ScriptCapitalG]", "IR"},
 CellID->587977465],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isNTCR", "\[ScriptCapitalG]"], "[", 
   RowBox[{"c", ",", "a"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["isNTCR", "\[ScriptCapitalR]"], "[", 
   RowBox[{"c", ",", "a"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isNTCR \[ScriptCapitalG] unary)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:950195558", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649648433886787*^9, 3.649648448713189*^9}},
 CellTags->{
  "ID:950195558", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isNTCR \[ScriptCapitalG] unary", "IR"},
 CellID->950195558],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isMNTCR", "\[ScriptCapitalG]"], "[", 
   RowBox[{"c", ",", "a"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["isMNTCR", "\[ScriptCapitalR]"], "[", 
   RowBox[{"c", ",", "a"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMNTCR \[ScriptCapitalG] unary)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2144197915", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649648455844616*^9, 3.649648460118751*^9}},
 CellTags->{
  "ID:2144197915", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isMNTCR \[ScriptCapitalG] unary", "IR"},
 CellID->2144197915],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainSets", "[", "\[ScriptCapitalG]", "]"}]], 
   RowBox[{"[", "c", "]"}]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["red", "\[ScriptCapitalG]"], "[", "c", "]"}], "\[Equal]", 
    
    RowBox[{
     UnderscriptBox["red", "\[ScriptCapitalR]"], "[", "c", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(red \[ScriptCapitalG] set)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1924053921", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649651622538926*^9, 3.649651640030013*^9}},
 CellTags->{
  "ID:1924053921", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "red \[ScriptCapitalG] set", "IR"},
 CellID->1924053921],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["\[Tilde]", "\[ScriptCapitalG]"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["\[Tilde]", "\[ScriptCapitalR]"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Tilde] \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1439528152", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.650185327682371*^9, 3.650185350714218*^9}},
 CellTags->{
  "ID:1439528152", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Tilde] \[ScriptCapitalG]", "IR"},
 CellID->1439528152],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1542915564]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->304038122],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DomainSets \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DomainSets \[ScriptCapitalG])\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["S", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["S", "ExpressionVariable"], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["S", "ExpressionVariable"], "]"}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["S", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["S", "ExpressionVariable"], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSets", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["S", "ExpressionVariable"], "]"}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1148013153"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1148013153-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1148013153-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1148013153", "Proof|ID:1148013153-1"},
 CellID->1967681402],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1148013153-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2032932287]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DomainTuples \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DomainTuples \[ScriptCapitalG])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["T", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["T", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["T", "ExpressionVariable"], "]"}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1668771272"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1668771272-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1668771272-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1668771272", "Proof|ID:1668771272-1"},
 CellID->525775183],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1668771272-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->397952331]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Sum] \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Sum] \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["f", "ExpressionVariable"]}, {
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox[
              SubscriptBox["\[Sum]", 
               RowBox[{"GroebnerRing", "[", 
                 StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
              
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   StyleBox["n", "ExpressionVariable"]}]}}]], 
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "[", 
               StyleBox["i", "ExpressionVariable"], "]"}]}], "\[Equal]", 
           RowBox[{
             UnderscriptBox[
              SubscriptBox["\[Sum]", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   StyleBox["n", "ExpressionVariable"]}]}}]], 
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "[", 
               StyleBox["i", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["f", "ExpressionVariable"]}, {
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox[
              SubscriptBox["\[Sum]", 
               RowBox[{"GroebnerRing", "[", 
                 StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
              
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   StyleBox["n", "ExpressionVariable"]}]}}]], 
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "[", 
               StyleBox["i", "ExpressionVariable"], "]"}]}], "\[Equal]", 
           RowBox[{
             UnderscriptBox[
              SubscriptBox["\[Sum]", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   StyleBox["n", "ExpressionVariable"]}]}}]], 
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "[", 
               StyleBox["i", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1928426019"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1928426019-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1928426019-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1928426019", "Proof|ID:1928426019-1"},
 CellID->33419350],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1928426019-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->944826184]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reduceM \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reduceM \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["T", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["reduceM", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["reduceM", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:581321955"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p581321955-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p581321955-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:581321955", "Proof|ID:581321955-1"},
 CellID->1728877054],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p581321955-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->596281154]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[RightArrow] \[ScriptCapitalG] multiplier)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[RightArrow] \[ScriptCapitalG] multiplier)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             StyleBox["m", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             StyleBox["c", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"(", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Rule]", 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}]], 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 StyleBox["b", "ExpressionVariable"]}], ")"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Rule]", 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}]], 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 StyleBox["b", "ExpressionVariable"]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             StyleBox["m", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             StyleBox["c", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"(", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Rule]", 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}]], 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 StyleBox["b", "ExpressionVariable"]}], ")"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Rule]", 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}]], 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 StyleBox["b", "ExpressionVariable"]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:405682622"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p405682622-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p405682622-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:405682622", "Proof|ID:405682622-1"},
 CellID->109823534],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p405682622-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1039693341]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[RightArrow] \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[RightArrow] \[ScriptCapitalG])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             StyleBox["c", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"(", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Rule]", 
                   RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 StyleBox["b", "ExpressionVariable"]}], ")"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Rule]", 
                   RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 StyleBox["b", "ExpressionVariable"]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             StyleBox["c", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"(", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Rule]", 
                   RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 StyleBox["b", "ExpressionVariable"]}], ")"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[Rule]", 
                   RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 StyleBox["b", "ExpressionVariable"]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1165811909"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1165811909-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1165811909-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1165811909", "Proof|ID:1165811909-1"},
 CellID->679771556],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1165811909-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->483984180]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[LeftRightArrow] \[ScriptCapitalG])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[LeftRightArrow] \[ScriptCapitalG])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             StyleBox["c", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"(", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[LeftRightArrow]", 
                   RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 StyleBox["b", "ExpressionVariable"]}], ")"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[LeftRightArrow]", 
                   RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 StyleBox["b", "ExpressionVariable"]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             StyleBox["c", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"(", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[LeftRightArrow]", 
                   RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 StyleBox["b", "ExpressionVariable"]}], ")"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{"(", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox[
                  SubscriptBox["\[LeftRightArrow]", 
                   RowBox[{"{", 
                    StyleBox["c", "ExpressionVariable"], "}"}]], 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
                 StyleBox["b", "ExpressionVariable"]}], 
               ")"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:306221528"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p306221528-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p306221528-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:306221528", "Proof|ID:306221528-1"},
 CellID->52377820],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p306221528-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->543843663]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible \[ScriptCapitalG] multiplier)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible \[ScriptCapitalG] multiplier)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             StyleBox["m", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             StyleBox["m", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:113746156"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p113746156-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p113746156-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:113746156", "Proof|ID:113746156-1"},
 CellID->1673450432],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p113746156-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1955298365]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1372373678"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1372373678-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1372373678-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1372373678", "Proof|ID:1372373678-1"},
 CellID->1812050497],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1372373678-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2059822108]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(symred \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(symred \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             StyleBox["m", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["symred", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["symred", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             StyleBox["m", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["symred", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["symred", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:12812135"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p12812135-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p12812135-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:12812135", "Proof|ID:12812135-1"},
 CellID->2042538189],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p12812135-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1790190522]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(areCorrelative \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(areCorrelative \[ScriptCapitalG])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areCorrelative", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["areCorrelative", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areCorrelative", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["areCorrelative", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:168109679"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p168109679-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p168109679-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:168109679", "Proof|ID:168109679-1"},
 CellID->1279405920],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p168109679-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2104251234]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isNTCR \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isNTCR \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isNTCR", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isNTCR", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isNTCR", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isNTCR", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1427340407"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1427340407-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1427340407-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1427340407", "Proof|ID:1427340407-1"},
 CellID->1645359805],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1427340407-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->688496565]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMNTCR \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMNTCR \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMNTCR", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isMNTCR", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMNTCR", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isMNTCR", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:587977465"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p587977465-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p587977465-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:587977465", "Proof|ID:587977465-1"},
 CellID->429530329],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p587977465-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->34258095]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isNTCR \[ScriptCapitalG] unary)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isNTCR \[ScriptCapitalG] unary)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isNTCR", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isNTCR", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isNTCR", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isNTCR", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:950195558"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p950195558-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p950195558-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:950195558", "Proof|ID:950195558-1"},
 CellID->1606573343],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p950195558-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->187923333]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMNTCR \[ScriptCapitalG] unary)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMNTCR \[ScriptCapitalG] unary)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMNTCR", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isMNTCR", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMNTCR", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isMNTCR", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2144197915"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2144197915-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2144197915-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2144197915", "Proof|ID:2144197915-1"},
 CellID->874871898],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2144197915-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1569130461]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(red \[ScriptCapitalG] set)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(red \[ScriptCapitalG] set)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["c", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainSets", "[", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
                "]"}]], "[", 
             StyleBox["c", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["red", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["c", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["red", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["c", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainSets", "[", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
                "]"}]], "[", 
             StyleBox["c", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["red", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["c", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["red", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], "[", 
               StyleBox["c", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1924053921"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1924053921-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1924053921-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1924053921", "Proof|ID:1924053921-1"},
 CellID->23710079],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1924053921-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->52275310]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Tilde] \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Tilde] \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Tilde]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             StyleBox["b", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Tilde]", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["b", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Tilde]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             StyleBox["b", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Tilde]", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"]], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1439528152"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1439528152-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1439528152-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1439528152", "Proof|ID:1439528152-1"},
 CellID->1503880702],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1439528152-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1443984090]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Main Result (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.649584330013981*^9, 3.649584338795059*^9}, {
  3.649649474557115*^9, 3.64964947828494*^9}},
 CellID->57707614],

Cell["", "OpenEnvironment",
 CellID->770279207],

Cell[CellGroupData[{

Cell["\<\
Lemma: \[ScriptCapitalG] Satisfies all Reduction Ring Axioms\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6495837230200653`*^9, 3.649583738129737*^9}},
 CellID->634165676],

Cell[BoxData[
 RowBox[{"isPartIrreflOrder", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalG]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalG]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isPartIrreflOrder \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1674560446", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649583868391027*^9, 3.649583874472446*^9}},
 CellTags->{
  "ID:1674560446", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isPartIrreflOrder \[ScriptCapitalG]"},
 CellID->1674560446],

Cell[BoxData[
 RowBox[{"isNoetherian", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalG]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalG]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isNoetherian \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:167203289", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64958388239186*^9, 3.649583885802244*^9}},
 CellTags->{
  "ID:167203289", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isNoetherian \[ScriptCapitalG]"},
 CellID->167203289],

Cell[BoxData[
 RowBox[{"Multipliers", "[", "\[ScriptCapitalG]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Multipliers \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:569855140", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649583761619521*^9, 3.649583786325988*^9}},
 CellTags->{
  "ID:569855140", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "Multipliers \[ScriptCapitalG]"},
 CellID->569855140],

Cell[BoxData[
 RowBox[{"R0", "[", "\[ScriptCapitalG]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R0 \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:180032137", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649583898319715*^9, 3.649583900274131*^9}},
 CellTags->{
  "ID:180032137", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R0 \[ScriptCapitalG]"},
 CellID->180032137],

Cell[BoxData[
 RowBox[{"R1", "[", "\[ScriptCapitalG]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R1 \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:428662253", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649583913758506*^9, 3.6495839206281357`*^9}},
 CellTags->{
  "ID:428662253", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R1 \[ScriptCapitalG]"},
 CellID->428662253],

Cell[BoxData[
 RowBox[{"R2", "[", "\[ScriptCapitalG]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R2 \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:530735576", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6495839146456842`*^9, 3.64958392176947*^9}},
 CellTags->{
  "ID:530735576", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R2 \[ScriptCapitalG]"},
 CellID->530735576],

Cell[BoxData[
 RowBox[{"R3", "[", "\[ScriptCapitalG]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R3 \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1081126646", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649583915280944*^9, 3.649583922726552*^9}},
 CellTags->{
  "ID:1081126646", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R3 \[ScriptCapitalG]"},
 CellID->1081126646],

Cell[BoxData[
 RowBox[{"R4", "[", "\[ScriptCapitalG]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R4 \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1591716135", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6495839159468*^9, 3.64958392357417*^9}},
 CellTags->{
  "ID:1591716135", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R4 \[ScriptCapitalG]"},
 CellID->1591716135],

Cell[BoxData[
 RowBox[{"R5", "[", "\[ScriptCapitalG]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R5 \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:960649057", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649583916574894*^9, 3.649583924826133*^9}},
 CellTags->{
  "ID:960649057", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R5 \[ScriptCapitalG]"},
 CellID->960649057],

Cell[BoxData[
 RowBox[{"R6", "[", "\[ScriptCapitalG]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R6 \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:227638759", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6495839171666393`*^9, 3.6495839256364517`*^9}},
 CellTags->{
  "ID:227638759", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R6 \[ScriptCapitalG]"},
 CellID->227638759],

Cell[BoxData[
 RowBox[{"R7", "[", "\[ScriptCapitalG]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R7 \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2067143528", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649583917795322*^9, 3.6495839266305656`*^9}},
 CellTags->{
  "ID:2067143528", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R7 \[ScriptCapitalG]"},
 CellID->2067143528],

Cell[BoxData[
 RowBox[{"R8", "[", "\[ScriptCapitalG]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R8 \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1699723496", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649583918639182*^9, 3.6495839276985817`*^9}},
 CellTags->{
  "ID:1699723496", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R8 \[ScriptCapitalG]"},
 CellID->1699723496],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1268327292]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1372217053],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isPartIrreflOrder \[ScriptCapitalG])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isPartIrreflOrder \[ScriptCapitalG])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isPartIrreflOrder", "[", 
           RowBox[{
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             ",", 
             RowBox[{"GroebnerRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isPartIrreflOrder", "[", 
           RowBox[{
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             ",", 
             RowBox[{"GroebnerRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1674560446"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1674560446-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1674560446-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1674560446", "Proof|ID:1674560446-1"},
 CellID->997501851],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1674560446-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{3.649584767064315*^9},
 CellID->1684644361]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isNoetherian \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isNoetherian \[ScriptCapitalG])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isNoetherian", "[", 
           RowBox[{
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             ",", 
             RowBox[{"GroebnerRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isNoetherian", "[", 
           RowBox[{
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             ",", 
             RowBox[{"GroebnerRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:167203289"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p167203289-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p167203289-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:167203289", "Proof|ID:167203289-1"},
 CellID->24119846],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p167203289-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1782458510]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Multipliers \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Multipliers \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"Multipliers", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"Multipliers", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:569855140"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p569855140-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p569855140-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:569855140", "Proof|ID:569855140-1"},
 CellID->1448471252],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p569855140-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1587438225]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R0 \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R0 \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R0", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R0", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:180032137"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p180032137-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p180032137-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:180032137", "Proof|ID:180032137-1"},
 CellID->270496012],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p180032137-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->943115763]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R1 \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R1 \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R1", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R1", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:428662253"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p428662253-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p428662253-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:428662253", "Proof|ID:428662253-1"},
 CellID->1471451373],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p428662253-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1601646706]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R2 \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R2 \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R2", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R2", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:530735576"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p530735576-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p530735576-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:530735576", "Proof|ID:530735576-1"},
 CellID->457832475],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p530735576-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{3.649586205236094*^9},
 CellID->1408840479]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R3 \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R3 \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R3", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R3", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1081126646"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1081126646-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1081126646-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1081126646", "Proof|ID:1081126646-1"},
 CellID->345429782],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1081126646-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1414623488]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R4 \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R4 \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R4", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R4", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1591716135"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1591716135-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1591716135-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1591716135", "Proof|ID:1591716135-1"},
 CellID->2122576129],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1591716135-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1462233688]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R5 \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R5 \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R5", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R5", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:960649057"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p960649057-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p960649057-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:960649057", "Proof|ID:960649057-1"},
 CellID->1652070226],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p960649057-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1035512283]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R6 \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R6 \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R6", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R6", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:227638759"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p227638759-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p227638759-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:227638759", "Proof|ID:227638759-1"},
 CellID->300639375],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p227638759-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->822053261]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R7 \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R7 \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R7", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R7", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2067143528"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2067143528-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2067143528-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2067143528", "Proof|ID:2067143528-1"},
 CellID->271136742],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2067143528-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->1516945570]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R8 \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R8 \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R8", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R8", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1699723496"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1699723496-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1699723496-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1699723496", "Proof|ID:1699723496-1"},
 CellID->1455152719],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1699723496-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1812443882]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->693241098],

Cell[CellGroupData[{

Cell["Theorem: \[ScriptCapitalG] is Reduction Ring", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6494909024993277`*^9, 3.649490904820784*^9}},
 CellID->1607583592],

Cell[BoxData[
 RowBox[{"isReductionRing", "[", "\[ScriptCapitalG]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalG] isReductionRing)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2054051596", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649490909835929*^9, 3.6494909203175573`*^9}, {
  3.6495837515745087`*^9, 3.649583780937685*^9}},
 CellTags->{
  "ID:2054051596", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isReductionRing"},
 CellID->2054051596],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1928703596]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1784515901],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[ScriptCapitalG] isReductionRing)\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(\[ScriptCapitalG] isReductionRing)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isReductionRing", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isReductionRing", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2054051596"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2054051596-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2054051596-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2054051596", "Proof|ID:2054051596-1"},
 CellID->1699885972],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2054051596-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1787222986]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Gr\[ODoubleDot]bner Rings of Algorithmic Reduction Rings are Algorithmic \
Reduction Rings (\[Checkmark])\
\>", "Subsection",
 CellChangeTimes->{{3.6491361687449713`*^9, 3.649136191284404*^9}, {
  3.649652566505192*^9, 3.6496525703823223`*^9}},
 CellID->428105198],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "isAlgoReductionRing", "[", "\[ScriptCapitalR]", 
   "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1031954878}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649583745457139*^9, 3.649583749538539*^9}, {
  3.649649906527032*^9, 3.649649907304862*^9}},
 CellTags->{
  "ID:1031954878", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1031954878],

Cell["", "Text",
 CellID->2043725198],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"\[ScriptCapitalG]", "=", 
   RowBox[{
   "GroebnerRing", "[", "\[ScriptCapitalR]", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 660860156}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649583773886771*^9, 3.6495837835491056`*^9}},
 CellTags->{
  "ID:660860156", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->660860156],

Cell["", "OpenEnvironment",
 CellID->57540137],

Cell[CellGroupData[{

Cell["\<\
Lemma: \[ScriptCapitalG] Satisfies all Axioms of Algorithmic Reduction Rings\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6496496053509703`*^9, 3.649649626696354*^9}},
 CellID->1770996009],

Cell[BoxData[
 RowBox[{"R9", "[", "\[ScriptCapitalG]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R9 \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:939639226", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649649705049707*^9, 3.64964970980896*^9}},
 CellTags->{
  "ID:939639226", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R9 \[ScriptCapitalG]"},
 CellID->939639226],

Cell[BoxData[
 RowBox[{"R10", "[", "\[ScriptCapitalG]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R10 \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:37694441", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649649705714438*^9, 3.6496497115381403`*^9}},
 CellTags->{
  "ID:37694441", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R10 \[ScriptCapitalG]"},
 CellID->37694441],

Cell[BoxData[
 RowBox[{"R11", "[", "\[ScriptCapitalG]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R11 \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:576375692", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6496497063501987`*^9, 3.649649712778902*^9}},
 CellTags->{
  "ID:576375692", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R11 \[ScriptCapitalG]"},
 CellID->576375692],

Cell[BoxData[
 RowBox[{"R12", "[", "\[ScriptCapitalG]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R12 \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1226510984", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649649707008462*^9, 3.649649713846966*^9}},
 CellTags->{
  "ID:1226510984", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R12 \[ScriptCapitalG]"},
 CellID->1226510984],

Cell[BoxData[
 RowBox[{"R13", "[", "\[ScriptCapitalG]", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R13 \[ScriptCapitalG])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1781039403", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6496497076735764`*^9, 3.649649715101719*^9}},
 CellTags->{
  "ID:1781039403", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R13 \[ScriptCapitalG]"},
 CellID->1781039403],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->20576105]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1461086965],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R9 \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R9 \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R9", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R9", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:939639226"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p939639226-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p939639226-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:939639226", "Proof|ID:939639226-1"},
 CellID->1140411407],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p939639226-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1138027217]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R10 \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R10 \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R10", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R10", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:37694441"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p37694441-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p37694441-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:37694441", "Proof|ID:37694441-1"},
 CellID->1985690430],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p37694441-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->266751900]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R11 \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R11 \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R11", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R11", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:576375692"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p576375692-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p576375692-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:576375692", "Proof|ID:576375692-1"},
 CellID->1030879810],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p576375692-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->854397617]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R12 \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R12 \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R12", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R12", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1226510984"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1226510984-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1226510984-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1226510984", "Proof|ID:1226510984-1"},
 CellID->1342766186],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1226510984-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->352367246]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R13 \[ScriptCapitalG])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R13 \[ScriptCapitalG])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R13", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"R13", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1781039403"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1781039403-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1781039403-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1781039403", "Proof|ID:1781039403-1"},
 CellID->234828850],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1781039403-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->4013719]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->495058079],

Cell[CellGroupData[{

Cell["\<\
Theorem: \[ScriptCapitalG] is Domain and Algorithmic Reduction Ring\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.592038636712653*^9, 3.592038644229086*^9}, {
  3.648287204762354*^9, 3.648287207090493*^9}, {3.648287632737198*^9, 
  3.648287636061654*^9}, {3.6567607737492585`*^9, 3.656760776744464*^9}},
 CellID->280489971],

Cell[BoxData[
 RowBox[{"isDomain", "[", "\[ScriptCapitalG]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalG] isDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:166003593", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6567598684134665`*^9, 3.656759873093475*^9}},
 CellTags->{
  "ID:166003593", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isDomain"},
 CellID->166003593],

Cell[BoxData[
 RowBox[{"isAlgoReductionRing", "[", "\[ScriptCapitalG]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalG] isAlgoReductionRing)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1167967104", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.592038649805953*^9, 3.592038654809882*^9}, {
  3.648287625167223*^9, 3.6482876294589663`*^9}, {3.649136212588141*^9, 
  3.649136223268352*^9}, {3.649649596004368*^9, 3.649649598553174*^9}},
 CellTags->{
  "ID:1167967104", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isAlgoReductionRing"},
 CellID->1167967104],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1939621786]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->96131666],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[ScriptCapitalG] isDomain)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[ScriptCapitalG] isDomain)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isDomain", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isDomain", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:166003593"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p166003593-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p166003593-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:166003593", "Proof|ID:166003593-1"},
 CellID->324790663],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p166003593-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->780913280]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[ScriptCapitalG] isAlgoReductionRing)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[ScriptCapitalG] isAlgoReductionRing)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isAlgoReductionRing", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isAlgoReductionRing", "[", 
           RowBox[{"GroebnerRing", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1167967104"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1167967104-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1167967104-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1167967104", "Proof|ID:1167967104-1"},
 CellID->1160302337],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1167967104-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1073405309]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Theorem: Correctness of Auxiliary Functions (\[Checkmark])\
\>", "Section",
 CellChangeTimes->{{3.649049530592454*^9, 3.6490495394669228`*^9}, {
  3.649581172998035*^9, 3.6495811767444897`*^9}},
 CellID->789808767],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "isAlgoReductionRing", "[", "\[ScriptCapitalR]", 
   "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 2087655986}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.649049549598763*^9},
 CellTags->{
  "ID:2087655986", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->2087655986],

Cell["", "Text",
 CellID->850461231],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"\[ScriptCapitalG]", "=", 
   RowBox[{
   "GroebnerRing", "[", "\[ScriptCapitalR]", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 955133330}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.6490495593636303`*^9},
 CellTags->{
  "ID:955133330", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->955133330],

Cell[CellGroupData[{

Cell["rd, rdT & trd (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6482872435761223`*^9, 3.648287243953402*^9}, {
  3.648287841400806*^9, 3.648287843171938*^9}, {3.648290458467927*^9, 
  3.648290458728138*^9}, {3.649501258621811*^9, 3.6495012608383827`*^9}, {
  3.649504389604821*^9, 3.64950439316046*^9}},
 CellID->767117005],

Cell[CellGroupData[{

Cell["Properties of rd (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.649140807921864*^9, 3.6491408110632477`*^9}, {
  3.649415996348549*^9, 3.649416000308515*^9}},
 CellID->1079351161],

Cell["", "OpenEnvironment",
 CellID->689062295],

Cell[CellGroupData[{

Cell["Lemma: Properties of rd", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.649414594042631*^9, 3.649414596746483*^9}},
 CellID->2101062054],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
   RowBox[{"[", 
    RowBox[{"x", ",", "y"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 977980890}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6494146069916763`*^9, 3.649414650193687*^9}},
 CellTags->{
  "ID:977980890", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->977980890],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"z", "=", 
   RowBox[{
    UnderscriptBox["rd", "\[ScriptCapitalG]"], "[", 
    RowBox[{"x", ",", "y"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1748570587}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649414630428619*^9, 3.6494146551374817`*^9}, {
  3.649415035756579*^9, 3.649415037565723*^9}},
 CellTags->{
  "ID:1748570587", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1748570587],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
  RowBox[{"[", "z", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership rd)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:699276041", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649414659591115*^9, 3.649414665985291*^9}},
 CellTags->{
  "ID:699276041", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership rd"},
 CellID->699276041],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalG]"], "[", 
   RowBox[{"x", ",", "y"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", 
      RowBox[{"{", "y", "}"}]], "\[ScriptCapitalG]"], "z"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reducible rd)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:867263181", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649414732825232*^9, 3.649414751531933*^9}, {
  3.649414943448641*^9, 3.6494149528150454`*^9}, {3.661851996951408*^9, 
  3.6618519995098124`*^9}},
 CellTags->{
  "ID:867263181", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "reducible rd"},
 CellID->867263181],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Not]", 
    RowBox[{
     UnderscriptBox["isReducible", "\[ScriptCapitalG]"], "[", 
     RowBox[{"x", ",", "y"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"z", "\[Equal]", "x"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(irreducible rd)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1431447619", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649414714307765*^9, 3.649414726950214*^9}, {
  3.649414915611474*^9, 3.649414940390338*^9}},
 CellTags->{
  "ID:1431447619", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "irreducible rd"},
 CellID->1431447619],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalG]"], "z"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalG]"], "[", 
   RowBox[{"x", ",", "y"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Succeeds] rd)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1490981815", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649415570185904*^9, 3.649415587821615*^9}},
 CellTags->{
  "ID:1490981815", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Succeeds] rd"},
 CellID->1490981815],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->61218476]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1894724465],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership rd)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership rd)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"GroebnerRing", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           "[", 
           RowBox[{
             UnderscriptBox["rd", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"GroebnerRing", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           "[", 
           RowBox[{
             UnderscriptBox["rd", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:699276041"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p699276041-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p699276041-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:699276041", "Proof|ID:699276041-1"},
 CellID->141722599],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p699276041-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->687345439]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reducible rd)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reducible rd)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               RowBox[{"{", 
                 StyleBox["y", "ExpressionVariable"], "}"}]], 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             RowBox[{
               UnderscriptBox["rd", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               RowBox[{"{", 
                 StyleBox["y", "ExpressionVariable"], "}"}]], 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             RowBox[{
               UnderscriptBox["rd", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:867263181"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p867263181-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p867263181-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:867263181", "Proof|ID:867263181-1"},
 CellID->231570427],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p867263181-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->9502323]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(irreducible rd)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(irreducible rd)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["rd", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             StyleBox[
             "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["isReducible", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["rd", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             StyleBox[
             "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1431447619"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1431447619-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1431447619-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1431447619", "Proof|ID:1431447619-1"},
 CellID->1892915845],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1431447619-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1652171802]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Succeeds] rd)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Succeeds] rd)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             RowBox[{
               UnderscriptBox["rd", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             RowBox[{
               UnderscriptBox["rd", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1490981815"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1490981815-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1490981815-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1490981815", "Proof|ID:1490981815-1"},
 CellID->642445226],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1490981815-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->379929766]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Properties of rdT (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.649140800187357*^9, 3.649140803870407*^9}, {
  3.649495418654203*^9, 3.649495419093399*^9}, {3.649501229578828*^9, 
  3.649501233412159*^9}},
 CellID->1481458283],

Cell["", "OpenEnvironment",
 CellID->1984772789],

Cell[CellGroupData[{

Cell["Definition: rdTBase1, rdTBase2, rdTRec", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.649489003954275*^9, 3.6494890165425663`*^9}, {
  3.649495424617282*^9, 3.649495444878903*^9}},
 CellID->777068689],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"Q", ",", "x", ",", "C"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 243233209}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649489025890895*^9, 3.64948902835375*^9}},
 CellTags->{
  "ID:243233209", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->243233209],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["rdTBase1", "\[ScriptCapitalG]"], "[", 
   RowBox[{"Q", ",", "x", ",", "C"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
     RowBox[{"i", ">", 
      RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
    RowBox[{"Q", "[", "i", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(rdTBase1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:974939989", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649489032259221*^9, 3.649489056194591*^9}, {
   3.649489087220467*^9, 3.6494890890022573`*^9}, {3.649495451320136*^9, 
   3.649495452949689*^9}, 3.6494962436065207`*^9},
 CellTags->{
  "ID:974939989", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdTBase1"},
 CellID->974939989],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["rdTBase2", "\[ScriptCapitalG]"], "[", 
   RowBox[{"Q", ",", "x", ",", "C"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"\[Not]", 
          RowBox[{"i", ">", 
           RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]}], 
         TagBox[")",
          "AutoParentheses"]}], "\[And]", 
        RowBox[{"x", 
         UnderscriptBox["\[Succeeds]", "\[ScriptCapitalG]"], 
         RowBox[{
          UnderscriptBox["rd", "\[ScriptCapitalG]"], "[", 
          RowBox[{"x", ",", 
           SubscriptBox["C", "i"]}], "]"}]}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[Implies]", 
      RowBox[{"Q", "[", "i", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(rdTBase2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1014318336", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649489061183864*^9, 3.649489092765766*^9}, {
   3.649493226269697*^9, 3.649493238247983*^9}, {3.6494954560055847`*^9, 
   3.649495457635394*^9}, {3.6494955170492764`*^9, 3.64949551730477*^9}, 
   3.6494962463664637`*^9},
 CellTags->{
  "ID:1014318336", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdTBase2"},
 CellID->1014318336],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["rdTRec", "\[ScriptCapitalG]"], "[", 
   RowBox[{"Q", ",", "x", ",", "C"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"\[Not]", 
          RowBox[{"i", ">", 
           RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]}], 
         TagBox[")",
          "AutoParentheses"]}], "\[And]", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"\[Not]", 
          RowBox[{"x", 
           UnderscriptBox["\[Succeeds]", "\[ScriptCapitalG]"], 
           RowBox[{
            UnderscriptBox["rd", "\[ScriptCapitalG]"], "[", 
            RowBox[{"x", ",", 
             SubscriptBox["C", "i"]}], "]"}]}]}], 
         TagBox[")",
          "AutoParentheses"]}], "\[And]", 
        RowBox[{"Q", "[", 
         RowBox[{"i", "+", "1"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[Implies]", 
      RowBox[{"Q", "[", "i", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(rdTRec)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:107497168", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649489083937497*^9, 3.649489106231461*^9}, {
  3.649493244447679*^9, 3.6494932545961018`*^9}, {3.649495459917575*^9, 
  3.649495461469407*^9}, {3.649496248133994*^9, 3.649496250119527*^9}},
 CellTags->{
  "ID:107497168", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdTRec"},
 CellID->107497168],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1555492801]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1152308520],

Cell["", "OpenEnvironment",
 CellID->1118809324],

Cell[CellGroupData[{

Cell["Lemma: Induction Rule for rdT", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.649488109607807*^9, 3.649488115528118*^9}, 
   3.649494353257667*^9, {3.649495500265779*^9, 3.64949550070938*^9}, {
   3.661852031770669*^9, 3.6618520324570704`*^9}},
 CellID->61842186],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
    RowBox[{"[", "x", "]"}]}], ",", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
    RowBox[{"[", "C", "]"}]}], ",", "Q"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 839458854}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649488399801593*^9, 3.649488400805118*^9}, {
  3.649496126678619*^9, 3.649496141627182*^9}, {3.649496827895253*^9, 
  3.649496832287697*^9}},
 CellTags->{
  "ID:839458854", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->839458854],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox[
          UnderscriptBox["\[ForAll]", 
           RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
          RowBox[{"i", ">", 
           RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
         
         RowBox[{"Q", "[", "i", "]"}]}]},
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
         
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{"x", 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalG]"], 
             RowBox[{
              UnderscriptBox["rd", "\[ScriptCapitalG]"], "[", 
              RowBox[{"x", ",", 
               SubscriptBox["C", "i"]}], "]"}]}], 
            TagBox[")",
             "AutoParentheses"]}], "\[Implies]", 
           RowBox[{"Q", "[", "i", "]"}]}], 
          TagBox[")",
           "AutoParentheses"]}]}]},
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
         
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{"\[Not]", 
               RowBox[{"x", 
                UnderscriptBox["\[Succeeds]", "\[ScriptCapitalG]"], 
                RowBox[{
                 UnderscriptBox["rd", "\[ScriptCapitalG]"], "[", 
                 RowBox[{"x", ",", 
                  SubscriptBox["C", "i"]}], "]"}]}]}], 
              TagBox[")",
               "AutoParentheses"]}], "\[And]", 
             RowBox[{"Q", "[", 
              RowBox[{"i", "+", "1"}], "]"}]}], 
            TagBox[")",
             "AutoParentheses"]}], "\[Implies]", 
           RowBox[{"Q", "[", "i", "]"}]}], 
          TagBox[")",
           "AutoParentheses"]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{"i", "\[Element]", 
      SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{"j", "\[Element]", 
       SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{
          RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], 
          "\[Equal]", "j"}], "\[And]", 
         RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]}], 
        TagBox[")",
         "AutoParentheses"]}], "\[Implies]", 
       RowBox[{"Q", "[", "i", "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1436407557", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649488868934737*^9, 3.649488888778269*^9}, {
  3.6494891905884447`*^9, 3.649489196478567*^9}, {3.6494892683364553`*^9, 
  3.6494892802167683`*^9}, {3.649489578240676*^9, 3.649489628465835*^9}, {
  3.649491219144045*^9, 3.649491224462866*^9}, {3.649491962095607*^9, 
  3.6494919747678757`*^9}, {3.649495549781068*^9, 3.6494955500492*^9}, {
  3.6494958501362343`*^9, 3.649495882533833*^9}, {3.649496149186172*^9, 
  3.649496190847515*^9}},
 CellTags->{
  "ID:1436407557", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "4"},
 CellID->1436407557],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["rdTBase1", "\[ScriptCapitalG]"], "[", 
         RowBox[{"Q", ",", "x", ",", "C"}], "]"}]},
       {
        RowBox[{
         UnderscriptBox["rdTBase2", "\[ScriptCapitalG]"], "[", 
         RowBox[{"Q", ",", "x", ",", "C"}], "]"}]},
       {
        RowBox[{
         UnderscriptBox["rdTRec", "\[ScriptCapitalG]"], "[", 
         RowBox[{"Q", ",", "x", ",", "C"}], "]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{"i", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
    RowBox[{"Q", "[", "i", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(rdT induction)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2098894842", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6494885806769342`*^9, 3.64948866505118*^9}, {
  3.649489117048843*^9, 3.649489125323048*^9}, {3.6494892535307217`*^9, 
  3.649489264900292*^9}, {3.6494955595390368`*^9, 3.6494955687688713`*^9}, {
  3.649496196585382*^9, 3.649496240327798*^9}},
 CellTags->{
  "ID:2098894842", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdT induction"},
 CellID->2098894842],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->687243513]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->463103006],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             StyleBox["Q", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
                RowBox[{
                  StyleBox["i", "ExpressionVariable"], ">", 
                  RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]], 
               RowBox[{
                 StyleBox["Q", "ExpressionVariable"], "[", 
                 StyleBox["i", "ExpressionVariable"], "]"}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                   RowBox[{
                    UnderscriptBox["rd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["Q", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["rd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["Q", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[Element]", 
                    SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "\[Equal]", 
                    StyleBox["j", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["Q", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             StyleBox["Q", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
                RowBox[{
                  StyleBox["i", "ExpressionVariable"], ">", 
                  RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]], 
               RowBox[{
                 StyleBox["Q", "ExpressionVariable"], "[", 
                 StyleBox["i", "ExpressionVariable"], "]"}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                   RowBox[{
                    UnderscriptBox["rd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["Q", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["rd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["Q", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[Element]", 
                    SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "\[Equal]", 
                    StyleBox["j", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["Q", "ExpressionVariable"], "[", 
                   StyleBox["i", "ExpressionVariable"], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1436407557"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1436407557-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1436407557-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1436407557", "Proof|ID:1436407557-1"},
 CellID->1376449459],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1436407557-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->792869139]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(rdT induction)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(rdT induction)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             StyleBox["Q", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["rdTBase1", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["Q", "ExpressionVariable"], ",", 
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["C", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["rdTBase2", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["Q", "ExpressionVariable"], ",", 
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["C", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["rdTRec", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["Q", "ExpressionVariable"], ",", 
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["C", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               StyleBox["Q", "ExpressionVariable"], "[", 
               StyleBox["i", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             StyleBox["Q", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["rdTBase1", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["Q", "ExpressionVariable"], ",", 
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["C", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["rdTBase2", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["Q", "ExpressionVariable"], ",", 
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["C", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["rdTRec", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["Q", "ExpressionVariable"], ",", 
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["C", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[Element]", 
                   "\[DoubleStruckCapitalN]"}]}}]], 
             RowBox[{
               StyleBox["Q", "ExpressionVariable"], "[", 
               StyleBox["i", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2098894842"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2098894842-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2098894842-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2098894842", "Proof|ID:2098894842-1"},
 CellID->1924563166],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2098894842-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2060479915]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->2066597320],

Cell[CellGroupData[{

Cell["Lemma: Alternative Description of rdT", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6494881478734703`*^9, 3.6494881561511087`*^9}, 
   3.6494943506817007`*^9, {3.649496554717108*^9, 3.649496555242135*^9}},
 CellID->422282357],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"x", ",", "C", ",", "i"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["rdT", "\[ScriptCapitalG]"], "[", 
     RowBox[{"x", ",", "C", ",", "i"}], "]"}], "\[Equal]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Piecewise]", GridBox[{
        {"x", "\[DoubleLeftArrow]", 
         RowBox[{"i", ">", 
          RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]},
        {
         RowBox[{
          UnderscriptBox["rd", "\[ScriptCapitalG]"], "[", 
          RowBox[{"x", ",", 
           SubscriptBox["C", "i"]}], "]"}], "\[DoubleLeftArrow]", 
         RowBox[{"x", 
          UnderscriptBox["\[Succeeds]", "\[ScriptCapitalG]"], 
          RowBox[{
           UnderscriptBox["rd", "\[ScriptCapitalG]"], "[", 
           RowBox[{"x", ",", 
            SubscriptBox["C", "i"]}], "]"}]}]},
        {
         RowBox[{
          UnderscriptBox["rdT", "\[ScriptCapitalG]"], "[", 
          RowBox[{"x", ",", "C", ",", 
           RowBox[{"i", "+", "1"}]}], "]"}], "\[DoubleLeftArrow]", "otherwise"}
       }]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(rdT alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:160961870", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6494881705489483`*^9, 3.6494882653255167`*^9}, {
  3.649488497684524*^9, 3.649488505651424*^9}, {3.649495597577969*^9, 
  3.649495608326531*^9}},
 CellTags->{
  "ID:160961870", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdT alternative", "IR"},
 CellID->160961870],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->347215208]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->60793602],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(rdT alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(rdT alternative)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             StyleBox["C", "ExpressionVariable"]}, {
             StyleBox["i", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["rdT", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["C", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{"\[Piecewise]", 
             GridBox[{{
                StyleBox["x", "ExpressionVariable"], "\[DoubleLeftArrow]", 
                RowBox[{
                  StyleBox["i", "ExpressionVariable"], ">", 
                  RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                RowBox[{
                  UnderscriptBox["rd", 
                   RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}], 
                "\[DoubleLeftArrow]", 
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], 
                  UnderscriptBox["\[Succeeds]", 
                   RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                  RowBox[{
                    UnderscriptBox["rd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}]}, {
                RowBox[{
                  UnderscriptBox["rdT", 
                   RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]}], "]"}], 
                "\[DoubleLeftArrow]", 
                "True"}}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             StyleBox["C", "ExpressionVariable"]}, {
             StyleBox["i", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["rdT", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["C", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{"\[Piecewise]", 
             GridBox[{{
                StyleBox["x", "ExpressionVariable"], "\[DoubleLeftArrow]", 
                RowBox[{
                  StyleBox["i", "ExpressionVariable"], ">", 
                  RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                RowBox[{
                  UnderscriptBox["rd", 
                   RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}], 
                "\[DoubleLeftArrow]", 
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], 
                  UnderscriptBox["\[Succeeds]", 
                   RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                  RowBox[{
                    UnderscriptBox["rd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}]}, {
                RowBox[{
                  UnderscriptBox["rdT", 
                   RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}]}], "]"}], 
                "\[DoubleLeftArrow]", 
                "True"}}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:160961870"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p160961870-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p160961870-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:160961870", "Proof|ID:160961870-1"},
 CellID->1503881909],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p160961870-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1283045382]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1367402977],

Cell[CellGroupData[{

Cell["Theorem: Properties of rdT", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6494940254689083`*^9, 3.649494031622435*^9}, {
  3.649495618070856*^9, 3.649495621997849*^9}},
 CellID->1990167663],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
    RowBox[{"[", "x", "]"}]}], ",", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
    RowBox[{"[", "C", "]"}]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 168345814}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64949463003113*^9, 3.649494637251851*^9}, {
  3.6495013925168467`*^9, 3.649501392886013*^9}},
 CellTags->{
  "ID:168345814", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->168345814],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "i", "\[Element]", "\[DoubleStruckCapitalN]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1634249728}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.649494652462694*^9},
 CellTags->{
  "ID:1634249728", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1634249728],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"y", "=", 
   RowBox[{
    UnderscriptBox["rdT", "\[ScriptCapitalG]"], "[", 
    RowBox[{"x", ",", "C", ",", "i"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 512270654}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649494657436282*^9, 3.649494664574581*^9}, {
  3.64949562505724*^9, 3.649495625573783*^9}},
 CellTags->{
  "ID:512270654", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->512270654],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
  RowBox[{"[", "y", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership rdT)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2105002699", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649494048099573*^9, 3.6494940703282547`*^9}, {
  3.649494642933016*^9, 3.649494662730872*^9}},
 CellTags->{
  "ID:2105002699", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership rdT"},
 CellID->2105002699],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[Exists]", 
      RowBox[{
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
     RowBox[{"j", "\[GreaterEqual]", "i"}]], 
    RowBox[{
     UnderscriptBox["isReducible", "\[ScriptCapitalG]"], "[", 
     RowBox[{"x", ",", 
      SubscriptBox["C", "j"]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", 
    UnderscriptBox[
     SubscriptBox["\[Rule]", "C"], "\[ScriptCapitalG]"], "y"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(reducible rdT)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1505002987", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649498338436418*^9, 3.649498358192528*^9}},
 CellTags->{
  "ID:1505002987", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "reducible rdT"},
 CellID->1505002987],

Cell[BoxData[
 RowBox[{
  RowBox[{"\[Not]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[Exists]", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
      RowBox[{"j", "\[GreaterEqual]", "i"}]], 
     RowBox[{
      UnderscriptBox["isReducible", "\[ScriptCapitalG]"], "[", 
      RowBox[{"x", ",", 
       SubscriptBox["C", "j"]}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}], "\[Implies]", 
  RowBox[{"y", "\[Equal]", "x"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(irreducible rdT)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:514228174", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64949836290654*^9, 3.64949837502607*^9}},
 CellTags->{
  "ID:514228174", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "irreducible rdT"},
 CellID->514228174],

Cell[BoxData[
 RowBox[{"x", 
  UnderscriptBox[
   SubsuperscriptBox["\[Rule]", "C", "*"], "\[ScriptCapitalG]"], 
  "y"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[RightArrow]* rdT)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:208729427", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6494946934415007`*^9, 3.649494703319659*^9}},
 CellTags->{
  "ID:208729427", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow]* rdT"},
 CellID->208729427],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalG]"], "y"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[Exists]", 
      RowBox[{
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
     RowBox[{"j", "\[GreaterEqual]", "i"}]], 
    RowBox[{
     UnderscriptBox["isReducible", "\[ScriptCapitalG]"], "[", 
     RowBox[{"x", ",", 
      SubscriptBox["C", "j"]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Succeeds] rdT)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:278468994", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649494795718233*^9, 3.64949479932408*^9}, {
  3.649494832042161*^9, 3.649494885335971*^9}, {3.64949565822147*^9, 
  3.6494957183460007`*^9}},
 CellTags->{
  "ID:278468994", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Succeeds] rdT"},
 CellID->278468994],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1542263296]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->365655572],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership rdT)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership rdT)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"GroebnerRing", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           "[", 
           RowBox[{
             UnderscriptBox["rdT", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["C", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"GroebnerRing", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           "[", 
           RowBox[{
             UnderscriptBox["rdT", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["C", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2105002699"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2105002699-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2105002699-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2105002699", "Proof|ID:2105002699-1"},
 CellID->871739323],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2105002699-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->793373954]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(reducible rdT)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(reducible rdT)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
              RowBox[{
                StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                StyleBox["i", "ExpressionVariable"]}]], 
             RowBox[{
               UnderscriptBox["isReducible", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 SubscriptBox[
                  StyleBox["C", "ExpressionVariable"], 
                  StyleBox["j", "ExpressionVariable"]]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               StyleBox["C", "ExpressionVariable"]], 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             RowBox[{
               UnderscriptBox["rdT", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["C", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
              RowBox[{
                StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                StyleBox["i", "ExpressionVariable"]}]], 
             RowBox[{
               UnderscriptBox["isReducible", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 SubscriptBox[
                  StyleBox["C", "ExpressionVariable"], 
                  StyleBox["j", "ExpressionVariable"]]}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox[
              SubscriptBox["\[Rule]", 
               StyleBox["C", "ExpressionVariable"]], 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             RowBox[{
               UnderscriptBox["rdT", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["C", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1505002987"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1505002987-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1505002987-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1505002987", "Proof|ID:1505002987-1"},
 CellID->327721344],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1505002987-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->272053484]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(irreducible rdT)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(irreducible rdT)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[Exists]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["i", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox["isReducible", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}]}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["rdT", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["C", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             StyleBox[
             "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           RowBox[{"\[Not]", 
             RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[Exists]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["i", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox["isReducible", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}]}], 
               TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["rdT", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["C", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             StyleBox[
             "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:514228174"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p514228174-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p514228174-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:514228174", "Proof|ID:514228174-1"},
 CellID->1831820924],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p514228174-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1587171133]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[RightArrow]* rdT)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[RightArrow]* rdT)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           StyleBox["x", "ExpressionVariable"], 
           UnderscriptBox[
            TemplateBox[{"\[Rule]", 
              StyleBox["C", "ExpressionVariable"], "*"}, "Subsuperscript", 
             SyntaxForm -> SubsuperscriptBox], 
            RowBox[{"GroebnerRing", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           RowBox[{
             UnderscriptBox["rdT", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["C", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           StyleBox["x", "ExpressionVariable"], 
           UnderscriptBox[
            TemplateBox[{"\[Rule]", 
              StyleBox["C", "ExpressionVariable"], "*"}, "Subsuperscript", 
             SyntaxForm -> SubsuperscriptBox], 
            RowBox[{"GroebnerRing", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           RowBox[{
             UnderscriptBox["rdT", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["C", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:208729427"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p208729427-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p208729427-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:208729427", "Proof|ID:208729427-1"},
 CellID->970814565],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p208729427-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->10817078]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Succeeds] rdT)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Succeeds] rdT)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             RowBox[{
               UnderscriptBox["rdT", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["C", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"]}], "]"}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
              RowBox[{
                StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                StyleBox["i", "ExpressionVariable"]}]], 
             RowBox[{
               UnderscriptBox["isReducible", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 SubscriptBox[
                  StyleBox["C", "ExpressionVariable"], 
                  StyleBox["j", "ExpressionVariable"]]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             RowBox[{
               UnderscriptBox["rdT", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["C", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"]}], "]"}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
              RowBox[{
                StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                StyleBox["i", "ExpressionVariable"]}]], 
             RowBox[{
               UnderscriptBox["isReducible", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 SubscriptBox[
                  StyleBox["C", "ExpressionVariable"], 
                  StyleBox["j", "ExpressionVariable"]]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:278468994"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p278468994-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p278468994-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:278468994", "Proof|ID:278468994-1"},
 CellID->120821588],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p278468994-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->401261460]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1171561374],

Cell[CellGroupData[{

Cell[TextData[{
 "Corollary: Properties of rdT for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "=", "1"}], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.649501710925831*^9, 3.649501722919757*^9}, {
  3.649503757822504*^9, 3.649503758083252*^9}},
 CellID->2018878554],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
     RowBox[{"[", "x", "]"}]}], ",", 
    RowBox[{
     UnderscriptBox["\[Element]", 
      RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
     RowBox[{"[", "C", "]"}]}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"x", 
      UnderscriptBox["\[Succeeds]", "\[ScriptCapitalG]"], 
      RowBox[{
       UnderscriptBox["rdT", "\[ScriptCapitalG]"], "[", 
       RowBox[{"x", ",", "C", ",", "1"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"x", "\[Element]", 
      RowBox[{
       UnderscriptBox["red", "\[ScriptCapitalG]"], "[", "C", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(red rdT)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:96175791", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649501769685473*^9, 3.64950185471982*^9}},
 CellTags->{
  "ID:96175791", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "red rdT"},
 CellID->96175791],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->632845748]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->568618794],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(red rdT)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(red rdT)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             RowBox[{
               UnderscriptBox["rdT", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["C", "ExpressionVariable"], ",", "1"}], "]"}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["red", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 StyleBox["C", "ExpressionVariable"], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             RowBox[{
               UnderscriptBox["rdT", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["C", "ExpressionVariable"], ",", "1"}], "]"}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Element]", 
               RowBox[{
                 UnderscriptBox["red", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 StyleBox["C", "ExpressionVariable"], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:96175791"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p96175791-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p96175791-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:96175791", "Proof|ID:96175791-1"},
 CellID->56129168],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p96175791-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->211899362]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Properties of trd (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.649140793490432*^9, 3.649140796062387*^9}, {
  3.649504381616991*^9, 3.649504385774609*^9}},
 CellID->25217793],

Cell["", "OpenEnvironment",
 CellID->583061016],

Cell[CellGroupData[{

Cell["Lemma: Induction Schema for trd", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6495014551928997`*^9, 3.649501460461907*^9}},
 CellID->967911099],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", 
      RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
     RowBox[{"[", "C", "]"}]}], ",", "Q"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
             RowBox[{"[", "x", "]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{
               RowBox[{"x", 
                UnderscriptBox["\[Succeeds]", "\[ScriptCapitalG]"], 
                RowBox[{
                 UnderscriptBox["rdT", "\[ScriptCapitalG]"], "[", 
                 RowBox[{"x", ",", "C", ",", "1"}], "]"}]}], "\[And]", 
               RowBox[{"Q", "[", 
                RowBox[{
                 UnderscriptBox["rdT", "\[ScriptCapitalG]"], "[", 
                 RowBox[{"x", ",", "C", ",", "1"}], "]"}], "]"}]}], 
              TagBox[")",
               "AutoParentheses"]}], "\[Implies]", 
             RowBox[{"Q", "[", "x", "]"}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
             RowBox[{"[", "x", "]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{"\[Not]", 
               RowBox[{"x", 
                UnderscriptBox["\[Succeeds]", "\[ScriptCapitalG]"], 
                RowBox[{
                 UnderscriptBox["rdT", "\[ScriptCapitalG]"], "[", 
                 RowBox[{"x", ",", "C", ",", "1"}], "]"}]}]}], 
              TagBox[")",
               "AutoParentheses"]}], "\[Implies]", 
             RowBox[{"Q", "[", "x", "]"}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
        RowBox[{"[", "x", "]"}]}]], 
      RowBox[{"Q", "[", "x", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(trd induction)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:909149002", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649501465757121*^9, 3.649501614703261*^9}},
 CellTags->{
  "ID:909149002", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "trd induction"},
 CellID->909149002],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1310698312]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1214519135],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(trd induction)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(trd induction)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             StyleBox["Q", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["rdT", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", "1"}], "]"}]}], 
                   "\[And]", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    UnderscriptBox["rdT", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", "1"}], "]"}], 
                    "]"}]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["Q", "ExpressionVariable"], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["rdT", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", "1"}], "]"}]}], 
                    
                    TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["Q", "ExpressionVariable"], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["Q", "ExpressionVariable"], "[", 
               StyleBox["x", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["C", "ExpressionVariable"], "]"}]}, {
             StyleBox["Q", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["rdT", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", "1"}], "]"}]}], 
                   "\[And]", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    UnderscriptBox["rdT", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", "1"}], "]"}], 
                    "]"}]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["Q", "ExpressionVariable"], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["rdT", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", "1"}], "]"}]}], 
                    
                    TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["Q", "ExpressionVariable"], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["Q", "ExpressionVariable"], "[", 
               StyleBox["x", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:909149002"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p909149002-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p909149002-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:909149002", "Proof|ID:909149002-1"},
 CellID->35905978],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p909149002-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->62162963]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->473680231],

Cell[CellGroupData[{

Cell["Theorem: Properties of trd", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648287291226268*^9, 3.648287295037788*^9}, {
  3.649501241986742*^9, 3.649501243474699*^9}},
 CellID->1909232321],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
    RowBox[{"[", "B", "]"}]}], ",", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
    RowBox[{"[", "x", "]"}]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1097836943}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648287364703764*^9, 3.648287371578514*^9}, {
  3.648287888379895*^9, 3.6482878976328897`*^9}, {3.6495032144336777`*^9, 
  3.649503218357135*^9}},
 CellTags->{
  "ID:1097836943", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1097836943],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"y", "=", 
   RowBox[{
    UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
    RowBox[{"x", ",", "B"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 206560148}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648287382587772*^9, 3.648287392295504*^9}},
 CellTags->{
  "ID:206560148", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->206560148],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
  RowBox[{"[", "y", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership trd)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1315173810", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648287397637533*^9, 3.648287403816825*^9}},
 CellTags->{
  "ID:1315173810", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership trd"},
 CellID->1315173810],

Cell[BoxData[
 RowBox[{"x", 
  UnderscriptBox[
   SubsuperscriptBox["\[Rule]", "B", "*"], "\[ScriptCapitalG]"], 
  "y"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[RightArrow]* trd)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1071214578", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648287416131401*^9, 3.6482874247803507`*^9}},
 CellTags->{
  "ID:1071214578", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow]* trd"},
 CellID->1071214578],

Cell[BoxData[
 RowBox[{"\[Not]", 
  RowBox[{"y", "\[Element]", 
   RowBox[{
    UnderscriptBox["red", "\[ScriptCapitalG]"], "[", "B", 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(trd \[NotElement] red)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:657218052", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648287433699761*^9, 3.64828745972101*^9}},
 CellTags->{
  "ID:657218052", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "trd \[NotElement] red"},
 CellID->657218052],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->672864406]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1965420673],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership trd)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership trd)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"GroebnerRing", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           "[", 
           RowBox[{
             UnderscriptBox["trd", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"GroebnerRing", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           "[", 
           RowBox[{
             UnderscriptBox["trd", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1315173810"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1315173810-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1315173810-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1315173810", "Proof|ID:1315173810-1"},
 CellID->56915654],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1315173810-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1118371038]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[RightArrow]* trd)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[RightArrow]* trd)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["x", "ExpressionVariable"], 
           UnderscriptBox[
            TemplateBox[{"\[Rule]", 
              StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
             SyntaxForm -> SubsuperscriptBox], 
            RowBox[{"GroebnerRing", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           RowBox[{
             UnderscriptBox["trd", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["x", "ExpressionVariable"], 
           UnderscriptBox[
            TemplateBox[{"\[Rule]", 
              StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
             SyntaxForm -> SubsuperscriptBox], 
            RowBox[{"GroebnerRing", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           RowBox[{
             UnderscriptBox["trd", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1071214578"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1071214578-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1071214578-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1071214578", "Proof|ID:1071214578-1"},
 CellID->204815329],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1071214578-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->387630611]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(trd \[NotElement] red)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(trd \[NotElement] red)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["trd", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["B", "ExpressionVariable"]}], "]"}], "\[Element]", 
               RowBox[{
                 UnderscriptBox["red", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}], 
             TagBox[")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["trd", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["B", "ExpressionVariable"]}], "]"}], "\[Element]", 
               RowBox[{
                 UnderscriptBox["red", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 StyleBox["B", "ExpressionVariable"], "]"}]}], 
             TagBox[
             ")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:657218052"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p657218052-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p657218052-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:657218052", "Proof|ID:657218052-1"},
 CellID->1364381745],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p657218052-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->659473488]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["cpmTuple & lmTuple (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.648287249003149*^9, 3.648287251645232*^9}, {
  3.649504703863977*^9, 3.649504705978282*^9}, {3.6495771637500772`*^9, 
  3.649577167422658*^9}},
 CellID->1574109816],

Cell[CellGroupData[{

Cell["\<\
Definition and Lemmas: MMPairs (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.648289473220914*^9, 3.6482894744121428`*^9}, {
  3.648290287161599*^9, 3.6482902896178427`*^9}, {3.649137745355567*^9, 
  3.649137754069605*^9}},
 CellID->715367204],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"\[ScriptCapitalD]", ",", "x", ",", "y"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 480907095}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648289036588304*^9, 3.648289041406444*^9}, {
   3.648289374318573*^9, 3.6482893761718407`*^9}, 3.648371175417143*^9, {
   3.6483722757873163`*^9, 3.64837227616118*^9}},
 CellTags->{
  "ID:480907095", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->480907095],

Cell["", "OpenEnvironment",
 CellID->1014570953],

Cell[CellGroupData[{

Cell["Definition: Domain of MNTCR-Multiplier-Pairs", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648288280763516*^9, 3.648288285089871*^9}, {
  3.648288895980811*^9, 3.6482889035196943`*^9}, {3.648371200932086*^9, 
  3.6483712406211576`*^9}, {3.648372281064569*^9, 3.648372285859482*^9}},
 CellID->1638005863],

Cell[BoxData[
 RowBox[{
  RowBox[{"MMPairs", "[", 
   RowBox[{"\[ScriptCapitalD]", ",", "x", ",", "y"}], "]"}], ":=", 
  UnderscriptBox["\[CapitalDelta]", "N"]}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1000930828}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648288944103046*^9, 3.648288960823629*^9}, {
   3.6482890450074244`*^9, 3.648289045749378*^9}, {3.648289378017729*^9, 
   3.64828937924704*^9}, 3.648371177926499*^9, {3.648371243908985*^9, 
   3.648371244859507*^9}, {3.648372272649798*^9, 3.648372272908317*^9}, {
   3.649136446111774*^9, 3.6491364466319427`*^9}},
 CellTags->{
  "ID:1000930828", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1000930828],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "z"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", "N"], 
      RowBox[{"[", "z", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"isTuple", "[", "z", "]"}]},
         {
          RowBox[{
           RowBox[{"\[LeftBracketingBar]", "z", "\[RightBracketingBar]"}], 
           "\[Equal]", "2"}]},
         {
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
           RowBox[{"[", 
            SubscriptBox["z", "1"], "]"}]}]},
         {
          RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", "\[ScriptCapitalD]", "]"}]], 
           RowBox[{"[", 
            SubscriptBox["z", "2"], "]"}]}]},
         {
          RowBox[{
           RowBox[{"\[LeftBracketingBar]", 
            SubscriptBox["z", "2"], "\[RightBracketingBar]"}], "\[Equal]", 
           "2"}]},
         {
          RowBox[{
           UnderscriptBox["\[Exists]", 
            RowBox[{"i", ",", 
             RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
              UnderscriptBox["multN", "\[ScriptCapitalD]"], "[", "x", 
              "]"}]}]], 
           RowBox[{
            UnderscriptBox["areCPMultipliers", "\[ScriptCapitalD]"], "[", 
            RowBox[{
             SubscriptBox[
              RowBox[{"(", 
               SubscriptBox["z", "2"], ")"}], "1"], ",", 
             SubscriptBox[
              RowBox[{"(", 
               SubscriptBox["z", "2"], ")"}], "2"], ",", "x", ",", "y", ",", 
             SubscriptBox["z", "1"], ",", "i", ",", "j"}], "]"}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(MMPairs \[Element])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1774407783", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64828897921805*^9, 3.648289011304599*^9}, {
   3.648289047853326*^9, 3.648289053751738*^9}, {3.648289091096932*^9, 
   3.648289221939014*^9}, 3.648289290842573*^9, {3.648289386148521*^9, 
   3.648289390844359*^9}, {3.6482896194962883`*^9, 3.6482896225425797`*^9}, {
   3.6482896739550247`*^9, 3.648289678033723*^9}, {3.648289731958438*^9, 
   3.6482897337042847`*^9}, {3.648371349733585*^9, 3.648371350069037*^9}, {
   3.648372288523469*^9, 3.648372293730843*^9}, {3.648883542356123*^9, 
   3.6488835461225443`*^9}, {3.64888589285956*^9, 3.648885897201105*^9}},
 CellTags->{
  "ID:1774407783", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element]"},
 CellID->1774407783],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->731225092]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->813407975],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"\[ScriptCapitalM]", "=", 
   RowBox[{"MMPairs", "[", 
    RowBox[{"\[ScriptCapitalD]", ",", "x", ",", "y"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 193942992}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648289566435585*^9, 3.6482895780140944`*^9}, {
  3.6482896634589987`*^9, 3.6482896647158413`*^9}, {3.648290236717498*^9, 
  3.6482902384631844`*^9}, {3.648371382479793*^9, 3.6483713940917263`*^9}, {
  3.648372304607876*^9, 3.648372313227274*^9}},
 CellTags->{
  "ID:193942992", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->193942992],

Cell["", "OpenEnvironment",
 CellID->1112882839],

Cell[CellGroupData[{

Cell["\<\
Lemma: Alternative Descriptions of \[Element] for MMPairs\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6482895218947477`*^9, 3.648289527961623*^9}, {
  3.648289886343557*^9, 3.6482898971692743`*^9}, {3.6483713981451883`*^9, 
  3.6483714114782667`*^9}, {3.648964279027513*^9, 3.648964298164603*^9}},
 CellID->1753178796],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalM]"], 
    RowBox[{"[", "z", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"ProdDomain", "[", 
     RowBox[{"\[ScriptCapitalD]", ",", 
      RowBox[{"DomainTuples", "[", "\[ScriptCapitalD]", "]"}]}], "]"}]], 
   RowBox[{"[", "z", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(MMPairs \[Element] ProdDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:44826735", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661662002387803*^9, 3.661662032090255*^9}, {
  3.6616621913197346`*^9, 3.6616621960153427`*^9}},
 CellTags->{
  "ID:44826735", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] ProdDomain"},
 CellID->44826735],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalM]", "]"}]], 
    RowBox[{"[", "z", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", 
     RowBox[{"ProdDomain", "[", 
      RowBox[{"\[ScriptCapitalD]", ",", 
       RowBox[{"DomainTuples", "[", "\[ScriptCapitalD]", "]"}]}], "]"}], 
     "]"}]], 
   RowBox[{"[", "z", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(MMPairs \[Element] ProdDomain DomainTuples)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:155756485", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6616620426670732`*^9, 3.661662060139104*^9}, {
  3.6616622009761515`*^9, 3.661662205328559*^9}},
 CellTags->{
  "ID:155756485", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] ProdDomain DomainTuples"},
 CellID->155756485],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "z"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalM]"], 
      RowBox[{"[", "z", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
        RowBox[{"[", 
         RowBox[{"a", ",", "m1", ",", "m2"}], "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[And]", 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{"a", "\[Equal]", 
             SubscriptBox["z", "1"]}]},
           {
            RowBox[{"m1", "\[Equal]", 
             SubscriptBox[
              RowBox[{"(", 
               SubscriptBox["z", "2"], ")"}], "1"]}]},
           {
            RowBox[{"m2", "\[Equal]", 
             SubscriptBox[
              RowBox[{"(", 
               SubscriptBox["z", "2"], ")"}], "2"]}]},
           {
            RowBox[{"z", "\[Equal]", 
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"a", ",", 
               RowBox[{"\[LeftAngleBracket]", 
                RowBox[{"m1", ",", "m2"}], "\[RightAngleBracket]"}]}], 
              "\[RightAngleBracket]"}]}]},
           {
            RowBox[{
             UnderscriptBox["\[Exists]", 
              RowBox[{"i", ",", 
               RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
               RowBox[{
                UnderscriptBox["multN", "\[ScriptCapitalD]"], "[", "x", 
                "]"}]}]], 
             RowBox[{
              UnderscriptBox["areCPMultipliers", "\[ScriptCapitalD]"], "[", 
              RowBox[{
              "m1", ",", "m2", ",", "x", ",", "y", ",", "a", ",", "i", ",", 
               "j"}], "]"}]}]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(MMPairs \[Element] alternative I)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1019394084", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648289796218246*^9, 3.6482898056500483`*^9}, {
  3.6482900183066*^9, 3.648290089677733*^9}, {3.6482902418258877`*^9, 
  3.648290254215626*^9}, {3.64837141455124*^9, 3.648371415962041*^9}, {
  3.648371490922429*^9, 3.6483714913263693`*^9}, {3.648372328232237*^9, 
  3.648372338952961*^9}, {3.648883552391817*^9, 3.648883554165332*^9}, {
  3.648885910969905*^9, 3.6488859135132627`*^9}},
 CellTags->{
  "ID:1019394084", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] alternative I", "IR"},
 CellID->1019394084],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"a", ",", "m1", ",", "m2"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalM]"], 
      RowBox[{"[", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"a", ",", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"m1", ",", "m2"}], "\[RightAngleBracket]"}]}], 
        "\[RightAngleBracket]"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
           RowBox[{"[", "a", "]"}]}]},
         {
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
           RowBox[{"[", "m1", "]"}]}]},
         {
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
           RowBox[{"[", "m2", "]"}]}]},
         {
          RowBox[{
           UnderscriptBox["\[Exists]", 
            RowBox[{"i", ",", 
             RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
              UnderscriptBox["multN", "\[ScriptCapitalD]"], "[", "x", 
              "]"}]}]], 
           RowBox[{
            UnderscriptBox["areCPMultipliers", "\[ScriptCapitalD]"], "[", 
            RowBox[{
            "m1", ",", "m2", ",", "x", ",", "y", ",", "a", ",", "i", ",", 
             "j"}], "]"}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(MMPairs \[Element] alternative II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1458538511", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648289551502146*^9, 3.648289607438637*^9}, {
   3.6482896691554937`*^9, 3.648289729080203*^9}, 3.6482897866253366`*^9, {
   3.64829025734799*^9, 3.648290274018618*^9}, {3.6483714863422737`*^9, 
   3.648371487605693*^9}, {3.648372331121264*^9, 3.648372332380221*^9}, {
   3.648883558938706*^9, 3.6488835606334963`*^9}, {3.648885916867549*^9, 
   3.648885919297824*^9}},
 CellTags->{
  "ID:1458538511", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] alternative II", "IR"},
 CellID->1458538511],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "z"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalM]"], 
      RowBox[{"[", "z", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
        RowBox[{"[", 
         RowBox[{"a", ",", "m1", ",", "m2"}], "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[And]", 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{"a", "\[Equal]", 
             SubscriptBox["z", "1"]}]},
           {
            RowBox[{"m1", "\[Equal]", 
             SubscriptBox[
              RowBox[{"(", 
               SubscriptBox["z", "2"], ")"}], "1"]}]},
           {
            RowBox[{"m2", "\[Equal]", 
             SubscriptBox[
              RowBox[{"(", 
               SubscriptBox["z", "2"], ")"}], "2"]}]},
           {
            RowBox[{"z", "\[Equal]", 
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"a", ",", 
               RowBox[{"\[LeftAngleBracket]", 
                RowBox[{"m1", ",", "m2"}], "\[RightAngleBracket]"}]}], 
              "\[RightAngleBracket]"}]}]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(MMPairs \[Element] alternative III)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:81000274", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648465578940819*^9, 3.6484655889367943`*^9}},
 CellTags->{
  "ID:81000274", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] alternative III"},
 CellID->81000274],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->30776632]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1853514680],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(MMPairs \[Element] ProdDomain)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(MMPairs \[Element] ProdDomain)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             StyleBox["y", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"MMPairs", "[", 
                  RowBox[{
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                    
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ProdDomain", "[", 
              RowBox[{
                StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], 
                  "]"}]}], "]"}]], "[", 
           StyleBox["z", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             StyleBox["y", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"MMPairs", "[", 
                  RowBox[{
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                    
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ProdDomain", "[", 
              RowBox[{
                StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                RowBox[{"DomainTuples", "[", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], 
                  "]"}]}], "]"}]], "[", 
           StyleBox["z", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:44826735"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p44826735-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p44826735-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:44826735", "Proof|ID:44826735-1"},
 CellID->33075853],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p44826735-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->151639289]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(MMPairs \[Element] ProdDomain DomainTuples)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(MMPairs \[Element] ProdDomain DomainTuples)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             StyleBox["y", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                    
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", 
              RowBox[{"ProdDomain", "[", 
                RowBox[{
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                  RowBox[{"DomainTuples", "[", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], 
                    "]"}]}], "]"}], "]"}]], "[", 
           StyleBox["z", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             StyleBox["y", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                    
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", 
              RowBox[{"ProdDomain", "[", 
                RowBox[{
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                  RowBox[{"DomainTuples", "[", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], 
                    "]"}]}], "]"}], "]"}]], "[", 
           StyleBox["z", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:155756485"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p155756485-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p155756485-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:155756485", "Proof|ID:155756485-1"},
 CellID->228098881],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p155756485-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->255646132]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(MMPairs \[Element] alternative I)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(MMPairs \[Element] alternative I)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
            StyleBox["x", "ExpressionVariable"], ",", 
            StyleBox["y", "ExpressionVariable"], ",", 
            StyleBox["z", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"MMPairs", "[", 
                  RowBox[{
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                    
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                 SubscriptBox[
                  StyleBox["z", "ExpressionVariable"], "1"]}], "\[And]", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], "\[Equal]", 
                 SubscriptBox[
                  SubscriptBox[
                   StyleBox["z", "ExpressionVariable"], "2"], "1"]}], 
               "\[And]", 
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], "\[Equal]", 
                 SubscriptBox[
                  SubscriptBox[
                   StyleBox["z", "ExpressionVariable"], "2"], "2"]}], 
               "\[And]", 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["areCPMultipliers", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
            StyleBox["x", "ExpressionVariable"], ",", 
            StyleBox["y", "ExpressionVariable"], ",", 
            StyleBox["z", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"MMPairs", "[", 
                  RowBox[{
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                    
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                 SubscriptBox[
                  StyleBox["z", "ExpressionVariable"], "1"]}], "\[And]", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], "\[Equal]", 
                 SubscriptBox[
                  SubscriptBox[
                   StyleBox["z", "ExpressionVariable"], "2"], "1"]}], 
               "\[And]", 
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], "\[Equal]", 
                 SubscriptBox[
                  SubscriptBox[
                   StyleBox["z", "ExpressionVariable"], "2"], "2"]}], 
               "\[And]", 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["areCPMultipliers", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1019394084"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1019394084-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1019394084-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1019394084", "Proof|ID:1019394084-1"},
 CellID->679633020],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1019394084-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1919531878]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(MMPairs \[Element] alternative II)\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(MMPairs \[Element] alternative II)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
            StyleBox["x", "ExpressionVariable"], ",", 
            StyleBox["y", "ExpressionVariable"], ",", 
            StyleBox["a", "ExpressionVariable"], ",", 
            StyleBox["m1", "ExpressionVariable"], ",", 
            StyleBox["m2", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"MMPairs", "[", 
                  RowBox[{
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                    
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
               "]"}], ")"}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                 StyleBox["a", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                 StyleBox["m1", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                 StyleBox["m2", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["areCPMultipliers", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
            StyleBox["x", "ExpressionVariable"], ",", 
            StyleBox["y", "ExpressionVariable"], ",", 
            StyleBox["a", "ExpressionVariable"], ",", 
            StyleBox["m1", "ExpressionVariable"], ",", 
            StyleBox["m2", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"MMPairs", "[", 
                  RowBox[{
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                    
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
               "]"}], ")"}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                 StyleBox["a", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                 StyleBox["m1", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                 StyleBox["m2", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["areCPMultipliers", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1458538511"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1458538511-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1458538511-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1458538511", "Proof|ID:1458538511-1"},
 CellID->1939038746],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1458538511-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1078322293]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(MMPairs \[Element] alternative III)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(MMPairs \[Element] alternative III)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
            StyleBox["x", "ExpressionVariable"], ",", 
            StyleBox["y", "ExpressionVariable"], ",", 
            StyleBox["z", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"MMPairs", "[", 
                RowBox[{
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                  StyleBox["x", "ExpressionVariable"], ",", 
                  StyleBox["y", "ExpressionVariable"]}], "]"}]], "[", 
             StyleBox["z", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                 SubscriptBox[
                  StyleBox["z", "ExpressionVariable"], "1"]}], "\[And]", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], "\[Equal]", 
                 SubscriptBox[
                  SubscriptBox[
                   StyleBox["z", "ExpressionVariable"], "2"], "1"]}], 
               "\[And]", 
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], "\[Equal]", 
                 SubscriptBox[
                  SubscriptBox[
                   StyleBox["z", "ExpressionVariable"], "2"], "2"]}], 
               "\[And]", 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], 
                   "\[RightAngleBracket]"}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
            StyleBox["x", "ExpressionVariable"], ",", 
            StyleBox["y", "ExpressionVariable"], ",", 
            StyleBox["z", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"MMPairs", "[", 
                RowBox[{
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                  StyleBox["x", "ExpressionVariable"], ",", 
                  StyleBox["y", "ExpressionVariable"]}], "]"}]], "[", 
             StyleBox["z", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                 SubscriptBox[
                  StyleBox["z", "ExpressionVariable"], "1"]}], "\[And]", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], "\[Equal]", 
                 SubscriptBox[
                  SubscriptBox[
                   StyleBox["z", "ExpressionVariable"], "2"], "1"]}], 
               "\[And]", 
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], "\[Equal]", 
                 SubscriptBox[
                  SubscriptBox[
                   StyleBox["z", "ExpressionVariable"], "2"], "2"]}], 
               "\[And]", 
               RowBox[{
                 StyleBox["z", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], 
                   "\[RightAngleBracket]"}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:81000274"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p81000274-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p81000274-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:81000274", "Proof|ID:81000274-1"},
 CellID->637304832],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p81000274-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1448112406]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->591758444],

Cell[CellGroupData[{

Cell["Lemma: Shape of Tuples of MMPairs", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648291078025511*^9, 3.648291097605349*^9}, {
   3.648291329688858*^9, 3.648291335232222*^9}, {3.648291567808166*^9, 
   3.648291569696764*^9}, 3.6483715942406588`*^9, {3.648964266823621*^9, 
   3.648964267046389*^9}},
 CellID->273179543],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalM]", "]"}]], 
    RowBox[{"[", "T", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Or]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"T", "\[Equal]", 
         RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]},
       {
        RowBox[{
         UnderscriptBox["\[Exists]", 
          RowBox[{
           RowBox[{
            UnderscriptBox["\[Element]", "\[ScriptCapitalD]"], 
            RowBox[{"[", 
             RowBox[{"a", ",", "m1", ",", "m2"}], "]"}]}], ",", 
           RowBox[{"t", "..."}]}]], 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"\[And]", 
           RowBox[{"\[Piecewise]", GridBox[{
              {
               RowBox[{
                UnderscriptBox["\[Element]", "\[ScriptCapitalM]"], 
                RowBox[{"[", 
                 RowBox[{"\[LeftAngleBracket]", 
                  RowBox[{"a", ",", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{"m1", ",", "m2"}], "\[RightAngleBracket]"}]}], 
                  "\[RightAngleBracket]"}], "]"}]}]},
              {
               RowBox[{
                UnderscriptBox["\[Element]", 
                 RowBox[{"DomainTuples", "[", "\[ScriptCapitalM]", "]"}]], 
                RowBox[{"[", 
                 RowBox[{"\[LeftAngleBracket]", 
                  RowBox[{"t", "..."}], "\[RightAngleBracket]"}], "]"}]}]},
              {
               RowBox[{"T", "\[Equal]", 
                RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                  RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{"a", ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{"m1", ",", "m2"}], "\[RightAngleBracket]"}]}], 
                   "\[RightAngleBracket]"}], ",", 
                  RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}]}]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
          TagBox[")",
           "AutoParentheses"]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DomainTuples NPairs shape)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1140506798", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648291107587175*^9, 3.648291206385516*^9}, {
  3.648291246861165*^9, 3.648291267998356*^9}, {3.6483715964628553`*^9, 
  3.648371608349247*^9}, {3.648372380601595*^9, 3.648372388641426*^9}},
 CellTags->{
  "ID:1140506798", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NPairs shape"},
 CellID->1140506798],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1529545215]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1737102059],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DomainTuples NPairs shape)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DomainTuples NPairs shape)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             StyleBox["y", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                    
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["T", "ExpressionVariable"], "\[Equal]", 
             RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
           "\[Or]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}, {
                 StyleBox[
                  RowBox[{"t", "..."}], "ExpressionVariable"]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                    
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
                 "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainTuples", "[", 
                    RowBox[{"MMPairs", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                    
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"], 
                   "\[RightAngleBracket]"}], "]"}], "\[And]", 
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}, {
             StyleBox["x", "ExpressionVariable"]}, {
             StyleBox["y", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                    
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
               StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["T", "ExpressionVariable"], "\[Equal]", 
             RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
           "\[Or]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}, {
                 StyleBox[
                  RowBox[{"t", "..."}], "ExpressionVariable"]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                    
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
                 "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainTuples", "[", 
                    RowBox[{"MMPairs", "[", 
                    RowBox[{
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                    
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"], 
                   "\[RightAngleBracket]"}], "]"}], "\[And]", 
               RowBox[{
                 StyleBox["T", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1140506798"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1140506798-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1140506798-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1140506798", "Proof|ID:1140506798-1"},
 CellID->344640348],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1140506798-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->601798179]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Properties of cpmTuple (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.649504715245604*^9, 3.649504718974131*^9}, {
  3.6495615756535597`*^9, 3.6495615796473722`*^9}, {3.6495677203283777`*^9, 
  3.64956772128797*^9}, {3.649572373269287*^9, 3.649572375862732*^9}},
 CellID->330309293],

Cell["", "OpenEnvironment",
 CellID->1405519764],

Cell[CellGroupData[{

Cell["Lemma: Properties of cpmTuple with Two Arguments", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6495663684031076`*^9, 3.649566371624662*^9}},
 CellID->1810578538],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
    RowBox[{"[", 
     RowBox[{"x", ",", "y"}], "]"}]}]], 
  RowBox[{
   RowBox[{"x", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalG]"]}], "\[And]", 
   RowBox[{"y", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalG]"]}], "\[And]", 
   RowBox[{"x", "\[NotEqual]", "y"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1527335579}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6495676452457457`*^9, 3.6495676722245207`*^9}},
 CellTags->{
  "ID:1527335579", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1527335579],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   RowBox[{"DomainTuples", "[", 
    RowBox[{"MMPairs", "[", 
     RowBox[{"\[ScriptCapitalG]", ",", "x", ",", "y"}], "]"}], "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["cpmTuple", "\[ScriptCapitalG]"], "[", 
    RowBox[{"x", ",", "y"}], "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership cpmTuple 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1794762591", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649567679542677*^9, 3.649567684514145*^9}},
 CellTags->{
  "ID:1794762591", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership cpmTuple 1"},
 CellID->1794762591],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
     RowBox[{"[", "a", "]"}]}]], 
   RowBox[{
    UnderscriptBox["isMNTCR", "\[ScriptCapitalG]"], "[", 
    RowBox[{"x", ",", "y", ",", "a", ",", "1", ",", "1"}], "]"}]], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
     RowBox[{"[", 
      RowBox[{"b", ",", "mx", ",", "my"}], "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"a", 
       UnderscriptBox["\[Tilde]", "\[ScriptCapitalG]"], "b"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[And]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"b", ",", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"mx", ",", "my"}], "\[RightAngleBracket]"}]}], 
        "\[RightAngleBracket]"}], 
       TagBox["\:22ff",
        Identity,
        SyntaxForm->"a+b"], 
       RowBox[{
        UnderscriptBox["cpmTuple", "\[ScriptCapitalG]"], "[", 
        RowBox[{"x", ",", "y"}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[And]", 
     RowBox[{
      UnderscriptBox["areCPMultipliers", "\[ScriptCapitalG]"], "[", 
      RowBox[{
      "mx", ",", "my", ",", "x", ",", "y", ",", "b", ",", "1", ",", "1"}], 
      "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cpmTuple content 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1790156146", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649567693968985*^9, 3.6495677141391172`*^9}, 
   3.6495703542195883`*^9, {3.650188225474638*^9, 3.650188265457691*^9}},
 CellTags->{
  "ID:1790156146", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpmTuple content 1"},
 CellID->1790156146],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1107193959]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1694269586],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership cpmTuple 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership cpmTuple 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
                RowBox[{"MMPairs", "[", 
                  RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
             RowBox[{
               UnderscriptBox["cpmTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
                RowBox[{"MMPairs", "[", 
                  RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
             RowBox[{
               UnderscriptBox["cpmTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1794762591"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1794762591-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1794762591-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1794762591", "Proof|ID:1794762591-1"},
 CellID->1374603973],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1794762591-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->759652547]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cpmTuple content 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cpmTuple content 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1"}], 
               "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["mx", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["my", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["\[Tilde]", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["mx", "ExpressionVariable"], ",", 
                    StyleBox["my", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
                 "\:22ff", 
                 RowBox[{
                   UnderscriptBox["cpmTuple", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], "\[And]", 
               
               RowBox[{
                 UnderscriptBox["areCPMultipliers", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["mx", "ExpressionVariable"], ",", 
                   StyleBox["my", "ExpressionVariable"], ",", 
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["b", "ExpressionVariable"], ",", "1", ",", "1"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1"}], 
               "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["mx", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["my", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["\[Tilde]", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["mx", "ExpressionVariable"], ",", 
                    StyleBox["my", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
                 "\:22ff", 
                 RowBox[{
                   UnderscriptBox["cpmTuple", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], "\[And]", 
               
               RowBox[{
                 UnderscriptBox["areCPMultipliers", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["mx", "ExpressionVariable"], ",", 
                   StyleBox["my", "ExpressionVariable"], ",", 
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["b", "ExpressionVariable"], ",", "1", ",", "1"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1790156146"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1790156146-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1790156146-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1790156146", "Proof|ID:1790156146-1"},
 CellID->1785625847],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1790156146-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1110321225]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1055128601],

Cell[CellGroupData[{

Cell["Lemma: Properties of cpmTuple with Three Arguments", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.649559569000122*^9, 3.649559573765818*^9}, {
  3.649566349734375*^9, 3.649566354126457*^9}, {3.649567604494186*^9, 
  3.649567605274745*^9}},
 CellID->2110554202],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
    RowBox[{"[", "x", "]"}]}]], 
  RowBox[{"x", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalG]"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 884601666}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649559578430938*^9, 3.6495596034090433`*^9}, {
  3.649567607596136*^9, 3.649567609472474*^9}},
 CellTags->{
  "ID:884601666", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->884601666],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{
     UnderscriptBox["multN", "\[ScriptCapitalG]"], "[", "x", "]"}]}]], 
  RowBox[{"j", ">", "i"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1188084204}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64955962212295*^9, 3.6495596429642563`*^9}, {
  3.649560011425576*^9, 3.6495600130815687`*^9}},
 CellTags->{
  "ID:1188084204", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1188084204],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   RowBox[{"DomainTuples", "[", 
    RowBox[{"MMPairs", "[", 
     RowBox[{"\[ScriptCapitalG]", ",", "x", ",", "x"}], "]"}], "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["cpmTuple", "\[ScriptCapitalG]"], "[", 
    RowBox[{"x", ",", "i", ",", "j"}], "]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership cpmTuple 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1708769881", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6495596628027773`*^9, 3.649559677699349*^9}, 
   3.649566317812933*^9, {3.649567613887226*^9, 3.6495676186003847`*^9}},
 CellTags->{
  "ID:1708769881", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership cpmTuple 2"},
 CellID->1708769881],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
     RowBox[{"[", "a", "]"}]}]], 
   RowBox[{
    UnderscriptBox["isMNTCR", "\[ScriptCapitalG]"], "[", 
    RowBox[{"x", ",", "x", ",", "a", ",", "i", ",", "j"}], "]"}]], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
     RowBox[{"[", 
      RowBox[{"b", ",", "m1", ",", "m2"}], "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"a", 
       UnderscriptBox["\[Tilde]", "\[ScriptCapitalG]"], "b"}], 
      TagBox[")",
       "AutoParentheses"]}], "\[And]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"b", ",", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"m1", ",", "m2"}], "\[RightAngleBracket]"}]}], 
        "\[RightAngleBracket]"}], 
       TagBox["\:22ff",
        Identity,
        SyntaxForm->"a+b"], 
       RowBox[{
        UnderscriptBox["cpmTuple", "\[ScriptCapitalG]"], "[", 
        RowBox[{"x", ",", "i", ",", "j"}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[And]", 
     RowBox[{
      UnderscriptBox["areCPMultipliers", "\[ScriptCapitalG]"], "[", 
      RowBox[{
      "m1", ",", "m2", ",", "x", ",", "x", ",", "b", ",", "i", ",", "j"}], 
      "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cpmTuple content 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:125591800", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649559687689522*^9, 3.6495597656305103`*^9}, 
   3.649566319593309*^9, {3.649567615696044*^9, 3.649567631782394*^9}, {
   3.650190110835445*^9, 3.650190134154395*^9}},
 CellTags->{
  "ID:125591800", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpmTuple content 2"},
 CellID->125591800],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->91348497]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->201729108],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership cpmTuple 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership cpmTuple 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["j", "ExpressionVariable"], ">", 
               StyleBox["i", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
                RowBox[{"MMPairs", "[", 
                  RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
             RowBox[{
               UnderscriptBox["cpmTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["j", "ExpressionVariable"], ">", 
               StyleBox["i", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
                RowBox[{"MMPairs", "[", 
                  RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
             RowBox[{
               UnderscriptBox["cpmTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1708769881"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1708769881-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1708769881-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1708769881", "Proof|ID:1708769881-1"},
 CellID->1434737069],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1708769881-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1718897090]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cpmTuple content 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cpmTuple content 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["j", "ExpressionVariable"], ">", 
               StyleBox["i", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["\[Tilde]", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
                 "\:22ff", 
                 RowBox[{
                   UnderscriptBox["cpmTuple", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[And]", 
               
               RowBox[{
                 UnderscriptBox["areCPMultipliers", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["j", "ExpressionVariable"], ">", 
               StyleBox["i", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["\[Tilde]", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
                 "\:22ff", 
                 RowBox[{
                   UnderscriptBox["cpmTuple", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[And]", 
               
               RowBox[{
                 UnderscriptBox["areCPMultipliers", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:125591800"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p125591800-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p125591800-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:125591800", "Proof|ID:125591800-1"},
 CellID->1320574433],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p125591800-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->561308411]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Properties of lmTupleAux (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.648289971615705*^9, 3.648289976763379*^9}, {
  3.648371634322257*^9, 3.648371637052066*^9}, {3.6495647979934998`*^9, 
  3.6495648023004637`*^9}, {3.649567726254777*^9, 3.649567727251872*^9}, {
  3.649575490337186*^9, 3.649575493270914*^9}},
 CellID->356782994],

Cell["", "OpenEnvironment",
 CellID->1415256475],

Cell[CellGroupData[{

Cell["Lemma: Induction Schema for lmTupleAux", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.649561626906538*^9, 3.649561634610799*^9}, {
  3.649562048846133*^9, 3.649562052796047*^9}, {3.649562590758659*^9, 
  3.649562594398451*^9}},
 CellID->350414352],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"Q", ",", "n"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1122306642}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.6495616451781683`*^9, 3.6495617653674994`*^9, {3.6495618126886473`*^9, 
   3.649561812910289*^9}},
 CellTags->{
  "ID:1122306642", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1122306642],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"Q", "[", 
     RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], "]"}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{"i", ",", 
         RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]], 
       RowBox[{"j", ">", "i"}]], 
      RowBox[{
       UnderscriptBox[
        UnderscriptBox["\[ForAll]", 
         RowBox[{"r", "..."}]], 
        RowBox[{
         UnderscriptBox["\[Element]", 
          RowBox[{"DomainTuples", "[", 
           RowBox[{"NatPairs", "[", 
            RowBox[{"n", ",", 
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              TagBox["<",
               Identity], 
              TagBox[")",
               "AutoParentheses"]}]}], "]"}], "]"}]], 
         RowBox[{"[", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"r", "..."}], "\[RightAngleBracket]"}], "]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{"Q", "[", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"r", "..."}], "\[RightAngleBracket]"}], "]"}], 
         "\[Implies]", 
         RowBox[{"Q", "[", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"i", ",", "j"}], "\[RightAngleBracket]"}], ",", 
            RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], "]"}]}], 
        TagBox[")",
         "AutoParentheses"]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{"isTuple", "[", "C", "]"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[Element]", 
        RowBox[{"DomainTuples", "[", 
         RowBox[{"NatPairs", "[", 
          RowBox[{"n", ",", 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            TagBox["<",
             Identity], 
            TagBox[")",
             "AutoParentheses"]}]}], "]"}], "]"}]], 
       RowBox[{"[", "C", "]"}]}], "\[Implies]", 
      RowBox[{"Q", "[", "C", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(5)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:885588057", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649561768494605*^9, 3.6495619251292686`*^9}},
 CellTags->{
  "ID:885588057", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "5"},
 CellID->885588057],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"Q", "[", 
     RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], "]"}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{"i", ",", 
         RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]], 
       RowBox[{"j", ">", "i"}]], 
      RowBox[{
       UnderscriptBox[
        UnderscriptBox["\[ForAll]", 
         RowBox[{"r", "..."}]], 
        RowBox[{
         UnderscriptBox["\[Element]", 
          RowBox[{"DomainTuples", "[", 
           RowBox[{"NatPairs", "[", 
            RowBox[{"n", ",", 
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              TagBox["<",
               Identity], 
              TagBox[")",
               "AutoParentheses"]}]}], "]"}], "]"}]], 
         RowBox[{"[", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"r", "..."}], "\[RightAngleBracket]"}], "]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{"Q", "[", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"r", "..."}], "\[RightAngleBracket]"}], "]"}], 
         "\[Implies]", 
         RowBox[{"Q", "[", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"i", ",", "j"}], "\[RightAngleBracket]"}], ",", 
            RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], "]"}]}], 
        TagBox[")",
         "AutoParentheses"]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", 
        RowBox[{"NatPairs", "[", 
         RowBox[{"n", ",", 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           TagBox["<",
            Identity], 
           TagBox[")",
            "AutoParentheses"]}]}], "]"}], "]"}]], 
      RowBox[{"[", "C", "]"}]}]], 
    RowBox[{"Q", "[", "C", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lmTupleAux induction)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:438489157", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649561938796507*^9, 3.649561958487187*^9}},
 CellTags->{
  "ID:438489157", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTupleAux induction"},
 CellID->438489157],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->371522104]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1219086707],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(5)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(5)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["Q", "ExpressionVariable"], ",", 
            StyleBox["n", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["Q", "ExpressionVariable"], "[", 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], "]"}],
              "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]}}]], 
                RowBox[{
                  StyleBox["j", "ExpressionVariable"], ">", 
                  StyleBox["i", "ExpressionVariable"]}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   RowBox[{
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", "<"}], "]"}], 
                    "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}]], 
                 RowBox[{
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "]"}]}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isTuple", "[", 
                   StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", "<"}], "]"}], 
                    "]"}]], "[", 
                 StyleBox["C", "ExpressionVariable"], "]"}], "\[Implies]", 
               RowBox[{
                 StyleBox["Q", "ExpressionVariable"], "[", 
                 StyleBox["C", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["Q", "ExpressionVariable"], ",", 
            StyleBox["n", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["Q", "ExpressionVariable"], "[", 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], "]"}],
              "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]}}]], 
                RowBox[{
                  StyleBox["j", "ExpressionVariable"], ">", 
                  StyleBox["i", "ExpressionVariable"]}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   RowBox[{
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", "<"}], "]"}], 
                    "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}]], 
                 RowBox[{
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "]"}]}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isTuple", "[", 
                   StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", "<"}], "]"}], 
                    "]"}]], "[", 
                 StyleBox["C", "ExpressionVariable"], "]"}], "\[Implies]", 
               RowBox[{
                 StyleBox["Q", "ExpressionVariable"], "[", 
                 StyleBox["C", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:885588057"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p885588057-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p885588057-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:885588057", "Proof|ID:885588057-1"},
 CellID->1319009677],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p885588057-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1824975630]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lmTupleAux induction)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lmTupleAux induction)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["Q", "ExpressionVariable"], ",", 
            StyleBox["n", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["Q", "ExpressionVariable"], "[", 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], "]"}],
              "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]}}]], 
                RowBox[{
                  StyleBox["j", "ExpressionVariable"], ">", 
                  StyleBox["i", "ExpressionVariable"]}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   RowBox[{
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", "<"}], "]"}], 
                    "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}]], 
                 RowBox[{
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "]"}]}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", "<"}], "]"}], 
                    "]"}]], "[", 
                   StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["Q", "ExpressionVariable"], "[", 
               StyleBox["C", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["Q", "ExpressionVariable"], ",", 
            StyleBox["n", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["Q", "ExpressionVariable"], "[", 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], "]"}],
              "\[And]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]}}]], 
                RowBox[{
                  StyleBox["j", "ExpressionVariable"], ">", 
                  StyleBox["i", "ExpressionVariable"]}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   RowBox[{
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", "<"}], "]"}], 
                    "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}]], 
                 RowBox[{
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "]"}]}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", "<"}], "]"}], 
                    "]"}]], "[", 
                   StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["Q", "ExpressionVariable"], "[", 
               StyleBox["C", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:438489157"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p438489157-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p438489157-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:438489157", "Proof|ID:438489157-1"},
 CellID->1261187836],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p438489157-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1114401215]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->122261323],

Cell[CellGroupData[{

Cell["Theorem: Properties of lmTupleAux", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.649562792267437*^9, 3.649562797965267*^9}},
 CellID->609860341],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
    RowBox[{"[", "x", "]"}]}]], 
  RowBox[{"x", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalG]"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1795822474}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649562838714448*^9, 3.649562876664954*^9}},
 CellTags->{
  "ID:1795822474", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1795822474],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", 
     RowBox[{"NatPairs", "[", 
      RowBox[{
       RowBox[{
        UnderscriptBox["multN", "\[ScriptCapitalG]"], "[", "x", "]"}], ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["<",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}], "]"}], "]"}]], 
   RowBox[{"[", "P", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 2110280918}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649562882551968*^9, 3.649562908483432*^9}},
 CellTags->{
  "ID:2110280918", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->2110280918],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   RowBox[{"DomainTuples", "[", 
    RowBox[{"MMPairs", "[", 
     RowBox[{"\[ScriptCapitalG]", ",", "x", ",", "x"}], "]"}], "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["lmTupleAux", "\[ScriptCapitalG]"], "[", 
    RowBox[{"x", ",", "P"}], "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership lmTupleAux)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:481303170", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649562918918655*^9, 3.6495629429972486`*^9}},
 CellTags->{
  "ID:481303170", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership lmTupleAux"},
 CellID->481303170],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", "j"}]], 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"i", ",", "j"}], "\[RightAngleBracket]"}], 
    TagBox["\:22ff",
     Identity,
     SyntaxForm->"a+b"], "P"}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
      RowBox[{"[", "a", "]"}]}]], 
    RowBox[{
     UnderscriptBox["isMNTCR", "\[ScriptCapitalG]"], "[", 
     RowBox[{"x", ",", "x", ",", "a", ",", "i", ",", "j"}], "]"}]], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
      RowBox[{"[", 
       RowBox[{"b", ",", "m1", ",", "m2"}], "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"a", 
        UnderscriptBox["\[Tilde]", "\[ScriptCapitalG]"], "b"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"b", ",", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"m1", ",", "m2"}], "\[RightAngleBracket]"}]}], 
         "\[RightAngleBracket]"}], 
        TagBox["\:22ff",
         Identity,
         SyntaxForm->"a+b"], 
        RowBox[{
         UnderscriptBox["lmTupleAux", "\[ScriptCapitalG]"], "[", 
         RowBox[{"x", ",", "P"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[And]", 
      RowBox[{
       UnderscriptBox["areCPMultipliers", "\[ScriptCapitalG]"], "[", 
       RowBox[{
       "m1", ",", "m2", ",", "x", ",", "x", ",", "b", ",", "i", ",", "j"}], 
       "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lmTupleAux content)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1389102731", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649562959347188*^9, 3.6495630601355886`*^9}, {
  3.649564062713225*^9, 3.649564071726591*^9}, {3.650190784681386*^9, 
  3.6501908103828583`*^9}},
 CellTags->{
  "ID:1389102731", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTupleAux content"},
 CellID->1389102731],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->178925209]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->370655995],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership lmTupleAux)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership lmTupleAux)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["multN", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], ",", "<"}], 
                    "]"}], "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]]}],
            "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
                RowBox[{"MMPairs", "[", 
                  RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
             RowBox[{
               UnderscriptBox["lmTupleAux", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["P", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["multN", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], ",", "<"}], 
                    "]"}], "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]]}],
            "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
                RowBox[{"MMPairs", "[", 
                  RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
             RowBox[{
               UnderscriptBox["lmTupleAux", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["P", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:481303170"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p481303170-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p481303170-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:481303170", "Proof|ID:481303170-1"},
 CellID->1309244696],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p481303170-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1292585652]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lmTupleAux content)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lmTupleAux content)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["multN", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], ",", "<"}], 
                    "]"}], "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "\:22ff", 
               StyleBox["P", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["\[Tilde]", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
                 "\:22ff", 
                 RowBox[{
                   UnderscriptBox["lmTupleAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"]}], "]"}]}], "\[And]", 
               
               RowBox[{
                 UnderscriptBox["areCPMultipliers", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["multN", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], ",", "<"}], 
                    "]"}], "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "\:22ff", 
               StyleBox["P", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["b", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["\[Tilde]", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 StyleBox["b", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
                 "\:22ff", 
                 RowBox[{
                   UnderscriptBox["lmTupleAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"]}], "]"}]}], "\[And]", 
               
               RowBox[{
                 UnderscriptBox["areCPMultipliers", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["m1", "ExpressionVariable"], ",", 
                   StyleBox["m2", "ExpressionVariable"], ",", 
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1389102731"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1389102731-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1389102731-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1389102731", "Proof|ID:1389102731-1"},
 CellID->648616519],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1389102731-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->742679995]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Properties of lmTuple (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6482894611591682`*^9, 3.648289467728517*^9}, {
  3.648371640075479*^9, 3.6483716416504793`*^9}, {3.649577155658865*^9, 
  3.649577159444893*^9}},
 CellID->392268996],

Cell["", "OpenEnvironment",
 CellID->738030033],

Cell[CellGroupData[{

Cell["Theorem: Properties of lmTuple", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648287858471087*^9, 3.648287863048792*^9}, {
  3.648371646901845*^9, 3.648371648358472*^9}},
 CellID->262315235],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
    RowBox[{"[", 
     RowBox[{"x", ",", "y"}], "]"}]}]], 
  RowBox[{
   RowBox[{"x", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalG]"]}], "\[And]", 
   RowBox[{"y", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalG]"]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1615013989}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6482879387622957`*^9, 3.648287946257235*^9}, {
  3.648288476751683*^9, 3.64828851655273*^9}},
 CellTags->{
  "ID:1615013989", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1615013989],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"z", "=", 
   RowBox[{
    UnderscriptBox["lmTuple", "\[ScriptCapitalG]"], "[", 
    RowBox[{"x", ",", "y"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1776007069}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648288381563387*^9, 3.648288391156662*^9}},
 CellTags->{
  "ID:1776007069", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1776007069],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   RowBox[{"DomainTuples", "[", 
    RowBox[{"MMPairs", "[", 
     RowBox[{"\[ScriptCapitalG]", ",", "x", ",", "y"}], "]"}], "]"}]], 
  RowBox[{"[", "z", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership lmTuple)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1846257369", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648289345774753*^9, 3.648289364453143*^9}, {
  3.648289398284092*^9, 3.648289398780602*^9}, {3.6483716543719873`*^9, 
  3.648371656736062*^9}, {3.6483724345153513`*^9, 3.648372439538706*^9}},
 CellTags->{
  "ID:1846257369", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership lmTuple"},
 CellID->1846257369],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", "\[NotEqual]", "y"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
     RowBox[{"[", "a", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["isMNTCR", "\[ScriptCapitalG]"], "[", 
      RowBox[{"x", ",", "y", ",", "a", ",", "1", ",", "1"}], "]"}], 
     "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
         RowBox[{"[", 
          RowBox[{"b", ",", "mx", ",", "my"}], "]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{"a", 
          UnderscriptBox["\[Tilde]", "\[ScriptCapitalG]"], "b"}], "\[And]", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"b", ",", 
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"mx", ",", "my"}], "\[RightAngleBracket]"}]}], 
            "\[RightAngleBracket]"}], 
           TagBox["\:22ff",
            Identity,
            SyntaxForm->"a+b"], "z"}], 
          TagBox[")",
           "AutoParentheses"]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lmTuple \[NotEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:851213532", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648288722093814*^9, 3.648288746496537*^9}, {
   3.649060936457008*^9, 3.649060938147976*^9}, {3.649061938314665*^9, 
   3.649061955999022*^9}, {3.6501919860315*^9, 3.650192029098269*^9}, 
   3.65019207600251*^9},
 CellTags->{
  "ID:851213532", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTuple \[NotEqual]"},
 CellID->851213532],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", "\[Equal]", "y"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"k", ",", 
       RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{
        UnderscriptBox["multN", "\[ScriptCapitalG]"], "[", "y", "]"}]}]], 
     RowBox[{"l", ">", "k"}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
       RowBox[{"[", "a", "]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        UnderscriptBox["isMNTCR", "\[ScriptCapitalG]"], "[", 
        RowBox[{"y", ",", "y", ",", "a", ",", "k", ",", "l"}], "]"}], 
       "\[Implies]", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         UnderscriptBox["\[Exists]", 
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
           RowBox[{"[", 
            RowBox[{"b", ",", "m1", ",", "m2"}], "]"}]}]], 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{"\[And]", 
            RowBox[{"\[Piecewise]", GridBox[{
               {
                RowBox[{"a", 
                 UnderscriptBox["\[Tilde]", "\[ScriptCapitalG]"], "b"}]},
               {
                RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                  RowBox[{"b", ",", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{"m1", ",", "m2"}], "\[RightAngleBracket]"}]}], 
                  "\[RightAngleBracket]"}], 
                 TagBox["\:22ff",
                  Identity,
                  SyntaxForm->"a+b"], "z"}]},
               {
                RowBox[{
                 UnderscriptBox["areCPMultipliers", "\[ScriptCapitalG]"], "[", 
                 RowBox[{
                 "m1", ",", "m2", ",", "y", ",", "y", ",", "b", ",", "k", ",",
                   "l"}], "]"}]}
              },
              
              GridBoxAlignment->{
               "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
           TagBox[")",
            "AutoParentheses"]}], 
          TagBox[")",
           "AutoParentheses"]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(lmTuple =)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:959030563", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648288396556753*^9, 3.6482884249536533`*^9}, {
   3.648288571693883*^9, 3.648288579129424*^9}, {3.648288634032104*^9, 
   3.648288711707099*^9}, 3.6482887487479153`*^9, 3.6482887994691668`*^9, 
   3.6482918097638206`*^9, {3.649060940351858*^9, 3.6490609420056877`*^9}, {
   3.649061900618836*^9, 3.6490619013214407`*^9}, {3.649061963775522*^9, 
   3.6490619709234743`*^9}, {3.649062020001384*^9, 3.649062059408987*^9}, {
   3.649062124210009*^9, 3.649062134598445*^9}, {3.650192048562769*^9, 
   3.650192078910121*^9}},
 CellTags->{
  "ID:959030563", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTuple ="},
 CellID->959030563],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->554761415]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1435911470],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership lmTuple)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership lmTuple)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
                RowBox[{"MMPairs", "[", 
                  RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
             RowBox[{
               UnderscriptBox["lmTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
                RowBox[{"MMPairs", "[", 
                  RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}]], "[", 
             RowBox[{
               UnderscriptBox["lmTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1846257369"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1846257369-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1846257369-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1846257369", "Proof|ID:1846257369-1"},
 CellID->166489159],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1846257369-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->595798944]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lmTuple \[NotEqual])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lmTuple \[NotEqual])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isMNTCR", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1"}], 
                 "]"}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["mx", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["my", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Tilde]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    StyleBox["b", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["mx", "ExpressionVariable"], ",", 
                    StyleBox["my", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
                    "\:22ff", 
                    RowBox[{
                    UnderscriptBox["lmTuple", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["isMNTCR", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", "1", ",", "1"}], 
                 "]"}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["mx", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["my", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Tilde]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    StyleBox["b", "ExpressionVariable"]}], "\[And]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["mx", "ExpressionVariable"], ",", 
                    StyleBox["my", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
                    "\:22ff", 
                    RowBox[{
                    UnderscriptBox["lmTuple", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:851213532"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p851213532-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p851213532-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:851213532", "Proof|ID:851213532-1"},
 CellID->412269868],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p851213532-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->811817271]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(lmTuple =)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(lmTuple =)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Equal]", 
               StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}]}, {
                  RowBox[{
                    RowBox[{
                    StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}]}}]], 
              RowBox[{
                StyleBox["l", "ExpressionVariable"], ">", 
                StyleBox["k", "ExpressionVariable"]}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["isMNTCR", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Tilde]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    StyleBox["b", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
                    "\:22ff", 
                    RowBox[{
                    UnderscriptBox["lmTuple", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], "\[And]", 
                    
                    RowBox[{
                    UnderscriptBox["areCPMultipliers", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Equal]", 
               StyleBox["y", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}]}, {
                  RowBox[{
                    RowBox[{
                    StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}]}}]], 
              RowBox[{
                StyleBox["l", "ExpressionVariable"], ">", 
                StyleBox["k", "ExpressionVariable"]}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["isMNTCR", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], "]"}], 
                 "\[Implies]", 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Tilde]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    StyleBox["b", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
                    "\:22ff", 
                    RowBox[{
                    UnderscriptBox["lmTuple", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], "\[And]", 
                    
                    RowBox[{
                    UnderscriptBox["areCPMultipliers", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:959030563"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p959030563-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p959030563-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:959030563", "Proof|ID:959030563-1"},
 CellID->1319693045],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p959030563-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.650259203200597*^9, 3.650259205168311*^9}},
 CellID->316689583]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["chainCrit (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.64828725647401*^9, 3.648287258037859*^9}, {
  3.6488926391504107`*^9, 3.6488926427375927`*^9}},
 CellID->458833326],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
   RowBox[{"[", "B", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1440157218}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64880834040443*^9, 3.6488083455746393`*^9}},
 CellTags->{
  "ID:1440157218", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1440157218],

Cell[CellGroupData[{

Cell["TODO-Tuples (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.648887797764741*^9, 3.6488878007405167`*^9}, {
  3.648892619701137*^9, 3.648892623912984*^9}},
 CellID->162772541],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"i", ",", "j"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 805845835}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648954702668837*^9, 3.6489547041980124`*^9}},
 CellTags->{
  "ID:805845835", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->805845835],

Cell["", "OpenEnvironment",
 CellID->509087297],

Cell[CellGroupData[{

Cell["Definition: TODO-Tuples", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6488083288931723`*^9, 3.648808334196032*^9}, {
  3.6488084188862343`*^9, 3.648808419955503*^9}},
 CellID->1841460198],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "P"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isTODOTuple", "\[ScriptCapitalG]"], "[", 
     RowBox[{"P", ",", "B", ",", "i", ",", "j"}], "]"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{"k", ",", 
        RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          RowBox[{"processed", "[", 
           RowBox[{"k", ",", "l", ",", "P"}], "]"}], "\[And]", 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{
            RowBox[{"k", "\[NotEqual]", "i"}], "\[Or]", 
            RowBox[{"l", "\[NotEqual]", "j"}]}], 
           TagBox[")",
            "AutoParentheses"]}], "\[And]", 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{
            RowBox[{"k", "\[NotEqual]", "j"}], "\[Or]", 
            RowBox[{"l", "\[NotEqual]", "i"}]}], 
           TagBox[")",
            "AutoParentheses"]}]}], 
         TagBox[")",
          "AutoParentheses"]}], "\[Implies]", 
        RowBox[{
         UnderscriptBox["cpConnectible", "\[ScriptCapitalG]"], "[", 
         RowBox[{"B", ",", 
          SubscriptBox["B", "k"], ",", 
          SubscriptBox["B", "l"]}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isTODOTuple)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1576526273", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64880835039001*^9, 3.6488084665929728`*^9}, {
  3.64880853237426*^9, 3.648808537451559*^9}, {3.6488878196897707`*^9, 
  3.64888782187798*^9}, {3.6488880107435904`*^9, 3.6488880147404613`*^9}, {
  3.648954580947956*^9, 3.648954628207912*^9}, {3.648954709278783*^9, 
  3.648954709997858*^9}},
 CellTags->{
  "ID:1576526273", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple"},
 CellID->1576526273],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1611579671]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1825735770],

Cell["", "OpenEnvironment",
 CellID->501261685],

Cell[CellGroupData[{

Cell["Lemma: Simple Properties of TODO-Tuples", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648887825344915*^9, 3.648887831687454*^9}, {
  3.648890599236978*^9, 3.648890611969298*^9}, {3.6490568942517776`*^9, 
  3.649056901003626*^9}},
 CellID->2002554147],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["isTODOTuple", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   RowBox[{"allPairs", "[", 
    RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], "]"}], 
   ",", "B", ",", "i", ",", "j"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isTODOTuple allPairs)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:857941005", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648890622065085*^9, 3.6488906304957027`*^9}, {
   3.648954669809811*^9, 3.648954675052272*^9}, 3.648954711952321*^9},
 CellTags->{
  "ID:857941005", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple allPairs"},
 CellID->857941005],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], ",", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         TagBox["\[LessEqual]",
          Identity], 
         TagBox[")",
          "AutoParentheses"]}]}], "]"}], "]"}]], 
    RowBox[{"[", "P", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isTODOTuple", "\[ScriptCapitalG]"], "[", 
     RowBox[{"P", ",", "B", ",", "i", ",", "j"}], "]"}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
       RowBox[{"[", "x", "]"}]}]], 
     RowBox[{
      UnderscriptBox["isTODOTuple", "\[ScriptCapitalG]"], "[", 
      RowBox[{
       RowBox[{"update", "[", 
        RowBox[{"P", ",", 
         RowBox[{
          RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], "+",
           "1"}]}], "]"}], ",", 
       RowBox[{"B", 
        TagBox["\:293a",
         Identity,
         SyntaxForm->"a*b"], "x"}], ",", "i", ",", "j"}], "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isTODOTuple update)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1202189993", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648887837259262*^9, 3.6488878966525993`*^9}, {
  3.648887969384142*^9, 3.648888030819659*^9}, {3.648890799497517*^9, 
  3.648890809569756*^9}, {3.6489547149315157`*^9, 3.648954719095574*^9}},
 CellTags->{
  "ID:1202189993", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple update"},
 CellID->1202189993],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"k", ",", 
    RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"t", "..."}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{
          UnderscriptBox["cpConnectible", "\[ScriptCapitalG]"], "[", 
          RowBox[{"B", ",", 
           SubscriptBox["B", "k"], ",", 
           SubscriptBox["B", "l"]}], "]"}], "\[Or]", 
         RowBox[{"\[Not]", 
          RowBox[{"processed", "[", 
           RowBox[{"k", ",", "l", ",", 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"t", "..."}], "\[RightAngleBracket]"}]}], "]"}]}], 
         "\[Or]", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{
           RowBox[{"k", "\[Equal]", "i"}], "\[And]", 
           RowBox[{"l", "\[Equal]", "j"}]}], 
          TagBox[")",
           "AutoParentheses"]}], "\[Or]", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{
           RowBox[{"k", "\[Equal]", "j"}], "\[And]", 
           RowBox[{"l", "\[Equal]", "i"}]}], 
          TagBox[")",
           "AutoParentheses"]}]}], 
        TagBox[")",
         "AutoParentheses"]}], "\[And]", 
       RowBox[{
        UnderscriptBox["isTODOTuple", "\[ScriptCapitalG]"], "[", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"k", ",", "l"}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], ",", "B", ",", 
         "i", ",", "j"}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{
      UnderscriptBox["isTODOTuple", "\[ScriptCapitalG]"], "[", 
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"t", "..."}], "\[RightAngleBracket]"}], ",", "B", ",", "i", 
       ",", "j"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isTODOTuple rest)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1698314044", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649040975171628*^9, 3.649041086062841*^9}, {
  3.64904190687289*^9, 3.649041909741294*^9}, {3.649053084856895*^9, 
  3.6490531055222893`*^9}, {3.6490584753200073`*^9, 3.6490584936256113`*^9}},
 CellTags->{
  "ID:1698314044", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple rest"},
 CellID->1698314044],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"k", ",", "l"}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"t", "..."}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["isTODOTuple", "\[ScriptCapitalG]"], "[", 
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"k", ",", "l"}], "\[RightAngleBracket]"}], ",", 
         RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], ",", "B", ",", "i",
        ",", "j"}], "]"}], "\[Implies]", 
     RowBox[{
      UnderscriptBox["isTODOTuple", "\[ScriptCapitalG]"], "[", 
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"i", ",", "j"}], "\[RightAngleBracket]"}], ",", 
         RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], ",", "B", ",", "k",
        ",", "l"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isTODOTuple exchange)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:812012342", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649056789444809*^9, 3.649056825641636*^9}},
 CellTags->{
  "ID:812012342", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple exchange"},
 CellID->812012342],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1611580180]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->44008041],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isTODOTuple allPairs)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isTODOTuple allPairs)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           UnderscriptBox["isTODOTuple", 
            RowBox[{"GroebnerRing", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           "[", 
           RowBox[{
             RowBox[{"allPairs", "[", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], "]"}], ",", 
             StyleBox["B", "ExpressionVariable"], ",", 
             StyleBox["i", "ExpressionVariable"], ",", 
             StyleBox["j", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           UnderscriptBox["isTODOTuple", 
            RowBox[{"GroebnerRing", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           "[", 
           RowBox[{
             RowBox[{"allPairs", "[", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}], "]"}], ",", 
             StyleBox["B", "ExpressionVariable"], ",", 
             StyleBox["i", "ExpressionVariable"], ",", 
             StyleBox["j", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:857941005"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p857941005-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p857941005-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:857941005", "Proof|ID:857941005-1"},
 CellID->1051929619],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p857941005-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->546271768]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isTODOTuple update)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isTODOTuple update)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["isTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 RowBox[{"update", "[", 
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]}], "]"}], ",", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "\:293a", 
                   StyleBox["x", "ExpressionVariable"]}], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["isTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 RowBox[{"update", "[", 
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]}], "]"}], ",", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "\:293a", 
                   StyleBox["x", "ExpressionVariable"]}], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1202189993"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1202189993-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1202189993-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1202189993", "Proof|ID:1202189993-1"},
 CellID->1110047828],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1202189993-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->652755217]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isTODOTuple rest)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isTODOTuple rest)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             StyleBox[
              RowBox[{"t", "..."}], "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                   SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["l", "ExpressionVariable"]]}], "]"}], "\[Or]", 
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{"processed", "[", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "]"}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   StyleBox["k", "ExpressionVariable"], "\[Equal]", 
                   StyleBox["i", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["l", "ExpressionVariable"], "\[Equal]", 
                   StyleBox["j", "ExpressionVariable"]}]}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   StyleBox["k", "ExpressionVariable"], "\[Equal]", 
                   StyleBox["j", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["l", "ExpressionVariable"], "\[Equal]", 
                   StyleBox["i", "ExpressionVariable"]}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{"\[LeftAngleBracket]", 
                 StyleBox[
                  RowBox[{"t", "..."}], "ExpressionVariable"], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             StyleBox[
              RowBox[{"t", "..."}], "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["cpConnectible", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], ",", 
                   SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                   SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["l", "ExpressionVariable"]]}], "]"}], "\[Or]", 
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{"processed", "[", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "]"}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   StyleBox["k", "ExpressionVariable"], "\[Equal]", 
                   StyleBox["i", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["l", "ExpressionVariable"], "\[Equal]", 
                   StyleBox["j", "ExpressionVariable"]}]}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   StyleBox["k", "ExpressionVariable"], "\[Equal]", 
                   StyleBox["j", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   StyleBox["l", "ExpressionVariable"], "\[Equal]", 
                   StyleBox["i", "ExpressionVariable"]}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{"\[LeftAngleBracket]", 
                 StyleBox[
                  RowBox[{"t", "..."}], "ExpressionVariable"], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1698314044"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1698314044-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1698314044-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1698314044", "Proof|ID:1698314044-1"},
 CellID->610967613],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1698314044-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->709078776]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isTODOTuple exchange)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isTODOTuple exchange)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}, {
             StyleBox["k", "ExpressionVariable"]}, {
             StyleBox["l", "ExpressionVariable"]}, {
             StyleBox[
              RowBox[{"t", "..."}], "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"], ",", 
               StyleBox["l", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}, {
             StyleBox["k", "ExpressionVariable"]}, {
             StyleBox["l", "ExpressionVariable"]}, {
             StyleBox[
              RowBox[{"t", "..."}], "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"], ",", 
               StyleBox["l", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:812012342"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p812012342-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p812012342-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:812012342", "Proof|ID:812012342-1"},
 CellID->195013346],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p812012342-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->914583248]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Properties of chainCrit (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.648887787593692*^9, 3.648887791246503*^9}, {
  3.648892630843829*^9, 3.6488926345390787`*^9}},
 CellID->304206187],

Cell["", "OpenEnvironment",
 CellID->611079668],

Cell[CellGroupData[{

Cell["Theorem: Properties of chainCrit", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648807884247645*^9, 3.648807890010829*^9}},
 CellID->875163593],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "P"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", 
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{"k", ",", "l"}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
       RowBox[{"[", "a", "]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{
          UnderscriptBox["isTODOTuple", "\[ScriptCapitalG]"], "[", 
          RowBox[{"P", ",", "B", ",", "i", ",", "j"}], "]"}], "\[And]", 
         RowBox[{
          UnderscriptBox["isMNTCR", "\[ScriptCapitalG]"], "[", 
          RowBox[{
           SubscriptBox["B", "i"], ",", 
           SubscriptBox["B", "j"], ",", "a", ",", "k", ",", "l"}], "]"}], 
         "\[And]", 
         RowBox[{
          UnderscriptBox["chainCrit", "\[ScriptCapitalG]"], "[", 
          RowBox[{"i", ",", "j", ",", "a", ",", "B", ",", "P"}], "]"}]}], 
        TagBox[")",
         "AutoParentheses"]}], "\[Implies]", 
       RowBox[{
        UnderscriptBox["cpConnectible", "\[ScriptCapitalG]"], "[", 
        RowBox[{
         SubscriptBox["B", "i"], ",", 
         SubscriptBox["B", "j"], ",", "a", ",", "k", ",", "l", ",", "B"}], 
        "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(chainCrit 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1260897708", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648807900393601*^9, 3.648807954315282*^9}, {
   3.648807989781859*^9, 3.6488080884002028`*^9}, 3.648808347242332*^9, {
   3.6488085693788347`*^9, 3.648808619024288*^9}, {3.648808649860874*^9, 
   3.648808657342062*^9}, {3.648808818334157*^9, 3.648808846624042*^9}, {
   3.6488090472749767`*^9, 3.64880905692369*^9}, {3.6488091076800137`*^9, 
   3.648809107883355*^9}, {3.648809826433372*^9, 3.648809834703271*^9}, {
   3.648810456523892*^9, 3.648810511962919*^9}, {3.648811392386992*^9, 
   3.648811398471401*^9}, {3.648955290103998*^9, 3.64895529076923*^9}, 
   3.6489556627982397`*^9},
 CellTags->{
  "ID:1260897708", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "chainCrit 1"},
 CellID->1260897708],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1094665073]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1919870362],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(chainCrit 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(chainCrit 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["P", "ExpressionVariable"]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             StyleBox["k", "ExpressionVariable"]}, {
             StyleBox["l", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["P", "ExpressionVariable"], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], ",", 
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox["j", "ExpressionVariable"]], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["k", "ExpressionVariable"], ",", 
                 StyleBox["l", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["chainCrit", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["P", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["cpConnectible", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"], ",", 
               StyleBox["l", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["P", "ExpressionVariable"]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             StyleBox["k", "ExpressionVariable"]}, {
             StyleBox["l", "ExpressionVariable"]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["P", "ExpressionVariable"], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isMNTCR", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], ",", 
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox["j", "ExpressionVariable"]], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["k", "ExpressionVariable"], ",", 
                 StyleBox["l", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["chainCrit", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["P", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["cpConnectible", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"], ",", 
               StyleBox["l", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1260897708"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1260897708-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1260897708-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1260897708", "Proof|ID:1260897708-1"},
 CellID->721986367],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1260897708-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {61., 67.}}]], "ProofInfoBody",
 CellID->113158442]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6488087045099277`*^9, 3.648808705808342*^9}, 
   3.648810420982791*^9, 3.6532840710346985`*^9, 3.6661730742453003`*^9, 
   3.666439911023279*^9},
 CellID->1158351766],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
    RowBox[{"[", "B", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["P", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
    RowBox[{"[", "a", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["k", "0"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["l", "0"], TraditionalForm]]],
 " a.b.f., assume\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTODOTuple", "[", 
    RowBox[{"P", ",", "B", ",", "i", ",", "j"}], "]"}], TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "j"], ",", "a", ",", 
     SubscriptBox["k", "0"], ",", 
     SubscriptBox["l", "0"]}], "]"}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"chainCrit", "[", 
    RowBox[{"i", ",", "j", ",", "a", ",", "B", ",", "P"}], "]"}], 
   TraditionalForm]]],
 "\nand prove\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"cpConnectible", "[", 
    RowBox[{
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "j"], ",", "a", ",", 
     SubscriptBox["k", "0"], ",", 
     SubscriptBox["l", "0"], ",", "B"}], "]"}], TraditionalForm]]],
 "\n\nProving (G), by the definition of cpConnectible, amounts to finding ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
    RowBox[{"[", 
     RowBox[{"m1", ",", "m2"}], "]"}]}], TraditionalForm]]],
 " with\n(G.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCPMultipliers", "[", 
    RowBox[{"m1", ",", "m2", ",", 
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "j"], ",", "a", ",", 
     SubscriptBox["k", "0"], ",", 
     SubscriptBox["l", "0"]}], "]"}], TraditionalForm]]],
 "\n(G.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areConnectibleMultipliers", "[", 
    RowBox[{"m1", ",", "m2", ",", 
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "j"], ",", "a", ",", "B"}], "]"}], TraditionalForm]]],
 "\n\nFrom (4), by the definition of chainCrit, we know\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "\[NotEqual]", "j"}], TraditionalForm]]],
 "\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"processed", "[", 
    RowBox[{"i", ",", "i", ",", "P"}], "]"}], TraditionalForm]]],
 "\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"processed", "[", 
    RowBox[{"j", ",", "j", ",", "P"}], "]"}], TraditionalForm]]],
 "\n(0.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", "\[NotEqual]", "i"}], TraditionalForm]]],
 "\n(0.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", "\[NotEqual]", "j"}], TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"rdm", "[", 
     RowBox[{"a", ",", 
      SubscriptBox["B", "k"]}], "]"}], "\[NotEqual]", "0"}], 
   TraditionalForm]]],
 "\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"processed", "[", 
    RowBox[{"k", ",", "k", ",", "P"}], "]"}], TraditionalForm]]],
 "\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"processed", "[", 
    RowBox[{"i", ",", "k", ",", "P"}], "]"}], TraditionalForm]]],
 "\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"processed", "[", 
    RowBox[{"j", ",", "k", ",", "P"}], "]"}], TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 ". We now expand the definition of isTODOTuple: We know\n(13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{"k", ",", 
       RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"processed", "[", 
        RowBox[{"k", ",", "l", ",", "P"}], "]"}], "\[And]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"k", "\[NotEqual]", "i"}], "\[Or]", 
         RowBox[{"l", "\[NotEqual]", "j"}]}], ")"}], "\[And]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"k", "\[NotEqual]", "j"}], "\[Or]", 
         RowBox[{"l", "\[NotEqual]", "i"}]}], ")"}]}], ")"}]}], "\[Implies]", 
    
    RowBox[{"cpConnectible", "[", 
     RowBox[{"B", ",", 
      SubscriptBox["B", "k"], ",", 
      SubscriptBox["B", "l"]}], "]"}]}], TraditionalForm]]],
 "\n\nHence, from (5), (6), (8), (9) and (10), using (13), (4), (0.1) and \
(0.2), we obtain\n(5\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"cpConnectible", "[", 
    RowBox[{"B", ",", 
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "i"]}], "]"}], TraditionalForm]]],
 "\n(6\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"cpConnectible", "[", 
    RowBox[{"B", ",", 
     SubscriptBox["B", "j"], ",", 
     SubscriptBox["B", "j"]}], "]"}], TraditionalForm]]],
 "\n(8\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"cpConnectible", "[", 
    RowBox[{"B", ",", 
     SubscriptBox["B", "k"], ",", 
     SubscriptBox["B", "k"]}], "]"}], TraditionalForm]]],
 "\n(9\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"cpConnectible", "[", 
    RowBox[{"B", ",", 
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "k"]}], "]"}], TraditionalForm]]],
 "\n(10\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"cpConnectible", "[", 
    RowBox[{"B", ",", 
     SubscriptBox["B", "j"], ",", 
     SubscriptBox["B", "k"]}], "]"}], TraditionalForm]]],
 "\n\nWe can now use (75) from \[OpenCurlyDoubleQuote]ReductionRings.nb\
\[CloseCurlyDoubleQuote] and infer\n(14)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isLocConnective", "[", 
    RowBox[{"B", ",", 
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "k"]}], "]"}], TraditionalForm]]],
 "\n(15)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isLocConnective", "[", 
    RowBox[{"B", ",", 
     SubscriptBox["B", "j"], ",", 
     SubscriptBox["B", "k"]}], "]"}], TraditionalForm]]],
 "\n\nFrom (2), by the definitions of isMNTCR, isNTCR and isCommonReducible, \
we know\n(16)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     SubscriptBox["B", "i"], ",", "m1"}], "]"}], TraditionalForm]]],
 "\n(17)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     SubscriptBox["B", "j"], ",", "m2"}], "]"}], TraditionalForm]]],
 "\n(18)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areIrrelative", "[", 
    RowBox[{
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "j"], ",", "m1", ",", "m2", ",", "k", ",", "l"}], 
    "]"}], TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
    RowBox[{"[", 
     RowBox[{"m1", ",", "m2"}], "]"}]}], TraditionalForm]]],
 ". We claim that these constants witness our existential goal, i.e. we show \
both subgoals.\n\n(G.1):\nBy the definition of areCPMultipliers we have to \
show\n(G.1.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "j"], ",", "a", ",", "k", ",", "l"}], "]"}], 
   TraditionalForm]]],
 "\n(G.1.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areIrrelative", "[", 
    RowBox[{
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "j"], ",", "m1", ",", "m2", ",", "k", ",", "l"}], 
    "]"}], TraditionalForm]]],
 "\n(G.1.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     SubscriptBox["B", "i"], ",", "m1"}], "]"}], TraditionalForm]]],
 "\n(G.1.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     SubscriptBox["B", "j"], ",", "m2"}], "]"}], TraditionalForm]]],
 "\nAll of the four subgoals are already known, hence we are done.\n\n(G.2):\n\
By the definition of areConnectibleMultipliers we have to show\n(G.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     RowBox[{"m1", ",", 
      SubscriptBox["B", "i"]}]], 
    RowBox[{"a", "-", 
     RowBox[{"m1", "*", 
      SubscriptBox["B", "i"]}]}]}], TraditionalForm]]],
 "\n(G.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     RowBox[{"m2", ",", 
      SubscriptBox["B", "j"]}]], 
    RowBox[{"a", "-", 
     RowBox[{"m2", "*", 
      SubscriptBox["B", "j"]}]}]}], TraditionalForm]]],
 "\n(G.5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "-", 
     RowBox[{"m1", "*", 
      SubscriptBox["B", "i"]}]}], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"a", ",", "\[Succeeds]"}]], 
    RowBox[{"a", "-", 
     RowBox[{"m2", "*", 
      SubscriptBox["B", "j"]}]}]}], TraditionalForm]]],
 "\n\n(G.3) and (G.4) follow readily from (16) and (17), hence what remains \
to be shown is (G.5); for this, let ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b1", ":=", 
    RowBox[{"a", "-", 
     RowBox[{"m1", "*", 
      SubscriptBox["B", "i"]}]}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b2", ":=", 
    RowBox[{"a", "-", 
     RowBox[{"m2", "*", 
      SubscriptBox["B", "j"]}]}]}], TraditionalForm]]],
 ". We have to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b1", 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"a", ",", "\[Succeeds]"}]], "b2"}], TraditionalForm]]],
 ".\nFrom (7) we know\n(19)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     SubscriptBox["B", "k"]], "b"}], TraditionalForm]]],
 "\nfor some ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 ", and from (G.3) and (G.4) we also know\n(20)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     SubscriptBox["B", "i"]], "b1"}], TraditionalForm]]],
 "\n(21)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     SubscriptBox["B", "j"]], "b2"}], TraditionalForm]]],
 "\n\nHence, we can now use (14) and (15) and infer\n(22)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"b1", 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"a", ",", "\[Succeeds]"}]], "b"}], TraditionalForm]]],
 "\n(23)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"b2", 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"a", ",", "\[Succeeds]"}]], "b"}], TraditionalForm]]],
 "\nOur goal follows readily from transitivity of ",
 Cell[BoxData[
  FormBox[
   OverscriptBox[
    SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
    RowBox[{"a", ",", "\[Succeeds]"}]], TraditionalForm]]],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.648808713091893*^9, 3.648809370451869*^9}, {
   3.6488094312517776`*^9, 3.648809477926857*^9}, {3.648809538441759*^9, 
   3.648809806930128*^9}, {3.6488098379234037`*^9, 3.6488100023536987`*^9}, {
   3.648810040741338*^9, 3.648810091230895*^9}, {3.648810143328574*^9, 
   3.6488102122415133`*^9}, {3.6488102423474197`*^9, 
   3.6488104209829187`*^9}, {3.6488105766469917`*^9, 3.648810576907086*^9}, {
   3.648955643914068*^9, 3.6489558144187593`*^9}, {3.6489558622121696`*^9, 
   3.6489558699969683`*^9}, 3.6532840710346985`*^9, 3.6661730742453003`*^9, 
   3.666439911023386*^9},
 CellID->931213996],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.648810403910111*^9, 3.648810420983054*^9}, 
   3.6532840710346985`*^9, 3.6661730742453003`*^9, 3.66643991102349*^9},
 TextAlignment->Right,
 CellID->922384546]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Theorem: Correctness of Function GB (\[Checkmark])\
\>", "Section",
 CellChangeTimes->{{3.6457666136819563`*^9, 3.645766623347131*^9}, {
  3.649049597488245*^9, 3.6490495986199713`*^9}, {3.649144855875936*^9, 
  3.649144856282055*^9}, {3.649393156315942*^9, 3.6493931601691504`*^9}},
 CellID->284021016],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "isAlgoReductionRing", "[", "\[ScriptCapitalR]", 
   "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 957934274}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.592023915573992*^9, 3.592023917728783*^9}, {
  3.592026332744973*^9, 3.592026338353141*^9}, {3.6482873124400253`*^9, 
  3.648287317030466*^9}},
 CellTags->{
  "ID:957934274", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->957934274],

Cell["", "Text",
 CellID->1053932410],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"\[ScriptCapitalG]", "=", 
   RowBox[{
   "GroebnerRing", "[", "\[ScriptCapitalR]", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1363987639}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.592023926115796*^9, 3.592023944789061*^9}, {
  3.648287327064844*^9, 3.6482873318724813`*^9}},
 CellTags->{
  "ID:1363987639", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1363987639],

Cell[CellGroupData[{

Cell["\<\
Termination & Induction Schemas (\[Checkmark])\
\>", "Subsection",
 CellChangeTimes->{{3.648290845311919*^9, 3.648290847063879*^9}, {
  3.648456259975161*^9, 3.6484562637064257`*^9}, {3.648532181065514*^9, 
  3.648532185321212*^9}},
 CellID->149761983],

Cell["", "OpenEnvironment",
 CellID->117373454],

Cell[CellGroupData[{

Cell["Definition: Termination Order for Function GBAux", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.661603551018162*^9, 3.6616035544969673`*^9}, {
  3.6616599023749113`*^9, 3.6616599091453238`*^9}},
 CellID->71182710],

Cell[BoxData[
 RowBox[{
  RowBox[{"GBTermOrder", "[", "\[ScriptCapitalR]", "]"}], ":=", 
  RowBox[{"LexOrder3", "[", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}], ",", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       TagBox[
        UnderscriptBox["\[SquareSubset]", "\[ScriptCapitalG]"],
        Identity], 
       TagBox[")",
        "AutoParentheses"]}]}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"DomainTuples", "[", 
       RowBox[{"ProdDomain", "[", 
        RowBox[{
        "\[DoubleStruckCapitalZ]Dom", ",", "\[DoubleStruckCapitalZ]Dom"}], 
        "]"}], "]"}], ",", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       TagBox[
        UnderscriptBox["\[Succeeds]", 
         RowBox[{"Tuples", "[", "]"}]],
        Identity], 
       TagBox[")",
        "AutoParentheses"]}]}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"DomainTuples", "[", 
       RowBox[{"ProdDomain", "[", 
        RowBox[{"\[ScriptCapitalG]", ",", 
         RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]}], "]"}], 
       "]"}], ",", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       TagBox[
        UnderscriptBox["\[Succeeds]", 
         RowBox[{"Tuples", "[", "]"}]],
        Identity], 
       TagBox[")",
        "AutoParentheses"]}]}], "\[RightAngleBracket]"}]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBTermOrder)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:184081942", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6616035580069737`*^9, 3.66160357875501*^9}, {
  3.6616036199702826`*^9, 3.661603621187085*^9}},
 CellTags->{
  "ID:184081942", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBTermOrder"},
 CellID->184081942],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->356448921]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->504344449],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"\[ScriptCapitalL]", "=", 
   RowBox[{
   "GBTermOrder", "[", "\[ScriptCapitalR]", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1341865321}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648373867979803*^9, 3.648373934291519*^9}, 
   3.6615936841325183`*^9, 3.66159372500459*^9, {3.6615944780937133`*^9, 
   3.6615945335518103`*^9}, {3.6615945675598702`*^9, 
   3.6615945721618786`*^9}, {3.661659766860073*^9, 3.661659772491683*^9}},
 CellTags->{
  "ID:1341865321", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1341865321],

Cell[CellGroupData[{

Cell["Termination (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.649049649516821*^9, 3.649049656153554*^9}},
 CellID->553484892],

Cell["", "OpenEnvironment",
 CellID->1370619040],

Cell[CellGroupData[{

Cell["\<\
Lemma: \[ScriptCapitalL] is Noetherian Partial Irreflexive Order\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648373988696175*^9, 3.648374020419423*^9}, 
   3.648374887059141*^9},
 CellID->1537800075],

Cell[BoxData[
 RowBox[{"isDomain", "[", "\[ScriptCapitalL]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalL] isDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:320402459", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661595105558015*^9, 3.661595113264429*^9}},
 CellTags->{
  "ID:320402459", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] isDomain"},
 CellID->320402459],

Cell[BoxData[
 RowBox[{"isPartIrreflOrder", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalL]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalL]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalL] isPartIrreflOrder)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1861353391", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6483740236054688`*^9, 3.648374046502315*^9}},
 CellTags->{
  "ID:1861353391", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] isPartIrreflOrder"},
 CellID->1861353391],

Cell[BoxData[
 RowBox[{"isNoetherian", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalL]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalL]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalL] isNoetherian)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:461530885", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6483740611948757`*^9, 3.648374064928801*^9}},
 CellTags->{
  "ID:461530885", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] isNoetherian"},
 CellID->461530885],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalL]"], 
    RowBox[{"[", "x", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{
     UnderscriptBox["\[Element]", 
      RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
     RowBox[{"[", "C", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"\[Or]", 
     RowBox[{"\[Piecewise]", GridBox[{
        {
         RowBox[{"x", "\[Equal]", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"C", ",", 
            RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", 
            RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
           "\[RightAngleBracket]"}]}]},
        {
         RowBox[{
          UnderscriptBox[
           UnderscriptBox["\[Exists]", 
            RowBox[{"t", ",", 
             RowBox[{"t", "..."}]}]], 
           RowBox[{
            UnderscriptBox["\[Element]", 
             RowBox[{"DomainTuples", "[", 
              RowBox[{"ProdDomain", "[", 
               RowBox[{
               "\[DoubleStruckCapitalZ]Dom", ",", 
                "\[DoubleStruckCapitalZ]Dom"}], "]"}], "]"}]], 
            RowBox[{"[", 
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"t", ",", 
               RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], "]"}]}]], 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{"x", "\[Equal]", 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"C", ",", 
              RowBox[{"\[LeftAngleBracket]", 
               RowBox[{"t", ",", 
                RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], ",", 
              RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
             "\[RightAngleBracket]"}]}], 
           TagBox[")",
            "AutoParentheses"]}]}]},
        {
         RowBox[{
          UnderscriptBox["\[Exists]", 
           RowBox[{
            UnderscriptBox["\[Element]", 
             RowBox[{"DomainTuples", "[", 
              RowBox[{"ProdDomain", "[", 
               RowBox[{
               "\[DoubleStruckCapitalZ]Dom", ",", 
                "\[DoubleStruckCapitalZ]Dom"}], "]"}], "]"}]], 
            RowBox[{"[", "T", "]"}]}]], 
          RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[Exists]", 
             RowBox[{"t", ",", 
              RowBox[{"t", "..."}]}]], 
            RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
               RowBox[{"ProdDomain", "[", 
                RowBox[{"\[ScriptCapitalG]", ",", 
                 RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]}], 
                "]"}], "]"}]], 
             RowBox[{"[", 
              RowBox[{"\[LeftAngleBracket]", 
               RowBox[{"t", ",", 
                RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], "]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{"x", "\[Equal]", 
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"C", ",", "T", ",", 
               RowBox[{"\[LeftAngleBracket]", 
                RowBox[{"t", ",", 
                 RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}]}], 
              "\[RightAngleBracket]"}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}},
          "RowsIndexed" -> {}}]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalL] shape)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1658780438", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648440247429385*^9, 3.648440349010583*^9}, {
  3.661600848887615*^9, 3.661600890009288*^9}},
 CellTags->{
  "ID:1658780438", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] shape"},
 CellID->1658780438],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1282280602]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->272090212],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[ScriptCapitalL] isDomain)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[ScriptCapitalL] isDomain)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isDomain", "[", 
           RowBox[{"GBTermOrder", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isDomain", "[", 
           RowBox[{"GBTermOrder", "[", 
             StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:320402459"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p320402459-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p320402459-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:320402459", "Proof|ID:320402459-1"},
 CellID->957846942],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p320402459-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->270950904]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[ScriptCapitalL] isPartIrreflOrder)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[ScriptCapitalL] isPartIrreflOrder)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isPartIrreflOrder", "[", 
           RowBox[{
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"GBTermOrder", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             ",", 
             RowBox[{"GBTermOrder", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isPartIrreflOrder", "[", 
           RowBox[{
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"GBTermOrder", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             ",", 
             RowBox[{"GBTermOrder", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1861353391"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1861353391-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1861353391-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1861353391", "Proof|ID:1861353391-1"},
 CellID->71550445],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1861353391-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->684980337]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[ScriptCapitalL] isNoetherian)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[ScriptCapitalL] isNoetherian)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isNoetherian", "[", 
           RowBox[{
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"GBTermOrder", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             ",", 
             RowBox[{"GBTermOrder", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isNoetherian", "[", 
           RowBox[{
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"GBTermOrder", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             ",", 
             RowBox[{"GBTermOrder", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:461530885"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p461530885-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p461530885-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:461530885", "Proof|ID:461530885-1"},
 CellID->318070455],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p461530885-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->316861376]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[ScriptCapitalL] shape)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[ScriptCapitalL] shape)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GBTermOrder", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                 StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["C", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
                   ",", 
                   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
                 "\[RightAngleBracket]"}]}], "\[Or]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[Exists]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], ",", 
                   StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}]], 
                RowBox[{
                  UnderscriptBox["\[Element]", 
                   RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    
                    RowBox[{
                    "\[DoubleStruckCapitalZ]Dom", ",", 
                    "\[DoubleStruckCapitalZ]Dom"}], "]"}], "]"}]], "[", 
                  RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "]"}]], 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}],
                    "\[RightAngleBracket]"}]}]}], "\[Or]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    
                    RowBox[{
                    "\[DoubleStruckCapitalZ]Dom", ",", 
                    "\[DoubleStruckCapitalZ]Dom"}], "]"}], "]"}]], "[", 
                    StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[Exists]", 
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}]], 
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]}], "]"}], "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "]"}]], 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], 
                    "\[RightAngleBracket]"}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GBTermOrder", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                 StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["C", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
                   ",", 
                   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
                 "\[RightAngleBracket]"}]}], "\[Or]", 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[Exists]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], ",", 
                   StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}]], 
                RowBox[{
                  UnderscriptBox["\[Element]", 
                   RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    
                    RowBox[{
                    "\[DoubleStruckCapitalZ]Dom", ",", 
                    "\[DoubleStruckCapitalZ]Dom"}], "]"}], "]"}]], "[", 
                  RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "]"}]], 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}],
                    "\[RightAngleBracket]"}]}]}], "\[Or]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    
                    RowBox[{
                    "\[DoubleStruckCapitalZ]Dom", ",", 
                    "\[DoubleStruckCapitalZ]Dom"}], "]"}], "]"}]], "[", 
                    StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[Exists]", 
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}]], 
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]}], "]"}], "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "]"}]], 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], "\[Equal]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], 
                    "\[RightAngleBracket]"}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1658780438"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1658780438-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1658780438-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1658780438", "Proof|ID:1658780438-1"},
 CellID->65069410],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1658780438-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->71741560]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->154526518],

Cell[CellGroupData[{

Cell["\<\
Theorem: \[ScriptCapitalL] Decreases in Every Recursive Call of Function GBAux\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6483748996545897`*^9, 3.648374916662917*^9}},
 CellID->2131923963],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
   RowBox[{"[", "B", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1606683850}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6482879174497128`*^9, 3.648287927382084*^9}},
 CellTags->{
  "ID:1606683850", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1606683850],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"t", ",", 
    RowBox[{"t", "..."}], ",", "T", ",", "S"}]], 
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"B", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", ",", 
       RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], ",", "T"}], 
    "\[RightAngleBracket]"}], 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalL]"], 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"B", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", "..."}], "\[RightAngleBracket]"}], ",", "S"}], 
    "\[RightAngleBracket]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalL] \[Succeeds] 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:973931584", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648374995856173*^9, 3.648375067626556*^9}, {
  3.648375650681089*^9, 3.648375702533682*^9}, {3.648375809768749*^9, 
  3.6483758228240643`*^9}, {3.648446503826695*^9, 3.648446529606009*^9}},
 CellTags->{
  "ID:973931584", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] \[Succeeds] 1"},
 CellID->973931584],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"T", ",", "t", ",", 
    RowBox[{"t", "..."}]}]], 
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"B", ",", "T", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", ",", 
       RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}]}], 
    "\[RightAngleBracket]"}], 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalL]"], 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"B", ",", "T", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", "..."}], "\[RightAngleBracket]"}]}], 
    "\[RightAngleBracket]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalL] \[Succeeds] 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1350364326", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648375112848361*^9, 3.6483751798361273`*^9}, {
  3.648375708143948*^9, 3.648375722983048*^9}, {3.648446540740839*^9, 
  3.6484465444400787`*^9}},
 CellTags->{
  "ID:1350364326", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] \[Succeeds] 2"},
 CellID->1350364326],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"h", ",", "T1", ",", "T2", ",", "S1", ",", "S2"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"B", 
     UnderscriptBox["\[SquareSubset]", "\[ScriptCapitalG]"], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"B", 
       TagBox["\:293a",
        Identity,
        SyntaxForm->"a*b"], "h"}], 
      TagBox[")",
       "AutoParentheses"]}]}], "\[Implies]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"B", ",", "T1", ",", "S1"}], "\[RightAngleBracket]"}], 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalL]"], 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       RowBox[{"B", 
        TagBox["\:293a",
         Identity,
         SyntaxForm->"a*b"], "h"}], ",", "T2", ",", "S2"}], 
      "\[RightAngleBracket]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalL] \[Succeeds] 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:529848255", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648446256529964*^9, 3.648446337169415*^9}},
 CellTags->{
  "ID:529848255", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] \[Succeeds] 3"},
 CellID->529848255],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
     RowBox[{"[", 
      RowBox[{"a", ",", "mi", ",", "mj"}], "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     UnderscriptBox["let", 
      RowBox[{"h", "=", 
       RowBox[{
        RowBox[{
         UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
         RowBox[{
          RowBox[{"a", 
           UnderscriptBox["-", "\[ScriptCapitalG]"], 
           RowBox[{"mi", 
            UnderscriptBox["*", "\[ScriptCapitalG]"], 
            SubscriptBox["B", "i"]}]}], ",", "B"}], "]"}], 
        UnderscriptBox["-", "\[ScriptCapitalG]"], 
        RowBox[{
         UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
         RowBox[{
          RowBox[{"a", 
           UnderscriptBox["-", "\[ScriptCapitalG]"], 
           RowBox[{"mj", 
            UnderscriptBox["*", "\[ScriptCapitalG]"], 
            SubscriptBox["B", "j"]}]}], ",", "B"}], "]"}]}]}]], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"\[Not]", 
         RowBox[{"h", "\[Equal]", 
          UnderscriptBox["0", "\[ScriptCapitalG]"]}]}], 
        TagBox[")",
         "AutoParentheses"]}], "\[Implies]", 
       RowBox[{"B", 
        UnderscriptBox["\[SquareSubset]", "\[ScriptCapitalG]"], 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"B", 
          TagBox["\:293a",
           Identity,
           SyntaxForm->"a*b"], "h"}], 
         TagBox[")",
          "AutoParentheses"]}]}]}], ")"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[SquareSubset] append trd)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:592062547", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6483762780771847`*^9, 3.6483764397799473`*^9}, {
  3.64837680883632*^9, 3.648376827453269*^9}, {3.6483770881968946`*^9, 
  3.648377088414504*^9}, {3.648380031315686*^9, 3.6483800407623672`*^9}, {
  3.64838015565136*^9, 3.648380197689966*^9}, {3.648380232496937*^9, 
  3.648380240096492*^9}, {3.6484463683744698`*^9, 3.648446370623953*^9}, {
  3.648446403856524*^9, 3.6484464612854443`*^9}, {3.648446635276716*^9, 
  3.6484466464714413`*^9}, {3.648447525230394*^9, 3.648447532496539*^9}},
 CellTags->{
  "ID:592062547", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[SquareSubset] append trd"},
 CellID->592062547],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->611030949]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1735955644],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[ScriptCapitalL] \[Succeeds] 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[ScriptCapitalL] \[Succeeds] 1)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["t", "ExpressionVariable"]}, {
             StyleBox[
              RowBox[{"t", "..."}], "ExpressionVariable"]}, {
             StyleBox["T", "ExpressionVariable"]}, {
             StyleBox["S", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], ",", 
                   StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["T", "ExpressionVariable"]}], 
             "\[RightAngleBracket]"}], 
           UnderscriptBox["\[Succeeds]", 
            RowBox[{"GBTermOrder", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 StyleBox[
                  RowBox[{"t", "..."}], "ExpressionVariable"], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["S", "ExpressionVariable"]}], 
             "\[RightAngleBracket]"}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["t", "ExpressionVariable"]}, {
             StyleBox[
              RowBox[{"t", "..."}], "ExpressionVariable"]}, {
             StyleBox["T", "ExpressionVariable"]}, {
             StyleBox["S", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], ",", 
                   StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["T", "ExpressionVariable"]}], 
             "\[RightAngleBracket]"}], 
           UnderscriptBox["\[Succeeds]", 
            RowBox[{"GBTermOrder", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 StyleBox[
                  RowBox[{"t", "..."}], "ExpressionVariable"], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["S", "ExpressionVariable"]}], 
             "\[RightAngleBracket]"}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:973931584"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p973931584-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p973931584-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:973931584", "Proof|ID:973931584-1"},
 CellID->262124527],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p973931584-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->11227126]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[ScriptCapitalL] \[Succeeds] 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[ScriptCapitalL] \[Succeeds] 2)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["T", "ExpressionVariable"]}, {
             StyleBox["t", "ExpressionVariable"]}, {
             StyleBox[
              RowBox[{"t", "..."}], "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], ",", 
                   StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
           UnderscriptBox["\[Succeeds]", 
            RowBox[{"GBTermOrder", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 StyleBox[
                  RowBox[{"t", "..."}], "ExpressionVariable"], 
                 "\[RightAngleBracket]"}]}], 
             "\[RightAngleBracket]"}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["T", "ExpressionVariable"]}, {
             StyleBox["t", "ExpressionVariable"]}, {
             StyleBox[
              RowBox[{"t", "..."}], "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], ",", 
                   StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
           UnderscriptBox["\[Succeeds]", 
            RowBox[{"GBTermOrder", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["T", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 StyleBox[
                  RowBox[{"t", "..."}], "ExpressionVariable"], 
                 "\[RightAngleBracket]"}]}], 
             "\[RightAngleBracket]"}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1350364326"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1350364326-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1350364326-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1350364326", "Proof|ID:1350364326-1"},
 CellID->47470875],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1350364326-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->396433889]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[ScriptCapitalL] \[Succeeds] 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[ScriptCapitalL] \[Succeeds] 3)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["h", "ExpressionVariable"]}, {
             StyleBox["T1", "ExpressionVariable"]}, {
             StyleBox["T2", "ExpressionVariable"]}, {
             StyleBox["S1", "ExpressionVariable"]}, {
             StyleBox["S2", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["B", "ExpressionVariable"], 
             UnderscriptBox["\[SquareSubset]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], "\:293a", 
               StyleBox["h", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["T1", "ExpressionVariable"], ",", 
                 StyleBox["S1", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"GBTermOrder", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "\:293a", 
                   StyleBox["h", "ExpressionVariable"]}], ",", 
                 StyleBox["T2", "ExpressionVariable"], ",", 
                 StyleBox["S2", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["h", "ExpressionVariable"]}, {
             StyleBox["T1", "ExpressionVariable"]}, {
             StyleBox["T2", "ExpressionVariable"]}, {
             StyleBox["S1", "ExpressionVariable"]}, {
             StyleBox["S2", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["B", "ExpressionVariable"], 
             UnderscriptBox["\[SquareSubset]", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], "\:293a", 
               StyleBox["h", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["T1", "ExpressionVariable"], ",", 
                 StyleBox["S1", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"GBTermOrder", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "\:293a", 
                   StyleBox["h", "ExpressionVariable"]}], ",", 
                 StyleBox["T2", "ExpressionVariable"], ",", 
                 StyleBox["S2", "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:529848255"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p529848255-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p529848255-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:529848255", "Proof|ID:529848255-1"},
 CellID->196790701],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p529848255-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->478260070]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[SquareSubset] append trd)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[SquareSubset] append trd)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mi", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mj", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["let", 
            GridBox[{{
               RowBox[{
                 StyleBox["h", "ExpressionVariable"], "=", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}]}]}}]], 
           RowBox[{
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   StyleBox["h", "ExpressionVariable"], "\[Equal]", 
                   UnderscriptBox["0", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]]}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], 
               UnderscriptBox["\[SquareSubset]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], "\:293a", 
                 StyleBox[
                 "h", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mi", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mj", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["let", 
            GridBox[{{
               RowBox[{
                 StyleBox["h", "ExpressionVariable"], "=", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}]}]}}]], 
           RowBox[{
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   StyleBox["h", "ExpressionVariable"], "\[Equal]", 
                   UnderscriptBox["0", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]]}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], 
               UnderscriptBox["\[SquareSubset]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], "\:293a", 
                 StyleBox[
                 "h", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:592062547"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p592062547-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p592062547-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:592062547", "Proof|ID:592062547-1"},
 CellID->1513019922],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p592062547-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{3.6616611046758256`*^9},
 CellID->2131077418]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Induction (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.649049664438673*^9, 3.6490496699751596`*^9}},
 CellID->1185760220],

Cell["", "OpenEnvironment",
 CellID->2057862047],

Cell[CellGroupData[{

Cell["Definition: Auxiliary Notions for Induction", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648443305708139*^9, 3.648443315618433*^9}},
 CellID->886140291],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"Q", ",", "C", ",", "i", ",", "j"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1848949834}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648465421289154*^9, 3.648465423577944*^9}, {
  3.648522681972294*^9, 3.648522683109395*^9}},
 CellTags->{
  "ID:1848949834", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1848949834],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"a", ",", "mi", ",", "mj", ",", "h", ",", "P"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["GBAuxTest", "\[ScriptCapitalG]"], "[", 
     RowBox[{
     "C", ",", "a", ",", "mi", ",", "mj", ",", "h", ",", "P", ",", "i", ",", 
      "j"}], "]"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Or]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"h", "\[Equal]", 
           UnderscriptBox["0", "\[ScriptCapitalG]"]}]},
         {
          RowBox[{
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{
              SubscriptBox["C", "i"], "\[Equal]", 
              SubscriptBox["C", "j"]}], "\[And]", 
             RowBox[{"mi", "\[Equal]", "mj"}]}], 
            TagBox[")",
             "AutoParentheses"]}], "\[Or]", 
           RowBox[{
            UnderscriptBox["chainCrit", "\[ScriptCapitalG]"], "[", 
            RowBox[{"i", ",", "j", ",", "a", ",", "C", ",", "P"}], "]"}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxTest)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1362258751", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6483728529342413`*^9, 3.648373064555149*^9}, {
   3.648373099372447*^9, 3.648373134718347*^9}, {3.648373309042426*^9, 
   3.6483733362562447`*^9}, {3.648377243450535*^9, 3.6483772452785053`*^9}, {
   3.648442267114406*^9, 3.6484422754862633`*^9}, {3.648443137967462*^9, 
   3.64844314487714*^9}, {3.648465324094946*^9, 3.648465373986161*^9}, {
   3.64846543139015*^9, 3.648465431789815*^9}, {3.6485291827110023`*^9, 
   3.648529218167674*^9}, 3.648529595500146*^9, 3.648879875717589*^9},
 CellTags->{
  "ID:1362258751", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxTest"},
 CellID->1362258751],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["GBAuxBase", "\[ScriptCapitalG]"], "[", 
   RowBox[{"Q", ",", "C", ",", "i", ",", "j"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{"Q", "[", 
   RowBox[{"C", ",", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "i", ",", 
    "j"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxBase)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1347349813", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6483726043471127`*^9, 3.64837265844622*^9}, {
  3.648373115983614*^9, 3.648373117470492*^9}, {3.648465436218758*^9, 
  3.64846543917804*^9}, {3.648522630541092*^9, 3.648522649675871*^9}, {
  3.648522687529417*^9, 3.648522695874789*^9}},
 CellTags->{
  "ID:1347349813", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxBase"},
 CellID->1347349813],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["GBAuxRec1", "\[ScriptCapitalG]"], "[", 
   RowBox[{"Q", ",", "C", ",", "i", ",", "j"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"R", "..."}]], 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", 
        RowBox[{"NatPairs", "[", 
         RowBox[{
          RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], ",", 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           TagBox["\[LessEqual]",
            Identity], 
           TagBox[")",
            "AutoParentheses"]}]}], "]"}], "]"}]], 
      RowBox[{"[", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"R", "..."}], "\[RightAngleBracket]"}], "]"}]}]], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{"k", ",", 
        RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
      RowBox[{"l", "\[GreaterEqual]", "k"}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{
           RowBox[{
            SubscriptBox["C", "k"], "\[Equal]", 
            UnderscriptBox["0", "\[ScriptCapitalG]"]}], "\[Or]", 
           RowBox[{
            SubscriptBox["C", "l"], "\[Equal]", 
            UnderscriptBox["0", "\[ScriptCapitalG]"]}]}], 
          TagBox[")",
           "AutoParentheses"]}], "\[And]", 
         RowBox[{"Q", "[", 
          RowBox[{"C", ",", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"R", "..."}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "k", 
           ",", "l"}], "]"}], "\[And]", 
         RowBox[{"Q", "[", 
          RowBox[{"C", ",", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"R", "..."}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "i", 
           ",", "j"}], "]"}]}], 
        TagBox[")",
         "AutoParentheses"]}], "\[Implies]", 
       RowBox[{"Q", "[", 
        RowBox[{"C", ",", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"k", ",", "l"}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"R", "..."}]}], "\[RightAngleBracket]"}], ",", 
         RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "i", 
         ",", "j"}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxRec1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:270708621", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6482947109402227`*^9, 3.648294756792984*^9}, {
  3.648294980187599*^9, 3.648295007643915*^9}, {3.6483719045475273`*^9, 
  3.6483719411698217`*^9}, {3.6483719716510153`*^9, 3.6483719773004093`*^9}, {
  3.648372705250742*^9, 3.648372782166644*^9}, {3.648373119313552*^9, 
  3.6483731211364927`*^9}, {3.648465446395516*^9, 3.648465460366131*^9}, {
  3.648522741717949*^9, 3.6485227784666653`*^9}, {3.648529873193864*^9, 
  3.648529927343768*^9}, {3.64853011832135*^9, 3.648530147124453*^9}, {
  3.648982414542975*^9, 3.648982414997814*^9}, {3.648982520456401*^9, 
  3.648982549370079*^9}, {3.64898263166751*^9, 3.6489826320016108`*^9}, {
  3.648985811744608*^9, 3.648985821327817*^9}},
 CellTags->{
  "ID:270708621", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec1"},
 CellID->270708621],

Cell[TextData[{
 "In the premise of GBAuxRec2 we could even assume that ",
 Cell[BoxData[
  FormBox["Q", TraditionalForm]]],
 " holds for ",
 StyleBox["all",
  FontSlant->"Italic"],
 " ",
 Cell[BoxData[
  FormBox[
   RowBox[{"MMPairs", "[", 
    RowBox[{"\[ScriptCapitalG]", ",", 
     SubscriptBox["B", "k"], ",", 
     SubscriptBox["B", "l"]}], "]"}], TraditionalForm]]],
 "-tuples, not only for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"lmTuples", "[", 
    RowBox[{
     SubscriptBox["B", "k"], ",", 
     SubscriptBox["B", "l"]}], "]"}], TraditionalForm]]],
 "."
}], "EnvironmentText",
 CellChangeTimes->{{3.6484561018470173`*^9, 3.648456141794731*^9}, {
   3.648456180465296*^9, 3.648456191101942*^9}, 3.6485300013689947`*^9},
 CellID->1510189147],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["GBAuxRec2", "\[ScriptCapitalG]"], "[", 
   RowBox[{"Q", ",", "C", ",", "i", ",", "j"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"R", "..."}]], 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", 
        RowBox[{"NatPairs", "[", 
         RowBox[{
          RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], ",", 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           TagBox["\[LessEqual]",
            Identity], 
           TagBox[")",
            "AutoParentheses"]}]}], "]"}], "]"}]], 
      RowBox[{"[", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"R", "..."}], "\[RightAngleBracket]"}], "]"}]}]], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{"k", ",", 
        RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
      RowBox[{"l", "\[GreaterEqual]", "k"}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{
          SubscriptBox["C", "k"], "\[NotEqual]", 
          UnderscriptBox["0", "\[ScriptCapitalG]"]}], "\[And]", 
         RowBox[{
          SubscriptBox["C", "l"], "\[NotEqual]", 
          UnderscriptBox["0", "\[ScriptCapitalG]"]}], "\[And]", 
         RowBox[{"Q", "[", 
          RowBox[{"C", ",", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"R", "..."}], "\[RightAngleBracket]"}], ",", 
           RowBox[{
            UnderscriptBox["lmTuple", "\[ScriptCapitalG]"], "[", 
            RowBox[{
             SubscriptBox["C", "k"], ",", 
             SubscriptBox["C", "l"]}], "]"}], ",", "k", ",", "l"}], "]"}]}], 
        TagBox[")",
         "AutoParentheses"]}], "\[Implies]", 
       RowBox[{"Q", "[", 
        RowBox[{"C", ",", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"k", ",", "l"}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"R", "..."}]}], "\[RightAngleBracket]"}], ",", 
         RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "i", 
         ",", "j"}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxRec2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1542247232", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648529841084799*^9, 3.6485298453296824`*^9}, {
  3.648529960540332*^9, 3.648529985881311*^9}, {3.648530149926775*^9, 
  3.648530159797118*^9}},
 CellTags->{
  "ID:1542247232", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec2"},
 CellID->1542247232],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["GBAuxRec3", "\[ScriptCapitalG]"], "[", 
   RowBox[{"Q", ",", "C", ",", "i", ",", "j"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", 
        RowBox[{"NatPairs", "[", 
         RowBox[{
          RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], ",", 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           TagBox["\[LessEqual]",
            Identity], 
           TagBox[")",
            "AutoParentheses"]}]}], "]"}], "]"}]], 
      RowBox[{"[", "P", "]"}]}]], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{"a", ",", "mi", ",", "mj"}]], 
      RowBox[{
       UnderscriptBox["\[Element]", 
        RowBox[{"MMPairs", "[", 
         RowBox[{"\[ScriptCapitalG]", ",", 
          SubscriptBox["C", "i"], ",", 
          SubscriptBox["C", "j"]}], "]"}]], 
       RowBox[{"[", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"a", ",", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
         "\[RightAngleBracket]"}], "]"}]}]], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{"R", "..."}]], 
       RowBox[{
        UnderscriptBox["\[Element]", 
         RowBox[{"DomainTuples", "[", 
          RowBox[{"MMPairs", "[", 
           RowBox[{"\[ScriptCapitalG]", ",", 
            SubscriptBox["C", "i"], ",", 
            SubscriptBox["C", "j"]}], "]"}], "]"}]], 
        RowBox[{"[", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"R", "..."}], "\[RightAngleBracket]"}], "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          RowBox[{"Q", "[", 
           RowBox[{"C", ",", "P", ",", 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"R", "..."}], "\[RightAngleBracket]"}], ",", "i", ",", 
            "j"}], "]"}], "\[And]", 
          RowBox[{
           UnderscriptBox["GBAuxTest", "\[ScriptCapitalG]"], "[", 
           RowBox[{"C", ",", "a", ",", "mi", ",", "mj", ",", 
            RowBox[{
             RowBox[{
              UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
              RowBox[{
               RowBox[{"a", 
                UnderscriptBox["-", "\[ScriptCapitalG]"], 
                RowBox[{"mi", 
                 UnderscriptBox["*", "\[ScriptCapitalG]"], 
                 SubscriptBox["C", "i"]}]}], ",", "C"}], "]"}], 
             UnderscriptBox["-", "\[ScriptCapitalG]"], 
             RowBox[{
              UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
              RowBox[{
               RowBox[{"a", 
                UnderscriptBox["-", "\[ScriptCapitalG]"], 
                RowBox[{"mj", 
                 UnderscriptBox["*", "\[ScriptCapitalG]"], 
                 SubscriptBox["C", "j"]}]}], ",", "C"}], "]"}]}], ",", "P", 
            ",", "i", ",", "j"}], "]"}]}], 
         TagBox[")",
          "AutoParentheses"]}], "\[Implies]", 
        RowBox[{"Q", "[", 
         RowBox[{"C", ",", "P", ",", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"a", ",", 
              RowBox[{"\[LeftAngleBracket]", 
               RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
             "\[RightAngleBracket]"}], ",", 
            RowBox[{"R", "..."}]}], "\[RightAngleBracket]"}], ",", "i", ",", 
          "j"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxRec3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:580452282", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648294762307164*^9, 3.648294810941917*^9}, {
   3.648294848980021*^9, 3.648294855333346*^9}, {3.648295009888998*^9, 
   3.648295031541606*^9}, {3.648371996415988*^9, 3.648372031762331*^9}, {
   3.648373145410158*^9, 3.64837323064084*^9}, {3.6483734022302856`*^9, 
   3.6483734478727913`*^9}, 3.64845597519707*^9, {3.64846546295147*^9, 
   3.648465489155237*^9}, {3.648522783467369*^9, 3.6485227865308113`*^9}, 
   3.648529807560176*^9, {3.648530224473877*^9, 3.648530236660924*^9}},
 CellTags->{
  "ID:580452282", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec3"},
 CellID->580452282],

Cell[TextData[{
 "In the premise of GBAuxRec4 we could even assume that ",
 Cell[BoxData[
  FormBox["Q", TraditionalForm]]],
 " holds for ",
 StyleBox["all",
  FontSlant->"Italic"],
 " ",
 Cell[BoxData[
  FormBox[
   RowBox[{"NatPairs", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], "+", 
      "1"}], ",", "\[LessEqual]"}], "]"}], TraditionalForm]]],
 "-tuples, not only for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"update", "[", 
    RowBox[{"P", ",", 
     RowBox[{
      RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], "+", 
      "1"}]}], "]"}], TraditionalForm]]],
 "."
}], "EnvironmentText",
 CellChangeTimes->{{3.648456017996563*^9, 3.64845607951654*^9}, {
   3.648456110537624*^9, 3.6484561109482117`*^9}, 3.648530016123304*^9},
 CellID->2063275027],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["GBAuxRec4", "\[ScriptCapitalG]"], "[", 
   RowBox[{"Q", ",", "C", ",", "i", ",", "j"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", 
        RowBox[{"NatPairs", "[", 
         RowBox[{
          RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], ",", 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           TagBox["\[LessEqual]",
            Identity], 
           TagBox[")",
            "AutoParentheses"]}]}], "]"}], "]"}]], 
      RowBox[{"[", "P", "]"}]}]], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{"a", ",", "mi", ",", "mj"}]], 
      RowBox[{
       UnderscriptBox["\[Element]", 
        RowBox[{"MMPairs", "[", 
         RowBox[{"\[ScriptCapitalG]", ",", 
          SubscriptBox["C", "i"], ",", 
          SubscriptBox["C", "j"]}], "]"}]], 
       RowBox[{"[", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"a", ",", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
         "\[RightAngleBracket]"}], "]"}]}]], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{"R", "..."}]], 
       RowBox[{
        UnderscriptBox["\[Element]", 
         RowBox[{"DomainTuples", "[", 
          RowBox[{"MMPairs", "[", 
           RowBox[{"\[ScriptCapitalG]", ",", 
            SubscriptBox["C", "i"], ",", 
            SubscriptBox["C", "j"]}], "]"}], "]"}]], 
        RowBox[{"[", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"R", "..."}], "\[RightAngleBracket]"}], "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox["let", 
         RowBox[{"h", "=", 
          RowBox[{
           RowBox[{
            UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
            RowBox[{
             RowBox[{"a", 
              UnderscriptBox["-", "\[ScriptCapitalG]"], 
              RowBox[{"mi", 
               UnderscriptBox["*", "\[ScriptCapitalG]"], 
               SubscriptBox["C", "i"]}]}], ",", "C"}], "]"}], 
           UnderscriptBox["-", "\[ScriptCapitalG]"], 
           RowBox[{
            UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
            RowBox[{
             RowBox[{"a", 
              UnderscriptBox["-", "\[ScriptCapitalG]"], 
              RowBox[{"mj", 
               UnderscriptBox["*", "\[ScriptCapitalG]"], 
               SubscriptBox["C", "j"]}]}], ",", "C"}], "]"}]}]}]], 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{
            RowBox[{"Q", "[", 
             RowBox[{
              RowBox[{"C", 
               TagBox["\:293a",
                Identity,
                SyntaxForm->"a*b"], "h"}], ",", 
              RowBox[{"update", "[", 
               RowBox[{"P", ",", 
                RowBox[{
                 RowBox[{
                 "\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], "+", 
                 "1"}]}], "]"}], ",", 
              RowBox[{"\[LeftAngleBracket]", 
               RowBox[{"R", "..."}], "\[RightAngleBracket]"}], ",", "i", ",", 
              "j"}], "]"}], "\[And]", 
            RowBox[{"\[Not]", 
             RowBox[{
              UnderscriptBox["GBAuxTest", "\[ScriptCapitalG]"], "[", 
              RowBox[{
              "C", ",", "a", ",", "mi", ",", "mj", ",", "h", ",", "P", ",", 
               "i", ",", "j"}], "]"}]}]}], 
           TagBox[")",
            "AutoParentheses"]}], "\[Implies]", 
          RowBox[{"Q", "[", 
           RowBox[{"C", ",", "P", ",", 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
              RowBox[{"\[LeftAngleBracket]", 
               RowBox[{"a", ",", 
                RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
               "\[RightAngleBracket]"}], ",", 
              RowBox[{"R", "..."}]}], "\[RightAngleBracket]"}], ",", "i", ",",
             "j"}], "]"}]}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxRec4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:598818422", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648373235143937*^9, 3.648373300681739*^9}, {
   3.648373346775165*^9, 3.6483734172283278`*^9}, 3.6483737340452213`*^9, 
   3.6484559620248127`*^9, {3.6484654921844397`*^9, 3.648465527743559*^9}, {
   3.648522789281321*^9, 3.648522792473755*^9}, 3.648529809442276*^9},
 CellTags->{
  "ID:598818422", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec4"},
 CellID->598818422],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2012993027]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1217706098],

Cell["", "OpenEnvironment",
 CellID->1578468637],

Cell[CellGroupData[{

Cell["Theorem: Induction Schema for Function GBAux", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648292691808813*^9, 3.648292704984788*^9}, 
   3.648373572833436*^9},
 CellID->336243787],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", "Q"]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1392375551}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648294632352338*^9, 3.648294633369321*^9}},
 CellTags->{
  "ID:1392375551", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1392375551],

Cell[BoxData[
 RowBox[{
  TagBox["(",
   "AutoParentheses"], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"\[And]", 
     RowBox[{"\[Piecewise]", GridBox[{
        {
         RowBox[{
          UnderscriptBox["\[ForAll]", 
           RowBox[{
            UnderscriptBox["\[Element]", 
             RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
            RowBox[{"[", "C", "]"}]}]], 
          RowBox[{"Q", "[", 
           RowBox[{"C", ",", 
            RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", 
            RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
           "]"}]}]},
        {
         RowBox[{
          UnderscriptBox["\[ForAll]", 
           RowBox[{
            UnderscriptBox["\[Element]", 
             RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
            RowBox[{"[", "C", "]"}]}]], 
          RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             RowBox[{"t", ",", 
              RowBox[{"t", "..."}]}]], 
            RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
               RowBox[{"ProdDomain", "[", 
                RowBox[{
                "\[DoubleStruckCapitalZ]Dom", ",", 
                 "\[DoubleStruckCapitalZ]Dom"}], "]"}], "]"}]], 
             RowBox[{"[", 
              RowBox[{"\[LeftAngleBracket]", 
               RowBox[{"t", ",", 
                RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], "]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{
               UnderscriptBox["\[ForAll]", 
                RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainTuples", "[", 
                   RowBox[{"ProdDomain", "[", 
                    RowBox[{"\[ScriptCapitalG]", ",", 
                    RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]}],
                     "]"}], "]"}]], 
                 RowBox[{"[", "T", "]"}]}]], 
               RowBox[{"Q", "[", 
                RowBox[{"C", ",", 
                 RowBox[{"\[LeftAngleBracket]", 
                  RowBox[{"t", "..."}], "\[RightAngleBracket]"}], ",", "T"}], 
                "]"}]}], 
              TagBox[")",
               "AutoParentheses"]}], "\[Implies]", 
             RowBox[{"Q", "[", 
              RowBox[{"C", ",", 
               RowBox[{"\[LeftAngleBracket]", 
                RowBox[{"t", ",", 
                 RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], ",", 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
              "]"}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}]},
        {
         RowBox[{
          UnderscriptBox["\[ForAll]", 
           RowBox[{
            UnderscriptBox["\[Element]", 
             RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
            RowBox[{"[", "C", "]"}]}]], 
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
               RowBox[{"ProdDomain", "[", 
                RowBox[{
                "\[DoubleStruckCapitalZ]Dom", ",", 
                 "\[DoubleStruckCapitalZ]Dom"}], "]"}], "]"}]], 
             RowBox[{"[", "P", "]"}]}]], 
           RowBox[{
            UnderscriptBox[
             UnderscriptBox["\[ForAll]", 
              RowBox[{"t", ",", 
               RowBox[{"t", "..."}]}]], 
             RowBox[{
              UnderscriptBox["\[Element]", 
               RowBox[{"DomainTuples", "[", 
                RowBox[{"ProdDomain", "[", 
                 RowBox[{"\[ScriptCapitalG]", ",", 
                  RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]}], 
                 "]"}], "]"}]], 
              RowBox[{"[", 
               RowBox[{"\[LeftAngleBracket]", 
                RowBox[{"t", ",", 
                 RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], "]"}]}]], 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             RowBox[{
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               RowBox[{
                RowBox[{"Q", "[", 
                 RowBox[{"C", ",", "P", ",", 
                  RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{"t", "..."}], "\[RightAngleBracket]"}]}], "]"}], 
                "\[And]", 
                RowBox[{
                 TagBox["(",
                  "AutoParentheses"], 
                 RowBox[{
                  UnderscriptBox[
                   UnderscriptBox["\[ForAll]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
                    RowBox[{"[", "h", "]"}]}], ",", 
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    RowBox[{
                    "\[DoubleStruckCapitalZ]Dom", ",", 
                    "\[DoubleStruckCapitalZ]Dom"}], "]"}], "]"}]], 
                    RowBox[{"[", "T", "]"}]}], ",", 
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    RowBox[{"\[ScriptCapitalG]", ",", 
                    RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]}],
                     "]"}], "]"}]], 
                    RowBox[{"[", "S", "]"}]}]}]], 
                   RowBox[{"C", 
                    UnderscriptBox["\[SquareSubset]", "\[ScriptCapitalG]"], 
                    RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    RowBox[{"C", 
                    TagBox["\:293a",
                    Identity,
                    SyntaxForm->"a*b"], "h"}], 
                    TagBox[")",
                    "AutoParentheses"]}]}]], 
                  RowBox[{"Q", "[", 
                   RowBox[{
                    RowBox[{"C", 
                    TagBox["\:293a",
                    Identity,
                    SyntaxForm->"a*b"], "h"}], ",", "T", ",", "S"}], "]"}]}], 
                 
                 TagBox[")",
                  "AutoParentheses"]}]}], 
               TagBox[")",
                "AutoParentheses"]}], "\[Implies]", 
              RowBox[{"Q", "[", 
               RowBox[{"C", ",", "P", ",", 
                RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{"t", ",", 
                  RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}]}], "]"}]}], 
             TagBox[")",
              "AutoParentheses"]}]}]}]}]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}},
          "RowsIndexed" -> {}}]}]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[Implies]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalL]"], 
       RowBox[{"[", "x", "]"}]}]], 
     RowBox[{"Q", "[", 
      RowBox[{
       SubscriptBox["x", "1"], ",", 
       SubscriptBox["x", "2"], ",", 
       SubscriptBox["x", "3"]}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}], 
  TagBox[")",
   "AutoParentheses"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:994411308", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmViYGAQBWIQLXTylvENvdeO0yp2WYDoDUrJTiDavPZ2MIgWWfIr
HETzyEX2guidywtmgOgH0l7zQPTS72xLQbRPzpoVIHplW/t5EN1wegKYNrz4
j+UmkF4WocEKogOebhQB0QVccaIg2uFuoRyI1mAxUQfRZUcKqkF0n+87MO3S
uXDNLSB94J3WBhA9/bjfRxA9Z0XmZxCd0lL+HUTXnSxivA2kw1Yw84DocwJp
vCCaxWe2CIg+2bwYTGfdUJcH0Xus7yiA6H+SU8pBNNPRyRUgumz/xK6HQPpP
9+1uEP0q9Xo/iG5693AiiN5lvmSCoAfQniqxhSDadKnkOhD9SLdzA4gGAMh3
rZQ=
  "],
 CellTags->{
  "ID:994411308", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "1"},
 CellID->994411308],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
           RowBox[{"[", "C", "]"}]}]], 
         RowBox[{
          UnderscriptBox[
           UnderscriptBox["\[ForAll]", 
            RowBox[{"i", ",", 
             RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
             "\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
           RowBox[{"j", "\[GreaterEqual]", "i"}]], 
          RowBox[{"Q", "[", 
           RowBox[{"C", ",", 
            RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", 
            RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "i",
             ",", "j"}], "]"}]}]}]},
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
           RowBox[{"[", "C", "]"}]}]], 
         RowBox[{
          UnderscriptBox[
           UnderscriptBox["\[ForAll]", 
            RowBox[{"i", ",", 
             RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
             "\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
           RowBox[{"j", "\[GreaterEqual]", "i"}]], 
          RowBox[{
           UnderscriptBox[
            UnderscriptBox["\[ForAll]", 
             RowBox[{"p", "..."}]], 
            RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
               RowBox[{"NatPairs", "[", 
                RowBox[{
                 RowBox[{
                 "\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], ",", 
                 
                 RowBox[{
                  TagBox["(",
                   "AutoParentheses"], 
                  TagBox["\[LessEqual]",
                   Identity], 
                  TagBox[")",
                   "AutoParentheses"]}]}], "]"}], "]"}]], 
             RowBox[{"[", 
              RowBox[{"\[LeftAngleBracket]", 
               RowBox[{"p", "..."}], "\[RightAngleBracket]"}], "]"}]}]], 
           RowBox[{
            UnderscriptBox[
             UnderscriptBox["\[ForAll]", 
              RowBox[{"k", ",", 
               RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
               RowBox[{
               "\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
             RowBox[{"l", "\[GreaterEqual]", "k"}]], 
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             RowBox[{
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               RowBox[{
                UnderscriptBox[
                 UnderscriptBox["\[ForAll]", 
                  RowBox[{"i0", ",", 
                   RowBox[{"j0", "=", "1"}], ",", "\[Ellipsis]", ",", 
                   RowBox[{
                   "\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
                 
                 RowBox[{"j0", "\[GreaterEqual]", "i0"}]], 
                RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"MMPairs", "[", 
                    RowBox[{"\[ScriptCapitalG]", ",", 
                    SubscriptBox["C", "i0"], ",", 
                    SubscriptBox["C", "j0"]}], "]"}], "]"}]], 
                   RowBox[{"[", "M", "]"}]}]], 
                 RowBox[{"Q", "[", 
                  RowBox[{"C", ",", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{"p", "..."}], "\[RightAngleBracket]"}], ",", "M", 
                   ",", "i0", ",", "j0"}], "]"}]}]}], 
               TagBox[")",
                "AutoParentheses"]}], "\[Implies]", 
              RowBox[{"Q", "[", 
               RowBox[{"C", ",", 
                RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                  RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{"i", ",", "j"}], "\[RightAngleBracket]"}], ",", 
                  RowBox[{"p", "..."}]}], "\[RightAngleBracket]"}], ",", 
                RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", 
                "k", ",", "l"}], "]"}]}], 
             TagBox[")",
              "AutoParentheses"]}]}]}]}]}]},
       {
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          UnderscriptBox["\[ForAll]", 
           RowBox[{
            UnderscriptBox["\[Element]", 
             RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
            RowBox[{"[", "C", "]"}]}]], 
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
               RowBox[{"NatPairs", "[", 
                RowBox[{
                 RowBox[{
                 "\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], ",", 
                 
                 RowBox[{
                  TagBox["(",
                   "AutoParentheses"], 
                  TagBox["\[LessEqual]",
                   Identity], 
                  TagBox[")",
                   "AutoParentheses"]}]}], "]"}], "]"}]], 
             RowBox[{"[", "P", "]"}]}]], 
           RowBox[{
            UnderscriptBox[
             UnderscriptBox["\[ForAll]", 
              RowBox[{"i", ",", 
               RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
               RowBox[{
               "\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
             RowBox[{"j", "\[GreaterEqual]", "i"}]], 
            RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               RowBox[{"a", ",", "mi", ",", "mj"}]], 
              RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"MMPairs", "[", 
                 RowBox[{"\[ScriptCapitalG]", ",", 
                  SubscriptBox["C", "i"], ",", 
                  SubscriptBox["C", "j"]}], "]"}]], 
               RowBox[{"[", 
                RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{"a", ",", 
                  RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
                 "\[RightAngleBracket]"}], "]"}]}]], 
             RowBox[{
              UnderscriptBox[
               UnderscriptBox["\[ForAll]", 
                RowBox[{"m", "..."}]], 
               RowBox[{
                UnderscriptBox["\[Element]", 
                 RowBox[{"DomainTuples", "[", 
                  RowBox[{"MMPairs", "[", 
                   RowBox[{"\[ScriptCapitalG]", ",", 
                    SubscriptBox["C", "i"], ",", 
                    SubscriptBox["C", "j"]}], "]"}], "]"}]], 
                RowBox[{"[", 
                 RowBox[{"\[LeftAngleBracket]", 
                  RowBox[{"m", "..."}], "\[RightAngleBracket]"}], "]"}]}]], 
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               RowBox[{
                UnderscriptBox["let", 
                 RowBox[{"h", "=", 
                  RowBox[{
                   RowBox[{
                    UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
                    RowBox[{
                    RowBox[{"a", 
                    UnderscriptBox["-", "\[ScriptCapitalG]"], 
                    RowBox[{"mi", 
                    UnderscriptBox["*", "\[ScriptCapitalG]"], 
                    SubscriptBox["C", "i"]}]}], ",", "C"}], "]"}], 
                   UnderscriptBox["-", "\[ScriptCapitalG]"], 
                   RowBox[{
                    UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
                    RowBox[{
                    RowBox[{"a", 
                    UnderscriptBox["-", "\[ScriptCapitalG]"], 
                    RowBox[{"mj", 
                    UnderscriptBox["*", "\[ScriptCapitalG]"], 
                    SubscriptBox["C", "j"]}]}], ",", "C"}], "]"}]}]}]], 
                RowBox[{
                 TagBox["(",
                  "AutoParentheses"], 
                 RowBox[{"\[And]", 
                  RowBox[{"\[Piecewise]", GridBox[{
                    {
                    RowBox[{
                    RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["GBAuxTest", "\[ScriptCapitalG]"], "[", 
                    RowBox[{
                    "C", ",", "a", ",", "mi", ",", "mj", ",", "h", ",", "P", 
                    ",", "i", ",", "j"}], "]"}], "\[And]", 
                    RowBox[{"Q", "[", 
                    RowBox[{"C", ",", "P", ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{"m", "..."}], "\[RightAngleBracket]"}], ",", "i", 
                    ",", "j"}], "]"}]}], 
                    TagBox[")",
                    "AutoParentheses"]}], "\[Implies]", 
                    RowBox[{"Q", "[", 
                    RowBox[{"C", ",", "P", ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{"a", ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
                    "\[RightAngleBracket]"}], ",", 
                    RowBox[{"m", "..."}]}], "\[RightAngleBracket]"}], ",", 
                    "i", ",", "j"}], "]"}]}]},
                    {
                    RowBox[{
                    RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    RowBox[{
                    RowBox[{"\[Not]", 
                    RowBox[{
                    UnderscriptBox["GBAuxTest", "\[ScriptCapitalG]"], "[", 
                    RowBox[{
                    "C", ",", "a", ",", "mi", ",", "mj", ",", "h", ",", "P", 
                    ",", "i", ",", "j"}], "]"}]}], "\[And]", 
                    RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    "\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], 
                    "+", "1"}], ",", 
                    RowBox[{
                    TagBox["(",
                    "AutoParentheses"], 
                    TagBox["\[LessEqual]",
                    Identity], 
                    TagBox[")",
                    "AutoParentheses"]}]}], "]"}], "]"}]], 
                    RowBox[{"[", "P0", "]"}]}]], 
                    RowBox[{"Q", "[", 
                    RowBox[{
                    RowBox[{"C", 
                    TagBox["\:293a",
                    Identity,
                    SyntaxForm->"a*b"], "h"}], ",", "P0", ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{"m", "..."}], "\[RightAngleBracket]"}], ",", "i", 
                    ",", "j"}], "]"}]}], 
                    TagBox[")",
                    "AutoParentheses"]}]}], 
                    TagBox[")",
                    "AutoParentheses"]}], "\[Implies]", 
                    RowBox[{"Q", "[", 
                    RowBox[{"C", ",", "P", ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{"a", ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
                    "\[RightAngleBracket]"}], ",", 
                    RowBox[{"m", "..."}]}], "\[RightAngleBracket]"}], ",", 
                    "i", ",", "j"}], "]"}]}]}
                    },
                    
                    GridBoxAlignment->{
                    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                    "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
                 TagBox[")",
                  "AutoParentheses"]}]}], 
               TagBox[")",
                "AutoParentheses"]}]}]}]}]}]}], 
         TagBox[")",
          "AutoParentheses"]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalL]"], 
     RowBox[{"[", "x", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{"i", ",", 
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", 
         SubscriptBox["x", "1"], "\[RightBracketingBar]"}]}]], 
      RowBox[{"j", "\[GreaterEqual]", "i"}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{
          UnderscriptBox["\[Element]", 
           RowBox[{"DomainTuples", "[", 
            RowBox[{"NatPairs", "[", 
             RowBox[{
              RowBox[{"\[LeftBracketingBar]", 
               SubscriptBox["x", "1"], "\[RightBracketingBar]"}], ",", 
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               TagBox["\[LessEqual]",
                Identity], 
               TagBox[")",
                "AutoParentheses"]}]}], "]"}], "]"}]], 
          RowBox[{"[", 
           SubscriptBox["x", "2"], "]"}]}], "\[And]", 
         RowBox[{
          UnderscriptBox["\[Element]", 
           RowBox[{"DomainTuples", "[", 
            RowBox[{"MMPairs", "[", 
             RowBox[{"\[ScriptCapitalG]", ",", 
              SubscriptBox[
               RowBox[{"(", 
                SubscriptBox["x", "1"], ")"}], "i"], ",", 
              SubscriptBox[
               RowBox[{"(", 
                SubscriptBox["x", "1"], ")"}], "j"]}], "]"}], "]"}]], 
          RowBox[{"[", 
           SubscriptBox["x", "3"], "]"}]}]}], 
        TagBox[")",
         "AutoParentheses"]}], "\[Implies]", 
       RowBox[{"Q", "[", 
        RowBox[{
         SubscriptBox["x", "1"], ",", 
         SubscriptBox["x", "2"], ",", 
         SubscriptBox["x", "3"], ",", "i", ",", "j"}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1185976259", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64844146937317*^9, 3.648441611487358*^9}, {
   3.6484417115889397`*^9, 3.648441741362029*^9}, {3.648441776136695*^9, 
   3.648441791388002*^9}, {3.64844182993484*^9, 3.6484420076388702`*^9}, {
   3.64844203970548*^9, 3.648442061002399*^9}, {3.648442101518256*^9, 
   3.648442121837467*^9}, {3.6484423631405153`*^9, 3.648442367308179*^9}, {
   3.648443161742689*^9, 3.6484431665923147`*^9}, {3.6484432078709917`*^9, 
   3.648443250402603*^9}, {3.648443439362527*^9, 3.648443505523432*^9}, {
   3.648443541573373*^9, 3.648443596665642*^9}, {3.648443820782257*^9, 
   3.648443835822658*^9}, {3.64845261288382*^9, 3.64845262415308*^9}, {
   3.6484526587906847`*^9, 3.6484526946810102`*^9}, {3.6484528014059143`*^9, 
   3.648452949498069*^9}, {3.648453047428954*^9, 3.648453233183402*^9}, {
   3.648453274973114*^9, 3.648453344328157*^9}, {3.648453532007277*^9, 
   3.648453532675378*^9}, {3.6484543966050797`*^9, 3.64845441115761*^9}, {
   3.64845446255431*^9, 3.6484544898493147`*^9}, {3.648455927231773*^9, 
   3.648455932094496*^9}, 3.648458735606452*^9, {3.6485246336168756`*^9, 
   3.648524671146842*^9}, {3.6485247143257647`*^9, 3.648524869217218*^9}, {
   3.648526223806601*^9, 3.648526247788383*^9}, {3.6485265176650133`*^9, 
   3.648526527782436*^9}, {3.6489827580465403`*^9, 3.648982765763702*^9}, {
   3.648982972556623*^9, 3.6489829732132187`*^9}, {3.648983972012768*^9, 
   3.648984009579343*^9}},
 CellTags->{
  "ID:1185976259", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "2"},
 CellID->1185976259],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
      RowBox[{"[", "C", "]"}]}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", 
        RowBox[{"DomainTuples", "[", 
         RowBox[{"NatPairs", "[", 
          RowBox[{
           RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], 
           ",", 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            TagBox["\[LessEqual]",
             Identity], 
            TagBox[")",
             "AutoParentheses"]}]}], "]"}], "]"}]], 
       RowBox[{"[", "P", "]"}]}]], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{"i", ",", 
         RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
       RowBox[{"j", "\[GreaterEqual]", "i"}]], 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         UnderscriptBox["\[Element]", 
          RowBox[{"DomainTuples", "[", 
           RowBox[{"MMPairs", "[", 
            RowBox[{"\[ScriptCapitalG]", ",", 
             SubscriptBox["C", "i"], ",", 
             SubscriptBox["C", "j"]}], "]"}], "]"}]], 
         RowBox[{"[", "M", "]"}]}]], 
       RowBox[{"Q", "[", 
        RowBox[{"C", ",", "P", ",", "M", ",", "i", ",", "j"}], "]"}]}]}]}]}], 
   
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalL]"], 
      RowBox[{"[", "x", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{"i", ",", 
         RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", 
          SubscriptBox["x", "1"], "\[RightBracketingBar]"}]}]], 
       RowBox[{"j", "\[GreaterEqual]", "i"}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", 
             RowBox[{"NatPairs", "[", 
              RowBox[{
               RowBox[{"\[LeftBracketingBar]", 
                SubscriptBox["x", "1"], "\[RightBracketingBar]"}], ",", 
               RowBox[{
                TagBox["(",
                 "AutoParentheses"], 
                TagBox["\[LessEqual]",
                 Identity], 
                TagBox[")",
                 "AutoParentheses"]}]}], "]"}], "]"}]], 
           RowBox[{"[", 
            SubscriptBox["x", "2"], "]"}]}], "\[And]", 
          RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", 
             RowBox[{"MMPairs", "[", 
              RowBox[{"\[ScriptCapitalG]", ",", 
               SubscriptBox[
                RowBox[{"(", 
                 SubscriptBox["x", "1"], ")"}], "i"], ",", 
               SubscriptBox[
                RowBox[{"(", 
                 SubscriptBox["x", "1"], ")"}], "j"]}], "]"}], "]"}]], 
           RowBox[{"[", 
            SubscriptBox["x", "3"], "]"}]}]}], 
         TagBox[")",
          "AutoParentheses"]}], "\[Implies]", 
        RowBox[{"Q", "[", 
         RowBox[{
          SubscriptBox["x", "1"], ",", 
          SubscriptBox["x", "2"], ",", 
          SubscriptBox["x", "3"], ",", "i", ",", "j"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1125858125", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6485242857317133`*^9, 3.648524290566832*^9}, {
  3.648524330799892*^9, 3.6485243702367764`*^9}, {3.648528075441805*^9, 
  3.64852810542109*^9}},
 CellTags->{
  "ID:1125858125", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "3"},
 CellID->1125858125],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
      RowBox[{"[", "C", "]"}]}]], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{"i", ",", 
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
      RowBox[{"j", "\[GreaterEqual]", "i"}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"\[And]", 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{
            UnderscriptBox["GBAuxBase", "\[ScriptCapitalG]"], "[", 
            RowBox[{"Q", ",", "C", ",", "i", ",", "j"}], "]"}]},
          {
           RowBox[{
            UnderscriptBox["GBAuxRec1", "\[ScriptCapitalG]"], "[", 
            RowBox[{"Q", ",", "C", ",", "i", ",", "j"}], "]"}]},
          {
           RowBox[{
            UnderscriptBox["GBAuxRec2", "\[ScriptCapitalG]"], "[", 
            RowBox[{"Q", ",", "C", ",", "i", ",", "j"}], "]"}]},
          {
           RowBox[{
            UnderscriptBox["GBAuxRec3", "\[ScriptCapitalG]"], "[", 
            RowBox[{"Q", ",", "C", ",", "i", ",", "j"}], "]"}]},
          {
           RowBox[{
            UnderscriptBox["GBAuxRec4", "\[ScriptCapitalG]"], "[", 
            RowBox[{"Q", ",", "C", ",", "i", ",", "j"}], "]"}]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
      RowBox[{"[", "C", "]"}]}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", 
        RowBox[{"DomainTuples", "[", 
         RowBox[{"NatPairs", "[", 
          RowBox[{
           RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], 
           ",", 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            TagBox["\[LessEqual]",
             Identity], 
            TagBox[")",
             "AutoParentheses"]}]}], "]"}], "]"}]], 
       RowBox[{"[", "P", "]"}]}]], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{"i", ",", 
         RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
       RowBox[{"j", "\[GreaterEqual]", "i"}]], 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         UnderscriptBox["\[Element]", 
          RowBox[{"DomainTuples", "[", 
           RowBox[{"MMPairs", "[", 
            RowBox[{"\[ScriptCapitalG]", ",", 
             SubscriptBox["C", "i"], ",", 
             SubscriptBox["C", "j"]}], "]"}], "]"}]], 
         RowBox[{"[", "M", "]"}]}]], 
       RowBox[{"Q", "[", 
        RowBox[{"C", ",", "P", ",", "M", ",", "i", ",", "j"}], "]"}]}]}]}]}], 
   
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAux induction)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1780855977", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmViYGAQA2IQPYuN/+0EvdeOSuHLuScC6T0ZLQIgetXnL6Ig+pKg
pByIFtt81ABEz5ncZgGiIy5JR4PozarLkkB0D2NFO4i2mrapB0SvqVHvA9EK
kvN2g2jedwf2gegvDbuPgegJJl/B9IXun6fB4reCzoDo4FdC/0H0Fp0MxklA
uuF0uiCIrq7wEwXRkwzvtW8H0v/q1TpA9JMy1wlgfpHxdBDdWf3jHohueCrz
CkR/eef4EUSf+BYGpm0qWL+D6FOZK8F0XvfBmB1A+v21mFgQ3THxhe0NIB0m
JWEHorslsv7eB9LFNsr/QbTMCyErRv3XjnIx921B9PblDr4guslaKwBEF375
d5YFSO+at/ociAYAuL+w/g==
  "],
 CellTags->{
  "ID:1780855977", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAux induction"},
 CellID->1780855977],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
      RowBox[{"[", "C", "]"}]}]], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{"i", ",", 
        RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
      RowBox[{"j", "\[GreaterEqual]", "i"}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"\[And]", 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{
            UnderscriptBox["GBAuxBase", "\[ScriptCapitalG]"], "[", 
            RowBox[{"Q", ",", "C", ",", "i", ",", "j"}], "]"}]},
          {
           RowBox[{
            UnderscriptBox["GBAuxRec1", "\[ScriptCapitalG]"], "[", 
            RowBox[{"Q", ",", "C", ",", "i", ",", "j"}], "]"}]},
          {
           RowBox[{
            UnderscriptBox["GBAuxRec2", "\[ScriptCapitalG]"], "[", 
            RowBox[{"Q", ",", "C", ",", "i", ",", "j"}], "]"}]},
          {
           RowBox[{
            UnderscriptBox["GBAuxRec3", "\[ScriptCapitalG]"], "[", 
            RowBox[{"Q", ",", "C", ",", "i", ",", "j"}], "]"}]},
          {
           RowBox[{
            UnderscriptBox["GBAuxRec4", "\[ScriptCapitalG]"], "[", 
            RowBox[{"Q", ",", "C", ",", "i", ",", "j"}], "]"}]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
      RowBox[{"[", "C", "]"}]}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", 
        RowBox[{"DomainTuples", "[", 
         RowBox[{"NatPairs", "[", 
          RowBox[{
           RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], 
           ",", 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            TagBox["\[LessEqual]",
             Identity], 
            TagBox[")",
             "AutoParentheses"]}]}], "]"}], "]"}]], 
       RowBox[{"[", "P", "]"}]}]], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{"j", "\[GreaterEqual]", "i"}], "\[Implies]", 
         RowBox[{
          UnderscriptBox["\[ForAll]", 
           RowBox[{
            UnderscriptBox["\[Element]", 
             RowBox[{"DomainTuples", "[", 
              RowBox[{"MMPairs", "[", 
               RowBox[{"\[ScriptCapitalG]", ",", 
                SubscriptBox["C", "i"], ",", 
                SubscriptBox["C", "j"]}], "]"}], "]"}]], 
            RowBox[{"[", "M", "]"}]}]], 
          RowBox[{"Q", "[", 
           RowBox[{"C", ",", "P", ",", "M", ",", "i", ",", "j"}], "]"}]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAux induction II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:380025998", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648538093864394*^9, 3.648538119965755*^9}},
 CellTags->{
  "ID:380025998", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAux induction II"},
 CellID->380025998],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2126131274]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1859434148],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["Q", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 StyleBox["Q", "ExpressionVariable"], "[", 
                 RowBox[{
                   StyleBox["C", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
                   ",", 
                   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
                 "]"}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}]], 
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    
                    RowBox[{
                    "\[DoubleStruckCapitalZ]Dom", ",", 
                    "\[DoubleStruckCapitalZ]Dom"}], "]"}], "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "]"}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]}], "]"}], "]"}]], "[", 
                    StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["T", "ExpressionVariable"]}], "]"}]}], 
                   "\[Implies]", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}],
                     "]"}]}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    
                    RowBox[{"\[DoubleStruckCapitalZ]Dom", ",", 
                    "\[DoubleStruckCapitalZ]Dom"}], "]"}], "]"}]], "[", 
                    StyleBox["P", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}]], 
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]}], "]"}], "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "]"}]], 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["h", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    
                    RowBox[{"\[DoubleStruckCapitalZ]Dom", ",", 
                    "\[DoubleStruckCapitalZ]Dom"}], "]"}], "]"}]], "[", 
                    StyleBox["T", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]}], "]"}], "]"}]], "[", 
                    StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], 
                    UnderscriptBox["\[SquareSubset]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], "\:293a", 
                    StyleBox["h", "ExpressionVariable"]}]}]], 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], "\:293a", 
                    StyleBox["h", "ExpressionVariable"]}], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["S", "ExpressionVariable"]}], "]"}]}]}], 
                    "\[Implies]", 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "]"}]}]}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GBTermOrder", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["Q", "ExpressionVariable"], "[", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["x", "ExpressionVariable"], "1"], ",", 
                 SubscriptBox[
                  StyleBox["x", "ExpressionVariable"], "2"], ",", 
                 SubscriptBox[
                  StyleBox["x", "ExpressionVariable"], "3"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["Q", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 StyleBox["Q", "ExpressionVariable"], "[", 
                 RowBox[{
                   StyleBox["C", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
                   ",", 
                   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
                 "]"}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}]], 
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    
                    RowBox[{
                    "\[DoubleStruckCapitalZ]Dom", ",", 
                    "\[DoubleStruckCapitalZ]Dom"}], "]"}], "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "]"}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]}], "]"}], "]"}]], "[", 
                    StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["T", "ExpressionVariable"]}], "]"}]}], 
                   "\[Implies]", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}],
                     "]"}]}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    
                    RowBox[{"\[DoubleStruckCapitalZ]Dom", ",", 
                    "\[DoubleStruckCapitalZ]Dom"}], "]"}], "]"}]], "[", 
                    StyleBox["P", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}]], 
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]}], "]"}], "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "]"}]], 
                   RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["h", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    
                    RowBox[{"\[DoubleStruckCapitalZ]Dom", ",", 
                    "\[DoubleStruckCapitalZ]Dom"}], "]"}], "]"}]], "[", 
                    StyleBox["T", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]}], "]"}], "]"}]], "[", 
                    StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], 
                    UnderscriptBox["\[SquareSubset]", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], "\:293a", 
                    StyleBox["h", "ExpressionVariable"]}]}]], 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], "\:293a", 
                    StyleBox["h", "ExpressionVariable"]}], ",", 
                    StyleBox["T", "ExpressionVariable"], ",", 
                    StyleBox["S", "ExpressionVariable"]}], "]"}]}]}], 
                    "\[Implies]", 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"t", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "]"}]}]}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GBTermOrder", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["Q", "ExpressionVariable"], "[", 
               RowBox[{
                 SubscriptBox[
                  StyleBox["x", "ExpressionVariable"], "1"], ",", 
                 SubscriptBox[
                  StyleBox["x", "ExpressionVariable"], "2"], ",", 
                 SubscriptBox[
                  StyleBox["x", "ExpressionVariable"], "3"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:994411308"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p994411308-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p994411308-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:994411308", "Proof|ID:994411308-1"},
 CellID->87121011],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p994411308-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1110560]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["Q", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["i", "ExpressionVariable"]}]], 
                 RowBox[{
                   StyleBox["Q", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
                    ",", 
                    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
                    ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["i", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    RowBox[{
                    StyleBox[
                    RowBox[{"p", "..."}], "ExpressionVariable"]}]], 
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                    "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"p", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}]], 
                   RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    StyleBox["l", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["k", "ExpressionVariable"]}]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i0", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j0", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    StyleBox["j0", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["i0", "ExpressionVariable"]}]], 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i0", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["j0", "ExpressionVariable"]]}], "]"}], "]"}]], 
                    "[", 
                    StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"p", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["M", "ExpressionVariable"], ",", 
                    StyleBox["i0", "ExpressionVariable"], ",", 
                    StyleBox["j0", "ExpressionVariable"]}], "]"}]}]}], 
                    "\[Implies]", 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"p", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
                    ",", 
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], "]"}]}]}]}]}]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                    "]"}]], "[", 
                    StyleBox["P", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["i", "ExpressionVariable"]}]], 
                   RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"]}]], 
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    RowBox[{
                    StyleBox[
                    RowBox[{"m", "..."}], "ExpressionVariable"]}]], 
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
                    "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"m", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}]], 
                    RowBox[{
                    UnderscriptBox["let", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["h", "ExpressionVariable"], "=", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["C", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["C", "ExpressionVariable"]}], "]"}]}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["GBAuxTest", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"], ",", 
                    StyleBox["h", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"m", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], 
                    "\[Implies]", 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"m", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[And]", 
                    
                    RowBox[{
                    RowBox[{
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    UnderscriptBox["GBAuxTest", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"], ",", 
                    StyleBox["h", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}], ",", 
                    "\[LessEqual]"}], "]"}], "]"}]], "[", 
                    StyleBox["P0", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], "\:293a", 
                    StyleBox["h", "ExpressionVariable"]}], ",", 
                    StyleBox["P0", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"m", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}], 
                    "\[Implies]", 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"m", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GBTermOrder", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                RowBox[{
                  StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                  StyleBox["i", "ExpressionVariable"]}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                    "]"}]], "[", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], "]"}], 
                   "\[And]", 
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
                    "[", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "3"], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["Q", "ExpressionVariable"], "[", 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], ",", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], ",", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "3"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["Q", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["i", "ExpressionVariable"]}]], 
                 RowBox[{
                   StyleBox["Q", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
                    ",", 
                    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
                    ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["i", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    RowBox[{
                    StyleBox[
                    RowBox[{"p", "..."}], "ExpressionVariable"]}]], 
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                    "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"p", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}]], 
                   RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    StyleBox["l", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["k", "ExpressionVariable"]}]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i0", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j0", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    StyleBox["j0", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["i0", "ExpressionVariable"]}]], 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i0", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["j0", "ExpressionVariable"]]}], "]"}], "]"}]], 
                    "[", 
                    StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"p", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["M", "ExpressionVariable"], ",", 
                    StyleBox["i0", "ExpressionVariable"], ",", 
                    StyleBox["j0", "ExpressionVariable"]}], "]"}]}]}], 
                    "\[Implies]", 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"p", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
                    ",", 
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], "]"}]}]}]}]}]}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                    "]"}]], "[", 
                    StyleBox["P", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["i", "ExpressionVariable"]}]], 
                   RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"]}]], 
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    RowBox[{
                    StyleBox[
                    RowBox[{"m", "..."}], "ExpressionVariable"]}]], 
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
                    "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"m", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}]], 
                    RowBox[{
                    UnderscriptBox["let", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["h", "ExpressionVariable"], "=", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["C", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["C", "ExpressionVariable"]}], "]"}]}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["GBAuxTest", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"], ",", 
                    StyleBox["h", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"m", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], 
                    "\[Implies]", 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"m", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[And]", 
                    
                    RowBox[{
                    RowBox[{
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    UnderscriptBox["GBAuxTest", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"], ",", 
                    StyleBox["h", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}], ",", 
                    "\[LessEqual]"}], "]"}], "]"}]], "[", 
                    StyleBox["P0", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], "\:293a", 
                    StyleBox["h", "ExpressionVariable"]}], ",", 
                    StyleBox["P0", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"m", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}], 
                    "\[Implies]", 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"m", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GBTermOrder", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                RowBox[{
                  StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                  StyleBox["i", "ExpressionVariable"]}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                    "]"}]], "[", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], "]"}], 
                   "\[And]", 
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
                    "[", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "3"], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["Q", "ExpressionVariable"], "[", 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], ",", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], ",", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "3"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1185976259"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1185976259-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1185976259-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1185976259", "Proof|ID:1185976259-1"},
 CellID->50921615],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1185976259-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellChangeTimes->{3.661667103548617*^9},
 CellID->188807672]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["Q", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                    "]"}]], "[", 
                    StyleBox["P", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["i", "ExpressionVariable"]}]], 
                   RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
                    "[", 
                    StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    StyleBox["M", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}]}]}], 
             ")"}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GBTermOrder", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                RowBox[{
                  StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                  StyleBox["i", "ExpressionVariable"]}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                    "]"}]], "[", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], "]"}], 
                   "\[And]", 
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
                    "[", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "3"], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["Q", "ExpressionVariable"], "[", 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], ",", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], ",", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "3"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["Q", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                    "]"}]], "[", 
                    StyleBox["P", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["i", "ExpressionVariable"]}]], 
                   RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
                    "[", 
                    StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    StyleBox["M", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}]}]}], 
             ")"}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"GBTermOrder", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                RowBox[{
                  StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                  StyleBox["i", "ExpressionVariable"]}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                    "]"}]], "[", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], "]"}], 
                   "\[And]", 
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
                    "[", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "3"], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["Q", "ExpressionVariable"], "[", 
                   RowBox[{
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "1"], ",", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "2"], ",", 
                    SubscriptBox[
                    StyleBox["x", "ExpressionVariable"], "3"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1125858125"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1125858125-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1125858125-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1125858125", "Proof|ID:1125858125-1"},
 CellID->52228001],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1125858125-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->673041172]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAux induction)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAux induction)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["Q", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                   StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                RowBox[{
                  StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                  StyleBox["i", "ExpressionVariable"]}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["GBAuxBase", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["GBAuxRec1", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["GBAuxRec2", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["GBAuxRec3", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["GBAuxRec4", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                   StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                    "]"}]], "[", 
                    StyleBox["P", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["i", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
                    "[", 
                    StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    StyleBox["M", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["Q", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                   StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                RowBox[{
                  StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                  StyleBox["i", "ExpressionVariable"]}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["GBAuxBase", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["GBAuxRec1", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["GBAuxRec2", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["GBAuxRec3", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["GBAuxRec4", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                   StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                    "]"}]], "[", 
                    StyleBox["P", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["i", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
                    "[", 
                    StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    StyleBox["M", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1780855977"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1780855977-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1780855977-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1780855977", "Proof|ID:1780855977-1"},
 CellID->3158281],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1780855977-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->98918190]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAux induction II)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAux induction II)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["Q", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                   StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                RowBox[{
                  StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                  StyleBox["i", "ExpressionVariable"]}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["GBAuxBase", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["GBAuxRec1", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["GBAuxRec2", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["GBAuxRec3", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["GBAuxRec4", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                   StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                    "]"}]], "[", 
                    StyleBox["P", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
                    "[", 
                    StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    StyleBox["M", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["Q", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                   StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox[
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                RowBox[{
                  StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                  StyleBox["i", "ExpressionVariable"]}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["GBAuxBase", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["GBAuxRec1", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["GBAuxRec2", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["GBAuxRec3", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["GBAuxRec4", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["Q", "ExpressionVariable"], ",", 
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                   StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                    "]"}]], "[", 
                    StyleBox["P", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                 RowBox[{
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                   RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
                    "[", 
                    StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    StyleBox["M", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:380025998"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p380025998-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p380025998-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:380025998", "Proof|ID:380025998-1"},
 CellID->1967872054],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p380025998-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1105788270]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (1)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.648443894752098*^9, 3.648443897373335*^9}, 
   3.648444825761845*^9, 3.648446229006565*^9, 3.6532840524862657`*^9, 
   3.666173039238839*^9, 3.66643988505309*^9},
 CellID->1976979955],

Cell[TextData[{
 "We prove our goal by Noetherian induction, hence we choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
    RowBox[{"[", "C", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "T", "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "S", "]"}], TraditionalForm]]],
 " a.b.f., assume\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
      RowBox[{"[", "C0", "]"}]}]], 
    RowBox[{"Q", "[", 
     RowBox[{"C0", ",", 
      RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", 
      RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], "]"}]}], 
   TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[Element]", 
        RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
       RowBox[{"[", "C0", "]"}]}], ",", "t", ",", 
      RowBox[{"t", "..."}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{"isTuple", "[", "S0", "]"}]], 
        RowBox[{"Q", "[", 
         RowBox[{"C0", ",", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"t", "..."}], "\[RightAngleBracket]"}], ",", "S0"}], 
         "]"}]}], ")"}], "\[Implies]", 
      RowBox[{"Q", "[", 
       RowBox[{"C0", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"t", ",", 
          RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], ",", 
        RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], "]"}]}], 
     ")"}]}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["\[Element]", 
         RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
        RowBox[{"[", "C0", "]"}]}], ",", 
       RowBox[{"isTuple", "[", "P", "]"}], ",", "t", ",", 
       RowBox[{"t", "..."}]}]], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"Q", "[", 
        RowBox[{"C0", ",", "P", ",", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"t", "..."}], "\[RightAngleBracket]"}]}], "]"}], "\[And]", 
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
           RowBox[{"[", "h", "]"}]}], ",", 
          RowBox[{"isTuple", "[", 
           RowBox[{"T0", ",", "S0"}], "]"}]}]], 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"C0", "\[SquareSubset]", 
           RowBox[{"C0", 
            TagBox["\:293a",
             Identity,
             SyntaxForm->"a*b"], "h"}]}], "\[Implies]", 
          RowBox[{"Q", "[", 
           RowBox[{
            RowBox[{"C0", 
             TagBox["\:293a",
              Identity,
              SyntaxForm->"a*b"], "h"}], ",", "T0", ",", "S0"}], "]"}]}], 
         ")"}]}]}], ")"}]}], "\[Implies]", 
    RowBox[{"Q", "[", 
     RowBox[{"C0", ",", "P", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"t", ",", 
        RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}]}], "]"}]}], 
   TraditionalForm]]],
 "\n(IH)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["\[Element]", 
         RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
        RowBox[{"[", "C0", "]"}]}], ",", 
       RowBox[{"isTuple", "[", 
        RowBox[{"T0", ",", "S0"}], "]"}]}]], 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"C", ",", "T", ",", "S"}], "\[RightAngleBracket]"}], 
      "\[Succeeds]", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"C0", ",", "T0", ",", "S0"}], "\[RightAngleBracket]"}]}]}], 
    "\[Implies]", 
    RowBox[{"Q", "[", 
     RowBox[{"C0", ",", "T0", ",", "S0"}], "]"}]}], TraditionalForm]]],
 "\nand show\n(G.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"Q", "[", 
    RowBox[{"C", ",", "T", ",", "S"}], "]"}], TraditionalForm]]],
 "\n\nWe know that ",
 Cell[BoxData[
  FormBox["S", TraditionalForm]]],
 " is either \[LeftAngleBracket]\[RightAngleBracket] or of the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"t", ",", 
     RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], TraditionalForm]]],
 ", hence we distinguish two cases.\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"S", "=", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 ".\nWe do the same for ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 ".\n\nCase I.I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"T", "=", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 ".\nOur goal follows readily from (1).\n\nCase I.II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"T", "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"t", ",", 
      RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}]}], TraditionalForm]]],
 " for some ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", ",", 
    RowBox[{"t", "..."}]}], TraditionalForm]]],
 "\nWe instantiate (2) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C0", ":=", "C"}], ",", 
    RowBox[{"t", ":=", "t"}], ",", 
    RowBox[{
     RowBox[{"t", "..."}], ":=", 
     RowBox[{"t", "..."}]}]}], TraditionalForm]]],
 " and infer\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{"isTuple", "[", "S0", "]"}]], 
      RowBox[{"Q", "[", 
       RowBox[{"C", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"t", "..."}], "\[RightAngleBracket]"}], ",", "S0"}], "]"}]}],
      ")"}], "\[Implies]", 
    RowBox[{"Q", "[", 
     RowBox[{"C", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"t", ",", 
        RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], "]"}]}], 
   TraditionalForm]]],
 "\nHence, our current goal can be reduced to\n(G.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"Q", "[", 
    RowBox[{"C", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", "..."}], "\[RightAngleBracket]"}], ",", "S0"}], "]"}], 
   TraditionalForm]]],
 "\nfor some a.b.f. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "S0", "]"}], TraditionalForm]]],
 ".\nWe instantiate (IH) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C0", ":=", "C"}], ",", 
    RowBox[{"T0", ":=", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", "..."}], "\[RightAngleBracket]"}]}], ",", 
    RowBox[{"S0", ":=", "S0"}]}], TraditionalForm]]],
 " and obtain\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"C", ",", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"t", ",", 
         RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], ",", 
       RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
      "\[RightAngleBracket]"}], "\[Succeeds]", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"C", ",", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"t", "..."}], "\[RightAngleBracket]"}], ",", "S0"}], 
      "\[RightAngleBracket]"}]}], "\[Implies]", 
    RowBox[{"Q", "[", 
     RowBox[{"C", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"t", "..."}], "\[RightAngleBracket]"}], ",", "S0"}], "]"}]}], 
   TraditionalForm]]],
 "\nmeaning that our goal can be further reduced to\n(G.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"C", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"t", ",", 
        RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], "\[Succeeds]", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"C", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"t", "..."}], "\[RightAngleBracket]"}], ",", "S0"}], 
     "\[RightAngleBracket]"}]}], TraditionalForm]]],
 "\nThis, however, is already contained in (\[ScriptCapitalL] \[Succeeds] 1), \
hence we are done.\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"S", "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"t", ",", 
      RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}]}], TraditionalForm]]],
 " for some ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", ",", 
    RowBox[{"t", "..."}]}], TraditionalForm]]],
 "\nWe instantiate (3) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C0", ":=", "C"}], ",", 
    RowBox[{"P", ":=", "T"}], ",", 
    RowBox[{"t", ":=", "t"}], ",", 
    RowBox[{
     RowBox[{"t", "..."}], ":=", 
     RowBox[{"t", "..."}]}]}], TraditionalForm]]],
 " and obtain\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"Q", "[", 
       RowBox[{"C", ",", "T", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"t", "..."}], "\[RightAngleBracket]"}]}], "]"}], "\[And]", 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         RowBox[{
          UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
          RowBox[{"[", "h", "]"}]}], ",", 
         RowBox[{"isTuple", "[", 
          RowBox[{"T0", ",", "S0"}], "]"}]}]], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"C", "\[SquareSubset]", 
          RowBox[{"C", 
           TagBox["\:293a",
            Identity,
            SyntaxForm->"a*b"], "h"}]}], "\[Implies]", 
         RowBox[{"Q", "[", 
          RowBox[{
           RowBox[{"C", 
            TagBox["\:293a",
             Identity,
             SyntaxForm->"a*b"], "h"}], ",", "T0", ",", "S0"}], "]"}]}], 
        ")"}]}]}], ")"}], "\[Implies]", 
    RowBox[{"Q", "[", 
     RowBox[{"C", ",", "T", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"t", ",", 
        RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}]}], "]"}]}], 
   TraditionalForm]]],
 "\nmeaning that our goal can be reduced to\n(G.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"Q", "[", 
    RowBox[{"C", ",", "T", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", "..."}], "\[RightAngleBracket]"}]}], "]"}], 
   TraditionalForm]]],
 "\n(G.5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
       RowBox[{"[", "h", "]"}]}], ",", 
      RowBox[{"isTuple", "[", 
       RowBox[{"T0", ",", "S0"}], "]"}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"C", "\[SquareSubset]", 
       RowBox[{"C", 
        TagBox["\:293a",
         Identity,
         SyntaxForm->"a*b"], "h"}]}], "\[Implies]", 
      RowBox[{"Q", "[", 
       RowBox[{
        RowBox[{"C", 
         TagBox["\:293a",
          Identity,
          SyntaxForm->"a*b"], "h"}], ",", "T0", ",", "S0"}], "]"}]}], ")"}]}],
    TraditionalForm]]],
 "\n\n(G.4):\nWe instantiate (IH) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C0", ":=", "C"}], ",", 
    RowBox[{"T0", ":=", "T"}], ",", 
    RowBox[{"S0", ":=", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", "..."}], "\[RightAngleBracket]"}]}]}], TraditionalForm]]],
 " and therefore reduce our goal to\n(G.6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"C", ",", "T", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"t", ",", 
        RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], "\[Succeeds]", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"C", ",", "T", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"t", "..."}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}]}], TraditionalForm]]],
 "\nThis, however, is already contained in (\[ScriptCapitalL] \[Succeeds] 2), \
hence we are done.\n\n(G.5):\nWe choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
    RowBox[{"[", "h", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "T0", "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "S0", "]"}], TraditionalForm]]],
 " a.b.f., assume\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"C", "\[SquareSubset]", 
    RowBox[{"C", 
     TagBox["\:293a",
      Identity,
      SyntaxForm->"a*b"], "h"}]}], TraditionalForm]]],
 "\nand show\n(G.7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"Q", "[", 
    RowBox[{
     RowBox[{"C", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "h"}], ",", "T0", ",", "S0"}], "]"}], 
   TraditionalForm]]],
 "\n\nWe instantiate (IH) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", ":=", 
     RowBox[{"C", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "h"}]}], ",", 
    RowBox[{"T0", ":=", "T0"}], ",", 
    RowBox[{"S0", ":=", "S0"}]}], TraditionalForm]]],
 " and therefore reduce our goal to\n(G.8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"C", ",", "T", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"t", ",", 
        RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], "\[Succeeds]", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"C", 
       TagBox["\:293a",
        Identity,
        SyntaxForm->"a*b"], "h"}], ",", "T0", ",", "S0"}], 
     "\[RightAngleBracket]"}]}], TraditionalForm]]],
 "\nThis, however, is already contained in (\[ScriptCapitalL] \[Succeeds] 3), \
hence we are done."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.648444846072212*^9, 3.6484454393976183`*^9}, {
   3.648445491751487*^9, 3.648445743875372*^9}, {3.64844580993049*^9, 
   3.648446017786531*^9}, {3.648446054434463*^9, 3.648446229006695*^9}, 
   3.6532840524862657`*^9, 3.666173039238839*^9, 3.666439885053197*^9},
 CellID->1320712767],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellChangeTimes->{{3.648446216083562*^9, 3.648446229006835*^9}, 
   3.6532840524862657`*^9, 3.666173039238839*^9, 3.666439885053299*^9},
 TextAlignment->Right,
 CellID->1296315400]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (2)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.648450687815901*^9, 3.648450691263989*^9}, 
   3.648455108460121*^9, 3.6532840455442533`*^9, 3.666173031763426*^9, 
   3.666439878103608*^9},
 CellID->1173610870],

Cell[TextData[{
 "We assume\n(0)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
      RowBox[{"[", "C", "]"}]}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
     RowBox[{"Q", "[", 
      RowBox[{"C", ",", 
       RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", 
       RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "i", ",",
        "j"}], "]"}]}]}], TraditionalForm]]],
 "\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", 
        RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
       RowBox[{"[", "C", "]"}]}]], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", 
         RowBox[{"NatPairs", "[", 
          RowBox[{
           RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], 
           ",", "\[LessEqual]"}], "]"}]], 
        RowBox[{"[", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"i", ",", "j"}], "\[RightAngleBracket]"}], "]"}]}]], 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         UnderscriptBox["\[Element]", 
          RowBox[{"DomainTuples", "[", 
           RowBox[{"NatPairs", "[", 
            RowBox[{
             RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], 
             ",", "\[LessEqual]"}], "]"}], "]"}]], 
         RowBox[{"[", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"p", "..."}], "\[RightAngleBracket]"}], "]"}]}]], 
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{"k", ",", 
          RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
          RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
        RowBox[{"(", 
         RowBox[{
          UnderscriptBox["\[ForAll]", 
           RowBox[{
            UnderscriptBox["\[Element]", 
             RowBox[{"DomainTuples", "[", 
              RowBox[{"MMPairs", "[", 
               RowBox[{"\[ScriptCapitalG]", ",", 
                SubscriptBox["C", "i"], ",", 
                SubscriptBox["C", "j"]}], "]"}], "]"}]], 
            RowBox[{"[", "M", "]"}]}]], 
          RowBox[{"Q", "[", 
           RowBox[{"C", ",", 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"p", "..."}], "\[RightAngleBracket]"}], ",", "M", ",", 
            "i", ",", "j"}], "]"}]}], ")"}]}]}]}]}], "\[Implies]", 
    RowBox[{"Q", "[", 
     RowBox[{"C", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"i", ",", "j"}], "\[RightAngleBracket]"}], ",", 
        RowBox[{"p", "..."}]}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "k", ",", 
      "l"}], "]"}]}], TraditionalForm]]],
 "\n(2)\n\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", 
        RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
       RowBox[{"[", "C0", "]"}]}]], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", 
         RowBox[{"DomainTuples", "[", 
          RowBox[{"NatPairs", "[", 
           RowBox[{
            RowBox[{"\[LeftBracketingBar]", "C0", "\[RightBracketingBar]"}], 
            ",", "\[LessEqual]"}], "]"}], "]"}]], 
        RowBox[{"[", "P", "]"}]}]], 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{"i", ",", 
         RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
       RowBox[{
        UnderscriptBox[
         UnderscriptBox["\[ForAll]", 
          RowBox[{"a", ",", "mi", ",", "mj"}]], 
         RowBox[{
          UnderscriptBox["\[Element]", 
           RowBox[{"MMPairs", "[", 
            RowBox[{"\[ScriptCapitalG]", ",", 
             SubscriptBox["C0", "i"], ",", 
             SubscriptBox["C0", "j"]}], "]"}]], 
          RowBox[{"[", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"a", ",", 
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
            "\[RightAngleBracket]"}], "]"}]}]], 
        RowBox[{
         UnderscriptBox[
          UnderscriptBox["\[ForAll]", 
           RowBox[{"m", "..."}]], 
          RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", 
             RowBox[{"MMPairs", "[", 
              RowBox[{"\[ScriptCapitalG]", ",", 
               SubscriptBox["C0", "i"], ",", 
               SubscriptBox["C0", "j"]}], "]"}], "]"}]], 
           RowBox[{"[", 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"m", "..."}], "\[RightAngleBracket]"}], "]"}]}]], 
         UnderscriptBox["let", 
          RowBox[{"h", "=", 
           RowBox[{
            RowBox[{"trd", "[", "]"}], "-", 
            RowBox[{"trd", "[", "]"}]}]}]]}]}]}]}]}], "\[And]", 
    RowBox[{"{", GridBox[{
       {
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"GBAuxTest", "[", 
            RowBox[{
            "C0", ",", "a", ",", "mi", ",", "mj", ",", "h", ",", "P", ",", 
             "i", ",", "j"}], "]"}], "\[And]", 
           RowBox[{"Q", "[", 
            RowBox[{"C0", ",", "P", ",", 
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"m", "..."}], "\[RightAngleBracket]"}], ",", "i", ",", 
             "j"}], "]"}]}], ")"}], "\[Implies]", 
         RowBox[{"Q", "[", 
          RowBox[{"C0", ",", "P", ",", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"a", ",", 
               RowBox[{"\[LeftAngleBracket]", 
                RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
              "\[RightAngleBracket]"}], ",", 
             RowBox[{"m", "..."}]}], "\[RightAngleBracket]"}], ",", "i", ",", 
           "j"}], "]"}]}]},
       {
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"\[Not]", 
            RowBox[{"GBAuxTest", "[", 
             RowBox[{
             "C0", ",", "a", ",", "mi", ",", "mj", ",", "h", ",", "P", ",", 
              "i", ",", "j"}], "]"}]}], "\[And]", 
           RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
              UnderscriptBox["\[Element]", 
               RowBox[{"DomainTuples", "[", 
                RowBox[{"NatPairs", "[", 
                 RowBox[{
                  RowBox[{
                   RowBox[{
                   "\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], 
                   "+", "1"}], ",", "\[LessEqual]"}], "]"}], "]"}]], 
              RowBox[{"[", "P0", "]"}]}]], 
            RowBox[{"Q", "[", 
             RowBox[{
              RowBox[{"C0", 
               TagBox["\:293a",
                Identity,
                SyntaxForm->"a*b"], "h"}], ",", "P0", ",", 
              RowBox[{"\[LeftAngleBracket]", 
               RowBox[{"m", "..."}], "\[RightAngleBracket]"}], ",", "i", ",", 
              "j"}], "]"}]}]}], ")"}], "\[Implies]", 
         RowBox[{"Q", "[", 
          RowBox[{"C0", ",", "P", ",", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"a", ",", 
               RowBox[{"\[LeftAngleBracket]", 
                RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
              "\[RightAngleBracket]"}], ",", 
             RowBox[{"m", "..."}]}], "\[RightAngleBracket]"}], ",", "i", ",", 
           "j"}], "]"}]}]}
      }]}]}], TraditionalForm]]],
 "\nand reduce our universally quantified goal according to formula (1), i.e. \
we need to show the three subgoals\n(G.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
      RowBox[{"[", "C", "]"}]}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
     RowBox[{"Q", "[", 
      RowBox[{"C", ",", 
       RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", 
       RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "i", ",",
        "j"}], "]"}]}]}], TraditionalForm]]],
 "\n(G.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[Element]", 
        RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
       RowBox[{"[", "C", "]"}]}], ",", "t", ",", 
      RowBox[{"t", "..."}]}]], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{"isTuple", "[", "T", "]"}]], 
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{"i", ",", 
          RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
          RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{
            UnderscriptBox["\[Element]", 
             RowBox[{"DomainTuples", "[", 
              RowBox[{"NatPairs", "[", 
               RowBox[{
                RowBox[{
                "\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], ",", 
                "\[LessEqual]"}], "]"}], "]"}]], 
            RowBox[{"[", 
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"t", "..."}], "\[RightAngleBracket]"}], "]"}]}], 
           "\[And]", 
           RowBox[{
            UnderscriptBox["\[Element]", 
             RowBox[{"DomainTuples", "[", 
              RowBox[{"MMPairs", "[", 
               RowBox[{"\[ScriptCapitalG]", ",", 
                SubscriptBox["C", "i"], ",", 
                SubscriptBox["C", "j"]}], "]"}], "]"}]], 
            RowBox[{"[", "T", "]"}]}]}], "\[Implies]", 
          RowBox[{"Q", "[", 
           RowBox[{"C", ",", 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"t", "..."}], "\[RightAngleBracket]"}], ",", "T", ",", 
            "i", ",", "j"}], "]"}]}], ")"}]}]}], "\[Implies]", 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{"i", ",", 
         RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          UnderscriptBox["\[Element]", 
           RowBox[{"DomainTuples", "[", 
            RowBox[{"NatPairs", "[", 
             RowBox[{
              RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], 
              ",", "\[LessEqual]"}], "]"}], "]"}]], 
          RowBox[{"[", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"t", ",", 
             RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], "]"}]}], 
         "\[Implies]", 
         RowBox[{"Q", "[", 
          RowBox[{"C", ",", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"t", ",", 
             RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "i", 
           ",", "j"}], "]"}]}], ")"}]}]}], ")"}]}], TraditionalForm]]],
 "\n(G.3)\t...\n\n(G.1) is precisely the content of (0).\n\n(G.2):\nWe choose \
",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
    RowBox[{"[", "C", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", ",", 
    RowBox[{"t", "..."}]}], TraditionalForm]]],
 " a.b.f., assume\n(2.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], ",", 
        "\[LessEqual]"}], "]"}], "]"}]], 
    RowBox[{"[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", ",", 
       RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], "]"}]}], 
   TraditionalForm]]],
 "\n(2.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{"isTuple", "[", "T", "]"}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{
         UnderscriptBox["\[Element]", 
          RowBox[{"DomainTuples", "[", 
           RowBox[{"NatPairs", "[", 
            RowBox[{
             RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], 
             ",", "\[LessEqual]"}], "]"}], "]"}]], 
         RowBox[{"[", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"t", "..."}], "\[RightAngleBracket]"}], "]"}]}], "\[And]", 
        
        RowBox[{
         UnderscriptBox["\[Element]", 
          RowBox[{"DomainTuples", "[", 
           RowBox[{"MMPairs", "[", 
            RowBox[{"\[ScriptCapitalG]", ",", 
             SubscriptBox["C", "i"], ",", 
             SubscriptBox["C", "j"]}], "]"}], "]"}]], 
         RowBox[{"[", "T", "]"}]}]}], "\[Implies]", 
       RowBox[{"Q", "[", 
        RowBox[{"C", ",", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"t", "..."}], "\[RightAngleBracket]"}], ",", "T", ",", "i", 
         ",", "j"}], "]"}]}], ")"}]}]}], TraditionalForm]]],
 "\nand prove\n(G.2.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"Q", "[", 
    RowBox[{"C", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", ",", 
       RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "i", ",", 
     "j"}], "]"}], TraditionalForm]]],
 "\n\nFrom (2.1) we can infer\n(2.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"NatPairs", "[", 
      RowBox[{
       RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], ",", 
       "\[LessEqual]"}], "]"}]], 
    RowBox[{"[", "t", "]"}]}], TraditionalForm]]],
 "\n(2.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], ",", 
        "\[LessEqual]"}], "]"}], "]"}]], 
    RowBox[{"[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", "..."}], "\[RightAngleBracket]"}], "]"}]}], 
   TraditionalForm]]],
 "\nand further the existence of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", ",", 
    RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 " with\n(2.5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"l", "\[GreaterEqual]", "k"}], TraditionalForm]]],
 "\n(2.6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"k", ",", "l"}], "\[RightAngleBracket]"}]}], TraditionalForm]]],
 "\nWe can instantiate (1) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", ":=", "C"}], ",", 
    RowBox[{"i", "=", "k"}], ",", 
    RowBox[{"j", "=", "l"}], ",", 
    RowBox[{
     RowBox[{"p", "..."}], "=", 
     RowBox[{"t", "..."}]}], ",", 
    RowBox[{"k", "=", "i"}], ",", 
    RowBox[{"l", "=", "j"}]}], TraditionalForm]]],
 " and obtain\n(2.7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", 
         RowBox[{"DomainTuples", "[", 
          RowBox[{"MMPairs", "[", 
           RowBox[{"\[ScriptCapitalG]", ",", 
            SubscriptBox["C", "k"], ",", 
            SubscriptBox["C", "l"]}], "]"}], "]"}]], 
        RowBox[{"[", "M", "]"}]}]], 
      RowBox[{"Q", "[", 
       RowBox[{"C", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"t", "..."}], "\[RightAngleBracket]"}], ",", "M", ",", "k", 
        ",", "l"}], "]"}]}], ")"}], "\[Implies]", 
    RowBox[{"Q", "[", 
     RowBox[{"C", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"k", ",", "l"}], "\[RightAngleBracket]"}], ",", 
        RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "i", ",", 
      "j"}], "]"}]}], TraditionalForm]]],
 "\nmeaning that we can reduce our goal to\n(G.2.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"Q", "[", 
    RowBox[{"C", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", "..."}], "\[RightAngleBracket]"}], ",", "M", ",", "k", ",",
      "l"}], "]"}], TraditionalForm]]],
 "\nfor some a.b.f. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"MMPairs", "[", 
       RowBox[{"\[ScriptCapitalG]", ",", 
        SubscriptBox["C", "k"], ",", 
        SubscriptBox["C", "l"]}], "]"}], "]"}]], 
    RowBox[{"[", "M", "]"}]}], TraditionalForm]]],
 ".\nWe now use (2.2), instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"T", ":=", "M"}], ",", 
    RowBox[{"i", ":=", "k"}], ",", 
    RowBox[{"j", ":=", "l"}]}], TraditionalForm]]],
 " and therefore reduce our goal to\n(G.2.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], ",", 
        "\[LessEqual]"}], "]"}], "]"}]], 
    RowBox[{"[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", "..."}], "\[RightAngleBracket]"}], "]"}]}], 
   TraditionalForm]]],
 "\n(G.2.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"MMPairs", "[", 
       RowBox[{"\[ScriptCapitalG]", ",", 
        SubscriptBox["C", "k"], ",", 
        SubscriptBox["C", "l"]}], "]"}], "]"}]], 
    RowBox[{"[", "M", "]"}]}], TraditionalForm]]],
 "\nBoth subgoals are already known, hence we are done.\n\n\n(G.3):\nWe have \
to show\n(G.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        UnderscriptBox["\[Element]", 
         RowBox[{"DomainTuples", "[", 
          RowBox[{"NatPairs", "[", 
           RowBox[{
            RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], 
            ",", "\[LessEqual]"}], "]"}], "]"}]], 
        RowBox[{"[", "P", "]"}]}], "\[And]", 
       RowBox[{
        UnderscriptBox["\[Element]", 
         RowBox[{"DomainTuples", "[", 
          RowBox[{"MMPairs", "[", 
           RowBox[{"\[ScriptCapitalG]", ",", 
            SubscriptBox["C", "i"], ",", 
            SubscriptBox["C", "j"]}], "]"}], "]"}]], 
        RowBox[{"[", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"t", ",", 
           RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], "]"}]}]}], 
      ")"}]}], "\[Implies]", 
    RowBox[{"Q", "[", 
     RowBox[{"C", ",", "P", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"t", ",", 
        RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], ",", "i", ",", 
      "j"}], "]"}]}], TraditionalForm]]],
 "\nunder the assumptions\n(IH.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        UnderscriptBox["\[Element]", 
         RowBox[{"DomainTuples", "[", 
          RowBox[{"NatPairs", "[", 
           RowBox[{
            RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], 
            ",", "\[LessEqual]"}], "]"}], "]"}]], 
        RowBox[{"[", "P", "]"}]}], "\[And]", 
       RowBox[{
        UnderscriptBox["\[Element]", 
         RowBox[{"DomainTuples", "[", 
          RowBox[{"MMPairs", "[", 
           RowBox[{"\[ScriptCapitalG]", ",", 
            SubscriptBox["C", "i"], ",", 
            SubscriptBox["C", "j"]}], "]"}], "]"}]], 
        RowBox[{"[", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"t", "..."}], "\[RightAngleBracket]"}], "]"}]}]}], ")"}]}], 
    "\[Implies]", 
    RowBox[{"Q", "[", 
     RowBox[{"C", ",", "P", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"t", "..."}], "\[RightAngleBracket]"}], ",", "i", ",", "j"}], 
     "]"}]}], TraditionalForm]]],
 "\n(IH.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
       RowBox[{"[", "h", "]"}]}]], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{"isTuple", "[", 
        RowBox[{"T", ",", "S"}], "]"}]], 
      RowBox[{"C", "\[SquareSubset]", 
       RowBox[{"C", 
        TagBox["\:293a",
         Identity,
         SyntaxForm->"a*b"], "h"}]}]}]}], "\[Implies]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{"i", ",", 
         RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", 
          RowBox[{"C", 
           TagBox["\:293a",
            Identity,
            SyntaxForm->"a*b"], "h"}], "\[RightBracketingBar]"}]}]], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          UnderscriptBox["\[Element]", 
           RowBox[{"DomainTuples", "[", 
            RowBox[{"NatPairs", "[", 
             RowBox[{
              RowBox[{"\[LeftBracketingBar]", 
               RowBox[{"C", 
                TagBox["\:293a",
                 Identity,
                 SyntaxForm->"a*b"], "h"}], "\[RightBracketingBar]"}], ",", 
              "\[LessEqual]"}], "]"}], "]"}]], 
          RowBox[{"[", "T", "]"}]}], "\[And]", 
         RowBox[{
          UnderscriptBox["\[Element]", 
           RowBox[{"DomainTuples", "[", 
            RowBox[{"MMPairs", "[", 
             RowBox[{"\[ScriptCapitalG]", ",", 
              SubscriptBox[
               RowBox[{"(", 
                RowBox[{"C", 
                 TagBox["\:293a",
                  Identity,
                  SyntaxForm->"a*b"], "h"}], ")"}], "i"], ",", 
              SubscriptBox[
               RowBox[{"(", 
                RowBox[{"C", 
                 TagBox["\:293a",
                  Identity,
                  SyntaxForm->"a*b"], "h"}], ")"}], "j"]}], "]"}], "]"}]], 
          RowBox[{"[", "S", "]"}]}]}], ")"}]}], "\[Implies]", 
      RowBox[{"Q", "[", 
       RowBox[{
        RowBox[{"C", 
         TagBox["\:293a",
          Identity,
          SyntaxForm->"a*b"], "h"}], ",", "T", ",", "S", ",", "i", ",", "j"}],
        "]"}]}], ")"}]}], TraditionalForm]]],
 "\nfor some a.b.f. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
    RowBox[{"[", "C", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTuple", "[", "P", "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", ",", 
    RowBox[{"t", "..."}]}], TraditionalForm]]],
 "\n\nFor proving (G.1) we choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 " a.b.f., assume\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], ",", 
        "\[LessEqual]"}], "]"}], "]"}]], 
    RowBox[{"[", "P", "]"}]}], TraditionalForm]]],
 "\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"MMPairs", "[", 
       RowBox[{"\[ScriptCapitalG]", ",", 
        SubscriptBox["C", "i"], ",", 
        SubscriptBox["C", "j"]}], "]"}], "]"}]], 
    RowBox[{"[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", ",", 
       RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], "]"}]}], 
   TraditionalForm]]],
 "\nand show\n(G.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"Q", "[", 
    RowBox[{"C", ",", "P", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", ",", 
       RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], ",", "i", ",", "j"}],
     "]"}], TraditionalForm]]],
 "\n\n(4) is equivalent to\n(4.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"MMPairs", "[", 
      RowBox[{"\[ScriptCapitalG]", ",", 
       SubscriptBox["C", "i"], ",", 
       SubscriptBox["C", "j"]}], "]"}]], 
    RowBox[{"[", "t", "]"}]}], TraditionalForm]]],
 "\n(4.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"MMPairs", "[", 
       RowBox[{"\[ScriptCapitalG]", ",", 
        SubscriptBox["C", "i"], ",", 
        SubscriptBox["C", "j"]}], "]"}], "]"}]], 
    RowBox[{"[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", "..."}], "\[RightAngleBracket]"}], "]"}]}], 
   TraditionalForm]]],
 "\nand from (4.1) we can infer the existence of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "mi", ",", "mj"}], "]"}]}], TraditionalForm]]],
 " with\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"t", "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"a", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}]}], TraditionalForm]]],
 "\n\nWe now instantiate assumption (2) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C0", ":=", "C"}], ",", 
    RowBox[{"P", ":=", "P"}], ",", 
    RowBox[{"i", ":=", "i"}], ",", 
    RowBox[{"j", ":=", "j"}], ",", 
    RowBox[{"a", ":=", "a"}], ",", 
    RowBox[{"mi", ":=", "mi"}], ",", 
    RowBox[{"mj", ":=", "mj"}], ",", 
    RowBox[{
     RowBox[{"m", "..."}], ":=", 
     RowBox[{"t", "..."}]}]}], TraditionalForm]]],
 ", and since all conditions are easily seen to hold we obtain\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"GBAuxTest", "[", 
       RowBox[{
       "C", ",", "a", ",", "mi", ",", "mj", ",", "h", ",", "P", ",", "i", ",",
         "j"}], "]"}], "\[And]", 
      RowBox[{"Q", "[", 
       RowBox[{"C", ",", "P", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"t", "..."}], "\[RightAngleBracket]"}], ",", "i", ",", "j"}],
        "]"}]}], ")"}], "\[Implies]", 
    RowBox[{"Q", "[", 
     RowBox[{"C", ",", "P", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"a", ",", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
         "\[RightAngleBracket]"}], ",", 
        RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], ",", "i", ",", 
      "j"}], "]"}]}], TraditionalForm]]],
 "\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"\[Not]", 
       RowBox[{"GBAuxTest", "[", 
        RowBox[{
        "C", ",", "a", ",", "mi", ",", "mj", ",", "h", ",", "P", ",", "i", 
         ",", "j"}], "]"}]}], "\[And]", 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         UnderscriptBox["\[Element]", 
          RowBox[{"DomainTuples", "[", 
           RowBox[{"NatPairs", "[", 
            RowBox[{
             RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], 
             ",", "\[LessEqual]"}], "]"}], "]"}]], 
         RowBox[{"[", "P0", "]"}]}]], 
       RowBox[{"Q", "[", 
        RowBox[{
         RowBox[{"C", 
          TagBox["\:293a",
           Identity,
           SyntaxForm->"a*b"], "h"}], ",", "P0", ",", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"t", "..."}], "\[RightAngleBracket]"}], ",", "i", ",", 
         "j"}], "]"}]}]}], ")"}], "\[Implies]", 
    RowBox[{"Q", "[", 
     RowBox[{"C", ",", "P", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"a", ",", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
         "\[RightAngleBracket]"}], ",", 
        RowBox[{"t", "..."}]}], "\[RightAngleBracket]"}], ",", "i", ",", 
      "j"}], "]"}]}], TraditionalForm]]],
 "\nwhere ",
 Cell[BoxData[
  FormBox[
   RowBox[{"h", "=", 
    RowBox[{
     RowBox[{"trd", "[", 
      RowBox[{
       RowBox[{"a", "-", 
        RowBox[{"mi", "*", 
         SubscriptBox["C", "i"]}]}], ",", "C"}], "]"}], "-", 
     RowBox[{"trd", "[", 
      RowBox[{
       RowBox[{"a", "-", 
        RowBox[{"mj", "*", 
         SubscriptBox["C", "j"]}]}], ",", "C"}], "]"}]}]}], TraditionalForm]]],
 ".\n\nWe now distinguish two cases, based on whether ",
 Cell[BoxData[
  FormBox[
   RowBox[{"GBAuxTest", "[", 
    RowBox[{
    "C", ",", "a", ",", "mi", ",", "mj", ",", "h", ",", "P", ",", "i", ",", 
     "j"}], "]"}], TraditionalForm]]],
 " holds or not.\n\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"GBAuxTest", "[", 
    RowBox[{
    "C", ",", "a", ",", "mi", ",", "mj", ",", "h", ",", "P", ",", "i", ",", 
     "j"}], "]"}], TraditionalForm]]],
 ".\nIn this case, using (6) and (5), we can reduce our goal (G.4) to\n(G.5)\t\
",
 Cell[BoxData[
  FormBox[
   RowBox[{"Q", "[", 
    RowBox[{"C", ",", "P", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", "..."}], "\[RightAngleBracket]"}], ",", "i", ",", "j"}], 
    "]"}], TraditionalForm]]],
 "\n\nBy (IH.1), our goal can be reduced to\n(G.6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], ",", 
        "\[LessEqual]"}], "]"}], "]"}]], 
    RowBox[{"[", "P", "]"}]}], TraditionalForm]]],
 "\n(G.7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"MMPairs", "[", 
       RowBox[{"\[ScriptCapitalG]", ",", 
        SubscriptBox["C", "i"], ",", 
        SubscriptBox["C", "j"]}], "]"}], "]"}]], 
    RowBox[{"[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", "..."}], "\[RightAngleBracket]"}], "]"}]}], 
   TraditionalForm]]],
 "\nBoth subgoals are already known ((3) and (4.2)), hence we are done.\n\n\
Case II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"GBAuxTest", "[", 
     RowBox[{
     "C", ",", "a", ",", "mi", ",", "mj", ",", "h", ",", "P", ",", "i", ",", 
      "j"}], "]"}]}], TraditionalForm]]],
 ".\nIn this case, using (7) and (5), we can reduce our goal to\n(G.8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"Q", "[", 
    RowBox[{
     RowBox[{"C", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "h"}], ",", "P0", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", "..."}], "\[RightAngleBracket]"}], ",", "i", ",", "j"}], 
    "]"}], TraditionalForm]]],
 "\nfor some a.b.f. ",
 Cell[BoxData[
  FormBox["P0", TraditionalForm]]],
 " with\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], "+", 
         "1"}], ",", "\[LessEqual]"}], "]"}], "]"}]], 
    RowBox[{"[", "P0", "]"}]}], TraditionalForm]]],
 "\n\nWe no instantiate (IH.2) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"h", ":=", "h"}], ",", 
    RowBox[{"T", ":=", "P0"}], ",", 
    RowBox[{"S", ":=", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", "..."}], "\[RightAngleBracket]"}]}]}], TraditionalForm]]],
 " and hence obtain\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"C", "\[SquareSubset]", 
      RowBox[{"C", 
       TagBox["\:293a",
        Identity,
        SyntaxForm->"a*b"], "h"}]}], "\[Implies]", 
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", 
        RowBox[{"C", 
         TagBox["\:293a",
          Identity,
          SyntaxForm->"a*b"], "h"}], "\[RightBracketingBar]"}]}]]}], "..."}], 
   TraditionalForm]]],
 "\nWe prove the premise of (9):\nFrom our case assumption we in particular \
know ",
 Cell[BoxData[
  FormBox[
   RowBox[{"h", "=", 
    RowBox[{
     RowBox[{
      RowBox[{"trd", "[", 
       RowBox[{
        RowBox[{"a", "-", 
         RowBox[{"mi", "*", 
          SubscriptBox["C", "i"]}]}], ",", "C"}], "]"}], "-", 
      RowBox[{"trd", "[", 
       RowBox[{
        RowBox[{"a", "-", 
         RowBox[{"mj", "*", 
          SubscriptBox["C", "j"]}]}], ",", "C"}], "]"}]}], "\[NotEqual]", 
     "0"}]}], TraditionalForm]]],
 ", hence the premise follows readily from (\[SquareSubset] append trd). \
Therefore, we can infer the conclusion of the implication, which we \
immediately instantiate by ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"i", ":=", "i"}], ",", 
    RowBox[{"j", ":=", "j"}]}], TraditionalForm]]],
 ". The condition on ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", "j"}], TraditionalForm]]],
 " is easily seen to hold, hence we can reduce our goal to\n(G.9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", 
         RowBox[{"C", 
          TagBox["\:293a",
           Identity,
           SyntaxForm->"a*b"], "h"}], "\[RightBracketingBar]"}], ",", 
        "\[LessEqual]"}], "]"}], "]"}]], 
    RowBox[{"[", "P0", "]"}]}], TraditionalForm]]],
 "\n(G.10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"MMPairs", "[", 
       RowBox[{"\[ScriptCapitalG]", ",", 
        SubscriptBox[
         RowBox[{"(", 
          RowBox[{"C", 
           TagBox["\:293a",
            Identity,
            SyntaxForm->"a*b"], "h"}], ")"}], "i"], ",", 
        SubscriptBox[
         RowBox[{"(", 
          RowBox[{"C", 
           TagBox["\:293a",
            Identity,
            SyntaxForm->"a*b"], "h"}], ")"}], "j"]}], "]"}], "]"}]], 
    RowBox[{"[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"t", "..."}], "\[RightAngleBracket]"}], "]"}]}], 
   TraditionalForm]]],
 "\n\n(G.9) follows immediately from (8), since ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[LeftBracketingBar]", 
     RowBox[{"C", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "h"}], "\[RightBracketingBar]"}], "=", 
    RowBox[{
     RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], "+", 
     "1"}]}], TraditionalForm]]],
 ".\n(G.10) follows immediately from (4.2), using ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox[
     RowBox[{"(", 
      RowBox[{"C", 
       TagBox["\:293a",
        Identity,
        SyntaxForm->"a*b"], "h"}], ")"}], "i"], "=", 
    SubscriptBox["C", "i"]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox[
     RowBox[{"(", 
      RowBox[{"C", 
       TagBox["\:293a",
        Identity,
        SyntaxForm->"a*b"], "h"}], ")"}], "j"], "=", 
    SubscriptBox["C", "j"]}], TraditionalForm]]],
 " due to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.648450705352235*^9, 3.6484509097733603`*^9}, {
   3.648451007418483*^9, 3.648451804310611*^9}, {3.648451853997559*^9, 
   3.648452010021192*^9}, {3.648452061143044*^9, 3.648452061143176*^9}, {
   3.648452107875497*^9, 3.648452378341394*^9}, {3.648452718837966*^9, 
   3.648452787672501*^9}, {3.6484532477577887`*^9, 3.6484532670301437`*^9}, {
   3.648453358385453*^9, 3.648453506759138*^9}, {3.648453544015195*^9, 
   3.648453628808275*^9}, {3.648453683178843*^9, 3.64845398295958*^9}, {
   3.648454021153479*^9, 3.6484540346367903`*^9}, {3.648454075818853*^9, 
   3.648454238872732*^9}, {3.648454299256297*^9, 3.648454338826187*^9}, {
   3.64845452977999*^9, 3.648454550230345*^9}, {3.648454693693725*^9, 
   3.648455044033642*^9}, {3.6484550748314238`*^9, 3.648455108460256*^9}, {
   3.648455145973271*^9, 3.648455205114283*^9}, {3.6484552624540777`*^9, 
   3.648455316068695*^9}, {3.648525089998456*^9, 3.648525396120215*^9}, {
   3.648525470323164*^9, 3.6485255981227903`*^9}, {3.648525631623167*^9, 
   3.648525841224107*^9}, {3.6485260926064*^9, 3.648526173330715*^9}, {
   3.648526330249802*^9, 3.648526480806861*^9}, 3.6532840455442533`*^9, 
   3.666173031763426*^9, 3.666439878103758*^9},
 CellID->2136547636],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6484550954501333`*^9, 3.648455108460411*^9}, 
   3.6532840455442533`*^9, 3.666173031763426*^9, 3.666439878103909*^9},
 TextAlignment->Right,
 CellID->1971546439]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->2063122267],

Cell[CellGroupData[{

Cell["Lemma: Alternative Description of GBAux", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648531796667028*^9, 3.648531816301549*^9}},
 CellID->70806236],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
   "C", ",", "P", ",", "i", ",", "j", ",", "a", ",", "mi", ",", "mj", ",", 
    RowBox[{"m", "..."}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
     RowBox[{"C", ",", "P", ",", "i", ",", "j", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"a", ",", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
         "\[RightAngleBracket]"}], ",", 
        RowBox[{"m", "..."}]}], "\[RightAngleBracket]"}]}], "]"}], "\[Equal]", 
    RowBox[{
     UnderscriptBox["let", 
      RowBox[{"h", "=", 
       RowBox[{
        RowBox[{
         UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
         RowBox[{
          RowBox[{"a", 
           UnderscriptBox["-", "\[ScriptCapitalG]"], 
           RowBox[{"mi", 
            UnderscriptBox["*", "\[ScriptCapitalG]"], 
            SubscriptBox["C", "i"]}]}], ",", "C"}], "]"}], 
        UnderscriptBox["-", "\[ScriptCapitalG]"], 
        RowBox[{
         UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
         RowBox[{
          RowBox[{"a", 
           UnderscriptBox["-", "\[ScriptCapitalG]"], 
           RowBox[{"mj", 
            UnderscriptBox["*", "\[ScriptCapitalG]"], 
            SubscriptBox["C", "j"]}]}], ",", "C"}], "]"}]}]}]], 
     RowBox[{"\[Piecewise]", GridBox[{
        {
         RowBox[{
          UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
          RowBox[{"C", ",", "P", ",", "i", ",", "j", ",", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"m", "..."}], "\[RightAngleBracket]"}]}], "]"}], 
         "\[DoubleLeftArrow]", 
         RowBox[{
          UnderscriptBox["GBAuxTest", "\[ScriptCapitalG]"], "[", 
          RowBox[{
          "C", ",", "a", ",", "mi", ",", "mj", ",", "h", ",", "P", ",", "i", 
           ",", "j"}], "]"}]},
        {
         RowBox[{
          UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
          RowBox[{
           RowBox[{"C", 
            TagBox["\:293a",
             Identity,
             SyntaxForm->"a*b"], "h"}], ",", 
           RowBox[{"update", "[", 
            RowBox[{"P", ",", 
             RowBox[{
              RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}], 
              "+", "1"}]}], "]"}], ",", "i", ",", "j", ",", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"m", "..."}], "\[RightAngleBracket]"}]}], "]"}], 
         "\[DoubleLeftArrow]", "otherwise"}
       }]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAux 3 alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1379909256", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648531848101619*^9, 3.648532038401133*^9}},
 CellTags->{
  "ID:1379909256", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAux 3 alternative", "IR"},
 CellID->1379909256],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1623219371]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2040040187],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAux 3 alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAux 3 alternative)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["C", "ExpressionVariable"]}, {
             StyleBox["P", "ExpressionVariable"]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["mi", "ExpressionVariable"]}, {
             StyleBox["mj", "ExpressionVariable"]}, {
             StyleBox[
              RowBox[{"m", "..."}], "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["GBAux", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["C", "ExpressionVariable"], ",", 
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"m", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["let", 
              GridBox[{{
                 RowBox[{
                   StyleBox["h", "ExpressionVariable"], "=", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["C", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["C", "ExpressionVariable"]}], "]"}]}]}]}}]], 
             RowBox[{"\[Piecewise]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"m", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "]"}], "\[DoubleLeftArrow]", 
                  RowBox[{
                    UnderscriptBox["GBAuxTest", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"], ",", 
                    StyleBox["h", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], "\:293a", 
                    StyleBox["h", "ExpressionVariable"]}], ",", 
                    RowBox[{"update", "[", 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]}], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"m", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "]"}], "\[DoubleLeftArrow]", 
                  "True"}}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["C", "ExpressionVariable"]}, {
             StyleBox["P", "ExpressionVariable"]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["mi", "ExpressionVariable"]}, {
             StyleBox["mj", "ExpressionVariable"]}, {
             StyleBox[
              RowBox[{"m", "..."}], "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["GBAux", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["C", "ExpressionVariable"], ",", 
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"m", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["let", 
              GridBox[{{
                 RowBox[{
                   StyleBox["h", "ExpressionVariable"], "=", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["C", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["C", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["C", "ExpressionVariable"]}], "]"}]}]}]}}]], 
             RowBox[{"\[Piecewise]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"m", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "]"}], "\[DoubleLeftArrow]", 
                  RowBox[{
                    UnderscriptBox["GBAuxTest", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"], ",", 
                    StyleBox["h", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], "\:293a", 
                    StyleBox["h", "ExpressionVariable"]}], ",", 
                    RowBox[{"update", "[", 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["C", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]}], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"m", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "]"}], "\[DoubleLeftArrow]", 
                  "True"}}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1379909256"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1379909256-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1379909256-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1379909256", "Proof|ID:1379909256-1"},
 CellID->1806109079],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1379909256-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1501295721]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Shape & Membership (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6485400336208153`*^9, 3.648540034516116*^9}, {
  3.6489006069007797`*^9, 3.648900610975916*^9}, {3.649140128762188*^9, 
  3.649140131172592*^9}},
 CellID->145398451],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
   RowBox[{"[", "B", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 884344895}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64913808950629*^9, 3.6491380952257*^9}},
 CellTags->{
  "ID:884344895", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->884344895],

Cell["", "OpenEnvironment",
 CellID->491130265],

Cell[CellGroupData[{

Cell["Lemma: Preparations for Shape of GBAux", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.649138110956957*^9, 3.6491381143294497`*^9}},
 CellID->2121949508],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
  RowBox[{"j", "\[GreaterEqual]", "i"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 2012261150}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648533283809424*^9, 3.648533304320109*^9}},
 CellTags->{
  "ID:2012261150", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->2012261150],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxBase", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Lambda]", 
     RowBox[{"x1", ",", "x2", ",", "x3", ",", "x4", ",", "x5"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", 
         RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
        RowBox[{"[", "U", "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
         RowBox[{"x1", ",", "x2", ",", "x4", ",", "x5", ",", "x3"}], "]"}], 
        "\[Equal]", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"x1", 
          TagBox["\:22c8",
           Identity,
           SyntaxForm->"a*b"], "U"}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], ",", "B", ",", "i", ",", "j"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(shape GBAuxBase)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1467104417", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649138139820804*^9, 3.649138194786585*^9}, {
  3.649139364041464*^9, 3.649139364299399*^9}},
 CellTags->{
  "ID:1467104417", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxBase"},
 CellID->1467104417],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxRec1", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Lambda]", 
     RowBox[{"x1", ",", "x2", ",", "x3", ",", "x4", ",", "x5"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", 
         RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
        RowBox[{"[", "U", "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
         RowBox[{"x1", ",", "x2", ",", "x4", ",", "x5", ",", "x3"}], "]"}], 
        "\[Equal]", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"x1", 
          TagBox["\:22c8",
           Identity,
           SyntaxForm->"a*b"], "U"}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], ",", "B", ",", "i", ",", "j"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(shape GBAuxRec1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1354352171", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649138221291503*^9, 3.649138225947319*^9}, 
   3.649139367552837*^9},
 CellTags->{
  "ID:1354352171", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec1"},
 CellID->1354352171],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxRec2", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Lambda]", 
     RowBox[{"x1", ",", "x2", ",", "x3", ",", "x4", ",", "x5"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", 
         RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
        RowBox[{"[", "U", "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
         RowBox[{"x1", ",", "x2", ",", "x4", ",", "x5", ",", "x3"}], "]"}], 
        "\[Equal]", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"x1", 
          TagBox["\:22c8",
           Identity,
           SyntaxForm->"a*b"], "U"}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], ",", "B", ",", "i", ",", "j"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(shape GBAuxRec2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1322915202", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649138228999214*^9, 3.649138232690217*^9}, 
   3.649139368882846*^9},
 CellTags->{
  "ID:1322915202", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec2"},
 CellID->1322915202],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxRec3", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Lambda]", 
     RowBox[{"x1", ",", "x2", ",", "x3", ",", "x4", ",", "x5"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", 
         RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
        RowBox[{"[", "U", "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
         RowBox[{"x1", ",", "x2", ",", "x4", ",", "x5", ",", "x3"}], "]"}], 
        "\[Equal]", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"x1", 
          TagBox["\:22c8",
           Identity,
           SyntaxForm->"a*b"], "U"}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], ",", "B", ",", "i", ",", "j"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(shape GBAuxRec3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:620036821", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649138230033346*^9, 3.649138235085866*^9}, 
   3.649139369805578*^9},
 CellTags->{
  "ID:620036821", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec3"},
 CellID->620036821],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxRec4", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Lambda]", 
     RowBox[{"x1", ",", "x2", ",", "x3", ",", "x4", ",", "x5"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", 
         RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
        RowBox[{"[", "U", "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
         RowBox[{"x1", ",", "x2", ",", "x4", ",", "x5", ",", "x3"}], "]"}], 
        "\[Equal]", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"x1", 
          TagBox["\:22c8",
           Identity,
           SyntaxForm->"a*b"], "U"}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], ",", "B", ",", "i", ",", "j"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(shape GBAuxRec4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1111467670", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649138231030615*^9, 3.649138236891571*^9}, 
   3.649139370693074*^9},
 CellTags->{
  "ID:1111467670", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec4"},
 CellID->1111467670],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->529886557]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2052416756],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(shape GBAuxBase)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(shape GBAuxBase)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxBase", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["U", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], "\:22c8", 
                    StyleBox["U", "ExpressionVariable"]}]}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxBase", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["U", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], "\:22c8", 
                    StyleBox["U", "ExpressionVariable"]}]}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1467104417"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1467104417-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1467104417-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1467104417", "Proof|ID:1467104417-1"},
 CellID->1997727746],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1467104417-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1261042889]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(shape GBAuxRec1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(shape GBAuxRec1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec1", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["U", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], "\:22c8", 
                    StyleBox["U", "ExpressionVariable"]}]}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec1", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["U", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], "\:22c8", 
                    StyleBox["U", "ExpressionVariable"]}]}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1354352171"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1354352171-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1354352171-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1354352171", "Proof|ID:1354352171-1"},
 CellID->478354620],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1354352171-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->460981908]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(shape GBAuxRec2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(shape GBAuxRec2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec2", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["U", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], "\:22c8", 
                    StyleBox["U", "ExpressionVariable"]}]}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec2", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["U", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], "\:22c8", 
                    StyleBox["U", "ExpressionVariable"]}]}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1322915202"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1322915202-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1322915202-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1322915202", "Proof|ID:1322915202-1"},
 CellID->1180582011],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1322915202-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->195380916]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(shape GBAuxRec3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(shape GBAuxRec3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec3", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["U", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], "\:22c8", 
                    StyleBox["U", "ExpressionVariable"]}]}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec3", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["U", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], "\:22c8", 
                    StyleBox["U", "ExpressionVariable"]}]}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:620036821"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p620036821-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p620036821-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:620036821", "Proof|ID:620036821-1"},
 CellID->1505716064],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p620036821-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->116667553]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(shape GBAuxRec4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(shape GBAuxRec4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec4", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["U", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], "\:22c8", 
                    StyleBox["U", "ExpressionVariable"]}]}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec4", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                    StyleBox["U", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], "\:22c8", 
                    StyleBox["U", "ExpressionVariable"]}]}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1111467670"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1111467670-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1111467670-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1111467670", "Proof|ID:1111467670-1"},
 CellID->284241391],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1111467670-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->723159660]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->525493891],

Cell[CellGroupData[{

Cell["Theorem: Shape of Tuples Returned by GBAux", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648540068953787*^9, 3.6485400969074907`*^9}},
 CellID->167641070],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], ",", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         TagBox["\[LessEqual]",
          Identity], 
         TagBox[")",
          "AutoParentheses"]}]}], "]"}], "]"}]], 
    RowBox[{"[", "P", "]"}]}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{"i", ",", 
      RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
    RowBox[{"j", "\[GreaterEqual]", "i"}]], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", 
        RowBox[{"MMPairs", "[", 
         RowBox[{"\[ScriptCapitalG]", ",", 
          SubscriptBox["B", "i"], ",", 
          SubscriptBox["B", "j"]}], "]"}], "]"}]], 
      RowBox[{"[", "M", "]"}]}]], 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", 
        RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
       RowBox[{"[", "U", "]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
        RowBox[{"B", ",", "P", ",", "i", ",", "j", ",", "M"}], "]"}], 
       "\[Equal]", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"B", 
         TagBox["\:22c8",
          Identity,
          SyntaxForm->"a*b"], "U"}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(shape GBAux)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1314333166", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6485401030034237`*^9, 3.6485401168591957`*^9}, {
   3.648540183179439*^9, 3.648540230223254*^9}, {3.6488988937626257`*^9, 
   3.648898945653164*^9}, {3.648899186672265*^9, 3.648899190743669*^9}, {
   3.648899637130867*^9, 3.648899641772415*^9}, 3.6491380976361637`*^9},
 CellTags->{
  "ID:1314333166", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAux"},
 CellID->1314333166],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->946812009]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1790647428],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(shape GBAux)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(shape GBAux)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
               "[", 
               StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                   StyleBox["U", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["GBAux", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], ",", 
                   StyleBox["P", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"], ",", 
                   StyleBox["M", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], "\:22c8", 
                 StyleBox[
                 "U", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
               "[", 
               StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                   StyleBox["U", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["GBAux", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], ",", 
                   StyleBox["P", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"], ",", 
                   StyleBox["M", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], "\:22c8", 
                 StyleBox[
                 "U", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1314333166"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1314333166-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1314333166-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1314333166", "Proof|ID:1314333166-1"},
 CellID->888484883],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1314333166-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->842826484]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->486203571],

Cell[CellGroupData[{

Cell["Corollary: Membership of Results Returned by GBAux", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648534330270444*^9, 3.648534337385338*^9}, {
  3.648540079106245*^9, 3.6485400877207336`*^9}, {3.649140089197413*^9, 
  3.6491400916652613`*^9}},
 CellID->1209390519],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], ",", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         TagBox["\[LessEqual]",
          Identity], 
         TagBox[")",
          "AutoParentheses"]}]}], "]"}], "]"}]], 
    RowBox[{"[", "P", "]"}]}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{"i", ",", 
      RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
    RowBox[{"j", "\[GreaterEqual]", "i"}]], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", 
        RowBox[{"MMPairs", "[", 
         RowBox[{"\[ScriptCapitalG]", ",", 
          SubscriptBox["B", "i"], ",", 
          SubscriptBox["B", "j"]}], "]"}], "]"}]], 
      RowBox[{"[", "M", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
      RowBox[{"[", 
       RowBox[{
        UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
        RowBox[{"B", ",", "P", ",", "i", ",", "j", ",", "M"}], "]"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership GBAux)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1167572947", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6485343800952797`*^9, 3.648534421408045*^9}, {
  3.6485346319485903`*^9, 3.6485346838253717`*^9}},
 CellTags->{
  "ID:1167572947", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership GBAux"},
 CellID->1167572947],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->765917740]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->858293749],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership GBAux)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership GBAux)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
               "[", 
               StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
                "]"}]], "[", 
             RowBox[{
               UnderscriptBox["GBAux", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["P", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"], ",", 
                 StyleBox["M", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
               "[", 
               StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
                "]"}]], "[", 
             RowBox[{
               UnderscriptBox["GBAux", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["P", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"], ",", 
                 StyleBox["M", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1167572947"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1167572947-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1167572947-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1167572947", "Proof|ID:1167572947-1"},
 CellID->897701299],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1167572947-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->786989361]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Ideal (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.648290871487434*^9, 3.6482908724600363`*^9}, {
  3.6491448401591053`*^9, 3.649144844201685*^9}},
 CellID->1967810377],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
   RowBox[{"[", "B", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 208228172}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649142826920586*^9, 3.6491428413506308`*^9}},
 CellTags->{
  "ID:208228172", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->208228172],

Cell["", "OpenEnvironment",
 CellID->833515163],

Cell[CellGroupData[{

Cell["Lemma", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{
  3.6484383645551233`*^9, {3.649134660565919*^9, 3.6491346634061947`*^9}, {
   3.649143287216885*^9, 3.649143307479622*^9}},
 CellID->1166686321],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{"a", ",", "mi", ",", "mj"}]], 
    RowBox[{
     UnderscriptBox["\[Element]", 
      RowBox[{"MMPairs", "[", 
       RowBox[{"\[ScriptCapitalG]", ",", 
        SubscriptBox["B", "i"], ",", 
        SubscriptBox["B", "j"]}], "]"}]], 
     RowBox[{"[", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"a", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
       "\[RightAngleBracket]"}], "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"a", 
       UnderscriptBox[
        SubsuperscriptBox["\[Rule]", "B", "*"], "\[ScriptCapitalG]"], 
       RowBox[{
        UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
        RowBox[{
         RowBox[{"a", 
          UnderscriptBox["-", "\[ScriptCapitalG]"], 
          RowBox[{"mi", 
           UnderscriptBox["*", "\[ScriptCapitalG]"], 
           SubscriptBox["B", "i"]}]}], ",", "B"}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[And]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"a", 
       UnderscriptBox[
        SubsuperscriptBox["\[Rule]", "B", "*"], "\[ScriptCapitalG]"], 
       RowBox[{
        UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
        RowBox[{
         RowBox[{"a", 
          UnderscriptBox["-", "\[ScriptCapitalG]"], 
          RowBox[{"mj", 
           UnderscriptBox["*", "\[ScriptCapitalG]"], 
           SubscriptBox["B", "j"]}]}], ",", "B"}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(MMPairs \[RightArrow]*)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:781335069", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648379852232953*^9, 3.648379885900406*^9}, {
  3.64837993151435*^9, 3.6483799907113647`*^9}, {3.6483800501427937`*^9, 
  3.6483801074915257`*^9}, {3.648380172605957*^9, 3.6483801757092524`*^9}, {
  3.6491345826022053`*^9, 3.649134641910013*^9}},
 CellTags->{
  "ID:781335069", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[RightArrow]*"},
 CellID->781335069],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1385002694]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1935100623],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(MMPairs \[RightArrow]*)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(MMPairs \[RightArrow]*)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["mi", "ExpressionVariable"]}, {
             StyleBox["mj", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"MMPairs", "[", 
                RowBox[{
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                  SubscriptBox[
                   StyleBox["B", "ExpressionVariable"], 
                   StyleBox["i", "ExpressionVariable"]], ",", 
                  SubscriptBox[
                   StyleBox["B", "ExpressionVariable"], 
                   StyleBox["j", "ExpressionVariable"]]}], "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
             "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
               RowBox[{
                 UnderscriptBox["trd", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                   StyleBox["B", "ExpressionVariable"]}], "]"}]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
               RowBox[{
                 UnderscriptBox["trd", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                   StyleBox["B", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["mi", "ExpressionVariable"]}, {
             StyleBox["mj", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"MMPairs", "[", 
                RowBox[{
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                  SubscriptBox[
                   StyleBox["B", "ExpressionVariable"], 
                   StyleBox["i", "ExpressionVariable"]], ",", 
                  SubscriptBox[
                   StyleBox["B", "ExpressionVariable"], 
                   StyleBox["j", "ExpressionVariable"]]}], "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
             "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
               RowBox[{
                 UnderscriptBox["trd", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                   StyleBox["B", "ExpressionVariable"]}], "]"}]}], "\[And]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox[
                TemplateBox[{"\[Rule]", 
                  StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript",
                  SyntaxForm -> SubsuperscriptBox], 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
               RowBox[{
                 UnderscriptBox["trd", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                   StyleBox["B", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:781335069"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p781335069-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p781335069-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:781335069", "Proof|ID:781335069-1"},
 CellID->283149330],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p781335069-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->296243687]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1412924928],

Cell[CellGroupData[{

Cell["Lemma: Preparations for Ideal of GBAux", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{3.6491433098756237`*^9},
 CellID->843145011],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
  RowBox[{"j", "\[GreaterEqual]", "i"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1452629404}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648533283809424*^9, 3.648533304320109*^9}},
 CellTags->{
  "ID:1452629404", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1452629404],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxBase", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Lambda]", 
     RowBox[{"x1", ",", "x2", ",", "x3", ",", "x4", ",", "x5"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["ideal", "\[ScriptCapitalG]"], "[", 
       RowBox[{
        UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
        RowBox[{"x1", ",", "x2", ",", "x4", ",", "x5", ",", "x3"}], "]"}], 
       "]"}], "\[Equal]", 
      RowBox[{
       UnderscriptBox["ideal", "\[ScriptCapitalG]"], "[", "x1", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], ",", "B", ",", "i", ",", "j"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(ideal GBAuxBase)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1597922168", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649143334947369*^9, 3.649143365097886*^9}},
 CellTags->{
  "ID:1597922168", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxBase"},
 CellID->1597922168],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxRec1", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Lambda]", 
     RowBox[{"x1", ",", "x2", ",", "x3", ",", "x4", ",", "x5"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["ideal", "\[ScriptCapitalG]"], "[", 
       RowBox[{
        UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
        RowBox[{"x1", ",", "x2", ",", "x4", ",", "x5", ",", "x3"}], "]"}], 
       "]"}], "\[Equal]", 
      RowBox[{
       UnderscriptBox["ideal", "\[ScriptCapitalG]"], "[", "x1", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], ",", "B", ",", "i", ",", "j"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(ideal GBAuxRec1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:348489962", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649143373019745*^9, 3.649143377234912*^9}},
 CellTags->{
  "ID:348489962", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec1"},
 CellID->348489962],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxRec2", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Lambda]", 
     RowBox[{"x1", ",", "x2", ",", "x3", ",", "x4", ",", "x5"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["ideal", "\[ScriptCapitalG]"], "[", 
       RowBox[{
        UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
        RowBox[{"x1", ",", "x2", ",", "x4", ",", "x5", ",", "x3"}], "]"}], 
       "]"}], "\[Equal]", 
      RowBox[{
       UnderscriptBox["ideal", "\[ScriptCapitalG]"], "[", "x1", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], ",", "B", ",", "i", ",", "j"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(ideal GBAuxRec2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1787031010", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649143381251823*^9, 3.6491433850258493`*^9}},
 CellTags->{
  "ID:1787031010", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec2"},
 CellID->1787031010],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxRec3", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Lambda]", 
     RowBox[{"x1", ",", "x2", ",", "x3", ",", "x4", ",", "x5"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["ideal", "\[ScriptCapitalG]"], "[", 
       RowBox[{
        UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
        RowBox[{"x1", ",", "x2", ",", "x4", ",", "x5", ",", "x3"}], "]"}], 
       "]"}], "\[Equal]", 
      RowBox[{
       UnderscriptBox["ideal", "\[ScriptCapitalG]"], "[", "x1", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], ",", "B", ",", "i", ",", "j"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(ideal GBAuxRec3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:963974147", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649143382286908*^9, 3.649143386168551*^9}},
 CellTags->{
  "ID:963974147", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec3"},
 CellID->963974147],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxRec4", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Lambda]", 
     RowBox[{"x1", ",", "x2", ",", "x3", ",", "x4", ",", "x5"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["ideal", "\[ScriptCapitalG]"], "[", 
       RowBox[{
        UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
        RowBox[{"x1", ",", "x2", ",", "x4", ",", "x5", ",", "x3"}], "]"}], 
       "]"}], "\[Equal]", 
      RowBox[{
       UnderscriptBox["ideal", "\[ScriptCapitalG]"], "[", "x1", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], ",", "B", ",", "i", ",", "j"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(ideal GBAuxRec4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:473957484", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6491433832842283`*^9, 3.649143387789941*^9}},
 CellTags->{
  "ID:473957484", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec4"},
 CellID->473957484],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1659585585]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->154264307],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(ideal GBAuxBase)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(ideal GBAuxBase)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxBase", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "]"}], 
                   "\[Equal]", 
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["x1", "ExpressionVariable"], "]"}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxBase", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "]"}], 
                   "\[Equal]", 
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["x1", "ExpressionVariable"], "]"}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1597922168"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1597922168-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1597922168-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1597922168", "Proof|ID:1597922168-1"},
 CellID->1879777321],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1597922168-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->208056954]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(ideal GBAuxRec1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(ideal GBAuxRec1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec1", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "]"}], 
                   "\[Equal]", 
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["x1", "ExpressionVariable"], "]"}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec1", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "]"}], 
                   "\[Equal]", 
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["x1", "ExpressionVariable"], "]"}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:348489962"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p348489962-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p348489962-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:348489962", "Proof|ID:348489962-1"},
 CellID->917719347],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p348489962-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1071294072]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(ideal GBAuxRec2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(ideal GBAuxRec2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec2", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "]"}], 
                   "\[Equal]", 
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["x1", "ExpressionVariable"], "]"}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec2", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "]"}], 
                   "\[Equal]", 
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["x1", "ExpressionVariable"], "]"}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1787031010"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1787031010-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1787031010-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1787031010", "Proof|ID:1787031010-1"},
 CellID->1427595038],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1787031010-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2125736511]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(ideal GBAuxRec3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(ideal GBAuxRec3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec3", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "]"}], 
                   "\[Equal]", 
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["x1", "ExpressionVariable"], "]"}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec3", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "]"}], 
                   "\[Equal]", 
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["x1", "ExpressionVariable"], "]"}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:963974147"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p963974147-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p963974147-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:963974147", "Proof|ID:963974147-1"},
 CellID->1497403325],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p963974147-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->996071133]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(ideal GBAuxRec4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(ideal GBAuxRec4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec4", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "]"}], 
                   "\[Equal]", 
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["x1", "ExpressionVariable"], "]"}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec4", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Lambda]", 
                  RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"]}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["x1", "ExpressionVariable"], ",", 
                    StyleBox["x2", "ExpressionVariable"], ",", 
                    StyleBox["x4", "ExpressionVariable"], ",", 
                    StyleBox["x5", "ExpressionVariable"], ",", 
                    StyleBox["x3", "ExpressionVariable"]}], "]"}], "]"}], 
                   "\[Equal]", 
                   RowBox[{
                    UnderscriptBox["ideal", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["x1", "ExpressionVariable"], "]"}]}]}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:473957484"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p473957484-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p473957484-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:473957484", "Proof|ID:473957484-1"},
 CellID->1947252190],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p473957484-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->246554488]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1579370208],

Cell[CellGroupData[{

Cell["Theorem: Ideal of Output Returned by GBAux", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.649144546992681*^9, 3.6491445734316483`*^9}},
 CellID->2113590487],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], ",", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         TagBox["\[LessEqual]",
          Identity], 
         TagBox[")",
          "AutoParentheses"]}]}], "]"}], "]"}]], 
    RowBox[{"[", "P", "]"}]}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{"i", ",", 
      RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
    RowBox[{"j", "\[GreaterEqual]", "i"}]], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", 
        RowBox[{"MMPairs", "[", 
         RowBox[{"\[ScriptCapitalG]", ",", 
          SubscriptBox["B", "i"], ",", 
          SubscriptBox["B", "j"]}], "]"}], "]"}]], 
      RowBox[{"[", "M", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["ideal", "\[ScriptCapitalG]"], "[", 
       RowBox[{
        UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
        RowBox[{"B", ",", "P", ",", "i", ",", "j", ",", "M"}], "]"}], "]"}], 
      "\[Equal]", 
      RowBox[{
       UnderscriptBox["ideal", "\[ScriptCapitalG]"], "[", "B", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(ideal GBAux)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1826585682", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649144584883565*^9, 3.6491446203102627`*^9}},
 CellTags->{
  "ID:1826585682", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAux"},
 CellID->1826585682],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->670628316]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1004533723],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(ideal GBAux)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(ideal GBAux)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
               "[", 
               StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["ideal", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 UnderscriptBox["GBAux", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], ",", 
                   StyleBox["P", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"], ",", 
                   StyleBox["M", "ExpressionVariable"]}], "]"}], "]"}], 
             "\[Equal]", 
             RowBox[{
               UnderscriptBox["ideal", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
               "[", 
               StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["ideal", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 UnderscriptBox["GBAux", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], ",", 
                   StyleBox["P", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"], ",", 
                   StyleBox["M", "ExpressionVariable"]}], "]"}], "]"}], 
             "\[Equal]", 
             RowBox[{
               UnderscriptBox["ideal", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1826585682"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1826585682-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1826585682-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1826585682", "Proof|ID:1826585682-1"},
 CellID->98529934],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1826585682-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1922222818]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Connectibility of Critical Pairs (\[Checkmark])\
\>", "Subsection",
 CellChangeTimes->{{3.648290877546577*^9, 3.648290880431984*^9}, {
  3.649134508775758*^9, 3.649134523684967*^9}},
 CellID->23100310],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
   RowBox[{"[", "B", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 416121228}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64880834040443*^9, 3.6488083455746393`*^9}},
 CellTags->{
  "ID:416121228", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->416121228],

Cell[CellGroupData[{

Cell["Definitions", "Subsubsection",
 CellChangeTimes->{{3.649049779328154*^9, 3.649049780953072*^9}, {
   3.6490498459719057`*^9, 3.649049849577444*^9}, 3.64939314937488*^9},
 CellID->1168597562],

Cell["", "OpenEnvironment",
 CellID->1967738288],

Cell[CellGroupData[{

Cell["\<\
Definition: Inductive Assertions About Connectibility of Critical Pairs\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648539789912755*^9, 3.648539812168797*^9}, {
   3.648540434456492*^9, 3.648540442640421*^9}, 3.6488927639910316`*^9},
 CellID->1074460594],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"C", ",", "P", ",", "M", ",", "i", ",", "j"}]]], "GlobalDeclaration",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 2050404037}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.648892770300284*^9},
 CellTags->{
  "ID:2050404037", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->2050404037],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isCPTODOTuple", "\[ScriptCapitalG]"], "[", 
   RowBox[{"M", ",", "P", ",", "C", ",", "i", ",", "j"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Or]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"\[Not]", 
         RowBox[{"processed", "[", 
          RowBox[{"i", ",", "j", ",", "P"}], "]"}]}]},
       {
        RowBox[{
         UnderscriptBox[
          UnderscriptBox["\[ForAll]", 
           RowBox[{"k", ",", 
            RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
            RowBox[{
             UnderscriptBox["multN", "\[ScriptCapitalG]"], "[", 
             SubscriptBox["C", "i"], "]"}]}]], 
          RowBox[{
           RowBox[{
            SubscriptBox["C", "i"], "\[Equal]", 
            SubscriptBox["C", "j"]}], "\[Implies]", 
           RowBox[{"l", ">", "k"}]}]], 
         RowBox[{
          UnderscriptBox[
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
             RowBox[{"[", "a", "]"}]}]], 
           RowBox[{
            UnderscriptBox["isMNTCR", "\[ScriptCapitalG]"], "[", 
            RowBox[{
             SubscriptBox["C", "i"], ",", 
             SubscriptBox["C", "j"], ",", "a", ",", "k", ",", "l"}], "]"}]], 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           RowBox[{
            RowBox[{
             TagBox["(",
              "AutoParentheses"], 
             RowBox[{"\[Not]", 
              RowBox[{
               UnderscriptBox["\[Exists]", 
                RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
                 RowBox[{"[", 
                  RowBox[{"b", ",", "mi", ",", "mj"}], "]"}]}]], 
               RowBox[{
                TagBox["(",
                 "AutoParentheses"], 
                RowBox[{
                 RowBox[{"a", 
                  UnderscriptBox["\[Tilde]", "\[ScriptCapitalG]"], "b"}], 
                 "\[And]", 
                 RowBox[{
                  RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{"b", ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
                   "\[RightAngleBracket]"}], 
                  TagBox["\:22ff",
                   Identity,
                   SyntaxForm->"a+b"], "M"}], "\[And]", 
                 RowBox[{
                  UnderscriptBox["areCPMultipliers", "\[ScriptCapitalG]"], 
                  "[", 
                  RowBox[{"mi", ",", "mj", ",", 
                   SubscriptBox["C", "i"], ",", 
                   SubscriptBox["C", "j"], ",", "b", ",", "k", ",", "l"}], 
                  "]"}]}], 
                TagBox[")",
                 "AutoParentheses"]}]}]}], 
             TagBox[")",
              "AutoParentheses"]}], "\[Implies]", 
            RowBox[{
             UnderscriptBox["cpConnectible", "\[ScriptCapitalG]"], "[", 
             RowBox[{
              SubscriptBox["C", "i"], ",", 
              SubscriptBox["C", "j"], ",", "a", ",", "k", ",", "l", ",", 
              "C"}], "]"}]}], 
           TagBox[")",
            "AutoParentheses"]}]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isCPTODOTuple)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1837575531", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649048243562953*^9, 3.6490482818572683`*^9}, {
   3.6490483215989122`*^9, 3.6490485715276833`*^9}, {3.649048616638103*^9, 
   3.649048632247025*^9}, {3.649048847952099*^9, 3.649048850388538*^9}, {
   3.64904895364186*^9, 3.6490489737592*^9}, {3.649050281374505*^9, 
   3.649050290966857*^9}, {3.649050528523202*^9, 3.649050538481936*^9}, 
   3.6490518546839657`*^9, {3.6490635626012363`*^9, 3.649063563376974*^9}, {
   3.6502680424163313`*^9, 3.650268092331319*^9}},
 CellTags->{
  "ID:1837575531", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple"},
 CellID->1837575531],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["GBAuxCPConnectible1", "\[ScriptCapitalG]"], "[", 
   RowBox[{"C", ",", "P", ",", "M", ",", "i", ",", "j"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isTODOTuple", "\[ScriptCapitalG]"], "[", 
     RowBox[{"P", ",", "C", ",", "i", ",", "j"}], "]"}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{"a", ",", "mi", ",", "mj"}]], 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"a", ",", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
         "\[RightAngleBracket]"}], 
        TagBox["\:22ff",
         Identity,
         SyntaxForm->"a+b"], "M"}]], 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{"k", ",", 
         RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{
          UnderscriptBox["multN", "\[ScriptCapitalG]"], "[", 
          SubscriptBox["C", "i"], "]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{
          UnderscriptBox["areCPMultipliers", "\[ScriptCapitalG]"], "[", 
          RowBox[{"mi", ",", "mj", ",", 
           SubscriptBox["C", "i"], ",", 
           SubscriptBox["C", "j"], ",", "a", ",", "k", ",", "l"}], "]"}], 
         "\[Implies]", 
         RowBox[{
          UnderscriptBox["cpConnectible", "\[ScriptCapitalG]"], "[", 
          RowBox[{
           SubscriptBox["C", "i"], ",", 
           SubscriptBox["C", "j"], ",", "a", ",", "k", ",", "l", ",", 
           RowBox[{
            UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
            RowBox[{"C", ",", "P", ",", "i", ",", "j", ",", "M"}], "]"}]}], 
          "]"}]}], 
        TagBox[")",
         "AutoParentheses"]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:346323949", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648539832323435*^9, 3.648539893231*^9}, {
   3.648539939249311*^9, 3.64853995864078*^9}, {3.6485403218133802`*^9, 
   3.6485403259559507`*^9}, {3.648540445596977*^9, 3.648540455022246*^9}, 
   3.648892683923718*^9, {3.6488927261799393`*^9, 3.648892772753129*^9}, {
   3.648892803766676*^9, 3.648892946188971*^9}, {3.6488929846028976`*^9, 
   3.648892997454808*^9}, {3.6488946306877737`*^9, 3.648894649937767*^9}, {
   3.648957634551903*^9, 3.6489576352537317`*^9}},
 CellTags->{
  "ID:346323949", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1"},
 CellID->346323949],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["GBAuxCPConnectible2", "\[ScriptCapitalG]"], "[", 
   RowBox[{"C", ",", "P", ",", "M", ",", "i", ",", "j"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["isTODOTuple", "\[ScriptCapitalG]"], "[", 
       RowBox[{"P", ",", "C", ",", "i", ",", "j"}], "]"}], "\[And]", 
      RowBox[{
       UnderscriptBox["isCPTODOTuple", "\[ScriptCapitalG]"], "[", 
       RowBox[{"M", ",", "P", ",", "C", ",", "i", ",", "j"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{"i0", ",", "j0"}]], 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"i0", ",", "j0"}], "\[RightAngleBracket]"}], 
        TagBox["\:22ff",
         Identity,
         SyntaxForm->"a+b"], "P"}]], 
      RowBox[{
       UnderscriptBox["cpConnectible", "\[ScriptCapitalG]"], "[", 
       RowBox[{
        RowBox[{
         UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
         RowBox[{"C", ",", "P", ",", "i", ",", "j", ",", "M"}], "]"}], ",", 
        SubscriptBox["C", "i0"], ",", 
        SubscriptBox["C", "j0"]}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1643593823", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648892972799567*^9, 3.648892979660475*^9}, 
   3.648893052501822*^9, {3.6488930913635902`*^9, 3.648893138247327*^9}, {
   3.648957637759717*^9, 3.648957638426692*^9}, {3.6489742246404257`*^9, 
   3.648974250087842*^9}, {3.6489744207244596`*^9, 3.6489744243406343`*^9}, 
   3.6489768544778233`*^9, 3.648978600781934*^9, {3.648980534533905*^9, 
   3.648980571905306*^9}, {3.6489806380569143`*^9, 3.648980659043194*^9}, {
   3.6489872204512453`*^9, 3.6489872267961884`*^9}, {3.649048829706746*^9, 
   3.649048854189986*^9}, {3.6490489813583193`*^9, 3.649048981650765*^9}},
 CellTags->{
  "ID:1643593823", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2"},
 CellID->1643593823],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["GBAuxCPConnectible3", "\[ScriptCapitalG]"], "[", 
   RowBox[{"C", ",", "P", ",", "M", ",", "i", ",", "j"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["isTODOTuple", "\[ScriptCapitalG]"], "[", 
       RowBox[{"P", ",", "C", ",", "i", ",", "j"}], "]"}], "\[And]", 
      RowBox[{
       UnderscriptBox["isCPTODOTuple", "\[ScriptCapitalG]"], "[", 
       RowBox[{"M", ",", "P", ",", "C", ",", "i", ",", "j"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{"i0", ",", 
         RowBox[{"j0", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"\[LeftBracketingBar]", 
          RowBox[{
           UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
           RowBox[{"C", ",", "P", ",", "i", ",", "j", ",", "M"}], "]"}], 
          "\[RightBracketingBar]"}]}]], 
       RowBox[{
        RowBox[{"j0", "\[GreaterEqual]", "i0"}], "\[And]", 
        RowBox[{"j0", ">", 
         RowBox[{"\[LeftBracketingBar]", "C", "\[RightBracketingBar]"}]}]}]], 
      
      RowBox[{
       UnderscriptBox["cpConnectible", "\[ScriptCapitalG]"], "[", 
       RowBox[{
        RowBox[{
         UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
         RowBox[{"C", ",", "P", ",", "i", ",", "j", ",", "M"}], "]"}], ",", 
        SubscriptBox[
         RowBox[{
          UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
          RowBox[{"C", ",", "P", ",", "i", ",", "j", ",", "M"}], "]"}], "i0"],
         ",", 
        SubscriptBox[
         RowBox[{
          UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
          RowBox[{"C", ",", "P", ",", "i", ",", "j", ",", "M"}], "]"}], 
         "j0"]}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1213243088", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6491463457820787`*^9, 3.649146458363299*^9}, 
   3.6491465237642193`*^9, 3.649146614051526*^9, {3.6491477005949383`*^9, 
   3.64914770451724*^9}},
 CellTags->{
  "ID:1213243088", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3"},
 CellID->1213243088],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->388797222]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->424254111]
}, Closed]],

Cell[CellGroupData[{

Cell["GBAuxTest (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.649049763144199*^9, 3.649049771880019*^9}, {
   3.649049805782078*^9, 3.649049816464239*^9}, 3.6490499177537727`*^9},
 CellID->2121626066],

Cell["", "OpenEnvironment",
 CellID->977102521],

Cell[CellGroupData[{

Cell["Lemma: Connectibility Below if trd-Difference is 0", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6488803229934673`*^9, 3.648880325399979*^9}, 
   3.648880467208131*^9, {3.648967256722734*^9, 3.648967281098626*^9}, {
   3.6489689830861883`*^9, 3.6489689874025297`*^9}},
 CellID->2132467789],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{"a", ",", "mi", ",", "mj"}]], 
    RowBox[{
     UnderscriptBox["\[Element]", 
      RowBox[{"MMPairs", "[", 
       RowBox[{"\[ScriptCapitalG]", ",", 
        SubscriptBox["B", "i"], ",", 
        SubscriptBox["B", "j"]}], "]"}]], 
     RowBox[{"[", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"a", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
       "\[RightAngleBracket]"}], "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        RowBox[{
         UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
         RowBox[{
          RowBox[{"a", 
           UnderscriptBox["-", "\[ScriptCapitalG]"], 
           RowBox[{"mi", 
            UnderscriptBox["*", "\[ScriptCapitalG]"], 
            SubscriptBox["B", "i"]}]}], ",", "B"}], "]"}], 
        UnderscriptBox["-", "\[ScriptCapitalG]"], 
        RowBox[{
         UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
         RowBox[{
          RowBox[{"a", 
           UnderscriptBox["-", "\[ScriptCapitalG]"], 
           RowBox[{"mj", 
            UnderscriptBox["*", "\[ScriptCapitalG]"], 
            SubscriptBox["B", "j"]}]}], ",", "B"}], "]"}]}], "\[Equal]", 
       UnderscriptBox["0", "\[ScriptCapitalG]"]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"a", 
         UnderscriptBox["-", "\[ScriptCapitalG]"], 
         RowBox[{"mi", 
          UnderscriptBox["*", "\[ScriptCapitalG]"], 
          SubscriptBox["B", "i"]}]}], 
        TagBox[")",
         "AutoParentheses"]}], 
       UnderscriptBox[
        OverscriptBox[
         SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
         RowBox[{"a", ",", 
          RowBox[{
           TagBox["(",
            "AutoParentheses"], 
           TagBox["\[Succeeds]",
            Identity], 
           TagBox[")",
            "AutoParentheses"]}]}]], "\[ScriptCapitalG]"], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"a", 
         UnderscriptBox["-", "\[ScriptCapitalG]"], 
         RowBox[{"mj", 
          UnderscriptBox["*", "\[ScriptCapitalG]"], 
          SubscriptBox["B", "j"]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(h=0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:949923506", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6488803305576677`*^9, 3.648880457765586*^9}, {
  3.648883808648327*^9, 3.648883822817978*^9}, {3.648966796770176*^9, 
  3.648966808568452*^9}, {3.64896690693497*^9, 3.6489669313922243`*^9}, {
  3.648967060459652*^9, 3.648967068946157*^9}, {3.64896894757365*^9, 
  3.64896897580332*^9}},
 CellTags->{
  "ID:949923506", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "h=0"},
 CellID->949923506],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->543412832]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->505215769],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(h=0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(h=0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["mi", "ExpressionVariable"]}, {
             StyleBox["mj", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"MMPairs", "[", 
                  RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
               "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["-", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 RowBox[{
                   UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}], 
               "\[Equal]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
               RowBox[{
                 StyleBox["mi", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]]}]}], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
               RowBox[{
                 StyleBox["mj", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox[
                  "j", "ExpressionVariable"]]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             StyleBox["a", "ExpressionVariable"]}, {
             StyleBox["mi", "ExpressionVariable"]}, {
             StyleBox["mj", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"MMPairs", "[", 
                  RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], 
               "]"}], "\[And]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["-", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 RowBox[{
                   UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}], 
               "\[Equal]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
               RowBox[{
                 StyleBox["mi", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]]}]}], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 StyleBox["B", "ExpressionVariable"], "*"}, "Subsuperscript", 
                SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
               RowBox[{
                 StyleBox["mj", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox[
                  "j", "ExpressionVariable"]]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:949923506"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p949923506-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p949923506-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:949923506", "Proof|ID:949923506-1"},
 CellID->221566149],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p949923506-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->87967590]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1859584630],

Cell[CellGroupData[{

Cell["Lemma: cpConnectible with trd-Difference Appended", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648967215244953*^9, 3.6489672426871767`*^9}},
 CellID->852372469],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"i", ",", 
   RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
   "\[LeftBracketingBar]", "B", 
    "\[RightBracketingBar]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1322820809}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.648968860352275*^9},
 CellTags->{
  "ID:1322820809", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1322820809],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "mi", ",", "mj"}], "]"}]}], ",", "k", ",", 
   "l"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1772563804}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648968869403737*^9, 3.648968877229372*^9}, {
  3.648968929079021*^9, 3.6489689297463217`*^9}},
 CellTags->{
  "ID:1772563804", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1772563804],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["areCPMultipliers", "\[ScriptCapitalG]"], "[", 
   RowBox[{"mi", ",", "mj", ",", 
    SubscriptBox["B", "i"], ",", 
    SubscriptBox["B", "j"], ",", "a", ",", "k", ",", "l"}], "]"}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 667808408}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648968913381483*^9, 3.64896893339734*^9}},
 CellTags->{
  "ID:667808408", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->667808408],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["-", "\[ScriptCapitalG]"], 
    RowBox[{"mi", 
     UnderscriptBox["*", "\[ScriptCapitalG]"], 
     SubscriptBox["B", "i"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], 
  UnderscriptBox[
   OverscriptBox[
    SubsuperscriptBox["\[LeftRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"B", 
       TagBox["\:293a",
        Identity,
        SyntaxForm->"a*b"], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
          RowBox[{
           RowBox[{"a", 
            UnderscriptBox["-", "\[ScriptCapitalG]"], 
            RowBox[{"mi", 
             UnderscriptBox["*", "\[ScriptCapitalG]"], 
             SubscriptBox["B", "i"]}]}], ",", "B"}], "]"}], 
         UnderscriptBox["-", "\[ScriptCapitalG]"], 
         RowBox[{
          UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
          RowBox[{
           RowBox[{"a", 
            UnderscriptBox["-", "\[ScriptCapitalG]"], 
            RowBox[{"mj", 
             UnderscriptBox["*", "\[ScriptCapitalG]"], 
             SubscriptBox["B", "j"]}]}], ",", "B"}], "]"}]}], ")"}]}], 
      TagBox[")",
       "AutoParentheses"]}], "*"], 
    RowBox[{"a", ",", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      TagBox["\[Succeeds]",
       Identity], 
      TagBox[")",
       "AutoParentheses"]}]}]], "\[ScriptCapitalG]"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["-", "\[ScriptCapitalG]"], 
    RowBox[{"mj", 
     UnderscriptBox["*", "\[ScriptCapitalG]"], 
     SubscriptBox["B", "j"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(h\[NotEqual]0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1601898370", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6489668221198587`*^9, 3.648966838655945*^9}, {
  3.6489669349344788`*^9, 3.64896694682698*^9}, {3.648967073525838*^9, 
  3.6489670748549213`*^9}},
 CellTags->{
  "ID:1601898370", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "h\[NotEqual]0"},
 CellID->1601898370],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["cpConnectible", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   SubscriptBox["B", "i"], ",", 
   SubscriptBox["B", "j"], ",", "a", ",", "k", ",", "l", ",", 
   RowBox[{"B", 
    TagBox["\:293a",
     Identity,
     SyntaxForm->"a*b"], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
       RowBox[{
        RowBox[{"a", 
         UnderscriptBox["-", "\[ScriptCapitalG]"], 
         RowBox[{"mi", 
          UnderscriptBox["*", "\[ScriptCapitalG]"], 
          SubscriptBox["B", "i"]}]}], ",", "B"}], "]"}], 
      UnderscriptBox["-", "\[ScriptCapitalG]"], 
      RowBox[{
       UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
       RowBox[{
        RowBox[{"a", 
         UnderscriptBox["-", "\[ScriptCapitalG]"], 
         RowBox[{"mj", 
          UnderscriptBox["*", "\[ScriptCapitalG]"], 
          SubscriptBox["B", "j"]}]}], ",", "B"}], "]"}]}], ")"}]}]}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cpConnectible append h)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:622683853", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6489669572055902`*^9, 3.648967047339808*^9}, {
  3.6489670930456457`*^9, 3.64896720910145*^9}, {3.648967781427302*^9, 
  3.648967811701037*^9}, {3.64896885888028*^9, 3.6489689416919127`*^9}},
 CellTags->{
  "ID:622683853", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpConnectible append h"},
 CellID->622683853],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2105705679]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->645399240],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(h\[NotEqual]0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(h\[NotEqual]0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mi", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mj", "ExpressionVariable"], "]"}]}, {
             StyleBox["k", "ExpressionVariable"]}, {
             StyleBox["l", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areCPMultipliers", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["mi", "ExpressionVariable"], ",", 
               StyleBox["mj", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"], ",", 
               StyleBox["l", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
               RowBox[{
                 StyleBox["mi", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]]}]}], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "\:293a", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}]}], "*"}, 
                "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
               RowBox[{
                 StyleBox["mj", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox[
                  "j", "ExpressionVariable"]]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mi", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mj", "ExpressionVariable"], "]"}]}, {
             StyleBox["k", "ExpressionVariable"]}, {
             StyleBox["l", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areCPMultipliers", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["mi", "ExpressionVariable"], ",", 
               StyleBox["mj", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"], ",", 
               StyleBox["l", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
               RowBox[{
                 StyleBox["mi", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]]}]}], 
             UnderscriptBox[
              OverscriptBox[
               TemplateBox[{"\[LeftRightArrow]", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "\:293a", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}]}], "*"}, 
                "Subsuperscript", SyntaxForm -> SubsuperscriptBox], 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", "\[Succeeds]"}]], 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
               RowBox[{
                 StyleBox["mj", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox[
                  "j", "ExpressionVariable"]]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1601898370"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1601898370-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1601898370-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1601898370", "Proof|ID:1601898370-1"},
 CellID->76678778],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1601898370-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->653057816]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cpConnectible append h)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cpConnectible append h)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mi", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mj", "ExpressionVariable"], "]"}]}, {
             StyleBox["k", "ExpressionVariable"]}, {
             StyleBox["l", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areCPMultipliers", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["mi", "ExpressionVariable"], ",", 
               StyleBox["mj", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"], ",", 
               StyleBox["l", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["cpConnectible", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"], ",", 
               StyleBox["l", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], "\:293a", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mi", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mj", "ExpressionVariable"], "]"}]}, {
             StyleBox["k", "ExpressionVariable"]}, {
             StyleBox["l", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["areCPMultipliers", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["mi", "ExpressionVariable"], ",", 
               StyleBox["mj", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"], ",", 
               StyleBox["l", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["cpConnectible", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"], ",", 
               StyleBox["l", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], "\:293a", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:622683853"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p622683853-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p622683853-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:622683853", "Proof|ID:622683853-1"},
 CellID->836026655],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p622683853-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1838968259]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1568209243],

Cell[CellGroupData[{

Cell["Lemma: Properties of GBAuxTest", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6488786849432497`*^9, 3.648878690153943*^9}},
 CellID->296316007],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "P"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", 
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
      RowBox[{"[", 
       RowBox[{"a", ",", "mi", ",", "mj"}], "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["GBAuxTest", "\[ScriptCapitalG]"], "[", 
       RowBox[{"B", ",", "a", ",", "mi", ",", "mj", ",", 
        RowBox[{
         RowBox[{
          UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
          RowBox[{
           RowBox[{"a", 
            UnderscriptBox["-", "\[ScriptCapitalG]"], 
            RowBox[{"mi", 
             UnderscriptBox["*", "\[ScriptCapitalG]"], 
             SubscriptBox["B", "i"]}]}], ",", "B"}], "]"}], 
         UnderscriptBox["-", "\[ScriptCapitalG]"], 
         RowBox[{
          UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
          RowBox[{
           RowBox[{"a", 
            UnderscriptBox["-", "\[ScriptCapitalG]"], 
            RowBox[{"mj", 
             UnderscriptBox["*", "\[ScriptCapitalG]"], 
             SubscriptBox["B", "j"]}]}], ",", "B"}], "]"}]}], ",", "P", ",", 
        "i", ",", "j"}], "]"}], 
      TagBox["\[DoubleLongLeftRightArrow]",
       Identity,
       SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["chainCrit", "\[ScriptCapitalG]"], "[", 
         RowBox[{"i", ",", "j", ",", "a", ",", "B", ",", "P"}], "]"}], 
        "\[Or]", 
        RowBox[{
         RowBox[{
          RowBox[{
           UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
           RowBox[{
            RowBox[{"a", 
             UnderscriptBox["-", "\[ScriptCapitalG]"], 
             RowBox[{"mi", 
              UnderscriptBox["*", "\[ScriptCapitalG]"], 
              SubscriptBox["B", "i"]}]}], ",", "B"}], "]"}], 
          UnderscriptBox["-", "\[ScriptCapitalG]"], 
          RowBox[{
           UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
           RowBox[{
            RowBox[{"a", 
             UnderscriptBox["-", "\[ScriptCapitalG]"], 
             RowBox[{"mj", 
              UnderscriptBox["*", "\[ScriptCapitalG]"], 
              SubscriptBox["B", "j"]}]}], ",", "B"}], "]"}]}], "\[Equal]", 
         UnderscriptBox["0", "\[ScriptCapitalG]"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxTest alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1834517245", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6488797339143877`*^9, 3.6488798098902273`*^9}, {
  3.648880538605566*^9, 3.648880569345483*^9}, {3.648884642491103*^9, 
  3.648884696988369*^9}, {3.6488848135282173`*^9, 3.648884861647065*^9}},
 CellTags->{
  "ID:1834517245", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxTest alternative"},
 CellID->1834517245],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "P"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", 
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
      RowBox[{"[", 
       RowBox[{"a", ",", "mi", ",", "mj"}], "]"}]}]], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{"k", ",", 
       RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{
        UnderscriptBox["multN", "\[ScriptCapitalG]"], "[", 
        SubscriptBox["B", "i"], "]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{
          UnderscriptBox["isTODOTuple", "\[ScriptCapitalG]"], "[", 
          RowBox[{"P", ",", "B", ",", "i", ",", "j"}], "]"}], "\[And]", 
         RowBox[{
          UnderscriptBox["areCPMultipliers", "\[ScriptCapitalG]"], "[", 
          RowBox[{"mi", ",", "mj", ",", 
           SubscriptBox["B", "i"], ",", 
           SubscriptBox["B", "j"], ",", "a", ",", "k", ",", "l"}], "]"}], 
         "\[And]", 
         RowBox[{
          UnderscriptBox["GBAuxTest", "\[ScriptCapitalG]"], "[", 
          RowBox[{"B", ",", "a", ",", "mi", ",", "mj", ",", 
           RowBox[{
            RowBox[{
             UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
             RowBox[{
              RowBox[{"a", 
               UnderscriptBox["-", "\[ScriptCapitalG]"], 
               RowBox[{"mi", 
                UnderscriptBox["*", "\[ScriptCapitalG]"], 
                SubscriptBox["B", "i"]}]}], ",", "B"}], "]"}], 
            UnderscriptBox["-", "\[ScriptCapitalG]"], 
            RowBox[{
             UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
             RowBox[{
              RowBox[{"a", 
               UnderscriptBox["-", "\[ScriptCapitalG]"], 
               RowBox[{"mj", 
                UnderscriptBox["*", "\[ScriptCapitalG]"], 
                SubscriptBox["B", "j"]}]}], ",", "B"}], "]"}]}], ",", "P", 
           ",", "i", ",", "j"}], "]"}]}], 
        TagBox[")",
         "AutoParentheses"]}], "\[Implies]", 
       RowBox[{
        UnderscriptBox["cpConnectible", "\[ScriptCapitalG]"], "[", 
        RowBox[{
         SubscriptBox["B", "i"], ",", 
         SubscriptBox["B", "j"], ",", "a", ",", "k", ",", "l", ",", "B"}], 
        "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxTest 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1923946854", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648807900393601*^9, 3.648807954315282*^9}, {
   3.648807989781859*^9, 3.6488080884002028`*^9}, 3.648808347242332*^9, {
   3.6488085693788347`*^9, 3.648808619024288*^9}, {3.648808649860874*^9, 
   3.648808657342062*^9}, {3.648808818334157*^9, 3.648808846624042*^9}, {
   3.6488090472749767`*^9, 3.64880905692369*^9}, {3.6488091076800137`*^9, 
   3.648809107883355*^9}, {3.648809826433372*^9, 3.648809834703271*^9}, {
   3.648810456523892*^9, 3.648810511962919*^9}, {3.648811392386992*^9, 
   3.648811398471401*^9}, {3.6488787192736673`*^9, 3.6488788832768307`*^9}, {
   3.648878913420063*^9, 3.648878915895842*^9}, 3.648878951986597*^9, {
   3.648879119534898*^9, 3.6488791239294777`*^9}, {3.648885771010607*^9, 
   3.648885845197895*^9}, {3.648885961308158*^9, 3.6488859690755987`*^9}, {
   3.648886817157611*^9, 3.6488868326995363`*^9}, {3.648957654853435*^9, 
   3.648957655555709*^9}, 3.648957896702519*^9},
 CellTags->{
  "ID:1923946854", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxTest 1"},
 CellID->1923946854],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->857381505]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1607665888],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAuxTest alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAuxTest alternative)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["P", "ExpressionVariable"]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mi", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mj", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["GBAuxTest", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["mi", "ExpressionVariable"], ",", 
               StyleBox["mj", "ExpressionVariable"], ",", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["-", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 RowBox[{
                   UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}], ",", 
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["chainCrit", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["B", "ExpressionVariable"], ",", 
                   StyleBox["P", "ExpressionVariable"]}], "]"}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}], 
                 "\[Equal]", 
                 UnderscriptBox["0", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]]}]}], ")"}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["P", "ExpressionVariable"]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mi", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mj", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["GBAuxTest", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["mi", "ExpressionVariable"], ",", 
               StyleBox["mj", "ExpressionVariable"], ",", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["-", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                 RowBox[{
                   UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}], ",", 
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["chainCrit", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["B", "ExpressionVariable"], ",", 
                   StyleBox["P", "ExpressionVariable"]}], "]"}], "\[Or]", 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}], 
                 "\[Equal]", 
                 UnderscriptBox["0", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]]}]}], ")"}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1834517245"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1834517245-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1834517245-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1834517245", "Proof|ID:1834517245-1"},
 CellID->1216317993],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1834517245-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1362723759]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAuxTest 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAuxTest 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["P", "ExpressionVariable"]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mi", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mj", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["P", "ExpressionVariable"], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCPMultipliers", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["mi", "ExpressionVariable"], ",", 
                 StyleBox["mj", "ExpressionVariable"], ",", 
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], ",", 
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox["j", "ExpressionVariable"]], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["k", "ExpressionVariable"], ",", 
                 StyleBox["l", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["GBAuxTest", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["mi", "ExpressionVariable"], ",", 
                 StyleBox["mj", "ExpressionVariable"], ",", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}], ",", 
                 StyleBox["P", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["cpConnectible", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"], ",", 
               StyleBox["l", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             StyleBox["P", "ExpressionVariable"]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mi", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mj", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["P", "ExpressionVariable"], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCPMultipliers", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["mi", "ExpressionVariable"], ",", 
                 StyleBox["mj", "ExpressionVariable"], ",", 
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox["i", "ExpressionVariable"]], ",", 
                 SubscriptBox[
                  StyleBox["B", "ExpressionVariable"], 
                  StyleBox["j", "ExpressionVariable"]], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["k", "ExpressionVariable"], ",", 
                 StyleBox["l", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["GBAuxTest", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["mi", "ExpressionVariable"], ",", 
                 StyleBox["mj", "ExpressionVariable"], ",", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}], ",", 
                 StyleBox["P", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["cpConnectible", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"], ",", 
               StyleBox["l", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1923946854"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1923946854-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1923946854-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1923946854", "Proof|ID:1923946854-1"},
 CellID->1529476345],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1923946854-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1417488253]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6488087045099277`*^9, 3.648808705808342*^9}, 
   3.648810420982791*^9, 3.6532840329862313`*^9, 3.6661730156631975`*^9, 
   3.666439865175506*^9},
 CellID->2144907034],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
    RowBox[{"[", "B", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], ",", 
        "\[LessEqual]"}], "]"}], "]"}]], 
    RowBox[{"[", "P", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", ",", "mi", ",", "mj", ",", "k", ",", "l"}], TraditionalForm]]],
 " a.b.f., assume\n(1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTODOTuple", "[", 
    RowBox[{"P", ",", "B"}], "]"}], TraditionalForm]]],
 "\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "j"], ",", "a", ",", "k", ",", "l"}], "]"}], 
   TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"GBAuxTest", "[", 
    RowBox[{"B", ",", "a", ",", "mi", ",", "mj", ",", 
     RowBox[{"h1", "-", "h2"}], ",", "P", ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n(4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"MMPairs", "[", 
      RowBox[{"\[ScriptCapitalG]", ",", 
       SubscriptBox["B", "i"], ",", 
       SubscriptBox["B", "j"]}], "]"}]], 
    RowBox[{"[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"a", ",", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
      "\[RightAngleBracket]"}], "]"}]}], TraditionalForm]]],
 "\nand prove\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"cpConnectible", "[", 
    RowBox[{
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "j"], ",", "a", ",", "k", ",", "l", ",", "B"}], "]"}], 
   TraditionalForm]]],
 "\nwhere ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"h1", ":=", 
     RowBox[{"trd", "[", 
      RowBox[{
       RowBox[{"a", "-", 
        RowBox[{"mi", "*", 
         SubscriptBox["B", "i"]}]}], ",", "B"}], "]"}]}], ",", 
    RowBox[{"h2", ":=", 
     RowBox[{"trd", "[", 
      RowBox[{
       RowBox[{"a", "-", 
        RowBox[{"mj", "*", 
         SubscriptBox["B", "j"]}]}], ",", "B"}], "]"}]}]}], TraditionalForm]]],
 ".\n\nFrom (3), (GBAuxTest alternative)), we know\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"chainCrit", "[", 
      RowBox[{"i", ",", "j", ",", "a", ",", "B", ",", "P"}], "]"}], "\[Or]", 
     RowBox[{"h1", "-", "h2"}]}], "=", "0"}], TraditionalForm]]],
 "\nbased upon which we distinguish two cases.\n\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"chainCrit", "[", 
    RowBox[{"i", ",", "j", ",", "a", ",", "B", ",", "P"}], "]"}], 
   TraditionalForm]]],
 ".\nIn that case we can make use of (chainCrit 1), instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"B", ":=", "B"}], ",", 
    RowBox[{"P", ":=", "P"}], ",", 
    RowBox[{"i", ":=", "i"}], ",", 
    RowBox[{"j", ":=", "j"}], ",", 
    RowBox[{"a", ":=", "a"}], ",", 
    RowBox[{"k", ":=", "k"}], ",", 
    RowBox[{"l", ":=", "l"}]}], TraditionalForm]]],
 ". All premises are easily seen to be fulfilled, hence our goal follows \
readily.\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"h1", "-", "h2"}], "=", "0"}], TraditionalForm]]],
 ".\nProving (G), by the definition of cpConnectible, amounts to finding ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
    RowBox[{"[", 
     RowBox[{"m1", ",", "m2"}], "]"}]}], TraditionalForm]]],
 " with\n(G.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCPMultipliers", "[", 
    RowBox[{"m1", ",", "m2", ",", 
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "j"], ",", "a", ",", "k", ",", "l"}], "]"}], 
   TraditionalForm]]],
 "\n(G.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areConnectibleMultipliers", "[", 
    RowBox[{"m1", ",", "m2", ",", 
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "j"], ",", "a", ",", "B"}], "]"}], TraditionalForm]]],
 "\nWe claim that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"m1", ":=", "mi"}], ",", 
    RowBox[{"m2", ":=", "mj"}]}], TraditionalForm]]],
 " are suitable witnesses. In fact, (G.1) is equivalent to\n(G.1.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "j"], ",", "a", ",", "k", ",", "l"}], "]"}], 
   TraditionalForm]]],
 "\n(G.1.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areIrrelative", "[", 
    RowBox[{
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "j"], ",", "mi", ",", "mj", ",", "k", ",", "l"}], 
    "]"}], TraditionalForm]]],
 "\n(G.1.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     SubscriptBox["B", "i"], ",", "mi"}], "]"}], TraditionalForm]]],
 "\n(G.1.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isReducible", "[", 
    RowBox[{"a", ",", 
     SubscriptBox["B", "j"], ",", "mj"}], "]"}], TraditionalForm]]],
 "\nand all of the four subgoals follow readily from (2). Therefore, we only \
have to consider (G.2), which is equivalent to\n(G.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     RowBox[{"mi", ",", 
      SubscriptBox["B", "i"]}]], 
    RowBox[{"a", "-", 
     RowBox[{"mi", "*", 
      SubscriptBox["B", "i"]}]}]}], TraditionalForm]]],
 "\n(G.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     RowBox[{"mj", ",", 
      SubscriptBox["B", "j"]}]], 
    RowBox[{"a", "-", 
     RowBox[{"mj", "*", 
      SubscriptBox["B", "j"]}]}]}], TraditionalForm]]],
 "\n(G.5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "-", 
     RowBox[{"mi", "*", 
      SubscriptBox["B", "i"]}]}], 
    OverscriptBox[
     SubsuperscriptBox["\[LeftRightArrow]", "B", "*"], 
     RowBox[{"a", ",", "\[Succeeds]"}]], 
    RowBox[{"a", "-", 
     RowBox[{"mj", "*", 
      SubscriptBox["B", "j"]}]}]}], TraditionalForm]]],
 "\n(G.3) and (G.4) follow readily from the already-proved (G.1.3) and \
(G.1.4), respectively, hence we only have to consider (G.5); this, however, \
is already contained in a lemma (using our case assumption), so we are done."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.648808713091893*^9, 3.648809370451869*^9}, {
   3.6488094312517776`*^9, 3.648809477926857*^9}, {3.648809538441759*^9, 
   3.648809806930128*^9}, {3.6488098379234037`*^9, 3.6488100023536987`*^9}, {
   3.648810040741338*^9, 3.648810091230895*^9}, {3.648810143328574*^9, 
   3.6488102122415133`*^9}, {3.6488102423474197`*^9, 
   3.6488104209829187`*^9}, {3.6488105766469917`*^9, 3.648810576907086*^9}, {
   3.648879033219363*^9, 3.648879633808186*^9}, {3.648879907354638*^9, 
   3.648880179010604*^9}, {3.648880210087184*^9, 3.648880270198082*^9}, {
   3.6488804861687183`*^9, 3.648880498786293*^9}, {3.64888054978872*^9, 
   3.648880663186264*^9}, 3.6532840329862313`*^9, 3.6661730156631975`*^9, 
   3.666439865175652*^9},
 CellID->1186451214],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.648810403910111*^9, 3.648810420983054*^9}, 
   3.6532840329862313`*^9, 3.6661730156631975`*^9, 3.666439865175797*^9},
 TextAlignment->Right,
 CellID->1306340414]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["GBAuxCPConnectible1 (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.649049862765684*^9, 3.64904987208377*^9}},
 CellID->1918221565],

Cell["", "OpenEnvironment",
 CellID->1414827049],

Cell[CellGroupData[{

Cell["\<\
Lemma: GBAuxCPConnectible1 is Preserved in Every Recursive Call of GBAux\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648893322616929*^9, 3.6488933424651117`*^9}, {
  3.648965324456442*^9, 3.648965324604393*^9}, {3.6489738188893003`*^9, 
  3.6489738223947477`*^9}},
 CellID->1588362347],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"C", ",", "P", ",", "i", ",", "j"}]], 
  RowBox[{
   UnderscriptBox["GBAuxCPConnectible1", "\[ScriptCapitalG]"], "[", 
   RowBox[{"C", ",", "P", ",", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "i", ",", 
    "j"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible1 empty)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:258768016", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.6488936857271976`*^9},
 CellTags->{
  "ID:258768016", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 empty"},
 CellID->258768016],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
  RowBox[{"j", "\[GreaterEqual]", "i"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1002747542}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648893361683807*^9, 3.648893374767713*^9}},
 CellTags->{
  "ID:1002747542", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1002747542],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxRec3", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   UnderscriptBox["GBAuxCPConnectible1", "\[ScriptCapitalG]"], ",", "B", ",", 
   "i", ",", "j"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible1 GBAuxRec3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:673303920", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648893465238351*^9, 3.648893469223743*^9}},
 CellTags->{
  "ID:673303920", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 GBAuxRec3"},
 CellID->673303920],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxRec4", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   UnderscriptBox["GBAuxCPConnectible1", "\[ScriptCapitalG]"], ",", "B", ",", 
   "i", ",", "j"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible1 GBAuxRec4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1502806330", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648893466052404*^9, 3.648893470772069*^9}},
 CellTags->{
  "ID:1502806330", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 GBAuxRec4"},
 CellID->1502806330],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2052180156]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->865462022],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAuxCPConnectible1 empty)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAuxCPConnectible1 empty)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["C", "ExpressionVariable"]}, {
             StyleBox["P", "ExpressionVariable"]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           UnderscriptBox["GBAuxCPConnectible1", 
            RowBox[{"GroebnerRing", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           "[", 
           RowBox[{
             StyleBox["C", "ExpressionVariable"], ",", 
             StyleBox["P", "ExpressionVariable"], ",", 
             RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", 
             StyleBox["i", "ExpressionVariable"], ",", 
             StyleBox["j", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             StyleBox["C", "ExpressionVariable"]}, {
             StyleBox["P", "ExpressionVariable"]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           UnderscriptBox["GBAuxCPConnectible1", 
            RowBox[{"GroebnerRing", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           "[", 
           RowBox[{
             StyleBox["C", "ExpressionVariable"], ",", 
             StyleBox["P", "ExpressionVariable"], ",", 
             RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", 
             StyleBox["i", "ExpressionVariable"], ",", 
             StyleBox["j", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:258768016"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p258768016-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p258768016-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:258768016", "Proof|ID:258768016-1"},
 CellID->670832970],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p258768016-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.6489740074507217`*^9, 3.648974010233613*^9}},
 CellID->1269726962]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAuxCPConnectible1 GBAuxRec3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAuxCPConnectible1 GBAuxRec3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec3", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible1", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec3", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible1", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:673303920"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p673303920-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p673303920-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:673303920", "Proof|ID:673303920-1"},
 CellID->2063565608],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p673303920-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->30292895]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAuxCPConnectible1 GBAuxRec4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAuxCPConnectible1 GBAuxRec4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec4", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible1", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec4", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible1", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1502806330"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1502806330-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1502806330-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1502806330", "Proof|ID:1502806330-1"},
 CellID->1751898811],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1502806330-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->1643547098]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1047637483],

Cell[CellGroupData[{

Cell["Theorem: GBAuxCPConnectible1 is Valid", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6489795782583647`*^9, 3.648979609423831*^9}},
 CellID->130625381],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", 
     RowBox[{"NatPairs", "[", 
      RowBox[{
       RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[LessEqual]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}], "]"}], "]"}]], 
   RowBox[{"[", "P", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1133484500}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649068950467733*^9, 3.649068956664016*^9}},
 CellTags->{
  "ID:1133484500", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1133484500],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
  RowBox[{"j", "\[GreaterEqual]", "i"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 291397089}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649068965201282*^9, 3.649068978947176*^9}},
 CellTags->{
  "ID:291397089", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->291397089],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"MMPairs", "[", 
       RowBox[{"\[ScriptCapitalG]", ",", 
        SubscriptBox["B", "i"], ",", 
        SubscriptBox["B", "j"]}], "]"}], "]"}]], 
    RowBox[{"[", "M", "]"}]}]], 
  RowBox[{
   UnderscriptBox["GBAuxCPConnectible1", "\[ScriptCapitalG]"], "[", 
   RowBox[{"B", ",", "P", ",", "M", ",", "i", ",", "j"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible1 GBAux)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1494094918", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6489796284821167`*^9, 3.6489796426801367`*^9}, {
  3.64906895881315*^9, 3.6490689814290943`*^9}},
 CellTags->{
  "ID:1494094918", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 GBAux"},
 CellID->1494094918],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isTODOTuple", "\[ScriptCapitalG]"], "[", 
     RowBox[{"P", ",", "B", ",", "i", ",", "j"}], "]"}], "\[And]", 
    RowBox[{
     SubscriptBox["B", "i"], "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalG]"]}], "\[And]", 
    RowBox[{
     SubscriptBox["B", "j"], "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalG]"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["cpConnectible", "\[ScriptCapitalG]"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
     RowBox[{"B", ",", "P", ",", "i", ",", "j", ",", 
      RowBox[{
       UnderscriptBox["lmTuple", "\[ScriptCapitalG]"], "[", 
       RowBox[{
        SubscriptBox["B", "i"], ",", 
        SubscriptBox["B", "j"]}], "]"}]}], "]"}], ",", 
    SubscriptBox["B", "i"], ",", 
    SubscriptBox["B", "j"]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(cpConnectible GBAux lmTuple)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:402861438", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649068987764974*^9, 3.649069042417233*^9}, {
  3.649069500692317*^9, 3.649069517159972*^9}},
 CellTags->{
  "ID:402861438", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpConnectible GBAux lmTuple"},
 CellID->402861438],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->492085035]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1004237531],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAuxCPConnectible1 GBAux)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAuxCPConnectible1 GBAux)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
               "[", 
               StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxCPConnectible1", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["M", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
               "[", 
               StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxCPConnectible1", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["M", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1494094918"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1494094918-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1494094918-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1494094918", "Proof|ID:1494094918-1"},
 CellID->1479447805],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1494094918-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->879762366]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(cpConnectible GBAux lmTuple)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(cpConnectible GBAux lmTuple)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
               StyleBox["i", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["P", "ExpressionVariable"], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["cpConnectible", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["GBAux", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], ",", 
                   StyleBox["P", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"], ",", 
                   RowBox[{
                    UnderscriptBox["lmTuple", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}]}], "]"}], 
               ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
               StyleBox["i", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["P", "ExpressionVariable"], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["cpConnectible", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["GBAux", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], ",", 
                   StyleBox["P", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"], ",", 
                   RowBox[{
                    UnderscriptBox["lmTuple", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}]}], "]"}], 
               ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:402861438"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p402861438-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p402861438-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:402861438", "Proof|ID:402861438-1"},
 CellID->111398506],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p402861438-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->23314887]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["isCPTODOTuple (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.64904990545263*^9, 3.649049908948646*^9}, {
  3.649133521079775*^9, 3.649133525125955*^9}},
 CellID->910786473],

Cell["", "OpenEnvironment",
 CellID->1347382991],

Cell[CellGroupData[{

Cell["Lemma: Basic Properties of isCPTODOTuple", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.649049936720628*^9, 3.6490499472490177`*^9}},
 CellID->1658075726],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"i", ",", 
   RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
   "\[LeftBracketingBar]", "B", 
    "\[RightBracketingBar]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1173305577}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649049953383305*^9, 3.649049960864114*^9}},
 CellTags->{
  "ID:1173305577", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1173305577],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", "P"]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1545923689}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649049996029291*^9, 3.649049997171155*^9}},
 CellTags->{
  "ID:1545923689", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1545923689],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"processed", "[", 
     RowBox[{"i", ",", "j", ",", "P"}], "]"}], "\[And]", 
    RowBox[{
     UnderscriptBox["isCPTODOTuple", "\[ScriptCapitalG]"], "[", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "P", ",", 
      "B", ",", "i", ",", "j"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["cpConnectible", "\[ScriptCapitalG]"], "[", 
   RowBox[{"B", ",", 
    SubscriptBox["B", "i"], ",", 
    SubscriptBox["B", "j"]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isCPTODOTuple empty)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1684239249", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649050184773789*^9, 3.649050223928323*^9}, {
  3.649051507411686*^9, 3.64905150892496*^9}},
 CellTags->{
  "ID:1684239249", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple empty"},
 CellID->1684239249],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"k", ",", "l", ",", 
    RowBox[{"r", "..."}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isCPTODOTuple", "\[ScriptCapitalG]"], "[", 
     RowBox[{"P", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"r", "..."}], "\[RightAngleBracket]"}], ",", "B", ",", "i", 
      ",", "j"}], "]"}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["isCPTODOTuple", "\[ScriptCapitalG]"], "[", 
     RowBox[{"P", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"k", ",", "l"}], "\[RightAngleBracket]"}], ",", 
        RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], ",", "B", ",", "i", 
      ",", "j"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isCPTODOTuple prepend)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:547535692", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649050140957348*^9, 3.6490501670929737`*^9}, {
  3.649051003424913*^9, 3.64905100545575*^9}},
 CellTags->{
  "ID:547535692", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple prepend"},
 CellID->547535692],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"k", ",", "l", ",", 
    RowBox[{"r", "..."}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["isTODOTuple", "\[ScriptCapitalG]"], "[", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"k", ",", "l"}], "\[RightAngleBracket]"}], ",", 
          RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], ",", "B", ",", 
        "i", ",", "j"}], "]"}], "\[And]", 
      RowBox[{
       UnderscriptBox["isCPTODOTuple", "\[ScriptCapitalG]"], "[", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"k", ",", "l"}], "\[RightAngleBracket]"}], ",", 
          RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], ",", "B", ",", 
        "i", ",", "j"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["isTODOTuple", "\[ScriptCapitalG]"], "[", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"r", "..."}], "\[RightAngleBracket]"}], ",", "B", ",", "k", 
      ",", "l"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isCPTODOTuple isTODOTuple)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:812188873", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649050007749566*^9, 3.6490501043062773`*^9}},
 CellTags->{
  "ID:812188873", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple isTODOTuple"},
 CellID->812188873],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     SubscriptBox["B", "i"], "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalG]"]}], "\[And]", 
    RowBox[{
     SubscriptBox["B", "j"], "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalG]"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isCPTODOTuple", "\[ScriptCapitalG]"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["lmTuple", "\[ScriptCapitalG]"], "[", 
     RowBox[{
      SubscriptBox["B", "i"], ",", 
      SubscriptBox["B", "j"]}], "]"}], ",", "P", ",", "B", ",", "i", ",", 
    "j"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isCPTODOTuple lmTuple)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1105161129", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64905033811631*^9, 3.649050386272149*^9}},
 CellTags->{
  "ID:1105161129", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple lmTuple"},
 CellID->1105161129],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     SubscriptBox["B", "i"], "\[Equal]", 
     UnderscriptBox["0", "\[ScriptCapitalG]"]}], "\[Or]", 
    RowBox[{
     SubscriptBox["B", "j"], "\[Equal]", 
     UnderscriptBox["0", "\[ScriptCapitalG]"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", "M"], 
    RowBox[{
     UnderscriptBox["isCPTODOTuple", "\[ScriptCapitalG]"], "[", 
     RowBox[{"M", ",", "P", ",", "B", ",", "i", ",", "j"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isCPTODOTuple 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:981490890", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649050391251197*^9, 3.649050408189808*^9}, {
  3.649050451145409*^9, 3.6490504565862913`*^9}, {3.649051112527898*^9, 
  3.649051123427689*^9}},
 CellTags->{
  "ID:981490890", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple 0"},
 CellID->981490890],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1131566164]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->584510172],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isCPTODOTuple empty)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isCPTODOTuple empty)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             StyleBox["P", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"processed", "[", 
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"], ",", 
                 StyleBox["P", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isCPTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", 
                 StyleBox["P", "ExpressionVariable"], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["cpConnectible", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             StyleBox["P", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"processed", "[", 
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"], ",", 
                 StyleBox["P", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isCPTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", 
                 StyleBox["P", "ExpressionVariable"], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["cpConnectible", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], ",", 
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1684239249"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1684239249-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1684239249-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1684239249", "Proof|ID:1684239249-1"},
 CellID->74113726],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1684239249-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->623669078]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isCPTODOTuple prepend)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isCPTODOTuple prepend)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             StyleBox["P", "ExpressionVariable"]}, {
             StyleBox["k", "ExpressionVariable"]}, {
             StyleBox["l", "ExpressionVariable"]}, {
             StyleBox[
              RowBox[{"r", "..."}], "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isCPTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isCPTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             StyleBox["P", "ExpressionVariable"]}, {
             StyleBox["k", "ExpressionVariable"]}, {
             StyleBox["l", "ExpressionVariable"]}, {
             StyleBox[
              RowBox[{"r", "..."}], "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isCPTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isCPTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], ",", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:547535692"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p547535692-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p547535692-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:547535692", "Proof|ID:547535692-1"},
 CellID->1769495093],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p547535692-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->423907628]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isCPTODOTuple isTODOTuple)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isCPTODOTuple isTODOTuple)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             StyleBox["k", "ExpressionVariable"]}, {
             StyleBox["l", "ExpressionVariable"]}, {
             StyleBox[
              RowBox[{"r", "..."}], "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isCPTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{"\[LeftAngleBracket]", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"], ",", 
               StyleBox["l", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             StyleBox["k", "ExpressionVariable"]}, {
             StyleBox["l", "ExpressionVariable"]}, {
             StyleBox[
              RowBox[{"r", "..."}], "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["isTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isCPTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{"\[LeftAngleBracket]", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["k", "ExpressionVariable"], ",", 
               StyleBox["l", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:812188873"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p812188873-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p812188873-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:812188873", "Proof|ID:812188873-1"},
 CellID->1479812468],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p812188873-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->759303520]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isCPTODOTuple lmTuple)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isCPTODOTuple lmTuple)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             StyleBox["P", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isCPTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lmTuple", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                   SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], ",", 
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             StyleBox["P", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isCPTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lmTuple", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                   SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], ",", 
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1105161129"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1105161129-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1105161129-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1105161129", "Proof|ID:1105161129-1"},
 CellID->2103372990],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1105161129-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1126448357]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isCPTODOTuple 0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isCPTODOTuple 0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             StyleBox["P", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], "\[Equal]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[Or]", 
             RowBox[{
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]], "\[Equal]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              RowBox[{
                StyleBox["M", "ExpressionVariable"]}]], 
             RowBox[{
               UnderscriptBox["isCPTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["M", "ExpressionVariable"], ",", 
                 StyleBox["P", "ExpressionVariable"], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             StyleBox["P", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["i", "ExpressionVariable"]], "\[Equal]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}], "\[Or]", 
             RowBox[{
               SubscriptBox[
                StyleBox["B", "ExpressionVariable"], 
                StyleBox["j", "ExpressionVariable"]], "\[Equal]", 
               UnderscriptBox["0", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                  "]"}]]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              RowBox[{
                StyleBox["M", "ExpressionVariable"]}]], 
             RowBox[{
               UnderscriptBox["isCPTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["M", "ExpressionVariable"], ",", 
                 StyleBox["P", "ExpressionVariable"], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:981490890"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p981490890-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p981490890-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:981490890", "Proof|ID:981490890-1"},
 CellID->296074352],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p981490890-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1927354151]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->584142620],

Cell[CellGroupData[{

Cell["Theorem: isCPTODOTuple", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.649050654898967*^9, 3.6490506622411003`*^9}, 
   3.6491321668679934`*^9},
 CellID->2080239230],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", 
     RowBox[{"NatPairs", "[", 
      RowBox[{
       RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[LessEqual]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}], "]"}], "]"}]], 
   RowBox[{"[", "P", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 681053692}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6491273228612547`*^9, 3.64912733427265*^9}, {
  3.649127410184004*^9, 3.649127429279245*^9}, {3.649127615675201*^9, 
  3.6491276375021067`*^9}},
 CellTags->{
  "ID:681053692", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->681053692],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"i", ",", 
   RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{
   "\[LeftBracketingBar]", "B", 
    "\[RightBracketingBar]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 995853659}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649127337840259*^9, 3.649127345755106*^9}},
 CellTags->{
  "ID:995853659", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->995853659],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "mi", ",", "mj"}], "]"}]}], ",", 
   RowBox[{"r", "..."}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 218025475}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6491274323804283`*^9, 3.649127471552177*^9}, {
  3.649129161256913*^9, 3.649129171781845*^9}},
 CellTags->{
  "ID:218025475", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->218025475],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isCPTODOTuple", "\[ScriptCapitalG]"], "[", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"a", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
       "\[RightAngleBracket]"}], ",", 
      RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], ",", "P", ",", "B", 
    ",", "i", ",", "j"}], "]"}], "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 2052980984}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649127349383751*^9, 3.64912737796692*^9}, {
  3.6491274637700443`*^9, 3.649127491906928*^9}, {3.649131303560721*^9, 
  3.649131308514442*^9}},
 CellTags->{
  "ID:2052980984", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->2052980984],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"h", "=", 
   RowBox[{
    RowBox[{
     UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
     RowBox[{
      RowBox[{"a", 
       UnderscriptBox["-", "\[ScriptCapitalG]"], 
       RowBox[{"mi", 
        UnderscriptBox["*", "\[ScriptCapitalG]"], 
        SubscriptBox["B", "i"]}]}], ",", "B"}], "]"}], 
    UnderscriptBox["-", "\[ScriptCapitalG]"], 
    RowBox[{
     UnderscriptBox["trd", "\[ScriptCapitalG]"], "[", 
     RowBox[{
      RowBox[{"a", 
       UnderscriptBox["-", "\[ScriptCapitalG]"], 
       RowBox[{"mj", 
        UnderscriptBox["*", "\[ScriptCapitalG]"], 
        SubscriptBox["B", "j"]}]}], ",", "B"}], 
     "]"}]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 925810378}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649127527688979*^9, 3.649127571573359*^9}},
 CellTags->{
  "ID:925810378", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->925810378],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isTODOTuple", "\[ScriptCapitalG]"], "[", 
     RowBox[{"P", ",", "B", ",", "i", ",", "j"}], "]"}], "\[And]", 
    RowBox[{
     UnderscriptBox["GBAuxTest", "\[ScriptCapitalG]"], "[", 
     RowBox[{
     "B", ",", "a", ",", "mi", ",", "mj", ",", "h", ",", "P", ",", "i", ",", 
      "j"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["isCPTODOTuple", "\[ScriptCapitalG]"], "[", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"r", "..."}], "\[RightAngleBracket]"}], ",", "P", ",", "B", ",", 
    "i", ",", "j"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isCPTODOTuple 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:423392620", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649127499523368*^9, 3.649127514474915*^9}, {
  3.649127573851403*^9, 3.649127596414538*^9}, {3.649131296831114*^9, 
  3.649131305405117*^9}},
 CellTags->{
  "ID:423392620", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple 1"},
 CellID->423392620],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["isCPTODOTuple", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"r", "..."}], "\[RightAngleBracket]"}], ",", 
   RowBox[{"update", "[", 
    RowBox[{"P", ",", 
     RowBox[{
      RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], "+", 
      "1"}]}], "]"}], ",", 
   RowBox[{"B", 
    TagBox["\:293a",
     Identity,
     SyntaxForm->"a*b"], "h"}], ",", "i", ",", "j"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isCPTODOTuple 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1533257011", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649127604174775*^9, 3.649127609976733*^9}, {
  3.6491276409749613`*^9, 3.649127651182201*^9}},
 CellTags->{
  "ID:1533257011", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple 2"},
 CellID->1533257011],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->434593514]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->387312174],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isCPTODOTuple 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isCPTODOTuple 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mi", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mj", "ExpressionVariable"], "]"}]}, {
             StyleBox[
              RowBox[{"r", "..."}], "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isCPTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["isTODOTuple", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], ",", 
                   StyleBox["B", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["GBAuxTest", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["mi", "ExpressionVariable"], ",", 
                   StyleBox["mj", "ExpressionVariable"], ",", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}], ",", 
                   StyleBox["P", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], "]"}]}], 
             "\[Implies]", 
             RowBox[{
               UnderscriptBox["isCPTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox["P", "ExpressionVariable"], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mi", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mj", "ExpressionVariable"], "]"}]}, {
             StyleBox[
              RowBox[{"r", "..."}], "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isCPTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["isTODOTuple", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], ",", 
                   StyleBox["B", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["GBAuxTest", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], ",", 
                   StyleBox["a", "ExpressionVariable"], ",", 
                   StyleBox["mi", "ExpressionVariable"], ",", 
                   StyleBox["mj", "ExpressionVariable"], ",", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}], ",", 
                   StyleBox["P", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], "]"}]}], 
             "\[Implies]", 
             RowBox[{
               UnderscriptBox["isCPTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox["P", "ExpressionVariable"], ",", 
                 StyleBox["B", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:423392620"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p423392620-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p423392620-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:423392620", "Proof|ID:423392620-1"},
 CellID->117206802],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p423392620-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->6506205]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isCPTODOTuple 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isCPTODOTuple 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mi", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mj", "ExpressionVariable"], "]"}]}, {
             StyleBox[
              RowBox[{"r", "..."}], "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isCPTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isCPTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{"\[LeftAngleBracket]", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"], 
                 "\[RightAngleBracket]"}], ",", 
               RowBox[{"update", "[", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], ",", 
                   RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]}], "]"}], ",", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], "\:293a", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}]}], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mi", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["mj", "ExpressionVariable"], "]"}]}, {
             StyleBox[
              RowBox[{"r", "..."}], "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isCPTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], ",", 
                    StyleBox["mj", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], ",", 
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["isCPTODOTuple", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               RowBox[{"\[LeftAngleBracket]", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"], 
                 "\[RightAngleBracket]"}], ",", 
               RowBox[{"update", "[", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], ",", 
                   RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]}], "]"}], ",", 
               RowBox[{
                 StyleBox["B", "ExpressionVariable"], "\:293a", 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mi", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                   RowBox[{
                    UnderscriptBox["trd", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    RowBox[{
                    StyleBox["mj", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}]}], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "]"}]}]}], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1533257011"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1533257011-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1533257011-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1533257011", "Proof|ID:1533257011-1"},
 CellID->141690382],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1533257011-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->49123522]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Proof of (isCPTODOTuple 2)",
  FontSlant->"Italic"],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.649127733968898*^9, 3.649127740990582*^9}, 
   3.649129154800241*^9, 3.6491296791195583`*^9, 3.653284023439015*^9, 
   3.666173002449974*^9, 3.666439855534905*^9},
 CellID->1556779726],

Cell[TextData[{
 "Choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
    RowBox[{"[", "B", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], ",", 
        "\[LessEqual]"}], "]"}], "]"}]], 
    RowBox[{"[", "P", "]"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
     RowBox[{"[", 
      RowBox[{"a", ",", "mi", ",", "mj"}], "]"}]}], ",", 
    RowBox[{"r", "..."}]}], TraditionalForm]]],
 " a.b.f., assume\n(2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isTODOTuple", "[", 
    RowBox[{"P", ",", "B", ",", "i", ",", "j"}], "]"}], TraditionalForm]]],
 "\n(3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isCPTOTOTuple", "[", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"a", ",", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
        "\[RightAngleBracket]"}], ",", 
       RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], ",", "P", ",", "B", 
     ",", "i", ",", "j"}], "]"}], TraditionalForm]]],
 "\nand prove\n(G)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"isCPTODOTuple", "[", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"r", "..."}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"update", "[", 
      RowBox[{"P", ",", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], "+", 
        "1"}]}], "]"}], ",", 
     RowBox[{"B", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "h"}], ",", "i", ",", "j"}], "]"}], 
   TraditionalForm]]],
 "\n\nWe expand the definition of isCPTODOTuple: We know\n\
(3\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"\[Not]", 
      RowBox[{"processed", "[", 
       RowBox[{"i", ",", "j", ",", "P"}], "]"}]}], "\[Or]", 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{"k", ",", 
         RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"multN", "[", 
          SubscriptBox["B", "i"], "]"}]}]], 
       RowBox[{
        SubscriptBox["B", "i"], "=", 
        RowBox[{
         SubscriptBox["B", "j"], "\[Implies]", 
         RowBox[{"l", ">", "k"}]}]}]], 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{"isMNTCR", "[", 
         RowBox[{
          SubscriptBox["B", "i"], ",", 
          SubscriptBox["B", "j"], ",", "a", ",", "i", ",", "j"}], "]"}]], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[Not]", 
          RowBox[{
           UnderscriptBox["\[Exists]", 
            RowBox[{"m1", ",", "m2"}]], 
           RowBox[{
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"a", ",", 
              RowBox[{"\[LeftAngleBracket]", 
               RowBox[{"m1", ",", "m2"}], "\[RightAngleBracket]"}]}], 
             "\[RightAngleBracket]"}], 
            TagBox["\:22ff",
             Identity,
             SyntaxForm->"a+b"], 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
              RowBox[{"\[LeftAngleBracket]", 
               RowBox[{"a", ",", 
                RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
               "\[RightAngleBracket]"}], ",", 
              RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}]}]}]}], 
         "\[And]", 
         RowBox[{"areCPMultipliers", "[", 
          RowBox[{"m1", ",", "m2", ",", 
           SubscriptBox["B", "i"], ",", 
           SubscriptBox["B", "j"], ",", "a", ",", "k", ",", "l"}], "]"}]}], 
        ")"}]}]}]}], "\[Implies]", 
    RowBox[{"cpConnectible", "[", 
     RowBox[{
      SubscriptBox["B", "i"], ",", 
      SubscriptBox["B", "j"], ",", "a", ",", "k", ",", "l", ",", "B"}], 
     "]"}]}], TraditionalForm]]],
 "\nand have to show\n(G\[CloseCurlyQuote])\n\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"\[Not]", 
      RowBox[{"processed", "[", 
       RowBox[{"i", ",", "j", ",", 
        RowBox[{"update", "[", 
         RowBox[{"P", ",", 
          RowBox[{
           RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], 
           "+", "1"}]}], "]"}]}], "]"}]}], "\[Or]", 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[ForAll]", 
        RowBox[{"k", ",", 
         RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
         RowBox[{"multN", "[", 
          SubscriptBox[
           RowBox[{"(", 
            RowBox[{"B", 
             TagBox["\:293a",
              Identity,
              SyntaxForm->"a*b"], "h"}], ")"}], "i"], "]"}]}]], 
       RowBox[{
        SubscriptBox[
         RowBox[{"(", 
          RowBox[{"B", 
           TagBox["\:293a",
            Identity,
            SyntaxForm->"a*b"], "h"}], ")"}], "i"], "=", 
        RowBox[{
         SubscriptBox[
          RowBox[{"(", 
           RowBox[{"B", 
            TagBox["\:293a",
             Identity,
             SyntaxForm->"a*b"], "h"}], ")"}], "j"], "\[Implies]", 
         RowBox[{"l", ">", "k"}]}]}]], 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{"isMNTCR", "[", 
         RowBox[{
          SubscriptBox[
           RowBox[{"(", 
            RowBox[{"B", 
             TagBox["\:293a",
              Identity,
              SyntaxForm->"a*b"], "h"}], ")"}], "i"], ",", 
          SubscriptBox[
           RowBox[{"(", 
            RowBox[{"B", 
             TagBox["\:293a",
              Identity,
              SyntaxForm->"a*b"], "h"}], ")"}], "j"], ",", "a", ",", "i", ",",
           "j"}], "]"}]], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[Not]", 
          RowBox[{
           UnderscriptBox["\[Exists]", 
            RowBox[{"m1", ",", "m2"}]], 
           RowBox[{
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"a", ",", 
              RowBox[{"\[LeftAngleBracket]", 
               RowBox[{"m1", ",", "m2"}], "\[RightAngleBracket]"}]}], 
             "\[RightAngleBracket]"}], 
            TagBox["\:22ff",
             Identity,
             SyntaxForm->"a+b"], 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"r", "..."}], "\[RightAngleBracket]"}]}]}]}], "\[And]", 
         RowBox[{"areCPMultipliers", "[", 
          RowBox[{"m1", ",", "m2", ",", 
           SubscriptBox[
            RowBox[{"(", 
             RowBox[{"B", 
              TagBox["\:293a",
               Identity,
               SyntaxForm->"a*b"], "h"}], ")"}], "i"], ",", 
           SubscriptBox[
            RowBox[{"(", 
             RowBox[{"B", 
              TagBox["\:293a",
               Identity,
               SyntaxForm->"a*b"], "h"}], ")"}], "j"], ",", "a", ",", "k", 
           ",", "l"}], "]"}]}], ")"}]}]}]}], "\[Implies]", 
    RowBox[{"cpConnectible", "[", 
     RowBox[{
      SubscriptBox[
       RowBox[{"(", 
        RowBox[{"B", 
         TagBox["\:293a",
          Identity,
          SyntaxForm->"a*b"], "h"}], ")"}], "i"], ",", 
      SubscriptBox[
       RowBox[{"(", 
        RowBox[{"B", 
         TagBox["\:293a",
          Identity,
          SyntaxForm->"a*b"], "h"}], ")"}], "j"], ",", "a", ",", "k", ",", 
      "l", ",", 
      RowBox[{"B", 
       TagBox["\:293a",
        Identity,
        SyntaxForm->"a*b"], "h"}]}], "]"}]}], TraditionalForm]]],
 "\n\nBased on (3\[CloseCurlyQuote]) we distinguish two cases:\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"processed", "[", 
     RowBox[{"i", ",", "j", ",", "P"}], "]"}]}], TraditionalForm]]],
 ".\nWe show ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[Not]", 
    RowBox[{"processed", "[", 
     RowBox[{"i", ",", "j", ",", 
      RowBox[{"update", "[", 
       RowBox[{"P", ",", 
        RowBox[{
         RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], "+", 
         "1"}]}], "]"}]}], "]"}]}], TraditionalForm]]],
 ".\nOur case assumption is equivalent to\n(5)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"i", ",", "j"}], "\[RightAngleBracket]"}], "\:22ff", " ", "P"}],
     "\[Or]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"j", ",", "i"}], "\[RightAngleBracket]"}], "\:22ff", " ", 
     "P"}]}], TraditionalForm]]],
 "\nwhereas our goal is equivalent to\n(G.1)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"i", ",", "j"}], "\[RightAngleBracket]"}], "\:22ff", " ", 
     RowBox[{"update", "[", 
      RowBox[{"P", ",", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], "+", 
        "1"}]}], "]"}]}], "\[Or]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"j", ",", "i"}], "\[RightAngleBracket]"}], "\:22ff", " ", 
     RowBox[{"update", "[", 
      RowBox[{"P", ",", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], "+", 
        "1"}]}], "]"}]}]}], TraditionalForm]]],
 "\nBased on (5) we distinguish two cases; however, since both of them are \
completely analogous, we only consider the first one here: we assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"i", ",", "j"}], "\[RightAngleBracket]"}], "\:22ff", " ", "P"}], 
   TraditionalForm]]],
 " and show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"i", ",", "j"}], "\[RightAngleBracket]"}], "\:22ff", " ", 
    RowBox[{"update", "[", 
     RowBox[{"P", ",", 
      RowBox[{
       RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], "+", 
       "1"}]}], "]"}]}], TraditionalForm]]],
 ". This, however, is already contained in a lemma, hence we are done.\n\n\
Case II: Other disjunct.\nWe know\n(6)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{"k", ",", 
        RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"multN", "[", 
         SubscriptBox["B", "i"], "]"}]}]], 
      RowBox[{
       SubscriptBox["B", "i"], "=", 
       RowBox[{
        SubscriptBox["B", "j"], "\[Implies]", 
        RowBox[{"l", ">", "k"}]}]}]], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{"isMNTCR", "[", 
        RowBox[{
         SubscriptBox["B", "i"], ",", 
         SubscriptBox["B", "j"], ",", "a", ",", "i", ",", "j"}], "]"}]], 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"\[Not]", 
         RowBox[{
          UnderscriptBox["\[Exists]", 
           RowBox[{"m1", ",", "m2"}]], 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"a", ",", 
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"m1", ",", "m2"}], "\[RightAngleBracket]"}]}], 
            "\[RightAngleBracket]"}], 
           TagBox["\:22ff",
            Identity,
            SyntaxForm->"a+b"], 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"a", ",", 
               RowBox[{"\[LeftAngleBracket]", 
                RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
              "\[RightAngleBracket]"}], ",", 
             RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}]}]}]}], "\[And]", 
        RowBox[{"areCPMultipliers", "[", 
         RowBox[{"m1", ",", "m2", ",", 
          SubscriptBox["B", "i"], ",", 
          SubscriptBox["B", "j"], ",", "a", ",", "k", ",", "l"}], "]"}]}], 
       ")"}]}]}], "\[Implies]", 
    RowBox[{"cpConnectible", "[", 
     RowBox[{
      SubscriptBox["B", "i"], ",", 
      SubscriptBox["B", "j"], ",", "a", ",", "k", ",", "l", ",", "B"}], 
     "]"}]}], TraditionalForm]]],
 "\nand choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", ",", 
    RowBox[{"l", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"multN", "[", 
     SubscriptBox[
      RowBox[{"(", 
       RowBox[{"B", 
        TagBox["\:293a",
         Identity,
         SyntaxForm->"a*b"], "h"}], ")"}], "i"], "]"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"isMNTCR", "[", 
    RowBox[{
     SubscriptBox[
      RowBox[{"(", 
       RowBox[{"B", 
        TagBox["\:293a",
         Identity,
         SyntaxForm->"a*b"], "h"}], ")"}], "i"], ",", 
     SubscriptBox[
      RowBox[{"(", 
       RowBox[{"B", 
        TagBox["\:293a",
         Identity,
         SyntaxForm->"a*b"], "h"}], ")"}], "j"], ",", 
     SubscriptBox["a", "0"], ",", "k", ",", "l"}], "]"}], TraditionalForm]]],
 " a.b.f. with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox[
     RowBox[{"(", 
      RowBox[{"B", 
       TagBox["\:293a",
        Identity,
        SyntaxForm->"a*b"], "h"}], ")"}], "i"], "=", 
    RowBox[{
     SubscriptBox[
      RowBox[{"(", 
       RowBox[{"B", 
        TagBox["\:293a",
         Identity,
         SyntaxForm->"a*b"], "h"}], ")"}], "j"], "\[Implies]", 
     RowBox[{"l", ">", "k"}]}]}], TraditionalForm]]],
 ", assume\n(7)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{"m1", ",", "m2"}]], 
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         SubscriptBox["a", "0"], ",", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"m1", ",", "m2"}], "\[RightAngleBracket]"}]}], 
        "\[RightAngleBracket]"}], 
       TagBox["\:22ff",
        Identity,
        SyntaxForm->"a+b"], 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"r", "..."}], "\[RightAngleBracket]"}]}]}]}], "\[And]", 
    RowBox[{"areCPMultipliers", "[", 
     RowBox[{"m1", ",", "m2", ",", 
      SubscriptBox[
       RowBox[{"(", 
        RowBox[{"B", 
         TagBox["\:293a",
          Identity,
          SyntaxForm->"a*b"], "h"}], ")"}], "i"], ",", 
      SubscriptBox[
       RowBox[{"(", 
        RowBox[{"B", 
         TagBox["\:293a",
          Identity,
          SyntaxForm->"a*b"], "h"}], ")"}], "j"], ",", 
      SubscriptBox["a", "0"], ",", "k", ",", "l"}], "]"}]}], 
   TraditionalForm]]],
 "\nand show\n(G.2)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"cpConnectible", "[", 
    RowBox[{
     SubscriptBox[
      RowBox[{"(", 
       RowBox[{"B", 
        TagBox["\:293a",
         Identity,
         SyntaxForm->"a*b"], "h"}], ")"}], "i"], ",", 
     SubscriptBox[
      RowBox[{"(", 
       RowBox[{"B", 
        TagBox["\:293a",
         Identity,
         SyntaxForm->"a*b"], "h"}], ")"}], "j"], ",", 
     SubscriptBox["a", "0"], ",", "k", ",", "l", ",", 
     RowBox[{"B", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "h"}]}], "]"}], TraditionalForm]]],
 "\n\nFirst of all, since ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 ", we replace every occurrence of ",
 Cell[BoxData[
  FormBox[
   SubscriptBox[
    RowBox[{"(", 
     RowBox[{"B", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "h"}], ")"}], 
    RowBox[{"i", "|", "j"}]], TraditionalForm]]],
 " by ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["B", 
    RowBox[{"i", "|", "j"}]], TraditionalForm]]],
 ", i.e. we know\n(7\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[Not]", 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{"m1", ",", "m2"}]], 
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         SubscriptBox["a", "0"], ",", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"m1", ",", "m2"}], "\[RightAngleBracket]"}]}], 
        "\[RightAngleBracket]"}], 
       TagBox["\:22ff",
        Identity,
        SyntaxForm->"a+b"], 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"r", "..."}], "\[RightAngleBracket]"}]}]}]}], "\[And]", 
    RowBox[{"areCPMultipliers", "[", 
     RowBox[{"m1", ",", "m2", ",", 
      SubscriptBox["B", "i"], ",", 
      SubscriptBox["B", "j"], ",", 
      SubscriptBox["a", "0"], ",", "k", ",", "l"}], "]"}]}], 
   TraditionalForm]]],
 "\nand show\n(G.3)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"cpConnectible", "[", 
    RowBox[{
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "j"], ",", 
     SubscriptBox["a", "0"], ",", "k", ",", "l", ",", 
     RowBox[{"B", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "h"}]}], "]"}], TraditionalForm]]],
 "\n\nWe now instantiate (6) with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"k", ":=", "k"}], ",", 
    RowBox[{"l", ":=", "l"}], ",", 
    RowBox[{"a", ":=", 
     SubscriptBox["a", "0"]}]}], TraditionalForm]]],
 ". All premises are easily seen to hold, hence we obtain\n(8)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"\[Not]", 
       RowBox[{
        UnderscriptBox["\[Exists]", 
         RowBox[{"m1", ",", "m2"}]], 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           SubscriptBox["a", "0"], ",", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"m1", ",", "m2"}], "\[RightAngleBracket]"}]}], 
          "\[RightAngleBracket]"}], 
         TagBox["\:22ff",
          Identity,
          SyntaxForm->"a+b"], 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"a", ",", 
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
            "\[RightAngleBracket]"}], ",", 
           RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}]}]}]}], "\[And]", 
      RowBox[{"areCPMultipliers", "[", 
       RowBox[{"m1", ",", "m2", ",", 
        SubscriptBox["B", "i"], ",", 
        SubscriptBox["B", "j"], ",", 
        SubscriptBox["a", "0"], ",", "k", ",", "l"}], "]"}]}], ")"}], 
    "\[Implies]", 
    RowBox[{"cpConnectible", "[", 
     RowBox[{
      SubscriptBox["B", "i"], ",", 
      SubscriptBox["B", "j"], ",", 
      SubscriptBox["a", "0"], ",", "k", ",", "l", ",", "B"}], "]"}]}], 
   TraditionalForm]]],
 "\n\nIf the premise of (8) holds we are done immediately (using \
(cpConnectible append)), so we assume its negation, i.e. we assume there are ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m1", ",", "m2"}], TraditionalForm]]],
 " with\n(9)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      SubscriptBox["a", "0"], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"m1", ",", "m2"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], 
    TagBox["\:22ff",
     Identity,
     SyntaxForm->"a+b"], 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"a", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
       "\[RightAngleBracket]"}], ",", 
      RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}]}], TraditionalForm]]],
 "\n(10)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCPMultipliers", "[", 
    RowBox[{"m1", ",", "m2", ",", 
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "j"], ",", 
     SubscriptBox["a", "0"], ",", "k", ",", "l"}], "]"}], TraditionalForm]]],
 "\n\n(9) is equivalent to\n(9\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      SubscriptBox["a", "0"], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"m1", ",", "m2"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], "=", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"a", ",", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"mi", ",", "mj"}], "\[RightAngleBracket]"}]}], 
      "\[RightAngleBracket]"}], "\[Or]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        SubscriptBox["a", "0"], ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"m1", ",", "m2"}], "\[RightAngleBracket]"}]}], 
       "\[RightAngleBracket]"}], 
      TagBox["\:22ff",
       Identity,
       SyntaxForm->"a+b"], 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"r", "..."}], "\[RightAngleBracket]"}]}]}]}], 
   TraditionalForm]]],
 "\nThe second case of (9\[CloseCurlyQuote]) contradicts \
(7\[CloseCurlyQuote]), hence we may assume\n(11)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"m1", "=", "mi"}], TraditionalForm]]],
 "\n(12)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"m2", "=", "mj"}], TraditionalForm]]],
 "\n(13)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["a", "0"], "=", "a"}], TraditionalForm]]],
 "\nand, hence,\n(10\[CloseCurlyQuote])\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"areCPMultipliers", "[", 
    RowBox[{"mi", ",", "mj", ",", 
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "j"], ",", "a", ",", "k", ",", "l"}], "]"}], 
   TraditionalForm]]],
 "\nand we prove\n(G.4)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{"cpConnectible", "[", 
    RowBox[{
     SubscriptBox["B", "i"], ",", 
     SubscriptBox["B", "j"], ",", "a", ",", "k", ",", "l", ",", 
     RowBox[{"B", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "h"}]}], "]"}], TraditionalForm]]],
 "\n\nNow we use (cpConnectible append h), instantiated with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"B", ":=", "B"}], ",", 
    RowBox[{"i", ":=", "i"}], ",", 
    RowBox[{"j", ":=", "j"}], ",", 
    RowBox[{"a", ":=", "a"}], ",", 
    RowBox[{"mi", ":=", "mi"}], ",", 
    RowBox[{"k", ":=", "k"}], ",", 
    RowBox[{"l", ":=", "l"}]}], TraditionalForm]]],
 ", and obtain\n(14)\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"areCPMultipliers", "[", 
     RowBox[{"mi", ",", "mj", ",", 
      SubscriptBox["B", "i"], ",", 
      SubscriptBox["B", "j"], ",", "a", ",", "k", ",", "l"}], "]"}], 
    "\[Implies]", 
    RowBox[{"cpConnectible", "[", 
     RowBox[{
      SubscriptBox["B", "i"], ",", 
      SubscriptBox["B", "j"], ",", "a", ",", "k", ",", "l", ",", 
      RowBox[{"B", 
       TagBox["\:293a",
        Identity,
        SyntaxForm->"a*b"], "h"}]}], "]"}]}], TraditionalForm]]],
 ".\nThe premise of (14) holds due to (10\[CloseCurlyQuote]), so we are \
done."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.649127751618095*^9, 3.64912827906776*^9}, {
   3.649128336159978*^9, 3.649128426025847*^9}, 3.649128472464522*^9, {
   3.649128565224992*^9, 3.649128609201151*^9}, {3.649128653975984*^9, 
   3.6491287768968363`*^9}, {3.6491288099044027`*^9, 
   3.6491288725457363`*^9}, {3.649128921239829*^9, 3.649129154800371*^9}, {
   3.649129688108307*^9, 3.649129972960252*^9}, {3.649130022999176*^9, 
   3.64913010703096*^9}, {3.6491301698939447`*^9, 3.649130494167123*^9}, {
   3.6491305289750547`*^9, 3.6491307116344624`*^9}, {3.649130753961557*^9, 
   3.649130843626989*^9}, {3.6491308768754263`*^9, 3.649131002215599*^9}, 
   3.653284023439015*^9, 3.666173002449974*^9, 3.666439855535009*^9},
 CellID->1788850434],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.6491291381822443`*^9, 3.6491291548005*^9}, 
   3.653284023439015*^9, 3.666173002449974*^9, 3.66643985553511*^9},
 TextAlignment->Right,
 CellID->1397741689]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["GBAuxCPConnectible2 (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.649049889286924*^9, 3.649049894733399*^9}, {
  3.6491344836093197`*^9, 3.649134487667061*^9}},
 CellID->1336194338],

Cell["", "OpenEnvironment",
 CellID->43048684],

Cell[CellGroupData[{

Cell["\<\
Lemma: GBAuxCPConnectible2 is Preserved in Every Recursive Call of GBAux\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{3.648973827294092*^9, 3.649049884315959*^9},
 CellID->1822040650],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
  RowBox[{"j", "\[GreaterEqual]", "i"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1154842295}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648973832890819*^9, 3.6489738542324467`*^9}},
 CellTags->{
  "ID:1154842295", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1154842295],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxBase", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   UnderscriptBox["GBAuxCPConnectible2", "\[ScriptCapitalG]"], ",", "B", ",", 
   "i", ",", "j"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible2 GBAuxBase)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1427865659", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648973888911702*^9, 3.648973892492886*^9}},
 CellTags->{
  "ID:1427865659", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxBase"},
 CellID->1427865659],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxRec1", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   UnderscriptBox["GBAuxCPConnectible2", "\[ScriptCapitalG]"], ",", "B", ",", 
   "i", ",", "j"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible2 GBAuxRec1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:802696830", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6489738880350847`*^9, 3.648973894887519*^9}},
 CellTags->{
  "ID:802696830", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec1"},
 CellID->802696830],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxRec2", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   UnderscriptBox["GBAuxCPConnectible2", "\[ScriptCapitalG]"], ",", "B", ",", 
   "i", ",", "j"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible2 GBAuxRec2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:668126868", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6489738872505198`*^9, 3.64897389628764*^9}},
 CellTags->{
  "ID:668126868", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec2"},
 CellID->668126868],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxRec3", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   UnderscriptBox["GBAuxCPConnectible2", "\[ScriptCapitalG]"], ",", "B", ",", 
   "i", ",", "j"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible2 GBAuxRec3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1848124281", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648973861563809*^9, 3.648973865763729*^9}},
 CellTags->{
  "ID:1848124281", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec3"},
 CellID->1848124281],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxRec4", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   UnderscriptBox["GBAuxCPConnectible2", "\[ScriptCapitalG]"], ",", "B", ",", 
   "i", ",", "j"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible2 GBAuxRec4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:6143564", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6489738861383533`*^9, 3.648973898313078*^9}},
 CellTags->{
  "ID:6143564", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec4"},
 CellID->6143564],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1511341304]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1991967224],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAuxCPConnectible2 GBAuxBase)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAuxCPConnectible2 GBAuxBase)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxBase", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible2", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxBase", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible2", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1427865659"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1427865659-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1427865659-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1427865659", "Proof|ID:1427865659-1"},
 CellID->544375535],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1427865659-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->473240177]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAuxCPConnectible2 GBAuxRec1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAuxCPConnectible2 GBAuxRec1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec1", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible2", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec1", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible2", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:802696830"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p802696830-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p802696830-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:802696830", "Proof|ID:802696830-1"},
 CellID->335517243],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p802696830-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->108528534]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAuxCPConnectible2 GBAuxRec2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAuxCPConnectible2 GBAuxRec2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec2", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible2", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec2", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible2", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:668126868"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p668126868-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p668126868-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:668126868", "Proof|ID:668126868-1"},
 CellID->705494055],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p668126868-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1531879012]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAuxCPConnectible2 GBAuxRec3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAuxCPConnectible2 GBAuxRec3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec3", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible2", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec3", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible2", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1848124281"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1848124281-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1848124281-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1848124281", "Proof|ID:1848124281-1"},
 CellID->885145762],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1848124281-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1378460394]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAuxCPConnectible2 GBAuxRec4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAuxCPConnectible2 GBAuxRec4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec4", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible2", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec4", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible2", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:6143564"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p6143564-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p6143564-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:6143564", "Proof|ID:6143564-1"},
 CellID->1629480533],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p6143564-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1662275265]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->1477099468],

Cell[CellGroupData[{

Cell["Theorem: GBAuxCPConnectible2 is Valid", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.649134019154538*^9, 3.6491340274186163`*^9}},
 CellID->1770683854],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", 
     RowBox[{"NatPairs", "[", 
      RowBox[{
       RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], ",", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        TagBox["\[LessEqual]",
         Identity], 
        TagBox[")",
         "AutoParentheses"]}]}], "]"}], "]"}]], 
   RowBox[{"[", "P", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1620493315}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.649151420524907*^9},
 CellTags->{
  "ID:1620493315", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1620493315],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
  RowBox[{"j", "\[GreaterEqual]", "i"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1672348970}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.649151430466724*^9},
 CellTags->{
  "ID:1672348970", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1672348970],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", 
     RowBox[{"MMPairs", "[", 
      RowBox[{"\[ScriptCapitalG]", ",", 
       SubscriptBox["B", "i"], ",", 
       SubscriptBox["B", "j"]}], "]"}], "]"}]], 
   RowBox[{"[", "M", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1826259623}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649151434056748*^9, 3.649151439821035*^9}},
 CellTags->{
  "ID:1826259623", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1826259623],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxCPConnectible2", "\[ScriptCapitalG]"], "[", 
  RowBox[{"B", ",", "P", ",", "M", ",", "i", ",", "j"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible2 GBAux)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1814594394", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649134044985539*^9, 3.6491340934050493`*^9}, 
   3.649134296950012*^9, {3.6491514187979527`*^9, 3.6491514381286707`*^9}},
 CellTags->{
  "ID:1814594394", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAux"},
 CellID->1814594394],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalG]"], 
   RowBox[{"[", "h", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1024326782}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649151482726398*^9, 3.649151490154344*^9}},
 CellTags->{
  "ID:1024326782", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1024326782],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"G", "=", 
   RowBox[{
    UnderscriptBox["GBAux", "\[ScriptCapitalG]"], "[", 
    RowBox[{
     RowBox[{"B", 
      TagBox["\:293a",
       Identity,
       SyntaxForm->"a*b"], "h"}], ",", 
     RowBox[{"update", "[", 
      RowBox[{"P", ",", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], "+", 
        "1"}]}], "]"}], ",", "i", ",", "j", ",", "M"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 585977074}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649151466608324*^9, 3.649151520208201*^9}},
 CellTags->{
  "ID:585977074", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->585977074],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isTODOTuple", "\[ScriptCapitalG]"], "[", 
     RowBox[{
      RowBox[{"update", "[", 
       RowBox[{"P", ",", 
        RowBox[{
         RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], "+", 
         "1"}]}], "]"}], ",", 
      RowBox[{"B", 
       TagBox["\:293a",
        Identity,
        SyntaxForm->"a*b"], "h"}], ",", "i", ",", "j"}], "]"}], "\[And]", 
    RowBox[{
     UnderscriptBox["isCPTODOTuple", "\[ScriptCapitalG]"], "[", 
     RowBox[{"M", ",", 
      RowBox[{"update", "[", 
       RowBox[{"P", ",", 
        RowBox[{
         RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], "+", 
         "1"}]}], "]"}], ",", 
      RowBox[{"B", 
       TagBox["\:293a",
        Identity,
        SyntaxForm->"a*b"], "h"}], ",", "i", ",", "j"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"i0", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{
      RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], "+", 
      "1"}]}]], 
   RowBox[{
    UnderscriptBox["cpConnectible", "\[ScriptCapitalG]"], "[", 
    RowBox[{"G", ",", 
     SubscriptBox["G", "i0"], ",", 
     SubscriptBox["G", 
      RowBox[{
       RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], "+", 
       "1"}]]}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible2 update)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1959330209", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649151538597663*^9, 3.6491516432341127`*^9}},
 CellTags->{
  "ID:1959330209", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 update"},
 CellID->1959330209],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1750384990]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->885022220],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAuxCPConnectible2 GBAux)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAuxCPConnectible2 GBAux)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
               "[", 
               StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxCPConnectible2", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["M", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
               "[", 
               StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxCPConnectible2", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["M", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1814594394"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1814594394-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1814594394-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1814594394", "Proof|ID:1814594394-1"},
 CellID->378539769],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1814594394-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->287249254]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAuxCPConnectible2 update)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAuxCPConnectible2 update)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
               "[", 
               StyleBox["M", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["h", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
               StyleBox["i", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 RowBox[{"update", "[", 
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]}], "]"}], ",", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "\:293a", 
                   StyleBox["h", "ExpressionVariable"]}], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isCPTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["M", "ExpressionVariable"], ",", 
                 RowBox[{"update", "[", 
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]}], "]"}], ",", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "\:293a", 
                   StyleBox["h", "ExpressionVariable"]}], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i0", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]}]}}]], 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\:293a", 
                    StyleBox["h", "ExpressionVariable"]}], ",", 
                    RowBox[{"update", "[", 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]}], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"], ",", 
                    StyleBox["M", "ExpressionVariable"]}], "]"}], ",", 
                 SubscriptBox[
                  RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\:293a", 
                    StyleBox["h", "ExpressionVariable"]}], ",", 
                    RowBox[{"update", "[", 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]}], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"], ",", 
                    StyleBox["M", "ExpressionVariable"]}], "]"}], 
                  StyleBox["i0", "ExpressionVariable"]], ",", 
                 SubscriptBox[
                  RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\:293a", 
                    StyleBox["h", "ExpressionVariable"]}], ",", 
                    RowBox[{"update", "[", 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]}], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"], ",", 
                    StyleBox["M", "ExpressionVariable"]}], "]"}], 
                  RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
               "[", 
               StyleBox["M", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["h", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
               StyleBox["i", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["isTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 RowBox[{"update", "[", 
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]}], "]"}], ",", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "\:293a", 
                   StyleBox["h", "ExpressionVariable"]}], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["isCPTODOTuple", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 StyleBox["M", "ExpressionVariable"], ",", 
                 RowBox[{"update", "[", 
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]}], "]"}], ",", 
                 RowBox[{
                   StyleBox["B", "ExpressionVariable"], "\:293a", 
                   StyleBox["h", "ExpressionVariable"]}], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i0", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]}]}}]], 
             RowBox[{
               UnderscriptBox["cpConnectible", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\:293a", 
                    StyleBox["h", "ExpressionVariable"]}], ",", 
                    RowBox[{"update", "[", 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]}], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"], ",", 
                    StyleBox["M", "ExpressionVariable"]}], "]"}], ",", 
                 SubscriptBox[
                  RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\:293a", 
                    StyleBox["h", "ExpressionVariable"]}], ",", 
                    RowBox[{"update", "[", 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]}], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"], ",", 
                    StyleBox["M", "ExpressionVariable"]}], "]"}], 
                  StyleBox["i0", "ExpressionVariable"]], ",", 
                 SubscriptBox[
                  RowBox[{
                    UnderscriptBox["GBAux", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\:293a", 
                    StyleBox["h", "ExpressionVariable"]}], ",", 
                    RowBox[{"update", "[", 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]}], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"], ",", 
                    StyleBox["M", "ExpressionVariable"]}], "]"}], 
                  RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", "1"}]]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1959330209"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1959330209-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1959330209-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1959330209", "Proof|ID:1959330209-1"},
 CellID->1849443430],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1959330209-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->59470111]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["GBAuxCPConnectible3 (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.649146651073*^9, 3.649146657104793*^9}, {
  3.649388389052167*^9, 3.6493883931167927`*^9}},
 CellID->1419927671],

Cell["", "OpenEnvironment",
 CellID->843615990],

Cell[CellGroupData[{

Cell["\<\
Lemma: GBAuxCPConnectible3 is Preserved in Every Recursive Call of GBAux\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.649146671978734*^9, 3.649146674079792*^9}},
 CellID->2024040760],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{"i", ",", 
    RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
  RowBox[{"j", "\[GreaterEqual]", "i"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 33155789}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648973832890819*^9, 3.6489738542324467`*^9}},
 CellTags->{
  "ID:33155789", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->33155789],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxBase", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   UnderscriptBox["GBAuxCPConnectible3", "\[ScriptCapitalG]"], ",", "B", ",", 
   "i", ",", "j"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible3 GBAuxBase)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1691022439", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649146693731202*^9, 3.649146696053953*^9}},
 CellTags->{
  "ID:1691022439", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxBase"},
 CellID->1691022439],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxRec1", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   UnderscriptBox["GBAuxCPConnectible3", "\[ScriptCapitalG]"], ",", "B", ",", 
   "i", ",", "j"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible3 GBAuxRec1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:544907973", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649146701616212*^9, 3.649146707693144*^9}},
 CellTags->{
  "ID:544907973", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec1"},
 CellID->544907973],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxRec2", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   UnderscriptBox["GBAuxCPConnectible3", "\[ScriptCapitalG]"], ",", "B", ",", 
   "i", ",", "j"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible3 GBAuxRec2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1039424559", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649146710244626*^9, 3.64914671398155*^9}},
 CellTags->{
  "ID:1039424559", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec2"},
 CellID->1039424559],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxRec3", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   UnderscriptBox["GBAuxCPConnectible3", "\[ScriptCapitalG]"], ",", "B", ",", 
   "i", ",", "j"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible3 GBAuxRec3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1704484269", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649146711279146*^9, 3.649146715749052*^9}},
 CellTags->{
  "ID:1704484269", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec3"},
 CellID->1704484269],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["GBAuxRec4", "\[ScriptCapitalG]"], "[", 
  RowBox[{
   UnderscriptBox["GBAuxCPConnectible3", "\[ScriptCapitalG]"], ",", "B", ",", 
   "i", ",", "j"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible3 GBAuxRec4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2114764931", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649146712063108*^9, 3.649146717370462*^9}},
 CellTags->{
  "ID:2114764931", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec4"},
 CellID->2114764931],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->2000303424]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->187817550],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAuxCPConnectible3 GBAuxBase)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAuxCPConnectible3 GBAuxBase)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxBase", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible3", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxBase", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible3", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1691022439"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1691022439-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1691022439-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1691022439", "Proof|ID:1691022439-1"},
 CellID->1214006285],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1691022439-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1934036983]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAuxCPConnectible3 GBAuxRec1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAuxCPConnectible3 GBAuxRec1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec1", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible3", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec1", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible3", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:544907973"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p544907973-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p544907973-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:544907973", "Proof|ID:544907973-1"},
 CellID->708615290],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p544907973-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{3.649149033390418*^9},
 CellID->2061640519]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAuxCPConnectible3 GBAuxRec2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAuxCPConnectible3 GBAuxRec2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec2", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible3", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec2", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible3", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1039424559"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1039424559-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1039424559-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1039424559", "Proof|ID:1039424559-1"},
 CellID->495379081],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1039424559-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->725364984]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAuxCPConnectible3 GBAuxRec3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAuxCPConnectible3 GBAuxRec3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec3", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible3", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec3", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible3", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1704484269"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1704484269-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1704484269-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1704484269", "Proof|ID:1704484269-1"},
 CellID->122315265],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1704484269-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1471508628]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAuxCPConnectible3 GBAuxRec4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAuxCPConnectible3 GBAuxRec4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec4", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible3", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxRec4", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GBAuxCPConnectible3", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2114764931"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2114764931-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2114764931-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2114764931", "Proof|ID:2114764931-1"},
 CellID->1252700938],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2114764931-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->1737567296]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->822190239],

Cell[CellGroupData[{

Cell["Theorem: GBAuxCPConnectible3 is Valid", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.649388112903565*^9, 3.6493881200366592`*^9}},
 CellID->1425736080],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", 
      RowBox[{"NatPairs", "[", 
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}], ",", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         TagBox["\[LessEqual]",
          Identity], 
         TagBox[")",
          "AutoParentheses"]}]}], "]"}], "]"}]], 
    RowBox[{"[", "P", "]"}]}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{"i", ",", 
      RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{"\[LeftBracketingBar]", "B", "\[RightBracketingBar]"}]}]], 
    RowBox[{"j", "\[GreaterEqual]", "i"}]], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", 
        RowBox[{"MMPairs", "[", 
         RowBox[{"\[ScriptCapitalG]", ",", 
          SubscriptBox["B", "i"], ",", 
          SubscriptBox["B", "j"]}], "]"}], "]"}]], 
      RowBox[{"[", "M", "]"}]}]], 
    RowBox[{
     UnderscriptBox["GBAuxCPConnectible3", "\[ScriptCapitalG]"], "[", 
     RowBox[{"B", ",", "P", ",", "M", ",", "i", ",", "j"}], 
     "]"}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GBAuxCPConnectible3 GBAux)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1693763802", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649388127482648*^9, 3.649388207531262*^9}},
 CellTags->{
  "ID:1693763802", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAux"},
 CellID->1693763802],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1297770576]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->435073891],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GBAuxCPConnectible3 GBAux)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GBAuxCPConnectible3 GBAux)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
               "[", 
               StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxCPConnectible3", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["M", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"NatPairs", "[", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["B", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ",", "\[LessEqual]"}], "]"}], 
                  "]"}]], "[", 
               StyleBox["P", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["B", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"MMPairs", "[", 
                    RowBox[{
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    SubscriptBox[
                    StyleBox["B", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]]}], "]"}], "]"}]], 
               "[", 
               StyleBox["M", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
             StyleBox["i", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["GBAuxCPConnectible3", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["M", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1693763802"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1693763802-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1693763802-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1693763802", "Proof|ID:1693763802-1"},
 CellID->874031660],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1693763802-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1515571182]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Main Result (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6482872713883038`*^9, 3.648287271574242*^9}, {
  3.649049627270584*^9, 3.649049629179393*^9}, {3.6493925313119307`*^9, 
  3.6493925352967052`*^9}},
 CellID->994676795],

Cell["", "OpenEnvironment",
 CellID->819031807],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: ",
 Cell[BoxData[
  FormBox[
   RowBox[{"GB", "[", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], "]"}], 
   TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648538785357751*^9, 3.648538791189683*^9}},
 CellID->948746204],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["GB", "\[ScriptCapitalG]"], "[", 
   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
  RowBox[{
  "\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(GB empty)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:439420249", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648538794211424*^9, 3.6485388185009947`*^9}},
 CellTags->{
  "ID:439420249", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GB empty", "IR"},
 CellID->439420249],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->145231829]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2134187147],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(GB empty)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(GB empty)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["GB", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], "]"}], 
           "\[Equal]", 
           
           RowBox[{"\[LeftAngleBracket]", 
             "\[RightAngleBracket]"}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["GB", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], "]"}], 
           "\[Equal]", 
           
           RowBox[{"\[LeftAngleBracket]", 
             "\[RightAngleBracket]"}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:439420249"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p439420249-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p439420249-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:439420249", "Proof|ID:439420249-1"},
 CellID->1181943649],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p439420249-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1443792220]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->470033776],

Cell[CellGroupData[{

Cell["Theorem: Properties of Function GB", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.648538514188077*^9, 3.648538520008029*^9}},
 CellID->1353328601],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
   RowBox[{"[", "B", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 298471155}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648538524031863*^9, 3.648538538484898*^9}},
 CellTags->{
  "ID:298471155", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->298471155],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"G", "=", 
   RowBox[{
    UnderscriptBox["GB", "\[ScriptCapitalG]"], "[", "B", 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\GroebnerRings.nb", 1615450915}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6485385477741613`*^9, 3.648538556420373*^9}},
 CellTags->{
  "ID:1615450915", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1615450915],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Exists]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
    RowBox[{"[", "U", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"G", "\[Equal]", 
    RowBox[{"B", 
     TagBox["\:22c8",
      Identity,
      SyntaxForm->"a*b"], "U"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(shape GB)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:209199628", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.649144873175481*^9, 3.649144896287072*^9}},
 CellTags->{
  "ID:209199628", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GB"},
 CellID->209199628],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalG]", "]"}]], 
  RowBox[{"[", "G", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership GB)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:448006251", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.648538540775413*^9, 3.648538566260298*^9}},
 CellTags->{
  "ID:448006251", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership GB", "IR"},
 CellID->448006251],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["ideal", "\[ScriptCapitalG]"], "[", "G", "]"}], "\[Equal]", 
  
  RowBox[{
   UnderscriptBox["ideal", "\[ScriptCapitalG]"], "[", "B", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(ideal GB)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:739795809", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.64853859009898*^9, 3.648538604953113*^9}},
 CellTags->{
  "ID:739795809", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GB", "IR"},
 CellID->739795809],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["isGroebnerBasis", "\[ScriptCapitalG]"], "[", "G", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isGroebnerBasis GB)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1550516051", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6485386084175167`*^9, 3.648538617628399*^9}},
 CellTags->{
  "ID:1550516051", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isGroebnerBasis GB"},
 CellID->1550516051],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->760997017]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->669844270],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(shape GB)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(shape GB)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                 StyleBox["U", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               UnderscriptBox["GB", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], "\:22c8", 
               StyleBox[
               "U", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainTuples", "[", 
                    RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
                 StyleBox["U", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               UnderscriptBox["GB", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{
               StyleBox["B", "ExpressionVariable"], "\:22c8", 
               StyleBox[
               "U", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:209199628"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p209199628-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p209199628-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:209199628", "Proof|ID:209199628-1"},
 CellID->2008671786],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p209199628-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{3.64914545113444*^9, 3.649145481929274*^9},
 CellID->2044012590]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership GB)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership GB)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
              "]"}]], "[", 
           RowBox[{
             UnderscriptBox["GB", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             StyleBox["B", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}], 
              "]"}]], "[", 
           RowBox[{
             UnderscriptBox["GB", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             StyleBox["B", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:448006251"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p448006251-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p448006251-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:448006251", "Proof|ID:448006251-1"},
 CellID->2018395045],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p448006251-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->1640934080]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(ideal GB)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(ideal GB)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["ideal", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GB", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}], "]"}], "\[Equal]", 
           
           RowBox[{
             UnderscriptBox["ideal", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             StyleBox["B", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["ideal", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             RowBox[{
               UnderscriptBox["GB", 
                RowBox[{"GroebnerRing", "[", 
                  StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["B", "ExpressionVariable"], "]"}], "]"}], "\[Equal]", 
           
           RowBox[{
             UnderscriptBox["ideal", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             StyleBox["B", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:739795809"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p739795809-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p739795809-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:739795809", "Proof|ID:739795809-1"},
 CellID->144088343],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p739795809-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->4322593]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isGroebnerBasis GB)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isGroebnerBasis GB)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["isGroebnerBasis", 
            RowBox[{"GroebnerRing", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           "[", 
           RowBox[{
             UnderscriptBox["GB", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             StyleBox["B", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainTuples", "[", 
                  RowBox[{"GroebnerRing", "[", 
                    StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], 
                    "]"}], "]"}]], "[", 
               StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["isGroebnerBasis", 
            RowBox[{"GroebnerRing", "[", 
              StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
           "[", 
           RowBox[{
             UnderscriptBox["GB", 
              RowBox[{"GroebnerRing", "[", 
                StyleBox["\[ScriptCapitalR]", "ExpressionVariable"], "]"}]], 
             "[", 
             StyleBox["B", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1550516051"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1550516051-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1550516051-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1550516051", "Proof|ID:1550516051-1"},
 CellID->856857959],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1550516051-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {61., 67.}}]], "ProofInfoBody",
 CellID->359992752]
}, Closed]]
}, Closed]]
}, Closed]]
}, Open  ]]
},
WindowSize->{1110, 760},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
ShowSelection->True,
CounterAssignments->{{"TheoremaFormulaCounter" -> 412}},
Magnification->1.,
FrontEndVersion->"9.0 for Linux x86 (64-bit) (February 7, 2013)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Theorema"}, "TheoremaCore.nb"]]], 
   
   Cell["Modify the definitions below to change the default appearance of all \
cells in a given style. \nMake modifications to any definition using commands \
in the Format menu.", "Text"], 
   Cell[
    CellGroupData[{
      Cell["Notebook Options Settings", "Section"], 
      Cell[
      "The options defined for the style below will be used at the Notebook \
level.", "Text"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Notebook"], InputAliases -> {"and" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[And]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "or" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", "\[Or]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "not" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Not]", "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "impl" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[Implies]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[DoubleLongLeftRightArrow]", 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                "\[SelectionPlaceholder]", "\[Equal]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox[
                  
                  RowBox[{
                   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                  "a:=b"], "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", ":=", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "andn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[And]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "orn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Or]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "equivn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[DoubleLeftRightArrow]", "\[Piecewise]", 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "cdist" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Piecewise]", 
                 
                 GridBox[{{
                   "\[SelectionPlaceholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}, {
                   "\[Placeholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "far" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exr" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "farc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exrc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "let" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["let", 
                  RowBox[{"\[Placeholder]", "=", "\[Placeholder]"}]], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "app" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293a", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "prep" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293b", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "join" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22c8", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "elemT" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22ff", Identity, SyntaxForm -> "a+b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "subop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "supop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SuperscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "subsupop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubsuperscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "oop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "uoop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               UnderoverscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "osop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox[
                OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
                "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "op" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               TagBox["\[SelectionPlaceholder]", Identity], 
               TagBox[")", "AutoParentheses"]}], "[]" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:e114", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:e115", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "()" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:fd3e", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:fd3f", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "<>" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27e8", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27e9", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "{}" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:e117", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:e118", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "[[]]" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27e6", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27e7", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "(())" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:2e28", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:2e29", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "<<>>" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27ea", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27eb", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "{{}}" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:2983", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:2984", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "(" -> 
            TagBox["(", "AutoParentheses"], ")" -> 
            TagBox[")", "AutoParentheses"]}, Background -> 
          RGBColor[0.957, 0.96, 0.97]], 
         Cell[
          StyleData["Notebook", "Printout"], Background -> None]}, Open]]}, 
     Open]], 
   Cell[
    CellGroupData[{
      Cell["Style Environment Names", "Section"], 
      Cell[
       StyleData[All, "Working"], DockedCells -> {
         Cell[
         "Theorema 2.0", "SBO", FontTracking -> 8, 
          CellMargins -> {{0, 0}, {0, 0}}, 
          CellFrameMargins -> {{0, 50}, {0, 0}}, FontSize -> 28, FontSlant -> 
          "Plain", FontColor -> RGBColor[0.29, 0.46, 0.6], TextAlignment -> 
          Right, Background -> RGBColor[0.36, 0.54, 0.69]]}], 
      Cell[
       StyleData[
       All, "Layout", StyleDefinitions -> StyleData[All, "Working"]], 
       MenuSortingValue -> 1200]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Styles for Title and Section Cells", "Section"], 
      Cell[
       CellGroupData[{
         Cell["Title", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], CellMargins -> {{27, 27}, {10, 30}}, 
             TextAlignment -> Center, MenuSortingValue -> 1100, 
             MenuCommandKey -> "1", FontFamily -> "Bitstream Charter", 
             FontSize -> 36, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Title", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{6, 10}, {5, 30}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], CellMargins -> {{27, 27}, {20, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1150, FontFamily -> 
             "Bitstream Charter", FontSize -> 24, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{6, 10}, {5, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], CellMargins -> {{27, 27}, {8, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Author"], CellMargins -> {{27, 27}, {18, 20}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Sectioning", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], ShowGroupOpener -> True, MenuSortingValue -> 
             1250, MenuCommandKey -> "4", FontFamily -> "Bitstream Charter", 
             FontSize -> 20, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Section", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{2, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {8, 12}}, 
             MenuSortingValue -> 1300, MenuCommandKey -> "5", FontFamily -> 
             "Bitstream Charter", FontSize -> 14, FontWeight -> "Bold", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Subsection", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{9, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {2, 10}}, 
             MenuSortingValue -> 1350, MenuCommandKey -> "6", FontFamily -> 
             "Bitstream Charter", FontSize -> 12, FontWeight -> "Bold", 
             FontSlant -> "Italic", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsection", "Printout"], ShowGroupOpener -> False,
              CellMargins -> {{35, 0}, {7, 14}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsubsection"], CellDingbat -> None, 
             ShowGroupOpener -> True, 
             CellMargins -> {{27, Inherited}, {Inherited, Inherited}}, 
             FontFamily -> "Bitstream Charter", FontSize -> 11, FontWeight -> 
             "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsubsection", "Printout"], 
             CellMargins -> {{50, 0}, {7, 14}}]}, Closed]]}, Open]]}, Open]], 
   
   Cell[
    CellGroupData[{
      Cell["Styles for Publication\[Hyphen]ready Documents", "Section"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Author"], CellMargins -> {{27, 30}, {20, 5}}, 
          MenuSortingValue -> 1900, FontFamily -> "Helvetica", FontSize -> 14,
           FontSlant -> "Italic"], 
         Cell[
          StyleData["Author", "Printout"], CellMargins -> {{6, 10}, {0, 10}}, 
          FontSize -> 10]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Abstract"], CellFrame -> {{0, 0}, {0.5, 0.5}}, 
          CellMargins -> {{27, 140}, {20, 20}}, LineSpacing -> {0.9, 0}, 
          MenuSortingValue -> 2050, FontFamily -> "Times", FontSize -> 12], 
         Cell[
          StyleData["Abstract", "Printout"], Hyphenation -> True, 
          LineSpacing -> {1, 2}, FontSize -> 10]}, Closed]]}, Closed]], 
   Cell[
    CellGroupData[{
      Cell["Text", "Section"], 
      Cell[
       StyleData["Text"], CellMargins -> {{27, 27}, {Inherited, Inherited}}, 
       TextJustification -> 1., ParagraphIndent -> 15, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["EnvironmentText"], ShowCellBracket -> Automatic, 
       CellMargins -> {{82, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextJustification -> 1., ParagraphSpacing -> {0.3, 0.}, 
       MenuSortingValue -> 1480, MenuCommandKey -> "T", FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Item"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "Item", 
          CounterAssignments -> {{"Subitem", 0}, {"Subsubitem", 0}}, 
          MenuSortingValue -> 1600, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Item", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemParagraph"], CellMargins -> {{81, 10}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          DefaultNewCellStyle -> "Item", MenuSortingValue -> 1600, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemParagraph", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "Subitem", 
          CounterAssignments -> {{"Subsubitem", 0}}, MenuSortingValue -> 1610,
           FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subitem", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemParagraph"], CellMargins -> {{105, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subitem", 
          MenuSortingValue -> 1610, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemParagraph", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subsubitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1620, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subsubitem", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemParagraph"], 
          CellMargins -> {{129, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1625, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemParagraph", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "ItemNumbered", 
          CounterAssignments -> {{"SubitemNumbered", 0}, {
            "SubsubitemNumbered", 0}}, MenuSortingValue -> 1630, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemNumbered", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "SubitemNumbered", 
          CounterAssignments -> {{"SubsubitemNumbered", 0}}, MenuSortingValue -> 
          1640, FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemNumbered", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], ".", 
              CounterBox["SubsubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> 
          "SubitemNumbered", CounterIncrements -> "SubsubitemNumbered", 
          MenuSortingValue -> 1645, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemNumbered", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Theorema", "Section"], 
      Cell[
       StyleData["GlobalDeclaration"], CellFrame -> {{0, 2}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 3}, {2, 2}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       InitializationCell -> True, CellFrameMargins -> 10, CellFrameColor -> 
       RGBColor[0.73, 0.65, 0.61], CellFrameLabelMargins -> {{33, 7}, {1, 1}},
        ShowAutoStyles -> False, MenuSortingValue -> 1441, MenuCommandKey -> 
       "G", FontSize -> 14], 
      Cell[
       StyleData["OpenEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {-4, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       StyleData["EnvironmentHeader"], 
       CellFrame -> {{False, False}, {False, False}}, 
       CellMargins -> {{76, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 6, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       ShowAutoStyles -> False, MenuSortingValue -> 1450, FontSize -> 14, 
       FontWeight -> "Bold", FontVariations -> {"CapsType" -> "SmallCaps"}, 
       Background -> RGBColor[0.89, 0.8, 0.69]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FormalTextInputFormula"], ShowCellBracket -> Automatic, 
          CellMargins -> {{96, 10}, {2, 2}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
          InitializationCell -> True, CellFrameLabelMargins -> 53, 
          ShowAutoStyles -> False, LimitsPositioningTokens -> {}, 
          MenuSortingValue -> 1470, MenuCommandKey -> "F", FontWeight -> 
          "Bold", FontColor -> RGBColor[0.25, 0.19, 0.2]], 
         Cell[
          StyleData["FormalTextInputFormula", "Layout"], Editable -> False], 
         Cell[
          StyleData["FormalTextInputFormula", "Presentation"], Editable -> 
          False]}, Open]], 
      Cell[
       StyleData["EndEnvironmentMarker"], 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextAlignment -> Right, MenuSortingValue -> 1480], 
      Cell[
       StyleData["CloseEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {False, True}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {Inherited, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Computation"], CellFrame -> {{1.5, 0}, {0, 0}}, 
          ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 7}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
          CellFrameMargins -> {{8, 8}, {3, 4}}, CellFrameColor -> 
          RGBColor[0.486, 0.33, 0.32], ShowAutoStyles -> False, 
          MenuSortingValue -> 1510, FontWeight -> "Bold", FontColor -> 
          RGBColor[0.23, 0.25, 0.34], Background -> 
          RGBColor[0.88, 0.88, 0.92]], 
         Cell[
          StyleData["Computation", "Layout"], Editable -> False], 
         Cell[
          StyleData["Computation", "Presentation"], Editable -> False]}, 
        Open]], 
      Cell[
       StyleData["ComputationResult"], CellFrame -> {{1.5, 0}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
       CellFrameMargins -> {{8, 8}, {4, 3}}, CellFrameColor -> 
       RGBColor[0.486, 0.33, 0.32], StyleMenuListing -> None, FontColor -> 
       RGBColor[0.23, 0.25, 0.34], Background -> RGBColor[0.88, 0.88, 0.92]], 
      
      Cell[
       StyleData["ComputationInfo"], ShowCellBracket -> False, 
       CellMargins -> {{120, 10}, {10, 2}}, StyleMenuListing -> None, 
       FontFamily -> $CellContext`Arial], 
      Cell[
       StyleData["ComputationInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{144, 10}, {10, 2}}, StyleMenuListing -> None], 
      Cell[
       StyleData["ProofInfo"], ShowCellBracket -> False, 
       CellMargins -> {{76, 10}, {6, 6}}, CellFrameLabelMargins -> 15, 
       StyleMenuListing -> None, FontFamily -> "Arial"], 
      Cell[
       StyleData["ProofInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{100, 10}, {0, 0}}, StyleMenuListing -> None], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FrameLabel"], LanguageCategory -> None, StyleMenuListing -> 
          None, FontFamily -> "Times New Roman", FontSize -> 10, FontWeight -> 
          "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
          FontVariations -> {
           "Masked" -> False, "Outline" -> False, "Shadow" -> False, 
            "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
          RGBColor[0.55, 0.44, 0.42], Background -> None], 
         Cell[
          StyleData["FrameLabel", "Printout"]]}, Open]], 
      Cell[
       StyleData["OpenArchive"], CellFrame -> {{False, False}, {False, True}},
        ShowCellBracket -> Automatic, CellMargins -> {{70, 70}, {0, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 20}, CellFrameMargins -> 3, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], CellFrameLabelMargins -> 1, 
       DefaultFormatType -> DefaultInputFormatType, ShowAutoStyles -> False, 
       MenuSortingValue -> 1210, FontWeight -> Bold, Background -> 
       RGBColor[0.73, 0.74, 0.84]], 
      Cell[
       StyleData["ArchiveInfo"], ShowCellBracket -> Automatic, 
       CellMargins -> {{70, 70}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 5, CellFrameLabelMargins -> 3, ShowCellTags -> 
       False, ShowAutoStyles -> False, MenuSortingValue -> 1220, FontWeight -> 
       "Bold", Background -> RGBColor[0.88, 0.88, 0.92]], 
      Cell[
       StyleData["CloseArchive"], 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> 
       Automatic, CellMargins -> {{0, 0}, {5, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 25}, CellFrameMargins -> 2, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], TextAlignment -> Center, 
       MenuSortingValue -> 1230, FontFamily -> "Arial Black", FontSize -> 24, 
       FontWeight -> "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       Background -> RGBColor[0.73, 0.65, 0.61]], 
      Cell[
       StyleData["IncludeArchive"], CellFrame -> {{4, 0}, {False, False}}, 
       ShowCellBracket -> Automatic, 
       CellMargins -> {{81, 100}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameColor -> RGBColor[0.55, 0.44, 0.42], 
       CellFrameLabels -> {{"\[RightSkeleton]", None}, {None, None}}, 
       CellFrameLabelMargins -> 25, MenuSortingValue -> 1445], 
      Cell[
       StyleData["ExpressionVariable"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionABF"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionMeta"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["AutoParentheses"], Editable -> False, SpanMaxSize -> 3, 
          FontOpacity -> 0.15], 
         Cell[
          StyleData["AutoParentheses", "Layout"], FontOpacity -> 0], 
         Cell[
          StyleData["AutoParentheses", "SlideShow"], FontOpacity -> 0]}, 
        Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Emphasized"], MenuCommandKey -> "E", FontSlant -> 
          "Italic"], 
         Cell[
          StyleData["Emphasized", "SlideShow"], FontSlant -> "Plain", 
          FontColor -> RGBColor[0.36, 0.54, 0.69]]}, Open]], 
      Cell[
       StyleData["GenericButton"], FontFamily -> "Bitstream Charter", 
       FontWeight -> "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
       FontVariations -> {
        "Masked" -> False, "Outline" -> False, "Shadow" -> False, 
         "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
       RGBColor[0.23, 0.25, 0.34]], 
      Cell[
       StyleData["TheoremaForm"], "TwoByteSyntaxCharacterAutoReplacement" -> 
       True, LineSpacing -> {1.25, 0, 2}, SingleLetterItalics -> True, 
       StyleMenuListing -> None, FontFamily -> "Courier"], 
      Cell[
       StyleData["ComponentActive"], FontFamily -> "Times New Roman"], 
      Cell[
       StyleData["ComponentInactive"], FontFamily -> "Times New Roman", 
       FontOpacity -> 0.4], 
      Cell[
       StyleData["ComponentEmpty"], FontFamily -> "Times New Roman", 
       FontVariations -> {"StrikeThrough" -> True}], 
      Cell[
       StyleData["TabView"], FontFamily -> "Times New Roman", FontSize -> 10], 
      Cell[
       StyleData["TabViewLabel"], FontFamily -> "Times New Roman", FontSize -> 
       10, FontWeight -> "Bold"], 
      Cell[
       StyleData["FormReference", StyleDefinitions -> StyleData["Text"]], 
       FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> "Plain",
        FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["FormReferenceHover", StyleDefinitions -> StyleData["Text"]],
        FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> 
       "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> True}, 
       FontColor -> RGBColor[0.404, 0.43, 0.545]], 
      Cell[
       StyleData["DisplayFormula"], CellMargins -> {{36, 6}, {7, 7}}, 
       ConversionRules :> {
        "HTML" -> {{"", ""}, {"<p class=\"Equation\">", "</p>"}}}, 
       CellHorizontalScrolling -> True, DefaultFormatType -> TraditionalForm, 
       InputAutoReplacements -> {}, TextAlignment -> Center, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       ButtonBoxOptions -> {Appearance -> {Automatic, None}}]}, Open]]}, 
  Visible -> False, FrontEndVersion -> 
  "9.0 for Linux x86 (64-bit) (February 7, 2013)", StyleDefinitions -> 
  "StylesheetFormatting.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "ID:1079943823"->{
  Cell[73877, 1840, 1134, 34, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1079943823", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "print"},
   CellID->1079943823]},
 "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"->{
  Cell[73877, 1840, 1134, 34, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1079943823", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "print"},
   CellID->1079943823],
  Cell[76120, 1918, 1968, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1033951332", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed"},
   CellID->1033951332],
  Cell[78791, 2005, 846, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1967331668", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1967331668],
  Cell[79640, 2027, 1127, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1698292943", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs 1"},
   CellID->1698292943],
  Cell[80770, 2060, 1720, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1961920283", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs 2"},
   CellID->1961920283],
  Cell[82493, 2110, 1561, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1516706954", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "allPairs"},
   CellID->1516706954],
  Cell[84057, 2152, 1235, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:329670188", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "update"},
   CellID->329670188],
  Cell[86332, 2227, 748, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1283917386", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1283917386],
  Cell[87083, 2247, 810, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:14652249", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->14652249],
  Cell[87896, 2269, 2437, 75, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1012271681", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element]"},
   CellID->1012271681],
  Cell[112282, 3004, 742, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:988862829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->988862829],
  Cell[113027, 3024, 1339, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:16377100", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element] ProdDomain"},
   CellID->16377100],
  Cell[114369, 3068, 1472, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:136514135", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element] ProdDomain DomainTuples"},
   CellID->136514135],
  Cell[115844, 3116, 1986, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1872634730", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element] alternative", "IR"},
   CellID->1872634730],
  Cell[117833, 3174, 1959, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1931599387", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs< \[Element] alternative", "IR"},
   CellID->1931599387],
  Cell[119795, 3236, 2024, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2124565744", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs\[LessEqual] \[Element] alternative", "IR"},
   CellID->2124565744],
  Cell[121822, 3300, 1538, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1917398594", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs< \[Element] alternative II"},
   CellID->1917398594],
  Cell[123363, 3349, 1649, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:272474466", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs\[LessEqual] \[Element] alternative II"},
   CellID->272474466],
  Cell[125015, 3401, 1398, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2005363738", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs< \[LessEqual]"},
   CellID->2005363738],
  Cell[126416, 3448, 1524, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1661715038", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NatPairs< \[LessEqual]"},
   CellID->1661715038],
  Cell[127943, 3497, 726, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1020080047", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1020080047],
  Cell[128672, 3518, 1316, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1327808237", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs +1"},
   CellID->1327808237],
  Cell[129991, 3562, 1426, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:798140928", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NatPairs +1"},
   CellID->798140928],
  Cell[131420, 3608, 791, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2111464127", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->2111464127],
  Cell[132214, 3630, 1264, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1834895057", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[GreaterEqual]"},
   CellID->1834895057],
  Cell[133481, 3672, 1388, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1703465728", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NatPairs \[GreaterEqual]"},
   CellID->1703465728],
  Cell[198273, 5464, 954, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:971135058", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->971135058],
  Cell[199230, 5488, 1116, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1202991311", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership pairs unary", "IR"},
   CellID->1202991311],
  Cell[200349, 5523, 1522, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1932739315", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs unary content'"},
   CellID->1932739315],
  Cell[201874, 5568, 1264, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1628109432", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs unary content"},
   CellID->1628109432],
  Cell[203141, 5608, 1189, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:540671405", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership pairs binary", "IR"},
   CellID->540671405],
  Cell[204333, 5645, 1465, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:881225230", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs binary content'"},
   CellID->881225230],
  Cell[205801, 5689, 1247, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:717573619", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs binary content"},
   CellID->717573619],
  Cell[207051, 5729, 859, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:794667329", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "allPairs 0", "IR"},
   CellID->794667329],
  Cell[207913, 5755, 1126, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1385740688", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership allPairs", "IR"},
   CellID->1385740688],
  Cell[209042, 5790, 1232, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1430860208", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "allPairs content"},
   CellID->1430860208],
  Cell[247887, 6913, 823, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:963656485", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->963656485],
  Cell[248713, 6935, 1072, 30, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1750952092", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1750952092],
  Cell[249788, 6967, 805, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1500640019", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1500640019],
  Cell[250596, 6990, 1160, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:723173326", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership update", "IR"},
   CellID->723173326],
  Cell[251759, 7026, 1301, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:392184336", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "update content I"},
   CellID->392184336],
  Cell[253063, 7073, 1174, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:341058317", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "update content II"},
   CellID->341058317],
  Cell[268783, 7520, 1965, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1597870023", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed alternative", "IR"},
   CellID->1597870023],
  Cell[270751, 7585, 1099, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:40246499", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed symmetric"},
   CellID->40246499],
  Cell[271853, 7619, 728, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1131819573", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1131819573],
  Cell[272584, 7639, 1106, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1083804603", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed allPairs"},
   CellID->1083804603],
  Cell[273693, 7673, 2454, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:316507172", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed update"},
   CellID->316507172],
  Cell[276150, 7748, 2158, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1179241697", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed prepend"},
   CellID->1179241697],
  Cell[313515, 8786, 651, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1573502139", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1573502139],
  Cell[314169, 8804, 789, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:377704681", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->377704681],
  Cell[314961, 8825, 1996, 37, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1468120413", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1468120413],
  Cell[316960, 8864, 1325, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1774031579", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing \[Element]"},
   CellID->1774031579],
  Cell[318288, 8902, 855, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:182210603", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing 0"},
   CellID->182210603],
  Cell[319146, 8927, 855, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:814933967", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing 1"},
   CellID->814933967],
  Cell[320004, 8952, 906, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1684568525", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing +"},
   CellID->1684568525],
  Cell[320913, 8979, 902, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:584522550", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing -"},
   CellID->584522550],
  Cell[321818, 9006, 908, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1458964616", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing - unary"},
   CellID->1458964616],
  Cell[322729, 9033, 906, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1496488880", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing \[RawStar]"},
   CellID->1496488880],
  Cell[323638, 9060, 1432, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:342394137", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing \[Succeeds]"},
   CellID->342394137],
  Cell[325073, 9105, 887, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1566608311", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing multN"},
   CellID->1566608311],
  Cell[325963, 9132, 1098, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:628841792", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing mult 1"},
   CellID->628841792],
  Cell[327064, 9166, 1120, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:378340035", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing mult 2"},
   CellID->378340035],
  Cell[328187, 9200, 1065, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:822076309", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lcrd"},
   CellID->822076309],
  Cell[329255, 9231, 969, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:849156542", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing rdm"},
   CellID->849156542],
  Cell[330227, 9260, 890, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:809363439", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lcrd unary"},
   CellID->809363439],
  Cell[331120, 9286, 987, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1671864047", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing cpm"},
   CellID->1671864047],
  Cell[332514, 9332, 1070, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:558323540", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing rd"},
   CellID->558323540],
  Cell[334129, 9385, 1675, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1403530194", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing rdT"},
   CellID->1403530194],
  Cell[336453, 9456, 1732, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:689058884", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing trd"},
   CellID->689058884],
  Cell[339865, 9557, 2982, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1528624471", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lmTuple"},
   CellID->1528624471],
  Cell[342850, 9624, 986, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1192872489", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lmTupleAux 1"},
   CellID->1192872489],
  Cell[343839, 9653, 1942, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1479034574", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lmTupleAux 2"},
   CellID->1479034574],
  Cell[347635, 9767, 1809, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1529751135", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing cpmTuple 1"},
   CellID->1529751135],
  Cell[351052, 9876, 1705, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:551185449", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing cpmTuple 2"},
   CellID->551185449],
  Cell[353203, 9938, 1246, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1533920470", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GB"},
   CellID->1533920470],
  Cell[354452, 9972, 1112, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1976234692", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GBAux 1"},
   CellID->1976234692],
  Cell[355567, 10003, 2792, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:144716839", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GBAux 2"},
   CellID->144716839],
  Cell[358362, 10078, 4571, 115, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:189044945", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GBAux 3"},
   CellID->189044945],
  Cell[364060, 10227, 3441, 90, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1766960499", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing chainCrit"},
   CellID->1766960499],
  Cell[369194, 10368, 1132, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1777604038", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing ard"},
   CellID->1777604038],
  Cell[370329, 10400, 1161, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:620529741", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing ardAux 1"},
   CellID->620529741],
  Cell[371493, 10431, 2316, 71, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:174904624", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing ardAux 2"},
   CellID->174904624],
  Cell[377682, 10622, 1036, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:336511876", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalQ]"},
   CellID->336511876],
  Cell[392232, 11064, 999, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:193404746", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]"},
   CellID->193404746],
  Cell[407831, 11552, 968, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1821700886", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]24"},
   CellID->1821700886],
  Cell[424655, 12079, 1211, 28, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1158015694", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalQ]xy"},
   CellID->1158015694],
  Cell[457422, 13081, 1229, 28, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1921709458", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalZ]xy"},
   CellID->1921709458],
  Cell[458654, 13111, 1203, 28, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:281007089", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]xy"},
   CellID->281007089],
  Cell[481261, 13836, 998, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1512392075", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalZ]24xy"},
   CellID->1512392075],
  Cell[482262, 13863, 1045, 27, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:543229319", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]24xy"},
   CellID->543229319],
  Cell[517693, 14993, 965, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:280174026", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isCommRing1"},
   CellID->280174026],
  Cell[522070, 15126, 712, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1057748777", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1057748777],
  Cell[522824, 15149, 743, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:464516108", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->464516108],
  Cell[524671, 15207, 1456, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1148013153", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainSets \[ScriptCapitalG]", "IR"},
   CellID->1148013153],
  Cell[526130, 15254, 1464, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1668771272", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples \[ScriptCapitalG]", "IR"},
   CellID->1668771272],
  Cell[527597, 15301, 1491, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1928426019", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Sum] \[ScriptCapitalG]", "IR"},
   CellID->1928426019],
  Cell[529091, 15346, 692, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:546598180", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->546598180],
  Cell[529786, 15365, 1600, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:581321955", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "reduceM \[ScriptCapitalG]", "IR"},
   CellID->581321955],
  Cell[531389, 15411, 1595, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:405682622", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow] \[ScriptCapitalG] multiplier", "IR"},
   
   CellID->405682622],
  Cell[532987, 15463, 1613, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1165811909", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow] \[ScriptCapitalG]", "IR"},
   CellID->1165811909],
  Cell[534603, 15514, 1641, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:306221528", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[LeftRightArrow] \[ScriptCapitalG]", "IR"},
   CellID->306221528],
  Cell[536247, 15565, 1076, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:113746156", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isReducible \[ScriptCapitalG] multiplier", "IR"},
   CellID->113746156],
  Cell[537326, 15594, 1028, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1372373678", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isReducible \[ScriptCapitalG]", "IR"},
   CellID->1372373678],
  Cell[538357, 15622, 1051, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:12812135", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "symred \[ScriptCapitalG]", "IR"},
   CellID->12812135],
  Cell[539411, 15652, 1057, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:168109679", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "areCorrelative \[ScriptCapitalG]", "IR"},
   CellID->168109679],
  Cell[540471, 15680, 1255, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1427340407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isNTCR \[ScriptCapitalG]", "IR"},
   CellID->1427340407],
  Cell[541729, 15717, 1254, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:587977465", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isMNTCR \[ScriptCapitalG]", "IR"},
   CellID->587977465],
  Cell[542986, 15754, 1017, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:950195558", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isNTCR \[ScriptCapitalG] unary", "IR"},
   CellID->950195558],
  Cell[544006, 15782, 1024, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2144197915", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isMNTCR \[ScriptCapitalG] unary", "IR"},
   CellID->2144197915],
  Cell[545033, 15810, 1207, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1924053921", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "red \[ScriptCapitalG] set", "IR"},
   CellID->1924053921],
  Cell[546243, 15847, 1148, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1439528152", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Tilde] \[ScriptCapitalG]", "IR"},
   CellID->1439528152],
  Cell[629817, 18138, 1044, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1674560446", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isPartIrreflOrder \[ScriptCapitalG]"},
   CellID->1674560446],
  Cell[630864, 18170, 1025, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:167203289", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isNoetherian \[ScriptCapitalG]"},
   CellID->167203289],
  Cell[631892, 18202, 821, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:569855140", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "Multipliers \[ScriptCapitalG]"},
   CellID->569855140],
  Cell[632716, 18225, 791, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:180032137", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R0 \[ScriptCapitalG]"},
   CellID->180032137],
  Cell[633510, 18247, 793, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:428662253", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R1 \[ScriptCapitalG]"},
   CellID->428662253],
  Cell[634306, 18269, 792, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:530735576", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R2 \[ScriptCapitalG]"},
   CellID->530735576],
  Cell[635101, 18291, 794, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1081126646", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R3 \[ScriptCapitalG]"},
   CellID->1081126646],
  Cell[635898, 18313, 791, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1591716135", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R4 \[ScriptCapitalG]"},
   CellID->1591716135],
  Cell[636692, 18335, 791, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:960649057", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R5 \[ScriptCapitalG]"},
   CellID->960649057],
  Cell[637486, 18357, 795, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:227638759", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R6 \[ScriptCapitalG]"},
   CellID->227638759],
  Cell[638284, 18379, 796, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2067143528", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R7 \[ScriptCapitalG]"},
   CellID->2067143528],
  Cell[639083, 18401, 796, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1699723496", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R8 \[ScriptCapitalG]"},
   CellID->1699723496],
  Cell[676070, 19520, 889, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2054051596", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isReductionRing"},
   CellID->2054051596],
  Cell[680367, 19653, 769, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1031954878", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1031954878],
  Cell[681179, 19678, 743, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:660860156", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->660860156],
  Cell[682439, 19717, 790, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:939639226", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R9 \[ScriptCapitalG]"},
   CellID->939639226],
  Cell[683232, 19739, 793, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:37694441", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R10 \[ScriptCapitalG]"},
   CellID->37694441],
  Cell[684028, 19761, 796, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:576375692", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R11 \[ScriptCapitalG]"},
   CellID->576375692],
  Cell[684827, 19783, 797, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1226510984", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R12 \[ScriptCapitalG]"},
   CellID->1226510984],
  Cell[685627, 19805, 799, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1781039403", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R13 \[ScriptCapitalG]"},
   CellID->1781039403],
  Cell[701645, 20291, 814, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:166003593", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isDomain"},
   CellID->166003593],
  Cell[702462, 20314, 1003, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1167967104", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isAlgoReductionRing"},
   CellID->1167967104],
  Cell[709746, 20535, 696, 19, 33, "GlobalDeclaration",
   CellTags->{
    "ID:2087655986", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->2087655986],
  Cell[710484, 20559, 719, 19, 33, "GlobalDeclaration",
   CellTags->{
    "ID:955133330", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->955133330],
  Cell[712234, 20612, 767, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:977980890", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->977980890],
  Cell[713004, 20634, 816, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1748570587", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1748570587],
  Cell[713823, 20657, 822, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:699276041", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership rd"},
   CellID->699276041],
  Cell[714648, 20681, 1184, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:867263181", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "reducible rd"},
   CellID->867263181],
  Cell[715835, 20717, 1168, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1431447619", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "irreducible rd"},
   CellID->1431447619],
  Cell[717006, 20755, 1058, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1490981815", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Succeeds] rd"},
   CellID->1490981815],
  Cell[740976, 21412, 680, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:243233209", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->243233209],
  Cell[741659, 21431, 1505, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:974939989", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdTBase1"},
   CellID->974939989],
  Cell[743167, 21476, 2294, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1014318336", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdTBase2"},
   CellID->1014318336],
  Cell[745464, 21547, 2498, 76, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:107497168", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdTRec"},
   CellID->107497168],
  Cell[748678, 21650, 1011, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:839458854", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->839458854],
  Cell[749692, 21678, 4623, 135, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1436407557", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "4"},
   CellID->1436407557],
  Cell[754318, 21815, 1921, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2098894842", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdT induction"},
   CellID->2098894842],
  Cell[778497, 22421, 2089, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:160961870", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdT alternative", "IR"},
   CellID->160961870],
  Cell[789369, 22705, 953, 25, 70, "GlobalDeclaration",
   CellTags->{
    "ID:168345814", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->168345814],
  Cell[790325, 22732, 684, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1634249728", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1634249728],
  Cell[791012, 22752, 821, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:512270654", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->512270654],
  Cell[791836, 22775, 878, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2105002699", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership rdT"},
   CellID->2105002699],
  Cell[792717, 22800, 1447, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1505002987", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "reducible rdT"},
   CellID->1505002987],
  Cell[794167, 22846, 1325, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:514228174", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "irreducible rdT"},
   CellID->514228174],
  Cell[795495, 22886, 849, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:208729427", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow]\[RawStar] rdT"},
   CellID->208729427],
  Cell[796347, 22911, 1542, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:278468994", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Succeeds] rdT"},
   CellID->278468994],
  Cell[832600, 23862, 1689, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:96175791", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "red rdT"},
   CellID->96175791],
  Cell[841022, 24103, 3592, 106, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:909149002", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "trd induction"},
   CellID->909149002],
  Cell[856495, 24506, 1008, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1097836943", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1097836943],
  Cell[857506, 24534, 763, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:206560148", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->206560148],
  Cell[858272, 24556, 827, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1315173810", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership trd"},
   CellID->1315173810],
  Cell[859102, 24580, 852, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1071214578", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow]\[RawStar] trd"},
   CellID->1071214578],
  Cell[859957, 24605, 880, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:657218052", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "trd \[NotElement] red"},
   CellID->657218052],
  Cell[877161, 25078, 822, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:480907095", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->480907095],
  Cell[878608, 25117, 1036, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1000930828", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1000930828],
  Cell[879647, 25142, 3575, 99, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1774407783", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element]"},
   CellID->1774407783],
  Cell[883356, 25250, 975, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:193942992", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->193942992],
  Cell[884981, 25295, 1175, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:44826735", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] ProdDomain"},
   CellID->44826735],
  Cell[886159, 25328, 1305, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:155756485", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] ProdDomain DomainTuples"},
   CellID->155756485],
  Cell[887467, 25365, 3415, 96, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1019394084", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] alternative I", "IR"},
   CellID->1019394084],
  Cell[890885, 25463, 2963, 82, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1458538511", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] alternative II", "IR"},
   CellID->1458538511],
  Cell[893851, 25547, 2529, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:81000274", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] alternative III"},
   CellID->81000274],
  Cell[934186, 26564, 3381, 88, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1140506798", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NPairs shape"},
   CellID->1140506798],
  Cell[948750, 26926, 1029, 27, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1527335579", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1527335579],
  Cell[949782, 26955, 1056, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1794762591", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership cpmTuple 1"},
   CellID->1794762591],
  Cell[950841, 26985, 2346, 70, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1790156146", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpmTuple content 1"},
   CellID->1790156146],
  Cell[973212, 27548, 892, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:884601666", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->884601666],
  Cell[974107, 27573, 906, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1188084204", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1188084204],
  Cell[975016, 27598, 1146, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1708769881", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership cpmTuple 2"},
   CellID->1708769881],
  Cell[976165, 27630, 2401, 71, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:125591800", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpmTuple content 2"},
   CellID->125591800],
  Cell[1000831, 28245, 731, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1122306642", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1122306642],
  Cell[1001565, 28266, 3193, 102, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:885588057", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "5"},
   CellID->885588057],
  Cell[1004761, 28370, 3039, 95, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:438489157", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTupleAux induction"},
   CellID->438489157],
  Cell[1026756, 28950, 844, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1795822474", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1795822474],
  Cell[1027603, 28974, 1070, 31, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2110280918", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->2110280918],
  Cell[1028676, 29007, 1057, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:481303170", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership lmTupleAux"},
   CellID->481303170],
  Cell[1029736, 29037, 2692, 81, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1389102731", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTupleAux content"},
   CellID->1389102731],
  Cell[1053379, 29632, 1026, 27, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1615013989", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1615013989],
  Cell[1054408, 29661, 770, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1776007069", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1776007069],
  Cell[1055181, 29683, 1095, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1846257369", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership lmTuple"},
   CellID->1846257369],
  Cell[1056279, 29712, 2420, 72, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:851213532", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTuple \[NotEqual]"},
   CellID->851213532],
  Cell[1058702, 29786, 3744, 103, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:959030563", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTuple ="},
   CellID->959030563],
  Cell[1096874, 30714, 785, 20, 44, "GlobalDeclaration",
   CellTags->{
    "ID:1440157218", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1440157218],
  Cell[1097875, 30743, 673, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:805845835", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->805845835],
  Cell[1099056, 30779, 2884, 85, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1576526273", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple"},
   CellID->1576526273],
  Cell[1102644, 30891, 1045, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:857941005", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple allPairs"},
   CellID->857941005],
  Cell[1103692, 30919, 2192, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1202189993", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple update"},
   CellID->1202189993],
  Cell[1105887, 30984, 3124, 89, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1698314044", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple rest"},
   CellID->1698314044],
  Cell[1109014, 31075, 1743, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:812012342", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple exchange"},
   CellID->812012342],
  Cell[1143456, 31943, 2816, 72, 99, "FormalTextInputFormula",
   CellTags->{
    "ID:1260897708", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "chainCrit 1"},
   CellID->1260897708],
  Cell[1169532, 32690, 817, 21, 33, "GlobalDeclaration",
   CellTags->{
    "ID:957934274", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->957934274],
  Cell[1170392, 32716, 795, 20, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1363987639", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1363987639],
  Cell[1172007, 32765, 2288, 68, 130, "FormalTextInputFormula",
   CellTags->{
    "ID:184081942", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBTermOrder"},
   CellID->184081942],
  Cell[1174429, 32842, 946, 22, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1341865321", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1341865321],
  Cell[1176068, 32891, 812, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:320402459", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] isDomain"},
   CellID->320402459],
  Cell[1176883, 32914, 1046, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1861353391", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] isPartIrreflOrder"},
   CellID->1861353391],
  Cell[1177932, 32946, 1028, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:461530885", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] isNoetherian"},
   CellID->461530885],
  Cell[1178963, 32978, 4409, 118, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1658780438", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] shape"},
   CellID->1658780438],
  Cell[1206465, 33710, 786, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1606683850", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1606683850],
  Cell[1207254, 33732, 1508, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:973931584", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] \[Succeeds] 1"},
   CellID->973931584],
  Cell[1208765, 33772, 1457, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1350364326", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] \[Succeeds] 2"},
   CellID->1350364326],
  Cell[1210225, 33812, 1671, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:529848255", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] \[Succeeds] 3"},
   CellID->529848255],
  Cell[1211899, 33864, 3057, 83, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:592062547", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[SquareSubset] append trd"},
   CellID->592062547],
  Cell[1246777, 34752, 743, 18, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1848949834", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1848949834],
  Cell[1247523, 34772, 2717, 74, 114, "FormalTextInputFormula",
   CellTags->{
    "ID:1362258751", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxTest"},
   CellID->1362258751],
  Cell[1250243, 34848, 1436, 37, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1347349813", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxBase"},
   CellID->1347349813],
  Cell[1251682, 34887, 4352, 115, 154, "FormalTextInputFormula",
   CellTags->{
    "ID:270708621", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec1"},
   CellID->270708621],
  Cell[1256792, 35031, 3492, 99, 154, "FormalTextInputFormula",
   CellTags->{
    "ID:1542247232", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec2"},
   CellID->1542247232],
  Cell[1260287, 35132, 5126, 136, 201, "FormalTextInputFormula",
   CellTags->{
    "ID:580452282", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec3"},
   CellID->580452282],
  Cell[1266250, 35299, 5545, 153, 226, "FormalTextInputFormula",
   CellTags->{
    "ID:598818422", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec4"},
   CellID->598818422],
  Cell[1272430, 35478, 651, 16, 32, "GlobalDeclaration",
   CellTags->{
    "ID:1392375551", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1392375551],
  Cell[1273084, 35496, 8640, 228, 316, "FormalTextInputFormula",
   CellTags->{
    "ID:994411308", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "1"},
   CellID->994411308],
  Cell[1281727, 35726, 17123, 416, 446, "FormalTextInputFormula",
   CellTags->{
    "ID:1185976259", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "2"},
   CellID->1185976259],
  Cell[1298853, 36144, 4696, 139, 173, "FormalTextInputFormula",
   CellTags->{
    "ID:1125858125", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "3"},
   CellID->1125858125],
  Cell[1303552, 36285, 4442, 125, 222, "FormalTextInputFormula",
   CellTags->{
    "ID:1780855977", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAux induction"},
   CellID->1780855977],
  Cell[1307997, 36412, 4432, 126, 281, "FormalTextInputFormula",
   CellTags->{
    "ID:380025998", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAux induction II"},
   CellID->380025998],
  Cell[1496312, 41028, 3435, 92, 164, "FormalTextInputFormula",
   CellTags->{
    "ID:1379909256", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAux 3 alternative", "IR"},
   CellID->1379909256],
  Cell[1516392, 41500, 778, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:884344895", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->884344895],
  Cell[1517642, 41538, 855, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2012261150", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->2012261150],
  Cell[1518500, 41561, 1873, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1467104417", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxBase"},
   CellID->1467104417],
  Cell[1520376, 41620, 1850, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1354352171", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec1"},
   CellID->1354352171],
  Cell[1522229, 41679, 1850, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1322915202", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec2"},
   CellID->1322915202],
  Cell[1524082, 41738, 1847, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:620036821", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec3"},
   CellID->620036821],
  Cell[1525932, 41797, 1850, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1111467670", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec4"},
   CellID->1111467670],
  Cell[1573348, 42958, 2753, 80, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1314333166", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAux"},
   CellID->1314333166],
  Cell[1586472, 43301, 2267, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1167572947", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership GBAux"},
   CellID->1167572947],
  Cell[1597975, 43601, 783, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:208228172", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->208228172],
  Cell[1599274, 43641, 2864, 81, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:781335069", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[RightArrow]\[RawStar]"},
   CellID->781335069],
  Cell[1615472, 44044, 855, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1452629404", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1452629404],
  Cell[1616330, 44067, 1429, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1597922168", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxBase"},
   CellID->1597922168],
  Cell[1617762, 44109, 1426, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:348489962", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec1"},
   CellID->348489962],
  Cell[1619191, 44151, 1431, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1787031010", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec2"},
   CellID->1787031010],
  Cell[1620625, 44193, 1426, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:963974147", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec3"},
   CellID->963974147],
  Cell[1622054, 44235, 1428, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:473957484", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec4"},
   CellID->473957484],
  Cell[1668282, 45367, 2261, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1826585682", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAux"},
   CellID->1826585682],
  Cell[1680381, 45684, 782, 20, 44, "GlobalDeclaration",
   CellTags->{
    "ID:416121228", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->416121228],
  Cell[1681970, 45732, 682, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2050404037", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->2050404037],
  Cell[1682655, 45752, 4775, 124, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1837575531", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple"},
   CellID->1837575531],
  Cell[1687433, 45878, 3358, 90, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:346323949", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1"},
   CellID->346323949],
  Cell[1690794, 45970, 2895, 77, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1643593823", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2"},
   CellID->1643593823],
  Cell[1693692, 46049, 3067, 85, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1213243088", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3"},
   CellID->1213243088],
  Cell[1697747, 46169, 3785, 112, 130, "FormalTextInputFormula",
   CellTags->{
    "ID:949923506", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "h=0"},
   CellID->949923506],
  Cell[1717818, 46671, 776, 21, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1322820809", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1322820809],
  Cell[1718597, 46694, 872, 23, 44, "GlobalDeclaration",
   CellTags->{
    "ID:1772563804", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1772563804],
  Cell[1719472, 46719, 853, 22, 34, "GlobalDeclaration",
   CellTags->{
    "ID:667808408", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->667808408],
  Cell[1720328, 46743, 2593, 81, 69, "FormalTextInputFormula",
   CellTags->{
    "ID:1601898370", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "h\[NotEqual]0"},
   CellID->1601898370],
  Cell[1722924, 46826, 1850, 50, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:622683853", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpConnectible append h"},
   CellID->622683853],
  Cell[1755998, 47610, 3585, 95, 93, "FormalTextInputFormula",
   CellTags->{
    "ID:1834517245", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxTest alternative"},
   CellID->1834517245],
  Cell[1759586, 47707, 4183, 100, 183, "FormalTextInputFormula",
   CellTags->{
    "ID:1923946854", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxTest 1"},
   CellID->1923946854],
  Cell[1810002, 48942, 1031, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:258768016", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 empty"},
   CellID->258768016],
  Cell[1811036, 48971, 855, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1002747542", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1002747542],
  Cell[1811894, 48994, 946, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:673303920", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 GBAuxRec3"},
   CellID->673303920],
  Cell[1812843, 49020, 949, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1502806330", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 GBAuxRec4"},
   CellID->1502806330],
  Cell[1830657, 49496, 1065, 30, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1133484500", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1133484500],
  Cell[1831725, 49528, 852, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:291397089", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->291397089],
  Cell[1832580, 49551, 1277, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1494094918", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 GBAux"},
   CellID->1494094918],
  Cell[1833860, 49587, 1776, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:402861438", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpConnectible GBAux lmTuple"},
   CellID->402861438],
  Cell[1856359, 50150, 800, 21, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1173305577", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1173305577],
  Cell[1857162, 50173, 651, 16, 30, "GlobalDeclaration",
   CellTags->{
    "ID:1545923689", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1545923689],
  Cell[1857816, 50191, 1393, 39, 43, "FormalTextInputFormula",
   CellTags->{
    "ID:1684239249", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple empty"},
   CellID->1684239249],
  Cell[1859212, 50232, 1629, 45, 43, "FormalTextInputFormula",
   CellTags->{
    "ID:547535692", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple prepend"},
   CellID->547535692],
  Cell[1860844, 50279, 2158, 60, 89, "FormalTextInputFormula",
   CellTags->{
    "ID:812188873", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple isTODOTuple"},
   CellID->812188873],
  Cell[1863005, 50341, 1407, 41, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1105161129", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple lmTuple"},
   CellID->1105161129],
  Cell[1864415, 50384, 1476, 44, 43, "FormalTextInputFormula",
   CellTags->{
    "ID:981490890", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple 0"},
   CellID->981490890],
  Cell[1907942, 51481, 1163, 32, 44, "GlobalDeclaration",
   CellTags->{
    "ID:681053692", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->681053692],
  Cell[1909108, 51515, 797, 21, 33, "GlobalDeclaration",
   CellTags->{
    "ID:995853659", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->995853659],
  Cell[1909908, 51538, 876, 23, 44, "GlobalDeclaration",
   CellTags->{
    "ID:218025475", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->218025475],
  Cell[1910787, 51563, 1195, 30, 34, "GlobalDeclaration",
   CellTags->{
    "ID:2052980984", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->2052980984],
  Cell[1911985, 51595, 1289, 36, 48, "GlobalDeclaration",
   CellTags->{
    "ID:925810378", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->925810378],
  Cell[1913277, 51633, 1524, 42, 81, "FormalTextInputFormula",
   CellTags->{
    "ID:423392620", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple 1"},
   CellID->423392620],
  Cell[1914804, 51677, 1248, 35, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:1533257011", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple 2"},
   CellID->1533257011],
  Cell[1975295, 53267, 857, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1154842295", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1154842295],
  Cell[1976155, 53290, 949, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1427865659", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxBase"},
   CellID->1427865659],
  Cell[1977107, 53316, 948, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:802696830", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec1"},
   CellID->802696830],
  Cell[1978058, 53342, 947, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:668126868", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec2"},
   CellID->668126868],
  Cell[1979008, 53368, 949, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1848124281", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec3"},
   CellID->1848124281],
  Cell[1979960, 53394, 942, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:6143564", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec4"},
   CellID->6143564],
  Cell[2011814, 54230, 1041, 30, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1620493315", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1620493315],
  Cell[2012858, 54262, 831, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1672348970", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1672348970],
  Cell[2013692, 54285, 912, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1826259623", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1826259623],
  Cell[2014607, 54311, 979, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1814594394", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAux"},
   CellID->1814594394],
  Cell[2015589, 54337, 743, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1024326782", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1024326782],
  Cell[2016335, 54358, 1057, 30, 70, "GlobalDeclaration",
   CellTags->{
    "ID:585977074", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->585977074],
  Cell[2017395, 54390, 2212, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1959330209", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 update"},
   CellID->1959330209],
  Cell[2048224, 55130, 851, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:33155789", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->33155789],
  Cell[2049078, 55153, 949, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1691022439", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxBase"},
   CellID->1691022439],
  Cell[2050030, 55179, 946, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:544907973", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec1"},
   CellID->544907973],
  Cell[2050979, 55205, 948, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1039424559", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec2"},
   CellID->1039424559],
  Cell[2051930, 55231, 949, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1704484269", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec3"},
   CellID->1704484269],
  Cell[2052882, 55257, 949, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2114764931", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec4"},
   CellID->2114764931],
  Cell[2084817, 56096, 1993, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1693763802", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAux"},
   CellID->1693763802],
  Cell[2096171, 56400, 946, 26, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:439420249", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GB empty", "IR"},
   CellID->439420249],
  Cell[2101083, 56554, 781, 20, 44, "GlobalDeclaration",
   CellTags->{
    "ID:298471155", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->298471155],
  Cell[2101867, 56576, 747, 20, 43, "GlobalDeclaration",
   CellTags->{
    "ID:1615450915", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1615450915],
  Cell[2102617, 56598, 1118, 35, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:209199628", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GB"},
   CellID->209199628],
  Cell[2103738, 56635, 868, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:448006251", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership GB", "IR"},
   CellID->448006251],
  Cell[2104609, 56660, 910, 26, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:739795809", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GB", "IR"},
   CellID->739795809],
  Cell[2105522, 56688, 832, 22, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1550516051", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isGroebnerBasis GB"},
   CellID->1550516051]},
 "print"->{
  Cell[73877, 1840, 1134, 34, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:1079943823", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "print"},
   CellID->1079943823]},
 "ID:1033951332"->{
  Cell[76120, 1918, 1968, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1033951332", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed"},
   CellID->1033951332]},
 "processed"->{
  Cell[76120, 1918, 1968, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1033951332", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed"},
   CellID->1033951332]},
 "ID:1967331668"->{
  Cell[78791, 2005, 846, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1967331668", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1967331668]},
 "ID:1698292943"->{
  Cell[79640, 2027, 1127, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1698292943", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs 1"},
   CellID->1698292943]},
 "pairs 1"->{
  Cell[79640, 2027, 1127, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1698292943", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs 1"},
   CellID->1698292943]},
 "ID:1961920283"->{
  Cell[80770, 2060, 1720, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1961920283", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs 2"},
   CellID->1961920283]},
 "pairs 2"->{
  Cell[80770, 2060, 1720, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1961920283", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs 2"},
   CellID->1961920283]},
 "ID:1516706954"->{
  Cell[82493, 2110, 1561, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1516706954", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "allPairs"},
   CellID->1516706954]},
 "allPairs"->{
  Cell[82493, 2110, 1561, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1516706954", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "allPairs"},
   CellID->1516706954]},
 "ID:329670188"->{
  Cell[84057, 2152, 1235, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:329670188", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "update"},
   CellID->329670188]},
 "update"->{
  Cell[84057, 2152, 1235, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:329670188", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "update"},
   CellID->329670188]},
 "ID:1283917386"->{
  Cell[86332, 2227, 748, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1283917386", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1283917386]},
 "ID:14652249"->{
  Cell[87083, 2247, 810, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:14652249", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->14652249]},
 "ID:1012271681"->{
  Cell[87896, 2269, 2437, 75, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1012271681", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element]"},
   CellID->1012271681]},
 "NatPairs \[Element]"->{
  Cell[87896, 2269, 2437, 75, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1012271681", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element]"},
   CellID->1012271681]},
 "ID:988862829"->{
  Cell[112282, 3004, 742, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:988862829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->988862829]},
 "ID:16377100"->{
  Cell[113027, 3024, 1339, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:16377100", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element] ProdDomain"},
   CellID->16377100]},
 "NatPairs \[Element] ProdDomain"->{
  Cell[113027, 3024, 1339, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:16377100", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element] ProdDomain"},
   CellID->16377100]},
 "ID:136514135"->{
  Cell[114369, 3068, 1472, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:136514135", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element] ProdDomain DomainTuples"},
   CellID->136514135]},
 "NatPairs \[Element] ProdDomain DomainTuples"->{
  Cell[114369, 3068, 1472, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:136514135", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element] ProdDomain DomainTuples"},
   CellID->136514135]},
 "ID:1872634730"->{
  Cell[115844, 3116, 1986, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1872634730", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element] alternative", "IR"},
   CellID->1872634730]},
 "NatPairs \[Element] alternative"->{
  Cell[115844, 3116, 1986, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1872634730", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element] alternative", "IR"},
   CellID->1872634730]},
 "IR"->{
  Cell[115844, 3116, 1986, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1872634730", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element] alternative", "IR"},
   CellID->1872634730],
  Cell[117833, 3174, 1959, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1931599387", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs< \[Element] alternative", "IR"},
   CellID->1931599387],
  Cell[119795, 3236, 2024, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2124565744", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs\[LessEqual] \[Element] alternative", "IR"},
   CellID->2124565744],
  Cell[199230, 5488, 1116, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1202991311", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership pairs unary", "IR"},
   CellID->1202991311],
  Cell[203141, 5608, 1189, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:540671405", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership pairs binary", "IR"},
   CellID->540671405],
  Cell[207051, 5729, 859, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:794667329", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "allPairs 0", "IR"},
   CellID->794667329],
  Cell[207913, 5755, 1126, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1385740688", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership allPairs", "IR"},
   CellID->1385740688],
  Cell[250596, 6990, 1160, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:723173326", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership update", "IR"},
   CellID->723173326],
  Cell[268783, 7520, 1965, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1597870023", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed alternative", "IR"},
   CellID->1597870023],
  Cell[524671, 15207, 1456, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1148013153", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainSets \[ScriptCapitalG]", "IR"},
   CellID->1148013153],
  Cell[526130, 15254, 1464, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1668771272", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples \[ScriptCapitalG]", "IR"},
   CellID->1668771272],
  Cell[527597, 15301, 1491, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1928426019", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Sum] \[ScriptCapitalG]", "IR"},
   CellID->1928426019],
  Cell[529786, 15365, 1600, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:581321955", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "reduceM \[ScriptCapitalG]", "IR"},
   CellID->581321955],
  Cell[531389, 15411, 1595, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:405682622", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow] \[ScriptCapitalG] multiplier", "IR"},
   
   CellID->405682622],
  Cell[532987, 15463, 1613, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1165811909", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow] \[ScriptCapitalG]", "IR"},
   CellID->1165811909],
  Cell[534603, 15514, 1641, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:306221528", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[LeftRightArrow] \[ScriptCapitalG]", "IR"},
   CellID->306221528],
  Cell[536247, 15565, 1076, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:113746156", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isReducible \[ScriptCapitalG] multiplier", "IR"},
   CellID->113746156],
  Cell[537326, 15594, 1028, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1372373678", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isReducible \[ScriptCapitalG]", "IR"},
   CellID->1372373678],
  Cell[538357, 15622, 1051, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:12812135", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "symred \[ScriptCapitalG]", "IR"},
   CellID->12812135],
  Cell[539411, 15652, 1057, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:168109679", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "areCorrelative \[ScriptCapitalG]", "IR"},
   CellID->168109679],
  Cell[540471, 15680, 1255, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1427340407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isNTCR \[ScriptCapitalG]", "IR"},
   CellID->1427340407],
  Cell[541729, 15717, 1254, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:587977465", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isMNTCR \[ScriptCapitalG]", "IR"},
   CellID->587977465],
  Cell[542986, 15754, 1017, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:950195558", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isNTCR \[ScriptCapitalG] unary", "IR"},
   CellID->950195558],
  Cell[544006, 15782, 1024, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2144197915", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isMNTCR \[ScriptCapitalG] unary", "IR"},
   CellID->2144197915],
  Cell[545033, 15810, 1207, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1924053921", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "red \[ScriptCapitalG] set", "IR"},
   CellID->1924053921],
  Cell[546243, 15847, 1148, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1439528152", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Tilde] \[ScriptCapitalG]", "IR"},
   CellID->1439528152],
  Cell[778497, 22421, 2089, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:160961870", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdT alternative", "IR"},
   CellID->160961870],
  Cell[887467, 25365, 3415, 96, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1019394084", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] alternative I", "IR"},
   CellID->1019394084],
  Cell[890885, 25463, 2963, 82, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1458538511", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] alternative II", "IR"},
   CellID->1458538511],
  Cell[1496312, 41028, 3435, 92, 164, "FormalTextInputFormula",
   CellTags->{
    "ID:1379909256", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAux 3 alternative", "IR"},
   CellID->1379909256],
  Cell[2096171, 56400, 946, 26, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:439420249", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GB empty", "IR"},
   CellID->439420249],
  Cell[2103738, 56635, 868, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:448006251", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership GB", "IR"},
   CellID->448006251],
  Cell[2104609, 56660, 910, 26, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:739795809", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GB", "IR"},
   CellID->739795809]},
 "ID:1931599387"->{
  Cell[117833, 3174, 1959, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1931599387", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs< \[Element] alternative", "IR"},
   CellID->1931599387]},
 "NatPairs< \[Element] alternative"->{
  Cell[117833, 3174, 1959, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1931599387", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs< \[Element] alternative", "IR"},
   CellID->1931599387]},
 "ID:2124565744"->{
  Cell[119795, 3236, 2024, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2124565744", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs\[LessEqual] \[Element] alternative", "IR"},
   CellID->2124565744]},
 "NatPairs\[LessEqual] \[Element] alternative"->{
  Cell[119795, 3236, 2024, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2124565744", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs\[LessEqual] \[Element] alternative", "IR"},
   CellID->2124565744]},
 "ID:1917398594"->{
  Cell[121822, 3300, 1538, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1917398594", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs< \[Element] alternative II"},
   CellID->1917398594]},
 "NatPairs< \[Element] alternative II"->{
  Cell[121822, 3300, 1538, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1917398594", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs< \[Element] alternative II"},
   CellID->1917398594]},
 "ID:272474466"->{
  Cell[123363, 3349, 1649, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:272474466", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs\[LessEqual] \[Element] alternative II"},
   CellID->272474466]},
 "NatPairs\[LessEqual] \[Element] alternative II"->{
  Cell[123363, 3349, 1649, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:272474466", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs\[LessEqual] \[Element] alternative II"},
   CellID->272474466]},
 "ID:2005363738"->{
  Cell[125015, 3401, 1398, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2005363738", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs< \[LessEqual]"},
   CellID->2005363738]},
 "NatPairs< \[LessEqual]"->{
  Cell[125015, 3401, 1398, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2005363738", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs< \[LessEqual]"},
   CellID->2005363738]},
 "ID:1661715038"->{
  Cell[126416, 3448, 1524, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1661715038", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NatPairs< \[LessEqual]"},
   CellID->1661715038]},
 "DomainTuples NatPairs< \[LessEqual]"->{
  Cell[126416, 3448, 1524, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1661715038", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NatPairs< \[LessEqual]"},
   CellID->1661715038]},
 "ID:1020080047"->{
  Cell[127943, 3497, 726, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1020080047", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1020080047]},
 "ID:1327808237"->{
  Cell[128672, 3518, 1316, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1327808237", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs +1"},
   CellID->1327808237]},
 "NatPairs +1"->{
  Cell[128672, 3518, 1316, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1327808237", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs +1"},
   CellID->1327808237]},
 "ID:798140928"->{
  Cell[129991, 3562, 1426, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:798140928", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NatPairs +1"},
   CellID->798140928]},
 "DomainTuples NatPairs +1"->{
  Cell[129991, 3562, 1426, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:798140928", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NatPairs +1"},
   CellID->798140928]},
 "ID:2111464127"->{
  Cell[131420, 3608, 791, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2111464127", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->2111464127]},
 "ID:1834895057"->{
  Cell[132214, 3630, 1264, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1834895057", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[GreaterEqual]"},
   CellID->1834895057]},
 "NatPairs \[GreaterEqual]"->{
  Cell[132214, 3630, 1264, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1834895057", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[GreaterEqual]"},
   CellID->1834895057]},
 "ID:1703465728"->{
  Cell[133481, 3672, 1388, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1703465728", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NatPairs \[GreaterEqual]"},
   CellID->1703465728]},
 "DomainTuples NatPairs \[GreaterEqual]"->{
  Cell[133481, 3672, 1388, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1703465728", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NatPairs \[GreaterEqual]"},
   CellID->1703465728]},
 "Proof|ID:16377100"->{
  Cell[135026, 3726, 3491, 97, 70, "ProofInfo",
   CellTags->{"Proof|ID:16377100", "Proof|ID:16377100-1"},
   CellID->22673357]},
 "Proof|ID:16377100-1"->{
  Cell[135026, 3726, 3491, 97, 70, "ProofInfo",
   CellTags->{"Proof|ID:16377100", "Proof|ID:16377100-1"},
   CellID->22673357]},
 "Proof|ID:136514135"->{
  Cell[139036, 3842, 3771, 102, 70, "ProofInfo",
   CellTags->{"Proof|ID:136514135", "Proof|ID:136514135-1"},
   CellID->585962876]},
 "Proof|ID:136514135-1"->{
  Cell[139036, 3842, 3771, 102, 70, "ProofInfo",
   CellTags->{"Proof|ID:136514135", "Proof|ID:136514135-1"},
   CellID->585962876]},
 "Proof|ID:1872634730"->{
  Cell[143326, 3963, 5199, 130, 70, "ProofInfo",
   CellTags->{"Proof|ID:1872634730", "Proof|ID:1872634730-1"},
   CellID->1738547973]},
 "Proof|ID:1872634730-1"->{
  Cell[143326, 3963, 5199, 130, 70, "ProofInfo",
   CellTags->{"Proof|ID:1872634730", "Proof|ID:1872634730-1"},
   CellID->1738547973]},
 "Proof|ID:1931599387"->{
  Cell[149047, 4112, 4821, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:1931599387", "Proof|ID:1931599387-1"},
   CellID->60767287]},
 "Proof|ID:1931599387-1"->{
  Cell[149047, 4112, 4821, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:1931599387", "Proof|ID:1931599387-1"},
   CellID->60767287]},
 "Proof|ID:2124565744"->{
  Cell[154390, 4255, 4913, 125, 70, "ProofInfo",
   CellTags->{"Proof|ID:2124565744", "Proof|ID:2124565744-1"},
   CellID->2051885642]},
 "Proof|ID:2124565744-1"->{
  Cell[154390, 4255, 4913, 125, 70, "ProofInfo",
   CellTags->{"Proof|ID:2124565744", "Proof|ID:2124565744-1"},
   CellID->2051885642]},
 "Proof|ID:1917398594"->{
  Cell[159825, 4399, 4582, 125, 70, "ProofInfo",
   CellTags->{"Proof|ID:1917398594", "Proof|ID:1917398594-1"},
   CellID->1425710701]},
 "Proof|ID:1917398594-1"->{
  Cell[159825, 4399, 4582, 125, 70, "ProofInfo",
   CellTags->{"Proof|ID:1917398594", "Proof|ID:1917398594-1"},
   CellID->1425710701]},
 "Proof|ID:272474466"->{
  Cell[164927, 4543, 4654, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:272474466", "Proof|ID:272474466-1"},
   CellID->136931469]},
 "Proof|ID:272474466-1"->{
  Cell[164927, 4543, 4654, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:272474466", "Proof|ID:272474466-1"},
   CellID->136931469]},
 "Proof|ID:2005363738"->{
  Cell[170102, 4686, 3074, 89, 70, "ProofInfo",
   CellTags->{"Proof|ID:2005363738", "Proof|ID:2005363738-1"},
   CellID->911690168]},
 "Proof|ID:2005363738-1"->{
  Cell[170102, 4686, 3074, 89, 70, "ProofInfo",
   CellTags->{"Proof|ID:2005363738", "Proof|ID:2005363738-1"},
   CellID->911690168]},
 "Proof|ID:1661715038"->{
  Cell[173697, 4794, 3351, 95, 70, "ProofInfo",
   CellTags->{"Proof|ID:1661715038", "Proof|ID:1661715038-1"},
   CellID->1036844371]},
 "Proof|ID:1661715038-1"->{
  Cell[173697, 4794, 3351, 95, 70, "ProofInfo",
   CellTags->{"Proof|ID:1661715038", "Proof|ID:1661715038-1"},
   CellID->1036844371]},
 "Proof|ID:1327808237"->{
  Cell[177569, 4908, 3961, 105, 70, "ProofInfo",
   CellTags->{"Proof|ID:1327808237", "Proof|ID:1327808237-1"},
   CellID->2053606910]},
 "Proof|ID:1327808237-1"->{
  Cell[177569, 4908, 3961, 105, 70, "ProofInfo",
   CellTags->{"Proof|ID:1327808237", "Proof|ID:1327808237-1"},
   CellID->2053606910]},
 "Proof|ID:798140928"->{
  Cell[182051, 5032, 4205, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:798140928", "Proof|ID:798140928-1"},
   CellID->974929944]},
 "Proof|ID:798140928-1"->{
  Cell[182051, 5032, 4205, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:798140928", "Proof|ID:798140928-1"},
   CellID->974929944]},
 "Proof|ID:1834895057"->{
  Cell[186776, 5159, 4843, 123, 70, "ProofInfo",
   CellTags->{"Proof|ID:1834895057", "Proof|ID:1834895057-1"},
   CellID->1231124039]},
 "Proof|ID:1834895057-1"->{
  Cell[186776, 5159, 4843, 123, 70, "ProofInfo",
   CellTags->{"Proof|ID:1834895057", "Proof|ID:1834895057-1"},
   CellID->1231124039]},
 "Proof|ID:1703465728"->{
  Cell[192141, 5301, 5113, 129, 70, "ProofInfo",
   CellTags->{"Proof|ID:1703465728", "Proof|ID:1703465728-1"},
   CellID->1999261425]},
 "Proof|ID:1703465728-1"->{
  Cell[192141, 5301, 5113, 129, 70, "ProofInfo",
   CellTags->{"Proof|ID:1703465728", "Proof|ID:1703465728-1"},
   CellID->1999261425]},
 "ID:971135058"->{
  Cell[198273, 5464, 954, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:971135058", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->971135058]},
 "ID:1202991311"->{
  Cell[199230, 5488, 1116, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1202991311", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership pairs unary", "IR"},
   CellID->1202991311]},
 "membership pairs unary"->{
  Cell[199230, 5488, 1116, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1202991311", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership pairs unary", "IR"},
   CellID->1202991311]},
 "ID:1932739315"->{
  Cell[200349, 5523, 1522, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1932739315", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs unary content'"},
   CellID->1932739315]},
 "pairs unary content'"->{
  Cell[200349, 5523, 1522, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1932739315", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs unary content'"},
   CellID->1932739315]},
 "ID:1628109432"->{
  Cell[201874, 5568, 1264, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1628109432", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs unary content"},
   CellID->1628109432]},
 "pairs unary content"->{
  Cell[201874, 5568, 1264, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1628109432", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs unary content"},
   CellID->1628109432]},
 "ID:540671405"->{
  Cell[203141, 5608, 1189, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:540671405", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership pairs binary", "IR"},
   CellID->540671405]},
 "membership pairs binary"->{
  Cell[203141, 5608, 1189, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:540671405", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership pairs binary", "IR"},
   CellID->540671405]},
 "ID:881225230"->{
  Cell[204333, 5645, 1465, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:881225230", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs binary content'"},
   CellID->881225230]},
 "pairs binary content'"->{
  Cell[204333, 5645, 1465, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:881225230", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs binary content'"},
   CellID->881225230]},
 "ID:717573619"->{
  Cell[205801, 5689, 1247, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:717573619", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs binary content"},
   CellID->717573619]},
 "pairs binary content"->{
  Cell[205801, 5689, 1247, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:717573619", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs binary content"},
   CellID->717573619]},
 "ID:794667329"->{
  Cell[207051, 5729, 859, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:794667329", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "allPairs 0", "IR"},
   CellID->794667329]},
 "allPairs 0"->{
  Cell[207051, 5729, 859, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:794667329", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "allPairs 0", "IR"},
   CellID->794667329]},
 "ID:1385740688"->{
  Cell[207913, 5755, 1126, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1385740688", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership allPairs", "IR"},
   CellID->1385740688]},
 "membership allPairs"->{
  Cell[207913, 5755, 1126, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1385740688", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership allPairs", "IR"},
   CellID->1385740688]},
 "ID:1430860208"->{
  Cell[209042, 5790, 1232, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1430860208", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "allPairs content"},
   CellID->1430860208]},
 "allPairs content"->{
  Cell[209042, 5790, 1232, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1430860208", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "allPairs content"},
   CellID->1430860208]},
 "Proof|ID:1202991311"->{
  Cell[210429, 5838, 2859, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:1202991311", "Proof|ID:1202991311-1"},
   CellID->1926557973]},
 "Proof|ID:1202991311-1"->{
  Cell[210429, 5838, 2859, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:1202991311", "Proof|ID:1202991311-1"},
   CellID->1926557973]},
 "Proof|ID:1932739315"->{
  Cell[213810, 5940, 4716, 125, 70, "ProofInfo",
   CellTags->{"Proof|ID:1932739315", "Proof|ID:1932739315-1"},
   CellID->1207788650]},
 "Proof|ID:1932739315-1"->{
  Cell[213810, 5940, 4716, 125, 70, "ProofInfo",
   CellTags->{"Proof|ID:1932739315", "Proof|ID:1932739315-1"},
   CellID->1207788650]},
 "Proof|ID:1628109432"->{
  Cell[219047, 6084, 3584, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:1628109432", "Proof|ID:1628109432-1"},
   CellID->135860045]},
 "Proof|ID:1628109432-1"->{
  Cell[219047, 6084, 3584, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:1628109432", "Proof|ID:1628109432-1"},
   CellID->135860045]},
 "Proof|ID:540671405"->{
  Cell[223153, 6202, 3330, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:540671405", "Proof|ID:540671405-1"},
   CellID->1030769530]},
 "Proof|ID:540671405-1"->{
  Cell[223153, 6202, 3330, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:540671405", "Proof|ID:540671405-1"},
   CellID->1030769530]},
 "Proof|ID:881225230"->{
  Cell[227046, 6314, 4859, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:881225230", "Proof|ID:881225230-1"},
   CellID->960050214]},
 "Proof|ID:881225230-1"->{
  Cell[227046, 6314, 4859, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:881225230", "Proof|ID:881225230-1"},
   CellID->960050214]},
 "Proof|ID:717573619"->{
  Cell[232425, 6459, 4076, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:717573619", "Proof|ID:717573619-1"},
   CellID->1036460169]},
 "Proof|ID:717573619-1"->{
  Cell[232425, 6459, 4076, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:717573619", "Proof|ID:717573619-1"},
   CellID->1036460169]},
 "Proof|ID:794667329"->{
  Cell[237022, 6588, 1945, 62, 70, "ProofInfo",
   CellTags->{"Proof|ID:794667329", "Proof|ID:794667329-1"},
   CellID->1738668792]},
 "Proof|ID:794667329-1"->{
  Cell[237022, 6588, 1945, 62, 70, "ProofInfo",
   CellTags->{"Proof|ID:794667329", "Proof|ID:794667329-1"},
   CellID->1738668792]},
 "Proof|ID:1385740688"->{
  Cell[239488, 6669, 2885, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:1385740688", "Proof|ID:1385740688-1"},
   CellID->1882316559]},
 "Proof|ID:1385740688-1"->{
  Cell[239488, 6669, 2885, 83, 70, "ProofInfo",
   CellTags->{"Proof|ID:1385740688", "Proof|ID:1385740688-1"},
   CellID->1882316559]},
 "Proof|ID:1430860208"->{
  Cell[242894, 6771, 4035, 109, 70, "ProofInfo",
   CellTags->{"Proof|ID:1430860208", "Proof|ID:1430860208-1"},
   CellID->1876195014]},
 "Proof|ID:1430860208-1"->{
  Cell[242894, 6771, 4035, 109, 70, "ProofInfo",
   CellTags->{"Proof|ID:1430860208", "Proof|ID:1430860208-1"},
   CellID->1876195014]},
 "ID:963656485"->{
  Cell[247887, 6913, 823, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:963656485", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->963656485]},
 "ID:1750952092"->{
  Cell[248713, 6935, 1072, 30, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1750952092", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1750952092]},
 "ID:1500640019"->{
  Cell[249788, 6967, 805, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1500640019", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1500640019]},
 "ID:723173326"->{
  Cell[250596, 6990, 1160, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:723173326", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership update", "IR"},
   CellID->723173326]},
 "membership update"->{
  Cell[250596, 6990, 1160, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:723173326", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership update", "IR"},
   CellID->723173326]},
 "ID:392184336"->{
  Cell[251759, 7026, 1301, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:392184336", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "update content I"},
   CellID->392184336]},
 "update content I"->{
  Cell[251759, 7026, 1301, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:392184336", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "update content I"},
   CellID->392184336]},
 "ID:341058317"->{
  Cell[253063, 7073, 1174, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:341058317", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "update content II"},
   CellID->341058317]},
 "update content II"->{
  Cell[253063, 7073, 1174, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:341058317", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "update content II"},
   CellID->341058317]},
 "Proof|ID:723173326"->{
  Cell[254394, 7120, 3938, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:723173326", "Proof|ID:723173326-1"},
   CellID->1574692164]},
 "Proof|ID:723173326-1"->{
  Cell[254394, 7120, 3938, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:723173326", "Proof|ID:723173326-1"},
   CellID->1574692164]},
 "Proof|ID:392184336"->{
  Cell[258852, 7245, 3907, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:392184336", "Proof|ID:392184336-1"},
   CellID->393000099]},
 "Proof|ID:392184336-1"->{
  Cell[258852, 7245, 3907, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:392184336", "Proof|ID:392184336-1"},
   CellID->393000099]},
 "Proof|ID:341058317"->{
  Cell[263280, 7368, 4348, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:341058317", "Proof|ID:341058317-1"},
   CellID->1522358468]},
 "Proof|ID:341058317-1"->{
  Cell[263280, 7368, 4348, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:341058317", "Proof|ID:341058317-1"},
   CellID->1522358468]},
 "ID:1597870023"->{
  Cell[268783, 7520, 1965, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1597870023", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed alternative", "IR"},
   CellID->1597870023]},
 "processed alternative"->{
  Cell[268783, 7520, 1965, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1597870023", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed alternative", "IR"},
   CellID->1597870023]},
 "ID:40246499"->{
  Cell[270751, 7585, 1099, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:40246499", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed symmetric"},
   CellID->40246499]},
 "processed symmetric"->{
  Cell[270751, 7585, 1099, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:40246499", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed symmetric"},
   CellID->40246499]},
 "ID:1131819573"->{
  Cell[271853, 7619, 728, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1131819573", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1131819573]},
 "ID:1083804603"->{
  Cell[272584, 7639, 1106, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1083804603", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed allPairs"},
   CellID->1083804603]},
 "processed allPairs"->{
  Cell[272584, 7639, 1106, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1083804603", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed allPairs"},
   CellID->1083804603]},
 "ID:316507172"->{
  Cell[273693, 7673, 2454, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:316507172", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed update"},
   CellID->316507172]},
 "processed update"->{
  Cell[273693, 7673, 2454, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:316507172", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed update"},
   CellID->316507172]},
 "ID:1179241697"->{
  Cell[276150, 7748, 2158, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1179241697", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed prepend"},
   CellID->1179241697]},
 "processed prepend"->{
  Cell[276150, 7748, 2158, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1179241697", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed prepend"},
   CellID->1179241697]},
 "Proof|ID:1597870023"->{
  Cell[278465, 7821, 5020, 129, 70, "ProofInfo",
   CellTags->{"Proof|ID:1597870023", "Proof|ID:1597870023-1"},
   CellID->932327505]},
 "Proof|ID:1597870023-1"->{
  Cell[278465, 7821, 5020, 129, 70, "ProofInfo",
   CellTags->{"Proof|ID:1597870023", "Proof|ID:1597870023-1"},
   CellID->932327505]},
 "Proof|ID:40246499"->{
  Cell[284006, 7969, 3135, 86, 70, "ProofInfo",
   CellTags->{"Proof|ID:40246499", "Proof|ID:40246499-1"},
   CellID->1343087848]},
 "Proof|ID:40246499-1"->{
  Cell[284006, 7969, 3135, 86, 70, "ProofInfo",
   CellTags->{"Proof|ID:40246499", "Proof|ID:40246499-1"},
   CellID->1343087848]},
 "Proof|ID:1083804603"->{
  Cell[287661, 8074, 3849, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:1083804603", "Proof|ID:1083804603-1"},
   CellID->465417913]},
 "Proof|ID:1083804603-1"->{
  Cell[287661, 8074, 3849, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:1083804603", "Proof|ID:1083804603-1"},
   CellID->465417913]},
 "Proof|ID:316507172"->{
  Cell[292030, 8199, 6411, 158, 70, "ProofInfo",
   CellTags->{"Proof|ID:316507172", "Proof|ID:316507172-1"},
   CellID->660680371]},
 "Proof|ID:316507172-1"->{
  Cell[292030, 8199, 6411, 158, 70, "ProofInfo",
   CellTags->{"Proof|ID:316507172", "Proof|ID:316507172-1"},
   CellID->660680371]},
 "Proof|ID:1179241697"->{
  Cell[298962, 8376, 6163, 149, 70, "ProofInfo",
   CellTags->{"Proof|ID:1179241697", "Proof|ID:1179241697-1"},
   CellID->2041135352]},
 "Proof|ID:1179241697-1"->{
  Cell[298962, 8376, 6163, 149, 70, "ProofInfo",
   CellTags->{"Proof|ID:1179241697", "Proof|ID:1179241697-1"},
   CellID->2041135352]},
 "ID:1573502139"->{
  Cell[313515, 8786, 651, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1573502139", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1573502139]},
 "ID:377704681"->{
  Cell[314169, 8804, 789, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:377704681", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->377704681]},
 "ID:1468120413"->{
  Cell[314961, 8825, 1996, 37, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1468120413", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1468120413]},
 "ID:1774031579"->{
  Cell[316960, 8864, 1325, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1774031579", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing \[Element]"},
   CellID->1774031579]},
 "GroebnerRing \[Element]"->{
  Cell[316960, 8864, 1325, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1774031579", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing \[Element]"},
   CellID->1774031579]},
 "ID:182210603"->{
  Cell[318288, 8902, 855, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:182210603", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing 0"},
   CellID->182210603]},
 "GroebnerRing 0"->{
  Cell[318288, 8902, 855, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:182210603", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing 0"},
   CellID->182210603]},
 "ID:814933967"->{
  Cell[319146, 8927, 855, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:814933967", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing 1"},
   CellID->814933967]},
 "GroebnerRing 1"->{
  Cell[319146, 8927, 855, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:814933967", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing 1"},
   CellID->814933967]},
 "ID:1684568525"->{
  Cell[320004, 8952, 906, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1684568525", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing +"},
   CellID->1684568525]},
 "GroebnerRing +"->{
  Cell[320004, 8952, 906, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1684568525", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing +"},
   CellID->1684568525]},
 "ID:584522550"->{
  Cell[320913, 8979, 902, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:584522550", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing -"},
   CellID->584522550]},
 "GroebnerRing -"->{
  Cell[320913, 8979, 902, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:584522550", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing -"},
   CellID->584522550]},
 "ID:1458964616"->{
  Cell[321818, 9006, 908, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1458964616", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing - unary"},
   CellID->1458964616]},
 "GroebnerRing - unary"->{
  Cell[321818, 9006, 908, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1458964616", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing - unary"},
   CellID->1458964616]},
 "ID:1496488880"->{
  Cell[322729, 9033, 906, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1496488880", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing \[RawStar]"},
   CellID->1496488880]},
 "GroebnerRing \[RawStar]"->{
  Cell[322729, 9033, 906, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1496488880", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing \[RawStar]"},
   CellID->1496488880]},
 "ID:342394137"->{
  Cell[323638, 9060, 1432, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:342394137", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing \[Succeeds]"},
   CellID->342394137]},
 "GroebnerRing \[Succeeds]"->{
  Cell[323638, 9060, 1432, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:342394137", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing \[Succeeds]"},
   CellID->342394137]},
 "ID:1566608311"->{
  Cell[325073, 9105, 887, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1566608311", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing multN"},
   CellID->1566608311]},
 "GroebnerRing multN"->{
  Cell[325073, 9105, 887, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1566608311", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing multN"},
   CellID->1566608311]},
 "ID:628841792"->{
  Cell[325963, 9132, 1098, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:628841792", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing mult 1"},
   CellID->628841792]},
 "GroebnerRing mult 1"->{
  Cell[325963, 9132, 1098, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:628841792", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing mult 1"},
   CellID->628841792]},
 "ID:378340035"->{
  Cell[327064, 9166, 1120, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:378340035", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing mult 2"},
   CellID->378340035]},
 "GroebnerRing mult 2"->{
  Cell[327064, 9166, 1120, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:378340035", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing mult 2"},
   CellID->378340035]},
 "ID:822076309"->{
  Cell[328187, 9200, 1065, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:822076309", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lcrd"},
   CellID->822076309]},
 "GroebnerRing lcrd"->{
  Cell[328187, 9200, 1065, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:822076309", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lcrd"},
   CellID->822076309]},
 "ID:849156542"->{
  Cell[329255, 9231, 969, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:849156542", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing rdm"},
   CellID->849156542]},
 "GroebnerRing rdm"->{
  Cell[329255, 9231, 969, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:849156542", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing rdm"},
   CellID->849156542]},
 "ID:809363439"->{
  Cell[330227, 9260, 890, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:809363439", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lcrd unary"},
   CellID->809363439]},
 "GroebnerRing lcrd unary"->{
  Cell[330227, 9260, 890, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:809363439", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lcrd unary"},
   CellID->809363439]},
 "ID:1671864047"->{
  Cell[331120, 9286, 987, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1671864047", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing cpm"},
   CellID->1671864047]},
 "GroebnerRing cpm"->{
  Cell[331120, 9286, 987, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1671864047", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing cpm"},
   CellID->1671864047]},
 "ID:558323540"->{
  Cell[332514, 9332, 1070, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:558323540", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing rd"},
   CellID->558323540]},
 "GroebnerRing rd"->{
  Cell[332514, 9332, 1070, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:558323540", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing rd"},
   CellID->558323540]},
 "ID:1403530194"->{
  Cell[334129, 9385, 1675, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1403530194", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing rdT"},
   CellID->1403530194]},
 "GroebnerRing rdT"->{
  Cell[334129, 9385, 1675, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1403530194", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing rdT"},
   CellID->1403530194]},
 "ID:689058884"->{
  Cell[336453, 9456, 1732, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:689058884", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing trd"},
   CellID->689058884]},
 "GroebnerRing trd"->{
  Cell[336453, 9456, 1732, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:689058884", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing trd"},
   CellID->689058884]},
 "ID:1528624471"->{
  Cell[339865, 9557, 2982, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1528624471", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lmTuple"},
   CellID->1528624471]},
 "GroebnerRing lmTuple"->{
  Cell[339865, 9557, 2982, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1528624471", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lmTuple"},
   CellID->1528624471]},
 "ID:1192872489"->{
  Cell[342850, 9624, 986, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1192872489", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lmTupleAux 1"},
   CellID->1192872489]},
 "GroebnerRing lmTupleAux 1"->{
  Cell[342850, 9624, 986, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1192872489", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lmTupleAux 1"},
   CellID->1192872489]},
 "ID:1479034574"->{
  Cell[343839, 9653, 1942, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1479034574", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lmTupleAux 2"},
   CellID->1479034574]},
 "GroebnerRing lmTupleAux 2"->{
  Cell[343839, 9653, 1942, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1479034574", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lmTupleAux 2"},
   CellID->1479034574]},
 "ID:1529751135"->{
  Cell[347635, 9767, 1809, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1529751135", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing cpmTuple 1"},
   CellID->1529751135]},
 "GroebnerRing cpmTuple 1"->{
  Cell[347635, 9767, 1809, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1529751135", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing cpmTuple 1"},
   CellID->1529751135]},
 "ID:551185449"->{
  Cell[351052, 9876, 1705, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:551185449", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing cpmTuple 2"},
   CellID->551185449]},
 "GroebnerRing cpmTuple 2"->{
  Cell[351052, 9876, 1705, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:551185449", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing cpmTuple 2"},
   CellID->551185449]},
 "ID:1533920470"->{
  Cell[353203, 9938, 1246, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1533920470", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GB"},
   CellID->1533920470]},
 "GroebnerRing GB"->{
  Cell[353203, 9938, 1246, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1533920470", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GB"},
   CellID->1533920470]},
 "ID:1976234692"->{
  Cell[354452, 9972, 1112, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1976234692", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GBAux 1"},
   CellID->1976234692]},
 "GroebnerRing GBAux 1"->{
  Cell[354452, 9972, 1112, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1976234692", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GBAux 1"},
   CellID->1976234692]},
 "ID:144716839"->{
  Cell[355567, 10003, 2792, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:144716839", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GBAux 2"},
   CellID->144716839]},
 "GroebnerRing GBAux 2"->{
  Cell[355567, 10003, 2792, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:144716839", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GBAux 2"},
   CellID->144716839]},
 "ID:189044945"->{
  Cell[358362, 10078, 4571, 115, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:189044945", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GBAux 3"},
   CellID->189044945]},
 "GroebnerRing GBAux 3"->{
  Cell[358362, 10078, 4571, 115, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:189044945", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GBAux 3"},
   CellID->189044945]},
 "ID:1766960499"->{
  Cell[364060, 10227, 3441, 90, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1766960499", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing chainCrit"},
   CellID->1766960499]},
 "GroebnerRing chainCrit"->{
  Cell[364060, 10227, 3441, 90, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1766960499", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing chainCrit"},
   CellID->1766960499]},
 "ID:1777604038"->{
  Cell[369194, 10368, 1132, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1777604038", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing ard"},
   CellID->1777604038]},
 "GroebnerRing ard"->{
  Cell[369194, 10368, 1132, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1777604038", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing ard"},
   CellID->1777604038]},
 "ID:620529741"->{
  Cell[370329, 10400, 1161, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:620529741", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing ardAux 1"},
   CellID->620529741]},
 "GroebnerRing ardAux 1"->{
  Cell[370329, 10400, 1161, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:620529741", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing ardAux 1"},
   CellID->620529741]},
 "ID:174904624"->{
  Cell[371493, 10431, 2316, 71, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:174904624", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing ardAux 2"},
   CellID->174904624]},
 "GroebnerRing ardAux 2"->{
  Cell[371493, 10431, 2316, 71, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:174904624", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing ardAux 2"},
   CellID->174904624]},
 "ID:336511876"->{
  Cell[377682, 10622, 1036, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:336511876", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalQ]"},
   CellID->336511876]},
 "\[ScriptCapitalG] \[ScriptCapitalQ]"->{
  Cell[377682, 10622, 1036, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:336511876", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalQ]"},
   CellID->336511876]},
 "ID:193404746"->{
  Cell[392232, 11064, 999, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:193404746", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]"},
   CellID->193404746]},
 "\[ScriptCapitalG] \[ScriptCapitalZ]"->{
  Cell[392232, 11064, 999, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:193404746", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]"},
   CellID->193404746]},
 "ID:1821700886"->{
  Cell[407831, 11552, 968, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1821700886", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]24"},
   CellID->1821700886]},
 "\[ScriptCapitalG] \[ScriptCapitalZ]24"->{
  Cell[407831, 11552, 968, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1821700886", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]24"},
   CellID->1821700886]},
 "ID:1158015694"->{
  Cell[424655, 12079, 1211, 28, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1158015694", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalQ]xy"},
   CellID->1158015694]},
 "\[ScriptCapitalG] \[ScriptCapitalQ]xy"->{
  Cell[424655, 12079, 1211, 28, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1158015694", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalQ]xy"},
   CellID->1158015694]},
 "ID:1921709458"->{
  Cell[457422, 13081, 1229, 28, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1921709458", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalZ]xy"},
   CellID->1921709458]},
 "\[ScriptCapitalZ]xy"->{
  Cell[457422, 13081, 1229, 28, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1921709458", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalZ]xy"},
   CellID->1921709458]},
 "ID:281007089"->{
  Cell[458654, 13111, 1203, 28, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:281007089", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]xy"},
   CellID->281007089]},
 "\[ScriptCapitalG] \[ScriptCapitalZ]xy"->{
  Cell[458654, 13111, 1203, 28, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:281007089", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]xy"},
   CellID->281007089]},
 "ID:1512392075"->{
  Cell[481261, 13836, 998, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1512392075", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalZ]24xy"},
   CellID->1512392075]},
 "\[ScriptCapitalZ]24xy"->{
  Cell[481261, 13836, 998, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1512392075", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalZ]24xy"},
   CellID->1512392075]},
 "ID:543229319"->{
  Cell[482262, 13863, 1045, 27, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:543229319", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]24xy"},
   CellID->543229319]},
 "\[ScriptCapitalG] \[ScriptCapitalZ]24xy"->{
  Cell[482262, 13863, 1045, 27, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:543229319", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]24xy"},
   CellID->543229319]},
 "ID:280174026"->{
  Cell[517693, 14993, 965, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:280174026", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isCommRing1"},
   CellID->280174026]},
 "\[ScriptCapitalG] isCommRing1"->{
  Cell[517693, 14993, 965, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:280174026", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isCommRing1"},
   CellID->280174026]},
 "Proof|ID:280174026"->{
  Cell[518815, 15029, 2376, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:280174026", "Proof|ID:280174026-1"},
   CellID->79132233]},
 "Proof|ID:280174026-1"->{
  Cell[518815, 15029, 2376, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:280174026", "Proof|ID:280174026-1"},
   CellID->79132233]},
 "ID:1057748777"->{
  Cell[522070, 15126, 712, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1057748777", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1057748777]},
 "ID:464516108"->{
  Cell[522824, 15149, 743, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:464516108", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->464516108]},
 "ID:1148013153"->{
  Cell[524671, 15207, 1456, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1148013153", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainSets \[ScriptCapitalG]", "IR"},
   CellID->1148013153]},
 "DomainSets \[ScriptCapitalG]"->{
  Cell[524671, 15207, 1456, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1148013153", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainSets \[ScriptCapitalG]", "IR"},
   CellID->1148013153]},
 "ID:1668771272"->{
  Cell[526130, 15254, 1464, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1668771272", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples \[ScriptCapitalG]", "IR"},
   CellID->1668771272]},
 "DomainTuples \[ScriptCapitalG]"->{
  Cell[526130, 15254, 1464, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1668771272", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples \[ScriptCapitalG]", "IR"},
   CellID->1668771272]},
 "ID:1928426019"->{
  Cell[527597, 15301, 1491, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1928426019", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Sum] \[ScriptCapitalG]", "IR"},
   CellID->1928426019]},
 "\[Sum] \[ScriptCapitalG]"->{
  Cell[527597, 15301, 1491, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1928426019", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Sum] \[ScriptCapitalG]", "IR"},
   CellID->1928426019]},
 "ID:546598180"->{
  Cell[529091, 15346, 692, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:546598180", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->546598180]},
 "ID:581321955"->{
  Cell[529786, 15365, 1600, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:581321955", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "reduceM \[ScriptCapitalG]", "IR"},
   CellID->581321955]},
 "reduceM \[ScriptCapitalG]"->{
  Cell[529786, 15365, 1600, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:581321955", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "reduceM \[ScriptCapitalG]", "IR"},
   CellID->581321955]},
 "ID:405682622"->{
  Cell[531389, 15411, 1595, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:405682622", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow] \[ScriptCapitalG] multiplier", "IR"},
   
   CellID->405682622]},
 "\[RightArrow] \[ScriptCapitalG] multiplier"->{
  Cell[531389, 15411, 1595, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:405682622", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow] \[ScriptCapitalG] multiplier", "IR"},
   
   CellID->405682622]},
 "ID:1165811909"->{
  Cell[532987, 15463, 1613, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1165811909", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow] \[ScriptCapitalG]", "IR"},
   CellID->1165811909]},
 "\[RightArrow] \[ScriptCapitalG]"->{
  Cell[532987, 15463, 1613, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1165811909", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow] \[ScriptCapitalG]", "IR"},
   CellID->1165811909]},
 "ID:306221528"->{
  Cell[534603, 15514, 1641, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:306221528", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[LeftRightArrow] \[ScriptCapitalG]", "IR"},
   CellID->306221528]},
 "\[LeftRightArrow] \[ScriptCapitalG]"->{
  Cell[534603, 15514, 1641, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:306221528", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[LeftRightArrow] \[ScriptCapitalG]", "IR"},
   CellID->306221528]},
 "ID:113746156"->{
  Cell[536247, 15565, 1076, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:113746156", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isReducible \[ScriptCapitalG] multiplier", "IR"},
   CellID->113746156]},
 "isReducible \[ScriptCapitalG] multiplier"->{
  Cell[536247, 15565, 1076, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:113746156", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isReducible \[ScriptCapitalG] multiplier", "IR"},
   CellID->113746156]},
 "ID:1372373678"->{
  Cell[537326, 15594, 1028, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1372373678", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isReducible \[ScriptCapitalG]", "IR"},
   CellID->1372373678]},
 "isReducible \[ScriptCapitalG]"->{
  Cell[537326, 15594, 1028, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1372373678", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isReducible \[ScriptCapitalG]", "IR"},
   CellID->1372373678]},
 "ID:12812135"->{
  Cell[538357, 15622, 1051, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:12812135", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "symred \[ScriptCapitalG]", "IR"},
   CellID->12812135]},
 "symred \[ScriptCapitalG]"->{
  Cell[538357, 15622, 1051, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:12812135", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "symred \[ScriptCapitalG]", "IR"},
   CellID->12812135]},
 "ID:168109679"->{
  Cell[539411, 15652, 1057, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:168109679", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "areCorrelative \[ScriptCapitalG]", "IR"},
   CellID->168109679]},
 "areCorrelative \[ScriptCapitalG]"->{
  Cell[539411, 15652, 1057, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:168109679", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "areCorrelative \[ScriptCapitalG]", "IR"},
   CellID->168109679]},
 "ID:1427340407"->{
  Cell[540471, 15680, 1255, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1427340407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isNTCR \[ScriptCapitalG]", "IR"},
   CellID->1427340407]},
 "isNTCR \[ScriptCapitalG]"->{
  Cell[540471, 15680, 1255, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1427340407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isNTCR \[ScriptCapitalG]", "IR"},
   CellID->1427340407]},
 "ID:587977465"->{
  Cell[541729, 15717, 1254, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:587977465", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isMNTCR \[ScriptCapitalG]", "IR"},
   CellID->587977465]},
 "isMNTCR \[ScriptCapitalG]"->{
  Cell[541729, 15717, 1254, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:587977465", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isMNTCR \[ScriptCapitalG]", "IR"},
   CellID->587977465]},
 "ID:950195558"->{
  Cell[542986, 15754, 1017, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:950195558", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isNTCR \[ScriptCapitalG] unary", "IR"},
   CellID->950195558]},
 "isNTCR \[ScriptCapitalG] unary"->{
  Cell[542986, 15754, 1017, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:950195558", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isNTCR \[ScriptCapitalG] unary", "IR"},
   CellID->950195558]},
 "ID:2144197915"->{
  Cell[544006, 15782, 1024, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2144197915", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isMNTCR \[ScriptCapitalG] unary", "IR"},
   CellID->2144197915]},
 "isMNTCR \[ScriptCapitalG] unary"->{
  Cell[544006, 15782, 1024, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2144197915", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isMNTCR \[ScriptCapitalG] unary", "IR"},
   CellID->2144197915]},
 "ID:1924053921"->{
  Cell[545033, 15810, 1207, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1924053921", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "red \[ScriptCapitalG] set", "IR"},
   CellID->1924053921]},
 "red \[ScriptCapitalG] set"->{
  Cell[545033, 15810, 1207, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1924053921", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "red \[ScriptCapitalG] set", "IR"},
   CellID->1924053921]},
 "ID:1439528152"->{
  Cell[546243, 15847, 1148, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1439528152", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Tilde] \[ScriptCapitalG]", "IR"},
   CellID->1439528152]},
 "\[Tilde] \[ScriptCapitalG]"->{
  Cell[546243, 15847, 1148, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1439528152", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Tilde] \[ScriptCapitalG]", "IR"},
   CellID->1439528152]},
 "Proof|ID:1148013153"->{
  Cell[547548, 15892, 3649, 100, 70, "ProofInfo",
   CellTags->{"Proof|ID:1148013153", "Proof|ID:1148013153-1"},
   CellID->1967681402]},
 "Proof|ID:1148013153-1"->{
  Cell[547548, 15892, 3649, 100, 70, "ProofInfo",
   CellTags->{"Proof|ID:1148013153", "Proof|ID:1148013153-1"},
   CellID->1967681402]},
 "Proof|ID:1668771272"->{
  Cell[551719, 16011, 3660, 100, 70, "ProofInfo",
   CellTags->{"Proof|ID:1668771272", "Proof|ID:1668771272-1"},
   CellID->525775183]},
 "Proof|ID:1668771272-1"->{
  Cell[551719, 16011, 3660, 100, 70, "ProofInfo",
   CellTags->{"Proof|ID:1668771272", "Proof|ID:1668771272-1"},
   CellID->525775183]},
 "Proof|ID:1928426019"->{
  Cell[555900, 16130, 4913, 129, 70, "ProofInfo",
   CellTags->{"Proof|ID:1928426019", "Proof|ID:1928426019-1"},
   CellID->33419350]},
 "Proof|ID:1928426019-1"->{
  Cell[555900, 16130, 4913, 129, 70, "ProofInfo",
   CellTags->{"Proof|ID:1928426019", "Proof|ID:1928426019-1"},
   CellID->33419350]},
 "Proof|ID:581321955"->{
  Cell[561334, 16278, 5328, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:581321955", "Proof|ID:581321955-1"},
   CellID->1728877054]},
 "Proof|ID:581321955-1"->{
  Cell[561334, 16278, 5328, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:581321955", "Proof|ID:581321955-1"},
   CellID->1728877054]},
 "Proof|ID:405682622"->{
  Cell[567182, 16429, 5297, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:405682622", "Proof|ID:405682622-1"},
   CellID->109823534]},
 "Proof|ID:405682622-1"->{
  Cell[567182, 16429, 5297, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:405682622", "Proof|ID:405682622-1"},
   CellID->109823534]},
 "Proof|ID:1165811909"->{
  Cell[573000, 16582, 4958, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:1165811909", "Proof|ID:1165811909-1"},
   CellID->679771556]},
 "Proof|ID:1165811909-1"->{
  Cell[573000, 16582, 4958, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:1165811909", "Proof|ID:1165811909-1"},
   CellID->679771556]},
 "Proof|ID:306221528"->{
  Cell[578479, 16729, 5004, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:306221528", "Proof|ID:306221528-1"},
   CellID->52377820]},
 "Proof|ID:306221528-1"->{
  Cell[578479, 16729, 5004, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:306221528", "Proof|ID:306221528-1"},
   CellID->52377820]},
 "Proof|ID:113746156"->{
  Cell[584003, 16876, 4030, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:113746156", "Proof|ID:113746156-1"},
   CellID->1673450432]},
 "Proof|ID:113746156-1"->{
  Cell[584003, 16876, 4030, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:113746156", "Proof|ID:113746156-1"},
   CellID->1673450432]},
 "Proof|ID:1372373678"->{
  Cell[588554, 17001, 3663, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:1372373678", "Proof|ID:1372373678-1"},
   CellID->1812050497]},
 "Proof|ID:1372373678-1"->{
  Cell[588554, 17001, 3663, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:1372373678", "Proof|ID:1372373678-1"},
   CellID->1812050497]},
 "Proof|ID:12812135"->{
  Cell[592739, 17119, 4295, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:12812135", "Proof|ID:12812135-1"},
   CellID->2042538189]},
 "Proof|ID:12812135-1"->{
  Cell[592739, 17119, 4295, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:12812135", "Proof|ID:12812135-1"},
   CellID->2042538189]},
 "Proof|ID:168109679"->{
  Cell[597554, 17248, 4018, 105, 70, "ProofInfo",
   CellTags->{"Proof|ID:168109679", "Proof|ID:168109679-1"},
   CellID->1279405920]},
 "Proof|ID:168109679-1"->{
  Cell[597554, 17248, 4018, 105, 70, "ProofInfo",
   CellTags->{"Proof|ID:168109679", "Proof|ID:168109679-1"},
   CellID->1279405920]},
 "Proof|ID:1427340407"->{
  Cell[602093, 17372, 4647, 117, 70, "ProofInfo",
   CellTags->{"Proof|ID:1427340407", "Proof|ID:1427340407-1"},
   CellID->1645359805]},
 "Proof|ID:1427340407-1"->{
  Cell[602093, 17372, 4647, 117, 70, "ProofInfo",
   CellTags->{"Proof|ID:1427340407", "Proof|ID:1427340407-1"},
   CellID->1645359805]},
 "Proof|ID:587977465"->{
  Cell[607261, 17508, 4643, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:587977465", "Proof|ID:587977465-1"},
   CellID->429530329]},
 "Proof|ID:587977465-1"->{
  Cell[607261, 17508, 4643, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:587977465", "Proof|ID:587977465-1"},
   CellID->429530329]},
 "Proof|ID:950195558"->{
  Cell[612423, 17643, 3644, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:950195558", "Proof|ID:950195558-1"},
   CellID->1606573343]},
 "Proof|ID:950195558-1"->{
  Cell[612423, 17643, 3644, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:950195558", "Proof|ID:950195558-1"},
   CellID->1606573343]},
 "Proof|ID:2144197915"->{
  Cell[616587, 17761, 3658, 100, 70, "ProofInfo",
   CellTags->{"Proof|ID:2144197915", "Proof|ID:2144197915-1"},
   CellID->874871898]},
 "Proof|ID:2144197915-1"->{
  Cell[616587, 17761, 3658, 100, 70, "ProofInfo",
   CellTags->{"Proof|ID:2144197915", "Proof|ID:2144197915-1"},
   CellID->874871898]},
 "Proof|ID:1924053921"->{
  Cell[620767, 17880, 3877, 103, 70, "ProofInfo",
   CellTags->{"Proof|ID:1924053921", "Proof|ID:1924053921-1"},
   CellID->23710079]},
 "Proof|ID:1924053921-1"->{
  Cell[620767, 17880, 3877, 103, 70, "ProofInfo",
   CellTags->{"Proof|ID:1924053921", "Proof|ID:1924053921-1"},
   CellID->23710079]},
 "Proof|ID:1439528152"->{
  Cell[625164, 18002, 3434, 93, 70, "ProofInfo",
   CellTags->{"Proof|ID:1439528152", "Proof|ID:1439528152-1"},
   CellID->1503880702]},
 "Proof|ID:1439528152-1"->{
  Cell[625164, 18002, 3434, 93, 70, "ProofInfo",
   CellTags->{"Proof|ID:1439528152", "Proof|ID:1439528152-1"},
   CellID->1503880702]},
 "ID:1674560446"->{
  Cell[629817, 18138, 1044, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1674560446", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isPartIrreflOrder \[ScriptCapitalG]"},
   CellID->1674560446]},
 "isPartIrreflOrder \[ScriptCapitalG]"->{
  Cell[629817, 18138, 1044, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1674560446", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isPartIrreflOrder \[ScriptCapitalG]"},
   CellID->1674560446]},
 "ID:167203289"->{
  Cell[630864, 18170, 1025, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:167203289", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isNoetherian \[ScriptCapitalG]"},
   CellID->167203289]},
 "isNoetherian \[ScriptCapitalG]"->{
  Cell[630864, 18170, 1025, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:167203289", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isNoetherian \[ScriptCapitalG]"},
   CellID->167203289]},
 "ID:569855140"->{
  Cell[631892, 18202, 821, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:569855140", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "Multipliers \[ScriptCapitalG]"},
   CellID->569855140]},
 "Multipliers \[ScriptCapitalG]"->{
  Cell[631892, 18202, 821, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:569855140", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "Multipliers \[ScriptCapitalG]"},
   CellID->569855140]},
 "ID:180032137"->{
  Cell[632716, 18225, 791, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:180032137", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R0 \[ScriptCapitalG]"},
   CellID->180032137]},
 "R0 \[ScriptCapitalG]"->{
  Cell[632716, 18225, 791, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:180032137", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R0 \[ScriptCapitalG]"},
   CellID->180032137]},
 "ID:428662253"->{
  Cell[633510, 18247, 793, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:428662253", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R1 \[ScriptCapitalG]"},
   CellID->428662253]},
 "R1 \[ScriptCapitalG]"->{
  Cell[633510, 18247, 793, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:428662253", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R1 \[ScriptCapitalG]"},
   CellID->428662253]},
 "ID:530735576"->{
  Cell[634306, 18269, 792, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:530735576", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R2 \[ScriptCapitalG]"},
   CellID->530735576]},
 "R2 \[ScriptCapitalG]"->{
  Cell[634306, 18269, 792, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:530735576", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R2 \[ScriptCapitalG]"},
   CellID->530735576]},
 "ID:1081126646"->{
  Cell[635101, 18291, 794, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1081126646", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R3 \[ScriptCapitalG]"},
   CellID->1081126646]},
 "R3 \[ScriptCapitalG]"->{
  Cell[635101, 18291, 794, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1081126646", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R3 \[ScriptCapitalG]"},
   CellID->1081126646]},
 "ID:1591716135"->{
  Cell[635898, 18313, 791, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1591716135", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R4 \[ScriptCapitalG]"},
   CellID->1591716135]},
 "R4 \[ScriptCapitalG]"->{
  Cell[635898, 18313, 791, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1591716135", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R4 \[ScriptCapitalG]"},
   CellID->1591716135]},
 "ID:960649057"->{
  Cell[636692, 18335, 791, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:960649057", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R5 \[ScriptCapitalG]"},
   CellID->960649057]},
 "R5 \[ScriptCapitalG]"->{
  Cell[636692, 18335, 791, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:960649057", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R5 \[ScriptCapitalG]"},
   CellID->960649057]},
 "ID:227638759"->{
  Cell[637486, 18357, 795, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:227638759", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R6 \[ScriptCapitalG]"},
   CellID->227638759]},
 "R6 \[ScriptCapitalG]"->{
  Cell[637486, 18357, 795, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:227638759", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R6 \[ScriptCapitalG]"},
   CellID->227638759]},
 "ID:2067143528"->{
  Cell[638284, 18379, 796, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2067143528", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R7 \[ScriptCapitalG]"},
   CellID->2067143528]},
 "R7 \[ScriptCapitalG]"->{
  Cell[638284, 18379, 796, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2067143528", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R7 \[ScriptCapitalG]"},
   CellID->2067143528]},
 "ID:1699723496"->{
  Cell[639083, 18401, 796, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1699723496", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R8 \[ScriptCapitalG]"},
   CellID->1699723496]},
 "R8 \[ScriptCapitalG]"->{
  Cell[639083, 18401, 796, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1699723496", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R8 \[ScriptCapitalG]"},
   CellID->1699723496]},
 "Proof|ID:1674560446"->{
  Cell[640037, 18432, 2856, 81, 70, "ProofInfo",
   CellTags->{"Proof|ID:1674560446", "Proof|ID:1674560446-1"},
   CellID->997501851]},
 "Proof|ID:1674560446-1"->{
  Cell[640037, 18432, 2856, 81, 70, "ProofInfo",
   CellTags->{"Proof|ID:1674560446", "Proof|ID:1674560446-1"},
   CellID->997501851]},
 "Proof|ID:167203289"->{
  Cell[643457, 18533, 2818, 79, 70, "ProofInfo",
   CellTags->{"Proof|ID:167203289", "Proof|ID:167203289-1"},
   CellID->24119846]},
 "Proof|ID:167203289-1"->{
  Cell[643457, 18533, 2818, 79, 70, "ProofInfo",
   CellTags->{"Proof|ID:167203289", "Proof|ID:167203289-1"},
   CellID->24119846]},
 "Proof|ID:569855140"->{
  Cell[646796, 18631, 2386, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:569855140", "Proof|ID:569855140-1"},
   CellID->1448471252]},
 "Proof|ID:569855140-1"->{
  Cell[646796, 18631, 2386, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:569855140", "Proof|ID:569855140-1"},
   CellID->1448471252]},
 "Proof|ID:180032137"->{
  Cell[649703, 18718, 2349, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:180032137", "Proof|ID:180032137-1"},
   CellID->270496012]},
 "Proof|ID:180032137-1"->{
  Cell[649703, 18718, 2349, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:180032137", "Proof|ID:180032137-1"},
   CellID->270496012]},
 "Proof|ID:428662253"->{
  Cell[652572, 18805, 2350, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:428662253", "Proof|ID:428662253-1"},
   CellID->1471451373]},
 "Proof|ID:428662253-1"->{
  Cell[652572, 18805, 2350, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:428662253", "Proof|ID:428662253-1"},
   CellID->1471451373]},
 "Proof|ID:530735576"->{
  Cell[655443, 18892, 2349, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:530735576", "Proof|ID:530735576-1"},
   CellID->457832475]},
 "Proof|ID:530735576-1"->{
  Cell[655443, 18892, 2349, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:530735576", "Proof|ID:530735576-1"},
   CellID->457832475]},
 "Proof|ID:1081126646"->{
  Cell[658355, 18980, 2358, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1081126646", "Proof|ID:1081126646-1"},
   CellID->345429782]},
 "Proof|ID:1081126646-1"->{
  Cell[658355, 18980, 2358, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1081126646", "Proof|ID:1081126646-1"},
   CellID->345429782]},
 "Proof|ID:1591716135"->{
  Cell[661235, 19068, 2359, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1591716135", "Proof|ID:1591716135-1"},
   CellID->2122576129]},
 "Proof|ID:1591716135-1"->{
  Cell[661235, 19068, 2359, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1591716135", "Proof|ID:1591716135-1"},
   CellID->2122576129]},
 "Proof|ID:960649057"->{
  Cell[664116, 19156, 2350, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:960649057", "Proof|ID:960649057-1"},
   CellID->1652070226]},
 "Proof|ID:960649057-1"->{
  Cell[664116, 19156, 2350, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:960649057", "Proof|ID:960649057-1"},
   CellID->1652070226]},
 "Proof|ID:227638759"->{
  Cell[666987, 19243, 2349, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:227638759", "Proof|ID:227638759-1"},
   CellID->300639375]},
 "Proof|ID:227638759-1"->{
  Cell[666987, 19243, 2349, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:227638759", "Proof|ID:227638759-1"},
   CellID->300639375]},
 "Proof|ID:2067143528"->{
  Cell[669856, 19330, 2358, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:2067143528", "Proof|ID:2067143528-1"},
   CellID->271136742]},
 "Proof|ID:2067143528-1"->{
  Cell[669856, 19330, 2358, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:2067143528", "Proof|ID:2067143528-1"},
   CellID->271136742]},
 "Proof|ID:1699723496"->{
  Cell[672736, 19418, 2359, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1699723496", "Proof|ID:1699723496-1"},
   CellID->1455152719]},
 "Proof|ID:1699723496-1"->{
  Cell[672736, 19418, 2359, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1699723496", "Proof|ID:1699723496-1"},
   CellID->1455152719]},
 "ID:2054051596"->{
  Cell[676070, 19520, 889, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2054051596", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isReductionRing"},
   CellID->2054051596]},
 "\[ScriptCapitalG] isReductionRing"->{
  Cell[676070, 19520, 889, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2054051596", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isReductionRing"},
   CellID->2054051596]},
 "Proof|ID:2054051596"->{
  Cell[677117, 19553, 2427, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:2054051596", "Proof|ID:2054051596-1"},
   CellID->1699885972]},
 "Proof|ID:2054051596-1"->{
  Cell[677117, 19553, 2427, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:2054051596", "Proof|ID:2054051596-1"},
   CellID->1699885972]},
 "ID:1031954878"->{
  Cell[680367, 19653, 769, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1031954878", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1031954878]},
 "ID:660860156"->{
  Cell[681179, 19678, 743, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:660860156", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->660860156]},
 "ID:939639226"->{
  Cell[682439, 19717, 790, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:939639226", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R9 \[ScriptCapitalG]"},
   CellID->939639226]},
 "R9 \[ScriptCapitalG]"->{
  Cell[682439, 19717, 790, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:939639226", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R9 \[ScriptCapitalG]"},
   CellID->939639226]},
 "ID:37694441"->{
  Cell[683232, 19739, 793, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:37694441", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R10 \[ScriptCapitalG]"},
   CellID->37694441]},
 "R10 \[ScriptCapitalG]"->{
  Cell[683232, 19739, 793, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:37694441", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R10 \[ScriptCapitalG]"},
   CellID->37694441]},
 "ID:576375692"->{
  Cell[684028, 19761, 796, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:576375692", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R11 \[ScriptCapitalG]"},
   CellID->576375692]},
 "R11 \[ScriptCapitalG]"->{
  Cell[684028, 19761, 796, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:576375692", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R11 \[ScriptCapitalG]"},
   CellID->576375692]},
 "ID:1226510984"->{
  Cell[684827, 19783, 797, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1226510984", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R12 \[ScriptCapitalG]"},
   CellID->1226510984]},
 "R12 \[ScriptCapitalG]"->{
  Cell[684827, 19783, 797, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1226510984", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R12 \[ScriptCapitalG]"},
   CellID->1226510984]},
 "ID:1781039403"->{
  Cell[685627, 19805, 799, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1781039403", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R13 \[ScriptCapitalG]"},
   CellID->1781039403]},
 "R13 \[ScriptCapitalG]"->{
  Cell[685627, 19805, 799, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1781039403", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R13 \[ScriptCapitalG]"},
   CellID->1781039403]},
 "Proof|ID:939639226"->{
  Cell[686582, 19836, 2358, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:939639226", "Proof|ID:939639226-1"},
   CellID->1140411407]},
 "Proof|ID:939639226-1"->{
  Cell[686582, 19836, 2358, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:939639226", "Proof|ID:939639226-1"},
   CellID->1140411407]},
 "Proof|ID:37694441"->{
  Cell[689461, 19923, 2357, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:37694441", "Proof|ID:37694441-1"},
   CellID->1985690430]},
 "Proof|ID:37694441-1"->{
  Cell[689461, 19923, 2357, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:37694441", "Proof|ID:37694441-1"},
   CellID->1985690430]},
 "Proof|ID:576375692"->{
  Cell[692337, 20010, 2362, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:576375692", "Proof|ID:576375692-1"},
   CellID->1030879810]},
 "Proof|ID:576375692-1"->{
  Cell[692337, 20010, 2362, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:576375692", "Proof|ID:576375692-1"},
   CellID->1030879810]},
 "Proof|ID:1226510984"->{
  Cell[695219, 20097, 2371, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1226510984", "Proof|ID:1226510984-1"},
   CellID->1342766186]},
 "Proof|ID:1226510984-1"->{
  Cell[695219, 20097, 2371, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1226510984", "Proof|ID:1226510984-1"},
   CellID->1342766186]},
 "Proof|ID:1781039403"->{
  Cell[698111, 20185, 2370, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1781039403", "Proof|ID:1781039403-1"},
   CellID->234828850]},
 "Proof|ID:1781039403-1"->{
  Cell[698111, 20185, 2370, 69, 70, "ProofInfo",
   CellTags->{"Proof|ID:1781039403", "Proof|ID:1781039403-1"},
   CellID->234828850]},
 "ID:166003593"->{
  Cell[701645, 20291, 814, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:166003593", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isDomain"},
   CellID->166003593]},
 "\[ScriptCapitalG] isDomain"->{
  Cell[701645, 20291, 814, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:166003593", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isDomain"},
   CellID->166003593]},
 "ID:1167967104"->{
  Cell[702462, 20314, 1003, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1167967104", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isAlgoReductionRing"},
   CellID->1167967104]},
 "\[ScriptCapitalG] isAlgoReductionRing"->{
  Cell[702462, 20314, 1003, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1167967104", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isAlgoReductionRing"},
   CellID->1167967104]},
 "Proof|ID:166003593"->{
  Cell[703621, 20349, 2381, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:166003593", "Proof|ID:166003593-1"},
   CellID->324790663]},
 "Proof|ID:166003593-1"->{
  Cell[703621, 20349, 2381, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:166003593", "Proof|ID:166003593-1"},
   CellID->324790663]},
 "Proof|ID:1167967104"->{
  Cell[706522, 20436, 2451, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:1167967104", "Proof|ID:1167967104-1"},
   CellID->1160302337]},
 "Proof|ID:1167967104-1"->{
  Cell[706522, 20436, 2451, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:1167967104", "Proof|ID:1167967104-1"},
   CellID->1160302337]},
 "ID:2087655986"->{
  Cell[709746, 20535, 696, 19, 33, "GlobalDeclaration",
   CellTags->{
    "ID:2087655986", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->2087655986]},
 "ID:955133330"->{
  Cell[710484, 20559, 719, 19, 33, "GlobalDeclaration",
   CellTags->{
    "ID:955133330", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->955133330]},
 "ID:977980890"->{
  Cell[712234, 20612, 767, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:977980890", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->977980890]},
 "ID:1748570587"->{
  Cell[713004, 20634, 816, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1748570587", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1748570587]},
 "ID:699276041"->{
  Cell[713823, 20657, 822, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:699276041", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership rd"},
   CellID->699276041]},
 "membership rd"->{
  Cell[713823, 20657, 822, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:699276041", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership rd"},
   CellID->699276041]},
 "ID:867263181"->{
  Cell[714648, 20681, 1184, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:867263181", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "reducible rd"},
   CellID->867263181]},
 "reducible rd"->{
  Cell[714648, 20681, 1184, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:867263181", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "reducible rd"},
   CellID->867263181]},
 "ID:1431447619"->{
  Cell[715835, 20717, 1168, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1431447619", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "irreducible rd"},
   CellID->1431447619]},
 "irreducible rd"->{
  Cell[715835, 20717, 1168, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1431447619", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "irreducible rd"},
   CellID->1431447619]},
 "ID:1490981815"->{
  Cell[717006, 20755, 1058, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1490981815", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Succeeds] rd"},
   CellID->1490981815]},
 "\[Succeeds] rd"->{
  Cell[717006, 20755, 1058, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1490981815", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Succeeds] rd"},
   CellID->1490981815]},
 "Proof|ID:699276041"->{
  Cell[718220, 20796, 4223, 112, 70, "ProofInfo",
   CellTags->{"Proof|ID:699276041", "Proof|ID:699276041-1"},
   CellID->141722599]},
 "Proof|ID:699276041-1"->{
  Cell[718220, 20796, 4223, 112, 70, "ProofInfo",
   CellTags->{"Proof|ID:699276041", "Proof|ID:699276041-1"},
   CellID->141722599]},
 "Proof|ID:867263181"->{
  Cell[722963, 20927, 5331, 136, 70, "ProofInfo",
   CellTags->{"Proof|ID:867263181", "Proof|ID:867263181-1"},
   CellID->231570427]},
 "Proof|ID:867263181-1"->{
  Cell[722963, 20927, 5331, 136, 70, "ProofInfo",
   CellTags->{"Proof|ID:867263181", "Proof|ID:867263181-1"},
   CellID->231570427]},
 "Proof|ID:1431447619"->{
  Cell[728812, 21082, 5188, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:1431447619", "Proof|ID:1431447619-1"},
   CellID->1892915845]},
 "Proof|ID:1431447619-1"->{
  Cell[728812, 21082, 5188, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:1431447619", "Proof|ID:1431447619-1"},
   CellID->1892915845]},
 "Proof|ID:1490981815"->{
  Cell[734522, 21236, 5158, 133, 70, "ProofInfo",
   CellTags->{"Proof|ID:1490981815", "Proof|ID:1490981815-1"},
   CellID->642445226]},
 "Proof|ID:1490981815-1"->{
  Cell[734522, 21236, 5158, 133, 70, "ProofInfo",
   CellTags->{"Proof|ID:1490981815", "Proof|ID:1490981815-1"},
   CellID->642445226]},
 "ID:243233209"->{
  Cell[740976, 21412, 680, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:243233209", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->243233209]},
 "ID:974939989"->{
  Cell[741659, 21431, 1505, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:974939989", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdTBase1"},
   CellID->974939989]},
 "rdTBase1"->{
  Cell[741659, 21431, 1505, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:974939989", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdTBase1"},
   CellID->974939989]},
 "ID:1014318336"->{
  Cell[743167, 21476, 2294, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1014318336", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdTBase2"},
   CellID->1014318336]},
 "rdTBase2"->{
  Cell[743167, 21476, 2294, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1014318336", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdTBase2"},
   CellID->1014318336]},
 "ID:107497168"->{
  Cell[745464, 21547, 2498, 76, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:107497168", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdTRec"},
   CellID->107497168]},
 "rdTRec"->{
  Cell[745464, 21547, 2498, 76, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:107497168", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdTRec"},
   CellID->107497168]},
 "ID:839458854"->{
  Cell[748678, 21650, 1011, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:839458854", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->839458854]},
 "ID:1436407557"->{
  Cell[749692, 21678, 4623, 135, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1436407557", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "4"},
   CellID->1436407557]},
 "4"->{
  Cell[749692, 21678, 4623, 135, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1436407557", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "4"},
   CellID->1436407557]},
 "ID:2098894842"->{
  Cell[754318, 21815, 1921, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2098894842", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdT induction"},
   CellID->2098894842]},
 "rdT induction"->{
  Cell[754318, 21815, 1921, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2098894842", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdT induction"},
   CellID->2098894842]},
 "Proof|ID:1436407557"->{
  Cell[756395, 21881, 13718, 320, 70, "ProofInfo",
   CellTags->{"Proof|ID:1436407557", "Proof|ID:1436407557-1"},
   CellID->1376449459]},
 "Proof|ID:1436407557-1"->{
  Cell[756395, 21881, 13718, 320, 70, "ProofInfo",
   CellTags->{"Proof|ID:1436407557", "Proof|ID:1436407557-1"},
   CellID->1376449459]},
 "Proof|ID:2098894842"->{
  Cell[770634, 22220, 6819, 167, 70, "ProofInfo",
   CellTags->{"Proof|ID:2098894842", "Proof|ID:2098894842-1"},
   CellID->1924563166]},
 "Proof|ID:2098894842-1"->{
  Cell[770634, 22220, 6819, 167, 70, "ProofInfo",
   CellTags->{"Proof|ID:2098894842", "Proof|ID:2098894842-1"},
   CellID->1924563166]},
 "ID:160961870"->{
  Cell[778497, 22421, 2089, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:160961870", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdT alternative", "IR"},
   CellID->160961870]},
 "rdT alternative"->{
  Cell[778497, 22421, 2089, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:160961870", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdT alternative", "IR"},
   CellID->160961870]},
 "Proof|ID:160961870"->{
  Cell[780741, 22491, 7622, 180, 70, "ProofInfo",
   CellTags->{"Proof|ID:160961870", "Proof|ID:160961870-1"},
   CellID->1503881909]},
 "Proof|ID:160961870-1"->{
  Cell[780741, 22491, 7622, 180, 70, "ProofInfo",
   CellTags->{"Proof|ID:160961870", "Proof|ID:160961870-1"},
   CellID->1503881909]},
 "ID:168345814"->{
  Cell[789369, 22705, 953, 25, 70, "GlobalDeclaration",
   CellTags->{
    "ID:168345814", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->168345814]},
 "ID:1634249728"->{
  Cell[790325, 22732, 684, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1634249728", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1634249728]},
 "ID:512270654"->{
  Cell[791012, 22752, 821, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:512270654", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->512270654]},
 "ID:2105002699"->{
  Cell[791836, 22775, 878, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2105002699", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership rdT"},
   CellID->2105002699]},
 "membership rdT"->{
  Cell[791836, 22775, 878, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2105002699", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership rdT"},
   CellID->2105002699]},
 "ID:1505002987"->{
  Cell[792717, 22800, 1447, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1505002987", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "reducible rdT"},
   CellID->1505002987]},
 "reducible rdT"->{
  Cell[792717, 22800, 1447, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1505002987", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "reducible rdT"},
   CellID->1505002987]},
 "ID:514228174"->{
  Cell[794167, 22846, 1325, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:514228174", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "irreducible rdT"},
   CellID->514228174]},
 "irreducible rdT"->{
  Cell[794167, 22846, 1325, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:514228174", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "irreducible rdT"},
   CellID->514228174]},
 "ID:208729427"->{
  Cell[795495, 22886, 849, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:208729427", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow]\[RawStar] rdT"},
   CellID->208729427]},
 "\[RightArrow]\[RawStar] rdT"->{
  Cell[795495, 22886, 849, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:208729427", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow]\[RawStar] rdT"},
   CellID->208729427]},
 "ID:278468994"->{
  Cell[796347, 22911, 1542, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:278468994", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Succeeds] rdT"},
   CellID->278468994]},
 "\[Succeeds] rdT"->{
  Cell[796347, 22911, 1542, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:278468994", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Succeeds] rdT"},
   CellID->278468994]},
 "Proof|ID:2105002699"->{
  Cell[798046, 22967, 4748, 123, 70, "ProofInfo",
   CellTags->{"Proof|ID:2105002699", "Proof|ID:2105002699-1"},
   CellID->871739323]},
 "Proof|ID:2105002699-1"->{
  Cell[798046, 22967, 4748, 123, 70, "ProofInfo",
   CellTags->{"Proof|ID:2105002699", "Proof|ID:2105002699-1"},
   CellID->871739323]},
 "Proof|ID:1505002987"->{
  Cell[803315, 23109, 7238, 179, 70, "ProofInfo",
   CellTags->{"Proof|ID:1505002987", "Proof|ID:1505002987-1"},
   CellID->327721344]},
 "Proof|ID:1505002987-1"->{
  Cell[803315, 23109, 7238, 179, 70, "ProofInfo",
   CellTags->{"Proof|ID:1505002987", "Proof|ID:1505002987-1"},
   CellID->327721344]},
 "Proof|ID:514228174"->{
  Cell[811074, 23307, 7185, 176, 70, "ProofInfo",
   CellTags->{"Proof|ID:514228174", "Proof|ID:514228174-1"},
   CellID->1831820924]},
 "Proof|ID:514228174-1"->{
  Cell[811074, 23307, 7185, 176, 70, "ProofInfo",
   CellTags->{"Proof|ID:514228174", "Proof|ID:514228174-1"},
   CellID->1831820924]},
 "Proof|ID:208729427"->{
  Cell[818780, 23502, 5099, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:208729427", "Proof|ID:208729427-1"},
   CellID->970814565]},
 "Proof|ID:208729427-1"->{
  Cell[818780, 23502, 5099, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:208729427", "Proof|ID:208729427-1"},
   CellID->970814565]},
 "Proof|ID:278468994"->{
  Cell[824398, 23649, 7103, 174, 70, "ProofInfo",
   CellTags->{"Proof|ID:278468994", "Proof|ID:278468994-1"},
   CellID->120821588]},
 "Proof|ID:278468994-1"->{
  Cell[824398, 23649, 7103, 174, 70, "ProofInfo",
   CellTags->{"Proof|ID:278468994", "Proof|ID:278468994-1"},
   CellID->120821588]},
 "ID:96175791"->{
  Cell[832600, 23862, 1689, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:96175791", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "red rdT"},
   CellID->96175791]},
 "red rdT"->{
  Cell[832600, 23862, 1689, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:96175791", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "red rdT"},
   CellID->96175791]},
 "Proof|ID:96175791"->{
  Cell[834445, 23926, 5391, 136, 70, "ProofInfo",
   CellTags->{"Proof|ID:96175791", "Proof|ID:96175791-1"},
   CellID->56129168]},
 "Proof|ID:96175791-1"->{
  Cell[834445, 23926, 5391, 136, 70, "ProofInfo",
   CellTags->{"Proof|ID:96175791", "Proof|ID:96175791-1"},
   CellID->56129168]},
 "ID:909149002"->{
  Cell[841022, 24103, 3592, 106, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:909149002", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "trd induction"},
   CellID->909149002]},
 "trd induction"->{
  Cell[841022, 24103, 3592, 106, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:909149002", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "trd induction"},
   CellID->909149002]},
 "Proof|ID:909149002"->{
  Cell[844772, 24220, 10722, 252, 70, "ProofInfo",
   CellTags->{"Proof|ID:909149002", "Proof|ID:909149002-1"},
   CellID->35905978]},
 "Proof|ID:909149002-1"->{
  Cell[844772, 24220, 10722, 252, 70, "ProofInfo",
   CellTags->{"Proof|ID:909149002", "Proof|ID:909149002-1"},
   CellID->35905978]},
 "ID:1097836943"->{
  Cell[856495, 24506, 1008, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1097836943", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1097836943]},
 "ID:206560148"->{
  Cell[857506, 24534, 763, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:206560148", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->206560148]},
 "ID:1315173810"->{
  Cell[858272, 24556, 827, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1315173810", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership trd"},
   CellID->1315173810]},
 "membership trd"->{
  Cell[858272, 24556, 827, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1315173810", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership trd"},
   CellID->1315173810]},
 "ID:1071214578"->{
  Cell[859102, 24580, 852, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1071214578", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow]\[RawStar] trd"},
   CellID->1071214578]},
 "\[RightArrow]\[RawStar] trd"->{
  Cell[859102, 24580, 852, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1071214578", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow]\[RawStar] trd"},
   CellID->1071214578]},
 "ID:657218052"->{
  Cell[859957, 24605, 880, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:657218052", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "trd \[NotElement] red"},
   CellID->657218052]},
 "trd \[NotElement] red"->{
  Cell[859957, 24605, 880, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:657218052", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "trd \[NotElement] red"},
   CellID->657218052]},
 "Proof|ID:1315173810"->{
  Cell[860994, 24640, 4359, 115, 70, "ProofInfo",
   CellTags->{"Proof|ID:1315173810", "Proof|ID:1315173810-1"},
   CellID->56915654]},
 "Proof|ID:1315173810-1"->{
  Cell[860994, 24640, 4359, 115, 70, "ProofInfo",
   CellTags->{"Proof|ID:1315173810", "Proof|ID:1315173810-1"},
   CellID->56915654]},
 "Proof|ID:1071214578"->{
  Cell[865875, 24774, 4720, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:1071214578", "Proof|ID:1071214578-1"},
   CellID->204815329]},
 "Proof|ID:1071214578-1"->{
  Cell[865875, 24774, 4720, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:1071214578", "Proof|ID:1071214578-1"},
   CellID->204815329]},
 "Proof|ID:657218052"->{
  Cell[871116, 24914, 4965, 127, 70, "ProofInfo",
   CellTags->{"Proof|ID:657218052", "Proof|ID:657218052-1"},
   CellID->1364381745]},
 "Proof|ID:657218052-1"->{
  Cell[871116, 24914, 4965, 127, 70, "ProofInfo",
   CellTags->{"Proof|ID:657218052", "Proof|ID:657218052-1"},
   CellID->1364381745]},
 "ID:480907095"->{
  Cell[877161, 25078, 822, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:480907095", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->480907095]},
 "ID:1000930828"->{
  Cell[878608, 25117, 1036, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1000930828", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1000930828]},
 "ID:1774407783"->{
  Cell[879647, 25142, 3575, 99, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1774407783", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element]"},
   CellID->1774407783]},
 "MMPairs \[Element]"->{
  Cell[879647, 25142, 3575, 99, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1774407783", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element]"},
   CellID->1774407783]},
 "ID:193942992"->{
  Cell[883356, 25250, 975, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:193942992", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->193942992]},
 "ID:44826735"->{
  Cell[884981, 25295, 1175, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:44826735", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] ProdDomain"},
   CellID->44826735]},
 "MMPairs \[Element] ProdDomain"->{
  Cell[884981, 25295, 1175, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:44826735", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] ProdDomain"},
   CellID->44826735]},
 "ID:155756485"->{
  Cell[886159, 25328, 1305, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:155756485", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] ProdDomain DomainTuples"},
   CellID->155756485]},
 "MMPairs \[Element] ProdDomain DomainTuples"->{
  Cell[886159, 25328, 1305, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:155756485", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] ProdDomain DomainTuples"},
   CellID->155756485]},
 "ID:1019394084"->{
  Cell[887467, 25365, 3415, 96, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1019394084", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] alternative I", "IR"},
   CellID->1019394084]},
 "MMPairs \[Element] alternative I"->{
  Cell[887467, 25365, 3415, 96, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1019394084", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] alternative I", "IR"},
   CellID->1019394084]},
 "ID:1458538511"->{
  Cell[890885, 25463, 2963, 82, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1458538511", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] alternative II", "IR"},
   CellID->1458538511]},
 "MMPairs \[Element] alternative II"->{
  Cell[890885, 25463, 2963, 82, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1458538511", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] alternative II", "IR"},
   CellID->1458538511]},
 "ID:81000274"->{
  Cell[893851, 25547, 2529, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:81000274", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] alternative III"},
   CellID->81000274]},
 "MMPairs \[Element] alternative III"->{
  Cell[893851, 25547, 2529, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:81000274", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] alternative III"},
   CellID->81000274]},
 "Proof|ID:44826735"->{
  Cell[896536, 25637, 3851, 100, 70, "ProofInfo",
   CellTags->{"Proof|ID:44826735", "Proof|ID:44826735-1"},
   CellID->33075853]},
 "Proof|ID:44826735-1"->{
  Cell[896536, 25637, 3851, 100, 70, "ProofInfo",
   CellTags->{"Proof|ID:44826735", "Proof|ID:44826735-1"},
   CellID->33075853]},
 "Proof|ID:155756485"->{
  Cell[900906, 25756, 4135, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:155756485", "Proof|ID:155756485-1"},
   CellID->228098881]},
 "Proof|ID:155756485-1"->{
  Cell[900906, 25756, 4135, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:155756485", "Proof|ID:155756485-1"},
   CellID->228098881]},
 "Proof|ID:1019394084"->{
  Cell[905561, 25881, 10471, 239, 70, "ProofInfo",
   CellTags->{"Proof|ID:1019394084", "Proof|ID:1019394084-1"},
   CellID->679633020]},
 "Proof|ID:1019394084-1"->{
  Cell[905561, 25881, 10471, 239, 70, "ProofInfo",
   CellTags->{"Proof|ID:1019394084", "Proof|ID:1019394084-1"},
   CellID->679633020]},
 "Proof|ID:1458538511"->{
  Cell[916554, 26139, 8823, 199, 70, "ProofInfo",
   CellTags->{"Proof|ID:1458538511", "Proof|ID:1458538511-1"},
   CellID->1939038746]},
 "Proof|ID:1458538511-1"->{
  Cell[916554, 26139, 8823, 199, 70, "ProofInfo",
   CellTags->{"Proof|ID:1458538511", "Proof|ID:1458538511-1"},
   CellID->1939038746]},
 "Proof|ID:81000274"->{
  Cell[925899, 26357, 7154, 171, 70, "ProofInfo",
   CellTags->{"Proof|ID:81000274", "Proof|ID:81000274-1"},
   CellID->637304832]},
 "Proof|ID:81000274-1"->{
  Cell[925899, 26357, 7154, 171, 70, "ProofInfo",
   CellTags->{"Proof|ID:81000274", "Proof|ID:81000274-1"},
   CellID->637304832]},
 "ID:1140506798"->{
  Cell[934186, 26564, 3381, 88, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1140506798", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NPairs shape"},
   CellID->1140506798]},
 "DomainTuples NPairs shape"->{
  Cell[934186, 26564, 3381, 88, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1140506798", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NPairs shape"},
   CellID->1140506798]},
 "Proof|ID:1140506798"->{
  Cell[937725, 26663, 9712, 221, 70, "ProofInfo",
   CellTags->{"Proof|ID:1140506798", "Proof|ID:1140506798-1"},
   CellID->344640348]},
 "Proof|ID:1140506798-1"->{
  Cell[937725, 26663, 9712, 221, 70, "ProofInfo",
   CellTags->{"Proof|ID:1140506798", "Proof|ID:1140506798-1"},
   CellID->344640348]},
 "ID:1527335579"->{
  Cell[948750, 26926, 1029, 27, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1527335579", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1527335579]},
 "ID:1794762591"->{
  Cell[949782, 26955, 1056, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1794762591", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership cpmTuple 1"},
   CellID->1794762591]},
 "membership cpmTuple 1"->{
  Cell[949782, 26955, 1056, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1794762591", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership cpmTuple 1"},
   CellID->1794762591]},
 "ID:1790156146"->{
  Cell[950841, 26985, 2346, 70, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1790156146", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpmTuple content 1"},
   CellID->1790156146]},
 "cpmTuple content 1"->{
  Cell[950841, 26985, 2346, 70, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1790156146", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpmTuple content 1"},
   CellID->1790156146]},
 "Proof|ID:1794762591"->{
  Cell[953345, 27066, 6393, 157, 70, "ProofInfo",
   CellTags->{"Proof|ID:1794762591", "Proof|ID:1794762591-1"},
   CellID->1374603973]},
 "Proof|ID:1794762591-1"->{
  Cell[953345, 27066, 6393, 157, 70, "ProofInfo",
   CellTags->{"Proof|ID:1794762591", "Proof|ID:1794762591-1"},
   CellID->1374603973]},
 "Proof|ID:1790156146"->{
  Cell[960259, 27242, 11875, 271, 70, "ProofInfo",
   CellTags->{"Proof|ID:1790156146", "Proof|ID:1790156146-1"},
   CellID->1785625847]},
 "Proof|ID:1790156146-1"->{
  Cell[960259, 27242, 11875, 271, 70, "ProofInfo",
   CellTags->{"Proof|ID:1790156146", "Proof|ID:1790156146-1"},
   CellID->1785625847]},
 "ID:884601666"->{
  Cell[973212, 27548, 892, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:884601666", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->884601666]},
 "ID:1188084204"->{
  Cell[974107, 27573, 906, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1188084204", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1188084204]},
 "ID:1708769881"->{
  Cell[975016, 27598, 1146, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1708769881", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership cpmTuple 2"},
   CellID->1708769881]},
 "membership cpmTuple 2"->{
  Cell[975016, 27598, 1146, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1708769881", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership cpmTuple 2"},
   CellID->1708769881]},
 "ID:125591800"->{
  Cell[976165, 27630, 2401, 71, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:125591800", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpmTuple content 2"},
   CellID->125591800]},
 "cpmTuple content 2"->{
  Cell[976165, 27630, 2401, 71, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:125591800", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpmTuple content 2"},
   CellID->125591800]},
 "Proof|ID:1708769881"->{
  Cell[978721, 27712, 7143, 175, 70, "ProofInfo",
   CellTags->{"Proof|ID:1708769881", "Proof|ID:1708769881-1"},
   CellID->1434737069]},
 "Proof|ID:1708769881-1"->{
  Cell[978721, 27712, 7143, 175, 70, "ProofInfo",
   CellTags->{"Proof|ID:1708769881", "Proof|ID:1708769881-1"},
   CellID->1434737069]},
 "Proof|ID:125591800"->{
  Cell[986386, 27906, 12998, 294, 70, "ProofInfo",
   CellTags->{"Proof|ID:125591800", "Proof|ID:125591800-1"},
   CellID->1320574433]},
 "Proof|ID:125591800-1"->{
  Cell[986386, 27906, 12998, 294, 70, "ProofInfo",
   CellTags->{"Proof|ID:125591800", "Proof|ID:125591800-1"},
   CellID->1320574433]},
 "ID:1122306642"->{
  Cell[1000831, 28245, 731, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1122306642", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1122306642]},
 "ID:885588057"->{
  Cell[1001565, 28266, 3193, 102, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:885588057", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "5"},
   CellID->885588057]},
 "5"->{
  Cell[1001565, 28266, 3193, 102, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:885588057", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "5"},
   CellID->885588057]},
 "ID:438489157"->{
  Cell[1004761, 28370, 3039, 95, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:438489157", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTupleAux induction"},
   CellID->438489157]},
 "lmTupleAux induction"->{
  Cell[1004761, 28370, 3039, 95, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:438489157", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTupleAux induction"},
   CellID->438489157]},
 "Proof|ID:885588057"->{
  Cell[1007957, 28476, 8784, 214, 70, "ProofInfo",
   CellTags->{"Proof|ID:885588057", "Proof|ID:885588057-1"},
   CellID->1319009677]},
 "Proof|ID:885588057-1"->{
  Cell[1007957, 28476, 8784, 214, 70, "ProofInfo",
   CellTags->{"Proof|ID:885588057", "Proof|ID:885588057-1"},
   CellID->1319009677]},
 "Proof|ID:438489157"->{
  Cell[1017262, 28709, 8534, 208, 70, "ProofInfo",
   CellTags->{"Proof|ID:438489157", "Proof|ID:438489157-1"},
   CellID->1261187836]},
 "Proof|ID:438489157-1"->{
  Cell[1017262, 28709, 8534, 208, 70, "ProofInfo",
   CellTags->{"Proof|ID:438489157", "Proof|ID:438489157-1"},
   CellID->1261187836]},
 "ID:1795822474"->{
  Cell[1026756, 28950, 844, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1795822474", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1795822474]},
 "ID:2110280918"->{
  Cell[1027603, 28974, 1070, 31, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2110280918", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->2110280918]},
 "ID:481303170"->{
  Cell[1028676, 29007, 1057, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:481303170", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership lmTupleAux"},
   CellID->481303170]},
 "membership lmTupleAux"->{
  Cell[1028676, 29007, 1057, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:481303170", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership lmTupleAux"},
   CellID->481303170]},
 "ID:1389102731"->{
  Cell[1029736, 29037, 2692, 81, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1389102731", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTupleAux content"},
   CellID->1389102731]},
 "lmTupleAux content"->{
  Cell[1029736, 29037, 2692, 81, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1389102731", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTupleAux content"},
   CellID->1389102731]},
 "Proof|ID:481303170"->{
  Cell[1032584, 29129, 6078, 150, 70, "ProofInfo",
   CellTags->{"Proof|ID:481303170", "Proof|ID:481303170-1"},
   CellID->1309244696]},
 "Proof|ID:481303170-1"->{
  Cell[1032584, 29129, 6078, 150, 70, "ProofInfo",
   CellTags->{"Proof|ID:481303170", "Proof|ID:481303170-1"},
   CellID->1309244696]},
 "Proof|ID:1389102731"->{
  Cell[1039183, 29298, 12904, 291, 70, "ProofInfo",
   CellTags->{"Proof|ID:1389102731", "Proof|ID:1389102731-1"},
   CellID->648616519]},
 "Proof|ID:1389102731-1"->{
  Cell[1039183, 29298, 12904, 291, 70, "ProofInfo",
   CellTags->{"Proof|ID:1389102731", "Proof|ID:1389102731-1"},
   CellID->648616519]},
 "ID:1615013989"->{
  Cell[1053379, 29632, 1026, 27, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1615013989", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1615013989]},
 "ID:1776007069"->{
  Cell[1054408, 29661, 770, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1776007069", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1776007069]},
 "ID:1846257369"->{
  Cell[1055181, 29683, 1095, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1846257369", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership lmTuple"},
   CellID->1846257369]},
 "membership lmTuple"->{
  Cell[1055181, 29683, 1095, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1846257369", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership lmTuple"},
   CellID->1846257369]},
 "ID:851213532"->{
  Cell[1056279, 29712, 2420, 72, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:851213532", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTuple \[NotEqual]"},
   CellID->851213532]},
 "lmTuple \[NotEqual]"->{
  Cell[1056279, 29712, 2420, 72, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:851213532", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTuple \[NotEqual]"},
   CellID->851213532]},
 "ID:959030563"->{
  Cell[1058702, 29786, 3744, 103, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:959030563", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTuple ="},
   CellID->959030563]},
 "lmTuple ="->{
  Cell[1058702, 29786, 3744, 103, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:959030563", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTuple ="},
   CellID->959030563]},
 "Proof|ID:1846257369"->{
  Cell[1062603, 29900, 6074, 151, 70, "ProofInfo",
   CellTags->{"Proof|ID:1846257369", "Proof|ID:1846257369-1"},
   CellID->166489159]},
 "Proof|ID:1846257369-1"->{
  Cell[1062603, 29900, 6074, 151, 70, "ProofInfo",
   CellTags->{"Proof|ID:1846257369", "Proof|ID:1846257369-1"},
   CellID->166489159]},
 "Proof|ID:851213532"->{
  Cell[1069198, 30070, 11096, 256, 70, "ProofInfo",
   CellTags->{"Proof|ID:851213532", "Proof|ID:851213532-1"},
   CellID->412269868]},
 "Proof|ID:851213532-1"->{
  Cell[1069198, 30070, 11096, 256, 70, "ProofInfo",
   CellTags->{"Proof|ID:851213532", "Proof|ID:851213532-1"},
   CellID->412269868]},
 "Proof|ID:959030563"->{
  Cell[1080814, 30345, 15263, 342, 70, "ProofInfo",
   CellTags->{"Proof|ID:959030563", "Proof|ID:959030563-1"},
   CellID->1319693045]},
 "Proof|ID:959030563-1"->{
  Cell[1080814, 30345, 15263, 342, 70, "ProofInfo",
   CellTags->{"Proof|ID:959030563", "Proof|ID:959030563-1"},
   CellID->1319693045]},
 "ID:1440157218"->{
  Cell[1096874, 30714, 785, 20, 44, "GlobalDeclaration",
   CellTags->{
    "ID:1440157218", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1440157218]},
 "ID:805845835"->{
  Cell[1097875, 30743, 673, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:805845835", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->805845835]},
 "ID:1576526273"->{
  Cell[1099056, 30779, 2884, 85, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1576526273", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple"},
   CellID->1576526273]},
 "isTODOTuple"->{
  Cell[1099056, 30779, 2884, 85, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1576526273", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple"},
   CellID->1576526273]},
 "ID:857941005"->{
  Cell[1102644, 30891, 1045, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:857941005", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple allPairs"},
   CellID->857941005]},
 "isTODOTuple allPairs"->{
  Cell[1102644, 30891, 1045, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:857941005", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple allPairs"},
   CellID->857941005]},
 "ID:1202189993"->{
  Cell[1103692, 30919, 2192, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1202189993", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple update"},
   CellID->1202189993]},
 "isTODOTuple update"->{
  Cell[1103692, 30919, 2192, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1202189993", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple update"},
   CellID->1202189993]},
 "ID:1698314044"->{
  Cell[1105887, 30984, 3124, 89, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1698314044", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple rest"},
   CellID->1698314044]},
 "isTODOTuple rest"->{
  Cell[1105887, 30984, 3124, 89, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1698314044", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple rest"},
   CellID->1698314044]},
 "ID:812012342"->{
  Cell[1109014, 31075, 1743, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:812012342", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple exchange"},
   CellID->812012342]},
 "isTODOTuple exchange"->{
  Cell[1109014, 31075, 1743, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:812012342", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple exchange"},
   CellID->812012342]},
 "Proof|ID:857941005"->{
  Cell[1110913, 31135, 4116, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:857941005", "Proof|ID:857941005-1"},
   CellID->1051929619]},
 "Proof|ID:857941005-1"->{
  Cell[1110913, 31135, 4116, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:857941005", "Proof|ID:857941005-1"},
   CellID->1051929619]},
 "Proof|ID:1202189993"->{
  Cell[1115549, 31260, 7401, 177, 70, "ProofInfo",
   CellTags->{"Proof|ID:1202189993", "Proof|ID:1202189993-1"},
   CellID->1110047828]},
 "Proof|ID:1202189993-1"->{
  Cell[1115549, 31260, 7401, 177, 70, "ProofInfo",
   CellTags->{"Proof|ID:1202189993", "Proof|ID:1202189993-1"},
   CellID->1110047828]},
 "Proof|ID:1698314044"->{
  Cell[1123471, 31456, 11356, 261, 70, "ProofInfo",
   CellTags->{"Proof|ID:1698314044", "Proof|ID:1698314044-1"},
   CellID->610967613]},
 "Proof|ID:1698314044-1"->{
  Cell[1123471, 31456, 11356, 261, 70, "ProofInfo",
   CellTags->{"Proof|ID:1698314044", "Proof|ID:1698314044-1"},
   CellID->610967613]},
 "Proof|ID:812012342"->{
  Cell[1135348, 31736, 6913, 166, 70, "ProofInfo",
   CellTags->{"Proof|ID:812012342", "Proof|ID:812012342-1"},
   CellID->195013346]},
 "Proof|ID:812012342-1"->{
  Cell[1135348, 31736, 6913, 166, 70, "ProofInfo",
   CellTags->{"Proof|ID:812012342", "Proof|ID:812012342-1"},
   CellID->195013346]},
 "ID:1260897708"->{
  Cell[1143456, 31943, 2816, 72, 99, "FormalTextInputFormula",
   CellTags->{
    "ID:1260897708", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "chainCrit 1"},
   CellID->1260897708]},
 "chainCrit 1"->{
  Cell[1143456, 31943, 2816, 72, 99, "FormalTextInputFormula",
   CellTags->{
    "ID:1260897708", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "chainCrit 1"},
   CellID->1260897708]},
 "Proof|ID:1260897708"->{
  Cell[1146430, 32026, 9956, 231, 28, "ProofInfo",
   CellTags->{"Proof|ID:1260897708", "Proof|ID:1260897708-1"},
   CellID->721986367]},
 "Proof|ID:1260897708-1"->{
  Cell[1146430, 32026, 9956, 231, 28, "ProofInfo",
   CellTags->{"Proof|ID:1260897708", "Proof|ID:1260897708-1"},
   CellID->721986367]},
 "ID:957934274"->{
  Cell[1169532, 32690, 817, 21, 33, "GlobalDeclaration",
   CellTags->{
    "ID:957934274", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->957934274]},
 "ID:1363987639"->{
  Cell[1170392, 32716, 795, 20, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1363987639", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1363987639]},
 "ID:184081942"->{
  Cell[1172007, 32765, 2288, 68, 130, "FormalTextInputFormula",
   CellTags->{
    "ID:184081942", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBTermOrder"},
   CellID->184081942]},
 "GBTermOrder"->{
  Cell[1172007, 32765, 2288, 68, 130, "FormalTextInputFormula",
   CellTags->{
    "ID:184081942", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBTermOrder"},
   CellID->184081942]},
 "ID:1341865321"->{
  Cell[1174429, 32842, 946, 22, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1341865321", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1341865321]},
 "ID:320402459"->{
  Cell[1176068, 32891, 812, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:320402459", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] isDomain"},
   CellID->320402459]},
 "\[ScriptCapitalL] isDomain"->{
  Cell[1176068, 32891, 812, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:320402459", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] isDomain"},
   CellID->320402459]},
 "ID:1861353391"->{
  Cell[1176883, 32914, 1046, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1861353391", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] isPartIrreflOrder"},
   CellID->1861353391]},
 "\[ScriptCapitalL] isPartIrreflOrder"->{
  Cell[1176883, 32914, 1046, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1861353391", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] isPartIrreflOrder"},
   CellID->1861353391]},
 "ID:461530885"->{
  Cell[1177932, 32946, 1028, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:461530885", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] isNoetherian"},
   CellID->461530885]},
 "\[ScriptCapitalL] isNoetherian"->{
  Cell[1177932, 32946, 1028, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:461530885", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] isNoetherian"},
   CellID->461530885]},
 "ID:1658780438"->{
  Cell[1178963, 32978, 4409, 118, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1658780438", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] shape"},
   CellID->1658780438]},
 "\[ScriptCapitalL] shape"->{
  Cell[1178963, 32978, 4409, 118, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1658780438", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] shape"},
   CellID->1658780438]},
 "Proof|ID:320402459"->{
  Cell[1183529, 33107, 2379, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:320402459", "Proof|ID:320402459-1"},
   CellID->957846942]},
 "Proof|ID:320402459-1"->{
  Cell[1183529, 33107, 2379, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:320402459", "Proof|ID:320402459-1"},
   CellID->957846942]},
 "Proof|ID:1861353391"->{
  Cell[1186428, 33194, 2859, 81, 70, "ProofInfo",
   CellTags->{"Proof|ID:1861353391", "Proof|ID:1861353391-1"},
   CellID->71550445]},
 "Proof|ID:1861353391-1"->{
  Cell[1186428, 33194, 2859, 81, 70, "ProofInfo",
   CellTags->{"Proof|ID:1861353391", "Proof|ID:1861353391-1"},
   CellID->71550445]},
 "Proof|ID:461530885"->{
  Cell[1189808, 33294, 2823, 79, 70, "ProofInfo",
   CellTags->{"Proof|ID:461530885", "Proof|ID:461530885-1"},
   CellID->318070455]},
 "Proof|ID:461530885-1"->{
  Cell[1189808, 33294, 2823, 79, 70, "ProofInfo",
   CellTags->{"Proof|ID:461530885", "Proof|ID:461530885-1"},
   CellID->318070455]},
 "Proof|ID:1658780438"->{
  Cell[1193151, 33392, 12299, 283, 70, "ProofInfo",
   CellTags->{"Proof|ID:1658780438", "Proof|ID:1658780438-1"},
   CellID->65069410]},
 "Proof|ID:1658780438-1"->{
  Cell[1193151, 33392, 12299, 283, 70, "ProofInfo",
   CellTags->{"Proof|ID:1658780438", "Proof|ID:1658780438-1"},
   CellID->65069410]},
 "ID:1606683850"->{
  Cell[1206465, 33710, 786, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1606683850", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1606683850]},
 "ID:973931584"->{
  Cell[1207254, 33732, 1508, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:973931584", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] \[Succeeds] 1"},
   CellID->973931584]},
 "\[ScriptCapitalL] \[Succeeds] 1"->{
  Cell[1207254, 33732, 1508, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:973931584", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] \[Succeeds] 1"},
   CellID->973931584]},
 "ID:1350364326"->{
  Cell[1208765, 33772, 1457, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1350364326", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] \[Succeeds] 2"},
   CellID->1350364326]},
 "\[ScriptCapitalL] \[Succeeds] 2"->{
  Cell[1208765, 33772, 1457, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1350364326", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] \[Succeeds] 2"},
   CellID->1350364326]},
 "ID:529848255"->{
  Cell[1210225, 33812, 1671, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:529848255", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] \[Succeeds] 3"},
   CellID->529848255]},
 "\[ScriptCapitalL] \[Succeeds] 3"->{
  Cell[1210225, 33812, 1671, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:529848255", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] \[Succeeds] 3"},
   CellID->529848255]},
 "ID:592062547"->{
  Cell[1211899, 33864, 3057, 83, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:592062547", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[SquareSubset] append trd"},
   CellID->592062547]},
 "\[SquareSubset] append trd"->{
  Cell[1211899, 33864, 3057, 83, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:592062547", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[SquareSubset] append trd"},
   CellID->592062547]},
 "Proof|ID:973931584"->{
  Cell[1215113, 33958, 5404, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:973931584", "Proof|ID:973931584-1"},
   CellID->262124527]},
 "Proof|ID:973931584-1"->{
  Cell[1215113, 33958, 5404, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:973931584", "Proof|ID:973931584-1"},
   CellID->262124527]},
 "Proof|ID:1350364326"->{
  Cell[1221036, 34112, 5278, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:1350364326", "Proof|ID:1350364326-1"},
   CellID->47470875]},
 "Proof|ID:1350364326-1"->{
  Cell[1221036, 34112, 5278, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:1350364326", "Proof|ID:1350364326-1"},
   CellID->47470875]},
 "Proof|ID:529848255"->{
  Cell[1226835, 34263, 5850, 141, 70, "ProofInfo",
   CellTags->{"Proof|ID:529848255", "Proof|ID:529848255-1"},
   CellID->196790701]},
 "Proof|ID:529848255-1"->{
  Cell[1226835, 34263, 5850, 141, 70, "ProofInfo",
   CellTags->{"Proof|ID:529848255", "Proof|ID:529848255-1"},
   CellID->196790701]},
 "Proof|ID:592062547"->{
  Cell[1233205, 34423, 12382, 288, 70, "ProofInfo",
   CellTags->{"Proof|ID:592062547", "Proof|ID:592062547-1"},
   CellID->1513019922]},
 "Proof|ID:592062547-1"->{
  Cell[1233205, 34423, 12382, 288, 70, "ProofInfo",
   CellTags->{"Proof|ID:592062547", "Proof|ID:592062547-1"},
   CellID->1513019922]},
 "ID:1848949834"->{
  Cell[1246777, 34752, 743, 18, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1848949834", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1848949834]},
 "ID:1362258751"->{
  Cell[1247523, 34772, 2717, 74, 114, "FormalTextInputFormula",
   CellTags->{
    "ID:1362258751", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxTest"},
   CellID->1362258751]},
 "GBAuxTest"->{
  Cell[1247523, 34772, 2717, 74, 114, "FormalTextInputFormula",
   CellTags->{
    "ID:1362258751", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxTest"},
   CellID->1362258751]},
 "ID:1347349813"->{
  Cell[1250243, 34848, 1436, 37, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1347349813", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxBase"},
   CellID->1347349813]},
 "GBAuxBase"->{
  Cell[1250243, 34848, 1436, 37, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1347349813", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxBase"},
   CellID->1347349813]},
 "ID:270708621"->{
  Cell[1251682, 34887, 4352, 115, 154, "FormalTextInputFormula",
   CellTags->{
    "ID:270708621", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec1"},
   CellID->270708621]},
 "GBAuxRec1"->{
  Cell[1251682, 34887, 4352, 115, 154, "FormalTextInputFormula",
   CellTags->{
    "ID:270708621", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec1"},
   CellID->270708621]},
 "ID:1542247232"->{
  Cell[1256792, 35031, 3492, 99, 154, "FormalTextInputFormula",
   CellTags->{
    "ID:1542247232", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec2"},
   CellID->1542247232]},
 "GBAuxRec2"->{
  Cell[1256792, 35031, 3492, 99, 154, "FormalTextInputFormula",
   CellTags->{
    "ID:1542247232", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec2"},
   CellID->1542247232]},
 "ID:580452282"->{
  Cell[1260287, 35132, 5126, 136, 201, "FormalTextInputFormula",
   CellTags->{
    "ID:580452282", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec3"},
   CellID->580452282]},
 "GBAuxRec3"->{
  Cell[1260287, 35132, 5126, 136, 201, "FormalTextInputFormula",
   CellTags->{
    "ID:580452282", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec3"},
   CellID->580452282]},
 "ID:598818422"->{
  Cell[1266250, 35299, 5545, 153, 226, "FormalTextInputFormula",
   CellTags->{
    "ID:598818422", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec4"},
   CellID->598818422]},
 "GBAuxRec4"->{
  Cell[1266250, 35299, 5545, 153, 226, "FormalTextInputFormula",
   CellTags->{
    "ID:598818422", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec4"},
   CellID->598818422]},
 "ID:1392375551"->{
  Cell[1272430, 35478, 651, 16, 32, "GlobalDeclaration",
   CellTags->{
    "ID:1392375551", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1392375551]},
 "ID:994411308"->{
  Cell[1273084, 35496, 8640, 228, 316, "FormalTextInputFormula",
   CellTags->{
    "ID:994411308", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "1"},
   CellID->994411308]},
 "1"->{
  Cell[1273084, 35496, 8640, 228, 316, "FormalTextInputFormula",
   CellTags->{
    "ID:994411308", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "1"},
   CellID->994411308]},
 "ID:1185976259"->{
  Cell[1281727, 35726, 17123, 416, 446, "FormalTextInputFormula",
   CellTags->{
    "ID:1185976259", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "2"},
   CellID->1185976259]},
 "2"->{
  Cell[1281727, 35726, 17123, 416, 446, "FormalTextInputFormula",
   CellTags->{
    "ID:1185976259", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "2"},
   CellID->1185976259]},
 "ID:1125858125"->{
  Cell[1298853, 36144, 4696, 139, 173, "FormalTextInputFormula",
   CellTags->{
    "ID:1125858125", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "3"},
   CellID->1125858125]},
 "3"->{
  Cell[1298853, 36144, 4696, 139, 173, "FormalTextInputFormula",
   CellTags->{
    "ID:1125858125", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "3"},
   CellID->1125858125]},
 "ID:1780855977"->{
  Cell[1303552, 36285, 4442, 125, 222, "FormalTextInputFormula",
   CellTags->{
    "ID:1780855977", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAux induction"},
   CellID->1780855977]},
 "GBAux induction"->{
  Cell[1303552, 36285, 4442, 125, 222, "FormalTextInputFormula",
   CellTags->{
    "ID:1780855977", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAux induction"},
   CellID->1780855977]},
 "ID:380025998"->{
  Cell[1307997, 36412, 4432, 126, 281, "FormalTextInputFormula",
   CellTags->{
    "ID:380025998", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAux induction II"},
   CellID->380025998]},
 "GBAux induction II"->{
  Cell[1307997, 36412, 4432, 126, 281, "FormalTextInputFormula",
   CellTags->{
    "ID:380025998", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAux induction II"},
   CellID->380025998]},
 "Proof|ID:994411308"->{
  Cell[1312587, 36549, 23368, 514, 28, "ProofInfo",
   CellTags->{"Proof|ID:994411308", "Proof|ID:994411308-1"},
   CellID->87121011]},
 "Proof|ID:994411308-1"->{
  Cell[1312587, 36549, 23368, 514, 28, "ProofInfo",
   CellTags->{"Proof|ID:994411308", "Proof|ID:994411308-1"},
   CellID->87121011]},
 "Proof|ID:1185976259"->{
  Cell[1336473, 37082, 53148, 1138, 26, "ProofInfo",
   CellTags->{"Proof|ID:1185976259", "Proof|ID:1185976259-1"},
   CellID->50921615]},
 "Proof|ID:1185976259-1"->{
  Cell[1336473, 37082, 53148, 1138, 26, "ProofInfo",
   CellTags->{"Proof|ID:1185976259", "Proof|ID:1185976259-1"},
   CellID->50921615]},
 "Proof|ID:1125858125"->{
  Cell[1390184, 38240, 16212, 370, 26, "ProofInfo",
   CellTags->{"Proof|ID:1125858125", "Proof|ID:1125858125-1"},
   CellID->52228001]},
 "Proof|ID:1125858125-1"->{
  Cell[1390184, 38240, 16212, 370, 26, "ProofInfo",
   CellTags->{"Proof|ID:1125858125", "Proof|ID:1125858125-1"},
   CellID->52228001]},
 "Proof|ID:1780855977"->{
  Cell[1406917, 38629, 17071, 377, 26, "ProofInfo",
   CellTags->{"Proof|ID:1780855977", "Proof|ID:1780855977-1"},
   CellID->3158281]},
 "Proof|ID:1780855977-1"->{
  Cell[1406917, 38629, 17071, 377, 26, "ProofInfo",
   CellTags->{"Proof|ID:1780855977", "Proof|ID:1780855977-1"},
   CellID->3158281]},
 "Proof|ID:380025998"->{
  Cell[1424508, 39025, 17319, 382, 26, "ProofInfo",
   CellTags->{"Proof|ID:380025998", "Proof|ID:380025998-1"},
   CellID->1967872054]},
 "Proof|ID:380025998-1"->{
  Cell[1424508, 39025, 17319, 382, 26, "ProofInfo",
   CellTags->{"Proof|ID:380025998", "Proof|ID:380025998-1"},
   CellID->1967872054]},
 "ID:1379909256"->{
  Cell[1496312, 41028, 3435, 92, 164, "FormalTextInputFormula",
   CellTags->{
    "ID:1379909256", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAux 3 alternative", "IR"},
   CellID->1379909256]},
 "GBAux 3 alternative"->{
  Cell[1496312, 41028, 3435, 92, 164, "FormalTextInputFormula",
   CellTags->{
    "ID:1379909256", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAux 3 alternative", "IR"},
   CellID->1379909256]},
 "Proof|ID:1379909256"->{
  Cell[1499905, 41131, 15695, 342, 28, "ProofInfo",
   CellTags->{"Proof|ID:1379909256", "Proof|ID:1379909256-1"},
   CellID->1806109079]},
 "Proof|ID:1379909256-1"->{
  Cell[1499905, 41131, 15695, 342, 28, "ProofInfo",
   CellTags->{"Proof|ID:1379909256", "Proof|ID:1379909256-1"},
   CellID->1806109079]},
 "ID:884344895"->{
  Cell[1516392, 41500, 778, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:884344895", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->884344895]},
 "ID:2012261150"->{
  Cell[1517642, 41538, 855, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2012261150", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->2012261150]},
 "ID:1467104417"->{
  Cell[1518500, 41561, 1873, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1467104417", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxBase"},
   CellID->1467104417]},
 "shape GBAuxBase"->{
  Cell[1518500, 41561, 1873, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1467104417", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxBase"},
   CellID->1467104417]},
 "ID:1354352171"->{
  Cell[1520376, 41620, 1850, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1354352171", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec1"},
   CellID->1354352171]},
 "shape GBAuxRec1"->{
  Cell[1520376, 41620, 1850, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1354352171", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec1"},
   CellID->1354352171]},
 "ID:1322915202"->{
  Cell[1522229, 41679, 1850, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1322915202", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec2"},
   CellID->1322915202]},
 "shape GBAuxRec2"->{
  Cell[1522229, 41679, 1850, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1322915202", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec2"},
   CellID->1322915202]},
 "ID:620036821"->{
  Cell[1524082, 41738, 1847, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:620036821", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec3"},
   CellID->620036821]},
 "shape GBAuxRec3"->{
  Cell[1524082, 41738, 1847, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:620036821", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec3"},
   CellID->620036821]},
 "ID:1111467670"->{
  Cell[1525932, 41797, 1850, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1111467670", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec4"},
   CellID->1111467670]},
 "shape GBAuxRec4"->{
  Cell[1525932, 41797, 1850, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1111467670", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec4"},
   CellID->1111467670]},
 "Proof|ID:1467104417"->{
  Cell[1527939, 41865, 8473, 197, 70, "ProofInfo",
   CellTags->{"Proof|ID:1467104417", "Proof|ID:1467104417-1"},
   CellID->1997727746]},
 "Proof|ID:1467104417-1"->{
  Cell[1527939, 41865, 8473, 197, 70, "ProofInfo",
   CellTags->{"Proof|ID:1467104417", "Proof|ID:1467104417-1"},
   CellID->1997727746]},
 "Proof|ID:1354352171"->{
  Cell[1536934, 42081, 8472, 197, 70, "ProofInfo",
   CellTags->{"Proof|ID:1354352171", "Proof|ID:1354352171-1"},
   CellID->478354620]},
 "Proof|ID:1354352171-1"->{
  Cell[1536934, 42081, 8472, 197, 70, "ProofInfo",
   CellTags->{"Proof|ID:1354352171", "Proof|ID:1354352171-1"},
   CellID->478354620]},
 "Proof|ID:1322915202"->{
  Cell[1545927, 42297, 8473, 197, 70, "ProofInfo",
   CellTags->{"Proof|ID:1322915202", "Proof|ID:1322915202-1"},
   CellID->1180582011]},
 "Proof|ID:1322915202-1"->{
  Cell[1545927, 42297, 8473, 197, 70, "ProofInfo",
   CellTags->{"Proof|ID:1322915202", "Proof|ID:1322915202-1"},
   CellID->1180582011]},
 "Proof|ID:620036821"->{
  Cell[1554921, 42513, 8464, 196, 70, "ProofInfo",
   CellTags->{"Proof|ID:620036821", "Proof|ID:620036821-1"},
   CellID->1505716064]},
 "Proof|ID:620036821-1"->{
  Cell[1554921, 42513, 8464, 196, 70, "ProofInfo",
   CellTags->{"Proof|ID:620036821", "Proof|ID:620036821-1"},
   CellID->1505716064]},
 "Proof|ID:1111467670"->{
  Cell[1563905, 42728, 8472, 197, 70, "ProofInfo",
   CellTags->{"Proof|ID:1111467670", "Proof|ID:1111467670-1"},
   CellID->284241391]},
 "Proof|ID:1111467670-1"->{
  Cell[1563905, 42728, 8472, 197, 70, "ProofInfo",
   CellTags->{"Proof|ID:1111467670", "Proof|ID:1111467670-1"},
   CellID->284241391]},
 "ID:1314333166"->{
  Cell[1573348, 42958, 2753, 80, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1314333166", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAux"},
   CellID->1314333166]},
 "shape GBAux"->{
  Cell[1573348, 42958, 2753, 80, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1314333166", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAux"},
   CellID->1314333166]},
 "Proof|ID:1314333166"->{
  Cell[1576258, 43049, 9134, 217, 70, "ProofInfo",
   CellTags->{"Proof|ID:1314333166", "Proof|ID:1314333166-1"},
   CellID->888484883]},
 "Proof|ID:1314333166-1"->{
  Cell[1576258, 43049, 9134, 217, 70, "ProofInfo",
   CellTags->{"Proof|ID:1314333166", "Proof|ID:1314333166-1"},
   CellID->888484883]},
 "ID:1167572947"->{
  Cell[1586472, 43301, 2267, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1167572947", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership GBAux"},
   CellID->1167572947]},
 "membership GBAux"->{
  Cell[1586472, 43301, 2267, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1167572947", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership GBAux"},
   CellID->1167572947]},
 "Proof|ID:1167572947"->{
  Cell[1588895, 43377, 8362, 199, 70, "ProofInfo",
   CellTags->{"Proof|ID:1167572947", "Proof|ID:1167572947-1"},
   CellID->897701299]},
 "Proof|ID:1167572947-1"->{
  Cell[1588895, 43377, 8362, 199, 70, "ProofInfo",
   CellTags->{"Proof|ID:1167572947", "Proof|ID:1167572947-1"},
   CellID->897701299]},
 "ID:208228172"->{
  Cell[1597975, 43601, 783, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:208228172", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->208228172]},
 "ID:781335069"->{
  Cell[1599274, 43641, 2864, 81, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:781335069", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[RightArrow]\[RawStar]"},
   CellID->781335069]},
 "MMPairs \[RightArrow]\[RawStar]"->{
  Cell[1599274, 43641, 2864, 81, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:781335069", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[RightArrow]\[RawStar]"},
   CellID->781335069]},
 "Proof|ID:781335069"->{
  Cell[1602296, 43733, 12233, 278, 70, "ProofInfo",
   CellTags->{"Proof|ID:781335069", "Proof|ID:781335069-1"},
   CellID->283149330]},
 "Proof|ID:781335069-1"->{
  Cell[1602296, 43733, 12233, 278, 70, "ProofInfo",
   CellTags->{"Proof|ID:781335069", "Proof|ID:781335069-1"},
   CellID->283149330]},
 "ID:1452629404"->{
  Cell[1615472, 44044, 855, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1452629404", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1452629404]},
 "ID:1597922168"->{
  Cell[1616330, 44067, 1429, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1597922168", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxBase"},
   CellID->1597922168]},
 "ideal GBAuxBase"->{
  Cell[1616330, 44067, 1429, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1597922168", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxBase"},
   CellID->1597922168]},
 "ID:348489962"->{
  Cell[1617762, 44109, 1426, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:348489962", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec1"},
   CellID->348489962]},
 "ideal GBAuxRec1"->{
  Cell[1617762, 44109, 1426, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:348489962", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec1"},
   CellID->348489962]},
 "ID:1787031010"->{
  Cell[1619191, 44151, 1431, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1787031010", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec2"},
   CellID->1787031010]},
 "ideal GBAuxRec2"->{
  Cell[1619191, 44151, 1431, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1787031010", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec2"},
   CellID->1787031010]},
 "ID:963974147"->{
  Cell[1620625, 44193, 1426, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:963974147", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec3"},
   CellID->963974147]},
 "ideal GBAuxRec3"->{
  Cell[1620625, 44193, 1426, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:963974147", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec3"},
   CellID->963974147]},
 "ID:473957484"->{
  Cell[1622054, 44235, 1428, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:473957484", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec4"},
   CellID->473957484]},
 "ideal GBAuxRec4"->{
  Cell[1622054, 44235, 1428, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:473957484", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec4"},
   CellID->473957484]},
 "Proof|ID:1597922168"->{
  Cell[1623639, 44286, 8323, 195, 70, "ProofInfo",
   CellTags->{"Proof|ID:1597922168", "Proof|ID:1597922168-1"},
   CellID->1879777321]},
 "Proof|ID:1597922168-1"->{
  Cell[1623639, 44286, 8323, 195, 70, "ProofInfo",
   CellTags->{"Proof|ID:1597922168", "Proof|ID:1597922168-1"},
   CellID->1879777321]},
 "Proof|ID:348489962"->{
  Cell[1632483, 44500, 8313, 194, 70, "ProofInfo",
   CellTags->{"Proof|ID:348489962", "Proof|ID:348489962-1"},
   CellID->917719347]},
 "Proof|ID:348489962-1"->{
  Cell[1632483, 44500, 8313, 194, 70, "ProofInfo",
   CellTags->{"Proof|ID:348489962", "Proof|ID:348489962-1"},
   CellID->917719347]},
 "Proof|ID:1787031010"->{
  Cell[1641317, 44713, 8323, 195, 70, "ProofInfo",
   CellTags->{"Proof|ID:1787031010", "Proof|ID:1787031010-1"},
   CellID->1427595038]},
 "Proof|ID:1787031010-1"->{
  Cell[1641317, 44713, 8323, 195, 70, "ProofInfo",
   CellTags->{"Proof|ID:1787031010", "Proof|ID:1787031010-1"},
   CellID->1427595038]},
 "Proof|ID:963974147"->{
  Cell[1650162, 44927, 8314, 194, 70, "ProofInfo",
   CellTags->{"Proof|ID:963974147", "Proof|ID:963974147-1"},
   CellID->1497403325]},
 "Proof|ID:963974147-1"->{
  Cell[1650162, 44927, 8314, 194, 70, "ProofInfo",
   CellTags->{"Proof|ID:963974147", "Proof|ID:963974147-1"},
   CellID->1497403325]},
 "Proof|ID:473957484"->{
  Cell[1658996, 45140, 8314, 194, 70, "ProofInfo",
   CellTags->{"Proof|ID:473957484", "Proof|ID:473957484-1"},
   CellID->1947252190]},
 "Proof|ID:473957484-1"->{
  Cell[1658996, 45140, 8314, 194, 70, "ProofInfo",
   CellTags->{"Proof|ID:473957484", "Proof|ID:473957484-1"},
   CellID->1947252190]},
 "ID:1826585682"->{
  Cell[1668282, 45367, 2261, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1826585682", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAux"},
   CellID->1826585682]},
 "ideal GBAux"->{
  Cell[1668282, 45367, 2261, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1826585682", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAux"},
   CellID->1826585682]},
 "Proof|ID:1826585682"->{
  Cell[1670700, 45444, 8933, 213, 70, "ProofInfo",
   CellTags->{"Proof|ID:1826585682", "Proof|ID:1826585682-1"},
   CellID->98529934]},
 "Proof|ID:1826585682-1"->{
  Cell[1670700, 45444, 8933, 213, 70, "ProofInfo",
   CellTags->{"Proof|ID:1826585682", "Proof|ID:1826585682-1"},
   CellID->98529934]},
 "ID:416121228"->{
  Cell[1680381, 45684, 782, 20, 44, "GlobalDeclaration",
   CellTags->{
    "ID:416121228", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->416121228]},
 "ID:2050404037"->{
  Cell[1681970, 45732, 682, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:2050404037", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->2050404037]},
 "ID:1837575531"->{
  Cell[1682655, 45752, 4775, 124, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1837575531", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple"},
   CellID->1837575531]},
 "isCPTODOTuple"->{
  Cell[1682655, 45752, 4775, 124, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1837575531", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple"},
   CellID->1837575531]},
 "ID:346323949"->{
  Cell[1687433, 45878, 3358, 90, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:346323949", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1"},
   CellID->346323949]},
 "GBAuxCPConnectible1"->{
  Cell[1687433, 45878, 3358, 90, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:346323949", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1"},
   CellID->346323949]},
 "ID:1643593823"->{
  Cell[1690794, 45970, 2895, 77, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1643593823", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2"},
   CellID->1643593823]},
 "GBAuxCPConnectible2"->{
  Cell[1690794, 45970, 2895, 77, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1643593823", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2"},
   CellID->1643593823]},
 "ID:1213243088"->{
  Cell[1693692, 46049, 3067, 85, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1213243088", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3"},
   CellID->1213243088]},
 "GBAuxCPConnectible3"->{
  Cell[1693692, 46049, 3067, 85, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1213243088", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3"},
   CellID->1213243088]},
 "ID:949923506"->{
  Cell[1697747, 46169, 3785, 112, 130, "FormalTextInputFormula",
   CellTags->{
    "ID:949923506", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "h=0"},
   CellID->949923506]},
 "h=0"->{
  Cell[1697747, 46169, 3785, 112, 130, "FormalTextInputFormula",
   CellTags->{
    "ID:949923506", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "h=0"},
   CellID->949923506]},
 "Proof|ID:949923506"->{
  Cell[1701688, 46292, 15153, 346, 28, "ProofInfo",
   CellTags->{"Proof|ID:949923506", "Proof|ID:949923506-1"},
   CellID->221566149]},
 "Proof|ID:949923506-1"->{
  Cell[1701688, 46292, 15153, 346, 28, "ProofInfo",
   CellTags->{"Proof|ID:949923506", "Proof|ID:949923506-1"},
   CellID->221566149]},
 "ID:1322820809"->{
  Cell[1717818, 46671, 776, 21, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1322820809", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1322820809]},
 "ID:1772563804"->{
  Cell[1718597, 46694, 872, 23, 44, "GlobalDeclaration",
   CellTags->{
    "ID:1772563804", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1772563804]},
 "ID:667808408"->{
  Cell[1719472, 46719, 853, 22, 34, "GlobalDeclaration",
   CellTags->{
    "ID:667808408", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->667808408]},
 "ID:1601898370"->{
  Cell[1720328, 46743, 2593, 81, 69, "FormalTextInputFormula",
   CellTags->{
    "ID:1601898370", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "h\[NotEqual]0"},
   CellID->1601898370]},
 "h\[NotEqual]0"->{
  Cell[1720328, 46743, 2593, 81, 69, "FormalTextInputFormula",
   CellTags->{
    "ID:1601898370", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "h\[NotEqual]0"},
   CellID->1601898370]},
 "ID:622683853"->{
  Cell[1722924, 46826, 1850, 50, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:622683853", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpConnectible append h"},
   CellID->622683853]},
 "cpConnectible append h"->{
  Cell[1722924, 46826, 1850, 50, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:622683853", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpConnectible append h"},
   CellID->622683853]},
 "Proof|ID:1601898370"->{
  Cell[1724931, 46887, 15843, 359, 28, "ProofInfo",
   CellTags->{"Proof|ID:1601898370", "Proof|ID:1601898370-1"},
   CellID->76678778]},
 "Proof|ID:1601898370-1"->{
  Cell[1724931, 46887, 15843, 359, 28, "ProofInfo",
   CellTags->{"Proof|ID:1601898370", "Proof|ID:1601898370-1"},
   CellID->76678778]},
 "Proof|ID:622683853"->{
  Cell[1741295, 47265, 13743, 312, 26, "ProofInfo",
   CellTags->{"Proof|ID:622683853", "Proof|ID:622683853-1"},
   CellID->836026655]},
 "Proof|ID:622683853-1"->{
  Cell[1741295, 47265, 13743, 312, 26, "ProofInfo",
   CellTags->{"Proof|ID:622683853", "Proof|ID:622683853-1"},
   CellID->836026655]},
 "ID:1834517245"->{
  Cell[1755998, 47610, 3585, 95, 93, "FormalTextInputFormula",
   CellTags->{
    "ID:1834517245", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxTest alternative"},
   CellID->1834517245]},
 "GBAuxTest alternative"->{
  Cell[1755998, 47610, 3585, 95, 93, "FormalTextInputFormula",
   CellTags->{
    "ID:1834517245", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxTest alternative"},
   CellID->1834517245]},
 "ID:1923946854"->{
  Cell[1759586, 47707, 4183, 100, 183, "FormalTextInputFormula",
   CellTags->{
    "ID:1923946854", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxTest 1"},
   CellID->1923946854]},
 "GBAuxTest 1"->{
  Cell[1759586, 47707, 4183, 100, 183, "FormalTextInputFormula",
   CellTags->{
    "ID:1923946854", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxTest 1"},
   CellID->1923946854]},
 "Proof|ID:1834517245"->{
  Cell[1763926, 47818, 18116, 407, 28, "ProofInfo",
   CellTags->{"Proof|ID:1834517245", "Proof|ID:1834517245-1"},
   CellID->1216317993]},
 "Proof|ID:1834517245-1"->{
  Cell[1763926, 47818, 18116, 407, 28, "ProofInfo",
   CellTags->{"Proof|ID:1834517245", "Proof|ID:1834517245-1"},
   CellID->1216317993]},
 "Proof|ID:1923946854"->{
  Cell[1782564, 48244, 18127, 405, 26, "ProofInfo",
   CellTags->{"Proof|ID:1923946854", "Proof|ID:1923946854-1"},
   CellID->1529476345]},
 "Proof|ID:1923946854-1"->{
  Cell[1782564, 48244, 18127, 405, 26, "ProofInfo",
   CellTags->{"Proof|ID:1923946854", "Proof|ID:1923946854-1"},
   CellID->1529476345]},
 "ID:258768016"->{
  Cell[1810002, 48942, 1031, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:258768016", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 empty"},
   CellID->258768016]},
 "GBAuxCPConnectible1 empty"->{
  Cell[1810002, 48942, 1031, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:258768016", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 empty"},
   CellID->258768016]},
 "ID:1002747542"->{
  Cell[1811036, 48971, 855, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1002747542", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1002747542]},
 "ID:673303920"->{
  Cell[1811894, 48994, 946, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:673303920", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 GBAuxRec3"},
   CellID->673303920]},
 "GBAuxCPConnectible1 GBAuxRec3"->{
  Cell[1811894, 48994, 946, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:673303920", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 GBAuxRec3"},
   CellID->673303920]},
 "ID:1502806330"->{
  Cell[1812843, 49020, 949, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1502806330", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 GBAuxRec4"},
   CellID->1502806330]},
 "GBAuxCPConnectible1 GBAuxRec4"->{
  Cell[1812843, 49020, 949, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1502806330", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 GBAuxRec4"},
   CellID->1502806330]},
 "Proof|ID:258768016"->{
  Cell[1813949, 49055, 3543, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:258768016", "Proof|ID:258768016-1"},
   CellID->670832970]},
 "Proof|ID:258768016-1"->{
  Cell[1813949, 49055, 3543, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:258768016", "Proof|ID:258768016-1"},
   CellID->670832970]},
 "Proof|ID:673303920"->{
  Cell[1818081, 49167, 5540, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:673303920", "Proof|ID:673303920-1"},
   CellID->2063565608]},
 "Proof|ID:673303920-1"->{
  Cell[1818081, 49167, 5540, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:673303920", "Proof|ID:673303920-1"},
   CellID->2063565608]},
 "Proof|ID:1502806330"->{
  Cell[1824140, 49324, 5549, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:1502806330", "Proof|ID:1502806330-1"},
   CellID->1751898811]},
 "Proof|ID:1502806330-1"->{
  Cell[1824140, 49324, 5549, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:1502806330", "Proof|ID:1502806330-1"},
   CellID->1751898811]},
 "ID:1133484500"->{
  Cell[1830657, 49496, 1065, 30, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1133484500", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1133484500]},
 "ID:291397089"->{
  Cell[1831725, 49528, 852, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:291397089", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->291397089]},
 "ID:1494094918"->{
  Cell[1832580, 49551, 1277, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1494094918", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 GBAux"},
   CellID->1494094918]},
 "GBAuxCPConnectible1 GBAux"->{
  Cell[1832580, 49551, 1277, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1494094918", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 GBAux"},
   CellID->1494094918]},
 "ID:402861438"->{
  Cell[1833860, 49587, 1776, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:402861438", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpConnectible GBAux lmTuple"},
   CellID->402861438]},
 "cpConnectible GBAux lmTuple"->{
  Cell[1833860, 49587, 1776, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:402861438", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpConnectible GBAux lmTuple"},
   CellID->402861438]},
 "Proof|ID:1494094918"->{
  Cell[1835793, 49647, 7827, 187, 70, "ProofInfo",
   CellTags->{"Proof|ID:1494094918", "Proof|ID:1494094918-1"},
   CellID->1479447805]},
 "Proof|ID:1494094918-1"->{
  Cell[1835793, 49647, 7827, 187, 70, "ProofInfo",
   CellTags->{"Proof|ID:1494094918", "Proof|ID:1494094918-1"},
   CellID->1479447805]},
 "Proof|ID:402861438"->{
  Cell[1844141, 49853, 11025, 256, 70, "ProofInfo",
   CellTags->{"Proof|ID:402861438", "Proof|ID:402861438-1"},
   CellID->111398506]},
 "Proof|ID:402861438-1"->{
  Cell[1844141, 49853, 11025, 256, 70, "ProofInfo",
   CellTags->{"Proof|ID:402861438", "Proof|ID:402861438-1"},
   CellID->111398506]},
 "ID:1173305577"->{
  Cell[1856359, 50150, 800, 21, 33, "GlobalDeclaration",
   CellTags->{
    "ID:1173305577", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1173305577]},
 "ID:1545923689"->{
  Cell[1857162, 50173, 651, 16, 30, "GlobalDeclaration",
   CellTags->{
    "ID:1545923689", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1545923689]},
 "ID:1684239249"->{
  Cell[1857816, 50191, 1393, 39, 43, "FormalTextInputFormula",
   CellTags->{
    "ID:1684239249", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple empty"},
   CellID->1684239249]},
 "isCPTODOTuple empty"->{
  Cell[1857816, 50191, 1393, 39, 43, "FormalTextInputFormula",
   CellTags->{
    "ID:1684239249", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple empty"},
   CellID->1684239249]},
 "ID:547535692"->{
  Cell[1859212, 50232, 1629, 45, 43, "FormalTextInputFormula",
   CellTags->{
    "ID:547535692", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple prepend"},
   CellID->547535692]},
 "isCPTODOTuple prepend"->{
  Cell[1859212, 50232, 1629, 45, 43, "FormalTextInputFormula",
   CellTags->{
    "ID:547535692", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple prepend"},
   CellID->547535692]},
 "ID:812188873"->{
  Cell[1860844, 50279, 2158, 60, 89, "FormalTextInputFormula",
   CellTags->{
    "ID:812188873", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple isTODOTuple"},
   CellID->812188873]},
 "isCPTODOTuple isTODOTuple"->{
  Cell[1860844, 50279, 2158, 60, 89, "FormalTextInputFormula",
   CellTags->{
    "ID:812188873", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple isTODOTuple"},
   CellID->812188873]},
 "ID:1105161129"->{
  Cell[1863005, 50341, 1407, 41, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1105161129", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple lmTuple"},
   CellID->1105161129]},
 "isCPTODOTuple lmTuple"->{
  Cell[1863005, 50341, 1407, 41, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:1105161129", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple lmTuple"},
   CellID->1105161129]},
 "ID:981490890"->{
  Cell[1864415, 50384, 1476, 44, 43, "FormalTextInputFormula",
   CellTags->{
    "ID:981490890", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple 0"},
   CellID->981490890]},
 "isCPTODOTuple 0"->{
  Cell[1864415, 50384, 1476, 44, 43, "FormalTextInputFormula",
   CellTags->{
    "ID:981490890", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple 0"},
   CellID->981490890]},
 "Proof|ID:1684239249"->{
  Cell[1866048, 50439, 6983, 169, 28, "ProofInfo",
   CellTags->{"Proof|ID:1684239249", "Proof|ID:1684239249-1"},
   CellID->74113726]},
 "Proof|ID:1684239249-1"->{
  Cell[1866048, 50439, 6983, 169, 28, "ProofInfo",
   CellTags->{"Proof|ID:1684239249", "Proof|ID:1684239249-1"},
   CellID->74113726]},
 "Proof|ID:547535692"->{
  Cell[1873552, 50627, 7700, 184, 26, "ProofInfo",
   CellTags->{"Proof|ID:547535692", "Proof|ID:547535692-1"},
   CellID->1769495093]},
 "Proof|ID:547535692-1"->{
  Cell[1873552, 50627, 7700, 184, 26, "ProofInfo",
   CellTags->{"Proof|ID:547535692", "Proof|ID:547535692-1"},
   CellID->1769495093]},
 "Proof|ID:812188873"->{
  Cell[1881772, 50830, 9452, 220, 26, "ProofInfo",
   CellTags->{"Proof|ID:812188873", "Proof|ID:812188873-1"},
   CellID->1479812468]},
 "Proof|ID:812188873-1"->{
  Cell[1881772, 50830, 9452, 220, 26, "ProofInfo",
   CellTags->{"Proof|ID:812188873", "Proof|ID:812188873-1"},
   CellID->1479812468]},
 "Proof|ID:1105161129"->{
  Cell[1891744, 51069, 7709, 187, 26, "ProofInfo",
   CellTags->{"Proof|ID:1105161129", "Proof|ID:1105161129-1"},
   CellID->2103372990]},
 "Proof|ID:1105161129-1"->{
  Cell[1891744, 51069, 7709, 187, 26, "ProofInfo",
   CellTags->{"Proof|ID:1105161129", "Proof|ID:1105161129-1"},
   CellID->2103372990]},
 "Proof|ID:981490890"->{
  Cell[1899975, 51275, 6987, 172, 26, "ProofInfo",
   CellTags->{"Proof|ID:981490890", "Proof|ID:981490890-1"},
   CellID->296074352]},
 "Proof|ID:981490890-1"->{
  Cell[1899975, 51275, 6987, 172, 26, "ProofInfo",
   CellTags->{"Proof|ID:981490890", "Proof|ID:981490890-1"},
   CellID->296074352]},
 "ID:681053692"->{
  Cell[1907942, 51481, 1163, 32, 44, "GlobalDeclaration",
   CellTags->{
    "ID:681053692", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->681053692]},
 "ID:995853659"->{
  Cell[1909108, 51515, 797, 21, 33, "GlobalDeclaration",
   CellTags->{
    "ID:995853659", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->995853659]},
 "ID:218025475"->{
  Cell[1909908, 51538, 876, 23, 44, "GlobalDeclaration",
   CellTags->{
    "ID:218025475", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->218025475]},
 "ID:2052980984"->{
  Cell[1910787, 51563, 1195, 30, 34, "GlobalDeclaration",
   CellTags->{
    "ID:2052980984", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->2052980984]},
 "ID:925810378"->{
  Cell[1911985, 51595, 1289, 36, 48, "GlobalDeclaration",
   CellTags->{
    "ID:925810378", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->925810378]},
 "ID:423392620"->{
  Cell[1913277, 51633, 1524, 42, 81, "FormalTextInputFormula",
   CellTags->{
    "ID:423392620", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple 1"},
   CellID->423392620]},
 "isCPTODOTuple 1"->{
  Cell[1913277, 51633, 1524, 42, 81, "FormalTextInputFormula",
   CellTags->{
    "ID:423392620", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple 1"},
   CellID->423392620]},
 "ID:1533257011"->{
  Cell[1914804, 51677, 1248, 35, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:1533257011", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple 2"},
   CellID->1533257011]},
 "isCPTODOTuple 2"->{
  Cell[1914804, 51677, 1248, 35, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:1533257011", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple 2"},
   CellID->1533257011]},
 "Proof|ID:423392620"->{
  Cell[1916208, 51723, 17593, 390, 28, "ProofInfo",
   CellTags->{"Proof|ID:423392620", "Proof|ID:423392620-1"},
   CellID->117206802]},
 "Proof|ID:423392620-1"->{
  Cell[1916208, 51723, 17593, 390, 28, "ProofInfo",
   CellTags->{"Proof|ID:423392620", "Proof|ID:423392620-1"},
   CellID->117206802]},
 "Proof|ID:1533257011"->{
  Cell[1934319, 52132, 15598, 353, 26, "ProofInfo",
   CellTags->{"Proof|ID:1533257011", "Proof|ID:1533257011-1"},
   CellID->141690382]},
 "Proof|ID:1533257011-1"->{
  Cell[1934319, 52132, 15598, 353, 26, "ProofInfo",
   CellTags->{"Proof|ID:1533257011", "Proof|ID:1533257011-1"},
   CellID->141690382]},
 "ID:1154842295"->{
  Cell[1975295, 53267, 857, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1154842295", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1154842295]},
 "ID:1427865659"->{
  Cell[1976155, 53290, 949, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1427865659", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxBase"},
   CellID->1427865659]},
 "GBAuxCPConnectible2 GBAuxBase"->{
  Cell[1976155, 53290, 949, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1427865659", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxBase"},
   CellID->1427865659]},
 "ID:802696830"->{
  Cell[1977107, 53316, 948, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:802696830", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec1"},
   CellID->802696830]},
 "GBAuxCPConnectible2 GBAuxRec1"->{
  Cell[1977107, 53316, 948, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:802696830", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec1"},
   CellID->802696830]},
 "ID:668126868"->{
  Cell[1978058, 53342, 947, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:668126868", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec2"},
   CellID->668126868]},
 "GBAuxCPConnectible2 GBAuxRec2"->{
  Cell[1978058, 53342, 947, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:668126868", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec2"},
   CellID->668126868]},
 "ID:1848124281"->{
  Cell[1979008, 53368, 949, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1848124281", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec3"},
   CellID->1848124281]},
 "GBAuxCPConnectible2 GBAuxRec3"->{
  Cell[1979008, 53368, 949, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1848124281", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec3"},
   CellID->1848124281]},
 "ID:6143564"->{
  Cell[1979960, 53394, 942, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:6143564", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec4"},
   CellID->6143564]},
 "GBAuxCPConnectible2 GBAuxRec4"->{
  Cell[1979960, 53394, 942, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:6143564", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec4"},
   CellID->6143564]},
 "Proof|ID:1427865659"->{
  Cell[1981060, 53429, 5548, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:1427865659", "Proof|ID:1427865659-1"},
   CellID->544375535]},
 "Proof|ID:1427865659-1"->{
  Cell[1981060, 53429, 5548, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:1427865659", "Proof|ID:1427865659-1"},
   CellID->544375535]},
 "Proof|ID:802696830"->{
  Cell[1987129, 53587, 5539, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:802696830", "Proof|ID:802696830-1"},
   CellID->335517243]},
 "Proof|ID:802696830-1"->{
  Cell[1987129, 53587, 5539, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:802696830", "Proof|ID:802696830-1"},
   CellID->335517243]},
 "Proof|ID:668126868"->{
  Cell[1993188, 53744, 5539, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:668126868", "Proof|ID:668126868-1"},
   CellID->705494055]},
 "Proof|ID:668126868-1"->{
  Cell[1993188, 53744, 5539, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:668126868", "Proof|ID:668126868-1"},
   CellID->705494055]},
 "Proof|ID:1848124281"->{
  Cell[1999248, 53901, 5548, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:1848124281", "Proof|ID:1848124281-1"},
   CellID->885145762]},
 "Proof|ID:1848124281-1"->{
  Cell[1999248, 53901, 5548, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:1848124281", "Proof|ID:1848124281-1"},
   CellID->885145762]},
 "Proof|ID:6143564"->{
  Cell[2005318, 54059, 5530, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:6143564", "Proof|ID:6143564-1"},
   CellID->1629480533]},
 "Proof|ID:6143564-1"->{
  Cell[2005318, 54059, 5530, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:6143564", "Proof|ID:6143564-1"},
   CellID->1629480533]},
 "ID:1620493315"->{
  Cell[2011814, 54230, 1041, 30, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1620493315", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1620493315]},
 "ID:1672348970"->{
  Cell[2012858, 54262, 831, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1672348970", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1672348970]},
 "ID:1826259623"->{
  Cell[2013692, 54285, 912, 24, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1826259623", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1826259623]},
 "ID:1814594394"->{
  Cell[2014607, 54311, 979, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1814594394", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAux"},
   CellID->1814594394]},
 "GBAuxCPConnectible2 GBAux"->{
  Cell[2014607, 54311, 979, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1814594394", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAux"},
   CellID->1814594394]},
 "ID:1024326782"->{
  Cell[2015589, 54337, 743, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1024326782", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1024326782]},
 "ID:585977074"->{
  Cell[2016335, 54358, 1057, 30, 70, "GlobalDeclaration",
   CellTags->{
    "ID:585977074", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->585977074]},
 "ID:1959330209"->{
  Cell[2017395, 54390, 2212, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1959330209", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 update"},
   CellID->1959330209]},
 "GBAuxCPConnectible2 update"->{
  Cell[2017395, 54390, 2212, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1959330209", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 update"},
   CellID->1959330209]},
 "Proof|ID:1814594394"->{
  Cell[2019764, 54466, 7826, 187, 70, "ProofInfo",
   CellTags->{"Proof|ID:1814594394", "Proof|ID:1814594394-1"},
   CellID->378539769]},
 "Proof|ID:1814594394-1"->{
  Cell[2019764, 54466, 7826, 187, 70, "ProofInfo",
   CellTags->{"Proof|ID:1814594394", "Proof|ID:1814594394-1"},
   CellID->378539769]},
 "Proof|ID:1959330209"->{
  Cell[2028111, 54672, 18875, 415, 70, "ProofInfo",
   CellTags->{"Proof|ID:1959330209", "Proof|ID:1959330209-1"},
   CellID->1849443430]},
 "Proof|ID:1959330209-1"->{
  Cell[2028111, 54672, 18875, 415, 70, "ProofInfo",
   CellTags->{"Proof|ID:1959330209", "Proof|ID:1959330209-1"},
   CellID->1849443430]},
 "ID:33155789"->{
  Cell[2048224, 55130, 851, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:33155789", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->33155789]},
 "ID:1691022439"->{
  Cell[2049078, 55153, 949, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1691022439", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxBase"},
   CellID->1691022439]},
 "GBAuxCPConnectible3 GBAuxBase"->{
  Cell[2049078, 55153, 949, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1691022439", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxBase"},
   CellID->1691022439]},
 "ID:544907973"->{
  Cell[2050030, 55179, 946, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:544907973", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec1"},
   CellID->544907973]},
 "GBAuxCPConnectible3 GBAuxRec1"->{
  Cell[2050030, 55179, 946, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:544907973", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec1"},
   CellID->544907973]},
 "ID:1039424559"->{
  Cell[2050979, 55205, 948, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1039424559", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec2"},
   CellID->1039424559]},
 "GBAuxCPConnectible3 GBAuxRec2"->{
  Cell[2050979, 55205, 948, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1039424559", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec2"},
   CellID->1039424559]},
 "ID:1704484269"->{
  Cell[2051930, 55231, 949, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1704484269", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec3"},
   CellID->1704484269]},
 "GBAuxCPConnectible3 GBAuxRec3"->{
  Cell[2051930, 55231, 949, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1704484269", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec3"},
   CellID->1704484269]},
 "ID:2114764931"->{
  Cell[2052882, 55257, 949, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2114764931", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec4"},
   CellID->2114764931]},
 "GBAuxCPConnectible3 GBAuxRec4"->{
  Cell[2052882, 55257, 949, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2114764931", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec4"},
   CellID->2114764931]},
 "Proof|ID:1691022439"->{
  Cell[2053988, 55292, 5549, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:1691022439", "Proof|ID:1691022439-1"},
   CellID->1214006285]},
 "Proof|ID:1691022439-1"->{
  Cell[2053988, 55292, 5549, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:1691022439", "Proof|ID:1691022439-1"},
   CellID->1214006285]},
 "Proof|ID:544907973"->{
  Cell[2060059, 55450, 5539, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:544907973", "Proof|ID:544907973-1"},
   CellID->708615290]},
 "Proof|ID:544907973-1"->{
  Cell[2060059, 55450, 5539, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:544907973", "Proof|ID:544907973-1"},
   CellID->708615290]},
 "Proof|ID:1039424559"->{
  Cell[2066161, 55608, 5548, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:1039424559", "Proof|ID:1039424559-1"},
   CellID->495379081]},
 "Proof|ID:1039424559-1"->{
  Cell[2066161, 55608, 5548, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:1039424559", "Proof|ID:1039424559-1"},
   CellID->495379081]},
 "Proof|ID:1704484269"->{
  Cell[2072230, 55766, 5548, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:1704484269", "Proof|ID:1704484269-1"},
   CellID->122315265]},
 "Proof|ID:1704484269-1"->{
  Cell[2072230, 55766, 5548, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:1704484269", "Proof|ID:1704484269-1"},
   CellID->122315265]},
 "Proof|ID:2114764931"->{
  Cell[2078300, 55924, 5549, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:2114764931", "Proof|ID:2114764931-1"},
   CellID->1252700938]},
 "Proof|ID:2114764931-1"->{
  Cell[2078300, 55924, 5549, 139, 70, "ProofInfo",
   CellTags->{"Proof|ID:2114764931", "Proof|ID:2114764931-1"},
   CellID->1252700938]},
 "ID:1693763802"->{
  Cell[2084817, 56096, 1993, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1693763802", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAux"},
   CellID->1693763802]},
 "GBAuxCPConnectible3 GBAux"->{
  Cell[2084817, 56096, 1993, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1693763802", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAux"},
   CellID->1693763802]},
 "Proof|ID:1693763802"->{
  Cell[2086967, 56163, 7826, 187, 70, "ProofInfo",
   CellTags->{"Proof|ID:1693763802", "Proof|ID:1693763802-1"},
   CellID->874031660]},
 "Proof|ID:1693763802-1"->{
  Cell[2086967, 56163, 7826, 187, 70, "ProofInfo",
   CellTags->{"Proof|ID:1693763802", "Proof|ID:1693763802-1"},
   CellID->874031660]},
 "ID:439420249"->{
  Cell[2096171, 56400, 946, 26, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:439420249", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GB empty", "IR"},
   CellID->439420249]},
 "GB empty"->{
  Cell[2096171, 56400, 946, 26, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:439420249", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GB empty", "IR"},
   CellID->439420249]},
 "Proof|ID:439420249"->{
  Cell[2097274, 56437, 2847, 84, 28, "ProofInfo",
   CellTags->{"Proof|ID:439420249", "Proof|ID:439420249-1"},
   CellID->1181943649]},
 "Proof|ID:439420249-1"->{
  Cell[2097274, 56437, 2847, 84, 28, "ProofInfo",
   CellTags->{"Proof|ID:439420249", "Proof|ID:439420249-1"},
   CellID->1181943649]},
 "ID:298471155"->{
  Cell[2101083, 56554, 781, 20, 44, "GlobalDeclaration",
   CellTags->{
    "ID:298471155", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->298471155]},
 "ID:1615450915"->{
  Cell[2101867, 56576, 747, 20, 43, "GlobalDeclaration",
   CellTags->{
    "ID:1615450915", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
   CellID->1615450915]},
 "ID:209199628"->{
  Cell[2102617, 56598, 1118, 35, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:209199628", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GB"},
   CellID->209199628]},
 "shape GB"->{
  Cell[2102617, 56598, 1118, 35, 44, "FormalTextInputFormula",
   CellTags->{
    "ID:209199628", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GB"},
   CellID->209199628]},
 "ID:448006251"->{
  Cell[2103738, 56635, 868, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:448006251", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership GB", "IR"},
   CellID->448006251]},
 "membership GB"->{
  Cell[2103738, 56635, 868, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:448006251", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership GB", "IR"},
   CellID->448006251]},
 "ID:739795809"->{
  Cell[2104609, 56660, 910, 26, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:739795809", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GB", "IR"},
   CellID->739795809]},
 "ideal GB"->{
  Cell[2104609, 56660, 910, 26, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:739795809", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GB", "IR"},
   CellID->739795809]},
 "ID:1550516051"->{
  Cell[2105522, 56688, 832, 22, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1550516051", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isGroebnerBasis GB"},
   CellID->1550516051]},
 "isGroebnerBasis GB"->{
  Cell[2105522, 56688, 832, 22, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:1550516051", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isGroebnerBasis GB"},
   CellID->1550516051]},
 "Proof|ID:209199628"->{
  Cell[2106510, 56721, 4471, 118, 28, "ProofInfo",
   CellTags->{"Proof|ID:209199628", "Proof|ID:209199628-1"},
   CellID->2008671786]},
 "Proof|ID:209199628-1"->{
  Cell[2106510, 56721, 4471, 118, 28, "ProofInfo",
   CellTags->{"Proof|ID:209199628", "Proof|ID:209199628-1"},
   CellID->2008671786]},
 "Proof|ID:448006251"->{
  Cell[2111565, 56859, 3740, 100, 26, "ProofInfo",
   CellTags->{"Proof|ID:448006251", "Proof|ID:448006251-1"},
   CellID->2018395045]},
 "Proof|ID:448006251-1"->{
  Cell[2111565, 56859, 3740, 100, 26, "ProofInfo",
   CellTags->{"Proof|ID:448006251", "Proof|ID:448006251-1"},
   CellID->2018395045]},
 "Proof|ID:739795809"->{
  Cell[2115826, 56978, 4257, 114, 26, "ProofInfo",
   CellTags->{"Proof|ID:739795809", "Proof|ID:739795809-1"},
   CellID->144088343]},
 "Proof|ID:739795809-1"->{
  Cell[2115826, 56978, 4257, 114, 26, "ProofInfo",
   CellTags->{"Proof|ID:739795809", "Proof|ID:739795809-1"},
   CellID->144088343]},
 "Proof|ID:1550516051"->{
  Cell[2120601, 57111, 3656, 99, 26, "ProofInfo",
   CellTags->{"Proof|ID:1550516051", "Proof|ID:1550516051-1"},
   CellID->856857959]},
 "Proof|ID:1550516051-1"->{
  Cell[2120601, 57111, 3656, 99, 26, "ProofInfo",
   CellTags->{"Proof|ID:1550516051", "Proof|ID:1550516051-1"},
   CellID->856857959]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"ID:1079943823", 2162409, 57994},
 {"Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", 2162719, 58002},
 {"print", 2234054, 59858},
 {"ID:1033951332", 2234296, 59865},
 {"processed", 2234538, 59872},
 {"ID:1967331668", 2234784, 59879},
 {"ID:1698292943", 2235011, 59886},
 {"pairs 1", 2235249, 59893},
 {"ID:1961920283", 2235493, 59900},
 {"pairs 2", 2235731, 59907},
 {"ID:1516706954", 2235975, 59914},
 {"allPairs", 2236215, 59921},
 {"ID:329670188", 2236459, 59928},
 {"update", 2236693, 59935},
 {"ID:1283917386", 2236934, 59942},
 {"ID:14652249", 2237159, 59949},
 {"ID:1012271681", 2237382, 59956},
 {"NatPairs \[Element]", 2237644, 59963},
 {"ID:988862829", 2237899, 59970},
 {"ID:16377100", 2238123, 59977},
 {"NatPairs \[Element] ProdDomain", 2238404, 59984},
 {"ID:136514135", 2238667, 59991},
 {"NatPairs \[Element] ProdDomain DomainTuples", 2238976, 59998},
 {"ID:1872634730", 2239255, 60005},
 {"NatPairs \[Element] alternative", 2239548, 60012},
 {"IR", 2239812, 60019},
 {"ID:1931599387", 2248036, 60219},
 {"NatPairs< \[Element] alternative", 2248331, 60226},
 {"ID:2124565744", 2248607, 60233},
 {"NatPairs\[LessEqual] \[Element] alternative", 2248924, 60240},
 {"ID:1917398594", 2249211, 60247},
 {"NatPairs< \[Element] alternative II", 2249506, 60254},
 {"ID:272474466", 2249778, 60261},
 {"NatPairs\[LessEqual] \[Element] alternative II", 2250093, 60268},
 {"ID:2005363738", 2250375, 60275},
 {"NatPairs< \[LessEqual]", 2250644, 60282},
 {"ID:1661715038", 2250904, 60289},
 {"DomainTuples NatPairs< \[LessEqual]", 2251199, 60296},
 {"ID:1020080047", 2251472, 60303},
 {"ID:1327808237", 2251700, 60310},
 {"NatPairs +1", 2251947, 60317},
 {"ID:798140928", 2252195, 60324},
 {"DomainTuples NatPairs +1", 2252466, 60331},
 {"ID:2111464127", 2252726, 60338},
 {"ID:1834895057", 2252954, 60345},
 {"NatPairs \[GreaterEqual]", 2253227, 60352},
 {"ID:1703465728", 2253489, 60359},
 {"DomainTuples NatPairs \[GreaterEqual]", 2253788, 60366},
 {"Proof|ID:16377100", 2254067, 60373},
 {"Proof|ID:16377100-1", 2254223, 60377},
 {"Proof|ID:136514135", 2254378, 60381},
 {"Proof|ID:136514135-1", 2254539, 60385},
 {"Proof|ID:1872634730", 2254699, 60389},
 {"Proof|ID:1872634730-1", 2254864, 60393},
 {"Proof|ID:1931599387", 2255027, 60397},
 {"Proof|ID:1931599387-1", 2255190, 60401},
 {"Proof|ID:2124565744", 2255351, 60405},
 {"Proof|ID:2124565744-1", 2255516, 60409},
 {"Proof|ID:1917398594", 2255679, 60413},
 {"Proof|ID:1917398594-1", 2255844, 60417},
 {"Proof|ID:272474466", 2256006, 60421},
 {"Proof|ID:272474466-1", 2256167, 60425},
 {"Proof|ID:2005363738", 2256327, 60429},
 {"Proof|ID:2005363738-1", 2256490, 60433},
 {"Proof|ID:1661715038", 2256651, 60437},
 {"Proof|ID:1661715038-1", 2256815, 60441},
 {"Proof|ID:1327808237", 2256977, 60445},
 {"Proof|ID:1327808237-1", 2257142, 60449},
 {"Proof|ID:798140928", 2257304, 60453},
 {"Proof|ID:798140928-1", 2257465, 60457},
 {"Proof|ID:1834895057", 2257625, 60461},
 {"Proof|ID:1834895057-1", 2257790, 60465},
 {"Proof|ID:1703465728", 2257953, 60469},
 {"Proof|ID:1703465728-1", 2258118, 60473},
 {"ID:971135058", 2258274, 60477},
 {"ID:1202991311", 2258500, 60484},
 {"membership pairs unary", 2258775, 60491},
 {"ID:1932739315", 2259041, 60498},
 {"pairs unary content'", 2259306, 60505},
 {"ID:1628109432", 2259564, 60512},
 {"pairs unary content", 2259827, 60519},
 {"ID:540671405", 2260083, 60526},
 {"membership pairs binary", 2260358, 60533},
 {"ID:881225230", 2260622, 60540},
 {"pairs binary content'", 2260887, 60547},
 {"ID:717573619", 2261143, 60554},
 {"pairs binary content", 2261406, 60561},
 {"ID:794667329", 2261661, 60568},
 {"allPairs 0", 2261909, 60575},
 {"ID:1385740688", 2262160, 60582},
 {"membership allPairs", 2262429, 60589},
 {"ID:1430860208", 2262692, 60596},
 {"allPairs content", 2262949, 60603},
 {"Proof|ID:1202991311", 2263209, 60610},
 {"Proof|ID:1202991311-1", 2263373, 60614},
 {"Proof|ID:1932739315", 2263535, 60618},
 {"Proof|ID:1932739315-1", 2263700, 60622},
 {"Proof|ID:1628109432", 2263863, 60626},
 {"Proof|ID:1628109432-1", 2264026, 60630},
 {"Proof|ID:540671405", 2264186, 60634},
 {"Proof|ID:540671405-1", 2264347, 60638},
 {"Proof|ID:881225230", 2264506, 60642},
 {"Proof|ID:881225230-1", 2264667, 60646},
 {"Proof|ID:717573619", 2264826, 60650},
 {"Proof|ID:717573619-1", 2264988, 60654},
 {"Proof|ID:794667329", 2265148, 60658},
 {"Proof|ID:794667329-1", 2265309, 60662},
 {"Proof|ID:1385740688", 2265469, 60666},
 {"Proof|ID:1385740688-1", 2265633, 60670},
 {"Proof|ID:1430860208", 2265795, 60674},
 {"Proof|ID:1430860208-1", 2265960, 60678},
 {"ID:963656485", 2266116, 60682},
 {"ID:1750952092", 2266342, 60689},
 {"ID:1500640019", 2266571, 60696},
 {"ID:723173326", 2266798, 60703},
 {"membership update", 2267061, 60710},
 {"ID:392184336", 2267319, 60717},
 {"update content I", 2267574, 60724},
 {"ID:341058317", 2267825, 60731},
 {"update content II", 2268082, 60738},
 {"Proof|ID:723173326", 2268340, 60745},
 {"Proof|ID:723173326-1", 2268502, 60749},
 {"Proof|ID:392184336", 2268662, 60753},
 {"Proof|ID:392184336-1", 2268823, 60757},
 {"Proof|ID:341058317", 2268982, 60761},
 {"Proof|ID:341058317-1", 2269144, 60765},
 {"ID:1597870023", 2269299, 60769},
 {"processed alternative", 2269572, 60776},
 {"ID:40246499", 2269835, 60783},
 {"processed symmetric", 2270094, 60790},
 {"ID:1131819573", 2270347, 60797},
 {"ID:1083804603", 2270575, 60804},
 {"processed allPairs", 2270836, 60811},
 {"ID:316507172", 2271091, 60818},
 {"processed update", 2271346, 60825},
 {"ID:1179241697", 2271598, 60832},
 {"processed prepend", 2271857, 60839},
 {"Proof|ID:1597870023", 2272118, 60846},
 {"Proof|ID:1597870023-1", 2272282, 60850},
 {"Proof|ID:40246499", 2272442, 60854},
 {"Proof|ID:40246499-1", 2272600, 60858},
 {"Proof|ID:1083804603", 2272758, 60862},
 {"Proof|ID:1083804603-1", 2272922, 60866},
 {"Proof|ID:316507172", 2273083, 60870},
 {"Proof|ID:316507172-1", 2273244, 60874},
 {"Proof|ID:1179241697", 2273404, 60878},
 {"Proof|ID:1179241697-1", 2273569, 60882},
 {"ID:1573502139", 2273726, 60886},
 {"ID:377704681", 2273953, 60893},
 {"ID:1468120413", 2274179, 60900},
 {"ID:1774031579", 2274408, 60907},
 {"GroebnerRing \[Element]", 2274679, 60914},
 {"ID:182210603", 2274939, 60921},
 {"GroebnerRing 0", 2275189, 60928},
 {"ID:814933967", 2275437, 60935},
 {"GroebnerRing 1", 2275687, 60942},
 {"ID:1684568525", 2275936, 60949},
 {"GroebnerRing +", 2276188, 60956},
 {"ID:584522550", 2276438, 60963},
 {"GroebnerRing -", 2276688, 60970},
 {"ID:1458964616", 2276937, 60977},
 {"GroebnerRing - unary", 2277201, 60984},
 {"ID:1496488880", 2277458, 60991},
 {"GroebnerRing \[RawStar]", 2277728, 60998},
 {"ID:342394137", 2277987, 61005},
 {"GroebnerRing \[Succeeds]", 2278258, 61012},
 {"ID:1566608311", 2278518, 61019},
 {"GroebnerRing multN", 2278778, 61026},
 {"ID:628841792", 2279032, 61033},
 {"GroebnerRing mult 1", 2279293, 61040},
 {"ID:378340035", 2279547, 61047},
 {"GroebnerRing mult 2", 2279808, 61054},
 {"ID:822076309", 2280062, 61061},
 {"GroebnerRing lcrd", 2280319, 61068},
 {"ID:849156542", 2280571, 61075},
 {"GroebnerRing rdm", 2280825, 61082},
 {"ID:809363439", 2281075, 61089},
 {"GroebnerRing lcrd unary", 2281343, 61096},
 {"ID:1671864047", 2281601, 61103},
 {"GroebnerRing cpm", 2281857, 61110},
 {"ID:558323540", 2282109, 61117},
 {"GroebnerRing rd", 2282362, 61124},
 {"ID:1403530194", 2282613, 61131},
 {"GroebnerRing rdT", 2282870, 61138},
 {"ID:689058884", 2283123, 61145},
 {"GroebnerRing trd", 2283378, 61152},
 {"ID:1528624471", 2283630, 61159},
 {"GroebnerRing lmTuple", 2283895, 61166},
 {"ID:1192872489", 2284153, 61173},
 {"GroebnerRing lmTupleAux 1", 2284427, 61180},
 {"ID:1479034574", 2284689, 61187},
 {"GroebnerRing lmTupleAux 2", 2284964, 61194},
 {"ID:1529751135", 2285227, 61201},
 {"GroebnerRing cpmTuple 1", 2285498, 61208},
 {"ID:551185449", 2285758, 61215},
 {"GroebnerRing cpmTuple 2", 2286027, 61222},
 {"ID:1533920470", 2286286, 61229},
 {"GroebnerRing GB", 2286541, 61236},
 {"ID:1976234692", 2286794, 61243},
 {"GroebnerRing GBAux 1", 2287059, 61250},
 {"ID:144716839", 2287316, 61257},
 {"GroebnerRing GBAux 2", 2287580, 61264},
 {"ID:189044945", 2287836, 61271},
 {"GroebnerRing GBAux 3", 2288101, 61278},
 {"ID:1766960499", 2288359, 61285},
 {"GroebnerRing chainCrit", 2288629, 61292},
 {"ID:1777604038", 2288890, 61299},
 {"GroebnerRing ard", 2289148, 61306},
 {"ID:620529741", 2289402, 61313},
 {"GroebnerRing ardAux 1", 2289668, 61320},
 {"ID:174904624", 2289925, 61327},
 {"GroebnerRing ardAux 2", 2290191, 61334},
 {"ID:336511876", 2290448, 61341},
 {"\[ScriptCapitalG] \[ScriptCapitalQ]", 2290742, 61348},
 {"ID:193404746", 2291013, 61355},
 {"\[ScriptCapitalG] \[ScriptCapitalZ]", 2291306, 61362},
 {"ID:1821700886", 2291577, 61369},
 {"\[ScriptCapitalG] \[ScriptCapitalZ]24", 2291876, 61376},
 {"ID:1158015694", 2292151, 61383},
 {"\[ScriptCapitalG] \[ScriptCapitalQ]xy", 2292451, 61390},
 {"ID:1921709458", 2292727, 61397},
 {"\[ScriptCapitalZ]xy", 2292991, 61404},
 {"ID:281007089", 2293248, 61411},
 {"\[ScriptCapitalG] \[ScriptCapitalZ]xy", 2293546, 61418},
 {"ID:1512392075", 2293820, 61425},
 {"\[ScriptCapitalZ]24xy", 2294087, 61432},
 {"ID:543229319", 2294345, 61439},
 {"\[ScriptCapitalG] \[ScriptCapitalZ]24xy", 2294647, 61446},
 {"ID:280174026", 2294922, 61453},
 {"\[ScriptCapitalG] isCommRing1", 2295203, 61460},
 {"Proof|ID:280174026", 2295473, 61467},
 {"Proof|ID:280174026-1", 2295633, 61471},
 {"ID:1057748777", 2295786, 61475},
 {"ID:464516108", 2296014, 61482},
 {"ID:1148013153", 2296241, 61489},
 {"DomainSets \[ScriptCapitalG]", 2296529, 61496},
 {"ID:1668771272", 2296802, 61503},
 {"DomainTuples \[ScriptCapitalG]", 2297094, 61510},
 {"ID:1928426019", 2297369, 61517},
 {"\[Sum] \[ScriptCapitalG]", 2297649, 61524},
 {"ID:546598180", 2297917, 61531},
 {"ID:581321955", 2298143, 61538},
 {"reduceM \[ScriptCapitalG]", 2298423, 61545},
 {"ID:405682622", 2298690, 61552},
 {"\[RightArrow] \[ScriptCapitalG] multiplier", 2299008, 61560},
 {"ID:1165811909", 2299297, 61568},
 {"\[RightArrow] \[ScriptCapitalG]", 2299591, 61575},
 {"ID:306221528", 2299866, 61582},
 {"\[LeftRightArrow] \[ScriptCapitalG]", 2300166, 61589},
 {"ID:113746156", 2300443, 61596},
 {"isReducible \[ScriptCapitalG] multiplier", 2300753, 61603},
 {"ID:1372373678", 2301036, 61610},
 {"isReducible \[ScriptCapitalG]", 2301326, 61617},
 {"ID:12812135", 2301598, 61624},
 {"symred \[ScriptCapitalG]", 2301874, 61631},
 {"ID:168109679", 2302138, 61638},
 {"areCorrelative \[ScriptCapitalG]", 2302432, 61645},
 {"ID:1427340407", 2302707, 61652},
 {"isNTCR \[ScriptCapitalG]", 2302987, 61659},
 {"ID:587977465", 2303255, 61666},
 {"isMNTCR \[ScriptCapitalG]", 2303535, 61673},
 {"ID:950195558", 2303802, 61680},
 {"isNTCR \[ScriptCapitalG] unary", 2304092, 61687},
 {"ID:2144197915", 2304365, 61694},
 {"isMNTCR \[ScriptCapitalG] unary", 2304659, 61701},
 {"ID:1924053921", 2304935, 61708},
 {"red \[ScriptCapitalG] set", 2305217, 61715},
 {"ID:1439528152", 2305487, 61722},
 {"\[Tilde] \[ScriptCapitalG]", 2305771, 61729},
 {"Proof|ID:1148013153", 2306048, 61736},
 {"Proof|ID:1148013153-1", 2306214, 61740},
 {"Proof|ID:1668771272", 2306378, 61744},
 {"Proof|ID:1668771272-1", 2306543, 61748},
 {"Proof|ID:1928426019", 2306706, 61752},
 {"Proof|ID:1928426019-1", 2306870, 61756},
 {"Proof|ID:581321955", 2307031, 61760},
 {"Proof|ID:581321955-1", 2307194, 61764},
 {"Proof|ID:405682622", 2307355, 61768},
 {"Proof|ID:405682622-1", 2307517, 61772},
 {"Proof|ID:1165811909", 2307678, 61776},
 {"Proof|ID:1165811909-1", 2307843, 61780},
 {"Proof|ID:306221528", 2308005, 61784},
 {"Proof|ID:306221528-1", 2308166, 61788},
 {"Proof|ID:113746156", 2308325, 61792},
 {"Proof|ID:113746156-1", 2308488, 61796},
 {"Proof|ID:1372373678", 2308650, 61800},
 {"Proof|ID:1372373678-1", 2308815, 61804},
 {"Proof|ID:12812135", 2308976, 61808},
 {"Proof|ID:12812135-1", 2309136, 61812},
 {"Proof|ID:168109679", 2309295, 61816},
 {"Proof|ID:168109679-1", 2309458, 61820},
 {"Proof|ID:1427340407", 2309620, 61824},
 {"Proof|ID:1427340407-1", 2309786, 61828},
 {"Proof|ID:587977465", 2309949, 61832},
 {"Proof|ID:587977465-1", 2310111, 61836},
 {"Proof|ID:950195558", 2310271, 61840},
 {"Proof|ID:950195558-1", 2310433, 61844},
 {"Proof|ID:2144197915", 2310594, 61848},
 {"Proof|ID:2144197915-1", 2310759, 61852},
 {"Proof|ID:1924053921", 2310922, 61856},
 {"Proof|ID:1924053921-1", 2311086, 61860},
 {"Proof|ID:1439528152", 2311248, 61864},
 {"Proof|ID:1439528152-1", 2311413, 61868},
 {"ID:1674560446", 2311570, 61872},
 {"isPartIrreflOrder \[ScriptCapitalG]", 2311866, 61879},
 {"ID:167203289", 2312139, 61886},
 {"isNoetherian \[ScriptCapitalG]", 2312423, 61893},
 {"ID:569855140", 2312689, 61900},
 {"Multipliers \[ScriptCapitalG]", 2312970, 61907},
 {"ID:180032137", 2313234, 61914},
 {"R0 \[ScriptCapitalG]", 2313497, 61921},
 {"ID:428662253", 2313752, 61928},
 {"R1 \[ScriptCapitalG]", 2314015, 61935},
 {"ID:530735576", 2314270, 61942},
 {"R2 \[ScriptCapitalG]", 2314533, 61949},
 {"ID:1081126646", 2314789, 61956},
 {"R3 \[ScriptCapitalG]", 2315054, 61963},
 {"ID:1591716135", 2315312, 61970},
 {"R4 \[ScriptCapitalG]", 2315577, 61977},
 {"ID:960649057", 2315834, 61984},
 {"R5 \[ScriptCapitalG]", 2316097, 61991},
 {"ID:227638759", 2316352, 61998},
 {"R6 \[ScriptCapitalG]", 2316615, 62005},
 {"ID:2067143528", 2316871, 62012},
 {"R7 \[ScriptCapitalG]", 2317136, 62019},
 {"ID:1699723496", 2317394, 62026},
 {"R8 \[ScriptCapitalG]", 2317659, 62033},
 {"Proof|ID:1674560446", 2317923, 62040},
 {"Proof|ID:1674560446-1", 2318087, 62044},
 {"Proof|ID:167203289", 2318248, 62048},
 {"Proof|ID:167203289-1", 2318408, 62052},
 {"Proof|ID:569855140", 2318566, 62056},
 {"Proof|ID:569855140-1", 2318728, 62060},
 {"Proof|ID:180032137", 2318888, 62064},
 {"Proof|ID:180032137-1", 2319049, 62068},
 {"Proof|ID:428662253", 2319208, 62072},
 {"Proof|ID:428662253-1", 2319370, 62076},
 {"Proof|ID:530735576", 2319530, 62080},
 {"Proof|ID:530735576-1", 2319691, 62084},
 {"Proof|ID:1081126646", 2319851, 62088},
 {"Proof|ID:1081126646-1", 2320015, 62092},
 {"Proof|ID:1591716135", 2320177, 62096},
 {"Proof|ID:1591716135-1", 2320342, 62100},
 {"Proof|ID:960649057", 2320504, 62104},
 {"Proof|ID:960649057-1", 2320666, 62108},
 {"Proof|ID:227638759", 2320826, 62112},
 {"Proof|ID:227638759-1", 2320987, 62116},
 {"Proof|ID:2067143528", 2321147, 62120},
 {"Proof|ID:2067143528-1", 2321311, 62124},
 {"Proof|ID:1699723496", 2321473, 62128},
 {"Proof|ID:1699723496-1", 2321638, 62132},
 {"ID:2054051596", 2321795, 62136},
 {"\[ScriptCapitalG] isReductionRing", 2322086, 62143},
 {"Proof|ID:2054051596", 2322363, 62150},
 {"Proof|ID:2054051596-1", 2322528, 62154},
 {"ID:1031954878", 2322685, 62158},
 {"ID:660860156", 2322913, 62165},
 {"ID:939639226", 2323139, 62172},
 {"R9 \[ScriptCapitalG]", 2323402, 62179},
 {"ID:37694441", 2323656, 62186},
 {"R10 \[ScriptCapitalG]", 2323919, 62193},
 {"ID:576375692", 2324173, 62200},
 {"R11 \[ScriptCapitalG]", 2324438, 62207},
 {"ID:1226510984", 2324695, 62214},
 {"R12 \[ScriptCapitalG]", 2324962, 62221},
 {"ID:1781039403", 2325221, 62228},
 {"R13 \[ScriptCapitalG]", 2325488, 62235},
 {"Proof|ID:939639226", 2325752, 62242},
 {"Proof|ID:939639226-1", 2325914, 62246},
 {"Proof|ID:37694441", 2326073, 62250},
 {"Proof|ID:37694441-1", 2326232, 62254},
 {"Proof|ID:576375692", 2326390, 62258},
 {"Proof|ID:576375692-1", 2326552, 62262},
 {"Proof|ID:1226510984", 2326713, 62266},
 {"Proof|ID:1226510984-1", 2326878, 62270},
 {"Proof|ID:1781039403", 2327041, 62274},
 {"Proof|ID:1781039403-1", 2327205, 62278},
 {"ID:166003593", 2327360, 62282},
 {"\[ScriptCapitalG] isDomain", 2327635, 62289},
 {"ID:1167967104", 2327897, 62296},
 {"\[ScriptCapitalG] isAlgoReductionRing", 2328197, 62303},
 {"Proof|ID:166003593", 2328478, 62310},
 {"Proof|ID:166003593-1", 2328639, 62314},
 {"Proof|ID:1167967104", 2328799, 62318},
 {"Proof|ID:1167967104-1", 2328964, 62322},
 {"ID:2087655986", 2329121, 62326},
 {"ID:955133330", 2329349, 62333},
 {"ID:977980890", 2329575, 62340},
 {"ID:1748570587", 2329802, 62347},
 {"ID:699276041", 2330030, 62354},
 {"membership rd", 2330279, 62361},
 {"ID:867263181", 2330527, 62368},
 {"reducible rd", 2330775, 62375},
 {"ID:1431447619", 2331024, 62382},
 {"irreducible rd", 2331278, 62389},
 {"ID:1490981815", 2331531, 62396},
 {"\[Succeeds] rd", 2331785, 62403},
 {"Proof|ID:699276041", 2332043, 62410},
 {"Proof|ID:699276041-1", 2332205, 62414},
 {"Proof|ID:867263181", 2332365, 62418},
 {"Proof|ID:867263181-1", 2332527, 62422},
 {"Proof|ID:1431447619", 2332688, 62426},
 {"Proof|ID:1431447619-1", 2332854, 62430},
 {"Proof|ID:1490981815", 2333018, 62434},
 {"Proof|ID:1490981815-1", 2333183, 62438},
 {"ID:243233209", 2333339, 62442},
 {"ID:974939989", 2333565, 62449},
 {"rdTBase1", 2333805, 62456},
 {"ID:1014318336", 2334050, 62463},
 {"rdTBase2", 2334292, 62470},
 {"ID:107497168", 2334538, 62477},
 {"rdTRec", 2334774, 62484},
 {"ID:839458854", 2335016, 62491},
 {"ID:1436407557", 2335244, 62498},
 {"4", 2335473, 62505},
 {"ID:2098894842", 2335714, 62512},
 {"rdT induction", 2335966, 62519},
 {"Proof|ID:1436407557", 2336224, 62526},
 {"Proof|ID:1436407557-1", 2336391, 62530},
 {"Proof|ID:2098894842", 2336556, 62534},
 {"Proof|ID:2098894842-1", 2336722, 62538},
 {"ID:160961870", 2336879, 62542},
 {"rdT alternative", 2337139, 62549},
 {"Proof|ID:160961870", 2337402, 62556},
 {"Proof|ID:160961870-1", 2337565, 62560},
 {"ID:168345814", 2337720, 62564},
 {"ID:1634249728", 2337947, 62571},
 {"ID:512270654", 2338175, 62578},
 {"ID:2105002699", 2338402, 62585},
 {"membership rdT", 2338655, 62592},
 {"ID:1505002987", 2338907, 62599},
 {"reducible rdT", 2339159, 62606},
 {"ID:514228174", 2339410, 62613},
 {"irreducible rdT", 2339664, 62620},
 {"ID:208729427", 2339915, 62627},
 {"\[RightArrow]\[RawStar] rdT", 2340192, 62634},
 {"ID:278468994", 2340454, 62641},
 {"\[Succeeds] rdT", 2340708, 62648},
 {"Proof|ID:2105002699", 2340966, 62655},
 {"Proof|ID:2105002699-1", 2341131, 62659},
 {"Proof|ID:1505002987", 2341294, 62663},
 {"Proof|ID:1505002987-1", 2341459, 62667},
 {"Proof|ID:514228174", 2341621, 62671},
 {"Proof|ID:514228174-1", 2341784, 62675},
 {"Proof|ID:208729427", 2341945, 62679},
 {"Proof|ID:208729427-1", 2342107, 62683},
 {"Proof|ID:278468994", 2342267, 62687},
 {"Proof|ID:278468994-1", 2342429, 62691},
 {"ID:96175791", 2342582, 62695},
 {"red rdT", 2342818, 62702},
 {"Proof|ID:96175791", 2343064, 62709},
 {"Proof|ID:96175791-1", 2343222, 62713},
 {"ID:909149002", 2343373, 62717},
 {"trd induction", 2343624, 62724},
 {"Proof|ID:909149002", 2343880, 62731},
 {"Proof|ID:909149002-1", 2344042, 62735},
 {"ID:1097836943", 2344197, 62739},
 {"ID:206560148", 2344426, 62746},
 {"ID:1315173810", 2344653, 62753},
 {"membership trd", 2344906, 62760},
 {"ID:1071214578", 2345158, 62767},
 {"\[RightArrow]\[RawStar] trd", 2345437, 62774},
 {"ID:657218052", 2345701, 62781},
 {"trd \[NotElement] red", 2345966, 62788},
 {"Proof|ID:1315173810", 2346229, 62795},
 {"Proof|ID:1315173810-1", 2346393, 62799},
 {"Proof|ID:1071214578", 2346555, 62803},
 {"Proof|ID:1071214578-1", 2346720, 62807},
 {"Proof|ID:657218052", 2346882, 62811},
 {"Proof|ID:657218052-1", 2347045, 62815},
 {"ID:480907095", 2347200, 62819},
 {"ID:1000930828", 2347427, 62826},
 {"ID:1774407783", 2347657, 62833},
 {"MMPairs \[Element]", 2347919, 62840},
 {"ID:193942992", 2348175, 62847},
 {"ID:44826735", 2348400, 62854},
 {"MMPairs \[Element] ProdDomain", 2348680, 62861},
 {"ID:155756485", 2348943, 62868},
 {"MMPairs \[Element] ProdDomain DomainTuples", 2349251, 62875},
 {"ID:1019394084", 2349530, 62882},
 {"MMPairs \[Element] alternative I", 2349826, 62889},
 {"ID:1458538511", 2350103, 62896},
 {"MMPairs \[Element] alternative II", 2350401, 62903},
 {"ID:81000274", 2350677, 62910},
 {"MMPairs \[Element] alternative III", 2350967, 62917},
 {"Proof|ID:44826735", 2351240, 62924},
 {"Proof|ID:44826735-1", 2351398, 62928},
 {"Proof|ID:155756485", 2351555, 62932},
 {"Proof|ID:155756485-1", 2351717, 62936},
 {"Proof|ID:1019394084", 2351878, 62940},
 {"Proof|ID:1019394084-1", 2352044, 62944},
 {"Proof|ID:1458538511", 2352208, 62948},
 {"Proof|ID:1458538511-1", 2352374, 62952},
 {"Proof|ID:81000274", 2352536, 62956},
 {"Proof|ID:81000274-1", 2352695, 62960},
 {"ID:1140506798", 2352848, 62964},
 {"DomainTuples NPairs shape", 2353124, 62971},
 {"Proof|ID:1140506798", 2353394, 62978},
 {"Proof|ID:1140506798-1", 2353559, 62982},
 {"ID:1527335579", 2353716, 62986},
 {"ID:1794762591", 2353946, 62993},
 {"membership cpmTuple 1", 2354214, 63000},
 {"ID:1790156146", 2354474, 63007},
 {"cpmTuple content 1", 2354736, 63014},
 {"Proof|ID:1794762591", 2354999, 63021},
 {"Proof|ID:1794762591-1", 2355165, 63025},
 {"Proof|ID:1790156146", 2355329, 63029},
 {"Proof|ID:1790156146-1", 2355496, 63033},
 {"ID:884601666", 2355654, 63037},
 {"ID:1188084204", 2355881, 63044},
 {"ID:1708769881", 2356110, 63051},
 {"membership cpmTuple 2", 2356378, 63058},
 {"ID:125591800", 2356637, 63065},
 {"cpmTuple content 2", 2356897, 63072},
 {"Proof|ID:1708769881", 2357158, 63079},
 {"Proof|ID:1708769881-1", 2357324, 63083},
 {"Proof|ID:125591800", 2357487, 63087},
 {"Proof|ID:125591800-1", 2357651, 63091},
 {"ID:1122306642", 2357808, 63095},
 {"ID:885588057", 2358037, 63102},
 {"5", 2358265, 63109},
 {"ID:438489157", 2358504, 63116},
 {"lmTupleAux induction", 2358769, 63123},
 {"Proof|ID:885588057", 2359032, 63130},
 {"Proof|ID:885588057-1", 2359196, 63134},
 {"Proof|ID:438489157", 2359358, 63138},
 {"Proof|ID:438489157-1", 2359522, 63142},
 {"ID:1795822474", 2359679, 63146},
 {"ID:2110280918", 2359909, 63153},
 {"ID:481303170", 2360139, 63160},
 {"membership lmTupleAux", 2360406, 63167},
 {"ID:1389102731", 2360665, 63174},
 {"lmTupleAux content", 2360928, 63181},
 {"Proof|ID:481303170", 2361191, 63188},
 {"Proof|ID:481303170-1", 2361355, 63192},
 {"Proof|ID:1389102731", 2361518, 63196},
 {"Proof|ID:1389102731-1", 2361685, 63200},
 {"ID:1615013989", 2361844, 63204},
 {"ID:1776007069", 2362075, 63211},
 {"ID:1846257369", 2362305, 63218},
 {"membership lmTuple", 2362568, 63225},
 {"ID:851213532", 2362825, 63232},
 {"lmTuple \[NotEqual]", 2363088, 63239},
 {"ID:959030563", 2363344, 63246},
 {"lmTuple =", 2363588, 63253},
 {"Proof|ID:1846257369", 2363842, 63260},
 {"Proof|ID:1846257369-1", 2364008, 63264},
 {"Proof|ID:851213532", 2364171, 63268},
 {"Proof|ID:851213532-1", 2364335, 63272},
 {"Proof|ID:959030563", 2364497, 63276},
 {"Proof|ID:959030563-1", 2364662, 63280},
 {"ID:1440157218", 2364820, 63284},
 {"ID:805845835", 2365049, 63291},
 {"ID:1576526273", 2365277, 63298},
 {"isTODOTuple", 2365526, 63305},
 {"ID:857941005", 2365776, 63312},
 {"isTODOTuple allPairs", 2366041, 63319},
 {"ID:1202189993", 2366299, 63326},
 {"isTODOTuple update", 2366562, 63333},
 {"ID:1698314044", 2366820, 63340},
 {"isTODOTuple rest", 2367079, 63347},
 {"ID:812012342", 2367334, 63354},
 {"isTODOTuple exchange", 2367599, 63361},
 {"Proof|ID:857941005", 2367862, 63368},
 {"Proof|ID:857941005-1", 2368026, 63372},
 {"Proof|ID:1202189993", 2368189, 63376},
 {"Proof|ID:1202189993-1", 2368356, 63380},
 {"Proof|ID:1698314044", 2368521, 63384},
 {"Proof|ID:1698314044-1", 2368688, 63388},
 {"Proof|ID:812012342", 2368852, 63392},
 {"Proof|ID:812012342-1", 2369015, 63396},
 {"ID:1260897708", 2369171, 63400},
 {"chainCrit 1", 2369420, 63407},
 {"Proof|ID:1260897708", 2369677, 63414},
 {"Proof|ID:1260897708-1", 2369843, 63418},
 {"ID:957934274", 2370000, 63422},
 {"ID:1363987639", 2370228, 63429},
 {"ID:184081942", 2370457, 63436},
 {"GBTermOrder", 2370705, 63443},
 {"ID:1341865321", 2370955, 63450},
 {"ID:320402459", 2371184, 63457},
 {"\[ScriptCapitalL] isDomain", 2371460, 63464},
 {"ID:1861353391", 2371723, 63471},
 {"\[ScriptCapitalL] isPartIrreflOrder", 2372020, 63478},
 {"ID:461530885", 2372294, 63485},
 {"\[ScriptCapitalL] isNoetherian", 2372579, 63492},
 {"ID:1658780438", 2372847, 63499},
 {"\[ScriptCapitalL] shape", 2373121, 63506},
 {"Proof|ID:320402459", 2373390, 63513},
 {"Proof|ID:320402459-1", 2373552, 63517},
 {"Proof|ID:1861353391", 2373713, 63521},
 {"Proof|ID:1861353391-1", 2373877, 63525},
 {"Proof|ID:461530885", 2374038, 63529},
 {"Proof|ID:461530885-1", 2374200, 63533},
 {"Proof|ID:1658780438", 2374361, 63537},
 {"Proof|ID:1658780438-1", 2374527, 63541},
 {"ID:1606683850", 2374685, 63545},
 {"ID:973931584", 2374914, 63552},
 {"\[ScriptCapitalL] \[Succeeds] 1", 2375201, 63559},
 {"ID:1350364326", 2375470, 63566},
 {"\[ScriptCapitalL] \[Succeeds] 2", 2375759, 63573},
 {"ID:529848255", 2376029, 63580},
 {"\[ScriptCapitalL] \[Succeeds] 3", 2376316, 63587},
 {"ID:592062547", 2376584, 63594},
 {"\[SquareSubset] append trd", 2376861, 63601},
 {"Proof|ID:973931584", 2377130, 63608},
 {"Proof|ID:973931584-1", 2377293, 63612},
 {"Proof|ID:1350364326", 2377455, 63616},
 {"Proof|ID:1350364326-1", 2377620, 63620},
 {"Proof|ID:529848255", 2377782, 63624},
 {"Proof|ID:529848255-1", 2377945, 63628},
 {"Proof|ID:592062547", 2378106, 63632},
 {"Proof|ID:592062547-1", 2378271, 63636},
 {"ID:1848949834", 2378429, 63640},
 {"ID:1362258751", 2378659, 63647},
 {"GBAuxTest", 2378905, 63654},
 {"ID:1347349813", 2379155, 63661},
 {"GBAuxBase", 2379400, 63668},
 {"ID:270708621", 2379648, 63675},
 {"GBAuxRec1", 2379893, 63682},
 {"ID:1542247232", 2380142, 63689},
 {"GBAuxRec2", 2380388, 63696},
 {"ID:580452282", 2380637, 63703},
 {"GBAuxRec3", 2380882, 63710},
 {"ID:598818422", 2381130, 63717},
 {"GBAuxRec4", 2381375, 63724},
 {"ID:1392375551", 2381624, 63731},
 {"ID:994411308", 2381853, 63738},
 {"1", 2382082, 63745},
 {"ID:1185976259", 2382323, 63752},
 {"2", 2382555, 63759},
 {"ID:1125858125", 2382799, 63766},
 {"3", 2383030, 63773},
 {"ID:1780855977", 2383273, 63780},
 {"GBAux induction", 2383532, 63787},
 {"ID:380025998", 2383788, 63794},
 {"GBAux induction II", 2384051, 63801},
 {"Proof|ID:994411308", 2384314, 63808},
 {"Proof|ID:994411308-1", 2384477, 63812},
 {"Proof|ID:1185976259", 2384639, 63816},
 {"Proof|ID:1185976259-1", 2384806, 63820},
 {"Proof|ID:1125858125", 2384971, 63824},
 {"Proof|ID:1125858125-1", 2385137, 63828},
 {"Proof|ID:1780855977", 2385301, 63832},
 {"Proof|ID:1780855977-1", 2385466, 63836},
 {"Proof|ID:380025998", 2385628, 63840},
 {"Proof|ID:380025998-1", 2385793, 63844},
 {"ID:1379909256", 2385951, 63848},
 {"GBAux 3 alternative", 2386223, 63855},
 {"Proof|ID:1379909256", 2386495, 63862},
 {"Proof|ID:1379909256-1", 2386663, 63866},
 {"ID:884344895", 2386822, 63870},
 {"ID:2012261150", 2387050, 63877},
 {"ID:1467104417", 2387280, 63884},
 {"shape GBAuxBase", 2387537, 63891},
 {"ID:1354352171", 2387792, 63898},
 {"shape GBAuxRec1", 2388049, 63905},
 {"ID:1322915202", 2388304, 63912},
 {"shape GBAuxRec2", 2388561, 63919},
 {"ID:620036821", 2388815, 63926},
 {"shape GBAuxRec3", 2389070, 63933},
 {"ID:1111467670", 2389323, 63940},
 {"shape GBAuxRec4", 2389580, 63947},
 {"Proof|ID:1467104417", 2389841, 63954},
 {"Proof|ID:1467104417-1", 2390008, 63958},
 {"Proof|ID:1354352171", 2390173, 63962},
 {"Proof|ID:1354352171-1", 2390339, 63966},
 {"Proof|ID:1322915202", 2390503, 63970},
 {"Proof|ID:1322915202-1", 2390670, 63974},
 {"Proof|ID:620036821", 2390834, 63978},
 {"Proof|ID:620036821-1", 2390998, 63982},
 {"Proof|ID:1111467670", 2391161, 63986},
 {"Proof|ID:1111467670-1", 2391327, 63990},
 {"ID:1314333166", 2391485, 63994},
 {"shape GBAux", 2391734, 64001},
 {"Proof|ID:1314333166", 2391991, 64008},
 {"Proof|ID:1314333166-1", 2392157, 64012},
 {"ID:1167572947", 2392315, 64016},
 {"membership GBAux", 2392574, 64023},
 {"Proof|ID:1167572947", 2392836, 64030},
 {"Proof|ID:1167572947-1", 2393002, 64034},
 {"ID:208228172", 2393159, 64038},
 {"ID:781335069", 2393386, 64045},
 {"MMPairs \[RightArrow]\[RawStar]", 2393673, 64052},
 {"Proof|ID:781335069", 2393947, 64059},
 {"Proof|ID:781335069-1", 2394111, 64063},
 {"ID:1452629404", 2394268, 64067},
 {"ID:1597922168", 2394498, 64074},
 {"ideal GBAuxBase", 2394755, 64081},
 {"ID:348489962", 2395009, 64088},
 {"ideal GBAuxRec1", 2395264, 64095},
 {"ID:1787031010", 2395517, 64102},
 {"ideal GBAuxRec2", 2395774, 64109},
 {"ID:963974147", 2396028, 64116},
 {"ideal GBAuxRec3", 2396283, 64123},
 {"ID:473957484", 2396535, 64130},
 {"ideal GBAuxRec4", 2396790, 64137},
 {"Proof|ID:1597922168", 2397049, 64144},
 {"Proof|ID:1597922168-1", 2397216, 64148},
 {"Proof|ID:348489962", 2397380, 64152},
 {"Proof|ID:348489962-1", 2397543, 64156},
 {"Proof|ID:1787031010", 2397705, 64160},
 {"Proof|ID:1787031010-1", 2397872, 64164},
 {"Proof|ID:963974147", 2398036, 64168},
 {"Proof|ID:963974147-1", 2398200, 64172},
 {"Proof|ID:473957484", 2398362, 64176},
 {"Proof|ID:473957484-1", 2398526, 64180},
 {"ID:1826585682", 2398683, 64184},
 {"ideal GBAux", 2398932, 64191},
 {"Proof|ID:1826585682", 2399189, 64198},
 {"Proof|ID:1826585682-1", 2399354, 64202},
 {"ID:416121228", 2399510, 64206},
 {"ID:2050404037", 2399738, 64213},
 {"ID:1837575531", 2399968, 64220},
 {"isCPTODOTuple", 2400222, 64227},
 {"ID:346323949", 2400475, 64234},
 {"GBAuxCPConnectible1", 2400738, 64241},
 {"ID:1643593823", 2400995, 64248},
 {"GBAuxCPConnectible2", 2401260, 64255},
 {"ID:1213243088", 2401519, 64262},
 {"GBAuxCPConnectible3", 2401784, 64269},
 {"ID:949923506", 2402042, 64276},
 {"h=0", 2402275, 64283},
 {"Proof|ID:949923506", 2402523, 64290},
 {"Proof|ID:949923506-1", 2402687, 64294},
 {"ID:1322820809", 2402844, 64298},
 {"ID:1772563804", 2403074, 64305},
 {"ID:667808408", 2403303, 64312},
 {"ID:1601898370", 2403531, 64319},
 {"h\[NotEqual]0", 2403784, 64326},
 {"ID:622683853", 2404036, 64333},
 {"cpConnectible append h", 2404305, 64340},
 {"Proof|ID:1601898370", 2404571, 64347},
 {"Proof|ID:1601898370-1", 2404737, 64351},
 {"Proof|ID:622683853", 2404900, 64355},
 {"Proof|ID:622683853-1", 2405064, 64359},
 {"ID:1834517245", 2405221, 64363},
 {"GBAuxTest alternative", 2405490, 64370},
 {"ID:1923946854", 2405751, 64377},
 {"GBAuxTest 1", 2406002, 64384},
 {"Proof|ID:1834517245", 2406261, 64391},
 {"Proof|ID:1834517245-1", 2406429, 64395},
 {"Proof|ID:1923946854", 2406595, 64399},
 {"Proof|ID:1923946854-1", 2406763, 64403},
 {"ID:258768016", 2406922, 64407},
 {"GBAuxCPConnectible1 empty", 2407197, 64414},
 {"ID:1002747542", 2407460, 64421},
 {"ID:673303920", 2407689, 64428},
 {"GBAuxCPConnectible1 GBAuxRec3", 2407971, 64435},
 {"ID:1502806330", 2408237, 64442},
 {"GBAuxCPConnectible1 GBAuxRec4", 2408521, 64449},
 {"Proof|ID:258768016", 2408794, 64456},
 {"Proof|ID:258768016-1", 2408956, 64460},
 {"Proof|ID:673303920", 2409116, 64464},
 {"Proof|ID:673303920-1", 2409280, 64468},
 {"Proof|ID:1502806330", 2409443, 64472},
 {"Proof|ID:1502806330-1", 2409610, 64476},
 {"ID:1133484500", 2409769, 64480},
 {"ID:291397089", 2409999, 64487},
 {"ID:1494094918", 2410227, 64494},
 {"GBAuxCPConnectible1 GBAux", 2410504, 64501},
 {"ID:402861438", 2410768, 64508},
 {"cpConnectible GBAux lmTuple", 2411047, 64515},
 {"Proof|ID:1494094918", 2411318, 64522},
 {"Proof|ID:1494094918-1", 2411485, 64526},
 {"Proof|ID:402861438", 2411649, 64530},
 {"Proof|ID:402861438-1", 2411813, 64534},
 {"ID:1173305577", 2411970, 64538},
 {"ID:1545923689", 2412200, 64545},
 {"ID:1684239249", 2412430, 64552},
 {"isCPTODOTuple empty", 2412695, 64559},
 {"ID:547535692", 2412953, 64566},
 {"isCPTODOTuple prepend", 2413220, 64573},
 {"ID:812188873", 2413478, 64580},
 {"isCPTODOTuple isTODOTuple", 2413753, 64587},
 {"ID:1105161129", 2414016, 64594},
 {"isCPTODOTuple lmTuple", 2414285, 64601},
 {"ID:981490890", 2414545, 64608},
 {"isCPTODOTuple 0", 2414800, 64615},
 {"Proof|ID:1684239249", 2415059, 64622},
 {"Proof|ID:1684239249-1", 2415224, 64626},
 {"Proof|ID:547535692", 2415386, 64630},
 {"Proof|ID:547535692-1", 2415550, 64634},
 {"Proof|ID:812188873", 2415712, 64638},
 {"Proof|ID:812188873-1", 2415876, 64642},
 {"Proof|ID:1105161129", 2416039, 64646},
 {"Proof|ID:1105161129-1", 2416206, 64650},
 {"Proof|ID:981490890", 2416370, 64654},
 {"Proof|ID:981490890-1", 2416533, 64658},
 {"ID:681053692", 2416688, 64662},
 {"ID:995853659", 2416916, 64669},
 {"ID:218025475", 2417143, 64676},
 {"ID:2052980984", 2417371, 64683},
 {"ID:925810378", 2417601, 64690},
 {"ID:423392620", 2417829, 64697},
 {"isCPTODOTuple 1", 2418084, 64704},
 {"ID:1533257011", 2418337, 64711},
 {"isCPTODOTuple 2", 2418594, 64718},
 {"Proof|ID:423392620", 2418854, 64725},
 {"Proof|ID:423392620-1", 2419018, 64729},
 {"Proof|ID:1533257011", 2419181, 64733},
 {"Proof|ID:1533257011-1", 2419348, 64737},
 {"ID:1154842295", 2419507, 64741},
 {"ID:1427865659", 2419737, 64748},
 {"GBAuxCPConnectible2 GBAuxBase", 2420021, 64755},
 {"ID:802696830", 2420288, 64762},
 {"GBAuxCPConnectible2 GBAuxRec1", 2420570, 64769},
 {"ID:668126868", 2420835, 64776},
 {"GBAuxCPConnectible2 GBAuxRec2", 2421117, 64783},
 {"ID:1848124281", 2421383, 64790},
 {"GBAuxCPConnectible2 GBAuxRec3", 2421667, 64797},
 {"ID:6143564", 2421932, 64804},
 {"GBAuxCPConnectible2 GBAuxRec4", 2422210, 64811},
 {"Proof|ID:1427865659", 2422478, 64818},
 {"Proof|ID:1427865659-1", 2422644, 64822},
 {"Proof|ID:802696830", 2422807, 64826},
 {"Proof|ID:802696830-1", 2422970, 64830},
 {"Proof|ID:668126868", 2423131, 64834},
 {"Proof|ID:668126868-1", 2423294, 64838},
 {"Proof|ID:1848124281", 2423456, 64842},
 {"Proof|ID:1848124281-1", 2423622, 64846},
 {"Proof|ID:6143564", 2423783, 64850},
 {"Proof|ID:6143564-1", 2423941, 64854},
 {"ID:1620493315", 2424094, 64858},
 {"ID:1672348970", 2424325, 64865},
 {"ID:1826259623", 2424555, 64872},
 {"ID:1814594394", 2424785, 64879},
 {"GBAuxCPConnectible2 GBAux", 2425061, 64886},
 {"ID:1024326782", 2425325, 64893},
 {"ID:585977074", 2425554, 64900},
 {"ID:1959330209", 2425783, 64907},
 {"GBAuxCPConnectible2 update", 2426062, 64914},
 {"Proof|ID:1814594394", 2426334, 64921},
 {"Proof|ID:1814594394-1", 2426500, 64925},
 {"Proof|ID:1959330209", 2426664, 64929},
 {"Proof|ID:1959330209-1", 2426832, 64933},
 {"ID:33155789", 2426990, 64937},
 {"ID:1691022439", 2427216, 64944},
 {"GBAuxCPConnectible3 GBAuxBase", 2427500, 64951},
 {"ID:544907973", 2427767, 64958},
 {"GBAuxCPConnectible3 GBAuxRec1", 2428049, 64965},
 {"ID:1039424559", 2428315, 64972},
 {"GBAuxCPConnectible3 GBAuxRec2", 2428599, 64979},
 {"ID:1704484269", 2428867, 64986},
 {"GBAuxCPConnectible3 GBAuxRec3", 2429151, 64993},
 {"ID:2114764931", 2429419, 65000},
 {"GBAuxCPConnectible3 GBAuxRec4", 2429703, 65007},
 {"Proof|ID:1691022439", 2429977, 65014},
 {"Proof|ID:1691022439-1", 2430144, 65018},
 {"Proof|ID:544907973", 2430308, 65022},
 {"Proof|ID:544907973-1", 2430471, 65026},
 {"Proof|ID:1039424559", 2430633, 65030},
 {"Proof|ID:1039424559-1", 2430799, 65034},
 {"Proof|ID:1704484269", 2430963, 65038},
 {"Proof|ID:1704484269-1", 2431129, 65042},
 {"Proof|ID:2114764931", 2431293, 65046},
 {"Proof|ID:2114764931-1", 2431460, 65050},
 {"ID:1693763802", 2431619, 65054},
 {"GBAuxCPConnectible3 GBAux", 2431896, 65061},
 {"Proof|ID:1693763802", 2432167, 65068},
 {"Proof|ID:1693763802-1", 2432333, 65072},
 {"ID:439420249", 2432490, 65076},
 {"GB empty", 2432736, 65083},
 {"Proof|ID:439420249", 2432992, 65090},
 {"Proof|ID:439420249-1", 2433155, 65094},
 {"ID:298471155", 2433310, 65098},
 {"ID:1615450915", 2433538, 65105},
 {"ID:209199628", 2433767, 65112},
 {"shape GB", 2434008, 65119},
 {"ID:448006251", 2434253, 65126},
 {"membership GB", 2434509, 65133},
 {"ID:739795809", 2434764, 65140},
 {"ideal GB", 2435010, 65147},
 {"ID:1550516051", 2435261, 65154},
 {"isGroebnerBasis GB", 2435523, 65161},
 {"Proof|ID:209199628", 2435785, 65168},
 {"Proof|ID:209199628-1", 2435949, 65172},
 {"Proof|ID:448006251", 2436111, 65176},
 {"Proof|ID:448006251-1", 2436275, 65180},
 {"Proof|ID:739795809", 2436437, 65184},
 {"Proof|ID:739795809-1", 2436600, 65188},
 {"Proof|ID:1550516051", 2436762, 65192},
 {"Proof|ID:1550516051-1", 2436927, 65196}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[579, 22, 179, 4, 84, "Title",
 CellID->1451885783],
Cell[CellGroupData[{
Cell[783, 30, 216, 4, 73, "Section",
 CellID->1605568989],
Cell[CellGroupData[{
Cell[1024, 38, 121, 2, 38, "Subsection",
 CellID->499915303],
Cell[1148, 42, 956, 21, 64, "Text",
 CellID->29975522],
Cell[2107, 65, 461, 10, 43, "Text",
 CellID->57224452],
Cell[CellGroupData[{
Cell[2593, 79, 159, 4, 24, "Item",
 CellID->45190291],
Cell[2755, 85, 225, 7, 25, "Item",
 CellID->90776693],
Cell[2983, 94, 245, 5, 24, "Item",
 CellID->8961680]
}, Open  ]],
Cell[3243, 102, 649, 11, 43, "Text",
 CellID->71065762],
Cell[3895, 115, 1227, 19, 119, "Text",
 CellID->10851318]
}, Closed]],
Cell[CellGroupData[{
Cell[5159, 139, 121, 2, 30, "Subsection",
 CellID->16638123],
Cell[5283, 143, 393, 10, 32, "Input",
 CellID->498486718],
Cell[5679, 155, 463, 10, 32, "Input",
 CellID->119228788],
Cell[6145, 167, 408, 9, 32, "Input",
 CellID->297313380],
Cell[6556, 178, 405, 9, 32, "Input",
 CellID->86850003],
Cell[6964, 189, 355, 8, 32, "Input",
 CellID->131034841],
Cell[7322, 199, 367, 8, 32, "Input",
 CellID->477360655],
Cell[7692, 209, 337, 8, 32, "Input",
 CellID->123628739],
Cell[8032, 219, 390, 9, 32, "Input",
 CellID->115117422],
Cell[CellGroupData[{
Cell[8447, 232, 1652, 35, 55, "Input",
 CellID->20666029],
Cell[10102, 269, 405, 7, 32, "Output",
 CellID->468046677]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[10556, 282, 124, 2, 30, "Subsection",
 CellID->1386333768],
Cell[CellGroupData[{
Cell[10705, 288, 336, 8, 70, "Input",
 CellID->1039312110],
Cell[CellGroupData[{
Cell[11066, 300, 951, 20, 70, "Print",
 CellID->880067753],
Cell[12020, 322, 969, 20, 70, "Print",
 CellID->604617084],
Cell[12992, 344, 755, 16, 70, "Print",
 CellID->964310657],
Cell[13750, 362, 758, 16, 70, "Print",
 CellID->1623802069],
Cell[14511, 380, 760, 16, 70, "Print",
 CellID->1392221509],
Cell[15274, 398, 760, 16, 70, "Print",
 CellID->1474649765],
Cell[16037, 416, 984, 20, 70, "Print",
 CellID->1623284855],
Cell[17024, 438, 956, 20, 70, "Print",
 CellID->2059021584],
Cell[17983, 460, 965, 20, 70, "Print",
 CellID->468682694],
Cell[18951, 482, 759, 16, 70, "Print",
 CellID->768345119],
Cell[19713, 500, 757, 16, 70, "Print",
 CellID->203140207],
Cell[20473, 518, 759, 16, 70, "Print",
 CellID->839315771],
Cell[21235, 536, 760, 16, 70, "Print",
 CellID->1909090729],
Cell[21998, 554, 759, 16, 70, "Print",
 CellID->781859586],
Cell[22760, 572, 759, 16, 70, "Print",
 CellID->792515171],
Cell[23522, 590, 760, 16, 70, "Print",
 CellID->2077551358],
Cell[24285, 608, 959, 20, 70, "Print",
 CellID->385817707],
Cell[25247, 630, 968, 20, 70, "Print",
 CellID->1914681352],
Cell[26218, 652, 755, 16, 70, "Print",
 CellID->309156947],
Cell[26976, 670, 761, 16, 70, "Print",
 CellID->538743880],
Cell[27740, 688, 764, 16, 70, "Print",
 CellID->1509225573],
Cell[28507, 706, 759, 16, 70, "Print",
 CellID->599128823],
Cell[29269, 724, 763, 16, 70, "Print",
 CellID->448464536],
Cell[30035, 742, 931, 20, 70, "Print",
 CellID->726826737],
Cell[30969, 764, 755, 16, 70, "Print",
 CellID->519974804],
Cell[31727, 782, 758, 16, 70, "Print",
 CellID->1538620868],
Cell[32488, 800, 760, 16, 70, "Print",
 CellID->2116298062],
Cell[33251, 818, 764, 16, 70, "Print",
 CellID->1975937998],
Cell[34018, 836, 760, 16, 70, "Print",
 CellID->1696687010],
Cell[34781, 854, 760, 16, 70, "Print",
 CellID->1940201374],
Cell[35544, 872, 764, 16, 70, "Print",
 CellID->1823397800],
Cell[36311, 890, 759, 16, 70, "Print",
 CellID->446834775],
Cell[37073, 908, 557, 11, 70, "Print",
 CellID->1578517663],
Cell[37633, 921, 549, 11, 70, "Print",
 CellID->31752406],
Cell[38185, 934, 748, 16, 70, "Print",
 CellID->1289398078],
Cell[38936, 952, 750, 16, 70, "Print",
 CellID->1707398801],
Cell[39689, 970, 751, 16, 70, "Print",
 CellID->480147046],
Cell[40443, 988, 756, 16, 70, "Print",
 CellID->1571230922],
Cell[41202, 1006, 752, 16, 70, "Print",
 CellID->1160857634],
Cell[41957, 1024, 756, 16, 70, "Print",
 CellID->1266278869],
Cell[42716, 1042, 752, 16, 70, "Print",
 CellID->1832033513],
Cell[43471, 1060, 756, 16, 70, "Print",
 CellID->1459714591],
Cell[44230, 1078, 751, 16, 70, "Print",
 CellID->372738200],
Cell[44984, 1096, 752, 16, 70, "Print",
 CellID->1119015511],
Cell[45739, 1114, 755, 16, 70, "Print",
 CellID->266406868],
Cell[46497, 1132, 752, 16, 70, "Print",
 CellID->1662811202],
Cell[47252, 1150, 751, 16, 70, "Print",
 CellID->905324562],
Cell[48006, 1168, 751, 16, 70, "Print",
 CellID->648137395],
Cell[48760, 1186, 755, 16, 70, "Print",
 CellID->285997184],
Cell[49518, 1204, 751, 16, 70, "Print",
 CellID->959440624],
Cell[50272, 1222, 752, 16, 70, "Print",
 CellID->1729084317],
Cell[51027, 1240, 755, 16, 70, "Print",
 CellID->318527011],
Cell[51785, 1258, 756, 16, 70, "Print",
 CellID->1995813270],
Cell[52544, 1276, 751, 16, 70, "Print",
 CellID->683642649],
Cell[53298, 1294, 754, 16, 70, "Print",
 CellID->2093499806],
Cell[54055, 1312, 758, 16, 70, "Print",
 CellID->1424876236],
Cell[54816, 1330, 757, 16, 70, "Print",
 CellID->187493898],
Cell[55576, 1348, 753, 16, 70, "Print",
 CellID->138253922],
Cell[56332, 1366, 754, 16, 70, "Print",
 CellID->2019201381],
Cell[57089, 1384, 757, 16, 70, "Print",
 CellID->170458589],
Cell[57849, 1402, 753, 16, 70, "Print",
 CellID->346169200],
Cell[58605, 1420, 753, 16, 70, "Print",
 CellID->859857300],
Cell[59361, 1438, 753, 16, 70, "Print",
 CellID->101944774],
Cell[60117, 1456, 757, 16, 70, "Print",
 CellID->313520785],
Cell[60877, 1474, 542, 11, 70, "Print",
 CellID->1157430104]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[61468, 1491, 338, 7, 70, "Input",
 CellID->892463300],
Cell[CellGroupData[{
Cell[61831, 1502, 620, 12, 70, "Print",
 CellID->564482905],
Cell[62454, 1516, 344, 7, 70, "Print",
 CellID->306351101],
Cell[62801, 1525, 344, 7, 70, "Print",
 CellID->726477269],
Cell[63148, 1534, 343, 7, 70, "Print",
 CellID->4348995],
Cell[63494, 1543, 344, 7, 70, "Print",
 CellID->41231562],
Cell[63841, 1552, 143, 2, 70, "Print",
 CellID->360733829],
Cell[63987, 1556, 135, 2, 70, "Print",
 CellID->816780095],
Cell[64125, 1560, 334, 7, 70, "Print",
 CellID->352919787],
Cell[64462, 1569, 334, 7, 70, "Print",
 CellID->523041642],
Cell[64799, 1578, 336, 7, 70, "Print",
 CellID->379549855],
Cell[65138, 1587, 337, 7, 70, "Print",
 CellID->68832474],
Cell[65478, 1596, 338, 7, 70, "Print",
 CellID->173122215],
Cell[65819, 1605, 336, 7, 70, "Print",
 CellID->530215840],
Cell[66158, 1614, 338, 7, 70, "Print",
 CellID->298553875],
Cell[66499, 1623, 336, 7, 70, "Print",
 CellID->499074680],
Cell[66838, 1632, 337, 7, 70, "Print",
 CellID->23924730],
Cell[67178, 1641, 147, 2, 70, "Print",
 CellID->429839575],
Cell[67328, 1645, 144, 2, 70, "Print",
 CellID->115076950],
Cell[67475, 1649, 146, 2, 70, "Print",
 CellID->11557326],
Cell[67624, 1653, 148, 2, 70, "Print",
 CellID->386496720],
Cell[67775, 1657, 134, 2, 70, "Print",
 CellID->20784557],
Cell[67912, 1661, 329, 7, 70, "Print",
 CellID->67136478],
Cell[68244, 1670, 125, 2, 70, "Print",
 CellID->27448579]
}, Open  ]],
Cell[68384, 1675, 175, 3, 70, "Output",
 CellID->368709728]
}, Closed]],
Cell[68574, 1681, 325, 6, 70, "Input",
 CellID->19428626],
Cell[CellGroupData[{
Cell[68924, 1691, 541, 12, 70, "Input",
 CellID->1909803754],
Cell[69468, 1705, 779, 12, 70, "Output",
 CellID->797779344]
}, Open  ]],
Cell[CellGroupData[{
Cell[70284, 1722, 406, 12, 70, "Input",
 CellID->18261744],
Cell[70693, 1736, 94, 2, 70, "Output",
 CellID->772937485]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[70836, 1744, 128, 2, 30, "Subsection",
 CellID->1454182661],
Cell[70967, 1748, 984, 25, 99, "Input",
 CellID->2074904605],
Cell[71954, 1775, 416, 9, 55, "Input",
 CellID->1980565193],
Cell[72373, 1786, 169, 3, 32, "Input",
 CellID->444074165],
Cell[72545, 1791, 257, 5, 32, "Input",
 CellID->798575911]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[72851, 1802, 234, 4, 43, "Section",
 CellID->1452921687],
Cell[CellGroupData[{
Cell[73110, 1810, 229, 7, 38, "Subsection",
 CellID->287260853],
Cell[73342, 1819, 47, 1, 12, "OpenEnvironment",
 CellID->207462652],
Cell[CellGroupData[{
Cell[73414, 1824, 460, 14, 43, "EnvironmentHeader",
 CellID->608117277],
Cell[73877, 1840, 1134, 34, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:1079943823", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "print"},
 CellID->1079943823],
Cell[75014, 1876, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1723749034]
}, Open  ]],
Cell[75095, 1880, 48, 1, 16, "CloseEnvironment",
 CellID->712958051]
}, Closed]],
Cell[CellGroupData[{
Cell[75180, 1886, 340, 8, 30, "Subsection",
 CellID->334464829],
Cell[CellGroupData[{
Cell[75545, 1898, 122, 2, 28, "Subsubsection",
 CellID->657395164],
Cell[75670, 1902, 47, 1, 70, "OpenEnvironment",
 CellID->704307368],
Cell[CellGroupData[{
Cell[75742, 1907, 375, 9, 70, "EnvironmentHeader",
 CellID->2046755789],
Cell[76120, 1918, 1968, 60, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1033951332", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed"},
 CellID->1033951332],
Cell[78091, 1980, 65, 1, 70, "EndEnvironmentMarker",
 CellID->547704634]
}, Open  ]],
Cell[78171, 1984, 47, 1, 70, "CloseEnvironment",
 CellID->95587551],
Cell[78221, 1987, 47, 1, 70, "OpenEnvironment",
 CellID->546908141],
Cell[CellGroupData[{
Cell[78293, 1992, 495, 11, 70, "EnvironmentHeader",
 CellID->227588461],
Cell[78791, 2005, 846, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1967331668", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1967331668],
Cell[79640, 2027, 1127, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1698292943", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs 1"},
 CellID->1698292943],
Cell[80770, 2060, 1720, 48, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1961920283", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs 2"},
 CellID->1961920283],
Cell[82493, 2110, 1561, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1516706954", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "allPairs"},
 CellID->1516706954],
Cell[84057, 2152, 1235, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:329670188", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "update"},
 CellID->329670188],
Cell[85295, 2193, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1008147046]
}, Open  ]],
Cell[85376, 2197, 48, 1, 70, "CloseEnvironment",
 CellID->118409702],
Cell[85427, 2200, 428, 9, 70, "Text",
 CellID->214949191],
Cell[85858, 2211, 47, 1, 70, "OpenEnvironment",
 CellID->657408265],
Cell[CellGroupData[{
Cell[85930, 2216, 399, 9, 70, "EnvironmentHeader",
 CellID->1299046839],
Cell[86332, 2227, 748, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1283917386", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1283917386],
Cell[87083, 2247, 810, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:14652249", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->14652249],
Cell[87896, 2269, 2437, 75, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1012271681", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element]"},
 CellID->1012271681],
Cell[90336, 2346, 65, 1, 70, "EndEnvironmentMarker",
 CellID->750257285]
}, Open  ]],
Cell[90416, 2350, 48, 1, 70, "CloseEnvironment",
 CellID->672258777]
}, Closed]],
Cell[CellGroupData[{
Cell[90501, 2356, 116, 2, 22, "Subsubsection",
 CellID->200708094],
Cell[CellGroupData[{
Cell[90642, 2362, 514, 11, 31, "Computation",
 CellID->1171053509],
Cell[91159, 2375, 732, 12, 24, "ComputationResult",
 CellID->2030739290]
}, Open  ]],
Cell[CellGroupData[{
Cell[91928, 2392, 377, 10, 28, "ComputationInfo",
 CellID->238027729],
Cell[92308, 2404, 650, 17, 71, "ComputationInfoBody",
 CellID->572441673]
}, Closed]],
Cell[CellGroupData[{
Cell[92995, 2426, 367, 9, 28, "Computation",
 CellID->1147098130],
Cell[93365, 2437, 398, 7, 24, "ComputationResult",
 CellID->1543197942]
}, Open  ]],
Cell[CellGroupData[{
Cell[93800, 2449, 378, 10, 28, "ComputationInfo",
 CellID->1024977621],
Cell[94181, 2461, 653, 17, 71, "ComputationInfoBody",
 CellID->1352090126]
}, Closed]],
Cell[CellGroupData[{
Cell[94871, 2483, 561, 13, 28, "Computation",
 CellID->1987206771],
Cell[95435, 2498, 250, 5, 24, "ComputationResult",
 CellID->1879346961]
}, Open  ]],
Cell[CellGroupData[{
Cell[95722, 2508, 377, 10, 28, "ComputationInfo",
 CellID->578228279],
Cell[96102, 2520, 650, 17, 71, "ComputationInfoBody",
 CellID->441030501]
}, Closed]],
Cell[CellGroupData[{
Cell[96789, 2542, 155, 3, 28, "Computation",
 CellID->618819632],
Cell[96947, 2547, 336, 8, 24, "ComputationResult",
 CellID->1355411954]
}, Open  ]],
Cell[CellGroupData[{
Cell[97320, 2560, 377, 10, 28, "ComputationInfo",
 CellID->1346320737],
Cell[97700, 2572, 650, 17, 71, "ComputationInfoBody",
 CellID->1995525275]
}, Closed]],
Cell[CellGroupData[{
Cell[98387, 2594, 130, 3, 28, "Computation",
 CellID->1130747252],
Cell[98520, 2599, 399, 9, 24, "ComputationResult",
 CellID->1139002741]
}, Open  ]],
Cell[CellGroupData[{
Cell[98956, 2613, 378, 10, 28, "ComputationInfo",
 CellID->1111910764],
Cell[99337, 2625, 650, 17, 71, "ComputationInfoBody",
 CellID->1023788914]
}, Closed]],
Cell[CellGroupData[{
Cell[100024, 2647, 130, 3, 28, "Computation",
 CellID->1098463539],
Cell[100157, 2652, 610, 14, 24, "ComputationResult",
 CellID->539950968]
}, Open  ]],
Cell[CellGroupData[{
Cell[100804, 2671, 378, 10, 28, "ComputationInfo",
 CellID->1782642311],
Cell[101185, 2683, 652, 17, 71, "ComputationInfoBody",
 CellID->219375137]
}, Closed]],
Cell[CellGroupData[{
Cell[101874, 2705, 226, 4, 28, "Computation",
 CellID->1917426498],
Cell[102103, 2711, 1013, 21, 24, "ComputationResult",
 CellID->58665367]
}, Open  ]],
Cell[CellGroupData[{
Cell[103153, 2737, 378, 10, 28, "ComputationInfo",
 CellID->1124327149],
Cell[103534, 2749, 651, 17, 71, "ComputationInfoBody",
 CellID->1789828913]
}, Closed]],
Cell[CellGroupData[{
Cell[104222, 2771, 152, 3, 28, "Computation",
 CellID->191290873],
Cell[104377, 2776, 1249, 27, 24, "ComputationResult",
 CellID->935380317]
}, Open  ]],
Cell[CellGroupData[{
Cell[105663, 2808, 377, 10, 28, "ComputationInfo",
 CellID->1954139160],
Cell[106043, 2820, 650, 17, 71, "ComputationInfoBody",
 CellID->1250936321]
}, Closed]],
Cell[CellGroupData[{
Cell[106730, 2842, 151, 3, 28, "Computation",
 CellID->125498398],
Cell[106884, 2847, 1158, 25, 24, "ComputationResult",
 CellID->412770255]
}, Open  ]],
Cell[CellGroupData[{
Cell[108079, 2877, 379, 10, 28, "ComputationInfo",
 CellID->1021847407],
Cell[108461, 2889, 648, 17, 71, "ComputationInfoBody",
 CellID->34256619]
}, Closed]],
Cell[CellGroupData[{
Cell[109146, 2911, 380, 9, 28, "Computation",
 CellID->1562111062],
Cell[109529, 2922, 908, 20, 24, "ComputationResult",
 CellID->350255224]
}, Open  ]],
Cell[CellGroupData[{
Cell[110474, 2947, 377, 10, 28, "ComputationInfo",
 CellID->965480556],
Cell[110854, 2959, 650, 17, 71, "ComputationInfoBody",
 CellID->230386765]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[111553, 2982, 182, 3, 22, "Subsubsection",
 CellID->2027758943],
Cell[111738, 2987, 48, 1, 70, "OpenEnvironment",
 CellID->1347355957],
Cell[CellGroupData[{
Cell[111811, 2992, 468, 10, 70, "EnvironmentHeader",
 CellID->1285769589],
Cell[112282, 3004, 742, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:988862829", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->988862829],
Cell[113027, 3024, 1339, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:16377100", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element] ProdDomain"},
 CellID->16377100],
Cell[114369, 3068, 1472, 46, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:136514135", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element] ProdDomain DomainTuples"},
 CellID->136514135],
Cell[115844, 3116, 1986, 56, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1872634730", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[Element] alternative", "IR"},
 CellID->1872634730],
Cell[117833, 3174, 1959, 60, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1931599387", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs< \[Element] alternative", "IR"},
 CellID->1931599387],
Cell[119795, 3236, 2024, 62, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2124565744", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs\[LessEqual] \[Element] alternative", "IR"},
 CellID->2124565744],
Cell[121822, 3300, 1538, 47, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1917398594", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs< \[Element] alternative II"},
 CellID->1917398594],
Cell[123363, 3349, 1649, 50, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:272474466", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs\[LessEqual] \[Element] alternative II"},
 CellID->272474466],
Cell[125015, 3401, 1398, 45, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2005363738", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs< \[LessEqual]"},
 CellID->2005363738],
Cell[126416, 3448, 1524, 47, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1661715038", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NatPairs< \[LessEqual]"},
 CellID->1661715038],
Cell[127943, 3497, 726, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1020080047", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1020080047],
Cell[128672, 3518, 1316, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1327808237", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs +1"},
 CellID->1327808237],
Cell[129991, 3562, 1426, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:798140928", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NatPairs +1"},
 CellID->798140928],
Cell[131420, 3608, 791, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:2111464127", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->2111464127],
Cell[132214, 3630, 1264, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1834895057", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "NatPairs \[GreaterEqual]"},
 CellID->1834895057],
Cell[133481, 3672, 1388, 43, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1703465728", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NatPairs \[GreaterEqual]"},
 CellID->1703465728],
Cell[134872, 3717, 66, 1, 70, "EndEnvironmentMarker",
 CellID->2009530504]
}, Open  ]],
Cell[134953, 3721, 48, 1, 70, "CloseEnvironment",
 CellID->954284704],
Cell[CellGroupData[{
Cell[135026, 3726, 3491, 97, 70, "ProofInfo",
 CellTags->{"Proof|ID:16377100", "Proof|ID:16377100-1"},
 CellID->22673357],
Cell[138520, 3825, 479, 12, 70, "ProofInfoBody",
 CellID->196582360]
}, Closed]],
Cell[CellGroupData[{
Cell[139036, 3842, 3771, 102, 70, "ProofInfo",
 CellTags->{"Proof|ID:136514135", "Proof|ID:136514135-1"},
 CellID->585962876],
Cell[142810, 3946, 479, 12, 70, "ProofInfoBody",
 CellID->33667364]
}, Closed]],
Cell[CellGroupData[{
Cell[143326, 3963, 5199, 130, 70, "ProofInfo",
 CellTags->{"Proof|ID:1872634730", "Proof|ID:1872634730-1"},
 CellID->1738547973],
Cell[148528, 4095, 482, 12, 70, "ProofInfoBody",
 CellID->1666042888]
}, Closed]],
Cell[CellGroupData[{
Cell[149047, 4112, 4821, 124, 70, "ProofInfo",
 CellTags->{"Proof|ID:1931599387", "Proof|ID:1931599387-1"},
 CellID->60767287],
Cell[153871, 4238, 482, 12, 70, "ProofInfoBody",
 CellID->2095809430]
}, Closed]],
Cell[CellGroupData[{
Cell[154390, 4255, 4913, 125, 70, "ProofInfo",
 CellTags->{"Proof|ID:2124565744", "Proof|ID:2124565744-1"},
 CellID->2051885642],
Cell[159306, 4382, 482, 12, 70, "ProofInfoBody",
 CellID->1918391475]
}, Closed]],
Cell[CellGroupData[{
Cell[159825, 4399, 4582, 125, 70, "ProofInfo",
 CellTags->{"Proof|ID:1917398594", "Proof|ID:1917398594-1"},
 CellID->1425710701],
Cell[164410, 4526, 480, 12, 70, "ProofInfoBody",
 CellID->98700243]
}, Closed]],
Cell[CellGroupData[{
Cell[164927, 4543, 4654, 124, 70, "ProofInfo",
 CellTags->{"Proof|ID:272474466", "Proof|ID:272474466-1"},
 CellID->136931469],
Cell[169584, 4669, 481, 12, 70, "ProofInfoBody",
 CellID->1186718445]
}, Closed]],
Cell[CellGroupData[{
Cell[170102, 4686, 3074, 89, 70, "ProofInfo",
 CellTags->{"Proof|ID:2005363738", "Proof|ID:2005363738-1"},
 CellID->911690168],
Cell[173179, 4777, 481, 12, 70, "ProofInfoBody",
 CellID->109415471]
}, Closed]],
Cell[CellGroupData[{
Cell[173697, 4794, 3351, 95, 70, "ProofInfo",
 CellTags->{"Proof|ID:1661715038", "Proof|ID:1661715038-1"},
 CellID->1036844371],
Cell[177051, 4891, 481, 12, 70, "ProofInfoBody",
 CellID->123023466]
}, Closed]],
Cell[CellGroupData[{
Cell[177569, 4908, 3961, 105, 70, "ProofInfo",
 CellTags->{"Proof|ID:1327808237", "Proof|ID:1327808237-1"},
 CellID->2053606910],
Cell[181533, 5015, 481, 12, 70, "ProofInfoBody",
 CellID->184747131]
}, Closed]],
Cell[CellGroupData[{
Cell[182051, 5032, 4205, 108, 70, "ProofInfo",
 CellTags->{"Proof|ID:798140928", "Proof|ID:798140928-1"},
 CellID->974929944],
Cell[186259, 5142, 480, 12, 70, "ProofInfoBody",
 CellID->963692269]
}, Closed]],
Cell[CellGroupData[{
Cell[186776, 5159, 4843, 123, 70, "ProofInfo",
 CellTags->{"Proof|ID:1834895057", "Proof|ID:1834895057-1"},
 CellID->1231124039],
Cell[191622, 5284, 482, 12, 70, "ProofInfoBody",
 CellID->2145280537]
}, Closed]],
Cell[CellGroupData[{
Cell[192141, 5301, 5113, 129, 70, "ProofInfo",
 CellTags->{"Proof|ID:1703465728", "Proof|ID:1703465728-1"},
 CellID->1999261425],
Cell[197257, 5432, 482, 12, 70, "ProofInfoBody",
 CellID->1185705224]
}, Closed]],
Cell[197754, 5447, 48, 1, 70, "OpenEnvironment",
 CellID->2014150930],
Cell[CellGroupData[{
Cell[197827, 5452, 443, 10, 70, "EnvironmentHeader",
 CellID->643050674],
Cell[198273, 5464, 954, 22, 70, "GlobalDeclaration",
 CellTags->{
  "ID:971135058", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->971135058],
Cell[199230, 5488, 1116, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1202991311", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership pairs unary", "IR"},
 CellID->1202991311],
Cell[200349, 5523, 1522, 43, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1932739315", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs unary content'"},
 CellID->1932739315],
Cell[201874, 5568, 1264, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1628109432", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs unary content"},
 CellID->1628109432],
Cell[203141, 5608, 1189, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:540671405", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership pairs binary", "IR"},
 CellID->540671405],
Cell[204333, 5645, 1465, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:881225230", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs binary content'"},
 CellID->881225230],
Cell[205801, 5689, 1247, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:717573619", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "pairs binary content"},
 CellID->717573619],
Cell[207051, 5729, 859, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:794667329", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "allPairs 0", "IR"},
 CellID->794667329],
Cell[207913, 5755, 1126, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1385740688", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership allPairs", "IR"},
 CellID->1385740688],
Cell[209042, 5790, 1232, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1430860208", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "allPairs content"},
 CellID->1430860208],
Cell[210277, 5829, 65, 1, 70, "EndEnvironmentMarker",
 CellID->968800477]
}, Open  ]],
Cell[210357, 5833, 47, 1, 70, "CloseEnvironment",
 CellID->51649926],
Cell[CellGroupData[{
Cell[210429, 5838, 2859, 83, 70, "ProofInfo",
 CellTags->{"Proof|ID:1202991311", "Proof|ID:1202991311-1"},
 CellID->1926557973],
Cell[213291, 5923, 482, 12, 70, "ProofInfoBody",
 CellID->1260065281]
}, Closed]],
Cell[CellGroupData[{
Cell[213810, 5940, 4716, 125, 70, "ProofInfo",
 CellTags->{"Proof|ID:1932739315", "Proof|ID:1932739315-1"},
 CellID->1207788650],
Cell[218529, 6067, 481, 12, 70, "ProofInfoBody",
 CellID->758612481]
}, Closed]],
Cell[CellGroupData[{
Cell[219047, 6084, 3584, 99, 70, "ProofInfo",
 CellTags->{"Proof|ID:1628109432", "Proof|ID:1628109432-1"},
 CellID->135860045],
Cell[222634, 6185, 482, 12, 70, "ProofInfoBody",
 CellID->1770197684]
}, Closed]],
Cell[CellGroupData[{
Cell[223153, 6202, 3330, 92, 70, "ProofInfo",
 CellTags->{"Proof|ID:540671405", "Proof|ID:540671405-1"},
 CellID->1030769530],
Cell[226486, 6296, 523, 13, 70, "ProofInfoBody",
 CellID->1057299996]
}, Closed]],
Cell[CellGroupData[{
Cell[227046, 6314, 4859, 126, 70, "ProofInfo",
 CellTags->{"Proof|ID:881225230", "Proof|ID:881225230-1"},
 CellID->960050214],
Cell[231908, 6442, 480, 12, 70, "ProofInfoBody",
 CellID->347641866]
}, Closed]],
Cell[CellGroupData[{
Cell[232425, 6459, 4076, 110, 70, "ProofInfo",
 CellTags->{"Proof|ID:717573619", "Proof|ID:717573619-1"},
 CellID->1036460169],
Cell[236504, 6571, 481, 12, 70, "ProofInfoBody",
 CellID->1623918929]
}, Closed]],
Cell[CellGroupData[{
Cell[237022, 6588, 1945, 62, 70, "ProofInfo",
 CellTags->{"Proof|ID:794667329", "Proof|ID:794667329-1"},
 CellID->1738668792],
Cell[238970, 6652, 481, 12, 70, "ProofInfoBody",
 CellID->2133578009]
}, Closed]],
Cell[CellGroupData[{
Cell[239488, 6669, 2885, 83, 70, "ProofInfo",
 CellTags->{"Proof|ID:1385740688", "Proof|ID:1385740688-1"},
 CellID->1882316559],
Cell[242376, 6754, 481, 12, 70, "ProofInfoBody",
 CellID->782413460]
}, Closed]],
Cell[CellGroupData[{
Cell[242894, 6771, 4035, 109, 70, "ProofInfo",
 CellTags->{"Proof|ID:1430860208", "Proof|ID:1430860208-1"},
 CellID->1876195014],
Cell[246932, 6882, 482, 12, 70, "ProofInfoBody",
 CellID->1624086186]
}, Closed]],
Cell[247429, 6897, 48, 1, 70, "OpenEnvironment",
 CellID->2123044522],
Cell[CellGroupData[{
Cell[247502, 6902, 382, 9, 70, "EnvironmentHeader",
 CellID->774910772],
Cell[247887, 6913, 823, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:963656485", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->963656485],
Cell[248713, 6935, 1072, 30, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1750952092", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1750952092],
Cell[249788, 6967, 805, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1500640019", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1500640019],
Cell[250596, 6990, 1160, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:723173326", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership update", "IR"},
 CellID->723173326],
Cell[251759, 7026, 1301, 45, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:392184336", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "update content I"},
 CellID->392184336],
Cell[253063, 7073, 1174, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:341058317", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "update content II"},
 CellID->341058317],
Cell[254240, 7111, 65, 1, 70, "EndEnvironmentMarker",
 CellID->522299237]
}, Open  ]],
Cell[254320, 7115, 49, 1, 70, "CloseEnvironment",
 CellID->1129995189],
Cell[CellGroupData[{
Cell[254394, 7120, 3938, 106, 70, "ProofInfo",
 CellTags->{"Proof|ID:723173326", "Proof|ID:723173326-1"},
 CellID->1574692164],
Cell[258335, 7228, 480, 12, 70, "ProofInfoBody",
 CellID->310043985]
}, Closed]],
Cell[CellGroupData[{
Cell[258852, 7245, 3907, 104, 70, "ProofInfo",
 CellTags->{"Proof|ID:392184336", "Proof|ID:392184336-1"},
 CellID->393000099],
Cell[262762, 7351, 481, 12, 70, "ProofInfoBody",
 CellID->1623641860]
}, Closed]],
Cell[CellGroupData[{
Cell[263280, 7368, 4348, 116, 70, "ProofInfo",
 CellTags->{"Proof|ID:341058317", "Proof|ID:341058317-1"},
 CellID->1522358468],
Cell[267631, 7486, 480, 12, 70, "ProofInfoBody",
 CellID->975211483]
}, Closed]],
Cell[268126, 7501, 48, 1, 70, "OpenEnvironment",
 CellID->1143846827],
Cell[CellGroupData[{
Cell[268199, 7506, 581, 12, 70, "EnvironmentHeader",
 CellID->1184365251],
Cell[268783, 7520, 1965, 63, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1597870023", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed alternative", "IR"},
 CellID->1597870023],
Cell[270751, 7585, 1099, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:40246499", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed symmetric"},
 CellID->40246499],
Cell[271853, 7619, 728, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1131819573", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1131819573],
Cell[272584, 7639, 1106, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1083804603", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed allPairs"},
 CellID->1083804603],
Cell[273693, 7673, 2454, 73, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:316507172", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed update"},
 CellID->316507172],
Cell[276150, 7748, 2158, 62, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1179241697", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "processed prepend"},
 CellID->1179241697],
Cell[278311, 7812, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1183261273]
}, Open  ]],
Cell[278392, 7816, 48, 1, 70, "CloseEnvironment",
 CellID->186310157],
Cell[CellGroupData[{
Cell[278465, 7821, 5020, 129, 70, "ProofInfo",
 CellTags->{"Proof|ID:1597870023", "Proof|ID:1597870023-1"},
 CellID->932327505],
Cell[283488, 7952, 481, 12, 70, "ProofInfoBody",
 CellID->559044054]
}, Closed]],
Cell[CellGroupData[{
Cell[284006, 7969, 3135, 86, 70, "ProofInfo",
 CellTags->{"Proof|ID:40246499", "Proof|ID:40246499-1"},
 CellID->1343087848],
Cell[287144, 8057, 480, 12, 70, "ProofInfoBody",
 CellID->1360525692]
}, Closed]],
Cell[CellGroupData[{
Cell[287661, 8074, 3849, 106, 70, "ProofInfo",
 CellTags->{"Proof|ID:1083804603", "Proof|ID:1083804603-1"},
 CellID->465417913],
Cell[291513, 8182, 480, 12, 70, "ProofInfoBody",
 CellID->20057220]
}, Closed]],
Cell[CellGroupData[{
Cell[292030, 8199, 6411, 158, 70, "ProofInfo",
 CellTags->{"Proof|ID:316507172", "Proof|ID:316507172-1"},
 CellID->660680371],
Cell[298444, 8359, 481, 12, 70, "ProofInfoBody",
 CellID->1967876509]
}, Closed]],
Cell[CellGroupData[{
Cell[298962, 8376, 6163, 149, 70, "ProofInfo",
 CellTags->{"Proof|ID:1179241697", "Proof|ID:1179241697-1"},
 CellID->2041135352],
Cell[305128, 8527, 482, 12, 70, "ProofInfoBody",
 CellID->1629853247]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[305683, 8547, 309, 5, 43, "Section",
 CellID->509234429],
Cell[CellGroupData[{
Cell[306017, 8556, 129, 2, 70, "Subsection",
 CellID->1820866479],
Cell[306149, 8560, 675, 15, 70, "Text",
 CellID->1002761981],
Cell[CellGroupData[{
Cell[306849, 8579, 219, 5, 70, "Item",
 CellID->1613933805],
Cell[307071, 8586, 910, 31, 70, "Item",
 CellID->307340730],
Cell[307984, 8619, 959, 29, 70, "Item",
 CellID->1910860741],
Cell[308946, 8650, 1233, 32, 70, "Item",
 CellID->1871827641]
}, Open  ]],
Cell[310194, 8685, 472, 10, 70, "Text",
 CellID->701994113],
Cell[CellGroupData[{
Cell[310691, 8699, 323, 10, 70, "Item",
 CellID->687289781],
Cell[311017, 8711, 329, 10, 70, "Item",
 CellID->1002810321],
Cell[311349, 8723, 366, 11, 70, "Item",
 CellID->101699731],
Cell[311718, 8736, 896, 20, 70, "Item",
 CellID->1471005405]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[312663, 8762, 184, 3, 70, "Subsection",
 CellID->1076888715],
Cell[312850, 8767, 47, 1, 70, "OpenEnvironment",
 CellID->508559908],
Cell[CellGroupData[{
Cell[312922, 8772, 590, 12, 70, "EnvironmentHeader",
 CellID->388885171],
Cell[313515, 8786, 651, 16, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1573502139", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1573502139],
Cell[314169, 8804, 789, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:377704681", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->377704681],
Cell[314961, 8825, 1996, 37, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1468120413", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1468120413],
Cell[316960, 8864, 1325, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1774031579", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing \[Element]"},
 CellID->1774031579],
Cell[318288, 8902, 855, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:182210603", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing 0"},
 CellID->182210603],
Cell[319146, 8927, 855, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:814933967", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing 1"},
 CellID->814933967],
Cell[320004, 8952, 906, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1684568525", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing +"},
 CellID->1684568525],
Cell[320913, 8979, 902, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:584522550", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing -"},
 CellID->584522550],
Cell[321818, 9006, 908, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1458964616", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing - unary"},
 CellID->1458964616],
Cell[322729, 9033, 906, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1496488880", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing \[RawStar]"},
 CellID->1496488880],
Cell[323638, 9060, 1432, 43, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:342394137", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing \[Succeeds]"},
 CellID->342394137],
Cell[325073, 9105, 887, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1566608311", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing multN"},
 CellID->1566608311],
Cell[325963, 9132, 1098, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:628841792", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing mult 1"},
 CellID->628841792],
Cell[327064, 9166, 1120, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:378340035", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing mult 2"},
 CellID->378340035],
Cell[328187, 9200, 1065, 29, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:822076309", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lcrd"},
 CellID->822076309],
Cell[329255, 9231, 969, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:849156542", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing rdm"},
 CellID->849156542],
Cell[330227, 9260, 890, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:809363439", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lcrd unary"},
 CellID->809363439],
Cell[331120, 9286, 987, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1671864047", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing cpm"},
 CellID->1671864047],
Cell[332110, 9315, 401, 15, 70, "EnvironmentText",
 CellID->1362566519],
Cell[332514, 9332, 1070, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:558323540", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing rd"},
 CellID->558323540],
Cell[333587, 9365, 539, 18, 70, "EnvironmentText",
 CellID->1551627756],
Cell[334129, 9385, 1675, 47, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1403530194", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing rdT"},
 CellID->1403530194],
Cell[335807, 9434, 643, 20, 70, "EnvironmentText",
 CellID->2034063019],
Cell[336453, 9456, 1732, 48, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:689058884", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing trd"},
 CellID->689058884],
Cell[338188, 9506, 1674, 49, 70, "EnvironmentText",
 CellID->666716832],
Cell[339865, 9557, 2982, 65, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1528624471", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lmTuple"},
 CellID->1528624471],
Cell[342850, 9624, 986, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1192872489", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lmTupleAux 1"},
 CellID->1192872489],
Cell[343839, 9653, 1942, 48, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1479034574", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing lmTupleAux 2"},
 CellID->1479034574],
Cell[345784, 9703, 1848, 62, 70, "EnvironmentText",
 CellID->1669742674],
Cell[347635, 9767, 1809, 51, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1529751135", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing cpmTuple 1"},
 CellID->1529751135],
Cell[349447, 9820, 1602, 54, 70, "EnvironmentText",
 CellID->203149419],
Cell[351052, 9876, 1705, 46, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:551185449", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing cpmTuple 2"},
 CellID->551185449],
Cell[352760, 9924, 440, 12, 70, "EnvironmentText",
 CellID->1547004803],
Cell[353203, 9938, 1246, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1533920470", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GB"},
 CellID->1533920470],
Cell[354452, 9972, 1112, 29, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1976234692", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GBAux 1"},
 CellID->1976234692],
Cell[355567, 10003, 2792, 73, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:144716839", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GBAux 2"},
 CellID->144716839],
Cell[358362, 10078, 4571, 115, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:189044945", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing GBAux 3"},
 CellID->189044945],
Cell[362936, 10195, 1121, 30, 70, "EnvironmentText",
 CellID->468041515],
Cell[364060, 10227, 3441, 90, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1766960499", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing chainCrit"},
 CellID->1766960499],
Cell[367504, 10319, 1687, 47, 70, "EnvironmentText",
 CellID->458982664],
Cell[369194, 10368, 1132, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1777604038", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing ard"},
 CellID->1777604038],
Cell[370329, 10400, 1161, 29, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:620529741", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing ardAux 1"},
 CellID->620529741],
Cell[371493, 10431, 2316, 71, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:174904624", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GroebnerRing ardAux 2"},
 CellID->174904624],
Cell[373812, 10504, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1222673264]
}, Open  ]],
Cell[373893, 10508, 49, 1, 70, "CloseEnvironment",
 CellID->1724139512]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[373991, 10515, 112, 2, 43, "Section",
 CellID->963895538],
Cell[374106, 10519, 410, 9, 32, "Input",
 CellID->1816127517],
Cell[374519, 10530, 410, 9, 32, "Input",
 CellID->1873029144],
Cell[374932, 10541, 424, 9, 32, "Input",
 CellID->1110067096],
Cell[375359, 10552, 412, 9, 32, "Input",
 CellID->1197775081],
Cell[375774, 10563, 292, 8, 32, "Input",
 CellID->851552697],
Cell[CellGroupData[{
Cell[376091, 10575, 178, 3, 32, "Input",
 CellID->430172879],
Cell[376272, 10580, 601, 12, 23, "Print",
 CellID->1649399640]
}, Open  ]],
Cell[CellGroupData[{
Cell[376910, 10597, 180, 3, 38, "Subsection",
 CellID->553055740],
Cell[377093, 10602, 47, 1, 12, "OpenEnvironment",
 CellID->572285523],
Cell[CellGroupData[{
Cell[377165, 10607, 514, 13, 43, "EnvironmentHeader",
 CellID->2012283505],
Cell[377682, 10622, 1036, 25, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:336511876", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalQ]"},
 CellID->336511876],
Cell[378721, 10649, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1595859076]
}, Open  ]],
Cell[378802, 10653, 48, 1, 16, "CloseEnvironment",
 CellID->914612762],
Cell[CellGroupData[{
Cell[378875, 10658, 610, 12, 55, "Computation",
 CellID->460713032],
Cell[379488, 10672, 605, 12, 41, "ComputationResult",
 CellID->1219447368]
}, Open  ]],
Cell[CellGroupData[{
Cell[380130, 10689, 377, 10, 28, "ComputationInfo",
 CellID->1022441481],
Cell[380510, 10701, 652, 17, 235, "ComputationInfoBody",
 CellID->1436640939]
}, Closed]],
Cell[CellGroupData[{
Cell[381199, 10723, 436, 9, 52, "Computation",
 CellID->1058478970],
Cell[381638, 10734, 521, 10, 24, "ComputationResult",
 CellID->1917317166]
}, Open  ]],
Cell[CellGroupData[{
Cell[382196, 10749, 377, 10, 28, "ComputationInfo",
 CellID->2127071249],
Cell[382576, 10761, 655, 17, 235, "ComputationInfoBody",
 CellID->1965686465]
}, Closed]],
Cell[CellGroupData[{
Cell[383268, 10783, 441, 9, 53, "Computation",
 CellID->1725275638],
Cell[383712, 10794, 565, 14, 42, "ComputationResult",
 CellID->1471124563]
}, Open  ]],
Cell[CellGroupData[{
Cell[384314, 10813, 378, 10, 28, "ComputationInfo",
 CellID->1462783354],
Cell[384695, 10825, 651, 17, 235, "ComputationInfoBody",
 CellID->31606615]
}, Closed]],
Cell[CellGroupData[{
Cell[385383, 10847, 353, 8, 53, "Computation",
 CellID->681863368],
Cell[385739, 10857, 623, 15, 42, "ComputationResult",
 CellID->2046403607]
}, Open  ]],
Cell[CellGroupData[{
Cell[386399, 10877, 377, 10, 28, "ComputationInfo",
 CellID->1100457786],
Cell[386779, 10889, 652, 17, 235, "ComputationInfoBody",
 CellID->1053359543]
}, Closed]],
Cell[CellGroupData[{
Cell[387468, 10911, 418, 11, 52, "Computation",
 CellID->380754896],
Cell[387889, 10924, 555, 13, 42, "ComputationResult",
 CellID->695736833]
}, Open  ]],
Cell[CellGroupData[{
Cell[388481, 10942, 377, 10, 28, "ComputationInfo",
 CellID->1874687287],
Cell[388861, 10954, 653, 17, 235, "ComputationInfoBody",
 CellID->623992335]
}, Closed]],
Cell[CellGroupData[{
Cell[389551, 10976, 421, 11, 52, "Computation",
 CellID->1319269262],
Cell[389975, 10989, 351, 8, 42, "ComputationResult",
 CellID->815793440]
}, Open  ]],
Cell[CellGroupData[{
Cell[390363, 11002, 379, 10, 28, "ComputationInfo",
 CellID->537945533],
Cell[390745, 11014, 653, 17, 235, "ComputationInfoBody",
 CellID->1252991721]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[391447, 11037, 132, 2, 30, "Subsection",
 CellID->1783762617],
Cell[391582, 11041, 47, 1, 12, "OpenEnvironment",
 CellID->413357741],
Cell[CellGroupData[{
Cell[391654, 11046, 575, 16, 43, "EnvironmentHeader",
 CellID->2097093576],
Cell[392232, 11064, 999, 25, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:193404746", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]"},
 CellID->193404746],
Cell[393234, 11091, 65, 1, 26, "EndEnvironmentMarker",
 CellID->953586710]
}, Open  ]],
Cell[393314, 11095, 49, 1, 16, "CloseEnvironment",
 CellID->1137065047],
Cell[CellGroupData[{
Cell[393388, 11100, 305, 7, 44, "Computation",
 CellID->1290937647],
Cell[393696, 11109, 641, 15, 24, "ComputationResult",
 CellID->1065167918]
}, Open  ]],
Cell[CellGroupData[{
Cell[394374, 11129, 379, 10, 28, "ComputationInfo",
 CellID->745975735],
Cell[394756, 11141, 653, 17, 235, "ComputationInfoBody",
 CellID->1868486125]
}, Closed]],
Cell[CellGroupData[{
Cell[395446, 11163, 256, 6, 41, "Computation",
 CellID->34460033],
Cell[395705, 11171, 462, 11, 24, "ComputationResult",
 CellID->2084155435]
}, Open  ]],
Cell[CellGroupData[{
Cell[396204, 11187, 375, 10, 28, "ComputationInfo",
 CellID->813692243],
Cell[396582, 11199, 653, 17, 235, "ComputationInfoBody",
 CellID->2017922153]
}, Closed]],
Cell[CellGroupData[{
Cell[397272, 11221, 300, 7, 41, "Computation",
 CellID->210527065],
Cell[397575, 11230, 485, 12, 24, "ComputationResult",
 CellID->233209813]
}, Open  ]],
Cell[CellGroupData[{
Cell[398097, 11247, 377, 10, 28, "ComputationInfo",
 CellID->2009197914],
Cell[398477, 11259, 652, 17, 235, "ComputationInfoBody",
 CellID->1673135784]
}, Closed]],
Cell[CellGroupData[{
Cell[399166, 11281, 492, 10, 39, "Computation",
 CellID->2004775239],
Cell[399661, 11293, 491, 12, 24, "ComputationResult",
 CellID->2097566239]
}, Open  ]],
Cell[CellGroupData[{
Cell[400189, 11310, 380, 10, 28, "ComputationInfo",
 CellID->1339960540],
Cell[400572, 11322, 652, 17, 235, "ComputationInfoBody",
 CellID->136544494]
}, Closed]],
Cell[CellGroupData[{
Cell[401261, 11344, 413, 10, 39, "Computation",
 CellID->314892606],
Cell[401677, 11356, 331, 7, 24, "ComputationResult",
 CellID->1102276527]
}, Open  ]],
Cell[CellGroupData[{
Cell[402045, 11368, 376, 10, 28, "ComputationInfo",
 CellID->825370302],
Cell[402424, 11380, 654, 17, 235, "ComputationInfoBody",
 CellID->1666078827]
}, Closed]],
Cell[CellGroupData[{
Cell[403115, 11402, 311, 7, 39, "Computation",
 CellID->1654602455],
Cell[403429, 11411, 374, 9, 24, "ComputationResult",
 CellID->1491910187]
}, Open  ]],
Cell[CellGroupData[{
Cell[403840, 11425, 378, 10, 28, "ComputationInfo",
 CellID->1152043447],
Cell[404221, 11437, 678, 17, 133, "ComputationInfoBody",
 CellID->643984759]
}, Closed]],
Cell[CellGroupData[{
Cell[404936, 11459, 523, 11, 39, "Computation",
 CellID->123561440],
Cell[405462, 11472, 333, 7, 24, "ComputationResult",
 CellID->355642102]
}, Open  ]],
Cell[CellGroupData[{
Cell[405832, 11484, 377, 10, 28, "ComputationInfo",
 CellID->1426675444],
Cell[406212, 11496, 653, 17, 235, "ComputationInfoBody",
 CellID->657498851]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[406914, 11519, 211, 5, 32, "Subsection",
 CellID->1940243944],
Cell[407128, 11526, 47, 1, 12, "OpenEnvironment",
 CellID->921750930],
Cell[CellGroupData[{
Cell[407200, 11531, 628, 19, 43, "EnvironmentHeader",
 CellID->369221017],
Cell[407831, 11552, 968, 25, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1821700886", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]24"},
 CellID->1821700886],
Cell[408802, 11579, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1831110154]
}, Open  ]],
Cell[408883, 11583, 49, 1, 16, "CloseEnvironment",
 CellID->1312555615],
Cell[CellGroupData[{
Cell[408957, 11588, 290, 7, 42, "Computation",
 CellID->210353575],
Cell[409250, 11597, 374, 8, 24, "ComputationResult",
 CellID->1123616661]
}, Open  ]],
Cell[CellGroupData[{
Cell[409661, 11610, 377, 10, 28, "ComputationInfo",
 CellID->1279883406],
Cell[410041, 11622, 653, 17, 250, "ComputationInfoBody",
 CellID->320835099]
}, Closed]],
Cell[CellGroupData[{
Cell[410731, 11644, 342, 7, 39, "Computation",
 CellID->2041351713],
Cell[411076, 11653, 452, 9, 24, "ComputationResult",
 CellID->1960744693]
}, Open  ]],
Cell[CellGroupData[{
Cell[411565, 11667, 377, 10, 28, "ComputationInfo",
 CellID->955142166],
Cell[411945, 11679, 653, 17, 232, "ComputationInfoBody",
 CellID->1202136422]
}, Closed]],
Cell[CellGroupData[{
Cell[412635, 11701, 389, 8, 39, "Computation",
 CellID->1718024098],
Cell[413027, 11711, 450, 9, 24, "ComputationResult",
 CellID->1268604381]
}, Open  ]],
Cell[CellGroupData[{
Cell[413514, 11725, 377, 10, 28, "ComputationInfo",
 CellID->147506085],
Cell[413894, 11737, 652, 17, 250, "ComputationInfoBody",
 CellID->795415407]
}, Closed]],
Cell[CellGroupData[{
Cell[414583, 11759, 330, 8, 39, "Computation",
 CellID->1660233602],
Cell[414916, 11769, 238, 6, 24, "ComputationResult",
 CellID->1236673481]
}, Open  ]],
Cell[CellGroupData[{
Cell[415191, 11780, 380, 10, 28, "ComputationInfo",
 CellID->1531881759],
Cell[415574, 11792, 655, 17, 250, "ComputationInfoBody",
 CellID->1891696089]
}, Closed]],
Cell[416244, 11812, 252, 6, 24, "Text",
 CellID->162878423],
Cell[CellGroupData[{
Cell[416521, 11822, 315, 7, 42, "Computation",
 CellID->561089538],
Cell[416839, 11831, 386, 8, 24, "ComputationResult",
 CellID->1487225701]
}, Open  ]],
Cell[CellGroupData[{
Cell[417262, 11844, 379, 10, 28, "ComputationInfo",
 CellID->1246547808],
Cell[417644, 11856, 651, 17, 250, "ComputationInfoBody",
 CellID->553142869]
}, Closed]],
Cell[CellGroupData[{
Cell[418332, 11878, 331, 8, 39, "Computation",
 CellID->372196881],
Cell[418666, 11888, 241, 6, 24, "ComputationResult",
 CellID->1782329282]
}, Open  ]],
Cell[CellGroupData[{
Cell[418944, 11899, 379, 10, 28, "ComputationInfo",
 CellID->1465396252],
Cell[419326, 11911, 652, 17, 250, "ComputationInfoBody",
 CellID->1202561127]
}, Closed]],
Cell[CellGroupData[{
Cell[420015, 11933, 364, 7, 39, "Computation",
 CellID->1596303038],
Cell[420382, 11942, 486, 10, 24, "ComputationResult",
 CellID->96031009]
}, Open  ]],
Cell[CellGroupData[{
Cell[420905, 11957, 378, 10, 28, "ComputationInfo",
 CellID->1550093037],
Cell[421286, 11969, 655, 17, 250, "ComputationInfoBody",
 CellID->1545764092]
}, Closed]],
Cell[CellGroupData[{
Cell[421978, 11991, 333, 8, 39, "Computation",
 CellID->364287853],
Cell[422314, 12001, 236, 6, 24, "ComputationResult",
 CellID->289900299]
}, Open  ]],
Cell[CellGroupData[{
Cell[422587, 12012, 377, 10, 28, "ComputationInfo",
 CellID->1727929368],
Cell[422967, 12024, 654, 17, 250, "ComputationInfoBody",
 CellID->1122644056]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[423670, 12047, 297, 7, 30, "Subsection",
 CellID->438160971],
Cell[423970, 12056, 48, 1, 12, "OpenEnvironment",
 CellID->1899836489],
Cell[CellGroupData[{
Cell[424043, 12061, 609, 16, 43, "EnvironmentHeader",
 CellID->418711622],
Cell[424655, 12079, 1211, 28, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1158015694", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalQ]xy"},
 CellID->1158015694],
Cell[425869, 12109, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1667815660]
}, Open  ]],
Cell[425950, 12113, 49, 1, 16, "CloseEnvironment",
 CellID->1792593212],
Cell[CellGroupData[{
Cell[426024, 12118, 615, 20, 57, "Computation",
 CellID->749980045],
Cell[426642, 12140, 1387, 37, 42, "ComputationResult",
 CellID->683271910]
}, Open  ]],
Cell[CellGroupData[{
Cell[428066, 12182, 378, 10, 28, "ComputationInfo",
 CellID->588655274],
Cell[428447, 12194, 654, 17, 430, "ComputationInfoBody",
 CellID->1614854082]
}, Closed]],
Cell[CellGroupData[{
Cell[429138, 12216, 217, 5, 39, "Computation",
 CellID->444957564],
Cell[429358, 12223, 205, 5, 24, "ComputationResult",
 CellID->1389037508]
}, Open  ]],
Cell[CellGroupData[{
Cell[429600, 12233, 376, 10, 28, "ComputationInfo",
 CellID->101779177],
Cell[429979, 12245, 651, 17, 430, "ComputationInfoBody",
 CellID->882684762]
}, Closed]],
Cell[CellGroupData[{
Cell[430667, 12267, 783, 26, 54, "Computation",
 CellID->84261983],
Cell[431453, 12295, 1412, 36, 42, "ComputationResult",
 CellID->1405364557]
}, Open  ]],
Cell[CellGroupData[{
Cell[432902, 12336, 376, 10, 28, "ComputationInfo",
 CellID->1568740038],
Cell[433281, 12348, 650, 17, 430, "ComputationInfoBody",
 CellID->630430316]
}, Closed]],
Cell[CellGroupData[{
Cell[433968, 12370, 394, 14, 42, "Computation",
 CellID->1496730429],
Cell[434365, 12386, 697, 18, 24, "ComputationResult",
 CellID->1282426654]
}, Open  ]],
Cell[CellGroupData[{
Cell[435099, 12409, 380, 10, 28, "ComputationInfo",
 CellID->2059153055],
Cell[435482, 12421, 653, 17, 430, "ComputationInfoBody",
 CellID->1534939845]
}, Closed]],
Cell[CellGroupData[{
Cell[436172, 12443, 572, 13, 44, "Computation",
 CellID->1444448452],
Cell[436747, 12458, 473, 11, 29, "ComputationResult",
 CellID->1031808536]
}, Open  ]],
Cell[CellGroupData[{
Cell[437257, 12474, 377, 10, 28, "ComputationInfo",
 CellID->948688360],
Cell[437637, 12486, 650, 17, 445, "ComputationInfoBody",
 CellID->21100703]
}, Closed]],
Cell[CellGroupData[{
Cell[438324, 12508, 788, 18, 44, "Computation",
 CellID->563077933],
Cell[439115, 12528, 343, 9, 42, "ComputationResult",
 CellID->1664994205]
}, Open  ]],
Cell[CellGroupData[{
Cell[439495, 12542, 377, 10, 28, "ComputationInfo",
 CellID->1036358125],
Cell[439875, 12554, 654, 17, 445, "ComputationInfoBody",
 CellID->1303560488]
}, Closed]],
Cell[CellGroupData[{
Cell[440566, 12576, 2157, 60, 54, "Computation",
 CellID->707667009],
Cell[442726, 12638, 176, 4, 24, "ComputationResult",
 CellID->1825061739]
}, Open  ]],
Cell[CellGroupData[{
Cell[442939, 12647, 377, 10, 28, "ComputationInfo",
 CellID->1750862647],
Cell[443319, 12659, 652, 17, 445, "ComputationInfoBody",
 CellID->1389165957]
}, Closed]],
Cell[CellGroupData[{
Cell[444008, 12681, 1421, 29, 45, "Computation",
 CellID->496046602],
Cell[445432, 12712, 319, 6, 24, "ComputationResult",
 CellID->1815416274]
}, Open  ]],
Cell[CellGroupData[{
Cell[445788, 12723, 379, 10, 28, "ComputationInfo",
 CellID->2055356184],
Cell[446170, 12735, 651, 17, 445, "ComputationInfoBody",
 CellID->866446722]
}, Closed]],
Cell[CellGroupData[{
Cell[446858, 12757, 366, 7, 41, "Computation",
 CellID->2143832469],
Cell[447227, 12766, 1335, 31, 42, "ComputationResult",
 CellID->369046373]
}, Open  ]],
Cell[CellGroupData[{
Cell[448599, 12802, 379, 10, 28, "ComputationInfo",
 CellID->173456624],
Cell[448981, 12814, 652, 17, 445, "ComputationInfoBody",
 CellID->391417162]
}, Closed]],
Cell[CellGroupData[{
Cell[449670, 12836, 676, 15, 44, "Computation",
 CellID->722194560],
Cell[450349, 12853, 1905, 60, 89, "ComputationResult",
 CellID->778638775]
}, Open  ]],
Cell[CellGroupData[{
Cell[452291, 12918, 375, 10, 28, "ComputationInfo",
 CellID->21965254],
Cell[452669, 12930, 651, 17, 445, "ComputationInfoBody",
 CellID->709996602]
}, Closed]],
Cell[453335, 12950, 466, 10, 44, "Text",
 CellID->153491769],
Cell[CellGroupData[{
Cell[453826, 12964, 917, 28, 90, "Input",
 CellID->1101254799],
Cell[454746, 12994, 1562, 48, 62, "Output",
 CellID->1633409070]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[456357, 13048, 335, 7, 30, "Subsection",
 CellID->2093452402],
Cell[456695, 13057, 48, 1, 12, "OpenEnvironment",
 CellID->1713907510],
Cell[CellGroupData[{
Cell[456768, 13062, 651, 17, 43, "EnvironmentHeader",
 CellID->810757233],
Cell[457422, 13081, 1229, 28, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1921709458", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalZ]xy"},
 CellID->1921709458],
Cell[458654, 13111, 1203, 28, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:281007089", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]xy"},
 CellID->281007089],
Cell[459860, 13141, 65, 1, 26, "EndEnvironmentMarker",
 CellID->282724972]
}, Open  ]],
Cell[459940, 13145, 49, 1, 16, "CloseEnvironment",
 CellID->1301683811],
Cell[CellGroupData[{
Cell[460014, 13150, 441, 14, 43, "Computation",
 CellID->928226933],
Cell[460458, 13166, 978, 24, 24, "ComputationResult",
 CellID->1031522555]
}, Open  ]],
Cell[CellGroupData[{
Cell[461473, 13195, 376, 10, 28, "ComputationInfo",
 CellID->494727030],
Cell[461852, 13207, 652, 17, 445, "ComputationInfoBody",
 CellID->1681718319]
}, Closed]],
Cell[CellGroupData[{
Cell[462541, 13229, 215, 5, 39, "Computation",
 CellID->1154153169],
Cell[462759, 13236, 128, 3, 24, "ComputationResult",
 CellID->565150268]
}, Open  ]],
Cell[CellGroupData[{
Cell[462924, 13244, 379, 10, 28, "ComputationInfo",
 CellID->250297112],
Cell[463306, 13256, 655, 17, 445, "ComputationInfoBody",
 CellID->1398378781]
}, Closed]],
Cell[CellGroupData[{
Cell[463998, 13278, 448, 15, 40, "Computation",
 CellID->1686609323],
Cell[464449, 13295, 877, 22, 24, "ComputationResult",
 CellID->1452783109]
}, Open  ]],
Cell[CellGroupData[{
Cell[465363, 13322, 377, 10, 28, "ComputationInfo",
 CellID->765097563],
Cell[465743, 13334, 654, 17, 445, "ComputationInfoBody",
 CellID->806774256]
}, Closed]],
Cell[CellGroupData[{
Cell[466434, 13356, 476, 16, 42, "Computation",
 CellID->493960096],
Cell[466913, 13374, 875, 22, 24, "ComputationResult",
 CellID->1934080482]
}, Open  ]],
Cell[CellGroupData[{
Cell[467825, 13401, 377, 10, 28, "ComputationInfo",
 CellID->1616194477],
Cell[468205, 13413, 651, 17, 445, "ComputationInfoBody",
 CellID->462334974]
}, Closed]],
Cell[CellGroupData[{
Cell[468893, 13435, 470, 16, 42, "Computation",
 CellID->1132140614],
Cell[469366, 13453, 877, 22, 24, "ComputationResult",
 CellID->1903936635]
}, Open  ]],
Cell[CellGroupData[{
Cell[470280, 13480, 378, 10, 28, "ComputationInfo",
 CellID->2064683746],
Cell[470661, 13492, 653, 17, 445, "ComputationInfoBody",
 CellID->2072598081]
}, Closed]],
Cell[CellGroupData[{
Cell[471351, 13514, 447, 11, 44, "Computation",
 CellID->737891318],
Cell[471801, 13527, 302, 8, 24, "ComputationResult",
 CellID->456067319]
}, Open  ]],
Cell[CellGroupData[{
Cell[472140, 13540, 377, 10, 28, "ComputationInfo",
 CellID->1015696530],
Cell[472520, 13552, 652, 17, 445, "ComputationInfoBody",
 CellID->1495241193]
}, Closed]],
Cell[CellGroupData[{
Cell[473209, 13574, 583, 14, 44, "Computation",
 CellID->1018642778],
Cell[473795, 13590, 463, 13, 29, "ComputationResult",
 CellID->81922570]
}, Open  ]],
Cell[CellGroupData[{
Cell[474295, 13608, 380, 10, 28, "ComputationInfo",
 CellID->1625548729],
Cell[474678, 13620, 653, 17, 445, "ComputationInfoBody",
 CellID->1981492116]
}, Closed]],
Cell[CellGroupData[{
Cell[475368, 13642, 606, 16, 49, "Input",
 CellID->1620459251],
Cell[475977, 13660, 327, 11, 37, "Output",
 CellID->1167063876]
}, Open  ]],
Cell[CellGroupData[{
Cell[476341, 13676, 630, 15, 47, "Computation",
 CellID->1874818521],
Cell[476974, 13693, 749, 23, 29, "ComputationResult",
 CellID->1919959885]
}, Open  ]],
Cell[CellGroupData[{
Cell[477760, 13721, 377, 10, 28, "ComputationInfo",
 CellID->580296865],
Cell[478140, 13733, 654, 17, 445, "ComputationInfoBody",
 CellID->112711656]
}, Closed]],
Cell[CellGroupData[{
Cell[478831, 13755, 636, 18, 55, "Input",
 CellID->334307522],
Cell[479470, 13775, 632, 20, 37, "Output",
 CellID->1032387437]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[480151, 13801, 357, 8, 32, "Subsection",
 CellID->484498399],
Cell[480511, 13811, 47, 1, 12, "OpenEnvironment",
 CellID->438098684],
Cell[CellGroupData[{
Cell[480583, 13816, 675, 18, 43, "EnvironmentHeader",
 CellID->269683380],
Cell[481261, 13836, 998, 25, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1512392075", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalZ]24xy"},
 CellID->1512392075],
Cell[482262, 13863, 1045, 27, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:543229319", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] \[ScriptCapitalZ]24xy"},
 CellID->543229319],
Cell[483310, 13892, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1764652756]
}, Open  ]],
Cell[483391, 13896, 49, 1, 16, "CloseEnvironment",
 CellID->1776414057],
Cell[CellGroupData[{
Cell[483465, 13901, 595, 13, 47, "Computation",
 CellID->2076222122],
Cell[484063, 13916, 742, 13, 24, "ComputationResult",
 CellID->1961765370]
}, Open  ]],
Cell[CellGroupData[{
Cell[484842, 13934, 377, 10, 28, "ComputationInfo",
 CellID->231019000],
Cell[485222, 13946, 653, 17, 460, "ComputationInfoBody",
 CellID->1617801885]
}, Closed]],
Cell[CellGroupData[{
Cell[485912, 13968, 466, 12, 44, "Computation",
 CellID->1987464113],
Cell[486381, 13982, 732, 21, 29, "ComputationResult",
 CellID->1211232653]
}, Open  ]],
Cell[CellGroupData[{
Cell[487150, 14008, 378, 10, 28, "ComputationInfo",
 CellID->1858134368],
Cell[487531, 14020, 654, 17, 460, "ComputationInfoBody",
 CellID->881951091]
}, Closed]],
Cell[488200, 14040, 730, 22, 34, "Computation",
 CellID->410409823],
Cell[CellGroupData[{
Cell[488955, 14066, 434, 11, 47, "Computation",
 CellID->388032786],
Cell[489392, 14079, 1000, 31, 29, "ComputationResult",
 CellID->1924293860]
}, Open  ]],
Cell[CellGroupData[{
Cell[490429, 14115, 376, 10, 28, "ComputationInfo",
 CellID->842604710],
Cell[490808, 14127, 652, 17, 460, "ComputationInfoBody",
 CellID->1428414825]
}, Closed]],
Cell[CellGroupData[{
Cell[491497, 14149, 435, 11, 44, "Computation",
 CellID->1393649830],
Cell[491935, 14162, 879, 26, 29, "ComputationResult",
 CellID->1840469971]
}, Open  ]],
Cell[CellGroupData[{
Cell[492851, 14193, 377, 10, 28, "ComputationInfo",
 CellID->334182051],
Cell[493231, 14205, 653, 17, 460, "ComputationInfoBody",
 CellID->1317407945]
}, Closed]],
Cell[493899, 14225, 694, 21, 34, "Computation",
 CellID->311652298],
Cell[CellGroupData[{
Cell[494618, 14250, 467, 12, 47, "Computation",
 CellID->966365481],
Cell[495088, 14264, 817, 24, 29, "ComputationResult",
 CellID->1237511167]
}, Open  ]],
Cell[CellGroupData[{
Cell[495942, 14293, 376, 10, 28, "ComputationInfo",
 CellID->149562406],
Cell[496321, 14305, 654, 17, 460, "ComputationInfoBody",
 CellID->1554428390]
}, Closed]],
Cell[496990, 14325, 690, 21, 34, "Computation",
 CellID->1424170364],
Cell[CellGroupData[{
Cell[497705, 14350, 465, 12, 47, "Computation",
 CellID->248651614],
Cell[498173, 14364, 913, 28, 29, "ComputationResult",
 CellID->934060400]
}, Open  ]],
Cell[CellGroupData[{
Cell[499123, 14397, 376, 10, 28, "ComputationInfo",
 CellID->425268034],
Cell[499502, 14409, 653, 17, 460, "ComputationInfoBody",
 CellID->121844055]
}, Closed]],
Cell[500170, 14429, 983, 31, 34, "Computation",
 CellID->166800873],
Cell[CellGroupData[{
Cell[501178, 14464, 466, 12, 47, "Computation",
 CellID->1503343685],
Cell[501647, 14478, 1068, 32, 29, "ComputationResult",
 CellID->365732408]
}, Open  ]],
Cell[CellGroupData[{
Cell[502752, 14515, 377, 10, 28, "ComputationInfo",
 CellID->312984509],
Cell[503132, 14527, 653, 17, 460, "ComputationInfoBody",
 CellID->1443280734]
}, Closed]],
Cell[503800, 14547, 966, 31, 34, "Computation",
 CellID->1183610157],
Cell[504769, 14580, 175, 4, 24, "Text",
 CellID->547291024],
Cell[CellGroupData[{
Cell[504969, 14588, 598, 18, 49, "Computation",
 CellID->2006674578],
Cell[505570, 14608, 1518, 37, 24, "ComputationResult",
 CellID->658239665]
}, Open  ]],
Cell[CellGroupData[{
Cell[507125, 14650, 377, 10, 28, "ComputationInfo",
 CellID->133220999],
Cell[507505, 14662, 653, 17, 460, "ComputationInfoBody",
 CellID->93241346]
}, Closed]],
Cell[CellGroupData[{
Cell[508195, 14684, 344, 11, 41, "Computation",
 CellID->183209066],
Cell[508542, 14697, 396, 11, 24, "ComputationResult",
 CellID->1718893651]
}, Open  ]],
Cell[CellGroupData[{
Cell[508975, 14713, 376, 10, 28, "ComputationInfo",
 CellID->173442579],
Cell[509354, 14725, 651, 17, 460, "ComputationInfoBody",
 CellID->1487544269]
}, Closed]],
Cell[CellGroupData[{
Cell[510042, 14747, 319, 7, 39, "Computation",
 CellID->1367741268],
Cell[510364, 14756, 997, 25, 24, "ComputationResult",
 CellID->511099276]
}, Open  ]],
Cell[CellGroupData[{
Cell[511398, 14786, 377, 10, 28, "ComputationInfo",
 CellID->309629685],
Cell[511778, 14798, 655, 17, 460, "ComputationInfoBody",
 CellID->1668723514]
}, Closed]],
Cell[CellGroupData[{
Cell[512470, 14820, 346, 11, 41, "Computation",
 CellID->1250393302],
Cell[512819, 14833, 274, 8, 24, "ComputationResult",
 CellID->1927486347]
}, Open  ]],
Cell[CellGroupData[{
Cell[513130, 14846, 380, 10, 28, "ComputationInfo",
 CellID->2036526983],
Cell[513513, 14858, 653, 17, 460, "ComputationInfoBody",
 CellID->1762191997]
}, Closed]],
Cell[CellGroupData[{
Cell[514203, 14880, 346, 9, 44, "Computation",
 CellID->1733660062],
Cell[514552, 14891, 275, 8, 24, "ComputationResult",
 CellID->924097278]
}, Open  ]],
Cell[CellGroupData[{
Cell[514864, 14904, 377, 10, 28, "ComputationInfo",
 CellID->576174505],
Cell[515244, 14916, 653, 17, 460, "ComputationInfoBody",
 CellID->1929933941]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[515958, 14940, 356, 7, 43, "Section",
 CellID->1382014194],
Cell[516317, 14949, 590, 14, 70, "Text",
 CellID->1670632347],
Cell[CellGroupData[{
Cell[516932, 14967, 265, 6, 70, "Subsection",
 CellID->279781653],
Cell[517200, 14975, 48, 1, 70, "OpenEnvironment",
 CellID->1257814565],
Cell[CellGroupData[{
Cell[517273, 14980, 417, 11, 70, "EnvironmentHeader",
 CellID->915451344],
Cell[517693, 14993, 965, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:280174026", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isCommRing1"},
 CellID->280174026],
Cell[518661, 15020, 65, 1, 70, "EndEnvironmentMarker",
 CellID->774494674]
}, Open  ]],
Cell[518741, 15024, 49, 1, 70, "CloseEnvironment",
 CellID->1228140165],
Cell[CellGroupData[{
Cell[518815, 15029, 2376, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:280174026", "Proof|ID:280174026-1"},
 CellID->79132233],
Cell[521194, 15099, 480, 12, 70, "ProofInfoBody",
 CellID->273864868]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[521723, 15117, 344, 7, 70, "Subsection",
 CellID->732003135],
Cell[522070, 15126, 712, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1057748777", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1057748777],
Cell[522785, 15146, 36, 1, 70, "Text",
 CellID->223648837],
Cell[522824, 15149, 743, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:464516108", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->464516108],
Cell[CellGroupData[{
Cell[523592, 15172, 377, 7, 70, "Subsubsection",
 CellID->216509956],
Cell[523972, 15181, 285, 8, 70, "Text",
 CellID->754492038],
Cell[524260, 15191, 48, 1, 70, "OpenEnvironment",
 CellID->1092859258],
Cell[CellGroupData[{
Cell[524333, 15196, 335, 9, 70, "EnvironmentHeader",
 CellID->1393668874],
Cell[524671, 15207, 1456, 45, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1148013153", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainSets \[ScriptCapitalG]", "IR"},
 CellID->1148013153],
Cell[526130, 15254, 1464, 45, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1668771272", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples \[ScriptCapitalG]", "IR"},
 CellID->1668771272],
Cell[527597, 15301, 1491, 43, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1928426019", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Sum] \[ScriptCapitalG]", "IR"},
 CellID->1928426019],
Cell[529091, 15346, 692, 17, 70, "GlobalDeclaration",
 CellTags->{
  "ID:546598180", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->546598180],
Cell[529786, 15365, 1600, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:581321955", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "reduceM \[ScriptCapitalG]", "IR"},
 CellID->581321955],
Cell[531389, 15411, 1595, 50, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:405682622", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow] \[ScriptCapitalG] multiplier", "IR"},
 
 CellID->405682622],
Cell[532987, 15463, 1613, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1165811909", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow] \[ScriptCapitalG]", "IR"},
 CellID->1165811909],
Cell[534603, 15514, 1641, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:306221528", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[LeftRightArrow] \[ScriptCapitalG]", "IR"},
 CellID->306221528],
Cell[536247, 15565, 1076, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:113746156", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isReducible \[ScriptCapitalG] multiplier", "IR"},
 CellID->113746156],
Cell[537326, 15594, 1028, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1372373678", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isReducible \[ScriptCapitalG]", "IR"},
 CellID->1372373678],
Cell[538357, 15622, 1051, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:12812135", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "symred \[ScriptCapitalG]", "IR"},
 CellID->12812135],
Cell[539411, 15652, 1057, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:168109679", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "areCorrelative \[ScriptCapitalG]", "IR"},
 CellID->168109679],
Cell[540471, 15680, 1255, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1427340407", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isNTCR \[ScriptCapitalG]", "IR"},
 CellID->1427340407],
Cell[541729, 15717, 1254, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:587977465", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isMNTCR \[ScriptCapitalG]", "IR"},
 CellID->587977465],
Cell[542986, 15754, 1017, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:950195558", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isNTCR \[ScriptCapitalG] unary", "IR"},
 CellID->950195558],
Cell[544006, 15782, 1024, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2144197915", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isMNTCR \[ScriptCapitalG] unary", "IR"},
 CellID->2144197915],
Cell[545033, 15810, 1207, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1924053921", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "red \[ScriptCapitalG] set", "IR"},
 CellID->1924053921],
Cell[546243, 15847, 1148, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1439528152", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Tilde] \[ScriptCapitalG]", "IR"},
 CellID->1439528152],
Cell[547394, 15883, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1542915564]
}, Open  ]],
Cell[547475, 15887, 48, 1, 70, "CloseEnvironment",
 CellID->304038122],
Cell[CellGroupData[{
Cell[547548, 15892, 3649, 100, 70, "ProofInfo",
 CellTags->{"Proof|ID:1148013153", "Proof|ID:1148013153-1"},
 CellID->1967681402],
Cell[551200, 15994, 482, 12, 70, "ProofInfoBody",
 CellID->2032932287]
}, Closed]],
Cell[CellGroupData[{
Cell[551719, 16011, 3660, 100, 70, "ProofInfo",
 CellTags->{"Proof|ID:1668771272", "Proof|ID:1668771272-1"},
 CellID->525775183],
Cell[555382, 16113, 481, 12, 70, "ProofInfoBody",
 CellID->397952331]
}, Closed]],
Cell[CellGroupData[{
Cell[555900, 16130, 4913, 129, 70, "ProofInfo",
 CellTags->{"Proof|ID:1928426019", "Proof|ID:1928426019-1"},
 CellID->33419350],
Cell[560816, 16261, 481, 12, 70, "ProofInfoBody",
 CellID->944826184]
}, Closed]],
Cell[CellGroupData[{
Cell[561334, 16278, 5328, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:581321955", "Proof|ID:581321955-1"},
 CellID->1728877054],
Cell[566665, 16412, 480, 12, 70, "ProofInfoBody",
 CellID->596281154]
}, Closed]],
Cell[CellGroupData[{
Cell[567182, 16429, 5297, 134, 70, "ProofInfo",
 CellTags->{"Proof|ID:405682622", "Proof|ID:405682622-1"},
 CellID->109823534],
Cell[572482, 16565, 481, 12, 70, "ProofInfoBody",
 CellID->1039693341]
}, Closed]],
Cell[CellGroupData[{
Cell[573000, 16582, 4958, 128, 70, "ProofInfo",
 CellTags->{"Proof|ID:1165811909", "Proof|ID:1165811909-1"},
 CellID->679771556],
Cell[577961, 16712, 481, 12, 70, "ProofInfoBody",
 CellID->483984180]
}, Closed]],
Cell[CellGroupData[{
Cell[578479, 16729, 5004, 128, 70, "ProofInfo",
 CellTags->{"Proof|ID:306221528", "Proof|ID:306221528-1"},
 CellID->52377820],
Cell[583486, 16859, 480, 12, 70, "ProofInfoBody",
 CellID->543843663]
}, Closed]],
Cell[CellGroupData[{
Cell[584003, 16876, 4030, 106, 70, "ProofInfo",
 CellTags->{"Proof|ID:113746156", "Proof|ID:113746156-1"},
 CellID->1673450432],
Cell[588036, 16984, 481, 12, 70, "ProofInfoBody",
 CellID->1955298365]
}, Closed]],
Cell[CellGroupData[{
Cell[588554, 17001, 3663, 99, 70, "ProofInfo",
 CellTags->{"Proof|ID:1372373678", "Proof|ID:1372373678-1"},
 CellID->1812050497],
Cell[592220, 17102, 482, 12, 70, "ProofInfoBody",
 CellID->2059822108]
}, Closed]],
Cell[CellGroupData[{
Cell[592739, 17119, 4295, 110, 70, "ProofInfo",
 CellTags->{"Proof|ID:12812135", "Proof|ID:12812135-1"},
 CellID->2042538189],
Cell[597037, 17231, 480, 12, 70, "ProofInfoBody",
 CellID->1790190522]
}, Closed]],
Cell[CellGroupData[{
Cell[597554, 17248, 4018, 105, 70, "ProofInfo",
 CellTags->{"Proof|ID:168109679", "Proof|ID:168109679-1"},
 CellID->1279405920],
Cell[601575, 17355, 481, 12, 70, "ProofInfoBody",
 CellID->2104251234]
}, Closed]],
Cell[CellGroupData[{
Cell[602093, 17372, 4647, 117, 70, "ProofInfo",
 CellTags->{"Proof|ID:1427340407", "Proof|ID:1427340407-1"},
 CellID->1645359805],
Cell[606743, 17491, 481, 12, 70, "ProofInfoBody",
 CellID->688496565]
}, Closed]],
Cell[CellGroupData[{
Cell[607261, 17508, 4643, 116, 70, "ProofInfo",
 CellTags->{"Proof|ID:587977465", "Proof|ID:587977465-1"},
 CellID->429530329],
Cell[611907, 17626, 479, 12, 70, "ProofInfoBody",
 CellID->34258095]
}, Closed]],
Cell[CellGroupData[{
Cell[612423, 17643, 3644, 99, 70, "ProofInfo",
 CellTags->{"Proof|ID:950195558", "Proof|ID:950195558-1"},
 CellID->1606573343],
Cell[616070, 17744, 480, 12, 70, "ProofInfoBody",
 CellID->187923333]
}, Closed]],
Cell[CellGroupData[{
Cell[616587, 17761, 3658, 100, 70, "ProofInfo",
 CellTags->{"Proof|ID:2144197915", "Proof|ID:2144197915-1"},
 CellID->874871898],
Cell[620248, 17863, 482, 12, 70, "ProofInfoBody",
 CellID->1569130461]
}, Closed]],
Cell[CellGroupData[{
Cell[620767, 17880, 3877, 103, 70, "ProofInfo",
 CellTags->{"Proof|ID:1924053921", "Proof|ID:1924053921-1"},
 CellID->23710079],
Cell[624647, 17985, 480, 12, 70, "ProofInfoBody",
 CellID->52275310]
}, Closed]],
Cell[CellGroupData[{
Cell[625164, 18002, 3434, 93, 70, "ProofInfo",
 CellTags->{"Proof|ID:1439528152", "Proof|ID:1439528152-1"},
 CellID->1503880702],
Cell[628601, 18097, 482, 12, 70, "ProofInfoBody",
 CellID->1443984090]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[629132, 18115, 184, 3, 70, "Subsubsection",
 CellID->57707614],
Cell[629319, 18120, 47, 1, 70, "OpenEnvironment",
 CellID->770279207],
Cell[CellGroupData[{
Cell[629391, 18125, 423, 11, 70, "EnvironmentHeader",
 CellID->634165676],
Cell[629817, 18138, 1044, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1674560446", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isPartIrreflOrder \[ScriptCapitalG]"},
 CellID->1674560446],
Cell[630864, 18170, 1025, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:167203289", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isNoetherian \[ScriptCapitalG]"},
 CellID->167203289],
Cell[631892, 18202, 821, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:569855140", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "Multipliers \[ScriptCapitalG]"},
 CellID->569855140],
Cell[632716, 18225, 791, 20, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:180032137", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R0 \[ScriptCapitalG]"},
 CellID->180032137],
Cell[633510, 18247, 793, 20, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:428662253", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R1 \[ScriptCapitalG]"},
 CellID->428662253],
Cell[634306, 18269, 792, 20, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:530735576", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R2 \[ScriptCapitalG]"},
 CellID->530735576],
Cell[635101, 18291, 794, 20, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1081126646", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R3 \[ScriptCapitalG]"},
 CellID->1081126646],
Cell[635898, 18313, 791, 20, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1591716135", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R4 \[ScriptCapitalG]"},
 CellID->1591716135],
Cell[636692, 18335, 791, 20, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:960649057", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R5 \[ScriptCapitalG]"},
 CellID->960649057],
Cell[637486, 18357, 795, 20, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:227638759", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R6 \[ScriptCapitalG]"},
 CellID->227638759],
Cell[638284, 18379, 796, 20, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2067143528", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R7 \[ScriptCapitalG]"},
 CellID->2067143528],
Cell[639083, 18401, 796, 20, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1699723496", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R8 \[ScriptCapitalG]"},
 CellID->1699723496],
Cell[639882, 18423, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1268327292]
}, Open  ]],
Cell[639963, 18427, 49, 1, 70, "CloseEnvironment",
 CellID->1372217053],
Cell[CellGroupData[{
Cell[640037, 18432, 2856, 81, 70, "ProofInfo",
 CellTags->{"Proof|ID:1674560446", "Proof|ID:1674560446-1"},
 CellID->997501851],
Cell[642896, 18515, 524, 13, 70, "ProofInfoBody",
 CellID->1684644361]
}, Closed]],
Cell[CellGroupData[{
Cell[643457, 18533, 2818, 79, 70, "ProofInfo",
 CellTags->{"Proof|ID:167203289", "Proof|ID:167203289-1"},
 CellID->24119846],
Cell[646278, 18614, 481, 12, 70, "ProofInfoBody",
 CellID->1782458510]
}, Closed]],
Cell[CellGroupData[{
Cell[646796, 18631, 2386, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:569855140", "Proof|ID:569855140-1"},
 CellID->1448471252],
Cell[649185, 18701, 481, 12, 70, "ProofInfoBody",
 CellID->1587438225]
}, Closed]],
Cell[CellGroupData[{
Cell[649703, 18718, 2349, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:180032137", "Proof|ID:180032137-1"},
 CellID->270496012],
Cell[652055, 18788, 480, 12, 70, "ProofInfoBody",
 CellID->943115763]
}, Closed]],
Cell[CellGroupData[{
Cell[652572, 18805, 2350, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:428662253", "Proof|ID:428662253-1"},
 CellID->1471451373],
Cell[654925, 18875, 481, 12, 70, "ProofInfoBody",
 CellID->1601646706]
}, Closed]],
Cell[CellGroupData[{
Cell[655443, 18892, 2349, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:530735576", "Proof|ID:530735576-1"},
 CellID->457832475],
Cell[657795, 18962, 523, 13, 70, "ProofInfoBody",
 CellID->1408840479]
}, Closed]],
Cell[CellGroupData[{
Cell[658355, 18980, 2358, 69, 70, "ProofInfo",
 CellTags->{"Proof|ID:1081126646", "Proof|ID:1081126646-1"},
 CellID->345429782],
Cell[660716, 19051, 482, 12, 70, "ProofInfoBody",
 CellID->1414623488]
}, Closed]],
Cell[CellGroupData[{
Cell[661235, 19068, 2359, 69, 70, "ProofInfo",
 CellTags->{"Proof|ID:1591716135", "Proof|ID:1591716135-1"},
 CellID->2122576129],
Cell[663597, 19139, 482, 12, 70, "ProofInfoBody",
 CellID->1462233688]
}, Closed]],
Cell[CellGroupData[{
Cell[664116, 19156, 2350, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:960649057", "Proof|ID:960649057-1"},
 CellID->1652070226],
Cell[666469, 19226, 481, 12, 70, "ProofInfoBody",
 CellID->1035512283]
}, Closed]],
Cell[CellGroupData[{
Cell[666987, 19243, 2349, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:227638759", "Proof|ID:227638759-1"},
 CellID->300639375],
Cell[669339, 19313, 480, 12, 70, "ProofInfoBody",
 CellID->822053261]
}, Closed]],
Cell[CellGroupData[{
Cell[669856, 19330, 2358, 69, 70, "ProofInfo",
 CellTags->{"Proof|ID:2067143528", "Proof|ID:2067143528-1"},
 CellID->271136742],
Cell[672217, 19401, 482, 12, 70, "ProofInfoBody",
 CellID->1516945570]
}, Closed]],
Cell[CellGroupData[{
Cell[672736, 19418, 2359, 69, 70, "ProofInfo",
 CellTags->{"Proof|ID:1699723496", "Proof|ID:1699723496-1"},
 CellID->1455152719],
Cell[675098, 19489, 482, 12, 70, "ProofInfoBody",
 CellID->1812443882]
}, Closed]],
Cell[675595, 19504, 47, 1, 70, "OpenEnvironment",
 CellID->693241098],
Cell[CellGroupData[{
Cell[675667, 19509, 400, 9, 70, "EnvironmentHeader",
 CellID->1607583592],
Cell[676070, 19520, 889, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2054051596", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isReductionRing"},
 CellID->2054051596],
Cell[676962, 19544, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1928703596]
}, Open  ]],
Cell[677043, 19548, 49, 1, 70, "CloseEnvironment",
 CellID->1784515901],
Cell[CellGroupData[{
Cell[677117, 19553, 2427, 71, 70, "ProofInfo",
 CellTags->{"Proof|ID:2054051596", "Proof|ID:2054051596-1"},
 CellID->1699885972],
Cell[679547, 19626, 482, 12, 70, "ProofInfoBody",
 CellID->1787222986]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[680090, 19645, 274, 6, 70, "Subsection",
 CellID->428105198],
Cell[680367, 19653, 769, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1031954878", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1031954878],
Cell[681139, 19675, 37, 1, 70, "Text",
 CellID->2043725198],
Cell[681179, 19678, 743, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:660860156", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->660860156],
Cell[681925, 19699, 46, 1, 70, "OpenEnvironment",
 CellID->57540137],
Cell[CellGroupData[{
Cell[681996, 19704, 440, 11, 70, "EnvironmentHeader",
 CellID->1770996009],
Cell[682439, 19717, 790, 20, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:939639226", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R9 \[ScriptCapitalG]"},
 CellID->939639226],
Cell[683232, 19739, 793, 20, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:37694441", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R10 \[ScriptCapitalG]"},
 CellID->37694441],
Cell[684028, 19761, 796, 20, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:576375692", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R11 \[ScriptCapitalG]"},
 CellID->576375692],
Cell[684827, 19783, 797, 20, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1226510984", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R12 \[ScriptCapitalG]"},
 CellID->1226510984],
Cell[685627, 19805, 799, 20, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1781039403", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "R13 \[ScriptCapitalG]"},
 CellID->1781039403],
Cell[686429, 19827, 64, 1, 70, "EndEnvironmentMarker",
 CellID->20576105]
}, Open  ]],
Cell[686508, 19831, 49, 1, 70, "CloseEnvironment",
 CellID->1461086965],
Cell[CellGroupData[{
Cell[686582, 19836, 2358, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:939639226", "Proof|ID:939639226-1"},
 CellID->1140411407],
Cell[688943, 19906, 481, 12, 70, "ProofInfoBody",
 CellID->1138027217]
}, Closed]],
Cell[CellGroupData[{
Cell[689461, 19923, 2357, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:37694441", "Proof|ID:37694441-1"},
 CellID->1985690430],
Cell[691821, 19993, 479, 12, 70, "ProofInfoBody",
 CellID->266751900]
}, Closed]],
Cell[CellGroupData[{
Cell[692337, 20010, 2362, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:576375692", "Proof|ID:576375692-1"},
 CellID->1030879810],
Cell[694702, 20080, 480, 12, 70, "ProofInfoBody",
 CellID->854397617]
}, Closed]],
Cell[CellGroupData[{
Cell[695219, 20097, 2371, 69, 70, "ProofInfo",
 CellTags->{"Proof|ID:1226510984", "Proof|ID:1226510984-1"},
 CellID->1342766186],
Cell[697593, 20168, 481, 12, 70, "ProofInfoBody",
 CellID->352367246]
}, Closed]],
Cell[CellGroupData[{
Cell[698111, 20185, 2370, 69, 70, "ProofInfo",
 CellTags->{"Proof|ID:1781039403", "Proof|ID:1781039403-1"},
 CellID->234828850],
Cell[700484, 20256, 479, 12, 70, "ProofInfoBody",
 CellID->4013719]
}, Closed]],
Cell[700978, 20271, 47, 1, 70, "OpenEnvironment",
 CellID->495058079],
Cell[CellGroupData[{
Cell[701050, 20276, 592, 13, 70, "EnvironmentHeader",
 CellID->280489971],
Cell[701645, 20291, 814, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:166003593", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isDomain"},
 CellID->166003593],
Cell[702462, 20314, 1003, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1167967104", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalG] isAlgoReductionRing"},
 CellID->1167967104],
Cell[703468, 20340, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1939621786]
}, Open  ]],
Cell[703549, 20344, 47, 1, 70, "CloseEnvironment",
 CellID->96131666],
Cell[CellGroupData[{
Cell[703621, 20349, 2381, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:166003593", "Proof|ID:166003593-1"},
 CellID->324790663],
Cell[706005, 20419, 480, 12, 70, "ProofInfoBody",
 CellID->780913280]
}, Closed]],
Cell[CellGroupData[{
Cell[706522, 20436, 2451, 71, 70, "ProofInfo",
 CellTags->{"Proof|ID:1167967104", "Proof|ID:1167967104-1"},
 CellID->1160302337],
Cell[708976, 20509, 482, 12, 70, "ProofInfoBody",
 CellID->1073405309]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[709519, 20528, 224, 5, 43, "Section",
 CellID->789808767],
Cell[709746, 20535, 696, 19, 33, "GlobalDeclaration",
 CellTags->{
  "ID:2087655986", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->2087655986],
Cell[710445, 20556, 36, 1, 24, "Text",
 CellID->850461231],
Cell[710484, 20559, 719, 19, 33, "GlobalDeclaration",
 CellTags->{
  "ID:955133330", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->955133330],
Cell[CellGroupData[{
Cell[711228, 20582, 332, 5, 38, "Subsection",
 CellID->767117005],
Cell[CellGroupData[{
Cell[711585, 20591, 194, 3, 70, "Subsubsection",
 CellID->1079351161],
Cell[711782, 20596, 47, 1, 70, "OpenEnvironment",
 CellID->689062295],
Cell[CellGroupData[{
Cell[711854, 20601, 377, 9, 70, "EnvironmentHeader",
 CellID->2101062054],
Cell[712234, 20612, 767, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:977980890", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->977980890],
Cell[713004, 20634, 816, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1748570587", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1748570587],
Cell[713823, 20657, 822, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:699276041", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership rd"},
 CellID->699276041],
Cell[714648, 20681, 1184, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:867263181", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "reducible rd"},
 CellID->867263181],
Cell[715835, 20717, 1168, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1431447619", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "irreducible rd"},
 CellID->1431447619],
Cell[717006, 20755, 1058, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1490981815", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Succeeds] rd"},
 CellID->1490981815],
Cell[718067, 20787, 64, 1, 70, "EndEnvironmentMarker",
 CellID->61218476]
}, Open  ]],
Cell[718146, 20791, 49, 1, 70, "CloseEnvironment",
 CellID->1894724465],
Cell[CellGroupData[{
Cell[718220, 20796, 4223, 112, 70, "ProofInfo",
 CellTags->{"Proof|ID:699276041", "Proof|ID:699276041-1"},
 CellID->141722599],
Cell[722446, 20910, 480, 12, 70, "ProofInfoBody",
 CellID->687345439]
}, Closed]],
Cell[CellGroupData[{
Cell[722963, 20927, 5331, 136, 70, "ProofInfo",
 CellTags->{"Proof|ID:867263181", "Proof|ID:867263181-1"},
 CellID->231570427],
Cell[728297, 21065, 478, 12, 70, "ProofInfoBody",
 CellID->9502323]
}, Closed]],
Cell[CellGroupData[{
Cell[728812, 21082, 5188, 135, 70, "ProofInfo",
 CellTags->{"Proof|ID:1431447619", "Proof|ID:1431447619-1"},
 CellID->1892915845],
Cell[734003, 21219, 482, 12, 70, "ProofInfoBody",
 CellID->1652171802]
}, Closed]],
Cell[CellGroupData[{
Cell[734522, 21236, 5158, 133, 70, "ProofInfo",
 CellTags->{"Proof|ID:1490981815", "Proof|ID:1490981815-1"},
 CellID->642445226],
Cell[739683, 21371, 481, 12, 70, "ProofInfoBody",
 CellID->379929766]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[740213, 21389, 242, 4, 70, "Subsubsection",
 CellID->1481458283],
Cell[740458, 21395, 48, 1, 70, "OpenEnvironment",
 CellID->1984772789],
Cell[CellGroupData[{
Cell[740531, 21400, 442, 10, 70, "EnvironmentHeader",
 CellID->777068689],
Cell[740976, 21412, 680, 17, 70, "GlobalDeclaration",
 CellTags->{
  "ID:243233209", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->243233209],
Cell[741659, 21431, 1505, 43, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:974939989", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdTBase1"},
 CellID->974939989],
Cell[743167, 21476, 2294, 69, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1014318336", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdTBase2"},
 CellID->1014318336],
Cell[745464, 21547, 2498, 76, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:107497168", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdTRec"},
 CellID->107497168],
Cell[747965, 21625, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1555492801]
}, Open  ]],
Cell[748046, 21629, 49, 1, 70, "CloseEnvironment",
 CellID->1152308520],
Cell[748098, 21632, 48, 1, 70, "OpenEnvironment",
 CellID->1118809324],
Cell[CellGroupData[{
Cell[748171, 21637, 504, 11, 70, "EnvironmentHeader",
 CellID->61842186],
Cell[748678, 21650, 1011, 26, 70, "GlobalDeclaration",
 CellTags->{
  "ID:839458854", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->839458854],
Cell[749692, 21678, 4623, 135, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1436407557", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "4"},
 CellID->1436407557],
Cell[754318, 21815, 1921, 55, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2098894842", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdT induction"},
 CellID->2098894842],
Cell[756242, 21872, 65, 1, 70, "EndEnvironmentMarker",
 CellID->687243513]
}, Open  ]],
Cell[756322, 21876, 48, 1, 70, "CloseEnvironment",
 CellID->463103006],
Cell[CellGroupData[{
Cell[756395, 21881, 13718, 320, 70, "ProofInfo",
 CellTags->{"Proof|ID:1436407557", "Proof|ID:1436407557-1"},
 CellID->1376449459],
Cell[770116, 22203, 481, 12, 70, "ProofInfoBody",
 CellID->792869139]
}, Closed]],
Cell[CellGroupData[{
Cell[770634, 22220, 6819, 167, 70, "ProofInfo",
 CellTags->{"Proof|ID:2098894842", "Proof|ID:2098894842-1"},
 CellID->1924563166],
Cell[777456, 22389, 482, 12, 70, "ProofInfoBody",
 CellID->2060479915]
}, Closed]],
Cell[777953, 22404, 48, 1, 70, "OpenEnvironment",
 CellID->2066597320],
Cell[CellGroupData[{
Cell[778026, 22409, 468, 10, 70, "EnvironmentHeader",
 CellID->422282357],
Cell[778497, 22421, 2089, 59, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:160961870", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "rdT alternative", "IR"},
 CellID->160961870],
Cell[780589, 22482, 65, 1, 70, "EndEnvironmentMarker",
 CellID->347215208]
}, Open  ]],
Cell[780669, 22486, 47, 1, 70, "CloseEnvironment",
 CellID->60793602],
Cell[CellGroupData[{
Cell[780741, 22491, 7622, 180, 70, "ProofInfo",
 CellTags->{"Proof|ID:160961870", "Proof|ID:160961870-1"},
 CellID->1503881909],
Cell[788366, 22673, 481, 12, 70, "ProofInfoBody",
 CellID->1283045382]
}, Closed]],
Cell[788862, 22688, 48, 1, 70, "OpenEnvironment",
 CellID->1367402977],
Cell[CellGroupData[{
Cell[788935, 22693, 431, 10, 70, "EnvironmentHeader",
 CellID->1990167663],
Cell[789369, 22705, 953, 25, 70, "GlobalDeclaration",
 CellTags->{
  "ID:168345814", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->168345814],
Cell[790325, 22732, 684, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1634249728", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1634249728],
Cell[791012, 22752, 821, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:512270654", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->512270654],
Cell[791836, 22775, 878, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2105002699", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership rdT"},
 CellID->2105002699],
Cell[792717, 22800, 1447, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1505002987", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "reducible rdT"},
 CellID->1505002987],
Cell[794167, 22846, 1325, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:514228174", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "irreducible rdT"},
 CellID->514228174],
Cell[795495, 22886, 849, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:208729427", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow]\[RawStar] rdT"},
 CellID->208729427],
Cell[796347, 22911, 1542, 45, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:278468994", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[Succeeds] rdT"},
 CellID->278468994],
Cell[797892, 22958, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1542263296]
}, Open  ]],
Cell[797973, 22962, 48, 1, 70, "CloseEnvironment",
 CellID->365655572],
Cell[CellGroupData[{
Cell[798046, 22967, 4748, 123, 70, "ProofInfo",
 CellTags->{"Proof|ID:2105002699", "Proof|ID:2105002699-1"},
 CellID->871739323],
Cell[802797, 23092, 481, 12, 70, "ProofInfoBody",
 CellID->793373954]
}, Closed]],
Cell[CellGroupData[{
Cell[803315, 23109, 7238, 179, 70, "ProofInfo",
 CellTags->{"Proof|ID:1505002987", "Proof|ID:1505002987-1"},
 CellID->327721344],
Cell[810556, 23290, 481, 12, 70, "ProofInfoBody",
 CellID->272053484]
}, Closed]],
Cell[CellGroupData[{
Cell[811074, 23307, 7185, 176, 70, "ProofInfo",
 CellTags->{"Proof|ID:514228174", "Proof|ID:514228174-1"},
 CellID->1831820924],
Cell[818262, 23485, 481, 12, 70, "ProofInfoBody",
 CellID->1587171133]
}, Closed]],
Cell[CellGroupData[{
Cell[818780, 23502, 5099, 128, 70, "ProofInfo",
 CellTags->{"Proof|ID:208729427", "Proof|ID:208729427-1"},
 CellID->970814565],
Cell[823882, 23632, 479, 12, 70, "ProofInfoBody",
 CellID->10817078]
}, Closed]],
Cell[CellGroupData[{
Cell[824398, 23649, 7103, 174, 70, "ProofInfo",
 CellTags->{"Proof|ID:278468994", "Proof|ID:278468994-1"},
 CellID->120821588],
Cell[831504, 23825, 480, 12, 70, "ProofInfoBody",
 CellID->401261460]
}, Closed]],
Cell[831999, 23840, 48, 1, 70, "OpenEnvironment",
 CellID->1171561374],
Cell[CellGroupData[{
Cell[832072, 23845, 525, 15, 70, "EnvironmentHeader",
 CellID->2018878554],
Cell[832600, 23862, 1689, 53, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:96175791", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "red rdT"},
 CellID->96175791],
Cell[834292, 23917, 65, 1, 70, "EndEnvironmentMarker",
 CellID->632845748]
}, Open  ]],
Cell[834372, 23921, 48, 1, 70, "CloseEnvironment",
 CellID->568618794],
Cell[CellGroupData[{
Cell[834445, 23926, 5391, 136, 70, "ProofInfo",
 CellTags->{"Proof|ID:96175791", "Proof|ID:96175791-1"},
 CellID->56129168],
Cell[839839, 24064, 479, 12, 70, "ProofInfoBody",
 CellID->211899362]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[840367, 24082, 191, 3, 70, "Subsubsection",
 CellID->25217793],
Cell[840561, 24087, 47, 1, 70, "OpenEnvironment",
 CellID->583061016],
Cell[CellGroupData[{
Cell[840633, 24092, 386, 9, 70, "EnvironmentHeader",
 CellID->967911099],
Cell[841022, 24103, 3592, 106, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:909149002", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "trd induction"},
 CellID->909149002],
Cell[844617, 24211, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1310698312]
}, Open  ]],
Cell[844698, 24215, 49, 1, 70, "CloseEnvironment",
 CellID->1214519135],
Cell[CellGroupData[{
Cell[844772, 24220, 10722, 252, 70, "ProofInfo",
 CellTags->{"Proof|ID:909149002", "Proof|ID:909149002-1"},
 CellID->35905978],
Cell[855497, 24474, 479, 12, 70, "ProofInfoBody",
 CellID->62162963]
}, Closed]],
Cell[855991, 24489, 47, 1, 70, "OpenEnvironment",
 CellID->473680231],
Cell[CellGroupData[{
Cell[856063, 24494, 429, 10, 70, "EnvironmentHeader",
 CellID->1909232321],
Cell[856495, 24506, 1008, 26, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1097836943", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1097836943],
Cell[857506, 24534, 763, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:206560148", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->206560148],
Cell[858272, 24556, 827, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1315173810", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership trd"},
 CellID->1315173810],
Cell[859102, 24580, 852, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1071214578", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[RightArrow]\[RawStar] trd"},
 CellID->1071214578],
Cell[859957, 24605, 880, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:657218052", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "trd \[NotElement] red"},
 CellID->657218052],
Cell[860840, 24631, 65, 1, 70, "EndEnvironmentMarker",
 CellID->672864406]
}, Open  ]],
Cell[860920, 24635, 49, 1, 70, "CloseEnvironment",
 CellID->1965420673],
Cell[CellGroupData[{
Cell[860994, 24640, 4359, 115, 70, "ProofInfo",
 CellTags->{"Proof|ID:1315173810", "Proof|ID:1315173810-1"},
 CellID->56915654],
Cell[865356, 24757, 482, 12, 70, "ProofInfoBody",
 CellID->1118371038]
}, Closed]],
Cell[CellGroupData[{
Cell[865875, 24774, 4720, 121, 70, "ProofInfo",
 CellTags->{"Proof|ID:1071214578", "Proof|ID:1071214578-1"},
 CellID->204815329],
Cell[870598, 24897, 481, 12, 70, "ProofInfoBody",
 CellID->387630611]
}, Closed]],
Cell[CellGroupData[{
Cell[871116, 24914, 4965, 127, 70, "ProofInfo",
 CellTags->{"Proof|ID:657218052", "Proof|ID:657218052-1"},
 CellID->1364381745],
Cell[876084, 25043, 480, 12, 70, "ProofInfoBody",
 CellID->659473488]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[876625, 25062, 242, 4, 30, "Subsection",
 CellID->1574109816],
Cell[CellGroupData[{
Cell[876892, 25070, 266, 6, 70, "Subsubsection",
 CellID->715367204],
Cell[877161, 25078, 822, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:480907095", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->480907095],
Cell[877986, 25099, 48, 1, 70, "OpenEnvironment",
 CellID->1014570953],
Cell[CellGroupData[{
Cell[878059, 25104, 546, 11, 70, "EnvironmentHeader",
 CellID->1638005863],
Cell[878608, 25117, 1036, 23, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1000930828", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1000930828],
Cell[879647, 25142, 3575, 99, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1774407783", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element]"},
 CellID->1774407783],
Cell[883225, 25243, 65, 1, 70, "EndEnvironmentMarker",
 CellID->731225092]
}, Open  ]],
Cell[883305, 25247, 48, 1, 70, "CloseEnvironment",
 CellID->813407975],
Cell[883356, 25250, 975, 23, 70, "GlobalDeclaration",
 CellTags->{
  "ID:193942992", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->193942992],
Cell[884334, 25275, 48, 1, 70, "OpenEnvironment",
 CellID->1112882839],
Cell[CellGroupData[{
Cell[884407, 25280, 571, 13, 70, "EnvironmentHeader",
 CellID->1753178796],
Cell[884981, 25295, 1175, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:44826735", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] ProdDomain"},
 CellID->44826735],
Cell[886159, 25328, 1305, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:155756485", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] ProdDomain DomainTuples"},
 CellID->155756485],
Cell[887467, 25365, 3415, 96, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1019394084", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] alternative I", "IR"},
 CellID->1019394084],
Cell[890885, 25463, 2963, 82, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1458538511", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] alternative II", "IR"},
 CellID->1458538511],
Cell[893851, 25547, 2529, 79, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:81000274", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[Element] alternative III"},
 CellID->81000274],
Cell[896383, 25628, 64, 1, 70, "EndEnvironmentMarker",
 CellID->30776632]
}, Open  ]],
Cell[896462, 25632, 49, 1, 70, "CloseEnvironment",
 CellID->1853514680],
Cell[CellGroupData[{
Cell[896536, 25637, 3851, 100, 70, "ProofInfo",
 CellTags->{"Proof|ID:44826735", "Proof|ID:44826735-1"},
 CellID->33075853],
Cell[900390, 25739, 479, 12, 70, "ProofInfoBody",
 CellID->151639289]
}, Closed]],
Cell[CellGroupData[{
Cell[900906, 25756, 4135, 106, 70, "ProofInfo",
 CellTags->{"Proof|ID:155756485", "Proof|ID:155756485-1"},
 CellID->228098881],
Cell[905044, 25864, 480, 12, 70, "ProofInfoBody",
 CellID->255646132]
}, Closed]],
Cell[CellGroupData[{
Cell[905561, 25881, 10471, 239, 70, "ProofInfo",
 CellTags->{"Proof|ID:1019394084", "Proof|ID:1019394084-1"},
 CellID->679633020],
Cell[916035, 26122, 482, 12, 70, "ProofInfoBody",
 CellID->1919531878]
}, Closed]],
Cell[CellGroupData[{
Cell[916554, 26139, 8823, 199, 70, "ProofInfo",
 CellTags->{"Proof|ID:1458538511", "Proof|ID:1458538511-1"},
 CellID->1939038746],
Cell[925380, 26340, 482, 12, 70, "ProofInfoBody",
 CellID->1078322293]
}, Closed]],
Cell[CellGroupData[{
Cell[925899, 26357, 7154, 171, 70, "ProofInfo",
 CellTags->{"Proof|ID:81000274", "Proof|ID:81000274-1"},
 CellID->637304832],
Cell[933056, 26530, 480, 12, 70, "ProofInfoBody",
 CellID->1448112406]
}, Closed]],
Cell[933551, 26545, 47, 1, 70, "OpenEnvironment",
 CellID->591758444],
Cell[CellGroupData[{
Cell[933623, 26550, 560, 12, 70, "EnvironmentHeader",
 CellID->273179543],
Cell[934186, 26564, 3381, 88, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1140506798", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "DomainTuples NPairs shape"},
 CellID->1140506798],
Cell[937570, 26654, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1529545215]
}, Open  ]],
Cell[937651, 26658, 49, 1, 70, "CloseEnvironment",
 CellID->1737102059],
Cell[CellGroupData[{
Cell[937725, 26663, 9712, 221, 70, "ProofInfo",
 CellTags->{"Proof|ID:1140506798", "Proof|ID:1140506798-1"},
 CellID->344640348],
Cell[947440, 26886, 481, 12, 70, "ProofInfoBody",
 CellID->601798179]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[947970, 26904, 297, 4, 70, "Subsubsection",
 CellID->330309293],
Cell[948270, 26910, 48, 1, 70, "OpenEnvironment",
 CellID->1405519764],
Cell[CellGroupData[{
Cell[948343, 26915, 404, 9, 70, "EnvironmentHeader",
 CellID->1810578538],
Cell[948750, 26926, 1029, 27, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1527335579", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1527335579],
Cell[949782, 26955, 1056, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1794762591", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership cpmTuple 1"},
 CellID->1794762591],
Cell[950841, 26985, 2346, 70, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1790156146", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpmTuple content 1"},
 CellID->1790156146],
Cell[953190, 27057, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1107193959]
}, Open  ]],
Cell[953271, 27061, 49, 1, 70, "CloseEnvironment",
 CellID->1694269586],
Cell[CellGroupData[{
Cell[953345, 27066, 6393, 157, 70, "ProofInfo",
 CellTags->{"Proof|ID:1794762591", "Proof|ID:1794762591-1"},
 CellID->1374603973],
Cell[959741, 27225, 481, 12, 70, "ProofInfoBody",
 CellID->759652547]
}, Closed]],
Cell[CellGroupData[{
Cell[960259, 27242, 11875, 271, 70, "ProofInfo",
 CellTags->{"Proof|ID:1790156146", "Proof|ID:1790156146-1"},
 CellID->1785625847],
Cell[972137, 27515, 482, 12, 70, "ProofInfoBody",
 CellID->1110321225]
}, Closed]],
Cell[972634, 27530, 48, 1, 70, "OpenEnvironment",
 CellID->1055128601],
Cell[CellGroupData[{
Cell[972707, 27535, 502, 11, 70, "EnvironmentHeader",
 CellID->2110554202],
Cell[973212, 27548, 892, 23, 70, "GlobalDeclaration",
 CellTags->{
  "ID:884601666", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->884601666],
Cell[974107, 27573, 906, 23, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1188084204", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1188084204],
Cell[975016, 27598, 1146, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1708769881", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership cpmTuple 2"},
 CellID->1708769881],
Cell[976165, 27630, 2401, 71, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:125591800", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpmTuple content 2"},
 CellID->125591800],
Cell[978569, 27703, 64, 1, 70, "EndEnvironmentMarker",
 CellID->91348497]
}, Open  ]],
Cell[978648, 27707, 48, 1, 70, "CloseEnvironment",
 CellID->201729108],
Cell[CellGroupData[{
Cell[978721, 27712, 7143, 175, 70, "ProofInfo",
 CellTags->{"Proof|ID:1708769881", "Proof|ID:1708769881-1"},
 CellID->1434737069],
Cell[985867, 27889, 482, 12, 70, "ProofInfoBody",
 CellID->1718897090]
}, Closed]],
Cell[CellGroupData[{
Cell[986386, 27906, 12998, 294, 70, "ProofInfo",
 CellTags->{"Proof|ID:125591800", "Proof|ID:125591800-1"},
 CellID->1320574433],
Cell[999387, 28202, 480, 12, 70, "ProofInfoBody",
 CellID->561308411]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[999916, 28220, 347, 5, 70, "Subsubsection",
 CellID->356782994],
Cell[1000266, 28227, 48, 1, 70, "OpenEnvironment",
 CellID->1415256475],
Cell[CellGroupData[{
Cell[1000339, 28232, 489, 11, 70, "EnvironmentHeader",
 CellID->350414352],
Cell[1000831, 28245, 731, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1122306642", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1122306642],
Cell[1001565, 28266, 3193, 102, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:885588057", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "5"},
 CellID->885588057],
Cell[1004761, 28370, 3039, 95, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:438489157", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTupleAux induction"},
 CellID->438489157],
Cell[1007803, 28467, 65, 1, 70, "EndEnvironmentMarker",
 CellID->371522104]
}, Open  ]],
Cell[1007883, 28471, 49, 1, 70, "CloseEnvironment",
 CellID->1219086707],
Cell[CellGroupData[{
Cell[1007957, 28476, 8784, 214, 70, "ProofInfo",
 CellTags->{"Proof|ID:885588057", "Proof|ID:885588057-1"},
 CellID->1319009677],
Cell[1016744, 28692, 481, 12, 70, "ProofInfoBody",
 CellID->1824975630]
}, Closed]],
Cell[CellGroupData[{
Cell[1017262, 28709, 8534, 208, 70, "ProofInfo",
 CellTags->{"Proof|ID:438489157", "Proof|ID:438489157-1"},
 CellID->1261187836],
Cell[1025799, 28919, 481, 12, 70, "ProofInfoBody",
 CellID->1114401215]
}, Closed]],
Cell[1026295, 28934, 47, 1, 70, "OpenEnvironment",
 CellID->122261323],
Cell[CellGroupData[{
Cell[1026367, 28939, 386, 9, 70, "EnvironmentHeader",
 CellID->609860341],
Cell[1026756, 28950, 844, 22, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1795822474", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1795822474],
Cell[1027603, 28974, 1070, 31, 70, "GlobalDeclaration",
 CellTags->{
  "ID:2110280918", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->2110280918],
Cell[1028676, 29007, 1057, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:481303170", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership lmTupleAux"},
 CellID->481303170],
Cell[1029736, 29037, 2692, 81, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1389102731", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTupleAux content"},
 CellID->1389102731],
Cell[1032431, 29120, 65, 1, 70, "EndEnvironmentMarker",
 CellID->178925209]
}, Open  ]],
Cell[1032511, 29124, 48, 1, 70, "CloseEnvironment",
 CellID->370655995],
Cell[CellGroupData[{
Cell[1032584, 29129, 6078, 150, 70, "ProofInfo",
 CellTags->{"Proof|ID:481303170", "Proof|ID:481303170-1"},
 CellID->1309244696],
Cell[1038665, 29281, 481, 12, 70, "ProofInfoBody",
 CellID->1292585652]
}, Closed]],
Cell[CellGroupData[{
Cell[1039183, 29298, 12904, 291, 70, "ProofInfo",
 CellTags->{"Proof|ID:1389102731", "Proof|ID:1389102731-1"},
 CellID->648616519],
Cell[1052090, 29591, 481, 12, 70, "ProofInfoBody",
 CellID->742679995]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1052620, 29609, 249, 4, 70, "Subsubsection",
 CellID->392268996],
Cell[1052872, 29615, 47, 1, 70, "OpenEnvironment",
 CellID->738030033],
Cell[CellGroupData[{
Cell[1052944, 29620, 432, 10, 70, "EnvironmentHeader",
 CellID->262315235],
Cell[1053379, 29632, 1026, 27, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1615013989", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1615013989],
Cell[1054408, 29661, 770, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1776007069", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1776007069],
Cell[1055181, 29683, 1095, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1846257369", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership lmTuple"},
 CellID->1846257369],
Cell[1056279, 29712, 2420, 72, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:851213532", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTuple \[NotEqual]"},
 CellID->851213532],
Cell[1058702, 29786, 3744, 103, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:959030563", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "lmTuple ="},
 CellID->959030563],
Cell[1062449, 29891, 65, 1, 70, "EndEnvironmentMarker",
 CellID->554761415]
}, Open  ]],
Cell[1062529, 29895, 49, 1, 70, "CloseEnvironment",
 CellID->1435911470],
Cell[CellGroupData[{
Cell[1062603, 29900, 6074, 151, 70, "ProofInfo",
 CellTags->{"Proof|ID:1846257369", "Proof|ID:1846257369-1"},
 CellID->166489159],
Cell[1068680, 30053, 481, 12, 70, "ProofInfoBody",
 CellID->595798944]
}, Closed]],
Cell[CellGroupData[{
Cell[1069198, 30070, 11096, 256, 70, "ProofInfo",
 CellTags->{"Proof|ID:851213532", "Proof|ID:851213532-1"},
 CellID->412269868],
Cell[1080297, 30328, 480, 12, 70, "ProofInfoBody",
 CellID->811817271]
}, Closed]],
Cell[CellGroupData[{
Cell[1080814, 30345, 15263, 342, 70, "ProofInfo",
 CellTags->{"Proof|ID:959030563", "Proof|ID:959030563-1"},
 CellID->1319693045],
Cell[1096080, 30689, 546, 13, 70, "ProofInfoBody",
 CellID->316689583]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1096687, 30709, 184, 3, 30, "Subsection",
 CellID->458833326],
Cell[1096874, 30714, 785, 20, 44, "GlobalDeclaration",
 CellTags->{
  "ID:1440157218", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1440157218],
Cell[CellGroupData[{
Cell[1097684, 30738, 188, 3, 28, "Subsubsection",
 CellID->162772541],
Cell[1097875, 30743, 673, 17, 70, "GlobalDeclaration",
 CellTags->{
  "ID:805845835", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->805845835],
Cell[1098551, 30762, 47, 1, 70, "OpenEnvironment",
 CellID->509087297],
Cell[CellGroupData[{
Cell[1098623, 30767, 430, 10, 70, "EnvironmentHeader",
 CellID->1841460198],
Cell[1099056, 30779, 2884, 85, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1576526273", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple"},
 CellID->1576526273],
Cell[1101943, 30866, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1611579671]
}, Open  ]],
Cell[1102024, 30870, 49, 1, 70, "CloseEnvironment",
 CellID->1825735770],
Cell[1102076, 30873, 47, 1, 70, "OpenEnvironment",
 CellID->501261685],
Cell[CellGroupData[{
Cell[1102148, 30878, 493, 11, 70, "EnvironmentHeader",
 CellID->2002554147],
Cell[1102644, 30891, 1045, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:857941005", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple allPairs"},
 CellID->857941005],
Cell[1103692, 30919, 2192, 63, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1202189993", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple update"},
 CellID->1202189993],
Cell[1105887, 30984, 3124, 89, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1698314044", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple rest"},
 CellID->1698314044],
Cell[1109014, 31075, 1743, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:812012342", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isTODOTuple exchange"},
 CellID->812012342],
Cell[1110760, 31126, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1611580180]
}, Open  ]],
Cell[1110841, 31130, 47, 1, 70, "CloseEnvironment",
 CellID->44008041],
Cell[CellGroupData[{
Cell[1110913, 31135, 4116, 106, 70, "ProofInfo",
 CellTags->{"Proof|ID:857941005", "Proof|ID:857941005-1"},
 CellID->1051929619],
Cell[1115032, 31243, 480, 12, 70, "ProofInfoBody",
 CellID->546271768]
}, Closed]],
Cell[CellGroupData[{
Cell[1115549, 31260, 7401, 177, 70, "ProofInfo",
 CellTags->{"Proof|ID:1202189993", "Proof|ID:1202189993-1"},
 CellID->1110047828],
Cell[1122953, 31439, 481, 12, 70, "ProofInfoBody",
 CellID->652755217]
}, Closed]],
Cell[CellGroupData[{
Cell[1123471, 31456, 11356, 261, 70, "ProofInfo",
 CellTags->{"Proof|ID:1698314044", "Proof|ID:1698314044-1"},
 CellID->610967613],
Cell[1134830, 31719, 481, 12, 70, "ProofInfoBody",
 CellID->709078776]
}, Closed]],
Cell[CellGroupData[{
Cell[1135348, 31736, 6913, 166, 70, "ProofInfo",
 CellTags->{"Proof|ID:812012342", "Proof|ID:812012342-1"},
 CellID->195013346],
Cell[1142264, 31904, 480, 12, 70, "ProofInfoBody",
 CellID->914583248]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1142793, 31922, 200, 3, 22, "Subsubsection",
 CellID->304206187],
Cell[1142996, 31927, 47, 1, 12, "OpenEnvironment",
 CellID->611079668],
Cell[CellGroupData[{
Cell[1143068, 31932, 385, 9, 43, "EnvironmentHeader",
 CellID->875163593],
Cell[1143456, 31943, 2816, 72, 99, "FormalTextInputFormula",
 CellTags->{
  "ID:1260897708", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "chainCrit 1"},
 CellID->1260897708],
Cell[1146275, 32017, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1094665073]
}, Open  ]],
Cell[1146356, 32021, 49, 1, 16, "CloseEnvironment",
 CellID->1919870362],
Cell[CellGroupData[{
Cell[1146430, 32026, 9956, 231, 28, "ProofInfo",
 CellTags->{"Proof|ID:1260897708", "Proof|ID:1260897708-1"},
 CellID->721986367],
Cell[1156389, 32259, 481, 12, 70, "ProofInfoBody",
 CellID->113158442]
}, Closed]],
Cell[CellGroupData[{
Cell[1156907, 32276, 317, 9, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1158351766],
Cell[1157227, 32287, 11648, 380, 1456, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->931213996],
Cell[1168878, 32669, 265, 5, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->922384546]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1169216, 32682, 313, 6, 43, "Section",
 CellID->284021016],
Cell[1169532, 32690, 817, 21, 33, "GlobalDeclaration",
 CellTags->{
  "ID:957934274", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->957934274],
Cell[1170352, 32713, 37, 1, 24, "Text",
 CellID->1053932410],
Cell[1170392, 32716, 795, 20, 33, "GlobalDeclaration",
 CellTags->{
  "ID:1363987639", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1363987639],
Cell[CellGroupData[{
Cell[1171212, 32740, 262, 6, 38, "Subsection",
 CellID->149761983],
Cell[1171477, 32748, 47, 1, 12, "OpenEnvironment",
 CellID->117373454],
Cell[CellGroupData[{
Cell[1171549, 32753, 455, 10, 43, "EnvironmentHeader",
 CellID->71182710],
Cell[1172007, 32765, 2288, 68, 130, "FormalTextInputFormula",
 CellTags->{
  "ID:184081942", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBTermOrder"},
 CellID->184081942],
Cell[1174298, 32835, 65, 1, 26, "EndEnvironmentMarker",
 CellID->356448921]
}, Open  ]],
Cell[1174378, 32839, 48, 1, 16, "CloseEnvironment",
 CellID->504344449],
Cell[1174429, 32842, 946, 22, 33, "GlobalDeclaration",
 CellTags->{
  "ID:1341865321", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1341865321],
Cell[CellGroupData[{
Cell[1175400, 32868, 137, 2, 28, "Subsubsection",
 CellID->553484892],
Cell[1175540, 32872, 48, 1, 70, "OpenEnvironment",
 CellID->1370619040],
Cell[CellGroupData[{
Cell[1175613, 32877, 452, 12, 70, "EnvironmentHeader",
 CellID->1537800075],
Cell[1176068, 32891, 812, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:320402459", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] isDomain"},
 CellID->320402459],
Cell[1176883, 32914, 1046, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1861353391", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] isPartIrreflOrder"},
 CellID->1861353391],
Cell[1177932, 32946, 1028, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:461530885", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] isNoetherian"},
 CellID->461530885],
Cell[1178963, 32978, 4409, 118, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1658780438", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] shape"},
 CellID->1658780438],
Cell[1183375, 33098, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1282280602]
}, Open  ]],
Cell[1183456, 33102, 48, 1, 70, "CloseEnvironment",
 CellID->272090212],
Cell[CellGroupData[{
Cell[1183529, 33107, 2379, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:320402459", "Proof|ID:320402459-1"},
 CellID->957846942],
Cell[1185911, 33177, 480, 12, 70, "ProofInfoBody",
 CellID->270950904]
}, Closed]],
Cell[CellGroupData[{
Cell[1186428, 33194, 2859, 81, 70, "ProofInfo",
 CellTags->{"Proof|ID:1861353391", "Proof|ID:1861353391-1"},
 CellID->71550445],
Cell[1189290, 33277, 481, 12, 70, "ProofInfoBody",
 CellID->684980337]
}, Closed]],
Cell[CellGroupData[{
Cell[1189808, 33294, 2823, 79, 70, "ProofInfo",
 CellTags->{"Proof|ID:461530885", "Proof|ID:461530885-1"},
 CellID->318070455],
Cell[1192634, 33375, 480, 12, 70, "ProofInfoBody",
 CellID->316861376]
}, Closed]],
Cell[CellGroupData[{
Cell[1193151, 33392, 12299, 283, 70, "ProofInfo",
 CellTags->{"Proof|ID:1658780438", "Proof|ID:1658780438-1"},
 CellID->65069410],
Cell[1205453, 33677, 480, 12, 70, "ProofInfoBody",
 CellID->71741560]
}, Closed]],
Cell[1205948, 33692, 47, 1, 70, "OpenEnvironment",
 CellID->154526518],
Cell[CellGroupData[{
Cell[1206020, 33697, 442, 11, 70, "EnvironmentHeader",
 CellID->2131923963],
Cell[1206465, 33710, 786, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1606683850", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1606683850],
Cell[1207254, 33732, 1508, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:973931584", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] \[Succeeds] 1"},
 CellID->973931584],
Cell[1208765, 33772, 1457, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1350364326", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] \[Succeeds] 2"},
 CellID->1350364326],
Cell[1210225, 33812, 1671, 50, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:529848255", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[ScriptCapitalL] \[Succeeds] 3"},
 CellID->529848255],
Cell[1211899, 33864, 3057, 83, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:592062547", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "\[SquareSubset] append trd"},
 CellID->592062547],
Cell[1214959, 33949, 65, 1, 70, "EndEnvironmentMarker",
 CellID->611030949]
}, Open  ]],
Cell[1215039, 33953, 49, 1, 70, "CloseEnvironment",
 CellID->1735955644],
Cell[CellGroupData[{
Cell[1215113, 33958, 5404, 135, 70, "ProofInfo",
 CellTags->{"Proof|ID:973931584", "Proof|ID:973931584-1"},
 CellID->262124527],
Cell[1220520, 34095, 479, 12, 70, "ProofInfoBody",
 CellID->11227126]
}, Closed]],
Cell[CellGroupData[{
Cell[1221036, 34112, 5278, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:1350364326", "Proof|ID:1350364326-1"},
 CellID->47470875],
Cell[1226317, 34246, 481, 12, 70, "ProofInfoBody",
 CellID->396433889]
}, Closed]],
Cell[CellGroupData[{
Cell[1226835, 34263, 5850, 141, 70, "ProofInfo",
 CellTags->{"Proof|ID:529848255", "Proof|ID:529848255-1"},
 CellID->196790701],
Cell[1232688, 34406, 480, 12, 70, "ProofInfoBody",
 CellID->478260070]
}, Closed]],
Cell[CellGroupData[{
Cell[1233205, 34423, 12382, 288, 70, "ProofInfo",
 CellTags->{"Proof|ID:592062547", "Proof|ID:592062547-1"},
 CellID->1513019922],
Cell[1245590, 34713, 525, 13, 70, "ProofInfoBody",
 CellID->2131077418]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1246164, 34732, 138, 2, 22, "Subsubsection",
 CellID->1185760220],
Cell[1246305, 34736, 48, 1, 12, "OpenEnvironment",
 CellID->2057862047],
Cell[CellGroupData[{
Cell[1246378, 34741, 396, 9, 43, "EnvironmentHeader",
 CellID->886140291],
Cell[1246777, 34752, 743, 18, 33, "GlobalDeclaration",
 CellTags->{
  "ID:1848949834", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1848949834],
Cell[1247523, 34772, 2717, 74, 114, "FormalTextInputFormula",
 CellTags->{
  "ID:1362258751", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxTest"},
 CellID->1362258751],
Cell[1250243, 34848, 1436, 37, 32, "FormalTextInputFormula",
 CellTags->{
  "ID:1347349813", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxBase"},
 CellID->1347349813],
Cell[1251682, 34887, 4352, 115, 154, "FormalTextInputFormula",
 CellTags->{
  "ID:270708621", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec1"},
 CellID->270708621],
Cell[1256037, 35004, 752, 25, 25, "EnvironmentText",
 CellID->1510189147],
Cell[1256792, 35031, 3492, 99, 154, "FormalTextInputFormula",
 CellTags->{
  "ID:1542247232", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec2"},
 CellID->1542247232],
Cell[1260287, 35132, 5126, 136, 201, "FormalTextInputFormula",
 CellTags->{
  "ID:580452282", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec3"},
 CellID->580452282],
Cell[1265416, 35270, 831, 27, 25, "EnvironmentText",
 CellID->2063275027],
Cell[1266250, 35299, 5545, 153, 226, "FormalTextInputFormula",
 CellTags->{
  "ID:598818422", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxRec4"},
 CellID->598818422],
Cell[1271798, 35454, 66, 1, 26, "EndEnvironmentMarker",
 CellID->2012993027]
}, Open  ]],
Cell[1271879, 35458, 49, 1, 16, "CloseEnvironment",
 CellID->1217706098],
Cell[1271931, 35461, 48, 1, 12, "OpenEnvironment",
 CellID->1578468637],
Cell[CellGroupData[{
Cell[1272004, 35466, 423, 10, 43, "EnvironmentHeader",
 CellID->336243787],
Cell[1272430, 35478, 651, 16, 32, "GlobalDeclaration",
 CellTags->{
  "ID:1392375551", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1392375551],
Cell[1273084, 35496, 8640, 228, 316, "FormalTextInputFormula",
 CellTags->{
  "ID:994411308", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "1"},
 CellID->994411308],
Cell[1281727, 35726, 17123, 416, 446, "FormalTextInputFormula",
 CellTags->{
  "ID:1185976259", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "2"},
 CellID->1185976259],
Cell[1298853, 36144, 4696, 139, 173, "FormalTextInputFormula",
 CellTags->{
  "ID:1125858125", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "3"},
 CellID->1125858125],
Cell[1303552, 36285, 4442, 125, 222, "FormalTextInputFormula",
 CellTags->{
  "ID:1780855977", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAux induction"},
 CellID->1780855977],
Cell[1307997, 36412, 4432, 126, 281, "FormalTextInputFormula",
 CellTags->{
  "ID:380025998", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAux induction II"},
 CellID->380025998],
Cell[1312432, 36540, 66, 1, 26, "EndEnvironmentMarker",
 CellID->2126131274]
}, Open  ]],
Cell[1312513, 36544, 49, 1, 16, "CloseEnvironment",
 CellID->1859434148],
Cell[CellGroupData[{
Cell[1312587, 36549, 23368, 514, 28, "ProofInfo",
 CellTags->{"Proof|ID:994411308", "Proof|ID:994411308-1"},
 CellID->87121011],
Cell[1335958, 37065, 478, 12, 70, "ProofInfoBody",
 CellID->1110560]
}, Closed]],
Cell[CellGroupData[{
Cell[1336473, 37082, 53148, 1138, 26, "ProofInfo",
 CellTags->{"Proof|ID:1185976259", "Proof|ID:1185976259-1"},
 CellID->50921615],
Cell[1389624, 38222, 523, 13, 70, "ProofInfoBody",
 CellID->188807672]
}, Closed]],
Cell[CellGroupData[{
Cell[1390184, 38240, 16212, 370, 26, "ProofInfo",
 CellTags->{"Proof|ID:1125858125", "Proof|ID:1125858125-1"},
 CellID->52228001],
Cell[1406399, 38612, 481, 12, 70, "ProofInfoBody",
 CellID->673041172]
}, Closed]],
Cell[CellGroupData[{
Cell[1406917, 38629, 17071, 377, 26, "ProofInfo",
 CellTags->{"Proof|ID:1780855977", "Proof|ID:1780855977-1"},
 CellID->3158281],
Cell[1423991, 39008, 480, 12, 70, "ProofInfoBody",
 CellID->98918190]
}, Closed]],
Cell[CellGroupData[{
Cell[1424508, 39025, 17319, 382, 26, "ProofInfo",
 CellTags->{"Proof|ID:380025998", "Proof|ID:380025998-1"},
 CellID->1967872054],
Cell[1441830, 39409, 481, 12, 70, "ProofInfoBody",
 CellID->1105788270]
}, Closed]],
Cell[CellGroupData[{
Cell[1442348, 39426, 341, 9, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1976979955],
Cell[1442692, 39437, 14200, 444, 1166, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1320712767],
Cell[1456895, 39883, 265, 5, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10001.},
 CellID->1296315400]
}, Closed]],
Cell[CellGroupData[{
Cell[1457197, 39893, 320, 9, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1173610870],
Cell[1457520, 39904, 38040, 1098, 2562, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->2136547636],
Cell[1495563, 41004, 267, 5, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1971546439]
}, Closed]],
Cell[1495845, 41012, 48, 1, 12, "OpenEnvironment",
 CellID->2063122267],
Cell[CellGroupData[{
Cell[1495918, 41017, 391, 9, 43, "EnvironmentHeader",
 CellID->70806236],
Cell[1496312, 41028, 3435, 92, 164, "FormalTextInputFormula",
 CellTags->{
  "ID:1379909256", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAux 3 alternative", "IR"},
 CellID->1379909256],
Cell[1499750, 41122, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1623219371]
}, Open  ]],
Cell[1499831, 41126, 49, 1, 16, "CloseEnvironment",
 CellID->2040040187],
Cell[CellGroupData[{
Cell[1499905, 41131, 15695, 342, 28, "ProofInfo",
 CellTags->{"Proof|ID:1379909256", "Proof|ID:1379909256-1"},
 CellID->1806109079],
Cell[1515603, 41475, 482, 12, 70, "ProofInfoBody",
 CellID->1501295721]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1516146, 41494, 243, 4, 30, "Subsection",
 CellID->145398451],
Cell[1516392, 41500, 778, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:884344895", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->884344895],
Cell[1517173, 41522, 47, 1, 70, "OpenEnvironment",
 CellID->491130265],
Cell[CellGroupData[{
Cell[1517245, 41527, 394, 9, 70, "EnvironmentHeader",
 CellID->2121949508],
Cell[1517642, 41538, 855, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:2012261150", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->2012261150],
Cell[1518500, 41561, 1873, 57, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1467104417", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxBase"},
 CellID->1467104417],
Cell[1520376, 41620, 1850, 57, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1354352171", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec1"},
 CellID->1354352171],
Cell[1522229, 41679, 1850, 57, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1322915202", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec2"},
 CellID->1322915202],
Cell[1524082, 41738, 1847, 57, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:620036821", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec3"},
 CellID->620036821],
Cell[1525932, 41797, 1850, 57, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1111467670", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAuxRec4"},
 CellID->1111467670],
Cell[1527785, 41856, 65, 1, 70, "EndEnvironmentMarker",
 CellID->529886557]
}, Open  ]],
Cell[1527865, 41860, 49, 1, 70, "CloseEnvironment",
 CellID->2052416756],
Cell[CellGroupData[{
Cell[1527939, 41865, 8473, 197, 70, "ProofInfo",
 CellTags->{"Proof|ID:1467104417", "Proof|ID:1467104417-1"},
 CellID->1997727746],
Cell[1536415, 42064, 482, 12, 70, "ProofInfoBody",
 CellID->1261042889]
}, Closed]],
Cell[CellGroupData[{
Cell[1536934, 42081, 8472, 197, 70, "ProofInfo",
 CellTags->{"Proof|ID:1354352171", "Proof|ID:1354352171-1"},
 CellID->478354620],
Cell[1545409, 42280, 481, 12, 70, "ProofInfoBody",
 CellID->460981908]
}, Closed]],
Cell[CellGroupData[{
Cell[1545927, 42297, 8473, 197, 70, "ProofInfo",
 CellTags->{"Proof|ID:1322915202", "Proof|ID:1322915202-1"},
 CellID->1180582011],
Cell[1554403, 42496, 481, 12, 70, "ProofInfoBody",
 CellID->195380916]
}, Closed]],
Cell[CellGroupData[{
Cell[1554921, 42513, 8464, 196, 70, "ProofInfo",
 CellTags->{"Proof|ID:620036821", "Proof|ID:620036821-1"},
 CellID->1505716064],
Cell[1563388, 42711, 480, 12, 70, "ProofInfoBody",
 CellID->116667553]
}, Closed]],
Cell[CellGroupData[{
Cell[1563905, 42728, 8472, 197, 70, "ProofInfo",
 CellTags->{"Proof|ID:1111467670", "Proof|ID:1111467670-1"},
 CellID->284241391],
Cell[1572380, 42927, 481, 12, 70, "ProofInfoBody",
 CellID->723159660]
}, Closed]],
Cell[1572876, 42942, 47, 1, 70, "OpenEnvironment",
 CellID->525493891],
Cell[CellGroupData[{
Cell[1572948, 42947, 397, 9, 70, "EnvironmentHeader",
 CellID->167641070],
Cell[1573348, 42958, 2753, 80, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1314333166", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GBAux"},
 CellID->1314333166],
Cell[1576104, 43040, 65, 1, 70, "EndEnvironmentMarker",
 CellID->946812009]
}, Open  ]],
Cell[1576184, 43044, 49, 1, 70, "CloseEnvironment",
 CellID->1790647428],
Cell[CellGroupData[{
Cell[1576258, 43049, 9134, 217, 70, "ProofInfo",
 CellTags->{"Proof|ID:1314333166", "Proof|ID:1314333166-1"},
 CellID->888484883],
Cell[1585395, 43268, 481, 12, 70, "ProofInfoBody",
 CellID->842826484]
}, Closed]],
Cell[1585891, 43283, 47, 1, 70, "OpenEnvironment",
 CellID->486203571],
Cell[CellGroupData[{
Cell[1585963, 43288, 506, 11, 70, "EnvironmentHeader",
 CellID->1209390519],
Cell[1586472, 43301, 2267, 65, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1167572947", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership GBAux"},
 CellID->1167572947],
Cell[1588742, 43368, 65, 1, 70, "EndEnvironmentMarker",
 CellID->765917740]
}, Open  ]],
Cell[1588822, 43372, 48, 1, 70, "CloseEnvironment",
 CellID->858293749],
Cell[CellGroupData[{
Cell[1588895, 43377, 8362, 199, 70, "ProofInfo",
 CellTags->{"Proof|ID:1167572947", "Proof|ID:1167572947-1"},
 CellID->897701299],
Cell[1597260, 43578, 481, 12, 70, "ProofInfoBody",
 CellID->786989361]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1597790, 43596, 182, 3, 30, "Subsection",
 CellID->1967810377],
Cell[1597975, 43601, 783, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:208228172", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->208228172],
Cell[1598761, 43623, 47, 1, 70, "OpenEnvironment",
 CellID->833515163],
Cell[CellGroupData[{
Cell[1598833, 43628, 438, 11, 70, "EnvironmentHeader",
 CellID->1166686321],
Cell[1599274, 43641, 2864, 81, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:781335069", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "MMPairs \[RightArrow]\[RawStar]"},
 CellID->781335069],
Cell[1602141, 43724, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1385002694]
}, Open  ]],
Cell[1602222, 43728, 49, 1, 70, "CloseEnvironment",
 CellID->1935100623],
Cell[CellGroupData[{
Cell[1602296, 43733, 12233, 278, 70, "ProofInfo",
 CellTags->{"Proof|ID:781335069", "Proof|ID:781335069-1"},
 CellID->283149330],
Cell[1614532, 44013, 480, 12, 70, "ProofInfoBody",
 CellID->296243687]
}, Closed]],
Cell[1615027, 44028, 48, 1, 70, "OpenEnvironment",
 CellID->1412924928],
Cell[CellGroupData[{
Cell[1615100, 44033, 369, 9, 70, "EnvironmentHeader",
 CellID->843145011],
Cell[1615472, 44044, 855, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1452629404", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1452629404],
Cell[1616330, 44067, 1429, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1597922168", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxBase"},
 CellID->1597922168],
Cell[1617762, 44109, 1426, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:348489962", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec1"},
 CellID->348489962],
Cell[1619191, 44151, 1431, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1787031010", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec2"},
 CellID->1787031010],
Cell[1620625, 44193, 1426, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:963974147", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec3"},
 CellID->963974147],
Cell[1622054, 44235, 1428, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:473957484", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAuxRec4"},
 CellID->473957484],
Cell[1623485, 44277, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1659585585]
}, Open  ]],
Cell[1623566, 44281, 48, 1, 70, "CloseEnvironment",
 CellID->154264307],
Cell[CellGroupData[{
Cell[1623639, 44286, 8323, 195, 70, "ProofInfo",
 CellTags->{"Proof|ID:1597922168", "Proof|ID:1597922168-1"},
 CellID->1879777321],
Cell[1631965, 44483, 481, 12, 70, "ProofInfoBody",
 CellID->208056954]
}, Closed]],
Cell[CellGroupData[{
Cell[1632483, 44500, 8313, 194, 70, "ProofInfo",
 CellTags->{"Proof|ID:348489962", "Proof|ID:348489962-1"},
 CellID->917719347],
Cell[1640799, 44696, 481, 12, 70, "ProofInfoBody",
 CellID->1071294072]
}, Closed]],
Cell[CellGroupData[{
Cell[1641317, 44713, 8323, 195, 70, "ProofInfo",
 CellTags->{"Proof|ID:1787031010", "Proof|ID:1787031010-1"},
 CellID->1427595038],
Cell[1649643, 44910, 482, 12, 70, "ProofInfoBody",
 CellID->2125736511]
}, Closed]],
Cell[CellGroupData[{
Cell[1650162, 44927, 8314, 194, 70, "ProofInfo",
 CellTags->{"Proof|ID:963974147", "Proof|ID:963974147-1"},
 CellID->1497403325],
Cell[1658479, 45123, 480, 12, 70, "ProofInfoBody",
 CellID->996071133]
}, Closed]],
Cell[CellGroupData[{
Cell[1658996, 45140, 8314, 194, 70, "ProofInfo",
 CellTags->{"Proof|ID:473957484", "Proof|ID:473957484-1"},
 CellID->1947252190],
Cell[1667313, 45336, 480, 12, 70, "ProofInfoBody",
 CellID->246554488]
}, Closed]],
Cell[1667808, 45351, 48, 1, 70, "OpenEnvironment",
 CellID->1579370208],
Cell[CellGroupData[{
Cell[1667881, 45356, 398, 9, 70, "EnvironmentHeader",
 CellID->2113590487],
Cell[1668282, 45367, 2261, 66, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1826585682", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GBAux"},
 CellID->1826585682],
Cell[1670546, 45435, 65, 1, 70, "EndEnvironmentMarker",
 CellID->670628316]
}, Open  ]],
Cell[1670626, 45439, 49, 1, 70, "CloseEnvironment",
 CellID->1004533723],
Cell[CellGroupData[{
Cell[1670700, 45444, 8933, 213, 70, "ProofInfo",
 CellTags->{"Proof|ID:1826585682", "Proof|ID:1826585682-1"},
 CellID->98529934],
Cell[1679636, 45659, 482, 12, 70, "ProofInfoBody",
 CellID->1922222818]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1680167, 45677, 211, 5, 30, "Subsection",
 CellID->23100310],
Cell[1680381, 45684, 782, 20, 44, "GlobalDeclaration",
 CellTags->{
  "ID:416121228", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->416121228],
Cell[CellGroupData[{
Cell[1681188, 45708, 196, 3, 28, "Subsubsection",
 CellID->1168597562],
Cell[1681387, 45713, 48, 1, 70, "OpenEnvironment",
 CellID->1967738288],
Cell[CellGroupData[{
Cell[1681460, 45718, 507, 12, 70, "EnvironmentHeader",
 CellID->1074460594],
Cell[1681970, 45732, 682, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:2050404037", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->2050404037],
Cell[1682655, 45752, 4775, 124, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1837575531", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple"},
 CellID->1837575531],
Cell[1687433, 45878, 3358, 90, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:346323949", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1"},
 CellID->346323949],
Cell[1690794, 45970, 2895, 77, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1643593823", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2"},
 CellID->1643593823],
Cell[1693692, 46049, 3067, 85, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1213243088", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3"},
 CellID->1213243088],
Cell[1696762, 46136, 65, 1, 70, "EndEnvironmentMarker",
 CellID->388797222]
}, Open  ]],
Cell[1696842, 46140, 48, 1, 70, "CloseEnvironment",
 CellID->424254111]
}, Closed]],
Cell[CellGroupData[{
Cell[1696927, 46146, 210, 3, 22, "Subsubsection",
 CellID->2121626066],
Cell[1697140, 46151, 47, 1, 12, "OpenEnvironment",
 CellID->977102521],
Cell[CellGroupData[{
Cell[1697212, 46156, 532, 11, 43, "EnvironmentHeader",
 CellID->2132467789],
Cell[1697747, 46169, 3785, 112, 130, "FormalTextInputFormula",
 CellTags->{
  "ID:949923506", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "h=0"},
 CellID->949923506],
Cell[1701535, 46283, 65, 1, 26, "EndEnvironmentMarker",
 CellID->543412832]
}, Open  ]],
Cell[1701615, 46287, 48, 1, 16, "CloseEnvironment",
 CellID->505215769],
Cell[CellGroupData[{
Cell[1701688, 46292, 15153, 346, 28, "ProofInfo",
 CellTags->{"Proof|ID:949923506", "Proof|ID:949923506-1"},
 CellID->221566149],
Cell[1716844, 46640, 479, 12, 70, "ProofInfoBody",
 CellID->87967590]
}, Closed]],
Cell[1717338, 46655, 48, 1, 12, "OpenEnvironment",
 CellID->1859584630],
Cell[CellGroupData[{
Cell[1717411, 46660, 404, 9, 43, "EnvironmentHeader",
 CellID->852372469],
Cell[1717818, 46671, 776, 21, 33, "GlobalDeclaration",
 CellTags->{
  "ID:1322820809", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1322820809],
Cell[1718597, 46694, 872, 23, 44, "GlobalDeclaration",
 CellTags->{
  "ID:1772563804", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1772563804],
Cell[1719472, 46719, 853, 22, 34, "GlobalDeclaration",
 CellTags->{
  "ID:667808408", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->667808408],
Cell[1720328, 46743, 2593, 81, 69, "FormalTextInputFormula",
 CellTags->{
  "ID:1601898370", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "h\[NotEqual]0"},
 CellID->1601898370],
Cell[1722924, 46826, 1850, 50, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:622683853", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpConnectible append h"},
 CellID->622683853],
Cell[1724777, 46878, 66, 1, 26, "EndEnvironmentMarker",
 CellID->2105705679]
}, Open  ]],
Cell[1724858, 46882, 48, 1, 16, "CloseEnvironment",
 CellID->645399240],
Cell[CellGroupData[{
Cell[1724931, 46887, 15843, 359, 28, "ProofInfo",
 CellTags->{"Proof|ID:1601898370", "Proof|ID:1601898370-1"},
 CellID->76678778],
Cell[1740777, 47248, 481, 12, 70, "ProofInfoBody",
 CellID->653057816]
}, Closed]],
Cell[CellGroupData[{
Cell[1741295, 47265, 13743, 312, 26, "ProofInfo",
 CellTags->{"Proof|ID:622683853", "Proof|ID:622683853-1"},
 CellID->836026655],
Cell[1755041, 47579, 481, 12, 70, "ProofInfoBody",
 CellID->1838968259]
}, Closed]],
Cell[1755537, 47594, 48, 1, 12, "OpenEnvironment",
 CellID->1568209243],
Cell[CellGroupData[{
Cell[1755610, 47599, 385, 9, 43, "EnvironmentHeader",
 CellID->296316007],
Cell[1755998, 47610, 3585, 95, 93, "FormalTextInputFormula",
 CellTags->{
  "ID:1834517245", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxTest alternative"},
 CellID->1834517245],
Cell[1759586, 47707, 4183, 100, 183, "FormalTextInputFormula",
 CellTags->{
  "ID:1923946854", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxTest 1"},
 CellID->1923946854],
Cell[1763772, 47809, 65, 1, 26, "EndEnvironmentMarker",
 CellID->857381505]
}, Open  ]],
Cell[1763852, 47813, 49, 1, 16, "CloseEnvironment",
 CellID->1607665888],
Cell[CellGroupData[{
Cell[1763926, 47818, 18116, 407, 28, "ProofInfo",
 CellTags->{"Proof|ID:1834517245", "Proof|ID:1834517245-1"},
 CellID->1216317993],
Cell[1782045, 48227, 482, 12, 70, "ProofInfoBody",
 CellID->1362723759]
}, Closed]],
Cell[CellGroupData[{
Cell[1782564, 48244, 18127, 405, 26, "ProofInfo",
 CellTags->{"Proof|ID:1923946854", "Proof|ID:1923946854-1"},
 CellID->1529476345],
Cell[1800694, 48651, 482, 12, 70, "ProofInfoBody",
 CellID->1417488253]
}, Closed]],
Cell[CellGroupData[{
Cell[1801213, 48668, 317, 9, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->2144907034],
Cell[1801533, 48679, 7388, 226, 704, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1186451214],
Cell[1808924, 48907, 267, 5, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1306340414]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1809240, 48918, 145, 2, 22, "Subsubsection",
 CellID->1918221565],
Cell[1809388, 48922, 48, 1, 70, "OpenEnvironment",
 CellID->1414827049],
Cell[CellGroupData[{
Cell[1809461, 48927, 538, 13, 70, "EnvironmentHeader",
 CellID->1588362347],
Cell[1810002, 48942, 1031, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:258768016", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 empty"},
 CellID->258768016],
Cell[1811036, 48971, 855, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1002747542", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1002747542],
Cell[1811894, 48994, 946, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:673303920", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 GBAuxRec3"},
 CellID->673303920],
Cell[1812843, 49020, 949, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1502806330", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 GBAuxRec4"},
 CellID->1502806330],
Cell[1813795, 49046, 66, 1, 70, "EndEnvironmentMarker",
 CellID->2052180156]
}, Open  ]],
Cell[1813876, 49050, 48, 1, 70, "CloseEnvironment",
 CellID->865462022],
Cell[CellGroupData[{
Cell[1813949, 49055, 3543, 92, 70, "ProofInfo",
 CellTags->{"Proof|ID:258768016", "Proof|ID:258768016-1"},
 CellID->670832970],
Cell[1817495, 49149, 549, 13, 70, "ProofInfoBody",
 CellID->1269726962]
}, Closed]],
Cell[CellGroupData[{
Cell[1818081, 49167, 5540, 138, 70, "ProofInfo",
 CellTags->{"Proof|ID:673303920", "Proof|ID:673303920-1"},
 CellID->2063565608],
Cell[1823624, 49307, 479, 12, 70, "ProofInfoBody",
 CellID->30292895]
}, Closed]],
Cell[CellGroupData[{
Cell[1824140, 49324, 5549, 139, 70, "ProofInfo",
 CellTags->{"Proof|ID:1502806330", "Proof|ID:1502806330-1"},
 CellID->1751898811],
Cell[1829692, 49465, 482, 12, 70, "ProofInfoBody",
 CellID->1643547098]
}, Closed]],
Cell[1830189, 49480, 48, 1, 70, "OpenEnvironment",
 CellID->1047637483],
Cell[CellGroupData[{
Cell[1830262, 49485, 392, 9, 70, "EnvironmentHeader",
 CellID->130625381],
Cell[1830657, 49496, 1065, 30, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1133484500", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1133484500],
Cell[1831725, 49528, 852, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:291397089", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->291397089],
Cell[1832580, 49551, 1277, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1494094918", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible1 GBAux"},
 CellID->1494094918],
Cell[1833860, 49587, 1776, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:402861438", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "cpConnectible GBAux lmTuple"},
 CellID->402861438],
Cell[1835639, 49638, 65, 1, 70, "EndEnvironmentMarker",
 CellID->492085035]
}, Open  ]],
Cell[1835719, 49642, 49, 1, 70, "CloseEnvironment",
 CellID->1004237531],
Cell[CellGroupData[{
Cell[1835793, 49647, 7827, 187, 70, "ProofInfo",
 CellTags->{"Proof|ID:1494094918", "Proof|ID:1494094918-1"},
 CellID->1479447805],
Cell[1843623, 49836, 481, 12, 70, "ProofInfoBody",
 CellID->879762366]
}, Closed]],
Cell[CellGroupData[{
Cell[1844141, 49853, 11025, 256, 70, "ProofInfo",
 CellTags->{"Proof|ID:402861438", "Proof|ID:402861438-1"},
 CellID->111398506],
Cell[1855169, 50111, 479, 12, 70, "ProofInfoBody",
 CellID->23314887]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1855697, 50129, 187, 3, 22, "Subsubsection",
 CellID->910786473],
Cell[1855887, 50134, 48, 1, 12, "OpenEnvironment",
 CellID->1347382991],
Cell[CellGroupData[{
Cell[1855960, 50139, 396, 9, 43, "EnvironmentHeader",
 CellID->1658075726],
Cell[1856359, 50150, 800, 21, 33, "GlobalDeclaration",
 CellTags->{
  "ID:1173305577", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1173305577],
Cell[1857162, 50173, 651, 16, 30, "GlobalDeclaration",
 CellTags->{
  "ID:1545923689", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1545923689],
Cell[1857816, 50191, 1393, 39, 43, "FormalTextInputFormula",
 CellTags->{
  "ID:1684239249", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple empty"},
 CellID->1684239249],
Cell[1859212, 50232, 1629, 45, 43, "FormalTextInputFormula",
 CellTags->{
  "ID:547535692", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple prepend"},
 CellID->547535692],
Cell[1860844, 50279, 2158, 60, 89, "FormalTextInputFormula",
 CellTags->{
  "ID:812188873", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple isTODOTuple"},
 CellID->812188873],
Cell[1863005, 50341, 1407, 41, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:1105161129", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple lmTuple"},
 CellID->1105161129],
Cell[1864415, 50384, 1476, 44, 43, "FormalTextInputFormula",
 CellTags->{
  "ID:981490890", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple 0"},
 CellID->981490890],
Cell[1865894, 50430, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1131566164]
}, Open  ]],
Cell[1865975, 50434, 48, 1, 16, "CloseEnvironment",
 CellID->584510172],
Cell[CellGroupData[{
Cell[1866048, 50439, 6983, 169, 28, "ProofInfo",
 CellTags->{"Proof|ID:1684239249", "Proof|ID:1684239249-1"},
 CellID->74113726],
Cell[1873034, 50610, 481, 12, 70, "ProofInfoBody",
 CellID->623669078]
}, Closed]],
Cell[CellGroupData[{
Cell[1873552, 50627, 7700, 184, 26, "ProofInfo",
 CellTags->{"Proof|ID:547535692", "Proof|ID:547535692-1"},
 CellID->1769495093],
Cell[1881255, 50813, 480, 12, 70, "ProofInfoBody",
 CellID->423907628]
}, Closed]],
Cell[CellGroupData[{
Cell[1881772, 50830, 9452, 220, 26, "ProofInfo",
 CellTags->{"Proof|ID:812188873", "Proof|ID:812188873-1"},
 CellID->1479812468],
Cell[1891227, 51052, 480, 12, 70, "ProofInfoBody",
 CellID->759303520]
}, Closed]],
Cell[CellGroupData[{
Cell[1891744, 51069, 7709, 187, 26, "ProofInfo",
 CellTags->{"Proof|ID:1105161129", "Proof|ID:1105161129-1"},
 CellID->2103372990],
Cell[1899456, 51258, 482, 12, 70, "ProofInfoBody",
 CellID->1126448357]
}, Closed]],
Cell[CellGroupData[{
Cell[1899975, 51275, 6987, 172, 26, "ProofInfo",
 CellTags->{"Proof|ID:981490890", "Proof|ID:981490890-1"},
 CellID->296074352],
Cell[1906965, 51449, 481, 12, 70, "ProofInfoBody",
 CellID->1927354151]
}, Closed]],
Cell[1907461, 51464, 47, 1, 12, "OpenEnvironment",
 CellID->584142620],
Cell[CellGroupData[{
Cell[1907533, 51469, 406, 10, 43, "EnvironmentHeader",
 CellID->2080239230],
Cell[1907942, 51481, 1163, 32, 44, "GlobalDeclaration",
 CellTags->{
  "ID:681053692", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->681053692],
Cell[1909108, 51515, 797, 21, 33, "GlobalDeclaration",
 CellTags->{
  "ID:995853659", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->995853659],
Cell[1909908, 51538, 876, 23, 44, "GlobalDeclaration",
 CellTags->{
  "ID:218025475", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->218025475],
Cell[1910787, 51563, 1195, 30, 34, "GlobalDeclaration",
 CellTags->{
  "ID:2052980984", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->2052980984],
Cell[1911985, 51595, 1289, 36, 48, "GlobalDeclaration",
 CellTags->{
  "ID:925810378", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->925810378],
Cell[1913277, 51633, 1524, 42, 81, "FormalTextInputFormula",
 CellTags->{
  "ID:423392620", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple 1"},
 CellID->423392620],
Cell[1914804, 51677, 1248, 35, 34, "FormalTextInputFormula",
 CellTags->{
  "ID:1533257011", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isCPTODOTuple 2"},
 CellID->1533257011],
Cell[1916055, 51714, 65, 1, 26, "EndEnvironmentMarker",
 CellID->434593514]
}, Open  ]],
Cell[1916135, 51718, 48, 1, 16, "CloseEnvironment",
 CellID->387312174],
Cell[CellGroupData[{
Cell[1916208, 51723, 17593, 390, 28, "ProofInfo",
 CellTags->{"Proof|ID:423392620", "Proof|ID:423392620-1"},
 CellID->117206802],
Cell[1933804, 52115, 478, 12, 70, "ProofInfoBody",
 CellID->6506205]
}, Closed]],
Cell[CellGroupData[{
Cell[1934319, 52132, 15598, 353, 26, "ProofInfo",
 CellTags->{"Proof|ID:1533257011", "Proof|ID:1533257011-1"},
 CellID->141690382],
Cell[1949920, 52487, 480, 12, 70, "ProofInfoBody",
 CellID->49123522]
}, Closed]],
Cell[CellGroupData[{
Cell[1950437, 52504, 356, 9, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1556779726],
Cell[1950796, 52515, 23479, 716, 1363, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1788850434],
Cell[1974278, 53233, 262, 5, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1397741689]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1974589, 53244, 197, 3, 22, "Subsubsection",
 CellID->1336194338],
Cell[1974789, 53249, 46, 1, 70, "OpenEnvironment",
 CellID->43048684],
Cell[CellGroupData[{
Cell[1974860, 53254, 432, 11, 70, "EnvironmentHeader",
 CellID->1822040650],
Cell[1975295, 53267, 857, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1154842295", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1154842295],
Cell[1976155, 53290, 949, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1427865659", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxBase"},
 CellID->1427865659],
Cell[1977107, 53316, 948, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:802696830", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec1"},
 CellID->802696830],
Cell[1978058, 53342, 947, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:668126868", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec2"},
 CellID->668126868],
Cell[1979008, 53368, 949, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1848124281", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec3"},
 CellID->1848124281],
Cell[1979960, 53394, 942, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:6143564", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAuxRec4"},
 CellID->6143564],
Cell[1980905, 53420, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1511341304]
}, Open  ]],
Cell[1980986, 53424, 49, 1, 70, "CloseEnvironment",
 CellID->1991967224],
Cell[CellGroupData[{
Cell[1981060, 53429, 5548, 139, 70, "ProofInfo",
 CellTags->{"Proof|ID:1427865659", "Proof|ID:1427865659-1"},
 CellID->544375535],
Cell[1986611, 53570, 481, 12, 70, "ProofInfoBody",
 CellID->473240177]
}, Closed]],
Cell[CellGroupData[{
Cell[1987129, 53587, 5539, 138, 70, "ProofInfo",
 CellTags->{"Proof|ID:802696830", "Proof|ID:802696830-1"},
 CellID->335517243],
Cell[1992671, 53727, 480, 12, 70, "ProofInfoBody",
 CellID->108528534]
}, Closed]],
Cell[CellGroupData[{
Cell[1993188, 53744, 5539, 138, 70, "ProofInfo",
 CellTags->{"Proof|ID:668126868", "Proof|ID:668126868-1"},
 CellID->705494055],
Cell[1998730, 53884, 481, 12, 70, "ProofInfoBody",
 CellID->1531879012]
}, Closed]],
Cell[CellGroupData[{
Cell[1999248, 53901, 5548, 139, 70, "ProofInfo",
 CellTags->{"Proof|ID:1848124281", "Proof|ID:1848124281-1"},
 CellID->885145762],
Cell[2004799, 54042, 482, 12, 70, "ProofInfoBody",
 CellID->1378460394]
}, Closed]],
Cell[CellGroupData[{
Cell[2005318, 54059, 5530, 138, 70, "ProofInfo",
 CellTags->{"Proof|ID:6143564", "Proof|ID:6143564-1"},
 CellID->1629480533],
Cell[2010851, 54199, 479, 12, 70, "ProofInfoBody",
 CellID->1662275265]
}, Closed]],
Cell[2011345, 54214, 48, 1, 70, "OpenEnvironment",
 CellID->1477099468],
Cell[CellGroupData[{
Cell[2011418, 54219, 393, 9, 70, "EnvironmentHeader",
 CellID->1770683854],
Cell[2011814, 54230, 1041, 30, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1620493315", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1620493315],
Cell[2012858, 54262, 831, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1672348970", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1672348970],
Cell[2013692, 54285, 912, 24, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1826259623", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1826259623],
Cell[2014607, 54311, 979, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1814594394", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 GBAux"},
 CellID->1814594394],
Cell[2015589, 54337, 743, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1024326782", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1024326782],
Cell[2016335, 54358, 1057, 30, 70, "GlobalDeclaration",
 CellTags->{
  "ID:585977074", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->585977074],
Cell[2017395, 54390, 2212, 65, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1959330209", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible2 update"},
 CellID->1959330209],
Cell[2019610, 54457, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1750384990]
}, Open  ]],
Cell[2019691, 54461, 48, 1, 70, "CloseEnvironment",
 CellID->885022220],
Cell[CellGroupData[{
Cell[2019764, 54466, 7826, 187, 70, "ProofInfo",
 CellTags->{"Proof|ID:1814594394", "Proof|ID:1814594394-1"},
 CellID->378539769],
Cell[2027593, 54655, 481, 12, 70, "ProofInfoBody",
 CellID->287249254]
}, Closed]],
Cell[CellGroupData[{
Cell[2028111, 54672, 18875, 415, 70, "ProofInfo",
 CellTags->{"Proof|ID:1959330209", "Proof|ID:1959330209-1"},
 CellID->1849443430],
Cell[2046989, 55089, 480, 12, 70, "ProofInfoBody",
 CellID->59470111]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2047518, 55107, 194, 3, 22, "Subsubsection",
 CellID->1419927671],
Cell[2047715, 55112, 47, 1, 70, "OpenEnvironment",
 CellID->843615990],
Cell[CellGroupData[{
Cell[2047787, 55117, 434, 11, 70, "EnvironmentHeader",
 CellID->2024040760],
Cell[2048224, 55130, 851, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:33155789", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->33155789],
Cell[2049078, 55153, 949, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1691022439", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxBase"},
 CellID->1691022439],
Cell[2050030, 55179, 946, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:544907973", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec1"},
 CellID->544907973],
Cell[2050979, 55205, 948, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1039424559", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec2"},
 CellID->1039424559],
Cell[2051930, 55231, 949, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1704484269", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec3"},
 CellID->1704484269],
Cell[2052882, 55257, 949, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2114764931", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAuxRec4"},
 CellID->2114764931],
Cell[2053834, 55283, 66, 1, 70, "EndEnvironmentMarker",
 CellID->2000303424]
}, Open  ]],
Cell[2053915, 55287, 48, 1, 70, "CloseEnvironment",
 CellID->187817550],
Cell[CellGroupData[{
Cell[2053988, 55292, 5549, 139, 70, "ProofInfo",
 CellTags->{"Proof|ID:1691022439", "Proof|ID:1691022439-1"},
 CellID->1214006285],
Cell[2059540, 55433, 482, 12, 70, "ProofInfoBody",
 CellID->1934036983]
}, Closed]],
Cell[CellGroupData[{
Cell[2060059, 55450, 5539, 138, 70, "ProofInfo",
 CellTags->{"Proof|ID:544907973", "Proof|ID:544907973-1"},
 CellID->708615290],
Cell[2065601, 55590, 523, 13, 70, "ProofInfoBody",
 CellID->2061640519]
}, Closed]],
Cell[CellGroupData[{
Cell[2066161, 55608, 5548, 139, 70, "ProofInfo",
 CellTags->{"Proof|ID:1039424559", "Proof|ID:1039424559-1"},
 CellID->495379081],
Cell[2071712, 55749, 481, 12, 70, "ProofInfoBody",
 CellID->725364984]
}, Closed]],
Cell[CellGroupData[{
Cell[2072230, 55766, 5548, 139, 70, "ProofInfo",
 CellTags->{"Proof|ID:1704484269", "Proof|ID:1704484269-1"},
 CellID->122315265],
Cell[2077781, 55907, 482, 12, 70, "ProofInfoBody",
 CellID->1471508628]
}, Closed]],
Cell[CellGroupData[{
Cell[2078300, 55924, 5549, 139, 70, "ProofInfo",
 CellTags->{"Proof|ID:2114764931", "Proof|ID:2114764931-1"},
 CellID->1252700938],
Cell[2083852, 56065, 482, 12, 70, "ProofInfoBody",
 CellID->1737567296]
}, Closed]],
Cell[2084349, 56080, 47, 1, 70, "OpenEnvironment",
 CellID->822190239],
Cell[CellGroupData[{
Cell[2084421, 56085, 393, 9, 70, "EnvironmentHeader",
 CellID->1425736080],
Cell[2084817, 56096, 1993, 56, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1693763802", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GBAuxCPConnectible3 GBAux"},
 CellID->1693763802],
Cell[2086813, 56154, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1297770576]
}, Open  ]],
Cell[2086894, 56158, 48, 1, 70, "CloseEnvironment",
 CellID->435073891],
Cell[CellGroupData[{
Cell[2086967, 56163, 7826, 187, 70, "ProofInfo",
 CellTags->{"Proof|ID:1693763802", "Proof|ID:1693763802-1"},
 CellID->874031660],
Cell[2094796, 56352, 482, 12, 70, "ProofInfoBody",
 CellID->1515571182]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[2095339, 56371, 238, 4, 30, "Subsection",
 CellID->994676795],
Cell[2095580, 56377, 47, 1, 12, "OpenEnvironment",
 CellID->819031807],
Cell[CellGroupData[{
Cell[2095652, 56382, 516, 16, 43, "EnvironmentHeader",
 CellID->948746204],
Cell[2096171, 56400, 946, 26, 32, "FormalTextInputFormula",
 CellTags->{
  "ID:439420249", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "GB empty", "IR"},
 CellID->439420249],
Cell[2097120, 56428, 65, 1, 26, "EndEnvironmentMarker",
 CellID->145231829]
}, Open  ]],
Cell[2097200, 56432, 49, 1, 16, "CloseEnvironment",
 CellID->2134187147],
Cell[CellGroupData[{
Cell[2097274, 56437, 2847, 84, 28, "ProofInfo",
 CellTags->{"Proof|ID:439420249", "Proof|ID:439420249-1"},
 CellID->1181943649],
Cell[2100124, 56523, 481, 12, 70, "ProofInfoBody",
 CellID->1443792220]
}, Closed]],
Cell[2100620, 56538, 47, 1, 12, "OpenEnvironment",
 CellID->470033776],
Cell[CellGroupData[{
Cell[2100692, 56543, 388, 9, 43, "EnvironmentHeader",
 CellID->1353328601],
Cell[2101083, 56554, 781, 20, 44, "GlobalDeclaration",
 CellTags->{
  "ID:298471155", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->298471155],
Cell[2101867, 56576, 747, 20, 43, "GlobalDeclaration",
 CellTags->{
  "ID:1615450915", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb"},
 CellID->1615450915],
Cell[2102617, 56598, 1118, 35, 44, "FormalTextInputFormula",
 CellTags->{
  "ID:209199628", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "shape GB"},
 CellID->209199628],
Cell[2103738, 56635, 868, 23, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:448006251", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "membership GB", "IR"},
 CellID->448006251],
Cell[2104609, 56660, 910, 26, 32, "FormalTextInputFormula",
 CellTags->{
  "ID:739795809", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "ideal GB", "IR"},
 CellID->739795809],
Cell[2105522, 56688, 832, 22, 32, "FormalTextInputFormula",
 CellTags->{
  "ID:1550516051", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings.nb", "isGroebnerBasis GB"},
 CellID->1550516051],
Cell[2106357, 56712, 65, 1, 26, "EndEnvironmentMarker",
 CellID->760997017]
}, Open  ]],
Cell[2106437, 56716, 48, 1, 16, "CloseEnvironment",
 CellID->669844270],
Cell[CellGroupData[{
Cell[2106510, 56721, 4471, 118, 28, "ProofInfo",
 CellTags->{"Proof|ID:209199628", "Proof|ID:209199628-1"},
 CellID->2008671786],
Cell[2110984, 56841, 544, 13, 70, "ProofInfoBody",
 CellID->2044012590]
}, Closed]],
Cell[CellGroupData[{
Cell[2111565, 56859, 3740, 100, 26, "ProofInfo",
 CellTags->{"Proof|ID:448006251", "Proof|ID:448006251-1"},
 CellID->2018395045],
Cell[2115308, 56961, 481, 12, 58, "ProofInfoBody",
 CellID->1640934080]
}, Closed]],
Cell[CellGroupData[{
Cell[2115826, 56978, 4257, 114, 26, "ProofInfo",
 CellTags->{"Proof|ID:739795809", "Proof|ID:739795809-1"},
 CellID->144088343],
Cell[2120086, 57094, 478, 12, 70, "ProofInfoBody",
 CellID->4322593]
}, Closed]],
Cell[CellGroupData[{
Cell[2120601, 57111, 3656, 99, 26, "ProofInfo",
 CellTags->{"Proof|ID:1550516051", "Proof|ID:1550516051-1"},
 CellID->856857959],
Cell[2124260, 57212, 481, 12, 70, "ProofInfoBody",
 CellID->359992752]
}, Closed]]
}, Closed]]
}, Closed]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
