(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 9.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[   1625047,      41207]
NotebookOptionsPosition[   1379998,      34878]
NotebookOutlinePosition[   1547480,      38900]
CellTagsIndexPosition[   1531721,      38521]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["Reduction Ring of Residue Classes", "Title",
 CellChangeTimes->{{3.645521668675948*^9, 3.6455216923880177`*^9}, {
  3.660906476007756*^9, 3.660906507317011*^9}},
 CellID->913222369],

Cell[TextData[{
 "Represented as Integer-Intervals from 0 to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "-", "1"}], TraditionalForm]]]
}], "Subtitle",
 CellChangeTimes->{{3.6584662843993745`*^9, 3.658466295444194*^9}, {
  3.659081130626007*^9, 3.659081137100018*^9}, {3.6593375024630885`*^9, 
  3.6593375037266903`*^9}, {3.6593496661175876`*^9, 3.659349667942791*^9}},
 CellID->53451034],

Cell[CellGroupData[{

Cell["Preliminaries", "Section",
 CellChangeTimes->{{3.645521699970169*^9, 3.645521702321048*^9}},
 CellID->1686608038],

Cell[CellGroupData[{

Cell["Description", "Subsection",
 CellChangeTimes->{{3.645521707736018*^9, 3.6455217086366*^9}, {
  3.6609063458089275`*^9, 3.660906347743331*^9}},
 CellID->586723348],

Cell[TextData[{
 "This notebook contains the definition of functor ",
 Cell[BoxData[
  FormBox[
   RowBox[{"ReductionIQR", "[", "n", "]"}], TraditionalForm]]],
 " that, for given positive integer ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 ", constructs the domain of integer residue classes modulo ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 ", represented as integers between 0 and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "-", "1"}], TraditionalForm]]],
 ", together with the various operations needed for being an algorithmic \
reduction ring. The proof that the domains thus obtained really form \
algorithmic reduction rings is included in section \
\[OpenCurlyDoubleQuote]Reduction-Integer-Quotient-Rings Constitute Reduction \
Rings\[CloseCurlyDoubleQuote]."
}], "Text",
 CellChangeTimes->{{3.660906356307746*^9, 3.660906462342132*^9}, {
  3.660906518595831*^9, 3.660906561495906*^9}},
 CellID->128068127]
}, Closed]],

Cell[CellGroupData[{

Cell["Theory Packages", "Subsection",
 CellChangeTimes->{{3.6554653930425663`*^9, 3.6554653958505716`*^9}},
 CellID->6629529],

Cell[BoxData[
 RowBox[{"$PackageDirectory", "=", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{
    "$RootDirectory", ",", "\"\<scratch\>\"", ",", "\"\<amaletzk\>\"", ",", 
     "\"\<Theorema Environment Data\>\""}], "}"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.666438783233775*^9, 3.666438821083324*^9}},
 CellID->498486718],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<LogicSets.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554654003901796`*^9, 3.655465404602187*^9}, {
   3.658466629372182*^9, 3.658466659419835*^9}, {3.6584701435395627`*^9, 
   3.6584701465035677`*^9}, 3.666440012888694*^9},
 CellID->66724946],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<AlgebraicStructures.m\>\""}], 
    "}"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.6554654003901796`*^9, 3.655465404602187*^9}, {
   3.658466629372182*^9, 3.658466659419835*^9}, 3.666440014048706*^9},
 CellID->47868298],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<NatInt.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554654003901796`*^9, 3.655465404602187*^9}, 
   3.6584666501992188`*^9, 3.666440015088595*^9},
 CellID->77326002],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<NatIntExtended.m\>\""}], "}"}], 
   "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.6554654003901796`*^9, 3.655465404602187*^9}, 
   3.6584666501992188`*^9, {3.6590815005036564`*^9, 3.659081501704859*^9}, 
   3.666440016008758*^9},
 CellID->21887977],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<Tuples.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554654003901796`*^9, 3.655465404602187*^9}, 
   3.6584666501992188`*^9, {3.658480578335864*^9, 3.6584805793186655`*^9}, 
   3.6664400168955927`*^9},
 CellID->43455998],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<ReductionRings.m\>\""}], "}"}], 
   "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.6554654003901796`*^9, 3.655465404602187*^9}, {
   3.658466629372182*^9, 3.658466659419835*^9}, {3.6584701435395627`*^9, 
   3.6584701465035677`*^9}, {3.658475045508382*^9, 3.658475048097986*^9}, 
   3.666440017736711*^9},
 CellID->83415409],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", "]"}]], "Input",
 CellChangeTimes->{{3.6554654003901796`*^9, 3.655465404602187*^9}},
 CellID->170511016],

Cell[BoxData[
 RowBox[{"MakeTheoryPackage", "[", "]"}]], "Input",
 CellChangeTimes->{{3.6554654090013947`*^9, 3.655465412714201*^9}},
 CellID->669825776],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DeleteCases", "[", 
  RowBox[{
   RowBox[{"Map", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Quiet", "[", 
       RowBox[{"Check", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"With", "[", 
           RowBox[{
            RowBox[{"{", 
             RowBox[{"s", "=", 
              RowBox[{"\"\<VAR$\>\"", "<>", "#", "<>", "\"\<$TM\>\""}]}], 
             "}"}], ",", 
            RowBox[{"Context", "[", "s", "]"}]}], "]"}], ";", "#"}], ",", 
         "Null"}], "]"}], "]"}], "&"}], ",", 
     RowBox[{"{", 
      RowBox[{
      "\"\<i\>\"", ",", "\"\<j\>\"", ",", "\"\<n\>\"", ",", "\"\<m\>\"", ",", 
       "\"\<P\>\"", ",", "\"\<x\>\"", ",", "\"\<y\>\"", ",", "\"\<z\>\"", 
       ",", "\"\<a\>\"", ",", "\"\<A\>\"", ",", "\"\<b\>\"", ",", 
       "\"\<c\>\""}], "}"}]}], "]"}], ",", "Null"}], "]"}]], "Input",
 CellChangeTimes->{{3.6592618519471025`*^9, 3.6592618631011224`*^9}, {
   3.659681175379635*^9, 3.659681185706853*^9}, 3.659683710620489*^9, 
   3.65968441725893*^9, 3.6597695015379353`*^9, {3.6597698034056654`*^9, 
   3.659769808569275*^9}, {3.6597702854721127`*^9, 3.659770290573322*^9}, {
   3.659848272310069*^9, 3.6598483467534*^9}, {3.6598515848824887`*^9, 
   3.6598516763142495`*^9}, {3.6598555136659927`*^9, 3.659855522090008*^9}, {
   3.6598559051642804`*^9, 3.659855919516306*^9}, {3.6602862205309095`*^9, 
   3.6602862222781124`*^9}, {3.660479643354128*^9, 3.660479646380533*^9}},
 CellID->20666029],

Cell[BoxData[
 RowBox[{"{", "}"}]], "Output",
 CellChangeTimes->{3.661584065007867*^9, 3.662288182811904*^9, 
  3.666175502707775*^9},
 CellID->627834352]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Theory Analysis", "Subsection",
 CellChangeTimes->{{3.640668830437483*^9, 3.640668834318727*^9}},
 CellID->1386333768],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"AnalyzeProofs", "[", "\"\<Theory\>\"", "]"}], "//", 
  "Timing"}]], "Input",
 CellChangeTimes->{{3.6406688560104437`*^9, 3.640668867513785*^9}, 
   3.64067664060292*^9, 3.644222019629964*^9, {3.6594304603925624`*^9, 
   3.6594304632941675`*^9}, {3.6594318836874695`*^9, 3.6594318856686726`*^9}},
 
 CellID->1039312110],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\AlgebraicStructures\\\" (\"\>", 
   "\[InvisibleSpace]", "1", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"15\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures\" (", 1, " of ", "15", 
   "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQVps/3+O882tHABy/jbY=
  "],
 CellID->43308995],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"22\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "22"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQbp672OO882tHAB+xjhU=
  "],
 CellID->773653712],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\ConsistencyChecker\\\" (\"\>", 
   "\[InvisibleSpace]", "2", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"15\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ConsistencyChecker\" (", 2, " of ", "15", 
   "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ9ui65HHe+bUjAB5ujfY=
  "],
 CellID->547289795],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\Fields\\\" (\"\>", "\[InvisibleSpace]", 
   "3", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"15\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Fields\" (", 3, " of ", "15", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQFuG67HHe+bUjABlUjUM=
  "],
 CellID->583205520],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"36\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "36"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQLuS47nHe+bUjABxGjaI=
  "],
 CellID->917617342],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\Functors\\\" (\"\>", 
   "\[InvisibleSpace]", "4", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"15\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Functors\" (", 4, " of ", "15", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQbrsW4Hne+bUjAB9rjf8=
  "],
 CellID->622225809],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"34\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "34"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQdgsJ8zzv/NoRABoBjUM=
  "],
 CellID->60147518],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\GroebnerRings\\\" (\"\>", 
   "\[InvisibleSpace]", "5", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"15\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\GroebnerRings\" (", 5, " of ", "15", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/rl6h+d559eOACRGjq8=
  "],
 CellID->346776519],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"165\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "165"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQdlU/4nne+bUjABtSjYM=
  "],
 CellID->563606824],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"165\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "165"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQPvQ4w+u882tHACI7jmE=
  "],
 CellID->306995116],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"165\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "165"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ5lbg8z7v/NoRABUXjI4=
  "],
 CellID->85378455],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"165\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "165"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQFrm1w/u882tHAB5xjfs=
  "],
 CellID->612641366],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\IntegerQuotientRings\\\" (\"\>", 
   "\[InvisibleSpace]", "6", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"15\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings\" (", 6, " of ", "15", 
   "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQfu35wPu882tHACIijmk=
  "],
 CellID->212043760],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"78\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "78"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQfu35wPu882tHACIijmk=
  "],
 CellID->287935756],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"78\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "78"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQPuH83vu882tHACE6jk8=
  "],
 CellID->39772154],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\Integers\\\" (\"\>", 
   "\[InvisibleSpace]", "7", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"15\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Integers\" (", 7, " of ", "15", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQlnT44X3e+bUjABvjjaY=
  "],
 CellID->571996574],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"85\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "85"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQzrJh9Dnv/NoRABiKjP0=
  "],
 CellID->308344464],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"85\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "85"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQviuY6XPe+bUjAB1lja0=
  "],
 CellID->200909420],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\LogicSets\\\" (\"\>", 
   "\[InvisibleSpace]", "8", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"15\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets\" (", 8, " of ", "15", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQTire7HPe+bUjAB33jd4=
  "],
 CellID->27659054],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"161\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "161"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQFsg74HPe+bUjABuSjZQ=
  "],
 CellID->765332773],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"161\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "161"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/rApwPe882tHACHTjkk=
  "],
 CellID->40385639],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"161\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "161"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQXlHwyfe882tHACGRjmE=
  "],
 CellID->248554666],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"161\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "161"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQfqM9y++882tHAB/Djgk=
  "],
 CellID->762742487],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\NatIntExtended\\\" (\"\>", 
   "\[InvisibleSpace]", "9", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"15\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended\" (", 9, " of ", "15", "):"],
  
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtpXf6Xfe+bUjABqxjW0=
  "],
 CellID->249282650],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"212\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "212"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ3iN1xO+882tHAB7IjfI=
  "],
 CellID->8175745],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"212\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "212"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQNi3U9D/v/NoRABlUjSg=
  "],
 CellID->185713539],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"212\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "212"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQTvJp9j/v/NoRABvVjYo=
  "],
 CellID->51957385],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"212\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "212"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQLtty0f+882tHACEhjlQ=
  "],
 CellID->111507338],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"212\"\>"}],
  SequenceForm["Reading proof ", 201, " of ", "212"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQVuo2Czjv/NoRABnFjT0=
  "],
 CellID->77483203],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\NatInt\\\" (\"\>", "\[InvisibleSpace]", 
   "10", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"15\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt\" (", 10, " of ", "15", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQFvgfHHDe+bUjAB0Pjbw=
  "],
 CellID->172403181],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"260\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "260"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQnvArLuC882tHACEujkI=
  "],
 CellID->565330808],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"260\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "260"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQvnL8VsB559eOACTRjs8=
  "],
 CellID->1606675],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"260\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "260"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQfjU1JPC882tHACEEji4=
  "],
 CellID->20301363],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"260\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "260"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQXpS3JfC882tHAB/zjh8=
  "],
 CellID->203974489],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"260\"\>"}],
  SequenceForm["Reading proof ", 201, " of ", "260"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQLvLmCjrv/NoRABmHjSM=
  "],
 CellID->195794873],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"260\"\>"}],
  SequenceForm["Reading proof ", 251, " of ", "260"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ3rTbIOi882tHAB97jfk=
  "],
 CellID->23265144],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\Polynomials\\\" (\"\>", 
   "\[InvisibleSpace]", "11", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"15\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Polynomials\" (", 11, " of ", "15", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/mtmE3Te+bUjAB54jcs=
  "],
 CellID->73579430],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"343\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "343"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQFjHyCDrv/NoRABdcjOo=
  "],
 CellID->625647329],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"343\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "343"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQdqs5HXTe+bUjAB4Ejek=
  "],
 CellID->380032735],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"343\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "343"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQnnM4NPi882tHAB/mjhE=
  "],
 CellID->59964078],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"343\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "343"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQDul6GHze+bUjAB9fjhw=
  "],
 CellID->173396994],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"343\"\>"}],
  SequenceForm["Reading proof ", 201, " of ", "343"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQPnM9JuS882tHACIhjl0=
  "],
 CellID->189331182],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"343\"\>"}],
  SequenceForm["Reading proof ", 251, " of ", "343"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQdpF5F3Le+bUjABwwjaw=
  "],
 CellID->29652246],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"343\"\>"}],
  SequenceForm["Reading proof ", 301, " of ", "343"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ3vYiO/S882tHACJHjmg=
  "],
 CellID->54261457],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\PolyTuples\\\" (\"\>", 
   "\[InvisibleSpace]", "12", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"15\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\PolyTuples\" (", 12, " of ", "15", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQnmF/IfS882tHAB8WjgY=
  "],
 CellID->2598487],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"161\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "161"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQlrC+HXre+bUjABs8jY0=
  "],
 CellID->196075428],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"161\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "161"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ7hJwDjvv/NoRABoUjT0=
  "],
 CellID->524371889],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"161\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "161"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQVktZH3be+bUjABvIjZk=
  "],
 CellID->117993319],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"161\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "161"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQFt7LF37e+bUjABnejT8=
  "],
 CellID->320762704],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\ReductionRings\\\" (\"\>", 
   "\[InvisibleSpace]", "13", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"15\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings\" (", 13, " of ", "15", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQZt+kGX7e+bUjABnTjUM=
  "],
 CellID->355680031],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"250\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "250"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQbltrEn7e+bUjAB3qjcg=
  "],
 CellID->487316616],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"250\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "250"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQPvJzU/h559eOACTijtA=
  "],
 CellID->140301994],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"250\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "250"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQFjlgH3He+bUjABsmjXU=
  "],
 CellID->166787746],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"250\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "250"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQjmU9H3He+bUjABuxjZM=
  "],
 CellID->283202512],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"250\"\>"}],
  SequenceForm["Reading proof ", 201, " of ", "250"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuI75oded75tSMAGuaNaQ==
  "],
 CellID->285961962],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\Sequences\\\" (\"\>", 
   "\[InvisibleSpace]", "14", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"15\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Sequences\" (", 14, " of ", "15", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQlmDliDrv/NoRABTpjIk=
  "],
 CellID->15104475],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"48\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "48"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ5mXmjTrv/NoRABShjIE=
  "],
 CellID->323998079],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\Tuples\\\" (\"\>", "\[InvisibleSpace]", 
   "15", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"15\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples\" (", 15, " of ", "15", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQXvcgIeq882tHACGmjlI=
  "],
 CellID->370342082],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"218\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "218"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQnnEnK+q882tHACEWjkI=
  "],
 CellID->351985089],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"218\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "218"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ7tq/Meq882tHACGFjl4=
  "],
 CellID->391305160],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"218\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "218"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ9lT6GXXe+bUjABzijcg=
  "],
 CellID->68475953],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"218\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "218"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQLmOzjT7v/NoRABkjjR4=
  "],
 CellID->26466573],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"218\"\>"}],
  SequenceForm["Reading proof ", 201, " of ", "218"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQLshqjD7v/NoRAB1HjcA=
  "],
 CellID->1984322],

Cell[BoxData["\<\"Creating forward rules:\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQnnF/dvR559eOACJWjnc=
  "],
 CellID->604141779],

Cell[BoxData["\<\"Preparing ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/np3fvR559eOACVIjtY=
  "],
 CellID->915312300],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQnnFtY/R559eOACKbjoQ=
  "],
 CellID->281663586],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 51, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQfnHgWfR559eOACW/jvM=
  "],
 CellID->209327295],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 101, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQXrRaKua882tHAB45jc0=
  "],
 CellID->18937374],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 151, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQjp7mF3Pe+bUjABymjaU=
  "],
 CellID->153557070],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 201, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQfsDQHHPe+bUjAB3yjck=
  "],
 CellID->45465621],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 251, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQPvFqR8x559eOACTWjtA=
  "],
 CellID->39313665],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 301, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQbrr6Jua882tHACNLjqk=
  "],
 CellID->105735773],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 351, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtj6gEHve+bUjABu9jYI=
  "],
 CellID->77314492],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "401", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 401, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQjl0dHHve+bUjAB1sjcI=
  "],
 CellID->185751704],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "451", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 451, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQrp/WFnve+bUjAB8bjgI=
  "],
 CellID->85116431],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "501", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 501, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQXtS4M/a882tHACDSjkM=
  "],
 CellID->639108732],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "551", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 551, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQPpLzJva882tHACKBjoM=
  "],
 CellID->721277251],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "601", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 601, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ3nhdPu6882tHACANjg8=
  "],
 CellID->3339797],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "651", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 651, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQnu8UHHfe+bUjAByijZw=
  "],
 CellID->289009105],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "701", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 701, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQDlrTGXfe+bUjAB5kje8=
  "],
 CellID->199179739],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "751", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 751, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQXjJtX9x559eOACGYjmA=
  "],
 CellID->487360478],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "801", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 801, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ3vLvR9x559eOACZMjxI=
  "],
 CellID->84922310],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "851", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 851, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQLs7Sij/v/NoRABtpjXA=
  "],
 CellID->558497404],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "901", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 901, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQPhISH3/e+bUjAB6VjeA=
  "],
 CellID->255445829],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "951", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 951, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ9tg/Of6882tHAB7ajgM=
  "],
 CellID->696097528],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1001", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1001, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQTtE8GX/e+bUjABzkjb8=
  "],
 CellID->111397814],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1051", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1051, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQbpr8L/6882tHACH4jnw=
  "],
 CellID->766053093],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1101, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQZvtnlHDe+bUjABxCjaA=
  "],
 CellID->80891635],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1151, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuI/yI94bzza0cAH2aODw==
  "],
 CellID->242035726],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1201, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQLg6am3De+bUjAB14jcw=
  "],
 CellID->459982240],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1251, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/rHnYsJ559eOACW+ju8=
  "],
 CellID->52879335],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1301, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ9ljOlnje+bUjABrujWA=
  "],
 CellID->514239187],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1351, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQzp5kmXje+bUjAByljaE=
  "],
 CellID->478929526],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1401", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1401, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQjvibk3je+bUjACAsjiw=
  "],
 CellID->142907266],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1451", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1451, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQFnqxIPG882tHAB6hjgU=
  "],
 CellID->324881021],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1501", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1501, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQnhZ0NfG882tHAB/ljig=
  "],
 CellID->211471681],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1551", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1551, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQDrDlTDrv/NoRABhfjQI=
  "],
 CellID->154205215],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1601", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1601, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQzl1ul3Te+bUjAB1rjb4=
  "],
 CellID->795327148],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1651", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1651, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ3juxOOm882tHACCPji0=
  "],
 CellID->125127396],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1701", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1701, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQXlK9Mum882tHACBxjjQ=
  "],
 CellID->112536954],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1751", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1751, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/pZ6L+m882tHACOljqU=
  "],
 CellID->479851726],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1801", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1801, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQrr4glHze+bUjAB33jco=
  "],
 CellID->159558757],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1851", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1851, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQPrPLPfm882tHACEjjjo=
  "],
 CellID->18724986],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1901", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1901, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQfqFSm3ze+bUjAB8tjfY=
  "],
 CellID->55389970],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1951", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 1951, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQXsi/Mfm882tHAB2ajc4=
  "],
 CellID->185394026],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2001", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2001, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ7qt6knze+bUjACEhjlk=
  "],
 CellID->113003008],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2051", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2051, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQrn0qnnLe+bUjAB69jec=
  "],
 CellID->19983785],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2101, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQzgrwTjnv/NoRABtKjXM=
  "],
 CellID->741298015],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2151, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ7rGuSznv/NoRABz5jbM=
  "],
 CellID->168981775],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2201, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQdlPblHLe+bUjABtujYw=
  "],
 CellID->69169284],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2251, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQnjPxScp559eOACI3jn8=
  "],
 CellID->336736819],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2301, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ3lcjnnre+bUjAB3rjcA=
  "],
 CellID->291445319],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2351, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQFn3hmXre+bUjABzDjbU=
  "],
 CellID->296881490],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2401", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2401, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQzgmuST3v/NoRAByajao=
  "],
 CellID->860092272],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2451", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2451, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQPrRvXep559eOACNjjp0=
  "],
 CellID->830634336],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2501", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2501, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ3qD5KPW882tHAB/4jio=
  "],
 CellID->481433850],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2551", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2551, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQviQimnbe+bUjABusjWs=
  "],
 CellID->247307259],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2601", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2601, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQPlDvkXbe+bUjAB87jfc=
  "],
 CellID->133525692],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2651", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2651, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQrsyqSTvv/NoRAB2ojc8=
  "],
 CellID->431298360],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2701", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2701, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQNgrdkHbe+bUjABwVjac=
  "],
 CellID->268266100],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2751", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2751, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQDnF4nHbe+bUjAB3Ejec=
  "],
 CellID->58630601],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2801", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2801, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQnr9SKv2882tHAB5KjdM=
  "],
 CellID->265247545],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2851", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2851, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQjpjil37e+bUjABy3jas=
  "],
 CellID->58725565],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2901", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2901, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQXvy3Nf2882tHACM4jpY=
  "],
 CellID->81617263],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2951", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 2951, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQbkvbl37e+bUjAB+Fjhs=
  "],
 CellID->56313838],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "3001", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 3001, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQXnThc/p559eOACSJjtY=
  "],
 CellID->235693051],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "3051", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 3051, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQjtqtnnHe+bUjABzzja4=
  "],
 CellID->33877423],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "3101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 3101, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQdl4am3He+bUjABzljbc=
  "],
 CellID->198408801],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "3151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 3151, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ/tM/MeO882tHACNLjo4=
  "],
 CellID->407586333],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "3201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 3201, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQbvx1NOO882tHACOYjrI=
  "],
 CellID->21003792],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "3251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 3251, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQvvTkV8Z559eOACbEjyI=
  "],
 CellID->511717689],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "3301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 3301, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ7jmvl3ne+bUjAB8+jfw=
  "],
 CellID->374982656],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "3351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3371"}],
  SequenceForm["Creating rule ", 3351, " of ", 3371],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQdtmVlHne+bUjAB2jjdM=
  "],
 CellID->265656759],

Cell[BoxData["\<\"Creating thinned data ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQrjCsyDzv/NoRABwIjZQ=
  "],
 CellID->224441463],

Cell[BoxData["\<\"Done\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGBQB+KTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQ5tpQnXne+bUjABwjjag=
  "],
 CellID->209939310]
}, Open  ]]
}, Closed]],

Cell[BoxData[
 RowBox[{
  RowBox[{"ComputeFormulaPositions", "[", "]"}], ";"}]], "Input",
 CellChangeTimes->{{3.6406695270485573`*^9, 3.6406695415951242`*^9}, {
   3.64067664934828*^9, 3.640676663567005*^9}, 3.6412902845300865`*^9, 
   3.641290346822996*^9, {3.6442219079512787`*^9, 3.644221909273131*^9}},
 CellID->19428626],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CheckIntegrity", "[", 
  RowBox[{"\"\<CheckPositions\>\"", "\[Rule]", 
   RowBox[{"{", 
    RowBox[{"FileNameJoin", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"CurrentValue", "[", "\"\<NotebookDirectory\>\"", "]"}], ",", 
       RowBox[{"CurrentValue", "[", "\"\<NotebookFileName\>\"", "]"}]}], 
      "}"}], "]"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.64067003109768*^9, 3.6406700374311132`*^9}, {
   3.6406797903560762`*^9, 3.640679792950843*^9}, 3.6442220291527653`*^9},
 CellID->1909803754],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{
  3.640670040858255*^9, 3.640670099679185*^9, 3.640670397721546*^9, 
   3.640679793300535*^9, 3.6406798406436234`*^9, 3.640679891162312*^9, 
   3.640679971009297*^9, 3.64068009182163*^9, 3.640680173340745*^9, 
   3.641037395456129*^9, 3.641116403388618*^9, 3.641286881014108*^9, 
   3.642225466578094*^9, 3.642234672070362*^9, 3.642240151894621*^9, 
   3.6423063269273643`*^9, {3.642308832539488*^9, 3.642308857563932*^9}, 
   3.642321396697977*^9, 3.6423285030212173`*^9, 3.642493582768566*^9, 
   3.642763270724422*^9, 3.642834604308413*^9, 3.643428832901271*^9, 
   3.6434441292624083`*^9, 3.643532851309215*^9, 3.644918543299676*^9, 
   3.645518304732339*^9},
 CellID->636966437]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CheckConsistency", "[", 
  "\"\<isAlgoReductionRing \[ScriptCapitalZ]n\>\"", "]"}]], "Input",
 CellChangeTimes->{{3.6594307492592697`*^9, 3.6594307655300984`*^9}},
 CellID->97543220],

Cell[BoxData[
 TabViewBox[{{1,"\<\"outdated (11)\"\>"->
  TemplateBox[{TagBox[
     GridBox[{{
        TemplateBox[{
          TagBox[
           GridBox[{{
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["R2' \[ScriptCapitalZ]n", 0]], Evaluator -> 
                 Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
                "\"\[NumberSign]0\"", TooltipStyle -> "TextStyling"], 
               Annotation[#, "\[NumberSign]0", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}]}], 
                    "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                    StyleBox["c", "ExpressionVariable"]}], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], 
                    "]"}]]}]}]}], $CellContext`TheoremaForm], DisplayForm]}, {
             
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["R2' \[ScriptCapitalZ]n", 1]], Evaluator -> 
                 Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
                "\"\[NumberSign]1\"", TooltipStyle -> "TextStyling"], 
               Annotation[#, "\[NumberSign]1", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]]}], 
                    "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["mult", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}], 
                    "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                    StyleBox["c", "ExpressionVariable"]}], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], 
                    "]"}]]}]}]}]}], $CellContext`TheoremaForm], 
               DisplayForm]}}, AutoDelete -> False, 
            GridBoxAlignment -> {"Columns" -> {Center, Left}}, 
            GridBoxFrame -> {
             "ColumnsIndexed" -> {{{1, -1}, {1, -1}} -> True}}, 
            GridBoxItemSize -> {
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"], 
          TagBox[
           ButtonBox[
            
            PaneSelectorBox[{
             False -> "\"(R2' \[ScriptCapitalZ]n)\"", True -> 
              StyleBox["\"(R2' \[ScriptCapitalZ]n)\"", "HyperlinkActive"]}, 
             Dynamic[
              CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
             FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
            "Hyperlink", 
            ButtonData :> {
             "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:50258269"}, 
            ButtonNote -> 
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb"], 
           Annotation[#, {
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:50258269"}, 
            "Hyperlink"]& ]}, "Labeled", DisplayFunction -> (GridBox[{{
             ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
             TagBox[
              ItemBox[
               PaneBox[
                TagBox[#, "SkipImageSizeLevel"], 
                Alignment -> {Center, Baseline}, BaselinePosition -> 
                Baseline], DefaultBaseStyle -> "Labeled"], 
              "SkipImageSizeLevel"]}}, 
           GridBoxAlignment -> {
            "Columns" -> {{Center}}, "Rows" -> {{Center}}}, AutoDelete -> 
           False, GridBoxItemSize -> {
            "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
           BaselinePosition -> {2, 1}]& ), 
         InterpretationFunction -> (RowBox[{"Labeled", "[", 
            RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {
        TemplateBox[{
          TagBox[
           GridBox[{{
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["DomainSequences imgSeq,[IR]", 2]], Evaluator -> 
                 Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
                "\"\[NumberSign]2\"", TooltipStyle -> "TextStyling"], 
               Annotation[#, "\[NumberSign]2", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainSequences", "[", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}, {
                    StyleBox["g", "ExpressionVariable"]}, {
                    StyleBox["\[ScriptCapitalD]1", "ExpressionVariable"]}}]], 
                  
                  RowBox[{
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]1", "ExpressionVariable"]], 
                    "[", 
                    RowBox[{
                    StyleBox["g", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], "]"}]}], 
                    "\[Implies]", 
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainSequences", "[", 
                    StyleBox["\[ScriptCapitalD]1", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{"imgSeq", "[", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], ",", 
                    StyleBox["g", "ExpressionVariable"]}], "]"}], 
                    "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]}, {
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["DomainSequences imgSeq,[IR]", 3]], Evaluator -> 
                 Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
                "\"\[NumberSign]3\"", TooltipStyle -> "TextStyling"], 
               Annotation[#, "\[NumberSign]3", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   RowBox[{
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                    
                    StyleBox["f", "ExpressionVariable"], ",", 
                    StyleBox["g", "ExpressionVariable"]}]], 
                  RowBox[{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainSequences", "[", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], "\[Implies]", 
                    
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    StyleBox["\[ScriptCapitalD]1", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]1", "ExpressionVariable"]], 
                    "[", 
                    RowBox[{
                    StyleBox["g", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], "]"}]}], 
                    "\[Implies]", 
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainSequences", "[", 
                    StyleBox["\[ScriptCapitalD]1", "ExpressionVariable"], 
                    "]"}]], "[", 
                    RowBox[{"imgSeq", "[", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], ",", 
                    StyleBox["g", "ExpressionVariable"]}], "]"}], 
                    "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]}},
             AutoDelete -> False, 
            GridBoxAlignment -> {"Columns" -> {Center, Left}}, 
            GridBoxFrame -> {
             "ColumnsIndexed" -> {{{1, -1}, {1, -1}} -> True}}, 
            GridBoxItemSize -> {
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"], 
          TagBox[
           ButtonBox[
            
            PaneSelectorBox[{
             False -> "\"(DomainSequences imgSeq,[IR])\"", True -> 
              StyleBox[
               "\"(DomainSequences imgSeq,[IR])\"", "HyperlinkActive"]}, 
             Dynamic[
              CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
             FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
            "Hyperlink", 
            ButtonData :> {
             "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Sequences.nb", "ID:814769285"}, ButtonNote -> 
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Sequences.nb"], 
           Annotation[#, {
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Sequences.nb", "ID:814769285"}, 
            "Hyperlink"]& ]}, "Labeled", DisplayFunction -> (GridBox[{{
             ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
             TagBox[
              ItemBox[
               PaneBox[
                TagBox[#, "SkipImageSizeLevel"], 
                Alignment -> {Center, Baseline}, BaselinePosition -> 
                Baseline], DefaultBaseStyle -> "Labeled"], 
              "SkipImageSizeLevel"]}}, 
           GridBoxAlignment -> {
            "Columns" -> {{Center}}, "Rows" -> {{Center}}}, AutoDelete -> 
           False, GridBoxItemSize -> {
            "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
           BaselinePosition -> {2, 1}]& ), 
         InterpretationFunction -> (RowBox[{"Labeled", "[", 
            RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {
        TemplateBox[{
          TagBox[
           GridBox[{{
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["gcd unique \[DoubleStruckCapitalZ]", 4]], 
                 Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
                 "Preemptive"], "\"\[NumberSign]4\"", TooltipStyle -> 
                "TextStyling"], Annotation[#, "\[NumberSign]4", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}]}, {
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}]}, {
                    RowBox[{
                    StyleBox["g1", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
                  RowBox[{
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[NotEqual]", "0"}],
                     "\[Or]", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                    "0"}]}], "\[And]", 
                    RowBox[{"isGCD", "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["g1", "ExpressionVariable"]}], "]"}]}], 
                    "\[Implies]", 
                    RowBox[{
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    
                    StyleBox[
                    "g1", "ExpressionVariable"]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]}, {
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["gcd unique \[DoubleStruckCapitalZ]", 5]], 
                 Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
                 "Preemptive"], "\"\[NumberSign]5\"", TooltipStyle -> 
                "TextStyling"], Annotation[#, "\[NumberSign]5", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}]}, {
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}]}}]], 
                  RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[NotEqual]", "0"}],
                     "\[Or]", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                    "0"}]}], "\[Implies]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["d1", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
                    RowBox[{
                    RowBox[{"isGCD", "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["d1", "ExpressionVariable"]}], "]"}], 
                    "\[Implies]", 
                    RowBox[{
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    
                    StyleBox[
                    "d1", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]}}, AutoDelete -> False, 
            GridBoxAlignment -> {"Columns" -> {Center, Left}}, 
            GridBoxFrame -> {
             "ColumnsIndexed" -> {{{1, -1}, {1, -1}} -> True}}, 
            GridBoxItemSize -> {
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"], 
          TagBox[
           ButtonBox[
            
            PaneSelectorBox[{
             False -> "\"(gcd unique \[DoubleStruckCapitalZ])\"", True -> 
              StyleBox[
               "\"(gcd unique \[DoubleStruckCapitalZ])\"", 
                "HyperlinkActive"]}, 
             Dynamic[
              CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
             FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
            "Hyperlink", 
            ButtonData :> {
             "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:460724241"}, 
            ButtonNote -> 
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb"], 
           Annotation[#, {
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:460724241"}, 
            "Hyperlink"]& ]}, "Labeled", DisplayFunction -> (GridBox[{{
             ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
             TagBox[
              ItemBox[
               PaneBox[
                TagBox[#, "SkipImageSizeLevel"], 
                Alignment -> {Center, Baseline}, BaselinePosition -> 
                Baseline], DefaultBaseStyle -> "Labeled"], 
              "SkipImageSizeLevel"]}}, 
           GridBoxAlignment -> {
            "Columns" -> {{Center}}, "Rows" -> {{Center}}}, AutoDelete -> 
           False, GridBoxItemSize -> {
            "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
           BaselinePosition -> {2, 1}]& ), 
         InterpretationFunction -> (RowBox[{"Labeled", "[", 
            RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {
        TemplateBox[{
          TagBox[
           GridBox[{{
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["abstraction term isSet domain,[IR]", 6]], 
                 Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
                 "Preemptive"], "\"\[NumberSign]6\"", TooltipStyle -> 
                "TextStyling"], Annotation[#, "\[NumberSign]6", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{"isDomain", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]}, {
                    StyleBox["f", "ExpressionVariable"]}, {
                    StyleBox["P", "ExpressionVariable"]}}]], 
                  RowBox[{"isSet", "[", 
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}], "}"}], 
                    "]"}]}], $CellContext`TheoremaForm], DisplayForm]}, {
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["abstraction term isSet domain,[IR]", 7]], 
                 Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
                 "Preemptive"], "\"\[NumberSign]7\"", TooltipStyle -> 
                "TextStyling"], Annotation[#, "\[NumberSign]7", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   StyleBox["D", "ExpressionVariable"]], 
                  RowBox[{
                    RowBox[{"isDomain", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
                    
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"]}]], 
                    RowBox[{"isSet", "[", 
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}], "}"}], 
                    "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]}}, 
            AutoDelete -> False, 
            GridBoxAlignment -> {"Columns" -> {Center, Left}}, 
            GridBoxFrame -> {
             "ColumnsIndexed" -> {{{1, -1}, {1, -1}} -> True}}, 
            GridBoxItemSize -> {
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"], 
          TagBox[
           ButtonBox[
            
            PaneSelectorBox[{
             False -> "\"(abstraction term isSet domain,[IR])\"", True -> 
              StyleBox[
               "\"(abstraction term isSet domain,[IR])\"", 
                "HyperlinkActive"]}, 
             Dynamic[
              CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
             FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
            "Hyperlink", 
            ButtonData :> {
             "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:158487732"}, ButtonNote -> 
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
           Annotation[#, {
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:158487732"}, 
            "Hyperlink"]& ]}, "Labeled", DisplayFunction -> (GridBox[{{
             ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
             TagBox[
              ItemBox[
               PaneBox[
                TagBox[#, "SkipImageSizeLevel"], 
                Alignment -> {Center, Baseline}, BaselinePosition -> 
                Baseline], DefaultBaseStyle -> "Labeled"], 
              "SkipImageSizeLevel"]}}, 
           GridBoxAlignment -> {
            "Columns" -> {{Center}}, "Rows" -> {{Center}}}, AutoDelete -> 
           False, GridBoxItemSize -> {
            "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
           BaselinePosition -> {2, 1}]& ), 
         InterpretationFunction -> (RowBox[{"Labeled", "[", 
            RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {
        TemplateBox[{
          TagBox[
           GridBox[{{
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["abstraction term \[Element] domain", 8]], 
                 Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
                 "Preemptive"], "\"\[NumberSign]8\"", TooltipStyle -> 
                "TextStyling"], Annotation[#, "\[NumberSign]8", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{"isDomain", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]}, {
                    StyleBox["f", "ExpressionVariable"]}, {
                    StyleBox["P", "ExpressionVariable"]}, {
                    StyleBox["x", "ExpressionVariable"]}}]], 
                  RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}], "}"}]}], 
                    ")"}], "\[DoubleLeftRightArrow]", 
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}], "\[And]", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], 
                    "]"}]}]}]}]}]}], $CellContext`TheoremaForm], 
               DisplayForm]}, {
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["abstraction term \[Element] domain", 9]], 
                 Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
                 "Preemptive"], "\"\[NumberSign]9\"", TooltipStyle -> 
                "TextStyling"], Annotation[#, "\[NumberSign]9", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   StyleBox["D", "ExpressionVariable"]], 
                  RowBox[{
                    RowBox[{"isDomain", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}], "\[Implies]", 
                    
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], ",", 
                    StyleBox["P", "ExpressionVariable"]}]], 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    StyleBox["x", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}], "}"}]}], 
                    TagBox[
                    
                    RowBox[{":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                    "a:=b"], 
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}], "\[And]", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
               DisplayForm]}}, AutoDelete -> False, 
            GridBoxAlignment -> {"Columns" -> {Center, Left}}, 
            GridBoxFrame -> {
             "ColumnsIndexed" -> {{{1, -1}, {1, -1}} -> True}}, 
            GridBoxItemSize -> {
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"], 
          TagBox[
           ButtonBox[
            
            PaneSelectorBox[{
             False -> "\"(abstraction term \[Element] domain)\"", True -> 
              StyleBox[
               "\"(abstraction term \[Element] domain)\"", 
                "HyperlinkActive"]}, 
             Dynamic[
              CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
             FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
            "Hyperlink", 
            ButtonData :> {
             "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:704307368"}, ButtonNote -> 
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
           Annotation[#, {
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:704307368"}, 
            "Hyperlink"]& ]}, "Labeled", DisplayFunction -> (GridBox[{{
             ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
             TagBox[
              ItemBox[
               PaneBox[
                TagBox[#, "SkipImageSizeLevel"], 
                Alignment -> {Center, Baseline}, BaselinePosition -> 
                Baseline], DefaultBaseStyle -> "Labeled"], 
              "SkipImageSizeLevel"]}}, 
           GridBoxAlignment -> {
            "Columns" -> {{Center}}, "Rows" -> {{Center}}}, AutoDelete -> 
           False, GridBoxItemSize -> {
            "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
           BaselinePosition -> {2, 1}]& ), 
         InterpretationFunction -> (RowBox[{"Labeled", "[", 
            RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {
        TemplateBox[{
          TagBox[
           GridBox[{{
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket[
                   "> isWellFounded \[DoubleStruckCapitalZ]+", 10]], 
                 Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
                 "Preemptive"], "\"\[NumberSign]10\"", TooltipStyle -> 
                "TextStyling"], 
               Annotation[#, "\[NumberSign]10", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   StyleBox["a", "ExpressionVariable"]], 
                  RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}], "\[Implies]", 
                    RowBox[{"isWellFounded", "[", 
                    RowBox[{">", ",", 
                    SubscriptBox["\[DoubleStruckCapitalZ]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", "\[Ellipsis]", 
                    ",", "\[Infinity]"}]]}], 
                    "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]}, {
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket[
                   "> isWellFounded \[DoubleStruckCapitalZ]+", 11]], 
                 Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
                 "Preemptive"], "\"\[NumberSign]11\"", TooltipStyle -> 
                "TextStyling"], 
               Annotation[#, "\[NumberSign]11", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   StyleBox["a", "ExpressionVariable"]], 
                  RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}], "\[Implies]", 
                    RowBox[{"isWellFounded", "[", 
                    RowBox[{">", ",", 
                    SubscriptBox["\[DoubleStruckCapitalZ]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", "\[Ellipsis]", 
                    ",", "\[Infinity]"}]]}], 
                    "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]}}, 
            AutoDelete -> False, 
            GridBoxAlignment -> {"Columns" -> {Center, Left}}, 
            GridBoxFrame -> {
             "ColumnsIndexed" -> {{{1, -1}, {1, -1}} -> True}}, 
            GridBoxItemSize -> {
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"], 
          TagBox[
           ButtonBox[
            
            PaneSelectorBox[{
             False -> "\"(> isWellFounded \[DoubleStruckCapitalZ]+)\"", True -> 
              StyleBox[
               "\"(> isWellFounded \[DoubleStruckCapitalZ]+)\"", 
                "HyperlinkActive"]}, 
             Dynamic[
              CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
             FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
            "Hyperlink", 
            ButtonData :> {
             "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:329307398"}, ButtonNote -> 
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
           Annotation[#, {
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:329307398"}, 
            "Hyperlink"]& ]}, "Labeled", DisplayFunction -> (GridBox[{{
             ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
             TagBox[
              ItemBox[
               PaneBox[
                TagBox[#, "SkipImageSizeLevel"], 
                Alignment -> {Center, Baseline}, BaselinePosition -> 
                Baseline], DefaultBaseStyle -> "Labeled"], 
              "SkipImageSizeLevel"]}}, 
           GridBoxAlignment -> {
            "Columns" -> {{Center}}, "Rows" -> {{Center}}}, AutoDelete -> 
           False, GridBoxItemSize -> {
            "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
           BaselinePosition -> {2, 1}]& ), 
         InterpretationFunction -> (RowBox[{"Labeled", "[", 
            RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {
        TemplateBox[{
          TagBox[
           GridBox[{{
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["mult1 alternative \[ScriptCapitalZ]n", 12]], 
                 Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
                 "Preemptive"], "\"\[NumberSign]12\"", TooltipStyle -> 
                "TextStyling"], 
               Annotation[#, "\[NumberSign]12", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    RowBox[{
                    UnderscriptBox["mult", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}], 
                    "\[DoubleLeftRightArrow]", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ">", "0"}], "\[And]", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ">", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], " ", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}]}]}], 
                    ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]}, {
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["mult1 alternative \[ScriptCapitalZ]n", 13]], 
                 Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
                 "Preemptive"], "\"\[NumberSign]13\"", TooltipStyle -> 
                "TextStyling"], 
               Annotation[#, "\[NumberSign]13", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]]}], 
                    "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["mult", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}], 
                    "\[DoubleLeftRightArrow]", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ">", "0"}], "\[And]", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ">", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], " ", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}]}]}], 
                    ")"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]}}, 
            AutoDelete -> False, 
            GridBoxAlignment -> {"Columns" -> {Center, Left}}, 
            GridBoxFrame -> {
             "ColumnsIndexed" -> {{{1, -1}, {1, -1}} -> True}}, 
            GridBoxItemSize -> {
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"], 
          TagBox[
           ButtonBox[
            
            PaneSelectorBox[{
             False -> "\"(mult1 alternative \[ScriptCapitalZ]n)\"", True -> 
              StyleBox[
               "\"(mult1 alternative \[ScriptCapitalZ]n)\"", 
                "HyperlinkActive"]}, 
             Dynamic[
              CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
             FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
            "Hyperlink", 
            ButtonData :> {
             "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:57125150"}, 
            ButtonNote -> 
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb"], 
           Annotation[#, {
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:57125150"}, 
            "Hyperlink"]& ]}, "Labeled", DisplayFunction -> (GridBox[{{
             ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
             TagBox[
              ItemBox[
               PaneBox[
                TagBox[#, "SkipImageSizeLevel"], 
                Alignment -> {Center, Baseline}, BaselinePosition -> 
                Baseline], DefaultBaseStyle -> "Labeled"], 
              "SkipImageSizeLevel"]}}, 
           GridBoxAlignment -> {
            "Columns" -> {{Center}}, "Rows" -> {{Center}}}, AutoDelete -> 
           False, GridBoxItemSize -> {
            "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
           BaselinePosition -> {2, 1}]& ), 
         InterpretationFunction -> (RowBox[{"Labeled", "[", 
            RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {
        TemplateBox[{
          TagBox[
           GridBox[{{
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["set equality alternative", 14]], Evaluator -> 
                 Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
                "\"\[NumberSign]14\"", TooltipStyle -> "TextStyling"], 
               Annotation[#, "\[NumberSign]14", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{"isSet", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}, {
                    StyleBox["B", "ExpressionVariable"]}}]], 
                  RowBox[{
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["B", "ExpressionVariable"]}], 
                    "\[DoubleLeftRightArrow]", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"isSet", "[", 
                    StyleBox["B", "ExpressionVariable"], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    StyleBox["x", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}], ")"}], 
                    "\[DoubleLeftRightArrow]", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}], ")"}]}]}]}], 
                    ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]}, {
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["set equality alternative", 15]], Evaluator -> 
                 Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
                "\"\[NumberSign]15\"", TooltipStyle -> "TextStyling"], 
               Annotation[#, "\[NumberSign]15", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   StyleBox["A", "ExpressionVariable"]], 
                  RowBox[{
                    RowBox[{"isSet", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], "\[Implies]", 
                    
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    StyleBox["B", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["B", "ExpressionVariable"]}], 
                    "\[DoubleLeftRightArrow]", 
                    RowBox[{"(", 
                    RowBox[{
                    RowBox[{"isSet", "[", 
                    StyleBox["B", "ExpressionVariable"], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    StyleBox["x", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}], ")"}], 
                    "\[DoubleLeftRightArrow]", 
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}], ")"}]}]}]}], 
                    ")"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]}},
             AutoDelete -> False, 
            GridBoxAlignment -> {"Columns" -> {Center, Left}}, 
            GridBoxFrame -> {
             "ColumnsIndexed" -> {{{1, -1}, {1, -1}} -> True}}, 
            GridBoxItemSize -> {
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"], 
          TagBox[
           ButtonBox[
            
            PaneSelectorBox[{
             False -> "\"(set equality alternative)\"", True -> 
              StyleBox["\"(set equality alternative)\"", "HyperlinkActive"]}, 
             
             Dynamic[
              CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
             FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
            "Hyperlink", 
            ButtonData :> {
             "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:506079437"}, ButtonNote -> 
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
           Annotation[#, {
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:506079437"}, 
            "Hyperlink"]& ]}, "Labeled", DisplayFunction -> (GridBox[{{
             ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
             TagBox[
              ItemBox[
               PaneBox[
                TagBox[#, "SkipImageSizeLevel"], 
                Alignment -> {Center, Baseline}, BaselinePosition -> 
                Baseline], DefaultBaseStyle -> "Labeled"], 
              "SkipImageSizeLevel"]}}, 
           GridBoxAlignment -> {
            "Columns" -> {{Center}}, "Rows" -> {{Center}}}, AutoDelete -> 
           False, GridBoxItemSize -> {
            "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
           BaselinePosition -> {2, 1}]& ), 
         InterpretationFunction -> (RowBox[{"Labeled", "[", 
            RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {
        TemplateBox[{
          TagBox[
           GridBox[{{
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket[
                   "isFinite \[DoubleRightArrow] Dedekind finite", 16]], 
                 Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
                 "Preemptive"], "\"\[NumberSign]16\"", TooltipStyle -> 
                "TextStyling"], 
               Annotation[#, "\[NumberSign]16", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{"isSet", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    RowBox[{"isFinite", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], "\[Implies]", 
                    
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    StyleBox["f", "ExpressionVariable"]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], "\[Element]", 
                    
                    StyleBox["A", "ExpressionVariable"]}]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}, {
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}], 
                    "\[Implies]", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["y", "ExpressionVariable"]}]}]}]}], "\[Implies]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
               DisplayForm]}, {
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket[
                   "isFinite \[DoubleRightArrow] Dedekind finite", 17]], 
                 Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
                 "Preemptive"], "\"\[NumberSign]17\"", TooltipStyle -> 
                "TextStyling"], 
               Annotation[#, "\[NumberSign]17", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{"isSet", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}, {
                    StyleBox["f", "ExpressionVariable"]}}]], 
                  RowBox[{
                    RowBox[{
                    RowBox[{"isFinite", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], "\[Element]", 
                    
                    StyleBox["A", "ExpressionVariable"]}]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}, {
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}], 
                    "\[Implies]", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["y", "ExpressionVariable"]}]}]}]}], "\[Implies]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], 
                    "]"}]}]}]}]}]}], $CellContext`TheoremaForm], 
               DisplayForm]}}, AutoDelete -> False, 
            GridBoxAlignment -> {"Columns" -> {Center, Left}}, 
            GridBoxFrame -> {
             "ColumnsIndexed" -> {{{1, -1}, {1, -1}} -> True}}, 
            GridBoxItemSize -> {
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"], 
          TagBox[
           ButtonBox[
            
            PaneSelectorBox[{
             False -> "\"(isFinite \[DoubleRightArrow] Dedekind finite)\"", 
              True -> StyleBox[
               "\"(isFinite \[DoubleRightArrow] Dedekind finite)\"", 
                "HyperlinkActive"]}, 
             Dynamic[
              CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
             FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
            "Hyperlink", 
            ButtonData :> {
             "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:185397334"}, ButtonNote -> 
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
           Annotation[#, {
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:185397334"}, 
            "Hyperlink"]& ]}, "Labeled", DisplayFunction -> (GridBox[{{
             ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
             TagBox[
              ItemBox[
               PaneBox[
                TagBox[#, "SkipImageSizeLevel"], 
                Alignment -> {Center, Baseline}, BaselinePosition -> 
                Baseline], DefaultBaseStyle -> "Labeled"], 
              "SkipImageSizeLevel"]}}, 
           GridBoxAlignment -> {
            "Columns" -> {{Center}}, "Rows" -> {{Center}}}, AutoDelete -> 
           False, GridBoxItemSize -> {
            "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
           BaselinePosition -> {2, 1}]& ), 
         InterpretationFunction -> (RowBox[{"Labeled", "[", 
            RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {
        TemplateBox[{
          TagBox[
           GridBox[{{
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["UnionOf isSet,[IR]", 18]], Evaluator -> 
                 Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
                "\"\[NumberSign]18\"", TooltipStyle -> "TextStyling"], 
               Annotation[#, "\[NumberSign]18", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{"isSet", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}, {
                    StyleBox["B", "ExpressionVariable"]}, {
                    StyleBox["P", "ExpressionVariable"]}, {
                    RowBox[{"isSet", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{"isSet", "[", 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[Union]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]], 
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}], 
                    "]"}]}], $CellContext`TheoremaForm], DisplayForm]}, {
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["UnionOf isSet,[IR]", 19]], Evaluator -> 
                 Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
                "\"\[NumberSign]19\"", TooltipStyle -> "TextStyling"], 
               Annotation[#, "\[NumberSign]19", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{"isSet", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}, {
                    StyleBox["B", "ExpressionVariable"]}, {
                    StyleBox["P", "ExpressionVariable"]}}]], 
                  RowBox[{"isSet", "[", 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[Union]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]], 
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}], 
                    "]"}]}], $CellContext`TheoremaForm], DisplayForm]}}, 
            AutoDelete -> False, 
            GridBoxAlignment -> {"Columns" -> {Center, Left}}, 
            GridBoxFrame -> {
             "ColumnsIndexed" -> {{{1, -1}, {1, -1}} -> True}}, 
            GridBoxItemSize -> {
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"], 
          TagBox[
           ButtonBox[
            
            PaneSelectorBox[{
             False -> "\"(UnionOf isSet,[IR])\"", True -> 
              StyleBox["\"(UnionOf isSet,[IR])\"", "HyperlinkActive"]}, 
             Dynamic[
              CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
             FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
            "Hyperlink", 
            ButtonData :> {
             "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:164534974"}, ButtonNote -> 
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
           Annotation[#, {
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:164534974"}, 
            "Hyperlink"]& ]}, "Labeled", DisplayFunction -> (GridBox[{{
             ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
             TagBox[
              ItemBox[
               PaneBox[
                TagBox[#, "SkipImageSizeLevel"], 
                Alignment -> {Center, Baseline}, BaselinePosition -> 
                Baseline], DefaultBaseStyle -> "Labeled"], 
              "SkipImageSizeLevel"]}}, 
           GridBoxAlignment -> {
            "Columns" -> {{Center}}, "Rows" -> {{Center}}}, AutoDelete -> 
           False, GridBoxItemSize -> {
            "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
           BaselinePosition -> {2, 1}]& ), 
         InterpretationFunction -> (RowBox[{"Labeled", "[", 
            RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {
        TemplateBox[{
          TagBox[
           GridBox[{{
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["\[Element] UnionOf,[IR]", 20]], Evaluator -> 
                 Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
                "\"\[NumberSign]20\"", TooltipStyle -> "TextStyling"], 
               Annotation[#, "\[NumberSign]20", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{"isSet", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}, {
                    StyleBox["B", "ExpressionVariable"]}, {
                    StyleBox["P", "ExpressionVariable"]}, {
                    RowBox[{"isSet", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}, {
                    StyleBox["x", "ExpressionVariable"]}}]], 
                  RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[Union]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]], 
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}]}], ")"}], 
                    "\[DoubleLeftRightArrow]", 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], "\[And]", 
                    RowBox[{"isSet", "[", 
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], "]"}]}]], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], 
                    "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]}, {
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["\[Element] UnionOf,[IR]", 21]], Evaluator -> 
                 Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
                "\"\[NumberSign]21\"", TooltipStyle -> "TextStyling"], 
               Annotation[#, "\[NumberSign]21", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{"isSet", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}, {
                    StyleBox["B", "ExpressionVariable"]}, {
                    StyleBox["P", "ExpressionVariable"]}, {
                    StyleBox["x", "ExpressionVariable"]}}]], 
                  RowBox[{
                    RowBox[{"(", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[Union]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]], 
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}]}], ")"}], 
                    "\[DoubleLeftRightArrow]", 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], "\[And]", 
                    RowBox[{"isSet", "[", 
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], "]"}]}]], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], 
                    "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]}},
             AutoDelete -> False, 
            GridBoxAlignment -> {"Columns" -> {Center, Left}}, 
            GridBoxFrame -> {
             "ColumnsIndexed" -> {{{1, -1}, {1, -1}} -> True}}, 
            GridBoxItemSize -> {
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"], 
          TagBox[
           ButtonBox[
            
            PaneSelectorBox[{
             False -> "\"(\[Element] UnionOf,[IR])\"", True -> 
              StyleBox["\"(\[Element] UnionOf,[IR])\"", "HyperlinkActive"]}, 
             Dynamic[
              CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
             FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
            "Hyperlink", 
            ButtonData :> {
             "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:649544011"}, ButtonNote -> 
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
           Annotation[#, {
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:649544011"}, 
            "Hyperlink"]& ]}, "Labeled", DisplayFunction -> (GridBox[{{
             ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
             TagBox[
              ItemBox[
               PaneBox[
                TagBox[#, "SkipImageSizeLevel"], 
                Alignment -> {Center, Baseline}, BaselinePosition -> 
                Baseline], DefaultBaseStyle -> "Labeled"], 
              "SkipImageSizeLevel"]}}, 
           GridBoxAlignment -> {
            "Columns" -> {{Center}}, "Rows" -> {{Center}}}, AutoDelete -> 
           False, GridBoxItemSize -> {
            "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
           BaselinePosition -> {2, 1}]& ), 
         InterpretationFunction -> (RowBox[{"Labeled", "[", 
            RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}}, DefaultBaseStyle -> 
      "Column", GridBoxAlignment -> {"Columns" -> {{Left}}}, 
      GridBoxItemSize -> {
       "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Column"],
    "\"These formulas have been used as assumptions in a proof, but have \
themselves been proved in a different version. The first formula of each \
group is the proved version, the others are the used ones.\""},
   "Labeled",
   DisplayFunction->(GridBox[{{
       ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
       TagBox[
        ItemBox[
         PaneBox[
          TagBox[#, "SkipImageSizeLevel"], Alignment -> {Center, Baseline}, 
          BaselinePosition -> Baseline], DefaultBaseStyle -> "Labeled"], 
        "SkipImageSizeLevel"]}}, 
     GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
     AutoDelete -> False, 
     GridBoxItemSize -> {"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}},
      BaselinePosition -> {2, 1}]& ),
   InterpretationFunction->(RowBox[{"Labeled", "[", 
      RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {
  2,"\<\"duplicates (4)\"\>"->
  TemplateBox[{TagBox[
     GridBox[{{
        TemplateBox[{
          TagBox[
           GridBox[{{
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["isNoetherian", 22]], Evaluator -> Automatic, 
                 ImageSize -> {20, 20}, Method -> "Preemptive"], 
                "\"\[NumberSign]22\"", TooltipStyle -> "TextStyling"], 
               Annotation[#, "\[NumberSign]22", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   RowBox[{
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                    
                    StyleBox["\[Succeeds]", "ExpressionVariable"]}]], 
                  RowBox[{
                    RowBox[{"isNoetherian", "[", 
                    RowBox[{
                    StyleBox["\[Succeeds]", "ExpressionVariable"], ",", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                    "]"}], 
                    TagBox[
                    
                    RowBox[{
                    ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                    "a:=b"], 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainSets", "[", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{"nonemptySet", "[", 
                    StyleBox["S", "ExpressionVariable"], "]"}], "\[Implies]", 
                    
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], "\[Element]", 
                    StyleBox["S", "ExpressionVariable"]}]}}]], 
                    RowBox[{"isMin", "[", 
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], ",", 
                    StyleBox["S", "ExpressionVariable"], ",", 
                    StyleBox["\[Succeeds]", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}], $CellContext`TheoremaForm], 
               DisplayForm]}, {
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["isNoetherian", 23]], Evaluator -> Automatic, 
                 ImageSize -> {20, 20}, Method -> "Preemptive"], 
                "\"\[NumberSign]23\"", TooltipStyle -> "TextStyling"], 
               Annotation[#, "\[NumberSign]23", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   RowBox[{
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
                    
                    StyleBox["\[Succeeds]", "ExpressionVariable"]}]], 
                  RowBox[{
                    RowBox[{"isNoetherian", "[", 
                    RowBox[{
                    StyleBox["\[Succeeds]", "ExpressionVariable"], ",", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                    "]"}], 
                    TagBox[
                    
                    RowBox[{
                    ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                    "a:=b"], 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainSets", "[", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], 
                    "]"}]], "[", 
                    StyleBox["S", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{"nonemptySet", "[", 
                    StyleBox["S", "ExpressionVariable"], "]"}], "\[Implies]", 
                    
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], "\[Element]", 
                    StyleBox["S", "ExpressionVariable"]}]}}]], 
                    RowBox[{"isMin", "[", 
                    RowBox[{
                    StyleBox["s", "ExpressionVariable"], ",", 
                    StyleBox["S", "ExpressionVariable"], ",", 
                    StyleBox["\[Succeeds]", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}], $CellContext`TheoremaForm], 
               DisplayForm]}}, AutoDelete -> False, 
            GridBoxAlignment -> {"Columns" -> {Center, Left}}, 
            GridBoxFrame -> {
             "ColumnsIndexed" -> {{{1, -1}, {1, -1}} -> True}}, 
            GridBoxItemSize -> {
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"], 
          TagBox[
           ButtonBox[
            
            PaneSelectorBox[{
             False -> "\"(isNoetherian)\"", True -> 
              StyleBox["\"(isNoetherian)\"", "HyperlinkActive"]}, 
             Dynamic[
              CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
             FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
            "Hyperlink", 
            ButtonData :> {
             "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:929796587"}, 
            ButtonNote -> 
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb"], 
           Annotation[#, {
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:929796587"}, 
            "Hyperlink"]& ]}, "Labeled", DisplayFunction -> (GridBox[{{
             ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
             TagBox[
              ItemBox[
               PaneBox[
                TagBox[#, "SkipImageSizeLevel"], 
                Alignment -> {Center, Baseline}, BaselinePosition -> 
                Baseline], DefaultBaseStyle -> "Labeled"], 
              "SkipImageSizeLevel"]}}, 
           GridBoxAlignment -> {
            "Columns" -> {{Center}}, "Rows" -> {{Center}}}, AutoDelete -> 
           False, GridBoxItemSize -> {
            "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
           BaselinePosition -> {2, 1}]& ), 
         InterpretationFunction -> (RowBox[{"Labeled", "[", 
            RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {
        TemplateBox[{
          TagBox[
           GridBox[{{
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["isWellFounded", 24]], Evaluator -> Automatic,
                  ImageSize -> {20, 20}, Method -> "Preemptive"], 
                "\"\[NumberSign]24\"", TooltipStyle -> "TextStyling"], 
               Annotation[#, "\[NumberSign]24", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{"isSet", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}, {
                    StyleBox["\[SucceedsEqual]", "ExpressionVariable"]}}]], 
                  RowBox[{
                    RowBox[{"isWellFounded", "[", 
                    RowBox[{
                    StyleBox["\[SucceedsEqual]", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                    TagBox[
                    
                    RowBox[{
                    ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                    "a:=b"], 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{"nonemptySet", "[", 
                    StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\[SubsetEqual]", 
                    StyleBox["A", "ExpressionVariable"]}]], 
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}}]], 
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}]], 
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[SucceedsEqual]", 
                    StyleBox["y", "ExpressionVariable"]}], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]}, {
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["isWellFounded", 25]], Evaluator -> Automatic,
                  ImageSize -> {20, 20}, Method -> "Preemptive"], 
                "\"\[NumberSign]25\"", TooltipStyle -> "TextStyling"], 
               Annotation[#, "\[NumberSign]25", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{"isSet", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}, {
                    StyleBox["\[SucceedsEqual]", "ExpressionVariable"]}}]], 
                  RowBox[{
                    RowBox[{"isWellFounded", "[", 
                    RowBox[{
                    StyleBox["\[SucceedsEqual]", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                    TagBox[
                    
                    RowBox[{
                    ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                    "a:=b"], 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{"nonemptySet", "[", 
                    StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\[SubsetEqual]", 
                    StyleBox["A", "ExpressionVariable"]}]], 
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["x", "ExpressionVariable"]}]], 
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[SucceedsEqual]", 
                    StyleBox["y", "ExpressionVariable"]}], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]}, {
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["isWellFounded", 26]], Evaluator -> Automatic,
                  ImageSize -> {20, 20}, Method -> "Preemptive"], 
                "\"\[NumberSign]26\"", TooltipStyle -> "TextStyling"], 
               Annotation[#, "\[NumberSign]26", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{"isSet", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}, {
                    StyleBox["\[SucceedsEqual]", "ExpressionVariable"]}}]], 
                  RowBox[{
                    RowBox[{"isWellFounded", "[", 
                    RowBox[{
                    StyleBox["\[SucceedsEqual]", "ExpressionVariable"], ",", 
                    StyleBox["A", "ExpressionVariable"]}], "]"}], 
                    TagBox[
                    
                    RowBox[{
                    ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                    "a:=b"], 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{"nonemptySet", "[", 
                    StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\[SubsetEqual]", 
                    StyleBox["A", "ExpressionVariable"]}]], 
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["x", "ExpressionVariable"]}]], 
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[SucceedsEqual]", 
                    StyleBox["y", "ExpressionVariable"]}], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]}}, AutoDelete -> False, 
            GridBoxAlignment -> {"Columns" -> {Center, Left}}, 
            GridBoxFrame -> {
             "ColumnsIndexed" -> {{{1, -1}, {1, -1}} -> True}}, 
            GridBoxItemSize -> {
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"], 
          TagBox[
           ButtonBox[
            
            PaneSelectorBox[{
             False -> "\"(isWellFounded)\"", True -> 
              StyleBox["\"(isWellFounded)\"", "HyperlinkActive"]}, 
             Dynamic[
              CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
             FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
            "Hyperlink", 
            ButtonData :> {
             "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:173714274"}, ButtonNote -> 
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
           Annotation[#, {
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:173714274"}, 
            "Hyperlink"]& ]}, "Labeled", DisplayFunction -> (GridBox[{{
             ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
             TagBox[
              ItemBox[
               PaneBox[
                TagBox[#, "SkipImageSizeLevel"], 
                Alignment -> {Center, Baseline}, BaselinePosition -> 
                Baseline], DefaultBaseStyle -> "Labeled"], 
              "SkipImageSizeLevel"]}}, 
           GridBoxAlignment -> {
            "Columns" -> {{Center}}, "Rows" -> {{Center}}}, AutoDelete -> 
           False, GridBoxItemSize -> {
            "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
           BaselinePosition -> {2, 1}]& ), 
         InterpretationFunction -> (RowBox[{"Labeled", "[", 
            RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {
        TemplateBox[{
          TagBox[
           GridBox[{{
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["set equality", 27]], Evaluator -> Automatic, 
                 ImageSize -> {20, 20}, Method -> "Preemptive"], 
                "\"\[NumberSign]27\"", TooltipStyle -> "TextStyling"], 
               Annotation[#, "\[NumberSign]27", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["B", "ExpressionVariable"]}]], 
                  RowBox[{
                    RowBox[{"isSet", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], "\[Implies]", 
                    
                    RowBox[{
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["B", "ExpressionVariable"]}], 
                    TagBox[
                    
                    RowBox[{
                    ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                    "a:=b"], 
                    RowBox[{
                    RowBox[{"isSet", "[", 
                    StyleBox["B", "ExpressionVariable"], "]"}], "\[And]", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\[SubsetEqual]", 
                    StyleBox["B", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\[SubsetEqual]", 
                    StyleBox[
                    "A", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]}, {
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["set equality", 28]], Evaluator -> Automatic, 
                 ImageSize -> {20, 20}, Method -> "Preemptive"], 
                "\"\[NumberSign]28\"", TooltipStyle -> "TextStyling"], 
               Annotation[#, "\[NumberSign]28", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{"isSet", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}, {
                    StyleBox["B", "ExpressionVariable"]}}]], 
                  RowBox[{
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["B", "ExpressionVariable"]}], 
                    TagBox[
                    
                    RowBox[{
                    ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                    "a:=b"], 
                    RowBox[{
                    RowBox[{"isSet", "[", 
                    StyleBox["B", "ExpressionVariable"], "]"}], "\[And]", 
                    RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\[SubsetEqual]", 
                    StyleBox["B", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\[SubsetEqual]", 
                    StyleBox[
                    "A", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]}}, AutoDelete -> False, 
            GridBoxAlignment -> {"Columns" -> {Center, Left}}, 
            GridBoxFrame -> {
             "ColumnsIndexed" -> {{{1, -1}, {1, -1}} -> True}}, 
            GridBoxItemSize -> {
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"], 
          TagBox[
           ButtonBox[
            
            PaneSelectorBox[{
             False -> "\"(set equality)\"", True -> 
              StyleBox["\"(set equality)\"", "HyperlinkActive"]}, 
             Dynamic[
              CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
             FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
            "Hyperlink", 
            ButtonData :> {
             "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:1919461563"}, ButtonNote -> 
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
           Annotation[#, {
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:1919461563"}, 
            "Hyperlink"]& ]}, "Labeled", DisplayFunction -> (GridBox[{{
             ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
             TagBox[
              ItemBox[
               PaneBox[
                TagBox[#, "SkipImageSizeLevel"], 
                Alignment -> {Center, Baseline}, BaselinePosition -> 
                Baseline], DefaultBaseStyle -> "Labeled"], 
              "SkipImageSizeLevel"]}}, 
           GridBoxAlignment -> {
            "Columns" -> {{Center}}, "Rows" -> {{Center}}}, AutoDelete -> 
           False, GridBoxItemSize -> {
            "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
           BaselinePosition -> {2, 1}]& ), 
         InterpretationFunction -> (RowBox[{"Labeled", "[", 
            RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {
        TemplateBox[{
          TagBox[
           GridBox[{{
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["enumeration term \[Element],[IR]", 29]], 
                 Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
                 "Preemptive"], "\"\[NumberSign]29\"", TooltipStyle -> 
                "TextStyling"], 
               Annotation[#, "\[NumberSign]29", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"x", "..."}], "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"]}]], 
                  RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    RowBox[{"{", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"x", "..."}], "ExpressionVariable"]}], "}"}]}], 
                    TagBox[
                    
                    RowBox[{
                    ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                    "a:=b"], 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["x", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    RowBox[{"{", 
                    StyleBox[
                    RowBox[{"x", "..."}], "ExpressionVariable"], 
                    "}"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]}, {
              TagBox[
               TooltipBox[
                ButtonBox[
                "\"C\"", Appearance -> Automatic, ButtonFunction :> 
                 CopyToClipboard[
                   AngleBracket["enumeration term \[Element],[IR]", 30]], 
                 Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
                 "Preemptive"], "\"\[NumberSign]30\"", TooltipStyle -> 
                "TextStyling"], 
               Annotation[#, "\[NumberSign]30", "Tooltip"]& ], 
              TagBox[
               FormBox[
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"x", "..."}], "ExpressionVariable"]}]], 
                  RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    RowBox[{"{", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox[
                    RowBox[{"x", "..."}], "ExpressionVariable"]}], "}"}]}], 
                    TagBox[
                    
                    RowBox[{
                    ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                    "a:=b"], 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["x", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    RowBox[{"{", 
                    StyleBox[
                    RowBox[{"x", "..."}], "ExpressionVariable"], 
                    "}"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]}},
             AutoDelete -> False, 
            GridBoxAlignment -> {"Columns" -> {Center, Left}}, 
            GridBoxFrame -> {
             "ColumnsIndexed" -> {{{1, -1}, {1, -1}} -> True}}, 
            GridBoxItemSize -> {
             "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"], 
          TagBox[
           ButtonBox[
            
            PaneSelectorBox[{
             False -> "\"(enumeration term \[Element],[IR])\"", True -> 
              StyleBox[
               "\"(enumeration term \[Element],[IR])\"", "HyperlinkActive"]}, 
             
             Dynamic[
              CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
             FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
            "Hyperlink", 
            ButtonData :> {
             "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:1130747252"}, ButtonNote -> 
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
           Annotation[#, {
            "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:1130747252"}, 
            "Hyperlink"]& ]}, "Labeled", DisplayFunction -> (GridBox[{{
             ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
             TagBox[
              ItemBox[
               PaneBox[
                TagBox[#, "SkipImageSizeLevel"], 
                Alignment -> {Center, Baseline}, BaselinePosition -> 
                Baseline], DefaultBaseStyle -> "Labeled"], 
              "SkipImageSizeLevel"]}}, 
           GridBoxAlignment -> {
            "Columns" -> {{Center}}, "Rows" -> {{Center}}}, AutoDelete -> 
           False, GridBoxItemSize -> {
            "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}, 
           BaselinePosition -> {2, 1}]& ), 
         InterpretationFunction -> (RowBox[{"Labeled", "[", 
            RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}}, DefaultBaseStyle -> 
      "Column", GridBoxAlignment -> {"Columns" -> {{Left}}}, 
      GridBoxItemSize -> {
       "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Column"],
    "\"\""},
   "Labeled",
   DisplayFunction->(GridBox[{{
       ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
       TagBox[
        ItemBox[
         PaneBox[
          TagBox[#, "SkipImageSizeLevel"], Alignment -> {Center, Baseline}, 
          BaselinePosition -> Baseline], DefaultBaseStyle -> "Labeled"], 
        "SkipImageSizeLevel"]}}, 
     GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
     AutoDelete -> False, 
     GridBoxItemSize -> {"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}},
      BaselinePosition -> {2, 1}]& ),
   InterpretationFunction->(RowBox[{"Labeled", "[", 
      RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {
  3,"\<\"non-definitions (35)\"\>"->
  TemplateBox[{TagBox[
     GridBox[{{
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951073, TheoryTools`$LabelDelimiter, 
             "< \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], "<", 
                StyleBox["y", "ExpressionVariable"]}], 
              "\[DoubleLeftRightArrow]", 
              RowBox[{
                StyleBox["y", "ExpressionVariable"], ">", 
                StyleBox[
                "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(< \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox["\"(< \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1529886137"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1529886137"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951087, TheoryTools`$LabelDelimiter, 
             "> \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], ">", 
                StyleBox["y", "ExpressionVariable"]}], 
              "\[DoubleLeftRightArrow]", 
              RowBox[{"(", 
                RowBox[{
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["y", "ExpressionVariable"]}], "\[And]", 
                  RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["y", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}]}], 
                ")"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(> \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox["\"(> \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1992599802"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1992599802"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951113, TheoryTools`$LabelDelimiter, 
             "assoc * \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["z", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                RowBox[{"(", 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], " ", 
                    StyleBox["y", "ExpressionVariable"]}], ")"}], " ", 
                StyleBox["z", "ExpressionVariable"]}], "\[Equal]", 
              RowBox[{
                StyleBox["x", "ExpressionVariable"], " ", 
                RowBox[{"(", 
                  RowBox[{
                    StyleBox["y", "ExpressionVariable"], " ", 
                    StyleBox["z", "ExpressionVariable"]}], 
                  ")"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(assoc * \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox[
             "\"(assoc * \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:803296988"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:803296988"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951142, TheoryTools`$LabelDelimiter, 
             "comm * \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], " ", 
                StyleBox["y", "ExpressionVariable"]}], "\[Equal]", 
              RowBox[{
                StyleBox["y", "ExpressionVariable"], " ", 
                StyleBox[
                "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(comm * \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox[
             "\"(comm * \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:15375924"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:15375924"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951160, TheoryTools`$LabelDelimiter, 
             "distr * + \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["z", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], " ", 
                RowBox[{"(", 
                  RowBox[{
                    StyleBox["y", "ExpressionVariable"], "+", 
                    StyleBox["z", "ExpressionVariable"]}], ")"}]}], 
              "\[Equal]", 
              RowBox[{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], " ", 
                  StyleBox["y", "ExpressionVariable"]}], "+", 
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], " ", 
                  StyleBox[
                  "z", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(distr * + \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox[
             "\"(distr * + \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:651802243"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:651802243"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951194, TheoryTools`$LabelDelimiter, 
             "- inverse \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], "-", 
                RowBox[{
                  TagBox["(", "AutoParentheses"], 
                  StyleBox["x", "ExpressionVariable"], 
                  TagBox[")", "AutoParentheses"]}]}], "\[Equal]", 
              "0"}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(- inverse \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox[
             "\"(- inverse \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:96279916"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:96279916"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951205, TheoryTools`$LabelDelimiter, 
             "assoc + \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["z", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "+", 
                  StyleBox["y", "ExpressionVariable"]}], "+", 
                StyleBox["z", "ExpressionVariable"]}], "\[Equal]", 
              RowBox[{
                StyleBox["x", "ExpressionVariable"], "+", 
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "+", 
                  StyleBox[
                  "z", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(assoc + \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox[
             "\"(assoc + \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1865835468"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1865835468"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951234, TheoryTools`$LabelDelimiter, 
             "comm + \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], "+", 
                StyleBox["y", "ExpressionVariable"]}], "\[Equal]", 
              RowBox[{
                StyleBox["y", "ExpressionVariable"], "+", 
                StyleBox[
                "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(comm + \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox[
             "\"(comm + \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:158934172"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:158934172"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951252, TheoryTools`$LabelDelimiter, 
             "- \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], "-", 
                StyleBox["y", "ExpressionVariable"]}], "\[Equal]", 
              RowBox[{
                StyleBox["x", "ExpressionVariable"], "-", 
                RowBox[{
                  TagBox["(", "AutoParentheses"], 
                  StyleBox["y", "ExpressionVariable"], 
                  TagBox[
                  ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm],
          DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(- \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox["\"(- \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:445777866"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:445777866"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951270, TheoryTools`$LabelDelimiter, 
             "\[GreaterEqual] reflexive \[DoubleStruckCapitalZ]"]]], 
         Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
         "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", 
              StyleBox[
              "x", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(\[GreaterEqual] reflexive \[DoubleStruckCapitalZ])\"",
             True -> StyleBox[
             "\"(\[GreaterEqual] reflexive \[DoubleStruckCapitalZ])\"", 
              "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1192085943"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1192085943"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951277, TheoryTools`$LabelDelimiter, 
             "\[GreaterEqual] transitive \[DoubleStruckCapitalZ]"]]], 
         Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
         "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["z", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", 
                  StyleBox["y", "ExpressionVariable"]}], "\[And]", 
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[GreaterEqual]", 
                  StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
              RowBox[{
                StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", 
                StyleBox[
                "z", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> 
            "\"(\[GreaterEqual] transitive \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox[
             "\"(\[GreaterEqual] transitive \[DoubleStruckCapitalZ])\"", 
              "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:184532980"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:184532980"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951302, TheoryTools`$LabelDelimiter, 
             "1 > 0,[IR]"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{"1", ">", "0"}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(1 > 0,[IR])\"", True -> 
            StyleBox["\"(1 > 0,[IR])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:612656118"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:612656118"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951306, TheoryTools`$LabelDelimiter, 
             "0 \[Element] \[DoubleStruckCapitalZ],[IR]"]]], Evaluator -> 
         Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          
          RowBox[{"0", "\[Element]", 
            "\[DoubleStruckCapitalZ]"}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(0 \[Element] \[DoubleStruckCapitalZ],[IR])\"", True -> 
            StyleBox[
             "\"(0 \[Element] \[DoubleStruckCapitalZ],[IR])\"", 
              "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1914719525"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1914719525"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951310, TheoryTools`$LabelDelimiter, 
             "1 \[Element] \[DoubleStruckCapitalZ],[IR]"]]], Evaluator -> 
         Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          
          RowBox[{"1", "\[Element]", 
            "\[DoubleStruckCapitalZ]"}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(1 \[Element] \[DoubleStruckCapitalZ],[IR])\"", True -> 
            StyleBox[
             "\"(1 \[Element] \[DoubleStruckCapitalZ],[IR])\"", 
              "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:158149974"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:158149974"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951314, TheoryTools`$LabelDelimiter, 
             "\[DoubleStruckCapitalZ] isSet"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          
          RowBox[{"isSet", "[", "\[DoubleStruckCapitalZ]", 
            "]"}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(\[DoubleStruckCapitalZ] isSet)\"", True -> 
            StyleBox[
             "\"(\[DoubleStruckCapitalZ] isSet)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1844729221"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1844729221"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951321, TheoryTools`$LabelDelimiter, 
             "- unary closed \[DoubleStruckCapitalZ],[IR]"]]], Evaluator -> 
         Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{"-", 
                  StyleBox["x", "ExpressionVariable"]}], 
                TagBox[")", "AutoParentheses"]}], "\[Element]", 
              "\[DoubleStruckCapitalZ]"}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          PaneSelectorBox[{
           False -> "\"(- unary closed \[DoubleStruckCapitalZ],[IR])\"", True -> 
            StyleBox[
             "\"(- unary closed \[DoubleStruckCapitalZ],[IR])\"", 
              "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1979354739"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1979354739"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951332, TheoryTools`$LabelDelimiter, 
             "* closed \[DoubleStruckCapitalZ],[IR]"]]], Evaluator -> 
         Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], " ", 
                StyleBox["y", "ExpressionVariable"]}], "\[Element]", 
              "\[DoubleStruckCapitalZ]"}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(* closed \[DoubleStruckCapitalZ],[IR])\"", True -> 
            StyleBox[
             "\"(* closed \[DoubleStruckCapitalZ],[IR])\"", 
              "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:50692980"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:50692980"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951344, TheoryTools`$LabelDelimiter, 
             "\[GreaterEqual] antisymmetric \[DoubleStruckCapitalZ]"]]], 
         Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
         "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", 
                  StyleBox["y", "ExpressionVariable"]}], "\[And]", 
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[GreaterEqual]", 
                  StyleBox["x", "ExpressionVariable"]}]}], "\[Implies]", 
              RowBox[{
                StyleBox["x", "ExpressionVariable"], "\[Equal]", 
                StyleBox[
                "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> 
            "\"(\[GreaterEqual] antisymmetric \[DoubleStruckCapitalZ])\"", 
            True -> StyleBox[
             "\"(\[GreaterEqual] antisymmetric \[DoubleStruckCapitalZ])\"", 
              "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1497518355"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1497518355"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951368, TheoryTools`$LabelDelimiter, 
             "1+1,[IR]"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            RowBox[{"1", "+", "1"}], "\[Equal]", 
            "2"}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(1+1,[IR])\"", True -> 
            StyleBox["\"(1+1,[IR])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:98773918"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:98773918"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951376, TheoryTools`$LabelDelimiter, 
             "1 rneutral \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], " ", "1"}], "\[Equal]", 
              StyleBox[
              "x", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(1 rneutral \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox[
             "\"(1 rneutral \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
           
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:11612048"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:11612048"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951387, TheoryTools`$LabelDelimiter, 
             "abstraction term isSet,[IR]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isSet", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}, {
                StyleBox["f", "ExpressionVariable"]}, {
                StyleBox["P", "ExpressionVariable"]}}]], 
            RowBox[{"isSet", "[", 
              RowBox[{"{", 
                RowBox[{
                  RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}], 
                  UnderscriptBox["|", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                  RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}], "}"}], 
              "]"}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(abstraction term isSet,[IR])\"", True -> 
            StyleBox["\"(abstraction term isSet,[IR])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:1890519876"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:1890519876"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951407, TheoryTools`$LabelDelimiter, 
             "+ closed \[DoubleStruckCapitalZ],[IR]"]]], Evaluator -> 
         Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], "+", 
                StyleBox["y", "ExpressionVariable"]}], "\[Element]", 
              "\[DoubleStruckCapitalZ]"}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(+ closed \[DoubleStruckCapitalZ],[IR])\"", True -> 
            StyleBox[
             "\"(+ closed \[DoubleStruckCapitalZ],[IR])\"", 
              "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1136577066"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1136577066"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951419, TheoryTools`$LabelDelimiter, 
             "0 rneutral \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], "+", "0"}], "\[Equal]", 
              StyleBox[
              "x", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(0 rneutral \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox[
             "\"(0 rneutral \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
           
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:2817676"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:2817676"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951430, TheoryTools`$LabelDelimiter, 
             "\[GreaterEqual] total \[DoubleStruckCapitalZ]"]]], Evaluator -> 
         Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", 
                StyleBox["y", "ExpressionVariable"]}], "\[Or]", 
              RowBox[{
                StyleBox["y", "ExpressionVariable"], "\[GreaterEqual]", 
                StyleBox[
                "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(\[GreaterEqual] total \[DoubleStruckCapitalZ])\"", 
            True -> StyleBox[
             "\"(\[GreaterEqual] total \[DoubleStruckCapitalZ])\"", 
              "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1061215427"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1061215427"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951446, TheoryTools`$LabelDelimiter, 
             "induction on \[DoubleStruckCapitalN],[IR]"]]], Evaluator -> 
         Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                StyleBox["P", "ExpressionVariable"]}, {
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                RowBox[{
                  StyleBox["P", "ExpressionVariable"], "[", 
                  StyleBox["a", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "\[Element]", 
                    SubscriptBox["\[DoubleStruckCapitalZ]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", "\[Ellipsis]", 
                    ",", "\[Infinity]"}]]}]}}]], 
                  RowBox[{
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}], "\[Implies]", 
                    
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "+", "1"}], 
                    "]"}]}]}]}], "\[Implies]", 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "\[Element]", 
                    SubscriptBox["\[DoubleStruckCapitalZ]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", "\[Ellipsis]", 
                    ",", "\[Infinity]"}]]}]}}]], 
                RowBox[{
                  StyleBox["P", "ExpressionVariable"], "[", 
                  StyleBox["n", "ExpressionVariable"], 
                  "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(induction on \[DoubleStruckCapitalN],[IR])\"", True -> 
            StyleBox[
             "\"(induction on \[DoubleStruckCapitalN],[IR])\"", 
              "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:185621103"}, ButtonNote -> 
          
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:185621103"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951478, TheoryTools`$LabelDelimiter, 
             "\[GreaterEqual] + \[DoubleStruckCapitalZ]"]]], Evaluator -> 
         Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["z", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", 
                StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
              RowBox[{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "+", 
                  StyleBox["z", "ExpressionVariable"]}], "\[GreaterEqual]", 
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "+", 
                  StyleBox[
                  "z", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(\[GreaterEqual] + \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox[
             "\"(\[GreaterEqual] + \[DoubleStruckCapitalZ])\"", 
              "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:767068711"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:767068711"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951505, TheoryTools`$LabelDelimiter, 
             "power set isSet,[IR]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isSet", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"isSet", "[", 
              RowBox[{"Pow", "[", 
                StyleBox["A", "ExpressionVariable"], "]"}], 
              "]"}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(power set isSet,[IR])\"", True -> 
            StyleBox["\"(power set isSet,[IR])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:84772714"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:84772714"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951521, TheoryTools`$LabelDelimiter, 
             "\[GreaterEqual] discrete \[DoubleStruckCapitalZ]"]]], Evaluator -> 
         Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["y", "ExpressionVariable"], "\[GreaterEqual]", 
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "+", "1"}]}], "\[Or]", 
              
              RowBox[{
                StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", 
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "+", "1"}]}], "\[Or]", 
              
              RowBox[{
                StyleBox["x", "ExpressionVariable"], "\[Equal]", 
                StyleBox[
                "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(\[GreaterEqual] discrete \[DoubleStruckCapitalZ])\"", 
            True -> StyleBox[
             "\"(\[GreaterEqual] discrete \[DoubleStruckCapitalZ])\"", 
              "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:216380082"}, ButtonNote -> 
          
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:216380082"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951550, TheoryTools`$LabelDelimiter, 
             "zdf * \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], " ", 
                  StyleBox["y", "ExpressionVariable"]}], "\[Equal]", "0"}], 
              "\[Implies]", 
              RowBox[{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Equal]", "0"}], 
                "\[Or]", 
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[Equal]", 
                  "0"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(zdf * \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox[
             "\"(zdf * \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:18630049"}, ButtonNote -> 
          
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:18630049"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951576, TheoryTools`$LabelDelimiter, 
             "enumeration term isSet,[IR]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox[
              RowBox[{"x", "..."}], "ExpressionVariable"]], 
            RowBox[{"isSet", "[", 
              RowBox[{"{", 
                StyleBox[
                 RowBox[{"x", "..."}], "ExpressionVariable"], "}"}], 
              "]"}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(enumeration term isSet,[IR])\"", True -> 
            StyleBox["\"(enumeration term isSet,[IR])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:1045764632"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:1045764632"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951586, TheoryTools`$LabelDelimiter, 
             "predicate set isSet,[IR]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["f", "ExpressionVariable"], ",", 
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["Q", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{"isSet", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  UnderscriptBox[
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]], 
                  RowBox[{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], "\[Element]", 
                    
                    StyleBox["A", "ExpressionVariable"]}]}]}], "\[Implies]", 
              RowBox[{"isSet", "[", 
                RowBox[{"{", 
                  RowBox[{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], "}"}], 
                "]"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(predicate set isSet,[IR])\"", True -> 
            StyleBox["\"(predicate set isSet,[IR])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:45751977"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:45751977"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951622, TheoryTools`$LabelDelimiter, 
             "empty set"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["x", "ExpressionVariable"]], 
            RowBox[{"\[Not]", 
              RowBox[{
                TagBox["(", "AutoParentheses"], 
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[EmptySet]"}], 
                TagBox[
                ")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(empty set)\"", True -> 
            StyleBox["\"(empty set)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:1089581973"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:1089581973"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951632, TheoryTools`$LabelDelimiter, 
             "-1,[IR]"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            RowBox[{
              TagBox["(", "AutoParentheses"], 
              RowBox[{"-", "1"}], 
              TagBox[")", "AutoParentheses"]}], "\[Equal]", 
            RowBox[{"-", "1"}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(-1,[IR])\"", True -> 
            StyleBox["\"(-1,[IR])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:360506785"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:360506785"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951641, TheoryTools`$LabelDelimiter, 
             "union isSet,[IR]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isSet", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{"isSet", "[", 
              RowBox[{"Un", "[", 
                StyleBox["A", "ExpressionVariable"], "]"}], 
              "]"}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(union isSet,[IR])\"", True -> 
            StyleBox["\"(union isSet,[IR])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:22012017"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:22012017"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951657, TheoryTools`$LabelDelimiter, 
             "\[GreaterEqual] * \[DoubleStruckCapitalZ]"]]], Evaluator -> 
         Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", 
                  "0"}], "\[And]", 
                RowBox[{
                  StyleBox["y", "ExpressionVariable"], "\[GreaterEqual]", 
                  "0"}]}], "\[Implies]", 
              RowBox[{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], " ", 
                  StyleBox["y", "ExpressionVariable"]}], "\[GreaterEqual]", 
                "0"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(\[GreaterEqual] * \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox[
             "\"(\[GreaterEqual] * \[DoubleStruckCapitalZ])\"", 
              "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:166723864"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:166723864"}, 
          "Hyperlink"]& ]}}, AutoDelete -> False, 
      GridBoxAlignment -> {"Columns" -> {Center, Left, Center}}, 
      GridBoxFrame -> {"Columns" -> False, "Rows" -> False}, 
      GridBoxItemSize -> {
       "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"],"\"\""},
   
   "Labeled",
   DisplayFunction->(GridBox[{{
       ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
       TagBox[
        ItemBox[
         PaneBox[
          TagBox[#, "SkipImageSizeLevel"], Alignment -> {Center, Baseline}, 
          BaselinePosition -> Baseline], DefaultBaseStyle -> "Labeled"], 
        "SkipImageSizeLevel"]}}, 
     GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
     AutoDelete -> False, 
     GridBoxItemSize -> {"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}},
      BaselinePosition -> {2, 1}]& ),
   InterpretationFunction->(RowBox[{"Labeled", "[", 
      RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {4,"\<\"such (0)\"\>"->
  TemplateBox[{InterpretationBox[
     TagBox[
      GridBox[{{""}}, AutoDelete -> False, 
       GridBoxAlignment -> {"Columns" -> {Center, Center, Left, Center}}, 
       GridBoxFrame -> {"Columns" -> False, "Rows" -> False}, 
       GridBoxItemSize -> {
        "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"], 
     Grid[{}, Frame -> False, Alignment -> {{Center, Center, Left, Center}}, 
      Frame -> False]],"\"\""},
   "Labeled",
   DisplayFunction->(GridBox[{{
       ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
       TagBox[
        ItemBox[
         PaneBox[
          TagBox[#, "SkipImageSizeLevel"], Alignment -> {Center, Baseline}, 
          BaselinePosition -> Baseline], DefaultBaseStyle -> "Labeled"], 
        "SkipImageSizeLevel"]}}, 
     GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
     AutoDelete -> False, 
     GridBoxItemSize -> {"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}},
      BaselinePosition -> {2, 1}]& ),
   InterpretationFunction->(RowBox[{"Labeled", "[", 
      RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {5,"\<\"the (0)\"\>"->
  TemplateBox[{InterpretationBox[
     TagBox[
      GridBox[{{""}}, AutoDelete -> False, 
       GridBoxAlignment -> {"Columns" -> {Center, Center, Left, Center}}, 
       GridBoxFrame -> {"Columns" -> False, "Rows" -> False}, 
       GridBoxItemSize -> {
        "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"], 
     Grid[{}, Frame -> False, Alignment -> {{Center, Center, Left, Center}}, 
      Frame -> False]],"\"\""},
   "Labeled",
   DisplayFunction->(GridBox[{{
       ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
       TagBox[
        ItemBox[
         PaneBox[
          TagBox[#, "SkipImageSizeLevel"], Alignment -> {Center, Baseline}, 
          BaselinePosition -> Baseline], DefaultBaseStyle -> "Labeled"], 
        "SkipImageSizeLevel"]}}, 
     GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
     AutoDelete -> False, 
     GridBoxItemSize -> {"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}},
      BaselinePosition -> {2, 1}]& ),
   InterpretationFunction->(RowBox[{"Labeled", "[", 
      RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {
  6,"\<\"problematic definitions (3)\"\>"->
  TemplateBox[{TagBox[
     GridBox[{{
        TagBox[
         GridBox[{{
            ButtonBox[
            "\"C\"", Appearance -> Automatic, ButtonFunction :> 
             CopyToClipboard[
               AngleBracket[
                StringJoin[
                TheoryTools`Private`src$951683, TheoryTools`$LabelDelimiter, 
                 "gcd \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
             ImageSize -> {20, 20}, Method -> "Preemptive"], 
            TagBox[
             TooltipBox[
             "\"\[LightBulb]\"", "\"{1, 1}\"", TooltipStyle -> "TextStyling"],
              Annotation[#, "{1, 1}", "Tooltip"]& ], 
            TagBox[
             FormBox[
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}]}, {
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}]}}]], 
                RowBox[{
                  RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[NotEqual]", "0"}],
                     "\[Or]", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                    "0"}]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["g", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
                    RowBox[{
                    RowBox[{"isGCD", "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["g", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["g1", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
                    RowBox[{
                    RowBox[{"isGCD", "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["g1", "ExpressionVariable"]}], "]"}], 
                    "\[Implies]", 
                    RowBox[{
                    StyleBox["g1", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["g", "ExpressionVariable"]}]}]}]}]}]}], 
                  "\[Implies]", 
                  RowBox[{
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"]}], "]"}], ":=", 
                    RowBox[{
                    UnderscriptBox["the", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["g", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
                    RowBox[{"isGCD", "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["g", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
            TagBox[
             ButtonBox[
              
              PaneSelectorBox[{
               False -> "\"(gcd \[DoubleStruckCapitalZ])\"", True -> 
                StyleBox[
                 "\"(gcd \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
               Dynamic[
                CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
               FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
              "Hyperlink", 
              ButtonData :> {
               "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:30791925"}, 
              ButtonNote -> 
              "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb"], 
             Annotation[#, {
              "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:30791925"}, 
              "Hyperlink"]& ]}, {
            ButtonBox[
            "\"C\"", Appearance -> Automatic, ButtonFunction :> 
             CopyToClipboard[
               AngleBracket[
                StringJoin[
                TheoryTools`Private`src$951753, TheoryTools`$LabelDelimiter, 
                 "gcd 0 \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
             ImageSize -> {20, 20}, Method -> "Preemptive"], "\"\"", 
            TagBox[
             FormBox[
              RowBox[{
                RowBox[{"gcd", "[", 
                  RowBox[{"0", ",", "0"}], "]"}], ":=", 
                "0"}], $CellContext`TheoremaForm], DisplayForm], 
            TagBox[
             ButtonBox[
              
              PaneSelectorBox[{
               False -> "\"(gcd 0 \[DoubleStruckCapitalZ])\"", True -> 
                StyleBox[
                 "\"(gcd 0 \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
               Dynamic[
                CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
               FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
              "Hyperlink", 
              ButtonData :> {
               "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:645363742"}, 
              ButtonNote -> 
              "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb"], 
             Annotation[#, {
              "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:645363742"}, 
              "Hyperlink"]& ]}}, AutoDelete -> False, 
          GridBoxAlignment -> {"Columns" -> {Center, Center, Left, Center}}, 
          GridBoxFrame -> {"ColumnsIndexed" -> {{{1, -1}, {1, -1}} -> True}}, 
          GridBoxItemSize -> {
           "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"]}, {
        TagBox[
         GridBox[{{
            ButtonBox[
            "\"C\"", Appearance -> Automatic, ButtonFunction :> 
             CopyToClipboard[
               AngleBracket[
                StringJoin[
                TheoryTools`Private`src$951761, TheoryTools`$LabelDelimiter, 
                 "abs \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
             ImageSize -> {20, 20}, Method -> "Preemptive"], 
            TagBox[
             TooltipBox[
             "\"\[LightBulb]\"", "\"{1, 1}\"", TooltipStyle -> "TextStyling"],
              Annotation[#, "{1, 1}", "Tooltip"]& ], 
            TagBox[
             FormBox[
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}]}}]], 
                RowBox[{
                  RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["x", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], ":=", 
                  RowBox[{"\[Piecewise]", 
                    GridBox[{{
                    StyleBox["x", "ExpressionVariable"], "\[DoubleLeftArrow]", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", 
                    "0"}]}, {
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}], "\[DoubleLeftArrow]", 
                    "True"}}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
            TagBox[
             ButtonBox[
              
              PaneSelectorBox[{
               False -> "\"(abs \[DoubleStruckCapitalZ])\"", True -> 
                StyleBox[
                 "\"(abs \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
               Dynamic[
                CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
               FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
              "Hyperlink", 
              ButtonData :> {
               "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:9097152"}, 
              ButtonNote -> 
              "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb"], 
             Annotation[#, {
              "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:9097152"}, 
              "Hyperlink"]& ]}, {
            ButtonBox[
            "\"C\"", Appearance -> Automatic, ButtonFunction :> 
             CopyToClipboard[
               AngleBracket[
                StringJoin[
                TheoryTools`Private`src$951781, TheoryTools`$LabelDelimiter, 
                 "length join"]]], Evaluator -> Automatic, 
             ImageSize -> {20, 20}, Method -> "Preemptive"], "\"\"", 
            TagBox[
             FormBox[
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 RowBox[{
                   StyleBox["T", "ExpressionVariable"], ",", 
                   StyleBox["S", "ExpressionVariable"]}]], 
                RowBox[{
                  RowBox[{"isTuple", "[", 
                    RowBox[{
                    StyleBox["T", "ExpressionVariable"], "\:22c8", 
                    StyleBox["S", "ExpressionVariable"]}], "]"}], 
                  "\[Implies]", 
                  RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    StyleBox["T", "ExpressionVariable"], "\:22c8", 
                    StyleBox["S", "ExpressionVariable"]}], 
                    "\[RightBracketingBar]"}], ":=", 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
            TagBox[
             ButtonBox[
              
              PaneSelectorBox[{
               False -> "\"(length join)\"", True -> 
                StyleBox["\"(length join)\"", "HyperlinkActive"]}, 
               Dynamic[
                CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
               FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
              "Hyperlink", 
              ButtonData :> {
               "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb", "ID:1115259658"}, ButtonNote -> 
              "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb"], 
             Annotation[#, {
              "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb", "ID:1115259658"}, 
              "Hyperlink"]& ]}}, AutoDelete -> False, 
          GridBoxAlignment -> {"Columns" -> {Center, Center, Left, Center}}, 
          GridBoxFrame -> {"ColumnsIndexed" -> {{{1, -1}, {1, -1}} -> True}}, 
          GridBoxItemSize -> {
           "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"]}, {
        TagBox[
         GridBox[{{
            ButtonBox[
            "\"C\"", Appearance -> Automatic, ButtonFunction :> 
             CopyToClipboard[
               AngleBracket[
                StringJoin[
                TheoryTools`Private`src$951822, TheoryTools`$LabelDelimiter, 
                 "index join 1"]]], Evaluator -> Automatic, 
             ImageSize -> {20, 20}, Method -> "Preemptive"], 
            TagBox[
             TooltipBox[
             "\"\[LightBulb]\"", "\"{1, 1}\"", TooltipStyle -> "TextStyling"],
              Annotation[#, "{1, 1}", "Tooltip"]& ], 
            TagBox[
             FormBox[
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 RowBox[{
                   StyleBox["T", "ExpressionVariable"], ",", 
                   StyleBox["S", "ExpressionVariable"]}]], 
                RowBox[{
                  RowBox[{"isTuple", "[", 
                    RowBox[{
                    StyleBox["T", "ExpressionVariable"], "\:22c8", 
                    StyleBox["S", "ExpressionVariable"]}], "]"}], 
                  "\[Implies]", 
                  RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    SubscriptBox[
                    RowBox[{
                    StyleBox["T", "ExpressionVariable"], "\:22c8", 
                    StyleBox["S", "ExpressionVariable"]}], 
                    StyleBox["i", "ExpressionVariable"]], ":=", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox[
                    "i", "ExpressionVariable"]]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
            TagBox[
             ButtonBox[
              
              PaneSelectorBox[{
               False -> "\"(index join 1)\"", True -> 
                StyleBox["\"(index join 1)\"", "HyperlinkActive"]}, 
               Dynamic[
                CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
               FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
              "Hyperlink", 
              ButtonData :> {
               "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb", "ID:526825714"}, ButtonNote -> 
              "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb"], 
             Annotation[#, {
              "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb", "ID:526825714"}, 
              "Hyperlink"]& ]}, {
            ButtonBox[
            "\"C\"", Appearance -> Automatic, ButtonFunction :> 
             CopyToClipboard[
               AngleBracket[
                StringJoin[
                TheoryTools`Private`src$951865, TheoryTools`$LabelDelimiter, 
                 "index join 2"]]], Evaluator -> Automatic, 
             ImageSize -> {20, 20}, Method -> "Preemptive"], "\"\"", 
            TagBox[
             FormBox[
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 RowBox[{
                   StyleBox["T", "ExpressionVariable"], ",", 
                   StyleBox["S", "ExpressionVariable"]}]], 
                RowBox[{
                  RowBox[{"isTuple", "[", 
                    RowBox[{
                    StyleBox["T", "ExpressionVariable"], "\:22c8", 
                    StyleBox["S", "ExpressionVariable"]}], "]"}], 
                  "\[Implies]", 
                  RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["S", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    SubscriptBox[
                    RowBox[{
                    StyleBox["T", "ExpressionVariable"], "\:22c8", 
                    StyleBox["S", "ExpressionVariable"]}], 
                    RowBox[{
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}], "+", 
                    StyleBox["i", "ExpressionVariable"]}]], ":=", 
                    SubscriptBox[
                    StyleBox["S", "ExpressionVariable"], 
                    StyleBox[
                    "i", "ExpressionVariable"]]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
            TagBox[
             ButtonBox[
              
              PaneSelectorBox[{
               False -> "\"(index join 2)\"", True -> 
                StyleBox["\"(index join 2)\"", "HyperlinkActive"]}, 
               Dynamic[
                CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
               FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
              "Hyperlink", 
              ButtonData :> {
               "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb", "ID:189594342"}, ButtonNote -> 
              "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb"], 
             Annotation[#, {
              "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb", "ID:189594342"}, 
              "Hyperlink"]& ]}}, AutoDelete -> False, 
          GridBoxAlignment -> {"Columns" -> {Center, Center, Left, Center}}, 
          GridBoxFrame -> {"ColumnsIndexed" -> {{{1, -1}, {1, -1}} -> True}}, 
          GridBoxItemSize -> {
           "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"]}}, 
      DefaultBaseStyle -> "Column", 
      GridBoxAlignment -> {"Columns" -> {{Left}}}, 
      GridBoxItemSize -> {
       "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Column"],
    "\"\""},
   "Labeled",
   DisplayFunction->(GridBox[{{
       ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
       TagBox[
        ItemBox[
         PaneBox[
          TagBox[#, "SkipImageSizeLevel"], Alignment -> {Center, Baseline}, 
          BaselinePosition -> Baseline], DefaultBaseStyle -> "Labeled"], 
        "SkipImageSizeLevel"]}}, 
     GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
     AutoDelete -> False, 
     GridBoxItemSize -> {"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}},
      BaselinePosition -> {2, 1}]& ),
   InterpretationFunction->(RowBox[{"Labeled", "[", 
      RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {
  7,"\<\"higher-order (9)\"\>"->
  TemplateBox[{TagBox[
     GridBox[{{
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951917, TheoryTools`$LabelDelimiter, 
             "sum base,[IR]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         TooltipBox[
         "\"\[LightBulb]\"", "\"1\"", TooltipStyle -> "TextStyling"], 
         Annotation[#, "1", "Tooltip"]& ], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
               StyleBox["f", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox[
                 SubscriptBox["\[Sum]", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", "0"}]}}]], 
                RowBox[{
                  StyleBox["f", "ExpressionVariable"], "[", 
                  StyleBox["i", "ExpressionVariable"], "]"}]}], ":=", 
              UnderscriptBox["0", 
               StyleBox[
               "\[ScriptCapitalD]", 
                "ExpressionVariable"]]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(sum base,[IR])\"", True -> 
            StyleBox["\"(sum base,[IR])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:1971224175"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:1971224175"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951928, TheoryTools`$LabelDelimiter, 
             "abstraction term \[Element]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         TooltipBox[
         "\"\[LightBulb]\"", "\"1\"", TooltipStyle -> "TextStyling"], 
         Annotation[#, "1", "Tooltip"]& ], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isSet", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}, {
                StyleBox["f", "ExpressionVariable"]}, {
                StyleBox["P", "ExpressionVariable"]}, {
                StyleBox["x", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], "\[Element]", 
                RowBox[{"{", 
                  RowBox[{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}], "}"}]}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                RowBox[{
                  RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}], "\[And]", 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["y", "ExpressionVariable"], 
                    "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(abstraction term \[Element])\"", True -> 
            StyleBox["\"(abstraction term \[Element])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:1688878042"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:1688878042"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951961, TheoryTools`$LabelDelimiter, 
             "\[Beta] reduction 2"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         TooltipBox[
         "\"\[LightBulb]\"", "\"1\"", TooltipStyle -> "TextStyling"], 
         Annotation[#, "1", "Tooltip"]& ], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["f", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                RowBox[{"(", 
                  RowBox[{
                    UnderscriptBox["\[Lambda]", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}]], 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], ")"}], 
                "[", 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], ",", 
                  StyleBox["b", "ExpressionVariable"]}], "]"}], ":=", 
              RowBox[{
                StyleBox["f", "ExpressionVariable"], "[", 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], ",", 
                  StyleBox["b", "ExpressionVariable"]}], 
                "]"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(\[Beta] reduction 2)\"", True -> 
            StyleBox["\"(\[Beta] reduction 2)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:1792564651"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:1792564651"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$951979, TheoryTools`$LabelDelimiter, 
             "TupleOf step"]]], Evaluator -> Automatic, ImageSize -> {20, 20},
          Method -> "Preemptive"], 
        TagBox[
         TooltipBox[
         "\"\[LightBulb]\"", "\"1\"", TooltipStyle -> "TextStyling"], 
         Annotation[#, "1", "Tooltip"]& ], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                StyleBox["f", "ExpressionVariable"]}, {
                StyleBox["P", "ExpressionVariable"]}, {
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "\[Element]", 
                  SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
            RowBox[{
              RowBox[{"\[LeftAngleBracket]", 
                RowBox[{
                  RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], 
                  UnderscriptBox["|", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "+", "1"}]}]}}]], 
                  RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}]}], 
                "\[RightAngleBracket]"}], ":=", 
              RowBox[{"\[Piecewise]", 
                GridBox[{{
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}]}], 
                    "\[RightAngleBracket]"}], "\:293a", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "+", "1"}], "]"}]}], 
                   "\[DoubleLeftArrow]", 
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "+", "1"}], "]"}]}, {
                  
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}]}], 
                    "\[RightAngleBracket]"}], "\[DoubleLeftArrow]", 
                   "True"}}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(TupleOf step)\"", True -> 
            StyleBox["\"(TupleOf step)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb", "ID:396789611"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb", "ID:396789611"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952031, TheoryTools`$LabelDelimiter, 
             "sum step,[IR]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         TooltipBox[
         "\"\[LightBulb]\"", "\"1\"", TooltipStyle -> "TextStyling"], 
         Annotation[#, "1", "Tooltip"]& ], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}, {
                StyleBox["f", "ExpressionVariable"]}, {
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "\[Element]", 
                  SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox[
                 SubscriptBox["\[Sum]", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "+", "1"}]}]}}]], 
                RowBox[{
                  StyleBox["f", "ExpressionVariable"], "[", 
                  StyleBox["i", "ExpressionVariable"], "]"}]}], ":=", 
              RowBox[{
                RowBox[{
                  UnderscriptBox[
                   SubscriptBox["\[Sum]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]}}]], 
                  RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}]}], 
                UnderscriptBox["+", 
                 StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], 
                RowBox[{
                  StyleBox["f", "ExpressionVariable"], "[", 
                  RowBox[{
                    StyleBox["n", "ExpressionVariable"], "+", "1"}], 
                  "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(sum step,[IR])\"", True -> 
            StyleBox["\"(sum step,[IR])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:1789188784"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:1789188784"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952062, TheoryTools`$LabelDelimiter, 
             "\[Beta] reduction"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         TooltipBox[
         "\"\[LightBulb]\"", "\"1\"", TooltipStyle -> "TextStyling"], 
         Annotation[#, "1", "Tooltip"]& ], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["f", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                RowBox[{"(", 
                  RowBox[{
                    UnderscriptBox["\[Lambda]", 
                    StyleBox["x", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], ")"}], "[", 
                
                StyleBox["a", "ExpressionVariable"], "]"}], ":=", 
              RowBox[{
                StyleBox["f", "ExpressionVariable"], "[", 
                StyleBox["a", "ExpressionVariable"], 
                "]"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(\[Beta] reduction)\"", True -> 
            StyleBox["\"(\[Beta] reduction)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:1466460153"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:1466460153"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952075, TheoryTools`$LabelDelimiter, 
             "predicate set \[Element]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         TooltipBox[
         "\"\[LightBulb]\"", "\"1\"", TooltipStyle -> "TextStyling"], 
         Annotation[#, "1", "Tooltip"]& ], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["f", "ExpressionVariable"], ",", 
               StyleBox["P", "ExpressionVariable"], ",", 
               StyleBox["Q", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{"isSet", "[", 
                RowBox[{"{", 
                  RowBox[{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], "}"}], 
                "]"}], "\[Implies]", 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 StyleBox["y", "ExpressionVariable"]], 
                RowBox[{
                  RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Element]", 
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], "}"}]}], 
                  TagBox[
                   
                   RowBox[{
                    ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                   "a:=b"], 
                  RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["Q", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], "\[And]", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], 
                    "]"}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(predicate set \[Element])\"", True -> 
            StyleBox["\"(predicate set \[Element])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:92884649"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:92884649"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952125, TheoryTools`$LabelDelimiter, 
             "UnionOf"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         TooltipBox[
         "\"\[LightBulb]\"", "\"1\"", TooltipStyle -> "TextStyling"], 
         Annotation[#, "1", "Tooltip"]& ], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["A", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["P", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox[
                 UnderscriptBox["\[Union]", 
                  GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]], 
                RowBox[{
                  StyleBox["B", "ExpressionVariable"], "[", 
                  StyleBox["a", "ExpressionVariable"], "]"}]}], ":=", 
              RowBox[{"Un", "[", 
                RowBox[{"{", 
                  RowBox[{
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}], "}"}], 
                "]"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(UnionOf)\"", True -> 
            StyleBox["\"(UnionOf)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:60334524"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:60334524"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952152, TheoryTools`$LabelDelimiter, 
             "IntersectionOf"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         TooltipBox[
         "\"\[LightBulb]\"", "\"1\"", TooltipStyle -> "TextStyling"], 
         Annotation[#, "1", "Tooltip"]& ], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["A", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["P", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox[
                 UnderscriptBox["\[Intersection]", 
                  GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]], 
                RowBox[{
                  StyleBox["B", "ExpressionVariable"], "[", 
                  StyleBox["a", "ExpressionVariable"], "]"}]}], ":=", 
              RowBox[{"Inter", "[", 
                RowBox[{"{", 
                  RowBox[{
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}], "}"}], 
                "]"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(IntersectionOf)\"", True -> 
            StyleBox["\"(IntersectionOf)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:502023636"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:502023636"}, 
          "Hyperlink"]& ]}}, AutoDelete -> False, 
      GridBoxAlignment -> {"Columns" -> {Center, Center, Left, Center}}, 
      GridBoxFrame -> {"Columns" -> False, "Rows" -> False}, 
      GridBoxItemSize -> {
       "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"],"\"\""},
   
   "Labeled",
   DisplayFunction->(GridBox[{{
       ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
       TagBox[
        ItemBox[
         PaneBox[
          TagBox[#, "SkipImageSizeLevel"], Alignment -> {Center, Baseline}, 
          BaselinePosition -> Baseline], DefaultBaseStyle -> "Labeled"], 
        "SkipImageSizeLevel"]}}, 
     GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
     AutoDelete -> False, 
     GridBoxItemSize -> {"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}},
      BaselinePosition -> {2, 1}]& ),
   InterpretationFunction->(RowBox[{"Labeled", "[", 
      RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}, {
  8,"\<\"checked (108)\"\>"->
  TemplateBox[{TagBox[
     GridBox[{{
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952179, TheoryTools`$LabelDelimiter, 
             "algorithmic reduction ring"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["D", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"isAlgoReductionRing", "[", 
                StyleBox["D", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{"isReductionRing", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{"R9", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{"R10", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{"R11", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{"R12", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{"R13", "[", 
                  StyleBox["D", "ExpressionVariable"], 
                  "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(algorithmic reduction ring)\"", True -> 
            StyleBox["\"(algorithmic reduction ring)\"", "HyperlinkActive"]}, 
           
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:979757495"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:979757495"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952233, TheoryTools`$LabelDelimiter, 
             "reduction ring"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["D", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"isReductionRing", "[", 
                StyleBox["D", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{"isCommRing1", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{"isPartIrreflOrder", "[", 
                  RowBox[{
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], ",", 
                    StyleBox["D", "ExpressionVariable"]}], "]"}], "\[And]", 
                RowBox[{"isNoetherian", "[", 
                  RowBox[{
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], ",", 
                    StyleBox["D", "ExpressionVariable"]}], "]"}], "\[And]", 
                RowBox[{"Multipliers", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{"R0", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{"R1", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{"R2", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{"R3", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{"R4", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{"R5", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{"R6", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{"R7", "[", 
                  StyleBox["D", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{"R8", "[", 
                  StyleBox["D", "ExpressionVariable"], 
                  "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(reduction ring)\"", True -> 
            StyleBox["\"(reduction ring)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:623701544"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:623701544"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952339, TheoryTools`$LabelDelimiter, 
             "R9"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
         "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["D", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"R9", "[", 
                StyleBox["D", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox[
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                   RowBox[{
                    StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}]}]], 
                RowBox[{
                  UnderscriptBox[
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
                   RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["i", "ExpressionVariable"]}]], 
                  RowBox[{
                    UnderscriptBox["let", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["L", "ExpressionVariable"], "=", 
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["L", "ExpressionVariable"], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\:22ff", 
                    StyleBox["L", "ExpressionVariable"]}], "\[Implies]", 
                    RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[And]", 
                    
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    "\[Implies]", 
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Tilde]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["b", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\:22ff", 
                    StyleBox[
                    "L", "ExpressionVariable"]}]}]}]}]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(R9)\"", True -> 
            StyleBox["\"(R9)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1502179371"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1502179371"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952445, TheoryTools`$LabelDelimiter, 
             "RIQR lcrd"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalN]"}]}, {
                StyleBox["x", "ExpressionVariable"]}, {
                StyleBox["y", "ExpressionVariable"]}, {
                StyleBox["i", "ExpressionVariable"]}, {
                StyleBox["j", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["lcrd", 
                 RowBox[{"ReductionIQR", "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], ",", 
                  StyleBox["y", "ExpressionVariable"], ",", 
                  StyleBox["i", "ExpressionVariable"], ",", 
                  StyleBox["j", "ExpressionVariable"]}], "]"}], ":=", 
              RowBox[{"\[Piecewise]", 
                GridBox[{{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    SubscriptBox["min", 
                    UnderscriptBox["\[Succeeds]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
                    "\[RightAngleBracket]"}], "]"}], "\[RightAngleBracket]"}],
                    "\[DoubleLeftArrow]", 
                   RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["y", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["j", "ExpressionVariable"]}]}]}, {
                   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
                   "\[DoubleLeftArrow]", 
                   "True"}}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(RIQR lcrd)\"", True -> 
            StyleBox["\"(RIQR lcrd)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:943334598"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:943334598"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952511, TheoryTools`$LabelDelimiter, 
             "R10"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
         "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["D", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"R10", "[", 
                StyleBox["D", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox[
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}]], 
                RowBox[{
                  UnderscriptBox["let", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["L", "ExpressionVariable"], "=", 
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
                  RowBox[{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["L", "ExpressionVariable"], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\:22ff", 
                    StyleBox["L", "ExpressionVariable"]}], "\[Implies]", 
                    RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"]}], "]"}]}], "\[And]", 
                    
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"]}], "]"}], 
                    "\[Implies]", 
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Tilde]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["b", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\:22ff", 
                    StyleBox[
                    "L", "ExpressionVariable"]}]}]}]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(R10)\"", True -> 
            StyleBox["\"(R10)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1524689422"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1524689422"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952583, TheoryTools`$LabelDelimiter, 
             "RIQR lcrd unary"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalN]"}]}, {
                StyleBox["x", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["lcrd", 
                 RowBox[{"ReductionIQR", "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                StyleBox["x", "ExpressionVariable"], "]"}], ":=", 
              RowBox[{"\[LeftAngleBracket]", 
                RowBox[{"gcd", "[", 
                  RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], 
                "\[RightAngleBracket]"}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(RIQR lcrd unary)\"", True -> 
            StyleBox["\"(RIQR lcrd unary)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:1868934214"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:1868934214"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952606, TheoryTools`$LabelDelimiter, 
             "reducible2"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["D", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["isReducible", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], ",", 
                  StyleBox["c", "ExpressionVariable"], ",", 
                  StyleBox["m1", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], 
                  UnderscriptBox["\[Succeeds]", 
                   StyleBox["D", "ExpressionVariable"]], 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}], "\[And]", 
                RowBox[{
                  UnderscriptBox["mult", 
                   StyleBox["D", "ExpressionVariable"]], "[", 
                  RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], 
                  "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(reducible2)\"", True -> 
            StyleBox["\"(reducible2)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:115218056"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:115218056"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952635, TheoryTools`$LabelDelimiter, 
             "R11"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
         "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["D", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"R11", "[", 
                StyleBox["D", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  UnderscriptBox["let", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["R", "ExpressionVariable"], "=", 
                    RowBox[{
                    UnderscriptBox["rdm", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}]}}]], 
                  RowBox[{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["R", "ExpressionVariable"], "]"}], "\[And]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                    "\[Implies]", 
                    RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["R", "ExpressionVariable"]}], "]"}]}], "\[And]", 
                    
                    RowBox[{
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
                    RowBox[{
                    StyleBox["R", "ExpressionVariable"], "\[Equal]", 
                    UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(R11)\"", True -> 
            StyleBox["\"(R11)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:528478842"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:528478842"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952690, TheoryTools`$LabelDelimiter, 
             "RIQR 0"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalN]"}]}}]], 
            RowBox[{
              UnderscriptBox["0", 
               RowBox[{"ReductionIQR", "[", 
                 StyleBox["n", "ExpressionVariable"], "]"}]], ":=", 
              "0"}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(RIQR 0)\"", True -> 
            StyleBox["\"(RIQR 0)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:447457872"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:447457872"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952700, TheoryTools`$LabelDelimiter, 
             "RIQR mult"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalN]"}]}, {
                StyleBox["x", "ExpressionVariable"]}, {
                StyleBox["m", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["mult", 
                 RowBox[{"ReductionIQR", "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                RowBox[{
                  StyleBox["m", "ExpressionVariable"], ",", 
                  StyleBox["x", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  UnderscriptBox["mult", 
                   RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                  RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", "1"}], "]"}], 
                "\[Or]", 
                RowBox[{
                  UnderscriptBox["mult", 
                   RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                  RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"], ",", "2"}], 
                  "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(RIQR mult)\"", True -> 
            StyleBox["\"(RIQR mult)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:445166729"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:445166729"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952740, TheoryTools`$LabelDelimiter, 
             "tuple E"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{"isTuple", "[", 
                StyleBox["T", "ExpressionVariable"], "]"}], "\[Implies]", 
              RowBox[{
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], "\:22ff", 
                  StyleBox["T", "ExpressionVariable"]}], 
                TagBox[
                 
                 RowBox[{
                  ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                   "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                 "a:=b"], 
                RowBox[{
                  UnderscriptBox["\[Exists]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox[
                    "i", "ExpressionVariable"]]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(tuple E)\"", True -> 
            StyleBox["\"(tuple E)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb", "ID:1281000450"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb", "ID:1281000450"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952775, TheoryTools`$LabelDelimiter, 
             "reducible1"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["D", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["isReducible", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], ",", 
                  StyleBox["c", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  UnderscriptBox["isReducible", 
                   StyleBox["D", "ExpressionVariable"]], "[", 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m", "ExpressionVariable"]}], 
                  "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(reducible1)\"", True -> 
            StyleBox["\"(reducible1)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1537297721"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1537297721"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952792, TheoryTools`$LabelDelimiter, 
             "irrelative"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["D", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"], ",", 
               StyleBox["m2", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["areIrrelative", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["c1", "ExpressionVariable"], ",", 
                  StyleBox["c2", "ExpressionVariable"], ",", 
                  StyleBox["m1", "ExpressionVariable"], ",", 
                  StyleBox["m2", "ExpressionVariable"], ",", 
                  StyleBox["i", "ExpressionVariable"], ",", 
                  StyleBox["j", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  StyleBox["c1", "ExpressionVariable"], "\[Equal]", 
                  StyleBox["c2", "ExpressionVariable"]}], "\[Implies]", 
                RowBox[{
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["j", "ExpressionVariable"]}], "\[And]", 
                  RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], "\[And]", 
                  RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(irrelative)\"", True -> 
            StyleBox["\"(irrelative)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1867242863"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1867242863"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952837, TheoryTools`$LabelDelimiter, 
             "common reducible"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["D", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["isCommonReducible", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["c1", "ExpressionVariable"], ",", 
                  StyleBox["c2", "ExpressionVariable"], ",", 
                  StyleBox["a", "ExpressionVariable"], ",", 
                  StyleBox["i", "ExpressionVariable"], ",", 
                  StyleBox["j", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m3", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m4", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["m3", "ExpressionVariable"]}], "]"}], "\[And]", 
                  RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["m4", "ExpressionVariable"]}], "]"}], "\[And]", 
                  RowBox[{
                    UnderscriptBox["areIrrelative", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["m3", "ExpressionVariable"], ",", 
                    StyleBox["m4", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(common reducible)\"", True -> 
            StyleBox["\"(common reducible)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1907029362"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1907029362"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952880, TheoryTools`$LabelDelimiter, 
             "non-trivial common reducible"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["D", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["isNTCR", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["c1", "ExpressionVariable"], ",", 
                  StyleBox["c2", "ExpressionVariable"], ",", 
                  StyleBox["a", "ExpressionVariable"], ",", 
                  StyleBox["i", "ExpressionVariable"], ",", 
                  StyleBox["j", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  UnderscriptBox["isCommonReducible", 
                   StyleBox["D", "ExpressionVariable"]], "[", 
                  RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                RowBox[{"\[Not]", 
                  RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m3", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m4", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["areIrrelative", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["m3", "ExpressionVariable"], ",", 
                    StyleBox["m4", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["m3", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["m4", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m3", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c1", "ExpressionVariable"]}]}], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["m4", "ExpressionVariable"]}], "]"}], "\[Or]", 
                    RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m4", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c2", "ExpressionVariable"]}]}], ",", 
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["m3", "ExpressionVariable"]}], "]"}]}]}]}], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(non-trivial common reducible)\"", True -> 
            StyleBox[
             "\"(non-trivial common reducible)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:2013437879"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:2013437879"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$952965, TheoryTools`$LabelDelimiter, 
             "minimal non-trivial common reducible"]]], Evaluator -> 
         Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["D", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["isMNTCR", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["c1", "ExpressionVariable"], ",", 
                  StyleBox["c2", "ExpressionVariable"], ",", 
                  StyleBox["a", "ExpressionVariable"], ",", 
                  StyleBox["i", "ExpressionVariable"], ",", 
                  StyleBox["j", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  UnderscriptBox["isNTCR", 
                   StyleBox["D", "ExpressionVariable"]], "[", 
                  RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a0", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["a0", "ExpressionVariable"]}], "\[Implies]", 
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    UnderscriptBox["isNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["a0", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(minimal non-trivial common reducible)\"", True -> 
            StyleBox[
             "\"(minimal non-trivial common reducible)\"", 
              "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:991122907"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:991122907"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953012, TheoryTools`$LabelDelimiter, 
             "critical pair multipliers"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["D", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"], ",", 
               StyleBox["m2", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["areCPMultipliers", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["m1", "ExpressionVariable"], ",", 
                  StyleBox["m2", "ExpressionVariable"], ",", 
                  StyleBox["c1", "ExpressionVariable"], ",", 
                  StyleBox["c2", "ExpressionVariable"], ",", 
                  StyleBox["a", "ExpressionVariable"], ",", 
                  StyleBox["i", "ExpressionVariable"], ",", 
                  StyleBox["j", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  UnderscriptBox["isMNTCR", 
                   StyleBox["D", "ExpressionVariable"]], "[", 
                  RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                RowBox[{
                  UnderscriptBox["areIrrelative", 
                   StyleBox["D", "ExpressionVariable"]], "[", 
                  RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                RowBox[{
                  UnderscriptBox["isReducible", 
                   StyleBox["D", "ExpressionVariable"]], "[", 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["m1", "ExpressionVariable"]}], "]"}], "\[And]", 
                RowBox[{
                  UnderscriptBox["isReducible", 
                   StyleBox["D", "ExpressionVariable"]], "[", 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], 
                  "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(critical pair multipliers)\"", True -> 
            StyleBox["\"(critical pair multipliers)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:779388596"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:779388596"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953063, TheoryTools`$LabelDelimiter, 
             "R12"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
         "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["D", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"R12", "[", 
                StyleBox["D", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox[
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                   RowBox[{
                    StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}]}]], 
                RowBox[{
                  UnderscriptBox[
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "<", 
                    StyleBox["j", "ExpressionVariable"]}]], 
                  RowBox[{
                    UnderscriptBox["let", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["L", "ExpressionVariable"], "=", 
                    RowBox[{
                    UnderscriptBox["lcrd", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}}]], 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["L", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["p1", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["p2", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["cpm", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    StyleBox["L", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                    
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["p1", "ExpressionVariable"], ",", 
                    StyleBox["p2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["areCPMultipliers", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["p1", "ExpressionVariable"], ",", 
                    StyleBox["p2", "ExpressionVariable"], ",", 
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    StyleBox["L", "ExpressionVariable"], 
                    StyleBox["k", "ExpressionVariable"]], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(R12)\"", True -> 
            StyleBox["\"(R12)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:415149625"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:415149625"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953159, TheoryTools`$LabelDelimiter, 
             "RIQR multN"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalN]"}]}, {
                StyleBox["x", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["multN", 
                 RowBox[{"ReductionIQR", "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                StyleBox["x", "ExpressionVariable"], "]"}], ":=", 
              "2"}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(RIQR multN)\"", True -> 
            StyleBox["\"(RIQR multN)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:550500630"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:550500630"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953173, TheoryTools`$LabelDelimiter, 
             "RIQR cpm"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalN]"}]}, {
                StyleBox["x", "ExpressionVariable"]}, {
                StyleBox["y", "ExpressionVariable"]}, {
                StyleBox["m", "ExpressionVariable"]}, {
                StyleBox["i", "ExpressionVariable"]}, {
                StyleBox["j", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["cpm", 
                 RowBox[{"ReductionIQR", "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], ",", 
                  StyleBox["y", "ExpressionVariable"], ",", 
                  StyleBox["m", "ExpressionVariable"], ",", 
                  StyleBox["i", "ExpressionVariable"], ",", 
                  StyleBox["j", "ExpressionVariable"]}], "]"}], ":=", 
              RowBox[{"\[Piecewise]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["let", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], "=", 
                    RowBox[{
                    UnderscriptBox["rdm", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}]}}]], 
                    RowBox[{"\[Piecewise]", 
                    GridBox[{{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                    RowBox[{"quo", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], "]"}]}]}],
                     "\[RightAngleBracket]"}], "\[DoubleLeftArrow]", 
                    RowBox[{
                    UnderscriptBox["mult", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"], ",", "1"}], "]"}]}, {
                    
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["+", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                    RowBox[{"quo", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], "]"}]}], 
                    ",", 
                    StyleBox["m1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\[DoubleLeftArrow]", 
                    "True"}}]}]}], "\[DoubleLeftArrow]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["y", "ExpressionVariable"]}]}, {
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["rdm", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["rdm", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
                    "\[RightAngleBracket]"}], "\[DoubleLeftArrow]", 
                   "True"}}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(RIQR cpm)\"", True -> 
            StyleBox["\"(RIQR cpm)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:424764915"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:424764915"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953301, TheoryTools`$LabelDelimiter, 
             "isClosed"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
               StyleBox["\[SmallCircle]", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{"isClosed", "[", 
                RowBox[{
                  StyleBox["\[SmallCircle]", "ExpressionVariable"], ",", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  UnderscriptBox["\[Element]", 
                   StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                  
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[SmallCircle]", 
                    StyleBox["y", "ExpressionVariable"]}], 
                  "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isClosed)\"", True -> 
            StyleBox["\"(isClosed)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:1523090696"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:1523090696"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953321, TheoryTools`$LabelDelimiter, 
             "isNeutral"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
               StyleBox["n", "ExpressionVariable"], ",", 
               StyleBox["\[SmallCircle]", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{"isNeutral", "[", 
                RowBox[{
                  StyleBox["\[SmallCircle]", "ExpressionVariable"], ",", 
                  StyleBox["n", "ExpressionVariable"], ",", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  UnderscriptBox["\[Element]", 
                   StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                  
                  StyleBox["n", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[SmallCircle]", 
                    StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
                    StyleBox["x", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "\[SmallCircle]", 
                    StyleBox["x", "ExpressionVariable"]}], "\[Equal]", 
                    StyleBox[
                    "x", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isNeutral)\"", True -> 
            StyleBox["\"(isNeutral)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:665775613"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:665775613"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953362, TheoryTools`$LabelDelimiter, 
             "isAssoc"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
               StyleBox["\[SmallCircle]", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{"isAssoc", "[", 
                RowBox[{
                  StyleBox["\[SmallCircle]", "ExpressionVariable"], ",", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[SmallCircle]", 
                    StyleBox["y", "ExpressionVariable"]}], "\[SmallCircle]", 
                    StyleBox["z", "ExpressionVariable"]}], "\[Equal]", 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[SmallCircle]", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[SmallCircle]", 
                    StyleBox[
                    "z", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isAssoc)\"", True -> 
            StyleBox["\"(isAssoc)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:154995644"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:154995644"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953402, TheoryTools`$LabelDelimiter, 
             "isDistr"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
               StyleBox["\[SmallCircle]", "ExpressionVariable"], ",", 
               StyleBox["\[Diamond]", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{"isDistr", "[", 
                RowBox[{
                  StyleBox["\[SmallCircle]", "ExpressionVariable"], ",", 
                  StyleBox["\[Diamond]", "ExpressionVariable"], ",", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[SmallCircle]", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Diamond]", 
                    StyleBox["z", "ExpressionVariable"]}]}], "\[Equal]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[SmallCircle]", 
                    StyleBox["y", "ExpressionVariable"]}], "\[Diamond]", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[SmallCircle]", 
                    StyleBox["z", "ExpressionVariable"]}]}]}], "\[And]", 
                  RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Diamond]", 
                    StyleBox["z", "ExpressionVariable"]}], "\[SmallCircle]", 
                    StyleBox["x", "ExpressionVariable"]}], "\[Equal]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[SmallCircle]", 
                    StyleBox["x", "ExpressionVariable"]}], "\[Diamond]", 
                    RowBox[{
                    StyleBox["z", "ExpressionVariable"], "\[SmallCircle]", 
                    StyleBox[
                    "x", "ExpressionVariable"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isDistr)\"", True -> 
            StyleBox["\"(isDistr)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:1959021758"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:1959021758"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953480, TheoryTools`$LabelDelimiter, 
             "isComm"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
               StyleBox["\[SmallCircle]", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{"isComm", "[", 
                RowBox[{
                  StyleBox["\[SmallCircle]", "ExpressionVariable"], ",", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[SmallCircle]", 
                    StyleBox["y", "ExpressionVariable"]}], "\[Equal]", 
                  RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[SmallCircle]", 
                    StyleBox[
                    "x", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isComm)\"", True -> 
            StyleBox["\"(isComm)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:2052928557"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:2052928557"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953508, TheoryTools`$LabelDelimiter, 
             "isInverse"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
               StyleBox["n", "ExpressionVariable"], ",", 
               StyleBox["\[SmallCircle]", "ExpressionVariable"], ",", 
               StyleBox["\[Square]", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{"isInverse", "[", 
                RowBox[{
                  StyleBox["\[SmallCircle]", "ExpressionVariable"], ",", 
                  StyleBox["\[Square]", "ExpressionVariable"], ",", 
                  StyleBox["n", "ExpressionVariable"], ",", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    RowBox[{"\[Square]", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}], "\[And]", 
                  RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[SmallCircle]", 
                    RowBox[{"\[Square]", 
                    StyleBox["x", "ExpressionVariable"]}]}], "\[Equal]", 
                    StyleBox["n", "ExpressionVariable"]}], "\[And]", 
                  RowBox[{
                    RowBox[{
                    RowBox[{"\[Square]", 
                    StyleBox["x", "ExpressionVariable"]}], "\[SmallCircle]", 
                    StyleBox["x", "ExpressionVariable"]}], "\[Equal]", 
                    StyleBox[
                    "n", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isInverse)\"", True -> 
            StyleBox["\"(isInverse)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:1343138992"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:1343138992"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953558, TheoryTools`$LabelDelimiter, 
             "isRing"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"isRing", "[", 
                StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{"isDomain", "[", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], "]"}], 
                "\[And]", 
                RowBox[{"isClosed", "[", 
                  RowBox[{
                    UnderscriptBox["+", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], ",", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                  "]"}], "\[And]", 
                RowBox[{"isClosed", "[", 
                  RowBox[{
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], ",", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                  "]"}], "\[And]", 
                RowBox[{"isAssoc", "[", 
                  RowBox[{
                    UnderscriptBox["+", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], ",", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                  "]"}], "\[And]", 
                RowBox[{"isAssoc", "[", 
                  RowBox[{
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], ",", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                  "]"}], "\[And]", 
                RowBox[{"isNeutral", "[", 
                  RowBox[{
                    UnderscriptBox["+", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], ",", 
                    UnderscriptBox["0", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], ",", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                  "]"}], "\[And]", 
                RowBox[{"isComm", "[", 
                  RowBox[{
                    UnderscriptBox["+", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], ",", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                  "]"}], "\[And]", 
                RowBox[{"isDistr", "[", 
                  RowBox[{
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], ",", 
                    UnderscriptBox["+", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], ",", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                  "]"}], "\[And]", 
                RowBox[{"isInverse", "[", 
                  RowBox[{
                    UnderscriptBox["+", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], ",", 
                    UnderscriptBox["Minus", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], ",", 
                    UnderscriptBox["0", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], ",", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                  "]"}], "\[And]", 
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], 
                    StyleBox["y", "ExpressionVariable"]}], "\[Equal]", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], 
                    UnderscriptBox["+", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isRing)\"", True -> 
            StyleBox["\"(isRing)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:1169392713"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:1169392713"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953664, TheoryTools`$LabelDelimiter, 
             "isCommRing1"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"isCommRing1", "[", 
                StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{"isRing", "[", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], "]"}], 
                "\[And]", 
                RowBox[{"isNeutral", "[", 
                  RowBox[{
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], ",", 
                    UnderscriptBox["1", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], ",", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                  "]"}], "\[And]", 
                RowBox[{"isComm", "[", 
                  RowBox[{
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], ",", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                  "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isCommRing1)\"", True -> 
            StyleBox["\"(isCommRing1)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:812208573"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:812208573"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953700, TheoryTools`$LabelDelimiter, 
             "isDomain"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["D", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"isDomain", "[", 
                StyleBox["D", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{"isSet", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox[
                    "A", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isDomain)\"", True -> 
            StyleBox["\"(isDomain)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:170473529"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:170473529"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953718, TheoryTools`$LabelDelimiter, 
             "RIQR \[Element]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalN]"}]}, {
                StyleBox["x", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["\[Element]", 
                 RowBox[{"ReductionIQR", "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                StyleBox["x", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                StyleBox["x", "ExpressionVariable"], "\[Element]", 
                SubscriptBox["\[DoubleStruckCapitalN]", 
                 RowBox[{"0", ",", "\[Ellipsis]", ",", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], "-", 
                    "1"}]}]]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(RIQR \[Element])\"", True -> 
            StyleBox["\"(RIQR \[Element])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:1587087117"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:1587087117"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953738, TheoryTools`$LabelDelimiter, 
             "RIQR +"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalN]"}]}, {
                StyleBox["x", "ExpressionVariable"]}, {
                StyleBox["y", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], 
                UnderscriptBox["+", 
                 RowBox[{"ReductionIQR", "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]], 
                StyleBox["y", "ExpressionVariable"]}], ":=", 
              RowBox[{"mod", "[", 
                RowBox[{
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], "+", 
                    StyleBox["y", "ExpressionVariable"]}], ",", 
                  StyleBox["n", "ExpressionVariable"]}], 
                "]"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(RIQR +)\"", True -> 
            StyleBox["\"(RIQR +)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:1939288109"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:1939288109"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953766, TheoryTools`$LabelDelimiter, 
             "RIQR -"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalN]"}]}, {
                StyleBox["x", "ExpressionVariable"]}, {
                StyleBox["y", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], 
                UnderscriptBox["-", 
                 RowBox[{"ReductionIQR", "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]], 
                StyleBox["y", "ExpressionVariable"]}], ":=", 
              RowBox[{"mod", "[", 
                RowBox[{
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], "-", 
                    StyleBox["y", "ExpressionVariable"]}], ",", 
                  StyleBox["n", "ExpressionVariable"]}], 
                "]"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(RIQR -)\"", True -> 
            StyleBox["\"(RIQR -)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:915824852"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:915824852"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953794, TheoryTools`$LabelDelimiter, 
             "RIQR - unary"]]], Evaluator -> Automatic, ImageSize -> {20, 20},
          Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalN]"}]}, {
                StyleBox["x", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["Minus", 
                 RowBox[{"ReductionIQR", "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                StyleBox["x", "ExpressionVariable"], "]"}], ":=", 
              RowBox[{"mod", "[", 
                RowBox[{
                  RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{"-", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}], ",", 
                  StyleBox["n", "ExpressionVariable"]}], 
                "]"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(RIQR - unary)\"", True -> 
            StyleBox["\"(RIQR - unary)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:370600110"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:370600110"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953821, TheoryTools`$LabelDelimiter, 
             "RIQR *"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalN]"}]}, {
                StyleBox["x", "ExpressionVariable"]}, {
                StyleBox["y", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], 
                UnderscriptBox["*", 
                 RowBox[{"ReductionIQR", "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]], 
                StyleBox["y", "ExpressionVariable"]}], ":=", 
              RowBox[{"mod", "[", 
                RowBox[{
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], " ", 
                    StyleBox["y", "ExpressionVariable"]}], ",", 
                  StyleBox["n", "ExpressionVariable"]}], 
                "]"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(RIQR *)\"", True -> 
            StyleBox["\"(RIQR *)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:344363438"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:344363438"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953849, TheoryTools`$LabelDelimiter, 
             "isPartIrreflOrder"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
               StyleBox["\[Succeeds]", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{"isPartIrreflOrder", "[", 
                RowBox[{
                  StyleBox["\[Succeeds]", "ExpressionVariable"], ",", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Succeeds]", 
                    StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["y", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Succeeds]", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}]}]}], "\[And]", 
                  RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Succeeds]", 
                    StyleBox["y", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Succeeds]", 
                    StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Succeeds]", 
                    StyleBox[
                    "z", "ExpressionVariable"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isPartIrreflOrder)\"", True -> 
            StyleBox["\"(isPartIrreflOrder)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:1500614048"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:1500614048"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953918, TheoryTools`$LabelDelimiter, 
             "RIQR \[Succeeds]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalN]"}]}, {
                StyleBox["x", "ExpressionVariable"]}, {
                StyleBox["y", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], 
                UnderscriptBox["\[Succeeds]", 
                 RowBox[{"ReductionIQR", "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]], 
                StyleBox["y", "ExpressionVariable"]}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                StyleBox["x", "ExpressionVariable"], ">", 
                StyleBox[
                "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(RIQR \[Succeeds])\"", True -> 
            StyleBox["\"(RIQR \[Succeeds])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:1495376438"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:1495376438"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953937, TheoryTools`$LabelDelimiter, 
             "membership integer domain"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["\[Element]", 
                 SubscriptBox["\[DoubleStruckCapitalZ]", 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", "\[Ellipsis]", 
                    ",", 
                    StyleBox["b", "ExpressionVariable"]}]]], "[", 
                StyleBox["i", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                StyleBox["i", "ExpressionVariable"], "\[Element]", 
                SubscriptBox["\[DoubleStruckCapitalZ]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", "\[Ellipsis]", 
                   ",", 
                   StyleBox[
                   "b", "ExpressionVariable"]}]]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(membership integer domain)\"", True -> 
            StyleBox["\"(membership integer domain)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1548965507"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:1548965507"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953950, TheoryTools`$LabelDelimiter, 
             "membership integer domain II"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["\[Element]", 
                 SubscriptBox["\[DoubleStruckCapitalZ]", 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", "\[Ellipsis]", 
                    ",", "\[Infinity]"}]]], "[", 
                StyleBox["i", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                StyleBox["i", "ExpressionVariable"], "\[Element]", 
                SubscriptBox["\[DoubleStruckCapitalZ]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], ",", "\[Ellipsis]", 
                   ",", "\[Infinity]"}]]}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(membership integer domain II)\"", True -> 
            StyleBox[
             "\"(membership integer domain II)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:757090774"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:757090774"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953960, TheoryTools`$LabelDelimiter, 
             "isMin"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["A", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["\[Succeeds]", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{"isMin", "[", 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], ",", 
                  StyleBox["A", "ExpressionVariable"], ",", 
                  StyleBox["\[Succeeds]", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                RowBox[{"\[Not]", 
                  RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Succeeds]", 
                    StyleBox["b", "ExpressionVariable"]}], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isMin)\"", True -> 
            StyleBox["\"(isMin)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:1924210441"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:1924210441"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$953983, TheoryTools`$LabelDelimiter, 
             "DomainSets decision predicate"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
               StyleBox["x", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["\[Element]", 
                 RowBox[{"DomainSets", "[", 
                   StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], 
                   "]"}]], "[", 
                StyleBox["x", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{"isSet", "[", 
                  StyleBox["x", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Element]", 
                    StyleBox["x", "ExpressionVariable"]}]}}]], 
                  RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], 
                    "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(DomainSets decision predicate)\"", True -> 
            StyleBox[
             "\"(DomainSets decision predicate)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:88731859"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:88731859"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$954008, TheoryTools`$LabelDelimiter, 
             "multipliers"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["D", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"Multipliers", "[", 
                StyleBox["D", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox[
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}]], 
                RowBox[{
                  RowBox[{
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}], "\[And]", 
                  RowBox[{
                    RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                    "\[DoubleLeftRightArrow]", 
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
                    RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(multipliers)\"", True -> 
            StyleBox["\"(multipliers)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1960646232"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1960646232"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$954054, TheoryTools`$LabelDelimiter, 
             "R0"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
         "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["D", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"R0", "[", 
                StyleBox["D", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox[
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}]], 
                RowBox[{
                  UnderscriptBox["mult", 
                   StyleBox["D", "ExpressionVariable"]], "[", 
                  RowBox[{
                    UnderscriptBox["1", 
                    StyleBox["D", "ExpressionVariable"]], ",", 
                    StyleBox["c", "ExpressionVariable"]}], 
                  "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(R0)\"", True -> 
            StyleBox["\"(R0)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1496065688"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1496065688"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$954075, TheoryTools`$LabelDelimiter, 
             "RIQR 1"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalN]"}]}}]], 
            RowBox[{
              UnderscriptBox["1", 
               RowBox[{"ReductionIQR", "[", 
                 StyleBox["n", "ExpressionVariable"], "]"}]], ":=", 
              RowBox[{"mod", "[", 
                RowBox[{"1", ",", 
                  StyleBox["n", "ExpressionVariable"]}], 
                "]"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(RIQR 1)\"", True -> 
            StyleBox["\"(RIQR 1)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:815094494"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:815094494"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$954092, TheoryTools`$LabelDelimiter, 
             "RIQR mult 1"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalN]"}]}, {
                StyleBox["x", "ExpressionVariable"]}, {
                StyleBox["m", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["mult", 
                 RowBox[{"ReductionIQR", "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                RowBox[{
                  StyleBox["m", "ExpressionVariable"], ",", 
                  StyleBox["x", "ExpressionVariable"], ",", "1"}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  StyleBox["m", "ExpressionVariable"], "\[GreaterEqual]", 
                  "1"}], "\[And]", 
                RowBox[{
                  RowBox[{"quo", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}]}], "]"}], 
                  ">", 
                  StyleBox[
                  "m", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(RIQR mult 1)\"", True -> 
            StyleBox["\"(RIQR mult 1)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:1793329966"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:1793329966"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$954133, TheoryTools`$LabelDelimiter, 
             "R1"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
         "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["D", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"R1", "[", 
                StyleBox["D", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox[
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}]], 
                RowBox[{
                  RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                  "\[Implies]", 
                  RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["Minus", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(R1)\"", True -> 
            StyleBox["\"(R1)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:975892043"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:975892043"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$954167, TheoryTools`$LabelDelimiter, 
             "RIQR mult 2"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalN]"}]}, {
                StyleBox["x", "ExpressionVariable"]}, {
                StyleBox["m", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["mult", 
                 RowBox[{"ReductionIQR", "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                RowBox[{
                  StyleBox["m", "ExpressionVariable"], ",", 
                  StyleBox["x", "ExpressionVariable"], ",", "2"}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["mult", 
                 RowBox[{"ReductionIQR", "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                RowBox[{
                  RowBox[{
                    UnderscriptBox["Minus", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], ",", 
                  StyleBox["x", "ExpressionVariable"], ",", "1"}], 
                "]"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(RIQR mult 2)\"", True -> 
            StyleBox["\"(RIQR mult 2)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:126969281"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:126969281"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$954202, TheoryTools`$LabelDelimiter, 
             "R2"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
         "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["D", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"R2", "[", 
                StyleBox["D", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox[
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}]], 
                RowBox[{
                  RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                  "\[Implies]", 
                  RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(R2)\"", True -> 
            StyleBox["\"(R2)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1150364990"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1150364990"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$954236, TheoryTools`$LabelDelimiter, 
             "R3"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
         "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["D", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"R3", "[", 
                StyleBox["D", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox[
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}]}]], 
                RowBox[{
                  UnderscriptBox["\[Exists]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["T", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[And]", 
                    
                    RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["T", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    SubscriptBox[
                    StyleBox["T", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "\[Equal]", 
                    StyleBox[
                    "b", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(R3)\"", True -> 
            StyleBox["\"(R3)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:852396718"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:852396718"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$954308, TheoryTools`$LabelDelimiter, 
             "R4"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
         "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["D", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"R4", "[", 
                StyleBox["D", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[Implies]", 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    UnderscriptBox["0", 
                    StyleBox[
                    "D", "ExpressionVariable"]]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(R4)\"", True -> 
            StyleBox["\"(R4)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1907869025"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1907869025"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$954331, TheoryTools`$LabelDelimiter, 
             "reduction modulo element using multiplier"]]], Evaluator -> 
         Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                StyleBox["D", "ExpressionVariable"]}, {
                StyleBox["a", "ExpressionVariable"]}, {
                StyleBox["b", "ExpressionVariable"]}, {
                RowBox[{
                  UnderscriptBox["\[Element]", 
                   StyleBox["D", "ExpressionVariable"]], "[", 
                  StyleBox["c", "ExpressionVariable"], "]"}]}, {
                StyleBox["m", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["a", "ExpressionVariable"], 
                UnderscriptBox[
                 SubscriptBox["\[Rule]", 
                  RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}]], 
                 StyleBox["D", "ExpressionVariable"]], 
                StyleBox["b", "ExpressionVariable"]}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}], "\[And]", 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], 
                  UnderscriptBox["\[Succeeds]", 
                   StyleBox["D", "ExpressionVariable"]], 
                  StyleBox["b", "ExpressionVariable"]}], "\[And]", 
                RowBox[{
                  UnderscriptBox["mult", 
                   StyleBox["D", "ExpressionVariable"]], "[", 
                  RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], 
                  "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(reduction modulo element using multiplier)\"", True -> 
            StyleBox[
             "\"(reduction modulo element using multiplier)\"", 
              "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:935236171"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:935236171"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$954368, TheoryTools`$LabelDelimiter, 
             "R5"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
         "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["D", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"R5", "[", 
                StyleBox["D", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
                  RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["d", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["TM", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["TN", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    UnderscriptBox["areR5Witnesses", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["TM", "ExpressionVariable"], ",", 
                    StyleBox["TN", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["d", "ExpressionVariable"], ",", 
                    StyleBox["m", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(R5)\"", True -> 
            StyleBox["\"(R5)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1446921308"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1446921308"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$954423, TheoryTools`$LabelDelimiter, 
             "R5 witnesses"]]], Evaluator -> Automatic, ImageSize -> {20, 20},
          Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["D", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["d", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"], ",", 
               StyleBox["TM", "ExpressionVariable"], ",", 
               StyleBox["TN", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["areR5Witnesses", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["TM", "ExpressionVariable"], ",", 
                  StyleBox["TN", "ExpressionVariable"], ",", 
                  StyleBox["a", "ExpressionVariable"], ",", 
                  StyleBox["b", "ExpressionVariable"], ",", 
                  StyleBox["c", "ExpressionVariable"], ",", 
                  StyleBox["d", "ExpressionVariable"], ",", 
                  StyleBox["m", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TM", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                  RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["+", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["d", "ExpressionVariable"]}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["TM", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                    ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    StyleBox["TM", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}], "\[And]", 
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                  RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], 
                    UnderscriptBox["+", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["d", "ExpressionVariable"]}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["TN", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                    ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}], "\[And]", 
                RowBox[{
                  RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TM", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    SubscriptBox[
                    StyleBox["TM", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "\[Equal]", 
                  RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["+", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TN", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    SubscriptBox[
                    StyleBox["TN", "ExpressionVariable"], 
                    StyleBox[
                    "i", "ExpressionVariable"]]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(R5 witnesses)\"", True -> 
            StyleBox["\"(R5 witnesses)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1949920189"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1949920189"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$954541, TheoryTools`$LabelDelimiter, 
             "R6"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
         "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["D", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"R6", "[", 
                StyleBox["D", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m2", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["b1", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["b2", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["b1", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}]], 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["b2", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["areCorrelative", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], "]"}]}], 
                  "\[Implies]", 
                  RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["TL", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    UnderscriptBox["isR6Witness", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["TL", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["b1", "ExpressionVariable"], ",", 
                    StyleBox["b2", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(R6)\"", True -> 
            StyleBox["\"(R6)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:208630606"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:208630606"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$954606, TheoryTools`$LabelDelimiter, 
             "R6 witness"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["D", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"], ",", 
               StyleBox["m2", "ExpressionVariable"], ",", 
               StyleBox["b1", "ExpressionVariable"], ",", 
               StyleBox["b2", "ExpressionVariable"], ",", 
               StyleBox["TM", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["isR6Witness", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["TM", "ExpressionVariable"], ",", 
                  StyleBox["a", "ExpressionVariable"], ",", 
                  StyleBox["c", "ExpressionVariable"], ",", 
                  StyleBox["m1", "ExpressionVariable"], ",", 
                  StyleBox["m2", "ExpressionVariable"], ",", 
                  StyleBox["b1", "ExpressionVariable"], ",", 
                  StyleBox["b2", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TM", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                  RowBox[{
                    UnderscriptBox["symred", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["b1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["TM", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "-", "1"}]}], "]"}], 
                    ",", 
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["b1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["TM", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    SubscriptBox[
                    StyleBox["TM", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], "]"}]}], "\[And]", 
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TM", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    UnderscriptBox["reduceM", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["b1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["TM", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}]}]}], 
                "\[And]", 
                RowBox[{
                  RowBox[{
                    RowBox[{
                    UnderscriptBox[
                    SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["TM", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
                    SubscriptBox[
                    StyleBox["TM", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]]}], 
                    UnderscriptBox["+", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["m1", "ExpressionVariable"]}], "\[Equal]", 
                  StyleBox[
                  "m2", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(R6 witness)\"", True -> 
            StyleBox["\"(R6 witness)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:2027349511"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:2027349511"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$954700, TheoryTools`$LabelDelimiter, 
             "symmetric closure of reduction relation"]]], Evaluator -> 
         Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["D", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["r", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                StyleBox["a", "ExpressionVariable"], 
                UnderscriptBox[
                 SubscriptBox["\[LeftRightArrow]", 
                  StyleBox["r", "ExpressionVariable"]], 
                 StyleBox["D", "ExpressionVariable"]], 
                StyleBox["b", "ExpressionVariable"]}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], 
                  UnderscriptBox[
                   SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                   StyleBox["D", "ExpressionVariable"]], 
                  StyleBox["b", "ExpressionVariable"]}], "\[Or]", 
                RowBox[{
                  StyleBox["b", "ExpressionVariable"], 
                  UnderscriptBox[
                   SubscriptBox["\[Rule]", 
                    StyleBox["r", "ExpressionVariable"]], 
                   StyleBox["D", "ExpressionVariable"]], 
                  StyleBox[
                  "a", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(symmetric closure of reduction relation)\"", True -> 
            StyleBox[
             "\"(symmetric closure of reduction relation)\"", 
              "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:370202704"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:370202704"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$954726, TheoryTools`$LabelDelimiter, 
             "R7"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
         "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["D", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"R7", "[", 
                StyleBox["D", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
                  RowBox[{
                    RowBox[{
                    UnderscriptBox["isNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    "\[Implies]", 
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["l", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a0", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["a0", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    StyleBox["a0", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["a", "ExpressionVariable"]}]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["a0", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m0", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m0", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], 
                    "\[Implies]", 
                    RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["m0", "ExpressionVariable"]}], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}]}], 
                    "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["a0", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["a0", "ExpressionVariable"], 
                    UnderscriptBox["+", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["+", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["d", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], 
                    UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["c", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["d", "ExpressionVariable"]}], "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["b", "ExpressionVariable"]}], 
                    UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["c", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["d", "ExpressionVariable"]}]}]}]}], "\[And]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["c2", "ExpressionVariable"]}], "\[Implies]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a0", "ExpressionVariable"], ",", 
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["m1", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a0", "ExpressionVariable"], ",", 
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["l", "ExpressionVariable"]}], "]"}]}], 
                    "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["m1", "ExpressionVariable"]}], ",", 
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["m2", "ExpressionVariable"]}], ",", 
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["m1", "ExpressionVariable"]}], ",", 
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["m2", "ExpressionVariable"]}], ",", 
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(R7)\"", True -> 
            StyleBox["\"(R7)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1150983847"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1150983847"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$954963, TheoryTools`$LabelDelimiter, 
             "R8"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
         "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["D", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"R8", "[", 
                StyleBox["D", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c0", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    UnderscriptBox["isNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c0", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"]}], "]"}], 
                  "\[Implies]", 
                  RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a0", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["a0", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    StyleBox["a0", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["a", "ExpressionVariable"]}]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c0", "ExpressionVariable"], ",", 
                    StyleBox["a0", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m0", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m0", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], 
                    "\[Implies]", 
                    RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["m0", "ExpressionVariable"]}], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}]}], 
                    "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["a0", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["a0", "ExpressionVariable"], 
                    UnderscriptBox["+", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["+", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}]}]}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["b", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["d", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], 
                    UnderscriptBox[
                    SubscriptBox["\[Rule]", 
                    StyleBox["c", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["d", "ExpressionVariable"]}], "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["b", "ExpressionVariable"]}], 
                    UnderscriptBox[
                    SubscriptBox["\[LeftRightArrow]", 
                    StyleBox["c", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox[
                    "d", "ExpressionVariable"]}]}]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(R8)\"", True -> 
            StyleBox["\"(R8)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:2106857486"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:2106857486"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955086, TheoryTools`$LabelDelimiter, 
             "min pair"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"], ",", 
               StyleBox["\[PrecedesEqual]", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                SubscriptBox["min", 
                 StyleBox["\[PrecedesEqual]", "ExpressionVariable"]], "[", 
                RowBox[{"\[LeftAngleBracket]", 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], 
                  "\[RightAngleBracket]"}], "]"}], ":=", 
              RowBox[{"\[Piecewise]", 
                GridBox[{{
                   StyleBox["y", "ExpressionVariable"], "\[DoubleLeftArrow]", 
                   
                   RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[PrecedesEqual]", 
                    StyleBox["x", "ExpressionVariable"]}]}, {
                   StyleBox["x", "ExpressionVariable"], "\[DoubleLeftArrow]", 
                   "True"}}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(min pair)\"", True -> 
            StyleBox["\"(min pair)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb", "ID:459374459"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb", "ID:459374459"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955107, TheoryTools`$LabelDelimiter, 
             "\[Tilde] multiplier"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["D", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["m", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                StyleBox["a", "ExpressionVariable"], 
                UnderscriptBox[
                 SubscriptBox["\[Tilde]", 
                  StyleBox["m", "ExpressionVariable"]], 
                 StyleBox["D", "ExpressionVariable"]], 
                StyleBox["b", "ExpressionVariable"]}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                  RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["a", "ExpressionVariable"]}]}], "\[And]", 
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["y", "ExpressionVariable"]}], 
                    "\[DoubleLeftRightArrow]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["x", "ExpressionVariable"]}], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["y", "ExpressionVariable"]}]}]}]}], "\[And]", 
                RowBox[{
                  UnderscriptBox[
                   UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    StyleBox["D", "ExpressionVariable"]]}]], 
                  RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], 
                    "\[DoubleLeftRightArrow]", 
                    RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["n", "ExpressionVariable"]}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}]}]}]}], 
                "\[And]", 
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c2", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    "\[DoubleLeftRightArrow]", 
                    RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}]}], 
                "\[And]", 
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"]}], "]"}], 
                    "\[DoubleLeftRightArrow]", 
                    RowBox[{
                    UnderscriptBox["isMNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(\[Tilde] multiplier)\"", True -> 
            StyleBox["\"(\[Tilde] multiplier)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1818365967"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1818365967"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955224, TheoryTools`$LabelDelimiter, 
             "\[Tilde]"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["D", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                StyleBox["a", "ExpressionVariable"], 
                UnderscriptBox["\[Tilde]", 
                 StyleBox["D", "ExpressionVariable"]], 
                StyleBox["b", "ExpressionVariable"]}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    UnderscriptBox["isUnit", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox[
                    SubscriptBox["\[Tilde]", 
                    StyleBox["m", "ExpressionVariable"]], 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox[
                    "b", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(\[Tilde])\"", True -> 
            StyleBox["\"(\[Tilde])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1533920470"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1533920470"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955247, TheoryTools`$LabelDelimiter, 
             "isUnit"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}, {
                RowBox[{
                  UnderscriptBox["\[Element]", 
                   StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                  
                  StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["isUnit", 
                 StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                StyleBox["u", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["v", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    StyleBox["u", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], 
                    StyleBox["v", "ExpressionVariable"]}], "\[Equal]", 
                  UnderscriptBox["1", 
                   StyleBox[
                   "\[ScriptCapitalD]", 
                    "ExpressionVariable"]]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isUnit)\"", True -> 
            StyleBox["\"(isUnit)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:1005941388"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:1005941388"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955265, TheoryTools`$LabelDelimiter, 
             "unary non-trivial common reducible"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["D", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["isNTCR", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["c", "ExpressionVariable"], ",", 
                  StyleBox["a", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  UnderscriptBox["isReducible", 
                   StyleBox["D", "ExpressionVariable"]], "[", 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
                RowBox[{"\[Not]", 
                  RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m3", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["m4", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m3", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m4", "ExpressionVariable"]}], "]"}], "\[And]", 
                    RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m4", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m3", "ExpressionVariable"]}], "]"}]}]}], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(unary non-trivial common reducible)\"", True -> 
            StyleBox[
             "\"(unary non-trivial common reducible)\"", "HyperlinkActive"]}, 
           
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1293246145"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1293246145"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955317, TheoryTools`$LabelDelimiter, 
             "unary minimal non-trivial common reducible"]]], Evaluator -> 
         Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["D", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["isMNTCR", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["c", "ExpressionVariable"], ",", 
                  StyleBox["a", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  UnderscriptBox["isNTCR", 
                   StyleBox["D", "ExpressionVariable"]], "[", 
                  RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"]}], "]"}], "\[And]", 
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["a0", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["a0", "ExpressionVariable"]}], "\[Implies]", 
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    UnderscriptBox["isNTCR", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["a0", "ExpressionVariable"]}], "]"}], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(unary minimal non-trivial common reducible)\"", True -> 
            StyleBox[
             "\"(unary minimal non-trivial common reducible)\"", 
              "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:894750987"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:894750987"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955352, TheoryTools`$LabelDelimiter, 
             "isQuoRem \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["b", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["q", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["r", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{"isQuoRem", "[", 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], ",", 
                  StyleBox["b", "ExpressionVariable"], ",", 
                  StyleBox["q", "ExpressionVariable"], ",", 
                  StyleBox["r", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                  RowBox[{
                    RowBox[{
                    StyleBox["q", "ExpressionVariable"], " ", 
                    StyleBox["b", "ExpressionVariable"]}], "+", 
                    StyleBox["r", "ExpressionVariable"]}]}], "\[And]", 
                RowBox[{
                  RowBox[{
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], ">", "0"}], "\[And]", 
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], "\[GreaterEqual]", 
                    "0"}], "\[And]", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], ">", 
                    StyleBox["r", "ExpressionVariable"]}]}], "\[Or]", 
                  RowBox[{
                    RowBox[{"0", ">", 
                    StyleBox["b", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{"0", "\[GreaterEqual]", 
                    StyleBox["r", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], ">", 
                    StyleBox[
                    "b", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isQuoRem \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox[
             "\"(isQuoRem \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:397151377"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:397151377"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955420, TheoryTools`$LabelDelimiter, 
             "isRem \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["b", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["r", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{"isRem", "[", 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], ",", 
                  StyleBox["b", "ExpressionVariable"], ",", 
                  StyleBox["r", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["q", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}]}}]], 
                RowBox[{"isQuoRem", "[", 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["r", "ExpressionVariable"]}], 
                  "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isRem \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox[
             "\"(isRem \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:104191072"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:104191072"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955444, TheoryTools`$LabelDelimiter, 
             "integer interval"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"]}]], 
            RowBox[{
              SubscriptBox["\[DoubleStruckCapitalZ]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", "\[Ellipsis]", ",", 
                 StyleBox["b", "ExpressionVariable"]}]], ":=", 
              RowBox[{"{", 
                RowBox[{
                  GridBox[{{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}]}}], "|", 
                  RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["x", "ExpressionVariable"]}]}]}], 
                "}"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(integer interval)\"", True -> 
            StyleBox["\"(integer interval)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:371888462"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:371888462"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955464, TheoryTools`$LabelDelimiter, 
             "RIQR rdm 1"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalN]"}]}, {
                StyleBox["x", "ExpressionVariable"]}, {
                StyleBox["y", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["rdm", 
                 RowBox[{"ReductionIQR", "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], ",", 
                  StyleBox["y", "ExpressionVariable"]}], "]"}], ":=", 
              RowBox[{
                UnderscriptBox["let", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["g", "ExpressionVariable"], "=", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}]}}]], 
                RowBox[{
                  UnderscriptBox["rdm", 
                   RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                  RowBox[{
                    StyleBox["y", "ExpressionVariable"], ",", 
                    RowBox[{
                    RowBox[{"quo", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["g", "ExpressionVariable"]}], "]"}], "-", "1"}], 
                    ",", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "-", 
                    RowBox[{"mod", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["g", "ExpressionVariable"]}], "]"}]}]}], 
                  "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(RIQR rdm 1)\"", True -> 
            StyleBox["\"(RIQR rdm 1)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:909359224"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:909359224"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955524, TheoryTools`$LabelDelimiter, 
             "RIQR rdm 2"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalN]"}]}, {
                StyleBox["x", "ExpressionVariable"]}, {
                StyleBox["y", "ExpressionVariable"]}, {
                StyleBox["m", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["rdm", 
                 RowBox[{"ReductionIQR", "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], ",", 
                  StyleBox["m", "ExpressionVariable"], ",", 
                  StyleBox["y", "ExpressionVariable"]}], "]"}], ":=", 
              RowBox[{"\[Piecewise]", 
                GridBox[{{
                   StyleBox["m", "ExpressionVariable"], "\[DoubleLeftArrow]", 
                   
                   RowBox[{
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                    StyleBox["x", "ExpressionVariable"]}], "\[Equal]", 
                    StyleBox["y", "ExpressionVariable"]}]}, {
                   RowBox[{
                    UnderscriptBox["rdm", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], "-", "1"}], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], 
                   "\[DoubleLeftArrow]", 
                   "True"}}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(RIQR rdm 2)\"", True -> 
            StyleBox["\"(RIQR rdm 2)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:236078685"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\IntegerQuotientRings.nb", "ID:236078685"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955574, TheoryTools`$LabelDelimiter, 
             "isFinDomain"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["D", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"isFinDomain", "[", 
                StyleBox["D", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox[
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{"isSet", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{"isFinite", "[", 
                   StyleBox["A", "ExpressionVariable"], "]"}]], 
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox[
                    "A", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isFinDomain)\"", True -> 
            StyleBox["\"(isFinDomain)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:509889183"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:509889183"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955599, TheoryTools`$LabelDelimiter, 
             "\[SquareSubset]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                StyleBox["D", "ExpressionVariable"]}, {
                RowBox[{
                  UnderscriptBox["\[Element]", 
                   RowBox[{"DomainSets", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}, {
                RowBox[{
                  UnderscriptBox["\[Element]", 
                   RowBox[{"DomainSets", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                  StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["A", "ExpressionVariable"], 
                UnderscriptBox["\[SquareSubset]", 
                 StyleBox["D", "ExpressionVariable"]], 
                StyleBox["B", "ExpressionVariable"]}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  UnderscriptBox["red", 
                   StyleBox["D", "ExpressionVariable"]], "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}], "\[Subset]", 
                RowBox[{
                  UnderscriptBox["red", 
                   StyleBox["D", "ExpressionVariable"]], "[", 
                  StyleBox["B", "ExpressionVariable"], 
                  "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(\[SquareSubset])\"", True -> 
            StyleBox["\"(\[SquareSubset])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:808948861"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:808948861"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955632, TheoryTools`$LabelDelimiter, 
             "R13"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
         "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["D", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"R13", "[", 
                StyleBox["D", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{"\[Not]", 
                RowBox[{
                  TagBox["(", "AutoParentheses"], 
                  RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"DomainSequences", "[", 
                    RowBox[{"DomainSets", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}], "]"}]], "[", 
                    StyleBox["B", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], 
                    UnderscriptBox["\[SquareSubset]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "+", "1"}], 
                    "]"}]}]}]}], 
                  TagBox[
                  ")", "AutoParentheses"]}]}]}]}], $CellContext`TheoremaForm],
          DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(R13)\"", True -> 
            StyleBox["\"(R13)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1276925313"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1276925313"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955672, TheoryTools`$LabelDelimiter, 
             "imgSeq"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                StyleBox["f", "ExpressionVariable"]}, {
                StyleBox["h", "ExpressionVariable"]}, {
                RowBox[{
                  StyleBox["i", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalN]"}]}}]], 
            RowBox[{
              RowBox[{
                RowBox[{"imgSeq", "[", 
                  RowBox[{
                    StyleBox["f", "ExpressionVariable"], ",", 
                    StyleBox["h", "ExpressionVariable"]}], "]"}], "[", 
                StyleBox["i", "ExpressionVariable"], "]"}], ":=", 
              RowBox[{
                StyleBox["h", "ExpressionVariable"], "[", 
                RowBox[{
                  StyleBox["f", "ExpressionVariable"], "[", 
                  StyleBox["i", "ExpressionVariable"], "]"}], 
                "]"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(imgSeq)\"", True -> 
            StyleBox["\"(imgSeq)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Sequences.nb", "ID:809351385"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Sequences.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Sequences.nb", "ID:809351385"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955693, TheoryTools`$LabelDelimiter, 
             "DomainSequences decision predicate"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
               StyleBox["f", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["\[Element]", 
                 RowBox[{"DomainSequences", "[", 
                   StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], 
                   "]"}]], "[", 
                StyleBox["f", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
                RowBox[{
                  UnderscriptBox["\[Element]", 
                   StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                  
                  RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}], 
                  "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(DomainSequences decision predicate)\"", True -> 
            StyleBox[
             "\"(DomainSequences decision predicate)\"", "HyperlinkActive"]}, 
           
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Sequences.nb", "ID:458427713"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Sequences.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Sequences.nb", "ID:458427713"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955709, TheoryTools`$LabelDelimiter, 
             "\[SubsetEqual]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["A", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                RowBox[{"isSet", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{"isSet", "[", 
                  StyleBox["B", "ExpressionVariable"], "]"}]}], "\[Implies]", 
              
              RowBox[{
                RowBox[{
                  StyleBox["A", "ExpressionVariable"], "\[SubsetEqual]", 
                  StyleBox["B", "ExpressionVariable"]}], 
                TagBox[
                 
                 RowBox[{
                  ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                   "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                 "a:=b"], 
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox[
                    "B", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(\[SubsetEqual])\"", True -> 
            StyleBox["\"(\[SubsetEqual])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:558452934"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:558452934"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955744, TheoryTools`$LabelDelimiter, 
             "integer interval II"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["a", "ExpressionVariable"]], 
            RowBox[{
              SubscriptBox["\[DoubleStruckCapitalZ]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", "\[Ellipsis]", ",",
                  "\[Infinity]"}]], ":=", 
              RowBox[{"{", 
                RowBox[{
                  GridBox[{{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}]}}], "|", 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", 
                    StyleBox["a", "ExpressionVariable"]}]}], 
                "}"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(integer interval II)\"", True -> 
            StyleBox["\"(integer interval II)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:106146911"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:106146911"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955754, TheoryTools`$LabelDelimiter, 
             "quo \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["b", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["q", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}]}}]], 
                RowBox[{
                  RowBox[{"isQuo", "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"]}], "]"}], "\[And]", 
                  RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["q1", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}]}}]], 
                    RowBox[{
                    RowBox[{"isQuo", "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["q1", "ExpressionVariable"]}], "]"}], 
                    "\[Implies]", 
                    RowBox[{
                    StyleBox["q1", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["q", "ExpressionVariable"]}]}]}]}]}], 
              "\[Implies]", 
              RowBox[{
                RowBox[{"quo", "[", 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"]}], "]"}], ":=", 
                RowBox[{
                  UnderscriptBox["the", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["q", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}]}}]], 
                  RowBox[{"isQuo", "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(quo \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox["\"(quo \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:740152346"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:740152346"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955810, TheoryTools`$LabelDelimiter, 
             "integer domain IV +"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                StyleBox["i", "ExpressionVariable"], 
                UnderscriptBox["+", "\[DoubleStruckCapitalZ]"], 
                StyleBox["j", "ExpressionVariable"]}], ":=", 
              RowBox[{
                StyleBox["i", "ExpressionVariable"], "+", 
                StyleBox[
                "j", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(integer domain IV +)\"", True -> 
            StyleBox["\"(integer domain IV +)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:50992623"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:50992623"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955824, TheoryTools`$LabelDelimiter, 
             "integer domain IV 0"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["0", "\[DoubleStruckCapitalZ]"], ":=", 
            "0"}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(integer domain IV 0)\"", True -> 
            StyleBox["\"(integer domain IV 0)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:134494272"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatInt.nb", "ID:134494272"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955826, TheoryTools`$LabelDelimiter, 
             "reduction with several multipliers"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                StyleBox["D", "ExpressionVariable"]}, {
                StyleBox["M", "ExpressionVariable"]}, {
                StyleBox["a", "ExpressionVariable"]}, {
                StyleBox["c", "ExpressionVariable"]}, {
                RowBox[{
                  RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "0"}], ",", 
                  "\[Ellipsis]", ",", 
                  RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["M", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["reduceM", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], ",", 
                  StyleBox["c", "ExpressionVariable"], ",", 
                  StyleBox["M", "ExpressionVariable"], ",", 
                  StyleBox["k", "ExpressionVariable"]}], "]"}], ":=", 
              RowBox[{
                StyleBox["a", "ExpressionVariable"], 
                UnderscriptBox["-", 
                 StyleBox["D", "ExpressionVariable"]], 
                RowBox[{
                  UnderscriptBox[
                   SubscriptBox["\[Sum]", 
                    StyleBox["D", "ExpressionVariable"]], 
                   GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["k", "ExpressionVariable"]}]}}]], 
                  RowBox[{
                    SubscriptBox[
                    StyleBox["M", "ExpressionVariable"], 
                    StyleBox["j", "ExpressionVariable"]], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox[
                    "c", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(reduction with several multipliers)\"", True -> 
            StyleBox[
             "\"(reduction with several multipliers)\"", "HyperlinkActive"]}, 
           
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:478302118"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:478302118"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955860, TheoryTools`$LabelDelimiter, 
             "isTotalRelation"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
               StyleBox["\[Succeeds]", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{"isTotalRelation", "[", 
                RowBox[{
                  StyleBox["\[Succeeds]", "ExpressionVariable"], ",", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Succeeds]", 
                    StyleBox["y", "ExpressionVariable"]}], "\[Or]", 
                  RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Succeeds]", 
                    StyleBox["x", "ExpressionVariable"]}], "\[Or]", 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Equal]", 
                    StyleBox[
                    "y", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isTotalRelation)\"", True -> 
            StyleBox["\"(isTotalRelation)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:2054329379"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:2054329379"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955893, TheoryTools`$LabelDelimiter, 
             "symmetric closure of reducibility"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["D", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["symred", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], ",", 
                  StyleBox["b", "ExpressionVariable"], ",", 
                  StyleBox["c", "ExpressionVariable"], ",", 
                  StyleBox["m1", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  UnderscriptBox["mult", 
                   StyleBox["D", "ExpressionVariable"]], "[", 
                  RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], "\[And]", 
                RowBox[{
                  StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    StyleBox["D", "ExpressionVariable"]], 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}], "\[And]", 
                RowBox[{
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox["b", "ExpressionVariable"]}], "\[Or]", 
                  RowBox[{
                    StyleBox["b", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", 
                    StyleBox["D", "ExpressionVariable"]], 
                    StyleBox[
                    "a", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(symmetric closure of reducibility)\"", True -> 
            StyleBox[
             "\"(symmetric closure of reducibility)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:448370877"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:448370877"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955936, TheoryTools`$LabelDelimiter, 
             "correlative"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["D", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["m1", "ExpressionVariable"], ",", 
               StyleBox["m2", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["areCorrelative", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                RowBox[{
                  StyleBox["c", "ExpressionVariable"], ",", 
                  StyleBox["m1", "ExpressionVariable"], ",", 
                  StyleBox["m2", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{
                    UnderscriptBox["multN", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
                RowBox[{
                  RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], "]"}], "\[And]", 
                  RowBox[{
                    UnderscriptBox["mult", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["k", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(correlative)\"", True -> 
            StyleBox["\"(correlative)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1834165766"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1834165766"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955967, TheoryTools`$LabelDelimiter, 
             "nonemptySet"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["A", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"nonemptySet", "[", 
                StyleBox["A", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{"isSet", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{
                  UnderscriptBox["\[Exists]", 
                   StyleBox["a", "ExpressionVariable"]], 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox[
                    "A", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(nonemptySet)\"", True -> 
            StyleBox["\"(nonemptySet)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:324063439"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:324063439"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$955992, TheoryTools`$LabelDelimiter, 
             "\[VerticalBar] \[DoubleStruckCapitalZ]"]]], Evaluator -> 
         Automatic, ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["b", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["a", "ExpressionVariable"], "\[VerticalBar]", 
                StyleBox["b", "ExpressionVariable"]}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}]}}]], 
                RowBox[{
                  StyleBox["b", "ExpressionVariable"], "\[Equal]", 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], " ", 
                    StyleBox[
                    "c", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(\[VerticalBar] \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox[
             "\"(\[VerticalBar] \[DoubleStruckCapitalZ])\"", 
              "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:60870234"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:60870234"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956011, TheoryTools`$LabelDelimiter, 
             "isGCD \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["b", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["g", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalN]"}]}}]], 
            RowBox[{
              RowBox[{"isGCD", "[", 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], ",", 
                  StyleBox["b", "ExpressionVariable"], ",", 
                  StyleBox["g", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  StyleBox["g", "ExpressionVariable"], "\[VerticalBar]", 
                  StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                RowBox[{
                  StyleBox["g", "ExpressionVariable"], "\[VerticalBar]", 
                  StyleBox["b", "ExpressionVariable"]}], "\[And]", 
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}]}}]], 
                  RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["a", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox["b", "ExpressionVariable"]}]}], "\[Implies]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[VerticalBar]", 
                    StyleBox[
                    "g", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isGCD \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox[
             "\"(isGCD \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:339201942"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:339201942"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956059, TheoryTools`$LabelDelimiter, 
             "mod \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["b", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}]}}]], 
                RowBox[{
                  RowBox[{"isRem", "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["r", "ExpressionVariable"]}], "]"}], "\[And]", 
                  RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["r1", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}]}}]], 
                    RowBox[{
                    RowBox[{"isRem", "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["r1", "ExpressionVariable"]}], "]"}], 
                    "\[Implies]", 
                    RowBox[{
                    StyleBox["r1", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["r", "ExpressionVariable"]}]}]}]}]}], 
              "\[Implies]", 
              RowBox[{
                RowBox[{"mod", "[", 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"]}], "]"}], ":=", 
                RowBox[{
                  UnderscriptBox["the", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}]}}]], 
                  RowBox[{"isRem", "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["r", "ExpressionVariable"]}], 
                    "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(mod \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox["\"(mod \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:317913883"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:317913883"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956115, TheoryTools`$LabelDelimiter, 
             "\[Subset]"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["A", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                RowBox[{"isSet", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{"isSet", "[", 
                  StyleBox["B", "ExpressionVariable"], "]"}]}], "\[Implies]", 
              
              RowBox[{
                RowBox[{
                  StyleBox["A", "ExpressionVariable"], "\[Subset]", 
                  StyleBox["B", "ExpressionVariable"]}], 
                TagBox[
                 
                 RowBox[{
                  ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                   "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                 "a:=b"], 
                RowBox[{
                  RowBox[{
                    StyleBox["A", "ExpressionVariable"], "\[SubsetEqual]", 
                    StyleBox["B", "ExpressionVariable"]}], "\[And]", 
                  RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\[SubsetEqual]", 
                    StyleBox["A", "ExpressionVariable"]}], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(\[Subset])\"", True -> 
            StyleBox["\"(\[Subset])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:938973158"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:938973158"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956160, TheoryTools`$LabelDelimiter, 
             "isFunction"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["A", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"], ",", 
               StyleBox["f", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{"isFunction", "[", 
                RowBox[{
                  StyleBox["f", "ExpressionVariable"], ",", 
                  StyleBox["A", "ExpressionVariable"], ",", 
                  StyleBox["B", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                RowBox[{
                  RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}], "\[Element]", 
                  
                  StyleBox[
                  "B", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isFunction)\"", True -> 
            StyleBox["\"(isFunction)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:39301121"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:39301121"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956183, TheoryTools`$LabelDelimiter, 
             "isInjective"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["A", "ExpressionVariable"], ",", 
               StyleBox["f", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{"isInjective", "[", 
                RowBox[{
                  StyleBox["f", "ExpressionVariable"], ",", 
                  StyleBox["A", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["a1", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}, {
                    RowBox[{
                    StyleBox["a2", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                RowBox[{
                  RowBox[{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["a1", "ExpressionVariable"], "]"}], "\[Equal]", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["a2", "ExpressionVariable"], "]"}]}], 
                  "\[Implies]", 
                  RowBox[{
                    StyleBox["a1", "ExpressionVariable"], "\[Equal]", 
                    StyleBox[
                    "a2", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isInjective)\"", True -> 
            StyleBox["\"(isInjective)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:242320146"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:242320146"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956216, TheoryTools`$LabelDelimiter, 
             "isQuo \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["b", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}, {
                RowBox[{
                  StyleBox["q", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{"isQuo", "[", 
                RowBox[{
                  StyleBox["a", "ExpressionVariable"], ",", 
                  StyleBox["b", "ExpressionVariable"], ",", 
                  StyleBox["q", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[Exists]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["r", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalZ]"}]}}]], 
                RowBox[{"isQuoRem", "[", 
                  RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["r", "ExpressionVariable"]}], 
                  "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isQuo \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox[
             "\"(isQuo \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:79981586"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:79981586"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956240, TheoryTools`$LabelDelimiter, 
             "red"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
         "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                StyleBox["D", "ExpressionVariable"]}, {
                RowBox[{
                  UnderscriptBox["\[Element]", 
                   RowBox[{"DomainSets", "[", 
                    StyleBox["D", "ExpressionVariable"], "]"}]], "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{
              RowBox[{
                UnderscriptBox["red", 
                 StyleBox["D", "ExpressionVariable"]], "[", 
                StyleBox["A", "ExpressionVariable"], "]"}], ":=", 
              RowBox[{"{", 
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], 
                  UnderscriptBox["|", 
                   GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    UnderscriptBox["\[Exists]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{
                    UnderscriptBox["isReducible", 
                    StyleBox["D", "ExpressionVariable"]], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["a", "ExpressionVariable"]}], "]"}]}]}], 
                "}"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(red)\"", True -> 
            StyleBox["\"(red)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1593192085"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\ReductionRings.nb", "ID:1593192085"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956265, TheoryTools`$LabelDelimiter, 
             "power set \[Element]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isSet", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}, {
                StyleBox["x", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], "\[Element]", 
                RowBox[{"Pow", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{"isSet", "[", 
                  StyleBox["x", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[SubsetEqual]", 
                  StyleBox[
                  "A", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(power set \[Element])\"", True -> 
            StyleBox["\"(power set \[Element])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:384643200"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:384643200"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956295, TheoryTools`$LabelDelimiter, 
             "isTuple join,[IR]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["S", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{"isTuple", "[", 
                RowBox[{
                  StyleBox["T", "ExpressionVariable"], "\:22c8", 
                  StyleBox["S", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{"isTuple", "[", 
                  StyleBox["T", "ExpressionVariable"], "]"}], "\[And]", 
                RowBox[{"isTuple", "[", 
                  StyleBox["S", "ExpressionVariable"], 
                  "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isTuple join,[IR])\"", True -> 
            StyleBox["\"(isTuple join,[IR])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb", "ID:688753716"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb", "ID:688753716"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956326, TheoryTools`$LabelDelimiter, 
             "append"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["T", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                StyleBox["T", "ExpressionVariable"], "\:293a", 
                StyleBox["a", "ExpressionVariable"]}], ":=", 
              RowBox[{
                StyleBox["T", "ExpressionVariable"], "\:22c8", 
                RowBox[{"\[LeftAngleBracket]", 
                  StyleBox["a", "ExpressionVariable"], 
                  "\[RightAngleBracket]"}]}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(append)\"", True -> 
            StyleBox["\"(append)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb", "ID:1525755901"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\Tuples.nb", "ID:1525755901"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956341, TheoryTools`$LabelDelimiter, 
             "isFinite"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isSet", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{
              RowBox[{"isFinite", "[", 
                StyleBox["A", "ExpressionVariable"], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                StyleBox["A", "ExpressionVariable"], "\[Element]", 
                RowBox[{"finSubsets", "[", 
                  StyleBox["A", "ExpressionVariable"], 
                  "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isFinite)\"", True -> 
            StyleBox["\"(isFinite)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:754397355"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:754397355"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956362, TheoryTools`$LabelDelimiter, 
             "isIrreflexive"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
               StyleBox["\[Succeeds]", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{"isIrreflexive", "[", 
                RowBox[{
                  StyleBox["\[Succeeds]", "ExpressionVariable"], ",", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{"\[Not]", 
                  RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Succeeds]", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isIrreflexive)\"", True -> 
            StyleBox["\"(isIrreflexive)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:932142776"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:932142776"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956384, TheoryTools`$LabelDelimiter, 
             "isAsymmetric"]]], Evaluator -> Automatic, ImageSize -> {20, 20},
          Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
               StyleBox["\[Succeeds]", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{"isAsymmetric", "[", 
                RowBox[{
                  StyleBox["\[Succeeds]", "ExpressionVariable"], ",", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Succeeds]", 
                    StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
                  RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Succeeds]", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[
                    ")", "AutoParentheses"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isAsymmetric)\"", True -> 
            StyleBox["\"(isAsymmetric)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:291399428"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:291399428"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956416, TheoryTools`$LabelDelimiter, 
             "isTransitive"]]], Evaluator -> Automatic, ImageSize -> {20, 20},
          Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
               StyleBox["\[SucceedsEqual]", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{"isTransitive", "[", 
                RowBox[{
                  StyleBox["\[SucceedsEqual]", "ExpressionVariable"], ",", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[SucceedsEqual]", 
                    StyleBox["y", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[SucceedsEqual]", 
                    StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
                  RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[SucceedsEqual]", 
                    StyleBox[
                    "z", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isTransitive)\"", True -> 
            StyleBox["\"(isTransitive)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:535380192"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:535380192"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956456, TheoryTools`$LabelDelimiter, 
             "isTotalIrreflOrder"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["\[ScriptCapitalD]", "ExpressionVariable"], ",", 
               StyleBox["\[Succeeds]", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{"isTotalIrreflOrder", "[", 
                RowBox[{
                  StyleBox["\[Succeeds]", "ExpressionVariable"], ",", 
                  StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]}], 
                "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}, {
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
                RowBox[{
                  RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Succeeds]", 
                    StyleBox["y", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Succeeds]", 
                    StyleBox["x", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["y", "ExpressionVariable"]}]}], "\[And]", 
                  RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Succeeds]", 
                    StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
                    StyleBox["y", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{"\[Not]", 
                    RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Succeeds]", 
                    StyleBox["x", "ExpressionVariable"]}], 
                    TagBox[")", "AutoParentheses"]}]}]}]}], "\[And]", 
                  RowBox[{
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalD]", "ExpressionVariable"]], "[", 
                    StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
                    RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Succeeds]", 
                    StyleBox["y", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Succeeds]", 
                    StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Succeeds]", 
                    StyleBox[
                    "z", "ExpressionVariable"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isTotalIrreflOrder)\"", True -> 
            StyleBox["\"(isTotalIrreflOrder)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:2071876542"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\AlgebraicStructures.nb", "ID:2071876542"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956545, TheoryTools`$LabelDelimiter, 
             "sgn \[DoubleStruckCapitalZ]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  "\[DoubleStruckCapitalZ]"}]}}]], 
            RowBox[{
              RowBox[{"sgn", "[", 
                StyleBox["x", "ExpressionVariable"], "]"}], ":=", 
              RowBox[{"\[Piecewise]", 
                GridBox[{{"1", "\[DoubleLeftArrow]", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ">", "0"}]}, {
                   RowBox[{"-", "1"}], "\[DoubleLeftArrow]", 
                   RowBox[{"0", ">", 
                    StyleBox["x", "ExpressionVariable"]}]}, {
                  "0", "\[DoubleLeftArrow]", 
                   "True"}}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(sgn \[DoubleStruckCapitalZ])\"", True -> 
            StyleBox["\"(sgn \[DoubleStruckCapitalZ])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:223909756"}, 
          ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\NatIntExtended.nb", "ID:223909756"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956564, TheoryTools`$LabelDelimiter, 
             "\[Union]"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
               StyleBox["A", "ExpressionVariable"], ",", 
               StyleBox["B", "ExpressionVariable"]}]], 
            RowBox[{
              RowBox[{
                StyleBox["A", "ExpressionVariable"], "\[Union]", 
                StyleBox["B", "ExpressionVariable"]}], ":=", 
              RowBox[{"Un", "[", 
                RowBox[{"{", 
                  RowBox[{
                    StyleBox["A", "ExpressionVariable"], ",", 
                    StyleBox["B", "ExpressionVariable"]}], "}"}], 
                "]"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(\[Union])\"", True -> 
            StyleBox["\"(\[Union])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:197157243"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:197157243"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956582, TheoryTools`$LabelDelimiter, 
             "union \[Element]"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isSet", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}, {
                StyleBox["x", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                StyleBox["x", "ExpressionVariable"], "\[Element]", 
                RowBox[{"Un", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox[
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    StyleBox["y", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                 RowBox[{"isSet", "[", 
                   StyleBox["y", "ExpressionVariable"], "]"}]], 
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Element]", 
                  StyleBox[
                  "y", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(union \[Element])\"", True -> 
            StyleBox["\"(union \[Element])\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:914914299"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:914914299"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956611, TheoryTools`$LabelDelimiter, 
             "finSubsets"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isSet", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}}]], 
            RowBox[{
              RowBox[{"finSubsets", "[", 
                StyleBox["A", "ExpressionVariable"], "]"}], ":=", 
              RowBox[{"lfp", "[", 
                RowBox[{
                  RowBox[{"finConstr", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], ",", 
                  RowBox[{"Pow", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}], 
                "]"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(finSubsets)\"", True -> 
            StyleBox["\"(finSubsets)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:192198659"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:192198659"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956642, TheoryTools`$LabelDelimiter, 
             "isSetFunction"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isSet", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}, {
                StyleBox["f", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{"isSetFunction", "[", 
                RowBox[{
                  StyleBox["f", "ExpressionVariable"], ",", 
                  StyleBox["A", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                UnderscriptBox["\[ForAll]", 
                 GridBox[{{
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\[Element]", 
                    RowBox[{"Pow", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}]}}]], 
                RowBox[{
                  RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["B", "ExpressionVariable"], "]"}], "\[Element]", 
                  
                  RowBox[{"Pow", "[", 
                    StyleBox["A", "ExpressionVariable"], 
                    "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isSetFunction)\"", True -> 
            StyleBox["\"(isSetFunction)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:154696877"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:154696877"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956676, TheoryTools`$LabelDelimiter, 
             "isBoundedMonotone"]]], Evaluator -> Automatic, 
         ImageSize -> {20, 20}, Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isSet", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}, {
                StyleBox["f", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{"isBoundedMonotone", "[", 
                RowBox[{
                  StyleBox["f", "ExpressionVariable"], ",", 
                  StyleBox["A", "ExpressionVariable"]}], "]"}], 
              TagBox[
               
               RowBox[{
                ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                 "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
               "a:=b"], 
              RowBox[{
                RowBox[{
                  RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}], 
                  "\[SubsetEqual]", 
                  StyleBox["A", "ExpressionVariable"]}], "\[And]", 
                RowBox[{
                  UnderscriptBox["\[ForAll]", 
                   GridBox[{{
                    RowBox[{"isSet", "[", 
                    StyleBox["B", "ExpressionVariable"], "]"}]}, {
                    RowBox[{"isSet", "[", 
                    StyleBox["C", "ExpressionVariable"], "]"}]}}]], 
                  RowBox[{
                    RowBox[{
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\[SubsetEqual]", 
                    StyleBox["C", "ExpressionVariable"]}], "\[And]", 
                    RowBox[{
                    StyleBox["C", "ExpressionVariable"], "\[SubsetEqual]", 
                    StyleBox["A", "ExpressionVariable"]}]}], "\[Implies]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["B", "ExpressionVariable"], "]"}], 
                    "\[SubsetEqual]", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["C", "ExpressionVariable"], 
                    "]"}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(isBoundedMonotone)\"", True -> 
            StyleBox["\"(isBoundedMonotone)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:136577837"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:136577837"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956728, TheoryTools`$LabelDelimiter, 
             "finConstr"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, 
         Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isSet", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}, {
                StyleBox["B", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{
                RowBox[{"finConstr", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}], "[", 
                StyleBox["B", "ExpressionVariable"], "]"}], ":=", 
              RowBox[{
                RowBox[{"{", "\[EmptySet]", "}"}], "\[Union]", 
                RowBox[{
                  UnderscriptBox["\[Union]", 
                   GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                  RowBox[{"{", 
                    RowBox[{
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Union]", 
                    RowBox[{"{", 
                    StyleBox["a", "ExpressionVariable"], "}"}]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], "\[Element]", 
                    StyleBox["B", "ExpressionVariable"]}]}}]]}], 
                    "}"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(finConstr)\"", True -> 
            StyleBox["\"(finConstr)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:31326685"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:31326685"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956755, TheoryTools`$LabelDelimiter, 
             "lfp"]]], Evaluator -> Automatic, ImageSize -> {20, 20}, Method -> 
         "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             GridBox[{{
                RowBox[{"isSet", "[", 
                  StyleBox["A", "ExpressionVariable"], "]"}]}, {
                StyleBox["f", "ExpressionVariable"]}}]], 
            RowBox[{
              RowBox[{"lfp", "[", 
                RowBox[{
                  StyleBox["f", "ExpressionVariable"], ",", 
                  StyleBox["A", "ExpressionVariable"]}], "]"}], ":=", 
              RowBox[{
                UnderscriptBox[
                 UnderscriptBox["\[Intersection]", 
                  GridBox[{{
                    RowBox[{
                    StyleBox["B", "ExpressionVariable"], "\[Element]", 
                    RowBox[{"Pow", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}]}}]], 
                 RowBox[{
                   RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["B", "ExpressionVariable"], "]"}], 
                   "\[SubsetEqual]", 
                   StyleBox["B", "ExpressionVariable"]}]], 
                StyleBox[
                "B", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
         DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(lfp)\"", True -> 
            StyleBox["\"(lfp)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:10293804"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:10293804"}, 
          "Hyperlink"]& ]}, {
        ButtonBox[
        "\"C\"", Appearance -> Automatic, ButtonFunction :> CopyToClipboard[
           AngleBracket[
            StringJoin[
            TheoryTools`Private`src$956786, TheoryTools`$LabelDelimiter, 
             "intersection"]]], Evaluator -> Automatic, ImageSize -> {20, 20},
          Method -> "Preemptive"], 
        TagBox[
         FormBox[
          RowBox[{
            UnderscriptBox["\[ForAll]", 
             StyleBox["A", "ExpressionVariable"]], 
            RowBox[{
              RowBox[{"Inter", "[", 
                StyleBox["A", "ExpressionVariable"], "]"}], ":=", 
              RowBox[{"{", 
                RowBox[{
                  GridBox[{{
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    RowBox[{"Un", "[", 
                    StyleBox["A", "ExpressionVariable"], "]"}]}]}}], "|", 
                  RowBox[{
                    UnderscriptBox[
                    UnderscriptBox["\[ForAll]", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Element]", 
                    StyleBox["A", "ExpressionVariable"]}]}}]], 
                    RowBox[{"isSet", "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]], 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], "\[Element]", 
                    StyleBox["a", "ExpressionVariable"]}]}]}], 
                "}"}]}]}], $CellContext`TheoremaForm], DisplayForm], 
        TagBox[
         ButtonBox[
          
          PaneSelectorBox[{
           False -> "\"(intersection)\"", True -> 
            StyleBox["\"(intersection)\"", "HyperlinkActive"]}, 
           Dynamic[
            CurrentValue["MouseOver"]], BaseStyle -> {"Hyperlink"}, 
           FrameMargins -> 0, ImageSize -> Automatic], BaseStyle -> 
          "Hyperlink", 
          ButtonData :> {
           "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:87041636"}, ButtonNote -> 
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb"], 
         Annotation[#, {
          "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\LogicSets.nb", "ID:87041636"}, 
          "Hyperlink"]& ]}}, AutoDelete -> False, 
      GridBoxAlignment -> {"Columns" -> {Center, Left, Center}}, 
      GridBoxFrame -> {"Columns" -> False, "Rows" -> False}, 
      GridBoxItemSize -> {
       "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], "Grid"],"\"\""},
   
   "Labeled",
   DisplayFunction->(GridBox[{{
       ItemBox[#2, DefaultBaseStyle -> "LabeledLabel"]}, {
       TagBox[
        ItemBox[
         PaneBox[
          TagBox[#, "SkipImageSizeLevel"], Alignment -> {Center, Baseline}, 
          BaselinePosition -> Baseline], DefaultBaseStyle -> "Labeled"], 
        "SkipImageSizeLevel"]}}, 
     GridBoxAlignment -> {"Columns" -> {{Center}}, "Rows" -> {{Center}}}, 
     AutoDelete -> False, 
     GridBoxItemSize -> {"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}},
      BaselinePosition -> {2, 1}]& ),
   InterpretationFunction->(RowBox[{"Labeled", "[", 
      RowBox[{#, ",", #2, ",", "Top"}], "]"}]& )]}}, 1]], "Output",
 CellChangeTimes->{{3.6594308370406237`*^9, 3.659430882155903*^9}, {
   3.6594309165071635`*^9, 3.659430931217989*^9}, 3.6594309844296827`*^9, 
   3.659431391109604*^9},
 CellID->357032028]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Auxiliary Functions", "Subsection",
 CellChangeTimes->{{3.640668579267561*^9, 3.640668583347287*^9}},
 CellID->1454182661],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "LoadPackage", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"LoadPackage", "[", "name_String", "]"}], ":=", 
   RowBox[{"LoadPackage", "[", 
    RowBox[{"{", "name", "}"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"LoadPackage", "[", 
    RowBox[{"l", ":", 
     RowBox[{"{", "___String", "}"}]}], "]"}], ":=", 
   RowBox[{"Scan", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Get", "[", 
       RowBox[{"FileNameJoin", "[", 
        RowBox[{"{", 
         RowBox[{"$HomeDirectory", ",", "\"\<Theorema Notebooks\>\"", ",", 
          RowBox[{"#", "<>", "\"\<.m\>\""}]}], "}"}], "]"}], "]"}], "&"}], 
     ",", "l"}], "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.640668603629292*^9, 3.640668727460284*^9}, 
   3.641286816086794*^9, 3.6434459961973133`*^9, {3.655186919856226*^9, 
   3.655186928139841*^9}, {3.666440024555827*^9, 3.666440031843532*^9}},
 CellID->2074904605],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "FindFormula", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"FindFormula", "[", "label_String", "]"}], ":=", 
   RowBox[{"NotebookLocate", "[", "label", "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.640677756847624*^9, 3.6406777896243896`*^9}, 
   3.640677822869519*^9, {3.640677854902379*^9, 3.6406778579321823`*^9}},
 CellID->1980565193],

Cell[BoxData[
 RowBox[{"LoadPackage", "[", "\"\<ProofPalette\>\"", "]"}]], "Input",
 CellChangeTimes->{{3.640668742006607*^9, 3.640668751417616*^9}},
 CellID->444074165],

Cell[BoxData[
 RowBox[{"LoadPackage", "[", "\"\<TheoryTools/TheoryTools\>\"", 
  "]"}]], "Input",
 CellChangeTimes->{{3.640668759133181*^9, 3.640668768089197*^9}, {
   3.6551869356278543`*^9, 3.6551869365638556`*^9}, {3.658464808476575*^9, 
   3.6584648127665825`*^9}, 3.6664400359520693`*^9},
 CellID->219149603]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Definitions", "Section",
 CellChangeTimes->{{3.645521765718409*^9, 3.6455217675396633`*^9}},
 CellID->2045714290],

Cell[CellGroupData[{

Cell["Theoretical Considerations", "Subsection",
 CellChangeTimes->{{3.607744304977942*^9, 3.607744305837212*^9}, {
  3.645523141181061*^9, 3.645523146006649*^9}},
 CellID->28139740],

Cell[TextData[{
 "We mainly follow the theory in [Stifter1988], with some deviations. Most \
importantly, we use our definition of irrelativity, which is different from \
the one in [Stifter1988]: If ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " is a unit in ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalZ]", "n"], TraditionalForm]]],
 " (i.e. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"gcd", "(", 
     RowBox[{"c", ",", "n"}], ")"}], "=", "1"}], TraditionalForm]]],
 "), then every two multipliers for ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " are both irrelative and correlative.\nIn this subsection we summarize some \
useful lemmata that cannot be found in the literature. For this, let ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " always be a natural number, and let ",
 Cell[BoxData[
  FormBox[
   RowBox[{"LR", "(", "c", ")"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"LCR", "(", 
    RowBox[{
     SubscriptBox["c", "1"], ",", 
     SubscriptBox["c", "2"]}], ")"}], TraditionalForm]]],
 " be defined as in [Stifter1988]."
}], "Text",
 CellChangeTimes->{{3.607744312146592*^9, 3.607744343554891*^9}, {
  3.607744717345495*^9, 3.607744763503287*^9}, {3.607744828597034*^9, 
  3.607744840494492*^9}, {3.607752112522572*^9, 3.607752137272157*^9}, {
  3.610090214462811*^9, 3.610090234473949*^9}, {3.610090461600552*^9, 
  3.610090477899377*^9}, {3.610090510836376*^9, 3.610090561879706*^9}, {
  3.6325458148794823`*^9, 3.6325458398106823`*^9}},
 CellID->755849276],

Cell[TextData[{
 "Let in the sequel ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", ",", 
    RowBox[{"j", "\[Element]", 
     RowBox[{"{", 
      RowBox[{"+", 
       RowBox[{",", "-"}]}], "}"}]}]}], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.610090856900569*^9, 3.610090904571304*^9}},
 CellID->1751629107],

Cell[TextData[{
 StyleBox["Lemma 1",
  FontWeight->"Bold"],
 "\nFor all ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"0", "\[NotEqual]", "c"}], "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", "n"]}], TraditionalForm]]],
 ": ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubsuperscriptBox["M", "c", "+"], "=", 
    SubscriptBox["\[DoubleStruckCapitalN]", 
     RowBox[{"1", ",", "\[Ellipsis]", ",", 
      RowBox[{
       RowBox[{"n", "/", 
        RowBox[{"gcd", "(", 
         RowBox[{"c", ",", "n"}], ")"}]}], "-", "1"}]}]]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubsuperscriptBox["M", "c", "-"], "=", 
    SubscriptBox["\[DoubleStruckCapitalN]", 
     RowBox[{
      RowBox[{"n", "-", 
       RowBox[{"n", "/", 
        RowBox[{"gcd", "(", 
         RowBox[{"c", ",", "n"}], ")"}]}], "+", "1"}], ",", "\[Ellipsis]", 
      ",", 
      RowBox[{"n", "-", "1"}]}]]}], TraditionalForm]]]
}], "Text",
 CellChangeTimes->{{3.607750073639671*^9, 3.607750142895668*^9}, {
  3.607758204195314*^9, 3.607758225107777*^9}, {3.610084289954244*^9, 
  3.610084290239268*^9}, {3.610084324229548*^9, 3.610084324517004*^9}},
 CellID->223062300],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Proof.",
 FontSlant->"Italic"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.607751666670173*^9, 3.607751667567031*^9}, 
   3.607751849290398*^9, 3.631447218625719*^9, 3.635757577775964*^9, 
   3.645519955480811*^9, 3.645522488512352*^9, 3.658485969345147*^9, 
   3.6661731770402822`*^9, 3.666440088575489*^9},
 CellID->2050266752],

Cell[TextData[{
 "Let ",
 Cell[BoxData[
  FormBox[
   RowBox[{"g", ":=", 
    RowBox[{"gcd", "(", 
     RowBox[{"c", ",", "n"}], ")"}]}], TraditionalForm]]],
 " and let ",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", ">", "0"}], TraditionalForm]]],
 " be minimal with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"k", "\[CenterDot]", "c"}], 
    SubscriptBox["\[Congruent]", "n"], "0"}], TraditionalForm]]],
 ". We have to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", "=", 
    RowBox[{"n", "/", "g"}]}], TraditionalForm]]],
 ".\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"k", "\[CenterDot]", "c"}], 
    SubscriptBox["\[Congruent]", "n"], "0"}], TraditionalForm]]],
 " is equivalent to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"k", "\[CenterDot]", "c"}], "=", 
    RowBox[{"x", "\[CenterDot]", "n"}]}], TraditionalForm]]],
 " for some ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ", which means that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", "\[CenterDot]", "c"}], TraditionalForm]]],
 " is a common multiple of ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 ". This is in turn equivalent to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lcm", "(", 
     RowBox[{"c", ",", "n"}], ")"}], "|", 
    RowBox[{"k", "\[CenterDot]", "c"}]}], TraditionalForm]]],
 ", and due to the minimality of ",
 Cell[BoxData[
  FormBox["k", TraditionalForm]]],
 " we even have ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lcm", "(", 
     RowBox[{"c", ",", "n"}], ")"}], "=", 
    RowBox[{"k", "\[CenterDot]", "c"}]}], TraditionalForm]]],
 ". However, ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lcm", "(", 
     RowBox[{"c", ",", "n"}], ")"}], "=", 
    RowBox[{"c", "\[CenterDot]", 
     RowBox[{"n", "/", "g"}]}]}], TraditionalForm]]],
 ", which means ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"k", "\[CenterDot]", "c"}], "=", 
    RowBox[{"c", "\[CenterDot]", 
     RowBox[{"n", "/", "g"}]}]}], TraditionalForm]]],
 ", and therefore ",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", "=", 
    RowBox[{"n", "/", "g"}]}], TraditionalForm]]],
 ".\nThe equality for ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["M", "c", "-"], TraditionalForm]]],
 " follows immediately."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.607751684917519*^9, 3.607751804471418*^9}, 
   3.607751849290498*^9, {3.607751880798592*^9, 3.60775192705424*^9}, {
   3.60775198488578*^9, 3.607752053038202*^9}, {3.607758266464544*^9, 
   3.60775828962848*^9}, {3.610084335580052*^9, 3.610084335985556*^9}, 
   3.631447218625881*^9, 3.635757577776116*^9, 3.645519955481009*^9, 
   3.64552248851248*^9, 3.658485969345147*^9, 3.6661731770402822`*^9, 
   3.666440088575589*^9},
 CellID->1478572631],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.60774566648836*^9, 3.607745679046682*^9}, 
   3.631447218626021*^9, 3.6357575777762403`*^9, 3.645519955481133*^9, 
   3.645522488512589*^9, 3.658485969345147*^9, 3.6661731770402822`*^9, 
   3.666440088575684*^9},
 TextAlignment->Right,
 CellID->1061920785]
}, Closed]],

Cell[TextData[{
 StyleBox["Lemma 2",
  FontWeight->"Bold"],
 "\nFor all ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"0", "\[NotEqual]", "c"}], "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", "n"]}], TraditionalForm]]],
 ": Every number that is reached by ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"m", "\[CenterDot]", "c"}], ")"}], "mod", " ", "n"}], 
   TraditionalForm]]],
 " (for a multiplier ",
 Cell[BoxData[
  FormBox["m", TraditionalForm]]],
 " of ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 ") is a multiple of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"gcd", "(", 
    RowBox[{"c", ",", "n"}], ")"}], TraditionalForm]]],
 ", and all such multiples (strictly between 0 and ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 ") are reached only using multipliers from ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["M", "c", "+"], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.607745738017503*^9, 3.60774583100586*^9}, {
  3.607745892687047*^9, 3.607745902264309*^9}, {3.607746108470813*^9, 
  3.607746116280456*^9}, {3.607747812951564*^9, 3.607747814257609*^9}, {
  3.607751438681519*^9, 3.607751438793489*^9}, {3.60775232673273*^9, 
  3.607752327176647*^9}, {3.607757640336781*^9, 3.607757648870391*^9}},
 CellID->1284341245],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Proof.",
 FontSlant->"Italic"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.607744906395332*^9, 3.60774490762385*^9}, 
   3.607745679046487*^9, 3.631447229217573*^9, 3.63575758277463*^9, 
   3.645519960790904*^9, 3.645522492626865*^9, 3.658486262382062*^9, 
   3.666173172563074*^9, 3.666440084165202*^9},
 CellID->1045358942],

Cell[TextData[{
 "Let ",
 Cell[BoxData[
  FormBox[
   RowBox[{"g", ":=", 
    RowBox[{"gcd", "(", 
     RowBox[{"c", ",", "n"}], ")"}]}], TraditionalForm]]],
 ", and let ",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", "=", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"m", "\[CenterDot]", "c"}], ")"}], "mod", " ", "n"}]}], 
   TraditionalForm]]],
 ". Then ",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", 
    SubscriptBox["\[Congruent]", "n"], 
    RowBox[{"m", "\[CenterDot]", "c"}]}], TraditionalForm]]],
 ", or equivalently ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"m", "\[CenterDot]", "c"}], "-", "k"}], 
    SubscriptBox["\[Congruent]", "n"], "0"}], TraditionalForm]]],
 ". This congruence can be written as ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"m", "\[CenterDot]", "c"}], "-", "k"}], "=", 
    RowBox[{"x", "\[CenterDot]", "n"}]}], TraditionalForm]]],
 " for some ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ", and rearranging yields ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"m", "\[CenterDot]", "c"}], "-", 
     RowBox[{"x", "\[CenterDot]", "n"}]}], "=", "k"}], TraditionalForm]]],
 ", which is only possible if ",
 Cell[BoxData[
  FormBox["k", TraditionalForm]]],
 " is a multiple of ",
 Cell[BoxData[
  FormBox["g", TraditionalForm]]],
 ". Hence, all numbers that are reached by ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"m", "\[CenterDot]", "c"}], ")"}], "mod", " ", "n"}], 
   TraditionalForm]]],
 " are indeed multiples of ",
 Cell[BoxData[
  FormBox["g", TraditionalForm]]],
 ".\n\nFor the other inclusion we proceed as follows: Due to lemma 1, ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"n", "/", "g"}], "\[CenterDot]", "c"}], 
    SubscriptBox["\[Congruent]", "n"], "0"}], TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "/", "g"}], TraditionalForm]]],
 " is the smallest multiplier ",
 Cell[BoxData[
  FormBox[
   RowBox[{">", "0"}], TraditionalForm]]],
 " with this property. Hence, there are exactly ",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", "=", 
    RowBox[{
     RowBox[{"n", "/", "g"}], "-", "1"}]}], TraditionalForm]]],
 " multipliers ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["m", "1"], ",", "\[Ellipsis]", ",", 
    SubscriptBox["m", "k"]}], TraditionalForm]]],
 " with ",
 Cell[BoxData[
  FormBox[
   RowBox[{"0", "<", 
    SubscriptBox["m", "i"], "<", 
    RowBox[{"n", "/", "g"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["m", "i"], "\[CenterDot]", "c"}], 
    SubscriptBox["\[NotCongruent]", "n"], "0"}], TraditionalForm]]],
 ", i.e. all of them are in ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["M", "c", "+"], TraditionalForm]]],
 ". Since all of these mutlipliers are between 0 and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "/", "g"}], TraditionalForm]]],
 ", this also holds for their differences: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[LeftBracketingBar]", 
     RowBox[{
      SubscriptBox["m", "i"], "-", 
      SubscriptBox["m", "j"]}], "\[RightBracketingBar]"}], "<", 
    RowBox[{"n", "/", "g"}]}], TraditionalForm]]],
 ". We claim that using two different such multipliers leads to different \
results: Therefore, choose ",
 Cell[BoxData[
  FormBox[
   RowBox[{"0", "<", 
    SubscriptBox["m", "i"], "<", 
    SubscriptBox["m", "j"], "<", 
    RowBox[{"n", "/", "g"}]}], TraditionalForm]]],
 ", which means we also have ",
 Cell[BoxData[
  FormBox[
   RowBox[{"0", "<", 
    RowBox[{
     SubscriptBox["m", "j"], "-", 
     SubscriptBox["m", "i"]}], "<", 
    RowBox[{"n", "/", "g"}]}], TraditionalForm]]],
 ". We have to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["m", "i"], "\[CenterDot]", "c"}], 
    SubscriptBox["\[NotCongruent]", "n"], 
    RowBox[{
     SubscriptBox["m", "j"], "\[CenterDot]", "c"}]}], TraditionalForm]]],
 ". Assume otherwise: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["m", "i"], "\[CenterDot]", "c"}], 
    SubscriptBox["\[Congruent]", "n"], 
    RowBox[{
     SubscriptBox["m", "j"], "\[CenterDot]", "c"}]}], TraditionalForm]]],
 ". Then ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c", "\[CenterDot]", 
     RowBox[{"(", 
      RowBox[{
       SubscriptBox["m", "j"], "-", 
       SubscriptBox["m", "i"]}], ")"}]}], 
    SubscriptBox["\[Congruent]", "n"], "0"}], TraditionalForm]]],
 ", implying ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["m", "j"], "-", 
     SubscriptBox["m", "i"]}], "=", "0"}], TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["m", "j"], "-", 
     SubscriptBox["m", "i"]}], "\[GreaterEqual]", 
    RowBox[{"n", "/", "g"}]}], TraditionalForm]]],
 ". Both contradict our assumption, hence ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["m", "i"], "\[CenterDot]", "c"}], 
    SubscriptBox["\[NotCongruent]", "n"], 
    RowBox[{
     SubscriptBox["m", "j"], "\[CenterDot]", "c"}]}], TraditionalForm]]],
 ". Now, since ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"n", "/", "g"}], "-", "1"}], TraditionalForm]]],
 " ",
 StyleBox["different",
  FontSlant->"Italic"],
 " multiples of ",
 Cell[BoxData[
  FormBox["g", TraditionalForm]]],
 " are reached by ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["M", "c", "+"], TraditionalForm]]],
 ", and there are exactly ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"n", "/", "g"}], "-", "1"}], TraditionalForm]]],
 " multiples of ",
 Cell[BoxData[
  FormBox["g", TraditionalForm]]],
 " strictly between 0 and ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 ", we are done."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.607745155757226*^9, 3.607745219031551*^9}, {
   3.60774528285699*^9, 3.607745642535676*^9}, 3.607745679046587*^9, {
   3.607745869884848*^9, 3.607745957541901*^9}, {3.607746007881213*^9, 
   3.607746089201037*^9}, {3.607746174891746*^9, 3.607746268409374*^9}, {
   3.607746324812845*^9, 3.607746367446196*^9}, {3.607746398776642*^9, 
   3.607746403487398*^9}, {3.607746511887843*^9, 3.607746599924868*^9}, {
   3.607746631666414*^9, 3.607746640982344*^9}, {3.607746742868977*^9, 
   3.607746875640985*^9}, {3.607746964185346*^9, 3.607747200984911*^9}, {
   3.607747239581807*^9, 3.607747252023004*^9}, {3.607747300297113*^9, 
   3.60774732048318*^9}, {3.607747362197495*^9, 3.60774741041773*^9}, {
   3.607747527233121*^9, 3.607747553123629*^9}, {3.607747703769704*^9, 
   3.607747767737665*^9}, {3.607747818004253*^9, 3.607747842710461*^9}, {
   3.607749982464179*^9, 3.607750057915176*^9}, {3.607750131708203*^9, 
   3.607750643104117*^9}, {3.607757698056783*^9, 3.607757739254989*^9}, {
   3.607757770841603*^9, 3.607757773116665*^9}, 3.631447229217703*^9, 
   3.635757582774757*^9, 3.645519960791029*^9, 3.645522492626979*^9, 
   3.658486262382062*^9, 3.666173172563074*^9, 3.666440084165346*^9},
 CellID->1132937314],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.60774566648836*^9, 3.607745679046682*^9}, 
   3.631447229217832*^9, 3.635757582774881*^9, 3.645519960791153*^9, 
   3.645522492627089*^9, 3.658486262382062*^9, 3.666173172563074*^9, 
   3.666440084165484*^9},
 TextAlignment->Right,
 CellID->626642836]
}, Closed]],

Cell[TextData[{
 StyleBox["Lemma 8",
  FontWeight->"Bold"],
 "\nFor all ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", ",", 
    RowBox[{"c", "\[Element]", 
     RowBox[{
      SubscriptBox["\[DoubleStruckCapitalZ]", "n"], "\\", 
      RowBox[{"{", "0", "}"}]}]}]}], TraditionalForm]]],
 ": ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " is reducible modulo ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " iff ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"gcd", "(", 
     RowBox[{"c", ",", "n"}], ")"}], "\[LessEqual]", "a"}], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.607759731248342*^9, 3.607759929035131*^9}, {
  3.607770787234996*^9, 3.607770789931838*^9}, {3.607847544907764*^9, 
  3.607847689912589*^9}, {3.607847768335471*^9, 3.607847916935198*^9}, {
  3.607848146696307*^9, 3.607848146696415*^9}, {3.607848253548787*^9, 
  3.607848283437212*^9}, {3.607848584879094*^9, 3.607848586241907*^9}, {
  3.607849183978901*^9, 3.607849230442044*^9}},
 CellID->1661433738],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Proof.",
 FontSlant->"Italic"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.607744906395332*^9, 3.60774490762385*^9}, 
   3.607745679046487*^9, 3.6314472640560503`*^9, 3.635757612462878*^9, 
   3.6455199862134943`*^9, 3.645522518127338*^9, 3.6584867841933784`*^9, 
   3.666173168741068*^9, 3.6664400787594557`*^9},
 CellID->1599285460],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"a", "-", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"m", "\[CenterDot]", "c"}], ")"}], "mod", " ", "n"}]}], 
       ")"}], "mod", " ", "n"}], "<", "a"}], "\[DoubleLeftRightArrow]", 
    RowBox[{"0", "<", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"m", "\[CenterDot]", "c"}], ")"}], "mod", " ", "n"}], 
     "\[LessEqual]", "a"}], 
    OverscriptBox["\[DoubleLeftRightArrow]", 
     RowBox[{"Lemma", " ", "2"}]], 
    RowBox[{
     RowBox[{"gcd", "(", 
      RowBox[{"c", ",", "n"}], ")"}], "\[LessEqual]", "a"}]}], 
   TraditionalForm]]],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.607745155757226*^9, 3.607745219031551*^9}, {
   3.60774528285699*^9, 3.607745642535676*^9}, 3.607745679046587*^9, {
   3.607751443891252*^9, 3.607751651278602*^9}, {3.607752252481072*^9, 
   3.607752271487646*^9}, {3.607752372312653*^9, 3.607752474670013*^9}, 
   3.607755663145283*^9, {3.607755711149489*^9, 3.607755821508294*^9}, {
   3.607755856352401*^9, 3.607755858103678*^9}, {3.607755900057404*^9, 
   3.607755903360412*^9}, {3.607756041820266*^9, 3.607756235158769*^9}, {
   3.607756265338065*^9, 3.607756695214362*^9}, {3.60775675993099*^9, 
   3.607756780425748*^9}, {3.607756881761227*^9, 3.607756900874431*^9}, {
   3.607756961094098*^9, 3.607756983196787*^9}, {3.607757022344024*^9, 
   3.607757055333418*^9}, {3.607757085667351*^9, 3.607757111534936*^9}, {
   3.607757197977873*^9, 3.607757199145531*^9}, {3.607757330357854*^9, 
   3.60775743137903*^9}, {3.607757462666469*^9, 3.607757581889098*^9}, {
   3.607757876893729*^9, 3.607758111544848*^9}, {3.607758155455974*^9, 
   3.607758199218492*^9}, {3.607758323359177*^9, 3.607758339597516*^9}, {
   3.607758376435853*^9, 3.607758424225084*^9}, {3.607758460245303*^9, 
   3.607758501619999*^9}, {3.607758541655157*^9, 3.607758577727251*^9}, {
   3.607758631060465*^9, 3.607758811977696*^9}, {3.607847888655577*^9, 
   3.607848234525192*^9}, {3.607848288769402*^9, 3.607848498759167*^9}, 
   3.607849235615099*^9, {3.608005355412945*^9, 3.608005447124294*^9}, 
   3.631447264056226*^9, 3.635757612463045*^9, 3.6455199862136183`*^9, 
   3.645522518127452*^9, 3.6584867841933784`*^9, {3.659178188520217*^9, 
   3.659178188520217*^9}, 3.666173168741068*^9, 3.666440078759557*^9},
 CellID->1269727608],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.60774566648836*^9, 3.607745679046682*^9}, 
   3.631447264056397*^9, 3.635757612463209*^9, 3.645519986213771*^9, 
   3.6455225181275663`*^9, 3.6584867841933784`*^9, 3.666173168741068*^9, 
   3.666440078759654*^9},
 TextAlignment->Right,
 CellID->333263577]
}, Closed]],

Cell[TextData[{
 StyleBox["Lemma 7",
  FontWeight->"Bold"],
 "\nFor all ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", ",", "b", ",", 
    RowBox[{"c", "\[Element]", 
     RowBox[{
      SubscriptBox["\[DoubleStruckCapitalZ]", "n"], "\\", 
      RowBox[{"{", "0", "}"}]}]}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m", "\[Element]", 
    SubsuperscriptBox["M", "c", "+"]}], TraditionalForm]]],
 ", and let ",
 Cell[BoxData[
  FormBox[
   RowBox[{"g", ":=", 
    RowBox[{"gcd", "(", 
     RowBox[{"c", ",", "n"}], ")"}]}], TraditionalForm]]],
 ": If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     RowBox[{"(", 
      RowBox[{"m", ",", "c"}], ")"}]], "b"}], TraditionalForm]]],
 ", then also ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"m", "-", 
        RowBox[{"n", "/", "g"}]}], ",", "c"}], ")"}]], "b"}], 
   TraditionalForm]]],
 ". Moreover, ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"m", "-", 
     RowBox[{"n", "/", "g"}]}], "\[Element]", 
    SubsuperscriptBox["M", "c", "-"]}], TraditionalForm]]],
 ". Of course, if ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m", "\[Element]", 
    SubsuperscriptBox["M", "c", "-"]}], TraditionalForm]]],
 ", a similar statement holds."
}], "Text",
 CellChangeTimes->{{3.607759731248342*^9, 3.607759929035131*^9}, {
  3.607770787234996*^9, 3.607770789931838*^9}, {3.607847544907764*^9, 
  3.607847689912589*^9}, {3.607847768335471*^9, 3.607847916935198*^9}, {
  3.607848146696307*^9, 3.607848146696415*^9}, {3.607848253548787*^9, 
  3.607848283437212*^9}, {3.607848584879094*^9, 3.607848586241907*^9}, {
  3.6592539884323044`*^9, 3.6592540004599257`*^9}},
 CellID->645395591],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Proof.",
 FontSlant->"Italic"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.607744906395332*^9, 3.60774490762385*^9}, 
   3.607745679046487*^9, 3.631447258157751*^9, 3.635757607588822*^9, 
   3.6455199826626663`*^9, 3.645522514969411*^9, 3.6584867095316477`*^9, 
   3.6661731646850605`*^9, 3.6664400746642017`*^9},
 CellID->1699779808],

Cell[TextData[{
 "Assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m", "\[Element]", 
    SubsuperscriptBox["M", "c", "+"]}], TraditionalForm]]],
 ", i.e. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"0", "<", "m", "<", 
    RowBox[{"n", "/", "g"}]}], TraditionalForm]]],
 ".\nWe first show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"a", "-", 
     RowBox[{"m", "\[CenterDot]", "c"}]}], 
    SubscriptBox["\[Congruent]", "n"], 
    RowBox[{"a", "-", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"n", "-", 
        RowBox[{"n", "/", "g"}], "+", "m"}], ")"}], "\[CenterDot]", "c"}]}]}],
    TraditionalForm]]],
 ". We may cancel ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " on both sides and have to show ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"m", "\[CenterDot]", "c"}], 
    SubscriptBox["\[Congruent]", "n"], 
    RowBox[{
     RowBox[{"n", "\[CenterDot]", "c"}], "-", 
     RowBox[{"n", "\[CenterDot]", 
      RowBox[{"(", 
       RowBox[{"c", "/", "g"}], ")"}]}], "+", 
     RowBox[{"m", "\[CenterDot]", "c"}]}]}], TraditionalForm]]],
 ", where we may now cancel ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m", "\[CenterDot]", "c"}], TraditionalForm]]],
 " on both sides, leaving us with ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"n", "\[CenterDot]", 
     RowBox[{"(", 
      RowBox[{"c", "-", 
       RowBox[{"c", "/", "g"}]}], ")"}]}], 
    SubscriptBox["\[Congruent]", "n"], "0"}], TraditionalForm]]],
 ". Since ",
 Cell[BoxData[
  FormBox["g", TraditionalForm]]],
 " divides ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " the factor of ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " is an integer and the congruence, hence, trivially holds. This means that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", 
    SubscriptBox["\[Rule]", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"n", "-", 
        RowBox[{"n", "/", "g"}], "+", "m"}], ",", "c"}], ")"}]], "b"}], 
   TraditionalForm]]],
 ".\nWhat remains to be shown is ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"n", "-", 
     RowBox[{"n", "/", "g"}], "+", "m"}], "\[Element]", 
    SubsuperscriptBox["M", "c", "-"]}], TraditionalForm]]],
 ", i.e. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"n", "-", 
     RowBox[{"n", "/", "g"}]}], "<", 
    RowBox[{"n", "-", 
     RowBox[{"n", "/", "g"}], "+", "m"}], "<", "n"}], TraditionalForm]]],
 ", or, after cancellation of ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"-", "n"}], "/", "g"}], "<", 
    RowBox[{"m", "-", 
     RowBox[{"n", "/", "g"}]}], "<", "0"}], TraditionalForm]]],
 ". We now add ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "/", "g"}], TraditionalForm]]],
 " to all three sides and thus arrive at ",
 Cell[BoxData[
  FormBox[
   RowBox[{"0", "<", "m", "<", 
    RowBox[{"n", "/", "g"}]}], TraditionalForm]]],
 ", which is exactly our assumption."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.607745155757226*^9, 3.607745219031551*^9}, {
   3.60774528285699*^9, 3.607745642535676*^9}, 3.607745679046587*^9, {
   3.607751443891252*^9, 3.607751651278602*^9}, {3.607752252481072*^9, 
   3.607752271487646*^9}, {3.607752372312653*^9, 3.607752474670013*^9}, 
   3.607755663145283*^9, {3.607755711149489*^9, 3.607755821508294*^9}, {
   3.607755856352401*^9, 3.607755858103678*^9}, {3.607755900057404*^9, 
   3.607755903360412*^9}, {3.607756041820266*^9, 3.607756235158769*^9}, {
   3.607756265338065*^9, 3.607756695214362*^9}, {3.60775675993099*^9, 
   3.607756780425748*^9}, {3.607756881761227*^9, 3.607756900874431*^9}, {
   3.607756961094098*^9, 3.607756983196787*^9}, {3.607757022344024*^9, 
   3.607757055333418*^9}, {3.607757085667351*^9, 3.607757111534936*^9}, {
   3.607757197977873*^9, 3.607757199145531*^9}, {3.607757330357854*^9, 
   3.60775743137903*^9}, {3.607757462666469*^9, 3.607757581889098*^9}, {
   3.607757876893729*^9, 3.607758111544848*^9}, {3.607758155455974*^9, 
   3.607758199218492*^9}, {3.607758323359177*^9, 3.607758339597516*^9}, {
   3.607758376435853*^9, 3.607758424225084*^9}, {3.607758460245303*^9, 
   3.607758501619999*^9}, {3.607758541655157*^9, 3.607758577727251*^9}, {
   3.607758631060465*^9, 3.607758811977696*^9}, {3.607847888655577*^9, 
   3.607848234525192*^9}, {3.607848288769402*^9, 3.607848498759167*^9}, 
   3.6314472581579237`*^9, 3.6357576075889683`*^9, 3.645519982662792*^9, 
   3.6455225149695263`*^9, {3.6584866838384027`*^9, 3.6584867095316477`*^9}, 
   3.6661731646850605`*^9, 3.6664400746643457`*^9},
 CellID->969904702],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.60774566648836*^9, 3.607745679046682*^9}, 
   3.631447258158093*^9, 3.63575760758911*^9, 3.645519982662917*^9, 
   3.645522514969639*^9, 3.6584867095316477`*^9, 3.6661731646850605`*^9, 
   3.666440074664483*^9},
 TextAlignment->Right,
 CellID->198497244]
}, Closed]],

Cell[TextData[{
 StyleBox["Corollary 3 (Main Observation)",
  FontWeight->"Bold"],
 "\nFor all ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", ",", 
    RowBox[{"c", "\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalZ]", "n"]}]}], TraditionalForm]]],
 ": Whenever ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " can be reduced modulo ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 ", it can be reduced modulo ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " in two ",
 StyleBox["irrelative",
  FontSlant->"Italic"],
 " ways (to the same element)."
}], "Text",
 CellChangeTimes->{{3.607759731248342*^9, 3.607759929035131*^9}, {
   3.607770787234996*^9, 3.607770789931838*^9}, {3.607847544907764*^9, 
   3.607847689912589*^9}, {3.607847768335471*^9, 3.607847916935198*^9}, {
   3.607848146696307*^9, 3.607848146696415*^9}, {3.607848253548787*^9, 
   3.607848283437212*^9}, {3.607848525126623*^9, 3.607848611042939*^9}, 
   3.607848668769336*^9, {3.607848803540714*^9, 3.607848806200884*^9}, {
   3.610091916248549*^9, 3.610091923742123*^9}, 3.610092124698142*^9, {
   3.632551724483976*^9, 3.632551724959867*^9}},
 CellID->1934527770],

Cell[TextData[{
 StyleBox["Corollary 4",
  FontWeight->"Bold"],
 "\nFor reduction in ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalZ]", "n"], TraditionalForm]]],
 " it is sufficient to only consider the multipliers in ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["M", "c", "+"], TraditionalForm]]],
 ". If no suitable multiplier in ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["M", "c", "+"], TraditionalForm]]],
 " exists, there does not exist such multiplier in ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["M", "c", "-"], TraditionalForm]]],
 " either, and vice versa."
}], "Text",
 CellChangeTimes->{{3.607759731248342*^9, 3.607759929035131*^9}, {
   3.607770787234996*^9, 3.607770789931838*^9}, {3.607847544907764*^9, 
   3.607847689912589*^9}, {3.607847768335471*^9, 3.607847916935198*^9}, {
   3.607848146696307*^9, 3.607848146696415*^9}, {3.607848253548787*^9, 
   3.607848283437212*^9}, {3.607848525126623*^9, 3.607848611042939*^9}, 
   3.607848668769336*^9, {3.607848803540714*^9, 3.607848806200884*^9}, {
   3.607848979443524*^9, 3.607849053154793*^9}, 3.61009212688677*^9},
 CellID->1075691108],

Cell[TextData[{
 StyleBox["Lemma 3",
  FontWeight->"Bold"],
 "\nFor all ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"0", "\[NotEqual]", "c"}], "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", "n"]}], TraditionalForm]]],
 ": ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"LR", "(", "c", ")"}], "=", 
    RowBox[{"gcd", "(", 
     RowBox[{"c", ",", "n"}], ")"}]}], TraditionalForm]]],
 " can only be reduced to 0 modulo ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.607744796591528*^9, 3.607744804764381*^9}, {
  3.607744844553496*^9, 3.60774490203693*^9}, {3.607751440529573*^9, 
  3.607751440715636*^9}},
 CellID->53741691],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Proof.",
 FontSlant->"Italic"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.607744906395332*^9, 3.60774490762385*^9}, 
   3.607745679046487*^9, 3.631447234910952*^9, 3.635757586285667*^9, 
   3.6455199645672827`*^9, 3.6455224962852507`*^9, 3.658486371163053*^9, 
   3.666173160751853*^9, 3.66644007096802*^9},
 CellID->204886373],

Cell[TextData[{
 "Let ",
 Cell[BoxData[
  FormBox[
   RowBox[{"g", ":=", 
    RowBox[{"gcd", "(", 
     RowBox[{"c", ",", "n"}], ")"}]}], TraditionalForm]]],
 ". Then we have ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"g", "-", 
        RowBox[{"m", "\[CenterDot]", "c"}]}], ")"}], "mod", " ", "n"}], "<", 
     "g"}], "\[DoubleLeftRightArrow]", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"g", "-", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{"m", "\[CenterDot]", "c"}], ")"}], "mod", " ", "n"}]}], 
       ")"}], "mod", " ", "n"}], "<", "g"}], "\[DoubleLeftRightArrow]", 
    RowBox[{"0", "<", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"m", "\[CenterDot]", "c"}], ")"}], "mod", " ", "n"}], 
     "\[LessEqual]", "g"}]}], TraditionalForm]]],
 ". Since we already know from lemma 2 that everything that is reached by ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"m", "\[CenterDot]", "c"}], ")"}], "mod", " ", "n"}], 
   TraditionalForm]]],
 " is a multiple of ",
 Cell[BoxData[
  FormBox["g", TraditionalForm]]],
 ", the only possibility for the inequality above to hold is ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"m", "\[CenterDot]", "c"}], 
    SubscriptBox["\[Congruent]", "n"], "g"}], TraditionalForm]]],
 ", meaning ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"g", "-", 
     RowBox[{"m", "\[CenterDot]", "c"}]}], 
    SubscriptBox["\[Congruent]", "n"], "0"}], TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox["g", TraditionalForm]]],
 " can indeed only be reduced to 0 modulo ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " (and it really can be reduced to it, since every multiple of ",
 Cell[BoxData[
  FormBox["g", TraditionalForm]]],
 ", and in particular ",
 Cell[BoxData[
  FormBox["g", TraditionalForm]]],
 " itself, is reachable by multiplication with a multiplier of ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 ", thanks to lemma 2)."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.607745155757226*^9, 3.607745219031551*^9}, {
   3.60774528285699*^9, 3.607745642535676*^9}, 3.607745679046587*^9, {
   3.607751443891252*^9, 3.607751651278602*^9}, {3.607770934094486*^9, 
   3.607770935213505*^9}, 3.631447234911084*^9, 3.6357575862858343`*^9, 
   3.645519964567409*^9, 3.645522496285398*^9, 3.658486371163053*^9, 
   3.666173160751853*^9, 3.666440070968161*^9},
 CellID->1003232302],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.60774566648836*^9, 3.607745679046682*^9}, 
   3.631447234911216*^9, 3.635757586285993*^9, 3.6455199645675297`*^9, 
   3.64552249628554*^9, 3.658486371163053*^9, 3.666173160751853*^9, 
   3.666440070968294*^9},
 TextAlignment->Right,
 CellID->325823042]
}, Closed]],

Cell[TextData[{
 StyleBox["Corollary 1",
  FontWeight->"Bold"],
 "\nIn ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalZ]", "n"], TraditionalForm]]],
 ", every one-element set is a Gr\[ODoubleDot]bner basis."
}], "Text",
 CellChangeTimes->{{3.607759731248342*^9, 3.607759929035131*^9}, {
  3.607770757483037*^9, 3.607770819638068*^9}},
 CellID->1312586886],

Cell[TextData[{
 StyleBox["Lemma 4",
  FontWeight->"Bold"],
 "\nFor all ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"0", "\[NotEqual]", "c"}], "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", "n"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", "n"]}], TraditionalForm]]],
 ": ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"c", " ", 
      SuperscriptBox[
       UnderscriptBox["\[CapitalDelta]", "_"], "a"]}], 
     "\[DoubleLeftRightArrow]", "a"}], "=", 
    RowBox[{"LR", "(", "c", ")"}]}], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.607752088268376*^9, 3.607752089998628*^9}, {
  3.607752164253409*^9, 3.607752221582985*^9}},
 CellID->1202825855],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Proof.",
 FontSlant->"Italic"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.607744906395332*^9, 3.60774490762385*^9}, 
   3.607745679046487*^9, 3.631447243119026*^9, 3.6357575895323668`*^9, 
   3.645519969061494*^9, 3.6455225006030827`*^9, 3.6584864032221093`*^9, 
   3.6661731562902455`*^9, 3.666440067544301*^9},
 CellID->1262287499],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]\[DoubleLongLeftArrow]\[CloseCurlyDoubleQuote]:\n\
Because of lemma 3, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"LR", "(", "c", ")"}], TraditionalForm]]],
 " can be reduced to 0, which means ",
 Cell[BoxData[
  FormBox[
   RowBox[{"LR", "(", "c", ")"}], TraditionalForm]]],
 " is reducible modulo ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 ". Since it can only be reduced to 0, and 0 is irreducible, it is also \
non-trivial. Minimality, finally, follows from lemma 3 as well.\n\n\
\[OpenCurlyDoubleQuote]\[DoubleLongRightArrow]\[CloseCurlyDoubleQuote]:\n\
Since the ordering on ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalZ]", "n"], TraditionalForm]]],
 " is total there can only be one minimal non-trivial reducible for ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 ". Because of \[OpenCurlyDoubleQuote]\[DoubleLongLeftArrow]\
\[CloseCurlyDoubleQuote], this has to be ",
 Cell[BoxData[
  FormBox[
   RowBox[{"LR", "(", "c", ")"}], TraditionalForm]]],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.607745155757226*^9, 3.607745219031551*^9}, {
   3.60774528285699*^9, 3.607745642535676*^9}, 3.607745679046587*^9, {
   3.607751443891252*^9, 3.607751651278602*^9}, {3.607752252481072*^9, 
   3.607752271487646*^9}, {3.607752372312653*^9, 3.607752474670013*^9}, 
   3.607755663145283*^9, {3.607755711149489*^9, 3.607755821508294*^9}, {
   3.607755856352401*^9, 3.607755858103678*^9}, {3.607755900057404*^9, 
   3.607755903360412*^9}, 3.631447243119153*^9, 3.635757589532494*^9, 
   3.6455199690616198`*^9, 3.645522500603199*^9, 3.6584864032221093`*^9, 
   3.6661731562902455`*^9, 3.666440067544445*^9},
 CellID->921465591],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.60774566648836*^9, 3.607745679046682*^9}, 
   3.631447243119276*^9, 3.635757589532618*^9, 3.645519969061741*^9, 
   3.6455225006033087`*^9, 3.658486403237709*^9, 3.6661731562902455`*^9, 
   3.66644006754458*^9},
 TextAlignment->Right,
 CellID->377926640]
}, Closed]],

Cell[TextData[{
 StyleBox["Lemma 5",
  FontWeight->"Bold"],
 "\nFor all ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"0", "\[NotEqual]", 
     SubscriptBox["c", "1"], "\[NotEqual]", 
     SubscriptBox["c", "2"]}], "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", "n"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", "n"]}], TraditionalForm]]],
 ": ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      SubscriptBox["c", "1"], 
      SubsuperscriptBox[
       UnderscriptBox["\[CapitalDelta]", "_"], 
       RowBox[{"i", ",", "j"}], "a"], 
      SubscriptBox["c", "2"]}], "\[DoubleLeftRightArrow]", "a"}], "=", 
    RowBox[{"LCR", "(", 
     RowBox[{
      SubscriptBox["c", "1"], ",", 
      SubscriptBox["c", "2"]}], ")"}]}], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.607755931628775*^9, 3.607755996506927*^9}, {
  3.610090799753827*^9, 3.610090812789551*^9}, {3.610090845069742*^9, 
  3.610090845898544*^9}},
 CellID->663291260],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Proof.",
 FontSlant->"Italic"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.607744906395332*^9, 3.60774490762385*^9}, 
   3.607745679046487*^9, 3.631447248613316*^9, 3.635757595061301*^9, 
   3.645519973686322*^9, 3.645522505258046*^9, 3.6584864698216267`*^9, 
   3.666173152390239*^9, 3.6664400636634274`*^9},
 CellID->642765392],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]\[DoubleLongLeftArrow]\[CloseCurlyDoubleQuote]:\nLet \
",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " denote ",
 Cell[BoxData[
  FormBox[
   RowBox[{"LCR", "(", 
    RowBox[{
     SubscriptBox["c", "1"], ",", 
     SubscriptBox["c", "2"]}], ")"}], TraditionalForm]]],
 ". Clearly, ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " is reducible both modulo ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["c", "1"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["c", "2"], TraditionalForm]]],
 ", and since ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["c", "1"], TraditionalForm]]],
 " is different from ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["c", "2"], TraditionalForm]]],
 " we do not have to care about irrelativity of the multipliers, nor about \
the indices ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["j", TraditionalForm]]],
 " at all. By definition of ",
 Cell[BoxData[
  FormBox["LCR", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " is also the smallest common reducible among all elements of ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalZ]", "n"], TraditionalForm]]],
 ", meaning that we only have to show that it is non-trivial.\nLet ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["g", "1"], ":=", 
    RowBox[{"LR", "(", 
     SubscriptBox["c", "1"], ")"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["g", "2"], ":=", 
    RowBox[{"LR", "(", 
     SubscriptBox["c", "2"], ")"}]}], TraditionalForm]]],
 ", hence ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", "=", 
    RowBox[{"max", "(", 
     RowBox[{
      SubscriptBox["g", "1"], ",", 
      SubscriptBox["g", "2"]}], ")"}]}], TraditionalForm]]],
 ". We distinguish two cases:\n\nCase I: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["g", "1"], "=", 
    SubscriptBox["g", "2"]}], TraditionalForm]]],
 ".\nIn this case ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", "=", 
    RowBox[{
     SubscriptBox["g", "1"], "=", 
     SubscriptBox["g", "2"]}]}], TraditionalForm]]],
 ".\nBy lemma 3 we know that ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["g", "1"], TraditionalForm]]],
 " can only be reduced to 0 modulo ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["c", "1"], TraditionalForm]]],
 " and that ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["g", "2"], TraditionalForm]]],
 " can only be reduced to 0 modulo ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["c", "2"], TraditionalForm]]],
 ", therefore ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " can only be reduced to 0 modulo both ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["c", "1"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["c", "2"], TraditionalForm]]],
 ". 0 is irreducible, and non-triviality follows readily from the definition.\
\n\nCase II: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["g", "1"], "\[NotEqual]", 
    SubscriptBox["g", "2"]}], TraditionalForm]]],
 ".\nW.l.o.g. assume ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["g", "1"], "<", 
    SubscriptBox["g", "2"]}], TraditionalForm]]],
 ", i.e. ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", "=", 
    SubscriptBox["g", "2"]}], TraditionalForm]]],
 ". Still, repeating the argument from case I, ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " can only be reduced to 0 modulo ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["c", "2"], TraditionalForm]]],
 " (using multiplier ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "2"], TraditionalForm]]],
 "), but it might be reducible to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"d", "\[NotEqual]", "0"}], TraditionalForm]]],
 " modulo ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["c", "1"], TraditionalForm]]],
 ". We have to show now that ",
 Cell[BoxData[
  FormBox["d", TraditionalForm]]],
 " cannot be reduced modulo ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["c", "2"], TraditionalForm]]],
 " using ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "2"], TraditionalForm]]],
 ", or, in other words, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"d", "\[LessEqual]", 
    RowBox[{"d", "-", "b"}]}], TraditionalForm]]],
 " (because ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["m", "2"], 
     SubscriptBox["c", "2"]}], 
    SubscriptBox["\[Congruent]", "n"], "b"}], TraditionalForm]]],
 "). Assume the opposite: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"d", "-", "b"}], "<", "d"}], TraditionalForm]]],
 ". This is only possible if ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", "\[LessEqual]", "d"}], TraditionalForm]]],
 ", contradicting ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", 
    SubscriptBox["\[Rule]", 
     SubscriptBox["c", "1"]], "d"}], TraditionalForm]]],
 ".\n\n\[OpenCurlyDoubleQuote]\[DoubleLongRightArrow]\[CloseCurlyDoubleQuote]:\
\nSince the ordering on ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalZ]", "n"], TraditionalForm]]],
 " is total there can only be one minimal non-trivial common reducible for ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["c", "1"], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["c", "2"], TraditionalForm]]],
 ". Because of \[OpenCurlyDoubleQuote]\[DoubleLongLeftArrow]\
\[CloseCurlyDoubleQuote], this has to be ",
 Cell[BoxData[
  FormBox[
   RowBox[{"LCR", "(", 
    RowBox[{
     SubscriptBox["c", "1"], ",", 
     SubscriptBox["c", "2"]}], ")"}], TraditionalForm]]],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.607745155757226*^9, 3.607745219031551*^9}, {
   3.60774528285699*^9, 3.607745642535676*^9}, 3.607745679046587*^9, {
   3.607751443891252*^9, 3.607751651278602*^9}, {3.607752252481072*^9, 
   3.607752271487646*^9}, {3.607752372312653*^9, 3.607752474670013*^9}, 
   3.607755663145283*^9, {3.607755711149489*^9, 3.607755821508294*^9}, {
   3.607755856352401*^9, 3.607755858103678*^9}, {3.607755900057404*^9, 
   3.607755903360412*^9}, {3.607756041820266*^9, 3.607756235158769*^9}, {
   3.607756265338065*^9, 3.607756695214362*^9}, {3.60775675993099*^9, 
   3.607756780425748*^9}, {3.607756881761227*^9, 3.607756900874431*^9}, {
   3.607756961094098*^9, 3.607756983196787*^9}, {3.607757022344024*^9, 
   3.607757055333418*^9}, {3.607757085667351*^9, 3.607757111534936*^9}, {
   3.607757197977873*^9, 3.607757199145531*^9}, 3.6314472486134453`*^9, {
   3.632546014819695*^9, 3.632546029410584*^9}, 3.635757595061427*^9, 
   3.6455199736864643`*^9, 3.64552250525816*^9, 3.6584864698216267`*^9, 
   3.666173152390239*^9, 3.666440063663527*^9},
 CellID->696507084],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.60774566648836*^9, 3.607745679046682*^9}, 
   3.6314472486135683`*^9, 3.635757595061551*^9, 3.645519973686602*^9, 
   3.6455225052582703`*^9, 3.6584864698216267`*^9, 3.666173152390239*^9, 
   3.666440063663623*^9},
 TextAlignment->Right,
 CellID->1280121131]
}, Closed]],

Cell[TextData[{
 StyleBox["Lemma 6",
  FontWeight->"Bold"],
 "\nFor all ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"0", "\[NotEqual]", "c"}], "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", "n"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", "n"]}], TraditionalForm]]],
 ": ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"c", 
     SubsuperscriptBox[
      UnderscriptBox["\[CapitalDelta]", "_"], 
      RowBox[{"i", ",", "j"}], "a"], "c"}], "\[DoubleLeftRightArrow]", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"i", "\[NotEqual]", "j"}], "\[And]", "a"}], "=", 
      RowBox[{"LCR", "(", 
       RowBox[{"c", ",", "c"}], ")"}]}], ")"}]}], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.607757278284551*^9, 3.607757315466116*^9}, {
  3.610090742024662*^9, 3.61009075126897*^9}, {3.610091047483032*^9, 
  3.610091095622181*^9}, {3.610091194866964*^9, 3.610091227842945*^9}, {
  3.610091379146624*^9, 3.610091380837345*^9}, {3.632546298795471*^9, 
  3.632546302040881*^9}, {3.6325464549005136`*^9, 3.632546455268894*^9}, {
  3.632546531875464*^9, 3.6325465337469463`*^9}, {3.6325515669347067`*^9, 
  3.632551573116808*^9}, {3.658486519164513*^9, 3.6584865210521164`*^9}},
 CellID->478524917],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Proof.",
 FontSlant->"Italic"]], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.607744906395332*^9, 3.60774490762385*^9}, 
   3.607745679046487*^9, 3.631447253492997*^9, 3.6357576000857964`*^9, 
   3.6455199787343397`*^9, 3.6455225096046*^9, 3.6584865813306227`*^9, 
   3.6661731473660297`*^9, 3.666440058488175*^9},
 CellID->1022330126],

Cell[TextData[{
 "\[OpenCurlyDoubleQuote]\[DoubleLongLeftArrow]\[CloseCurlyDoubleQuote]:\nIn \
this case, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"g", ":=", 
    RowBox[{
     RowBox[{"LCR", "(", 
      RowBox[{"c", ",", "c"}], ")"}], "=", 
     RowBox[{
      RowBox[{"LR", "(", "c", ")"}], "=", 
      RowBox[{"gcd", "(", 
       RowBox[{"c", ",", "n"}], ")"}]}]}]}], TraditionalForm]]],
 ". By definition of ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[CapitalDelta]", "_"], TraditionalForm]]],
 " we first have to show that ",
 Cell[BoxData[
  FormBox["g", TraditionalForm]]],
 " is a common reducible modulo ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " w.r.t. ",
 Cell[BoxData[
  FormBox["i", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["j", TraditionalForm]]],
 ", i.e. it can be reduced in two ways with one multiplier in ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["M", "c", "i"], TraditionalForm]]],
 " and one in ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["M", "c", "j"], TraditionalForm]]],
 ", where ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "\[NotEqual]", "j"}], TraditionalForm]]],
 ". Assume w.l.o.g. that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "=", "+"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"j", "=", "-"}], TraditionalForm]]],
 ".\nFrom lemma 3 we know that ",
 Cell[BoxData[
  FormBox["g", TraditionalForm]]],
 " can only be reduced to 0 modulo ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 ", which means that we have to show that it can be reduced to 0 using one \
multiplier from ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["M", "c", "+"], TraditionalForm]]],
 " and one from ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["M", "c", "-"], TraditionalForm]]],
 ". So, we have to show that there exist ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["m", "1"], "\[Element]", 
    SubsuperscriptBox["M", "c", "+"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["m", "2"], "\[Element]", 
    SubsuperscriptBox["M", "c", "-"]}], TraditionalForm]]],
 " such that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["m", "1"], "c"}], 
    SubscriptBox["\[Congruent]", "n"], "g"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["m", "2"], "c"}], 
    SubscriptBox["\[Congruent]", "n"], "g"}], TraditionalForm]]],
 ". The first congruence follows immediately from lemma 2 (which also \
guarantees membership of ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "1"], TraditionalForm]]],
 " in ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["M", "c", "+"], TraditionalForm]]],
 "), and a witness for ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "2"], TraditionalForm]]],
 " is claimed to be ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "-", 
    RowBox[{"n", "/", "g"}], "+", 
    SubscriptBox["m", "1"]}], TraditionalForm]]],
 ": First of all, ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["m", "2"], TraditionalForm]]],
 " is certainly in ",
 Cell[BoxData[
  FormBox[
   SubsuperscriptBox["M", "c", "-"], TraditionalForm]]],
 ", since ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["m", "1"], "\[Element]", 
    SubsuperscriptBox["M", "c", "+"]}], TraditionalForm]]],
 ". Also: ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["m", "2"], "c"}], "=", 
    RowBox[{
     RowBox[{
      RowBox[{"(", 
       RowBox[{"n", "-", 
        RowBox[{"n", "/", "g"}], "+", 
        SubscriptBox["m", "1"]}], ")"}], "\[CenterDot]", "c"}], "=", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "\[CenterDot]", "c"}], "-", 
       RowBox[{"c", "\[CenterDot]", 
        RowBox[{"n", "/", "g"}]}], "+", 
       RowBox[{
        SubscriptBox["m", "1"], "c"}]}], "=", 
      RowBox[{
       RowBox[{
        RowBox[{"n", "\[CenterDot]", 
         RowBox[{"(", 
          RowBox[{"c", "-", 
           RowBox[{"c", "/", "g"}]}], ")"}]}], "+", 
        RowBox[{
         SubscriptBox["m", "1"], "c"}]}], 
       SubscriptBox["\[Congruent]", "n"], 
       RowBox[{
        SubscriptBox["m", "1"], "c"}], 
       SubscriptBox["\[Congruent]", "n"], "g"}]}]}]}], TraditionalForm]]],
 ".\nThe rest of the argument proceeds analoguously to case I in the proof of \
lemma 5: ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " can only be reduced to 0, 0 is irreducible, and non-triviality follows. \
Minimality follows immediately from the definition of ",
 Cell[BoxData[
  FormBox["LR", TraditionalForm]]],
 " and from lemma 2 anyway.\n\n\[OpenCurlyDoubleQuote]\[DoubleLongRightArrow]\
\[CloseCurlyDoubleQuote]:\nSince the ordering on ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalZ]", "n"], TraditionalForm]]],
 " is total there can at most be only one minimal non-trivial common \
reducible for ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 ". Because of \[OpenCurlyDoubleQuote]\[DoubleLongLeftArrow]\
\[CloseCurlyDoubleQuote], this has to be ",
 Cell[BoxData[
  FormBox[
   RowBox[{"LCR", "(", 
    RowBox[{"c", ",", "c"}], ")"}], TraditionalForm]]],
 ". The fact that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "\[NotEqual]", "j"}], TraditionalForm]]],
 " is contained in the definition of ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["\[CapitalDelta]", "_"], TraditionalForm]]],
 "."
}], "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.607745155757226*^9, 3.607745219031551*^9}, {
   3.60774528285699*^9, 3.607745642535676*^9}, 3.607745679046587*^9, {
   3.607751443891252*^9, 3.607751651278602*^9}, {3.607752252481072*^9, 
   3.607752271487646*^9}, {3.607752372312653*^9, 3.607752474670013*^9}, 
   3.607755663145283*^9, {3.607755711149489*^9, 3.607755821508294*^9}, {
   3.607755856352401*^9, 3.607755858103678*^9}, {3.607755900057404*^9, 
   3.607755903360412*^9}, {3.607756041820266*^9, 3.607756235158769*^9}, {
   3.607756265338065*^9, 3.607756695214362*^9}, {3.60775675993099*^9, 
   3.607756780425748*^9}, {3.607756881761227*^9, 3.607756900874431*^9}, {
   3.607756961094098*^9, 3.607756983196787*^9}, {3.607757022344024*^9, 
   3.607757055333418*^9}, {3.607757085667351*^9, 3.607757111534936*^9}, {
   3.607757197977873*^9, 3.607757199145531*^9}, {3.607757330357854*^9, 
   3.60775743137903*^9}, {3.607757462666469*^9, 3.607757581889098*^9}, {
   3.607757876893729*^9, 3.607758111544848*^9}, {3.607758155455974*^9, 
   3.607758199218492*^9}, {3.607758323359177*^9, 3.607758339597516*^9}, {
   3.607758376435853*^9, 3.607758424225084*^9}, {3.607758460245303*^9, 
   3.607758501619999*^9}, {3.607758541655157*^9, 3.607758577727251*^9}, {
   3.607758631060465*^9, 3.607758811977696*^9}, {3.610090761780022*^9, 
   3.610090763450218*^9}, {3.61009140862207*^9, 3.610091428031794*^9}, {
   3.610091466359445*^9, 3.610091620607135*^9}, 3.631447253493129*^9, {
   3.632546310915689*^9, 3.6325464223972893`*^9}, {3.6325464599374104`*^9, 
   3.632546491381032*^9}, {3.6325465422438517`*^9, 3.6325465755177917`*^9}, {
   3.6325466149662943`*^9, 3.632546615411107*^9}, {3.632546694757616*^9, 
   3.632546696051169*^9}, {3.632551592628064*^9, 3.632551603332053*^9}, {
   3.632551635562965*^9, 3.632551660176057*^9}, 3.6357576000859632`*^9, 
   3.6455199787344646`*^9, 3.645522509604714*^9, 3.6584865813306227`*^9, 
   3.6661731473660297`*^9, 3.666440058488317*^9},
 CellID->416978728],

Cell["\[EmptySquare]", "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellChangeTimes->{{3.60774566648836*^9, 3.607745679046682*^9}, 
   3.631447253493256*^9, 3.6357576000861263`*^9, 3.6455199787345877`*^9, 
   3.645522509604826*^9, 3.6584865813306227`*^9, 3.6661731473660297`*^9, 
   3.666440058488454*^9},
 TextAlignment->Right,
 CellID->1523883859]
}, Closed]],

Cell[TextData[{
 StyleBox["Corollary 2",
  FontWeight->"Bold"],
 "\nFor all ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"0", "\[NotEqual]", "c"}], "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", "n"]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", ",", 
    SubscriptBox["m", "1"], ",", 
    RowBox[{
     SubscriptBox["m", "2"], "\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalZ]", "n"]}]}], TraditionalForm]]],
 ": If ",
 Cell[BoxData[
  FormBox[
   RowBox[{"c", 
    SubsuperscriptBox[
     UnderscriptBox["\[CapitalDelta]", "_"], 
     RowBox[{"i", ",", "j"}], "a"], "c"}], TraditionalForm]]],
 ", then ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{
     SubscriptBox["m", "1"], ",", 
     SubscriptBox["m", "2"]}], ")"}], TraditionalForm]]],
 " constitute critical pair multipliers iff ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "\[NotEqual]", "j"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["m", "1"], "\[Element]", 
    SubsuperscriptBox["M", "c", "+"]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["m", "2"], "=", 
    RowBox[{"n", "-", 
     RowBox[{"n", "/", 
      RowBox[{"gcd", "(", 
       RowBox[{"c", ",", "n"}], ")"}]}], "+", 
     SubscriptBox["m", "1"]}]}], TraditionalForm]]],
 ", or vice versa (depending on whether ",
 Cell[BoxData[
  FormBox[
   RowBox[{"i", "=", "+"}], TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox[
   RowBox[{"j", "=", "+"}], TraditionalForm]]],
 "). In particular, there is at most one pair of critical pair multipliers \
for each ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 ", up to order."
}], "Text",
 CellChangeTimes->{{3.607759731248342*^9, 3.607759929035131*^9}, {
  3.607770787234996*^9, 3.607770789931838*^9}, {3.607847733407281*^9, 
  3.60784773901851*^9}, {3.607848267681558*^9, 3.6078482712395*^9}, {
  3.610091676260987*^9, 3.610091766895468*^9}},
 CellID->569233702]
}, Closed]],

Cell[CellGroupData[{

Cell["Auxiliary Functions", "Subsection",
 CellChangeTimes->{{3.6455218834382277`*^9, 3.6455218870151453`*^9}},
 CellID->1440146747],

Cell["", "OpenEnvironment",
 CellID->314144664],

Cell[CellGroupData[{

Cell["Postulate: mod and gcd", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.645521896436523*^9, 3.645521898483592*^9}, {
  3.658464984827286*^9, 3.6584649870580893`*^9}, {3.658465023695958*^9, 
  3.6584650352243786`*^9}},
 CellID->1280908502],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"x", ",", 
   RowBox[{
   "y", "\[Element]", "\[DoubleStruckCapitalZ]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 232123642}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.658464843406037*^9, 3.6584648480080447`*^9}},
 CellTags->{
  "ID:232123642", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->232123642],

Cell[BoxData[
 RowBox[{
  RowBox[{"gcd", "[", 
   RowBox[{"x", ",", "y"}], "]"}], "\[Equal]", 
  RowBox[{"\[Wolf]GCD", "[", 
   RowBox[{"x", ",", "y"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(gcd GCD)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2028940590", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.5941095148288*^9, 3.594109525534463*^9}, {
  3.619249442087902*^9, 3.619249453816342*^9}, {3.645521914181345*^9, 
  3.645521921550354*^9}, {3.658464852594453*^9, 3.658464855995259*^9}},
 CellTags->{
  "ID:2028940590", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "gcd GCD"},
 CellID->2028940590],

Cell[BoxData[
 RowBox[{
  RowBox[{"mod", "[", 
   RowBox[{"x", ",", "y"}], "]"}], "\[Equal]", 
  RowBox[{"\[Wolf]Mod", "[", 
   RowBox[{"x", ",", "y"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(mod Mod)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:166548655", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.658464860177066*^9, 3.658464870067484*^9}},
 CellTags->{
  "ID:166548655", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mod Mod"},
 CellID->166548655],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1224687606]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1474676546]
}, Closed]],

Cell[CellGroupData[{

Cell["Domain of Reduction-Integer-Quotient-Rings", "Subsection",
 CellChangeTimes->{{3.592043452298074*^9, 3.592043454336059*^9}, {
  3.645521943808042*^9, 3.645521959742894*^9}},
 CellID->1062008728],

Cell["", "OpenEnvironment",
 CellID->91468236],

Cell[CellGroupData[{

Cell["Definition: ReductionIQR", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.594104836760066*^9, 3.594104848260629*^9}, {
  3.6455219660728683`*^9, 3.645521968635228*^9}},
 CellID->1417579161],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "n", "\[Element]", "\[DoubleStruckCapitalN]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\IntegerQuotientRings.nb", 1363294779}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594104874065635*^9, 3.594104899354088*^9}, 
   3.594105369451941*^9},
 CellTags->{
  "ID:1363294779", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->1363294779],

Cell[BoxData[
 RowBox[{
  RowBox[{"ReductionIQR", "[", "n", "]"}], ":=", 
  UnderscriptBox["\[CapitalDelta]", "N"]}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\IntegerQuotientRings.nb", 1613423474}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594104919088824*^9, 3.594104923817059*^9}, {
   3.594104964092016*^9, 3.594104973801868*^9}, {3.594105046211753*^9, 
   3.594105053114089*^9}, 3.594105373292198*^9, 3.594110874575462*^9, {
   3.6584650557696147`*^9, 3.6584650743034477`*^9}},
 CellTags->{
  "ID:1613423474", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->1613423474],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"x", ",", "y", ",", "m", ",", "i", ",", "j"}]]], "GlobalDeclaration",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\IntegerQuotientRings.nb", 1525507482}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594105000212169*^9, 3.594105017698905*^9}, {
  3.595057138148614*^9, 3.595057139399322*^9}, {3.608549515038671*^9, 
  3.608549515214507*^9}},
 CellTags->{
  "ID:1525507482", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->1525507482],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Element]", "N"], 
   RowBox[{"[", "x", "]"}]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalZ]", 
     RowBox[{"0", ",", "\[Ellipsis]", ",", 
      RowBox[{"n", "-", "1"}]}]]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RIQR \[Element])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1587087117", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594105034842931*^9, 3.594105066511218*^9}, 
   3.594105374729335*^9, {3.595057141534171*^9, 3.595057142789765*^9}, {
   3.6584650641634293`*^9, 3.658465067813836*^9}, {3.6584671535047045`*^9, 
   3.658467163566722*^9}},
 CellTags->{
  "ID:1587087117", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR \[Element]"},
 CellID->1587087117],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["0", "N"], ":=", "0"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RIQR 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:447457872", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.594105114700154*^9, {3.594110969089559*^9, 3.594110972418441*^9}, {
   3.655187035031228*^9, 3.655187035374429*^9}},
 CellTags->{
  "ID:447457872", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR 0"},
 CellID->447457872],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["1", "N"], ":=", 
  RowBox[{"mod", "[", 
   RowBox[{"1", ",", "n"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RIQR 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:815094494", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.594105124094116*^9, {3.594110977291997*^9, 3.594110978252326*^9}, {
   3.655187038026434*^9, 3.655187038400834*^9}, {3.6584650847252655`*^9, 
   3.6584650880636716`*^9}},
 CellTags->{
  "ID:815094494", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR 1"},
 CellID->815094494],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["+", "N"], "y"}], ":=", 
  RowBox[{"mod", "[", 
   RowBox[{
    RowBox[{"x", "+", "y"}], ",", "n"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RIQR +)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1939288109", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594105131953688*^9, 3.594105135814897*^9}, {
  3.594110980759738*^9, 3.594110981202393*^9}, {3.6584650935704813`*^9, 
  3.6584650987808905`*^9}},
 CellTags->{
  "ID:1939288109", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR +"},
 CellID->1939288109],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["-", "N"], "y"}], ":=", 
  RowBox[{"mod", "[", 
   RowBox[{
    RowBox[{"x", "-", "y"}], ",", "n"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RIQR -)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:915824852", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.594105148747624*^9, {3.594110983893659*^9, 3.594110984373559*^9}, {
   3.658465101261295*^9, 3.6584651124933147`*^9}},
 CellTags->{
  "ID:915824852", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR -"},
 CellID->915824852],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["-", "N"], "x"}], ":=", 
  RowBox[{"mod", "[", 
   RowBox[{
    RowBox[{"-", "x"}], ",", "n"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RIQR - unary)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:370600110", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.594105158920952*^9, {3.594110986843653*^9, 3.594110987175801*^9}, {
   3.6584651142717175`*^9, 3.6584651191087265`*^9}},
 CellTags->{
  "ID:370600110", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR - unary"},
 CellID->370600110],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["*", "N"], "y"}], ":=", 
  RowBox[{"mod", "[", 
   RowBox[{
    RowBox[{"x", "*", "y"}], ",", "n"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RIQR *)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:344363438", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.594105168020003*^9, {3.594110989572531*^9, 3.594110989978691*^9}, {
   3.6584651355521555`*^9, 3.658465141682966*^9}},
 CellTags->{
  "ID:344363438", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR *"},
 CellID->344363438],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", 
    UnderscriptBox["\[Succeeds]", "N"], "y"}], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", ">", "y"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RIQR \[Succeeds])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1495376438", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594105181610315*^9, 3.594105187398871*^9}, {
  3.6584651465969744`*^9, 3.6584651487965784`*^9}},
 CellTags->{
  "ID:1495376438", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR \[Succeeds]"},
 CellID->1495376438],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["multN", "N"], "[", "x", "]"}], ":=", 
  "2"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RIQR multN)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:550500630", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594105203785643*^9, 3.5941052088691*^9}, {
   3.594110996684291*^9, 3.594110999599063*^9}, {3.607338461665509*^9, 
   3.607338466133825*^9}, 3.607678154116524*^9, {3.631855500712533*^9, 
   3.631855509767109*^9}, {3.631856528132408*^9, 3.6318565334834642`*^9}, {
   3.658465154256588*^9, 3.6584651737566223`*^9}},
 CellTags->{
  "ID:550500630", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR multN"},
 CellID->550500630],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "N"], "[", 
   RowBox[{"m", ",", "x"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["mult", "N"], "[", 
     RowBox[{"m", ",", "x", ",", "1"}], "]"}], "\[Or]", 
    RowBox[{
     UnderscriptBox["mult", "N"], "[", 
     RowBox[{"m", ",", "x", ",", "2"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RIQR mult)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:445166729", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594105224965305*^9, 3.594105236244283*^9}},
 CellTags->{
  "ID:445166729", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR mult"},
 CellID->445166729],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "N"], "[", 
   RowBox[{"m", ",", "x", ",", "1"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"\[And]", 
   RowBox[{"\[Piecewise]", GridBox[{
      {
       RowBox[{"m", "\[GreaterEqual]", "1"}]},
      {
       RowBox[{
        RowBox[{"quo", "[", 
         RowBox[{"n", ",", 
          RowBox[{"gcd", "[", 
           RowBox[{"n", ",", "x"}], "]"}]}], "]"}], ">", "m"}]}
     }]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RIQR mult 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1793329966", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594105281431392*^9, 3.594105293340124*^9}, {
  3.594109042289234*^9, 3.594109045043827*^9}, {3.594109113587823*^9, 
  3.594109169192286*^9}, {3.594109340937358*^9, 3.594109407134757*^9}, {
  3.594362832843971*^9, 3.594362835052852*^9}, {3.594362913058074*^9, 
  3.59436293110296*^9}, {3.607338474543362*^9, 3.607338475614283*^9}, {
  3.65907742119689*^9, 3.65907743863772*^9}, {3.659078367245951*^9, 
  3.659078369539155*^9}},
 CellTags->{
  "ID:1793329966", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR mult 1"},
 CellID->1793329966],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "N"], "[", 
   RowBox[{"m", ",", "x", ",", "2"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   UnderscriptBox["mult", "N"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["-", "N"], "m"}], ",", "x", ",", "1"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RIQR mult 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:126969281", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594105328470021*^9, 3.594105379261365*^9}, 
   3.5943628400215*^9},
 CellTags->{
  "ID:126969281", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR mult 2"},
 CellID->126969281],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["rdm", "N"], "[", 
   RowBox[{"x", ",", "y"}], "]"}], ":=", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["let", 
     RowBox[{"g", "=", 
      RowBox[{"gcd", "[", 
       RowBox[{"n", ",", "y"}], "]"}]}]], 
    RowBox[{
     UnderscriptBox["rdm", "N"], "[", 
     RowBox[{"y", ",", 
      RowBox[{
       RowBox[{"quo", "[", 
        RowBox[{"n", ",", "g"}], "]"}], "-", "1"}], ",", 
      RowBox[{"x", "-", 
       RowBox[{"mod", "[", 
        RowBox[{"x", ",", "g"}], "]"}]}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RIQR rdm 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:909359224", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594109556166621*^9, 3.594109636627035*^9}, 
   3.594109669500675*^9, {3.594110462658929*^9, 3.594110478119118*^9}, {
   3.607678172500238*^9, 3.607678178362167*^9}, {3.607679111984465*^9, 
   3.607679114811871*^9}, {3.607681838634085*^9, 3.607681902070806*^9}, {
   3.608005543695978*^9, 3.608005546646686*^9}, 3.608005612089332*^9, 
   3.658465246987151*^9, {3.6593391927147985`*^9, 3.659339208268026*^9}, {
   3.659339515137165*^9, 3.6593395957269063`*^9}, {3.659339646411395*^9, 
   3.659339677081049*^9}, {3.659339809557482*^9, 3.6593398253291097`*^9}, {
   3.659339868572386*^9, 3.6593399359957037`*^9}, {3.659340065430132*^9, 
   3.6593400658981323`*^9}, {3.6593401356146545`*^9, 
   3.6593401358174553`*^9}, {3.65934016690831*^9, 3.6593401670331097`*^9}, {
   3.6593422574521713`*^9, 3.6593422575925713`*^9}},
 CellTags->{
  "ID:909359224", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR rdm 1"},
 CellID->909359224],

Cell[TextData[{
 Cell[BoxData[
  FormBox[
   RowBox[{"rdm", "[", 
    RowBox[{"x", ",", "m", ",", "y"}], "]"}], TraditionalForm]]],
 " returns ",
 Cell[BoxData[
  FormBox[
   RowBox[{"0", "<", "m", "<", 
    RowBox[{"quo", "[", 
     RowBox[{"n", ",", 
      RowBox[{"gcd", "[", 
       RowBox[{"n", ",", "x"}], "]"}]}], "]"}]}], TraditionalForm]]],
 " s.t. ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"m", " ", "x"}], "=", "y"}], TraditionalForm]]],
 ". Such a multiplier always exists (and is unique) for non-zero multiples ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"gcd", "[", 
    RowBox[{"n", ",", "x"}], "]"}], TraditionalForm]]],
 ", thanks to (mult1 sufficient \[ScriptCapitalZ]n)."
}], "EnvironmentText",
 CellChangeTimes->{{3.607682125516959*^9, 3.607682235533348*^9}, {
  3.6080054936402*^9, 3.608005515717215*^9}, {3.608005553321889*^9, 
  3.60800562938355*^9}, {3.659338938449952*^9, 3.6593390723293867`*^9}, {
  3.6593394215526004`*^9, 3.659339437152628*^9}, {3.6593396805442553`*^9, 
  3.659339772663417*^9}},
 CellID->2109204388],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["rdm", "N"], "[", 
   RowBox[{"x", ",", "m", ",", "y"}], "]"}], ":=", 
  RowBox[{"\[Piecewise]", GridBox[{
     {"m", "\[DoubleLeftArrow]", 
      RowBox[{
       RowBox[{"m", 
        UnderscriptBox["*", "N"], "x"}], "\[Equal]", "y"}]},
     {
      RowBox[{
       UnderscriptBox["rdm", "N"], "[", 
       RowBox[{"x", ",", 
        RowBox[{"m", "-", "1"}], ",", "y"}], "]"}], "\[DoubleLeftArrow]", 
      "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RIQR rdm 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:236078685", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.607673136406166*^9, 3.607673259814656*^9}, {
   3.6076781203412*^9, 3.607678132705293*^9}, {3.607678919319852*^9, 
   3.607679055342464*^9}, {3.607679098082561*^9, 3.60767909974527*^9}, {
   3.607679136662236*^9, 3.607679140740476*^9}, {3.607679329165932*^9, 
   3.607679370609371*^9}, {3.607679401404744*^9, 3.607679466191444*^9}, {
   3.607679622805874*^9, 3.607679665045413*^9}, {3.607679750675228*^9, 
   3.607679756835971*^9}, {3.607679807565009*^9, 3.607679846276323*^9}, {
   3.607681814912329*^9, 3.607681821789185*^9}, {3.607765769544471*^9, 
   3.607765772396796*^9}, {3.60800561385939*^9, 3.608005616498851*^9}, {
   3.6584652627587786`*^9, 3.6584652783588057`*^9}, {3.659339077758197*^9, 
   3.6593391386919036`*^9}, 3.6593393144272127`*^9, {3.659339354722083*^9, 
   3.659339391600548*^9}, {3.6593398355159273`*^9, 3.6593398358279285`*^9}, {
   3.6593402060331783`*^9, 3.6593402062671785`*^9}, 3.6593404641158323`*^9, {
   3.659340494270685*^9, 3.6593405130999184`*^9}},
 CellTags->{
  "ID:236078685", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR rdm 2"},
 CellID->236078685],

Cell[TextData[{
 "Although in ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalZ]", "n"], TraditionalForm]]],
 " every one-element set is already a Gr\[ODoubleDot]bner basis, we cannot \
change the definition of ",
 Cell[BoxData[
  FormBox["lcrd", TraditionalForm]]],
 " to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"lcrd", "[", 
     RowBox[{"x", ",", "x"}], "]"}], "=", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 ": Although this would work in ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalZ]", "n"], TraditionalForm]]],
 ", it does not work in ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[DoubleStruckCapitalZ]", "n"], "[", "X", "]"}], 
   TraditionalForm]]],
 " any more, since in ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[DoubleStruckCapitalZ]", "n"], "[", "X", "]"}], 
   TraditionalForm]]],
 " there are one-element sets which are no Gr\[ODoubleDot]bner bases."
}], "EnvironmentText",
 CellChangeTimes->{{3.608006704413429*^9, 3.608006784348903*^9}, {
   3.608006829310563*^9, 3.60800687588128*^9}, 3.6325520774406023`*^9},
 CellID->61492128],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lcrd", "N"], "[", 
   RowBox[{"x", ",", "y", ",", "i", ",", "j"}], "]"}], ":=", 
  RowBox[{"\[Piecewise]", GridBox[{
     {
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        SubscriptBox["min", 
         UnderscriptBox["\[Succeeds]", "N"]], "[", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{"gcd", "[", 
           RowBox[{"n", ",", "x"}], "]"}], ",", 
          RowBox[{"gcd", "[", 
           RowBox[{"n", ",", "y"}], "]"}]}], "\[RightAngleBracket]"}], "]"}], 
       "\[RightAngleBracket]"}], "\[DoubleLeftArrow]", 
      RowBox[{
       RowBox[{"x", "\[NotEqual]", "y"}], "\[Or]", 
       RowBox[{"i", "\[NotEqual]", "j"}]}]},
     {
      RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
      "\[DoubleLeftArrow]", "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RIQR lcrd)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:943334598", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594109651423661*^9, 3.594109682707189*^9}, {
   3.594109991272098*^9, 3.594109991753599*^9}, {3.607338487602607*^9, 
   3.607338490925494*^9}, {3.607679185065008*^9, 3.607679187781972*^9}, {
   3.607679224144979*^9, 3.607679226603054*^9}, {3.608005918950354*^9, 
   3.608005944859809*^9}, {3.60800662527872*^9, 3.608006630263208*^9}, {
   3.608549519163122*^9, 3.608549519941134*^9}, {3.610085398584525*^9, 
   3.610085400542817*^9}, {3.610085447557275*^9, 3.610085560641809*^9}, {
   3.610090026337256*^9, 3.610090031220638*^9}, {3.61009223181641*^9, 
   3.610092289259481*^9}, {3.610092341839886*^9, 3.610092403598137*^9}, {
   3.610095950952421*^9, 3.610095974148316*^9}, {3.610097125800898*^9, 
   3.610097144227715*^9}, 3.632552201302597*^9, {3.6592494630717554`*^9, 
   3.6592494698733673`*^9}, {3.6592495479359045`*^9, 3.659249549963908*^9}, {
   3.659332265862277*^9, 3.659332285877112*^9}},
 CellTags->{
  "ID:943334598", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR lcrd"},
 CellID->943334598],

Cell[TextData[{
 "In the definition above note that ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["min", 
    UnderscriptBox["\[Succeeds]", "N"]], TraditionalForm]]],
 " in fact gives the ",
 StyleBox["maximum",
  FontWeight->"Bold"],
 " of its arguments!"
}], "EnvironmentText",
 CellChangeTimes->{{3.610092298554093*^9, 3.610092336035374*^9}, {
  3.610094214803531*^9, 3.610094216759786*^9}},
 CellID->119714818],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lcrd", "N"], "[", "x", "]"}], ":=", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"gcd", "[", 
    RowBox[{"n", ",", "x"}], "]"}], 
   "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RIQR lcrd unary)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1868934214", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594457418951249*^9, 3.594457437819868*^9}, {
   3.608005763790365*^9, 3.608005770926602*^9}, {3.608005877596179*^9, 
   3.608005878777312*^9}, {3.610097155619687*^9, 3.610097171256186*^9}, {
   3.658465295440836*^9, 3.6584652956592364`*^9}, 3.659249473586174*^9, {
   3.6592495585283227`*^9, 3.659249560197526*^9}},
 CellTags->{
  "ID:1868934214", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR lcrd unary"},
 CellID->1868934214],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["cpm", "N"], "[", 
   RowBox[{"x", ",", "y", ",", "m", ",", "i", ",", "j"}], "]"}], ":=", 
  RowBox[{"\[Piecewise]", GridBox[{
     {
      RowBox[{
       UnderscriptBox["let", 
        RowBox[{"m1", "=", 
         RowBox[{
          UnderscriptBox["rdm", "N"], "[", 
          RowBox[{"m", ",", "y"}], "]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"m1", ",", 
              RowBox[{"m1", 
               UnderscriptBox["-", "N"], 
               RowBox[{"quo", "[", 
                RowBox[{"n", ",", 
                 RowBox[{"gcd", "[", 
                  RowBox[{"n", ",", "y"}], "]"}]}], "]"}]}]}], 
             "\[RightAngleBracket]"}], "\[DoubleLeftArrow]", 
            RowBox[{
             UnderscriptBox["mult", "N"], "[", 
             RowBox[{"m1", ",", "y", ",", "1"}], "]"}]},
           {
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
              RowBox[{"m1", 
               UnderscriptBox["+", "N"], 
               RowBox[{"quo", "[", 
                RowBox[{"n", ",", 
                 RowBox[{"gcd", "[", 
                  RowBox[{"n", ",", "y"}], "]"}]}], "]"}]}], ",", "m1"}], 
             "\[RightAngleBracket]"}], "\[DoubleLeftArrow]", "otherwise"}
          }]}], 
        TagBox[")",
         "AutoParentheses"]}]}], "\[DoubleLeftArrow]", 
      RowBox[{"x", "\[Equal]", "y"}]},
     {
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{
         UnderscriptBox["rdm", "N"], "[", 
         RowBox[{"m", ",", "x"}], "]"}], ",", 
        RowBox[{
         UnderscriptBox["rdm", "N"], "[", 
         RowBox[{"m", ",", "y"}], "]"}]}], "\[RightAngleBracket]"}], 
      "\[DoubleLeftArrow]", "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RIQR cpm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:424764915", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594459276124169*^9, 3.594459296097016*^9}, {
   3.594459345205451*^9, 3.594459346743078*^9}, {3.595057156838874*^9, 
   3.59505715968297*^9}, {3.607338494540954*^9, 3.607338499801845*^9}, {
   3.607682504939297*^9, 3.607682543498815*^9}, {3.607759621800696*^9, 
   3.607759667002015*^9}, {3.607759963254842*^9, 3.60775996661566*^9}, {
   3.607760197716301*^9, 3.607760198051169*^9}, {3.607760242606078*^9, 
   3.607760245233631*^9}, {3.607760277479782*^9, 3.607760280099468*^9}, {
   3.607760392036752*^9, 3.607760394659064*^9}, {3.607760435154085*^9, 
   3.607760476408328*^9}, {3.607761454599673*^9, 3.607761454751417*^9}, {
   3.607761536236946*^9, 3.607761539689967*^9}, {3.607761857165654*^9, 
   3.607761892609887*^9}, 3.608005790326888*^9, {3.608005958611448*^9, 
   3.60800597189048*^9}, {3.608006637822584*^9, 3.60800664291807*^9}, {
   3.610084446135542*^9, 3.610084447420639*^9}, {3.610085300512918*^9, 
   3.610085301768853*^9}, {3.610093755769433*^9, 3.610093756366853*^9}, {
   3.610094055747824*^9, 3.61009408588284*^9}, {3.658465308045658*^9, 
   3.658465316594473*^9}, 3.659077812290577*^9, {3.6590813153469315`*^9, 
   3.659081321103341*^9}, {3.6591832456565156`*^9, 3.6591832525049276`*^9}, {
   3.6593343191024876`*^9, 3.659334354202549*^9}, {3.6593344038262362`*^9, 
   3.6593344061506405`*^9}, {3.659334438926298*^9, 3.6593344494563165`*^9}},
 CellTags->{
  "ID:424764915", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR cpm"},
 CellID->424764915],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->89671333]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1623525075]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Tests", "Section",
 CellChangeTimes->{{3.592043465736425*^9, 3.592043466408566*^9}},
 CellID->496748878],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 FormBox[
  SubscriptBox["\[DoubleStruckCapitalZ]", "24"], 
  TraditionalForm]]]], "Subsection",
 CellChangeTimes->{{3.6455225584350986`*^9, 3.645522561882739*^9}},
 CellID->225277583],

Cell["", "OpenEnvironment",
 CellID->2008906006],

Cell[CellGroupData[{

Cell[TextData[{
 "Definition: Reduction-Integer-Quotient-Ring ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalZ]", "24"], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.594110685370442*^9, 3.594110695537398*^9}, {
  3.6455227256334753`*^9, 3.645522741812976*^9}},
 CellID->957072583],

Cell[BoxData[
 RowBox[{"\[ScriptCapitalZ]24", ":=", 
  RowBox[{"ReductionIQR", "[", "24", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalZ]24)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:580511390", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594110710628823*^9, 3.594110733816923*^9}, {
  3.6457799262425823`*^9, 3.6457799291049128`*^9}},
 CellTags->{
  "ID:580511390", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[ScriptCapitalZ]24"},
 CellID->580511390],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1484947434]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->172883714],

Cell[CellGroupData[{

Cell["rdm", "Subsubsection",
 CellChangeTimes->{{3.645522551077071*^9, 3.6455225514107857`*^9}},
 CellID->1321677136],

Cell["\<\
rdm might return different results now, because its implementation has been \
changed in order to return the \[OpenCurlyDoubleQuote]optimal\
\[CloseCurlyDoubleQuote] multiplier, i.e. the multiplier that yields the \
least result.\
\>", "Text",
 CellChangeTimes->{{3.659424286429058*^9, 3.659424323089123*^9}},
 CellID->47140036],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["rdm", "\[ScriptCapitalZ]24"], "[", 
  RowBox[{"5", ",", "4"}], "]"}]], "Computation",
 CellChangeTimes->{{3.594111214625435*^9, 3.594111242477395*^9}, {
  3.594111275994052*^9, 3.594111283546915*^9}, {3.607683946688327*^9, 
  3.607683949037904*^9}, {3.607684931006103*^9, 3.607684936005513*^9}, {
  3.6457800910480747`*^9, 3.645780093118435*^9}},
 CellID->849198139],

Cell[BoxData[
 FormBox["1", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.607683902169916*^9, 3.607683949751948*^9, {3.607684931775777*^9, 
   3.607684936465714*^9}, 3.607760135664708*^9, 3.608006089125455*^9, 
   3.610096556145528*^9, 3.615713811887597*^9, 3.631447355256798*^9, 
   3.6357579474709997`*^9, 3.645522651892364*^9},
 CellID->1460150010]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c849198139"]]]]}, {None, None}},
 CellChangeTimes->{3.645522652307527*^9},
 CellID->1244254393],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c849198139-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645522652558651*^9},
 CellID->1827885036]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["rdm", "\[ScriptCapitalZ]24"], "[", 
  RowBox[{"22", ",", "23"}], "]"}]], "Computation",
 CellChangeTimes->{{3.645522703588517*^9, 3.6455227121886663`*^9}, 
   3.645780096796762*^9},
 CellID->570943662],

Cell[BoxData[
 FormBox["2", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645522713052698*^9},
 CellID->1414651157]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c570943662"]]]]}, {None, None}},
 CellChangeTimes->{3.645522713389352*^9},
 CellID->923325515],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c570943662-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645522713541551*^9},
 CellID->393895088]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["lcrd", "Subsubsection",
 CellChangeTimes->{{3.645522605224699*^9, 3.6455226058175697`*^9}},
 CellID->285408563],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalZ]24"], "[", 
  RowBox[{"7", ",", "6", ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.594110656059297*^9, 3.594110669946294*^9}, {
   3.594110755766045*^9, 3.594110765474773*^9}, {3.610096479918596*^9, 
   3.61009648145479*^9}, 3.645780099743249*^9},
 CellID->490840057],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", "6", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{
  3.607683878732878*^9, {3.607760118067693*^9, 3.607760131377644*^9}, 
   3.608006080420492*^9, {3.610096482103439*^9, 3.610096490111207*^9}, 
   3.615713802443147*^9, 3.615801268233673*^9, {3.631447337854094*^9, 
   3.6314473417219067`*^9}, 3.631449154710539*^9, 3.6357578686417503`*^9, 
   3.6455226574830513`*^9},
 CellID->1920791119]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c490840057"]]]]}, {None, None}},
 CellChangeTimes->{3.645522657943169*^9},
 CellID->1840123044],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c490840057-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645522658089881*^9},
 CellID->2050404037]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalZ]24"], "[", 
  RowBox[{"7", ",", "7", ",", "1", ",", "2"}], "]"}]], "Computation",
 CellChangeTimes->{{3.594110928386809*^9, 3.594110940511361*^9}, {
   3.610096510169598*^9, 3.610096517925804*^9}, 3.645780101144368*^9},
 CellID->260906778],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", "1", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{{3.607683889381804*^9, 3.607683897975522*^9}, 
   3.607760133784252*^9, 3.608006086767435*^9, 3.610096518762723*^9, 
   3.615713807969942*^9, 3.6314473489339437`*^9, 3.635757873148114*^9, 
   3.63575796433746*^9, 3.6455226603156233`*^9},
 CellID->1643593823]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c260906778"]]]]}, {None, None}},
 CellChangeTimes->{3.6455226607033377`*^9},
 CellID->1414827049],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c260906778-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645522660852499*^9},
 CellID->1588362347]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalZ]24"], "[", 
  RowBox[{"14", ",", "8", ",", "1", ",", "2"}], "]"}]], "Computation",
 CellChangeTimes->{{3.610096521505255*^9, 3.610096546447486*^9}, {
   3.6455227933432217`*^9, 3.6455228237173023`*^9}, 3.645780102508666*^9},
 CellID->2002455362],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", "8", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{
  3.610096547796865*^9, 3.615713809854416*^9, 3.631447352548233*^9, 
   3.635757883299739*^9, 3.645522663113171*^9, {3.645522794311151*^9, 
   3.645522824206806*^9}},
 CellID->867934889]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c2002455362"]]]]}, {None, None}},
 CellChangeTimes->{3.645522824608921*^9},
 CellID->1370538128],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c2002455362-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645522824770362*^9},
 CellID->1552535153]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalZ]24"], "[", 
  RowBox[{"10", ",", "14", ",", "1", ",", "2"}], "]"}]], "Computation",
 CellChangeTimes->{{3.607760141127187*^9, 3.607760154960184*^9}, {
   3.608006103869923*^9, 3.608006115786098*^9}, {3.610096559432817*^9, 
   3.610096560104798*^9}, 3.645780105338708*^9},
 CellID->1432418796],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", "2", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{
  3.607760155713274*^9, {3.60800609192918*^9, 3.608006116315926*^9}, 
   3.610096560796199*^9, 3.615713814015136*^9, 3.6314473578009663`*^9, 
   3.63575795026693*^9, 3.645522665213388*^9},
 CellID->1002747542]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1432418796"]]]]}, {None, None}},
 CellChangeTimes->{3.6455226655830927`*^9},
 CellID->1780779641],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1432418796-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6455226657247753`*^9},
 CellID->131149116]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalZ]24"], "[", 
  RowBox[{"8", ",", "11", ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.607760504024219*^9, 3.607760520591159*^9}, {
   3.607760605983363*^9, 3.607760606130817*^9}, {3.610096673843647*^9, 
   3.6100966743671*^9}, 3.645780106850081*^9},
 CellID->1091363187],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", "8", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.607760521597375*^9, 3.607760607109251*^9, 
  3.608006125104062*^9, 3.610096675086797*^9, 3.615713818414739*^9, 
  3.6314473631988487`*^9, 3.635757955019883*^9, 3.6455226672010527`*^9},
 CellID->673303920]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1091363187"]]]]}, {None, None}},
 CellChangeTimes->{3.645522667550952*^9},
 CellID->1502806330],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1091363187-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645522667704516*^9},
 CellID->1911587008]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["cpm", "Subsubsection",
 CellChangeTimes->{{3.645522615089941*^9, 3.645522615424081*^9}},
 CellID->840789324],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["cpm", "\[ScriptCapitalZ]24"], "[", 
  RowBox[{"7", ",", "6", ",", "6", ",", "1", ",", "1"}], "]"}]], "Computation",\

 CellChangeTimes->{
  3.60768400471224*^9, {3.610096501406932*^9, 3.610096502041631*^9}, 
   3.6457801106049747`*^9},
 CellID->710437623],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"4", ",", "1"}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{
  3.607684006206845*^9, {3.607760120005802*^9, 3.607760129642332*^9}, 
   3.608006084089877*^9, 3.61009650343228*^9, 3.615713804610086*^9, 
   3.615801270056391*^9, 3.631447345388959*^9, 3.631449157409095*^9, 
   3.635757870922874*^9, 3.6455226697702103`*^9},
 CellID->442317352]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c710437623"]]]]}, {None, None}},
 CellChangeTimes->{3.6455226701381702`*^9},
 CellID->1642092347],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c710437623-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645522670283531*^9},
 CellID->93258772]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["cpm", "\[ScriptCapitalZ]24"], "[", 
  RowBox[{"10", ",", "14", ",", "2", ",", "1", ",", "2"}], 
  "]"}]], "Computation",
 CellChangeTimes->{{3.607760162066146*^9, 3.607760176447883*^9}, 
   3.608006118593536*^9, {3.61009656723322*^9, 3.610096567675655*^9}, 
   3.645780112153504*^9},
 CellID->1351181837],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"5", ",", "7"}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{{3.607760177170075*^9, 3.607760201746509*^9}, 
   3.607760249517706*^9, 3.607760284420463*^9, 3.607760398410359*^9, 
   3.607760480845432*^9, {3.608006095078285*^9, 3.608006119102674*^9}, 
   3.610096569944367*^9, 3.61571381650738*^9, 3.6314473607526817`*^9, 
   3.635757952724628*^9, 3.645522671789156*^9},
 CellID->1528111983]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1351181837"]]]]}, {None, None}},
 CellChangeTimes->{3.645522672163439*^9},
 CellID->100484650],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1351181837-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6455226723076572`*^9},
 CellID->345002316]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["cpm", "\[ScriptCapitalZ]24"], "[", 
  RowBox[{"8", ",", "11", ",", "8", ",", "1", ",", "2"}], 
  "]"}]], "Computation",
 CellChangeTimes->{{3.607760530302698*^9, 3.607760533112867*^9}, {
   3.607760610564391*^9, 3.607760610677154*^9}, {3.610096677634167*^9, 
   3.610096678059831*^9}, 3.645780113930798*^9},
 CellID->140107252],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"1", ",", "5"}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.607760534170178*^9, 3.607760611178798*^9, 
  3.60800612725234*^9, 3.610096678752975*^9, 3.615713820489908*^9, 
  3.631447365952324*^9, 3.63575795726125*^9, 3.64552267395501*^9},
 CellID->1619126148]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c140107252"]]]]}, {None, None}},
 CellChangeTimes->{3.645522674308649*^9},
 CellID->315583679],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c140107252-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6455226744502296`*^9},
 CellID->2027695778]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 FormBox[
  SubscriptBox["\[DoubleStruckCapitalZ]", "13"], 
  TraditionalForm]]]], "Subsection",
 CellChangeTimes->{{3.645522568234461*^9, 3.645522572707737*^9}},
 CellID->625973727],

Cell["", "OpenEnvironment",
 CellID->187010305],

Cell[CellGroupData[{

Cell[TextData[{
 "Definition",
 ": Reduction-Integer-Quotient-Ring ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalZ]", "13"], TraditionalForm]]]
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6455228412604713`*^9, 3.645522843630392*^9}},
 CellID->862011211],

Cell[BoxData[
 RowBox[{"\[ScriptCapitalZ]13", ":=", 
  RowBox[{"ReductionIQR", "[", "13", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalZ]13)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:275453931", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.645522847911118*^9, 3.6455228688651857`*^9}, {
  3.6457799477099857`*^9, 3.6457799499969683`*^9}},
 CellTags->{
  "ID:275453931", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[ScriptCapitalZ]13"},
 CellID->275453931],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1991026502]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1516776967],

Cell[CellGroupData[{

Cell["rdm", "Subsubsection",
 CellChangeTimes->{{3.645522551077071*^9, 3.6455225514107857`*^9}},
 CellID->932766472],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["rdm", "\[ScriptCapitalZ]13"], "[", 
  RowBox[{"4", ",", "5"}], "]"}]], "Computation",
 CellChangeTimes->{{3.594111214625435*^9, 3.594111242477395*^9}, {
  3.594111275994052*^9, 3.594111283546915*^9}, {3.607683946688327*^9, 
  3.607683949037904*^9}, {3.607684931006103*^9, 3.607684936005513*^9}, {
  3.645522897915348*^9, 3.645522898100005*^9}, {3.645522963919161*^9, 
  3.645522964265613*^9}, {3.645780121992927*^9, 3.645780124486533*^9}},
 CellID->889827160],

Cell[BoxData[
 FormBox["3", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.607683902169916*^9, 3.607683949751948*^9, {3.607684931775777*^9, 
   3.607684936465714*^9}, 3.607760135664708*^9, 3.608006089125455*^9, 
   3.610096556145528*^9, 3.615713811887597*^9, 3.631447355256798*^9, 
   3.6357579474709997`*^9, 3.645522651892364*^9, 3.645522909015627*^9, 
   3.6455229652009697`*^9},
 CellID->1729630962]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c889827160"]]]]}, {None, None}},
 CellChangeTimes->{3.645522965625964*^9},
 CellID->722095009],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c889827160-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645522965764893*^9},
 CellID->996548225]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["rdm", "\[ScriptCapitalZ]13"], "[", 
  RowBox[{"12", ",", "6"}], "]"}]], "Computation",
 CellChangeTimes->{{3.645522703588517*^9, 3.6455227121886663`*^9}, {
   3.64552291476108*^9, 3.645522925188521*^9}, 3.6457801294097137`*^9},
 CellID->81011551],

Cell[BoxData[
 FormBox["1", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{
  3.645522713052698*^9, {3.6455229182761173`*^9, 3.645522926043532*^9}},
 CellID->1196950894]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c81011551"]]]]}, {None, None}},
 CellChangeTimes->{3.645522926445571*^9},
 CellID->1841495814],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c81011551-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.64552292658563*^9},
 CellID->658681144]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["rdm", "\[ScriptCapitalZ]13"], "[", 
  RowBox[{"6", ",", "12"}], "]"}]], "Computation",
 CellChangeTimes->{{3.645522948840391*^9, 3.645522950885263*^9}, 
   3.6457801309585543`*^9},
 CellID->763410701],

Cell[BoxData[
 FormBox["7", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645522951406818*^9},
 CellID->4820217]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c763410701"]]]]}, {None, None}},
 CellChangeTimes->{3.645522951880105*^9},
 CellID->2064421696],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c763410701-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645522952044159*^9},
 CellID->949097751]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["lcrd", "Subsubsection",
 CellChangeTimes->{{3.645522605224699*^9, 3.6455226058175697`*^9}},
 CellID->449176887],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalZ]13"], "[", 
  RowBox[{"7", ",", "6", ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.594110656059297*^9, 3.594110669946294*^9}, {
   3.594110755766045*^9, 3.594110765474773*^9}, {3.610096479918596*^9, 
   3.61009648145479*^9}, {3.645522972411415*^9, 3.645522972639352*^9}, 
   3.645780133949842*^9},
 CellID->1410083086],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", "1", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{
  3.607683878732878*^9, {3.607760118067693*^9, 3.607760131377644*^9}, 
   3.608006080420492*^9, {3.610096482103439*^9, 3.610096490111207*^9}, 
   3.615713802443147*^9, 3.615801268233673*^9, {3.631447337854094*^9, 
   3.6314473417219067`*^9}, 3.631449154710539*^9, 3.6357578686417503`*^9, 
   3.6455226574830513`*^9, 3.645522974733118*^9},
 CellID->490732742]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1410083086"]]]]}, {None, None}},
 CellChangeTimes->{3.645522975170835*^9},
 CellID->343184513],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1410083086-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645522975318368*^9},
 CellID->2078259407]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalZ]13"], "[", 
  RowBox[{"12", ",", "6", ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.594110928386809*^9, 3.594110940511361*^9}, {
   3.610096510169598*^9, 3.610096517925804*^9}, {3.645522977578589*^9, 
   3.6455229999226837`*^9}, 3.645780135460958*^9},
 CellID->158220634],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", "1", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{{3.607683889381804*^9, 3.607683897975522*^9}, 
   3.607760133784252*^9, 3.608006086767435*^9, 3.610096518762723*^9, 
   3.615713807969942*^9, 3.6314473489339437`*^9, 3.635757873148114*^9, 
   3.63575796433746*^9, 3.6455226603156233`*^9, {3.645522980131386*^9, 
   3.6455230007469177`*^9}},
 CellID->830283031]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c158220634"]]]]}, {None, None}},
 CellChangeTimes->{3.645523001179059*^9},
 CellID->1491837892],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c158220634-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.64552300132889*^9},
 CellID->1317010961]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalZ]13"], "[", 
  RowBox[{"4", ",", "8", ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.610096521505255*^9, 3.610096546447486*^9}, {
   3.6455227933432217`*^9, 3.6455228237173023`*^9}, {3.6455230123223667`*^9, 
   3.645523024169672*^9}, 3.645780137192944*^9},
 CellID->676617568],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", "1", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{
  3.610096547796865*^9, 3.615713809854416*^9, 3.631447352548233*^9, 
   3.635757883299739*^9, 3.645522663113171*^9, {3.645522794311151*^9, 
   3.645522824206806*^9}, {3.645523022341157*^9, 3.6455230246674213`*^9}},
 CellID->1584131273]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c676617568"]]]]}, {None, None}},
 CellChangeTimes->{3.645523025070633*^9},
 CellID->494933475],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c676617568-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645523025219475*^9},
 CellID->1500053850]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalZ]13"], "[", 
  RowBox[{"10", ",", "3", ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.607760141127187*^9, 3.607760154960184*^9}, {
   3.608006103869923*^9, 3.608006115786098*^9}, {3.610096559432817*^9, 
   3.610096560104798*^9}, {3.645523048405254*^9, 3.645523055656554*^9}, 
   3.645780138778131*^9},
 CellID->1992980618],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", "1", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{
  3.607760155713274*^9, {3.60800609192918*^9, 3.608006116315926*^9}, 
   3.610096560796199*^9, 3.615713814015136*^9, 3.6314473578009663`*^9, 
   3.63575795026693*^9, 3.645522665213388*^9, 3.645523056480823*^9},
 CellID->1494517660]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1992980618"]]]]}, {None, None}},
 CellChangeTimes->{3.645523056915079*^9},
 CellID->1378990324],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1992980618-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645523057058683*^9},
 CellID->1348975192]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["lcrd", "\[ScriptCapitalZ]13"], "[", 
  RowBox[{"2", ",", "11", ",", "1", ",", "1"}], "]"}]], "Computation",
 CellChangeTimes->{{3.607760504024219*^9, 3.607760520591159*^9}, {
   3.607760605983363*^9, 3.607760606130817*^9}, {3.610096673843647*^9, 
   3.6100966743671*^9}, {3.645523062979026*^9, 3.645523066839291*^9}, 
   3.645780140322226*^9},
 CellID->168529735],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", "1", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{
  3.607760521597375*^9, 3.607760607109251*^9, 3.608006125104062*^9, 
   3.610096675086797*^9, 3.615713818414739*^9, 3.6314473631988487`*^9, 
   3.635757955019883*^9, 3.6455226672010527`*^9, {3.6455230638708153`*^9, 
   3.645523067333045*^9}},
 CellID->699749025]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c168529735"]]]]}, {None, None}},
 CellChangeTimes->{3.6455230677451143`*^9},
 CellID->1668924093],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c168529735-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645523067891382*^9},
 CellID->225927190]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["cpm", "Subsubsection",
 CellChangeTimes->{{3.645522615089941*^9, 3.645522615424081*^9}},
 CellID->123861815],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["cpm", "\[ScriptCapitalZ]13"], "[", 
  RowBox[{"7", ",", "6", ",", "6", ",", "1", ",", "1"}], "]"}]], "Computation",\

 CellChangeTimes->{
  3.60768400471224*^9, {3.610096501406932*^9, 3.610096502041631*^9}, {
   3.6455230733196793`*^9, 3.645523073463696*^9}, {3.645780143453649*^9, 
   3.645780144891903*^9}},
 CellID->894685],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"2", ",", "1"}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{
  3.607684006206845*^9, {3.607760120005802*^9, 3.607760129642332*^9}, 
   3.608006084089877*^9, 3.61009650343228*^9, 3.615713804610086*^9, 
   3.615801270056391*^9, 3.631447345388959*^9, 3.631449157409095*^9, 
   3.635757870922874*^9, 3.6455226697702103`*^9, 3.645523077383296*^9},
 CellID->799097763]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c894685"]]]]}, {None, None}},
 CellChangeTimes->{3.645523077802918*^9},
 CellID->1700953392],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"c894685-display.m\"}]"]], \
{Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6455230779391403`*^9},
 CellID->1596585968]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["cpm", "\[ScriptCapitalZ]13"], "[", 
  RowBox[{"10", ",", "12", ",", "5", ",", "1", ",", "1"}], 
  "]"}]], "Computation",
 CellChangeTimes->{{3.607760162066146*^9, 3.607760176447883*^9}, 
   3.608006118593536*^9, {3.61009656723322*^9, 3.610096567675655*^9}, {
   3.645523080896915*^9, 3.645523107444398*^9}, 3.645780146558707*^9},
 CellID->788223205],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"3", ",", "8"}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{{3.607760177170075*^9, 3.607760201746509*^9}, 
   3.607760249517706*^9, 3.607760284420463*^9, 3.607760398410359*^9, 
   3.607760480845432*^9, {3.608006095078285*^9, 3.608006119102674*^9}, 
   3.610096569944367*^9, 3.61571381650738*^9, 3.6314473607526817`*^9, 
   3.635757952724628*^9, 3.645522671789156*^9, {3.645523097610463*^9, 
   3.645523108326117*^9}},
 CellID->157392879]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c788223205"]]]]}, {None, None}},
 CellChangeTimes->{3.6455231087499847`*^9},
 CellID->494916420],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c788223205-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645523108889827*^9},
 CellID->1558628290]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["cpm", "\[ScriptCapitalZ]13"], "[", 
  RowBox[{"6", ",", "7", ",", "8", ",", "1", ",", "1"}], "]"}]], "Computation",\

 CellChangeTimes->{{3.607760530302698*^9, 3.607760533112867*^9}, {
   3.607760610564391*^9, 3.607760610677154*^9}, {3.610096677634167*^9, 
   3.610096678059831*^9}, {3.6455230910499477`*^9, 3.6455231194438143`*^9}, 
   3.64578014792274*^9},
 CellID->2133581505],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"1", ",", "1"}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.607760534170178*^9, 3.607760611178798*^9, 
  3.60800612725234*^9, 3.610096678752975*^9, 3.615713820489908*^9, 
  3.631447365952324*^9, 3.63575795726125*^9, 3.64552267395501*^9, 
  3.645523119971776*^9},
 CellID->1665785973]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c2133581505"]]]]}, {None, None}},
 CellChangeTimes->{3.645523120368453*^9},
 CellID->661543208],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c2133581505-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6455231205111723`*^9},
 CellID->253352679]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Theorem: Reduction-Integer-Quotient-Rings Constitute Reduction Rings (\
\[Checkmark])\
\>", "Section",
 CellChangeTimes->{{3.592043458571405*^9, 3.592043460801741*^9}, {
  3.645521849621438*^9, 3.645521850950171*^9}, {3.645522035327516*^9, 
  3.645522048615306*^9}, {3.659424195915699*^9, 3.659424200158907*^9}},
 CellID->1529081576],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "n", "\[Element]", "\[DoubleStruckCapitalN]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\IntegerQuotientRings.nb", 7294098}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.658465412646842*^9, 3.6584654174360504`*^9}},
 CellTags->{
  "ID:7294098", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->7294098],

Cell["", "Text",
 CellID->231827409],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"\[ScriptCapitalZ]n", "=", 
   RowBox[{"ReductionIQR", "[", "n", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\IntegerQuotientRings.nb", 6216142}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6584654288708706`*^9, 3.6584654383868876`*^9}, {
  3.658465641191244*^9, 3.658465663358883*^9}, {3.6584657150749736`*^9, 
  3.6584657177425785`*^9}},
 CellTags->{
  "ID:6216142", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->6216142],

Cell[CellGroupData[{

Cell["Reduction Ring (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6584654567481194`*^9, 3.658465459228524*^9}, {
  3.6594241869300833`*^9, 3.659424191204491*^9}},
 CellID->509369914],

Cell[CellGroupData[{

Cell["Preparations (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6590812785308666`*^9, 3.6590812807928705`*^9}, {
  3.659088380907812*^9, 3.65908838516662*^9}},
 CellID->157842255],

Cell["", "OpenEnvironment",
 CellID->635903570],

Cell[CellGroupData[{

Cell["\<\
Lemma: Alternative Description of Notions in \[ScriptCapitalZ]n\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6584665122099757`*^9, 3.6584665235239964`*^9}, {
  3.658466695379898*^9, 3.6584666994359055`*^9}},
 CellID->369402263],

Cell[BoxData[
 RowBox[{"n", "\[NotEqual]", "0"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(modulus \[NotEqual] 0 \[ScriptCapitalZ]n)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:285451873", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.658468730460079*^9, 3.6584687318172817`*^9}},
 CellTags->{
  "ID:285451873", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "modulus \[NotEqual] 0 \[ScriptCapitalZ]n"},
 CellID->285451873],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "x"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
      RowBox[{"[", "x", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"x", "\[Element]", "\[DoubleStruckCapitalZ]"}]},
         {
          RowBox[{"x", "\[GreaterEqual]", "0"}]},
         {
          RowBox[{"n", ">", "x"}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] alternative \[ScriptCapitalZ]n)", "FrameLabel"],
         "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:270643577", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6584667228681464`*^9, 3.6584667532257996`*^9}},
 CellTags->{
  "ID:270643577", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Element] alternative \[ScriptCapitalZ]n"},
 
 CellID->270643577],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"ReductionIQR", "[", "1", "]"}]], 
    RowBox[{"[", "x", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", "\[Equal]", "0"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalZ]1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:916172587", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6590779556704283`*^9, 3.659077981285673*^9}},
 CellTags->{
  "ID:916172587", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[ScriptCapitalZ]1"},
 CellID->916172587],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
   RowBox[{"[", "x", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\IntegerQuotientRings.nb", 200088253}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6584676558229876`*^9, 3.658467659676194*^9}, {
  3.658467751130555*^9, 3.658467757417366*^9}},
 CellTags->{
  "ID:200088253", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->200088253],

Cell[BoxData[
 RowBox[{
  RowBox[{"\[LeftBracketingBar]", "x", "\[RightBracketingBar]"}], "\[Equal]", 
  "x"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(abs \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:62294620", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6590805534179926`*^9, 3.6590805585192013`*^9}},
 CellTags->{
  "ID:62294620", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "abs \[ScriptCapitalZ]n"},
 CellID->62294620],

Cell[BoxData[
 RowBox[{
  RowBox[{"mod", "[", 
   RowBox[{"x", ",", "n"}], "]"}], "\[Equal]", 
  "x"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(mod identity \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:90375616", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.658467668085609*^9, 3.6584676718764153`*^9}},
 CellTags->{
  "ID:90375616", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mod identity \[ScriptCapitalZ]n"},
 CellID->90375616],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["-", "\[ScriptCapitalZ]n"], "x"}], "\[Equal]", 
  RowBox[{"\[Piecewise]", GridBox[{
     {"x", "\[DoubleLeftArrow]", 
      RowBox[{"x", "\[Equal]", "0"}]},
     {
      RowBox[{"n", "-", "x"}], "\[DoubleLeftArrow]", "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(- unary alternative \[ScriptCapitalZ]n)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:575444601", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6584665289372053`*^9, 3.6584665474856377`*^9}, {
  3.658466604848939*^9, 3.6584666065181417`*^9}, {3.6584676629688*^9, 
  3.6584676639672017`*^9}, {3.6584689145746026`*^9, 3.658468926945424*^9}},
 CellTags->{
  "ID:575444601", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "- unary alternative \[ScriptCapitalZ]n"},
 CellID->575444601],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->4769084]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->179375776],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(modulus \[NotEqual] 0 \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(modulus \[NotEqual] 0 \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           StyleBox["n", "ExpressionVariable"], "\[NotEqual]", 
           "0"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           StyleBox["n", "ExpressionVariable"], "\[NotEqual]", 
           "0"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:285451873"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p285451873-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p285451873-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:285451873", "Proof|ID:285451873-1"},
 CellID->127340268],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p285451873-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->558949415]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] alternative \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] alternative \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             StyleBox["x", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "\[Element]", 
                 "\[DoubleStruckCapitalZ]"}], "\[And]", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", 
                 "0"}], "\[And]", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ">", 
                 StyleBox["x", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             StyleBox["x", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "\[Element]", 
                 "\[DoubleStruckCapitalZ]"}], "\[And]", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "\[GreaterEqual]", 
                 "0"}], "\[And]", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ">", 
                 StyleBox["x", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:270643577"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p270643577-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p270643577-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:270643577", "Proof|ID:270643577-1"},
 CellID->506264492],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p270643577-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->60003179]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[ScriptCapitalZ]1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[ScriptCapitalZ]1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", "1", "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["x", "ExpressionVariable"], "\[Equal]", 
           "0"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", "1", "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["x", "ExpressionVariable"], "\[Equal]", 
           "0"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:916172587"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p916172587-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p916172587-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:916172587", "Proof|ID:916172587-1"},
 CellID->265170607],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p916172587-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->351314708]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(abs \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(abs \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"\[LeftBracketingBar]", 
             StyleBox["x", "ExpressionVariable"], "\[RightBracketingBar]"}], 
           "\[Equal]", 
           StyleBox[
           "x", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"\[LeftBracketingBar]", 
             StyleBox["x", "ExpressionVariable"], "\[RightBracketingBar]"}], 
           "\[Equal]", 
           StyleBox[
           "x", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:62294620"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p62294620-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p62294620-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:62294620", "Proof|ID:62294620-1"},
 CellID->319229321],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p62294620-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->301185869]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(mod identity \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(mod identity \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"mod", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["n", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           StyleBox[
           "x", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"mod", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["n", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           StyleBox[
           "x", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:90375616"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p90375616-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p90375616-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:90375616", "Proof|ID:90375616-1"},
 CellID->190812059],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p90375616-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->751344191]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(- unary alternative \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(- unary alternative \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["Minus", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
           RowBox[{"\[Piecewise]", 
             GridBox[{{
                StyleBox["x", "ExpressionVariable"], "\[DoubleLeftArrow]", 
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Equal]", "0"}]}, {
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "-", 
                  StyleBox["x", "ExpressionVariable"]}], "\[DoubleLeftArrow]",
                 "True"}}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["Minus", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             StyleBox["x", "ExpressionVariable"], "]"}], "\[Equal]", 
           RowBox[{"\[Piecewise]", 
             GridBox[{{
                StyleBox["x", "ExpressionVariable"], "\[DoubleLeftArrow]", 
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], "\[Equal]", "0"}]}, {
                RowBox[{
                  StyleBox["n", "ExpressionVariable"], "-", 
                  StyleBox["x", "ExpressionVariable"]}], "\[DoubleLeftArrow]",
                 "True"}}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:575444601"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p575444601-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p575444601-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:575444601", "Proof|ID:575444601-1"},
 CellID->94536918],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p575444601-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->752059431]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->190598708],

Cell[CellGroupData[{

Cell["Lemma: Membership of Various Terms", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.659345735103279*^9, 3.6593457501417055`*^9}},
 CellID->637227383],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
   RowBox[{"[", "x", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\IntegerQuotientRings.nb", 974601634}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659345755898116*^9, 3.659345763339329*^9}},
 CellTags->{
  "ID:974601634", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->974601634],

Cell[BoxData[
 RowBox[{
  RowBox[{"gcd", "[", 
   RowBox[{"n", ",", "x"}], "]"}], "\[Element]", 
  "\[DoubleStruckCapitalN]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(gcd \[Element] \[DoubleStruckCapitalN] \[ScriptCapitalZ]n)", 
         "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:779775892", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6590834376441298`*^9, 3.6590834418561373`*^9}},
 CellTags->{
  "ID:779775892", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
   "gcd \[Element] \[DoubleStruckCapitalN] \[ScriptCapitalZ]n"},
 CellID->779775892],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
    RowBox[{"[", 
     RowBox[{"gcd", "[", 
      RowBox[{"n", ",", "x"}], "]"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership gcd \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:7430669", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659081174087683*^9, 3.6590811834164996`*^9}, {
  3.6590815637617674`*^9, 3.659081577879792*^9}},
 CellTags->{
  "ID:7430669", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership gcd \[ScriptCapitalZ]n", "IR"},
 CellID->7430669],

Cell[BoxData[
 RowBox[{
  RowBox[{"gcd", "[", 
   RowBox[{"n", ",", "x"}], "]"}], "\[NotEqual]", 
  UnderscriptBox["0", "\[ScriptCapitalZ]n"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(gcd \[NotEqual] 0 \[ScriptCapitalZ]n)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:243173858", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6590812421516027`*^9, 3.659081248095213*^9}},
 CellTags->{
  "ID:243173858", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "gcd \[NotEqual] 0 \[ScriptCapitalZ]n"},
 CellID->243173858],

Cell[BoxData[
 RowBox[{
  RowBox[{"quo", "[", 
   RowBox[{"n", ",", 
    RowBox[{"gcd", "[", 
     RowBox[{"n", ",", "x"}], "]"}]}], "]"}], "\[Element]", 
  "\[DoubleStruckCapitalN]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(quo gcd \[Element] \[DoubleStruckCapitalN] \[ScriptCapitalZ]n)", 
         "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:335259273", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659081598019428*^9, 3.6590816139782557`*^9}, {
  3.659081663461543*^9, 3.659081664101144*^9}},
 CellTags->{
  "ID:335259273", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
   "quo gcd \[Element] \[DoubleStruckCapitalN] \[ScriptCapitalZ]n"},
 CellID->335259273],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"gcd", "[", 
    RowBox[{"n", ",", "x"}], "]"}], "\[NotEqual]", "1"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
   RowBox[{"[", 
    RowBox[{"quo", "[", 
     RowBox[{"n", ",", 
      RowBox[{"gcd", "[", 
       RowBox[{"n", ",", "x"}], "]"}]}], "]"}], 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership quo gcd \[ScriptCapitalZ]n)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:582853848", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659242859833762*^9, 3.6592428872586102`*^9}},
 CellTags->{
  "ID:582853848", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership quo gcd \[ScriptCapitalZ]n"},
 CellID->582853848],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
  RowBox[{"[", 
   RowBox[{
    RowBox[{"quo", "[", 
     RowBox[{"n", ",", 
      RowBox[{"gcd", "[", 
       RowBox[{"n", ",", "x"}], "]"}]}], "]"}], "-", "1"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership quo gcd -1 \[ScriptCapitalZ]n)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:29867892", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6593448290692883`*^9, 3.659344831705693*^9}},
 CellTags->{
  "ID:29867892", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership quo gcd -1 \[ScriptCapitalZ]n"},
 CellID->29867892],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"quo", "[", 
    RowBox[{"n", ",", 
     RowBox[{"gcd", "[", 
      RowBox[{"n", ",", "x"}], "]"}]}], "]"}], 
   UnderscriptBox["*", "\[ScriptCapitalZ]n"], "x"}], "\[Equal]", 
  UnderscriptBox["0", "\[ScriptCapitalZ]n"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(quo gcd * \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:372275792", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659240768468488*^9, 3.6592407827113132`*^9}, {
  3.6592408353936057`*^9, 3.6592408368444085`*^9}},
 CellTags->{
  "ID:372275792", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "quo gcd * \[ScriptCapitalZ]n"},
 CellID->372275792],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "m", "\[Element]", "\[DoubleStruckCapitalN]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\IntegerQuotientRings.nb", 825367134}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659345850154482*^9, 3.6593458547252893`*^9}},
 CellTags->{
  "ID:825367134", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->825367134],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
  RowBox[{"[", 
   RowBox[{"quo", "[", 
    RowBox[{"x", ",", "m"}], "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership quo \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:55333112", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659169833784523*^9, 3.6591698754833965`*^9}, {
  3.6593458593116975`*^9, 3.659345861058901*^9}},
 CellTags->{
  "ID:55333112", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership quo \[ScriptCapitalZ]n", "IR"},
 CellID->55333112],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
  RowBox[{"[", 
   RowBox[{"x", "-", 
    RowBox[{"mod", "[", 
     RowBox[{"x", ",", "m"}], "]"}]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership -mod \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:474647872", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6593456342647023`*^9, 3.6593456578675437`*^9}, 
   3.659345857814095*^9},
 CellTags->{
  "ID:474647872", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership -mod \[ScriptCapitalZ]n", "IR"},
 CellID->474647872],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->285655755]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->305608247],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(gcd \[Element] \[DoubleStruckCapitalN] \
\[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(gcd \[Element] \[DoubleStruckCapitalN] \
\[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"gcd", "[", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], ",", 
               StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Element]", 
           "\[DoubleStruckCapitalN]"}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"gcd", "[", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], ",", 
               StyleBox["x", "ExpressionVariable"]}], "]"}], "\[Element]", 
           "\[DoubleStruckCapitalN]"}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:779775892"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p779775892-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p779775892-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:779775892", "Proof|ID:779775892-1"},
 CellID->297664998],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p779775892-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->10018482]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership gcd \[ScriptCapitalZ]n)\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(membership gcd \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{"gcd", "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{"gcd", "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:7430669"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p7430669-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p7430669-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:7430669", "Proof|ID:7430669-1"},
 CellID->340529413],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p7430669-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->114730061]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(gcd \[NotEqual] 0 \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(gcd \[NotEqual] 0 \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"gcd", "[", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], ",", 
               StyleBox["x", "ExpressionVariable"]}], "]"}], "\[NotEqual]", 
           UnderscriptBox["0", 
            RowBox[{"ReductionIQR", "[", 
              StyleBox["n", "ExpressionVariable"], 
              "]"}]]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"gcd", "[", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], ",", 
               StyleBox["x", "ExpressionVariable"]}], "]"}], "\[NotEqual]", 
           UnderscriptBox["0", 
            RowBox[{"ReductionIQR", "[", 
              StyleBox["n", "ExpressionVariable"], 
              "]"}]]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:243173858"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p243173858-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p243173858-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:243173858", "Proof|ID:243173858-1"},
 CellID->623193018],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p243173858-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->20166670]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(quo gcd \[Element] \[DoubleStruckCapitalN] \
\[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(quo gcd \[Element] \[DoubleStruckCapitalN] \
\[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"quo", "[", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], ",", 
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["x", "ExpressionVariable"]}], "]"}]}], "]"}], 
           "\[Element]", 
           "\[DoubleStruckCapitalN]"}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"quo", "[", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], ",", 
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["x", "ExpressionVariable"]}], "]"}]}], "]"}], 
           "\[Element]", 
           "\[DoubleStruckCapitalN]"}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:335259273"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p335259273-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p335259273-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:335259273", "Proof|ID:335259273-1"},
 CellID->301510995],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p335259273-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->678743316]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership quo gcd \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership quo gcd \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"gcd", "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], "\[NotEqual]", 
             "1"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{"quo", "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 RowBox[{"gcd", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"gcd", "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 StyleBox["x", "ExpressionVariable"]}], "]"}], "\[NotEqual]", 
             "1"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{"quo", "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 RowBox[{"gcd", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:582853848"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p582853848-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p582853848-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:582853848", "Proof|ID:582853848-1"},
 CellID->62215823],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p582853848-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->70354999]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership quo gcd -1 \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership quo gcd -1 \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ReductionIQR", "[", 
              StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
           RowBox[{
             RowBox[{"quo", "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 RowBox[{"gcd", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}]}], "]"}], 
             "-", "1"}], "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ReductionIQR", "[", 
              StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
           RowBox[{
             RowBox[{"quo", "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 RowBox[{"gcd", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}]}], "]"}], 
             "-", "1"}], "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:29867892"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p29867892-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p29867892-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:29867892", "Proof|ID:29867892-1"},
 CellID->753487385],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p29867892-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->115804636]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(quo gcd * \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(quo gcd * \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"quo", "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 RowBox[{"gcd", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}]}], "]"}], 
             UnderscriptBox["*", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             StyleBox["x", "ExpressionVariable"]}], "\[Equal]", 
           UnderscriptBox["0", 
            RowBox[{"ReductionIQR", "[", 
              StyleBox["n", "ExpressionVariable"], 
              "]"}]]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"quo", "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 RowBox[{"gcd", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["x", "ExpressionVariable"]}], "]"}]}], "]"}], 
             UnderscriptBox["*", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             StyleBox["x", "ExpressionVariable"]}], "\[Equal]", 
           UnderscriptBox["0", 
            RowBox[{"ReductionIQR", "[", 
              StyleBox["n", "ExpressionVariable"], 
              "]"}]]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:372275792"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p372275792-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p372275792-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:372275792", "Proof|ID:372275792-1"},
 CellID->9434432],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p372275792-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->166615473]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership quo \[ScriptCapitalZ]n)\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(membership quo \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ReductionIQR", "[", 
              StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
           RowBox[{"quo", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ReductionIQR", "[", 
              StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
           RowBox[{"quo", "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:55333112"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p55333112-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p55333112-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:55333112", "Proof|ID:55333112-1"},
 CellID->64245290],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p55333112-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->611966264]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership -mod \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership -mod \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ReductionIQR", "[", 
              StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "-", 
             RowBox[{"mod", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["m", "ExpressionVariable"]}], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ReductionIQR", "[", 
              StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "-", 
             RowBox[{"mod", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["m", "ExpressionVariable"]}], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:474647872"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p474647872-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p474647872-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:474647872", "Proof|ID:474647872-1"},
 CellID->209531985],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p474647872-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->74010259]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Commutative Ring with Identity (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.6584654925199823`*^9, 3.658465497901992*^9}, {
   3.6584736886589966`*^9, 3.6584736930426044`*^9}, 3.658559607210219*^9, {
   3.6585604697199345`*^9, 3.658560473963142*^9}},
 CellID->750673971],

Cell["", "OpenEnvironment",
 CellID->1269913440],

Cell[CellGroupData[{

Cell["\<\
Lemma: \[ScriptCapitalZ]n is a Commutative Ring with Unit\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.594446377117788*^9, 3.594446395467861*^9}, {
  3.594446459066728*^9, 3.594446505044395*^9}, {3.6584655612401032`*^9, 
  3.6584655647189093`*^9}, {3.6584658844962716`*^9, 3.658465902264703*^9}},
 CellID->1312707458],

Cell[BoxData[
 RowBox[{"isFinDomain", "[", "\[ScriptCapitalZ]n", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isFinDomain \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:208073319", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6586422724867306`*^9, 3.6586422787579412`*^9}},
 CellTags->{
  "ID:208073319", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isFinDomain \[ScriptCapitalZ]n"},
 CellID->208073319],

Cell[BoxData[
 RowBox[{"isDomain", "[", "\[ScriptCapitalZ]n", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isDomain \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:207295573", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6585595741381607`*^9, 3.6585595795669703`*^9}},
 CellTags->{
  "ID:207295573", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isDomain \[ScriptCapitalZ]n"},
 CellID->207295573],

Cell[BoxData[
 RowBox[{"isNeutral", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["*", "\[ScriptCapitalZ]n"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", 
   UnderscriptBox["1", "\[ScriptCapitalZ]n"], ",", "\[ScriptCapitalZ]n"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isNeutral 1 \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:189562335", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6584655800225363`*^9, 3.6584655991803703`*^9}, {
  3.658465823857965*^9, 3.658465825714368*^9}},
 CellTags->{
  "ID:189562335", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isNeutral 1 \[ScriptCapitalZ]n"},
 CellID->189562335],

Cell[BoxData[
 RowBox[{"isCommRing1", "[", "\[ScriptCapitalZ]n", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isCommRing1 \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:371480170", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594446399411061*^9, 3.594446412624801*^9}, {
   3.59444650948937*^9, 3.594446541430735*^9}, {3.607338509173718*^9, 
   3.607338510244437*^9}, {3.6584655514900866`*^9, 3.6584655546100917`*^9}, 
   3.658465827102771*^9},
 CellTags->{
  "ID:371480170", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isCommRing1 \[ScriptCapitalZ]n"},
 CellID->371480170],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->608276809]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1211509373],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isFinDomain \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isFinDomain \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"isFinDomain", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"isFinDomain", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:208073319"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p208073319-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p208073319-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:208073319", "Proof|ID:208073319-1"},
 CellID->169378186],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p208073319-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->439233552]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isDomain \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isDomain \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"isDomain", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"isDomain", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:207295573"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p207295573-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p207295573-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:207295573", "Proof|ID:207295573-1"},
 CellID->415181399],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p207295573-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->27765665]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isNeutral 1 \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isNeutral 1 \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"isNeutral", "[", 
           RowBox[{
             UnderscriptBox["*", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], ",", 
             UnderscriptBox["1", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], ",", 
             RowBox[{"ReductionIQR", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"isNeutral", "[", 
           RowBox[{
             UnderscriptBox["*", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], ",", 
             UnderscriptBox["1", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], ",", 
             RowBox[{"ReductionIQR", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:189562335"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p189562335-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p189562335-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:189562335", "Proof|ID:189562335-1"},
 CellID->43006419],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p189562335-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->144573181]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isCommRing1 \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isCommRing1 \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"isCommRing1", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"isCommRing1", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:371480170"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p371480170-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p371480170-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:371480170", "Proof|ID:371480170-1"},
 CellID->344386318],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p371480170-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {61., 67.}}]], "ProofInfoBody",
 CellID->576038596]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Order Relation (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6584655044072037`*^9, 3.6584655070904083`*^9}, {
  3.6584747734787035`*^9, 3.658474777799911*^9}},
 CellID->76760028],

Cell["", "OpenEnvironment",
 CellID->338374869],

Cell[CellGroupData[{

Cell["\<\
Lemma: Properties of the Order Relation on \[ScriptCapitalZ]n\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6584658903930817`*^9, 3.658465921764737*^9}},
 CellID->134166625],

Cell[BoxData[
 RowBox[{"isPartIrreflOrder", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]n"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalZ]n"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isPartIrreflOrder \[ScriptCapitalZ]n)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1342794225", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.59444662050856*^9, 3.594446683075286*^9}, 
   3.607338517954448*^9, {3.610185214816342*^9, 3.610185217262858*^9}, {
   3.6584659486435843`*^9, 3.6584659546027946`*^9}},
 CellTags->{
  "ID:1342794225", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isPartIrreflOrder \[ScriptCapitalZ]n"},
 CellID->1342794225],

Cell[BoxData[
 RowBox[{"isTotalIrreflOrder", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]n"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalZ]n"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isTotalIrreflOrder \[ScriptCapitalZ]n)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:145791983", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.658465938877967*^9, 3.658465960468405*^9}},
 CellTags->{
  "ID:145791983", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isTotalIrreflOrder \[ScriptCapitalZ]n"},
 CellID->145791983],

Cell[BoxData[
 RowBox[{"isNoetherian", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]n"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalZ]n"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isNoetherian \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1051492688", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.610185205833395*^9, 3.610185221938676*^9}, 
   3.658465964742812*^9},
 CellTags->{
  "ID:1051492688", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isNoetherian \[ScriptCapitalZ]n"},
 CellID->1051492688],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->31879075]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->83297866],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isPartIrreflOrder \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isPartIrreflOrder \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"isPartIrreflOrder", "[", 
           RowBox[{
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], ",", 
             RowBox[{"ReductionIQR", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"isPartIrreflOrder", "[", 
           RowBox[{
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], ",", 
             RowBox[{"ReductionIQR", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1342794225"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1342794225-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1342794225-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1342794225", "Proof|ID:1342794225-1"},
 CellID->256415475],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1342794225-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->3143692]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isTotalIrreflOrder \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isTotalIrreflOrder \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"isTotalIrreflOrder", "[", 
           RowBox[{
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], ",", 
             RowBox[{"ReductionIQR", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"isTotalIrreflOrder", "[", 
           RowBox[{
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], ",", 
             RowBox[{"ReductionIQR", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:145791983"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p145791983-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p145791983-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:145791983", "Proof|ID:145791983-1"},
 CellID->187860367],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p145791983-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->503595798]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isNoetherian \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isNoetherian \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"isNoetherian", "[", 
           RowBox[{
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], ",", 
             RowBox[{"ReductionIQR", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"isNoetherian", "[", 
           RowBox[{
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], ",", 
             RowBox[{"ReductionIQR", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1051492688"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1051492688-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1051492688-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1051492688", "Proof|ID:1051492688-1"},
 CellID->66415286],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1051492688-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->54951663]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Multipliers & R0-R4 (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.658465512035617*^9, 3.6584655171212254`*^9}, {
   3.6590885991833954`*^9, 3.6590886035358033`*^9}, 3.6590923477791805`*^9, {
   3.659167546751687*^9, 3.659167550854494*^9}},
 CellID->193070521],

Cell["", "OpenEnvironment",
 CellID->617760674],

Cell[CellGroupData[{

Cell["Lemma: Preparations and Properties of Multipliers", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6584807100714955`*^9, 3.658480713191501*^9}, {
  3.659184165470331*^9, 3.6591841728803444`*^9}},
 CellID->411365783],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"m", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", "f"], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "m"}]], 
       RowBox[{
        RowBox[{"f", "[", "i", "]"}], "\[Element]", 
        "\[DoubleStruckCapitalZ]"}]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        UnderscriptBox[
         SubscriptBox["\[Sum]", "\[ScriptCapitalZ]n"], 
         RowBox[{
          RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "m"}]], 
        RowBox[{"f", "[", "i", "]"}]}], "\[Equal]", 
       RowBox[{"mod", "[", 
        RowBox[{
         RowBox[{
          UnderscriptBox[
           SubscriptBox["\[Sum]", "\[DoubleStruckCapitalZ]Dom"], 
           RowBox[{
            RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "m"}]], 
          RowBox[{"f", "[", "i", "]"}]}], ",", "n"}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Sum] \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:307883626", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6584807184019103`*^9, 3.6584807748594093`*^9}, {
  3.6584810526077003`*^9, 3.6584810593469124`*^9}, {3.658481106225995*^9, 
  3.6584811242304263`*^9}, {3.6615839245402203`*^9, 3.6615839258350224`*^9}},
 CellTags->{
  "ID:307883626", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Sum] \[ScriptCapitalZ]n"},
 CellID->307883626],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
   RowBox[{"[", 
    RowBox[{"c", ",", "m"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\IntegerQuotientRings.nb", 209868567}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6592525141613145`*^9, 3.6592525228973303`*^9}},
 CellTags->{
  "ID:209868567", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->209868567],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
   RowBox[{"m", ",", "c", ",", "1"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"m", ">", "0"}], "\[And]", 
    RowBox[{"n", ">", 
     RowBox[{"m", "*", 
      RowBox[{"gcd", "[", 
       RowBox[{"n", ",", "c"}], "]"}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(mult1 alternative \[ScriptCapitalZ]n)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:57125150", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6590922570580215`*^9, 3.6590922826324663`*^9}, {
  3.659240663043503*^9, 3.6592406758823256`*^9}, {3.6592460786556263`*^9, 
  3.6592460802312293`*^9}},
 CellTags->{
  "ID:57125150", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 alternative \[ScriptCapitalZ]n"},
 CellID->57125150],

Cell[BoxData[
 RowBox[{
  RowBox[{"c", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
     RowBox[{"m", ",", "c", ",", "1"}], "]"}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"m", 
       UnderscriptBox["*", "\[ScriptCapitalZ]n"], "c"}], "\[NotEqual]", 
      UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R2' \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:50258269", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6590866616735926`*^9, 3.659086719315694*^9}, {
  3.659087547256948*^9, 3.6590875481929493`*^9}, {3.659092248088005*^9, 
  3.659092251223611*^9}, {3.6592406839007397`*^9, 3.6592407020123715`*^9}, {
  3.659252525580535*^9, 3.6592525331465483`*^9}},
 CellTags->{
  "ID:50258269", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R2' \[ScriptCapitalZ]n"},
 CellID->50258269],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->343575043]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->834291338],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Sum] \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Sum] \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             StyleBox["f", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   StyleBox["m", "ExpressionVariable"]}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "[", 
                 StyleBox["i", "ExpressionVariable"], "]"}], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox[
                SubscriptBox["\[Sum]", 
                 RowBox[{"ReductionIQR", "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]], 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["m", "ExpressionVariable"]}]}}]], 
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "[", 
                 StyleBox["i", "ExpressionVariable"], "]"}]}], "\[Equal]", 
             RowBox[{"mod", "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox[
                    SubscriptBox["\[Sum]", "\[DoubleStruckCapitalZ]"], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["m", "ExpressionVariable"]}]}}]], 
                   RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}]}], ",", 
                 StyleBox["n", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             StyleBox["f", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   StyleBox["m", "ExpressionVariable"]}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "[", 
                 StyleBox["i", "ExpressionVariable"], "]"}], "\[Element]", 
               "\[DoubleStruckCapitalZ]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox[
                SubscriptBox["\[Sum]", 
                 RowBox[{"ReductionIQR", "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]], 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["m", "ExpressionVariable"]}]}}]], 
               RowBox[{
                 StyleBox["f", "ExpressionVariable"], "[", 
                 StyleBox["i", "ExpressionVariable"], "]"}]}], "\[Equal]", 
             RowBox[{"mod", "[", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox[
                    SubscriptBox["\[Sum]", "\[DoubleStruckCapitalZ]"], 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["m", "ExpressionVariable"]}]}}]], 
                   RowBox[{
                    StyleBox["f", "ExpressionVariable"], "[", 
                    StyleBox["i", "ExpressionVariable"], "]"}]}], ",", 
                 StyleBox["n", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:307883626"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p307883626-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p307883626-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:307883626", "Proof|ID:307883626-1"},
 CellID->253583856],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p307883626-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->217854969]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(mult1 alternative \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(mult1 alternative \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ">", "0"}], "\[And]", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ">", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], " ", 
                   RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ">", "0"}], "\[And]", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ">", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], " ", 
                   RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:57125150"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p57125150-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p57125150-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:57125150", "Proof|ID:57125150-1"},
 CellID->163486749],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p57125150-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->65009874]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R2' \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R2' \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], 
               StyleBox["c", "ExpressionVariable"]}], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], 
                "]"}]]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], 
               StyleBox["c", "ExpressionVariable"]}], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], 
                "]"}]]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:50258269"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p50258269-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p50258269-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:50258269", "Proof|ID:50258269-1"},
 CellID->223632667],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p50258269-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->553065911]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->168314290],

Cell[CellGroupData[{

Cell["\<\
Lemma: \[ScriptCapitalZ]n Satisfies Multipliers and R0-R4\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6584660550407715`*^9, 3.658466068271595*^9}},
 CellID->566297413],

Cell[BoxData[
 RowBox[{"Multipliers", "[", "\[ScriptCapitalZ]n", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Multipliers \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1161287108", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594446695522677*^9, 3.594446698725513*^9}, {
  3.6584660825008197`*^9, 3.658466084029622*^9}},
 CellTags->{
  "ID:1161287108", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "Multipliers \[ScriptCapitalZ]n"},
 CellID->1161287108],

Cell[BoxData[
 RowBox[{"R0", "[", "\[ScriptCapitalZ]n", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R0 \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:124498584", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594446724741005*^9, 3.594446727041064*^9}, 
   3.65846608856923*^9},
 CellTags->{
  "ID:124498584", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R0 \[ScriptCapitalZ]n"},
 CellID->124498584],

Cell[BoxData[
 RowBox[{"R1", "[", "\[ScriptCapitalZ]n", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R1 \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1236424988", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594446732346009*^9, 3.594446745337828*^9}, 
   3.6584660896456327`*^9},
 CellTags->{
  "ID:1236424988", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R1 \[ScriptCapitalZ]n"},
 CellID->1236424988],

Cell[BoxData[
 RowBox[{"R2", "[", "\[ScriptCapitalZ]n", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R2 \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1558881148", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594446733343932*^9, 3.594446746186222*^9}, 
   3.6584660912846355`*^9},
 CellTags->{
  "ID:1558881148", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R2 \[ScriptCapitalZ]n"},
 CellID->1558881148],

Cell[BoxData[
 RowBox[{"R3", "[", "\[ScriptCapitalZ]n", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R3 \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1530993938", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594446734010182*^9, 3.59444674710826*^9}, 
   3.658466092220637*^9},
 CellTags->{
  "ID:1530993938", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R3 \[ScriptCapitalZ]n"},
 CellID->1530993938],

Cell[BoxData[
 RowBox[{"R4", "[", "\[ScriptCapitalZ]n", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R4 \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:589877437", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594446734603145*^9, 3.594446747919977*^9}, 
   3.6584660931722383`*^9},
 CellTags->{
  "ID:589877437", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R4 \[ScriptCapitalZ]n"},
 CellID->589877437],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->140752188]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->78307490],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Multipliers \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Multipliers \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"Multipliers", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"Multipliers", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1161287108"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1161287108-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1161287108-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1161287108", "Proof|ID:1161287108-1"},
 CellID->311987723],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1161287108-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->552671078]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R0 \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R0 \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R0", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R0", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:124498584"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p124498584-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p124498584-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:124498584", "Proof|ID:124498584-1"},
 CellID->24095429],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p124498584-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->237502882]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R1 \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R1 \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R1", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R1", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1236424988"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1236424988-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1236424988-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1236424988", "Proof|ID:1236424988-1"},
 CellID->296864625],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1236424988-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->44769427]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R2 \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R2 \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R2", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R2", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1558881148"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1558881148-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1558881148-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1558881148", "Proof|ID:1558881148-1"},
 CellID->239424613],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1558881148-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->286970037]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R3 \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R3 \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R3", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R3", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1530993938"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1530993938-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1530993938-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1530993938", "Proof|ID:1530993938-1"},
 CellID->344362146],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1530993938-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->62375380]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R4 \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R4 \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R4", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R4", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:589877437"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p589877437-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p589877437-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:589877437", "Proof|ID:589877437-1"},
 CellID->352205366],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p589877437-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->173788587]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Reducibility in \[ScriptCapitalZ]n (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.658485337306636*^9, 3.6584853434998465`*^9}, {
  3.659332104745194*^9, 3.6593321090040016`*^9}},
 CellID->191294579],

Cell["", "OpenEnvironment",
 CellID->40861546],

Cell[CellGroupData[{

Cell["Lemma: Properties of Multipliers", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.659240605978603*^9, 3.659240613887817*^9}},
 CellID->219633271],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
   RowBox[{"[", 
    RowBox[{"c", ",", "m"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\IntegerQuotientRings.nb", 104617955}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6592406498614798`*^9, 3.659240658659895*^9}},
 CellTags->{
  "ID:104617955", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->104617955],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
  RowBox[{"[", 
   RowBox[{"m", 
    UnderscriptBox["-", "\[ScriptCapitalZ]n"], 
    RowBox[{"quo", "[", 
     RowBox[{"n", ",", 
      RowBox[{"gcd", "[", 
       RowBox[{"n", ",", "c"}], "]"}]}], "]"}]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership -quo \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:608534217", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659184099107815*^9, 3.6591841092634325`*^9}},
 CellTags->{
  "ID:608534217", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership -quo \[ScriptCapitalZ]n", "IR"},
 CellID->608534217],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
  RowBox[{"[", 
   RowBox[{"m", 
    UnderscriptBox["+", "\[ScriptCapitalZ]n"], 
    RowBox[{"quo", "[", 
     RowBox[{"n", ",", 
      RowBox[{"gcd", "[", 
       RowBox[{"n", ",", "c"}], "]"}]}], "]"}]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership +quo \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:217754000", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659252463648426*^9, 3.6592524692800355`*^9}},
 CellTags->{
  "ID:217754000", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership +quo \[ScriptCapitalZ]n", "IR"},
 CellID->217754000],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
   RowBox[{"m", ",", "c", ",", "1"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
   RowBox[{
    RowBox[{"m", 
     UnderscriptBox["-", "\[ScriptCapitalZ]n"], 
     RowBox[{"quo", "[", 
      RowBox[{"n", ",", 
       RowBox[{"gcd", "[", 
        RowBox[{"n", ",", "c"}], "]"}]}], "]"}]}], ",", "c", ",", "2"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(mult2 -quo \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:415524168", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6591818485316615`*^9, 3.65918188743913*^9}, {
  3.6591832887905912`*^9, 3.6591832948122015`*^9}, {3.659242602147509*^9, 
  3.659242606655917*^9}, {3.6592453465589404`*^9, 3.659245350521347*^9}},
 CellTags->{
  "ID:415524168", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult2 -quo \[ScriptCapitalZ]n"},
 CellID->415524168],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
   RowBox[{"m", ",", "c", ",", "2"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
   RowBox[{
    RowBox[{"m", 
     UnderscriptBox["+", "\[ScriptCapitalZ]n"], 
     RowBox[{"quo", "[", 
      RowBox[{"n", ",", 
       RowBox[{"gcd", "[", 
        RowBox[{"n", ",", "c"}], "]"}]}], "]"}]}], ",", "c", ",", "1"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(mult1 +quo \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:51392774", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6591833201778464`*^9, 3.6591833238126526`*^9}, {
  3.6592487098984323`*^9, 3.659248716590844*^9}},
 CellTags->{
  "ID:51392774", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 +quo \[ScriptCapitalZ]n"},
 CellID->51392774],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"m", 
     UnderscriptBox["-", "\[ScriptCapitalZ]n"], 
     RowBox[{"quo", "[", 
      RowBox[{"n", ",", 
       RowBox[{"gcd", "[", 
        RowBox[{"n", ",", "c"}], "]"}]}], "]"}]}], ")"}], 
   UnderscriptBox["*", "\[ScriptCapitalZ]n"], "c"}], "\[Equal]", 
  RowBox[{"m", 
   UnderscriptBox["*", "\[ScriptCapitalZ]n"], 
   "c"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* -quo \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:488904469", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659183823937931*^9, 3.6591838445143676`*^9}, {
  3.6591841885739717`*^9, 3.659184195328784*^9}},
 CellTags->{
  "ID:488904469", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "* -quo \[ScriptCapitalZ]n"},
 CellID->488904469],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"m", 
     UnderscriptBox["+", "\[ScriptCapitalZ]n"], 
     RowBox[{"quo", "[", 
      RowBox[{"n", ",", 
       RowBox[{"gcd", "[", 
        RowBox[{"n", ",", "c"}], "]"}]}], "]"}]}], ")"}], 
   UnderscriptBox["*", "\[ScriptCapitalZ]n"], "c"}], "\[Equal]", 
  RowBox[{"m", 
   UnderscriptBox["*", "\[ScriptCapitalZ]n"], 
   "c"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* +quo \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:469346824", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6592489266764126`*^9, 3.659248932417223*^9}},
 CellTags->{
  "ID:469346824", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "* +quo \[ScriptCapitalZ]n"},
 CellID->469346824],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->246186662]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->384842955],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership -quo \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership -quo \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ReductionIQR", "[", 
              StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["-", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             RowBox[{"quo", "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 RowBox[{"gcd", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ReductionIQR", "[", 
              StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["-", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             RowBox[{"quo", "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 RowBox[{"gcd", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:608534217"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p608534217-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p608534217-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:608534217", "Proof|ID:608534217-1"},
 CellID->171963183],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p608534217-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{3.6592525893846464`*^9},
 CellID->552509088]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership +quo \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(membership +quo \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ReductionIQR", "[", 
              StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["+", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             RowBox[{"quo", "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 RowBox[{"gcd", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ReductionIQR", "[", 
              StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["+", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             RowBox[{"quo", "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 RowBox[{"gcd", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:217754000"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p217754000-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p217754000-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:217754000", "Proof|ID:217754000-1"},
 CellID->218735392],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p217754000-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->329603756]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(mult2 -quo \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(mult2 -quo \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                 RowBox[{"quo", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], "]"}]}], 
               ",", 
               StyleBox["c", "ExpressionVariable"], ",", "2"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                 RowBox[{"quo", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], "]"}]}], 
               ",", 
               StyleBox["c", "ExpressionVariable"], ",", "2"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:415524168"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p415524168-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p415524168-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:415524168", "Proof|ID:415524168-1"},
 CellID->322621179],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p415524168-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellChangeTimes->{3.659249609150412*^9, 3.659252448204399*^9},
 CellID->193335281]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(mult1 +quo \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(mult1 +quo \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", "2"}], "]"}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                 RowBox[{"quo", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], "]"}]}], 
               ",", 
               StyleBox["c", "ExpressionVariable"], ",", "1"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", "2"}], "]"}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                 RowBox[{"quo", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], "]"}]}], 
               ",", 
               StyleBox["c", "ExpressionVariable"], ",", "1"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:51392774"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p51392774-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p51392774-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:51392774", "Proof|ID:51392774-1"},
 CellID->336233968],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p51392774-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {61., 67.}}]], "ProofInfoBody",
 CellID->2402072]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* -quo \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* -quo \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["-", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], 
               RowBox[{"quo", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], "]"}]}], 
             UnderscriptBox["*", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             StyleBox["c", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["-", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], 
               RowBox[{"quo", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], "]"}]}], 
             UnderscriptBox["*", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             StyleBox["c", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:488904469"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p488904469-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p488904469-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:488904469", "Proof|ID:488904469-1"},
 CellID->318992087],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p488904469-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->558690740]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* +quo \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* +quo \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], 
               RowBox[{"quo", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], "]"}]}], 
             UnderscriptBox["*", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             StyleBox["c", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], 
               RowBox[{"quo", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], "]"}]}], 
             UnderscriptBox["*", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             StyleBox["c", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:469346824"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p469346824-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p469346824-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:469346824", "Proof|ID:469346824-1"},
 CellID->231394775],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p469346824-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->467357679]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->337203903],

Cell[CellGroupData[{

Cell["Lemma: \[Succeeds] and +", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.658485352095462*^9, 3.658485366278887*^9}},
 CellID->657296789],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\IntegerQuotientRings.nb", 195073923}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.658485545997203*^9, 3.6584855548746185`*^9}},
 CellTags->{
  "ID:195073923", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->195073923],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]n"], 
    RowBox[{"a", 
     UnderscriptBox["-", "\[ScriptCapitalZ]n"], "b"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"b", "\[NotEqual]", "0"}], "\[And]", 
    RowBox[{"a", "\[GreaterEqual]", "b"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Succeeds] - \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:18535645", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6584855664674387`*^9, 3.658485581786666*^9}, {
  3.658487329132537*^9, 3.6584873302557383`*^9}},
 CellTags->{
  "ID:18535645", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Succeeds] - \[ScriptCapitalZ]n"},
 CellID->18535645],

Cell[BoxData[
 RowBox[{
  RowBox[{"b", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
      RowBox[{"[", "m", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
       RowBox[{"m", ",", "b"}], "]"}], "\[Implies]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"a", 
          UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]n"], 
          RowBox[{"a", 
           UnderscriptBox["-", "\[ScriptCapitalZ]n"], 
           RowBox[{"m", 
            UnderscriptBox["*", "\[ScriptCapitalZ]n"], "b"}]}]}], 
         TagBox[")",
          "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"a", "\[GreaterEqual]", 
          RowBox[{"m", 
           UnderscriptBox["*", "\[ScriptCapitalZ]n"], "b"}]}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Succeeds] - * \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:155348029", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65917848064633*^9, 3.6591785451244435`*^9}},
 CellTags->{
  "ID:155348029", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Succeeds] - * \[ScriptCapitalZ]n"},
 CellID->155348029],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]n"], 
    RowBox[{"a", 
     UnderscriptBox["+", "\[ScriptCapitalZ]n"], "b"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", "+", "b"}], "\[GreaterEqual]", "n"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Succeeds] + \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:220895581", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6584853739083004`*^9, 3.6584854158763742`*^9}, {
  3.6584855577616234`*^9, 3.6584855589316254`*^9}, {3.6584856557629957`*^9, 
  3.658485660380604*^9}, {3.658485722017312*^9, 3.6584857808636155`*^9}},
 CellTags->{
  "ID:220895581", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Succeeds] + \[ScriptCapitalZ]n"},
 CellID->220895581],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->38173617]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->45785939],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Succeeds] - \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Succeeds] - \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], 
               StyleBox["b", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[NotEqual]", "0"}], 
               "\[And]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[GreaterEqual]", 
                 StyleBox["b", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["-", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], 
               StyleBox["b", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], "\[NotEqual]", "0"}], 
               "\[And]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[GreaterEqual]", 
                 StyleBox["b", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:18535645"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p18535645-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p18535645-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:18535645", "Proof|ID:18535645-1"},
 CellID->255392256],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p18535645-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->600071146]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Succeeds] - * \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Succeeds] - * \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["mult", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["b", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                    StyleBox["b", "ExpressionVariable"]}]}]}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[GreaterEqual]", 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                    StyleBox[
                    "b", "ExpressionVariable"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["b", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["mult", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["b", "ExpressionVariable"]}], "]"}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], 
                    UnderscriptBox["-", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                    StyleBox["b", "ExpressionVariable"]}]}]}], 
                 "\[DoubleLeftRightArrow]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[GreaterEqual]", 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                    StyleBox[
                    "b", "ExpressionVariable"]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:155348029"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p155348029-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p155348029-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:155348029", "Proof|ID:155348029-1"},
 CellID->101870138],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p155348029-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->250604527]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Succeeds] + \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Succeeds] + \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], 
               StyleBox["b", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "+", 
               StyleBox["b", "ExpressionVariable"]}], "\[GreaterEqual]", 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[Succeeds]", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], 
               StyleBox["b", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "+", 
               StyleBox["b", "ExpressionVariable"]}], "\[GreaterEqual]", 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:220895581"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p220895581-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p220895581-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:220895581", "Proof|ID:220895581-1"},
 CellID->11546098],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p220895581-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->21689854]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->32514940],

Cell[CellGroupData[{

Cell["Lemma: Multiples of gcd", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.659090918038869*^9, 3.6590909242320795`*^9}, {
  3.659091616690096*^9, 3.6590916239597087`*^9}},
 CellID->674314239],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
   RowBox[{"[", "c", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\IntegerQuotientRings.nb", 108166851}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659090954667733*^9, 3.6590910010154147`*^9}},
 CellTags->{
  "ID:108166851", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->108166851],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
    RowBox[{"[", "m", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"gcd", "[", 
     RowBox[{"n", ",", "c"}], "]"}], "\[VerticalBar]", 
    RowBox[{"m", 
     UnderscriptBox["*", "\[ScriptCapitalZ]n"], "c"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(gcd \[VerticalBar] * \[ScriptCapitalZ]n)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:166132041", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6590909763829713`*^9, 3.659091017259043*^9}},
 CellTags->{
  "ID:166132041", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "gcd \[VerticalBar] * \[ScriptCapitalZ]n"},
 CellID->166132041],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"c", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"m", 
       UnderscriptBox["*", "\[ScriptCapitalZ]n"], "c"}], 
      UnderscriptBox["|", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
        RowBox[{"[", "m", "]"}]}]], 
      RowBox[{
       UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
       RowBox[{"m", ",", "c", ",", "1"}], "]"}]}], "}"}], "\[SubsetEqual]", 
    RowBox[{"{", 
     RowBox[{"a", 
      UnderscriptBox["|", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
        RowBox[{"[", "a", "]"}]}]], 
      RowBox[{
       RowBox[{"a", "\[NotEqual]", 
        UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], "\[And]", 
       RowBox[{
        RowBox[{"gcd", "[", 
         RowBox[{"n", ",", "c"}], "]"}], "\[VerticalBar]", "a"}]}]}], "}"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(mult1 sufficient 1 \[ScriptCapitalZ]n)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:414892396", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659167623535022*^9, 3.6591677086185713`*^9}, {
  3.6591715383357334`*^9, 3.6591715481169505`*^9}},
 CellTags->{
  "ID:414892396", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 1 \[ScriptCapitalZ]n"},
 CellID->414892396],

Cell[BoxData[
 RowBox[{"isFinite", "[", 
  RowBox[{"{", 
   RowBox[{"a", 
    UnderscriptBox["|", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
      RowBox[{"[", "a", "]"}]}]], 
    RowBox[{
     RowBox[{"a", "\[NotEqual]", 
      UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], "\[And]", 
     RowBox[{
      RowBox[{"gcd", "[", 
       RowBox[{"n", ",", "c"}], "]"}], "\[VerticalBar]", "a"}]}]}], "}"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(mult1 sufficient 2 \[ScriptCapitalZ]n)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:48379652", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659167765792672*^9, 3.6591677726878843`*^9}},
 CellTags->{
  "ID:48379652", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 2 \[ScriptCapitalZ]n"},
 CellID->48379652],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"c", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
      RowBox[{"[", "a", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"a", "\[NotEqual]", 
         UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], "\[And]", 
        RowBox[{
         RowBox[{"gcd", "[", 
          RowBox[{"n", ",", "c"}], "]"}], "\[VerticalBar]", "a"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[Implies]", 
      RowBox[{
       UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
       RowBox[{
        RowBox[{"quo", "[", 
         RowBox[{"a", ",", 
          RowBox[{"gcd", "[", 
           RowBox[{"n", ",", "c"}], "]"}]}], "]"}], ",", "c", ",", "1"}], 
       "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(mult1 sufficient 3 \[ScriptCapitalZ]n)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:287224029", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6591678010175333`*^9, 3.6591678498768196`*^9}, {
  3.6591694016773643`*^9, 3.659169447979245*^9}, {3.6591695473670197`*^9, 
  3.6591695511578264`*^9}},
 CellTags->{
  "ID:287224029", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 3 \[ScriptCapitalZ]n"},
 CellID->287224029],

Cell[BoxData[
 RowBox[{
  TagBox["(",
   "AutoParentheses"], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
     RowBox[{"[", 
      RowBox[{"a1", ",", "a2"}], "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        RowBox[{"gcd", "[", 
         RowBox[{"n", ",", "c"}], "]"}], "\[VerticalBar]", "a1"}], "\[And]", 
       RowBox[{
        RowBox[{"gcd", "[", 
         RowBox[{"n", ",", "c"}], "]"}], "\[VerticalBar]", "a2"}], "\[And]", 
       RowBox[{
        RowBox[{"quo", "[", 
         RowBox[{"a1", ",", 
          RowBox[{"gcd", "[", 
           RowBox[{"n", ",", "c"}], "]"}]}], "]"}], "\[Equal]", 
        RowBox[{"quo", "[", 
         RowBox[{"a2", ",", 
          RowBox[{"gcd", "[", 
           RowBox[{"n", ",", "c"}], "]"}]}], "]"}]}]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{"a1", "\[Equal]", "a2"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}], 
  TagBox[")",
   "AutoParentheses"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(mult1 sufficient 4 \[ScriptCapitalZ]n)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:398117055", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659173211732873*^9, 3.659173247456936*^9}},
 CellTags->{
  "ID:398117055", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 4 \[ScriptCapitalZ]n"},
 CellID->398117055],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"c", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
      RowBox[{"[", 
       RowBox[{"m1", ",", "m2"}], "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
         RowBox[{"m1", ",", "c", ",", "1"}], "]"}], "\[And]", 
        RowBox[{
         UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
         RowBox[{"m2", ",", "c", ",", "1"}], "]"}], "\[And]", 
        RowBox[{
         RowBox[{"m1", 
          UnderscriptBox["*", "\[ScriptCapitalZ]n"], "c"}], "\[Equal]", 
         RowBox[{"m2", 
          UnderscriptBox["*", "\[ScriptCapitalZ]n"], "c"}]}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[Implies]", 
      RowBox[{"m1", "\[Equal]", "m2"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(mult1 * injective \[ScriptCapitalZ]n)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:568873110", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659174273831539*^9, 3.6591743345000453`*^9}},
 CellTags->{
  "ID:568873110", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 * injective \[ScriptCapitalZ]n"},
 CellID->568873110],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"c", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
     RowBox[{"[", "a", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{"a", "\[NotEqual]", 
        UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], "\[And]", 
       RowBox[{
        RowBox[{"gcd", "[", 
         RowBox[{"n", ",", "c"}], "]"}], "\[VerticalBar]", "a"}]}], 
      TagBox[")",
       "AutoParentheses"]}], "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
         RowBox[{"[", "m", "]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{
          UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
          RowBox[{"m", ",", "c", ",", "1"}], "]"}], "\[And]", 
         RowBox[{"a", "\[Equal]", 
          RowBox[{"m", 
           UnderscriptBox["*", "\[ScriptCapitalZ]n"], "c"}]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(mult1 sufficient \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:347198673", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6590910214866505`*^9, 3.6590911557716866`*^9}, {
  3.65909205157446*^9, 3.659092056051668*^9}},
 CellTags->{
  "ID:347198673", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient \[ScriptCapitalZ]n"},
 CellID->347198673],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->209695896]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->26902409],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(gcd \[VerticalBar] * \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(gcd \[VerticalBar] * \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"gcd", "[", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], "\[VerticalBar]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"gcd", "[", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], "\[VerticalBar]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:166132041"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p166132041-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p166132041-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:166132041", "Proof|ID:166132041-1"},
 CellID->111518977],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p166132041-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->518018963]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(mult1 sufficient 1 \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(mult1 sufficient 1 \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[Implies]", 
           RowBox[{
             RowBox[{"{", 
               RowBox[{
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                   StyleBox["c", "ExpressionVariable"]}], 
                 UnderscriptBox["|", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   UnderscriptBox["mult", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}]}], 
               "}"}], "\[SubsetEqual]", 
             RowBox[{"{", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["|", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
                   RowBox[{
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                    "\[VerticalBar]", 
                    StyleBox["a", "ExpressionVariable"]}]}]}], 
               "}"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[Implies]", 
           RowBox[{
             RowBox[{"{", 
               RowBox[{
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                   StyleBox["c", "ExpressionVariable"]}], 
                 UnderscriptBox["|", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   UnderscriptBox["mult", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}]}], 
               "}"}], "\[SubsetEqual]", 
             RowBox[{"{", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["|", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
                   RowBox[{
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                    "\[VerticalBar]", 
                    StyleBox["a", "ExpressionVariable"]}]}]}], 
               "}"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:414892396"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p414892396-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p414892396-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:414892396", "Proof|ID:414892396-1"},
 CellID->924193601],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p414892396-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->694816149]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(mult1 sufficient 2 \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(mult1 sufficient 2 \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isFinite", "[", 
           RowBox[{"{", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["|", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
                 RowBox[{
                   RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                   "\[VerticalBar]", 
                   StyleBox["a", "ExpressionVariable"]}]}]}], "}"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isFinite", "[", 
           RowBox[{"{", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["|", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
                 RowBox[{
                   RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                   "\[VerticalBar]", 
                   StyleBox["a", "ExpressionVariable"]}]}]}], "}"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:48379652"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p48379652-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p48379652-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:48379652", "Proof|ID:48379652-1"},
 CellID->384976109],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p48379652-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->382554458]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(mult1 sufficient 3 \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(mult1 sufficient 3 \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
                 RowBox[{
                   RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                   "\[VerticalBar]", 
                   StyleBox["a", "ExpressionVariable"]}]}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["mult", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 RowBox[{
                   RowBox[{"quo", "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], "]"}], 
                   ",", 
                   StyleBox["c", "ExpressionVariable"], ",", "1"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
                 RowBox[{
                   RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                   "\[VerticalBar]", 
                   StyleBox["a", "ExpressionVariable"]}]}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["mult", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 RowBox[{
                   RowBox[{"quo", "[", 
                    RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], "]"}], 
                   ",", 
                   StyleBox["c", "ExpressionVariable"], ",", "1"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:287224029"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p287224029-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p287224029-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:287224029", "Proof|ID:287224029-1"},
 CellID->144370346],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p287224029-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->265857267]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(mult1 sufficient 4 \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(mult1 sufficient 4 \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], 
               "\[VerticalBar]", 
               StyleBox["a1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], 
               "\[VerticalBar]", 
               StyleBox["a2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{"quo", "[", 
                 RowBox[{
                   StyleBox["a1", "ExpressionVariable"], ",", 
                   RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], "]"}], 
               "\[Equal]", 
               RowBox[{"quo", "[", 
                 RowBox[{
                   StyleBox["a2", "ExpressionVariable"], ",", 
                   RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], "]"}]}]}],
            "\[Implies]", 
           RowBox[{
             StyleBox["a1", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "a2", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], 
               "\[VerticalBar]", 
               StyleBox["a1", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], 
               "\[VerticalBar]", 
               StyleBox["a2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{"quo", "[", 
                 RowBox[{
                   StyleBox["a1", "ExpressionVariable"], ",", 
                   RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], "]"}], 
               "\[Equal]", 
               RowBox[{"quo", "[", 
                 RowBox[{
                   StyleBox["a2", "ExpressionVariable"], ",", 
                   RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], "]"}]}]}],
            "\[Implies]", 
           RowBox[{
             StyleBox["a1", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "a2", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:398117055"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p398117055-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p398117055-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:398117055", "Proof|ID:398117055-1"},
 CellID->168110018],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p398117055-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->367098984]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(mult1 * injective \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(mult1 * injective \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["mult", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}], 
                 "\[And]", 
                 RowBox[{
                   UnderscriptBox["mult", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   RowBox[{
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}], 
                 "\[And]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                    StyleBox["c", "ExpressionVariable"]}], "\[Equal]", 
                   RowBox[{
                    StyleBox["m2", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}], "\[Implies]", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], "\[Equal]", 
                 StyleBox[
                 "m2", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["m1", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["mult", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   RowBox[{
                    StyleBox["m1", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}], 
                 "\[And]", 
                 RowBox[{
                   UnderscriptBox["mult", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   RowBox[{
                    StyleBox["m2", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}], 
                 "\[And]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["m1", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                    StyleBox["c", "ExpressionVariable"]}], "\[Equal]", 
                   RowBox[{
                    StyleBox["m2", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                    StyleBox["c", "ExpressionVariable"]}]}]}], "\[Implies]", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], "\[Equal]", 
                 StyleBox[
                 "m2", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:568873110"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p568873110-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p568873110-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:568873110", "Proof|ID:568873110-1"},
 CellID->3033029],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p568873110-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->384543067]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(mult1 sufficient \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(mult1 sufficient \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
                 RowBox[{
                   RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                   "\[VerticalBar]", 
                   StyleBox["a", "ExpressionVariable"]}]}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["mult", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}], 
                   "\[And]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                    StyleBox[
                    "c", "ExpressionVariable"]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[NotEqual]", 
                   UnderscriptBox["0", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
                 RowBox[{
                   RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                   "\[VerticalBar]", 
                   StyleBox["a", "ExpressionVariable"]}]}], "\[Implies]", 
               RowBox[{
                 UnderscriptBox["\[Exists]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["mult", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}], 
                   "\[And]", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                    StyleBox[
                    "c", "ExpressionVariable"]}]}]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:347198673"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p347198673-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p347198673-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:347198673", "Proof|ID:347198673-1"},
 CellID->636035778],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p347198673-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->537098177]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->599981467],

Cell[CellGroupData[{

Cell["Lemma: Reducibility in \[ScriptCapitalZ]n", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.659177961397418*^9, 3.6591779683862305`*^9}},
 CellID->124267281],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "c"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\IntegerQuotientRings.nb", 25008217}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659177976420244*^9, 3.65917798534346*^9}},
 CellTags->{
  "ID:25008217", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->25008217],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalZ]n"], "[", 
   RowBox[{"a", ",", "c"}], "]"}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", "\[GreaterEqual]", 
    RowBox[{"gcd", "[", 
     RowBox[{"n", ",", "c"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReducible \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:761196101", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6591779900234685`*^9, 3.6591780163407145`*^9}},
 CellTags->{
  "ID:761196101", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isReducible \[ScriptCapitalZ]n"},
 CellID->761196101],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["isReducible", "\[ScriptCapitalZ]n"], "[", 
   RowBox[{"a", ",", "c"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{
       UnderscriptBox["multN", "\[ScriptCapitalZ]n"], "[", "c", "]"}]}]], 
    RowBox[{
     UnderscriptBox["\[Exists]", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
       RowBox[{"[", "m", "]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{
        UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
        RowBox[{"m", ",", "c", ",", "i"}], "]"}], "\[And]", 
       RowBox[{
        UnderscriptBox["isReducible", "\[ScriptCapitalZ]n"], "[", 
        RowBox[{"a", ",", "c", ",", "m"}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReducible' \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:216211003", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6592634752979546`*^9, 3.6592636063939857`*^9}},
 CellTags->{
  "ID:216211003", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isReducible' \[ScriptCapitalZ]n"},
 CellID->216211003],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"c", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], "\[And]", 
    RowBox[{
     UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
     RowBox[{"a", ",", "c"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"gcd", "[", 
       RowBox[{"n", ",", "c"}], "]"}], 
      UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]n"], 
      RowBox[{
       RowBox[{"gcd", "[", 
        RowBox[{"n", ",", "c"}], "]"}], 
       UnderscriptBox["-", "\[ScriptCapitalZ]n"], 
       RowBox[{"a", 
        UnderscriptBox["*", "\[ScriptCapitalZ]n"], "c"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"a", 
       UnderscriptBox["*", "\[ScriptCapitalZ]n"], "c"}], "\[Equal]", 
      RowBox[{"gcd", "[", 
       RowBox[{"n", ",", "c"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(gcd \[Succeeds] \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:431638777", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6592537123430195`*^9, 3.6592538235712147`*^9}, {
  3.659253858905277*^9, 3.6592538711356983`*^9}, {3.6592541441829777`*^9, 
  3.6592541504073887`*^9}, {3.6592541911078606`*^9, 3.6592541970514708`*^9}, {
  3.6592563431810226`*^9, 3.659256353976242*^9}},
 CellTags->{
  "ID:431638777", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "gcd \[Succeeds] \[ScriptCapitalZ]n"},
 CellID->431638777],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->379545476]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->25000160],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], "\[GreaterEqual]", 
             RowBox[{"gcd", "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], "\[GreaterEqual]", 
             RowBox[{"gcd", "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:761196101"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p761196101-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p761196101-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:761196101", "Proof|ID:761196101-1"},
 CellID->46321884],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p761196101-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->473334112]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReducible' \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReducible' \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["mult", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["isReducible", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isReducible", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{
                    UnderscriptBox["multN", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["mult", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"]}], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["isReducible", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["m", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:216211003"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p216211003-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p216211003-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:216211003", "Proof|ID:216211003-1"},
 CellID->182622923],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p216211003-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->72865]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(gcd \[Succeeds] \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(gcd \[Succeeds] \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], 
               UnderscriptBox["\[Succeeds]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], 
               RowBox[{
                 RowBox[{"gcd", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["-", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                   StyleBox["c", "ExpressionVariable"]}]}]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                 StyleBox["c", "ExpressionVariable"]}], "\[Equal]", 
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}], 
               UnderscriptBox["\[Succeeds]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], 
               RowBox[{
                 RowBox[{"gcd", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["-", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["*", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                   StyleBox["c", "ExpressionVariable"]}]}]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                 StyleBox["c", "ExpressionVariable"]}], "\[Equal]", 
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:431638777"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p431638777-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p431638777-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:431638777", "Proof|ID:431638777-1"},
 CellID->607642207],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p431638777-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->112691637]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->337195016],

Cell[CellGroupData[{

Cell["\<\
Lemma: Minimal Non-Trivial Common Reducibles in \[ScriptCapitalZ]n\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.659257490850641*^9, 3.659257504141864*^9}},
 CellID->379952792],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
   RowBox[{"[", "a", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\IntegerQuotientRings.nb", 219886779}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659267435839713*^9, 3.659267442719325*^9}},
 CellTags->{
  "ID:219886779", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->219886779],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
    RowBox[{"[", 
     RowBox[{"c1", ",", "c2"}], "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", 
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{
      UnderscriptBox["multN", "\[ScriptCapitalZ]n"], "[", "c1", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["isCommonReducible", "\[ScriptCapitalZ]n"], "[", 
      RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}], 
     "\[DoubleLongLeftRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       RowBox[{"a", "\[GreaterEqual]", 
        RowBox[{"gcd", "[", 
         RowBox[{"n", ",", "c1"}], "]"}]}], "\[And]", 
       RowBox[{"a", "\[GreaterEqual]", 
        RowBox[{"gcd", "[", 
         RowBox[{"n", ",", "c2"}], "]"}]}], "\[And]", 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{"c1", "\[Equal]", "c2"}], "\[Implies]", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"i", "\[NotEqual]", "j"}], 
          TagBox[")",
           "AutoParentheses"]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isCommonReducible \[ScriptCapitalZ]n)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:277360974", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659259342564495*^9, 3.659259348367705*^9}, {
  3.6592593811901627`*^9, 3.6592594453842754`*^9}, {3.659259714158148*^9, 
  3.6592597454830027`*^9}, {3.6592598507363873`*^9, 3.6592598527955914`*^9}, {
  3.6592604087513676`*^9, 3.6592604194227867`*^9}, {3.659260669678426*^9, 
  3.6592606812858467`*^9}, {3.6592607940116444`*^9, 3.65926083126451*^9}, {
  3.659260910465849*^9, 3.6592609252390747`*^9}, {3.659267446198131*^9, 
  3.6592674463541317`*^9}},
 CellTags->{
  "ID:277360974", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isCommonReducible \[ScriptCapitalZ]n"},
 CellID->277360974],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
     RowBox[{"[", "c", "]"}]}]], 
   RowBox[{"c", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalZ]n"]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isMNTCR", "\[ScriptCapitalZ]n"], "[", 
     RowBox[{"c", ",", "a"}], "]"}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", "\[Equal]", 
      RowBox[{"gcd", "[", 
       RowBox[{"n", ",", "c"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMNTCR unary \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:112584880", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6592575096642733`*^9, 3.659257559865162*^9}, {
  3.6592582129819098`*^9, 3.659258213715111*^9}, {3.659267447617734*^9, 
  3.659267447820534*^9}},
 CellTags->{
  "ID:112584880", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR unary \[ScriptCapitalZ]n"},
 CellID->112584880],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
     RowBox[{"[", "c", "]"}]}]], 
   RowBox[{"c", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalZ]n"]}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{"i", ",", 
      RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{
       UnderscriptBox["multN", "\[ScriptCapitalZ]n"], "[", "c", "]"}]}]], 
    RowBox[{"i", "\[NotEqual]", "j"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["isMNTCR", "\[ScriptCapitalZ]n"], "[", 
      RowBox[{"c", ",", "c", ",", "a", ",", "i", ",", "j"}], "]"}], 
     "\[DoubleLongLeftRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"a", "\[Equal]", 
       RowBox[{"gcd", "[", 
        RowBox[{"n", ",", "c"}], "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMNTCR identical \[ScriptCapitalZ]n)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:195146313", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659259093338457*^9, 3.6592591255213137`*^9}, {
  3.659259479875936*^9, 3.659259480359537*^9}, {3.6592607730764074`*^9, 
  3.659260775088811*^9}, {3.659260844664933*^9, 3.659260855179352*^9}, {
  3.6592674489593363`*^9, 3.6592674491153364`*^9}},
 CellTags->{
  "ID:195146313", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR identical \[ScriptCapitalZ]n"},
 CellID->195146313],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
     RowBox[{"[", 
      RowBox[{"c1", ",", "c2"}], "]"}]}]], 
   RowBox[{
    RowBox[{"c1", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], "\[And]", 
    RowBox[{"c2", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], "\[And]", 
    RowBox[{"c1", "\[NotEqual]", "c2"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", 
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{
      UnderscriptBox["multN", "\[ScriptCapitalZ]n"], "[", "c1", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      RowBox[{"gcd", "[", 
       RowBox[{"n", ",", "c1"}], "]"}], "\[GreaterEqual]", 
      RowBox[{"gcd", "[", 
       RowBox[{"n", ",", "c2"}], "]"}]}], "\[Implies]", 
     RowBox[{
      UnderscriptBox["isMNTCR", "\[ScriptCapitalZ]n"], "[", 
      RowBox[{"c1", ",", "c2", ",", 
       RowBox[{"gcd", "[", 
        RowBox[{"n", ",", "c1"}], "]"}], ",", "i", ",", "j"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMNTCR distinct' \[ScriptCapitalZ]n)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:6406785", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6592673864344263`*^9, 3.6592674186016827`*^9}, {
  3.6592674503633385`*^9, 3.659267476290584*^9}},
 CellTags->{
  "ID:6406785", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR distinct' \[ScriptCapitalZ]n"},
 CellID->6406785],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
     RowBox[{"[", 
      RowBox[{"c1", ",", "c2"}], "]"}]}]], 
   RowBox[{
    RowBox[{"c1", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], "\[And]", 
    RowBox[{"c2", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], "\[And]", 
    RowBox[{"c1", "\[NotEqual]", "c2"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", 
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{
      UnderscriptBox["multN", "\[ScriptCapitalZ]n"], "[", "c1", "]"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["isMNTCR", "\[ScriptCapitalZ]n"], "[", 
      RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}], 
     "\[DoubleLongLeftRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"a", "\[Equal]", 
       RowBox[{
        SubscriptBox["min", 
         UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]n"]], "[", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{"gcd", "[", 
           RowBox[{"n", ",", "c1"}], "]"}], ",", 
          RowBox[{"gcd", "[", 
           RowBox[{"n", ",", "c2"}], "]"}]}], "\[RightAngleBracket]"}], 
        "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMNTCR distinct \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:87852126", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6592589597878227`*^9, 3.659259048925179*^9}, {
  3.659260928733481*^9, 3.6592609406051016`*^9}, {3.6592674517517414`*^9, 
  3.6592674519233418`*^9}},
 CellTags->{
  "ID:87852126", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR distinct \[ScriptCapitalZ]n"},
 CellID->87852126],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
     RowBox[{"[", 
      RowBox[{"c1", ",", "c2"}], "]"}]}]], 
   RowBox[{
    RowBox[{"c1", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], "\[And]", 
    RowBox[{"c2", "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalZ]n"]}]}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{"i", ",", 
      RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{
       UnderscriptBox["multN", "\[ScriptCapitalZ]n"], "[", "c1", "]"}]}]], 
    RowBox[{
     RowBox[{"c1", "\[Equal]", "c2"}], "\[Implies]", 
     RowBox[{"i", "\[NotEqual]", "j"}]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["isMNTCR", "\[ScriptCapitalZ]n"], "[", 
      RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}], 
     "\[DoubleLongLeftRightArrow]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"a", "\[Equal]", 
       RowBox[{
        SubscriptBox["min", 
         UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]n"]], "[", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{"gcd", "[", 
           RowBox[{"n", ",", "c1"}], "]"}], ",", 
          RowBox[{"gcd", "[", 
           RowBox[{"n", ",", "c2"}], "]"}]}], "\[RightAngleBracket]"}], 
        "]"}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMNTCR \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:37400195", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659331831492114*^9, 3.6593318420533323`*^9}, {
  3.6593329903997498`*^9, 3.6593329943465567`*^9}},
 CellTags->{
  "ID:37400195", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR \[ScriptCapitalZ]n"},
 CellID->37400195],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->43844530]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->512005801],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isCommonReducible \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isCommonReducible \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isCommonReducible", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[GreaterEqual]", 
                 RowBox[{"gcd", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c1", "ExpressionVariable"]}], "]"}]}], "\[And]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[GreaterEqual]", 
                 RowBox[{"gcd", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"]}], "]"}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], "\[Equal]", 
                   StyleBox["c2", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["j", "ExpressionVariable"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isCommonReducible", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[GreaterEqual]", 
                 RowBox[{"gcd", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c1", "ExpressionVariable"]}], "]"}]}], "\[And]", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], "\[GreaterEqual]", 
                 RowBox[{"gcd", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"]}], "]"}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], "\[Equal]", 
                   StyleBox["c2", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                   StyleBox["j", "ExpressionVariable"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:277360974"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p277360974-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p277360974-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:277360974", "Proof|ID:277360974-1"},
 CellID->41026893],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p277360974-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->187450513]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMNTCR unary \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMNTCR unary \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMNTCR", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMNTCR", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:112584880"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p112584880-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p112584880-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:112584880", "Proof|ID:112584880-1"},
 CellID->228914149],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p112584880-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.659265521970551*^9, 3.659265523748954*^9}},
 CellID->255363086]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMNTCR identical \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMNTCR identical \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["j", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMNTCR", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["j", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMNTCR", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:195146313"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p195146313-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p195146313-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:195146313", "Proof|ID:195146313-1"},
 CellID->270250321],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p195146313-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellChangeTimes->{3.6592666859101954`*^9},
 CellID->366403582]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMNTCR distinct' \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMNTCR distinct' \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"]}], "]"}], 
               "\[GreaterEqual]", 
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"]}], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"]}], "]"}], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"]}], "]"}], 
               "\[GreaterEqual]", 
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"]}], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["isMNTCR", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               RowBox[{"gcd", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["c1", "ExpressionVariable"]}], "]"}], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:6406785"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p6406785-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p6406785-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:6406785", "Proof|ID:6406785-1"},
 CellID->295234282],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p6406785-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->90395247]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMNTCR distinct \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMNTCR distinct \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMNTCR", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Equal]", 
               RowBox[{
                 SubscriptBox["min", 
                  UnderscriptBox["\[Succeeds]", 
                   RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c1", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"]}], "]"}]}], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["c2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMNTCR", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Equal]", 
               RowBox[{
                 SubscriptBox["min", 
                  UnderscriptBox["\[Succeeds]", 
                   RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c1", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"]}], "]"}]}], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:87852126"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p87852126-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p87852126-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:87852126", "Proof|ID:87852126-1"},
 CellID->306702955],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p87852126-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->453994183]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMNTCR \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMNTCR \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["c2", "ExpressionVariable"]}], "\[Implies]", 
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                 StyleBox["j", "ExpressionVariable"]}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMNTCR", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Equal]", 
               RowBox[{
                 SubscriptBox["min", 
                  UnderscriptBox["\[Succeeds]", 
                   RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c1", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"]}], "]"}]}], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["c1", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["c2", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], "\[Equal]", 
                 StyleBox["c2", "ExpressionVariable"]}], "\[Implies]", 
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
                 StyleBox["j", "ExpressionVariable"]}]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isMNTCR", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["c1", "ExpressionVariable"], ",", 
                 StyleBox["c2", "ExpressionVariable"], ",", 
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], "\[Equal]", 
               RowBox[{
                 SubscriptBox["min", 
                  UnderscriptBox["\[Succeeds]", 
                   RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c1", "ExpressionVariable"]}], "]"}], ",", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["c2", "ExpressionVariable"]}], "]"}]}], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:37400195"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p37400195-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p37400195-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:37400195", "Proof|ID:37400195-1"},
 CellID->282288311],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p37400195-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->307661666]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["R5 & R6 (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6584655333618546`*^9, 3.658465534174056*^9}, {
  3.6585590483662367`*^9, 3.6585590507062407`*^9}, {3.65942401958659*^9, 
  3.659424023704997*^9}},
 CellID->73381205],

Cell["", "OpenEnvironment",
 CellID->91360536],

Cell[CellGroupData[{

Cell["Lemma: Consequences of Correlativity", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6594172313852525`*^9, 3.659417241104069*^9}},
 CellID->154663064],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
    RowBox[{"[", 
     RowBox[{"c", ",", "m1", ",", "m2"}], "]"}]}]], 
  RowBox[{"c", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalZ]n"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\IntegerQuotientRings.nb", 14129606}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659419836175034*^9, 3.659419842508645*^9}},
 CellTags->{
  "ID:14129606", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->14129606],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
     RowBox[{"m1", ",", "c", ",", "1"}], "]"}], "\[And]", 
    RowBox[{
     UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
     RowBox[{"m2", ",", "c", ",", "1"}], "]"}], "\[And]", 
    RowBox[{"m2", ">", "m1"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
   RowBox[{
    RowBox[{"m2", 
     UnderscriptBox["-", "\[ScriptCapitalZ]n"], "m1"}], ",", "c", ",", "1"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(mult1 - \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:126313666", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6594199712400713`*^9, 3.6594200036869283`*^9}},
 CellTags->{
  "ID:126313666", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 - \[ScriptCapitalZ]n"},
 CellID->126313666],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["areCorrelative", "\[ScriptCapitalZ]n"], "[", 
     RowBox[{"c", ",", "m1", ",", "m2"}], "]"}], "\[And]", 
    RowBox[{"m2", ">", "m1"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
   RowBox[{
    RowBox[{"m2", 
     UnderscriptBox["-", "\[ScriptCapitalZ]n"], "m1"}], ",", "c", ",", "1"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(areCorrelative 1 \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:888879353", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6594198570946703`*^9, 3.659419860089876*^9}, {
  3.6594199025375504`*^9, 3.659419910197164*^9}},
 CellTags->{
  "ID:888879353", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "areCorrelative 1 \[ScriptCapitalZ]n"},
 CellID->888879353],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["areCorrelative", "\[ScriptCapitalZ]n"], "[", 
     RowBox[{"c", ",", "m1", ",", "m2"}], "]"}], "\[And]", 
    RowBox[{"m1", "\[NotEqual]", "m2"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
   RowBox[{
    RowBox[{"m2", 
     UnderscriptBox["-", "\[ScriptCapitalZ]n"], "m1"}], ",", "c"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(areCorrelative 2 \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:23406014", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6594172533968906`*^9, 3.659417296063966*^9}, {
  3.6594173525672655`*^9, 3.6594174152325754`*^9}, {3.6594184353629694`*^9, 
  3.6594185026614876`*^9}, {3.6594198445210485`*^9, 3.659419847032653*^9}},
 CellTags->{
  "ID:23406014", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "areCorrelative 2 \[ScriptCapitalZ]n"},
 CellID->23406014],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->179715897]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->77124816],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(mult1 - \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(mult1 - \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["mult", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}], 
             "\[And]", 
             RowBox[{
               StyleBox["m2", "ExpressionVariable"], ">", 
               StyleBox["m1", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                 StyleBox["m1", "ExpressionVariable"]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", "1"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               UnderscriptBox["mult", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["mult", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"], ",", "1"}], "]"}], 
             "\[And]", 
             RowBox[{
               StyleBox["m2", "ExpressionVariable"], ">", 
               StyleBox["m1", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                 StyleBox["m1", "ExpressionVariable"]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", "1"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:126313666"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p126313666-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p126313666-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:126313666", "Proof|ID:126313666-1"},
 CellID->79969318],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p126313666-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->315040505]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(areCorrelative 1 \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(areCorrelative 1 \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCorrelative", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               StyleBox["m2", "ExpressionVariable"], ">", 
               StyleBox["m1", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                 StyleBox["m1", "ExpressionVariable"]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", "1"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCorrelative", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               StyleBox["m2", "ExpressionVariable"], ">", 
               StyleBox["m1", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                 StyleBox["m1", "ExpressionVariable"]}], ",", 
               StyleBox["c", "ExpressionVariable"], ",", "1"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:888879353"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p888879353-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p888879353-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:888879353", "Proof|ID:888879353-1"},
 CellID->341588],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p888879353-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->625993774]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(areCorrelative 2 \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(areCorrelative 2 \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCorrelative", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               StyleBox["m1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["m2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                 StyleBox["m1", "ExpressionVariable"]}], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["m2", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[And]", 
             RowBox[{
               UnderscriptBox["areCorrelative", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], ",", 
                 StyleBox["m1", "ExpressionVariable"], ",", 
                 StyleBox["m2", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               StyleBox["m1", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox["m2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["m2", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                 StyleBox["m1", "ExpressionVariable"]}], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:23406014"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p23406014-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p23406014-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:23406014", "Proof|ID:23406014-1"},
 CellID->386167612],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p23406014-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->145559439]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->909951923],

Cell[CellGroupData[{

Cell["Lemma: \[ScriptCapitalZ]n Satisfies R5 and R6", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.594446565624218*^9, 3.594446591442624*^9}, {
  3.6585590704558754`*^9, 3.6585590812354937`*^9}},
 CellID->1129973761],

Cell[BoxData[
 RowBox[{"R5", "[", "\[ScriptCapitalZ]n", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R5 \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:325383468", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594446735269612*^9, 3.594446748768272*^9}, {
  3.658466310841421*^9, 3.6584663122298236`*^9}},
 CellTags->{
  "ID:325383468", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R5 \[ScriptCapitalZ]n"},
 CellID->325383468],

Cell[BoxData[
 RowBox[{"R6", "[", "\[ScriptCapitalZ]n", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R6 \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1783775451", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594446735968205*^9, 3.594446749616791*^9}, 
   3.6584663148818283`*^9},
 CellTags->{
  "ID:1783775451", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R6 \[ScriptCapitalZ]n"},
 CellID->1783775451],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->955608461]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->521644330],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R5 \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R5 \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R5", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R5", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:325383468"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p325383468-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p325383468-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:325383468", "Proof|ID:325383468-1"},
 CellID->4554534],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p325383468-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->2823860]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R6 \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R6 \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R6", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R6", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1783775451"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1783775451-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1783775451-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1783775451", "Proof|ID:1783775451-1"},
 CellID->190836234],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1783775451-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->5677791]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["R7 & R8 (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.658559056727851*^9, 3.658559059426656*^9}, {
  3.6585628496787186`*^9, 3.658562854124727*^9}},
 CellID->124396989],

Cell["", "OpenEnvironment",
 CellID->12240399],

Cell[CellGroupData[{

Cell["Lemma: Preparation for R7 and R8", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.658559154742823*^9, 3.6585591613260345`*^9}},
 CellID->576684220],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "c1", ",", "c2"}], "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", "j"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["isNTCR", "\[ScriptCapitalZ]n"], "[", 
      RowBox[{"c1", ",", "c2", ",", "a", ",", "i", ",", "j"}], "]"}], 
     "\[Implies]", 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
         RowBox[{"[", "a0", "]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{
           RowBox[{"a", 
            UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]n"], "a0"}], 
           "\[Or]", 
           RowBox[{"a", "\[Equal]", "a0"}]}], 
          TagBox[")",
           "AutoParentheses"]}], "\[And]", 
         RowBox[{
          UnderscriptBox["isMNTCR", "\[ScriptCapitalZ]n"], "[", 
          RowBox[{"c1", ",", "c2", ",", "a0", ",", "i", ",", "j"}], "]"}]}], 
        TagBox[")",
         "AutoParentheses"]}]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isMNTCR existence \[ScriptCapitalZ]n)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:394176110", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6585591892524853`*^9, 3.658559306518691*^9}},
 CellTags->{
  "ID:394176110", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR existence \[ScriptCapitalZ]n"},
 CellID->394176110],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "c"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isNTCR", "\[ScriptCapitalZ]n"], "[", 
     RowBox[{"c", ",", "a"}], "]"}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
        RowBox[{"[", "a0", "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          RowBox[{"a", 
           UnderscriptBox["\[Succeeds]", "\[ScriptCapitalZ]n"], "a0"}], 
          "\[Or]", 
          RowBox[{"a", "\[Equal]", "a0"}]}], 
         TagBox[")",
          "AutoParentheses"]}], "\[And]", 
        RowBox[{
         UnderscriptBox["isMNTCR", "\[ScriptCapitalZ]n"], "[", 
         RowBox[{"c", ",", "a0"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(isMNTCR unary existence \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:232518088", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.658559314724305*^9, 3.658559332367936*^9}},
 CellTags->{
  "ID:232518088", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR unary existence \[ScriptCapitalZ]n"},
 CellID->232518088],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->135443090]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->257333253],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMNTCR existence \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMNTCR existence \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isNTCR", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["a0", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                   StyleBox["a0", "ExpressionVariable"]}], "\[Or]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                   StyleBox["a0", "ExpressionVariable"]}]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isMNTCR", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["a0", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c1", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c2", "ExpressionVariable"], "]"}]}, {
             StyleBox["i", "ExpressionVariable"]}, {
             StyleBox["j", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isNTCR", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["c1", "ExpressionVariable"], ",", 
               StyleBox["c2", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"], ",", 
               StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["a0", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                   StyleBox["a0", "ExpressionVariable"]}], "\[Or]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                   StyleBox["a0", "ExpressionVariable"]}]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isMNTCR", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], ",", 
                   StyleBox["c2", "ExpressionVariable"], ",", 
                   StyleBox["a0", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:394176110"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p394176110-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p394176110-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:394176110", "Proof|ID:394176110-1"},
 CellID->648217336],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p394176110-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->127336303]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isMNTCR unary existence \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isMNTCR unary existence \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isNTCR", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["a0", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                   StyleBox["a0", "ExpressionVariable"]}], "\[Or]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                   StyleBox["a0", "ExpressionVariable"]}]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isMNTCR", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["a0", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isNTCR", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   StyleBox["a0", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                   StyleBox["a0", "ExpressionVariable"]}], "\[Or]", 
                 RowBox[{
                   StyleBox["a", "ExpressionVariable"], "\[Equal]", 
                   StyleBox["a0", "ExpressionVariable"]}]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["isMNTCR", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["a0", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:232518088"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p232518088-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p232518088-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:232518088", "Proof|ID:232518088-1"},
 CellID->424300880],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p232518088-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->49071767]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->644279200],

Cell[CellGroupData[{

Cell["Lemma: \[ScriptCapitalZ]n Satisfies R7 and R8", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.658559087054304*^9, 3.658559092139913*^9}},
 CellID->334567729],

Cell[BoxData[
 RowBox[{"R7", "[", "\[ScriptCapitalZ]n", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R7 \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2147206476", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594446737113096*^9, 3.59444675053875*^9}, 
   3.6584663161766305`*^9},
 CellTags->{
  "ID:2147206476", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R7 \[ScriptCapitalZ]n"},
 CellID->2147206476],

Cell[BoxData[
 RowBox[{"R8", "[", "\[ScriptCapitalZ]n", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R8 \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1971443720", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594446737852884*^9, 3.594446751607988*^9}, 
   3.658466317019032*^9},
 CellTags->{
  "ID:1971443720", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R8 \[ScriptCapitalZ]n"},
 CellID->1971443720],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->79251623]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->112564214],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R7 \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R7 \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R7", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R7", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2147206476"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2147206476-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2147206476-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2147206476", "Proof|ID:2147206476-1"},
 CellID->31414230],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2147206476-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->347321696]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R8 \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R8 \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R8", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R8", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1971443720"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1971443720-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1971443720-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1971443720", "Proof|ID:1971443720-1"},
 CellID->239044020],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1971443720-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->28742966]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Main Result (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.658466133045909*^9, 3.6584661352455125`*^9}, {
  3.659424178116068*^9, 3.6594241823748755`*^9}},
 CellID->106088833],

Cell["", "OpenEnvironment",
 CellID->1543467287],

Cell[CellGroupData[{

Cell["\<\
Theorem: \[ScriptCapitalZ]n is a Reduction Ring\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.594446828495274*^9, 3.594446846865993*^9}, {
  3.607338531610847*^9, 3.607338531759572*^9}, {3.6584661807819924`*^9, 
  3.658466184916*^9}},
 CellID->1491276973],

Cell[BoxData[
 RowBox[{"isReductionRing", "[", "\[ScriptCapitalZ]n", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isReductionRing \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:727194704", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594446853531086*^9, 3.594446879122588*^9}, {
  3.607338525855809*^9, 3.607338526852516*^9}, {3.658466138365518*^9, 
  3.658466139941121*^9}},
 CellTags->{
  "ID:727194704", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isReductionRing \[ScriptCapitalZ]n"},
 CellID->727194704],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1443393611]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->2047014872],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isReductionRing \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isReductionRing \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"isReductionRing", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"isReductionRing", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:727194704"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p727194704-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p727194704-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:727194704", "Proof|ID:727194704-1"},
 CellID->278604702],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p727194704-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->8265060]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Algorithmic Reduction Ring (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.658465468698741*^9, 3.6584654734411488`*^9}, {
  3.659355566353952*^9, 3.6593555710807605`*^9}},
 CellID->24901450],

Cell[CellGroupData[{

Cell["R9-R13 (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.65846611104987*^9, 3.658466115090277*^9}, {
  3.659355552735128*^9, 3.659355556869135*^9}},
 CellID->415571348],

Cell["", "OpenEnvironment",
 CellID->319526061],

Cell[CellGroupData[{

Cell["Lemma: Auxiliary Results Needed for R12", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.659340555672393*^9, 3.659340563878008*^9}},
 CellID->8268143],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
   RowBox[{"[", 
    RowBox[{"x", ",", "y"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\IntegerQuotientRings.nb", 73858040}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659347104040884*^9, 3.6593471130577*^9}},
 CellTags->{
  "ID:73858040", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->73858040],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{"m", "\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]], 
   RowBox[{"n", ">", "m"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
        RowBox[{"[", "k", "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"m", "\[GreaterEqual]", "k"}], "\[And]", 
        RowBox[{
         RowBox[{"k", 
          UnderscriptBox["*", "\[ScriptCapitalZ]n"], "x"}], "\[Equal]", 
         "y"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
       RowBox[{"[", 
        RowBox[{
         UnderscriptBox["rdm", "\[ScriptCapitalZ]n"], "[", 
         RowBox[{"x", ",", "m", ",", "y"}], "]"}], "]"}]}], "\[And]", 
      RowBox[{"m", "\[GreaterEqual]", 
       RowBox[{
        UnderscriptBox["rdm", "\[ScriptCapitalZ]n"], "[", 
        RowBox[{"x", ",", "m", ",", "y"}], "]"}]}], "\[And]", 
      RowBox[{
       RowBox[{
        RowBox[{
         UnderscriptBox["rdm", "\[ScriptCapitalZ]n"], "[", 
         RowBox[{"x", ",", "m", ",", "y"}], "]"}], 
        UnderscriptBox["*", "\[ScriptCapitalZ]n"], "x"}], "\[Equal]", "y"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(rdm ternary \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:502941593", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659340578588834*^9, 3.659340723372688*^9}, 
   3.6593424094910383`*^9, {3.6593424650115356`*^9, 3.6593424973035927`*^9}, {
   3.6593425337764564`*^9, 3.6593425473640804`*^9}, {3.6593426200134077`*^9, 
   3.659342633891432*^9}, {3.6593429171099296`*^9, 3.659342919824334*^9}, {
   3.659342964050412*^9, 3.6593429655168147`*^9}, {3.6593437365793686`*^9, 
   3.659343767983224*^9}, 3.659347116333706*^9},
 CellTags->{
  "ID:502941593", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm ternary \[ScriptCapitalZ]n"},
 CellID->502941593],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Exists]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
    RowBox[{"[", "k", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"quo", "[", 
       RowBox[{"n", ",", 
        RowBox[{"gcd", "[", 
         RowBox[{"n", ",", "y"}], "]"}]}], "]"}], "-", "1"}], 
     "\[GreaterEqual]", "k"}], "\[And]", 
    RowBox[{
     RowBox[{"k", 
      UnderscriptBox["*", "\[ScriptCapitalZ]n"], "y"}], "\[Equal]", 
     RowBox[{"x", "-", 
      RowBox[{"mod", "[", 
       RowBox[{"x", ",", 
        RowBox[{"gcd", "[", 
         RowBox[{"n", ",", "y"}], "]"}]}], "]"}]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(rdm aux \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:691632805", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659347155146574*^9, 3.6593472328309135`*^9}},
 CellTags->{
  "ID:691632805", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm aux \[ScriptCapitalZ]n"},
 CellID->691632805],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalZ]n"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["rdm", "\[ScriptCapitalZ]n"], "[", 
    RowBox[{"x", ",", "y"}], "]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(membership rdm \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:290458459", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6593521274783115`*^9, 3.6593521406447344`*^9}},
 CellTags->{
  "ID:290458459", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership rdm \[ScriptCapitalZ]n"},
 CellID->290458459],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    UnderscriptBox["rdm", "\[ScriptCapitalZ]n"], "[", 
    RowBox[{"x", ",", "y"}], "]"}], "\[Equal]", 
   UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], "\[Or]", 
  RowBox[{
   UnderscriptBox["mult", "\[ScriptCapitalZ]n"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["rdm", "\[ScriptCapitalZ]n"], "[", 
     RowBox[{"x", ",", "y"}], "]"}], ",", "y", ",", "1"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(rdm cases \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:38241486", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659352073829817*^9, 3.6593521019878664`*^9}},
 CellTags->{
  "ID:38241486", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm cases \[ScriptCapitalZ]n"},
 CellID->38241486],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    UnderscriptBox["rdm", "\[ScriptCapitalZ]n"], "[", 
    RowBox[{"x", ",", "y"}], "]"}], 
   UnderscriptBox["*", "\[ScriptCapitalZ]n"], "y"}], "\[Equal]", 
  RowBox[{"x", "-", 
   RowBox[{"mod", "[", 
    RowBox[{"x", ",", 
     RowBox[{"gcd", "[", 
      RowBox[{"n", ",", "y"}], "]"}]}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(rdm * \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:140983642", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6593522055252485`*^9, 3.659352225259283*^9}},
 CellTags->{
  "ID:140983642", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm * \[ScriptCapitalZ]n"},
 CellID->140983642],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["rdm", "\[ScriptCapitalZ]n"], "[", 
   RowBox[{"x", ",", 
    UnderscriptBox["0", "\[ScriptCapitalZ]n"]}], "]"}], "\[Equal]", 
  UnderscriptBox["0", "\[ScriptCapitalZ]n"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(rdm 0 \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:33831774", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6593517688794813`*^9, 3.6593517839647074`*^9}},
 CellTags->{
  "ID:33831774", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm 0 \[ScriptCapitalZ]n"},
 CellID->33831774],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->16043932]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->445628779],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(rdm ternary \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(rdm ternary \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], ">", 
               StyleBox["m", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], "\[GreaterEqual]", 
                   StyleBox["k", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                    StyleBox["x", "ExpressionVariable"]}], "\[Equal]", 
                   StyleBox["y", "ExpressionVariable"]}]}]}]}], "\[Implies]", 
           
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 UnderscriptBox["rdm", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}], 
             "\[And]", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[GreaterEqual]", 
               RowBox[{
                 UnderscriptBox["rdm", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}]}], "\[And]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["rdm", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["*", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                 StyleBox["x", "ExpressionVariable"]}], "\[Equal]", 
               StyleBox[
               "y", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], ">", 
               StyleBox["m", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Exists]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                    StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], "\[GreaterEqual]", 
                   StyleBox["k", "ExpressionVariable"]}], "\[And]", 
                 RowBox[{
                   RowBox[{
                    StyleBox["k", "ExpressionVariable"], 
                    UnderscriptBox["*", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                    StyleBox["x", "ExpressionVariable"]}], "\[Equal]", 
                   StyleBox["y", "ExpressionVariable"]}]}]}]}], "\[Implies]", 
           
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 UnderscriptBox["rdm", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], "]"}], 
             "\[And]", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[GreaterEqual]", 
               RowBox[{
                 UnderscriptBox["rdm", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}]}], "\[And]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["rdm", 
                    RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["*", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                 StyleBox["x", "ExpressionVariable"]}], "\[Equal]", 
               StyleBox[
               "y", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:502941593"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p502941593-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p502941593-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:502941593", "Proof|ID:502941593-1"},
 CellID->32304572],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p502941593-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->251898129]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(rdm aux \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(rdm aux \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               RowBox[{
                 RowBox[{"quo", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], "]"}], 
                 "-", "1"}], "\[GreaterEqual]", 
               StyleBox["k", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                 StyleBox["y", "ExpressionVariable"]}], "\[Equal]", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "-", 
                 RowBox[{"mod", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               RowBox[{
                 RowBox[{"quo", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], "]"}], 
                 "-", "1"}], "\[GreaterEqual]", 
               StyleBox["k", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], 
                 StyleBox["y", "ExpressionVariable"]}], "\[Equal]", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], "-", 
                 RowBox[{"mod", "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    RowBox[{"gcd", "[", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:691632805"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p691632805-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p691632805-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:691632805", "Proof|ID:691632805-1"},
 CellID->23114620],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p691632805-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->91619570]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(membership rdm \[ScriptCapitalZ]n)\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(membership rdm \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ReductionIQR", "[", 
              StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
           RowBox[{
             UnderscriptBox["rdm", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"ReductionIQR", "[", 
              StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
           RowBox[{
             UnderscriptBox["rdm", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               StyleBox["y", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:290458459"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p290458459-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p290458459-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:290458459", "Proof|ID:290458459-1"},
 CellID->155701127],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p290458459-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellChangeTimes->{
  3.659352870308816*^9, {3.6593533117269917`*^9, 3.659353313115394*^9}},
 CellID->8836418]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(rdm cases \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(rdm cases \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["rdm", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[Or]", 
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["rdm", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], ",", 
               StyleBox["y", "ExpressionVariable"], ",", "1"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["rdm", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox["0", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]]}], "\[Or]", 
           RowBox[{
             UnderscriptBox["mult", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["rdm", 
                  RowBox[{"ReductionIQR", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
                 RowBox[{
                   StyleBox["x", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], "]"}], ",", 
               StyleBox["y", "ExpressionVariable"], ",", "1"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:38241486"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p38241486-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p38241486-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:38241486", "Proof|ID:38241486-1"},
 CellID->904484803],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p38241486-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->2214627]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(rdm * \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(rdm * \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["rdm", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["*", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             StyleBox["y", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "-", 
             RowBox[{"mod", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 RowBox[{"gcd", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["rdm", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["*", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], 
             StyleBox["y", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "-", 
             RowBox[{"mod", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 RowBox[{"gcd", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:140983642"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p140983642-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p140983642-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:140983642", "Proof|ID:140983642-1"},
 CellID->492763287],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p140983642-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->440487519]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(rdm 0 \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(rdm 0 \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["rdm", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "]"}], 
           "\[Equal]", 
           UnderscriptBox["0", 
            RowBox[{"ReductionIQR", "[", 
              StyleBox["n", "ExpressionVariable"], 
              "]"}]]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["rdm", 
              RowBox[{"ReductionIQR", "[", 
                StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], ",", 
               UnderscriptBox["0", 
                RowBox[{"ReductionIQR", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]]}], "]"}], 
           "\[Equal]", 
           UnderscriptBox["0", 
            RowBox[{"ReductionIQR", "[", 
              StyleBox["n", "ExpressionVariable"], 
              "]"}]]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:33831774"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p33831774-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p33831774-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:33831774", "Proof|ID:33831774-1"},
 CellID->447394967],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p33831774-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->35066950]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->840971231],

Cell[CellGroupData[{

Cell["Lemma: \[ScriptCapitalZ]n satisfies R9-R13", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.658466150455539*^9, 3.658466159487955*^9}},
 CellID->11062889],

Cell[BoxData[
 RowBox[{"R9", "[", "\[ScriptCapitalZ]n", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R9 \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1376076253", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594446738592992*^9, 3.594446753192721*^9}, 
   3.6584663773649383`*^9},
 CellTags->{
  "ID:1376076253", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R9 \[ScriptCapitalZ]n"},
 CellID->1376076253],

Cell[BoxData[
 RowBox[{"R10", "[", "\[ScriptCapitalZ]n", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R10 \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:125450892", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594446739590154*^9, 3.594446756253659*^9}, 
   3.6584663788157406`*^9},
 CellTags->{
  "ID:125450892", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R10 \[ScriptCapitalZ]n"},
 CellID->125450892],

Cell[BoxData[
 RowBox[{"R11", "[", "\[ScriptCapitalZ]n", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R11 \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:189921319", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594446742377207*^9, 3.594446758540019*^9}, 
   3.6584663798453426`*^9},
 CellTags->{
  "ID:189921319", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R11 \[ScriptCapitalZ]n"},
 CellID->189921319],

Cell[BoxData[
 RowBox[{"R12", "[", "\[ScriptCapitalZ]n", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R12 \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:667306596", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594446743161616*^9, 3.594446759689316*^9}, 
   3.6584663809217443`*^9},
 CellTags->{
  "ID:667306596", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R12 \[ScriptCapitalZ]n"},
 CellID->667306596],

Cell[BoxData[
 RowBox[{"R13", "[", "\[ScriptCapitalZ]n", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(R13 \[ScriptCapitalZ]n)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:194342088", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.658466413905203*^9, 3.6584664160736065`*^9}},
 CellTags->{
  "ID:194342088", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R13 \[ScriptCapitalZ]n"},
 CellID->194342088],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->72432757]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->394157218],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R9 \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R9 \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R9", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R9", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1376076253"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1376076253-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1376076253-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1376076253", "Proof|ID:1376076253-1"},
 CellID->793462086],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1376076253-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->629931647]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R10 \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R10 \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R10", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R10", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:125450892"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p125450892-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p125450892-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:125450892", "Proof|ID:125450892-1"},
 CellID->207385051],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p125450892-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->178882523]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R11 \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R11 \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R11", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R11", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:189921319"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p189921319-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p189921319-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:189921319", "Proof|ID:189921319-1"},
 CellID->221524119],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p189921319-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->453410172]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R12 \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R12 \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R12", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R12", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:667306596"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p667306596-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p667306596-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:667306596", "Proof|ID:667306596-1"},
 CellID->389476407],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p667306596-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {61., 67.}}]], "ProofInfoBody",
 CellID->30462548]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(R13 \[ScriptCapitalZ]n)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(R13 \[ScriptCapitalZ]n)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R13", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"R13", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:194342088"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p194342088-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p194342088-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:194342088", "Proof|ID:194342088-1"},
 CellID->102872355],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p194342088-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->107661161]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Main Result (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6584661194114847`*^9, 3.6584661215018883`*^9}, {
  3.659355558959539*^9, 3.6593555631247463`*^9}},
 CellID->365692985],

Cell["", "OpenEnvironment",
 CellID->289788666],

Cell[CellGroupData[{

Cell["\<\
Theorem: \[ScriptCapitalZ]n is an Algorithmic Reduction Ring\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6584661876772046`*^9, 3.6584661961792192`*^9}, 
   3.658466446030659*^9},
 CellID->315438148],

Cell[BoxData[
 RowBox[{"isAlgoReductionRing", "[", "\[ScriptCapitalZ]n", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isAlgoReductionRing \[ScriptCapitalZ]n)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:403270264", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6584662005160275`*^9, 3.6584662028092313`*^9}},
 CellTags->{
  "ID:403270264", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isAlgoReductionRing \[ScriptCapitalZ]n"},
 CellID->403270264],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->379063054]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->23977526],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isAlgoReductionRing \[ScriptCapitalZ]n)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isAlgoReductionRing \[ScriptCapitalZ]n)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"isAlgoReductionRing", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{"isAlgoReductionRing", "[", 
           RowBox[{"ReductionIQR", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:403270264"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p403270264-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p403270264-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:403270264", "Proof|ID:403270264-1"},
 CellID->78837715],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p403270264-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->30002591]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]]
}, Open  ]]
},
WindowSize->{1108, 762},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
ShowSelection->True,
CounterAssignments->{{"TheoremaFormulaCounter" -> 1}},
Magnification->1.,
FrontEndVersion->"9.0 for Linux x86 (64-bit) (February 7, 2013)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Theorema"}, "TheoremaCore.nb"]]], 
   
   Cell["Modify the definitions below to change the default appearance of all \
cells in a given style. \nMake modifications to any definition using commands \
in the Format menu.", "Text"], 
   Cell[
    CellGroupData[{
      Cell["Notebook Options Settings", "Section"], 
      Cell[
      "The options defined for the style below will be used at the Notebook \
level.", "Text"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Notebook"], InputAliases -> {"and" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[And]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "or" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", "\[Or]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "not" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Not]", "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "impl" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[Implies]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[DoubleLongLeftRightArrow]", 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                "\[SelectionPlaceholder]", "\[Equal]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox[
                  
                  RowBox[{
                   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                  "a:=b"], "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", ":=", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "andn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[And]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "orn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Or]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "equivn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[DoubleLeftRightArrow]", "\[Piecewise]", 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "cdist" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Piecewise]", 
                 
                 GridBox[{{
                   "\[SelectionPlaceholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}, {
                   "\[Placeholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "far" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exr" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "farc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exrc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "let" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["let", 
                  RowBox[{"\[Placeholder]", "=", "\[Placeholder]"}]], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "app" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293a", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "prep" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293b", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "join" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22c8", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "elemT" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22ff", Identity, SyntaxForm -> "a+b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "subop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "supop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SuperscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "subsupop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubsuperscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "oop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "uoop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               UnderoverscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "osop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox[
                OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
                "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "op" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               TagBox["\[SelectionPlaceholder]", Identity], 
               TagBox[")", "AutoParentheses"]}], "[]" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:e114", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:e115", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "()" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:fd3e", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:fd3f", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "<>" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27e8", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27e9", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "{}" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:e117", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:e118", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "[[]]" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27e6", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27e7", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "(())" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:2e28", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:2e29", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "<<>>" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27ea", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27eb", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "{{}}" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:2983", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:2984", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "(" -> 
            TagBox["(", "AutoParentheses"], ")" -> 
            TagBox[")", "AutoParentheses"]}, Background -> 
          RGBColor[0.957, 0.96, 0.97]], 
         Cell[
          StyleData["Notebook", "Printout"], Background -> None]}, Open]]}, 
     Open]], 
   Cell[
    CellGroupData[{
      Cell["Style Environment Names", "Section"], 
      Cell[
       StyleData[All, "Working"], DockedCells -> {
         Cell[
         "Theorema 2.0", "SBO", FontTracking -> 8, 
          CellMargins -> {{0, 0}, {0, 0}}, 
          CellFrameMargins -> {{0, 50}, {0, 0}}, FontSize -> 28, FontSlant -> 
          "Plain", FontColor -> RGBColor[0.29, 0.46, 0.6], TextAlignment -> 
          Right, Background -> RGBColor[0.36, 0.54, 0.69]]}], 
      Cell[
       StyleData[
       All, "Layout", StyleDefinitions -> StyleData[All, "Working"]], 
       MenuSortingValue -> 1200]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Styles for Title and Section Cells", "Section"], 
      Cell[
       CellGroupData[{
         Cell["Title", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], CellMargins -> {{27, 27}, {10, 30}}, 
             TextAlignment -> Center, MenuSortingValue -> 1100, 
             MenuCommandKey -> "1", FontFamily -> "Bitstream Charter", 
             FontSize -> 36, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Title", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{6, 10}, {5, 30}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], CellMargins -> {{27, 27}, {20, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1150, FontFamily -> 
             "Bitstream Charter", FontSize -> 24, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{6, 10}, {5, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], CellMargins -> {{27, 27}, {8, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Author"], CellMargins -> {{27, 27}, {18, 20}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Sectioning", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], ShowGroupOpener -> True, MenuSortingValue -> 
             1250, MenuCommandKey -> "4", FontFamily -> "Bitstream Charter", 
             FontSize -> 20, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Section", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{2, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {8, 12}}, 
             MenuSortingValue -> 1300, MenuCommandKey -> "5", FontFamily -> 
             "Bitstream Charter", FontSize -> 14, FontWeight -> "Bold", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Subsection", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{9, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {2, 10}}, 
             MenuSortingValue -> 1350, MenuCommandKey -> "6", FontFamily -> 
             "Bitstream Charter", FontSize -> 12, FontWeight -> "Bold", 
             FontSlant -> "Italic", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsection", "Printout"], ShowGroupOpener -> False,
              CellMargins -> {{35, 0}, {7, 14}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsubsection"], CellDingbat -> None, 
             ShowGroupOpener -> True, 
             CellMargins -> {{27, Inherited}, {Inherited, Inherited}}, 
             FontFamily -> "Bitstream Charter", FontSize -> 11, FontWeight -> 
             "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsubsection", "Printout"], 
             CellMargins -> {{50, 0}, {7, 14}}]}, Closed]]}, Open]]}, Open]], 
   
   Cell[
    CellGroupData[{
      Cell["Styles for Publication\[Hyphen]ready Documents", "Section"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Author"], CellMargins -> {{27, 30}, {20, 5}}, 
          MenuSortingValue -> 1900, FontFamily -> "Helvetica", FontSize -> 14,
           FontSlant -> "Italic"], 
         Cell[
          StyleData["Author", "Printout"], CellMargins -> {{6, 10}, {0, 10}}, 
          FontSize -> 10]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Abstract"], CellFrame -> {{0, 0}, {0.5, 0.5}}, 
          CellMargins -> {{27, 140}, {20, 20}}, LineSpacing -> {0.9, 0}, 
          MenuSortingValue -> 2050, FontFamily -> "Times", FontSize -> 12], 
         Cell[
          StyleData["Abstract", "Printout"], Hyphenation -> True, 
          LineSpacing -> {1, 2}, FontSize -> 10]}, Closed]]}, Closed]], 
   Cell[
    CellGroupData[{
      Cell["Text", "Section"], 
      Cell[
       StyleData["Text"], CellMargins -> {{27, 27}, {Inherited, Inherited}}, 
       TextJustification -> 1., ParagraphIndent -> 15, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["EnvironmentText"], ShowCellBracket -> Automatic, 
       CellMargins -> {{82, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextJustification -> 1., ParagraphSpacing -> {0.3, 0.}, 
       MenuSortingValue -> 1480, MenuCommandKey -> "T", FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Item"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "Item", 
          CounterAssignments -> {{"Subitem", 0}, {"Subsubitem", 0}}, 
          MenuSortingValue -> 1600, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Item", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemParagraph"], CellMargins -> {{81, 10}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          DefaultNewCellStyle -> "Item", MenuSortingValue -> 1600, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemParagraph", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "Subitem", 
          CounterAssignments -> {{"Subsubitem", 0}}, MenuSortingValue -> 1610,
           FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subitem", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemParagraph"], CellMargins -> {{105, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subitem", 
          MenuSortingValue -> 1610, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemParagraph", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subsubitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1620, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subsubitem", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemParagraph"], 
          CellMargins -> {{129, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1625, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemParagraph", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "ItemNumbered", 
          CounterAssignments -> {{"SubitemNumbered", 0}, {
            "SubsubitemNumbered", 0}}, MenuSortingValue -> 1630, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemNumbered", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "SubitemNumbered", 
          CounterAssignments -> {{"SubsubitemNumbered", 0}}, MenuSortingValue -> 
          1640, FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemNumbered", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], ".", 
              CounterBox["SubsubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> 
          "SubitemNumbered", CounterIncrements -> "SubsubitemNumbered", 
          MenuSortingValue -> 1645, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemNumbered", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Theorema", "Section"], 
      Cell[
       StyleData["GlobalDeclaration"], CellFrame -> {{0, 2}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 3}, {2, 2}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       InitializationCell -> True, CellFrameMargins -> 10, CellFrameColor -> 
       RGBColor[0.73, 0.65, 0.61], CellFrameLabelMargins -> {{33, 7}, {1, 1}},
        ShowAutoStyles -> False, MenuSortingValue -> 1441, MenuCommandKey -> 
       "G", FontSize -> 14], 
      Cell[
       StyleData["OpenEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {-4, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       StyleData["EnvironmentHeader"], 
       CellFrame -> {{False, False}, {False, False}}, 
       CellMargins -> {{76, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 6, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       ShowAutoStyles -> False, MenuSortingValue -> 1450, FontSize -> 14, 
       FontWeight -> "Bold", FontVariations -> {"CapsType" -> "SmallCaps"}, 
       Background -> RGBColor[0.89, 0.8, 0.69]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FormalTextInputFormula"], ShowCellBracket -> Automatic, 
          CellMargins -> {{96, 10}, {2, 2}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
          InitializationCell -> True, CellFrameLabelMargins -> 53, 
          ShowAutoStyles -> False, LimitsPositioningTokens -> {}, 
          MenuSortingValue -> 1470, MenuCommandKey -> "F", FontWeight -> 
          "Bold", FontColor -> RGBColor[0.25, 0.19, 0.2]], 
         Cell[
          StyleData["FormalTextInputFormula", "Layout"], Editable -> False], 
         Cell[
          StyleData["FormalTextInputFormula", "Presentation"], Editable -> 
          False]}, Open]], 
      Cell[
       StyleData["EndEnvironmentMarker"], 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextAlignment -> Right, MenuSortingValue -> 1480], 
      Cell[
       StyleData["CloseEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {False, True}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {Inherited, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Computation"], CellFrame -> {{1.5, 0}, {0, 0}}, 
          ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 7}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
          CellFrameMargins -> {{8, 8}, {3, 4}}, CellFrameColor -> 
          RGBColor[0.486, 0.33, 0.32], ShowAutoStyles -> False, 
          MenuSortingValue -> 1510, FontWeight -> "Bold", FontColor -> 
          RGBColor[0.23, 0.25, 0.34], Background -> 
          RGBColor[0.88, 0.88, 0.92]], 
         Cell[
          StyleData["Computation", "Layout"], Editable -> False], 
         Cell[
          StyleData["Computation", "Presentation"], Editable -> False]}, 
        Open]], 
      Cell[
       StyleData["ComputationResult"], CellFrame -> {{1.5, 0}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
       CellFrameMargins -> {{8, 8}, {4, 3}}, CellFrameColor -> 
       RGBColor[0.486, 0.33, 0.32], StyleMenuListing -> None, FontColor -> 
       RGBColor[0.23, 0.25, 0.34], Background -> RGBColor[0.88, 0.88, 0.92]], 
      
      Cell[
       StyleData["ComputationInfo"], ShowCellBracket -> False, 
       CellMargins -> {{120, 10}, {10, 2}}, StyleMenuListing -> None, 
       FontFamily -> $CellContext`Arial], 
      Cell[
       StyleData["ComputationInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{144, 10}, {10, 2}}, StyleMenuListing -> None], 
      Cell[
       StyleData["ProofInfo"], ShowCellBracket -> False, 
       CellMargins -> {{76, 10}, {6, 6}}, CellFrameLabelMargins -> 15, 
       StyleMenuListing -> None, FontFamily -> "Arial"], 
      Cell[
       StyleData["ProofInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{100, 10}, {0, 0}}, StyleMenuListing -> None], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FrameLabel"], LanguageCategory -> None, StyleMenuListing -> 
          None, FontFamily -> "Times New Roman", FontSize -> 10, FontWeight -> 
          "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
          FontVariations -> {
           "Masked" -> False, "Outline" -> False, "Shadow" -> False, 
            "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
          RGBColor[0.55, 0.44, 0.42], Background -> None], 
         Cell[
          StyleData["FrameLabel", "Printout"]]}, Open]], 
      Cell[
       StyleData["OpenArchive"], CellFrame -> {{False, False}, {False, True}},
        ShowCellBracket -> Automatic, CellMargins -> {{70, 70}, {0, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 20}, CellFrameMargins -> 3, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], CellFrameLabelMargins -> 1, 
       DefaultFormatType -> DefaultInputFormatType, ShowAutoStyles -> False, 
       MenuSortingValue -> 1210, FontWeight -> Bold, Background -> 
       RGBColor[0.73, 0.74, 0.84]], 
      Cell[
       StyleData["ArchiveInfo"], ShowCellBracket -> Automatic, 
       CellMargins -> {{70, 70}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 5, CellFrameLabelMargins -> 3, ShowCellTags -> 
       False, ShowAutoStyles -> False, MenuSortingValue -> 1220, FontWeight -> 
       "Bold", Background -> RGBColor[0.88, 0.88, 0.92]], 
      Cell[
       StyleData["CloseArchive"], 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> 
       Automatic, CellMargins -> {{0, 0}, {5, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 25}, CellFrameMargins -> 2, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], TextAlignment -> Center, 
       MenuSortingValue -> 1230, FontFamily -> "Arial Black", FontSize -> 24, 
       FontWeight -> "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       Background -> RGBColor[0.73, 0.65, 0.61]], 
      Cell[
       StyleData["IncludeArchive"], CellFrame -> {{4, 0}, {False, False}}, 
       ShowCellBracket -> Automatic, 
       CellMargins -> {{81, 100}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameColor -> RGBColor[0.55, 0.44, 0.42], 
       CellFrameLabels -> {{"\[RightSkeleton]", None}, {None, None}}, 
       CellFrameLabelMargins -> 25, MenuSortingValue -> 1445], 
      Cell[
       StyleData["ExpressionVariable"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionABF"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionMeta"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["AutoParentheses"], Editable -> False, SpanMaxSize -> 3, 
          FontOpacity -> 0.15], 
         Cell[
          StyleData["AutoParentheses", "Layout"], FontOpacity -> 0], 
         Cell[
          StyleData["AutoParentheses", "SlideShow"], FontOpacity -> 0]}, 
        Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Emphasized"], MenuCommandKey -> "E", FontSlant -> 
          "Italic"], 
         Cell[
          StyleData["Emphasized", "SlideShow"], FontSlant -> "Plain", 
          FontColor -> RGBColor[0.36, 0.54, 0.69]]}, Open]], 
      Cell[
       StyleData["GenericButton"], FontFamily -> "Bitstream Charter", 
       FontWeight -> "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
       FontVariations -> {
        "Masked" -> False, "Outline" -> False, "Shadow" -> False, 
         "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
       RGBColor[0.23, 0.25, 0.34]], 
      Cell[
       StyleData["TheoremaForm"], "TwoByteSyntaxCharacterAutoReplacement" -> 
       True, LineSpacing -> {1.25, 0, 2}, SingleLetterItalics -> True, 
       StyleMenuListing -> None, FontFamily -> "Courier"], 
      Cell[
       StyleData["ComponentActive"], FontFamily -> "Times New Roman"], 
      Cell[
       StyleData["ComponentInactive"], FontFamily -> "Times New Roman", 
       FontOpacity -> 0.4], 
      Cell[
       StyleData["ComponentEmpty"], FontFamily -> "Times New Roman", 
       FontVariations -> {"StrikeThrough" -> True}], 
      Cell[
       StyleData["TabView"], FontFamily -> "Times New Roman", FontSize -> 10], 
      Cell[
       StyleData["TabViewLabel"], FontFamily -> "Times New Roman", FontSize -> 
       10, FontWeight -> "Bold"], 
      Cell[
       StyleData["FormReference", StyleDefinitions -> StyleData["Text"]], 
       FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> "Plain",
        FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["FormReferenceHover", StyleDefinitions -> StyleData["Text"]],
        FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> 
       "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> True}, 
       FontColor -> RGBColor[0.404, 0.43, 0.545]], 
      Cell[
       StyleData["DisplayFormula"], CellMargins -> {{36, 6}, {7, 7}}, 
       ConversionRules :> {
        "HTML" -> {{"", ""}, {"<p class=\"Equation\">", "</p>"}}}, 
       CellHorizontalScrolling -> True, DefaultFormatType -> TraditionalForm, 
       InputAutoReplacements -> {}, TextAlignment -> Center, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       ButtonBoxOptions -> {Appearance -> {Automatic, None}}]}, Open]]}, 
  Visible -> False, FrontEndVersion -> 
  "9.0 for Linux x86 (64-bit) (February 7, 2013)", StyleDefinitions -> 
  "StylesheetFormatting.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "ID:232123642"->{
  Cell[777684, 17943, 721, 19, 32, "GlobalDeclaration",
   CellTags->{
    "ID:232123642", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->232123642]},
 "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"->{
  Cell[777684, 17943, 721, 19, 32, "GlobalDeclaration",
   CellTags->{
    "ID:232123642", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->232123642],
  Cell[778408, 17964, 1024, 26, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:2028940590", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "gcd GCD"},
   CellID->2028940590],
  Cell[779435, 17992, 879, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:166548655", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mod Mod"},
   CellID->166548655],
  Cell[781208, 18050, 748, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1363294779", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->1363294779],
  Cell[781959, 18071, 932, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1613423474", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->1613423474],
  Cell[782894, 18094, 818, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1525507482", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->1525507482],
  Cell[783715, 18116, 1427, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1587087117", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR \[Element]"},
   CellID->1587087117],
  Cell[785145, 18159, 855, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:447457872", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR 0"},
   CellID->447457872],
  Cell[786003, 18184, 963, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:815094494", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR 1"},
   CellID->815094494],
  Cell[786969, 18212, 988, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1939288109", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR +"},
   CellID->1939288109],
  Cell[787960, 18241, 960, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:915824852", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR -"},
   CellID->915824852],
  Cell[788923, 18270, 964, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:370600110", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR - unary"},
   CellID->370600110],
  Cell[789890, 18299, 960, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:344363438", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR \[RawStar]"},
   CellID->344363438],
  Cell[790853, 18328, 1255, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1495376438", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR \[Succeeds]"},
   CellID->1495376438],
  Cell[792111, 18370, 1094, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:550500630", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR multN"},
   CellID->550500630],
  Cell[793208, 18399, 1308, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:445166729", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR mult"},
   CellID->445166729],
  Cell[794519, 18442, 1693, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1793329966", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR mult 1"},
   CellID->1793329966],
  Cell[796215, 18490, 1174, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:126969281", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR mult 2"},
   CellID->126969281],
  Cell[797392, 18528, 2121, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:909359224", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR rdm 1"},
   CellID->909359224],
  Cell[800630, 18618, 2149, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:236078685", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR rdm 2"},
   CellID->236078685],
  Cell[803965, 18707, 2430, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:943334598", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR lcrd"},
   CellID->943334598],
  Cell[806814, 18779, 1225, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1868934214", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR lcrd unary"},
   CellID->1868934214],
  Cell[808042, 18811, 3929, 91, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:424764915", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR cpm"},
   CellID->424764915],
  Cell[813154, 18950, 893, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:580511390", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[ScriptCapitalZ]24"},
   CellID->580511390],
  Cell[834139, 19616, 895, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:275453931", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[ScriptCapitalZ]13"},
   CellID->275453931],
  Cell[856754, 20318, 715, 18, 30, "GlobalDeclaration",
   CellTags->{
    "ID:7294098", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->7294098],
  Cell[857511, 20341, 836, 20, 33, "GlobalDeclaration",
   CellTags->{
    "ID:6216142", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->6216142],
  Cell[859334, 20396, 844, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:285451873", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "modulus \[NotEqual] 0 \[ScriptCapitalZ]n"},
   CellID->285451873],
  Cell[860181, 20419, 1720, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:270643577", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Element] alternative \[ScriptCapitalZ]n"},
   
   CellID->270643577],
  Cell[861904, 20475, 1053, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:916172587", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[ScriptCapitalZ]1"},
   CellID->916172587],
  Cell[862960, 20508, 806, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:200088253", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->200088253],
  Cell[863769, 20530, 860, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:62294620", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "abs \[ScriptCapitalZ]n"},
   CellID->62294620],
  Cell[864632, 20554, 868, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:90375616", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mod identity \[ScriptCapitalZ]n"},
   CellID->90375616],
  Cell[865503, 20579, 1235, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:575444601", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "- unary alternative \[ScriptCapitalZ]n"},
   CellID->575444601],
  Cell[889054, 21274, 755, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:974601634", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->974601634],
  Cell[889812, 21295, 972, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:779775892", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
     "gcd \[Element] \[DoubleStruckCapitalN] \[ScriptCapitalZ]n"},
   CellID->779775892],
  Cell[890787, 21323, 1290, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:7430669", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership gcd \[ScriptCapitalZ]n", "IR"},
   CellID->7430669],
  Cell[892080, 21363, 931, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:243173858", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "gcd \[NotEqual] 0 \[ScriptCapitalZ]n"},
   CellID->243173858],
  Cell[893014, 21389, 1085, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:335259273", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
     "quo gcd \[Element] \[DoubleStruckCapitalN] \[ScriptCapitalZ]n"},
   CellID->335259273],
  Cell[894102, 21420, 1144, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:582853848", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership quo gcd \[ScriptCapitalZ]n"},
   CellID->582853848],
  Cell[895249, 21454, 1047, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:29867892", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership quo gcd -1 \[ScriptCapitalZ]n"},
   CellID->29867892],
  Cell[896299, 21485, 1083, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:372275792", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "quo gcd \[RawStar] \[ScriptCapitalZ]n"},
   CellID->372275792],
  Cell[897385, 21515, 721, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:825367134", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->825367134],
  Cell[898109, 21535, 989, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:55333112", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership quo \[ScriptCapitalZ]n", "IR"},
   CellID->55333112],
  Cell[899101, 21562, 1006, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:474647872", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership -mod \[ScriptCapitalZ]n", "IR"},
   CellID->474647872],
  Cell[938716, 22698, 842, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:208073319", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isFinDomain \[ScriptCapitalZ]n"},
   CellID->208073319],
  Cell[939561, 22721, 833, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:207295573", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isDomain \[ScriptCapitalZ]n"},
   CellID->207295573],
  Cell[940397, 22744, 1134, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:189562335", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isNeutral 1 \[ScriptCapitalZ]n"},
   CellID->189562335],
  Cell[941534, 22778, 1013, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:371480170", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isCommRing1 \[ScriptCapitalZ]n"},
   CellID->371480170],
  Cell[955646, 23210, 1196, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1342794225", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isPartIrreflOrder \[ScriptCapitalZ]n"},
   CellID->1342794225],
  Cell[956845, 23245, 1071, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:145791983", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isTotalIrreflOrder \[ScriptCapitalZ]n"},
   CellID->145791983],
  Cell[957919, 23278, 1073, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1051492688", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isNoetherian \[ScriptCapitalZ]n"},
   CellID->1051492688],
  Cell[969861, 23642, 2252, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:307883626", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Sum] \[ScriptCapitalZ]n"},
   CellID->307883626],
  Cell[972116, 23709, 784, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:209868567", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->209868567],
  Cell[972903, 23731, 1322, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:57125150", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 alternative \[ScriptCapitalZ]n"},
   CellID->57125150],
  Cell[974228, 23770, 1567, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:50258269", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R2' \[ScriptCapitalZ]n"},
   CellID->50258269],
  Cell[994504, 24319, 892, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1161287108", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "Multipliers \[ScriptCapitalZ]n"},
   CellID->1161287108],
  Cell[995399, 24343, 833, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:124498584", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R0 \[ScriptCapitalZ]n"},
   CellID->124498584],
  Cell[996235, 24366, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1236424988", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R1 \[ScriptCapitalZ]n"},
   CellID->1236424988],
  Cell[997077, 24389, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1558881148", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R2 \[ScriptCapitalZ]n"},
   CellID->1558881148],
  Cell[997919, 24412, 836, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1530993938", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R3 \[ScriptCapitalZ]n"},
   CellID->1530993938],
  Cell[998758, 24435, 836, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:589877437", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R4 \[ScriptCapitalZ]n"},
   CellID->589877437],
  Cell[1017695, 25028, 782, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:104617955", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->104617955],
  Cell[1018480, 25050, 1087, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:608534217", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership -quo \[ScriptCapitalZ]n", "IR"},
   CellID->608534217],
  Cell[1019570, 25082, 1087, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:217754000", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership +quo \[ScriptCapitalZ]n", "IR"},
   CellID->217754000],
  Cell[1020660, 25114, 1372, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:415524168", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult2 -quo \[ScriptCapitalZ]n"},
   CellID->415524168],
  Cell[1022035, 25151, 1273, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:51392774", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 +quo \[ScriptCapitalZ]n"},
   CellID->51392774],
  Cell[1023311, 25187, 1204, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:488904469", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[RawStar] -quo \[ScriptCapitalZ]n"},
   CellID->488904469],
  Cell[1024518, 25222, 1153, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:469346824", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[RawStar] +quo \[ScriptCapitalZ]n"},
   CellID->469346824],
  Cell[1059901, 26177, 782, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:195073923", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->195073923],
  Cell[1060686, 26199, 1318, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:18535645", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Succeeds] - \[ScriptCapitalZ]n"},
   CellID->18535645],
  Cell[1062007, 26239, 2191, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:155348029", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
     "\[Succeeds] - \[RawStar] \[ScriptCapitalZ]n"},
   CellID->155348029],
  Cell[1064201, 26308, 1384, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:220895581", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Succeeds] + \[ScriptCapitalZ]n"},
   CellID->220895581],
  Cell[1084956, 26862, 757, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:108166851", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->108166851],
  Cell[1085716, 26883, 1212, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:166132041", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
     "gcd \[VerticalBar] \[RawStar] \[ScriptCapitalZ]n"},
   CellID->166132041],
  Cell[1086931, 26920, 1980, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:414892396", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 1 \[ScriptCapitalZ]n"},
   CellID->414892396],
  Cell[1088914, 26980, 1227, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:48379652", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 2 \[ScriptCapitalZ]n"},
   CellID->48379652],
  Cell[1090144, 27016, 2105, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:287224029", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 3 \[ScriptCapitalZ]n"},
   CellID->287224029],
  Cell[1092252, 27083, 1918, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:398117055", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 4 \[ScriptCapitalZ]n"},
   CellID->398117055],
  Cell[1094173, 27144, 2066, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:568873110", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
     "mult1 \[RawStar] injective \[ScriptCapitalZ]n"},
   CellID->568873110],
  Cell[1096242, 27209, 2334, 72, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:347198673", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient \[ScriptCapitalZ]n"},
   CellID->347198673],
  Cell[1147555, 28553, 776, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:25008217", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->25008217],
  Cell[1148334, 28575, 1126, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:761196101", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isReducible \[ScriptCapitalZ]n"},
   CellID->761196101],
  Cell[1149463, 28608, 1774, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:216211003", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isReducible' \[ScriptCapitalZ]n"},
   CellID->216211003],
  Cell[1151240, 28661, 2214, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:431638777", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "gcd \[Succeeds] \[ScriptCapitalZ]n"},
   CellID->431638777],
  Cell[1174918, 29285, 755, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:219886779", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->219886779],
  Cell[1175676, 29306, 2638, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:277360974", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isCommonReducible \[ScriptCapitalZ]n"},
   CellID->277360974],
  Cell[1178317, 29381, 1584, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:112584880", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR unary \[ScriptCapitalZ]n"},
   CellID->112584880],
  Cell[1179904, 29429, 2030, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:195146313", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR identical \[ScriptCapitalZ]n"},
   CellID->195146313],
  Cell[1181937, 29488, 2011, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:6406785", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR distinct' \[ScriptCapitalZ]n"},
   CellID->6406785],
  Cell[1183951, 29546, 2356, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:87852126", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR distinct \[ScriptCapitalZ]n"},
   CellID->87852126],
  Cell[1186310, 29614, 2370, 68, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:37400195", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR \[ScriptCapitalZ]n"},
   CellID->37400195],
  Cell[1241040, 30985, 892, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:14129606", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->14129606],
  Cell[1241935, 31010, 1382, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:126313666", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 - \[ScriptCapitalZ]n"},
   CellID->126313666],
  Cell[1243320, 31051, 1331, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:888879353", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "areCorrelative 1 \[ScriptCapitalZ]n"},
   CellID->888879353],
  Cell[1244654, 31090, 1431, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:23406014", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "areCorrelative 2 \[ScriptCapitalZ]n"},
   CellID->23406014],
  Cell[1268094, 31693, 859, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:325383468", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R5 \[ScriptCapitalZ]n"},
   CellID->325383468],
  Cell[1268956, 31716, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1783775451", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R6 \[ScriptCapitalZ]n"},
   CellID->1783775451],
  Cell[1276341, 31949, 2235, 68, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:394176110", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR existence \[ScriptCapitalZ]n"},
   CellID->394176110],
  Cell[1278579, 32019, 2049, 64, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:232518088", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR unary existence \[ScriptCapitalZ]n"},
   CellID->232518088],
  Cell[1296232, 32480, 838, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2147206476", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R7 \[ScriptCapitalZ]n"},
   CellID->2147206476],
  Cell[1297073, 32503, 837, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1971443720", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R8 \[ScriptCapitalZ]n"},
   CellID->1971443720],
  Cell[1304614, 32741, 957, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:727194704", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isReductionRing \[ScriptCapitalZ]n"},
   CellID->727194704],
  Cell[1309548, 32896, 775, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:73858040", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->73858040],
  Cell[1310326, 32918, 2799, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:502941593", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm ternary \[ScriptCapitalZ]n"},
   CellID->502941593],
  Cell[1313128, 32999, 1498, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:691632805", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm aux \[ScriptCapitalZ]n"},
   CellID->691632805],
  Cell[1314629, 33046, 980, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:290458459", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership rdm \[ScriptCapitalZ]n"},
   CellID->290458459],
  Cell[1315612, 33073, 1186, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:38241486", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm cases \[ScriptCapitalZ]n"},
   CellID->38241486],
  Cell[1316801, 33107, 1103, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:140983642", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm \[RawStar] \[ScriptCapitalZ]n"},
   CellID->140983642],
  Cell[1317907, 33139, 979, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:33831774", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm 0 \[ScriptCapitalZ]n"},
   CellID->33831774],
  Cell[1356587, 34158, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1376076253", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R9 \[ScriptCapitalZ]n"},
   CellID->1376076253],
  Cell[1357429, 34181, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:125450892", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R10 \[ScriptCapitalZ]n"},
   CellID->125450892],
  Cell[1358271, 34204, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:189921319", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R11 \[ScriptCapitalZ]n"},
   CellID->189921319],
  Cell[1359113, 34227, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:667306596", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R12 \[ScriptCapitalZ]n"},
   CellID->667306596],
  Cell[1359955, 34250, 813, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:194342088", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R13 \[ScriptCapitalZ]n"},
   CellID->194342088],
  Cell[1375996, 34752, 875, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:403270264", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isAlgoReductionRing \[ScriptCapitalZ]n"},
   CellID->403270264]},
 "ID:2028940590"->{
  Cell[778408, 17964, 1024, 26, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:2028940590", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "gcd GCD"},
   CellID->2028940590]},
 "gcd GCD"->{
  Cell[778408, 17964, 1024, 26, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:2028940590", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "gcd GCD"},
   CellID->2028940590]},
 "ID:166548655"->{
  Cell[779435, 17992, 879, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:166548655", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mod Mod"},
   CellID->166548655]},
 "mod Mod"->{
  Cell[779435, 17992, 879, 24, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:166548655", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mod Mod"},
   CellID->166548655]},
 "ID:1363294779"->{
  Cell[781208, 18050, 748, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1363294779", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->1363294779]},
 "ID:1613423474"->{
  Cell[781959, 18071, 932, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1613423474", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->1613423474]},
 "ID:1525507482"->{
  Cell[782894, 18094, 818, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1525507482", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->1525507482]},
 "ID:1587087117"->{
  Cell[783715, 18116, 1427, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1587087117", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR \[Element]"},
   CellID->1587087117]},
 "RIQR \[Element]"->{
  Cell[783715, 18116, 1427, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1587087117", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR \[Element]"},
   CellID->1587087117]},
 "ID:447457872"->{
  Cell[785145, 18159, 855, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:447457872", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR 0"},
   CellID->447457872]},
 "RIQR 0"->{
  Cell[785145, 18159, 855, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:447457872", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR 0"},
   CellID->447457872]},
 "ID:815094494"->{
  Cell[786003, 18184, 963, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:815094494", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR 1"},
   CellID->815094494]},
 "RIQR 1"->{
  Cell[786003, 18184, 963, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:815094494", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR 1"},
   CellID->815094494]},
 "ID:1939288109"->{
  Cell[786969, 18212, 988, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1939288109", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR +"},
   CellID->1939288109]},
 "RIQR +"->{
  Cell[786969, 18212, 988, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1939288109", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR +"},
   CellID->1939288109]},
 "ID:915824852"->{
  Cell[787960, 18241, 960, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:915824852", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR -"},
   CellID->915824852]},
 "RIQR -"->{
  Cell[787960, 18241, 960, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:915824852", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR -"},
   CellID->915824852]},
 "ID:370600110"->{
  Cell[788923, 18270, 964, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:370600110", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR - unary"},
   CellID->370600110]},
 "RIQR - unary"->{
  Cell[788923, 18270, 964, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:370600110", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR - unary"},
   CellID->370600110]},
 "ID:344363438"->{
  Cell[789890, 18299, 960, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:344363438", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR \[RawStar]"},
   CellID->344363438]},
 "RIQR \[RawStar]"->{
  Cell[789890, 18299, 960, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:344363438", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR \[RawStar]"},
   CellID->344363438]},
 "ID:1495376438"->{
  Cell[790853, 18328, 1255, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1495376438", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR \[Succeeds]"},
   CellID->1495376438]},
 "RIQR \[Succeeds]"->{
  Cell[790853, 18328, 1255, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1495376438", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR \[Succeeds]"},
   CellID->1495376438]},
 "ID:550500630"->{
  Cell[792111, 18370, 1094, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:550500630", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR multN"},
   CellID->550500630]},
 "RIQR multN"->{
  Cell[792111, 18370, 1094, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:550500630", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR multN"},
   CellID->550500630]},
 "ID:445166729"->{
  Cell[793208, 18399, 1308, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:445166729", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR mult"},
   CellID->445166729]},
 "RIQR mult"->{
  Cell[793208, 18399, 1308, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:445166729", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR mult"},
   CellID->445166729]},
 "ID:1793329966"->{
  Cell[794519, 18442, 1693, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1793329966", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR mult 1"},
   CellID->1793329966]},
 "RIQR mult 1"->{
  Cell[794519, 18442, 1693, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1793329966", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR mult 1"},
   CellID->1793329966]},
 "ID:126969281"->{
  Cell[796215, 18490, 1174, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:126969281", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR mult 2"},
   CellID->126969281]},
 "RIQR mult 2"->{
  Cell[796215, 18490, 1174, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:126969281", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR mult 2"},
   CellID->126969281]},
 "ID:909359224"->{
  Cell[797392, 18528, 2121, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:909359224", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR rdm 1"},
   CellID->909359224]},
 "RIQR rdm 1"->{
  Cell[797392, 18528, 2121, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:909359224", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR rdm 1"},
   CellID->909359224]},
 "ID:236078685"->{
  Cell[800630, 18618, 2149, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:236078685", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR rdm 2"},
   CellID->236078685]},
 "RIQR rdm 2"->{
  Cell[800630, 18618, 2149, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:236078685", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR rdm 2"},
   CellID->236078685]},
 "ID:943334598"->{
  Cell[803965, 18707, 2430, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:943334598", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR lcrd"},
   CellID->943334598]},
 "RIQR lcrd"->{
  Cell[803965, 18707, 2430, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:943334598", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR lcrd"},
   CellID->943334598]},
 "ID:1868934214"->{
  Cell[806814, 18779, 1225, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1868934214", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR lcrd unary"},
   CellID->1868934214]},
 "RIQR lcrd unary"->{
  Cell[806814, 18779, 1225, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1868934214", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR lcrd unary"},
   CellID->1868934214]},
 "ID:424764915"->{
  Cell[808042, 18811, 3929, 91, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:424764915", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR cpm"},
   CellID->424764915]},
 "RIQR cpm"->{
  Cell[808042, 18811, 3929, 91, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:424764915", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR cpm"},
   CellID->424764915]},
 "ID:580511390"->{
  Cell[813154, 18950, 893, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:580511390", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[ScriptCapitalZ]24"},
   CellID->580511390]},
 "\[ScriptCapitalZ]24"->{
  Cell[813154, 18950, 893, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:580511390", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[ScriptCapitalZ]24"},
   CellID->580511390]},
 "ID:275453931"->{
  Cell[834139, 19616, 895, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:275453931", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[ScriptCapitalZ]13"},
   CellID->275453931]},
 "\[ScriptCapitalZ]13"->{
  Cell[834139, 19616, 895, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:275453931", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[ScriptCapitalZ]13"},
   CellID->275453931]},
 "ID:7294098"->{
  Cell[856754, 20318, 715, 18, 30, "GlobalDeclaration",
   CellTags->{
    "ID:7294098", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->7294098]},
 "ID:6216142"->{
  Cell[857511, 20341, 836, 20, 33, "GlobalDeclaration",
   CellTags->{
    "ID:6216142", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->6216142]},
 "ID:285451873"->{
  Cell[859334, 20396, 844, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:285451873", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "modulus \[NotEqual] 0 \[ScriptCapitalZ]n"},
   CellID->285451873]},
 "modulus \[NotEqual] 0 \[ScriptCapitalZ]n"->{
  Cell[859334, 20396, 844, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:285451873", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "modulus \[NotEqual] 0 \[ScriptCapitalZ]n"},
   CellID->285451873]},
 "ID:270643577"->{
  Cell[860181, 20419, 1720, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:270643577", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Element] alternative \[ScriptCapitalZ]n"},
   
   CellID->270643577]},
 "\[Element] alternative \[ScriptCapitalZ]n"->{
  Cell[860181, 20419, 1720, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:270643577", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Element] alternative \[ScriptCapitalZ]n"},
   
   CellID->270643577]},
 "ID:916172587"->{
  Cell[861904, 20475, 1053, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:916172587", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[ScriptCapitalZ]1"},
   CellID->916172587]},
 "\[ScriptCapitalZ]1"->{
  Cell[861904, 20475, 1053, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:916172587", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[ScriptCapitalZ]1"},
   CellID->916172587]},
 "ID:200088253"->{
  Cell[862960, 20508, 806, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:200088253", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->200088253]},
 "ID:62294620"->{
  Cell[863769, 20530, 860, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:62294620", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "abs \[ScriptCapitalZ]n"},
   CellID->62294620]},
 "abs \[ScriptCapitalZ]n"->{
  Cell[863769, 20530, 860, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:62294620", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "abs \[ScriptCapitalZ]n"},
   CellID->62294620]},
 "ID:90375616"->{
  Cell[864632, 20554, 868, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:90375616", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mod identity \[ScriptCapitalZ]n"},
   CellID->90375616]},
 "mod identity \[ScriptCapitalZ]n"->{
  Cell[864632, 20554, 868, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:90375616", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mod identity \[ScriptCapitalZ]n"},
   CellID->90375616]},
 "ID:575444601"->{
  Cell[865503, 20579, 1235, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:575444601", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "- unary alternative \[ScriptCapitalZ]n"},
   CellID->575444601]},
 "- unary alternative \[ScriptCapitalZ]n"->{
  Cell[865503, 20579, 1235, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:575444601", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "- unary alternative \[ScriptCapitalZ]n"},
   CellID->575444601]},
 "Proof|ID:285451873"->{
  Cell[866892, 20621, 2305, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:285451873", "Proof|ID:285451873-1"},
   CellID->127340268]},
 "Proof|ID:285451873-1"->{
  Cell[866892, 20621, 2305, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:285451873", "Proof|ID:285451873-1"},
   CellID->127340268]},
 "Proof|ID:270643577"->{
  Cell[869717, 20710, 3865, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:270643577", "Proof|ID:270643577-1"},
   CellID->506264492]},
 "Proof|ID:270643577-1"->{
  Cell[869717, 20710, 3865, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:270643577", "Proof|ID:270643577-1"},
   CellID->506264492]},
 "Proof|ID:916172587"->{
  Cell[874101, 20835, 2353, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:916172587", "Proof|ID:916172587-1"},
   CellID->265170607]},
 "Proof|ID:916172587-1"->{
  Cell[874101, 20835, 2353, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:916172587", "Proof|ID:916172587-1"},
   CellID->265170607]},
 "Proof|ID:62294620"->{
  Cell[876974, 20924, 3001, 86, 70, "ProofInfo",
   CellTags->{"Proof|ID:62294620", "Proof|ID:62294620-1"},
   CellID->319229321]},
 "Proof|ID:62294620-1"->{
  Cell[876974, 20924, 3001, 86, 70, "ProofInfo",
   CellTags->{"Proof|ID:62294620", "Proof|ID:62294620-1"},
   CellID->319229321]},
 "Proof|ID:90375616"->{
  Cell[880494, 21029, 3107, 89, 70, "ProofInfo",
   CellTags->{"Proof|ID:90375616", "Proof|ID:90375616-1"},
   CellID->190812059]},
 "Proof|ID:90375616-1"->{
  Cell[880494, 21029, 3107, 89, 70, "ProofInfo",
   CellTags->{"Proof|ID:90375616", "Proof|ID:90375616-1"},
   CellID->190812059]},
 "Proof|ID:575444601"->{
  Cell[884120, 21137, 3972, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:575444601", "Proof|ID:575444601-1"},
   CellID->94536918]},
 "Proof|ID:575444601-1"->{
  Cell[884120, 21137, 3972, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:575444601", "Proof|ID:575444601-1"},
   CellID->94536918]},
 "ID:974601634"->{
  Cell[889054, 21274, 755, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:974601634", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->974601634]},
 "ID:779775892"->{
  Cell[889812, 21295, 972, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:779775892", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
     "gcd \[Element] \[DoubleStruckCapitalN] \[ScriptCapitalZ]n"},
   CellID->779775892]},
 "gcd \[Element] \[DoubleStruckCapitalN] \[ScriptCapitalZ]n"->{
  Cell[889812, 21295, 972, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:779775892", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
     "gcd \[Element] \[DoubleStruckCapitalN] \[ScriptCapitalZ]n"},
   CellID->779775892]},
 "ID:7430669"->{
  Cell[890787, 21323, 1290, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:7430669", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership gcd \[ScriptCapitalZ]n", "IR"},
   CellID->7430669]},
 "membership gcd \[ScriptCapitalZ]n"->{
  Cell[890787, 21323, 1290, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:7430669", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership gcd \[ScriptCapitalZ]n", "IR"},
   CellID->7430669]},
 "IR"->{
  Cell[890787, 21323, 1290, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:7430669", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership gcd \[ScriptCapitalZ]n", "IR"},
   CellID->7430669],
  Cell[898109, 21535, 989, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:55333112", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership quo \[ScriptCapitalZ]n", "IR"},
   CellID->55333112],
  Cell[899101, 21562, 1006, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:474647872", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership -mod \[ScriptCapitalZ]n", "IR"},
   CellID->474647872],
  Cell[1018480, 25050, 1087, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:608534217", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership -quo \[ScriptCapitalZ]n", "IR"},
   CellID->608534217],
  Cell[1019570, 25082, 1087, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:217754000", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership +quo \[ScriptCapitalZ]n", "IR"},
   CellID->217754000]},
 "ID:243173858"->{
  Cell[892080, 21363, 931, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:243173858", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "gcd \[NotEqual] 0 \[ScriptCapitalZ]n"},
   CellID->243173858]},
 "gcd \[NotEqual] 0 \[ScriptCapitalZ]n"->{
  Cell[892080, 21363, 931, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:243173858", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "gcd \[NotEqual] 0 \[ScriptCapitalZ]n"},
   CellID->243173858]},
 "ID:335259273"->{
  Cell[893014, 21389, 1085, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:335259273", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
     "quo gcd \[Element] \[DoubleStruckCapitalN] \[ScriptCapitalZ]n"},
   CellID->335259273]},
 "quo gcd \[Element] \[DoubleStruckCapitalN] \[ScriptCapitalZ]n"->{
  Cell[893014, 21389, 1085, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:335259273", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
     "quo gcd \[Element] \[DoubleStruckCapitalN] \[ScriptCapitalZ]n"},
   CellID->335259273]},
 "ID:582853848"->{
  Cell[894102, 21420, 1144, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:582853848", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership quo gcd \[ScriptCapitalZ]n"},
   CellID->582853848]},
 "membership quo gcd \[ScriptCapitalZ]n"->{
  Cell[894102, 21420, 1144, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:582853848", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership quo gcd \[ScriptCapitalZ]n"},
   CellID->582853848]},
 "ID:29867892"->{
  Cell[895249, 21454, 1047, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:29867892", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership quo gcd -1 \[ScriptCapitalZ]n"},
   CellID->29867892]},
 "membership quo gcd -1 \[ScriptCapitalZ]n"->{
  Cell[895249, 21454, 1047, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:29867892", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership quo gcd -1 \[ScriptCapitalZ]n"},
   CellID->29867892]},
 "ID:372275792"->{
  Cell[896299, 21485, 1083, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:372275792", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "quo gcd \[RawStar] \[ScriptCapitalZ]n"},
   CellID->372275792]},
 "quo gcd \[RawStar] \[ScriptCapitalZ]n"->{
  Cell[896299, 21485, 1083, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:372275792", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "quo gcd \[RawStar] \[ScriptCapitalZ]n"},
   CellID->372275792]},
 "ID:825367134"->{
  Cell[897385, 21515, 721, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:825367134", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->825367134]},
 "ID:55333112"->{
  Cell[898109, 21535, 989, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:55333112", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership quo \[ScriptCapitalZ]n", "IR"},
   CellID->55333112]},
 "membership quo \[ScriptCapitalZ]n"->{
  Cell[898109, 21535, 989, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:55333112", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership quo \[ScriptCapitalZ]n", "IR"},
   CellID->55333112]},
 "ID:474647872"->{
  Cell[899101, 21562, 1006, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:474647872", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership -mod \[ScriptCapitalZ]n", "IR"},
   CellID->474647872]},
 "membership -mod \[ScriptCapitalZ]n"->{
  Cell[899101, 21562, 1006, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:474647872", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership -mod \[ScriptCapitalZ]n", "IR"},
   CellID->474647872]},
 "Proof|ID:779775892"->{
  Cell[900263, 21600, 3120, 88, 70, "ProofInfo",
   CellTags->{"Proof|ID:779775892", "Proof|ID:779775892-1"},
   CellID->297664998]},
 "Proof|ID:779775892-1"->{
  Cell[900263, 21600, 3120, 88, 70, "ProofInfo",
   CellTags->{"Proof|ID:779775892", "Proof|ID:779775892-1"},
   CellID->297664998]},
 "Proof|ID:7430669"->{
  Cell[903902, 21707, 3845, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:7430669", "Proof|ID:7430669-1"},
   CellID->340529413]},
 "Proof|ID:7430669-1"->{
  Cell[903902, 21707, 3845, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:7430669", "Proof|ID:7430669-1"},
   CellID->340529413]},
 "Proof|ID:243173858"->{
  Cell[908265, 21830, 3299, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:243173858", "Proof|ID:243173858-1"},
   CellID->623193018]},
 "Proof|ID:243173858-1"->{
  Cell[908265, 21830, 3299, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:243173858", "Proof|ID:243173858-1"},
   CellID->623193018]},
 "Proof|ID:335259273"->{
  Cell[912083, 21941, 3426, 96, 70, "ProofInfo",
   CellTags->{"Proof|ID:335259273", "Proof|ID:335259273-1"},
   CellID->301510995]},
 "Proof|ID:335259273-1"->{
  Cell[912083, 21941, 3426, 96, 70, "ProofInfo",
   CellTags->{"Proof|ID:335259273", "Proof|ID:335259273-1"},
   CellID->301510995]},
 "Proof|ID:582853848"->{
  Cell[916029, 22056, 4164, 112, 70, "ProofInfo",
   CellTags->{"Proof|ID:582853848", "Proof|ID:582853848-1"},
   CellID->62215823]},
 "Proof|ID:582853848-1"->{
  Cell[916029, 22056, 4164, 112, 70, "ProofInfo",
   CellTags->{"Proof|ID:582853848", "Proof|ID:582853848-1"},
   CellID->62215823]},
 "Proof|ID:29867892"->{
  Cell[920712, 22187, 3670, 100, 70, "ProofInfo",
   CellTags->{"Proof|ID:29867892", "Proof|ID:29867892-1"},
   CellID->753487385]},
 "Proof|ID:29867892-1"->{
  Cell[920712, 22187, 3670, 100, 70, "ProofInfo",
   CellTags->{"Proof|ID:29867892", "Proof|ID:29867892-1"},
   CellID->753487385]},
 "Proof|ID:372275792"->{
  Cell[924901, 22306, 3991, 107, 70, "ProofInfo",
   CellTags->{"Proof|ID:372275792", "Proof|ID:372275792-1"},
   CellID->9434432]},
 "Proof|ID:372275792-1"->{
  Cell[924901, 22306, 3991, 107, 70, "ProofInfo",
   CellTags->{"Proof|ID:372275792", "Proof|ID:372275792-1"},
   CellID->9434432]},
 "Proof|ID:55333112"->{
  Cell[929412, 22432, 3561, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:55333112", "Proof|ID:55333112-1"},
   CellID->64245290]},
 "Proof|ID:55333112-1"->{
  Cell[929412, 22432, 3561, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:55333112", "Proof|ID:55333112-1"},
   CellID->64245290]},
 "Proof|ID:474647872"->{
  Cell[933492, 22549, 3733, 101, 70, "ProofInfo",
   CellTags->{"Proof|ID:474647872", "Proof|ID:474647872-1"},
   CellID->209531985]},
 "Proof|ID:474647872-1"->{
  Cell[933492, 22549, 3733, 101, 70, "ProofInfo",
   CellTags->{"Proof|ID:474647872", "Proof|ID:474647872-1"},
   CellID->209531985]},
 "ID:208073319"->{
  Cell[938716, 22698, 842, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:208073319", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isFinDomain \[ScriptCapitalZ]n"},
   CellID->208073319]},
 "isFinDomain \[ScriptCapitalZ]n"->{
  Cell[938716, 22698, 842, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:208073319", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isFinDomain \[ScriptCapitalZ]n"},
   CellID->208073319]},
 "ID:207295573"->{
  Cell[939561, 22721, 833, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:207295573", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isDomain \[ScriptCapitalZ]n"},
   CellID->207295573]},
 "isDomain \[ScriptCapitalZ]n"->{
  Cell[939561, 22721, 833, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:207295573", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isDomain \[ScriptCapitalZ]n"},
   CellID->207295573]},
 "ID:189562335"->{
  Cell[940397, 22744, 1134, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:189562335", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isNeutral 1 \[ScriptCapitalZ]n"},
   CellID->189562335]},
 "isNeutral 1 \[ScriptCapitalZ]n"->{
  Cell[940397, 22744, 1134, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:189562335", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isNeutral 1 \[ScriptCapitalZ]n"},
   CellID->189562335]},
 "ID:371480170"->{
  Cell[941534, 22778, 1013, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:371480170", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isCommRing1 \[ScriptCapitalZ]n"},
   CellID->371480170]},
 "isCommRing1 \[ScriptCapitalZ]n"->{
  Cell[941534, 22778, 1013, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:371480170", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isCommRing1 \[ScriptCapitalZ]n"},
   CellID->371480170]},
 "Proof|ID:208073319"->{
  Cell[942704, 22813, 2387, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:208073319", "Proof|ID:208073319-1"},
   CellID->169378186]},
 "Proof|ID:208073319-1"->{
  Cell[942704, 22813, 2387, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:208073319", "Proof|ID:208073319-1"},
   CellID->169378186]},
 "Proof|ID:207295573"->{
  Cell[945611, 22903, 2367, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:207295573", "Proof|ID:207295573-1"},
   CellID->415181399]},
 "Proof|ID:207295573-1"->{
  Cell[945611, 22903, 2367, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:207295573", "Proof|ID:207295573-1"},
   CellID->415181399]},
 "Proof|ID:189562335"->{
  Cell[948497, 22992, 3014, 85, 70, "ProofInfo",
   CellTags->{"Proof|ID:189562335", "Proof|ID:189562335-1"},
   CellID->43006419]},
 "Proof|ID:189562335-1"->{
  Cell[948497, 22992, 3014, 85, 70, "ProofInfo",
   CellTags->{"Proof|ID:189562335", "Proof|ID:189562335-1"},
   CellID->43006419]},
 "Proof|ID:371480170"->{
  Cell[952031, 23096, 2387, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:371480170", "Proof|ID:371480170-1"},
   CellID->344386318]},
 "Proof|ID:371480170-1"->{
  Cell[952031, 23096, 2387, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:371480170", "Proof|ID:371480170-1"},
   CellID->344386318]},
 "ID:1342794225"->{
  Cell[955646, 23210, 1196, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1342794225", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isPartIrreflOrder \[ScriptCapitalZ]n"},
   CellID->1342794225]},
 "isPartIrreflOrder \[ScriptCapitalZ]n"->{
  Cell[955646, 23210, 1196, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1342794225", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isPartIrreflOrder \[ScriptCapitalZ]n"},
   CellID->1342794225]},
 "ID:145791983"->{
  Cell[956845, 23245, 1071, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:145791983", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isTotalIrreflOrder \[ScriptCapitalZ]n"},
   CellID->145791983]},
 "isTotalIrreflOrder \[ScriptCapitalZ]n"->{
  Cell[956845, 23245, 1071, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:145791983", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isTotalIrreflOrder \[ScriptCapitalZ]n"},
   CellID->145791983]},
 "ID:1051492688"->{
  Cell[957919, 23278, 1073, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1051492688", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isNoetherian \[ScriptCapitalZ]n"},
   CellID->1051492688]},
 "isNoetherian \[ScriptCapitalZ]n"->{
  Cell[957919, 23278, 1073, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1051492688", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isNoetherian \[ScriptCapitalZ]n"},
   CellID->1051492688]},
 "Proof|ID:1342794225"->{
  Cell[959146, 23320, 2790, 81, 70, "ProofInfo",
   CellTags->{"Proof|ID:1342794225", "Proof|ID:1342794225-1"},
   CellID->256415475]},
 "Proof|ID:1342794225-1"->{
  Cell[959146, 23320, 2790, 81, 70, "ProofInfo",
   CellTags->{"Proof|ID:1342794225", "Proof|ID:1342794225-1"},
   CellID->256415475]},
 "Proof|ID:145791983"->{
  Cell[962455, 23420, 2785, 80, 70, "ProofInfo",
   CellTags->{"Proof|ID:145791983", "Proof|ID:145791983-1"},
   CellID->187860367]},
 "Proof|ID:145791983-1"->{
  Cell[962455, 23420, 2785, 80, 70, "ProofInfo",
   CellTags->{"Proof|ID:145791983", "Proof|ID:145791983-1"},
   CellID->187860367]},
 "Proof|ID:1051492688"->{
  Cell[965760, 23519, 2761, 80, 70, "ProofInfo",
   CellTags->{"Proof|ID:1051492688", "Proof|ID:1051492688-1"},
   CellID->66415286]},
 "Proof|ID:1051492688-1"->{
  Cell[965760, 23519, 2761, 80, 70, "ProofInfo",
   CellTags->{"Proof|ID:1051492688", "Proof|ID:1051492688-1"},
   CellID->66415286]},
 "ID:307883626"->{
  Cell[969861, 23642, 2252, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:307883626", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Sum] \[ScriptCapitalZ]n"},
   CellID->307883626]},
 "\[Sum] \[ScriptCapitalZ]n"->{
  Cell[969861, 23642, 2252, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:307883626", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Sum] \[ScriptCapitalZ]n"},
   CellID->307883626]},
 "ID:209868567"->{
  Cell[972116, 23709, 784, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:209868567", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->209868567]},
 "ID:57125150"->{
  Cell[972903, 23731, 1322, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:57125150", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 alternative \[ScriptCapitalZ]n"},
   CellID->57125150]},
 "mult1 alternative \[ScriptCapitalZ]n"->{
  Cell[972903, 23731, 1322, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:57125150", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 alternative \[ScriptCapitalZ]n"},
   CellID->57125150]},
 "ID:50258269"->{
  Cell[974228, 23770, 1567, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:50258269", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R2' \[ScriptCapitalZ]n"},
   CellID->50258269]},
 "R2' \[ScriptCapitalZ]n"->{
  Cell[974228, 23770, 1567, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:50258269", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R2' \[ScriptCapitalZ]n"},
   CellID->50258269]},
 "Proof|ID:307883626"->{
  Cell[975951, 23824, 6345, 160, 70, "ProofInfo",
   CellTags->{"Proof|ID:307883626", "Proof|ID:307883626-1"},
   CellID->253583856]},
 "Proof|ID:307883626-1"->{
  Cell[975951, 23824, 6345, 160, 70, "ProofInfo",
   CellTags->{"Proof|ID:307883626", "Proof|ID:307883626-1"},
   CellID->253583856]},
 "Proof|ID:57125150"->{
  Cell[982816, 24003, 4916, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:57125150", "Proof|ID:57125150-1"},
   CellID->163486749]},
 "Proof|ID:57125150-1"->{
  Cell[982816, 24003, 4916, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:57125150", "Proof|ID:57125150-1"},
   CellID->163486749]},
 "Proof|ID:50258269"->{
  Cell[988250, 24150, 5262, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:50258269", "Proof|ID:50258269-1"},
   CellID->223632667]},
 "Proof|ID:50258269-1"->{
  Cell[988250, 24150, 5262, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:50258269", "Proof|ID:50258269-1"},
   CellID->223632667]},
 "ID:1161287108"->{
  Cell[994504, 24319, 892, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1161287108", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "Multipliers \[ScriptCapitalZ]n"},
   CellID->1161287108]},
 "Multipliers \[ScriptCapitalZ]n"->{
  Cell[994504, 24319, 892, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1161287108", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "Multipliers \[ScriptCapitalZ]n"},
   CellID->1161287108]},
 "ID:124498584"->{
  Cell[995399, 24343, 833, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:124498584", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R0 \[ScriptCapitalZ]n"},
   CellID->124498584]},
 "R0 \[ScriptCapitalZ]n"->{
  Cell[995399, 24343, 833, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:124498584", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R0 \[ScriptCapitalZ]n"},
   CellID->124498584]},
 "ID:1236424988"->{
  Cell[996235, 24366, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1236424988", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R1 \[ScriptCapitalZ]n"},
   CellID->1236424988]},
 "R1 \[ScriptCapitalZ]n"->{
  Cell[996235, 24366, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1236424988", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R1 \[ScriptCapitalZ]n"},
   CellID->1236424988]},
 "ID:1558881148"->{
  Cell[997077, 24389, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1558881148", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R2 \[ScriptCapitalZ]n"},
   CellID->1558881148]},
 "R2 \[ScriptCapitalZ]n"->{
  Cell[997077, 24389, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1558881148", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R2 \[ScriptCapitalZ]n"},
   CellID->1558881148]},
 "ID:1530993938"->{
  Cell[997919, 24412, 836, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1530993938", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R3 \[ScriptCapitalZ]n"},
   CellID->1530993938]},
 "R3 \[ScriptCapitalZ]n"->{
  Cell[997919, 24412, 836, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1530993938", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R3 \[ScriptCapitalZ]n"},
   CellID->1530993938]},
 "ID:589877437"->{
  Cell[998758, 24435, 836, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:589877437", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R4 \[ScriptCapitalZ]n"},
   CellID->589877437]},
 "R4 \[ScriptCapitalZ]n"->{
  Cell[998758, 24435, 836, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:589877437", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R4 \[ScriptCapitalZ]n"},
   CellID->589877437]},
 "Proof|ID:1161287108"->{
  Cell[999749, 24467, 2396, 72, 70, "ProofInfo",
   CellTags->{"Proof|ID:1161287108", "Proof|ID:1161287108-1"},
   CellID->311987723]},
 "Proof|ID:1161287108-1"->{
  Cell[999749, 24467, 2396, 72, 70, "ProofInfo",
   CellTags->{"Proof|ID:1161287108", "Proof|ID:1161287108-1"},
   CellID->311987723]},
 "Proof|ID:124498584"->{
  Cell[1002666, 24558, 2342, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:124498584", "Proof|ID:124498584-1"},
   CellID->24095429]},
 "Proof|ID:124498584-1"->{
  Cell[1002666, 24558, 2342, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:124498584", "Proof|ID:124498584-1"},
   CellID->24095429]},
 "Proof|ID:1236424988"->{
  Cell[1005528, 24647, 2352, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:1236424988", "Proof|ID:1236424988-1"},
   CellID->296864625]},
 "Proof|ID:1236424988-1"->{
  Cell[1005528, 24647, 2352, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:1236424988", "Proof|ID:1236424988-1"},
   CellID->296864625]},
 "Proof|ID:1558881148"->{
  Cell[1008400, 24737, 2352, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:1558881148", "Proof|ID:1558881148-1"},
   CellID->239424613]},
 "Proof|ID:1558881148-1"->{
  Cell[1008400, 24737, 2352, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:1558881148", "Proof|ID:1558881148-1"},
   CellID->239424613]},
 "Proof|ID:1530993938"->{
  Cell[1011273, 24827, 2352, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:1530993938", "Proof|ID:1530993938-1"},
   CellID->344362146]},
 "Proof|ID:1530993938-1"->{
  Cell[1011273, 24827, 2352, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:1530993938", "Proof|ID:1530993938-1"},
   CellID->344362146]},
 "Proof|ID:589877437"->{
  Cell[1014145, 24917, 2343, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:589877437", "Proof|ID:589877437-1"},
   CellID->352205366]},
 "Proof|ID:589877437-1"->{
  Cell[1014145, 24917, 2343, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:589877437", "Proof|ID:589877437-1"},
   CellID->352205366]},
 "ID:104617955"->{
  Cell[1017695, 25028, 782, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:104617955", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->104617955]},
 "ID:608534217"->{
  Cell[1018480, 25050, 1087, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:608534217", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership -quo \[ScriptCapitalZ]n", "IR"},
   CellID->608534217]},
 "membership -quo \[ScriptCapitalZ]n"->{
  Cell[1018480, 25050, 1087, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:608534217", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership -quo \[ScriptCapitalZ]n", "IR"},
   CellID->608534217]},
 "ID:217754000"->{
  Cell[1019570, 25082, 1087, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:217754000", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership +quo \[ScriptCapitalZ]n", "IR"},
   CellID->217754000]},
 "membership +quo \[ScriptCapitalZ]n"->{
  Cell[1019570, 25082, 1087, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:217754000", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership +quo \[ScriptCapitalZ]n", "IR"},
   CellID->217754000]},
 "ID:415524168"->{
  Cell[1020660, 25114, 1372, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:415524168", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult2 -quo \[ScriptCapitalZ]n"},
   CellID->415524168]},
 "mult2 -quo \[ScriptCapitalZ]n"->{
  Cell[1020660, 25114, 1372, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:415524168", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult2 -quo \[ScriptCapitalZ]n"},
   CellID->415524168]},
 "ID:51392774"->{
  Cell[1022035, 25151, 1273, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:51392774", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 +quo \[ScriptCapitalZ]n"},
   CellID->51392774]},
 "mult1 +quo \[ScriptCapitalZ]n"->{
  Cell[1022035, 25151, 1273, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:51392774", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 +quo \[ScriptCapitalZ]n"},
   CellID->51392774]},
 "ID:488904469"->{
  Cell[1023311, 25187, 1204, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:488904469", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[RawStar] -quo \[ScriptCapitalZ]n"},
   CellID->488904469]},
 "\[RawStar] -quo \[ScriptCapitalZ]n"->{
  Cell[1023311, 25187, 1204, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:488904469", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[RawStar] -quo \[ScriptCapitalZ]n"},
   CellID->488904469]},
 "ID:469346824"->{
  Cell[1024518, 25222, 1153, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:469346824", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[RawStar] +quo \[ScriptCapitalZ]n"},
   CellID->469346824]},
 "\[RawStar] +quo \[ScriptCapitalZ]n"->{
  Cell[1024518, 25222, 1153, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:469346824", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[RawStar] +quo \[ScriptCapitalZ]n"},
   CellID->469346824]},
 "Proof|ID:608534217"->{
  Cell[1025827, 25265, 4501, 117, 70, "ProofInfo",
   CellTags->{"Proof|ID:608534217", "Proof|ID:608534217-1"},
   CellID->171963183]},
 "Proof|ID:608534217-1"->{
  Cell[1025827, 25265, 4501, 117, 70, "ProofInfo",
   CellTags->{"Proof|ID:608534217", "Proof|ID:608534217-1"},
   CellID->171963183]},
 "Proof|ID:217754000"->{
  Cell[1030892, 25402, 4501, 117, 70, "ProofInfo",
   CellTags->{"Proof|ID:217754000", "Proof|ID:217754000-1"},
   CellID->218735392]},
 "Proof|ID:217754000-1"->{
  Cell[1030892, 25402, 4501, 117, 70, "ProofInfo",
   CellTags->{"Proof|ID:217754000", "Proof|ID:217754000-1"},
   CellID->218735392]},
 "Proof|ID:415524168"->{
  Cell[1035913, 25538, 5509, 140, 70, "ProofInfo",
   CellTags->{"Proof|ID:415524168", "Proof|ID:415524168-1"},
   CellID->322621179]},
 "Proof|ID:415524168-1"->{
  Cell[1035913, 25538, 5509, 140, 70, "ProofInfo",
   CellTags->{"Proof|ID:415524168", "Proof|ID:415524168-1"},
   CellID->322621179]},
 "Proof|ID:51392774"->{
  Cell[1042006, 25698, 5504, 140, 70, "ProofInfo",
   CellTags->{"Proof|ID:51392774", "Proof|ID:51392774-1"},
   CellID->336233968]},
 "Proof|ID:51392774-1"->{
  Cell[1042006, 25698, 5504, 140, 70, "ProofInfo",
   CellTags->{"Proof|ID:51392774", "Proof|ID:51392774-1"},
   CellID->336233968]},
 "Proof|ID:488904469"->{
  Cell[1048027, 25857, 5202, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:488904469", "Proof|ID:488904469-1"},
   CellID->318992087]},
 "Proof|ID:488904469-1"->{
  Cell[1048027, 25857, 5202, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:488904469", "Proof|ID:488904469-1"},
   CellID->318992087]},
 "Proof|ID:469346824"->{
  Cell[1053749, 26010, 5202, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:469346824", "Proof|ID:469346824-1"},
   CellID->231394775]},
 "Proof|ID:469346824-1"->{
  Cell[1053749, 26010, 5202, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:469346824", "Proof|ID:469346824-1"},
   CellID->231394775]},
 "ID:195073923"->{
  Cell[1059901, 26177, 782, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:195073923", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->195073923]},
 "ID:18535645"->{
  Cell[1060686, 26199, 1318, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:18535645", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Succeeds] - \[ScriptCapitalZ]n"},
   CellID->18535645]},
 "\[Succeeds] - \[ScriptCapitalZ]n"->{
  Cell[1060686, 26199, 1318, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:18535645", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Succeeds] - \[ScriptCapitalZ]n"},
   CellID->18535645]},
 "ID:155348029"->{
  Cell[1062007, 26239, 2191, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:155348029", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
     "\[Succeeds] - \[RawStar] \[ScriptCapitalZ]n"},
   CellID->155348029]},
 "\[Succeeds] - \[RawStar] \[ScriptCapitalZ]n"->{
  Cell[1062007, 26239, 2191, 67, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:155348029", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
     "\[Succeeds] - \[RawStar] \[ScriptCapitalZ]n"},
   CellID->155348029]},
 "ID:220895581"->{
  Cell[1064201, 26308, 1384, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:220895581", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Succeeds] + \[ScriptCapitalZ]n"},
   CellID->220895581]},
 "\[Succeeds] + \[ScriptCapitalZ]n"->{
  Cell[1064201, 26308, 1384, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:220895581", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Succeeds] + \[ScriptCapitalZ]n"},
   CellID->220895581]},
 "Proof|ID:18535645"->{
  Cell[1065739, 26357, 4870, 127, 70, "ProofInfo",
   CellTags->{"Proof|ID:18535645", "Proof|ID:18535645-1"},
   CellID->255392256]},
 "Proof|ID:18535645-1"->{
  Cell[1065739, 26357, 4870, 127, 70, "ProofInfo",
   CellTags->{"Proof|ID:18535645", "Proof|ID:18535645-1"},
   CellID->255392256]},
 "Proof|ID:155348029"->{
  Cell[1071128, 26503, 7643, 183, 70, "ProofInfo",
   CellTags->{"Proof|ID:155348029", "Proof|ID:155348029-1"},
   CellID->101870138]},
 "Proof|ID:155348029-1"->{
  Cell[1071128, 26503, 7643, 183, 70, "ProofInfo",
   CellTags->{"Proof|ID:155348029", "Proof|ID:155348029-1"},
   CellID->101870138]},
 "Proof|ID:220895581"->{
  Cell[1079291, 26705, 4665, 123, 70, "ProofInfo",
   CellTags->{"Proof|ID:220895581", "Proof|ID:220895581-1"},
   CellID->11546098]},
 "Proof|ID:220895581-1"->{
  Cell[1079291, 26705, 4665, 123, 70, "ProofInfo",
   CellTags->{"Proof|ID:220895581", "Proof|ID:220895581-1"},
   CellID->11546098]},
 "ID:108166851"->{
  Cell[1084956, 26862, 757, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:108166851", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->108166851]},
 "ID:166132041"->{
  Cell[1085716, 26883, 1212, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:166132041", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
     "gcd \[VerticalBar] \[RawStar] \[ScriptCapitalZ]n"},
   CellID->166132041]},
 "gcd \[VerticalBar] \[RawStar] \[ScriptCapitalZ]n"->{
  Cell[1085716, 26883, 1212, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:166132041", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
     "gcd \[VerticalBar] \[RawStar] \[ScriptCapitalZ]n"},
   CellID->166132041]},
 "ID:414892396"->{
  Cell[1086931, 26920, 1980, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:414892396", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 1 \[ScriptCapitalZ]n"},
   CellID->414892396]},
 "mult1 sufficient 1 \[ScriptCapitalZ]n"->{
  Cell[1086931, 26920, 1980, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:414892396", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 1 \[ScriptCapitalZ]n"},
   CellID->414892396]},
 "ID:48379652"->{
  Cell[1088914, 26980, 1227, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:48379652", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 2 \[ScriptCapitalZ]n"},
   CellID->48379652]},
 "mult1 sufficient 2 \[ScriptCapitalZ]n"->{
  Cell[1088914, 26980, 1227, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:48379652", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 2 \[ScriptCapitalZ]n"},
   CellID->48379652]},
 "ID:287224029"->{
  Cell[1090144, 27016, 2105, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:287224029", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 3 \[ScriptCapitalZ]n"},
   CellID->287224029]},
 "mult1 sufficient 3 \[ScriptCapitalZ]n"->{
  Cell[1090144, 27016, 2105, 65, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:287224029", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 3 \[ScriptCapitalZ]n"},
   CellID->287224029]},
 "ID:398117055"->{
  Cell[1092252, 27083, 1918, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:398117055", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 4 \[ScriptCapitalZ]n"},
   CellID->398117055]},
 "mult1 sufficient 4 \[ScriptCapitalZ]n"->{
  Cell[1092252, 27083, 1918, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:398117055", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 4 \[ScriptCapitalZ]n"},
   CellID->398117055]},
 "ID:568873110"->{
  Cell[1094173, 27144, 2066, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:568873110", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
     "mult1 \[RawStar] injective \[ScriptCapitalZ]n"},
   CellID->568873110]},
 "mult1 \[RawStar] injective \[ScriptCapitalZ]n"->{
  Cell[1094173, 27144, 2066, 63, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:568873110", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
     "mult1 \[RawStar] injective \[ScriptCapitalZ]n"},
   CellID->568873110]},
 "ID:347198673"->{
  Cell[1096242, 27209, 2334, 72, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:347198673", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient \[ScriptCapitalZ]n"},
   CellID->347198673]},
 "mult1 sufficient \[ScriptCapitalZ]n"->{
  Cell[1096242, 27209, 2334, 72, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:347198673", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient \[ScriptCapitalZ]n"},
   CellID->347198673]},
 "Proof|ID:166132041"->{
  Cell[1098731, 27292, 4070, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:166132041", "Proof|ID:166132041-1"},
   CellID->111518977]},
 "Proof|ID:166132041-1"->{
  Cell[1098731, 27292, 4070, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:166132041", "Proof|ID:166132041-1"},
   CellID->111518977]},
 "Proof|ID:414892396"->{
  Cell[1103321, 27421, 7509, 182, 70, "ProofInfo",
   CellTags->{"Proof|ID:414892396", "Proof|ID:414892396-1"},
   CellID->924193601]},
 "Proof|ID:414892396-1"->{
  Cell[1103321, 27421, 7509, 182, 70, "ProofInfo",
   CellTags->{"Proof|ID:414892396", "Proof|ID:414892396-1"},
   CellID->924193601]},
 "Proof|ID:48379652"->{
  Cell[1111350, 27622, 4796, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:48379652", "Proof|ID:48379652-1"},
   CellID->384976109]},
 "Proof|ID:48379652-1"->{
  Cell[1111350, 27622, 4796, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:48379652", "Proof|ID:48379652-1"},
   CellID->384976109]},
 "Proof|ID:287224029"->{
  Cell[1116665, 27765, 6459, 160, 70, "ProofInfo",
   CellTags->{"Proof|ID:287224029", "Proof|ID:287224029-1"},
   CellID->144370346]},
 "Proof|ID:287224029-1"->{
  Cell[1116665, 27765, 6459, 160, 70, "ProofInfo",
   CellTags->{"Proof|ID:287224029", "Proof|ID:287224029-1"},
   CellID->144370346]},
 "Proof|ID:398117055"->{
  Cell[1123644, 27944, 6702, 170, 70, "ProofInfo",
   CellTags->{"Proof|ID:398117055", "Proof|ID:398117055-1"},
   CellID->168110018]},
 "Proof|ID:398117055-1"->{
  Cell[1123644, 27944, 6702, 170, 70, "ProofInfo",
   CellTags->{"Proof|ID:398117055", "Proof|ID:398117055-1"},
   CellID->168110018]},
 "Proof|ID:568873110"->{
  Cell[1130866, 28133, 7651, 184, 70, "ProofInfo",
   CellTags->{"Proof|ID:568873110", "Proof|ID:568873110-1"},
   CellID->3033029]},
 "Proof|ID:568873110-1"->{
  Cell[1130866, 28133, 7651, 184, 70, "ProofInfo",
   CellTags->{"Proof|ID:568873110", "Proof|ID:568873110-1"},
   CellID->3033029]},
 "Proof|ID:347198673"->{
  Cell[1139037, 28336, 7549, 184, 70, "ProofInfo",
   CellTags->{"Proof|ID:347198673", "Proof|ID:347198673-1"},
   CellID->636035778]},
 "Proof|ID:347198673-1"->{
  Cell[1139037, 28336, 7549, 184, 70, "ProofInfo",
   CellTags->{"Proof|ID:347198673", "Proof|ID:347198673-1"},
   CellID->636035778]},
 "ID:25008217"->{
  Cell[1147555, 28553, 776, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:25008217", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->25008217]},
 "ID:761196101"->{
  Cell[1148334, 28575, 1126, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:761196101", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isReducible \[ScriptCapitalZ]n"},
   CellID->761196101]},
 "isReducible \[ScriptCapitalZ]n"->{
  Cell[1148334, 28575, 1126, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:761196101", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isReducible \[ScriptCapitalZ]n"},
   CellID->761196101]},
 "ID:216211003"->{
  Cell[1149463, 28608, 1774, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:216211003", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isReducible' \[ScriptCapitalZ]n"},
   CellID->216211003]},
 "isReducible' \[ScriptCapitalZ]n"->{
  Cell[1149463, 28608, 1774, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:216211003", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isReducible' \[ScriptCapitalZ]n"},
   CellID->216211003]},
 "ID:431638777"->{
  Cell[1151240, 28661, 2214, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:431638777", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "gcd \[Succeeds] \[ScriptCapitalZ]n"},
   CellID->431638777]},
 "gcd \[Succeeds] \[ScriptCapitalZ]n"->{
  Cell[1151240, 28661, 2214, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:431638777", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "gcd \[Succeeds] \[ScriptCapitalZ]n"},
   CellID->431638777]},
 "Proof|ID:761196101"->{
  Cell[1153609, 28738, 4382, 115, 70, "ProofInfo",
   CellTags->{"Proof|ID:761196101", "Proof|ID:761196101-1"},
   CellID->46321884]},
 "Proof|ID:761196101-1"->{
  Cell[1153609, 28738, 4382, 115, 70, "ProofInfo",
   CellTags->{"Proof|ID:761196101", "Proof|ID:761196101-1"},
   CellID->46321884]},
 "Proof|ID:216211003"->{
  Cell[1158511, 28872, 7302, 176, 70, "ProofInfo",
   CellTags->{"Proof|ID:216211003", "Proof|ID:216211003-1"},
   CellID->182622923]},
 "Proof|ID:216211003-1"->{
  Cell[1158511, 28872, 7302, 176, 70, "ProofInfo",
   CellTags->{"Proof|ID:216211003", "Proof|ID:216211003-1"},
   CellID->182622923]},
 "Proof|ID:431638777"->{
  Cell[1166329, 29067, 7589, 183, 70, "ProofInfo",
   CellTags->{"Proof|ID:431638777", "Proof|ID:431638777-1"},
   CellID->607642207]},
 "Proof|ID:431638777-1"->{
  Cell[1166329, 29067, 7589, 183, 70, "ProofInfo",
   CellTags->{"Proof|ID:431638777", "Proof|ID:431638777-1"},
   CellID->607642207]},
 "ID:219886779"->{
  Cell[1174918, 29285, 755, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:219886779", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->219886779]},
 "ID:277360974"->{
  Cell[1175676, 29306, 2638, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:277360974", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isCommonReducible \[ScriptCapitalZ]n"},
   CellID->277360974]},
 "isCommonReducible \[ScriptCapitalZ]n"->{
  Cell[1175676, 29306, 2638, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:277360974", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isCommonReducible \[ScriptCapitalZ]n"},
   CellID->277360974]},
 "ID:112584880"->{
  Cell[1178317, 29381, 1584, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:112584880", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR unary \[ScriptCapitalZ]n"},
   CellID->112584880]},
 "isMNTCR unary \[ScriptCapitalZ]n"->{
  Cell[1178317, 29381, 1584, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:112584880", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR unary \[ScriptCapitalZ]n"},
   CellID->112584880]},
 "ID:195146313"->{
  Cell[1179904, 29429, 2030, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:195146313", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR identical \[ScriptCapitalZ]n"},
   CellID->195146313]},
 "isMNTCR identical \[ScriptCapitalZ]n"->{
  Cell[1179904, 29429, 2030, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:195146313", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR identical \[ScriptCapitalZ]n"},
   CellID->195146313]},
 "ID:6406785"->{
  Cell[1181937, 29488, 2011, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:6406785", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR distinct' \[ScriptCapitalZ]n"},
   CellID->6406785]},
 "isMNTCR distinct' \[ScriptCapitalZ]n"->{
  Cell[1181937, 29488, 2011, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:6406785", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR distinct' \[ScriptCapitalZ]n"},
   CellID->6406785]},
 "ID:87852126"->{
  Cell[1183951, 29546, 2356, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:87852126", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR distinct \[ScriptCapitalZ]n"},
   CellID->87852126]},
 "isMNTCR distinct \[ScriptCapitalZ]n"->{
  Cell[1183951, 29546, 2356, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:87852126", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR distinct \[ScriptCapitalZ]n"},
   CellID->87852126]},
 "ID:37400195"->{
  Cell[1186310, 29614, 2370, 68, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:37400195", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR \[ScriptCapitalZ]n"},
   CellID->37400195]},
 "isMNTCR \[ScriptCapitalZ]n"->{
  Cell[1186310, 29614, 2370, 68, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:37400195", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR \[ScriptCapitalZ]n"},
   CellID->37400195]},
 "Proof|ID:277360974"->{
  Cell[1188835, 29693, 8376, 198, 70, "ProofInfo",
   CellTags->{"Proof|ID:277360974", "Proof|ID:277360974-1"},
   CellID->41026893]},
 "Proof|ID:277360974-1"->{
  Cell[1188835, 29693, 8376, 198, 70, "ProofInfo",
   CellTags->{"Proof|ID:277360974", "Proof|ID:277360974-1"},
   CellID->41026893]},
 "Proof|ID:112584880"->{
  Cell[1197731, 29910, 4953, 127, 70, "ProofInfo",
   CellTags->{"Proof|ID:112584880", "Proof|ID:112584880-1"},
   CellID->228914149]},
 "Proof|ID:112584880-1"->{
  Cell[1197731, 29910, 4953, 127, 70, "ProofInfo",
   CellTags->{"Proof|ID:112584880", "Proof|ID:112584880-1"},
   CellID->228914149]},
 "Proof|ID:195146313"->{
  Cell[1203270, 30057, 7323, 178, 70, "ProofInfo",
   CellTags->{"Proof|ID:195146313", "Proof|ID:195146313-1"},
   CellID->270250321]},
 "Proof|ID:195146313-1"->{
  Cell[1203270, 30057, 7323, 178, 70, "ProofInfo",
   CellTags->{"Proof|ID:195146313", "Proof|ID:195146313-1"},
   CellID->270250321]},
 "Proof|ID:6406785"->{
  Cell[1211157, 30255, 8299, 200, 70, "ProofInfo",
   CellTags->{"Proof|ID:6406785", "Proof|ID:6406785-1"},
   CellID->295234282]},
 "Proof|ID:6406785-1"->{
  Cell[1211157, 30255, 8299, 200, 70, "ProofInfo",
   CellTags->{"Proof|ID:6406785", "Proof|ID:6406785-1"},
   CellID->295234282]},
 "Proof|ID:87852126"->{
  Cell[1219973, 30474, 9482, 222, 70, "ProofInfo",
   CellTags->{"Proof|ID:87852126", "Proof|ID:87852126-1"},
   CellID->306702955]},
 "Proof|ID:87852126-1"->{
  Cell[1219973, 30474, 9482, 222, 70, "ProofInfo",
   CellTags->{"Proof|ID:87852126", "Proof|ID:87852126-1"},
   CellID->306702955]},
 "Proof|ID:37400195"->{
  Cell[1229974, 30715, 9832, 228, 70, "ProofInfo",
   CellTags->{"Proof|ID:37400195", "Proof|ID:37400195-1"},
   CellID->282288311]},
 "Proof|ID:37400195-1"->{
  Cell[1229974, 30715, 9832, 228, 70, "ProofInfo",
   CellTags->{"Proof|ID:37400195", "Proof|ID:37400195-1"},
   CellID->282288311]},
 "ID:14129606"->{
  Cell[1241040, 30985, 892, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:14129606", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->14129606]},
 "ID:126313666"->{
  Cell[1241935, 31010, 1382, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:126313666", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 - \[ScriptCapitalZ]n"},
   CellID->126313666]},
 "mult1 - \[ScriptCapitalZ]n"->{
  Cell[1241935, 31010, 1382, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:126313666", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 - \[ScriptCapitalZ]n"},
   CellID->126313666]},
 "ID:888879353"->{
  Cell[1243320, 31051, 1331, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:888879353", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "areCorrelative 1 \[ScriptCapitalZ]n"},
   CellID->888879353]},
 "areCorrelative 1 \[ScriptCapitalZ]n"->{
  Cell[1243320, 31051, 1331, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:888879353", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "areCorrelative 1 \[ScriptCapitalZ]n"},
   CellID->888879353]},
 "ID:23406014"->{
  Cell[1244654, 31090, 1431, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:23406014", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "areCorrelative 2 \[ScriptCapitalZ]n"},
   CellID->23406014]},
 "areCorrelative 2 \[ScriptCapitalZ]n"->{
  Cell[1244654, 31090, 1431, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:23406014", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "areCorrelative 2 \[ScriptCapitalZ]n"},
   CellID->23406014]},
 "Proof|ID:126313666"->{
  Cell[1246240, 31139, 6986, 170, 70, "ProofInfo",
   CellTags->{"Proof|ID:126313666", "Proof|ID:126313666-1"},
   CellID->79969318]},
 "Proof|ID:126313666-1"->{
  Cell[1246240, 31139, 6986, 170, 70, "ProofInfo",
   CellTags->{"Proof|ID:126313666", "Proof|ID:126313666-1"},
   CellID->79969318]},
 "Proof|ID:888879353"->{
  Cell[1253746, 31328, 6394, 156, 70, "ProofInfo",
   CellTags->{"Proof|ID:888879353", "Proof|ID:888879353-1"},
   CellID->341588]},
 "Proof|ID:888879353-1"->{
  Cell[1253746, 31328, 6394, 156, 70, "ProofInfo",
   CellTags->{"Proof|ID:888879353", "Proof|ID:888879353-1"},
   CellID->341588]},
 "Proof|ID:23406014"->{
  Cell[1260660, 31503, 6392, 156, 70, "ProofInfo",
   CellTags->{"Proof|ID:23406014", "Proof|ID:23406014-1"},
   CellID->386167612]},
 "Proof|ID:23406014-1"->{
  Cell[1260660, 31503, 6392, 156, 70, "ProofInfo",
   CellTags->{"Proof|ID:23406014", "Proof|ID:23406014-1"},
   CellID->386167612]},
 "ID:325383468"->{
  Cell[1268094, 31693, 859, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:325383468", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R5 \[ScriptCapitalZ]n"},
   CellID->325383468]},
 "R5 \[ScriptCapitalZ]n"->{
  Cell[1268094, 31693, 859, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:325383468", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R5 \[ScriptCapitalZ]n"},
   CellID->325383468]},
 "ID:1783775451"->{
  Cell[1268956, 31716, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1783775451", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R6 \[ScriptCapitalZ]n"},
   CellID->1783775451]},
 "R6 \[ScriptCapitalZ]n"->{
  Cell[1268956, 31716, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1783775451", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R6 \[ScriptCapitalZ]n"},
   CellID->1783775451]},
 "Proof|ID:325383468"->{
  Cell[1269951, 31748, 2341, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:325383468", "Proof|ID:325383468-1"},
   CellID->4554534]},
 "Proof|ID:325383468-1"->{
  Cell[1269951, 31748, 2341, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:325383468", "Proof|ID:325383468-1"},
   CellID->4554534]},
 "Proof|ID:1783775451"->{
  Cell[1272810, 31837, 2352, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:1783775451", "Proof|ID:1783775451-1"},
   CellID->190836234]},
 "Proof|ID:1783775451-1"->{
  Cell[1272810, 31837, 2352, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:1783775451", "Proof|ID:1783775451-1"},
   CellID->190836234]},
 "ID:394176110"->{
  Cell[1276341, 31949, 2235, 68, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:394176110", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR existence \[ScriptCapitalZ]n"},
   CellID->394176110]},
 "isMNTCR existence \[ScriptCapitalZ]n"->{
  Cell[1276341, 31949, 2235, 68, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:394176110", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR existence \[ScriptCapitalZ]n"},
   CellID->394176110]},
 "ID:232518088"->{
  Cell[1278579, 32019, 2049, 64, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:232518088", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR unary existence \[ScriptCapitalZ]n"},
   CellID->232518088]},
 "isMNTCR unary existence \[ScriptCapitalZ]n"->{
  Cell[1278579, 32019, 2049, 64, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:232518088", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR unary existence \[ScriptCapitalZ]n"},
   CellID->232518088]},
 "Proof|ID:394176110"->{
  Cell[1280784, 32094, 7689, 180, 70, "ProofInfo",
   CellTags->{"Proof|ID:394176110", "Proof|ID:394176110-1"},
   CellID->648217336]},
 "Proof|ID:394176110-1"->{
  Cell[1280784, 32094, 7689, 180, 70, "ProofInfo",
   CellTags->{"Proof|ID:394176110", "Proof|ID:394176110-1"},
   CellID->648217336]},
 "Proof|ID:232518088"->{
  Cell[1288993, 32293, 6269, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:232518088", "Proof|ID:232518088-1"},
   CellID->424300880]},
 "Proof|ID:232518088-1"->{
  Cell[1288993, 32293, 6269, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:232518088", "Proof|ID:232518088-1"},
   CellID->424300880]},
 "ID:2147206476"->{
  Cell[1296232, 32480, 838, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2147206476", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R7 \[ScriptCapitalZ]n"},
   CellID->2147206476]},
 "R7 \[ScriptCapitalZ]n"->{
  Cell[1296232, 32480, 838, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2147206476", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R7 \[ScriptCapitalZ]n"},
   CellID->2147206476]},
 "ID:1971443720"->{
  Cell[1297073, 32503, 837, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1971443720", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R8 \[ScriptCapitalZ]n"},
   CellID->1971443720]},
 "R8 \[ScriptCapitalZ]n"->{
  Cell[1297073, 32503, 837, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1971443720", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R8 \[ScriptCapitalZ]n"},
   CellID->1971443720]},
 "Proof|ID:2147206476"->{
  Cell[1298065, 32535, 2351, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:2147206476", "Proof|ID:2147206476-1"},
   CellID->31414230]},
 "Proof|ID:2147206476-1"->{
  Cell[1298065, 32535, 2351, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:2147206476", "Proof|ID:2147206476-1"},
   CellID->31414230]},
 "Proof|ID:1971443720"->{
  Cell[1300937, 32625, 2352, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:1971443720", "Proof|ID:1971443720-1"},
   CellID->239044020]},
 "Proof|ID:1971443720-1"->{
  Cell[1300937, 32625, 2352, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:1971443720", "Proof|ID:1971443720-1"},
   CellID->239044020]},
 "ID:727194704"->{
  Cell[1304614, 32741, 957, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:727194704", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isReductionRing \[ScriptCapitalZ]n"},
   CellID->727194704]},
 "isReductionRing \[ScriptCapitalZ]n"->{
  Cell[1304614, 32741, 957, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:727194704", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isReductionRing \[ScriptCapitalZ]n"},
   CellID->727194704]},
 "Proof|ID:727194704"->{
  Cell[1305729, 32776, 2403, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:727194704", "Proof|ID:727194704-1"},
   CellID->278604702]},
 "Proof|ID:727194704-1"->{
  Cell[1305729, 32776, 2403, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:727194704", "Proof|ID:727194704-1"},
   CellID->278604702]},
 "ID:73858040"->{
  Cell[1309548, 32896, 775, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:73858040", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
   CellID->73858040]},
 "ID:502941593"->{
  Cell[1310326, 32918, 2799, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:502941593", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm ternary \[ScriptCapitalZ]n"},
   CellID->502941593]},
 "rdm ternary \[ScriptCapitalZ]n"->{
  Cell[1310326, 32918, 2799, 79, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:502941593", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm ternary \[ScriptCapitalZ]n"},
   CellID->502941593]},
 "ID:691632805"->{
  Cell[1313128, 32999, 1498, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:691632805", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm aux \[ScriptCapitalZ]n"},
   CellID->691632805]},
 "rdm aux \[ScriptCapitalZ]n"->{
  Cell[1313128, 32999, 1498, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:691632805", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm aux \[ScriptCapitalZ]n"},
   CellID->691632805]},
 "ID:290458459"->{
  Cell[1314629, 33046, 980, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:290458459", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership rdm \[ScriptCapitalZ]n"},
   CellID->290458459]},
 "membership rdm \[ScriptCapitalZ]n"->{
  Cell[1314629, 33046, 980, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:290458459", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership rdm \[ScriptCapitalZ]n"},
   CellID->290458459]},
 "ID:38241486"->{
  Cell[1315612, 33073, 1186, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:38241486", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm cases \[ScriptCapitalZ]n"},
   CellID->38241486]},
 "rdm cases \[ScriptCapitalZ]n"->{
  Cell[1315612, 33073, 1186, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:38241486", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm cases \[ScriptCapitalZ]n"},
   CellID->38241486]},
 "ID:140983642"->{
  Cell[1316801, 33107, 1103, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:140983642", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm \[RawStar] \[ScriptCapitalZ]n"},
   CellID->140983642]},
 "rdm \[RawStar] \[ScriptCapitalZ]n"->{
  Cell[1316801, 33107, 1103, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:140983642", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm \[RawStar] \[ScriptCapitalZ]n"},
   CellID->140983642]},
 "ID:33831774"->{
  Cell[1317907, 33139, 979, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:33831774", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm 0 \[ScriptCapitalZ]n"},
   CellID->33831774]},
 "rdm 0 \[ScriptCapitalZ]n"->{
  Cell[1317907, 33139, 979, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:33831774", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm 0 \[ScriptCapitalZ]n"},
   CellID->33831774]},
 "Proof|ID:502941593"->{
  Cell[1319041, 33175, 9489, 223, 70, "ProofInfo",
   CellTags->{"Proof|ID:502941593", "Proof|ID:502941593-1"},
   CellID->32304572]},
 "Proof|ID:502941593-1"->{
  Cell[1319041, 33175, 9489, 223, 70, "ProofInfo",
   CellTags->{"Proof|ID:502941593", "Proof|ID:502941593-1"},
   CellID->32304572]},
 "Proof|ID:691632805"->{
  Cell[1329050, 33417, 6293, 157, 70, "ProofInfo",
   CellTags->{"Proof|ID:691632805", "Proof|ID:691632805-1"},
   CellID->23114620]},
 "Proof|ID:691632805-1"->{
  Cell[1329050, 33417, 6293, 157, 70, "ProofInfo",
   CellTags->{"Proof|ID:691632805", "Proof|ID:691632805-1"},
   CellID->23114620]},
 "Proof|ID:290458459"->{
  Cell[1335862, 33593, 4053, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:290458459", "Proof|ID:290458459-1"},
   CellID->155701127]},
 "Proof|ID:290458459-1"->{
  Cell[1335862, 33593, 4053, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:290458459", "Proof|ID:290458459-1"},
   CellID->155701127]},
 "Proof|ID:38241486"->{
  Cell[1340526, 33722, 5330, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:38241486", "Proof|ID:38241486-1"},
   CellID->904484803]},
 "Proof|ID:38241486-1"->{
  Cell[1340526, 33722, 5330, 135, 70, "ProofInfo",
   CellTags->{"Proof|ID:38241486", "Proof|ID:38241486-1"},
   CellID->904484803]},
 "Proof|ID:140983642"->{
  Cell[1346373, 33876, 5001, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:140983642", "Proof|ID:140983642-1"},
   CellID->492763287]},
 "Proof|ID:140983642-1"->{
  Cell[1346373, 33876, 5001, 128, 70, "ProofInfo",
   CellTags->{"Proof|ID:140983642", "Proof|ID:140983642-1"},
   CellID->492763287]},
 "Proof|ID:33831774"->{
  Cell[1351894, 34023, 3728, 102, 70, "ProofInfo",
   CellTags->{"Proof|ID:33831774", "Proof|ID:33831774-1"},
   CellID->447394967]},
 "Proof|ID:33831774-1"->{
  Cell[1351894, 34023, 3728, 102, 70, "ProofInfo",
   CellTags->{"Proof|ID:33831774", "Proof|ID:33831774-1"},
   CellID->447394967]},
 "ID:1376076253"->{
  Cell[1356587, 34158, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1376076253", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R9 \[ScriptCapitalZ]n"},
   CellID->1376076253]},
 "R9 \[ScriptCapitalZ]n"->{
  Cell[1356587, 34158, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1376076253", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R9 \[ScriptCapitalZ]n"},
   CellID->1376076253]},
 "ID:125450892"->{
  Cell[1357429, 34181, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:125450892", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R10 \[ScriptCapitalZ]n"},
   CellID->125450892]},
 "R10 \[ScriptCapitalZ]n"->{
  Cell[1357429, 34181, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:125450892", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R10 \[ScriptCapitalZ]n"},
   CellID->125450892]},
 "ID:189921319"->{
  Cell[1358271, 34204, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:189921319", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R11 \[ScriptCapitalZ]n"},
   CellID->189921319]},
 "R11 \[ScriptCapitalZ]n"->{
  Cell[1358271, 34204, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:189921319", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R11 \[ScriptCapitalZ]n"},
   CellID->189921319]},
 "ID:667306596"->{
  Cell[1359113, 34227, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:667306596", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R12 \[ScriptCapitalZ]n"},
   CellID->667306596]},
 "R12 \[ScriptCapitalZ]n"->{
  Cell[1359113, 34227, 839, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:667306596", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R12 \[ScriptCapitalZ]n"},
   CellID->667306596]},
 "ID:194342088"->{
  Cell[1359955, 34250, 813, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:194342088", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R13 \[ScriptCapitalZ]n"},
   CellID->194342088]},
 "R13 \[ScriptCapitalZ]n"->{
  Cell[1359955, 34250, 813, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:194342088", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R13 \[ScriptCapitalZ]n"},
   CellID->194342088]},
 "Proof|ID:1376076253"->{
  Cell[1360923, 34281, 2352, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:1376076253", "Proof|ID:1376076253-1"},
   CellID->793462086]},
 "Proof|ID:1376076253-1"->{
  Cell[1360923, 34281, 2352, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:1376076253", "Proof|ID:1376076253-1"},
   CellID->793462086]},
 "Proof|ID:125450892"->{
  Cell[1363796, 34371, 2347, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:125450892", "Proof|ID:125450892-1"},
   CellID->207385051]},
 "Proof|ID:125450892-1"->{
  Cell[1363796, 34371, 2347, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:125450892", "Proof|ID:125450892-1"},
   CellID->207385051]},
 "Proof|ID:189921319"->{
  Cell[1366663, 34460, 2347, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:189921319", "Proof|ID:189921319-1"},
   CellID->221524119]},
 "Proof|ID:189921319-1"->{
  Cell[1366663, 34460, 2347, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:189921319", "Proof|ID:189921319-1"},
   CellID->221524119]},
 "Proof|ID:667306596"->{
  Cell[1369530, 34549, 2347, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:667306596", "Proof|ID:667306596-1"},
   CellID->389476407]},
 "Proof|ID:667306596-1"->{
  Cell[1369530, 34549, 2347, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:667306596", "Proof|ID:667306596-1"},
   CellID->389476407]},
 "Proof|ID:194342088"->{
  Cell[1372396, 34638, 2347, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:194342088", "Proof|ID:194342088-1"},
   CellID->102872355]},
 "Proof|ID:194342088-1"->{
  Cell[1372396, 34638, 2347, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:194342088", "Proof|ID:194342088-1"},
   CellID->102872355]},
 "ID:403270264"->{
  Cell[1375996, 34752, 875, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:403270264", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isAlgoReductionRing \[ScriptCapitalZ]n"},
   CellID->403270264]},
 "isAlgoReductionRing \[ScriptCapitalZ]n"->{
  Cell[1375996, 34752, 875, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:403270264", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isAlgoReductionRing \[ScriptCapitalZ]n"},
   CellID->403270264]},
 "Proof|ID:403270264"->{
  Cell[1377026, 34785, 2426, 72, 28, "ProofInfo",
   CellTags->{"Proof|ID:403270264", "Proof|ID:403270264-1"},
   CellID->78837715]},
 "Proof|ID:403270264-1"->{
  Cell[1377026, 34785, 2426, 72, 28, "ProofInfo",
   CellTags->{"Proof|ID:403270264", "Proof|ID:403270264-1"},
   CellID->78837715]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"ID:232123642", 1417611, 35642},
 {"Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 1417920, 35650},
 {"ID:2028940590", 1446419, 36353},
 {"gcd GCD", 1446666, 36360},
 {"ID:166548655", 1446918, 36367},
 {"mod Mod", 1447162, 36374},
 {"ID:1363294779", 1447412, 36381},
 {"ID:1613423474", 1447648, 36388},
 {"ID:1525507482", 1447884, 36395},
 {"ID:1587087117", 1448120, 36402},
 {"RIQR \[Element]", 1448383, 36409},
 {"ID:447457872", 1448643, 36416},
 {"RIQR 0", 1448885, 36423},
 {"ID:815094494", 1449133, 36430},
 {"RIQR 1", 1449375, 36437},
 {"ID:1939288109", 1449624, 36444},
 {"RIQR +", 1449868, 36451},
 {"ID:915824852", 1450118, 36458},
 {"RIQR -", 1450360, 36465},
 {"ID:370600110", 1450608, 36472},
 {"RIQR - unary", 1450862, 36479},
 {"ID:344363438", 1451116, 36486},
 {"RIQR \[RawStar]", 1451376, 36493},
 {"ID:1495376438", 1451634, 36500},
 {"RIQR \[Succeeds]", 1451899, 36507},
 {"ID:550500630", 1452160, 36514},
 {"RIQR multN", 1452411, 36521},
 {"ID:445166729", 1452664, 36528},
 {"RIQR mult", 1452913, 36535},
 {"ID:1793329966", 1453166, 36542},
 {"RIQR mult 1", 1453421, 36549},
 {"ID:126969281", 1453677, 36556},
 {"RIQR mult 2", 1453930, 36563},
 {"ID:909359224", 1454184, 36570},
 {"RIQR rdm 1", 1454435, 36577},
 {"ID:236078685", 1454688, 36584},
 {"RIQR rdm 2", 1454939, 36591},
 {"ID:943334598", 1455192, 36598},
 {"RIQR lcrd", 1455441, 36605},
 {"ID:1868934214", 1455694, 36612},
 {"RIQR lcrd unary", 1455957, 36619},
 {"ID:424764915", 1456217, 36626},
 {"RIQR cpm", 1456464, 36633},
 {"ID:580511390", 1456715, 36640},
 {"\[ScriptCapitalZ]24", 1456983, 36647},
 {"ID:275453931", 1457244, 36654},
 {"\[ScriptCapitalZ]13", 1457512, 36661},
 {"ID:7294098", 1457771, 36668},
 {"ID:6216142", 1457998, 36675},
 {"ID:285451873", 1458227, 36682},
 {"modulus \[NotEqual] 0 \[ScriptCapitalZ]n", 1458537, 36689},
 {"ID:270643577", 1458819, 36696},
 {"\[Element] alternative \[ScriptCapitalZ]n", 1459136, 36704},
 {"ID:916172587", 1459424, 36712},
 {"\[ScriptCapitalZ]1", 1459691, 36719},
 {"ID:200088253", 1459952, 36726},
 {"ID:62294620", 1460184, 36733},
 {"abs \[ScriptCapitalZ]n", 1460456, 36740},
 {"ID:90375616", 1460717, 36747},
 {"mod identity \[ScriptCapitalZ]n", 1461007, 36754},
 {"ID:575444601", 1461278, 36761},
 {"- unary alternative \[ScriptCapitalZ]n", 1461585, 36768},
 {"Proof|ID:285451873", 1461872, 36775},
 {"Proof|ID:285451873-1", 1462033, 36779},
 {"Proof|ID:270643577", 1462192, 36783},
 {"Proof|ID:270643577-1", 1462354, 36787},
 {"Proof|ID:916172587", 1462514, 36791},
 {"Proof|ID:916172587-1", 1462675, 36795},
 {"Proof|ID:62294620", 1462833, 36799},
 {"Proof|ID:62294620-1", 1462991, 36803},
 {"Proof|ID:90375616", 1463147, 36807},
 {"Proof|ID:90375616-1", 1463305, 36811},
 {"Proof|ID:575444601", 1463462, 36815},
 {"Proof|ID:575444601-1", 1463623, 36819},
 {"ID:974601634", 1463776, 36823},
 {"ID:779775892", 1464009, 36830},
 {"gcd \[Element] \[DoubleStruckCapitalN] \[ScriptCapitalZ]n", 1464359, 36838},
 {"ID:7430669", 1464662, 36846},
 {"membership gcd \[ScriptCapitalZ]n", 1464961, 36853},
 {"IR", 1465229, 36860},
 {"ID:243173858", 1466557, 36891},
 {"gcd \[NotEqual] 0 \[ScriptCapitalZ]n", 1466859, 36898},
 {"ID:335259273", 1467137, 36905},
 {"quo gcd \[Element] \[DoubleStruckCapitalN] \[ScriptCapitalZ]n", 1467496, \
36913},
 {"ID:582853848", 1467806, 36921},
 {"membership quo gcd \[ScriptCapitalZ]n", 1468111, 36928},
 {"ID:29867892", 1468390, 36935},
 {"membership quo gcd -1 \[ScriptCapitalZ]n", 1468699, 36942},
 {"ID:372275792", 1468980, 36949},
 {"quo gcd \[RawStar] \[ScriptCapitalZ]n", 1469285, 36956},
 {"ID:825367134", 1469565, 36963},
 {"ID:55333112", 1469797, 36970},
 {"membership quo \[ScriptCapitalZ]n", 1470097, 36977},
 {"ID:474647872", 1470376, 36984},
 {"membership -mod \[ScriptCapitalZ]n", 1470681, 36991},
 {"Proof|ID:779775892", 1470970, 36998},
 {"Proof|ID:779775892-1", 1471131, 37002},
 {"Proof|ID:7430669", 1471288, 37006},
 {"Proof|ID:7430669-1", 1471444, 37010},
 {"Proof|ID:243173858", 1471600, 37014},
 {"Proof|ID:243173858-1", 1471761, 37018},
 {"Proof|ID:335259273", 1471920, 37022},
 {"Proof|ID:335259273-1", 1472081, 37026},
 {"Proof|ID:582853848", 1472240, 37030},
 {"Proof|ID:582853848-1", 1472401, 37034},
 {"Proof|ID:29867892", 1472559, 37038},
 {"Proof|ID:29867892-1", 1472718, 37042},
 {"Proof|ID:372275792", 1472876, 37046},
 {"Proof|ID:372275792-1", 1473036, 37050},
 {"Proof|ID:55333112", 1473193, 37054},
 {"Proof|ID:55333112-1", 1473350, 37058},
 {"Proof|ID:474647872", 1473506, 37062},
 {"Proof|ID:474647872-1", 1473668, 37066},
 {"ID:208073319", 1473822, 37070},
 {"isFinDomain \[ScriptCapitalZ]n", 1474112, 37077},
 {"ID:207295573", 1474384, 37084},
 {"isDomain \[ScriptCapitalZ]n", 1474668, 37091},
 {"ID:189562335", 1474937, 37098},
 {"isNeutral 1 \[ScriptCapitalZ]n", 1475228, 37105},
 {"ID:371480170", 1475501, 37112},
 {"isCommRing1 \[ScriptCapitalZ]n", 1475792, 37119},
 {"Proof|ID:208073319", 1476071, 37126},
 {"Proof|ID:208073319-1", 1476232, 37130},
 {"Proof|ID:207295573", 1476391, 37134},
 {"Proof|ID:207295573-1", 1476552, 37138},
 {"Proof|ID:189562335", 1476711, 37142},
 {"Proof|ID:189562335-1", 1476871, 37146},
 {"Proof|ID:371480170", 1477029, 37150},
 {"Proof|ID:371480170-1", 1477190, 37154},
 {"ID:1342794225", 1477344, 37158},
 {"isPartIrreflOrder \[ScriptCapitalZ]n", 1477649, 37165},
 {"ID:145791983", 1477930, 37172},
 {"isTotalIrreflOrder \[ScriptCapitalZ]n", 1478235, 37179},
 {"ID:1051492688", 1478516, 37186},
 {"isNoetherian \[ScriptCapitalZ]n", 1478811, 37193},
 {"Proof|ID:1342794225", 1479094, 37200},
 {"Proof|ID:1342794225-1", 1479258, 37204},
 {"Proof|ID:145791983", 1479419, 37208},
 {"Proof|ID:145791983-1", 1479580, 37212},
 {"Proof|ID:1051492688", 1479740, 37216},
 {"Proof|ID:1051492688-1", 1479903, 37220},
 {"ID:307883626", 1480057, 37224},
 {"\[Sum] \[ScriptCapitalZ]n", 1480338, 37231},
 {"ID:209868567", 1480606, 37238},
 {"ID:57125150", 1480838, 37245},
 {"mult1 alternative \[ScriptCapitalZ]n", 1481139, 37252},
 {"ID:50258269", 1481415, 37259},
 {"R2' \[ScriptCapitalZ]n", 1481688, 37266},
 {"Proof|ID:307883626", 1481957, 37273},
 {"Proof|ID:307883626-1", 1482119, 37277},
 {"Proof|ID:57125150", 1482278, 37281},
 {"Proof|ID:57125150-1", 1482437, 37285},
 {"Proof|ID:50258269", 1482594, 37289},
 {"Proof|ID:50258269-1", 1482753, 37293},
 {"ID:1161287108", 1482906, 37297},
 {"Multipliers \[ScriptCapitalZ]n", 1483198, 37304},
 {"ID:124498584", 1483472, 37311},
 {"R0 \[ScriptCapitalZ]n", 1483744, 37318},
 {"ID:1236424988", 1484008, 37325},
 {"R1 \[ScriptCapitalZ]n", 1484282, 37332},
 {"ID:1558881148", 1484548, 37339},
 {"R2 \[ScriptCapitalZ]n", 1484822, 37346},
 {"ID:1530993938", 1485088, 37353},
 {"R3 \[ScriptCapitalZ]n", 1485362, 37360},
 {"ID:589877437", 1485627, 37367},
 {"R4 \[ScriptCapitalZ]n", 1485899, 37374},
 {"Proof|ID:1161287108", 1486169, 37381},
 {"Proof|ID:1161287108-1", 1486333, 37385},
 {"Proof|ID:124498584", 1486494, 37389},
 {"Proof|ID:124498584-1", 1486655, 37393},
 {"Proof|ID:1236424988", 1486815, 37397},
 {"Proof|ID:1236424988-1", 1486980, 37401},
 {"Proof|ID:1558881148", 1487143, 37405},
 {"Proof|ID:1558881148-1", 1487308, 37409},
 {"Proof|ID:1530993938", 1487471, 37413},
 {"Proof|ID:1530993938-1", 1487636, 37417},
 {"Proof|ID:589877437", 1487798, 37421},
 {"Proof|ID:589877437-1", 1487960, 37425},
 {"ID:104617955", 1488114, 37429},
 {"ID:608534217", 1488348, 37436},
 {"membership -quo \[ScriptCapitalZ]n", 1488654, 37443},
 {"ID:217754000", 1488938, 37450},
 {"membership +quo \[ScriptCapitalZ]n", 1489244, 37457},
 {"ID:415524168", 1489528, 37464},
 {"mult2 -quo \[ScriptCapitalZ]n", 1489818, 37471},
 {"ID:51392774", 1490090, 37478},
 {"mult1 +quo \[ScriptCapitalZ]n", 1490378, 37485},
 {"ID:488904469", 1490649, 37492},
 {"\[RawStar] -quo \[ScriptCapitalZ]n", 1490949, 37499},
 {"ID:469346824", 1491227, 37506},
 {"\[RawStar] +quo \[ScriptCapitalZ]n", 1491527, 37513},
 {"Proof|ID:608534217", 1491811, 37520},
 {"Proof|ID:608534217-1", 1491974, 37524},
 {"Proof|ID:217754000", 1492135, 37528},
 {"Proof|ID:217754000-1", 1492298, 37532},
 {"Proof|ID:415524168", 1492459, 37536},
 {"Proof|ID:415524168-1", 1492622, 37540},
 {"Proof|ID:51392774", 1492782, 37544},
 {"Proof|ID:51392774-1", 1492942, 37548},
 {"Proof|ID:488904469", 1493101, 37552},
 {"Proof|ID:488904469-1", 1493264, 37556},
 {"Proof|ID:469346824", 1493425, 37560},
 {"Proof|ID:469346824-1", 1493588, 37564},
 {"ID:195073923", 1493743, 37568},
 {"ID:18535645", 1493976, 37575},
 {"\[Succeeds] - \[ScriptCapitalZ]n", 1494270, 37582},
 {"ID:155348029", 1494544, 37589},
 {"\[Succeeds] - \[RawStar] \[ScriptCapitalZ]n", 1494868, 37597},
 {"ID:220895581", 1495161, 37605},
 {"\[Succeeds] + \[ScriptCapitalZ]n", 1495457, 37612},
 {"Proof|ID:18535645", 1495738, 37619},
 {"Proof|ID:18535645-1", 1495898, 37623},
 {"Proof|ID:155348029", 1496057, 37627},
 {"Proof|ID:155348029-1", 1496220, 37631},
 {"Proof|ID:220895581", 1496381, 37635},
 {"Proof|ID:220895581-1", 1496543, 37639},
 {"ID:108166851", 1496697, 37643},
 {"ID:166132041", 1496931, 37650},
 {"gcd \[VerticalBar] \[RawStar] \[ScriptCapitalZ]n", 1497265, 37658},
 {"ID:414892396", 1497563, 37666},
 {"mult1 sufficient 1 \[ScriptCapitalZ]n", 1497869, 37673},
 {"ID:48379652", 1498149, 37680},
 {"mult1 sufficient 2 \[ScriptCapitalZ]n", 1498453, 37687},
 {"ID:287224029", 1498732, 37694},
 {"mult1 sufficient 3 \[ScriptCapitalZ]n", 1499038, 37701},
 {"ID:398117055", 1499319, 37708},
 {"mult1 sufficient 4 \[ScriptCapitalZ]n", 1499625, 37715},
 {"ID:568873110", 1499906, 37722},
 {"mult1 \[RawStar] injective \[ScriptCapitalZ]n", 1500234, 37730},
 {"ID:347198673", 1500529, 37738},
 {"mult1 sufficient \[ScriptCapitalZ]n", 1500831, 37745},
 {"Proof|ID:166132041", 1501116, 37752},
 {"Proof|ID:166132041-1", 1501279, 37756},
 {"Proof|ID:414892396", 1501440, 37760},
 {"Proof|ID:414892396-1", 1501603, 37764},
 {"Proof|ID:48379652", 1501763, 37768},
 {"Proof|ID:48379652-1", 1501923, 37772},
 {"Proof|ID:287224029", 1502082, 37776},
 {"Proof|ID:287224029-1", 1502245, 37780},
 {"Proof|ID:398117055", 1502406, 37784},
 {"Proof|ID:398117055-1", 1502569, 37788},
 {"Proof|ID:568873110", 1502730, 37792},
 {"Proof|ID:568873110-1", 1502891, 37796},
 {"Proof|ID:347198673", 1503050, 37800},
 {"Proof|ID:347198673-1", 1503213, 37804},
 {"ID:25008217", 1503367, 37808},
 {"ID:761196101", 1503599, 37815},
 {"isReducible \[ScriptCapitalZ]n", 1503891, 37822},
 {"ID:216211003", 1504165, 37829},
 {"isReducible' \[ScriptCapitalZ]n", 1504459, 37836},
 {"ID:431638777", 1504734, 37843},
 {"gcd \[Succeeds] \[ScriptCapitalZ]n", 1505034, 37850},
 {"Proof|ID:761196101", 1505318, 37857},
 {"Proof|ID:761196101-1", 1505480, 37861},
 {"Proof|ID:216211003", 1505640, 37865},
 {"Proof|ID:216211003-1", 1505803, 37869},
 {"Proof|ID:431638777", 1505964, 37873},
 {"Proof|ID:431638777-1", 1506127, 37877},
 {"ID:219886779", 1506282, 37881},
 {"ID:277360974", 1506516, 37888},
 {"isCommonReducible \[ScriptCapitalZ]n", 1506820, 37895},
 {"ID:112584880", 1507100, 37902},
 {"isMNTCR unary \[ScriptCapitalZ]n", 1507396, 37909},
 {"ID:195146313", 1507672, 37916},
 {"isMNTCR identical \[ScriptCapitalZ]n", 1507976, 37923},
 {"ID:6406785", 1508254, 37930},
 {"isMNTCR distinct' \[ScriptCapitalZ]n", 1508554, 37937},
 {"ID:87852126", 1508829, 37944},
 {"isMNTCR distinct \[ScriptCapitalZ]n", 1509129, 37951},
 {"ID:37400195", 1509405, 37958},
 {"isMNTCR \[ScriptCapitalZ]n", 1509687, 37965},
 {"Proof|ID:277360974", 1509961, 37972},
 {"Proof|ID:277360974-1", 1510123, 37976},
 {"Proof|ID:112584880", 1510283, 37980},
 {"Proof|ID:112584880-1", 1510446, 37984},
 {"Proof|ID:195146313", 1510607, 37988},
 {"Proof|ID:195146313-1", 1510770, 37992},
 {"Proof|ID:6406785", 1510929, 37996},
 {"Proof|ID:6406785-1", 1511086, 38000},
 {"Proof|ID:87852126", 1511242, 38004},
 {"Proof|ID:87852126-1", 1511402, 38008},
 {"Proof|ID:37400195", 1511560, 38012},
 {"Proof|ID:37400195-1", 1511720, 38016},
 {"ID:14129606", 1511872, 38020},
 {"ID:126313666", 1512104, 38027},
 {"mult1 - \[ScriptCapitalZ]n", 1512388, 38034},
 {"ID:888879353", 1512658, 38041},
 {"areCorrelative 1 \[ScriptCapitalZ]n", 1512960, 38048},
 {"ID:23406014", 1513238, 38055},
 {"areCorrelative 2 \[ScriptCapitalZ]n", 1513538, 38062},
 {"Proof|ID:126313666", 1513821, 38069},
 {"Proof|ID:126313666-1", 1513983, 38073},
 {"Proof|ID:888879353", 1514143, 38077},
 {"Proof|ID:888879353-1", 1514303, 38081},
 {"Proof|ID:23406014", 1514460, 38085},
 {"Proof|ID:23406014-1", 1514620, 38089},
 {"ID:325383468", 1514773, 38093},
 {"R5 \[ScriptCapitalZ]n", 1515046, 38100},
 {"ID:1783775451", 1515311, 38107},
 {"R6 \[ScriptCapitalZ]n", 1515586, 38114},
 {"Proof|ID:325383468", 1515858, 38121},
 {"Proof|ID:325383468-1", 1516018, 38125},
 {"Proof|ID:1783775451", 1516177, 38129},
 {"Proof|ID:1783775451-1", 1516342, 38133},
 {"ID:394176110", 1516498, 38137},
 {"isMNTCR existence \[ScriptCapitalZ]n", 1516802, 38144},
 {"ID:232518088", 1517082, 38151},
 {"isMNTCR unary existence \[ScriptCapitalZ]n", 1517398, 38158},
 {"Proof|ID:394176110", 1517690, 38165},
 {"Proof|ID:394176110-1", 1517853, 38169},
 {"Proof|ID:232518088", 1518014, 38173},
 {"Proof|ID:232518088-1", 1518177, 38177},
 {"ID:2147206476", 1518333, 38181},
 {"R7 \[ScriptCapitalZ]n", 1518608, 38188},
 {"ID:1971443720", 1518875, 38195},
 {"R8 \[ScriptCapitalZ]n", 1519150, 38202},
 {"Proof|ID:2147206476", 1519423, 38209},
 {"Proof|ID:2147206476-1", 1519587, 38213},
 {"Proof|ID:1971443720", 1519749, 38217},
 {"Proof|ID:1971443720-1", 1519914, 38221},
 {"ID:727194704", 1520070, 38225},
 {"isReductionRing \[ScriptCapitalZ]n", 1520369, 38232},
 {"Proof|ID:727194704", 1520652, 38239},
 {"Proof|ID:727194704-1", 1520814, 38243},
 {"ID:73858040", 1520967, 38247},
 {"ID:502941593", 1521199, 38254},
 {"rdm ternary \[ScriptCapitalZ]n", 1521491, 38261},
 {"ID:691632805", 1521765, 38268},
 {"rdm aux \[ScriptCapitalZ]n", 1522049, 38275},
 {"ID:290458459", 1522319, 38282},
 {"membership rdm \[ScriptCapitalZ]n", 1522616, 38289},
 {"ID:38241486", 1522891, 38296},
 {"rdm cases \[ScriptCapitalZ]n", 1523177, 38303},
 {"ID:140983642", 1523447, 38310},
 {"rdm \[RawStar] \[ScriptCapitalZ]n", 1523745, 38317},
 {"ID:33831774", 1524021, 38324},
 {"rdm 0 \[ScriptCapitalZ]n", 1524298, 38331},
 {"Proof|ID:502941593", 1524569, 38338},
 {"Proof|ID:502941593-1", 1524731, 38342},
 {"Proof|ID:691632805", 1524891, 38346},
 {"Proof|ID:691632805-1", 1525053, 38350},
 {"Proof|ID:290458459", 1525213, 38354},
 {"Proof|ID:290458459-1", 1525376, 38358},
 {"Proof|ID:38241486", 1525536, 38362},
 {"Proof|ID:38241486-1", 1525696, 38366},
 {"Proof|ID:140983642", 1525855, 38370},
 {"Proof|ID:140983642-1", 1526018, 38374},
 {"Proof|ID:33831774", 1526178, 38378},
 {"Proof|ID:33831774-1", 1526338, 38382},
 {"ID:1376076253", 1526492, 38386},
 {"R9 \[ScriptCapitalZ]n", 1526767, 38393},
 {"ID:125450892", 1527033, 38400},
 {"R10 \[ScriptCapitalZ]n", 1527308, 38407},
 {"ID:189921319", 1527573, 38414},
 {"R11 \[ScriptCapitalZ]n", 1527848, 38421},
 {"ID:667306596", 1528113, 38428},
 {"R12 \[ScriptCapitalZ]n", 1528388, 38435},
 {"ID:194342088", 1528653, 38442},
 {"R13 \[ScriptCapitalZ]n", 1528928, 38449},
 {"Proof|ID:1376076253", 1529200, 38456},
 {"Proof|ID:1376076253-1", 1529365, 38460},
 {"Proof|ID:125450892", 1529527, 38464},
 {"Proof|ID:125450892-1", 1529689, 38468},
 {"Proof|ID:189921319", 1529849, 38472},
 {"Proof|ID:189921319-1", 1530011, 38476},
 {"Proof|ID:667306596", 1530171, 38480},
 {"Proof|ID:667306596-1", 1530333, 38484},
 {"Proof|ID:194342088", 1530493, 38488},
 {"Proof|ID:194342088-1", 1530655, 38492},
 {"ID:403270264", 1530809, 38496},
 {"isAlgoReductionRing \[ScriptCapitalZ]n", 1531116, 38503},
 {"Proof|ID:403270264", 1531403, 38510},
 {"Proof|ID:403270264-1", 1531564, 38514}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[579, 22, 187, 3, 84, "Title",
 CellID->913222369],
Cell[769, 27, 389, 9, 52, "Subtitle",
 CellID->53451034],
Cell[CellGroupData[{
Cell[1183, 40, 119, 2, 73, "Section",
 CellID->1686608038],
Cell[CellGroupData[{
Cell[1327, 46, 168, 3, 70, "Subsection",
 CellID->586723348],
Cell[1498, 51, 937, 23, 70, "Text",
 CellID->128068127]
}, Closed]],
Cell[CellGroupData[{
Cell[2472, 79, 125, 2, 70, "Subsection",
 CellID->6629529],
Cell[2600, 83, 344, 8, 70, "Input",
 CellID->498486718],
Cell[2947, 93, 408, 9, 70, "Input",
 CellID->66724946],
Cell[3358, 104, 366, 8, 70, "Input",
 CellID->47868298],
Cell[3727, 114, 329, 8, 70, "Input",
 CellID->77326002],
Cell[4059, 124, 390, 9, 70, "Input",
 CellID->21887977],
Cell[4452, 135, 383, 9, 70, "Input",
 CellID->43455998],
Cell[4838, 146, 464, 10, 70, "Input",
 CellID->83415409],
Cell[5305, 158, 153, 3, 70, "Input",
 CellID->170511016],
Cell[5461, 163, 153, 3, 70, "Input",
 CellID->669825776],
Cell[CellGroupData[{
Cell[5639, 170, 1468, 32, 70, "Input",
 CellID->20666029],
Cell[7110, 204, 154, 4, 70, "Output",
 CellID->627834352]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[7313, 214, 124, 2, 70, "Subsection",
 CellID->1386333768],
Cell[CellGroupData[{
Cell[7462, 220, 353, 8, 70, "Input",
 CellID->1039312110],
Cell[CellGroupData[{
Cell[7840, 232, 1019, 21, 70, "Print",
 CellID->43308995],
Cell[8862, 255, 724, 15, 70, "Print",
 CellID->773653712],
Cell[9589, 272, 1018, 21, 70, "Print",
 CellID->547289795],
Cell[10610, 295, 990, 20, 70, "Print",
 CellID->583205520],
Cell[11603, 317, 724, 15, 70, "Print",
 CellID->917617342],
Cell[12330, 334, 994, 20, 70, "Print",
 CellID->622225809],
Cell[13327, 356, 723, 15, 70, "Print",
 CellID->60147518],
Cell[14053, 373, 1004, 20, 70, "Print",
 CellID->346776519],
Cell[15060, 395, 726, 15, 70, "Print",
 CellID->563606824],
Cell[15789, 412, 728, 15, 70, "Print",
 CellID->306995116],
Cell[16520, 429, 729, 15, 70, "Print",
 CellID->85378455],
Cell[17252, 446, 730, 15, 70, "Print",
 CellID->612641366],
Cell[17985, 463, 1022, 21, 70, "Print",
 CellID->212043760],
Cell[19010, 486, 724, 15, 70, "Print",
 CellID->287935756],
Cell[19737, 503, 725, 15, 70, "Print",
 CellID->39772154],
Cell[20465, 520, 994, 20, 70, "Print",
 CellID->571996574],
Cell[21462, 542, 724, 15, 70, "Print",
 CellID->308344464],
Cell[22189, 559, 726, 15, 70, "Print",
 CellID->200909420],
Cell[22918, 576, 995, 20, 70, "Print",
 CellID->27659054],
Cell[23916, 598, 726, 15, 70, "Print",
 CellID->765332773],
Cell[24645, 615, 727, 15, 70, "Print",
 CellID->40385639],
Cell[25375, 632, 730, 15, 70, "Print",
 CellID->248554666],
Cell[26108, 649, 730, 15, 70, "Print",
 CellID->762742487],
Cell[26841, 666, 1009, 21, 70, "Print",
 CellID->249282650],
Cell[27853, 689, 724, 15, 70, "Print",
 CellID->8175745],
Cell[28580, 706, 728, 15, 70, "Print",
 CellID->185713539],
Cell[29311, 723, 729, 15, 70, "Print",
 CellID->51957385],
Cell[30043, 740, 730, 15, 70, "Print",
 CellID->111507338],
Cell[30776, 757, 729, 15, 70, "Print",
 CellID->77483203],
Cell[31508, 774, 992, 20, 70, "Print",
 CellID->172403181],
Cell[32503, 796, 726, 15, 70, "Print",
 CellID->565330808],
Cell[33232, 813, 726, 15, 70, "Print",
 CellID->1606675],
Cell[33961, 830, 729, 15, 70, "Print",
 CellID->20301363],
Cell[34693, 847, 730, 15, 70, "Print",
 CellID->203974489],
Cell[35426, 864, 730, 15, 70, "Print",
 CellID->195794873],
Cell[36159, 881, 729, 15, 70, "Print",
 CellID->23265144],
Cell[36891, 898, 1001, 20, 70, "Print",
 CellID->73579430],
Cell[37895, 920, 726, 15, 70, "Print",
 CellID->625647329],
Cell[38624, 937, 728, 15, 70, "Print",
 CellID->380032735],
Cell[39355, 954, 729, 15, 70, "Print",
 CellID->59964078],
Cell[40087, 971, 730, 15, 70, "Print",
 CellID->173396994],
Cell[40820, 988, 730, 15, 70, "Print",
 CellID->189331182],
Cell[41553, 1005, 729, 15, 70, "Print",
 CellID->29652246],
Cell[42285, 1022, 729, 15, 70, "Print",
 CellID->54261457],
Cell[43017, 1039, 998, 20, 70, "Print",
 CellID->2598487],
Cell[44018, 1061, 726, 15, 70, "Print",
 CellID->196075428],
Cell[44747, 1078, 728, 15, 70, "Print",
 CellID->524371889],
Cell[45478, 1095, 730, 15, 70, "Print",
 CellID->117993319],
Cell[46211, 1112, 730, 15, 70, "Print",
 CellID->320762704],
Cell[46944, 1129, 1008, 20, 70, "Print",
 CellID->355680031],
Cell[47955, 1151, 726, 15, 70, "Print",
 CellID->487316616],
Cell[48684, 1168, 728, 15, 70, "Print",
 CellID->140301994],
Cell[49415, 1185, 730, 15, 70, "Print",
 CellID->166787746],
Cell[50148, 1202, 730, 15, 70, "Print",
 CellID->283202512],
Cell[50881, 1219, 730, 15, 70, "Print",
 CellID->285961962],
Cell[51614, 1236, 997, 20, 70, "Print",
 CellID->15104475],
Cell[52614, 1258, 724, 15, 70, "Print",
 CellID->323998079],
Cell[53341, 1275, 992, 20, 70, "Print",
 CellID->370342082],
Cell[54336, 1297, 726, 15, 70, "Print",
 CellID->351985089],
Cell[55065, 1314, 728, 15, 70, "Print",
 CellID->391305160],
Cell[55796, 1331, 729, 15, 70, "Print",
 CellID->68475953],
Cell[56528, 1348, 729, 15, 70, "Print",
 CellID->26466573],
Cell[57260, 1365, 728, 15, 70, "Print",
 CellID->1984322],
Cell[57991, 1382, 527, 10, 70, "Print",
 CellID->604141779],
Cell[58521, 1394, 517, 10, 70, "Print",
 CellID->915312300],
Cell[59041, 1406, 718, 15, 70, "Print",
 CellID->281663586],
Cell[59762, 1423, 720, 15, 70, "Print",
 CellID->209327295],
Cell[60485, 1440, 721, 15, 70, "Print",
 CellID->18937374],
Cell[61209, 1457, 722, 15, 70, "Print",
 CellID->153557070],
Cell[61934, 1474, 721, 15, 70, "Print",
 CellID->45465621],
Cell[62658, 1491, 721, 15, 70, "Print",
 CellID->39313665],
Cell[63382, 1508, 722, 15, 70, "Print",
 CellID->105735773],
Cell[64107, 1525, 721, 15, 70, "Print",
 CellID->77314492],
Cell[64831, 1542, 722, 15, 70, "Print",
 CellID->185751704],
Cell[65556, 1559, 721, 15, 70, "Print",
 CellID->85116431],
Cell[66280, 1576, 722, 15, 70, "Print",
 CellID->639108732],
Cell[67005, 1593, 722, 15, 70, "Print",
 CellID->721277251],
Cell[67730, 1610, 720, 15, 70, "Print",
 CellID->3339797],
Cell[68453, 1627, 722, 15, 70, "Print",
 CellID->289009105],
Cell[69178, 1644, 722, 15, 70, "Print",
 CellID->199179739],
Cell[69903, 1661, 722, 15, 70, "Print",
 CellID->487360478],
Cell[70628, 1678, 721, 15, 70, "Print",
 CellID->84922310],
Cell[71352, 1695, 722, 15, 70, "Print",
 CellID->558497404],
Cell[72077, 1712, 722, 15, 70, "Print",
 CellID->255445829],
Cell[72802, 1729, 722, 15, 70, "Print",
 CellID->696097528],
Cell[73527, 1746, 724, 15, 70, "Print",
 CellID->111397814],
Cell[74254, 1763, 724, 15, 70, "Print",
 CellID->766053093],
Cell[74981, 1780, 723, 15, 70, "Print",
 CellID->80891635],
Cell[75707, 1797, 724, 15, 70, "Print",
 CellID->242035726],
Cell[76434, 1814, 724, 15, 70, "Print",
 CellID->459982240],
Cell[77161, 1831, 723, 15, 70, "Print",
 CellID->52879335],
Cell[77887, 1848, 724, 15, 70, "Print",
 CellID->514239187],
Cell[78614, 1865, 724, 15, 70, "Print",
 CellID->478929526],
Cell[79341, 1882, 724, 15, 70, "Print",
 CellID->142907266],
Cell[80068, 1899, 724, 15, 70, "Print",
 CellID->324881021],
Cell[80795, 1916, 724, 15, 70, "Print",
 CellID->211471681],
Cell[81522, 1933, 724, 15, 70, "Print",
 CellID->154205215],
Cell[82249, 1950, 724, 15, 70, "Print",
 CellID->795327148],
Cell[82976, 1967, 724, 15, 70, "Print",
 CellID->125127396],
Cell[83703, 1984, 724, 15, 70, "Print",
 CellID->112536954],
Cell[84430, 2001, 724, 15, 70, "Print",
 CellID->479851726],
Cell[85157, 2018, 724, 15, 70, "Print",
 CellID->159558757],
Cell[85884, 2035, 723, 15, 70, "Print",
 CellID->18724986],
Cell[86610, 2052, 723, 15, 70, "Print",
 CellID->55389970],
Cell[87336, 2069, 724, 15, 70, "Print",
 CellID->185394026],
Cell[88063, 2086, 724, 15, 70, "Print",
 CellID->113003008],
Cell[88790, 2103, 723, 15, 70, "Print",
 CellID->19983785],
Cell[89516, 2120, 724, 15, 70, "Print",
 CellID->741298015],
Cell[90243, 2137, 724, 15, 70, "Print",
 CellID->168981775],
Cell[90970, 2154, 723, 15, 70, "Print",
 CellID->69169284],
Cell[91696, 2171, 724, 15, 70, "Print",
 CellID->336736819],
Cell[92423, 2188, 724, 15, 70, "Print",
 CellID->291445319],
Cell[93150, 2205, 724, 15, 70, "Print",
 CellID->296881490],
Cell[93877, 2222, 724, 15, 70, "Print",
 CellID->860092272],
Cell[94604, 2239, 724, 15, 70, "Print",
 CellID->830634336],
Cell[95331, 2256, 724, 15, 70, "Print",
 CellID->481433850],
Cell[96058, 2273, 724, 15, 70, "Print",
 CellID->247307259],
Cell[96785, 2290, 724, 15, 70, "Print",
 CellID->133525692],
Cell[97512, 2307, 724, 15, 70, "Print",
 CellID->431298360],
Cell[98239, 2324, 724, 15, 70, "Print",
 CellID->268266100],
Cell[98966, 2341, 723, 15, 70, "Print",
 CellID->58630601],
Cell[99692, 2358, 724, 15, 70, "Print",
 CellID->265247545],
Cell[100419, 2375, 723, 15, 70, "Print",
 CellID->58725565],
Cell[101145, 2392, 723, 15, 70, "Print",
 CellID->81617263],
Cell[101871, 2409, 723, 15, 70, "Print",
 CellID->56313838],
Cell[102597, 2426, 724, 15, 70, "Print",
 CellID->235693051],
Cell[103324, 2443, 723, 15, 70, "Print",
 CellID->33877423],
Cell[104050, 2460, 724, 15, 70, "Print",
 CellID->198408801],
Cell[104777, 2477, 724, 15, 70, "Print",
 CellID->407586333],
Cell[105504, 2494, 723, 15, 70, "Print",
 CellID->21003792],
Cell[106230, 2511, 724, 15, 70, "Print",
 CellID->511717689],
Cell[106957, 2528, 724, 15, 70, "Print",
 CellID->374982656],
Cell[107684, 2545, 724, 15, 70, "Print",
 CellID->265656759],
Cell[108411, 2562, 529, 10, 70, "Print",
 CellID->224441463],
Cell[108943, 2574, 508, 10, 70, "Print",
 CellID->209939310]
}, Open  ]]
}, Closed]],
Cell[109478, 2588, 325, 6, 70, "Input",
 CellID->19428626],
Cell[CellGroupData[{
Cell[109828, 2598, 541, 12, 70, "Input",
 CellID->1909803754],
Cell[110372, 2612, 735, 12, 70, "Output",
 CellID->636966437]
}, Open  ]],
Cell[CellGroupData[{
Cell[111144, 2629, 206, 4, 70, "Input",
 CellID->97543220],
Cell[111353, 2635, 605590, 13436, 70, "Output",
 CellID->357032028]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[716992, 16077, 128, 2, 70, "Subsection",
 CellID->1454182661],
Cell[717123, 16081, 983, 25, 70, "Input",
 CellID->2074904605],
Cell[718109, 16108, 416, 9, 70, "Input",
 CellID->1980565193],
Cell[718528, 16119, 169, 3, 70, "Input",
 CellID->444074165],
Cell[718700, 16124, 313, 6, 70, "Input",
 CellID->219149603]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[719062, 16136, 119, 2, 43, "Section",
 CellID->2045714290],
Cell[CellGroupData[{
Cell[719206, 16142, 182, 3, 38, "Subsection",
 CellID->28139740],
Cell[719391, 16147, 1564, 42, 83, "Text",
 CellID->755849276],
Cell[720958, 16191, 321, 12, 24, "Text",
 CellID->1751629107],
Cell[721282, 16205, 1182, 37, 46, "Text",
 CellID->223062300],
Cell[CellGroupData[{
Cell[722489, 16246, 394, 7, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->2050266752],
Cell[722886, 16255, 2836, 98, 88, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1478572631],
Cell[725725, 16355, 362, 7, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1061920785]
}, Closed]],
Cell[726102, 16365, 1311, 41, 65, "Text",
 CellID->1284341245],
Cell[CellGroupData[{
Cell[727438, 16410, 390, 7, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1045358942],
Cell[727831, 16419, 7085, 232, 191, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1132937314],
Cell[734919, 16653, 357, 7, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->626642836]
}, Closed]],
Cell[735291, 16663, 1011, 31, 44, "Text",
 CellID->1661433738],
Cell[CellGroupData[{
Cell[736327, 16698, 399, 7, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1599285460],
Cell[736729, 16707, 2432, 49, 30, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1269727608],
Cell[739164, 16758, 361, 7, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->333263577]
}, Closed]],
Cell[739540, 16768, 1762, 59, 65, "Text",
 CellID->645395591],
Cell[CellGroupData[{
Cell[741327, 16831, 399, 7, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1699779808],
Cell[741729, 16840, 4596, 131, 110, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->969904702],
Cell[746328, 16973, 360, 7, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->198497244]
}, Closed]],
Cell[746703, 16983, 1156, 31, 44, "Text",
 CellID->1934527770],
Cell[747862, 17016, 1142, 28, 64, "Text",
 CellID->1075691108],
Cell[749007, 17046, 695, 24, 44, "Text",
 CellID->53741691],
Cell[CellGroupData[{
Cell[749727, 17074, 393, 7, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->204886373],
Cell[750123, 17083, 2512, 77, 83, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1003232302],
Cell[752638, 17162, 358, 7, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->325823042]
}, Closed]],
Cell[753011, 17172, 375, 11, 43, "Text",
 CellID->1312586886],
Cell[753389, 17185, 789, 28, 43, "Text",
 CellID->1202825855],
Cell[CellGroupData[{
Cell[754203, 17217, 399, 7, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1262287499],
Cell[754605, 17226, 1747, 40, 121, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->921465591],
Cell[756355, 17268, 360, 7, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->377926640]
}, Closed]],
Cell[756730, 17278, 1068, 36, 47, "Text",
 CellID->663291260],
Cell[CellGroupData[{
Cell[757823, 17318, 394, 7, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->642765392],
Cell[758220, 17327, 6700, 225, 356, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->696507084],
Cell[764923, 17554, 364, 7, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1280121131]
}, Closed]],
Cell[765302, 17564, 1331, 37, 47, "Text",
 CellID->478524917],
Cell[CellGroupData[{
Cell[766658, 17605, 397, 7, 25, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1022330126],
Cell[767058, 17614, 7525, 221, 264, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->416978728],
Cell[774586, 17837, 366, 7, 24, "Text",
 CellGroupingRules->{GroupTogetherGrouping, 10000.},
 CellID->1523883859]
}, Closed]],
Cell[774967, 17847, 1992, 69, 67, "Text",
 CellID->569233702]
}, Closed]],
Cell[CellGroupData[{
Cell[776996, 17921, 132, 2, 30, "Subsection",
 CellID->1440146747],
Cell[777131, 17925, 47, 1, 12, "OpenEnvironment",
 CellID->314144664],
Cell[CellGroupData[{
Cell[777203, 17930, 478, 11, 43, "EnvironmentHeader",
 CellID->1280908502],
Cell[777684, 17943, 721, 19, 32, "GlobalDeclaration",
 CellTags->{
  "ID:232123642", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->232123642],
Cell[778408, 17964, 1024, 26, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:2028940590", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "gcd GCD"},
 CellID->2028940590],
Cell[779435, 17992, 879, 24, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:166548655", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mod Mod"},
 CellID->166548655],
Cell[780317, 18018, 66, 1, 26, "EndEnvironmentMarker",
 CellID->1224687606]
}, Open  ]],
Cell[780398, 18022, 49, 1, 16, "CloseEnvironment",
 CellID->1474676546]
}, Closed]],
Cell[CellGroupData[{
Cell[780484, 18028, 200, 3, 30, "Subsection",
 CellID->1062008728],
Cell[780687, 18033, 46, 1, 70, "OpenEnvironment",
 CellID->91468236],
Cell[CellGroupData[{
Cell[780758, 18038, 447, 10, 70, "EnvironmentHeader",
 CellID->1417579161],
Cell[781208, 18050, 748, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1363294779", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->1363294779],
Cell[781959, 18071, 932, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1613423474", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->1613423474],
Cell[782894, 18094, 818, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1525507482", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->1525507482],
Cell[783715, 18116, 1427, 41, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1587087117", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR \[Element]"},
 CellID->1587087117],
Cell[785145, 18159, 855, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:447457872", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR 0"},
 CellID->447457872],
Cell[786003, 18184, 963, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:815094494", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR 1"},
 CellID->815094494],
Cell[786969, 18212, 988, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1939288109", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR +"},
 CellID->1939288109],
Cell[787960, 18241, 960, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:915824852", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR -"},
 CellID->915824852],
Cell[788923, 18270, 964, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:370600110", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR - unary"},
 CellID->370600110],
Cell[789890, 18299, 960, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:344363438", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR \[RawStar]"},
 CellID->344363438],
Cell[790853, 18328, 1255, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1495376438", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR \[Succeeds]"},
 CellID->1495376438],
Cell[792111, 18370, 1094, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:550500630", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR multN"},
 CellID->550500630],
Cell[793208, 18399, 1308, 41, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:445166729", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR mult"},
 CellID->445166729],
Cell[794519, 18442, 1693, 46, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1793329966", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR mult 1"},
 CellID->1793329966],
Cell[796215, 18490, 1174, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:126969281", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR mult 2"},
 CellID->126969281],
Cell[797392, 18528, 2121, 53, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:909359224", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR rdm 1"},
 CellID->909359224],
Cell[799516, 18583, 1111, 33, 70, "EnvironmentText",
 CellID->2109204388],
Cell[800630, 18618, 2149, 48, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:236078685", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR rdm 2"},
 CellID->236078685],
Cell[802782, 18668, 1180, 37, 70, "EnvironmentText",
 CellID->61492128],
Cell[803965, 18707, 2430, 55, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:943334598", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR lcrd"},
 CellID->943334598],
Cell[806398, 18764, 413, 13, 70, "EnvironmentText",
 CellID->119714818],
Cell[806814, 18779, 1225, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1868934214", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR lcrd unary"},
 CellID->1868934214],
Cell[808042, 18811, 3929, 91, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:424764915", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "RIQR cpm"},
 CellID->424764915],
Cell[811974, 18904, 64, 1, 70, "EndEnvironmentMarker",
 CellID->89671333]
}, Open  ]],
Cell[812053, 18908, 49, 1, 70, "CloseEnvironment",
 CellID->1623525075]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[812151, 18915, 110, 2, 43, "Section",
 CellID->496748878],
Cell[CellGroupData[{
Cell[812286, 18921, 212, 5, 70, "Subsection",
 CellID->225277583],
Cell[812501, 18928, 48, 1, 70, "OpenEnvironment",
 CellID->2008906006],
Cell[CellGroupData[{
Cell[812574, 18933, 577, 15, 70, "EnvironmentHeader",
 CellID->957072583],
Cell[813154, 18950, 893, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:580511390", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[ScriptCapitalZ]24"},
 CellID->580511390],
Cell[814050, 18974, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1484947434]
}, Open  ]],
Cell[814131, 18978, 48, 1, 70, "CloseEnvironment",
 CellID->172883714],
Cell[CellGroupData[{
Cell[814204, 18983, 117, 2, 70, "Subsubsection",
 CellID->1321677136],
Cell[814324, 18987, 338, 7, 70, "Text",
 CellID->47140036],
Cell[CellGroupData[{
Cell[814687, 18998, 408, 8, 70, "Computation",
 CellID->849198139],
Cell[815098, 19008, 379, 7, 70, "ComputationResult",
 CellID->1460150010]
}, Open  ]],
Cell[CellGroupData[{
Cell[815514, 19020, 377, 10, 70, "ComputationInfo",
 CellID->1244254393],
Cell[815894, 19032, 647, 17, 70, "ComputationInfoBody",
 CellID->1827885036]
}, Closed]],
Cell[CellGroupData[{
Cell[816578, 19054, 243, 6, 70, "Computation",
 CellID->570943662],
Cell[816824, 19062, 140, 3, 70, "ComputationResult",
 CellID->1414651157]
}, Open  ]],
Cell[CellGroupData[{
Cell[817001, 19070, 376, 10, 70, "ComputationInfo",
 CellID->923325515],
Cell[817380, 19082, 646, 17, 70, "ComputationInfoBody",
 CellID->393895088]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[818075, 19105, 117, 2, 70, "Subsubsection",
 CellID->285408563],
Cell[CellGroupData[{
Cell[818217, 19111, 355, 7, 70, "Computation",
 CellID->490840057],
Cell[818575, 19120, 504, 12, 70, "ComputationResult",
 CellID->1920791119]
}, Open  ]],
Cell[CellGroupData[{
Cell[819116, 19137, 377, 10, 70, "ComputationInfo",
 CellID->1840123044],
Cell[819496, 19149, 647, 17, 70, "ComputationInfoBody",
 CellID->2050404037]
}, Closed]],
Cell[CellGroupData[{
Cell[820180, 19171, 306, 6, 70, "Computation",
 CellID->260906778],
Cell[820489, 19179, 424, 10, 70, "ComputationResult",
 CellID->1643593823]
}, Open  ]],
Cell[CellGroupData[{
Cell[820950, 19194, 379, 10, 70, "ComputationInfo",
 CellID->1414827049],
Cell[821332, 19206, 647, 17, 70, "ComputationInfoBody",
 CellID->1588362347]
}, Closed]],
Cell[CellGroupData[{
Cell[822016, 19228, 312, 6, 70, "Computation",
 CellID->2002455362],
Cell[822331, 19236, 353, 10, 70, "ComputationResult",
 CellID->867934889]
}, Open  ]],
Cell[CellGroupData[{
Cell[822721, 19251, 378, 10, 70, "ComputationInfo",
 CellID->1370538128],
Cell[823102, 19263, 648, 17, 70, "ComputationInfoBody",
 CellID->1552535153]
}, Closed]],
Cell[CellGroupData[{
Cell[823787, 19285, 359, 7, 70, "Computation",
 CellID->1432418796],
Cell[824149, 19294, 376, 10, 70, "ComputationResult",
 CellID->1002747542]
}, Open  ]],
Cell[CellGroupData[{
Cell[824562, 19309, 380, 10, 70, "ComputationInfo",
 CellID->1780779641],
Cell[824945, 19321, 649, 17, 70, "ComputationInfoBody",
 CellID->131149116]
}, Closed]],
Cell[CellGroupData[{
Cell[825631, 19343, 356, 7, 70, "Computation",
 CellID->1091363187],
Cell[825990, 19352, 372, 9, 70, "ComputationResult",
 CellID->673303920]
}, Open  ]],
Cell[CellGroupData[{
Cell[826399, 19366, 378, 10, 70, "ComputationInfo",
 CellID->1502806330],
Cell[826780, 19378, 648, 17, 70, "ComputationInfoBody",
 CellID->1911587008]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[827477, 19401, 114, 2, 70, "Subsubsection",
 CellID->840789324],
Cell[CellGroupData[{
Cell[827616, 19407, 297, 8, 70, "Computation",
 CellID->710437623],
Cell[827916, 19417, 467, 11, 70, "ComputationResult",
 CellID->442317352]
}, Open  ]],
Cell[CellGroupData[{
Cell[828420, 19433, 379, 10, 70, "ComputationInfo",
 CellID->1642092347],
Cell[828802, 19445, 645, 17, 70, "ComputationInfoBody",
 CellID->93258772]
}, Closed]],
Cell[CellGroupData[{
Cell[829484, 19467, 346, 8, 70, "Computation",
 CellID->1351181837],
Cell[829833, 19477, 515, 11, 70, "ComputationResult",
 CellID->1528111983]
}, Open  ]],
Cell[CellGroupData[{
Cell[830385, 19493, 377, 10, 70, "ComputationInfo",
 CellID->100484650],
Cell[830765, 19505, 649, 17, 70, "ComputationInfoBody",
 CellID->345002316]
}, Closed]],
Cell[CellGroupData[{
Cell[831451, 19527, 369, 8, 70, "Computation",
 CellID->140107252],
Cell[831823, 19537, 387, 9, 70, "ComputationResult",
 CellID->1619126148]
}, Open  ]],
Cell[CellGroupData[{
Cell[832247, 19551, 376, 10, 70, "ComputationInfo",
 CellID->315583679],
Cell[832626, 19563, 649, 17, 70, "ComputationInfoBody",
 CellID->2027695778]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[833336, 19587, 210, 5, 70, "Subsection",
 CellID->625973727],
Cell[833549, 19594, 47, 1, 70, "OpenEnvironment",
 CellID->187010305],
Cell[CellGroupData[{
Cell[833621, 19599, 515, 15, 70, "EnvironmentHeader",
 CellID->862011211],
Cell[834139, 19616, 895, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:275453931", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[ScriptCapitalZ]13"},
 CellID->275453931],
Cell[835037, 19640, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1991026502]
}, Open  ]],
Cell[835118, 19644, 49, 1, 70, "CloseEnvironment",
 CellID->1516776967],
Cell[CellGroupData[{
Cell[835192, 19649, 116, 2, 70, "Subsubsection",
 CellID->932766472],
Cell[CellGroupData[{
Cell[835333, 19655, 501, 9, 70, "Computation",
 CellID->889827160],
Cell[835837, 19666, 429, 8, 70, "ComputationResult",
 CellID->1729630962]
}, Open  ]],
Cell[CellGroupData[{
Cell[836303, 19679, 376, 10, 70, "ComputationInfo",
 CellID->722095009],
Cell[836682, 19691, 646, 17, 70, "ComputationInfoBody",
 CellID->996548225]
}, Closed]],
Cell[CellGroupData[{
Cell[837365, 19713, 288, 6, 70, "Computation",
 CellID->81011551],
Cell[837656, 19721, 191, 4, 70, "ComputationResult",
 CellID->1196950894]
}, Open  ]],
Cell[CellGroupData[{
Cell[837884, 19730, 376, 10, 70, "ComputationInfo",
 CellID->1841495814],
Cell[838263, 19742, 644, 17, 70, "ComputationInfoBody",
 CellID->658681144]
}, Closed]],
Cell[CellGroupData[{
Cell[838944, 19764, 242, 6, 70, "Computation",
 CellID->763410701],
Cell[839189, 19772, 137, 3, 70, "ComputationResult",
 CellID->4820217]
}, Open  ]],
Cell[CellGroupData[{
Cell[839363, 19780, 377, 10, 70, "ComputationInfo",
 CellID->2064421696],
Cell[839743, 19792, 646, 17, 70, "ComputationInfoBody",
 CellID->949097751]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[840438, 19815, 117, 2, 70, "Subsubsection",
 CellID->449176887],
Cell[CellGroupData[{
Cell[840580, 19821, 406, 8, 70, "Computation",
 CellID->1410083086],
Cell[840989, 19831, 525, 12, 70, "ComputationResult",
 CellID->490732742]
}, Open  ]],
Cell[CellGroupData[{
Cell[841551, 19848, 377, 10, 70, "ComputationInfo",
 CellID->343184513],
Cell[841931, 19860, 648, 17, 70, "ComputationInfoBody",
 CellID->2078259407]
}, Closed]],
Cell[CellGroupData[{
Cell[842616, 19882, 359, 7, 70, "Computation",
 CellID->158220634],
Cell[842978, 19891, 475, 11, 70, "ComputationResult",
 CellID->830283031]
}, Open  ]],
Cell[CellGroupData[{
Cell[843490, 19907, 377, 10, 70, "ComputationInfo",
 CellID->1491837892],
Cell[843870, 19919, 646, 17, 70, "ComputationInfoBody",
 CellID->1317010961]
}, Closed]],
Cell[CellGroupData[{
Cell[844553, 19941, 362, 7, 70, "Computation",
 CellID->676617568],
Cell[844918, 19950, 402, 10, 70, "ComputationResult",
 CellID->1584131273]
}, Open  ]],
Cell[CellGroupData[{
Cell[845357, 19965, 376, 10, 70, "ComputationInfo",
 CellID->494933475],
Cell[845736, 19977, 647, 17, 70, "ComputationInfoBody",
 CellID->1500053850]
}, Closed]],
Cell[CellGroupData[{
Cell[846420, 19999, 408, 8, 70, "Computation",
 CellID->1992980618],
Cell[846831, 20009, 398, 10, 70, "ComputationResult",
 CellID->1494517660]
}, Open  ]],
Cell[CellGroupData[{
Cell[847266, 20024, 378, 10, 70, "ComputationInfo",
 CellID->1378990324],
Cell[847647, 20036, 648, 17, 70, "ComputationInfoBody",
 CellID->1348975192]
}, Closed]],
Cell[CellGroupData[{
Cell[848332, 20058, 405, 8, 70, "Computation",
 CellID->168529735],
Cell[848740, 20068, 429, 11, 70, "ComputationResult",
 CellID->699749025]
}, Open  ]],
Cell[CellGroupData[{
Cell[849206, 20084, 379, 10, 70, "ComputationInfo",
 CellID->1668924093],
Cell[849588, 20096, 646, 17, 70, "ComputationInfoBody",
 CellID->225927190]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[850283, 20119, 114, 2, 70, "Subsubsection",
 CellID->123861815],
Cell[CellGroupData[{
Cell[850422, 20125, 368, 9, 70, "Computation",
 CellID->894685],
Cell[850793, 20136, 489, 11, 70, "ComputationResult",
 CellID->799097763]
}, Open  ]],
Cell[CellGroupData[{
Cell[851319, 20152, 374, 10, 70, "ComputationInfo",
 CellID->1700953392],
Cell[851696, 20164, 640, 16, 70, "ComputationInfoBody",
 CellID->1596585968]
}, Closed]],
Cell[CellGroupData[{
Cell[852373, 20185, 391, 8, 70, "Computation",
 CellID->788223205],
Cell[852767, 20195, 564, 12, 70, "ComputationResult",
 CellID->157392879]
}, Open  ]],
Cell[CellGroupData[{
Cell[853368, 20212, 378, 10, 70, "ComputationInfo",
 CellID->494916420],
Cell[853749, 20224, 647, 17, 70, "ComputationInfoBody",
 CellID->1558628290]
}, Closed]],
Cell[CellGroupData[{
Cell[854433, 20246, 421, 9, 70, "Computation",
 CellID->2133581505],
Cell[854857, 20257, 412, 10, 70, "ComputationResult",
 CellID->1665785973]
}, Open  ]],
Cell[CellGroupData[{
Cell[855306, 20272, 377, 10, 70, "ComputationInfo",
 CellID->661543208],
Cell[855686, 20284, 649, 17, 70, "ComputationInfoBody",
 CellID->253352679]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[856408, 20309, 343, 7, 43, "Section",
 CellID->1529081576],
Cell[856754, 20318, 715, 18, 30, "GlobalDeclaration",
 CellTags->{
  "ID:7294098", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->7294098],
Cell[857472, 20338, 36, 1, 24, "Text",
 CellID->231827409],
Cell[857511, 20341, 836, 20, 33, "GlobalDeclaration",
 CellTags->{
  "ID:6216142", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->6216142],
Cell[CellGroupData[{
Cell[858372, 20365, 190, 3, 38, "Subsection",
 CellID->509369914],
Cell[CellGroupData[{
Cell[858587, 20372, 190, 3, 70, "Subsubsection",
 CellID->157842255],
Cell[858780, 20377, 47, 1, 70, "OpenEnvironment",
 CellID->635903570],
Cell[CellGroupData[{
Cell[858852, 20382, 479, 12, 70, "EnvironmentHeader",
 CellID->369402263],
Cell[859334, 20396, 844, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:285451873", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "modulus \[NotEqual] 0 \[ScriptCapitalZ]n"},
 CellID->285451873],
Cell[860181, 20419, 1720, 54, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:270643577", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Element] alternative \[ScriptCapitalZ]n"},
 
 CellID->270643577],
Cell[861904, 20475, 1053, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:916172587", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[ScriptCapitalZ]1"},
 CellID->916172587],
Cell[862960, 20508, 806, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:200088253", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->200088253],
Cell[863769, 20530, 860, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:62294620", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "abs \[ScriptCapitalZ]n"},
 CellID->62294620],
Cell[864632, 20554, 868, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:90375616", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mod identity \[ScriptCapitalZ]n"},
 CellID->90375616],
Cell[865503, 20579, 1235, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:575444601", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "- unary alternative \[ScriptCapitalZ]n"},
 CellID->575444601],
Cell[866741, 20612, 63, 1, 70, "EndEnvironmentMarker",
 CellID->4769084]
}, Open  ]],
Cell[866819, 20616, 48, 1, 70, "CloseEnvironment",
 CellID->179375776],
Cell[CellGroupData[{
Cell[866892, 20621, 2305, 70, 70, "ProofInfo",
 CellTags->{"Proof|ID:285451873", "Proof|ID:285451873-1"},
 CellID->127340268],
Cell[869200, 20693, 480, 12, 70, "ProofInfoBody",
 CellID->558949415]
}, Closed]],
Cell[CellGroupData[{
Cell[869717, 20710, 3865, 106, 70, "ProofInfo",
 CellTags->{"Proof|ID:270643577", "Proof|ID:270643577-1"},
 CellID->506264492],
Cell[873585, 20818, 479, 12, 70, "ProofInfoBody",
 CellID->60003179]
}, Closed]],
Cell[CellGroupData[{
Cell[874101, 20835, 2353, 70, 70, "ProofInfo",
 CellTags->{"Proof|ID:916172587", "Proof|ID:916172587-1"},
 CellID->265170607],
Cell[876457, 20907, 480, 12, 70, "ProofInfoBody",
 CellID->351314708]
}, Closed]],
Cell[CellGroupData[{
Cell[876974, 20924, 3001, 86, 70, "ProofInfo",
 CellTags->{"Proof|ID:62294620", "Proof|ID:62294620-1"},
 CellID->319229321],
Cell[879978, 21012, 479, 12, 70, "ProofInfoBody",
 CellID->301185869]
}, Closed]],
Cell[CellGroupData[{
Cell[880494, 21029, 3107, 89, 70, "ProofInfo",
 CellTags->{"Proof|ID:90375616", "Proof|ID:90375616-1"},
 CellID->190812059],
Cell[883604, 21120, 479, 12, 70, "ProofInfoBody",
 CellID->751344191]
}, Closed]],
Cell[CellGroupData[{
Cell[884120, 21137, 3972, 104, 70, "ProofInfo",
 CellTags->{"Proof|ID:575444601", "Proof|ID:575444601-1"},
 CellID->94536918],
Cell[888095, 21243, 480, 12, 70, "ProofInfoBody",
 CellID->752059431]
}, Closed]],
Cell[888590, 21258, 47, 1, 70, "OpenEnvironment",
 CellID->190598708],
Cell[CellGroupData[{
Cell[888662, 21263, 389, 9, 70, "EnvironmentHeader",
 CellID->637227383],
Cell[889054, 21274, 755, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:974601634", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->974601634],
Cell[889812, 21295, 972, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:779775892", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
   "gcd \[Element] \[DoubleStruckCapitalN] \[ScriptCapitalZ]n"},
 CellID->779775892],
Cell[890787, 21323, 1290, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:7430669", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership gcd \[ScriptCapitalZ]n", "IR"},
 CellID->7430669],
Cell[892080, 21363, 931, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:243173858", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "gcd \[NotEqual] 0 \[ScriptCapitalZ]n"},
 CellID->243173858],
Cell[893014, 21389, 1085, 29, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:335259273", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
   "quo gcd \[Element] \[DoubleStruckCapitalN] \[ScriptCapitalZ]n"},
 CellID->335259273],
Cell[894102, 21420, 1144, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:582853848", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership quo gcd \[ScriptCapitalZ]n"},
 CellID->582853848],
Cell[895249, 21454, 1047, 29, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:29867892", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership quo gcd -1 \[ScriptCapitalZ]n"},
 CellID->29867892],
Cell[896299, 21485, 1083, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:372275792", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "quo gcd \[RawStar] \[ScriptCapitalZ]n"},
 CellID->372275792],
Cell[897385, 21515, 721, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:825367134", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->825367134],
Cell[898109, 21535, 989, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:55333112", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership quo \[ScriptCapitalZ]n", "IR"},
 CellID->55333112],
Cell[899101, 21562, 1006, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:474647872", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership -mod \[ScriptCapitalZ]n", "IR"},
 CellID->474647872],
Cell[900110, 21591, 65, 1, 70, "EndEnvironmentMarker",
 CellID->285655755]
}, Open  ]],
Cell[900190, 21595, 48, 1, 70, "CloseEnvironment",
 CellID->305608247],
Cell[CellGroupData[{
Cell[900263, 21600, 3120, 88, 70, "ProofInfo",
 CellTags->{"Proof|ID:779775892", "Proof|ID:779775892-1"},
 CellID->297664998],
Cell[903386, 21690, 479, 12, 70, "ProofInfoBody",
 CellID->10018482]
}, Closed]],
Cell[CellGroupData[{
Cell[903902, 21707, 3845, 104, 70, "ProofInfo",
 CellTags->{"Proof|ID:7430669", "Proof|ID:7430669-1"},
 CellID->340529413],
Cell[907750, 21813, 478, 12, 70, "ProofInfoBody",
 CellID->114730061]
}, Closed]],
Cell[CellGroupData[{
Cell[908265, 21830, 3299, 92, 70, "ProofInfo",
 CellTags->{"Proof|ID:243173858", "Proof|ID:243173858-1"},
 CellID->623193018],
Cell[911567, 21924, 479, 12, 70, "ProofInfoBody",
 CellID->20166670]
}, Closed]],
Cell[CellGroupData[{
Cell[912083, 21941, 3426, 96, 70, "ProofInfo",
 CellTags->{"Proof|ID:335259273", "Proof|ID:335259273-1"},
 CellID->301510995],
Cell[915512, 22039, 480, 12, 70, "ProofInfoBody",
 CellID->678743316]
}, Closed]],
Cell[CellGroupData[{
Cell[916029, 22056, 4164, 112, 70, "ProofInfo",
 CellTags->{"Proof|ID:582853848", "Proof|ID:582853848-1"},
 CellID->62215823],
Cell[920196, 22170, 479, 12, 70, "ProofInfoBody",
 CellID->70354999]
}, Closed]],
Cell[CellGroupData[{
Cell[920712, 22187, 3670, 100, 70, "ProofInfo",
 CellTags->{"Proof|ID:29867892", "Proof|ID:29867892-1"},
 CellID->753487385],
Cell[924385, 22289, 479, 12, 70, "ProofInfoBody",
 CellID->115804636]
}, Closed]],
Cell[CellGroupData[{
Cell[924901, 22306, 3991, 107, 70, "ProofInfo",
 CellTags->{"Proof|ID:372275792", "Proof|ID:372275792-1"},
 CellID->9434432],
Cell[928895, 22415, 480, 12, 70, "ProofInfoBody",
 CellID->166615473]
}, Closed]],
Cell[CellGroupData[{
Cell[929412, 22432, 3561, 98, 70, "ProofInfo",
 CellTags->{"Proof|ID:55333112", "Proof|ID:55333112-1"},
 CellID->64245290],
Cell[932976, 22532, 479, 12, 70, "ProofInfoBody",
 CellID->611966264]
}, Closed]],
Cell[CellGroupData[{
Cell[933492, 22549, 3733, 101, 70, "ProofInfo",
 CellTags->{"Proof|ID:474647872", "Proof|ID:474647872-1"},
 CellID->209531985],
Cell[937228, 22652, 479, 12, 70, "ProofInfoBody",
 CellID->74010259]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[937756, 22670, 294, 6, 70, "Subsubsection",
 CellID->750673971],
Cell[938053, 22678, 48, 1, 70, "OpenEnvironment",
 CellID->1269913440],
Cell[CellGroupData[{
Cell[938126, 22683, 587, 13, 70, "EnvironmentHeader",
 CellID->1312707458],
Cell[938716, 22698, 842, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:208073319", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isFinDomain \[ScriptCapitalZ]n"},
 CellID->208073319],
Cell[939561, 22721, 833, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:207295573", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isDomain \[ScriptCapitalZ]n"},
 CellID->207295573],
Cell[940397, 22744, 1134, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:189562335", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isNeutral 1 \[ScriptCapitalZ]n"},
 CellID->189562335],
Cell[941534, 22778, 1013, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:371480170", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isCommRing1 \[ScriptCapitalZ]n"},
 CellID->371480170],
Cell[942550, 22804, 65, 1, 70, "EndEnvironmentMarker",
 CellID->608276809]
}, Open  ]],
Cell[942630, 22808, 49, 1, 70, "CloseEnvironment",
 CellID->1211509373],
Cell[CellGroupData[{
Cell[942704, 22813, 2387, 71, 70, "ProofInfo",
 CellTags->{"Proof|ID:208073319", "Proof|ID:208073319-1"},
 CellID->169378186],
Cell[945094, 22886, 480, 12, 70, "ProofInfoBody",
 CellID->439233552]
}, Closed]],
Cell[CellGroupData[{
Cell[945611, 22903, 2367, 70, 70, "ProofInfo",
 CellTags->{"Proof|ID:207295573", "Proof|ID:207295573-1"},
 CellID->415181399],
Cell[947981, 22975, 479, 12, 70, "ProofInfoBody",
 CellID->27765665]
}, Closed]],
Cell[CellGroupData[{
Cell[948497, 22992, 3014, 85, 70, "ProofInfo",
 CellTags->{"Proof|ID:189562335", "Proof|ID:189562335-1"},
 CellID->43006419],
Cell[951514, 23079, 480, 12, 70, "ProofInfoBody",
 CellID->144573181]
}, Closed]],
Cell[CellGroupData[{
Cell[952031, 23096, 2387, 71, 70, "ProofInfo",
 CellTags->{"Proof|ID:371480170", "Proof|ID:371480170-1"},
 CellID->344386318],
Cell[954421, 23169, 480, 12, 70, "ProofInfoBody",
 CellID->576038596]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[954950, 23187, 194, 3, 70, "Subsubsection",
 CellID->76760028],
Cell[955147, 23192, 47, 1, 70, "OpenEnvironment",
 CellID->338374869],
Cell[CellGroupData[{
Cell[955219, 23197, 424, 11, 70, "EnvironmentHeader",
 CellID->134166625],
Cell[955646, 23210, 1196, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1342794225", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isPartIrreflOrder \[ScriptCapitalZ]n"},
 CellID->1342794225],
Cell[956845, 23245, 1071, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:145791983", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isTotalIrreflOrder \[ScriptCapitalZ]n"},
 CellID->145791983],
Cell[957919, 23278, 1073, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1051492688", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isNoetherian \[ScriptCapitalZ]n"},
 CellID->1051492688],
Cell[958995, 23311, 64, 1, 70, "EndEnvironmentMarker",
 CellID->31879075]
}, Open  ]],
Cell[959074, 23315, 47, 1, 70, "CloseEnvironment",
 CellID->83297866],
Cell[CellGroupData[{
Cell[959146, 23320, 2790, 81, 70, "ProofInfo",
 CellTags->{"Proof|ID:1342794225", "Proof|ID:1342794225-1"},
 CellID->256415475],
Cell[961939, 23403, 479, 12, 70, "ProofInfoBody",
 CellID->3143692]
}, Closed]],
Cell[CellGroupData[{
Cell[962455, 23420, 2785, 80, 70, "ProofInfo",
 CellTags->{"Proof|ID:145791983", "Proof|ID:145791983-1"},
 CellID->187860367],
Cell[965243, 23502, 480, 12, 70, "ProofInfoBody",
 CellID->503595798]
}, Closed]],
Cell[CellGroupData[{
Cell[965760, 23519, 2761, 80, 70, "ProofInfo",
 CellTags->{"Proof|ID:1051492688", "Proof|ID:1051492688-1"},
 CellID->66415286],
Cell[968524, 23601, 480, 12, 70, "ProofInfoBody",
 CellID->54951663]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[969053, 23619, 275, 4, 70, "Subsubsection",
 CellID->193070521],
Cell[969331, 23625, 47, 1, 70, "OpenEnvironment",
 CellID->617760674],
Cell[CellGroupData[{
Cell[969403, 23630, 455, 10, 70, "EnvironmentHeader",
 CellID->411365783],
Cell[969861, 23642, 2252, 65, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:307883626", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Sum] \[ScriptCapitalZ]n"},
 CellID->307883626],
Cell[972116, 23709, 784, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:209868567", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->209868567],
Cell[972903, 23731, 1322, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:57125150", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 alternative \[ScriptCapitalZ]n"},
 CellID->57125150],
Cell[974228, 23770, 1567, 43, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:50258269", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R2' \[ScriptCapitalZ]n"},
 CellID->50258269],
Cell[975798, 23815, 65, 1, 70, "EndEnvironmentMarker",
 CellID->343575043]
}, Open  ]],
Cell[975878, 23819, 48, 1, 70, "CloseEnvironment",
 CellID->834291338],
Cell[CellGroupData[{
Cell[975951, 23824, 6345, 160, 70, "ProofInfo",
 CellTags->{"Proof|ID:307883626", "Proof|ID:307883626-1"},
 CellID->253583856],
Cell[982299, 23986, 480, 12, 70, "ProofInfoBody",
 CellID->217854969]
}, Closed]],
Cell[CellGroupData[{
Cell[982816, 24003, 4916, 128, 70, "ProofInfo",
 CellTags->{"Proof|ID:57125150", "Proof|ID:57125150-1"},
 CellID->163486749],
Cell[987735, 24133, 478, 12, 70, "ProofInfoBody",
 CellID->65009874]
}, Closed]],
Cell[CellGroupData[{
Cell[988250, 24150, 5262, 134, 70, "ProofInfo",
 CellTags->{"Proof|ID:50258269", "Proof|ID:50258269-1"},
 CellID->223632667],
Cell[993515, 24286, 479, 12, 70, "ProofInfoBody",
 CellID->553065911]
}, Closed]],
Cell[994009, 24301, 47, 1, 70, "OpenEnvironment",
 CellID->168314290],
Cell[CellGroupData[{
Cell[994081, 24306, 420, 11, 70, "EnvironmentHeader",
 CellID->566297413],
Cell[994504, 24319, 892, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1161287108", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "Multipliers \[ScriptCapitalZ]n"},
 CellID->1161287108],
Cell[995399, 24343, 833, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:124498584", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R0 \[ScriptCapitalZ]n"},
 CellID->124498584],
Cell[996235, 24366, 839, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1236424988", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R1 \[ScriptCapitalZ]n"},
 CellID->1236424988],
Cell[997077, 24389, 839, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1558881148", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R2 \[ScriptCapitalZ]n"},
 CellID->1558881148],
Cell[997919, 24412, 836, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1530993938", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R3 \[ScriptCapitalZ]n"},
 CellID->1530993938],
Cell[998758, 24435, 836, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:589877437", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R4 \[ScriptCapitalZ]n"},
 CellID->589877437],
Cell[999597, 24458, 65, 1, 70, "EndEnvironmentMarker",
 CellID->140752188]
}, Open  ]],
Cell[999677, 24462, 47, 1, 70, "CloseEnvironment",
 CellID->78307490],
Cell[CellGroupData[{
Cell[999749, 24467, 2396, 72, 70, "ProofInfo",
 CellTags->{"Proof|ID:1161287108", "Proof|ID:1161287108-1"},
 CellID->311987723],
Cell[1002148, 24541, 481, 12, 70, "ProofInfoBody",
 CellID->552671078]
}, Closed]],
Cell[CellGroupData[{
Cell[1002666, 24558, 2342, 70, 70, "ProofInfo",
 CellTags->{"Proof|ID:124498584", "Proof|ID:124498584-1"},
 CellID->24095429],
Cell[1005011, 24630, 480, 12, 70, "ProofInfoBody",
 CellID->237502882]
}, Closed]],
Cell[CellGroupData[{
Cell[1005528, 24647, 2352, 71, 70, "ProofInfo",
 CellTags->{"Proof|ID:1236424988", "Proof|ID:1236424988-1"},
 CellID->296864625],
Cell[1007883, 24720, 480, 12, 70, "ProofInfoBody",
 CellID->44769427]
}, Closed]],
Cell[CellGroupData[{
Cell[1008400, 24737, 2352, 71, 70, "ProofInfo",
 CellTags->{"Proof|ID:1558881148", "Proof|ID:1558881148-1"},
 CellID->239424613],
Cell[1010755, 24810, 481, 12, 70, "ProofInfoBody",
 CellID->286970037]
}, Closed]],
Cell[CellGroupData[{
Cell[1011273, 24827, 2352, 71, 70, "ProofInfo",
 CellTags->{"Proof|ID:1530993938", "Proof|ID:1530993938-1"},
 CellID->344362146],
Cell[1013628, 24900, 480, 12, 70, "ProofInfoBody",
 CellID->62375380]
}, Closed]],
Cell[CellGroupData[{
Cell[1014145, 24917, 2343, 70, 70, "ProofInfo",
 CellTags->{"Proof|ID:589877437", "Proof|ID:589877437-1"},
 CellID->352205366],
Cell[1016491, 24989, 480, 12, 70, "ProofInfoBody",
 CellID->173788587]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1017020, 25007, 213, 3, 70, "Subsubsection",
 CellID->191294579],
Cell[1017236, 25012, 46, 1, 70, "OpenEnvironment",
 CellID->40861546],
Cell[CellGroupData[{
Cell[1017307, 25017, 385, 9, 70, "EnvironmentHeader",
 CellID->219633271],
Cell[1017695, 25028, 782, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:104617955", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->104617955],
Cell[1018480, 25050, 1087, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:608534217", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership -quo \[ScriptCapitalZ]n", "IR"},
 CellID->608534217],
Cell[1019570, 25082, 1087, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:217754000", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership +quo \[ScriptCapitalZ]n", "IR"},
 CellID->217754000],
Cell[1020660, 25114, 1372, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:415524168", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult2 -quo \[ScriptCapitalZ]n"},
 CellID->415524168],
Cell[1022035, 25151, 1273, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:51392774", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 +quo \[ScriptCapitalZ]n"},
 CellID->51392774],
Cell[1023311, 25187, 1204, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:488904469", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[RawStar] -quo \[ScriptCapitalZ]n"},
 CellID->488904469],
Cell[1024518, 25222, 1153, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:469346824", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[RawStar] +quo \[ScriptCapitalZ]n"},
 CellID->469346824],
Cell[1025674, 25256, 65, 1, 70, "EndEnvironmentMarker",
 CellID->246186662]
}, Open  ]],
Cell[1025754, 25260, 48, 1, 70, "CloseEnvironment",
 CellID->384842955],
Cell[CellGroupData[{
Cell[1025827, 25265, 4501, 117, 70, "ProofInfo",
 CellTags->{"Proof|ID:608534217", "Proof|ID:608534217-1"},
 CellID->171963183],
Cell[1030331, 25384, 524, 13, 70, "ProofInfoBody",
 CellID->552509088]
}, Closed]],
Cell[CellGroupData[{
Cell[1030892, 25402, 4501, 117, 70, "ProofInfo",
 CellTags->{"Proof|ID:217754000", "Proof|ID:217754000-1"},
 CellID->218735392],
Cell[1035396, 25521, 480, 12, 70, "ProofInfoBody",
 CellID->329603756]
}, Closed]],
Cell[CellGroupData[{
Cell[1035913, 25538, 5509, 140, 70, "ProofInfo",
 CellTags->{"Proof|ID:415524168", "Proof|ID:415524168-1"},
 CellID->322621179],
Cell[1041425, 25680, 544, 13, 70, "ProofInfoBody",
 CellID->193335281]
}, Closed]],
Cell[CellGroupData[{
Cell[1042006, 25698, 5504, 140, 70, "ProofInfo",
 CellTags->{"Proof|ID:51392774", "Proof|ID:51392774-1"},
 CellID->336233968],
Cell[1047513, 25840, 477, 12, 70, "ProofInfoBody",
 CellID->2402072]
}, Closed]],
Cell[CellGroupData[{
Cell[1048027, 25857, 5202, 134, 70, "ProofInfo",
 CellTags->{"Proof|ID:488904469", "Proof|ID:488904469-1"},
 CellID->318992087],
Cell[1053232, 25993, 480, 12, 70, "ProofInfoBody",
 CellID->558690740]
}, Closed]],
Cell[CellGroupData[{
Cell[1053749, 26010, 5202, 134, 70, "ProofInfo",
 CellTags->{"Proof|ID:469346824", "Proof|ID:469346824-1"},
 CellID->231394775],
Cell[1058954, 26146, 480, 12, 70, "ProofInfoBody",
 CellID->467357679]
}, Closed]],
Cell[1059449, 26161, 47, 1, 70, "OpenEnvironment",
 CellID->337203903],
Cell[CellGroupData[{
Cell[1059521, 26166, 377, 9, 70, "EnvironmentHeader",
 CellID->657296789],
Cell[1059901, 26177, 782, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:195073923", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->195073923],
Cell[1060686, 26199, 1318, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:18535645", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Succeeds] - \[ScriptCapitalZ]n"},
 CellID->18535645],
Cell[1062007, 26239, 2191, 67, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:155348029", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
   "\[Succeeds] - \[RawStar] \[ScriptCapitalZ]n"},
 CellID->155348029],
Cell[1064201, 26308, 1384, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:220895581", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "\[Succeeds] + \[ScriptCapitalZ]n"},
 CellID->220895581],
Cell[1065588, 26348, 64, 1, 70, "EndEnvironmentMarker",
 CellID->38173617]
}, Open  ]],
Cell[1065667, 26352, 47, 1, 70, "CloseEnvironment",
 CellID->45785939],
Cell[CellGroupData[{
Cell[1065739, 26357, 4870, 127, 70, "ProofInfo",
 CellTags->{"Proof|ID:18535645", "Proof|ID:18535645-1"},
 CellID->255392256],
Cell[1070612, 26486, 479, 12, 70, "ProofInfoBody",
 CellID->600071146]
}, Closed]],
Cell[CellGroupData[{
Cell[1071128, 26503, 7643, 183, 70, "ProofInfo",
 CellTags->{"Proof|ID:155348029", "Proof|ID:155348029-1"},
 CellID->101870138],
Cell[1078774, 26688, 480, 12, 70, "ProofInfoBody",
 CellID->250604527]
}, Closed]],
Cell[CellGroupData[{
Cell[1079291, 26705, 4665, 123, 70, "ProofInfo",
 CellTags->{"Proof|ID:220895581", "Proof|ID:220895581-1"},
 CellID->11546098],
Cell[1083959, 26830, 479, 12, 70, "ProofInfoBody",
 CellID->21689854]
}, Closed]],
Cell[1084453, 26845, 46, 1, 70, "OpenEnvironment",
 CellID->32514940],
Cell[CellGroupData[{
Cell[1084524, 26850, 429, 10, 70, "EnvironmentHeader",
 CellID->674314239],
Cell[1084956, 26862, 757, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:108166851", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->108166851],
Cell[1085716, 26883, 1212, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:166132041", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
   "gcd \[VerticalBar] \[RawStar] \[ScriptCapitalZ]n"},
 CellID->166132041],
Cell[1086931, 26920, 1980, 58, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:414892396", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 1 \[ScriptCapitalZ]n"},
 CellID->414892396],
Cell[1088914, 26980, 1227, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:48379652", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 2 \[ScriptCapitalZ]n"},
 CellID->48379652],
Cell[1090144, 27016, 2105, 65, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:287224029", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 3 \[ScriptCapitalZ]n"},
 CellID->287224029],
Cell[1092252, 27083, 1918, 59, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:398117055", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient 4 \[ScriptCapitalZ]n"},
 CellID->398117055],
Cell[1094173, 27144, 2066, 63, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:568873110", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", 
   "mult1 \[RawStar] injective \[ScriptCapitalZ]n"},
 CellID->568873110],
Cell[1096242, 27209, 2334, 72, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:347198673", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 sufficient \[ScriptCapitalZ]n"},
 CellID->347198673],
Cell[1098579, 27283, 65, 1, 70, "EndEnvironmentMarker",
 CellID->209695896]
}, Open  ]],
Cell[1098659, 27287, 47, 1, 70, "CloseEnvironment",
 CellID->26902409],
Cell[CellGroupData[{
Cell[1098731, 27292, 4070, 110, 70, "ProofInfo",
 CellTags->{"Proof|ID:166132041", "Proof|ID:166132041-1"},
 CellID->111518977],
Cell[1102804, 27404, 480, 12, 70, "ProofInfoBody",
 CellID->518018963]
}, Closed]],
Cell[CellGroupData[{
Cell[1103321, 27421, 7509, 182, 70, "ProofInfo",
 CellTags->{"Proof|ID:414892396", "Proof|ID:414892396-1"},
 CellID->924193601],
Cell[1110833, 27605, 480, 12, 70, "ProofInfoBody",
 CellID->694816149]
}, Closed]],
Cell[CellGroupData[{
Cell[1111350, 27622, 4796, 124, 70, "ProofInfo",
 CellTags->{"Proof|ID:48379652", "Proof|ID:48379652-1"},
 CellID->384976109],
Cell[1116149, 27748, 479, 12, 70, "ProofInfoBody",
 CellID->382554458]
}, Closed]],
Cell[CellGroupData[{
Cell[1116665, 27765, 6459, 160, 70, "ProofInfo",
 CellTags->{"Proof|ID:287224029", "Proof|ID:287224029-1"},
 CellID->144370346],
Cell[1123127, 27927, 480, 12, 70, "ProofInfoBody",
 CellID->265857267]
}, Closed]],
Cell[CellGroupData[{
Cell[1123644, 27944, 6702, 170, 70, "ProofInfo",
 CellTags->{"Proof|ID:398117055", "Proof|ID:398117055-1"},
 CellID->168110018],
Cell[1130349, 28116, 480, 12, 70, "ProofInfoBody",
 CellID->367098984]
}, Closed]],
Cell[CellGroupData[{
Cell[1130866, 28133, 7651, 184, 70, "ProofInfo",
 CellTags->{"Proof|ID:568873110", "Proof|ID:568873110-1"},
 CellID->3033029],
Cell[1138520, 28319, 480, 12, 70, "ProofInfoBody",
 CellID->384543067]
}, Closed]],
Cell[CellGroupData[{
Cell[1139037, 28336, 7549, 184, 70, "ProofInfo",
 CellTags->{"Proof|ID:347198673", "Proof|ID:347198673-1"},
 CellID->636035778],
Cell[1146589, 28522, 480, 12, 70, "ProofInfoBody",
 CellID->537098177]
}, Closed]],
Cell[1147084, 28537, 47, 1, 70, "OpenEnvironment",
 CellID->599981467],
Cell[CellGroupData[{
Cell[1147156, 28542, 396, 9, 70, "EnvironmentHeader",
 CellID->124267281],
Cell[1147555, 28553, 776, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:25008217", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->25008217],
Cell[1148334, 28575, 1126, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:761196101", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isReducible \[ScriptCapitalZ]n"},
 CellID->761196101],
Cell[1149463, 28608, 1774, 51, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:216211003", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isReducible' \[ScriptCapitalZ]n"},
 CellID->216211003],
Cell[1151240, 28661, 2214, 66, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:431638777", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "gcd \[Succeeds] \[ScriptCapitalZ]n"},
 CellID->431638777],
Cell[1153457, 28729, 65, 1, 70, "EndEnvironmentMarker",
 CellID->379545476]
}, Open  ]],
Cell[1153537, 28733, 47, 1, 70, "CloseEnvironment",
 CellID->25000160],
Cell[CellGroupData[{
Cell[1153609, 28738, 4382, 115, 70, "ProofInfo",
 CellTags->{"Proof|ID:761196101", "Proof|ID:761196101-1"},
 CellID->46321884],
Cell[1157994, 28855, 480, 12, 70, "ProofInfoBody",
 CellID->473334112]
}, Closed]],
Cell[CellGroupData[{
Cell[1158511, 28872, 7302, 176, 70, "ProofInfo",
 CellTags->{"Proof|ID:216211003", "Proof|ID:216211003-1"},
 CellID->182622923],
Cell[1165816, 29050, 476, 12, 70, "ProofInfoBody",
 CellID->72865]
}, Closed]],
Cell[CellGroupData[{
Cell[1166329, 29067, 7589, 183, 70, "ProofInfo",
 CellTags->{"Proof|ID:431638777", "Proof|ID:431638777-1"},
 CellID->607642207],
Cell[1173921, 29252, 480, 12, 70, "ProofInfoBody",
 CellID->112691637]
}, Closed]],
Cell[1174416, 29267, 47, 1, 70, "OpenEnvironment",
 CellID->337195016],
Cell[CellGroupData[{
Cell[1174488, 29272, 427, 11, 70, "EnvironmentHeader",
 CellID->379952792],
Cell[1174918, 29285, 755, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:219886779", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->219886779],
Cell[1175676, 29306, 2638, 73, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:277360974", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isCommonReducible \[ScriptCapitalZ]n"},
 CellID->277360974],
Cell[1178317, 29381, 1584, 46, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:112584880", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR unary \[ScriptCapitalZ]n"},
 CellID->112584880],
Cell[1179904, 29429, 2030, 57, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:195146313", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR identical \[ScriptCapitalZ]n"},
 CellID->195146313],
Cell[1181937, 29488, 2011, 56, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:6406785", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR distinct' \[ScriptCapitalZ]n"},
 CellID->6406785],
Cell[1183951, 29546, 2356, 66, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:87852126", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR distinct \[ScriptCapitalZ]n"},
 CellID->87852126],
Cell[1186310, 29614, 2370, 68, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:37400195", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR \[ScriptCapitalZ]n"},
 CellID->37400195],
Cell[1188683, 29684, 64, 1, 70, "EndEnvironmentMarker",
 CellID->43844530]
}, Open  ]],
Cell[1188762, 29688, 48, 1, 70, "CloseEnvironment",
 CellID->512005801],
Cell[CellGroupData[{
Cell[1188835, 29693, 8376, 198, 70, "ProofInfo",
 CellTags->{"Proof|ID:277360974", "Proof|ID:277360974-1"},
 CellID->41026893],
Cell[1197214, 29893, 480, 12, 70, "ProofInfoBody",
 CellID->187450513]
}, Closed]],
Cell[CellGroupData[{
Cell[1197731, 29910, 4953, 127, 70, "ProofInfo",
 CellTags->{"Proof|ID:112584880", "Proof|ID:112584880-1"},
 CellID->228914149],
Cell[1202687, 30039, 546, 13, 70, "ProofInfoBody",
 CellID->255363086]
}, Closed]],
Cell[CellGroupData[{
Cell[1203270, 30057, 7323, 178, 70, "ProofInfo",
 CellTags->{"Proof|ID:195146313", "Proof|ID:195146313-1"},
 CellID->270250321],
Cell[1210596, 30237, 524, 13, 70, "ProofInfoBody",
 CellID->366403582]
}, Closed]],
Cell[CellGroupData[{
Cell[1211157, 30255, 8299, 200, 70, "ProofInfo",
 CellTags->{"Proof|ID:6406785", "Proof|ID:6406785-1"},
 CellID->295234282],
Cell[1219459, 30457, 477, 12, 70, "ProofInfoBody",
 CellID->90395247]
}, Closed]],
Cell[CellGroupData[{
Cell[1219973, 30474, 9482, 222, 70, "ProofInfo",
 CellTags->{"Proof|ID:87852126", "Proof|ID:87852126-1"},
 CellID->306702955],
Cell[1229458, 30698, 479, 12, 70, "ProofInfoBody",
 CellID->453994183]
}, Closed]],
Cell[CellGroupData[{
Cell[1229974, 30715, 9832, 228, 70, "ProofInfo",
 CellTags->{"Proof|ID:37400195", "Proof|ID:37400195-1"},
 CellID->282288311],
Cell[1239809, 30945, 479, 12, 70, "ProofInfoBody",
 CellID->307661666]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1240337, 30963, 235, 4, 70, "Subsubsection",
 CellID->73381205],
Cell[1240575, 30969, 46, 1, 70, "OpenEnvironment",
 CellID->91360536],
Cell[CellGroupData[{
Cell[1240646, 30974, 391, 9, 70, "EnvironmentHeader",
 CellID->154663064],
Cell[1241040, 30985, 892, 23, 70, "GlobalDeclaration",
 CellTags->{
  "ID:14129606", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->14129606],
Cell[1241935, 31010, 1382, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:126313666", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "mult1 - \[ScriptCapitalZ]n"},
 CellID->126313666],
Cell[1243320, 31051, 1331, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:888879353", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "areCorrelative 1 \[ScriptCapitalZ]n"},
 CellID->888879353],
Cell[1244654, 31090, 1431, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:23406014", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "areCorrelative 2 \[ScriptCapitalZ]n"},
 CellID->23406014],
Cell[1246088, 31130, 65, 1, 70, "EndEnvironmentMarker",
 CellID->179715897]
}, Open  ]],
Cell[1246168, 31134, 47, 1, 70, "CloseEnvironment",
 CellID->77124816],
Cell[CellGroupData[{
Cell[1246240, 31139, 6986, 170, 70, "ProofInfo",
 CellTags->{"Proof|ID:126313666", "Proof|ID:126313666-1"},
 CellID->79969318],
Cell[1253229, 31311, 480, 12, 70, "ProofInfoBody",
 CellID->315040505]
}, Closed]],
Cell[CellGroupData[{
Cell[1253746, 31328, 6394, 156, 70, "ProofInfo",
 CellTags->{"Proof|ID:888879353", "Proof|ID:888879353-1"},
 CellID->341588],
Cell[1260143, 31486, 480, 12, 70, "ProofInfoBody",
 CellID->625993774]
}, Closed]],
Cell[CellGroupData[{
Cell[1260660, 31503, 6392, 156, 70, "ProofInfo",
 CellTags->{"Proof|ID:23406014", "Proof|ID:23406014-1"},
 CellID->386167612],
Cell[1267055, 31661, 479, 12, 70, "ProofInfoBody",
 CellID->145559439]
}, Closed]],
Cell[1267549, 31676, 47, 1, 70, "OpenEnvironment",
 CellID->909951923],
Cell[CellGroupData[{
Cell[1267621, 31681, 470, 10, 70, "EnvironmentHeader",
 CellID->1129973761],
Cell[1268094, 31693, 859, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:325383468", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R5 \[ScriptCapitalZ]n"},
 CellID->325383468],
Cell[1268956, 31716, 839, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1783775451", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R6 \[ScriptCapitalZ]n"},
 CellID->1783775451],
Cell[1269798, 31739, 65, 1, 70, "EndEnvironmentMarker",
 CellID->955608461]
}, Open  ]],
Cell[1269878, 31743, 48, 1, 70, "CloseEnvironment",
 CellID->521644330],
Cell[CellGroupData[{
Cell[1269951, 31748, 2341, 70, 70, "ProofInfo",
 CellTags->{"Proof|ID:325383468", "Proof|ID:325383468-1"},
 CellID->4554534],
Cell[1272295, 31820, 478, 12, 70, "ProofInfoBody",
 CellID->2823860]
}, Closed]],
Cell[CellGroupData[{
Cell[1272810, 31837, 2352, 71, 70, "ProofInfo",
 CellTags->{"Proof|ID:1783775451", "Proof|ID:1783775451-1"},
 CellID->190836234],
Cell[1275165, 31910, 479, 12, 70, "ProofInfoBody",
 CellID->5677791]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1275693, 31928, 184, 3, 70, "Subsubsection",
 CellID->124396989],
Cell[1275880, 31933, 46, 1, 70, "OpenEnvironment",
 CellID->12240399],
Cell[CellGroupData[{
Cell[1275951, 31938, 387, 9, 70, "EnvironmentHeader",
 CellID->576684220],
Cell[1276341, 31949, 2235, 68, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:394176110", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR existence \[ScriptCapitalZ]n"},
 CellID->394176110],
Cell[1278579, 32019, 2049, 64, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:232518088", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isMNTCR unary existence \[ScriptCapitalZ]n"},
 CellID->232518088],
Cell[1280631, 32085, 65, 1, 70, "EndEnvironmentMarker",
 CellID->135443090]
}, Open  ]],
Cell[1280711, 32089, 48, 1, 70, "CloseEnvironment",
 CellID->257333253],
Cell[CellGroupData[{
Cell[1280784, 32094, 7689, 180, 70, "ProofInfo",
 CellTags->{"Proof|ID:394176110", "Proof|ID:394176110-1"},
 CellID->648217336],
Cell[1288476, 32276, 480, 12, 70, "ProofInfoBody",
 CellID->127336303]
}, Closed]],
Cell[CellGroupData[{
Cell[1288993, 32293, 6269, 154, 70, "ProofInfo",
 CellTags->{"Proof|ID:232518088", "Proof|ID:232518088-1"},
 CellID->424300880],
Cell[1295265, 32449, 479, 12, 70, "ProofInfoBody",
 CellID->49071767]
}, Closed]],
Cell[1295759, 32464, 47, 1, 70, "OpenEnvironment",
 CellID->644279200],
Cell[CellGroupData[{
Cell[1295831, 32469, 398, 9, 70, "EnvironmentHeader",
 CellID->334567729],
Cell[1296232, 32480, 838, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2147206476", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R7 \[ScriptCapitalZ]n"},
 CellID->2147206476],
Cell[1297073, 32503, 837, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1971443720", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R8 \[ScriptCapitalZ]n"},
 CellID->1971443720],
Cell[1297913, 32526, 64, 1, 70, "EndEnvironmentMarker",
 CellID->79251623]
}, Open  ]],
Cell[1297992, 32530, 48, 1, 70, "CloseEnvironment",
 CellID->112564214],
Cell[CellGroupData[{
Cell[1298065, 32535, 2351, 71, 70, "ProofInfo",
 CellTags->{"Proof|ID:2147206476", "Proof|ID:2147206476-1"},
 CellID->31414230],
Cell[1300419, 32608, 481, 12, 70, "ProofInfoBody",
 CellID->347321696]
}, Closed]],
Cell[CellGroupData[{
Cell[1300937, 32625, 2352, 71, 70, "ProofInfo",
 CellTags->{"Proof|ID:1971443720", "Proof|ID:1971443720-1"},
 CellID->239044020],
Cell[1303292, 32698, 480, 12, 70, "ProofInfoBody",
 CellID->28742966]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1303821, 32716, 190, 3, 70, "Subsubsection",
 CellID->106088833],
Cell[1304014, 32721, 48, 1, 70, "OpenEnvironment",
 CellID->1543467287],
Cell[CellGroupData[{
Cell[1304087, 32726, 524, 13, 70, "EnvironmentHeader",
 CellID->1491276973],
Cell[1304614, 32741, 957, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:727194704", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isReductionRing \[ScriptCapitalZ]n"},
 CellID->727194704],
Cell[1305574, 32767, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1443393611]
}, Open  ]],
Cell[1305655, 32771, 49, 1, 70, "CloseEnvironment",
 CellID->2047014872],
Cell[CellGroupData[{
Cell[1305729, 32776, 2403, 71, 70, "ProofInfo",
 CellTags->{"Proof|ID:727194704", "Proof|ID:727194704-1"},
 CellID->278604702],
Cell[1308135, 32849, 478, 12, 70, "ProofInfoBody",
 CellID->8265060]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1308674, 32868, 201, 3, 30, "Subsection",
 CellID->24901450],
Cell[CellGroupData[{
Cell[1308900, 32875, 180, 3, 28, "Subsubsection",
 CellID->415571348],
Cell[1309083, 32880, 47, 1, 70, "OpenEnvironment",
 CellID->319526061],
Cell[CellGroupData[{
Cell[1309155, 32885, 390, 9, 70, "EnvironmentHeader",
 CellID->8268143],
Cell[1309548, 32896, 775, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:73858040", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb"},
 CellID->73858040],
Cell[1310326, 32918, 2799, 79, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:502941593", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm ternary \[ScriptCapitalZ]n"},
 CellID->502941593],
Cell[1313128, 32999, 1498, 45, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:691632805", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm aux \[ScriptCapitalZ]n"},
 CellID->691632805],
Cell[1314629, 33046, 980, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:290458459", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "membership rdm \[ScriptCapitalZ]n"},
 CellID->290458459],
Cell[1315612, 33073, 1186, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:38241486", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm cases \[ScriptCapitalZ]n"},
 CellID->38241486],
Cell[1316801, 33107, 1103, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:140983642", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm \[RawStar] \[ScriptCapitalZ]n"},
 CellID->140983642],
Cell[1317907, 33139, 979, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:33831774", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "rdm 0 \[ScriptCapitalZ]n"},
 CellID->33831774],
Cell[1318889, 33166, 64, 1, 70, "EndEnvironmentMarker",
 CellID->16043932]
}, Open  ]],
Cell[1318968, 33170, 48, 1, 70, "CloseEnvironment",
 CellID->445628779],
Cell[CellGroupData[{
Cell[1319041, 33175, 9489, 223, 70, "ProofInfo",
 CellTags->{"Proof|ID:502941593", "Proof|ID:502941593-1"},
 CellID->32304572],
Cell[1328533, 33400, 480, 12, 70, "ProofInfoBody",
 CellID->251898129]
}, Closed]],
Cell[CellGroupData[{
Cell[1329050, 33417, 6293, 157, 70, "ProofInfo",
 CellTags->{"Proof|ID:691632805", "Proof|ID:691632805-1"},
 CellID->23114620],
Cell[1335346, 33576, 479, 12, 70, "ProofInfoBody",
 CellID->91619570]
}, Closed]],
Cell[CellGroupData[{
Cell[1335862, 33593, 4053, 108, 70, "ProofInfo",
 CellTags->{"Proof|ID:290458459", "Proof|ID:290458459-1"},
 CellID->155701127],
Cell[1339918, 33703, 571, 14, 70, "ProofInfoBody",
 CellID->8836418]
}, Closed]],
Cell[CellGroupData[{
Cell[1340526, 33722, 5330, 135, 70, "ProofInfo",
 CellTags->{"Proof|ID:38241486", "Proof|ID:38241486-1"},
 CellID->904484803],
Cell[1345859, 33859, 477, 12, 70, "ProofInfoBody",
 CellID->2214627]
}, Closed]],
Cell[CellGroupData[{
Cell[1346373, 33876, 5001, 128, 70, "ProofInfo",
 CellTags->{"Proof|ID:140983642", "Proof|ID:140983642-1"},
 CellID->492763287],
Cell[1351377, 34006, 480, 12, 70, "ProofInfoBody",
 CellID->440487519]
}, Closed]],
Cell[CellGroupData[{
Cell[1351894, 34023, 3728, 102, 70, "ProofInfo",
 CellTags->{"Proof|ID:33831774", "Proof|ID:33831774-1"},
 CellID->447394967],
Cell[1355625, 34127, 478, 12, 70, "ProofInfoBody",
 CellID->35066950]
}, Closed]],
Cell[1356118, 34142, 47, 1, 70, "OpenEnvironment",
 CellID->840971231],
Cell[CellGroupData[{
Cell[1356190, 34147, 394, 9, 70, "EnvironmentHeader",
 CellID->11062889],
Cell[1356587, 34158, 839, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1376076253", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R9 \[ScriptCapitalZ]n"},
 CellID->1376076253],
Cell[1357429, 34181, 839, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:125450892", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R10 \[ScriptCapitalZ]n"},
 CellID->125450892],
Cell[1358271, 34204, 839, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:189921319", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R11 \[ScriptCapitalZ]n"},
 CellID->189921319],
Cell[1359113, 34227, 839, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:667306596", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R12 \[ScriptCapitalZ]n"},
 CellID->667306596],
Cell[1359955, 34250, 813, 20, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:194342088", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "R13 \[ScriptCapitalZ]n"},
 CellID->194342088],
Cell[1360771, 34272, 64, 1, 70, "EndEnvironmentMarker",
 CellID->72432757]
}, Open  ]],
Cell[1360850, 34276, 48, 1, 70, "CloseEnvironment",
 CellID->394157218],
Cell[CellGroupData[{
Cell[1360923, 34281, 2352, 71, 70, "ProofInfo",
 CellTags->{"Proof|ID:1376076253", "Proof|ID:1376076253-1"},
 CellID->793462086],
Cell[1363278, 34354, 481, 12, 70, "ProofInfoBody",
 CellID->629931647]
}, Closed]],
Cell[CellGroupData[{
Cell[1363796, 34371, 2347, 70, 70, "ProofInfo",
 CellTags->{"Proof|ID:125450892", "Proof|ID:125450892-1"},
 CellID->207385051],
Cell[1366146, 34443, 480, 12, 70, "ProofInfoBody",
 CellID->178882523]
}, Closed]],
Cell[CellGroupData[{
Cell[1366663, 34460, 2347, 70, 70, "ProofInfo",
 CellTags->{"Proof|ID:189921319", "Proof|ID:189921319-1"},
 CellID->221524119],
Cell[1369013, 34532, 480, 12, 70, "ProofInfoBody",
 CellID->453410172]
}, Closed]],
Cell[CellGroupData[{
Cell[1369530, 34549, 2347, 70, 70, "ProofInfo",
 CellTags->{"Proof|ID:667306596", "Proof|ID:667306596-1"},
 CellID->389476407],
Cell[1371880, 34621, 479, 12, 70, "ProofInfoBody",
 CellID->30462548]
}, Closed]],
Cell[CellGroupData[{
Cell[1372396, 34638, 2347, 70, 70, "ProofInfo",
 CellTags->{"Proof|ID:194342088", "Proof|ID:194342088-1"},
 CellID->102872355],
Cell[1374746, 34710, 480, 12, 70, "ProofInfoBody",
 CellID->107661161]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1375275, 34728, 192, 3, 22, "Subsubsection",
 CellID->365692985],
Cell[1375470, 34733, 47, 1, 12, "OpenEnvironment",
 CellID->289788666],
Cell[CellGroupData[{
Cell[1375542, 34738, 451, 12, 43, "EnvironmentHeader",
 CellID->315438148],
Cell[1375996, 34752, 875, 22, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:403270264", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings.nb", "isAlgoReductionRing \[ScriptCapitalZ]n"},
 CellID->403270264],
Cell[1376874, 34776, 65, 1, 26, "EndEnvironmentMarker",
 CellID->379063054]
}, Open  ]],
Cell[1376954, 34780, 47, 1, 16, "CloseEnvironment",
 CellID->23977526],
Cell[CellGroupData[{
Cell[1377026, 34785, 2426, 72, 28, "ProofInfo",
 CellTags->{"Proof|ID:403270264", "Proof|ID:403270264-1"},
 CellID->78837715],
Cell[1379455, 34859, 479, 12, 70, "ProofInfoBody",
 CellID->30002591]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
