(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 9.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[   1360892,      39187]
NotebookOptionsPosition[    968047,      28652]
NotebookOutlinePosition[   1253296,      35974]
CellTagsIndexPosition[   1223862,      35226]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["Numbers", "Title",
 CellChangeTimes->{{3.6603708881422796`*^9, 3.6603708893122816`*^9}},
 CellID->21108537],

Cell["Natural Numbers & Integers", "Subtitle",
 CellChangeTimes->{{3.660370894694291*^9, 3.6603708988906984`*^9}, {
  3.6603729490061283`*^9, 3.660372949271329*^9}},
 CellID->138585177],

Cell[CellGroupData[{

Cell["Preliminaries", "Section",
 CellChangeTimes->{{3.6443084245106*^9, 3.644308427083333*^9}},
 CellID->909929215],

Cell[CellGroupData[{

Cell["Description", "Subsection",
 CellChangeTimes->{{3.6604716785383368`*^9, 3.66047168621355*^9}, {
  3.6611652326197357`*^9, 3.66116523461654*^9}},
 CellID->731805898],

Cell["\<\
This notebook introduces the sets of natural numbers and integers, together \
with the most important arithmetical functions, in a systematic way. Indeed, \
the principle employed here can easily be extended to construct the \
rationals, reals, and even complex numbers in pretty much the same spirit on \
top of the integers as well.\
\>", "Text",
 CellChangeTimes->{{3.6604717039195814`*^9, 3.660471730034027*^9}, {
  3.6604717626380844`*^9, 3.6604718678758693`*^9}},
 CellID->581780680],

Cell[TextData[{
 "First, the set of set-theory natural numbers, natSet, is defined by a \
least-fixpoint-construction (see Section \[OpenCurlyDoubleQuote]Natural \
Numbers\[CloseCurlyDoubleQuote] for more details). Then, the usual operations \
on natural numbers, namely ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["-", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["\[GreaterEqual]", TraditionalForm]]],
 " are introduced as functorial functions of functor Nat.\nOn top of that, \
the set-theory integers are defined as equivalence classes of ordered pairs \
of natural numbers, with the intended meaning that a representation ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"m", ",", "n"}], ")"}], TraditionalForm]]],
 " stands for the integer ",
 Cell[BoxData[
  FormBox[
   RowBox[{"m", "-", "n"}], TraditionalForm]]],
 ", and the equivalence relation is of course ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{"m", ",", "n"}], ")"}], "~", 
       RowBox[{"(", 
        RowBox[{"k", ",", "l"}], ")"}]}], "\[DoubleLongLeftRightArrow]", 
      "m"}], "+", "l"}], "=", 
    RowBox[{"k", "+", "n"}]}], TraditionalForm]]],
 ".\nHaving constructed the set-theory integers, the integers \
\[DoubleStruckCapitalZ] are finally obtained by wrapping each ",
 "set-theory",
 "-integer, which is a set, into the \[OpenCurlyDoubleQuote]datatype\
\[CloseCurlyDoubleQuote] constructor Int, only to ensure that integers are no \
sets. This has the effect that functions can be overloaded for sets and \
integers without ambiguity or inconsistency.\nFinally, the set of natural \
numbers, ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalN]", "0"], TraditionalForm]]],
 ", is simply an interval of integers, as defined in notebook \
\[OpenCurlyDoubleQuote]NatInt.nb\[CloseCurlyDoubleQuote]."
}], "Text",
 CellChangeTimes->{{3.660471890464709*^9, 3.6604723627799387`*^9}, {
  3.6604725827871246`*^9, 3.66047259114874*^9}, {3.6613375269287434`*^9, 
  3.661337561233204*^9}},
 CellID->902968620],

Cell[TextData[{
 "It should be easy to see that one does not need to stop at the integers, \
but can go along introducing first the set-theory-rationals, then the \
set-theory-reals, and finally the set of complex numbers \
\[DoubleStruckCapitalC] are pairs of set-theory-reals. Ideally, complex \
numbers should not be plain ordered pairs from set theory, but in analogy to \
the Int-constructor should be represented as something like ",
 Cell[BoxData[
  FormBox[
   RowBox[{"Complex", "[", 
    RowBox[{"a", ",", "b"}], "]"}], TraditionalForm]]],
 ", where Complex is again a datatype-like free constructor (which, in \
particular, does not construct sets).\nHaving ",
 Cell[BoxData[
  FormBox["\[DoubleStruckCapitalC]", TraditionalForm]]],
 ", the set of real numbers \[DoubleStruckCapitalR] can easily be obtained by \
separation, and then the set of integers \[DoubleStruckCapitalZ] again by a \
least-fixpoint construction with \[DoubleStruckCapitalR] as the underlying \
universe and the productions ",
 Cell[BoxData[
  FormBox["0", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"+", "1"}], TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"-", "1"}], TraditionalForm]]],
 ". ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalN]", "0"], TraditionalForm]]],
 ", as well as ",
 Cell[BoxData[
  FormBox["\[DoubleStruckCapitalN]", TraditionalForm]]],
 " are again mere intervals of \[DoubleStruckCapitalZ], and \
\[DoubleStruckCapitalQ], finally, is obtained by separation from \
\[DoubleStruckCapitalR] as the set of all real numbers that are equal to some \
",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "/", "b"}], TraditionalForm]]],
 " for integers ",
 Cell[BoxData[
  FormBox["a", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 ".\nSummarizing, one first needs to go \[OpenCurlyDoubleQuote]upwards\
\[CloseCurlyDoubleQuote] in the hierarchy, starting from natSet and heading \
towards \[DoubleStruckCapitalC], and then go \[OpenCurlyDoubleQuote]downwards\
\[CloseCurlyDoubleQuote] again to construct \[DoubleStruckCapitalR], \
\[DoubleStruckCapitalZ], ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalN]", "0"], TraditionalForm]]],
 " and \[DoubleStruckCapitalQ] as subsets of \[DoubleStruckCapitalC].\nOf \
course, one could start with ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["\[DoubleStruckCapitalN]", "0"], TraditionalForm]]],
 " instead of the set-theory-naturals right way, and then immediately define \
\[DoubleStruckCapitalZ] instead of the ",
 "set-theory",
 "-integers, but the advantage of the proposed method lies in the fact that \
the number domains thus constructed are not independent \
\[OpenCurlyDoubleQuote]types\[CloseCurlyDoubleQuote], but satisfy ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["\[DoubleStruckCapitalN]", "0"], "\[Subset]", 
    "\[DoubleStruckCapitalZ]", "\[Subset]", "\[DoubleStruckCapitalQ]", 
    "\[Subset]", "\[DoubleStruckCapitalR]", "\[Subset]", 
    "\[DoubleStruckCapitalC]"}], TraditionalForm]]],
 " \[Dash] and this is particularly important taking into account that number \
literals, such as 0 and 1, as well as arithmetic operations, such as +, ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " and \[GreaterEqual], shall be usable in all these domains without the \
danger of ambiguity or even inconsistency. If number domains were independet \
\[OpenCurlyDoubleQuote]types\[CloseCurlyDoubleQuote], the consistency of the \
various definitions of, say, +, would have to be ",
 StyleBox["proved",
  FontSlant->"Italic"],
 ", and do not come \[OpenCurlyDoubleQuote]for free\[CloseCurlyDoubleQuote] \
as in the proposed method. The only drawback of that method is that, say, the \
natural number 0 actually is something ugly like ",
 Cell[BoxData[
  FormBox[
   RowBox[{"Complex", "[", 
    RowBox[{
     UnderscriptBox["0", "Real"], ",", 
     UnderscriptBox["0", "Real"]}], "]"}], TraditionalForm]]],
 ", where ",
 Cell[BoxData[
  FormBox[
   UnderscriptBox["0", "Real"], TraditionalForm]]],
 " in turn will probably be an equivalence class of Cauchy-sequences in \
\[DoubleStruckCapitalQ]. But, after all, once enough properties of 0 have \
been proved, its precise definition becomes immaterial anyway."
}], "Text",
 CellChangeTimes->{{3.660472377303564*^9, 3.6604725614306874`*^9}, {
  3.6604725946119456`*^9, 3.6604734603322663`*^9}, {3.660473726999135*^9, 
  3.6604737281847367`*^9}, {3.661337565819612*^9, 3.6613375739628263`*^9}},
 CellID->438330771]
}, Closed]],

Cell[CellGroupData[{

Cell["Theory Packages", "Subsection",
 CellChangeTimes->{{3.6566705778956037`*^9, 3.656670580376008*^9}},
 CellID->102004898],

Cell[BoxData[
 RowBox[{
  RowBox[{"$PackageDirectory", "=", 
   RowBox[{"FileNameJoin", "[", 
    RowBox[{"{", 
     RowBox[{
     "$RootDirectory", ",", "\"\<scratch\>\"", ",", "\"\<amaletzk\>\"", ",", 
      "\"\<Theorema Environment Data\>\""}], "}"}], "]"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.666438783233775*^9, 3.666438821083324*^9}, 
   3.667735111181109*^9},
 CellID->498486718],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<LogicSets.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.656670589533224*^9, 3.6566705941196322`*^9}, {
   3.656767086897564*^9, 3.656767101545989*^9}, 3.666439315853882*^9},
 CellID->406742757],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<AlgebraicStructures.m\>\""}], 
    "}"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.656670589533224*^9, 3.6566705941196322`*^9}, {
   3.656767086897564*^9, 3.656767101545989*^9}, {3.6611671321434727`*^9, 
   3.661167138742284*^9}, 3.6664393170944777`*^9},
 CellID->6977455],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<Numbers.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.656670589533224*^9, 3.6566705941196322`*^9}, {
  3.667557879147784*^9, 3.667557883027082*^9}},
 CellID->139193289],

Cell[BoxData[
 RowBox[{"MakeTheoryPackage", "[", "]"}]], "Input",
 CellChangeTimes->{{3.655458089638938*^9, 3.655458093336144*^9}},
 CellID->182306024],

Cell[BoxData[
 RowBox[{"UpdateTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<AlgebraicStructures.m\>\""}], 
    "}"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.6555627795009193`*^9, 3.6555627842121277`*^9}, 
   3.6675578944517393`*^9},
 CellID->20362838],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Names", "[", "\"\<Global`*$TM\>\"", "]"}]], "Input",
 CellChangeTimes->{{3.6592618519471025`*^9, 3.6592618631011224`*^9}, {
   3.659681175379635*^9, 3.659681185706853*^9}, 3.659683710620489*^9, 
   3.65968441725893*^9, 3.6597695015379353`*^9, {3.6597698034056654`*^9, 
   3.659769808569275*^9}, {3.6597702854721127`*^9, 3.659770290573322*^9}, {
   3.659848272310069*^9, 3.6598483467534*^9}, {3.6598515848824887`*^9, 
   3.6598516763142495`*^9}, {3.6598555136659927`*^9, 3.659855522090008*^9}, {
   3.6598559051642804`*^9, 3.659855919516306*^9}, {3.6602862205309095`*^9, 
   3.6602862222781124`*^9}, {3.660479643354128*^9, 3.660479646380533*^9}, 
   3.6675579076910954`*^9, 3.6677357112054243`*^9},
 CellID->20666029],

Cell[BoxData[
 RowBox[{"{", "}"}]], "Output",
 CellChangeTimes->{{3.6592618606675177`*^9, 3.659261863553523*^9}, {
   3.6596811716824284`*^9, 3.659681186315254*^9}, {3.6596812817094216`*^9, 
   3.6596813028006587`*^9}, 3.6596819060879183`*^9, 3.6596833252354116`*^9, 
   3.659683558659622*^9, {3.6596837085924854`*^9, 3.6596837109480896`*^9}, {
   3.6596844150905266`*^9, 3.6596844182417316`*^9}, 3.6597682216870866`*^9, {
   3.65976949874553*^9, 3.6597695020839357`*^9}, {3.6597698011592617`*^9, 
   3.6597698090216756`*^9}, {3.6597702823677073`*^9, 3.659770291290923*^9}, {
   3.6598482697828646`*^9, 3.659848283230088*^9}, {3.6598483257089624`*^9, 
   3.6598483475958014`*^9}, {3.659851582526885*^9, 3.659851591793301*^9}, {
   3.6598516645050287`*^9, 3.659851677655852*^9}, 3.659853278048464*^9, {
   3.6598555077535825`*^9, 3.659855523962011*^9}, 3.659855883433442*^9, 
   3.6598559207019076`*^9, {3.659856164017535*^9, 3.6598561825211678`*^9}, 
   3.659856443586827*^9, 3.659856548996212*^9, 3.6598567131251*^9, 
   3.659856751204767*^9, 3.659856874446984*^9, 3.6598573942116966`*^9, {
   3.6598577350138955`*^9, 3.659857756947534*^9}, 3.659858388284853*^9, {
   3.6602167142756796`*^9, 3.6602167244780974`*^9}, 3.6602196234977903`*^9, 
   3.6602242186346664`*^9, {3.6602862153049*^9, 3.6602862245557165`*^9}, 
   3.66028736404112*^9, {3.6602875017827625`*^9, 3.6602875069931717`*^9}, 
   3.6602875499460473`*^9, 3.6602883894745226`*^9, 3.6603715096819983`*^9, 
   3.6603718119505296`*^9, 3.660372887741821*^9, 3.660373305732955*^9, 
   3.660388963232644*^9, {3.660395289716957*^9, 3.6603953095757923`*^9}, 
   3.6604747426127186`*^9, 3.6604781134162397`*^9, 3.6604782239892344`*^9, 
   3.6604788179944777`*^9, 3.660479647441335*^9, 3.66048049274642*^9, 
   3.6604811507711754`*^9, 3.6605462624642744`*^9, 3.6606436252656946`*^9, 
   3.661171298335791*^9, 3.6612335103527374`*^9, 3.6612335492154055`*^9, {
   3.6612360111497307`*^9, 3.6612360190277443`*^9}, 3.6612381009570045`*^9, 
   3.6612461018061156`*^9, 3.6612473526349125`*^9, 3.6612485772252645`*^9, 
   3.6612487820390244`*^9, 3.6612512473731556`*^9, 3.661253894236805*^9, 
   3.661322107646716*^9, 3.6613247100732975`*^9, 3.6613251703551064`*^9, 
   3.6613376826308174`*^9, 3.6661760926336117`*^9, 3.66773384899341*^9, 
   3.667735711835741*^9, 3.6677357809818993`*^9, 3.667736293705586*^9, 
   3.6678154093896837`*^9, 3.667815558877899*^9, {3.66781576998433*^9, 
   3.6678157785347223`*^9}, {3.667815810166841*^9, 3.667815831623155*^9}},
 CellID->569800073]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Theory Analysis", "Subsection",
 CellChangeTimes->{{3.640668830437483*^9, 3.640668834318727*^9}},
 CellID->1386333768],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"First", "[", 
  RowBox[{"Timing", "[", 
   RowBox[{"AnalyzeProofs", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"NotebookDirectory", "[", "]"}], ",", "\"\<Recursive\>\""}], 
       "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"FileNameJoin", "[", 
         RowBox[{"{", 
          RowBox[{
          "$HomeDirectory", ",", "\"\<Theorema Notebooks\>\"", ",", 
           "\"\<Reduction Ring Theory\>\""}], "}"}], "]"}], ",", 
        "\"\<Recursive\>\""}], "}"}]}], "}"}], "]"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.653223166066796*^9, 3.653223174615611*^9}, {
  3.655715257190379*^9, 3.6557152590467825`*^9}, {3.6561617938702183`*^9, 
  3.6561618097442417`*^9}, {3.6562164498823805`*^9, 3.656216451816784*^9}, {
  3.6617535657951183`*^9, 3.661753573298732*^9}, {3.666931622014642*^9, 
  3.666931675694009*^9}, {3.6669317194951277`*^9, 3.666931756112032*^9}},
 CellID->239327779],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"\"\>", 
   "\[InvisibleSpace]", "\<\"/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/\"\>", "\[InvisibleSpace]", "\<\"\\\" (\"\>", 
   "\[InvisibleSpace]", "1", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"18\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"", 
   "/home/amaletzk/Theorema Notebooks/ElementaryTheories/", "\" (", 1, " of ",
    "18", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557931989496*^9},
 CellID->1996807139],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"\"\>", 
   "\[InvisibleSpace]", "\<\"/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/AlgebraicStructures\"\>", 
   "\[InvisibleSpace]", "\<\"\\\" (\"\>", "\[InvisibleSpace]", "2", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"18\"\>", 
   "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"", 
   "/home/amaletzk/Theorema Notebooks/ElementaryTheories/AlgebraicStructures",
    "\" (", 2, " of ", "18", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557932023834*^9},
 CellID->1264118277],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"60\"\>"}],
  SequenceForm["  Reading proof ", 1, " of ", "60"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579320259123`*^9},
 CellID->1741093491],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"60\"\>"}],
  SequenceForm["  Reading proof ", 51, " of ", "60"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557932598054*^9},
 CellID->1767534958],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"\"\>", 
   "\[InvisibleSpace]", "\<\"/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Functors\"\>", 
   "\[InvisibleSpace]", "\<\"\\\" (\"\>", "\[InvisibleSpace]", "3", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"18\"\>", 
   "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"", 
   "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Functors", "\" (", 3,
    " of ", "18", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557932704708*^9},
 CellID->1570628282],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"36\"\>"}],
  SequenceForm["  Reading proof ", 1, " of ", "36"],
  Editable->False]], "Print",
 CellChangeTimes->{3.66755793270676*^9},
 CellID->1431522614],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"\"\>", 
   "\[InvisibleSpace]", "\<\"/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/LogicSets\"\>", 
   "\[InvisibleSpace]", "\<\"\\\" (\"\>", "\[InvisibleSpace]", "4", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"18\"\>", 
   "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"", 
   "/home/amaletzk/Theorema Notebooks/ElementaryTheories/LogicSets", "\" (", 
   4, " of ", "18", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557933127553*^9},
 CellID->46013314],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"203\"\>"}],
  SequenceForm["  Reading proof ", 1, " of ", "203"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557933173345*^9},
 CellID->413353689],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"203\"\>"}],
  SequenceForm["  Reading proof ", 51, " of ", "203"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557933640779*^9},
 CellID->1417273774],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"203\"\>"}],
  SequenceForm["  Reading proof ", 101, " of ", "203"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557934155181*^9},
 CellID->1414858982],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"203\"\>"}],
  SequenceForm["  Reading proof ", 151, " of ", "203"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557934657551*^9},
 CellID->632179620],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"203\"\>"}],
  SequenceForm["  Reading proof ", 201, " of ", "203"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557935350555*^9},
 CellID->445177995],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"\"\>", 
   "\[InvisibleSpace]", "\<\"/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/NatInt\"\>", 
   "\[InvisibleSpace]", "\<\"\\\" (\"\>", "\[InvisibleSpace]", "5", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"18\"\>", 
   "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"", 
   "/home/amaletzk/Theorema Notebooks/ElementaryTheories/NatInt", "\" (", 5, 
   " of ", "18", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557935394539*^9},
 CellID->256427531],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"300\"\>"}],
  SequenceForm["  Reading proof ", 1, " of ", "300"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579353964987`*^9},
 CellID->1702810782],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"300\"\>"}],
  SequenceForm["  Reading proof ", 51, " of ", "300"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579359161377`*^9},
 CellID->610477982],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"300\"\>"}],
  SequenceForm["  Reading proof ", 101, " of ", "300"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557936398465*^9},
 CellID->437563033],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"300\"\>"}],
  SequenceForm["  Reading proof ", 151, " of ", "300"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557936932382*^9},
 CellID->881701306],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"300\"\>"}],
  SequenceForm["  Reading proof ", 201, " of ", "300"],
  Editable->False]], "Print",
 CellChangeTimes->{3.66755793744473*^9},
 CellID->1547923834],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"300\"\>"}],
  SequenceForm["  Reading proof ", 251, " of ", "300"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557937940258*^9},
 CellID->1241639991],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"\"\>", 
   "\[InvisibleSpace]", "\<\"/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/NatIntExtended\"\>", 
   "\[InvisibleSpace]", "\<\"\\\" (\"\>", "\[InvisibleSpace]", "6", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"18\"\>", 
   "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"", 
   "/home/amaletzk/Theorema Notebooks/ElementaryTheories/NatIntExtended", 
   "\" (", 6, " of ", "18", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579384183273`*^9},
 CellID->1907575484],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"244\"\>"}],
  SequenceForm["  Reading proof ", 1, " of ", "244"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557938479732*^9},
 CellID->25744908],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"244\"\>"}],
  SequenceForm["  Reading proof ", 51, " of ", "244"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557938945589*^9},
 CellID->1066639784],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"244\"\>"}],
  SequenceForm["  Reading proof ", 101, " of ", "244"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557939478771*^9},
 CellID->1142307976],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"244\"\>"}],
  SequenceForm["  Reading proof ", 151, " of ", "244"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557940040897*^9},
 CellID->559336593],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"244\"\>"}],
  SequenceForm["  Reading proof ", 201, " of ", "244"],
  Editable->False]], "Print",
 CellChangeTimes->{3.66755794064913*^9},
 CellID->1038330147],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"\"\>", 
   "\[InvisibleSpace]", "\<\"/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers\"\>", 
   "\[InvisibleSpace]", "\<\"\\\" (\"\>", "\[InvisibleSpace]", "7", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"18\"\>", 
   "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"", 
   "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers", "\" (", 7, 
   " of ", "18", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579411499043`*^9},
 CellID->818971569],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"179\"\>"}],
  SequenceForm["  Reading proof ", 1, " of ", "179"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557941201511*^9},
 CellID->1237330945],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"179\"\>"}],
  SequenceForm["  Reading proof ", 51, " of ", "179"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557941699833*^9},
 CellID->265312282],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"179\"\>"}],
  SequenceForm["  Reading proof ", 101, " of ", "179"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557942211952*^9},
 CellID->2034025804],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"179\"\>"}],
  SequenceForm["  Reading proof ", 151, " of ", "179"],
  Editable->False]], "Print",
 CellChangeTimes->{3.66755794271898*^9},
 CellID->299202341],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"\"\>", 
   "\[InvisibleSpace]", "\<\"/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Sequences\"\>", 
   "\[InvisibleSpace]", "\<\"\\\" (\"\>", "\[InvisibleSpace]", "8", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"18\"\>", 
   "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"", 
   "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Sequences", "\" (", 
   8, " of ", "18", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579430335293`*^9},
 CellID->225271960],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"48\"\>"}],
  SequenceForm["  Reading proof ", 1, " of ", "48"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557943086713*^9},
 CellID->26177497],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"\"\>", 
   "\[InvisibleSpace]", "\<\"/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Tuples\"\>", 
   "\[InvisibleSpace]", "\<\"\\\" (\"\>", "\[InvisibleSpace]", "9", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"18\"\>", 
   "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"", 
   "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Tuples", "\" (", 9, 
   " of ", "18", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557943539546*^9},
 CellID->995664070],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"271\"\>"}],
  SequenceForm["  Reading proof ", 1, " of ", "271"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579435941553`*^9},
 CellID->1802485517],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"271\"\>"}],
  SequenceForm["  Reading proof ", 51, " of ", "271"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557944061057*^9},
 CellID->102721391],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"271\"\>"}],
  SequenceForm["  Reading proof ", 101, " of ", "271"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557944588616*^9},
 CellID->772960722],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"271\"\>"}],
  SequenceForm["  Reading proof ", 151, " of ", "271"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557945121207*^9},
 CellID->545971667],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"271\"\>"}],
  SequenceForm["  Reading proof ", 201, " of ", "271"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579456217823`*^9},
 CellID->1544609497],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"271\"\>"}],
  SequenceForm["  Reading proof ", 251, " of ", "271"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557946139826*^9},
 CellID->1356748633],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"\"\>", 
   "\[InvisibleSpace]", "\<\"/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory\"\>", "\[InvisibleSpace]", "\<\"\\\" (\"\>", "\[InvisibleSpace]", "10",
    "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"18\"\>", 
   "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"", 
   "/home/amaletzk/Theorema Notebooks/Reduction Ring Theory", "\" (", 10, 
   " of ", "18", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557946367264*^9},
 CellID->872782520],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"\"\>", 
   "\[InvisibleSpace]", "\<\"/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Fields\"\>", "\[InvisibleSpace]", "\<\"\\\" (\"\>", 
   "\[InvisibleSpace]", "11", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"18\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"", 
   "/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Fields", "\" (", 
   11, " of ", "18", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557946369769*^9},
 CellID->2040460043],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"36\"\>"}],
  SequenceForm["  Reading proof ", 1, " of ", "36"],
  Editable->False]], "Print",
 CellChangeTimes->{3.66755794637176*^9},
 CellID->1123372340],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"\"\>", 
   "\[InvisibleSpace]", "\<\"/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/GroebnerRings\"\>", "\[InvisibleSpace]", "\<\"\\\" (\"\>", 
   "\[InvisibleSpace]", "12", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"18\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"", 
   "/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/GroebnerRings", 
   "\" (", 12, " of ", "18", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557946806138*^9},
 CellID->13998605],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"170\"\>"}],
  SequenceForm["  Reading proof ", 1, " of ", "170"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557946854967*^9},
 CellID->244795172],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"170\"\>"}],
  SequenceForm["  Reading proof ", 51, " of ", "170"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557947684452*^9},
 CellID->303869597],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"170\"\>"}],
  SequenceForm["  Reading proof ", 101, " of ", "170"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557948389496*^9},
 CellID->1680057154],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"170\"\>"}],
  SequenceForm["  Reading proof ", 151, " of ", "170"],
  Editable->False]], "Print",
 CellChangeTimes->{3.66755794903671*^9},
 CellID->415121188],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"\"\>", 
   "\[InvisibleSpace]", "\<\"/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings\"\>", "\[InvisibleSpace]", "\<\"\\\" (\"\>", 
   "\[InvisibleSpace]", "13", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"18\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"", 
   "/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/IntegerQuotientRings", "\" (", 13, " of ", "18", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557949320003*^9},
 CellID->290229500],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"78\"\>"}],
  SequenceForm["  Reading proof ", 1, " of ", "78"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579493792048`*^9},
 CellID->917949173],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"78\"\>"}],
  SequenceForm["  Reading proof ", 51, " of ", "78"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579499501667`*^9},
 CellID->861684230],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"\"\>", 
   "\[InvisibleSpace]", "\<\"/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Integers\"\>", "\[InvisibleSpace]", "\<\"\\\" (\"\>", 
   "\[InvisibleSpace]", "14", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"18\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"", 
   "/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Integers", "\" (",
    14, " of ", "18", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557950343968*^9},
 CellID->1480145550],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"85\"\>"}],
  SequenceForm["  Reading proof ", 1, " of ", "85"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557950393228*^9},
 CellID->159503147],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"85\"\>"}],
  SequenceForm["  Reading proof ", 51, " of ", "85"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557950912925*^9},
 CellID->344990273],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"\"\>", 
   "\[InvisibleSpace]", "\<\"/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Overview\"\>", "\[InvisibleSpace]", "\<\"\\\" (\"\>", 
   "\[InvisibleSpace]", "15", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"18\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"", 
   "/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Overview", "\" (",
    15, " of ", "18", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557951327238*^9},
 CellID->89052781],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"\"\>", 
   "\[InvisibleSpace]", "\<\"/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/Polynomials\"\>", "\[InvisibleSpace]", "\<\"\\\" (\"\>", 
   "\[InvisibleSpace]", "16", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"18\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"", 
   "/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/Polynomials", 
   "\" (", 16, " of ", "18", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579513750677`*^9},
 CellID->742993044],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"342\"\>"}],
  SequenceForm["  Reading proof ", 1, " of ", "342"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557951377377*^9},
 CellID->1964139600],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"342\"\>"}],
  SequenceForm["  Reading proof ", 51, " of ", "342"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579520189543`*^9},
 CellID->437821055],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"342\"\>"}],
  SequenceForm["  Reading proof ", 101, " of ", "342"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557952704279*^9},
 CellID->1141074587],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"342\"\>"}],
  SequenceForm["  Reading proof ", 151, " of ", "342"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557953370863*^9},
 CellID->1905329242],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"342\"\>"}],
  SequenceForm["  Reading proof ", 201, " of ", "342"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557954078314*^9},
 CellID->694875228],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"342\"\>"}],
  SequenceForm["  Reading proof ", 251, " of ", "342"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557954820507*^9},
 CellID->343196592],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"342\"\>"}],
  SequenceForm["  Reading proof ", 301, " of ", "342"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557955594817*^9},
 CellID->1299769609],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"\"\>", 
   "\[InvisibleSpace]", "\<\"/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples\"\>", "\[InvisibleSpace]", "\<\"\\\" (\"\>", 
   "\[InvisibleSpace]", "17", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"18\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"", 
   "/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/PolyTuples", 
   "\" (", 17, " of ", "18", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579561765413`*^9},
 CellID->1747622676],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"161\"\>"}],
  SequenceForm["  Reading proof ", 1, " of ", "161"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579562286263`*^9},
 CellID->1931994882],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"161\"\>"}],
  SequenceForm["  Reading proof ", 51, " of ", "161"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557956781704*^9},
 CellID->1285042923],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"161\"\>"}],
  SequenceForm["  Reading proof ", 101, " of ", "161"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557957455945*^9},
 CellID->656249480],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"161\"\>"}],
  SequenceForm["  Reading proof ", 151, " of ", "161"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557958069603*^9},
 CellID->1610366942],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"\"\>", 
   "\[InvisibleSpace]", "\<\"/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/ReductionRings\"\>", "\[InvisibleSpace]", "\<\"\\\" (\"\>", 
   "\[InvisibleSpace]", "18", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"18\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"", 
   "/home/amaletzk/Theorema Notebooks/Reduction Ring Theory/ReductionRings", 
   "\" (", 18, " of ", "18", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557958215578*^9},
 CellID->673784444],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"252\"\>"}],
  SequenceForm["  Reading proof ", 1, " of ", "252"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579582723827`*^9},
 CellID->1385058587],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"252\"\>"}],
  SequenceForm["  Reading proof ", 51, " of ", "252"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557958961918*^9},
 CellID->1517845881],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"252\"\>"}],
  SequenceForm["  Reading proof ", 101, " of ", "252"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557959584638*^9},
 CellID->2077601540],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"252\"\>"}],
  SequenceForm["  Reading proof ", 151, " of ", "252"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579605306883`*^9},
 CellID->2134222863],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"252\"\>"}],
  SequenceForm["  Reading proof ", 201, " of ", "252"],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557961333009*^9},
 CellID->196398467],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"252\"\>"}],
  SequenceForm["  Reading proof ", 251, " of ", "252"],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579619366493`*^9},
 CellID->2072983810],

Cell[BoxData["\<\"Creating forward rules:\"\>"], "Print",
 CellChangeTimes->{3.667557961997048*^9},
 CellID->2123204866],

Cell[BoxData["\<\"Preparing ...\"\>"], "Print",
 CellChangeTimes->{3.667557962051532*^9},
 CellID->284217937],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557962466156*^9},
 CellID->399737596],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 51, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579637196913`*^9},
 CellID->805060146],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 101, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557964965556*^9},
 CellID->263663299],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 151, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557966204782*^9},
 CellID->1733700269],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 201, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557967453908*^9},
 CellID->504235321],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 251, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579686828737`*^9},
 CellID->1411520420],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 301, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557969905527*^9},
 CellID->257856182],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 351, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.66755797113521*^9},
 CellID->1540842051],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "401", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 401, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579723638887`*^9},
 CellID->487865859],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "451", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 451, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557973607349*^9},
 CellID->602281323],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "501", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 501, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557974877145*^9},
 CellID->684893606],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "551", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 551, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557976135664*^9},
 CellID->1229499825],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "601", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 601, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557977412961*^9},
 CellID->736007362],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "651", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 651, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579786602097`*^9},
 CellID->2047264904],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "701", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 701, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579799240007`*^9},
 CellID->579248345],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "751", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 751, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557981185368*^9},
 CellID->301728890],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "801", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 801, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557982442458*^9},
 CellID->2091818475],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "851", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 851, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579837043953`*^9},
 CellID->1034002157],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "901", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 901, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579849501667`*^9},
 CellID->405159898],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "951", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 951, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557986200819*^9},
 CellID->880592609],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1001", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1001, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.66755798745453*^9},
 CellID->1428824071],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1051", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1051, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557988707573*^9},
 CellID->1545734938],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1101, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557989951604*^9},
 CellID->704434208],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1151, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557991195387*^9},
 CellID->904655799],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1201, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557992430087*^9},
 CellID->2006111296],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1251, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675579936686363`*^9},
 CellID->1851230075],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1301, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557994901169*^9},
 CellID->848859071],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1351, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557996148089*^9},
 CellID->1691606509],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1401", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1401, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557997400416*^9},
 CellID->293097688],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1451", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1451, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557998655999*^9},
 CellID->1366882988],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1501", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1501, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667557999907773*^9},
 CellID->154429773],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1551", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1551, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558001157391*^9},
 CellID->744698092],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1601", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1601, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558002408182*^9},
 CellID->2124648104],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1651", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1651, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558003659582*^9},
 CellID->1405866756],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1701", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1701, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558004912191*^9},
 CellID->1678907350],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1751", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1751, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675580061652603`*^9},
 CellID->1176964354],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1801", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1801, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558007400305*^9},
 CellID->32915441],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1851", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1851, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558008620175*^9},
 CellID->1952766498],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1901", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1901, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675580098444633`*^9},
 CellID->79551376],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1951", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 1951, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.66755801106316*^9},
 CellID->2039798323],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2001", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2001, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558012284706*^9},
 CellID->920011543],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2051", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2051, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675580135096607`*^9},
 CellID->1341994267],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2101, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558014766714*^9},
 CellID->1365657716],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2151, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675580160182943`*^9},
 CellID->750840837],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2201, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.66755801726816*^9},
 CellID->819852299],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2251, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558018516541*^9},
 CellID->90096710],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2301, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558019760078*^9},
 CellID->512166966],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2351, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558021005906*^9},
 CellID->1395279649],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2401", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2401, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558022236671*^9},
 CellID->907983991],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2451", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2451, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558023459772*^9},
 CellID->1271409132],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2501", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2501, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558024697452*^9},
 CellID->1007395009],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2551", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2551, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558025912683*^9},
 CellID->1008558891],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2601", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2601, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558027129643*^9},
 CellID->668606491],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2651", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2651, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558028346962*^9},
 CellID->346902053],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2701", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2701, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558029562542*^9},
 CellID->47434879],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2751", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2751, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558030780005*^9},
 CellID->339282318],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2801", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2801, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558031997993*^9},
 CellID->1946506496],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2851", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2851, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558033222933*^9},
 CellID->1917192120],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2901", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2901, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675580344506283`*^9},
 CellID->1728132324],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "2951", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 2951, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.66755803568465*^9},
 CellID->739649454],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "3001", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 3001, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558036908524*^9},
 CellID->1897894093],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "3051", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 3051, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675580381318073`*^9},
 CellID->1157191333],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "3101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 3101, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558039357925*^9},
 CellID->448292855],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "3151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 3151, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558040584526*^9},
 CellID->438328241],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "3201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3248"}],
  SequenceForm["Creating rule ", 3201, " of ", 3248],
  Editable->False]], "Print",
 CellChangeTimes->{3.6675580418098173`*^9},
 CellID->2142336410],

Cell[BoxData["\<\"Creating thinned data ...\"\>"], "Print",
 CellChangeTimes->{3.667558042966743*^9},
 CellID->1773987139],

Cell[BoxData["\<\"Creating theory data ...\"\>"], "Print",
 CellChangeTimes->{3.667558043056696*^9},
 CellID->434266741],

Cell[BoxData["\<\"Creating backward rules ...\"\>"], "Print",
 CellChangeTimes->{3.6675580431054773`*^9},
 CellID->2009684850],

Cell[BoxData["\<\"Creating forward rules ...\"\>"], "Print",
 CellChangeTimes->{3.667558043295766*^9},
 CellID->804196891],

Cell[BoxData["\<\"Preparing ...\"\>"], "Print",
 CellChangeTimes->{3.66755804334501*^9},
 CellID->689469934],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "15"}],
  SequenceForm["Creating rule ", 1, " of ", 15],
  Editable->False]], "Print",
 CellChangeTimes->{3.667558043348111*^9},
 CellID->628793925],

Cell[BoxData["\<\"Done\"\>"], "Print",
 CellChangeTimes->{3.667558043350856*^9},
 CellID->1777775172]
}, Open  ]],

Cell[BoxData["97.19599999999999795363692101091146469116`8.008248305637245"], \
"Output",
 CellChangeTimes->{3.6675580433529043`*^9},
 CellID->1546107737]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"AnalyzeProofs", "[", "]"}], "//", "Timing"}]], "Input",
 CellChangeTimes->{{3.64655328989894*^9, 3.646553296127252*^9}, {
  3.6554463677093964`*^9, 3.655446368723398*^9}, {3.6569180725458093`*^9, 
  3.656918075915415*^9}},
 CellID->1859739660],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \\\"\"\>", 
   "\[InvisibleSpace]", "\<\"/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers\"\>", 
   "\[InvisibleSpace]", "\<\"\\\" (\"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"1\"\>", 
   "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"", 
   "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers", "\" (", 1, 
   " of ", "1", "):"],
  Editable->False]], "Print",
 CellChangeTimes->{3.656918076523816*^9, 3.6574441827918043`*^9, 
  3.6580584628557053`*^9, 3.6580602063453693`*^9, 3.6581216439454794`*^9, 
  3.6581293293669815`*^9, 3.660478117987048*^9, 3.660643509123491*^9, 
  3.660643569698397*^9, 3.6612485584740314`*^9, 3.6613175323732796`*^9, 
  3.667733839680173*^9, 3.667815772251609*^9},
 CellID->281141794],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"165\"\>"}],
  SequenceForm["  Reading proof ", 1, " of ", "165"],
  Editable->False]], "Print",
 CellChangeTimes->{3.656918076523816*^9, 3.6574441827918043`*^9, 
  3.6580584628557053`*^9, 3.6580602063453693`*^9, 3.6581216439454794`*^9, 
  3.6581293293669815`*^9, 3.660478117987048*^9, 3.660643509123491*^9, 
  3.660643569698397*^9, 3.6612485584740314`*^9, 3.6613175323732796`*^9, 
  3.667733839680173*^9, 3.667815772254448*^9},
 CellID->1164344103],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"165\"\>"}],
  SequenceForm["  Reading proof ", 51, " of ", "165"],
  Editable->False]], "Print",
 CellChangeTimes->{3.656918076523816*^9, 3.6574441827918043`*^9, 
  3.6580584628557053`*^9, 3.6580602063453693`*^9, 3.6581216439454794`*^9, 
  3.6581293293669815`*^9, 3.660478117987048*^9, 3.660643509123491*^9, 
  3.660643569698397*^9, 3.6612485584740314`*^9, 3.6613175323732796`*^9, 
  3.667733839680173*^9, 3.667815772645495*^9},
 CellID->1263200977],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"165\"\>"}],
  SequenceForm["  Reading proof ", 101, " of ", "165"],
  Editable->False]], "Print",
 CellChangeTimes->{3.656918076523816*^9, 3.6574441827918043`*^9, 
  3.6580584628557053`*^9, 3.6580602063453693`*^9, 3.6581216439454794`*^9, 
  3.6581293293669815`*^9, 3.660478117987048*^9, 3.660643509123491*^9, 
  3.660643569698397*^9, 3.6612485584740314`*^9, 3.6613175323732796`*^9, 
  3.667733839680173*^9, 3.667815773059823*^9},
 CellID->61583932],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"165\"\>"}],
  SequenceForm["  Reading proof ", 151, " of ", "165"],
  Editable->False]], "Print",
 CellChangeTimes->{3.656918076523816*^9, 3.6574441827918043`*^9, 
  3.6580584628557053`*^9, 3.6580602063453693`*^9, 3.6581216439454794`*^9, 
  3.6581293293669815`*^9, 3.660478117987048*^9, 3.660643509123491*^9, 
  3.660643569698397*^9, 3.6612485584740314`*^9, 3.6613175323732796`*^9, 
  3.667733839680173*^9, 3.667815773467414*^9},
 CellID->1482990827],

Cell[BoxData["\<\"Creating forward rules:\"\>"], "Print",
 CellChangeTimes->{3.656918076523816*^9, 3.6574441827918043`*^9, 
  3.6580584628557053`*^9, 3.6580602063453693`*^9, 3.6581216439454794`*^9, 
  3.6581293293669815`*^9, 3.660478117987048*^9, 3.660643509123491*^9, 
  3.660643569698397*^9, 3.6612485584740314`*^9, 3.6613175323732796`*^9, 
  3.667733839680173*^9, 3.667815773600748*^9},
 CellID->1725599731],

Cell[BoxData["\<\"Preparing ...\"\>"], "Print",
 CellChangeTimes->{3.656918076523816*^9, 3.6574441827918043`*^9, 
  3.6580584628557053`*^9, 3.6580602063453693`*^9, 3.6581216439454794`*^9, 
  3.6581293293669815`*^9, 3.660478117987048*^9, 3.660643509123491*^9, 
  3.660643569698397*^9, 3.6612485584740314`*^9, 3.6613175323732796`*^9, 
  3.667733839680173*^9, 3.667815773603826*^9},
 CellID->326369969],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "234"}],
  SequenceForm["Creating rule ", 1, " of ", 234],
  Editable->False]], "Print",
 CellChangeTimes->{3.656918076523816*^9, 3.6574441827918043`*^9, 
  3.6580584628557053`*^9, 3.6580602063453693`*^9, 3.6581216439454794`*^9, 
  3.6581293293669815`*^9, 3.660478117987048*^9, 3.660643509123491*^9, 
  3.660643569698397*^9, 3.6612485584740314`*^9, 3.6613175323732796`*^9, 
  3.667733839680173*^9, 3.667815773657662*^9},
 CellID->406809379],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "234"}],
  SequenceForm["Creating rule ", 51, " of ", 234],
  Editable->False]], "Print",
 CellChangeTimes->{3.656918076523816*^9, 3.6574441827918043`*^9, 
  3.6580584628557053`*^9, 3.6580602063453693`*^9, 3.6581216439454794`*^9, 
  3.6581293293669815`*^9, 3.660478117987048*^9, 3.660643509123491*^9, 
  3.660643569698397*^9, 3.6612485584740314`*^9, 3.6613175323732796`*^9, 
  3.667733839680173*^9, 3.667815773660552*^9},
 CellID->537433043],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "234"}],
  SequenceForm["Creating rule ", 101, " of ", 234],
  Editable->False]], "Print",
 CellChangeTimes->{3.656918076523816*^9, 3.6574441827918043`*^9, 
  3.6580584628557053`*^9, 3.6580602063453693`*^9, 3.6581216439454794`*^9, 
  3.6581293293669815`*^9, 3.660478117987048*^9, 3.660643509123491*^9, 
  3.660643569698397*^9, 3.6612485584740314`*^9, 3.6613175323732796`*^9, 
  3.667733839680173*^9, 3.667815773663134*^9},
 CellID->241512502],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "234"}],
  SequenceForm["Creating rule ", 151, " of ", 234],
  Editable->False]], "Print",
 CellChangeTimes->{3.656918076523816*^9, 3.6574441827918043`*^9, 
  3.6580584628557053`*^9, 3.6580602063453693`*^9, 3.6581216439454794`*^9, 
  3.6581293293669815`*^9, 3.660478117987048*^9, 3.660643509123491*^9, 
  3.660643569698397*^9, 3.6612485584740314`*^9, 3.6613175323732796`*^9, 
  3.667733839680173*^9, 3.66781577370909*^9},
 CellID->993359368],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "234"}],
  SequenceForm["Creating rule ", 201, " of ", 234],
  Editable->False]], "Print",
 CellChangeTimes->{3.656918076523816*^9, 3.6574441827918043`*^9, 
  3.6580584628557053`*^9, 3.6580602063453693`*^9, 3.6581216439454794`*^9, 
  3.6581293293669815`*^9, 3.660478117987048*^9, 3.660643509123491*^9, 
  3.660643569698397*^9, 3.6612485584740314`*^9, 3.6613175323732796`*^9, 
  3.667733839680173*^9, 3.667815773711849*^9},
 CellID->2116037914],

Cell[BoxData["\<\"Creating thinned data ...\"\>"], "Print",
 CellChangeTimes->{3.656918076523816*^9, 3.6574441827918043`*^9, 
  3.6580584628557053`*^9, 3.6580602063453693`*^9, 3.6581216439454794`*^9, 
  3.6581293293669815`*^9, 3.660478117987048*^9, 3.660643509123491*^9, 
  3.660643569698397*^9, 3.6612485584740314`*^9, 3.6613175323732796`*^9, 
  3.667733839680173*^9, 3.667815773714547*^9},
 CellID->222181995],

Cell[BoxData["\<\"Creating theory data ...\"\>"], "Print",
 CellChangeTimes->{3.656918076523816*^9, 3.6574441827918043`*^9, 
  3.6580584628557053`*^9, 3.6580602063453693`*^9, 3.6581216439454794`*^9, 
  3.6581293293669815`*^9, 3.660478117987048*^9, 3.660643509123491*^9, 
  3.660643569698397*^9, 3.6612485584740314`*^9, 3.6613175323732796`*^9, 
  3.667733839680173*^9, 3.667815773761757*^9},
 CellID->1898721665],

Cell[BoxData["\<\"Creating backward rules ...\"\>"], "Print",
 CellChangeTimes->{3.656918076523816*^9, 3.6574441827918043`*^9, 
  3.6580584628557053`*^9, 3.6580602063453693`*^9, 3.6581216439454794`*^9, 
  3.6581293293669815`*^9, 3.660478117987048*^9, 3.660643509123491*^9, 
  3.660643569698397*^9, 3.6612485584740314`*^9, 3.6613175323732796`*^9, 
  3.667733839680173*^9, 3.66781577376449*^9},
 CellID->194167885],

Cell[BoxData["\<\"Creating forward rules ...\"\>"], "Print",
 CellChangeTimes->{3.656918076523816*^9, 3.6574441827918043`*^9, 
  3.6580584628557053`*^9, 3.6580602063453693`*^9, 3.6581216439454794`*^9, 
  3.6581293293669815`*^9, 3.660478117987048*^9, 3.660643509123491*^9, 
  3.660643569698397*^9, 3.6612485584740314`*^9, 3.6613175323732796`*^9, 
  3.667733839680173*^9, 3.667815773767055*^9},
 CellID->578987836],

Cell[BoxData["\<\"Preparing ...\"\>"], "Print",
 CellChangeTimes->{3.656918076523816*^9, 3.6574441827918043`*^9, 
  3.6580584628557053`*^9, 3.6580602063453693`*^9, 3.6581216439454794`*^9, 
  3.6581293293669815`*^9, 3.660478117987048*^9, 3.660643509123491*^9, 
  3.660643569698397*^9, 3.6612485584740314`*^9, 3.6613175323732796`*^9, 
  3.667733839680173*^9, 3.667815773769598*^9},
 CellID->446001340],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "3"}],
  SequenceForm["Creating rule ", 1, " of ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.656918076523816*^9, 3.6574441827918043`*^9, 
  3.6580584628557053`*^9, 3.6580602063453693`*^9, 3.6581216439454794`*^9, 
  3.6581293293669815`*^9, 3.660478117987048*^9, 3.660643509123491*^9, 
  3.660643569698397*^9, 3.6612485584740314`*^9, 3.6613175323732796`*^9, 
  3.667733839680173*^9, 3.66781577377179*^9},
 CellID->1806064551],

Cell[BoxData["\<\"Done\"\>"], "Print",
 CellChangeTimes->{3.656918076523816*^9, 3.6574441827918043`*^9, 
  3.6580584628557053`*^9, 3.6580602063453693`*^9, 3.6581216439454794`*^9, 
  3.6581293293669815`*^9, 3.660478117987048*^9, 3.660643509123491*^9, 
  3.660643569698397*^9, 3.6612485584740314`*^9, 3.6613175323732796`*^9, 
  3.667733839680173*^9, 3.667815773773621*^9},
 CellID->1973519727]
}, Open  ]],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "1.00800000000000000710542735760100185871`6.02406044538913", ",", "Null"}], 
  "}"}]], "Output",
 CellChangeTimes->{3.656918087194235*^9, 3.65744420357104*^9, 
  3.658058469236117*^9, 3.6580602126009808`*^9, 3.658121673944332*^9, 
  3.6581293365273943`*^9, 3.6604781227762566`*^9, 3.6606435165491037`*^9, 
  3.6606435714611998`*^9, 3.661248570002452*^9, 3.6613175449781017`*^9, 
  3.667733841692521*^9, 3.6678157737754707`*^9},
 CellID->92225566]
}, Closed]],

Cell[BoxData[
 RowBox[{
  RowBox[{"AnalyzeProofs", "[", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"FileNameJoin", "[", 
      RowBox[{"{", 
       RowBox[{
       "$HomeDirectory", ",", "\"\<Theorema Notebooks\>\"", ",", 
        "\"\<NatInt\>\""}], "}"}], "]"}], ",", 
     RowBox[{"FileNameJoin", "[", 
      RowBox[{"{", 
       RowBox[{
       "$HomeDirectory", ",", "\"\<Theorema Notebooks\>\"", ",", 
        "\"\<Reduction Ring Theory\>\"", ",", "\"\<ReductionRings\>\""}], 
       "}"}], "]"}], ",", 
     RowBox[{"FileNameJoin", "[", 
      RowBox[{"{", 
       RowBox[{
       "$HomeDirectory", ",", "\"\<Theorema Notebooks\>\"", ",", 
        "\"\<Reduction Ring Theory\>\"", ",", "\"\<Polynomials\>\""}], "}"}], 
      "]"}], ",", 
     RowBox[{"FileNameJoin", "[", 
      RowBox[{"{", 
       RowBox[{
       "$HomeDirectory", ",", "\"\<Theorema Notebooks\>\"", ",", 
        "\"\<Reduction Ring Theory\>\"", ",", "\"\<GroebnerRings\>\""}], 
       "}"}], "]"}]}], "}"}], "]"}], ";"}]], "Input",
 CellChangeTimes->{{3.6406688560104437`*^9, 3.640668867513785*^9}, 
   3.64067664060292*^9, 3.644222019629964*^9, 3.6443084442789383`*^9, {
   3.644637309909933*^9, 3.644637395211917*^9}, 3.6446374289086847`*^9, {
   3.649665456506633*^9, 3.64966546130441*^9}, {3.6554463708762016`*^9, 
   3.6554463720150037`*^9}},
 CellID->1039312110],

Cell[BoxData[
 RowBox[{
  RowBox[{"ComputeFormulaPositions", "[", "]"}], ";"}]], "Input",
 CellChangeTimes->{{3.644649344001361*^9, 3.6446493476777363`*^9}},
 CellID->2077551358],

Cell[BoxData[
 RowBox[{"CheckIntegrity", "[", 
  RowBox[{"\"\<CheckPositions\>\"", "\[Rule]", 
   RowBox[{"{", 
    RowBox[{"FileNameJoin", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"CurrentValue", "[", "\"\<NotebookDirectory\>\"", "]"}], ",", 
       RowBox[{"CurrentValue", "[", "\"\<NotebookFileName\>\"", "]"}]}], 
      "}"}], "]"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.64067003109768*^9, 3.6406700374311132`*^9}, {
   3.6406797903560762`*^9, 3.640679792950843*^9}, 3.6442220291527653`*^9},
 CellID->1909803754],

Cell[BoxData[{
 RowBox[{
  RowBox[{"ps", "=", 
   RowBox[{"GetProofSize", "[", "All", "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"ps0", "=", 
  RowBox[{"Map", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Total", "[", 
      RowBox[{"Most", "[", 
       RowBox[{"Last", "[", "#", "]"}], "]"}], "]"}], "&"}], ",", "ps"}], 
   "]"}]}]}], "Input",
 CellChangeTimes->{{3.647156479700988*^9, 3.647156479817154*^9}, {
   3.647156560839411*^9, 3.6471566309015427`*^9}, 3.6471570742150927`*^9, 
   3.6471571422253036`*^9, {3.647157728211722*^9, 3.647157728406296*^9}},
 CellID->1780799392],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Count", "[", 
  RowBox[{
   RowBox[{"NotebookGet", "[", "]"}], ",", 
   RowBox[{"Cell", "[", 
    RowBox[{"_", ",", "\"\<FormalTextInputFormula\>\"", ",", "___"}], "]"}], 
   ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "Infinity"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.6550382930110416`*^9, 3.655038363616766*^9}, {
   3.655442032762142*^9, 3.6554420535101786`*^9}, 3.655446359394582*^9, {
   3.6554580724009075`*^9, 3.655458084943329*^9}},
 CellID->63220416],

Cell[BoxData["123"], "Output",
 CellChangeTimes->{3.65545808552053*^9, 3.657513226876388*^9, 
  3.657537219869627*^9, 3.657609115071938*^9, 3.6580473422329206`*^9, 
  3.6586615277780223`*^9, 3.6587229310902023`*^9, 3.6602884137169657`*^9, 
  3.6603953176254063`*^9},
 CellID->141359559]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Auxiliary Functions", "Subsection",
 CellChangeTimes->{{3.640668579267561*^9, 3.640668583347287*^9}},
 CellID->1454182661],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "LoadPackage", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"LoadPackage", "[", "name_String", "]"}], ":=", 
   RowBox[{"LoadPackage", "[", 
    RowBox[{"{", "name", "}"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"LoadPackage", "[", 
    RowBox[{"l", ":", 
     RowBox[{"{", "___String", "}"}]}], "]"}], ":=", 
   RowBox[{"Scan", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Get", "[", 
       RowBox[{"FileNameJoin", "[", 
        RowBox[{"{", 
         RowBox[{"$HomeDirectory", ",", "\"\<Theorema Notebooks\>\"", ",", 
          RowBox[{"#", "<>", "\"\<.m\>\""}]}], "}"}], "]"}], "]"}], "&"}], 
     ",", "l"}], "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.640668603629292*^9, 3.640668727460284*^9}, 
   3.641286816086794*^9, 3.6434459961973133`*^9, {3.653223148703965*^9, 
   3.653223155365177*^9}, {3.666439327921836*^9, 3.666439333528713*^9}},
 CellID->2074904605],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "FindFormula", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"FindFormula", "[", "label_String", "]"}], ":=", 
   RowBox[{"NotebookLocate", "[", "label", "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.640677756847624*^9, 3.6406777896243896`*^9}, 
   3.640677822869519*^9, {3.640677854902379*^9, 3.6406778579321823`*^9}},
 CellID->1980565193],

Cell[BoxData[
 RowBox[{"LoadPackage", "[", "\"\<ProofPalette\>\"", "]"}]], "Input",
 CellChangeTimes->{{3.640668742006607*^9, 3.640668751417616*^9}},
 CellID->444074165],

Cell[BoxData[
 RowBox[{"LoadPackage", "[", "\"\<TheoryTools/TheoryTools\>\"", 
  "]"}]], "Input",
 CellChangeTimes->{{3.640668759133181*^9, 3.640668768089197*^9}, {
  3.6551873799276342`*^9, 3.655187380848036*^9}, {3.658224662152712*^9, 
  3.658224665147917*^9}, {3.666439336397678*^9, 3.6664393369334*^9}},
 CellID->584704106]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Set-Theory Natural Numbers (\[Checkmark])", "Section",
 CellChangeTimes->{{3.6597659478942933`*^9, 3.6597659515914993`*^9}, {
  3.660370931697556*^9, 3.660370934677161*^9}, {3.660386236206241*^9, 
  3.6603862401374483`*^9}, {3.6604739611867456`*^9, 3.660473962793549*^9}, {
  3.6613374496422076`*^9, 3.6613374542452154`*^9}},
 CellID->252853686],

Cell[CellGroupData[{

Cell["Natural Recursion (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.659765959609914*^9, 3.659765964929523*^9}, {
   3.6598554419526668`*^9, 3.659855450579482*^9}, {3.6598554965527625`*^9, 
   3.6598555007023697`*^9}, {3.660473968456359*^9, 3.660473969485961*^9}, {
   3.661337461267228*^9, 3.6613374676476393`*^9}, 3.667734292709393*^9},
 CellID->394311767],

Cell[CellGroupData[{

Cell["Informal Explanation", "Subsubsection",
 CellChangeTimes->{{3.6597675370778837`*^9, 3.6597675418982925`*^9}},
 CellID->250375803],

Cell[TextData[{
 "The set of ",
 "set-theory",
 " natural numbers is defined by a least-fixpoint construction as the \
smallest set containing \[EmptySet] and, for every element ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ", also its successor ",
 Cell[BoxData[
  FormBox[
   RowBox[{"succ", "[", "x", "]"}], TraditionalForm]]],
 ". Hence, set-theory natural numbers are just sets (that\[CloseCurlyQuote]s \
why the name \[OpenCurlyDoubleQuote]set-theory\[CloseCurlyDoubleQuote])."
}], "Text",
 CellChangeTimes->{{3.659767602598999*^9, 3.659767756212469*^9}, 
   3.660473768495208*^9, {3.66047394079751*^9, 3.660473959314743*^9}, {
   3.6604739937440033`*^9, 3.660474037954481*^9}, {3.661337476492855*^9, 
   3.6613374900492787`*^9}, 3.6613375787520347`*^9},
 CellID->65046519],

Cell[TextData[{
 "The construction of natural numbers in fact first defines the concept of \
natural recursion for arbitrary base ",
 Cell[BoxData[
  FormBox["b", TraditionalForm]]],
 " and combinator ",
 Cell[BoxData[
  FormBox["c", TraditionalForm]]],
 " by means of a set of ordered pairs (natRecSet). Then, natSet and natRec \
(i.e. natural recursion as a function) are defined in terms of natRecSet. The \
reason for this approach is that the (generalized) Axiom of Infinity only \
needs to be used once, whereas if first the set of natural numbers was \
defined and only then natural recursion, once the original Axiom of Infinity \
and still once the generalized one would have to be used (because natural \
recursion should work for arbitrary bases and combinators)."
}], "Text",
 CellChangeTimes->{{3.6597677701120934`*^9, 3.659767826849393*^9}, {
   3.659767941931795*^9, 3.659767971197447*^9}, {3.659769107495843*^9, 
   3.6597691289614806`*^9}, 3.6604716363558626`*^9, {3.6604737799456277`*^9, 
   3.6604738959318314`*^9}},
 CellID->451096809]
}, Closed]],

Cell[CellGroupData[{

Cell["Definitions", "Subsubsection",
 CellChangeTimes->{{3.659765975849542*^9, 3.659765979141148*^9}},
 CellID->415653668],

Cell["", "OpenEnvironment",
 CellID->562348793],

Cell[CellGroupData[{

Cell["Definition: natRecSet & natRec", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6597659926663713`*^9, 3.6597660066595964`*^9}, {
  3.659766545209542*^9, 3.6597665569563627`*^9}, {3.659767502321023*^9, 
  3.6597675106046376`*^9}, {3.659767847207429*^9, 3.6597678476754293`*^9}, {
  3.6604741793229294`*^9, 3.6604741811793327`*^9}, {3.661337580249637*^9, 
  3.661337609695689*^9}, {3.6677341426531553`*^9, 3.6677341470607643`*^9}},
 CellID->708005171],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"b", ",", "c"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          346267501}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.667734707803677*^9, 3.667734709809136*^9}},
 CellTags->{
  "ID:346267501", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->346267501],

Cell[BoxData[
 RowBox[{
  RowBox[{"natRecSet", "[", 
   RowBox[{"b", ",", "c"}], "]"}], ":=", 
  RowBox[{"indSet", "[", 
   RowBox[{
    RowBox[{"Pair", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", "b"}], "]"}], ",", 
    RowBox[{
     UnderscriptBox["\[Lambda]", "x"], 
     RowBox[{"Pair", "[", 
      RowBox[{
       RowBox[{"succ", "[", 
        RowBox[{"fst", "[", "x", "]"}], "]"}], ",", 
       RowBox[{"c", "[", 
        RowBox[{
         RowBox[{"fst", "[", "x", "]"}], ",", 
         RowBox[{"snd", "[", "x", "]"}]}], "]"}]}], "]"}]}]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(natRecSet)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:143274494", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659766335872774*^9, 3.6597665125742846`*^9}, {
  3.659766739289483*^9, 3.6597667413642864`*^9}, {3.6597667995679884`*^9, 
  3.6597668333888483`*^9}, {3.6597668701425123`*^9, 3.659766892482752*^9}, {
  3.659767038734009*^9, 3.659767046752423*^9}, {3.6597683884709797`*^9, 
  3.6597683943833904`*^9}, {3.659788352329484*^9, 3.659788356448891*^9}, {
  3.6677347129470263`*^9, 3.667734798580305*^9}},
 CellTags->{
  "ID:143274494", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet"},
 CellID->143274494],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "n"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", "y"], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"Pair", "[", 
         RowBox[{"n", ",", "y"}], "]"}], "\[Element]", 
        RowBox[{"natRecSet", "[", 
         RowBox[{"b", ",", "c"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       RowBox[{"natRec", "[", 
        RowBox[{"b", ",", "c"}], "]"}], "[", "n", "]"}], ":=", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox["\[CurlyEpsilon]", "y"], 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          RowBox[{"Pair", "[", 
           RowBox[{"n", ",", "y"}], "]"}], "\[Element]", 
          RowBox[{"natRecSet", "[", 
           RowBox[{"b", ",", "c"}], "]"}]}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(natRec)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:669214708", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6597672194759264`*^9, 3.659767313450491*^9}, {
  3.6598459458279734`*^9, 3.659845978728431*^9}, {3.6598550120441113`*^9, 
  3.6598550156321177`*^9}, {3.6677347169885607`*^9, 3.6677347172606707`*^9}},
 CellTags->{
  "ID:669214708", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRec"},
 CellID->669214708],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->382099679]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->47167836]
}, Closed]],

Cell[CellGroupData[{

Cell["Lemmas (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6597684617297087`*^9, 3.659768473460929*^9}, 
   3.6598500188321376`*^9, {3.659851417556595*^9, 3.6598514212070017`*^9}, {
   3.667734125148591*^9, 3.667734163348716*^9}, {3.667734277445238*^9, 
   3.6677342864460087`*^9}},
 CellID->217957457],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"b", ",", "c"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          97279932}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659768507016588*^9, 3.6597685083893905`*^9}, 
   3.6598500188321376`*^9},
 CellTags->{
  "ID:97279932", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->97279932],

Cell["", "OpenEnvironment",
 CellChangeTimes->{
  3.659850019643339*^9, {3.6677349831310472`*^9, 3.6677350125854263`*^9}},
 CellID->35112716],

Cell[CellGroupData[{

Cell["Lemma: Properties of natRecSet", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6597710327546263`*^9, 3.659771037887035*^9}, 
   3.659849976898264*^9, 3.659850019643339*^9, 3.6604748693941417`*^9, {
   3.667734983131611*^9, 3.667735012585555*^9}},
 CellID->85473914],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"F", "=", 
   RowBox[{"natRecSet", "[", 
    RowBox[{"b", ",", "c"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          504518491}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65977109181833*^9, 3.659771104001951*^9}, 
   3.659849976898264*^9, 3.659850019643339*^9, 3.6604748693941417`*^9, {
   3.667734983132208*^9, 3.667735012585742*^9}},
 CellTags->{
  "ID:504518491", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->504518491],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "P"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"P", "[", 
           RowBox[{"Pair", "[", 
            RowBox[{
             RowBox[{"{", "}"}], ",", "b"}], "]"}], "]"}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{"f", "\[Element]", "F"}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{"P", "[", "f", "]"}], "\[Implies]", 
             RowBox[{"P", "[", 
              RowBox[{"Pair", "[", 
               RowBox[{
                RowBox[{"succ", "[", 
                 RowBox[{"fst", "[", "f", "]"}], "]"}], ",", 
                RowBox[{"c", "[", 
                 RowBox[{
                  RowBox[{"fst", "[", "f", "]"}], ",", 
                  RowBox[{"snd", "[", "f", "]"}]}], "]"}]}], "]"}], "]"}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{"f", "\[Element]", "F"}]], 
      RowBox[{"P", "[", "f", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(natRecSet induction)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:79522091", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659771062316678*^9, 3.659771274573651*^9}, 
   3.659849976898264*^9, 3.6598500196589394`*^9, {3.6604748693941417`*^9, 
   3.660474888410575*^9}, {3.66773498313249*^9, 3.667735012585947*^9}},
 CellTags->{
  "ID:79522091", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet induction", "IR"},
 CellID->79522091],

Cell[BoxData[
 RowBox[{"isSet", "[", "F", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(natRecSet isSet)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:212626428", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659771524067889*^9, 3.659771527125495*^9}, 
   3.659849976898264*^9, 3.6598500196589394`*^9, 3.6604748693941417`*^9, {
   3.667734983132757*^9, 3.6677350125861397`*^9}},
 CellTags->{
  "ID:212626428", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet isSet", "IR"},
 CellID->212626428],

Cell[BoxData[
 RowBox[{
  RowBox[{"Pair", "[", 
   RowBox[{
    RowBox[{"{", "}"}], ",", "b"}], "]"}], "\[Element]", 
  "F"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(base \[Element] natRecSet)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:433946629", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659771625982868*^9, 3.659771631973279*^9}, 
   3.659849976898264*^9, 3.659850019674539*^9, {3.6604748693941417`*^9, 
   3.6604748936365843`*^9}, {3.667734983133024*^9, 3.667735012586334*^9}},
 CellTags->{
  "ID:433946629", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "base \[Element] natRecSet"},
 CellID->433946629],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"f", "\[Element]", "F"}]], 
  RowBox[{
   RowBox[{"Pair", "[", 
    RowBox[{
     RowBox[{"succ", "[", 
      RowBox[{"fst", "[", "f", "]"}], "]"}], ",", 
     RowBox[{"c", "[", 
      RowBox[{
       RowBox[{"fst", "[", "f", "]"}], ",", 
       RowBox[{"snd", "[", "f", "]"}]}], "]"}]}], "]"}], "\[Element]", 
   "F"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(natRecSet inductive)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:198301498", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6597716385096903`*^9, 3.659771671129348*^9}, 
   3.659849976898264*^9, 3.659850019674539*^9, {3.6604748693941417`*^9, 
   3.660474896116988*^9}, {3.667734983133293*^9, 3.667735012586535*^9}},
 CellTags->{
  "ID:198301498", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet inductive"},
 CellID->198301498],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"f", "\[Element]", "F"}]], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{
     RowBox[{"isSet", "[", "x", "]"}], ",", "y"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{"f", "\[Equal]", 
     RowBox[{"Pair", "[", 
      RowBox[{"x", ",", "y"}], "]"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(natRecSet shape)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:972417912", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659771551789138*^9, 3.6597716099928403`*^9}, 
   3.659849976898264*^9, 3.6598500196901393`*^9, 3.6604748693941417`*^9, {
   3.660476532986064*^9, 3.660476537088871*^9}, {3.667734983133559*^9, 
   3.667735012586729*^9}},
 CellTags->{
  "ID:972417912", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet shape"},
 CellID->972417912],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"f", "\[Element]", "F"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Or]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"f", "\[Equal]", 
         RowBox[{"Pair", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", "b"}], "]"}]}]},
       {
        RowBox[{
         UnderscriptBox["\[Exists]", 
          RowBox[{"g", "\[Element]", "F"}]], 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"f", "\[Equal]", 
           RowBox[{"Pair", "[", 
            RowBox[{
             RowBox[{"succ", "[", 
              RowBox[{"fst", "[", "g", "]"}], "]"}], ",", 
             RowBox[{"c", "[", 
              RowBox[{
               RowBox[{"fst", "[", "g", "]"}], ",", 
               RowBox[{"snd", "[", "g", "]"}]}], "]"}]}], "]"}]}], 
          TagBox[")",
           "AutoParentheses"]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(natRecSet cases)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:57153239", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659847130831457*^9, 3.6598471930453663`*^9}, {
   3.659848542970544*^9, 3.6598485472449512`*^9}, 3.659849976898264*^9, 
   3.6598500196901393`*^9, 3.6604748693941417`*^9, {3.660474901374198*^9, 
   3.660474904182203*^9}, {3.667734983133827*^9, 3.667735012586924*^9}},
 CellTags->{
  "ID:57153239", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet cases"},
 CellID->57153239],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "P"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{"x", ",", "y"}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{"\[Not]", 
               RowBox[{
                RowBox[{"Pair", "[", 
                 RowBox[{"x", ",", "y"}], "]"}], "\[Element]", "F"}]}], 
              TagBox[")",
               "AutoParentheses"]}], "\[Implies]", 
             RowBox[{"P", "[", 
              RowBox[{"x", ",", "y"}], "]"}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{"P", "[", 
           RowBox[{
            RowBox[{"{", "}"}], ",", "b"}], "]"}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             RowBox[{"isSet", "[", "x", "]"}], ",", "y"}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{
               RowBox[{"P", "[", 
                RowBox[{"x", ",", "y"}], "]"}], "\[And]", 
               RowBox[{
                RowBox[{"Pair", "[", 
                 RowBox[{"x", ",", "y"}], "]"}], "\[Element]", "F"}]}], 
              TagBox[")",
               "AutoParentheses"]}], "\[Implies]", 
             RowBox[{"P", "[", 
              RowBox[{
               RowBox[{"succ", "[", "x", "]"}], ",", 
               RowBox[{"c", "[", 
                RowBox[{"x", ",", "y"}], "]"}]}], "]"}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{"x", ",", "y"}]], 
      RowBox[{"P", "[", 
       RowBox[{"x", ",", "y"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(natRecSet induction')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:240965330", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6597712885668755`*^9, 3.659771308737711*^9}, {
   3.6597713442745733`*^9, 3.659771505831457*^9}, 3.659849976913864*^9, 
   3.659850019705739*^9, 3.6604748693941417`*^9, {3.66047490809781*^9, 
   3.660474911171015*^9}, {3.6604765680393257`*^9, 3.660476572454133*^9}, {
   3.667734983134095*^9, 3.667735012587118*^9}},
 CellTags->{
  "ID:240965330", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet induction'"},
 CellID->240965330],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{"x1", ",", "y1"}]], 
   RowBox[{
    RowBox[{"Pair", "[", 
     RowBox[{"x1", ",", "y1"}], "]"}], "\[Element]", "F"}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{"x2", ",", "y2"}]], 
    RowBox[{
     RowBox[{"Pair", "[", 
      RowBox[{"x2", ",", "y2"}], "]"}], "\[Element]", "F"}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{"x1", "\[Equal]", "x2"}], "\[Implies]", 
     RowBox[{"y1", "\[Equal]", "y2"}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(natRecSet functional)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:207844390", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6597716819401665`*^9, 3.6597717191294317`*^9}, {
   3.6598494355401125`*^9, 3.659849497052021*^9}, 3.659849976913864*^9, 
   3.659850019721339*^9, {3.6598507319445906`*^9, 3.659850770554658*^9}, 
   3.6604748693941417`*^9, {3.667734983134362*^9, 3.6677350125873117`*^9}},
 CellTags->{
  "ID:207844390", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet functional"},
 CellID->207844390],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"f", "\[Element]", "F"}]], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{"n", "\[Element]", "natSet"}]], 
   RowBox[{
    UnderscriptBox["\[Exists]", "y"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"f", "\[Equal]", 
      RowBox[{"Pair", "[", 
       RowBox[{"n", ",", "y"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(natRecSet shape')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:348247310", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659853297236498*^9, 3.6598533407605743`*^9}, 
   3.659853542887729*^9, 3.6677344312256517`*^9, {3.6677348936833773`*^9, 
   3.6677349156034393`*^9}, {3.6677349831346197`*^9, 3.6677350125875063`*^9}},
 CellTags->{
  "ID:348247310", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet shape'"},
 CellID->348247310],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"n", "\[Element]", "natSet"}]], 
  RowBox[{
   UnderscriptBox["\[Exists]", "y"], 
   RowBox[{
    RowBox[{"Pair", "[", 
     RowBox[{"n", ",", "y"}], "]"}], "\[Element]", 
    "F"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(natRecSet total)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:117940020", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.667734918066977*^9, {3.6677349831347647`*^9, 3.6677350125876513`*^9}},
 CellTags->{
  "ID:117940020", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet total"},
 CellID->117940020],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellChangeTimes->{
  3.659849976913864*^9, 3.659850019721339*^9, 3.6604748693941417`*^9, {
   3.667734983134832*^9, 3.667735012587718*^9}},
 CellID->893683]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellChangeTimes->{
  3.659850019721339*^9, {3.667734983134869*^9, 3.6677350125877523`*^9}},
 CellID->18123018],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(natRecSet induction)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(natRecSet induction)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["b", "ExpressionVariable"], ",", 
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               RowBox[{"Pair", "[", 
                 RowBox[{"\[EmptySet]", ",", 
                   StyleBox["b", "ExpressionVariable"]}], "]"}], "]"}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["f", "ExpressionVariable"], "\[Element]", 
                    RowBox[{"natRecSet", "[", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   StyleBox["f", "ExpressionVariable"], "]"}], "\[Implies]", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   RowBox[{"Pair", "[", 
                    RowBox[{
                    RowBox[{"succ", "[", 
                    RowBox[{"fst", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], "]"}], ",", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "[", 
                    RowBox[{
                    RowBox[{"fst", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"snd", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}], "]"}]}], 
                    "]"}], "]"}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["f", "ExpressionVariable"], "\[Element]", 
                   RowBox[{"natRecSet", "[", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}]}}]], 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}]}]}], 
       TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["b", "ExpressionVariable"], ",", 
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               RowBox[{"Pair", "[", 
                 RowBox[{"\[EmptySet]", ",", 
                   StyleBox["b", "ExpressionVariable"]}], "]"}], "]"}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    StyleBox["f", "ExpressionVariable"], "\[Element]", 
                    RowBox[{"natRecSet", "[", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   StyleBox["f", "ExpressionVariable"], "]"}], "\[Implies]", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   RowBox[{"Pair", "[", 
                    RowBox[{
                    RowBox[{"succ", "[", 
                    RowBox[{"fst", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], "]"}], ",", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "[", 
                    RowBox[{
                    RowBox[{"fst", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"snd", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}], "]"}]}], 
                    "]"}], "]"}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["f", "ExpressionVariable"], "\[Element]", 
                   RowBox[{"natRecSet", "[", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}]}}]], 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}]}]}], 
       TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:79522091"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p79522091-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p79522091-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:79522091", "Proof|ID:79522091-1"},
 CellID->1623641860],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p79522091-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1522358468]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(natRecSet isSet)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(natRecSet isSet)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["b", "ExpressionVariable"], ",", 
            StyleBox["c", "ExpressionVariable"]}]], 
         RowBox[{"isSet", "[", 
           RowBox[{"natRecSet", "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], "]"}]}], 
       TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["b", "ExpressionVariable"], ",", 
            StyleBox["c", "ExpressionVariable"]}]], 
         RowBox[{"isSet", "[", 
           RowBox[{"natRecSet", "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}], "]"}]}], 
       TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:212626428"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p212626428-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p212626428-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:212626428", "Proof|ID:212626428-1"},
 CellID->2101062054],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p212626428-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->699276041]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(base \[Element] natRecSet)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(base \[Element] natRecSet)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["b", "ExpressionVariable"], ",", 
            StyleBox["c", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"Pair", "[", 
             RowBox[{"\[EmptySet]", ",", 
               StyleBox["b", "ExpressionVariable"]}], "]"}], "\[Element]", 
           RowBox[{"natRecSet", "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}]}]}], 
       TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["b", "ExpressionVariable"], ",", 
            StyleBox["c", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"Pair", "[", 
             RowBox[{"\[EmptySet]", ",", 
               StyleBox["b", "ExpressionVariable"]}], "]"}], "\[Element]", 
           RowBox[{"natRecSet", "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}]}]}], 
       TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:433946629"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p433946629-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p433946629-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:433946629", "Proof|ID:433946629-1"},
 CellID->61218476],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p433946629-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->1894724465]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(natRecSet inductive)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(natRecSet inductive)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "\[Element]", 
               RowBox[{"natRecSet", "[", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}]}]}}]], 
         RowBox[{
           RowBox[{"Pair", "[", 
             RowBox[{
               RowBox[{"succ", "[", 
                 RowBox[{"fst", "[", 
                   StyleBox["f", "ExpressionVariable"], "]"}], "]"}], ",", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "[", 
                 RowBox[{
                   RowBox[{"fst", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{"snd", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}], "]"}]}], 
             "]"}], "\[Element]", 
           RowBox[{"natRecSet", "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}]}]}], 
       TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "\[Element]", 
               RowBox[{"natRecSet", "[", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}]}]}}]], 
         RowBox[{
           RowBox[{"Pair", "[", 
             RowBox[{
               RowBox[{"succ", "[", 
                 RowBox[{"fst", "[", 
                   StyleBox["f", "ExpressionVariable"], "]"}], "]"}], ",", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "[", 
                 RowBox[{
                   RowBox[{"fst", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}], ",", 
                   RowBox[{"snd", "[", 
                    StyleBox["f", "ExpressionVariable"], "]"}]}], "]"}]}], 
             "]"}], "\[Element]", 
           RowBox[{"natRecSet", "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], "]"}]}]}], 
       TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:198301498"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p198301498-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p198301498-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:198301498", "Proof|ID:198301498-1"},
 CellID->1431447619],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p198301498-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->867263181]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(natRecSet shape)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(natRecSet shape)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "\[Element]", 
               RowBox[{"natRecSet", "[", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{"isSet", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}, {
               StyleBox["y", "ExpressionVariable"]}}]], 
           RowBox[{
             StyleBox["f", "ExpressionVariable"], "\[Equal]", 
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "\[Element]", 
               RowBox[{"natRecSet", "[", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{"isSet", "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}]}, {
               StyleBox["y", "ExpressionVariable"]}}]], 
           RowBox[{
             StyleBox["f", "ExpressionVariable"], "\[Equal]", 
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:972417912"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p972417912-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p972417912-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:972417912", "Proof|ID:972417912-1"},
 CellID->196372021],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p972417912-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->128215566]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(natRecSet cases)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(natRecSet cases)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "\[Element]", 
               RowBox[{"natRecSet", "[", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["f", "ExpressionVariable"], "\[Equal]", 
             RowBox[{"Pair", "[", 
               RowBox[{"\[EmptySet]", ",", 
                 StyleBox["b", "ExpressionVariable"]}], "]"}]}], "\[Or]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["g", "ExpressionVariable"], "\[Element]", 
                   RowBox[{"natRecSet", "[", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}]}}]], 
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"Pair", "[", 
                 RowBox[{
                   RowBox[{"succ", "[", 
                    RowBox[{"fst", "[", 
                    StyleBox["g", "ExpressionVariable"], "]"}], "]"}], ",", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], "[", 
                    RowBox[{
                    RowBox[{"fst", "[", 
                    StyleBox["g", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"snd", "[", 
                    StyleBox["g", "ExpressionVariable"], "]"}]}], "]"}]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "\[Element]", 
               RowBox[{"natRecSet", "[", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["f", "ExpressionVariable"], "\[Equal]", 
             RowBox[{"Pair", "[", 
               RowBox[{"\[EmptySet]", ",", 
                 StyleBox["b", "ExpressionVariable"]}], "]"}]}], "\[Or]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   StyleBox["g", "ExpressionVariable"], "\[Element]", 
                   RowBox[{"natRecSet", "[", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}]}}]], 
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"Pair", "[", 
                 RowBox[{
                   RowBox[{"succ", "[", 
                    RowBox[{"fst", "[", 
                    StyleBox["g", "ExpressionVariable"], "]"}], "]"}], ",", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], "[", 
                    RowBox[{
                    RowBox[{"fst", "[", 
                    StyleBox["g", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{"snd", "[", 
                    StyleBox["g", "ExpressionVariable"], "]"}]}], "]"}]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:57153239"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p57153239-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p57153239-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:57153239", "Proof|ID:57153239-1"},
 CellID->736236675],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p57153239-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->197934379]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(natRecSet induction')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(natRecSet induction')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["b", "ExpressionVariable"], ",", 
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], ",", 
                  StyleBox["y", "ExpressionVariable"]}]], 
               RowBox[{
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    RowBox[{"Pair", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], 
                    "\[Element]", 
                    RowBox[{"natRecSet", "[", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}]}], 
             "\[And]", 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               RowBox[{"\[EmptySet]", ",", 
                 StyleBox["b", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{"isSet", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}, {
                   StyleBox["y", "ExpressionVariable"]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "\[And]", 
                   RowBox[{
                    RowBox[{"Pair", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], 
                    "\[Element]", 
                    RowBox[{"natRecSet", "[", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   RowBox[{
                    RowBox[{"succ", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
                   "]"}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              RowBox[{
                StyleBox["x", "ExpressionVariable"], ",", 
                StyleBox["y", "ExpressionVariable"]}]], 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["b", "ExpressionVariable"], ",", 
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                RowBox[{
                  StyleBox["x", "ExpressionVariable"], ",", 
                  StyleBox["y", "ExpressionVariable"]}]], 
               RowBox[{
                 RowBox[{"\[Not]", 
                   RowBox[{
                    TagBox["(", "AutoParentheses"], 
                    RowBox[{
                    RowBox[{"Pair", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], 
                    "\[Element]", 
                    RowBox[{"natRecSet", "[", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}], 
                    TagBox[")", "AutoParentheses"]}]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}]}], 
             "\[And]", 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               RowBox[{"\[EmptySet]", ",", 
                 StyleBox["b", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{"isSet", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}, {
                   StyleBox["y", "ExpressionVariable"]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], "\[And]", 
                   RowBox[{
                    RowBox[{"Pair", "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}], 
                    "\[Element]", 
                    RowBox[{"natRecSet", "[", 
                    RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}]}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   RowBox[{
                    RowBox[{"succ", "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["x", "ExpressionVariable"], ",", 
                    StyleBox["y", "ExpressionVariable"]}], "]"}]}], 
                   "]"}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              RowBox[{
                StyleBox["x", "ExpressionVariable"], ",", 
                StyleBox["y", "ExpressionVariable"]}]], 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               RowBox[{
                 StyleBox["x", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:240965330"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p240965330-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p240965330-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:240965330", "Proof|ID:240965330-1"},
 CellID->221211846],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p240965330-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.6615193402077456`*^9, 3.6615193427505503`*^9}},
 CellID->295739411]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(natRecSet functional)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(natRecSet functional)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["b", "ExpressionVariable"], ",", 
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["x1", "ExpressionVariable"], ",", 
            StyleBox["y1", "ExpressionVariable"], ",", 
            StyleBox["x2", "ExpressionVariable"], ",", 
            StyleBox["y2", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["x1", "ExpressionVariable"], ",", 
                   StyleBox["y1", "ExpressionVariable"]}], "]"}], 
               "\[Element]", 
               RowBox[{"natRecSet", "[", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[And]", 
             RowBox[{
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["x2", "ExpressionVariable"], ",", 
                   StyleBox["y2", "ExpressionVariable"]}], "]"}], 
               "\[Element]", 
               RowBox[{"natRecSet", "[", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[And]", 
             RowBox[{
               StyleBox["x1", "ExpressionVariable"], "\[Equal]", 
               StyleBox["x2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["y1", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "y2", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["b", "ExpressionVariable"], ",", 
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["x1", "ExpressionVariable"], ",", 
            StyleBox["y1", "ExpressionVariable"], ",", 
            StyleBox["x2", "ExpressionVariable"], ",", 
            StyleBox["y2", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["x1", "ExpressionVariable"], ",", 
                   StyleBox["y1", "ExpressionVariable"]}], "]"}], 
               "\[Element]", 
               RowBox[{"natRecSet", "[", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[And]", 
             RowBox[{
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["x2", "ExpressionVariable"], ",", 
                   StyleBox["y2", "ExpressionVariable"]}], "]"}], 
               "\[Element]", 
               RowBox[{"natRecSet", "[", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[And]", 
             RowBox[{
               StyleBox["x1", "ExpressionVariable"], "\[Equal]", 
               StyleBox["x2", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["y1", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "y2", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:207844390"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p207844390-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p207844390-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:207844390", "Proof|ID:207844390-1"},
 CellID->201549000],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p207844390-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->386042070]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(natRecSet shape')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(natRecSet shape')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "\[Element]", 
               RowBox[{"natRecSet", "[", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], "\[Element]", 
                 "natSet"}]}}]], 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              RowBox[{
                StyleBox["y", "ExpressionVariable"]}]], 
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "\[Element]", 
               RowBox[{"natRecSet", "[", 
                 RowBox[{
                   StyleBox["b", "ExpressionVariable"], ",", 
                   StyleBox["c", "ExpressionVariable"]}], "]"}]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], "\[Element]", 
                 "natSet"}]}}]], 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              RowBox[{
                StyleBox["y", "ExpressionVariable"]}]], 
             RowBox[{
               StyleBox["f", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["n", "ExpressionVariable"], ",", 
                   StyleBox["y", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:348247310"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p348247310-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p348247310-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:348247310", "Proof|ID:348247310-1"},
 CellID->22594877],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p348247310-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->171131707]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(natRecSet total)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(natRecSet total)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "natSet"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            RowBox[{
              StyleBox["y", "ExpressionVariable"]}]], 
           RowBox[{
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Element]", 
             RowBox[{"natRecSet", "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "natSet"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            RowBox[{
              StyleBox["y", "ExpressionVariable"]}]], 
           RowBox[{
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], ",", 
                 StyleBox["y", "ExpressionVariable"]}], "]"}], "\[Element]", 
             RowBox[{"natRecSet", "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:117940020"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p117940020-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p117940020-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:117940020", "Proof|ID:117940020-1"},
 CellID->128642575],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p117940020-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->27786247]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->50539674],

Cell[CellGroupData[{

Cell["\<\
Lemma: Characteristic Properties of natRec\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6598532443202047`*^9, 3.6598532492030134`*^9}, {
  3.667734644455624*^9, 3.667734647134903*^9}},
 CellID->155324481],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"natRec", "[", 
    RowBox[{"b", ",", "c"}], "]"}], "[", 
   RowBox[{"{", "}"}], "]"}], "\[Equal]", "b"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(natRec base)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:127781311", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65985344127155*^9, 3.6598534470903606`*^9}, {
  3.660479028251647*^9, 3.6604790285792475`*^9}},
 CellTags->{
  "ID:127781311", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRec base", "IR"},
 CellID->127781311],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"n", "\[Element]", "natSet"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"natRec", "[", 
      RowBox[{"b", ",", "c"}], "]"}], "[", 
     RowBox[{"succ", "[", "n", "]"}], "]"}], "\[Equal]", 
    RowBox[{"c", "[", 
     RowBox[{"n", ",", 
      RowBox[{
       RowBox[{"natRec", "[", 
        RowBox[{"b", ",", "c"}], "]"}], "[", "n", "]"}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(natRec step)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:186992137", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6598534522851696`*^9, 3.6598534922188396`*^9}, {
  3.6604790308880515`*^9, 3.660479032292054*^9}},
 CellTags->{
  "ID:186992137", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRec step", "IR"},
 CellID->186992137],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->55183838]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->389335050],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(natRec base)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(natRec base)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["b", "ExpressionVariable"], ",", 
            StyleBox["c", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"natRec", "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}], "[", 
             "\[EmptySet]", "]"}], "\[Equal]", 
           StyleBox[
           "b", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["b", "ExpressionVariable"], ",", 
            StyleBox["c", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"natRec", "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}], "[", 
             "\[EmptySet]", "]"}], "\[Equal]", 
           StyleBox[
           "b", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:127781311"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p127781311-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p127781311-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:127781311", "Proof|ID:127781311-1"},
 CellID->5572876],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p127781311-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->462709909]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(natRec step)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(natRec step)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "natSet"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"natRec", "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}], "[", 
             RowBox[{"succ", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}], "]"}], "\[Equal]", 
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "[", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], ",", 
               RowBox[{
                 RowBox[{"natRec", "[", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], "[", 
                 StyleBox["n", "ExpressionVariable"], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             StyleBox["b", "ExpressionVariable"]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               "natSet"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"natRec", "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}], "[", 
             RowBox[{"succ", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}], "]"}], "\[Equal]", 
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "[", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], ",", 
               RowBox[{
                 RowBox[{"natRec", "[", 
                   RowBox[{
                    StyleBox["b", "ExpressionVariable"], ",", 
                    StyleBox["c", "ExpressionVariable"]}], "]"}], "[", 
                 StyleBox["n", "ExpressionVariable"], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:186992137"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p186992137-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p186992137-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:186992137", "Proof|ID:186992137-1"},
 CellID->546211],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p186992137-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->116871600]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Nat Functor (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.659855469549115*^9, 3.6598554821851377`*^9}, {
   3.6598607936625404`*^9, 3.6598607962989454`*^9}, {3.660224276760369*^9, 
   3.660224280800776*^9}, 3.6603714769345407`*^9, {3.660386226503024*^9, 
   3.6603862304030313`*^9}, 3.66047969578582*^9, {3.660643498562272*^9, 
   3.6606435026182795`*^9}},
 CellID->748923816],

Cell[CellGroupData[{

Cell["Definitions", "Subsubsection",
 CellChangeTimes->{{3.6598614989807796`*^9, 3.6598615005719824`*^9}},
 CellID->424914336],

Cell["", "OpenEnvironment",
 CellID->79952930],

Cell[CellGroupData[{

Cell["Definition: Nat", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.65986080481656*^9, 3.6598608077805653`*^9}},
 CellID->635773539],

Cell[BoxData[
 RowBox[{"Nat", ":=", 
  UnderscriptBox["\[CapitalDelta]", "N"]}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          379686829}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659860813989376*^9, 3.6598608199485865`*^9}},
 CellTags->{
  "ID:379686829", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->379686829],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"a", ",", "b"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          778005153}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659860826438198*^9, 3.6598608282166014`*^9}},
 CellTags->{
  "ID:778005153", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->778005153],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", "N"], 
    RowBox[{"[", "a", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a:=b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", "\[Element]", "natSet"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Nat \[Element])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:97787251", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659860833817011*^9, 3.6598608518194427`*^9}},
 CellTags->{
  "ID:97787251", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat \[Element]"},
 CellID->97787251],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["0", "N"], ":=", 
  RowBox[{"{", "}"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Nat 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:720795295", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659861393992595*^9, 3.6598613970814004`*^9}, {
  3.6604797517899184`*^9, 3.660479752133119*^9}},
 CellTags->{
  "ID:720795295", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat 0"},
 CellID->720795295],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["1", "N"], ":=", 
  RowBox[{"succ", "[", 
   UnderscriptBox["0", "Nat"], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Nat 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:3778917", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6598614018082085`*^9, 3.659861406675417*^9}, {
  3.6604797543171225`*^9, 3.6604797576243286`*^9}, {3.6604808441526375`*^9, 
  3.6604808466486416`*^9}},
 CellTags->{
  "ID:3778917", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat 1"},
 CellID->3778917],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["Pred", "N"], ":=", 
  RowBox[{"natRec", "[", 
   RowBox[{
    RowBox[{"{", "}"}], ",", 
    RowBox[{
     UnderscriptBox["\[Lambda]", 
      RowBox[{"x", ",", "y"}]], "x"}]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Nat Pred)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:54060264", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6603719047776933`*^9, 3.6603719385995526`*^9}, {
  3.6604797626475377`*^9, 3.6604797629751377`*^9}},
 CellTags->{
  "ID:54060264", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat Pred"},
 CellID->54060264],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox["+", "N"], "b"}], ":=", 
  RowBox[{
   RowBox[{"natRec", "[", 
    RowBox[{"a", ",", 
     RowBox[{
      UnderscriptBox["\[Lambda]", 
       RowBox[{"x", ",", "y"}]], 
      RowBox[{"succ", "[", "y", "]"}]}]}], "]"}], "[", "b", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Nat +)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:205601935", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.6598608721618786`*^9, {3.6598609176983585`*^9, 3.6598609459656076`*^9}, {
   3.660479765627143*^9, 3.6604797660951433`*^9}},
 CellTags->{
  "ID:205601935", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat +"},
 CellID->205601935],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox["-", "N"], "b"}], ":=", 
  RowBox[{
   RowBox[{"natRec", "[", 
    RowBox[{"a", ",", 
     RowBox[{
      UnderscriptBox["\[Lambda]", 
       RowBox[{"x", ",", "y"}]], 
      RowBox[{
       UnderscriptBox["Pred", "N"], "[", "y", "]"}]}]}], "]"}], "[", "b", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Nat -)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:74238769", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660371961517993*^9, 3.6603720174908915`*^9}},
 CellTags->{
  "ID:74238769", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat -"},
 CellID->74238769],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox["*", "N"], "b"}], ":=", 
  RowBox[{
   RowBox[{"natRec", "[", 
    RowBox[{
     RowBox[{"{", "}"}], ",", 
     RowBox[{
      UnderscriptBox["\[Lambda]", 
       RowBox[{"x", ",", "y"}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"y", 
        UnderscriptBox["+", "Nat"], "a"}], 
       TagBox[")",
        "AutoParentheses"]}]}]}], "]"}], "[", "b", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Nat *)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:288612943", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.6598608753598843`*^9, {3.6598609525176196`*^9, 3.659860995994896*^9}, {
   3.6602144390188828`*^9, 3.660214440376085*^9}, {3.660479772194754*^9, 
   3.660479772506755*^9}},
 CellTags->{
  "ID:288612943", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat *"},
 CellID->288612943],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["\[GreaterEqual]", "N"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a:=b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"b", "\[Element]", 
    RowBox[{"succ", "[", "a", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Nat \[GreaterEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:155532301", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659860878043089*^9, 3.6598608813346944`*^9}, {
   3.659861001642106*^9, 3.6598610559400015`*^9}, {3.6598612468687367`*^9, 
   3.6598612844180026`*^9}, 3.6598613343068905`*^9, {3.6605430794074802`*^9, 
   3.6605430899062986`*^9}, {3.6605561203743978`*^9, 
   3.6605561248984056`*^9}, {3.6605561845851107`*^9, 
   3.6605561963475313`*^9}, {3.660556249730825*^9, 3.660556252944431*^9}, {
   3.6605567886815715`*^9, 3.6605567961695848`*^9}, {3.6605568675241103`*^9, 
   3.660556870846916*^9}},
 CellTags->{
  "ID:155532301", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat \[GreaterEqual]"},
 CellID->155532301],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->217986896]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->22540073]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Lemmas: General Properties of Nat and Pred (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.6598615067183933`*^9, 3.6598615075451946`*^9}, {
   3.6602242681803536`*^9, 3.6602242725483613`*^9}, 3.660286230969328*^9, {
   3.660372079922201*^9, 3.660372093634625*^9}, {3.660372166096752*^9, 
   3.660372168296356*^9}, 3.6604796973614225`*^9, {3.6604811344379473`*^9, 
   3.6604811389307547`*^9}},
 CellID->512561655],

Cell["", "OpenEnvironment",
 CellID->399148770],

Cell[CellGroupData[{

Cell["Lemma: Properties of Nat", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.659861538417649*^9, 3.659861541459654*^9}, {
  3.6598641978225203`*^9, 3.65986420340733*^9}},
 CellID->43728960],

Cell[BoxData[
 RowBox[{"isDomain", "[", "Nat", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Nat isDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:15394254", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659861551147271*^9, 3.659861554735277*^9}},
 CellTags->{
  "ID:15394254", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat isDomain"},
 CellID->15394254],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "P"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"P", "[", 
       UnderscriptBox["0", "Nat"], "]"}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          UnderscriptBox["\[Element]", "Nat"], 
          RowBox[{"[", "n", "]"}]}]], 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          RowBox[{"P", "[", "n", "]"}], "\[Implies]", 
          RowBox[{"P", "[", 
           RowBox[{"succ", "[", "n", "]"}], "]"}]}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       UnderscriptBox["\[Element]", "Nat"], 
       RowBox[{"[", "n", "]"}]}]], 
     RowBox[{"P", "[", "n", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Nat induction)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:478904098", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6598620361843233`*^9, 3.659862093576824*^9}, {
  3.6598626644118266`*^9, 3.6598626648018274`*^9}, {3.6604798233472443`*^9, 
  3.660479823877645*^9}},
 CellTags->{
  "ID:478904098", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat induction", "IR"},
 CellID->478904098],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "Nat"], 
  RowBox[{"[", 
   UnderscriptBox["0", "Nat"], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(0 \[Element] Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:60850406", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6598626328841715`*^9, 3.6598626427589884`*^9}},
 CellTags->{
  "ID:60850406", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 \[Element] Nat", "IR"},
 CellID->60850406],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "Nat"], 
    RowBox[{"[", "n", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[Element]", "Nat"], 
   RowBox[{"[", 
    RowBox[{"succ", "[", "n", "]"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(succ \[Element] Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:114723517", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6598627155331163`*^9, 3.6598627326931467`*^9}, {
  3.6604798287448535`*^9, 3.6604798292908545`*^9}},
 CellTags->{
  "ID:114723517", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ \[Element] Nat", "IR"},
 CellID->114723517],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "Nat"], 
  RowBox[{"[", 
   UnderscriptBox["1", "Nat"], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1 \[Element] Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:275276526", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660394149852155*^9, 3.660394151193757*^9}},
 CellTags->{
  "ID:275276526", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 \[Element] Nat", "IR"},
 CellID->275276526],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "Nat"], 
   RowBox[{"[", "n", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          65979872}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6605597804570274`*^9, 3.6605597867906384`*^9}},
 CellTags->{
  "ID:65979872", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->65979872],

Cell[BoxData[
 RowBox[{"isSet", "[", "n", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Nat shape)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:572443783", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6604798964645724`*^9, 3.6604799065421906`*^9}, 
   3.660559791642247*^9},
 CellTags->{
  "ID:572443783", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat shape"},
 CellID->572443783],

Cell[BoxData[
 RowBox[{"\[Or]", 
  RowBox[{"\[Piecewise]", GridBox[{
     {
      RowBox[{"n", "\[Equal]", 
       UnderscriptBox["0", "Nat"]}]},
     {
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{
         UnderscriptBox["\[Element]", "Nat"], 
         RowBox[{"[", "m", "]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"n", "\[Equal]", 
         RowBox[{"succ", "[", "m", "]"}]}], 
        TagBox[")",
         "AutoParentheses"]}]}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Nat cases)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:252333509", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659864102163152*^9, 3.6598641291667995`*^9}, {
  3.6604798348756647`*^9, 3.6604798353904653`*^9}, {3.6605597929838495`*^9, 
  3.660559795074253*^9}},
 CellTags->{
  "ID:252333509", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat cases"},
 CellID->252333509],

Cell[BoxData[
 RowBox[{
  RowBox[{"succ", "[", "n", "]"}], "\[NotEqual]", 
  UnderscriptBox["0", "Nat"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(succ \[NotEqual] 0 Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:77756457", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660559730209339*^9, 3.660559751581377*^9}, 
   3.66055979902106*^9},
 CellTags->{
  "ID:77756457", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ \[NotEqual] 0 Nat"},
 CellID->77756457],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "Nat"], 
    RowBox[{"[", "m", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"succ", "[", "n", "]"}], "\[Equal]", 
     RowBox[{"succ", "[", "m", "]"}]}], "\[Implies]", 
    RowBox[{"n", "\[Equal]", "m"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(succ injective Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:52299395", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6605598453219414`*^9, 3.6605598664173784`*^9}},
 CellTags->{
  "ID:52299395", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ injective Nat"},
 CellID->52299395],

Cell[BoxData[
 RowBox[{
  RowBox[{"succ", "[", "n", "]"}], "\[NotEqual]", 
  "n"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(succ no fixpoint Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:263281689", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660561008594985*^9, 3.66056101130939*^9}},
 CellTags->{
  "ID:263281689", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ no fixpoint Nat"},
 CellID->263281689],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["1", "Nat"], "\[NotEqual]", 
  UnderscriptBox["0", "Nat"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1 \[NotEqual] 0 Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:377798869", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612379653887663`*^9, 3.6612379733759804`*^9}},
 CellTags->{
  "ID:377798869", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 \[NotEqual] 0 Nat"},
 CellID->377798869],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"m", "\[Element]", "n"}]], 
  RowBox[{
   UnderscriptBox["\[Element]", "Nat"], 
   RowBox[{"[", "m", "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] closed Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:33739458", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6605569743218975`*^9, 3.660557011902364*^9}, {
  3.660559802312666*^9, 3.6605598071798744`*^9}},
 CellTags->{
  "ID:33739458", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[Element] closed Nat"},
 CellID->33739458],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"m", "\[Element]", "n"}]], 
  RowBox[{"m", "\[SubsetEqual]", "n"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Nat transitive)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:654854303", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6605478250114193`*^9, 3.6605478403462462`*^9}, {
  3.6605543677689176`*^9, 3.6605543722929254`*^9}, {3.660555151251094*^9, 
  3.660555155572301*^9}, {3.6605568165588207`*^9, 3.660556820833228*^9}, {
  3.6605569525442595`*^9, 3.660556955617465*^9}, {3.660556988049922*^9, 
  3.6605570271591907`*^9}, {3.66055981037788*^9, 3.660559817163892*^9}},
 CellTags->{
  "ID:654854303", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat transitive"},
 CellID->654854303],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->241267144]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->142415995],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Nat isDomain)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Nat isDomain)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isDomain", "[", "Nat", "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isDomain", "[", "Nat", "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:15394254"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p15394254-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p15394254-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:15394254", "Proof|ID:15394254-1"},
 CellID->198699355],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p15394254-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->314318358]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Nat induction)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Nat induction)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               UnderscriptBox["0", "Nat"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "Nat"], "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}], "\[Implies]", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   RowBox[{"succ", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "Nat"], "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               StyleBox["n", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               UnderscriptBox["0", "Nat"], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "Nat"], "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}], "\[Implies]", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   RowBox[{"succ", "[", 
                    StyleBox["n", "ExpressionVariable"], "]"}], "]"}]}]}]}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "Nat"], "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               StyleBox["n", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:478904098"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p478904098-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p478904098-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:478904098", "Proof|ID:478904098-1"},
 CellID->65480230],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p478904098-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->5553732]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(0 \[Element] Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(0 \[Element] Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", "Nat"], "[", 
         UnderscriptBox["0", "Nat"], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", "Nat"], "[", 
         UnderscriptBox["0", "Nat"], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:60850406"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p60850406-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p60850406-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:60850406", "Proof|ID:60850406-1"},
 CellID->291459920],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p60850406-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->94542070]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(succ \[Element] Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(succ \[Element] Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Nat"], "[", 
           RowBox[{"succ", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Nat"], "[", 
           RowBox[{"succ", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:114723517"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p114723517-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p114723517-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:114723517", "Proof|ID:114723517-1"},
 CellID->310507632],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p114723517-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->740339103]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(1 \[Element] Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(1 \[Element] Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", "Nat"], "[", 
         UnderscriptBox["1", "Nat"], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", "Nat"], "[", 
         UnderscriptBox["1", "Nat"], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:275276526"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p275276526-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p275276526-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:275276526", "Proof|ID:275276526-1"},
 CellID->58288735],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p275276526-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->411515486]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Nat shape)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Nat shape)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isSet", "[", 
           StyleBox["n", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isSet", "[", 
           StyleBox["n", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:572443783"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p572443783-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p572443783-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:572443783", "Proof|ID:572443783-1"},
 CellID->43340428],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p572443783-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->73111381]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Nat cases)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Nat cases)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["n", "ExpressionVariable"], "\[Equal]", 
             UnderscriptBox["0", "Nat"]}], "\[Or]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "Nat"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"succ", "[", 
                 StyleBox["m", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["n", "ExpressionVariable"], "\[Equal]", 
             UnderscriptBox["0", "Nat"]}], "\[Or]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "Nat"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"succ", "[", 
                 StyleBox["m", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:252333509"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p252333509-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p252333509-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:252333509", "Proof|ID:252333509-1"},
 CellID->179637041],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p252333509-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->128846745]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(succ \[NotEqual] 0 Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(succ \[NotEqual] 0 Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"succ", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], "\[NotEqual]", 
           UnderscriptBox["0", "Nat"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"succ", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], "\[NotEqual]", 
           UnderscriptBox["0", "Nat"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:77756457"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p77756457-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p77756457-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:77756457", "Proof|ID:77756457-1"},
 CellID->41859577],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p77756457-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->154826070]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(succ injective Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(succ injective Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"succ", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{"succ", "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             StyleBox["n", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "m", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"succ", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{"succ", "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}], "\[Implies]", 
           RowBox[{
             StyleBox["n", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "m", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:52299395"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p52299395-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p52299395-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:52299395", "Proof|ID:52299395-1"},
 CellID->590942562],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p52299395-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->325144463]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(succ no fixpoint Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(succ no fixpoint Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"succ", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], "\[NotEqual]", 
           StyleBox[
           "n", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"succ", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], "\[NotEqual]", 
           StyleBox[
           "n", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:263281689"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p263281689-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p263281689-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:263281689", "Proof|ID:263281689-1"},
 CellID->255034219],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p263281689-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->341857592]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(1 \[NotEqual] 0 Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(1 \[NotEqual] 0 Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["1", "Nat"], "\[NotEqual]", 
         UnderscriptBox["0", "Nat"]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["1", "Nat"], "\[NotEqual]", 
         UnderscriptBox["0", "Nat"]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:377798869"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p377798869-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p377798869-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:377798869", "Proof|ID:377798869-1"},
 CellID->103721996],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p377798869-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->516295583]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] closed Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] closed Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               StyleBox["m", "ExpressionVariable"]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Nat"], "[", 
           StyleBox["n", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               StyleBox["m", "ExpressionVariable"]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Nat"], "[", 
           StyleBox["n", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:33739458"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p33739458-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p33739458-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:33739458", "Proof|ID:33739458-1"},
 CellID->548186621],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p33739458-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->520182775]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Nat transitive)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Nat transitive)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               StyleBox["m", "ExpressionVariable"]}]}}]], 
         RowBox[{
           StyleBox["n", "ExpressionVariable"], "\[SubsetEqual]", 
           StyleBox[
           "m", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               StyleBox["m", "ExpressionVariable"]}]}}]], 
         RowBox[{
           StyleBox["n", "ExpressionVariable"], "\[SubsetEqual]", 
           StyleBox[
           "m", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:654854303"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p654854303-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p654854303-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:654854303", "Proof|ID:654854303-1"},
 CellID->8946125],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p654854303-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->402341539]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->385788823],

Cell[CellGroupData[{

Cell["Lemma: Properties of Pred", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.660372173459965*^9, 3.6603721787795744`*^9}},
 CellID->224309402],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["Pred", "Nat"], "[", 
   UnderscriptBox["0", "Nat"], "]"}], "\[Equal]", 
  UnderscriptBox["0", "Nat"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Pred 0 Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:240175329", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6603721845047846`*^9, 3.6603721990596104`*^9}},
 CellTags->{
  "ID:240175329", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Pred 0 Nat", "IR"},
 CellID->240175329],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "Nat"], 
   RowBox[{"[", "n", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          158284383}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.660372287593766*^9},
 CellTags->{
  "ID:158284383", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->158284383],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["Pred", "Nat"], "[", 
   RowBox[{"succ", "[", "n", "]"}], "]"}], "\[Equal]", 
  "n"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Pred succ Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:145236688", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6603722065320234`*^9, 3.6603722292934637`*^9}, {
  3.6603722861897635`*^9, 3.6603722907605715`*^9}, {3.6604809744908657`*^9, 
  3.6604809749900665`*^9}},
 CellTags->{
  "ID:145236688", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Pred succ Nat", "IR"},
 CellID->145236688],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "Nat"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["Pred", "Nat"], "[", "n", "]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Pred closed Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:225607097", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6603722923361745`*^9, 3.6603723042545953`*^9}},
 CellTags->{
  "ID:225607097", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Pred closed Nat", "IR"},
 CellID->225607097],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->562641991]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->3500975],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Pred 0 Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Pred 0 Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["Pred", "Nat"], "[", 
           UnderscriptBox["0", "Nat"], "]"}], "\[Equal]", 
         UnderscriptBox["0", "Nat"]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         RowBox[{
           UnderscriptBox["Pred", "Nat"], "[", 
           UnderscriptBox["0", "Nat"], "]"}], "\[Equal]", 
         UnderscriptBox["0", "Nat"]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:240175329"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p240175329-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p240175329-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:240175329", "Proof|ID:240175329-1"},
 CellID->415645142],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p240175329-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->7870725]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Pred succ Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Pred succ Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["Pred", "Nat"], "[", 
             RowBox[{"succ", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}], "]"}], "\[Equal]", 
           StyleBox[
           "n", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["Pred", "Nat"], "[", 
             RowBox[{"succ", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}], "]"}], "\[Equal]", 
           StyleBox[
           "n", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:145236688"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p145236688-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p145236688-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:145236688", "Proof|ID:145236688-1"},
 CellID->498357158],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p145236688-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->228630812]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Pred closed Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Pred closed Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Nat"], "[", 
           RowBox[{
             UnderscriptBox["Pred", "Nat"], "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Nat"], "[", 
           RowBox[{
             UnderscriptBox["Pred", "Nat"], "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:225607097"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p225607097-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p225607097-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:225607097", "Proof|ID:225607097-1"},
 CellID->787676860],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p225607097-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->22707532]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemmas: ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " (\[Checkmark])"
}], "Subsubsection",
 CellChangeTimes->{{3.6603721080802507`*^9, 3.660372123914278*^9}, {
  3.660384744078018*^9, 3.66038474507642*^9}, {3.660384799114915*^9, 
  3.660384803170922*^9}, {3.6604796988434258`*^9, 3.660479699670227*^9}, {
  3.6605455470566173`*^9, 3.660545551346625*^9}},
 CellID->254557108],

Cell["", "OpenEnvironment",
 CellID->427738811],

Cell[CellGroupData[{

Cell["Lemma: Properties of + in Nat", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6598615174824123`*^9, 3.6598615328016386`*^9}},
 CellID->928488810],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "Nat"], 
   RowBox[{"[", 
    RowBox[{"m", ",", "n"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          218406943}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659861702592337*^9, 3.659861711765153*^9}, {
  3.659862052408352*^9, 3.6598620528451524`*^9}},
 CellTags->{
  "ID:218406943", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->218406943],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["+", "Nat"], 
   UnderscriptBox["0", "Nat"]}], "\[Equal]", "m"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ 0 Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:257436422", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659861721000369*^9, 3.659861727739581*^9}, {
  3.6598620549979563`*^9, 3.6598620559027576`*^9}, {3.6598626818526573`*^9, 
  3.6598626848010626`*^9}},
 CellTags->{
  "ID:257436422", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ 0 Nat", "IR"},
 CellID->257436422],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["+", "Nat"], 
   RowBox[{"succ", "[", "n", "]"}]}], "\[Equal]", 
  RowBox[{"succ", "[", 
   RowBox[{"m", 
    UnderscriptBox["+", "Nat"], "n"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ succ Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:249353684", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659861762356042*^9, 3.6598617743212633`*^9}, {
  3.6598620569947596`*^9, 3.659862061128767*^9}, {3.659862689512271*^9, 
  3.6598626942234793`*^9}, {3.6605408762570095`*^9, 3.660540883308222*^9}},
 CellTags->{
  "ID:249353684", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ succ Nat", "IR"},
 CellID->249353684],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "Nat"], 
  RowBox[{"[", 
   RowBox[{"m", 
    UnderscriptBox["+", "Nat"], "n"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ closed Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:279658160", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6598635317129498`*^9, 3.659863541556567*^9}},
 CellTags->{
  "ID:279658160", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ closed Nat", "IR"},
 CellID->279658160],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["+", "Nat"], 
   UnderscriptBox["1", "Nat"]}], "\[Equal]", 
  RowBox[{"succ", "[", "m", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ 1 Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:647623443", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6603731111226134`*^9, 3.6603731236650352`*^9}, {
  3.6605408871770287`*^9, 3.6605408875358295`*^9}},
 CellTags->{
  "ID:647623443", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ 1 Nat"},
 CellID->647623443],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["0", "Nat"], 
   UnderscriptBox["+", "Nat"], "n"}], "\[Equal]", 
  "n"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(0 lneutral Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:18327536", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6598620121914806`*^9, 3.6598620204438953`*^9}, {
  3.6598620626419697`*^9, 3.659862063609171*^9}, {3.6598626962670827`*^9, 
  3.6598626972966843`*^9}},
 CellTags->{
  "ID:18327536", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 lneutral Nat", "IR"},
 CellID->18327536],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"succ", "[", "m", "]"}], 
   UnderscriptBox["+", "Nat"], "n"}], "\[Equal]", 
  RowBox[{"succ", "[", 
   RowBox[{"m", 
    UnderscriptBox["+", "Nat"], "n"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(succ + Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1914645", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659863157747093*^9, 3.6598631728947196`*^9}, {
  3.660540891217436*^9, 3.660540895258843*^9}},
 CellTags->{
  "ID:1914645", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ + Nat", "IR"},
 CellID->1914645],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"m", 
     UnderscriptBox["+", "Nat"], "n"}], "\[Equal]", 
    UnderscriptBox["0", "Nat"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"m", "\[Equal]", 
     UnderscriptBox["0", "Nat"]}], "\[And]", 
    RowBox[{"n", "\[Equal]", 
     UnderscriptBox["0", "Nat"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ = 0 Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:343580194", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659864007984787*^9, 3.659864039934643*^9}},
 CellTags->{
  "ID:343580194", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ = 0 Nat"},
 CellID->343580194],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["+", "Nat"], "n"}], "\[Equal]", 
  RowBox[{"n", 
   UnderscriptBox["+", "Nat"], "m"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ comm Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:834229092", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659862375531719*^9, 3.659862389586144*^9}, {
  3.659862698451086*^9, 3.659862699558688*^9}},
 CellTags->{
  "ID:834229092", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ comm Nat"},
 CellID->834229092],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "Nat"], 
   RowBox[{"[", "k", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          303913216}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.6598625181615696`*^9},
 CellTags->{
  "ID:303913216", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->303913216],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"m", 
     UnderscriptBox["+", "Nat"], "n"}], ")"}], 
   UnderscriptBox["+", "Nat"], "k"}], "\[Equal]", 
  RowBox[{"m", 
   UnderscriptBox["+", "Nat"], 
   RowBox[{"(", 
    RowBox[{"n", 
     UnderscriptBox["+", "Nat"], "k"}], ")"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ assoc Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:73705609", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65986246139307*^9, 3.659862521203575*^9}, {
  3.659862701555492*^9, 3.659862706875101*^9}, {3.659863461574227*^9, 
  3.6598634687658396`*^9}},
 CellTags->{
  "ID:73705609", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ assoc Nat"},
 CellID->73705609],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["+", "Nat"], 
   RowBox[{"(", 
    RowBox[{"n", 
     UnderscriptBox["+", "Nat"], "k"}], ")"}]}], "\[Equal]", 
  RowBox[{"n", 
   UnderscriptBox["+", "Nat"], 
   RowBox[{"(", 
    RowBox[{"m", 
     UnderscriptBox["+", "Nat"], "k"}], ")"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ comm-assoc Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:202544617", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661171812918295*^9, 3.661171834040732*^9}},
 CellTags->{
  "ID:202544617", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ comm-assoc Nat"},
 CellID->202544617],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "Nat"], 
    RowBox[{"[", "i", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"m", 
       UnderscriptBox["+", "Nat"], "n"}], ")"}], 
     UnderscriptBox["+", "Nat"], 
     RowBox[{"(", 
      RowBox[{"k", 
       UnderscriptBox["+", "Nat"], "i"}], ")"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"m", 
       UnderscriptBox["+", "Nat"], "k"}], ")"}], 
     UnderscriptBox["+", "Nat"], 
     RowBox[{"(", 
      RowBox[{"n", 
       UnderscriptBox["+", "Nat"], "i"}], ")"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ + + Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:555281086", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661233315975396*^9, 3.661233377673504*^9}},
 CellTags->{
  "ID:555281086", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ + + Nat"},
 CellID->555281086],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"m", 
     UnderscriptBox["+", "Nat"], "k"}], "\[Equal]", 
    RowBox[{"n", 
     UnderscriptBox["+", "Nat"], "k"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"m", "\[Equal]", "n"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ rcanc Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:5229733", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.659862527505986*^9, 3.659862548987224*^9}, {
  3.6598627083883038`*^9, 3.6598627099171066`*^9}},
 CellTags->{
  "ID:5229733", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ rcanc Nat"},
 CellID->5229733],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->23399794]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->559100389],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ 0 Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ 0 Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             UnderscriptBox["0", "Nat"]}], "\[Equal]", 
           StyleBox[
           "m", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             UnderscriptBox["0", "Nat"]}], "\[Equal]", 
           StyleBox[
           "m", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:257436422"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p257436422-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p257436422-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:257436422", "Proof|ID:257436422-1"},
 CellID->197432916],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p257436422-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->565004225]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ succ Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ succ Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             RowBox[{"succ", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], "\[Equal]", 
           RowBox[{"succ", "[", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             RowBox[{"succ", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], "\[Equal]", 
           RowBox[{"succ", "[", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:249353684"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p249353684-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p249353684-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:249353684", "Proof|ID:249353684-1"},
 CellID->376294367],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p249353684-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->108601403]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ closed Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ closed Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Nat"], "[", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Nat"], "[", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:279658160"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p279658160-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p279658160-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:279658160", "Proof|ID:279658160-1"},
 CellID->363010569],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p279658160-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->585218458]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ 1 Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ 1 Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             UnderscriptBox["1", "Nat"]}], "\[Equal]", 
           RowBox[{"succ", "[", 
             StyleBox["m", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             UnderscriptBox["1", "Nat"]}], "\[Equal]", 
           RowBox[{"succ", "[", 
             StyleBox["m", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:647623443"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p647623443-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p647623443-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:647623443", "Proof|ID:647623443-1"},
 CellID->35382149],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p647623443-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->257589816]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(0 lneutral Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(0 lneutral Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["0", "Nat"], 
             UnderscriptBox["+", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
           StyleBox[
           "n", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["0", "Nat"], 
             UnderscriptBox["+", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
           StyleBox[
           "n", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:18327536"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p18327536-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p18327536-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:18327536", "Proof|ID:18327536-1"},
 CellID->794865037],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p18327536-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->71581738]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(succ + Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(succ + Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"succ", "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], 
             UnderscriptBox["+", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{"succ", "[", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"succ", "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], 
             UnderscriptBox["+", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{"succ", "[", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1914645"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1914645-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1914645-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1914645", "Proof|ID:1914645-1"},
 CellID->218839119],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1914645-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->337393823]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ = 0 Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ = 0 Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
             UnderscriptBox["0", "Nat"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[Equal]", 
               UnderscriptBox["0", "Nat"]}], "\[And]", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Equal]", 
               UnderscriptBox["0", "Nat"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
             UnderscriptBox["0", "Nat"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[Equal]", 
               UnderscriptBox["0", "Nat"]}], "\[And]", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Equal]", 
               UnderscriptBox[
               "0", "Nat"]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:343580194"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p343580194-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p343580194-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:343580194", "Proof|ID:343580194-1"},
 CellID->734595749],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p343580194-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->407594587]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ comm Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ comm Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["n", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             StyleBox[
             "m", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["n", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             StyleBox[
             "m", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:834229092"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p834229092-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p834229092-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:834229092", "Proof|ID:834229092-1"},
 CellID->12781957],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p834229092-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->354412127]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ assoc Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ assoc Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             UnderscriptBox["+", "Nat"], 
             StyleBox["k", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox[
               "k", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             UnderscriptBox["+", "Nat"], 
             StyleBox["k", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox[
               "k", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:73705609"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p73705609-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p73705609-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:73705609", "Proof|ID:73705609-1"},
 CellID->337185825],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p73705609-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->62558116]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ comm-assoc Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ comm-assoc Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             StyleBox["n", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox[
               "k", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             StyleBox["n", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox[
               "k", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:202544617"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p202544617-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p202544617-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:202544617", "Proof|ID:202544617-1"},
 CellID->299537],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p202544617-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->173855211]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ + + Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ + + Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["i", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             UnderscriptBox["+", "Nat"], 
             RowBox[{
               StyleBox["k", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["i", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}], 
             UnderscriptBox["+", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox[
               "i", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["i", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             UnderscriptBox["+", "Nat"], 
             RowBox[{
               StyleBox["k", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["i", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}], 
             UnderscriptBox["+", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox[
               "i", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:555281086"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p555281086-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p555281086-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:555281086", "Proof|ID:555281086-1"},
 CellID->94141421],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p555281086-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->37231442]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ rcanc Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ rcanc Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:5229733"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p5229733-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p5229733-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:5229733", "Proof|ID:5229733-1"},
 CellID->540060300],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p5229733-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->235773962]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->325165612],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Properties of ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " in Nat"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6602118877327127`*^9, 3.6602118975919304`*^9}},
 CellID->901457357],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "Nat"], 
   RowBox[{"[", 
    RowBox[{"m", ",", "n"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          301992507}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6602119038787413`*^9, 3.660211913769159*^9}},
 CellTags->{
  "ID:301992507", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->301992507],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["*", "Nat"], 
   UnderscriptBox["0", "Nat"]}], "\[Equal]", 
  UnderscriptBox["0", "Nat"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* 0 Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:99163941", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6602119255003796`*^9, 3.6602119356413975`*^9}, {
  3.660212040317581*^9, 3.6602120447323885`*^9}},
 CellTags->{
  "ID:99163941", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* 0 Nat", "IR"},
 CellID->99163941],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["*", "Nat"], 
   RowBox[{"succ", "[", "n", "]"}]}], "\[Equal]", 
  RowBox[{
   RowBox[{"m", 
    UnderscriptBox["*", "Nat"], "n"}], 
   UnderscriptBox["+", "Nat"], "m"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* succ Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:371484270", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6602119408050065`*^9, 3.6602119614126425`*^9}, {
  3.6605416480315657`*^9, 3.6605416486555667`*^9}},
 CellTags->{
  "ID:371484270", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* succ Nat", "IR"},
 CellID->371484270],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "Nat"], 
  RowBox[{"[", 
   RowBox[{"m", 
    UnderscriptBox["*", "Nat"], "n"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* closed Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:142418160", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6602119979635067`*^9, 3.6602120078227243`*^9}},
 CellTags->{
  "ID:142418160", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* closed Nat", "IR"},
 CellID->142418160],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["*", "Nat"], 
   UnderscriptBox["1", "Nat"]}], "\[Equal]", "m"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1 rneutral Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:21383846", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6602171325758142`*^9, 3.6602171429966326`*^9}},
 CellTags->{
  "ID:21383846", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 rneutral Nat", "IR"},
 CellID->21383846],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["0", "Nat"], 
   UnderscriptBox["*", "Nat"], "n"}], "\[Equal]", 
  UnderscriptBox["0", "Nat"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(0 lannihilator Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:28953873", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660212031035565*^9, 3.6602120560892086`*^9}},
 CellTags->{
  "ID:28953873", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 lannihilator Nat", "IR"},
 CellID->28953873],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"succ", "[", "m", "]"}], 
   UnderscriptBox["*", "Nat"], "n"}], "\[Equal]", 
  RowBox[{
   RowBox[{"m", 
    UnderscriptBox["*", "Nat"], "n"}], 
   UnderscriptBox["+", "Nat"], "n"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(succ * Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:218775165", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6602120672276287`*^9, 3.660212084512459*^9}, {
  3.660541659544386*^9, 3.660541660090387*^9}},
 CellTags->{
  "ID:218775165", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ * Nat", "IR"},
 CellID->218775165],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"m", 
     UnderscriptBox["*", "Nat"], "n"}], "\[Equal]", 
    UnderscriptBox["0", "Nat"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"m", "\[Equal]", 
     UnderscriptBox["0", "Nat"]}], "\[Or]", 
    RowBox[{"n", "\[Equal]", 
     UnderscriptBox["0", "Nat"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* zdf Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:470439981", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660212101953289*^9, 3.6602121244641285`*^9}},
 CellTags->{
  "ID:470439981", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* zdf Nat"},
 CellID->470439981],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["*", "Nat"], "n"}], "\[Equal]", 
  RowBox[{"n", 
   UnderscriptBox["*", "Nat"], "m"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* comm Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:218136211", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660212143714563*^9, 3.6602121578169875`*^9}},
 CellTags->{
  "ID:218136211", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* comm Nat"},
 CellID->218136211],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "Nat"], 
   RowBox[{"[", "k", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          666029340}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660212289529019*^9, 3.660212296939032*^9}},
 CellTags->{
  "ID:666029340", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->666029340],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["*", "Nat"], 
   RowBox[{"(", 
    RowBox[{"n", 
     UnderscriptBox["+", "Nat"], "k"}], ")"}]}], "\[Equal]", 
  RowBox[{
   RowBox[{"m", 
    UnderscriptBox["*", "Nat"], "n"}], 
   UnderscriptBox["+", "Nat"], 
   RowBox[{"m", 
    UnderscriptBox["*", "Nat"], "k"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* distr + Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:43150914", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6602123189194703`*^9, 3.660212363660349*^9}, {
  3.6605452112806273`*^9, 3.6605452215142455`*^9}},
 CellTags->{
  "ID:43150914", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* distr + Nat"},
 CellID->43150914],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"m", 
     UnderscriptBox["*", "Nat"], "n"}], ")"}], 
   UnderscriptBox["*", "Nat"], "k"}], "\[Equal]", 
  RowBox[{"m", 
   UnderscriptBox["*", "Nat"], 
   RowBox[{"(", 
    RowBox[{"n", 
     UnderscriptBox["*", "Nat"], "k"}], ")"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* assoc Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:475421150", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660212305409847*^9, 3.6602123474051204`*^9}},
 CellTags->{
  "ID:475421150", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* assoc Nat"},
 CellID->475421150],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["*", "Nat"], 
   RowBox[{"(", 
    RowBox[{"n", 
     UnderscriptBox["*", "Nat"], "k"}], ")"}]}], "\[Equal]", 
  RowBox[{"n", 
   UnderscriptBox["*", "Nat"], 
   RowBox[{"(", 
    RowBox[{"m", 
     UnderscriptBox["*", "Nat"], "k"}], ")"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* comm-assoc Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:98508460", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661171094614033*^9, 3.6611711162512712`*^9}},
 CellTags->{
  "ID:98508460", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* comm-assoc Nat"},
 CellID->98508460],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"m", 
     UnderscriptBox["*", "Nat"], "k"}], "\[Equal]", 
    RowBox[{"n", 
     UnderscriptBox["*", "Nat"], "k"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"m", "\[Equal]", "n"}], "\[Or]", 
    RowBox[{"k", "\[Equal]", 
     UnderscriptBox["0", "Nat"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* rcanc Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:138249572", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6602123736599665`*^9, 3.6602123962020063`*^9}},
 CellTags->{
  "ID:138249572", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* rcanc Nat"},
 CellID->138249572],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->7112206]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->239279945],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* 0 Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* 0 Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             UnderscriptBox["0", "Nat"]}], "\[Equal]", 
           UnderscriptBox["0", "Nat"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             UnderscriptBox["0", "Nat"]}], "\[Equal]", 
           UnderscriptBox["0", "Nat"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:99163941"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p99163941-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p99163941-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:99163941", "Proof|ID:99163941-1"},
 CellID->138225265],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p99163941-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->392553738]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* succ Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* succ Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             RowBox[{"succ", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             UnderscriptBox["+", "Nat"], 
             StyleBox[
             "m", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             RowBox[{"succ", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             UnderscriptBox["+", "Nat"], 
             StyleBox[
             "m", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:371484270"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p371484270-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p371484270-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:371484270", "Proof|ID:371484270-1"},
 CellID->2648213],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p371484270-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->637195123]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* closed Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* closed Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Nat"], "[", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Nat"], "[", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:142418160"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p142418160-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p142418160-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:142418160", "Proof|ID:142418160-1"},
 CellID->32519450],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p142418160-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->748270549]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(1 rneutral Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(1 rneutral Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             UnderscriptBox["1", "Nat"]}], "\[Equal]", 
           StyleBox[
           "m", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             UnderscriptBox["1", "Nat"]}], "\[Equal]", 
           StyleBox[
           "m", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:21383846"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p21383846-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p21383846-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:21383846", "Proof|ID:21383846-1"},
 CellID->16031556],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p21383846-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->456731267]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(0 lannihilator Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(0 lannihilator Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["0", "Nat"], 
             UnderscriptBox["*", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
           UnderscriptBox["0", "Nat"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["0", "Nat"], 
             UnderscriptBox["*", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
           UnderscriptBox["0", "Nat"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:28953873"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p28953873-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p28953873-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:28953873", "Proof|ID:28953873-1"},
 CellID->86840982],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p28953873-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->14700220]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(succ * Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(succ * Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"succ", "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], 
             UnderscriptBox["*", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             UnderscriptBox["+", "Nat"], 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"succ", "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], 
             UnderscriptBox["*", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             UnderscriptBox["+", "Nat"], 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:218775165"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p218775165-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p218775165-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:218775165", "Proof|ID:218775165-1"},
 CellID->82184287],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p218775165-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->141785307]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* zdf Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* zdf Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
             UnderscriptBox["0", "Nat"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[Equal]", 
               UnderscriptBox["0", "Nat"]}], "\[Or]", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Equal]", 
               UnderscriptBox["0", "Nat"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
             UnderscriptBox["0", "Nat"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[Equal]", 
               UnderscriptBox["0", "Nat"]}], "\[Or]", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Equal]", 
               UnderscriptBox[
               "0", "Nat"]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:470439981"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p470439981-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p470439981-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:470439981", "Proof|ID:470439981-1"},
 CellID->167271436],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p470439981-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->20808447]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* comm Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* comm Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["n", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             StyleBox[
             "m", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["n", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             StyleBox[
             "m", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:218136211"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p218136211-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p218136211-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:218136211", "Proof|ID:218136211-1"},
 CellID->116241402],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p218136211-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->464340012]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* distr + Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* distr + Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             UnderscriptBox["+", "Nat"], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox[
               "k", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             UnderscriptBox["+", "Nat"], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox[
               "k", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:43150914"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p43150914-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p43150914-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:43150914", "Proof|ID:43150914-1"},
 CellID->92944893],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p43150914-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->422365210]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* assoc Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* assoc Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             UnderscriptBox["*", "Nat"], 
             StyleBox["k", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox[
               "k", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             UnderscriptBox["*", "Nat"], 
             StyleBox["k", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox[
               "k", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:475421150"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p475421150-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p475421150-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:475421150", "Proof|ID:475421150-1"},
 CellID->574072770],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p475421150-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->691990920]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* comm-assoc Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* comm-assoc Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             StyleBox["n", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox[
               "k", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             StyleBox["n", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox[
               "k", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:98508460"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p98508460-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p98508460-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:98508460", "Proof|ID:98508460-1"},
 CellID->81244651],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p98508460-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->176140155]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* rcanc Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* rcanc Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[Equal]", 
               StyleBox["n", "ExpressionVariable"]}], "\[Or]", 
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Equal]", 
               UnderscriptBox["0", "Nat"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[Equal]", 
               StyleBox["n", "ExpressionVariable"]}], "\[Or]", 
             RowBox[{
               StyleBox["k", "ExpressionVariable"], "\[Equal]", 
               UnderscriptBox[
               "0", "Nat"]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:138249572"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p138249572-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p138249572-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:138249572", "Proof|ID:138249572-1"},
 CellID->165947681],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p138249572-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->55566549]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Lemmas: \[GreaterEqual] (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.660372135037098*^9, 3.660372139155505*^9}, {
  3.660384755856039*^9, 3.6603847620804496`*^9}, {3.660479701854231*^9, 
  3.660479702712232*^9}, {3.660643488734255*^9, 3.6606434931178627`*^9}},
 CellID->359761349],

Cell["", "OpenEnvironment",
 CellID->54526147],

Cell[CellGroupData[{

Cell["Lemma: Properties of \[GreaterEqual] in Nat", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.660217934260623*^9, 3.6602179402822337`*^9}},
 CellID->1339458],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "Nat"], 
   RowBox[{"[", 
    RowBox[{"m", ",", "n"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          621974620}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660217944291441*^9, 3.660217951108653*^9}},
 CellTags->{
  "ID:621974620", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->621974620],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["0", "Nat"], 
    UnderscriptBox["\[GreaterEqual]", "Nat"], "n"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"n", "\[Equal]", 
    UnderscriptBox["0", "Nat"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(0 \[GreaterEqual] Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:380981542", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660218282172434*^9, 3.660218300252866*^9}},
 CellTags->{
  "ID:380981542", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 \[GreaterEqual] Nat", "IR"},
 CellID->380981542],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"succ", "[", "m", "]"}], 
    UnderscriptBox["\[GreaterEqual]", "Nat"], "n"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"m", 
     UnderscriptBox["\[GreaterEqual]", "Nat"], "n"}], "\[Or]", 
    RowBox[{
     RowBox[{"succ", "[", "m", "]"}], "\[Equal]", "n"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(succ \[GreaterEqual] Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:42977349", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6602183238245077`*^9, 3.660218381091008*^9}, {
  3.6605428076072025`*^9, 3.6605428108676085`*^9}},
 CellTags->{
  "ID:42977349", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ \[GreaterEqual] Nat", "IR"},
 CellID->42977349],

Cell[BoxData[
 RowBox[{"m", 
  UnderscriptBox["\[GreaterEqual]", "Nat"], "m"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] reflexive Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:428740350", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6602191811892138`*^9, 3.660219182296816*^9}, {
  3.660219611251769*^9, 3.6602196121409707`*^9}},
 CellTags->{
  "ID:428740350", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] reflexive Nat"},
 CellID->428740350],

Cell[BoxData[
 RowBox[{"m", 
  UnderscriptBox["\[GreaterEqual]", "Nat"], 
  UnderscriptBox["0", "Nat"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] 0 Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:297686800", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6602179704058867`*^9, 3.6602179800623035`*^9}, {
  3.660218102272918*^9, 3.6602181031309195`*^9}},
 CellTags->{
  "ID:297686800", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] 0 Nat", "IR"},
 CellID->297686800],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"m", 
    UnderscriptBox["\[GreaterEqual]", "Nat"], 
    RowBox[{"succ", "[", "n", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"m", 
     UnderscriptBox["\[GreaterEqual]", "Nat"], "n"}], "\[And]", 
    RowBox[{"m", "\[NotEqual]", "n"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] succ Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:585961470", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660218001902342*^9, 3.6602180219327774`*^9}, {
  3.660218105548924*^9, 3.6602181099949317`*^9}, {3.6605427555655117`*^9, 
  3.6605427562051125`*^9}},
 CellTags->{
  "ID:585961470", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] succ Nat", "IR"},
 CellID->585961470],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"m", 
     UnderscriptBox["\[GreaterEqual]", "Nat"], "n"}], "\[And]", 
    RowBox[{"n", 
     UnderscriptBox["\[GreaterEqual]", "Nat"], "m"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"m", "\[Equal]", "n"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] antisymmetric Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:894887193", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6602227012940016`*^9, 3.6602227148972254`*^9}},
 CellTags->{
  "ID:894887193", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] antisymmetric Nat"},
 CellID->894887193],

Cell[BoxData[
 RowBox[{
  RowBox[{"succ", "[", "m", "]"}], 
  UnderscriptBox["\[GreaterEqual]", "Nat"], "m"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(succ increasing Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:208533627", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6602197362869887`*^9, 3.660219743135401*^9}, {
  3.660542796281583*^9, 3.6605427967807837`*^9}},
 CellTags->{
  "ID:208533627", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ increasing Nat"},
 CellID->208533627],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"succ", "[", "m", "]"}], 
    UnderscriptBox["\[GreaterEqual]", "Nat"], 
    RowBox[{"succ", "[", "n", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"m", 
    UnderscriptBox["\[GreaterEqual]", "Nat"], "n"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(succ \[GreaterEqual] succ Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:112605784", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.66022063063096*^9, 3.6602206532509995`*^9}, {
  3.6605427829435596`*^9, 3.660542785486364*^9}},
 CellTags->{
  "ID:112605784", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ \[GreaterEqual] succ Nat", "IR"},
 CellID->112605784],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["\[GreaterEqual]", "Nat"], "n"}], "\[Or]", 
  RowBox[{"n", 
   UnderscriptBox["\[GreaterEqual]", "Nat"], 
   "m"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] total Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:304315076", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6602228755307074`*^9, 3.660222886778327*^9}},
 CellTags->{
  "ID:304315076", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] total Nat"},
 CellID->304315076],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["\[GreaterEqual]", "Nat"], "n"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "Nat"], 
      RowBox[{"[", "k", "]"}]}]], 
    RowBox[{"m", "\[Equal]", 
     RowBox[{"n", 
      UnderscriptBox["+", "Nat"], "k"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] \[Exists] Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:446896782", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660385453746865*^9, 3.6603854884101257`*^9}},
 CellTags->{
  "ID:446896782", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] \[Exists] Nat"},
 CellID->446896782],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["1", "Nat"], 
  UnderscriptBox["\[GreaterEqual]", "Nat"], 
  UnderscriptBox["0", "Nat"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1 \[GreaterEqual] 0 Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:196203313", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612358426372347`*^9, 3.661235863261471*^9}},
 CellTags->{
  "ID:196203313", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 \[GreaterEqual] 0 Nat"},
 CellID->196203313],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"m", 
    UnderscriptBox["\[GreaterEqual]", "Nat"], 
    UnderscriptBox["1", "Nat"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"m", "\[NotEqual]", 
    UnderscriptBox["0", "Nat"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] 1 Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:30448165", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660643144762251*^9, 3.6606431608926787`*^9}, {
  3.6606432148687735`*^9, 3.660643217754779*^9}},
 CellTags->{
  "ID:30448165", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] 1 Nat", "IR"},
 CellID->30448165],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "Nat"], 
   RowBox[{"[", "k", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          222994735}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660222750527688*^9, 3.660222757017299*^9}},
 CellTags->{
  "ID:222994735", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->222994735],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"m", 
     UnderscriptBox["\[GreaterEqual]", "Nat"], "n"}], "\[And]", 
    RowBox[{"n", 
     UnderscriptBox["\[GreaterEqual]", "Nat"], "k"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"m", 
   UnderscriptBox["\[GreaterEqual]", "Nat"], 
   "k"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] transitive Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:546802157", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.66022276297651*^9, 3.660222778280137*^9}},
 CellTags->{
  "ID:546802157", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] transitive Nat"},
 CellID->546802157],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["\[GreaterEqual]", "Nat"], "n"}], "\[Implies]", 
  RowBox[{
   RowBox[{"m", 
    UnderscriptBox["+", "Nat"], "k"}], 
   UnderscriptBox["\[GreaterEqual]", "Nat"], 
   RowBox[{"n", 
    UnderscriptBox["+", "Nat"], "k"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] + \[DoubleLeftArrow] Nat)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:352201475", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660223911844528*^9, 3.6602239332321653`*^9}},
 CellTags->{
  "ID:352201475", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] + \[DoubleLeftArrow] Nat"},
 CellID->352201475],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"m", 
     UnderscriptBox["+", "Nat"], "k"}], 
    UnderscriptBox["\[GreaterEqual]", "Nat"], 
    RowBox[{"n", 
     UnderscriptBox["+", "Nat"], "k"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"m", 
    UnderscriptBox["\[GreaterEqual]", "Nat"], "n"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] + Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:97128652", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612484955747213`*^9, 3.6612485314079843`*^9}},
 CellTags->{
  "ID:97128652", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] + Nat"},
 CellID->97128652],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["+", "Nat"], "n"}], 
  UnderscriptBox["\[GreaterEqual]", "Nat"], "m"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ \[GreaterEqual] Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:278931212", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6603848351197777`*^9, 3.6603848463517976`*^9}},
 CellTags->{
  "ID:278931212", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ \[GreaterEqual] Nat"},
 CellID->278931212],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["\[GreaterEqual]", "Nat"], "n"}], "\[Implies]", 
  RowBox[{
   RowBox[{"m", 
    UnderscriptBox["*", "Nat"], "k"}], 
   UnderscriptBox["\[GreaterEqual]", "Nat"], 
   RowBox[{"n", 
    UnderscriptBox["*", "Nat"], "k"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] * Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:335211897", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6605463089523563`*^9, 3.6605463320247965`*^9}},
 CellTags->{
  "ID:335211897", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] * Nat"},
 CellID->335211897],

Cell[BoxData[
 RowBox[{
  RowBox[{"k", "\[NotEqual]", 
   UnderscriptBox["0", "Nat"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"m", 
    UnderscriptBox["*", "Nat"], "k"}], 
   UnderscriptBox["\[GreaterEqual]", "Nat"], 
   "m"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* \[GreaterEqual] Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:32863246", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660643079382536*^9, 3.6606430973849673`*^9}},
 CellTags->{
  "ID:32863246", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* \[GreaterEqual] Nat"},
 CellID->32863246],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->251470773]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->123859362],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(0 \[GreaterEqual] Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(0 \[GreaterEqual] Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["0", "Nat"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["n", "ExpressionVariable"], "\[Equal]", 
             UnderscriptBox["0", "Nat"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["0", "Nat"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["n", "ExpressionVariable"], "\[Equal]", 
             UnderscriptBox["0", "Nat"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:380981542"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p380981542-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p380981542-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:380981542", "Proof|ID:380981542-1"},
 CellID->159585745],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p380981542-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->36940991]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(succ \[GreaterEqual] Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(succ \[GreaterEqual] Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"succ", "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["\[GreaterEqual]", "Nat"], 
                 StyleBox["n", "ExpressionVariable"]}], "\[Or]", 
               RowBox[{
                 RowBox[{"succ", "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}], "\[Equal]", 
                 StyleBox["n", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"succ", "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["\[GreaterEqual]", "Nat"], 
                 StyleBox["n", "ExpressionVariable"]}], "\[Or]", 
               RowBox[{
                 RowBox[{"succ", "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}], "\[Equal]", 
                 StyleBox["n", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:42977349"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p42977349-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p42977349-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:42977349", "Proof|ID:42977349-1"},
 CellID->255247870],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p42977349-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->60792811]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] reflexive Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] reflexive Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["m", "ExpressionVariable"], 
           UnderscriptBox["\[GreaterEqual]", "Nat"], 
           StyleBox[
           "m", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["m", "ExpressionVariable"], 
           UnderscriptBox["\[GreaterEqual]", "Nat"], 
           StyleBox[
           "m", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:428740350"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p428740350-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p428740350-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:428740350", "Proof|ID:428740350-1"},
 CellID->34057048],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p428740350-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->812129610]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] 0 Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] 0 Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["m", "ExpressionVariable"], 
           UnderscriptBox["\[GreaterEqual]", "Nat"], 
           UnderscriptBox["0", "Nat"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["m", "ExpressionVariable"], 
           UnderscriptBox["\[GreaterEqual]", "Nat"], 
           UnderscriptBox["0", "Nat"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:297686800"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p297686800-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p297686800-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:297686800", "Proof|ID:297686800-1"},
 CellID->45781616],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p297686800-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->541714067]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] succ Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] succ Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             RowBox[{"succ", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["\[GreaterEqual]", "Nat"], 
                 StyleBox["n", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
                 StyleBox["n", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             RowBox[{"succ", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["\[GreaterEqual]", "Nat"], 
                 StyleBox["n", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
                 StyleBox["n", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:585961470"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p585961470-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p585961470-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:585961470", "Proof|ID:585961470-1"},
 CellID->252534163],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p585961470-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->60797310]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] antisymmetric Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] antisymmetric Nat)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Nat"], 
               StyleBox["m", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Nat"], 
               StyleBox["m", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:894887193"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p894887193-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p894887193-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:894887193", "Proof|ID:894887193-1"},
 CellID->246933219],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p894887193-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->141573684]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(succ increasing Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(succ increasing Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"succ", "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], 
           UnderscriptBox["\[GreaterEqual]", "Nat"], 
           StyleBox[
           "m", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"succ", "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], 
           UnderscriptBox["\[GreaterEqual]", "Nat"], 
           StyleBox[
           "m", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:208533627"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p208533627-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p208533627-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:208533627", "Proof|ID:208533627-1"},
 CellID->10870655],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p208533627-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->71717834]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(succ \[GreaterEqual] succ Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(succ \[GreaterEqual] succ Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"succ", "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             RowBox[{"succ", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"succ", "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             RowBox[{"succ", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:112605784"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p112605784-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p112605784-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:112605784", "Proof|ID:112605784-1"},
 CellID->362360428],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p112605784-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->292064837]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] total Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] total Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Or]", 
           RowBox[{
             StyleBox["n", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox[
             "m", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Or]", 
           RowBox[{
             StyleBox["n", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox[
             "m", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:304315076"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p304315076-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p304315076-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:304315076", "Proof|ID:304315076-1"},
 CellID->318982970],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p304315076-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->873958818]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] \[Exists] Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] \[Exists] Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "Nat"], "[", 
                   StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[Equal]", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], 
                 UnderscriptBox["+", "Nat"], 
                 StyleBox[
                 "k", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "Nat"], "[", 
                   StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[Equal]", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], 
                 UnderscriptBox["+", "Nat"], 
                 StyleBox[
                 "k", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:446896782"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p446896782-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p446896782-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:446896782", "Proof|ID:446896782-1"},
 CellID->239855888],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p446896782-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->356600252]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(1 \[GreaterEqual] 0 Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(1 \[GreaterEqual] 0 Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["1", "Nat"], 
         UnderscriptBox["\[GreaterEqual]", "Nat"], 
         UnderscriptBox["0", "Nat"]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["1", "Nat"], 
         UnderscriptBox["\[GreaterEqual]", "Nat"], 
         UnderscriptBox["0", "Nat"]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:196203313"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p196203313-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p196203313-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:196203313", "Proof|ID:196203313-1"},
 CellID->28380975],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p196203313-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->13411257]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] 1 Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] 1 Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             UnderscriptBox["1", "Nat"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "Nat"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             UnderscriptBox["1", "Nat"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "Nat"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:30448165"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p30448165-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p30448165-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:30448165", "Proof|ID:30448165-1"},
 CellID->64482358],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p30448165-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->202694291]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] transitive Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] transitive Nat)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox[
             "k", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox[
             "k", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:546802157"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p546802157-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p546802157-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:546802157", "Proof|ID:546802157-1"},
 CellID->527891345],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p546802157-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->823045757]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] + \[DoubleLeftArrow] Nat)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] + \[DoubleLeftArrow] Nat)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox[
               "k", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox[
               "k", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:352201475"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p352201475-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p352201475-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:352201475", "Proof|ID:352201475-1"},
 CellID->76510990],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p352201475-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->109258721]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] + Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] + Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:97128652"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p97128652-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p97128652-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:97128652", "Proof|ID:97128652-1"},
 CellID->393536374],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p97128652-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->644252183]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ \[GreaterEqual] Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ \[GreaterEqual] Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], 
           UnderscriptBox["\[GreaterEqual]", "Nat"], 
           StyleBox[
           "m", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["+", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], 
           UnderscriptBox["\[GreaterEqual]", "Nat"], 
           StyleBox[
           "m", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:278931212"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p278931212-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p278931212-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:278931212", "Proof|ID:278931212-1"},
 CellID->260422124],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p278931212-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->478276088]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] * Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] * Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox[
               "k", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox[
               "k", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:335211897"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p335211897-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p335211897-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:335211897", "Proof|ID:335211897-1"},
 CellID->171667115],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p335211897-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->548791013]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* \[GreaterEqual] Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* \[GreaterEqual] Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["k", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "Nat"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox[
             "m", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["k", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "Nat"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox[
             "m", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:32863246"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p32863246-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p32863246-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:32863246", "Proof|ID:32863246-1"},
 CellID->145909442],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p32863246-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->204709576]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemmas: ",
 Cell[BoxData[
  FormBox["-", TraditionalForm]]],
 " (\[Checkmark])"
}], "Subsubsection",
 CellChangeTimes->{{3.660384785776891*^9, 3.660384789520898*^9}, {
  3.6603862184222097`*^9, 3.6603862225250173`*^9}, {3.660479704053835*^9, 
  3.660479704350235*^9}, {3.660546122298028*^9, 3.6605461264164352`*^9}},
 CellID->505759711],

Cell["", "OpenEnvironment",
 CellID->56489176],

Cell[CellGroupData[{

Cell["Lemma: Properties of - in Nat", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.660372373456317*^9, 3.6603723782767253`*^9}},
 CellID->227256827],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "Nat"], 
   RowBox[{"[", 
    RowBox[{"m", ",", "n"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          609763659}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6603723830503335`*^9, 3.660372390023546*^9}},
 CellTags->{
  "ID:609763659", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->609763659],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["-", "Nat"], 
   UnderscriptBox["0", "Nat"]}], "\[Equal]", "m"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(- 0 Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:19208428", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6603723940639524`*^9, 3.660372404578371*^9}},
 CellTags->{
  "ID:19208428", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- 0 Nat", "IR"},
 CellID->19208428],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["-", "Nat"], 
   RowBox[{"succ", "[", "n", "]"}]}], "\[Equal]", 
  RowBox[{
   UnderscriptBox["Pred", "Nat"], "[", 
   RowBox[{"m", 
    UnderscriptBox["-", "Nat"], "n"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(- succ Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:593421999", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660372409913581*^9, 3.660372427136011*^9}, {
  3.660542920349601*^9, 3.660542920864402*^9}},
 CellTags->{
  "ID:593421999", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- succ Nat", "IR"},
 CellID->593421999],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "Nat"], 
  RowBox[{"[", 
   RowBox[{"m", 
    UnderscriptBox["-", "Nat"], "n"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(- closed Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:153357458", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660373327511594*^9, 3.660373339710815*^9}},
 CellTags->{
  "ID:153357458", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- closed Nat", "IR"},
 CellID->153357458],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["-", "Nat"], 
   UnderscriptBox["1", "Nat"]}], "\[Equal]", 
  RowBox[{
   UnderscriptBox["Pred", "Nat"], "[", "m", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(- 1 Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:616125500", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660373582731042*^9, 3.6603735986742697`*^9}},
 CellTags->{
  "ID:616125500", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- 1 Nat"},
 CellID->616125500],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["0", "Nat"], 
   UnderscriptBox["-", "Nat"], "n"}], "\[Equal]", 
  UnderscriptBox["0", "Nat"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(0 - Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:491337343", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660373367868864*^9, 3.6603733799120855`*^9}},
 CellTags->{
  "ID:491337343", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 - Nat", "IR"},
 CellID->491337343],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"succ", "[", "m", "]"}], 
   UnderscriptBox["-", "Nat"], 
   RowBox[{"succ", "[", "n", "]"}]}], "\[Equal]", 
  RowBox[{"m", 
   UnderscriptBox["-", "Nat"], "n"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(succ - succ Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:621700658", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.660373386230097*^9, 3.6603734366025853`*^9}, {
  3.6603734830750666`*^9, 3.6603735074423094`*^9}, {3.6605429315348206`*^9, 
  3.6605429341868253`*^9}},
 CellTags->{
  "ID:621700658", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ - succ Nat", "IR"},
 CellID->621700658],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"m", 
     UnderscriptBox["+", "Nat"], "n"}], ")"}], 
   UnderscriptBox["-", "Nat"], "n"}], "\[Equal]", 
  "m"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ - Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:236021775", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6603851734143724`*^9, 3.6603851870955963`*^9}},
 CellTags->{
  "ID:236021775", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ - Nat"},
 CellID->236021775],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["\[GreaterEqual]", "Nat"], "n"}], "\[Implies]", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{"m", 
      UnderscriptBox["-", "Nat"], "n"}], ")"}], 
    UnderscriptBox["+", "Nat"], "n"}], "\[Equal]", 
   "m"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(- + Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:3006634", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6603850793306074`*^9, 3.6603851061626544`*^9}},
 CellTags->{
  "ID:3006634", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- + Nat"},
 CellID->3006634],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "Nat"], 
   RowBox[{"[", "k", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          80465146}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.661245641663306*^9},
 CellTags->{
  "ID:80465146", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->80465146],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"m", 
     UnderscriptBox["+", "Nat"], "k"}], ")"}], 
   UnderscriptBox["-", "Nat"], 
   RowBox[{"(", 
    RowBox[{"n", 
     UnderscriptBox["+", "Nat"], "k"}], ")"}]}], "\[Equal]", 
  RowBox[{"m", 
   UnderscriptBox["-", "Nat"], "n"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ - + Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:404996142", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661245622631273*^9, 3.6612456684641533`*^9}},
 CellTags->{
  "ID:404996142", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ - + Nat"},
 CellID->404996142],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["*", "Nat"], 
   RowBox[{"(", 
    RowBox[{"n", 
     UnderscriptBox["-", "Nat"], "k"}], ")"}]}], "\[Equal]", 
  RowBox[{
   RowBox[{"m", 
    UnderscriptBox["*", "Nat"], "n"}], 
   UnderscriptBox["-", "Nat"], 
   RowBox[{"m", 
    UnderscriptBox["*", "Nat"], "k"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* distr - Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:225131303", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612450361170425`*^9, 3.661245042918655*^9}, {
   3.6612451019335585`*^9, 3.661245127018402*^9}, 3.6612456318196893`*^9},
 CellTags->{
  "ID:225131303", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* distr - Nat"},
 CellID->225131303],

Cell[BoxData[
 RowBox[{
  RowBox[{"m", 
   UnderscriptBox["\[GreaterEqual]", "Nat"], "n"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"m", 
       UnderscriptBox["-", "Nat"], "n"}], ")"}], 
     UnderscriptBox["+", "Nat"], "k"}], "\[Equal]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"m", 
       UnderscriptBox["+", "Nat"], "k"}], ")"}], 
     UnderscriptBox["-", "Nat"], "n"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["((-)+ Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:142079924", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612466955909586`*^9, 3.6612467353866286`*^9}},
 CellTags->{
  "ID:142079924", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "(-)+ Nat"},
 CellID->142079924],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "Nat"], 
    RowBox[{"[", "i", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"m", 
       UnderscriptBox["\[GreaterEqual]", "Nat"], "i"}], "\[And]", 
      RowBox[{"k", 
       UnderscriptBox["\[GreaterEqual]", "Nat"], "n"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"m", 
         UnderscriptBox["-", "Nat"], "i"}], "\[Equal]", 
        RowBox[{"k", 
         UnderscriptBox["-", "Nat"], "n"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{"m", 
         UnderscriptBox["+", "Nat"], "n"}], "\[Equal]", 
        RowBox[{"k", 
         UnderscriptBox["+", "Nat"], "i"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(- = - Nat)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:367201929", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661318640861826*^9, 3.6613186962731237`*^9}, {
  3.661318816502535*^9, 3.66131883667337*^9}, {3.661319103090638*^9, 
  3.66131912129587*^9}, {3.66131925799891*^9, 3.661319262585318*^9}, {
  3.6613193362330475`*^9, 3.6613193504602723`*^9}},
 CellTags->{
  "ID:367201929", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- = - Nat"},
 CellID->367201929],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->833478354]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->435769634],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(- 0 Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(- 0 Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["-", "Nat"], 
             UnderscriptBox["0", "Nat"]}], "\[Equal]", 
           StyleBox[
           "m", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["-", "Nat"], 
             UnderscriptBox["0", "Nat"]}], "\[Equal]", 
           StyleBox[
           "m", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:19208428"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p19208428-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p19208428-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:19208428", "Proof|ID:19208428-1"},
 CellID->94129044],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p19208428-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->264683537]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(- succ Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(- succ Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["-", "Nat"], 
             RowBox[{"succ", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["Pred", "Nat"], "[", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["-", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["-", "Nat"], 
             RowBox[{"succ", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["Pred", "Nat"], "[", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["-", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:593421999"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p593421999-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p593421999-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:593421999", "Proof|ID:593421999-1"},
 CellID->286638854],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p593421999-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->97602613]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(- closed Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(- closed Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Nat"], "[", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["-", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Nat"], "[", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["-", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:153357458"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p153357458-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p153357458-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:153357458", "Proof|ID:153357458-1"},
 CellID->454920045],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p153357458-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->935874446]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(- 1 Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(- 1 Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["-", "Nat"], 
             UnderscriptBox["1", "Nat"]}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["Pred", "Nat"], "[", 
             StyleBox["m", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["-", "Nat"], 
             UnderscriptBox["1", "Nat"]}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["Pred", "Nat"], "[", 
             StyleBox["m", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:616125500"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p616125500-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p616125500-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:616125500", "Proof|ID:616125500-1"},
 CellID->14344385],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p616125500-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->49030196]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(0 - Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(0 - Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["0", "Nat"], 
             UnderscriptBox["-", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
           UnderscriptBox["0", "Nat"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["0", "Nat"], 
             UnderscriptBox["-", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
           UnderscriptBox["0", "Nat"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:491337343"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p491337343-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p491337343-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:491337343", "Proof|ID:491337343-1"},
 CellID->458038256],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p491337343-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->21026969]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(succ - succ Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(succ - succ Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"succ", "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], 
             UnderscriptBox["-", "Nat"], 
             RowBox[{"succ", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], "\[Equal]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["-", "Nat"], 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"succ", "[", 
               StyleBox["m", "ExpressionVariable"], "]"}], 
             UnderscriptBox["-", "Nat"], 
             RowBox[{"succ", "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}], "\[Equal]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["-", "Nat"], 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:621700658"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p621700658-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p621700658-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:621700658", "Proof|ID:621700658-1"},
 CellID->31914115],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p621700658-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->35212336]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ - Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ - Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             UnderscriptBox["-", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
           StyleBox[
           "m", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             UnderscriptBox["-", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
           StyleBox[
           "m", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:236021775"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p236021775-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p236021775-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:236021775", "Proof|ID:236021775-1"},
 CellID->322826090],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p236021775-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->105703489]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(- + Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(- + Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["-", "Nat"], 
                 StyleBox["n", "ExpressionVariable"]}], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
             StyleBox[
             "m", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["-", "Nat"], 
                 StyleBox["n", "ExpressionVariable"]}], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
             StyleBox[
             "m", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:3006634"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p3006634-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p3006634-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:3006634", "Proof|ID:3006634-1"},
 CellID->539834930],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p3006634-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->55369331]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ - + Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ - + Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}], 
             UnderscriptBox["-", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["-", "Nat"], 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}], 
             UnderscriptBox["-", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["-", "Nat"], 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:404996142"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p404996142-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p404996142-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:404996142", "Proof|ID:404996142-1"},
 CellID->457219728],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p404996142-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->822478605]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* distr - Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* distr - Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["-", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             UnderscriptBox["-", "Nat"], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox[
               "k", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["*", "Nat"], 
             RowBox[{
               StyleBox["n", "ExpressionVariable"], 
               UnderscriptBox["-", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}], 
             UnderscriptBox["-", "Nat"], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["*", "Nat"], 
               StyleBox[
               "k", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:225131303"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p225131303-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p225131303-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:225131303", "Proof|ID:225131303-1"},
 CellID->442897560],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p225131303-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->567742436]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"((-)+ Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"((-)+ Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["-", "Nat"], 
                 StyleBox["n", "ExpressionVariable"]}], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["+", "Nat"], 
                 StyleBox["k", "ExpressionVariable"]}], 
               UnderscriptBox["-", "Nat"], 
               StyleBox[
               "n", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["-", "Nat"], 
                 StyleBox["n", "ExpressionVariable"]}], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["k", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["+", "Nat"], 
                 StyleBox["k", "ExpressionVariable"]}], 
               UnderscriptBox["-", "Nat"], 
               StyleBox[
               "n", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:142079924"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p142079924-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p142079924-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:142079924", "Proof|ID:142079924-1"},
 CellID->230509363],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p142079924-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->21742626]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(- = - Nat)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(- = - Nat)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["i", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Nat"], 
               StyleBox["i", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["k", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["-", "Nat"], 
                 StyleBox["i", "ExpressionVariable"]}], "\[Equal]", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], 
                 UnderscriptBox["-", "Nat"], 
                 StyleBox["n", "ExpressionVariable"]}]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["+", "Nat"], 
                 StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], 
                 UnderscriptBox["+", "Nat"], 
                 StyleBox[
                 "i", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["k", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["i", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["m", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Nat"], 
               StyleBox["i", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["k", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Nat"], 
               StyleBox["n", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["-", "Nat"], 
                 StyleBox["i", "ExpressionVariable"]}], "\[Equal]", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], 
                 UnderscriptBox["-", "Nat"], 
                 StyleBox["n", "ExpressionVariable"]}]}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], 
                 UnderscriptBox["+", "Nat"], 
                 StyleBox["n", "ExpressionVariable"]}], "\[Equal]", 
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], 
                 UnderscriptBox["+", "Nat"], 
                 StyleBox[
                 "i", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:367201929"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p367201929-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p367201929-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:367201929", "Proof|ID:367201929-1"},
 CellID->109183652],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p367201929-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->145493550]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Set-Theory Integers (\[Checkmark])", "Section",
 CellChangeTimes->{{3.6602242099454517`*^9, 3.6602242110998535`*^9}, {
  3.660643635639713*^9, 3.660643636794115*^9}, {3.6613374256181655`*^9, 
  3.6613374322637773`*^9}, {3.6613375119829173`*^9, 3.661337514276121*^9}},
 CellID->193970655],

Cell[CellGroupData[{

Cell["PreInt Functor (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6611652790620174`*^9, 3.6611652851460285`*^9}, {
  3.661251788117905*^9, 3.6612517928135138`*^9}},
 CellID->884540480],

Cell[CellGroupData[{

Cell["Definitions", "Subsubsection",
 CellChangeTimes->{{3.661165294474845*^9, 3.6611652963624477`*^9}},
 CellID->684862369],

Cell["", "OpenEnvironment",
 CellID->475123522],

Cell[CellGroupData[{

Cell["Definition: PreInt Functor", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6611653028520594`*^9, 3.6611653065180655`*^9}},
 CellID->11831326],

Cell[BoxData[
 RowBox[{"PreInt", ":=", 
  UnderscriptBox["\[CapitalDelta]", "N"]}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          513393113}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611653097316713`*^9, 3.661165322523694*^9}},
 CellTags->{
  "ID:513393113", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->513393113],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"a", ",", "b"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          623651727}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661165325534499*^9, 3.6611653284985046`*^9}},
 CellTags->{
  "ID:623651727", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->623651727],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", "N"], 
    RowBox[{"[", "a", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a:=b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", "\[Element]", 
    RowBox[{"natSet", "\[Cross]", "natSet"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PreInt \[Element])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:100384801", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661165333350113*^9, 3.6611653582165565`*^9}},
 CellTags->{
  "ID:100384801", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "PreInt \[Element]"},
 CellID->100384801],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["0", "N"], ":=", 
  RowBox[{"Pair", "[", 
   RowBox[{
    UnderscriptBox["0", "Nat"], ",", 
    UnderscriptBox["0", "Nat"]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PreInt 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:398024454", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661166169261981*^9, 3.661166184940009*^9}},
 CellTags->{
  "ID:398024454", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "PreInt 0"},
 CellID->398024454],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["1", "N"], ":=", 
  RowBox[{"Pair", "[", 
   RowBox[{
    UnderscriptBox["1", "Nat"], ",", 
    UnderscriptBox["0", "Nat"]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PreInt 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:418726841", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611661916168203`*^9, 3.6611662055476446`*^9}},
 CellTags->{
  "ID:418726841", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "PreInt 1"},
 CellID->418726841],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox["+", "N"], "b"}], ":=", 
  RowBox[{"Pair", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"fst", "[", "a", "]"}], 
     UnderscriptBox["+", "Nat"], 
     RowBox[{"fst", "[", "b", "]"}]}], ",", 
    RowBox[{
     RowBox[{"snd", "[", "a", "]"}], 
     UnderscriptBox["+", "Nat"], 
     RowBox[{"snd", "[", "b", "]"}]}]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PreInt +)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:10580372", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661165404299038*^9, 3.6611654481975145`*^9}},
 CellTags->{
  "ID:10580372", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "PreInt +"},
 CellID->10580372],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["-", "N"], "a"}], ":=", 
  RowBox[{"Pair", "[", 
   RowBox[{
    RowBox[{"snd", "[", "a", "]"}], ",", 
    RowBox[{"fst", "[", "a", "]"}]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PreInt - unary)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:106096058", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611653838162017`*^9, 3.661165396982625*^9}},
 CellTags->{
  "ID:106096058", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "PreInt - unary"},
 CellID->106096058],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox["*", "N"], "b"}], ":=", 
  RowBox[{"Pair", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"fst", "[", "a", "]"}], 
      UnderscriptBox["*", "Nat"], 
      RowBox[{"fst", "[", "b", "]"}]}], 
     UnderscriptBox["+", "Nat"], 
     RowBox[{
      RowBox[{"snd", "[", "a", "]"}], 
      UnderscriptBox["*", "Nat"], 
      RowBox[{"snd", "[", "b", "]"}]}]}], ",", 
    RowBox[{
     RowBox[{
      RowBox[{"fst", "[", "a", "]"}], 
      UnderscriptBox["*", "Nat"], 
      RowBox[{"snd", "[", "b", "]"}]}], 
     UnderscriptBox["+", "Nat"], 
     RowBox[{
      RowBox[{"snd", "[", "a", "]"}], 
      UnderscriptBox["*", "Nat"], 
      RowBox[{"fst", "[", "b", "]"}]}]}]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PreInt *)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:12727554", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611654743275604`*^9, 3.661165518444438*^9}},
 CellTags->{
  "ID:12727554", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "PreInt *"},
 CellID->12727554],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"a", 
    UnderscriptBox["\[GreaterEqual]", "N"], "b"}], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a:=b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"fst", "[", "a", "]"}], 
     UnderscriptBox["+", "Nat"], 
     RowBox[{"snd", "[", "b", "]"}]}], 
    UnderscriptBox["\[GreaterEqual]", "Nat"], 
    RowBox[{
     RowBox[{"snd", "[", "a", "]"}], 
     UnderscriptBox["+", "Nat"], 
     RowBox[{"fst", "[", "b", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PreInt \[GreaterEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:329154583", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611655256672506`*^9, 3.6611655649793196`*^9}},
 CellTags->{
  "ID:329154583", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "PreInt \[GreaterEqual]"},
 CellID->329154583],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["equiv", "N"], "[", 
   RowBox[{"a", ",", "b"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a:=b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"fst", "[", "a", "]"}], 
     UnderscriptBox["+", "Nat"], 
     RowBox[{"snd", "[", "b", "]"}]}], "\[Equal]", 
    RowBox[{
     RowBox[{"snd", "[", "a", "]"}], 
     UnderscriptBox["+", "Nat"], 
     RowBox[{"fst", "[", "b", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PreInt equiv)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:468243965", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611655700805287`*^9, 3.6611656021697855`*^9}},
 CellTags->{
  "ID:468243965", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "PreInt equiv"},
 CellID->468243965],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->386145701]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->138864013]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Lemmas: General Properties of PreInt (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.661165686862334*^9, 3.6611657033515625`*^9}, {
  3.6611676580711966`*^9, 3.661167662579604*^9}},
 CellID->100030134],

Cell["", "OpenEnvironment",
 CellID->31288446],

Cell[CellGroupData[{

Cell["Lemma: General Properties of PreInt", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6611657094667735`*^9, 3.661165716611586*^9}},
 CellID->207945970],

Cell[BoxData[
 RowBox[{"isDomain", "[", "PreInt", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PreInt isDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:168445606", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611657188891897`*^9, 3.6611657230699973`*^9}},
 CellTags->{
  "ID:168445606", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "PreInt isDomain", "IR"},
 CellID->168445606],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "x"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", "PreInt"], 
      RowBox[{"[", "x", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "Nat"], 
        RowBox[{"[", 
         RowBox[{"m", ",", "n"}], "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"x", "\[Equal]", 
        RowBox[{"Pair", "[", 
         RowBox[{"m", ",", "n"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] PreInt alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:977301511", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661165737936824*^9, 3.6611657934729214`*^9}},
 CellTags->{
  "ID:977301511", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[Element] PreInt alternative", "IR"},
 CellID->977301511],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"m", ",", "n"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", "PreInt"], 
      RowBox[{"[", 
       RowBox[{"Pair", "[", 
        RowBox[{"m", ",", "n"}], "]"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[Element]", "Nat"], 
       RowBox[{"[", "m", "]"}]}], "\[And]", 
      RowBox[{
       UnderscriptBox["\[Element]", "Nat"], 
       RowBox[{"[", "n", "]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Pair \[Element] PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:353341722", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661165799900132*^9, 3.661165829243784*^9}},
 CellTags->{
  "ID:353341722", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Pair \[Element] PreInt", "IR"},
 CellID->353341722],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "PreInt"], 
  RowBox[{"[", 
   UnderscriptBox["0", "PreInt"], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(0 \[Element] PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:34141802", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661166230226888*^9, 3.661166243003311*^9}},
 CellTags->{
  "ID:34141802", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 \[Element] PreInt", "IR"},
 CellID->34141802],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "PreInt"], 
  RowBox[{"[", 
   UnderscriptBox["1", "PreInt"], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1 \[Element] PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:553448562", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611662489781213`*^9, 3.6611662504133234`*^9}},
 CellTags->{
  "ID:553448562", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 \[Element] PreInt", "IR"},
 CellID->553448562],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->18036666]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->54156207],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PreInt isDomain)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PreInt isDomain)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isDomain", "[", "PreInt", "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isDomain", "[", "PreInt", "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:168445606"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p168445606-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p168445606-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:168445606", "Proof|ID:168445606-1"},
 CellID->142038033],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p168445606-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->283190144]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] PreInt alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] PreInt alternative)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["x", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "Nat"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "Nat"], "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["n", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["x", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "Nat"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "Nat"], "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["n", "ExpressionVariable"]}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:977301511"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p977301511-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p977301511-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:977301511", "Proof|ID:977301511-1"},
 CellID->32364991],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p977301511-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->351963332]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Pair \[Element] PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Pair \[Element] PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["m", "ExpressionVariable"], ",", 
            StyleBox["n", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["n", "ExpressionVariable"]}], "]"}], "]"}], ")"}],
            "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", "Nat"], "[", 
                 StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", "Nat"], "[", 
                 StyleBox["n", "ExpressionVariable"], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["m", "ExpressionVariable"], ",", 
            StyleBox["n", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["n", "ExpressionVariable"]}], "]"}], "]"}], ")"}],
            "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", "Nat"], "[", 
                 StyleBox["m", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", "Nat"], "[", 
                 StyleBox["n", "ExpressionVariable"], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:353341722"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p353341722-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p353341722-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:353341722", "Proof|ID:353341722-1"},
 CellID->352067017],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p353341722-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->147706619]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(0 \[Element] PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(0 \[Element] PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", "PreInt"], "[", 
         UnderscriptBox["0", "PreInt"], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", "PreInt"], "[", 
         UnderscriptBox["0", "PreInt"], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:34141802"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p34141802-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p34141802-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:34141802", "Proof|ID:34141802-1"},
 CellID->53936406],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p34141802-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->312339613]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(1 \[Element] PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(1 \[Element] PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", "PreInt"], "[", 
         UnderscriptBox["1", "PreInt"], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", "PreInt"], "[", 
         UnderscriptBox["1", "PreInt"], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:553448562"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p553448562-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p553448562-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:553448562", "Proof|ID:553448562-1"},
 CellID->265994209],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p553448562-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->510886051]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Lemmas: equiv (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6612316669674993`*^9, 3.661231669416703*^9}, {
  3.661250068932085*^9, 3.6612500735184927`*^9}},
 CellID->428038701],

Cell["", "OpenEnvironment",
 CellID->234603150],

Cell[CellGroupData[{

Cell["Lemma: Properties of equiv in PreInt", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6612316758907146`*^9, 3.6612316832071276`*^9}},
 CellID->734012110],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"a1", ",", "a2", ",", "b1", ",", "b2"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["equiv", "PreInt"], "[", 
     RowBox[{
      RowBox[{"Pair", "[", 
       RowBox[{"a1", ",", "a2"}], "]"}], ",", 
      RowBox[{"Pair", "[", 
       RowBox[{"b1", ",", "b2"}], "]"}]}], "]"}], 
    "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"a1", 
       UnderscriptBox["+", "Nat"], "b2"}], "\[Equal]", 
      RowBox[{"a2", 
       UnderscriptBox["+", "Nat"], "b1"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(equiv Pair PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:220417861", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612316942373466`*^9, 3.661231737823824*^9}, {
  3.661232407847*^9, 3.6612324110918064`*^9}},
 CellTags->{
  "ID:220417861", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "equiv Pair PreInt", "IR"},
 CellID->220417861],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "Nat"], 
    RowBox[{"[", 
     RowBox[{"m", ",", "n"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["equiv", "PreInt"], "[", 
     RowBox[{
      RowBox[{"Pair", "[", 
       RowBox[{"m", ",", "n"}], "]"}], ",", 
      UnderscriptBox["0", "PreInt"]}], "]"}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"m", "\[Equal]", "n"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(equiv 0 PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:52539739", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661231802688738*^9, 3.6612318486776185`*^9}, {
  3.661234481139843*^9, 3.6612344990798745`*^9}},
 CellTags->{
  "ID:52539739", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "equiv 0 PreInt", "IR"},
 CellID->52539739],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "Nat"], 
    RowBox[{"[", 
     RowBox[{"m", ",", "n"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"m", 
     UnderscriptBox["\[GreaterEqual]", "Nat"], "n"}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["equiv", "PreInt"], "[", 
     RowBox[{
      RowBox[{"Pair", "[", 
       RowBox[{"m", ",", "n"}], "]"}], ",", 
      RowBox[{"Pair", "[", 
       RowBox[{
        RowBox[{"m", 
         UnderscriptBox["-", "Nat"], "n"}], ",", 
        UnderscriptBox["0", "Nat"]}], "]"}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(equiv Nat \[GreaterEqual] PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:94484613", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661238483751477*^9, 3.6612385318775616`*^9}},
 CellTags->{
  "ID:94484613", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "equiv Nat \[GreaterEqual] PreInt"},
 CellID->94484613],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "Nat"], 
    RowBox[{"[", 
     RowBox[{"m", ",", "n"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"n", 
     UnderscriptBox["\[GreaterEqual]", "Nat"], "m"}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["equiv", "PreInt"], "[", 
     RowBox[{
      RowBox[{"Pair", "[", 
       RowBox[{"m", ",", "n"}], "]"}], ",", 
      RowBox[{"Pair", "[", 
       RowBox[{
        UnderscriptBox["0", "Nat"], ",", 
        RowBox[{"n", 
         UnderscriptBox["-", "Nat"], "m"}]}], "]"}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(equiv Nat \[LessEqual] PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:321256465", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661238541705579*^9, 3.6612385585224085`*^9}},
 CellTags->{
  "ID:321256465", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "equiv Nat \[LessEqual] PreInt"},
 CellID->321256465],

Cell[BoxData[
 RowBox[{"\[Not]", 
  RowBox[{
   UnderscriptBox["equiv", "PreInt"], "[", 
   RowBox[{
    UnderscriptBox["1", "PreInt"], ",", 
    UnderscriptBox["0", "PreInt"]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Not] 1 equiv 0 PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:129809121", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612378791674147`*^9, 3.6612378954850435`*^9}},
 CellTags->{
  "ID:129809121", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[Not] 1 equiv 0 PreInt"},
 CellID->129809121],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "PreInt"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          413771523}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612317548122535`*^9, 3.661231763002268*^9}},
 CellTags->{
  "ID:413771523", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->413771523],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Exists]", 
   RowBox[{
    UnderscriptBox["\[Element]", "Nat"], 
    RowBox[{"[", "n", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Or]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["equiv", "PreInt"], "[", 
         RowBox[{"a", ",", 
          RowBox[{"Pair", "[", 
           RowBox[{"n", ",", 
            UnderscriptBox["0", "Nat"]}], "]"}]}], "]"}]},
       {
        RowBox[{
         UnderscriptBox["equiv", "PreInt"], "[", 
         RowBox[{"a", ",", 
          RowBox[{"Pair", "[", 
           RowBox[{
            UnderscriptBox["0", "Nat"], ",", "n"}], "]"}]}], "]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(equiv Nat PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:637862621", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612382126844006`*^9, 3.661238249579466*^9}, {
  3.661238835610095*^9, 3.6612388694309545`*^9}, {3.661238935842271*^9, 
  3.6612389395862775`*^9}},
 CellTags->{
  "ID:637862621", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "equiv Nat PreInt"},
 CellID->637862621],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["equiv", "PreInt"], "[", 
  RowBox[{"a", ",", "a"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(equiv reflexive PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:12369324", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612317673858757`*^9, 3.661231773797487*^9}},
 CellTags->{
  "ID:12369324", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "equiv reflexive PreInt"},
 CellID->12369324],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["equiv", "PreInt"], "[", 
   RowBox[{"a", ",", "b"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["equiv", "PreInt"], "[", 
   RowBox[{"b", ",", "a"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(equiv symmetric PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:19276", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612317792262964`*^9, 3.6612317935315213`*^9}},
 CellTags->{
  "ID:19276", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "equiv symmetric PreInt"},
 CellID->19276],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "PreInt"], 
    RowBox[{"[", "c", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["equiv", "PreInt"], "[", 
       RowBox[{"a", ",", "b"}], "]"}], "\[And]", 
      RowBox[{
       UnderscriptBox["equiv", "PreInt"], "[", 
       RowBox[{"b", ",", "c"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     UnderscriptBox["equiv", "PreInt"], "[", 
     RowBox[{"a", ",", "c"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(equiv transitive PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:164883544", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612318700184555`*^9, 3.661231903215314*^9}},
 CellTags->{
  "ID:164883544", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "equiv transitive PreInt"},
 CellID->164883544],

Cell[BoxData[
 RowBox[{"isEquivalence", "[", 
  RowBox[{
   UnderscriptBox["equiv", "PreInt"], ",", "PreInt"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(equiv isEquivalence PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:357440757", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612319508421974`*^9, 3.661231962292618*^9}},
 CellTags->{
  "ID:357440757", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "equiv isEquivalence PreInt"},
 CellID->357440757],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->24574481]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->259129778],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(equiv Pair PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(equiv Pair PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["a1", "ExpressionVariable"], ",", 
            StyleBox["a2", "ExpressionVariable"], ",", 
            StyleBox["b1", "ExpressionVariable"], ",", 
            StyleBox["b2", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["a1", "ExpressionVariable"], ",", 
                   StyleBox["a2", "ExpressionVariable"]}], "]"}], ",", 
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["b1", "ExpressionVariable"], ",", 
                   StyleBox["b2", "ExpressionVariable"]}], "]"}]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             RowBox[{
               StyleBox["a1", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["b2", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["a2", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox[
               "b1", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["a1", "ExpressionVariable"], ",", 
            StyleBox["a2", "ExpressionVariable"], ",", 
            StyleBox["b1", "ExpressionVariable"], ",", 
            StyleBox["b2", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["a1", "ExpressionVariable"], ",", 
                   StyleBox["a2", "ExpressionVariable"]}], "]"}], ",", 
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["b1", "ExpressionVariable"], ",", 
                   StyleBox["b2", "ExpressionVariable"]}], "]"}]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             RowBox[{
               StyleBox["a1", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["b2", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["a2", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox[
               "b1", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:220417861"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p220417861-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p220417861-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:220417861", "Proof|ID:220417861-1"},
 CellID->227048585],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p220417861-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->124141065]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(equiv 0 PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(equiv 0 PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["n", "ExpressionVariable"]}], "]"}], ",", 
               UnderscriptBox["0", "PreInt"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["n", "ExpressionVariable"]}], "]"}], ",", 
               UnderscriptBox["0", "PreInt"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:52539739"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p52539739-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p52539739-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:52539739", "Proof|ID:52539739-1"},
 CellID->142039401],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p52539739-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->135864141]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(equiv Nat \[GreaterEqual] PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(equiv Nat \[GreaterEqual] PreInt)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["n", "ExpressionVariable"]}], "]"}], ",", 
               RowBox[{"Pair", "[", 
                 RowBox[{
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["-", "Nat"], 
                    StyleBox["n", "ExpressionVariable"]}], ",", 
                   UnderscriptBox["0", "Nat"]}], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["n", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["n", "ExpressionVariable"]}], "]"}], ",", 
               RowBox[{"Pair", "[", 
                 RowBox[{
                   RowBox[{
                    StyleBox["m", "ExpressionVariable"], 
                    UnderscriptBox["-", "Nat"], 
                    StyleBox["n", "ExpressionVariable"]}], ",", 
                   UnderscriptBox["0", "Nat"]}], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:94484613"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p94484613-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p94484613-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:94484613", "Proof|ID:94484613-1"},
 CellID->325556718],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p94484613-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{3.6612441601529036`*^9},
 CellID->55143000]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(equiv Nat \[LessEqual] PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(equiv Nat \[LessEqual] PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["n", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["m", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["n", "ExpressionVariable"]}], "]"}], ",", 
               RowBox[{"Pair", "[", 
                 RowBox[{
                   UnderscriptBox["0", "Nat"], ",", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["-", "Nat"], 
                    StyleBox["m", "ExpressionVariable"]}]}], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["n", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox["m", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               RowBox[{"Pair", "[", 
                 RowBox[{
                   StyleBox["m", "ExpressionVariable"], ",", 
                   StyleBox["n", "ExpressionVariable"]}], "]"}], ",", 
               RowBox[{"Pair", "[", 
                 RowBox[{
                   UnderscriptBox["0", "Nat"], ",", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], 
                    UnderscriptBox["-", "Nat"], 
                    StyleBox["m", "ExpressionVariable"]}]}], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:321256465"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p321256465-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p321256465-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:321256465", "Proof|ID:321256465-1"},
 CellID->438580499],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p321256465-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->227406264]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Not] 1 equiv 0 PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Not] 1 equiv 0 PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"\[Not]", 
         RowBox[{
           TagBox["(", "AutoParentheses"], 
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               UnderscriptBox["1", "PreInt"], ",", 
               UnderscriptBox["0", "PreInt"]}], "]"}], 
           TagBox[")", "AutoParentheses"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"\[Not]", 
         RowBox[{
           TagBox["(", "AutoParentheses"], 
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               UnderscriptBox["1", "PreInt"], ",", 
               UnderscriptBox["0", "PreInt"]}], "]"}], 
           TagBox[")", "AutoParentheses"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:129809121"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p129809121-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p129809121-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:129809121", "Proof|ID:129809121-1"},
 CellID->242312613],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p129809121-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->768460961]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(equiv Nat PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(equiv Nat PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", "Nat"], "[", 
                 StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               UnderscriptBox["equiv", "PreInt"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 RowBox[{"Pair", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    UnderscriptBox["0", "Nat"]}], "]"}]}], "]"}], "\[Or]", 
             RowBox[{
               UnderscriptBox["equiv", "PreInt"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 RowBox[{"Pair", "[", 
                   RowBox[{
                    UnderscriptBox["0", "Nat"], ",", 
                    StyleBox["n", "ExpressionVariable"]}], "]"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{
                 UnderscriptBox["\[Element]", "Nat"], "[", 
                 StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             RowBox[{
               UnderscriptBox["equiv", "PreInt"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 RowBox[{"Pair", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    UnderscriptBox["0", "Nat"]}], "]"}]}], "]"}], "\[Or]", 
             RowBox[{
               UnderscriptBox["equiv", "PreInt"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 RowBox[{"Pair", "[", 
                   RowBox[{
                    UnderscriptBox["0", "Nat"], ",", 
                    StyleBox["n", "ExpressionVariable"]}], "]"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:637862621"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p637862621-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p637862621-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:637862621", "Proof|ID:637862621-1"},
 CellID->752123440],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p637862621-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->189898560]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(equiv reflexive PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(equiv reflexive PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["equiv", "PreInt"], "[", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], ",", 
             StyleBox["a", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["equiv", "PreInt"], "[", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], ",", 
             StyleBox["a", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:12369324"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p12369324-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p12369324-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:12369324", "Proof|ID:12369324-1"},
 CellID->28237667],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p12369324-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->609896313]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(equiv symmetric PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(equiv symmetric PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], ",", 
               StyleBox["a", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:19276"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p19276-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p19276-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:19276", "Proof|ID:19276-1"},
 CellID->146831318],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p19276-1-display.m\"}]"]],
      Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->835535524]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(equiv transitive PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(equiv transitive PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["equiv", "PreInt"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["b", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["equiv", "PreInt"], "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["equiv", "PreInt"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 StyleBox["b", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["equiv", "PreInt"], "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", 
                 StyleBox["c", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:164883544"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p164883544-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p164883544-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:164883544", "Proof|ID:164883544-1"},
 CellID->62903797],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p164883544-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->325542796]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(equiv isEquivalence PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(equiv isEquivalence PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isEquivalence", "[", 
         RowBox[{
           UnderscriptBox["equiv", "PreInt"], ",", "PreInt"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isEquivalence", "[", 
         RowBox[{
           UnderscriptBox["equiv", "PreInt"], ",", "PreInt"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:357440757"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p357440757-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p357440757-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:357440757", "Proof|ID:357440757-1"},
 CellID->81770248],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p357440757-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->82676088]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemmas: ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["-", TraditionalForm]]],
 " (\[Checkmark])"
}], "Subsubsection",
 CellChangeTimes->{{3.6611660653501987`*^9, 3.6611660717150097`*^9}, {
  3.661166661568646*^9, 3.661166661568646*^9}, {3.6611739243346043`*^9, 
  3.6611739243346043`*^9}, {3.661250382367835*^9, 3.661250386845043*^9}},
 CellID->55494429],

Cell["", "OpenEnvironment",
 CellID->135470470],

Cell[CellGroupData[{

Cell["Lemma: Properties of + in PreInt", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.661166081901828*^9, 3.661166091932645*^9}},
 CellID->348809526],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"a1", ",", "a2", ",", "b1", ",", "b2"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"Pair", "[", 
      RowBox[{"a1", ",", "a2"}], "]"}], 
     UnderscriptBox["+", "PreInt"], 
     RowBox[{"Pair", "[", 
      RowBox[{"b1", ",", "b2"}], "]"}]}], "\[Equal]", 
    RowBox[{"Pair", "[", 
     RowBox[{
      RowBox[{"a1", 
       UnderscriptBox["+", "Nat"], "b1"}], ",", 
      RowBox[{"a2", 
       UnderscriptBox["+", "Nat"], "b2"}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ Pair PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:125790650", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661167720986107*^9, 3.6611677748830013`*^9}, {
  3.6611679239110637`*^9, 3.6611679251746655`*^9}},
 CellTags->{
  "ID:125790650", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ Pair PreInt", "IR"},
 CellID->125790650],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "PreInt"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          86057977}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611661147554855`*^9, 3.6611661232887*^9}},
 CellTags->{
  "ID:86057977", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->86057977],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "PreInt"], 
  RowBox[{"[", 
   RowBox[{"a", 
    UnderscriptBox["+", "PreInt"], "b"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ closed PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:196108575", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661166125987505*^9, 3.6611661387015276`*^9}},
 CellTags->{
  "ID:196108575", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ closed PreInt"},
 CellID->196108575],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox["+", "PreInt"], 
   UnderscriptBox["0", "PreInt"]}], "\[Equal]", 
  "a"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(0 rneutral PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:132949647", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611663105202293`*^9, 3.6611663244510536`*^9}},
 CellTags->{
  "ID:132949647", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 rneutral PreInt"},
 CellID->132949647],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox["+", "PreInt"], "b"}], "\[Equal]", 
  RowBox[{"b", 
   UnderscriptBox["+", "PreInt"], "a"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ comm PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:25630206", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611663297082624`*^9, 3.661166340519082*^9}},
 CellTags->{
  "ID:25630206", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ comm PreInt"},
 CellID->25630206],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "PreInt"], 
   RowBox[{"[", "c", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          16604231}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611663732167397`*^9, 3.6611663806891527`*^9}},
 CellTags->{
  "ID:16604231", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->16604231],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"a", 
     UnderscriptBox["+", "PreInt"], "b"}], ")"}], 
   UnderscriptBox["+", "PreInt"], "c"}], "\[Equal]", 
  RowBox[{"a", 
   UnderscriptBox["+", "PreInt"], 
   RowBox[{"(", 
    RowBox[{"b", 
     UnderscriptBox["+", "PreInt"], "c"}], 
    ")"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ assoc PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:73792979", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661166388972767*^9, 3.661166423246027*^9}},
 CellTags->{
  "ID:73792979", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ assoc PreInt"},
 CellID->73792979],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["equiv", "PreInt"], "[", 
   RowBox[{
    RowBox[{"a", 
     UnderscriptBox["+", "PreInt"], "c"}], ",", 
    RowBox[{"b", 
     UnderscriptBox["+", "PreInt"], "c"}]}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["equiv", "PreInt"], "[", 
   RowBox[{"a", ",", "b"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ rcanc PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:243295365", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611664386900544`*^9, 3.6611664580808887`*^9}, {
  3.6613175822943673`*^9, 3.6613176117834187`*^9}},
 CellTags->{
  "ID:243295365", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ rcanc PreInt"},
 CellID->243295365],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["equiv", "PreInt"], "[", 
   RowBox[{"a", ",", "b"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["equiv", "PreInt"], "[", 
   RowBox[{
    RowBox[{"a", 
     UnderscriptBox["+", "PreInt"], "c"}], ",", 
    RowBox[{"b", 
     UnderscriptBox["+", "PreInt"], "c"}]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ invaiant equiv PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:47524840", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661232107218872*^9, 3.6612321435045366`*^9}},
 CellTags->{
  "ID:47524840", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ invaiant equiv PreInt"},
 CellID->47524840],

Cell[BoxData[
 RowBox[{"isInvariantFunction2", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["+", "PreInt"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", 
   UnderscriptBox["equiv", "PreInt"], ",", "PreInt"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ isInvariantFunction2 PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:107686175", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661232034520745*^9, 3.6612320743788147`*^9}},
 CellTags->{
  "ID:107686175", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ isInvariantFunction2 PreInt"},
 CellID->107686175],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->125892535]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->92071022],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ Pair PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ Pair PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["a1", "ExpressionVariable"], ",", 
            StyleBox["a2", "ExpressionVariable"], ",", 
            StyleBox["b1", "ExpressionVariable"], ",", 
            StyleBox["b2", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["a1", "ExpressionVariable"], ",", 
                 StyleBox["a2", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["+", "PreInt"], 
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["b1", "ExpressionVariable"], ",", 
                 StyleBox["b2", "ExpressionVariable"]}], "]"}]}], "\[Equal]", 
           RowBox[{"Pair", "[", 
             RowBox[{
               RowBox[{
                 StyleBox["a1", "ExpressionVariable"], 
                 UnderscriptBox["+", "Nat"], 
                 StyleBox["b1", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["a2", "ExpressionVariable"], 
                 UnderscriptBox["+", "Nat"], 
                 StyleBox["b2", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["a1", "ExpressionVariable"], ",", 
            StyleBox["a2", "ExpressionVariable"], ",", 
            StyleBox["b1", "ExpressionVariable"], ",", 
            StyleBox["b2", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["a1", "ExpressionVariable"], ",", 
                 StyleBox["a2", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["+", "PreInt"], 
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["b1", "ExpressionVariable"], ",", 
                 StyleBox["b2", "ExpressionVariable"]}], "]"}]}], "\[Equal]", 
           RowBox[{"Pair", "[", 
             RowBox[{
               RowBox[{
                 StyleBox["a1", "ExpressionVariable"], 
                 UnderscriptBox["+", "Nat"], 
                 StyleBox["b1", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["a2", "ExpressionVariable"], 
                 UnderscriptBox["+", "Nat"], 
                 StyleBox["b2", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:125790650"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p125790650-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p125790650-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:125790650", "Proof|ID:125790650-1"},
 CellID->219935054],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p125790650-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->267994709]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ closed PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ closed PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "PreInt"], "[", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["+", "PreInt"], 
             StyleBox["b", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "PreInt"], "[", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["+", "PreInt"], 
             StyleBox["b", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:196108575"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p196108575-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p196108575-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:196108575", "Proof|ID:196108575-1"},
 CellID->746037930],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p196108575-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->198300422]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(0 rneutral PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(0 rneutral PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["+", "PreInt"], 
             UnderscriptBox["0", "PreInt"]}], "\[Equal]", 
           StyleBox[
           "a", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["+", "PreInt"], 
             UnderscriptBox["0", "PreInt"]}], "\[Equal]", 
           StyleBox[
           "a", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:132949647"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p132949647-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p132949647-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:132949647", "Proof|ID:132949647-1"},
 CellID->16458106],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p132949647-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->24014104]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ comm PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ comm PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["+", "PreInt"], 
             StyleBox["b", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["b", "ExpressionVariable"], 
             UnderscriptBox["+", "PreInt"], 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["+", "PreInt"], 
             StyleBox["b", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["b", "ExpressionVariable"], 
             UnderscriptBox["+", "PreInt"], 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:25630206"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p25630206-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p25630206-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:25630206", "Proof|ID:25630206-1"},
 CellID->184131697],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p25630206-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->15556707]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ assoc PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ assoc PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", "PreInt"], 
               StyleBox["b", "ExpressionVariable"]}], 
             UnderscriptBox["+", "PreInt"], 
             StyleBox["c", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["+", "PreInt"], 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["+", "PreInt"], 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", "PreInt"], 
               StyleBox["b", "ExpressionVariable"]}], 
             UnderscriptBox["+", "PreInt"], 
             StyleBox["c", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["+", "PreInt"], 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["+", "PreInt"], 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:73792979"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p73792979-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p73792979-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:73792979", "Proof|ID:73792979-1"},
 CellID->273331503],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p73792979-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->675199731]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ rcanc PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ rcanc PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", "PreInt"], 
                 StyleBox["c", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["+", "PreInt"], 
                 StyleBox["c", "ExpressionVariable"]}]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", "PreInt"], 
                 StyleBox["c", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["+", "PreInt"], 
                 StyleBox["c", "ExpressionVariable"]}]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:243295365"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p243295365-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p243295365-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:243295365", "Proof|ID:243295365-1"},
 CellID->446127764],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p243295365-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->136545031]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ invaiant equiv PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ invaiant equiv PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", "PreInt"], 
                 StyleBox["c", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["+", "PreInt"], 
                 StyleBox["c", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["+", "PreInt"], 
                 StyleBox["c", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["+", "PreInt"], 
                 StyleBox["c", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:47524840"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p47524840-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p47524840-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:47524840", "Proof|ID:47524840-1"},
 CellID->317988883],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p47524840-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->49485106]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ isInvariantFunction2 PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ isInvariantFunction2 PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isInvariantFunction2", "[", 
         RowBox[{
           UnderscriptBox["+", "PreInt"], ",", 
           UnderscriptBox["equiv", "PreInt"], ",", "PreInt"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isInvariantFunction2", "[", 
         RowBox[{
           UnderscriptBox["+", "PreInt"], ",", 
           UnderscriptBox["equiv", "PreInt"], ",", "PreInt"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:107686175"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p107686175-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p107686175-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:107686175", "Proof|ID:107686175-1"},
 CellID->207023133],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p107686175-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->11877910]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->115221006],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Properties of ",
 Cell[BoxData[
  FormBox["-", TraditionalForm]]],
 " in PreInt"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.661173935941025*^9, 3.661173944021839*^9}},
 CellID->286130886],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"a1", ",", "a2"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["-", "PreInt"], 
     RowBox[{"Pair", "[", 
      RowBox[{"a1", ",", "a2"}], "]"}]}], "\[Equal]", 
    RowBox[{"Pair", "[", 
     RowBox[{"a2", ",", "a1"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(- unary Pair PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:360943410", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661173986828314*^9, 3.6611740104009557`*^9}},
 CellTags->{
  "ID:360943410", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- unary Pair PreInt", "IR"},
 CellID->360943410],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "PreInt"], 
   RowBox[{"[", "a", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          474936069}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661173952617454*^9, 3.6611739603706675`*^9}},
 CellTags->{
  "ID:474936069", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->474936069],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "PreInt"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["-", "PreInt"], "a"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(- unary closed PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:5265224", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611739643798747`*^9, 3.661173980541503*^9}},
 CellTags->{
  "ID:5265224", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- unary closed PreInt", "IR"},
 CellID->5265224],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["equiv", "PreInt"], "[", 
  RowBox[{
   RowBox[{"a", 
    UnderscriptBox["+", "PreInt"], 
    RowBox[{"(", 
     RowBox[{
      UnderscriptBox["-", "PreInt"], "a"}], ")"}]}], ",", 
   UnderscriptBox["0", "PreInt"]}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(- inverse PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:203475906", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611740404798083`*^9, 3.6611740577334385`*^9}, {
  3.661174239926159*^9, 3.661174251345379*^9}, {3.6612315754889383`*^9, 
  3.6612315845213547`*^9}},
 CellTags->{
  "ID:203475906", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- inverse PreInt"},
 CellID->203475906],

Cell[BoxData[
 RowBox[{"isInvariantFunction1", "[", 
  RowBox[{
   UnderscriptBox["Minus", "PreInt"], ",", 
   UnderscriptBox["equiv", "PreInt"], ",", "PreInt"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(- isInvariantFunction1 PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:611704112", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612328444605675`*^9, 3.6612328555677867`*^9}, {
  3.6612502236843567`*^9, 3.661250242107989*^9}},
 CellTags->{
  "ID:611704112", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- isInvariantFunction1 PreInt"},
 CellID->611704112],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->416398914]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->313664736],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(- unary Pair PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(- unary Pair PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["a1", "ExpressionVariable"], ",", 
            StyleBox["a2", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["Minus", "PreInt"], "[", 
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["a1", "ExpressionVariable"], ",", 
                 StyleBox["a2", "ExpressionVariable"]}], "]"}], "]"}], 
           "\[Equal]", 
           RowBox[{"Pair", "[", 
             RowBox[{
               StyleBox["a2", "ExpressionVariable"], ",", 
               StyleBox["a1", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["a1", "ExpressionVariable"], ",", 
            StyleBox["a2", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["Minus", "PreInt"], "[", 
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["a1", "ExpressionVariable"], ",", 
                 StyleBox["a2", "ExpressionVariable"]}], "]"}], "]"}], 
           "\[Equal]", 
           RowBox[{"Pair", "[", 
             RowBox[{
               StyleBox["a2", "ExpressionVariable"], ",", 
               StyleBox["a1", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:360943410"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p360943410-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p360943410-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:360943410", "Proof|ID:360943410-1"},
 CellID->307773152],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p360943410-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->129842997]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(- unary closed PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(- unary closed PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "PreInt"], "[", 
           RowBox[{
             UnderscriptBox["Minus", "PreInt"], "[", 
             StyleBox["a", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "PreInt"], "[", 
           RowBox[{
             UnderscriptBox["Minus", "PreInt"], "[", 
             StyleBox["a", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:5265224"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p5265224-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p5265224-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:5265224", "Proof|ID:5265224-1"},
 CellID->489898059],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p5265224-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->58997533]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(- inverse PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(- inverse PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["equiv", "PreInt"], "[", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", "PreInt"], 
               RowBox[{
                 UnderscriptBox["Minus", "PreInt"], "[", 
                 StyleBox["a", "ExpressionVariable"], "]"}]}], ",", 
             UnderscriptBox["0", "PreInt"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["equiv", "PreInt"], "[", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", "PreInt"], 
               RowBox[{
                 UnderscriptBox["Minus", "PreInt"], "[", 
                 StyleBox["a", "ExpressionVariable"], "]"}]}], ",", 
             UnderscriptBox["0", "PreInt"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:203475906"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p203475906-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p203475906-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:203475906", "Proof|ID:203475906-1"},
 CellID->819919783],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p203475906-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->420597009]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(- isInvariantFunction1 PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(- isInvariantFunction1 PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isInvariantFunction1", "[", 
         RowBox[{
           UnderscriptBox["Minus", "PreInt"], ",", 
           UnderscriptBox["equiv", "PreInt"], ",", "PreInt"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isInvariantFunction1", "[", 
         RowBox[{
           UnderscriptBox["Minus", "PreInt"], ",", 
           UnderscriptBox["equiv", "PreInt"], ",", "PreInt"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:611704112"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p611704112-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p611704112-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:611704112", "Proof|ID:611704112-1"},
 CellID->443769915],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p611704112-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->218910620]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemmas: ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " (\[Checkmark])"
}], "Subsubsection",
 CellChangeTimes->{{3.661173899577361*^9, 3.661173909155778*^9}, {
  3.6612330329732985`*^9, 3.6612330335348997`*^9}, {3.6612505022238455`*^9, 
  3.6612505065294533`*^9}},
 CellID->119023973],

Cell["", "OpenEnvironment",
 CellID->170850154],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Properties of ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " in PreInt"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6611665532888556`*^9, 3.6611665594508667`*^9}, {
  3.6611666671846557`*^9, 3.661166668573058*^9}},
 CellID->17732315],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"a1", ",", "a2", ",", "b1", ",", "b2"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"Pair", "[", 
      RowBox[{"a1", ",", "a2"}], "]"}], 
     UnderscriptBox["*", "PreInt"], 
     RowBox[{"Pair", "[", 
      RowBox[{"b1", ",", "b2"}], "]"}]}], "\[Equal]", 
    RowBox[{"Pair", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"a1", 
        UnderscriptBox["*", "Nat"], "b1"}], 
       UnderscriptBox["+", "Nat"], 
       RowBox[{"a2", 
        UnderscriptBox["*", "Nat"], "b2"}]}], ",", 
      RowBox[{
       RowBox[{"a1", 
        UnderscriptBox["*", "Nat"], "b2"}], 
       UnderscriptBox["+", "Nat"], 
       RowBox[{"a2", 
        UnderscriptBox["*", "Nat"], "b1"}]}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* Pair PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:455358938", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661167809733463*^9, 3.6611678458787265`*^9}, {
  3.6611679272338696`*^9, 3.661167928341471*^9}},
 CellTags->{
  "ID:455358938", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* Pair PreInt", "IR"},
 CellID->455358938],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "PreInt"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          206670450}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611665628828726`*^9, 3.661166585456112*^9}, {
  3.661166671193863*^9, 3.6611666713186626`*^9}},
 CellTags->{
  "ID:206670450", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->206670450],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "PreInt"], 
  RowBox[{"[", 
   RowBox[{"a", 
    UnderscriptBox["*", "PreInt"], "b"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* closed PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:80978087", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661166581977306*^9, 3.661166596890932*^9}, {
  3.661166674126668*^9, 3.6611666783230753`*^9}},
 CellTags->{
  "ID:80978087", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* closed PreInt", "IR"},
 CellID->80978087],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox["*", "PreInt"], 
   UnderscriptBox["0", "PreInt"]}], "\[Equal]", 
  UnderscriptBox["0", "PreInt"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(0 rannihilator PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:561254237", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611667239687557`*^9, 3.6611667377903795`*^9}},
 CellTags->{
  "ID:561254237", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 rannihilator PreInt"},
 CellID->561254237],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox["*", "PreInt"], 
   UnderscriptBox["1", "PreInt"]}], "\[Equal]", 
  "a"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1 rneutral PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:201948360", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611666059857483`*^9, 3.6611666227713776`*^9}, {
  3.6611666834866843`*^9, 3.6611666951867046`*^9}},
 CellTags->{
  "ID:201948360", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 rneutral PreInt"},
 CellID->201948360],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox["*", "PreInt"], "b"}], "\[Equal]", 
  RowBox[{"b", 
   UnderscriptBox["*", "PreInt"], "a"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* comm PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:152728157", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.66116670362632*^9, 3.6611667143279386`*^9}},
 CellTags->{
  "ID:152728157", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* comm PreInt"},
 CellID->152728157],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["equiv", "PreInt"], "[", 
   RowBox[{
    RowBox[{"a", 
     UnderscriptBox["*", "PreInt"], "b"}], ",", 
    UnderscriptBox["0", "PreInt"]}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["equiv", "PreInt"], "[", 
     RowBox[{"a", ",", 
      UnderscriptBox["0", "PreInt"]}], "]"}], "\[Or]", 
    RowBox[{
     UnderscriptBox["equiv", "PreInt"], "[", 
     RowBox[{"b", ",", 
      UnderscriptBox["0", "PreInt"]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* zdf PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:258603229", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611668407193604`*^9, 3.661166867130207*^9}, {
  3.661318007512314*^9, 3.6613180523819923`*^9}},
 CellTags->{
  "ID:258603229", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* zdf PreInt"},
 CellID->258603229],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "PreInt"], 
   RowBox[{"[", "c", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          200297603}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611667460115943`*^9, 3.661166753499607*^9}},
 CellTags->{
  "ID:200297603", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->200297603],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"a", 
     UnderscriptBox["*", "PreInt"], "b"}], ")"}], 
   UnderscriptBox["*", "PreInt"], "c"}], "\[Equal]", 
  RowBox[{"a", 
   UnderscriptBox["*", "PreInt"], 
   RowBox[{"(", 
    RowBox[{"b", 
     UnderscriptBox["*", "PreInt"], "c"}], 
    ")"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* assoc PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:142359696", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661166759505618*^9, 3.6611667837636604`*^9}},
 CellTags->{
  "ID:142359696", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* assoc PreInt"},
 CellID->142359696],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox["*", "PreInt"], 
   RowBox[{"(", 
    RowBox[{"b", 
     UnderscriptBox["+", "PreInt"], "c"}], ")"}]}], "\[Equal]", 
  RowBox[{
   RowBox[{"a", 
    UnderscriptBox["*", "PreInt"], "b"}], 
   UnderscriptBox["+", "PreInt"], 
   RowBox[{"a", 
    UnderscriptBox["*", "PreInt"], "c"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* distr + PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:262630475", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611667889740696`*^9, 3.6611668170229187`*^9}},
 CellTags->{
  "ID:262630475", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* distr + PreInt"},
 CellID->262630475],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["equiv", "PreInt"], "[", 
   RowBox[{"a", ",", "b"}], "]"}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["equiv", "PreInt"], "[", 
   RowBox[{
    RowBox[{"a", 
     UnderscriptBox["*", "PreInt"], "c"}], ",", 
    RowBox[{"b", 
     UnderscriptBox["*", "PreInt"], "c"}]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* invariant equiv PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:146416852", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612330433473167`*^9, 3.6612330713805656`*^9}},
 CellTags->{
  "ID:146416852", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* invariant equiv PreInt"},
 CellID->146416852],

Cell[BoxData[
 RowBox[{"isInvariantFunction2", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["*", "PreInt"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", 
   UnderscriptBox["equiv", "PreInt"], ",", "PreInt"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* isInvariantFunction2 PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:204021385", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661233097011411*^9, 3.6612330989770145`*^9}},
 CellTags->{
  "ID:204021385", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* isInvariantFunction2 PreInt"},
 CellID->204021385],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->102737931]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->282267077],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* Pair PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* Pair PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["a1", "ExpressionVariable"], ",", 
            StyleBox["a2", "ExpressionVariable"], ",", 
            StyleBox["b1", "ExpressionVariable"], ",", 
            StyleBox["b2", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["a1", "ExpressionVariable"], ",", 
                 StyleBox["a2", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["*", "PreInt"], 
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["b1", "ExpressionVariable"], ",", 
                 StyleBox["b2", "ExpressionVariable"]}], "]"}]}], "\[Equal]", 
           RowBox[{"Pair", "[", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a1", "ExpressionVariable"], 
                   UnderscriptBox["*", "Nat"], 
                   StyleBox["b1", "ExpressionVariable"]}], 
                 UnderscriptBox["+", "Nat"], 
                 RowBox[{
                   StyleBox["a2", "ExpressionVariable"], 
                   UnderscriptBox["*", "Nat"], 
                   StyleBox["b2", "ExpressionVariable"]}]}], ",", 
               RowBox[{
                 RowBox[{
                   StyleBox["a1", "ExpressionVariable"], 
                   UnderscriptBox["*", "Nat"], 
                   StyleBox["b2", "ExpressionVariable"]}], 
                 UnderscriptBox["+", "Nat"], 
                 RowBox[{
                   StyleBox["a2", "ExpressionVariable"], 
                   UnderscriptBox["*", "Nat"], 
                   StyleBox["b1", "ExpressionVariable"]}]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["a1", "ExpressionVariable"], ",", 
            StyleBox["a2", "ExpressionVariable"], ",", 
            StyleBox["b1", "ExpressionVariable"], ",", 
            StyleBox["b2", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["a1", "ExpressionVariable"], ",", 
                 StyleBox["a2", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["*", "PreInt"], 
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["b1", "ExpressionVariable"], ",", 
                 StyleBox["b2", "ExpressionVariable"]}], "]"}]}], "\[Equal]", 
           RowBox[{"Pair", "[", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   StyleBox["a1", "ExpressionVariable"], 
                   UnderscriptBox["*", "Nat"], 
                   StyleBox["b1", "ExpressionVariable"]}], 
                 UnderscriptBox["+", "Nat"], 
                 RowBox[{
                   StyleBox["a2", "ExpressionVariable"], 
                   UnderscriptBox["*", "Nat"], 
                   StyleBox["b2", "ExpressionVariable"]}]}], ",", 
               RowBox[{
                 RowBox[{
                   StyleBox["a1", "ExpressionVariable"], 
                   UnderscriptBox["*", "Nat"], 
                   StyleBox["b2", "ExpressionVariable"]}], 
                 UnderscriptBox["+", "Nat"], 
                 RowBox[{
                   StyleBox["a2", "ExpressionVariable"], 
                   UnderscriptBox["*", "Nat"], 
                   StyleBox["b1", "ExpressionVariable"]}]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:455358938"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p455358938-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p455358938-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:455358938", "Proof|ID:455358938-1"},
 CellID->297124414],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p455358938-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->122132350]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* closed PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* closed PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "PreInt"], "[", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["*", "PreInt"], 
             StyleBox["b", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "PreInt"], "[", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["*", "PreInt"], 
             StyleBox["b", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:80978087"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p80978087-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p80978087-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:80978087", "Proof|ID:80978087-1"},
 CellID->295496011],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p80978087-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->498778090]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(0 rannihilator PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(0 rannihilator PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["*", "PreInt"], 
             UnderscriptBox["0", "PreInt"]}], "\[Equal]", 
           UnderscriptBox["0", "PreInt"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["*", "PreInt"], 
             UnderscriptBox["0", "PreInt"]}], "\[Equal]", 
           UnderscriptBox["0", "PreInt"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:561254237"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p561254237-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p561254237-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:561254237", "Proof|ID:561254237-1"},
 CellID->690925667],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p561254237-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->25562229]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(1 rneutral PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(1 rneutral PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["*", "PreInt"], 
             UnderscriptBox["1", "PreInt"]}], "\[Equal]", 
           StyleBox[
           "a", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["*", "PreInt"], 
             UnderscriptBox["1", "PreInt"]}], "\[Equal]", 
           StyleBox[
           "a", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:201948360"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p201948360-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p201948360-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:201948360", "Proof|ID:201948360-1"},
 CellID->558544782],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p201948360-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->365680064]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* comm PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* comm PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["*", "PreInt"], 
             StyleBox["b", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["b", "ExpressionVariable"], 
             UnderscriptBox["*", "PreInt"], 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["*", "PreInt"], 
             StyleBox["b", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["b", "ExpressionVariable"], 
             UnderscriptBox["*", "PreInt"], 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:152728157"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p152728157-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p152728157-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:152728157", "Proof|ID:152728157-1"},
 CellID->24746466],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p152728157-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->497158307]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* zdf PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* zdf PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["*", "PreInt"], 
                 StyleBox["b", "ExpressionVariable"]}], ",", 
               UnderscriptBox["0", "PreInt"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["equiv", "PreInt"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 UnderscriptBox["0", "PreInt"]}], "]"}], "\[Or]", 
             RowBox[{
               UnderscriptBox["equiv", "PreInt"], "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", 
                 UnderscriptBox["0", "PreInt"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["*", "PreInt"], 
                 StyleBox["b", "ExpressionVariable"]}], ",", 
               UnderscriptBox["0", "PreInt"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["equiv", "PreInt"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 UnderscriptBox["0", "PreInt"]}], "]"}], "\[Or]", 
             RowBox[{
               UnderscriptBox["equiv", "PreInt"], "[", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], ",", 
                 UnderscriptBox["0", "PreInt"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:258603229"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p258603229-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p258603229-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:258603229", "Proof|ID:258603229-1"},
 CellID->149136758],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p258603229-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->40095241]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* assoc PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* assoc PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["*", "PreInt"], 
               StyleBox["b", "ExpressionVariable"]}], 
             UnderscriptBox["*", "PreInt"], 
             StyleBox["c", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["*", "PreInt"], 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["*", "PreInt"], 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["*", "PreInt"], 
               StyleBox["b", "ExpressionVariable"]}], 
             UnderscriptBox["*", "PreInt"], 
             StyleBox["c", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["*", "PreInt"], 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["*", "PreInt"], 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:142359696"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p142359696-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p142359696-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:142359696", "Proof|ID:142359696-1"},
 CellID->796805],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p142359696-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->116032710]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* distr + PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* distr + PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["*", "PreInt"], 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["+", "PreInt"], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["*", "PreInt"], 
               StyleBox["b", "ExpressionVariable"]}], 
             UnderscriptBox["+", "PreInt"], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["*", "PreInt"], 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["*", "PreInt"], 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["+", "PreInt"], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["*", "PreInt"], 
               StyleBox["b", "ExpressionVariable"]}], 
             UnderscriptBox["+", "PreInt"], 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["*", "PreInt"], 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:262630475"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p262630475-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p262630475-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:262630475", "Proof|ID:262630475-1"},
 CellID->457849091],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p262630475-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->105684652]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* invariant equiv PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* invariant equiv PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["*", "PreInt"], 
                 StyleBox["c", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["*", "PreInt"], 
                 StyleBox["c", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], 
                 UnderscriptBox["*", "PreInt"], 
                 StyleBox["c", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["b", "ExpressionVariable"], 
                 UnderscriptBox["*", "PreInt"], 
                 StyleBox["c", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:146416852"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p146416852-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p146416852-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:146416852", "Proof|ID:146416852-1"},
 CellID->83067894],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p146416852-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->109470325]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* isInvariantFunction2 PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* isInvariantFunction2 PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isInvariantFunction2", "[", 
         RowBox[{
           UnderscriptBox["*", "PreInt"], ",", 
           UnderscriptBox["equiv", "PreInt"], ",", "PreInt"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isInvariantFunction2", "[", 
         RowBox[{
           UnderscriptBox["*", "PreInt"], ",", 
           UnderscriptBox["equiv", "PreInt"], ",", "PreInt"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:204021385"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p204021385-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p204021385-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:204021385", "Proof|ID:204021385-1"},
 CellID->716469443],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p204021385-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->145105386]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Lemmas: \[GreaterEqual] (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6611738700309086`*^9, 3.661173874055716*^9}, {
  3.6611739130401845`*^9, 3.6611739172677917`*^9}, {3.661251780536292*^9, 
  3.661251784748299*^9}},
 CellID->293842696],

Cell["", "OpenEnvironment",
 CellID->614975653],

Cell[CellGroupData[{

Cell["\<\
Lemma: Properties of \[GreaterEqual] in PreInt\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6611743541339593`*^9, 3.661174360685971*^9}},
 CellID->100053836],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"a1", ",", "a2", ",", "b1", ",", "b2"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"Pair", "[", 
       RowBox[{"a1", ",", "a2"}], "]"}], 
      UnderscriptBox["\[GreaterEqual]", "PreInt"], 
      RowBox[{"Pair", "[", 
       RowBox[{"b1", ",", "b2"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"a1", 
       UnderscriptBox["+", "Nat"], "b2"}], 
      UnderscriptBox["\[GreaterEqual]", "Nat"], 
      RowBox[{"a2", 
       UnderscriptBox["+", "Nat"], "b1"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] Pair PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:129935970", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661174372214391*^9, 3.661174431863096*^9}},
 CellTags->{
  "ID:129935970", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] Pair PreInt", "IR"},
 CellID->129935970],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "Nat"], 
    RowBox[{"[", 
     RowBox[{"m", ",", "n"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"Pair", "[", 
       RowBox[{"m", ",", "n"}], "]"}], 
      UnderscriptBox["\[GreaterEqual]", "PreInt"], 
      UnderscriptBox["0", "PreInt"]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"m", 
      UnderscriptBox["\[GreaterEqual]", "Nat"], "n"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] 0 PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:95876614", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612367136551647`*^9, 3.6612367539488354`*^9}, {
  3.661236884458665*^9, 3.661236886720669*^9}},
 CellTags->{
  "ID:95876614", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] 0 PreInt", "IR"},
 CellID->95876614],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["1", "PreInt"], 
  UnderscriptBox["\[GreaterEqual]", "PreInt"], 
  UnderscriptBox["0", "PreInt"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1 \[GreaterEqual] 0 PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:221338547", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611750396245637`*^9, 3.661175050622583*^9}, {
  3.661175089466651*^9, 3.6611750916350546`*^9}},
 CellTags->{
  "ID:221338547", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 \[GreaterEqual] 0 PreInt"},
 CellID->221338547],

Cell[BoxData[
 RowBox[{"isInvariantRelation2", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[GreaterEqual]", "PreInt"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", 
   UnderscriptBox["equiv", "PreInt"], ",", "PreInt"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(\[GreaterEqual] isInvariantRelation2 PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:13569836", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612356440488853`*^9, 3.661235688009763*^9}},
 CellTags->{
  "ID:13569836", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] isInvariantRelation2 PreInt"},
 CellID->13569836],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "PreInt"], 
   RowBox[{"[", 
    RowBox[{"a", ",", "b"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          337456597}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611744591319437`*^9, 3.6611744669631577`*^9}},
 CellTags->{
  "ID:337456597", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->337456597],

Cell[BoxData[
 RowBox[{"a", 
  UnderscriptBox["\[GreaterEqual]", "PreInt"], 
  "a"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] reflexive PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:83533617", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611744765415745`*^9, 3.661174481798784*^9}},
 CellTags->{
  "ID:83533617", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] reflexive PreInt"},
 CellID->83533617],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", 
     UnderscriptBox["\[GreaterEqual]", "PreInt"], "b"}], "\[And]", 
    RowBox[{"b", 
     UnderscriptBox["\[GreaterEqual]", "PreInt"], "a"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   UnderscriptBox["equiv", "PreInt"], "[", 
   RowBox[{"a", ",", "b"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] antisymmetric PreInt)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:281735911", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611746011389933`*^9, 3.6611746231818323`*^9}},
 CellTags->{
  "ID:281735911", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] antisymmetric PreInt"},
 CellID->281735911],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox["\[GreaterEqual]", "PreInt"], "b"}], "\[Or]", 
  RowBox[{"b", 
   UnderscriptBox["\[GreaterEqual]", "PreInt"], 
   "a"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] total PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:115550209", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661174850146631*^9, 3.6611748617998514`*^9}},
 CellTags->{
  "ID:115550209", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] total PreInt"},
 CellID->115550209],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", 
     UnderscriptBox["\[GreaterEqual]", "PreInt"], 
     UnderscriptBox["0", "PreInt"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{
       UnderscriptBox["equiv", "PreInt"], "[", 
       RowBox[{"a", ",", 
        UnderscriptBox["0", "PreInt"]}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"a", 
   UnderscriptBox["\[GreaterEqual]", "PreInt"], 
   UnderscriptBox["1", "PreInt"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] discrete PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:215340491", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611749391759872`*^9, 3.6611749820916624`*^9}},
 CellTags->{
  "ID:215340491", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] discrete PreInt"},
 CellID->215340491],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", 
     UnderscriptBox["\[GreaterEqual]", "PreInt"], 
     UnderscriptBox["0", "PreInt"]}], "\[And]", 
    RowBox[{"b", 
     UnderscriptBox["\[GreaterEqual]", "PreInt"], 
     UnderscriptBox["0", "PreInt"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"a", 
    UnderscriptBox["*", "PreInt"], "b"}], 
   UnderscriptBox["\[GreaterEqual]", "PreInt"], 
   UnderscriptBox["0", "PreInt"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] * PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:92075332", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.66117473027602*^9, 3.6611747662964835`*^9}, {
  3.6612312653115935`*^9, 3.66123128033442*^9}},
 CellTags->{
  "ID:92075332", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] * PreInt"},
 CellID->92075332],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox["\[GreaterEqual]", "PreInt"], 
   UnderscriptBox["0", "PreInt"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "Nat"], 
      RowBox[{"[", "n", "]"}]}]], 
    RowBox[{
     UnderscriptBox["equiv", "PreInt"], "[", 
     RowBox[{"a", ",", 
      RowBox[{"Pair", "[", 
       RowBox[{"n", ",", 
        UnderscriptBox["0", "Nat"]}], "]"}]}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Nat \[SucceedsTilde] PreInt+)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:145078248", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6613240354185114`*^9, 3.6613240707993736`*^9}},
 CellTags->{
  "ID:145078248", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat \[SucceedsTilde] PreInt+"},
 CellID->145078248],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "PreInt"], 
   RowBox[{"[", "c", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          118070489}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6611746527906837`*^9, 3.6611746598574963`*^9}},
 CellTags->{
  "ID:118070489", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->118070489],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"a", 
     UnderscriptBox["\[GreaterEqual]", "PreInt"], "b"}], "\[And]", 
    RowBox[{"b", 
     UnderscriptBox["\[GreaterEqual]", "PreInt"], "c"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"a", 
   UnderscriptBox["\[GreaterEqual]", "PreInt"], 
   "c"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] transitive PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:307661531", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661174665021106*^9, 3.6611746843339396`*^9}},
 CellTags->{
  "ID:307661531", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] transitive PreInt"},
 CellID->307661531],

Cell[BoxData[
 RowBox[{
  RowBox[{"a", 
   UnderscriptBox["\[GreaterEqual]", "PreInt"], "b"}], "\[Implies]", 
  RowBox[{
   RowBox[{"a", 
    UnderscriptBox["+", "PreInt"], "c"}], 
   UnderscriptBox["\[GreaterEqual]", "PreInt"], 
   RowBox[{"b", 
    UnderscriptBox["+", "PreInt"], "c"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] + PreInt)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:647932785", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661174696548761*^9, 3.6611747187632*^9}},
 CellTags->{
  "ID:647932785", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] + PreInt"},
 CellID->647932785],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->15173820]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->362378933],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] Pair PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] Pair PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["a1", "ExpressionVariable"], ",", 
            StyleBox["a2", "ExpressionVariable"], ",", 
            StyleBox["b1", "ExpressionVariable"], ",", 
            StyleBox["b2", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["a1", "ExpressionVariable"], ",", 
                 StyleBox["a2", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["b1", "ExpressionVariable"], ",", 
                 StyleBox["b2", "ExpressionVariable"]}], "]"}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             RowBox[{
               StyleBox["a1", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["b2", "ExpressionVariable"]}], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             RowBox[{
               StyleBox["a2", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox[
               "b1", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["a1", "ExpressionVariable"], ",", 
            StyleBox["a2", "ExpressionVariable"], ",", 
            StyleBox["b1", "ExpressionVariable"], ",", 
            StyleBox["b2", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["a1", "ExpressionVariable"], ",", 
                 StyleBox["a2", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["b1", "ExpressionVariable"], ",", 
                 StyleBox["b2", "ExpressionVariable"]}], "]"}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             RowBox[{
               StyleBox["a1", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox["b2", "ExpressionVariable"]}], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             RowBox[{
               StyleBox["a2", "ExpressionVariable"], 
               UnderscriptBox["+", "Nat"], 
               StyleBox[
               "b1", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:129935970"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p129935970-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p129935970-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:129935970", "Proof|ID:129935970-1"},
 CellID->9772636],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p129935970-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->22268996]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] 0 PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] 0 PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["n", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             UnderscriptBox["0", "PreInt"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Nat"], "[", 
               StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"Pair", "[", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], ",", 
                 StyleBox["n", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             UnderscriptBox["0", "PreInt"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["m", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Nat"], 
             StyleBox[
             "n", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:95876614"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p95876614-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p95876614-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:95876614", "Proof|ID:95876614-1"},
 CellID->62135307],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p95876614-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->392128574]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(1 \[GreaterEqual] 0 PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(1 \[GreaterEqual] 0 PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["1", "PreInt"], 
         UnderscriptBox["\[GreaterEqual]", "PreInt"], 
         UnderscriptBox["0", "PreInt"]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["1", "PreInt"], 
         UnderscriptBox["\[GreaterEqual]", "PreInt"], 
         UnderscriptBox["0", "PreInt"]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:221338547"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p221338547-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p221338547-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:221338547", "Proof|ID:221338547-1"},
 CellID->51061701],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p221338547-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->92240380]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] isInvariantRelation2 PreInt)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] isInvariantRelation2 PreInt)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isInvariantRelation2", "[", 
         RowBox[{
           UnderscriptBox["\[GreaterEqual]", "PreInt"], ",", 
           UnderscriptBox["equiv", "PreInt"], ",", "PreInt"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isInvariantRelation2", "[", 
         RowBox[{
           UnderscriptBox["\[GreaterEqual]", "PreInt"], ",", 
           UnderscriptBox["equiv", "PreInt"], ",", "PreInt"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:13569836"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p13569836-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p13569836-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:13569836", "Proof|ID:13569836-1"},
 CellID->553106541],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p13569836-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->461282630]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] reflexive PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] reflexive PreInt)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["a", "ExpressionVariable"], 
           UnderscriptBox["\[GreaterEqual]", "PreInt"], 
           StyleBox[
           "a", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["a", "ExpressionVariable"], 
           UnderscriptBox["\[GreaterEqual]", "PreInt"], 
           StyleBox[
           "a", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:83533617"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p83533617-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p83533617-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:83533617", "Proof|ID:83533617-1"},
 CellID->1756479],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p83533617-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->169206792]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] antisymmetric PreInt)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] antisymmetric PreInt)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "PreInt"], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "PreInt"], 
               StyleBox["a", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "PreInt"], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "PreInt"], 
               StyleBox["a", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["equiv", "PreInt"], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], ",", 
               StyleBox["b", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:281735911"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p281735911-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p281735911-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:281735911", "Proof|ID:281735911-1"},
 CellID->44053081],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p281735911-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->621527603]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] total PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] total PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             StyleBox["b", "ExpressionVariable"]}], "\[Or]", 
           RowBox[{
             StyleBox["b", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             StyleBox["b", "ExpressionVariable"]}], "\[Or]", 
           RowBox[{
             StyleBox["b", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             StyleBox[
             "a", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:115550209"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p115550209-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p115550209-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:115550209", "Proof|ID:115550209-1"},
 CellID->877842890],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p115550209-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->46153360]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] discrete PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] discrete PreInt)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "PreInt"], 
               UnderscriptBox["0", "PreInt"]}], "\[And]", 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   UnderscriptBox["equiv", "PreInt"], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    UnderscriptBox["0", "PreInt"]}], "]"}], 
                 TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             UnderscriptBox["1", "PreInt"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "PreInt"], 
               UnderscriptBox["0", "PreInt"]}], "\[And]", 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   UnderscriptBox["equiv", "PreInt"], "[", 
                   RowBox[{
                    StyleBox["a", "ExpressionVariable"], ",", 
                    UnderscriptBox["0", "PreInt"]}], "]"}], 
                 TagBox[")", "AutoParentheses"]}]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             UnderscriptBox["1", "PreInt"]}]}]}], $CellContext`TheoremaForm]],
      "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:215340491"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p215340491-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p215340491-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:215340491", "Proof|ID:215340491-1"},
 CellID->147476088],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p215340491-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->342294249]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] * PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] * PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "PreInt"], 
               UnderscriptBox["0", "PreInt"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "PreInt"], 
               UnderscriptBox["0", "PreInt"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["*", "PreInt"], 
               StyleBox["b", "ExpressionVariable"]}], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             UnderscriptBox["0", "PreInt"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "PreInt"], 
               UnderscriptBox["0", "PreInt"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "PreInt"], 
               UnderscriptBox["0", "PreInt"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["*", "PreInt"], 
               StyleBox["b", "ExpressionVariable"]}], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             UnderscriptBox["0", "PreInt"]}]}]}], $CellContext`TheoremaForm]],
      "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:92075332"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p92075332-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p92075332-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:92075332", "Proof|ID:92075332-1"},
 CellID->47485746],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p92075332-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->108029984]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Nat \[SucceedsTilde] PreInt+)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Nat \[SucceedsTilde] PreInt+)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             UnderscriptBox["0", "PreInt"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "Nat"], "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["equiv", "PreInt"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 RowBox[{"Pair", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    UnderscriptBox["0", "Nat"]}], "]"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             UnderscriptBox["0", "PreInt"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "Nat"], "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["equiv", "PreInt"], "[", 
               RowBox[{
                 StyleBox["a", "ExpressionVariable"], ",", 
                 RowBox[{"Pair", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    UnderscriptBox["0", "Nat"]}], "]"}]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:145078248"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p145078248-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p145078248-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:145078248", "Proof|ID:145078248-1"},
 CellID->565621444],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p145078248-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->548997320]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] transitive PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] transitive PreInt)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "PreInt"], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "PreInt"], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "PreInt"], 
               StyleBox["b", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "PreInt"], 
               StyleBox["c", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:307661531"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p307661531-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p307661531-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:307661531", "Proof|ID:307661531-1"},
 CellID->58099428],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p307661531-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->476167925]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] + PreInt)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] + PreInt)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", "PreInt"], 
               StyleBox["c", "ExpressionVariable"]}], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["+", "PreInt"], 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             StyleBox["b", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", "PreInt"], 
               StyleBox["c", "ExpressionVariable"]}], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             RowBox[{
               StyleBox["b", "ExpressionVariable"], 
               UnderscriptBox["+", "PreInt"], 
               StyleBox[
               "c", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:647932785"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p647932785-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p647932785-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:647932785", "Proof|ID:647932785-1"},
 CellID->4451951],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p647932785-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->15683258]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Int Functor (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.661165952328*^9, 3.661165954605604*^9}, {
  3.6613374188165536`*^9, 3.6613374227321606`*^9}},
 CellID->414452305],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"\[ScriptCapitalQ]", "=", 
   RowBox[{"QuoDom", "[", 
    RowBox[{"PreInt", ",", 
     UnderscriptBox["equiv", "PreInt"]}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          463204962}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612518410643983`*^9, 3.6612518600808315`*^9}},
 CellTags->{
  "ID:463204962", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->463204962],

Cell[CellGroupData[{

Cell["Definitions", "Subsubsection",
 CellChangeTimes->{{3.6612522541375237`*^9, 3.6612522563995275`*^9}},
 CellID->269593602],

Cell["", "OpenEnvironment",
 CellID->614819668],

Cell[CellGroupData[{

Cell["Definition: Int Functor", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.661251814279151*^9, 3.6612518168999557`*^9}},
 CellID->201483006],

Cell[BoxData[
 RowBox[{"Int", ":=", 
  UnderscriptBox["\[CapitalDelta]", "N"]}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          32325634}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612518219387646`*^9, 3.6612518285063763`*^9}},
 CellTags->{
  "ID:32325634", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->32325634],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   UnderscriptBox["\[Element]", "N"], 
   TagBox[")",
    "AutoParentheses"]}], ":=", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   UnderscriptBox["\[Element]", "\[ScriptCapitalQ]"], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Int \[Element])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:112681936", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661251863856038*^9, 3.6612518783016634`*^9}, {
  3.6612526205685673`*^9, 3.6612526650286455`*^9}},
 CellTags->{
  "ID:112681936", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Int \[Element]"},
 CellID->112681936],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["0", "N"], ":=", 
  RowBox[{
   UnderscriptBox["EC", "\[ScriptCapitalQ]"], "[", 
   UnderscriptBox["0", "PreInt"], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Int 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:52976508", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612521807705946`*^9, 3.6612521958714213`*^9}},
 CellTags->{
  "ID:52976508", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Int 0"},
 CellID->52976508],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["1", "N"], ":=", 
  RowBox[{
   UnderscriptBox["EC", "\[ScriptCapitalQ]"], "[", 
   UnderscriptBox["1", "PreInt"], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Int 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:712864778", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612521997870283`*^9, 3.6612522022206326`*^9}},
 CellTags->{
  "ID:712864778", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Int 1"},
 CellID->712864778],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   TagBox[
    UnderscriptBox["+", "N"],
    Identity], 
   TagBox[")",
    "AutoParentheses"]}], ":=", 
  RowBox[{
   UnderscriptBox["Fun2", "\[ScriptCapitalQ]"], "[", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["+", "PreInt"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Int +)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:19419340", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661251895539694*^9, 3.6612519641954145`*^9}},
 CellTags->{
  "ID:19419340", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Int +"},
 CellID->19419340],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["Minus", "N"], ":=", 
  RowBox[{
   UnderscriptBox["Fun1", "\[ScriptCapitalQ]"], "[", 
   UnderscriptBox["Minus", "PreInt"], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Int - unary)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:280798627", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661251971683428*^9, 3.6612519933674655`*^9}},
 CellTags->{
  "ID:280798627", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Int - unary"},
 CellID->280798627],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   TagBox[
    UnderscriptBox["*", "N"],
    Identity], 
   TagBox[")",
    "AutoParentheses"]}], ":=", 
  RowBox[{
   UnderscriptBox["Fun2", "\[ScriptCapitalQ]"], "[", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["*", "PreInt"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Int *)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:328192039", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.661251973602231*^9, {3.661252008733493*^9, 3.661252015706705*^9}},
 CellTags->{
  "ID:328192039", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Int *"},
 CellID->328192039],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   TagBox[
    UnderscriptBox["\[GreaterEqual]", "N"],
    Identity], 
   TagBox[")",
    "AutoParentheses"]}], ":=", 
  RowBox[{
   UnderscriptBox["Rel2", "\[ScriptCapitalQ]"], "[", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[GreaterEqual]", "PreInt"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Int \[GreaterEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:819512290", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.6612519745070324`*^9, {3.661252018109109*^9, 3.6612520316967335`*^9}},
 CellTags->{
  "ID:819512290", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Int \[GreaterEqual]"},
 CellID->819512290],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->395448942]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->281466815]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Lemmas: General Properties of Int (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.6612522613915367`*^9, 3.66125226895755*^9}, {
  3.6612534295889883`*^9, 3.661253433301795*^9}, {3.661253565340427*^9, 
  3.661253569880035*^9}},
 CellID->95650746],

Cell["", "OpenEnvironment",
 CellID->570814532],

Cell[CellGroupData[{

Cell["Lemma: General Properties of Int", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6612522750103607`*^9, 3.661252281016371*^9}},
 CellID->30393050],

Cell[BoxData[
 RowBox[{"isDomain", "[", "Int", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Int isDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:354647619", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661252283933576*^9, 3.661252289643186*^9}},
 CellTags->{
  "ID:354647619", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Int isDomain", "IR"},
 CellID->354647619],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "x"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", "Int"], 
      RowBox[{"[", "x", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        UnderscriptBox["\[Element]", "PreInt"], 
        RowBox[{"[", "a", "]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"x", "\[Equal]", 
        RowBox[{
         UnderscriptBox["EC", "\[ScriptCapitalQ]"], "[", "a", "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] Int alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:565027904", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661252307162017*^9, 3.6612523807785463`*^9}},
 CellTags->{
  "ID:565027904", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[Element] Int alternative", "IR"},
 CellID->565027904],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "PreInt"], 
    RowBox[{"[", "a", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", "Int"], 
    RowBox[{"[", 
     RowBox[{
      UnderscriptBox["EC", "\[ScriptCapitalQ]"], "[", "a", "]"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(EC \[Element] Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:40757894", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661252394945371*^9, 3.6612524232126207`*^9}, {
  3.66125285544458*^9, 3.6612528683614025`*^9}},
 CellTags->{
  "ID:40757894", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "EC \[Element] Int", "IR"},
 CellID->40757894],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "Int"], 
  RowBox[{"[", 
   UnderscriptBox["0", "Int"], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(0 \[Element] Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:350791961", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661252437814246*^9, 3.6612524475954638`*^9}},
 CellTags->{
  "ID:350791961", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 \[Element] Int", "IR"},
 CellID->350791961],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "Int"], 
  RowBox[{"[", 
   UnderscriptBox["1", "Int"], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1 \[Element] Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:181240486", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661252457345481*^9, 3.661252458640283*^9}},
 CellTags->{
  "ID:181240486", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 \[Element] Int"},
 CellID->181240486],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["1", "Int"], "\[NotEqual]", 
  UnderscriptBox["0", "Int"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1 \[NotEqual] 0 Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:364462666", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612524604654865`*^9, 3.661252467625899*^9}},
 CellTags->{
  "ID:364462666", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 \[NotEqual] 0 Int"},
 CellID->364462666],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->388453292]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->418936054],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Int isDomain)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Int isDomain)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isDomain", "[", "Int", "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isDomain", "[", "Int", "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:354647619"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p354647619-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p354647619-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:354647619", "Proof|ID:354647619-1"},
 CellID->268377469],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p354647619-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->8704553]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] Int alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] Int alternative)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["x", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "PreInt"], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["EC", 
                  RowBox[{"QuoDom", "[", 
                    RowBox[{"PreInt", ",", 
                    UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
                 StyleBox["a", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["x", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "PreInt"], "[", 
                   StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["EC", 
                  RowBox[{"QuoDom", "[", 
                    RowBox[{"PreInt", ",", 
                    UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
                 StyleBox["a", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:565027904"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p565027904-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p565027904-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:565027904", "Proof|ID:565027904-1"},
 CellID->760024466],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p565027904-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->102735102]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(EC \[Element] Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(EC \[Element] Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Int"], "[", 
           RowBox[{
             UnderscriptBox["EC", 
              RowBox[{"QuoDom", "[", 
                RowBox[{"PreInt", ",", 
                  UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
             StyleBox["a", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Int"], "[", 
           RowBox[{
             UnderscriptBox["EC", 
              RowBox[{"QuoDom", "[", 
                RowBox[{"PreInt", ",", 
                  UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
             StyleBox["a", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:40757894"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p40757894-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p40757894-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:40757894", "Proof|ID:40757894-1"},
 CellID->35500606],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p40757894-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->168541484]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(0 \[Element] Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(0 \[Element] Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", "Int"], "[", 
         UnderscriptBox["0", "Int"], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", "Int"], "[", 
         UnderscriptBox["0", "Int"], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:350791961"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p350791961-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p350791961-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:350791961", "Proof|ID:350791961-1"},
 CellID->16051126],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p350791961-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->704723967]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(1 \[Element] Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(1 \[Element] Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", "Int"], "[", 
         UnderscriptBox["1", "Int"], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", "Int"], "[", 
         UnderscriptBox["1", "Int"], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:181240486"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p181240486-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p181240486-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:181240486", "Proof|ID:181240486-1"},
 CellID->157610485],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p181240486-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->95057389]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(1 \[NotEqual] 0 Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(1 \[NotEqual] 0 Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["1", "Int"], "\[NotEqual]", 
         UnderscriptBox["0", "Int"]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["1", "Int"], "\[NotEqual]", 
         UnderscriptBox["0", "Int"]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:364462666"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p364462666-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p364462666-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:364462666", "Proof|ID:364462666-1"},
 CellID->279790495],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p364462666-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->107741559]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemmas: ",
 Cell[BoxData[
  FormBox["+", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["-", TraditionalForm]]],
 " (\[Checkmark])"
}], "Subsubsection",
 CellChangeTimes->{{3.6612535894580693`*^9, 3.6612535975544834`*^9}, {
  3.6613374080213346`*^9, 3.6613374141521454`*^9}},
 CellID->74984569],

Cell["", "OpenEnvironment",
 CellID->234630122],

Cell[CellGroupData[{

Cell["Lemma: Properties of + in Int", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6612536111577077`*^9, 3.661253615915716*^9}},
 CellID->25494490],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "PreInt"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["EC", "\[ScriptCapitalQ]"], "[", "a", "]"}], 
     UnderscriptBox["+", "Int"], 
     RowBox[{
      UnderscriptBox["EC", "\[ScriptCapitalQ]"], "[", "b", "]"}]}], 
    "\[Equal]", 
    RowBox[{
     UnderscriptBox["EC", "\[ScriptCapitalQ]"], "[", 
     RowBox[{"a", 
      UnderscriptBox["+", "PreInt"], "b"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ EC Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:56650873", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661253633621747*^9, 3.6612536818726315`*^9}},
 CellTags->{
  "ID:56650873", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ EC Int", "IR"},
 CellID->56650873],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "Int"], 
   RowBox[{"[", 
    RowBox[{"x", ",", "y", ",", "z"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          13236461}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6612537109822826`*^9, 3.661253719000697*^9}, {
  3.6613168254918375`*^9, 3.6613168257102375`*^9}},
 CellTags->{
  "ID:13236461", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->13236461],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "Int"], 
  RowBox[{"[", 
   RowBox[{"x", 
    UnderscriptBox["+", "Int"], "y"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ closed Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:562490241", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661253722588703*^9, 3.661253732666321*^9}},
 CellTags->{
  "ID:562490241", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ closed Int", "IR"},
 CellID->562490241],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["+", "Int"], 
   UnderscriptBox["0", "Int"]}], "\[Equal]", "x"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(0 rneutral Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:624613835", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661316835569455*^9, 3.6613168501086807`*^9}},
 CellTags->{
  "ID:624613835", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 rneutral Int"},
 CellID->624613835],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["+", "Int"], "y"}], "\[Equal]", 
  RowBox[{"y", 
   UnderscriptBox["+", "Int"], "x"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ comm Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:115787764", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6613168558182907`*^9, 3.6613168652407064`*^9}},
 CellTags->{
  "ID:115787764", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ comm Int"},
 CellID->115787764],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"x", 
     UnderscriptBox["+", "Int"], "y"}], ")"}], 
   UnderscriptBox["+", "Int"], "z"}], "\[Equal]", 
  RowBox[{"x", 
   UnderscriptBox["+", "Int"], 
   RowBox[{"(", 
    RowBox[{"y", 
     UnderscriptBox["+", "Int"], "z"}], ")"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ assoc Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:11815767", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6613168678459115`*^9, 3.6613168744603233`*^9}, {
  3.661317365250186*^9, 3.661317375702204*^9}},
 CellTags->{
  "ID:11815767", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ assoc Int"},
 CellID->11815767],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"x", 
    UnderscriptBox["+", "Int"], "z"}], "\[Equal]", 
   RowBox[{"y", 
    UnderscriptBox["+", "Int"], "z"}]}], "\[Implies]", 
  RowBox[{"x", "\[Equal]", "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(+ rcanc Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:139539770", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661316888250747*^9, 3.6613168978447638`*^9}, {
  3.661317377948608*^9, 3.6613173827222166`*^9}},
 CellTags->{
  "ID:139539770", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ rcanc Int"},
 CellID->139539770],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->317123382]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->213410482],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ EC Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ EC Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["EC", 
                RowBox[{"QuoDom", "[", 
                  RowBox[{"PreInt", ",", 
                    UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}], 
             UnderscriptBox["+", "Int"], 
             RowBox[{
               UnderscriptBox["EC", 
                RowBox[{"QuoDom", "[", 
                  RowBox[{"PreInt", ",", 
                    UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["EC", 
              RowBox[{"QuoDom", "[", 
                RowBox[{"PreInt", ",", 
                  UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", "PreInt"], 
               StyleBox["b", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["EC", 
                RowBox[{"QuoDom", "[", 
                  RowBox[{"PreInt", ",", 
                    UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}], 
             UnderscriptBox["+", "Int"], 
             RowBox[{
               UnderscriptBox["EC", 
                RowBox[{"QuoDom", "[", 
                  RowBox[{"PreInt", ",", 
                    UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["EC", 
              RowBox[{"QuoDom", "[", 
                RowBox[{"PreInt", ",", 
                  UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["+", "PreInt"], 
               StyleBox["b", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:56650873"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p56650873-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p56650873-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:56650873", "Proof|ID:56650873-1"},
 CellID->106415148],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p56650873-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->80588368]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ closed Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ closed Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Int"], "[", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["+", "Int"], 
             StyleBox["y", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Int"], "[", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["+", "Int"], 
             StyleBox["y", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:562490241"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p562490241-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p562490241-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:562490241", "Proof|ID:562490241-1"},
 CellID->110816780],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p562490241-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->120666679]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(0 rneutral Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(0 rneutral Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["+", "Int"], 
             UnderscriptBox["0", "Int"]}], "\[Equal]", 
           StyleBox[
           "x", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["+", "Int"], 
             UnderscriptBox["0", "Int"]}], "\[Equal]", 
           StyleBox[
           "x", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:624613835"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p624613835-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p624613835-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:624613835", "Proof|ID:624613835-1"},
 CellID->23051911],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p624613835-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->26362014]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ comm Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ comm Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["+", "Int"], 
             StyleBox["y", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["y", "ExpressionVariable"], 
             UnderscriptBox["+", "Int"], 
             StyleBox[
             "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["+", "Int"], 
             StyleBox["y", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["y", "ExpressionVariable"], 
             UnderscriptBox["+", "Int"], 
             StyleBox[
             "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:115787764"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p115787764-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p115787764-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:115787764", "Proof|ID:115787764-1"},
 CellID->253640770],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p115787764-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->217136621]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ assoc Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ assoc Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["+", "Int"], 
               StyleBox["y", "ExpressionVariable"]}], 
             UnderscriptBox["+", "Int"], 
             StyleBox["z", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["+", "Int"], 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["+", "Int"], 
               StyleBox[
               "z", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["+", "Int"], 
               StyleBox["y", "ExpressionVariable"]}], 
             UnderscriptBox["+", "Int"], 
             StyleBox["z", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["+", "Int"], 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["+", "Int"], 
               StyleBox[
               "z", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:11815767"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p11815767-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p11815767-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:11815767", "Proof|ID:11815767-1"},
 CellID->44906707],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p11815767-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->31471341]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(+ rcanc Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(+ rcanc Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["+", "Int"], 
               StyleBox["z", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["+", "Int"], 
               StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["+", "Int"], 
               StyleBox["z", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["+", "Int"], 
               StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:139539770"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p139539770-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p139539770-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:139539770", "Proof|ID:139539770-1"},
 CellID->42190126],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p139539770-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->194323372]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->13262632],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Properties of ",
 Cell[BoxData[
  FormBox["-", TraditionalForm]]],
 " in Int"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.661321235884985*^9, 3.661321244074999*^9}},
 CellID->18951672],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "PreInt"], 
    RowBox[{"[", "a", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["-", "Int"], 
     RowBox[{
      UnderscriptBox["EC", "\[ScriptCapitalQ]"], "[", "a", "]"}]}], 
    "\[Equal]", 
    RowBox[{
     UnderscriptBox["EC", "\[ScriptCapitalQ]"], "[", 
     RowBox[{
      UnderscriptBox["-", "PreInt"], "a"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(- unary EC Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:331988502", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6613212579746237`*^9, 3.661321280485463*^9}},
 CellTags->{
  "ID:331988502", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- unary EC Int", "IR"},
 CellID->331988502],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "Int"], 
   RowBox[{"[", "x", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          138388913}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6613252078263717`*^9, 3.6613252139415827`*^9}},
 CellTags->{
  "ID:138388913", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->138388913],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "Int"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["-", "Int"], "x"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(- unary closed Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:789456798", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661325220321994*^9, 3.6613252294480095`*^9}},
 CellTags->{
  "ID:789456798", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- unary closed Int", "IR"},
 CellID->789456798],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["+", "Int"], 
   RowBox[{"(", 
    RowBox[{
     UnderscriptBox["-", "Int"], "x"}], ")"}]}], "\[Equal]", 
  UnderscriptBox["0", "Int"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(- inverse Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:561244613", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661325233504017*^9, 3.661325246062039*^9}},
 CellTags->{
  "ID:561244613", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- inverse Int"},
 CellID->561244613],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->183233282]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->412087319],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(- unary EC Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(- unary EC Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["Minus", "Int"], "[", 
             RowBox[{
               UnderscriptBox["EC", 
                RowBox[{"QuoDom", "[", 
                  RowBox[{"PreInt", ",", 
                    UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["EC", 
              RowBox[{"QuoDom", "[", 
                RowBox[{"PreInt", ",", 
                  UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
             RowBox[{
               UnderscriptBox["Minus", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["Minus", "Int"], "[", 
             RowBox[{
               UnderscriptBox["EC", 
                RowBox[{"QuoDom", "[", 
                  RowBox[{"PreInt", ",", 
                    UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["EC", 
              RowBox[{"QuoDom", "[", 
                RowBox[{"PreInt", ",", 
                  UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
             RowBox[{
               UnderscriptBox["Minus", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:331988502"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p331988502-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p331988502-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:331988502", "Proof|ID:331988502-1"},
 CellID->217485270],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p331988502-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->999928246]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(- unary closed Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(- unary closed Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Int"], "[", 
           RowBox[{
             UnderscriptBox["Minus", "Int"], "[", 
             StyleBox["x", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Int"], "[", 
           RowBox[{
             UnderscriptBox["Minus", "Int"], "[", 
             StyleBox["x", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:789456798"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p789456798-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p789456798-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:789456798", "Proof|ID:789456798-1"},
 CellID->617130052],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p789456798-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->194480606]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(- inverse Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(- inverse Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["+", "Int"], 
             RowBox[{
               UnderscriptBox["Minus", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}], "\[Equal]", 
           UnderscriptBox["0", "Int"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["+", "Int"], 
             RowBox[{
               UnderscriptBox["Minus", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}], "\[Equal]", 
           UnderscriptBox["0", "Int"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:561244613"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p561244613-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p561244613-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:561244613", "Proof|ID:561244613-1"},
 CellID->394996459],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p561244613-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->983259]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemmas: ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " (\[Checkmark])"
}], "Subsubsection",
 CellChangeTimes->{{3.661321306615509*^9, 3.6613213107651167`*^9}, {
  3.6613341831566634`*^9, 3.6613341873686705`*^9}},
 CellID->716281922],

Cell["", "OpenEnvironment",
 CellID->101122110],

Cell[CellGroupData[{

Cell[TextData[{
 "Lemma: Properties of ",
 Cell[BoxData[
  FormBox["*", TraditionalForm]]],
 " in Int"
}], "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6613213182531295`*^9, 3.661321328127947*^9}},
 CellID->317349246],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "PreInt"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["EC", "\[ScriptCapitalQ]"], "[", "a", "]"}], 
     UnderscriptBox["*", "Int"], 
     RowBox[{
      UnderscriptBox["EC", "\[ScriptCapitalQ]"], "[", "b", "]"}]}], 
    "\[Equal]", 
    RowBox[{
     UnderscriptBox["EC", "\[ScriptCapitalQ]"], "[", 
     RowBox[{"a", 
      UnderscriptBox["*", "PreInt"], "b"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* EC Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:947607354", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661321334477158*^9, 3.661321347471981*^9}},
 CellTags->{
  "ID:947607354", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* EC Int", "IR"},
 CellID->947607354],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "Int"], 
   RowBox[{"[", 
    RowBox[{"x", ",", "y", ",", "z"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          105543058}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.661321357440398*^9},
 CellTags->{
  "ID:105543058", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->105543058],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "Int"], 
  RowBox[{"[", 
   RowBox[{"x", 
    UnderscriptBox["*", "Int"], "y"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* closed Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:696766377", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6613255071596975`*^9, 3.6613255162389135`*^9}},
 CellTags->{
  "ID:696766377", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* closed Int", "IR"},
 CellID->696766377],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["*", "Int"], 
   UnderscriptBox["0", "Int"]}], "\[Equal]", 
  UnderscriptBox["0", "Int"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(0 rannihilator Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:202378771", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661325520872122*^9, 3.6613255309497395`*^9}},
 CellTags->{
  "ID:202378771", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 rannihilator Int"},
 CellID->202378771],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["*", "Int"], 
   UnderscriptBox["1", "Int"]}], "\[Equal]", "x"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1 rneutral Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:689609352", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661325534210145*^9, 3.6613255436169615`*^9}},
 CellTags->{
  "ID:689609352", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 rneutral Int"},
 CellID->689609352],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["*", "Int"], "y"}], "\[Equal]", 
  RowBox[{"y", 
   UnderscriptBox["*", "Int"], "x"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* comm Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:196343023", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6613255474077682`*^9, 3.661325556330984*^9}},
 CellTags->{
  "ID:196343023", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* comm Int"},
 CellID->196343023],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", 
     UnderscriptBox["*", "Int"], "y"}], "\[Equal]", 
    UnderscriptBox["0", "Int"]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", "\[Equal]", 
     UnderscriptBox["0", "Int"]}], "\[Or]", 
    RowBox[{"y", "\[Equal]", 
     UnderscriptBox["0", "Int"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* zdf Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:207834126", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6613213634620085`*^9, 3.6613213865812497`*^9}},
 CellTags->{
  "ID:207834126", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* zdf Int"},
 CellID->207834126],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"x", 
     UnderscriptBox["*", "Int"], "y"}], ")"}], 
   UnderscriptBox["*", "Int"], "z"}], "\[Equal]", 
  RowBox[{"x", 
   UnderscriptBox["*", "Int"], 
   RowBox[{"(", 
    RowBox[{"y", 
     UnderscriptBox["*", "Int"], "z"}], ")"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* assoc Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:10196611", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6613255598097897`*^9, 3.6613255800898256`*^9}},
 CellTags->{
  "ID:10196611", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* assoc Int"},
 CellID->10196611],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["*", "Int"], 
   RowBox[{"(", 
    RowBox[{"y", 
     UnderscriptBox["+", "Int"], "z"}], ")"}]}], "\[Equal]", 
  RowBox[{
   RowBox[{"x", 
    UnderscriptBox["*", "Int"], "y"}], 
   UnderscriptBox["+", "Int"], 
   RowBox[{"x", 
    UnderscriptBox["*", "Int"], "z"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(* distr + Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:777222529", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6613255838338323`*^9, 3.6613256094958773`*^9}},
 CellTags->{
  "ID:777222529", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "* distr + Int"},
 CellID->777222529],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->35189982]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->566077695],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* EC Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* EC Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["EC", 
                RowBox[{"QuoDom", "[", 
                  RowBox[{"PreInt", ",", 
                    UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}], 
             UnderscriptBox["*", "Int"], 
             RowBox[{
               UnderscriptBox["EC", 
                RowBox[{"QuoDom", "[", 
                  RowBox[{"PreInt", ",", 
                    UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["EC", 
              RowBox[{"QuoDom", "[", 
                RowBox[{"PreInt", ",", 
                  UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["*", "PreInt"], 
               StyleBox["b", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["EC", 
                RowBox[{"QuoDom", "[", 
                  RowBox[{"PreInt", ",", 
                    UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}], 
             UnderscriptBox["*", "Int"], 
             RowBox[{
               UnderscriptBox["EC", 
                RowBox[{"QuoDom", "[", 
                  RowBox[{"PreInt", ",", 
                    UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["EC", 
              RowBox[{"QuoDom", "[", 
                RowBox[{"PreInt", ",", 
                  UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
             RowBox[{
               StyleBox["a", "ExpressionVariable"], 
               UnderscriptBox["*", "PreInt"], 
               StyleBox["b", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:947607354"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p947607354-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p947607354-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:947607354", "Proof|ID:947607354-1"},
 CellID->558776955],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p947607354-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->522627315]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* closed Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* closed Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Int"], "[", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["*", "Int"], 
             StyleBox["y", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "Int"], "[", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["*", "Int"], 
             StyleBox["y", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:696766377"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p696766377-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p696766377-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:696766377", "Proof|ID:696766377-1"},
 CellID->81854393],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p696766377-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->106710227]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(0 rannihilator Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(0 rannihilator Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["*", "Int"], 
             UnderscriptBox["0", "Int"]}], "\[Equal]", 
           UnderscriptBox["0", "Int"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["*", "Int"], 
             UnderscriptBox["0", "Int"]}], "\[Equal]", 
           UnderscriptBox["0", "Int"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:202378771"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p202378771-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p202378771-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:202378771", "Proof|ID:202378771-1"},
 CellID->465404293],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p202378771-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->561652051]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(1 rneutral Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(1 rneutral Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["*", "Int"], 
             UnderscriptBox["1", "Int"]}], "\[Equal]", 
           StyleBox[
           "x", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["*", "Int"], 
             UnderscriptBox["1", "Int"]}], "\[Equal]", 
           StyleBox[
           "x", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:689609352"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p689609352-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p689609352-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:689609352", "Proof|ID:689609352-1"},
 CellID->485464469],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p689609352-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->294236516]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* comm Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* comm Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["*", "Int"], 
             StyleBox["y", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["y", "ExpressionVariable"], 
             UnderscriptBox["*", "Int"], 
             StyleBox[
             "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["*", "Int"], 
             StyleBox["y", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["y", "ExpressionVariable"], 
             UnderscriptBox["*", "Int"], 
             StyleBox[
             "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:196343023"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p196343023-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p196343023-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:196343023", "Proof|ID:196343023-1"},
 CellID->43558645],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p196343023-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->308643022]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* zdf Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* zdf Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["*", "Int"], 
               StyleBox["y", "ExpressionVariable"]}], "\[Equal]", 
             UnderscriptBox["0", "Int"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Equal]", 
               UnderscriptBox["0", "Int"]}], "\[Or]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Equal]", 
               UnderscriptBox["0", "Int"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["*", "Int"], 
               StyleBox["y", "ExpressionVariable"]}], "\[Equal]", 
             UnderscriptBox["0", "Int"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Equal]", 
               UnderscriptBox["0", "Int"]}], "\[Or]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], "\[Equal]", 
               UnderscriptBox[
               "0", "Int"]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:207834126"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p207834126-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p207834126-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:207834126", "Proof|ID:207834126-1"},
 CellID->5339075],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p207834126-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->181354457]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* assoc Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* assoc Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["*", "Int"], 
               StyleBox["y", "ExpressionVariable"]}], 
             UnderscriptBox["*", "Int"], 
             StyleBox["z", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["*", "Int"], 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["*", "Int"], 
               StyleBox[
               "z", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["*", "Int"], 
               StyleBox["y", "ExpressionVariable"]}], 
             UnderscriptBox["*", "Int"], 
             StyleBox["z", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["*", "Int"], 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["*", "Int"], 
               StyleBox[
               "z", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:10196611"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p10196611-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p10196611-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:10196611", "Proof|ID:10196611-1"},
 CellID->268794065],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p10196611-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->283564566]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(* distr + Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(* distr + Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["*", "Int"], 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["+", "Int"], 
               StyleBox["z", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["*", "Int"], 
               StyleBox["y", "ExpressionVariable"]}], 
             UnderscriptBox["+", "Int"], 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["*", "Int"], 
               StyleBox[
               "z", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["*", "Int"], 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["+", "Int"], 
               StyleBox["z", "ExpressionVariable"]}]}], "\[Equal]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["*", "Int"], 
               StyleBox["y", "ExpressionVariable"]}], 
             UnderscriptBox["+", "Int"], 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["*", "Int"], 
               StyleBox[
               "z", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:777222529"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p777222529-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p777222529-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:777222529", "Proof|ID:777222529-1"},
 CellID->343265473],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p777222529-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->885039319]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Lemmas: \[GreaterEqual] (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.661253289563143*^9, 3.6612532922463474`*^9}, {
  3.661337173861923*^9, 3.6613371781831303`*^9}},
 CellID->389529963],

Cell["", "OpenEnvironment",
 CellID->77983804],

Cell[CellGroupData[{

Cell["Lemma: Properties of \[GreaterEqual] in Int", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.661253297721957*^9, 3.661253303337967*^9}},
 CellID->210469497],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "PreInt"], 
    RowBox[{"[", 
     RowBox[{"a", ",", "b"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["EC", "\[ScriptCapitalQ]"], "[", "a", "]"}], 
      UnderscriptBox["\[GreaterEqual]", "Int"], 
      RowBox[{
       UnderscriptBox["EC", "\[ScriptCapitalQ]"], "[", "b", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"a", 
      UnderscriptBox["\[GreaterEqual]", "PreInt"], "b"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] EC Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:785793765", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661321439746143*^9, 3.6613214844870214`*^9}},
 CellTags->{
  "ID:785793765", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] EC Int", "IR"},
 CellID->785793765],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["1", "Int"], 
  UnderscriptBox["\[GreaterEqual]", "Int"], 
  UnderscriptBox["0", "Int"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(1 \[GreaterEqual] 0 Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:72386745", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661325701723239*^9, 3.6613257114888563`*^9}},
 CellTags->{
  "ID:72386745", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 \[GreaterEqual] 0 Int"},
 CellID->72386745],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "Int"], 
   RowBox[{"[", 
    RowBox[{"x", ",", "y", ",", "z"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
          100510373}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.66132150643626*^9, 3.6613215135498724`*^9}},
 CellTags->{
  "ID:100510373", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->100510373],

Cell[BoxData[
 RowBox[{"x", 
  UnderscriptBox["\[GreaterEqual]", "Int"], "x"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] reflexive Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:267773274", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6613256840484085`*^9, 3.6613256875116143`*^9}},
 CellTags->{
  "ID:267773274", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] reflexive Int"},
 CellID->267773274],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", 
     UnderscriptBox["\[GreaterEqual]", "Int"], "y"}], "\[And]", 
    RowBox[{"y", 
     UnderscriptBox["\[GreaterEqual]", "Int"], "x"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"x", "\[Equal]", "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] antisymmetric Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:69005952", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6613219234259925`*^9, 3.661321937762418*^9}},
 CellTags->{
  "ID:69005952", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] antisymmetric Int"},
 CellID->69005952],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", 
     UnderscriptBox["\[GreaterEqual]", "Int"], "y"}], "\[And]", 
    RowBox[{"y", 
     UnderscriptBox["\[GreaterEqual]", "Int"], "z"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"x", 
   UnderscriptBox["\[GreaterEqual]", "Int"], 
   "z"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] transitive Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:275182510", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6613258076318254`*^9, 3.6613258248542557`*^9}},
 CellTags->{
  "ID:275182510", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] transitive Int"},
 CellID->275182510],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["\[GreaterEqual]", "Int"], "y"}], "\[Or]", 
  RowBox[{"y", 
   UnderscriptBox["\[GreaterEqual]", "Int"], 
   "x"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] total Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:176048001", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6613257258408813`*^9, 3.6613257361993*^9}},
 CellTags->{
  "ID:176048001", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] total Int"},
 CellID->176048001],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", 
     UnderscriptBox["\[GreaterEqual]", "Int"], 
     UnderscriptBox["0", "Int"]}], "\[And]", 
    RowBox[{"x", "\[NotEqual]", 
     UnderscriptBox["0", "Int"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"x", 
   UnderscriptBox["\[GreaterEqual]", "Int"], 
   UnderscriptBox["1", "Int"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] discrete Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:566590035", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6613220014417295`*^9, 3.6613220275249753`*^9}},
 CellTags->{
  "ID:566590035", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] discrete Int"},
 CellID->566590035],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["\[GreaterEqual]", "Int"], "y"}], "\[Implies]", 
  RowBox[{
   RowBox[{"x", 
    UnderscriptBox["+", "Int"], "z"}], 
   UnderscriptBox["\[GreaterEqual]", "Int"], 
   RowBox[{"y", 
    UnderscriptBox["+", "Int"], "z"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] + Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:262876021", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6613258304858656`*^9, 3.6613258681131315`*^9}},
 CellTags->{
  "ID:262876021", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] + Int"},
 CellID->262876021],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"x", 
     UnderscriptBox["\[GreaterEqual]", "Int"], 
     UnderscriptBox["0", "Int"]}], "\[And]", 
    RowBox[{"y", 
     UnderscriptBox["\[GreaterEqual]", "Int"], 
     UnderscriptBox["0", "Int"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"x", 
    UnderscriptBox["*", "Int"], "y"}], 
   UnderscriptBox["\[GreaterEqual]", "Int"], 
   UnderscriptBox["0", "Int"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[GreaterEqual] * Int)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:34443687", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6613257592405405`*^9, 3.6613257866185884`*^9}},
 CellTags->{
  "ID:34443687", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] * Int"},
 CellID->34443687],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->341695448]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->127316464],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] EC Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] EC Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["EC", 
                RowBox[{"QuoDom", "[", 
                  RowBox[{"PreInt", ",", 
                    UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[GreaterEqual]", "Int"], 
             RowBox[{
               UnderscriptBox["EC", 
                RowBox[{"QuoDom", "[", 
                  RowBox[{"PreInt", ",", 
                    UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "PreInt"], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["EC", 
                RowBox[{"QuoDom", "[", 
                  RowBox[{"PreInt", ",", 
                    UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
               StyleBox["a", "ExpressionVariable"], "]"}], 
             UnderscriptBox["\[GreaterEqual]", "Int"], 
             RowBox[{
               UnderscriptBox["EC", 
                RowBox[{"QuoDom", "[", 
                  RowBox[{"PreInt", ",", 
                    UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
               StyleBox["b", "ExpressionVariable"], "]"}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["a", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "PreInt"], 
             StyleBox[
             "b", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:785793765"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p785793765-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p785793765-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:785793765", "Proof|ID:785793765-1"},
 CellID->518056285],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p785793765-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->605637846]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(1 \[GreaterEqual] 0 Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(1 \[GreaterEqual] 0 Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["1", "Int"], 
         UnderscriptBox["\[GreaterEqual]", "Int"], 
         UnderscriptBox["0", "Int"]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["1", "Int"], 
         UnderscriptBox["\[GreaterEqual]", "Int"], 
         UnderscriptBox["0", "Int"]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:72386745"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p72386745-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p72386745-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:72386745", "Proof|ID:72386745-1"},
 CellID->72615069],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p72386745-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->22111013]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] reflexive Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] reflexive Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["x", "ExpressionVariable"], 
           UnderscriptBox["\[GreaterEqual]", "Int"], 
           StyleBox[
           "x", "ExpressionVariable"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           StyleBox["x", "ExpressionVariable"], 
           UnderscriptBox["\[GreaterEqual]", "Int"], 
           StyleBox[
           "x", "ExpressionVariable"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:267773274"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p267773274-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p267773274-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:267773274", "Proof|ID:267773274-1"},
 CellID->50495462],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p267773274-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->442279872]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] antisymmetric Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] antisymmetric Int)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Int"], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Int"], 
               StyleBox["x", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Int"], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Int"], 
               StyleBox["x", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:69005952"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p69005952-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p69005952-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:69005952", "Proof|ID:69005952-1"},
 CellID->49834528],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p69005952-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->128346525]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] transitive Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] transitive Int)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Int"], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Int"], 
               StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Int"], 
             StyleBox[
             "z", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Int"], 
               StyleBox["y", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Int"], 
               StyleBox["z", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Int"], 
             StyleBox[
             "z", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:275182510"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p275182510-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p275182510-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:275182510", "Proof|ID:275182510-1"},
 CellID->94179691],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p275182510-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->25059238]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] total Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] total Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Int"], 
             StyleBox["y", "ExpressionVariable"]}], "\[Or]", 
           RowBox[{
             StyleBox["y", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Int"], 
             StyleBox[
             "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Int"], 
             StyleBox["y", "ExpressionVariable"]}], "\[Or]", 
           RowBox[{
             StyleBox["y", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Int"], 
             StyleBox[
             "x", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:176048001"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p176048001-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p176048001-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:176048001", "Proof|ID:176048001-1"},
 CellID->320141850],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p176048001-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->123691091]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] discrete Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] discrete Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Int"], 
               UnderscriptBox["0", "Int"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "Int"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Int"], 
             UnderscriptBox["1", "Int"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Int"], 
               UnderscriptBox["0", "Int"]}], "\[And]", 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", "Int"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Int"], 
             UnderscriptBox["1", "Int"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:566590035"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p566590035-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p566590035-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:566590035", "Proof|ID:566590035-1"},
 CellID->88604489],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p566590035-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->339162341]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] + Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] + Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Int"], 
             StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["+", "Int"], 
               StyleBox["z", "ExpressionVariable"]}], 
             UnderscriptBox["\[GreaterEqual]", "Int"], 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["+", "Int"], 
               StyleBox[
               "z", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Int"], 
             StyleBox["y", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["+", "Int"], 
               StyleBox["z", "ExpressionVariable"]}], 
             UnderscriptBox["\[GreaterEqual]", "Int"], 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["+", "Int"], 
               StyleBox[
               "z", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:262876021"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p262876021-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p262876021-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:262876021", "Proof|ID:262876021-1"},
 CellID->490373417],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p262876021-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->399550044]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[GreaterEqual] * Int)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[GreaterEqual] * Int)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Int"], 
               UnderscriptBox["0", "Int"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Int"], 
               UnderscriptBox["0", "Int"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["*", "Int"], 
               StyleBox["y", "ExpressionVariable"]}], 
             UnderscriptBox["\[GreaterEqual]", "Int"], 
             UnderscriptBox["0", "Int"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Int"], 
               UnderscriptBox["0", "Int"]}], "\[And]", 
             RowBox[{
               StyleBox["y", "ExpressionVariable"], 
               UnderscriptBox["\[GreaterEqual]", "Int"], 
               UnderscriptBox["0", "Int"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               StyleBox["x", "ExpressionVariable"], 
               UnderscriptBox["*", "Int"], 
               StyleBox["y", "ExpressionVariable"]}], 
             UnderscriptBox["\[GreaterEqual]", "Int"], 
             UnderscriptBox["0", "Int"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:34443687"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p34443687-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p34443687-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:34443687", "Proof|ID:34443687-1"},
 CellID->161887690],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p34443687-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->247568277]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->541251984],

Cell[CellGroupData[{

Cell["Lemma: Induction on Int", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6613349135051465`*^9, 3.6613349175767536`*^9}},
 CellID->7001469],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "Int"], 
     RowBox[{"[", "x", "]"}]}]], 
   RowBox[{"x", 
    UnderscriptBox["\[GreaterEqual]", "Int"], 
    UnderscriptBox["0", "Int"]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "Nat"], 
      RowBox[{"[", "n", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"x", "\[Equal]", 
      RowBox[{
       UnderscriptBox["EC", "\[ScriptCapitalQ]"], "[", 
       RowBox[{"Pair", "[", 
        RowBox[{"n", ",", 
         UnderscriptBox["0", "Nat"]}], "]"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Nat \[SucceedsTilde] Int+)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:134029858", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.661253195713378*^9, 3.6612532781907225`*^9}, {
   3.661253312105182*^9, 3.6612533452552404`*^9}, 3.6613215172782793`*^9, {
   3.6613217228876405`*^9, 3.6613217440100775`*^9}, {3.661334929323574*^9, 
   3.661334942349597*^9}},
 CellTags->{
  "ID:134029858", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat \[SucceedsTilde] Int+"},
 CellID->134029858],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "P"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", "Int"], 
             RowBox[{"[", "x", "]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{
              UnderscriptBox["0", "Int"], 
              UnderscriptBox["\[GreaterEqual]", "Int"], "x"}], "\[Implies]", 
             RowBox[{"P", "[", "x", "]"}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", "Int"], 
             RowBox[{"[", "x", "]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{
               RowBox[{"x", 
                UnderscriptBox["\[GreaterEqual]", "Int"], 
                UnderscriptBox["0", "Int"]}], "\[And]", 
               RowBox[{"P", "[", "x", "]"}]}], 
              TagBox[")",
               "AutoParentheses"]}], "\[Implies]", 
             RowBox[{"P", "[", 
              RowBox[{"x", 
               UnderscriptBox["+", "Int"], 
               UnderscriptBox["1", "Int"]}], "]"}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "Int"], 
        RowBox[{"[", "x", "]"}]}]], 
      RowBox[{"P", "[", "x", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Int+ induction)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:181242736", 
            "Source:/home/amaletzk/Theorema \
Notebooks/ElementaryTheories/Numbers.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6613349585736256`*^9, 3.661335083125244*^9}, {
  3.661336348134071*^9, 3.661336349382073*^9}},
 CellTags->{
  "ID:181242736", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Int+ induction"},
 CellID->181242736],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->331282669]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->16861641],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Nat \[SucceedsTilde] Int+)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Nat \[SucceedsTilde] Int+)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Int"], 
             UnderscriptBox["0", "Int"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "Nat"], "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["EC", 
                  RowBox[{"QuoDom", "[", 
                    RowBox[{"PreInt", ",", 
                    UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
                 RowBox[{"Pair", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    UnderscriptBox["0", "Nat"]}], "]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "Int"], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["x", "ExpressionVariable"], 
             UnderscriptBox["\[GreaterEqual]", "Int"], 
             UnderscriptBox["0", "Int"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "Nat"], "[", 
                   StyleBox["n", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["x", "ExpressionVariable"], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["EC", 
                  RowBox[{"QuoDom", "[", 
                    RowBox[{"PreInt", ",", 
                    UnderscriptBox["equiv", "PreInt"]}], "]"}]], "[", 
                 RowBox[{"Pair", "[", 
                   RowBox[{
                    StyleBox["n", "ExpressionVariable"], ",", 
                    UnderscriptBox["0", "Nat"]}], "]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:134029858"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p134029858-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p134029858-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:134029858", "Proof|ID:134029858-1"},
 CellID->379729521],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p134029858-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->330204082]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Int+ induction)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Int+ induction)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "Int"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["0", "Int"], 
                   UnderscriptBox["\[GreaterEqual]", "Int"], 
                   StyleBox["x", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "Int"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], 
                    UnderscriptBox["\[GreaterEqual]", "Int"], 
                    UnderscriptBox["0", "Int"]}], "\[And]", 
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], 
                    UnderscriptBox["+", "Int"], 
                    UnderscriptBox["1", "Int"]}], "]"}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "Int"], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               StyleBox["x", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "Int"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["0", "Int"], 
                   UnderscriptBox["\[GreaterEqual]", "Int"], 
                   StyleBox["x", "ExpressionVariable"]}], "\[Implies]", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}]}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "Int"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], 
                    UnderscriptBox["\[GreaterEqual]", "Int"], 
                    UnderscriptBox["0", "Int"]}], "\[And]", 
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    StyleBox["x", "ExpressionVariable"], "]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["x", "ExpressionVariable"], 
                    UnderscriptBox["+", "Int"], 
                    UnderscriptBox["1", "Int"]}], "]"}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "Int"], "[", 
                   StyleBox["x", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               StyleBox["x", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:181242736"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p181242736-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p181242736-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:181242736", "Proof|ID:181242736-1"},
 CellID->625873853],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p181242736-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->165801038]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]]
}, Open  ]]
},
WindowSize->{1106, 758},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
CounterAssignments->{{"TheoremaFormulaCounter" -> 4}},
Magnification->1.,
FrontEndVersion->"9.0 for Linux x86 (64-bit) (February 7, 2013)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Theorema"}, "TheoremaCore.nb"]]], 
   Cell["Modify the definitions below to change the default appearance of all \
cells in a given style. \nMake modifications to any definition using commands \
in the Format menu.", "Text"], 
   Cell[
    CellGroupData[{
      Cell["Notebook Options Settings", "Section"], 
      Cell[
      "The options defined for the style below will be used at the Notebook \
level.", "Text"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Notebook"], InputAliases -> {"and" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[And]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "or" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", "\[Or]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "not" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Not]", "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "impl" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[Implies]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[DoubleLongLeftRightArrow]", 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                "\[SelectionPlaceholder]", "\[Equal]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox[
                  
                  RowBox[{
                   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                  "a:=b"], "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", ":=", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "andn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[And]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "orn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Or]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "equivn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[DoubleLeftRightArrow]", "\[Piecewise]", 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "cdist" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Piecewise]", 
                 
                 GridBox[{{
                   "\[SelectionPlaceholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}, {
                   "\[Placeholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "far" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exr" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "farc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exrc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "let" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["let", 
                  RowBox[{"\[Placeholder]", "=", "\[Placeholder]"}]], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "app" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293a", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "prep" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293b", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "join" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22c8", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "elemT" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22ff", Identity, SyntaxForm -> "a+b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "subop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "supop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SuperscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "subsupop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubsuperscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "oop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "uoop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               UnderoverscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "osop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox[
                OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
                "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "op" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               TagBox["\[SelectionPlaceholder]", Identity], 
               TagBox[")", "AutoParentheses"]}], "[]" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:e114", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:e115", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "()" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:fd3e", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:fd3f", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "<>" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27e8", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27e9", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "{}" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:e117", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:e118", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "[[]]" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27e6", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27e7", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "(())" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:2e28", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:2e29", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "<<>>" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27ea", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27eb", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "{{}}" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:2983", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:2984", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "(" -> 
            TagBox["(", "AutoParentheses"], ")" -> 
            TagBox[")", "AutoParentheses"]}, Background -> 
          RGBColor[0.957, 0.96, 0.97]], 
         Cell[
          StyleData["Notebook", "Printout"], Background -> None]}, Open]]}, 
     Open]], 
   Cell[
    CellGroupData[{
      Cell["Style Environment Names", "Section"], 
      Cell[
       StyleData[All, "Working"], DockedCells -> {
         Cell[
         "Theorema 2.0", "SBO", FontTracking -> 8, 
          CellMargins -> {{0, 0}, {0, 0}}, 
          CellFrameMargins -> {{0, 50}, {0, 0}}, FontSize -> 28, FontSlant -> 
          "Plain", FontColor -> RGBColor[0.29, 0.46, 0.6], TextAlignment -> 
          Right, Background -> RGBColor[0.36, 0.54, 0.69]]}], 
      Cell[
       StyleData[
       All, "Layout", StyleDefinitions -> StyleData[All, "Working"]], 
       MenuSortingValue -> 1200]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Styles for Title and Section Cells", "Section"], 
      Cell[
       CellGroupData[{
         Cell["Title", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], CellMargins -> {{27, 27}, {10, 30}}, 
             TextAlignment -> Center, MenuSortingValue -> 1100, 
             MenuCommandKey -> "1", FontFamily -> "Bitstream Charter", 
             FontSize -> 36, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Title", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{6, 10}, {5, 30}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], CellMargins -> {{27, 27}, {20, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1150, FontFamily -> 
             "Bitstream Charter", FontSize -> 24, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{6, 10}, {5, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], CellMargins -> {{27, 27}, {8, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Author"], CellMargins -> {{27, 27}, {18, 20}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Sectioning", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], ShowGroupOpener -> True, MenuSortingValue -> 
             1250, MenuCommandKey -> "4", FontFamily -> "Bitstream Charter", 
             FontSize -> 20, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Section", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{2, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {8, 12}}, 
             MenuSortingValue -> 1300, MenuCommandKey -> "5", FontFamily -> 
             "Bitstream Charter", FontSize -> 14, FontWeight -> "Bold", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Subsection", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{9, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {2, 10}}, 
             MenuSortingValue -> 1350, MenuCommandKey -> "6", FontFamily -> 
             "Bitstream Charter", FontSize -> 12, FontWeight -> "Bold", 
             FontSlant -> "Italic", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsection", "Printout"], ShowGroupOpener -> False,
              CellMargins -> {{35, 0}, {7, 14}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsubsection"], CellDingbat -> None, 
             ShowGroupOpener -> True, 
             CellMargins -> {{27, Inherited}, {Inherited, Inherited}}, 
             FontFamily -> "Bitstream Charter", FontSize -> 11, FontWeight -> 
             "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsubsection", "Printout"], 
             CellMargins -> {{50, 0}, {7, 14}}]}, Closed]]}, Open]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Styles for Publication\[Hyphen]ready Documents", "Section"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Author"], CellMargins -> {{27, 30}, {20, 5}}, 
          MenuSortingValue -> 1900, FontFamily -> "Helvetica", FontSize -> 14,
           FontSlant -> "Italic"], 
         Cell[
          StyleData["Author", "Printout"], CellMargins -> {{6, 10}, {0, 10}}, 
          FontSize -> 10]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Abstract"], CellFrame -> {{0, 0}, {0.5, 0.5}}, 
          CellMargins -> {{27, 140}, {20, 20}}, LineSpacing -> {0.9, 0}, 
          MenuSortingValue -> 2050, FontFamily -> "Times", FontSize -> 12], 
         Cell[
          StyleData["Abstract", "Printout"], Hyphenation -> True, 
          LineSpacing -> {1, 2}, FontSize -> 10]}, Closed]]}, Closed]], 
   Cell[
    CellGroupData[{
      Cell["Text", "Section"], 
      Cell[
       StyleData["Text"], CellMargins -> {{27, 27}, {Inherited, Inherited}}, 
       TextJustification -> 1., ParagraphIndent -> 15, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["EnvironmentText"], ShowCellBracket -> Automatic, 
       CellMargins -> {{82, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextJustification -> 1., ParagraphSpacing -> {0.3, 0.}, 
       MenuSortingValue -> 1480, MenuCommandKey -> "T", FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Item"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "Item", 
          CounterAssignments -> {{"Subitem", 0}, {"Subsubitem", 0}}, 
          MenuSortingValue -> 1600, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Item", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemParagraph"], CellMargins -> {{81, 10}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          DefaultNewCellStyle -> "Item", MenuSortingValue -> 1600, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemParagraph", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "Subitem", 
          CounterAssignments -> {{"Subsubitem", 0}}, MenuSortingValue -> 1610,
           FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subitem", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemParagraph"], CellMargins -> {{105, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subitem", 
          MenuSortingValue -> 1610, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemParagraph", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subsubitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1620, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subsubitem", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemParagraph"], 
          CellMargins -> {{129, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1625, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemParagraph", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "ItemNumbered", 
          CounterAssignments -> {{"SubitemNumbered", 0}, {
            "SubsubitemNumbered", 0}}, MenuSortingValue -> 1630, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemNumbered", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "SubitemNumbered", 
          CounterAssignments -> {{"SubsubitemNumbered", 0}}, MenuSortingValue -> 
          1640, FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemNumbered", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], ".", 
              CounterBox["SubsubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> 
          "SubitemNumbered", CounterIncrements -> "SubsubitemNumbered", 
          MenuSortingValue -> 1645, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemNumbered", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Theorema", "Section"], 
      Cell[
       StyleData["GlobalDeclaration"], CellFrame -> {{0, 2}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 3}, {2, 2}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       InitializationCell -> True, CellFrameMargins -> 10, CellFrameColor -> 
       RGBColor[0.73, 0.65, 0.61], CellFrameLabelMargins -> {{33, 7}, {1, 1}},
        ShowAutoStyles -> False, MenuSortingValue -> 1441, MenuCommandKey -> 
       "G", FontSize -> 14], 
      Cell[
       StyleData["OpenEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {-4, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       StyleData["EnvironmentHeader"], 
       CellFrame -> {{False, False}, {False, False}}, 
       CellMargins -> {{76, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 6, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       ShowAutoStyles -> False, MenuSortingValue -> 1450, FontSize -> 14, 
       FontWeight -> "Bold", FontVariations -> {"CapsType" -> "SmallCaps"}, 
       Background -> RGBColor[0.89, 0.8, 0.69]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FormalTextInputFormula"], ShowCellBracket -> Automatic, 
          CellMargins -> {{96, 10}, {2, 2}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
          InitializationCell -> True, CellFrameLabelMargins -> 53, 
          ShowAutoStyles -> False, LimitsPositioningTokens -> {}, 
          MenuSortingValue -> 1470, MenuCommandKey -> "F", FontWeight -> 
          "Bold", FontColor -> RGBColor[0.25, 0.19, 0.2]], 
         Cell[
          StyleData["FormalTextInputFormula", "Layout"], Editable -> False], 
         Cell[
          StyleData["FormalTextInputFormula", "Presentation"], Editable -> 
          False]}, Open]], 
      Cell[
       StyleData["EndEnvironmentMarker"], 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextAlignment -> Right, MenuSortingValue -> 1480], 
      Cell[
       StyleData["CloseEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {False, True}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {Inherited, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Computation"], CellFrame -> {{1.5, 0}, {0, 0}}, 
          ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 7}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
          CellFrameMargins -> {{8, 8}, {3, 4}}, CellFrameColor -> 
          RGBColor[0.486, 0.33, 0.32], ShowAutoStyles -> False, 
          MenuSortingValue -> 1510, FontWeight -> "Bold", FontColor -> 
          RGBColor[0.23, 0.25, 0.34], Background -> 
          RGBColor[0.88, 0.88, 0.92]], 
         Cell[
          StyleData["Computation", "Layout"], Editable -> False], 
         Cell[
          StyleData["Computation", "Presentation"], Editable -> False]}, 
        Open]], 
      Cell[
       StyleData["ComputationResult"], CellFrame -> {{1.5, 0}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
       CellFrameMargins -> {{8, 8}, {4, 3}}, CellFrameColor -> 
       RGBColor[0.486, 0.33, 0.32], StyleMenuListing -> None, FontColor -> 
       RGBColor[0.23, 0.25, 0.34], Background -> RGBColor[0.88, 0.88, 0.92]], 
      Cell[
       StyleData["ComputationInfo"], ShowCellBracket -> False, 
       CellMargins -> {{120, 10}, {10, 2}}, StyleMenuListing -> None, 
       FontFamily -> $CellContext`Arial], 
      Cell[
       StyleData["ComputationInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{144, 10}, {10, 2}}, StyleMenuListing -> None], 
      Cell[
       StyleData["ProofInfo"], ShowCellBracket -> False, 
       CellMargins -> {{76, 10}, {6, 6}}, CellFrameLabelMargins -> 15, 
       StyleMenuListing -> None, FontFamily -> "Arial"], 
      Cell[
       StyleData["ProofInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{100, 10}, {0, 0}}, StyleMenuListing -> None], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FrameLabel"], LanguageCategory -> None, StyleMenuListing -> 
          None, FontFamily -> "Times New Roman", FontSize -> 10, FontWeight -> 
          "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
          FontVariations -> {
           "Masked" -> False, "Outline" -> False, "Shadow" -> False, 
            "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
          RGBColor[0.55, 0.44, 0.42], Background -> None], 
         Cell[
          StyleData["FrameLabel", "Printout"]]}, Open]], 
      Cell[
       StyleData["OpenArchive"], CellFrame -> {{False, False}, {False, True}},
        ShowCellBracket -> Automatic, CellMargins -> {{70, 70}, {0, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 20}, CellFrameMargins -> 3, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], CellFrameLabelMargins -> 1, 
       DefaultFormatType -> DefaultInputFormatType, ShowAutoStyles -> False, 
       MenuSortingValue -> 1210, FontWeight -> Bold, Background -> 
       RGBColor[0.73, 0.74, 0.84]], 
      Cell[
       StyleData["ArchiveInfo"], ShowCellBracket -> Automatic, 
       CellMargins -> {{70, 70}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 5, CellFrameLabelMargins -> 3, ShowCellTags -> 
       False, ShowAutoStyles -> False, MenuSortingValue -> 1220, FontWeight -> 
       "Bold", Background -> RGBColor[0.88, 0.88, 0.92]], 
      Cell[
       StyleData["CloseArchive"], 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> 
       Automatic, CellMargins -> {{0, 0}, {5, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 25}, CellFrameMargins -> 2, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], TextAlignment -> Center, 
       MenuSortingValue -> 1230, FontFamily -> "Arial Black", FontSize -> 24, 
       FontWeight -> "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       Background -> RGBColor[0.73, 0.65, 0.61]], 
      Cell[
       StyleData["IncludeArchive"], CellFrame -> {{4, 0}, {False, False}}, 
       ShowCellBracket -> Automatic, 
       CellMargins -> {{81, 100}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameColor -> RGBColor[0.55, 0.44, 0.42], 
       CellFrameLabels -> {{"\[RightSkeleton]", None}, {None, None}}, 
       CellFrameLabelMargins -> 25, MenuSortingValue -> 1445], 
      Cell[
       StyleData["ExpressionVariable"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionABF"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionMeta"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["AutoParentheses"], Editable -> False, SpanMaxSize -> 3, 
          FontOpacity -> 0.15], 
         Cell[
          StyleData["AutoParentheses", "Layout"], FontOpacity -> 0], 
         Cell[
          StyleData["AutoParentheses", "SlideShow"], FontOpacity -> 0]}, 
        Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Emphasized"], MenuCommandKey -> "E", FontSlant -> 
          "Italic"], 
         Cell[
          StyleData["Emphasized", "SlideShow"], FontSlant -> "Plain", 
          FontColor -> RGBColor[0.36, 0.54, 0.69]]}, Open]], 
      Cell[
       StyleData["GenericButton"], FontFamily -> "Bitstream Charter", 
       FontWeight -> "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
       FontVariations -> {
        "Masked" -> False, "Outline" -> False, "Shadow" -> False, 
         "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
       RGBColor[0.23, 0.25, 0.34]], 
      Cell[
       StyleData["TheoremaForm"], "TwoByteSyntaxCharacterAutoReplacement" -> 
       True, LineSpacing -> {1.25, 0, 2}, SingleLetterItalics -> True, 
       StyleMenuListing -> None, FontFamily -> "Courier"], 
      Cell[
       StyleData["ComponentActive"], FontFamily -> "Times New Roman"], 
      Cell[
       StyleData["ComponentInactive"], FontFamily -> "Times New Roman", 
       FontOpacity -> 0.4], 
      Cell[
       StyleData["ComponentEmpty"], FontFamily -> "Times New Roman", 
       FontVariations -> {"StrikeThrough" -> True}], 
      Cell[
       StyleData["TabView"], FontFamily -> "Times New Roman", FontSize -> 10], 
      Cell[
       StyleData["TabViewLabel"], FontFamily -> "Times New Roman", FontSize -> 
       10, FontWeight -> "Bold"], 
      Cell[
       StyleData["FormReference", StyleDefinitions -> StyleData["Text"]], 
       FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> "Plain",
        FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["FormReferenceHover", StyleDefinitions -> StyleData["Text"]],
        FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> 
       "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> True}, 
       FontColor -> RGBColor[0.404, 0.43, 0.545]], 
      Cell[
       StyleData["DisplayFormula"], CellMargins -> {{36, 6}, {7, 7}}, 
       ConversionRules :> {
        "HTML" -> {{"", ""}, {"<p class=\"Equation\">", "</p>"}}}, 
       CellHorizontalScrolling -> True, DefaultFormatType -> TraditionalForm, 
       InputAutoReplacements -> {}, TextAlignment -> Center, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       ButtonBoxOptions -> {Appearance -> {Automatic, None}}]}, Open]]}, 
  Visible -> False, FrontEndVersion -> 
  "9.0 for Linux x86 (64-bit) (February 7, 2013)", StyleDefinitions -> 
  "StylesheetFormatting.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "ID:346267501"->{
  Cell[87002, 2279, 635, 16, 31, "GlobalDeclaration",
   CellTags->{
    "ID:346267501", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->346267501]},
 "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"->{
  Cell[87002, 2279, 635, 16, 31, "GlobalDeclaration",
   CellTags->{
    "ID:346267501", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->346267501],
  Cell[87640, 2297, 1611, 43, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:143274494", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet"},
   CellID->143274494],
  Cell[89254, 2342, 2201, 72, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:669214708", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRec"},
   CellID->669214708],
  Cell[91937, 2433, 662, 17, 31, "GlobalDeclaration",
   CellTags->{
    "ID:97279932", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->97279932],
  Cell[93277, 2472, 812, 20, 32, "GlobalDeclaration",
   CellTags->{
    "ID:504518491", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->504518491],
  Cell[94092, 2494, 2480, 74, 46, "FormalTextInputFormula",
   CellTags->{
    "ID:79522091", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet induction", "IR"},
   CellID->79522091],
  Cell[96575, 2570, 884, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:212626428", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet isSet", "IR"},
   CellID->212626428],
  Cell[97462, 2594, 998, 26, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:433946629", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "base \[Element] natRecSet"},
   CellID->433946629],
  Cell[98463, 2622, 1248, 34, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:198301498", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet inductive"},
   CellID->198301498],
  Cell[99714, 2658, 1281, 37, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:972417912", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet shape"},
   CellID->972417912],
  Cell[100998, 2697, 2066, 59, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:57153239", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet cases"},
   CellID->57153239],
  Cell[103067, 2758, 3468, 104, 78, "FormalTextInputFormula",
   CellTags->{
    "ID:240965330", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet induction'"},
   CellID->240965330],
  Cell[106538, 2864, 1558, 44, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:207844390", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet functional"},
   CellID->207844390],
  Cell[108099, 2910, 1303, 37, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:348247310", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet shape'"},
   CellID->348247310],
  Cell[109405, 2949, 985, 29, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:117940020", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet total"},
   CellID->117940020],
  Cell[162135, 4398, 906, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:127781311", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRec base", "IR"},
   CellID->127781311],
  Cell[163044, 4425, 1267, 38, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:186992137", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRec step", "IR"},
   CellID->186992137],
  Cell[173000, 4728, 646, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:379686829", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->379686829],
  Cell[173649, 4746, 637, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:778005153", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->778005153],
  Cell[174289, 4764, 1189, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:97787251", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat \[Element]"},
   CellID->97787251],
  Cell[175481, 4806, 819, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:720795295", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat 0"},
   CellID->720795295],
  Cell[176303, 4831, 908, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:3778917", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat 1"},
   CellID->3778917],
  Cell[177214, 4858, 970, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:54060264", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat Pred"},
   CellID->54060264],
  Cell[178187, 4888, 1066, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:205601935", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat +"},
   CellID->205601935],
  Cell[179256, 4922, 1011, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:74238769", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat -"},
   CellID->74238769],
  Cell[180270, 4955, 1273, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:288612943", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat \[RawStar]"},
   CellID->288612943],
  Cell[181546, 4997, 1659, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:155532301", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat \[GreaterEqual]"},
   CellID->155532301],
  Cell[184310, 5083, 748, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:15394254", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat isDomain"},
   CellID->15394254],
  Cell[185061, 5105, 1924, 62, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:478904098", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat induction", "IR"},
   CellID->478904098],
  Cell[186988, 5169, 832, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:60850406", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 \[Element] Nat", "IR"},
   CellID->60850406],
  Cell[187823, 5194, 1033, 29, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:114723517", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ \[Element] Nat", "IR"},
   CellID->114723517],
  Cell[188859, 5225, 831, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:275276526", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[Element] Nat", "IR"},
   CellID->275276526],
  Cell[189693, 5250, 691, 18, 41, "GlobalDeclaration",
   CellTags->{
    "ID:65979872", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->65979872],
  Cell[190387, 5270, 770, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:572443783", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat shape"},
   CellID->572443783],
  Cell[191160, 5293, 1442, 43, 66, "FormalTextInputFormula",
   CellTags->{
    "ID:252333509", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat cases"},
   CellID->252333509],
  Cell[192605, 5338, 846, 23, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:77756457", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ \[NotEqual] 0 Nat"},
   CellID->77756457],
  Cell[193454, 5363, 1116, 34, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:52299395", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ injective Nat"},
   CellID->52299395],
  Cell[194573, 5399, 796, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:263281689", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ no fixpoint Nat"},
   CellID->263281689],
  Cell[195372, 5423, 817, 22, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:377798869", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[NotEqual] 0 Nat"},
   CellID->377798869],
  Cell[196192, 5447, 943, 26, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:33739458", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[Element] closed Nat"},
   CellID->33739458],
  Cell[197138, 5475, 1138, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:654854303", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat transitive"},
   CellID->654854303],
  Cell[238422, 6742, 861, 24, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:240175329", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Pred 0 Nat", "IR"},
   CellID->240175329],
  Cell[239286, 6768, 666, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:158284383", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->158284383],
  Cell[239955, 6788, 955, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:145236688", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Pred succ Nat", "IR"},
   CellID->145236688],
  Cell[240913, 6816, 870, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:225607097", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Pred closed Nat", "IR"},
   CellID->225607097],
  Cell[251476, 7155, 766, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:218406943", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->218406943],
  Cell[252245, 7177, 923, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:257436422", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ 0 Nat", "IR"},
   CellID->257436422],
  Cell[253171, 7204, 1063, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:249353684", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ succ Nat", "IR"},
   CellID->249353684],
  Cell[254237, 7234, 850, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:279658160", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ closed Nat", "IR"},
   CellID->279658160],
  Cell[255090, 7260, 899, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:647623443", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ 1 Nat"},
   CellID->647623443],
  Cell[255992, 7287, 939, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:18327536", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 lneutral Nat", "IR"},
   CellID->18327536],
  Cell[256934, 7315, 956, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1914645", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ + Nat", "IR"},
   CellID->1914645],
  Cell[257893, 7344, 1190, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:343580194", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ = 0 Nat"},
   CellID->343580194],
  Cell[259086, 7385, 889, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:834229092", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ comm Nat"},
   CellID->834229092],
  Cell[259978, 7412, 668, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:303913216", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->303913216],
  Cell[260649, 7432, 1094, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:73705609", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ assoc Nat"},
   CellID->73705609],
  Cell[261746, 7466, 1008, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:202544617", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ comm-assoc Nat"},
   CellID->202544617],
  Cell[262757, 7498, 1407, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:555281086", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ + + Nat"},
   CellID->555281086],
  Cell[264167, 7546, 1147, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:5229733", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ rcanc Nat"},
   CellID->5229733],
  Cell[312012, 8975, 717, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:301992507", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->301992507],
  Cell[312732, 8996, 895, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:99163941", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] 0 Nat", "IR"},
   CellID->99163941],
  Cell[313630, 9023, 987, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:371484270", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] succ Nat", "IR"},
   CellID->371484270],
  Cell[314620, 9053, 852, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:142418160", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] closed Nat", "IR"},
   CellID->142418160],
  Cell[315475, 9079, 832, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:21383846", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 rneutral Nat", "IR"},
   CellID->21383846],
  Cell[316310, 9104, 864, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:28953873", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 lannihilator Nat", "IR"},
   CellID->28953873],
  Cell[317177, 9130, 981, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:218775165", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ \[RawStar] Nat", "IR"},
   CellID->218775165],
  Cell[318161, 9160, 1191, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:470439981", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] zdf Nat"},
   CellID->470439981],
  Cell[319355, 9201, 842, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:218136211", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] comm Nat"},
   CellID->218136211],
  Cell[320200, 9227, 690, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:666029340", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->666029340],
  Cell[320893, 9247, 1080, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:43150914", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] distr + Nat"},
   CellID->43150914],
  Cell[321976, 9281, 1000, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:475421150", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] assoc Nat"},
   CellID->475421150],
  Cell[322979, 9313, 1007, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:98508460", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] comm-assoc Nat"},
   CellID->98508460],
  Cell[323989, 9345, 1194, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:138249572", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] rcanc Nat"},
   CellID->138249572],
  Cell[372145, 10782, 715, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:621974620", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->621974620],
  Cell[372863, 10803, 1126, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:380981542", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 \[GreaterEqual] Nat", "IR"},
   CellID->380981542],
  Cell[373992, 10840, 1292, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:42977349", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ \[GreaterEqual] Nat", "IR"},
   CellID->42977349],
  Cell[375287, 10881, 864, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:428740350", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] reflexive Nat"},
   CellID->428740350],
  Cell[376154, 10905, 882, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:297686800", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] 0 Nat", "IR"},
   CellID->297686800],
  Cell[377039, 10930, 1316, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:585961470", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] succ Nat", "IR"},
   CellID->585961470],
  Cell[378358, 10971, 1080, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:894887193", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] antisymmetric Nat"},
   CellID->894887193],
  Cell[379441, 11004, 875, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:208533627", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ increasing Nat"},
   CellID->208533627],
  Cell[380319, 11029, 1239, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:112605784", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ \[GreaterEqual] succ Nat", "IR"},
   CellID->112605784],
  Cell[381561, 11068, 901, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:304315076", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] total Nat"},
   CellID->304315076],
  Cell[382465, 11095, 1167, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:446896782", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] \[Exists] Nat"},
   CellID->446896782],
  Cell[383635, 11132, 853, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:196203313", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[GreaterEqual] 0 Nat"},
   CellID->196203313],
  Cell[384491, 11157, 1179, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:30448165", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] 1 Nat", "IR"},
   CellID->30448165],
  Cell[385673, 11195, 690, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:222994735", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->222994735],
  Cell[386366, 11215, 1107, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:546802157", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] transitive Nat"},
   CellID->546802157],
  Cell[387476, 11250, 1045, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:352201475", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] + \[DoubleLeftArrow] Nat"},
   CellID->352201475],
  Cell[388524, 11281, 1208, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:97128652", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] + Nat"},
   CellID->97128652],
  Cell[389735, 11321, 849, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:278931212", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ \[GreaterEqual] Nat"},
   CellID->278931212],
  Cell[390587, 11346, 1000, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:335211897", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] \[RawStar] Nat"},
   CellID->335211897],
  Cell[391590, 11376, 943, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:32863246", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] \[GreaterEqual] Nat"},
   CellID->32863246],
  Cell[461486, 13436, 717, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:609763659", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->609763659],
  Cell[462206, 13457, 816, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:19208428", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- 0 Nat", "IR"},
   CellID->19208428],
  Cell[463025, 13482, 987, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:593421999", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- succ Nat", "IR"},
   CellID->593421999],
  Cell[464015, 13512, 848, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:153357458", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- closed Nat", "IR"},
   CellID->153357458],
  Cell[464866, 13538, 871, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:616125500", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- 1 Nat"},
   CellID->616125500],
  Cell[465740, 13565, 845, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:491337343", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 - Nat", "IR"},
   CellID->491337343],
  Cell[466588, 13591, 1028, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:621700658", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ - succ Nat", "IR"},
   CellID->621700658],
  Cell[467619, 13621, 869, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:236021775", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ - Nat"},
   CellID->236021775],
  Cell[468491, 13649, 965, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:3006634", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- + Nat"},
   CellID->3006634],
  Cell[469459, 13680, 663, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:80465146", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->80465146],
  Cell[470125, 13700, 996, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:404996142", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ - + Nat"},
   CellID->404996142],
  Cell[471124, 13732, 1106, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:225131303", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] distr - Nat"},
   CellID->225131303],
  Cell[472233, 13766, 1206, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:142079924", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "(-)+ Nat"},
   CellID->142079924],
  Cell[473442, 13806, 2138, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:367201929", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- = - Nat"},
   CellID->367201929],
  Cell[524838, 15330, 649, 16, 30, "GlobalDeclaration",
   CellTags->{
    "ID:513393113", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->513393113],
  Cell[525490, 15348, 637, 16, 31, "GlobalDeclaration",
   CellTags->{
    "ID:623651727", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->623651727],
  Cell[526130, 15366, 1235, 41, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:100384801", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt \[Element]"},
   CellID->100384801],
  Cell[527368, 15409, 865, 25, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:398024454", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt 0"},
   CellID->398024454],
  Cell[528236, 15436, 869, 25, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:418726841", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt 1"},
   CellID->418726841],
  Cell[529108, 15463, 1072, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:10580372", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt +"},
   CellID->10580372],
  Cell[530183, 15497, 906, 26, 28, "FormalTextInputFormula",
   CellTags->{
    "ID:106096058", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt - unary"},
   CellID->106096058],
  Cell[531092, 15525, 1436, 44, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:12727554", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt \[RawStar]"},
   CellID->12727554],
  Cell[532531, 15571, 1463, 48, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:329154583", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt \[GreaterEqual]"},
   CellID->329154583],
  Cell[533997, 15621, 1332, 43, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:468243965", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt equiv"},
   CellID->468243965],
  Cell[536187, 15699, 770, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:168445606", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt isDomain", "IR"},
   CellID->168445606],
  Cell[536960, 15721, 1640, 54, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:977301511", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[Element] PreInt alternative", "IR"},
   CellID->977301511],
  Cell[538603, 15777, 1529, 50, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:353341722", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Pair \[Element] PreInt", "IR"},
   CellID->353341722],
  Cell[540135, 15829, 840, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:34141802", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 \[Element] PreInt", "IR"},
   CellID->34141802],
  Cell[540978, 15854, 847, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:553448562", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[Element] PreInt", "IR"},
   CellID->553448562],
  Cell[557926, 16373, 1504, 47, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:220417861", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv Pair PreInt", "IR"},
   CellID->220417861],
  Cell[559433, 16422, 1398, 44, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:52539739", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv 0 PreInt", "IR"},
   CellID->52539739],
  Cell[560834, 16468, 1416, 43, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:94484613", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv Nat \[GreaterEqual] PreInt"},
   CellID->94484613],
  Cell[562253, 16513, 1413, 43, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:321256465", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv Nat \[LessEqual] PreInt"},
   CellID->321256465],
  Cell[563669, 16558, 911, 25, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:129809121", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[Not] 1 equiv 0 PreInt"},
   CellID->129809121],
  Cell[564583, 16585, 720, 19, 41, "GlobalDeclaration",
   CellTags->{
    "ID:413771523", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->413771523],
  Cell[565306, 16606, 1703, 51, 69, "FormalTextInputFormula",
   CellTags->{
    "ID:637862621", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv Nat PreInt"},
   CellID->637862621],
  Cell[567012, 16659, 817, 22, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:12369324", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv reflexive PreInt"},
   CellID->12369324],
  Cell[567832, 16683, 930, 26, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:19276", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv symmetric PreInt"},
   CellID->19276],
  Cell[568765, 16711, 1420, 45, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:164883544", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv transitive PreInt"},
   CellID->164883544],
  Cell[570188, 16758, 849, 23, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:357440757", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv isEquivalence PreInt"},
   CellID->357440757],
  Cell[611236, 17955, 1361, 41, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:125790650", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ Pair PreInt", "IR"},
   CellID->125790650],
  Cell[612600, 17998, 715, 19, 41, "GlobalDeclaration",
   CellTags->{
    "ID:86057977", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->86057977],
  Cell[613318, 18019, 856, 24, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:196108575", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ closed PreInt"},
   CellID->196108575],
  Cell[614177, 18045, 844, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:132949647", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 rneutral PreInt"},
   CellID->132949647],
  Cell[615024, 18071, 851, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:25630206", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ comm PreInt"},
   CellID->25630206],
  Cell[615878, 18097, 694, 18, 41, "GlobalDeclaration",
   CellTags->{
    "ID:16604231", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->16604231],
  Cell[616575, 18117, 1018, 31, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:73792979", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ assoc PreInt"},
   CellID->73792979],
  Cell[617596, 18150, 1093, 31, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:243295365", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ rcanc PreInt"},
   CellID->243295365],
  Cell[618692, 18183, 1057, 31, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:47524840", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ invaiant equiv PreInt"},
   CellID->47524840],
  Cell[619752, 18216, 1030, 31, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:107686175", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ isInvariantFunction2 PreInt"},
   CellID->107686175],
  Cell[652392, 19175, 1112, 34, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:360943410", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- unary Pair PreInt", "IR"},
   CellID->360943410],
  Cell[653507, 19211, 695, 18, 41, "GlobalDeclaration",
   CellTags->{
    "ID:474936069", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->474936069],
  Cell[654205, 19231, 863, 24, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:5265224", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- unary closed PreInt", "IR"},
   CellID->5265224],
  Cell[655071, 19257, 1076, 30, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:203475906", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- inverse PreInt"},
   CellID->203475906],
  Cell[656150, 19289, 959, 25, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:611704112", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- isInvariantFunction1 PreInt"},
   CellID->611704112],
  Cell[670688, 19739, 1603, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:455358938", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] Pair PreInt", "IR"},
   CellID->455358938],
  Cell[672294, 19790, 771, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:206670450", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->206670450],
  Cell[673068, 19812, 908, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:80978087", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] closed PreInt", "IR"},
   CellID->80978087],
  Cell[673979, 19839, 878, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:561254237", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 rannihilator PreInt"},
   CellID->561254237],
  Cell[674860, 19865, 897, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:201948360", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 rneutral PreInt"},
   CellID->201948360],
  Cell[675760, 19892, 853, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:152728157", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] comm PreInt"},
   CellID->152728157],
  Cell[676616, 19918, 1344, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:258603229", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] zdf PreInt"},
   CellID->258603229],
  Cell[677963, 19961, 695, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:200297603", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->200297603],
  Cell[678661, 19981, 1023, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:142359696", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] assoc PreInt"},
   CellID->142359696],
  Cell[679687, 20014, 1053, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:262630475", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] distr + PreInt"},
   CellID->262630475],
  Cell[680743, 20047, 1064, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:146416852", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] invariant equiv PreInt"},
   CellID->146416852],
  Cell[681810, 20080, 1030, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:204021385", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] isInvariantFunction2 PreInt"},
   CellID->204021385],
  Cell[723510, 21297, 1590, 50, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:129935970", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] Pair PreInt", "IR"},
   CellID->129935970],
  Cell[725103, 21349, 1539, 49, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:95876614", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] 0 PreInt", "IR"},
   CellID->95876614],
  Cell[726645, 21400, 919, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:221338547", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[GreaterEqual] 0 PreInt"},
   CellID->221338547],
  Cell[727567, 21426, 1078, 32, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:13569836", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] isInvariantRelation2 PreInt"},
   CellID->13569836],
  Cell[728648, 21460, 722, 19, 41, "GlobalDeclaration",
   CellTags->{
    "ID:337456597", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->337456597],
  Cell[729373, 21481, 822, 22, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:83533617", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] reflexive PreInt"},
   CellID->83533617],
  Cell[730198, 21505, 1157, 34, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:281735911", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] antisymmetric PreInt"},
   CellID->281735911],
  Cell[731358, 21541, 913, 25, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:115550209", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] total PreInt"},
   CellID->115550209],
  Cell[732274, 21568, 1372, 42, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:215340491", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] discrete PreInt"},
   CellID->215340491],
  Cell[733649, 21612, 1298, 38, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:92075332", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] \[RawStar] PreInt"},
   CellID->92075332],
  Cell[734950, 21652, 1299, 39, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:145078248", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat \[SucceedsTilde] PreInt+"},
   CellID->145078248],
  Cell[736252, 21693, 697, 18, 41, "GlobalDeclaration",
   CellTags->{
    "ID:118070489", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->118070489],
  Cell[736952, 21713, 1125, 33, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:307661531", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] transitive PreInt"},
   CellID->307661531],
  Cell[738080, 21748, 1012, 28, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:647932785", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] + PreInt"},
   CellID->647932785],
  Cell[786089, 23143, 749, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:463204962", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->463204962],
  Cell[787445, 23186, 645, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:32325634", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->32325634],
  Cell[788093, 23204, 1067, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:112681936", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int \[Element]"},
   CellID->112681936],
  Cell[789163, 23239, 849, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:52976508", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int 0"},
   CellID->52976508],
  Cell[790015, 23265, 852, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:712864778", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int 1"},
   CellID->712864778],
  Cell[790870, 23291, 1098, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:19419340", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int +"},
   CellID->19419340],
  Cell[791971, 23331, 872, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:280798627", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int - unary"},
   CellID->280798627],
  Cell[792846, 23357, 1124, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:328192039", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int \[RawStar]"},
   CellID->328192039],
  Cell[793973, 23398, 1184, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:819512290", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int \[GreaterEqual]"},
   CellID->819512290],
  Cell[796056, 23473, 757, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:354647619", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int isDomain", "IR"},
   CellID->354647619],
  Cell[796816, 23495, 1619, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:565027904", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[Element] Int alternative", "IR"},
   CellID->565027904],
  Cell[798438, 23550, 1163, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:40757894", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "EC \[Element] Int", "IR"},
   CellID->40757894],
  Cell[799604, 23587, 833, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:350791961", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 \[Element] Int", "IR"},
   CellID->350791961],
  Cell[800440, 23612, 825, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:181240486", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[Element] Int"},
   CellID->181240486],
  Cell[801268, 23637, 815, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:364462666", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[NotEqual] 0 Int"},
   CellID->364462666],
  Cell[819770, 24222, 1339, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:56650873", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ EC Int", "IR"},
   CellID->56650873],
  Cell[821112, 24266, 777, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:13236461", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->13236461],
  Cell[821892, 24288, 848, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:562490241", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ closed Int", "IR"},
   CellID->562490241],
  Cell[822743, 24314, 827, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:624613835", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 rneutral Int"},
   CellID->624613835],
  Cell[823573, 24339, 844, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:115787764", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ comm Int"},
   CellID->115787764],
  Cell[824420, 24365, 1048, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:11815767", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ assoc Int"},
   CellID->11815767],
  Cell[825471, 24398, 961, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:139539770", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ rcanc Int"},
   CellID->139539770],
  Cell[850720, 25149, 1242, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:331988502", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- unary EC Int", "IR"},
   CellID->331988502],
  Cell[851965, 25190, 694, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:138388913", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->138388913],
  Cell[852662, 25210, 857, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:789456798", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- unary closed Int", "IR"},
   CellID->789456798],
  Cell[853522, 25236, 895, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:561244613", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- inverse Int"},
   CellID->561244613],
  Cell[865668, 25614, 1340, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:947607354", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] EC Int", "IR"},
   CellID->947607354],
  Cell[867011, 25658, 701, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:105543058", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->105543058],
  Cell[867715, 25679, 852, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:696766377", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] closed Int", "IR"},
   CellID->696766377],
  Cell[868570, 25705, 861, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:202378771", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 rannihilator Int"},
   CellID->202378771],
  Cell[869434, 25731, 827, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:689609352", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 rneutral Int"},
   CellID->689609352],
  Cell[870264, 25756, 842, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:196343023", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] comm Int"},
   CellID->196343023],
  Cell[871109, 25782, 1193, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:207834126", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] zdf Int"},
   CellID->207834126],
  Cell[872305, 25823, 999, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:10196611", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] assoc Int"},
   CellID->10196611],
  Cell[873307, 25855, 1032, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:777222529", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] distr + Int"},
   CellID->777222529],
  Cell[906257, 26835, 1550, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:785793765", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] EC Int", "IR"},
   CellID->785793765],
  Cell[907810, 26886, 850, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:72386745", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[GreaterEqual] 0 Int"},
   CellID->72386745],
  Cell[908663, 26911, 726, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:100510373", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->100510373],
  Cell[909392, 26932, 815, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:267773274", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] reflexive Int"},
   CellID->267773274],
  Cell[910210, 26955, 1075, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:69005952", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] antisymmetric Int"},
   CellID->69005952],
  Cell[911288, 26988, 1112, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:275182510", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] transitive Int"},
   CellID->275182510],
  Cell[912403, 27023, 899, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:176048001", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] total Int"},
   CellID->176048001],
  Cell[913305, 27050, 1156, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:566590035", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] discrete Int"},
   CellID->566590035],
  Cell[914464, 27086, 1000, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:262876021", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] + Int"},
   CellID->262876021],
  Cell[915467, 27116, 1224, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:34443687", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] \[RawStar] Int"},
   CellID->34443687],
  Cell[952069, 28193, 1720, 52, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:134029858", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat \[SucceedsTilde] Int+"},
   CellID->134029858],
  Cell[953792, 28247, 2965, 92, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:181242736", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int+ induction"},
   CellID->181242736]},
 "ID:143274494"->{
  Cell[87640, 2297, 1611, 43, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:143274494", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet"},
   CellID->143274494]},
 "natRecSet"->{
  Cell[87640, 2297, 1611, 43, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:143274494", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet"},
   CellID->143274494]},
 "ID:669214708"->{
  Cell[89254, 2342, 2201, 72, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:669214708", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRec"},
   CellID->669214708]},
 "natRec"->{
  Cell[89254, 2342, 2201, 72, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:669214708", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRec"},
   CellID->669214708]},
 "ID:97279932"->{
  Cell[91937, 2433, 662, 17, 31, "GlobalDeclaration",
   CellTags->{
    "ID:97279932", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->97279932]},
 "ID:504518491"->{
  Cell[93277, 2472, 812, 20, 32, "GlobalDeclaration",
   CellTags->{
    "ID:504518491", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->504518491]},
 "ID:79522091"->{
  Cell[94092, 2494, 2480, 74, 46, "FormalTextInputFormula",
   CellTags->{
    "ID:79522091", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet induction", "IR"},
   CellID->79522091]},
 "natRecSet induction"->{
  Cell[94092, 2494, 2480, 74, 46, "FormalTextInputFormula",
   CellTags->{
    "ID:79522091", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet induction", "IR"},
   CellID->79522091]},
 "IR"->{
  Cell[94092, 2494, 2480, 74, 46, "FormalTextInputFormula",
   CellTags->{
    "ID:79522091", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet induction", "IR"},
   CellID->79522091],
  Cell[96575, 2570, 884, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:212626428", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet isSet", "IR"},
   CellID->212626428],
  Cell[162135, 4398, 906, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:127781311", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRec base", "IR"},
   CellID->127781311],
  Cell[163044, 4425, 1267, 38, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:186992137", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRec step", "IR"},
   CellID->186992137],
  Cell[185061, 5105, 1924, 62, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:478904098", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat induction", "IR"},
   CellID->478904098],
  Cell[186988, 5169, 832, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:60850406", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 \[Element] Nat", "IR"},
   CellID->60850406],
  Cell[187823, 5194, 1033, 29, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:114723517", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ \[Element] Nat", "IR"},
   CellID->114723517],
  Cell[188859, 5225, 831, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:275276526", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[Element] Nat", "IR"},
   CellID->275276526],
  Cell[238422, 6742, 861, 24, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:240175329", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Pred 0 Nat", "IR"},
   CellID->240175329],
  Cell[239955, 6788, 955, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:145236688", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Pred succ Nat", "IR"},
   CellID->145236688],
  Cell[240913, 6816, 870, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:225607097", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Pred closed Nat", "IR"},
   CellID->225607097],
  Cell[252245, 7177, 923, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:257436422", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ 0 Nat", "IR"},
   CellID->257436422],
  Cell[253171, 7204, 1063, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:249353684", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ succ Nat", "IR"},
   CellID->249353684],
  Cell[254237, 7234, 850, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:279658160", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ closed Nat", "IR"},
   CellID->279658160],
  Cell[255992, 7287, 939, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:18327536", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 lneutral Nat", "IR"},
   CellID->18327536],
  Cell[256934, 7315, 956, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1914645", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ + Nat", "IR"},
   CellID->1914645],
  Cell[312732, 8996, 895, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:99163941", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] 0 Nat", "IR"},
   CellID->99163941],
  Cell[313630, 9023, 987, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:371484270", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] succ Nat", "IR"},
   CellID->371484270],
  Cell[314620, 9053, 852, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:142418160", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] closed Nat", "IR"},
   CellID->142418160],
  Cell[315475, 9079, 832, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:21383846", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 rneutral Nat", "IR"},
   CellID->21383846],
  Cell[316310, 9104, 864, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:28953873", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 lannihilator Nat", "IR"},
   CellID->28953873],
  Cell[317177, 9130, 981, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:218775165", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ \[RawStar] Nat", "IR"},
   CellID->218775165],
  Cell[372863, 10803, 1126, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:380981542", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 \[GreaterEqual] Nat", "IR"},
   CellID->380981542],
  Cell[373992, 10840, 1292, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:42977349", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ \[GreaterEqual] Nat", "IR"},
   CellID->42977349],
  Cell[376154, 10905, 882, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:297686800", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] 0 Nat", "IR"},
   CellID->297686800],
  Cell[377039, 10930, 1316, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:585961470", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] succ Nat", "IR"},
   CellID->585961470],
  Cell[380319, 11029, 1239, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:112605784", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ \[GreaterEqual] succ Nat", "IR"},
   CellID->112605784],
  Cell[384491, 11157, 1179, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:30448165", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] 1 Nat", "IR"},
   CellID->30448165],
  Cell[462206, 13457, 816, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:19208428", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- 0 Nat", "IR"},
   CellID->19208428],
  Cell[463025, 13482, 987, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:593421999", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- succ Nat", "IR"},
   CellID->593421999],
  Cell[464015, 13512, 848, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:153357458", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- closed Nat", "IR"},
   CellID->153357458],
  Cell[465740, 13565, 845, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:491337343", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 - Nat", "IR"},
   CellID->491337343],
  Cell[466588, 13591, 1028, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:621700658", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ - succ Nat", "IR"},
   CellID->621700658],
  Cell[536187, 15699, 770, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:168445606", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt isDomain", "IR"},
   CellID->168445606],
  Cell[536960, 15721, 1640, 54, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:977301511", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[Element] PreInt alternative", "IR"},
   CellID->977301511],
  Cell[538603, 15777, 1529, 50, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:353341722", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Pair \[Element] PreInt", "IR"},
   CellID->353341722],
  Cell[540135, 15829, 840, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:34141802", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 \[Element] PreInt", "IR"},
   CellID->34141802],
  Cell[540978, 15854, 847, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:553448562", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[Element] PreInt", "IR"},
   CellID->553448562],
  Cell[557926, 16373, 1504, 47, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:220417861", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv Pair PreInt", "IR"},
   CellID->220417861],
  Cell[559433, 16422, 1398, 44, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:52539739", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv 0 PreInt", "IR"},
   CellID->52539739],
  Cell[611236, 17955, 1361, 41, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:125790650", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ Pair PreInt", "IR"},
   CellID->125790650],
  Cell[652392, 19175, 1112, 34, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:360943410", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- unary Pair PreInt", "IR"},
   CellID->360943410],
  Cell[654205, 19231, 863, 24, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:5265224", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- unary closed PreInt", "IR"},
   CellID->5265224],
  Cell[670688, 19739, 1603, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:455358938", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] Pair PreInt", "IR"},
   CellID->455358938],
  Cell[673068, 19812, 908, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:80978087", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] closed PreInt", "IR"},
   CellID->80978087],
  Cell[723510, 21297, 1590, 50, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:129935970", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] Pair PreInt", "IR"},
   CellID->129935970],
  Cell[725103, 21349, 1539, 49, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:95876614", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] 0 PreInt", "IR"},
   CellID->95876614],
  Cell[796056, 23473, 757, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:354647619", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int isDomain", "IR"},
   CellID->354647619],
  Cell[796816, 23495, 1619, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:565027904", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[Element] Int alternative", "IR"},
   CellID->565027904],
  Cell[798438, 23550, 1163, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:40757894", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "EC \[Element] Int", "IR"},
   CellID->40757894],
  Cell[799604, 23587, 833, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:350791961", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 \[Element] Int", "IR"},
   CellID->350791961],
  Cell[819770, 24222, 1339, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:56650873", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ EC Int", "IR"},
   CellID->56650873],
  Cell[821892, 24288, 848, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:562490241", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ closed Int", "IR"},
   CellID->562490241],
  Cell[850720, 25149, 1242, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:331988502", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- unary EC Int", "IR"},
   CellID->331988502],
  Cell[852662, 25210, 857, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:789456798", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- unary closed Int", "IR"},
   CellID->789456798],
  Cell[865668, 25614, 1340, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:947607354", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] EC Int", "IR"},
   CellID->947607354],
  Cell[867715, 25679, 852, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:696766377", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] closed Int", "IR"},
   CellID->696766377],
  Cell[906257, 26835, 1550, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:785793765", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] EC Int", "IR"},
   CellID->785793765]},
 "ID:212626428"->{
  Cell[96575, 2570, 884, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:212626428", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet isSet", "IR"},
   CellID->212626428]},
 "natRecSet isSet"->{
  Cell[96575, 2570, 884, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:212626428", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet isSet", "IR"},
   CellID->212626428]},
 "ID:433946629"->{
  Cell[97462, 2594, 998, 26, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:433946629", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "base \[Element] natRecSet"},
   CellID->433946629]},
 "base \[Element] natRecSet"->{
  Cell[97462, 2594, 998, 26, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:433946629", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "base \[Element] natRecSet"},
   CellID->433946629]},
 "ID:198301498"->{
  Cell[98463, 2622, 1248, 34, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:198301498", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet inductive"},
   CellID->198301498]},
 "natRecSet inductive"->{
  Cell[98463, 2622, 1248, 34, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:198301498", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet inductive"},
   CellID->198301498]},
 "ID:972417912"->{
  Cell[99714, 2658, 1281, 37, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:972417912", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet shape"},
   CellID->972417912]},
 "natRecSet shape"->{
  Cell[99714, 2658, 1281, 37, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:972417912", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet shape"},
   CellID->972417912]},
 "ID:57153239"->{
  Cell[100998, 2697, 2066, 59, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:57153239", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet cases"},
   CellID->57153239]},
 "natRecSet cases"->{
  Cell[100998, 2697, 2066, 59, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:57153239", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet cases"},
   CellID->57153239]},
 "ID:240965330"->{
  Cell[103067, 2758, 3468, 104, 78, "FormalTextInputFormula",
   CellTags->{
    "ID:240965330", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet induction'"},
   CellID->240965330]},
 "natRecSet induction'"->{
  Cell[103067, 2758, 3468, 104, 78, "FormalTextInputFormula",
   CellTags->{
    "ID:240965330", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet induction'"},
   CellID->240965330]},
 "ID:207844390"->{
  Cell[106538, 2864, 1558, 44, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:207844390", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet functional"},
   CellID->207844390]},
 "natRecSet functional"->{
  Cell[106538, 2864, 1558, 44, 47, "FormalTextInputFormula",
   CellTags->{
    "ID:207844390", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet functional"},
   CellID->207844390]},
 "ID:348247310"->{
  Cell[108099, 2910, 1303, 37, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:348247310", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet shape'"},
   CellID->348247310]},
 "natRecSet shape'"->{
  Cell[108099, 2910, 1303, 37, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:348247310", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet shape'"},
   CellID->348247310]},
 "ID:117940020"->{
  Cell[109405, 2949, 985, 29, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:117940020", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet total"},
   CellID->117940020]},
 "natRecSet total"->{
  Cell[109405, 2949, 985, 29, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:117940020", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRecSet total"},
   CellID->117940020]},
 "Proof|ID:79522091"->{
  Cell[110776, 2994, 6385, 160, 28, "ProofInfo",
   CellTags->{"Proof|ID:79522091", "Proof|ID:79522091-1"},
   CellID->1623641860]},
 "Proof|ID:79522091-1"->{
  Cell[110776, 2994, 6385, 160, 28, "ProofInfo",
   CellTags->{"Proof|ID:79522091", "Proof|ID:79522091-1"},
   CellID->1623641860]},
 "Proof|ID:212626428"->{
  Cell[117681, 3173, 2402, 72, 26, "ProofInfo",
   CellTags->{"Proof|ID:212626428", "Proof|ID:212626428-1"},
   CellID->2101062054]},
 "Proof|ID:212626428-1"->{
  Cell[117681, 3173, 2402, 72, 26, "ProofInfo",
   CellTags->{"Proof|ID:212626428", "Proof|ID:212626428-1"},
   CellID->2101062054]},
 "Proof|ID:433946629"->{
  Cell[120603, 3264, 2684, 78, 26, "ProofInfo",
   CellTags->{"Proof|ID:433946629", "Proof|ID:433946629-1"},
   CellID->61218476]},
 "Proof|ID:433946629-1"->{
  Cell[120603, 3264, 2684, 78, 26, "ProofInfo",
   CellTags->{"Proof|ID:433946629", "Proof|ID:433946629-1"},
   CellID->61218476]},
 "Proof|ID:198301498"->{
  Cell[123808, 3361, 4108, 110, 26, "ProofInfo",
   CellTags->{"Proof|ID:198301498", "Proof|ID:198301498-1"},
   CellID->1431447619]},
 "Proof|ID:198301498-1"->{
  Cell[123808, 3361, 4108, 110, 26, "ProofInfo",
   CellTags->{"Proof|ID:198301498", "Proof|ID:198301498-1"},
   CellID->1431447619]},
 "Proof|ID:972417912"->{
  Cell[128436, 3490, 3607, 98, 26, "ProofInfo",
   CellTags->{"Proof|ID:972417912", "Proof|ID:972417912-1"},
   CellID->196372021]},
 "Proof|ID:972417912-1"->{
  Cell[128436, 3490, 3607, 98, 26, "ProofInfo",
   CellTags->{"Proof|ID:972417912", "Proof|ID:972417912-1"},
   CellID->196372021]},
 "Proof|ID:57153239"->{
  Cell[132563, 3607, 5226, 132, 26, "ProofInfo",
   CellTags->{"Proof|ID:57153239", "Proof|ID:57153239-1"},
   CellID->736236675]},
 "Proof|ID:57153239-1"->{
  Cell[132563, 3607, 5226, 132, 26, "ProofInfo",
   CellTags->{"Proof|ID:57153239", "Proof|ID:57153239-1"},
   CellID->736236675]},
 "Proof|ID:240965330"->{
  Cell[138308, 3758, 9017, 218, 26, "ProofInfo",
   CellTags->{"Proof|ID:240965330", "Proof|ID:240965330-1"},
   CellID->221211846]},
 "Proof|ID:240965330-1"->{
  Cell[138308, 3758, 9017, 218, 26, "ProofInfo",
   CellTags->{"Proof|ID:240965330", "Proof|ID:240965330-1"},
   CellID->221211846]},
 "Proof|ID:207844390"->{
  Cell[147915, 3996, 4960, 128, 26, "ProofInfo",
   CellTags->{"Proof|ID:207844390", "Proof|ID:207844390-1"},
   CellID->201549000]},
 "Proof|ID:207844390-1"->{
  Cell[147915, 3996, 4960, 128, 26, "ProofInfo",
   CellTags->{"Proof|ID:207844390", "Proof|ID:207844390-1"},
   CellID->201549000]},
 "Proof|ID:348247310"->{
  Cell[153395, 4143, 3860, 106, 26, "ProofInfo",
   CellTags->{"Proof|ID:348247310", "Proof|ID:348247310-1"},
   CellID->22594877]},
 "Proof|ID:348247310-1"->{
  Cell[153395, 4143, 3860, 106, 26, "ProofInfo",
   CellTags->{"Proof|ID:348247310", "Proof|ID:348247310-1"},
   CellID->22594877]},
 "Proof|ID:117940020"->{
  Cell[157775, 4268, 3333, 94, 26, "ProofInfo",
   CellTags->{"Proof|ID:117940020", "Proof|ID:117940020-1"},
   CellID->128642575]},
 "Proof|ID:117940020-1"->{
  Cell[157775, 4268, 3333, 94, 26, "ProofInfo",
   CellTags->{"Proof|ID:117940020", "Proof|ID:117940020-1"},
   CellID->128642575]},
 "ID:127781311"->{
  Cell[162135, 4398, 906, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:127781311", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRec base", "IR"},
   CellID->127781311]},
 "natRec base"->{
  Cell[162135, 4398, 906, 25, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:127781311", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRec base", "IR"},
   CellID->127781311]},
 "ID:186992137"->{
  Cell[163044, 4425, 1267, 38, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:186992137", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRec step", "IR"},
   CellID->186992137]},
 "natRec step"->{
  Cell[163044, 4425, 1267, 38, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:186992137", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "natRec step", "IR"},
   CellID->186992137]},
 "Proof|ID:127781311"->{
  Cell[164466, 4474, 2654, 80, 28, "ProofInfo",
   CellTags->{"Proof|ID:127781311", "Proof|ID:127781311-1"},
   CellID->5572876]},
 "Proof|ID:127781311-1"->{
  Cell[164466, 4474, 2654, 80, 28, "ProofInfo",
   CellTags->{"Proof|ID:127781311", "Proof|ID:127781311-1"},
   CellID->5572876]},
 "Proof|ID:186992137"->{
  Cell[167640, 4573, 3830, 104, 26, "ProofInfo",
   CellTags->{"Proof|ID:186992137", "Proof|ID:186992137-1"},
   CellID->546211]},
 "Proof|ID:186992137-1"->{
  Cell[167640, 4573, 3830, 104, 26, "ProofInfo",
   CellTags->{"Proof|ID:186992137", "Proof|ID:186992137-1"},
   CellID->546211]},
 "ID:379686829"->{
  Cell[173000, 4728, 646, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:379686829", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->379686829]},
 "ID:778005153"->{
  Cell[173649, 4746, 637, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:778005153", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->778005153]},
 "ID:97787251"->{
  Cell[174289, 4764, 1189, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:97787251", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat \[Element]"},
   CellID->97787251]},
 "Nat \[Element]"->{
  Cell[174289, 4764, 1189, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:97787251", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat \[Element]"},
   CellID->97787251]},
 "ID:720795295"->{
  Cell[175481, 4806, 819, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:720795295", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat 0"},
   CellID->720795295]},
 "Nat 0"->{
  Cell[175481, 4806, 819, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:720795295", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat 0"},
   CellID->720795295]},
 "ID:3778917"->{
  Cell[176303, 4831, 908, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:3778917", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat 1"},
   CellID->3778917]},
 "Nat 1"->{
  Cell[176303, 4831, 908, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:3778917", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat 1"},
   CellID->3778917]},
 "ID:54060264"->{
  Cell[177214, 4858, 970, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:54060264", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat Pred"},
   CellID->54060264]},
 "Nat Pred"->{
  Cell[177214, 4858, 970, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:54060264", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat Pred"},
   CellID->54060264]},
 "ID:205601935"->{
  Cell[178187, 4888, 1066, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:205601935", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat +"},
   CellID->205601935]},
 "Nat +"->{
  Cell[178187, 4888, 1066, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:205601935", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat +"},
   CellID->205601935]},
 "ID:74238769"->{
  Cell[179256, 4922, 1011, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:74238769", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat -"},
   CellID->74238769]},
 "Nat -"->{
  Cell[179256, 4922, 1011, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:74238769", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat -"},
   CellID->74238769]},
 "ID:288612943"->{
  Cell[180270, 4955, 1273, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:288612943", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat \[RawStar]"},
   CellID->288612943]},
 "Nat \[RawStar]"->{
  Cell[180270, 4955, 1273, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:288612943", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat \[RawStar]"},
   CellID->288612943]},
 "ID:155532301"->{
  Cell[181546, 4997, 1659, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:155532301", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat \[GreaterEqual]"},
   CellID->155532301]},
 "Nat \[GreaterEqual]"->{
  Cell[181546, 4997, 1659, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:155532301", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat \[GreaterEqual]"},
   CellID->155532301]},
 "ID:15394254"->{
  Cell[184310, 5083, 748, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:15394254", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat isDomain"},
   CellID->15394254]},
 "Nat isDomain"->{
  Cell[184310, 5083, 748, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:15394254", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat isDomain"},
   CellID->15394254]},
 "ID:478904098"->{
  Cell[185061, 5105, 1924, 62, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:478904098", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat induction", "IR"},
   CellID->478904098]},
 "Nat induction"->{
  Cell[185061, 5105, 1924, 62, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:478904098", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat induction", "IR"},
   CellID->478904098]},
 "ID:60850406"->{
  Cell[186988, 5169, 832, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:60850406", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 \[Element] Nat", "IR"},
   CellID->60850406]},
 "0 \[Element] Nat"->{
  Cell[186988, 5169, 832, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:60850406", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 \[Element] Nat", "IR"},
   CellID->60850406]},
 "ID:114723517"->{
  Cell[187823, 5194, 1033, 29, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:114723517", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ \[Element] Nat", "IR"},
   CellID->114723517]},
 "succ \[Element] Nat"->{
  Cell[187823, 5194, 1033, 29, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:114723517", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ \[Element] Nat", "IR"},
   CellID->114723517]},
 "ID:275276526"->{
  Cell[188859, 5225, 831, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:275276526", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[Element] Nat", "IR"},
   CellID->275276526]},
 "1 \[Element] Nat"->{
  Cell[188859, 5225, 831, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:275276526", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[Element] Nat", "IR"},
   CellID->275276526]},
 "ID:65979872"->{
  Cell[189693, 5250, 691, 18, 41, "GlobalDeclaration",
   CellTags->{
    "ID:65979872", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->65979872]},
 "ID:572443783"->{
  Cell[190387, 5270, 770, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:572443783", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat shape"},
   CellID->572443783]},
 "Nat shape"->{
  Cell[190387, 5270, 770, 21, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:572443783", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat shape"},
   CellID->572443783]},
 "ID:252333509"->{
  Cell[191160, 5293, 1442, 43, 66, "FormalTextInputFormula",
   CellTags->{
    "ID:252333509", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat cases"},
   CellID->252333509]},
 "Nat cases"->{
  Cell[191160, 5293, 1442, 43, 66, "FormalTextInputFormula",
   CellTags->{
    "ID:252333509", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat cases"},
   CellID->252333509]},
 "ID:77756457"->{
  Cell[192605, 5338, 846, 23, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:77756457", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ \[NotEqual] 0 Nat"},
   CellID->77756457]},
 "succ \[NotEqual] 0 Nat"->{
  Cell[192605, 5338, 846, 23, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:77756457", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ \[NotEqual] 0 Nat"},
   CellID->77756457]},
 "ID:52299395"->{
  Cell[193454, 5363, 1116, 34, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:52299395", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ injective Nat"},
   CellID->52299395]},
 "succ injective Nat"->{
  Cell[193454, 5363, 1116, 34, 41, "FormalTextInputFormula",
   CellTags->{
    "ID:52299395", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ injective Nat"},
   CellID->52299395]},
 "ID:263281689"->{
  Cell[194573, 5399, 796, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:263281689", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ no fixpoint Nat"},
   CellID->263281689]},
 "succ no fixpoint Nat"->{
  Cell[194573, 5399, 796, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:263281689", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ no fixpoint Nat"},
   CellID->263281689]},
 "ID:377798869"->{
  Cell[195372, 5423, 817, 22, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:377798869", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[NotEqual] 0 Nat"},
   CellID->377798869]},
 "1 \[NotEqual] 0 Nat"->{
  Cell[195372, 5423, 817, 22, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:377798869", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[NotEqual] 0 Nat"},
   CellID->377798869]},
 "ID:33739458"->{
  Cell[196192, 5447, 943, 26, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:33739458", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[Element] closed Nat"},
   CellID->33739458]},
 "\[Element] closed Nat"->{
  Cell[196192, 5447, 943, 26, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:33739458", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[Element] closed Nat"},
   CellID->33739458]},
 "ID:654854303"->{
  Cell[197138, 5475, 1138, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:654854303", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat transitive"},
   CellID->654854303]},
 "Nat transitive"->{
  Cell[197138, 5475, 1138, 27, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:654854303", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat transitive"},
   CellID->654854303]},
 "Proof|ID:15394254"->{
  Cell[198432, 5513, 1703, 54, 28, "ProofInfo",
   CellTags->{"Proof|ID:15394254", "Proof|ID:15394254-1"},
   CellID->198699355]},
 "Proof|ID:15394254-1"->{
  Cell[198432, 5513, 1703, 54, 28, "ProofInfo",
   CellTags->{"Proof|ID:15394254", "Proof|ID:15394254-1"},
   CellID->198699355]},
 "Proof|ID:478904098"->{
  Cell[200654, 5586, 4374, 118, 26, "ProofInfo",
   CellTags->{"Proof|ID:478904098", "Proof|ID:478904098-1"},
   CellID->65480230]},
 "Proof|ID:478904098-1"->{
  Cell[200654, 5586, 4374, 118, 26, "ProofInfo",
   CellTags->{"Proof|ID:478904098", "Proof|ID:478904098-1"},
   CellID->65480230]},
 "Proof|ID:60850406"->{
  Cell[205546, 5723, 1843, 58, 26, "ProofInfo",
   CellTags->{"Proof|ID:60850406", "Proof|ID:60850406-1"},
   CellID->291459920]},
 "Proof|ID:60850406-1"->{
  Cell[205546, 5723, 1843, 58, 26, "ProofInfo",
   CellTags->{"Proof|ID:60850406", "Proof|ID:60850406-1"},
   CellID->291459920]},
 "Proof|ID:114723517"->{
  Cell[207907, 5800, 2421, 72, 26, "ProofInfo",
   CellTags->{"Proof|ID:114723517", "Proof|ID:114723517-1"},
   CellID->310507632]},
 "Proof|ID:114723517-1"->{
  Cell[207907, 5800, 2421, 72, 26, "ProofInfo",
   CellTags->{"Proof|ID:114723517", "Proof|ID:114723517-1"},
   CellID->310507632]},
 "Proof|ID:275276526"->{
  Cell[210848, 5891, 1847, 58, 26, "ProofInfo",
   CellTags->{"Proof|ID:275276526", "Proof|ID:275276526-1"},
   CellID->58288735]},
 "Proof|ID:275276526-1"->{
  Cell[210848, 5891, 1847, 58, 26, "ProofInfo",
   CellTags->{"Proof|ID:275276526", "Proof|ID:275276526-1"},
   CellID->58288735]},
 "Proof|ID:572443783"->{
  Cell[213215, 5968, 2236, 68, 26, "ProofInfo",
   CellTags->{"Proof|ID:572443783", "Proof|ID:572443783-1"},
   CellID->43340428]},
 "Proof|ID:572443783-1"->{
  Cell[213215, 5968, 2236, 68, 26, "ProofInfo",
   CellTags->{"Proof|ID:572443783", "Proof|ID:572443783-1"},
   CellID->43340428]},
 "Proof|ID:252333509"->{
  Cell[215970, 6055, 3251, 92, 26, "ProofInfo",
   CellTags->{"Proof|ID:252333509", "Proof|ID:252333509-1"},
   CellID->179637041]},
 "Proof|ID:252333509-1"->{
  Cell[215970, 6055, 3251, 92, 26, "ProofInfo",
   CellTags->{"Proof|ID:252333509", "Proof|ID:252333509-1"},
   CellID->179637041]},
 "Proof|ID:77756457"->{
  Cell[219741, 6166, 2402, 72, 26, "ProofInfo",
   CellTags->{"Proof|ID:77756457", "Proof|ID:77756457-1"},
   CellID->41859577]},
 "Proof|ID:77756457-1"->{
  Cell[219741, 6166, 2402, 72, 26, "ProofInfo",
   CellTags->{"Proof|ID:77756457", "Proof|ID:77756457-1"},
   CellID->41859577]},
 "Proof|ID:52299395"->{
  Cell[222662, 6257, 3163, 90, 26, "ProofInfo",
   CellTags->{"Proof|ID:52299395", "Proof|ID:52299395-1"},
   CellID->590942562]},
 "Proof|ID:52299395-1"->{
  Cell[222662, 6257, 3163, 90, 26, "ProofInfo",
   CellTags->{"Proof|ID:52299395", "Proof|ID:52299395-1"},
   CellID->590942562]},
 "Proof|ID:263281689"->{
  Cell[226344, 6366, 2446, 74, 26, "ProofInfo",
   CellTags->{"Proof|ID:263281689", "Proof|ID:263281689-1"},
   CellID->255034219]},
 "Proof|ID:263281689-1"->{
  Cell[226344, 6366, 2446, 74, 26, "ProofInfo",
   CellTags->{"Proof|ID:263281689", "Proof|ID:263281689-1"},
   CellID->255034219]},
 "Proof|ID:377798869"->{
  Cell[229310, 6459, 1846, 58, 26, "ProofInfo",
   CellTags->{"Proof|ID:377798869", "Proof|ID:377798869-1"},
   CellID->103721996]},
 "Proof|ID:377798869-1"->{
  Cell[229310, 6459, 1846, 58, 26, "ProofInfo",
   CellTags->{"Proof|ID:377798869", "Proof|ID:377798869-1"},
   CellID->103721996]},
 "Proof|ID:33739458"->{
  Cell[231676, 6536, 2628, 76, 26, "ProofInfo",
   CellTags->{"Proof|ID:33739458", "Proof|ID:33739458-1"},
   CellID->548186621]},
 "Proof|ID:33739458-1"->{
  Cell[231676, 6536, 2628, 76, 26, "ProofInfo",
   CellTags->{"Proof|ID:33739458", "Proof|ID:33739458-1"},
   CellID->548186621]},
 "Proof|ID:654854303"->{
  Cell[234823, 6631, 2646, 78, 26, "ProofInfo",
   CellTags->{"Proof|ID:654854303", "Proof|ID:654854303-1"},
   CellID->8946125]},
 "Proof|ID:654854303-1"->{
  Cell[234823, 6631, 2646, 78, 26, "ProofInfo",
   CellTags->{"Proof|ID:654854303", "Proof|ID:654854303-1"},
   CellID->8946125]},
 "ID:240175329"->{
  Cell[238422, 6742, 861, 24, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:240175329", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Pred 0 Nat", "IR"},
   CellID->240175329]},
 "Pred 0 Nat"->{
  Cell[238422, 6742, 861, 24, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:240175329", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Pred 0 Nat", "IR"},
   CellID->240175329]},
 "ID:158284383"->{
  Cell[239286, 6768, 666, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:158284383", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->158284383]},
 "ID:145236688"->{
  Cell[239955, 6788, 955, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:145236688", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Pred succ Nat", "IR"},
   CellID->145236688]},
 "Pred succ Nat"->{
  Cell[239955, 6788, 955, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:145236688", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Pred succ Nat", "IR"},
   CellID->145236688]},
 "ID:225607097"->{
  Cell[240913, 6816, 870, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:225607097", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Pred closed Nat", "IR"},
   CellID->225607097]},
 "Pred closed Nat"->{
  Cell[240913, 6816, 870, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:225607097", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Pred closed Nat", "IR"},
   CellID->225607097]},
 "Proof|ID:240175329"->{
  Cell[241937, 6852, 1972, 62, 70, "ProofInfo",
   CellTags->{"Proof|ID:240175329", "Proof|ID:240175329-1"},
   CellID->415645142]},
 "Proof|ID:240175329-1"->{
  Cell[241937, 6852, 1972, 62, 70, "ProofInfo",
   CellTags->{"Proof|ID:240175329", "Proof|ID:240175329-1"},
   CellID->415645142]},
 "Proof|ID:145236688"->{
  Cell[244427, 6933, 2588, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:145236688", "Proof|ID:145236688-1"},
   CellID->498357158]},
 "Proof|ID:145236688-1"->{
  Cell[244427, 6933, 2588, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:145236688", "Proof|ID:145236688-1"},
   CellID->498357158]},
 "Proof|ID:225607097"->{
  Cell[247535, 7030, 2487, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:225607097", "Proof|ID:225607097-1"},
   CellID->787676860]},
 "Proof|ID:225607097-1"->{
  Cell[247535, 7030, 2487, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:225607097", "Proof|ID:225607097-1"},
   CellID->787676860]},
 "ID:218406943"->{
  Cell[251476, 7155, 766, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:218406943", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->218406943]},
 "ID:257436422"->{
  Cell[252245, 7177, 923, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:257436422", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ 0 Nat", "IR"},
   CellID->257436422]},
 "+ 0 Nat"->{
  Cell[252245, 7177, 923, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:257436422", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ 0 Nat", "IR"},
   CellID->257436422]},
 "ID:249353684"->{
  Cell[253171, 7204, 1063, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:249353684", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ succ Nat", "IR"},
   CellID->249353684]},
 "+ succ Nat"->{
  Cell[253171, 7204, 1063, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:249353684", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ succ Nat", "IR"},
   CellID->249353684]},
 "ID:279658160"->{
  Cell[254237, 7234, 850, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:279658160", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ closed Nat", "IR"},
   CellID->279658160]},
 "+ closed Nat"->{
  Cell[254237, 7234, 850, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:279658160", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ closed Nat", "IR"},
   CellID->279658160]},
 "ID:647623443"->{
  Cell[255090, 7260, 899, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:647623443", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ 1 Nat"},
   CellID->647623443]},
 "+ 1 Nat"->{
  Cell[255090, 7260, 899, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:647623443", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ 1 Nat"},
   CellID->647623443]},
 "ID:18327536"->{
  Cell[255992, 7287, 939, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:18327536", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 lneutral Nat", "IR"},
   CellID->18327536]},
 "0 lneutral Nat"->{
  Cell[255992, 7287, 939, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:18327536", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 lneutral Nat", "IR"},
   CellID->18327536]},
 "ID:1914645"->{
  Cell[256934, 7315, 956, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1914645", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ + Nat", "IR"},
   CellID->1914645]},
 "succ + Nat"->{
  Cell[256934, 7315, 956, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1914645", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ + Nat", "IR"},
   CellID->1914645]},
 "ID:343580194"->{
  Cell[257893, 7344, 1190, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:343580194", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ = 0 Nat"},
   CellID->343580194]},
 "+ = 0 Nat"->{
  Cell[257893, 7344, 1190, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:343580194", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ = 0 Nat"},
   CellID->343580194]},
 "ID:834229092"->{
  Cell[259086, 7385, 889, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:834229092", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ comm Nat"},
   CellID->834229092]},
 "+ comm Nat"->{
  Cell[259086, 7385, 889, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:834229092", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ comm Nat"},
   CellID->834229092]},
 "ID:303913216"->{
  Cell[259978, 7412, 668, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:303913216", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->303913216]},
 "ID:73705609"->{
  Cell[260649, 7432, 1094, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:73705609", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ assoc Nat"},
   CellID->73705609]},
 "+ assoc Nat"->{
  Cell[260649, 7432, 1094, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:73705609", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ assoc Nat"},
   CellID->73705609]},
 "ID:202544617"->{
  Cell[261746, 7466, 1008, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:202544617", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ comm-assoc Nat"},
   CellID->202544617]},
 "+ comm-assoc Nat"->{
  Cell[261746, 7466, 1008, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:202544617", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ comm-assoc Nat"},
   CellID->202544617]},
 "ID:555281086"->{
  Cell[262757, 7498, 1407, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:555281086", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ + + Nat"},
   CellID->555281086]},
 "+ + + Nat"->{
  Cell[262757, 7498, 1407, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:555281086", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ + + Nat"},
   CellID->555281086]},
 "ID:5229733"->{
  Cell[264167, 7546, 1147, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:5229733", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ rcanc Nat"},
   CellID->5229733]},
 "+ rcanc Nat"->{
  Cell[264167, 7546, 1147, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:5229733", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ rcanc Nat"},
   CellID->5229733]},
 "Proof|ID:257436422"->{
  Cell[265469, 7594, 2546, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:257436422", "Proof|ID:257436422-1"},
   CellID->197432916]},
 "Proof|ID:257436422-1"->{
  Cell[265469, 7594, 2546, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:257436422", "Proof|ID:257436422-1"},
   CellID->197432916]},
 "Proof|ID:249353684"->{
  Cell[268535, 7691, 3263, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:249353684", "Proof|ID:249353684-1"},
   CellID->376294367]},
 "Proof|ID:249353684-1"->{
  Cell[268535, 7691, 3263, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:249353684", "Proof|ID:249353684-1"},
   CellID->376294367]},
 "Proof|ID:279658160"->{
  Cell[272318, 7802, 2841, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:279658160", "Proof|ID:279658160-1"},
   CellID->363010569]},
 "Proof|ID:279658160-1"->{
  Cell[272318, 7802, 2841, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:279658160", "Proof|ID:279658160-1"},
   CellID->363010569]},
 "Proof|ID:647623443"->{
  Cell[275679, 7903, 2620, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:647623443", "Proof|ID:647623443-1"},
   CellID->35382149]},
 "Proof|ID:647623443-1"->{
  Cell[275679, 7903, 2620, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:647623443", "Proof|ID:647623443-1"},
   CellID->35382149]},
 "Proof|ID:18327536"->{
  Cell[278819, 8000, 2555, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:18327536", "Proof|ID:18327536-1"},
   CellID->794865037]},
 "Proof|ID:18327536-1"->{
  Cell[278819, 8000, 2555, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:18327536", "Proof|ID:18327536-1"},
   CellID->794865037]},
 "Proof|ID:1914645"->{
  Cell[281892, 8097, 3253, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:1914645", "Proof|ID:1914645-1"},
   CellID->218839119]},
 "Proof|ID:1914645-1"->{
  Cell[281892, 8097, 3253, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:1914645", "Proof|ID:1914645-1"},
   CellID->218839119]},
 "Proof|ID:343580194"->{
  Cell[285663, 8208, 3508, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:343580194", "Proof|ID:343580194-1"},
   CellID->734595749]},
 "Proof|ID:343580194-1"->{
  Cell[285663, 8208, 3508, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:343580194", "Proof|ID:343580194-1"},
   CellID->734595749]},
 "Proof|ID:834229092"->{
  Cell[289691, 8325, 3091, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:834229092", "Proof|ID:834229092-1"},
   CellID->12781957]},
 "Proof|ID:834229092-1"->{
  Cell[289691, 8325, 3091, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:834229092", "Proof|ID:834229092-1"},
   CellID->12781957]},
 "Proof|ID:73705609"->{
  Cell[293302, 8434, 3865, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:73705609", "Proof|ID:73705609-1"},
   CellID->337185825]},
 "Proof|ID:73705609-1"->{
  Cell[293302, 8434, 3865, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:73705609", "Proof|ID:73705609-1"},
   CellID->337185825]},
 "Proof|ID:202544617"->{
  Cell[297685, 8561, 3877, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:202544617", "Proof|ID:202544617-1"},
   CellID->299537]},
 "Proof|ID:202544617-1"->{
  Cell[297685, 8561, 3877, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:202544617", "Proof|ID:202544617-1"},
   CellID->299537]},
 "Proof|ID:555281086"->{
  Cell[302082, 8688, 4641, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:555281086", "Proof|ID:555281086-1"},
   CellID->94141421]},
 "Proof|ID:555281086-1"->{
  Cell[302082, 8688, 4641, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:555281086", "Proof|ID:555281086-1"},
   CellID->94141421]},
 "Proof|ID:5229733"->{
  Cell[307242, 8833, 3744, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:5229733", "Proof|ID:5229733-1"},
   CellID->540060300]},
 "Proof|ID:5229733-1"->{
  Cell[307242, 8833, 3744, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:5229733", "Proof|ID:5229733-1"},
   CellID->540060300]},
 "ID:301992507"->{
  Cell[312012, 8975, 717, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:301992507", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->301992507]},
 "ID:99163941"->{
  Cell[312732, 8996, 895, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:99163941", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] 0 Nat", "IR"},
   CellID->99163941]},
 "\[RawStar] 0 Nat"->{
  Cell[312732, 8996, 895, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:99163941", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] 0 Nat", "IR"},
   CellID->99163941]},
 "ID:371484270"->{
  Cell[313630, 9023, 987, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:371484270", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] succ Nat", "IR"},
   CellID->371484270]},
 "\[RawStar] succ Nat"->{
  Cell[313630, 9023, 987, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:371484270", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] succ Nat", "IR"},
   CellID->371484270]},
 "ID:142418160"->{
  Cell[314620, 9053, 852, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:142418160", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] closed Nat", "IR"},
   CellID->142418160]},
 "\[RawStar] closed Nat"->{
  Cell[314620, 9053, 852, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:142418160", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] closed Nat", "IR"},
   CellID->142418160]},
 "ID:21383846"->{
  Cell[315475, 9079, 832, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:21383846", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 rneutral Nat", "IR"},
   CellID->21383846]},
 "1 rneutral Nat"->{
  Cell[315475, 9079, 832, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:21383846", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 rneutral Nat", "IR"},
   CellID->21383846]},
 "ID:28953873"->{
  Cell[316310, 9104, 864, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:28953873", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 lannihilator Nat", "IR"},
   CellID->28953873]},
 "0 lannihilator Nat"->{
  Cell[316310, 9104, 864, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:28953873", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 lannihilator Nat", "IR"},
   CellID->28953873]},
 "ID:218775165"->{
  Cell[317177, 9130, 981, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:218775165", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ \[RawStar] Nat", "IR"},
   CellID->218775165]},
 "succ \[RawStar] Nat"->{
  Cell[317177, 9130, 981, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:218775165", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ \[RawStar] Nat", "IR"},
   CellID->218775165]},
 "ID:470439981"->{
  Cell[318161, 9160, 1191, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:470439981", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] zdf Nat"},
   CellID->470439981]},
 "\[RawStar] zdf Nat"->{
  Cell[318161, 9160, 1191, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:470439981", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] zdf Nat"},
   CellID->470439981]},
 "ID:218136211"->{
  Cell[319355, 9201, 842, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:218136211", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] comm Nat"},
   CellID->218136211]},
 "\[RawStar] comm Nat"->{
  Cell[319355, 9201, 842, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:218136211", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] comm Nat"},
   CellID->218136211]},
 "ID:666029340"->{
  Cell[320200, 9227, 690, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:666029340", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->666029340]},
 "ID:43150914"->{
  Cell[320893, 9247, 1080, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:43150914", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] distr + Nat"},
   CellID->43150914]},
 "\[RawStar] distr + Nat"->{
  Cell[320893, 9247, 1080, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:43150914", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] distr + Nat"},
   CellID->43150914]},
 "ID:475421150"->{
  Cell[321976, 9281, 1000, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:475421150", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] assoc Nat"},
   CellID->475421150]},
 "\[RawStar] assoc Nat"->{
  Cell[321976, 9281, 1000, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:475421150", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] assoc Nat"},
   CellID->475421150]},
 "ID:98508460"->{
  Cell[322979, 9313, 1007, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:98508460", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] comm-assoc Nat"},
   CellID->98508460]},
 "\[RawStar] comm-assoc Nat"->{
  Cell[322979, 9313, 1007, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:98508460", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] comm-assoc Nat"},
   CellID->98508460]},
 "ID:138249572"->{
  Cell[323989, 9345, 1194, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:138249572", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] rcanc Nat"},
   CellID->138249572]},
 "\[RawStar] rcanc Nat"->{
  Cell[323989, 9345, 1194, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:138249572", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] rcanc Nat"},
   CellID->138249572]},
 "Proof|ID:99163941"->{
  Cell[325337, 9395, 2499, 76, 70, "ProofInfo",
   CellTags->{"Proof|ID:99163941", "Proof|ID:99163941-1"},
   CellID->138225265]},
 "Proof|ID:99163941-1"->{
  Cell[325337, 9395, 2499, 76, 70, "ProofInfo",
   CellTags->{"Proof|ID:99163941", "Proof|ID:99163941-1"},
   CellID->138225265]},
 "Proof|ID:371484270"->{
  Cell[328355, 9490, 3424, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:371484270", "Proof|ID:371484270-1"},
   CellID->2648213]},
 "Proof|ID:371484270-1"->{
  Cell[328355, 9490, 3424, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:371484270", "Proof|ID:371484270-1"},
   CellID->2648213]},
 "Proof|ID:142418160"->{
  Cell[332299, 9607, 2840, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:142418160", "Proof|ID:142418160-1"},
   CellID->32519450]},
 "Proof|ID:142418160-1"->{
  Cell[332299, 9607, 2840, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:142418160", "Proof|ID:142418160-1"},
   CellID->32519450]},
 "Proof|ID:21383846"->{
  Cell[335659, 9708, 2554, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:21383846", "Proof|ID:21383846-1"},
   CellID->16031556]},
 "Proof|ID:21383846-1"->{
  Cell[335659, 9708, 2554, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:21383846", "Proof|ID:21383846-1"},
   CellID->16031556]},
 "Proof|ID:28953873"->{
  Cell[338732, 9805, 2520, 76, 70, "ProofInfo",
   CellTags->{"Proof|ID:28953873", "Proof|ID:28953873-1"},
   CellID->86840982]},
 "Proof|ID:28953873-1"->{
  Cell[338732, 9805, 2520, 76, 70, "ProofInfo",
   CellTags->{"Proof|ID:28953873", "Proof|ID:28953873-1"},
   CellID->86840982]},
 "Proof|ID:218775165"->{
  Cell[341770, 9900, 3425, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:218775165", "Proof|ID:218775165-1"},
   CellID->82184287]},
 "Proof|ID:218775165-1"->{
  Cell[341770, 9900, 3425, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:218775165", "Proof|ID:218775165-1"},
   CellID->82184287]},
 "Proof|ID:470439981"->{
  Cell[345715, 10017, 3506, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:470439981", "Proof|ID:470439981-1"},
   CellID->167271436]},
 "Proof|ID:470439981-1"->{
  Cell[345715, 10017, 3506, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:470439981", "Proof|ID:470439981-1"},
   CellID->167271436]},
 "Proof|ID:218136211"->{
  Cell[349740, 10134, 3092, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:218136211", "Proof|ID:218136211-1"},
   CellID->116241402]},
 "Proof|ID:218136211-1"->{
  Cell[349740, 10134, 3092, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:218136211", "Proof|ID:218136211-1"},
   CellID->116241402]},
 "Proof|ID:43150914"->{
  Cell[353352, 10243, 4114, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:43150914", "Proof|ID:43150914-1"},
   CellID->92944893]},
 "Proof|ID:43150914-1"->{
  Cell[353352, 10243, 4114, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:43150914", "Proof|ID:43150914-1"},
   CellID->92944893]},
 "Proof|ID:475421150"->{
  Cell[357985, 10376, 3870, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:475421150", "Proof|ID:475421150-1"},
   CellID->574072770]},
 "Proof|ID:475421150-1"->{
  Cell[357985, 10376, 3870, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:475421150", "Proof|ID:475421150-1"},
   CellID->574072770]},
 "Proof|ID:98508460"->{
  Cell[362375, 10503, 3874, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:98508460", "Proof|ID:98508460-1"},
   CellID->81244651]},
 "Proof|ID:98508460-1"->{
  Cell[362375, 10503, 3874, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:98508460", "Proof|ID:98508460-1"},
   CellID->81244651]},
 "Proof|ID:138249572"->{
  Cell[366768, 10630, 4076, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:138249572", "Proof|ID:138249572-1"},
   CellID->165947681]},
 "Proof|ID:138249572-1"->{
  Cell[366768, 10630, 4076, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:138249572", "Proof|ID:138249572-1"},
   CellID->165947681]},
 "ID:621974620"->{
  Cell[372145, 10782, 715, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:621974620", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->621974620]},
 "ID:380981542"->{
  Cell[372863, 10803, 1126, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:380981542", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 \[GreaterEqual] Nat", "IR"},
   CellID->380981542]},
 "0 \[GreaterEqual] Nat"->{
  Cell[372863, 10803, 1126, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:380981542", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 \[GreaterEqual] Nat", "IR"},
   CellID->380981542]},
 "ID:42977349"->{
  Cell[373992, 10840, 1292, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:42977349", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ \[GreaterEqual] Nat", "IR"},
   CellID->42977349]},
 "succ \[GreaterEqual] Nat"->{
  Cell[373992, 10840, 1292, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:42977349", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ \[GreaterEqual] Nat", "IR"},
   CellID->42977349]},
 "ID:428740350"->{
  Cell[375287, 10881, 864, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:428740350", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] reflexive Nat"},
   CellID->428740350]},
 "\[GreaterEqual] reflexive Nat"->{
  Cell[375287, 10881, 864, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:428740350", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] reflexive Nat"},
   CellID->428740350]},
 "ID:297686800"->{
  Cell[376154, 10905, 882, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:297686800", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] 0 Nat", "IR"},
   CellID->297686800]},
 "\[GreaterEqual] 0 Nat"->{
  Cell[376154, 10905, 882, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:297686800", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] 0 Nat", "IR"},
   CellID->297686800]},
 "ID:585961470"->{
  Cell[377039, 10930, 1316, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:585961470", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] succ Nat", "IR"},
   CellID->585961470]},
 "\[GreaterEqual] succ Nat"->{
  Cell[377039, 10930, 1316, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:585961470", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] succ Nat", "IR"},
   CellID->585961470]},
 "ID:894887193"->{
  Cell[378358, 10971, 1080, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:894887193", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] antisymmetric Nat"},
   CellID->894887193]},
 "\[GreaterEqual] antisymmetric Nat"->{
  Cell[378358, 10971, 1080, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:894887193", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] antisymmetric Nat"},
   CellID->894887193]},
 "ID:208533627"->{
  Cell[379441, 11004, 875, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:208533627", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ increasing Nat"},
   CellID->208533627]},
 "succ increasing Nat"->{
  Cell[379441, 11004, 875, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:208533627", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ increasing Nat"},
   CellID->208533627]},
 "ID:112605784"->{
  Cell[380319, 11029, 1239, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:112605784", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ \[GreaterEqual] succ Nat", "IR"},
   CellID->112605784]},
 "succ \[GreaterEqual] succ Nat"->{
  Cell[380319, 11029, 1239, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:112605784", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ \[GreaterEqual] succ Nat", "IR"},
   CellID->112605784]},
 "ID:304315076"->{
  Cell[381561, 11068, 901, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:304315076", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] total Nat"},
   CellID->304315076]},
 "\[GreaterEqual] total Nat"->{
  Cell[381561, 11068, 901, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:304315076", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] total Nat"},
   CellID->304315076]},
 "ID:446896782"->{
  Cell[382465, 11095, 1167, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:446896782", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] \[Exists] Nat"},
   CellID->446896782]},
 "\[GreaterEqual] \[Exists] Nat"->{
  Cell[382465, 11095, 1167, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:446896782", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] \[Exists] Nat"},
   CellID->446896782]},
 "ID:196203313"->{
  Cell[383635, 11132, 853, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:196203313", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[GreaterEqual] 0 Nat"},
   CellID->196203313]},
 "1 \[GreaterEqual] 0 Nat"->{
  Cell[383635, 11132, 853, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:196203313", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[GreaterEqual] 0 Nat"},
   CellID->196203313]},
 "ID:30448165"->{
  Cell[384491, 11157, 1179, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:30448165", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] 1 Nat", "IR"},
   CellID->30448165]},
 "\[GreaterEqual] 1 Nat"->{
  Cell[384491, 11157, 1179, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:30448165", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] 1 Nat", "IR"},
   CellID->30448165]},
 "ID:222994735"->{
  Cell[385673, 11195, 690, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:222994735", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->222994735]},
 "ID:546802157"->{
  Cell[386366, 11215, 1107, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:546802157", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] transitive Nat"},
   CellID->546802157]},
 "\[GreaterEqual] transitive Nat"->{
  Cell[386366, 11215, 1107, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:546802157", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] transitive Nat"},
   CellID->546802157]},
 "ID:352201475"->{
  Cell[387476, 11250, 1045, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:352201475", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] + \[DoubleLeftArrow] Nat"},
   CellID->352201475]},
 "\[GreaterEqual] + \[DoubleLeftArrow] Nat"->{
  Cell[387476, 11250, 1045, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:352201475", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] + \[DoubleLeftArrow] Nat"},
   CellID->352201475]},
 "ID:97128652"->{
  Cell[388524, 11281, 1208, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:97128652", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] + Nat"},
   CellID->97128652]},
 "\[GreaterEqual] + Nat"->{
  Cell[388524, 11281, 1208, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:97128652", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] + Nat"},
   CellID->97128652]},
 "ID:278931212"->{
  Cell[389735, 11321, 849, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:278931212", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ \[GreaterEqual] Nat"},
   CellID->278931212]},
 "+ \[GreaterEqual] Nat"->{
  Cell[389735, 11321, 849, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:278931212", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ \[GreaterEqual] Nat"},
   CellID->278931212]},
 "ID:335211897"->{
  Cell[390587, 11346, 1000, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:335211897", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] \[RawStar] Nat"},
   CellID->335211897]},
 "\[GreaterEqual] \[RawStar] Nat"->{
  Cell[390587, 11346, 1000, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:335211897", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] \[RawStar] Nat"},
   CellID->335211897]},
 "ID:32863246"->{
  Cell[391590, 11376, 943, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:32863246", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] \[GreaterEqual] Nat"},
   CellID->32863246]},
 "\[RawStar] \[GreaterEqual] Nat"->{
  Cell[391590, 11376, 943, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:32863246", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] \[GreaterEqual] Nat"},
   CellID->32863246]},
 "Proof|ID:380981542"->{
  Cell[392689, 11414, 2764, 80, 70, "ProofInfo",
   CellTags->{"Proof|ID:380981542", "Proof|ID:380981542-1"},
   CellID->159585745]},
 "Proof|ID:380981542-1"->{
  Cell[392689, 11414, 2764, 80, 70, "ProofInfo",
   CellTags->{"Proof|ID:380981542", "Proof|ID:380981542-1"},
   CellID->159585745]},
 "Proof|ID:42977349"->{
  Cell[395972, 11513, 3830, 102, 70, "ProofInfo",
   CellTags->{"Proof|ID:42977349", "Proof|ID:42977349-1"},
   CellID->255247870]},
 "Proof|ID:42977349-1"->{
  Cell[395972, 11513, 3830, 102, 70, "ProofInfo",
   CellTags->{"Proof|ID:42977349", "Proof|ID:42977349-1"},
   CellID->255247870]},
 "Proof|ID:428740350"->{
  Cell[400320, 11634, 2457, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:428740350", "Proof|ID:428740350-1"},
   CellID->34057048]},
 "Proof|ID:428740350-1"->{
  Cell[400320, 11634, 2457, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:428740350", "Proof|ID:428740350-1"},
   CellID->34057048]},
 "Proof|ID:297686800"->{
  Cell[403297, 11727, 2399, 72, 70, "ProofInfo",
   CellTags->{"Proof|ID:297686800", "Proof|ID:297686800-1"},
   CellID->45781616]},
 "Proof|ID:297686800-1"->{
  Cell[403297, 11727, 2399, 72, 70, "ProofInfo",
   CellTags->{"Proof|ID:297686800", "Proof|ID:297686800-1"},
   CellID->45781616]},
 "Proof|ID:585961470"->{
  Cell[406216, 11818, 3771, 102, 70, "ProofInfo",
   CellTags->{"Proof|ID:585961470", "Proof|ID:585961470-1"},
   CellID->252534163]},
 "Proof|ID:585961470-1"->{
  Cell[406216, 11818, 3771, 102, 70, "ProofInfo",
   CellTags->{"Proof|ID:585961470", "Proof|ID:585961470-1"},
   CellID->252534163]},
 "Proof|ID:894887193"->{
  Cell[410506, 11939, 3574, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:894887193", "Proof|ID:894887193-1"},
   CellID->246933219]},
 "Proof|ID:894887193-1"->{
  Cell[410506, 11939, 3574, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:894887193", "Proof|ID:894887193-1"},
   CellID->246933219]},
 "Proof|ID:208533627"->{
  Cell[414600, 12057, 2521, 76, 70, "ProofInfo",
   CellTags->{"Proof|ID:208533627", "Proof|ID:208533627-1"},
   CellID->10870655]},
 "Proof|ID:208533627-1"->{
  Cell[414600, 12057, 2521, 76, 70, "ProofInfo",
   CellTags->{"Proof|ID:208533627", "Proof|ID:208533627-1"},
   CellID->10870655]},
 "Proof|ID:112605784"->{
  Cell[417640, 12152, 3416, 96, 70, "ProofInfo",
   CellTags->{"Proof|ID:112605784", "Proof|ID:112605784-1"},
   CellID->362360428]},
 "Proof|ID:112605784-1"->{
  Cell[417640, 12152, 3416, 96, 70, "ProofInfo",
   CellTags->{"Proof|ID:112605784", "Proof|ID:112605784-1"},
   CellID->362360428]},
 "Proof|ID:304315076"->{
  Cell[421576, 12267, 3172, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:304315076", "Proof|ID:304315076-1"},
   CellID->318982970]},
 "Proof|ID:304315076-1"->{
  Cell[421576, 12267, 3172, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:304315076", "Proof|ID:304315076-1"},
   CellID->318982970]},
 "Proof|ID:446896782"->{
  Cell[425268, 12376, 3861, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:446896782", "Proof|ID:446896782-1"},
   CellID->239855888]},
 "Proof|ID:446896782-1"->{
  Cell[425268, 12376, 3861, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:446896782", "Proof|ID:446896782-1"},
   CellID->239855888]},
 "Proof|ID:196203313"->{
  Cell[429649, 12501, 1927, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:196203313", "Proof|ID:196203313-1"},
   CellID->28380975]},
 "Proof|ID:196203313-1"->{
  Cell[429649, 12501, 1927, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:196203313", "Proof|ID:196203313-1"},
   CellID->28380975]},
 "Proof|ID:30448165"->{
  Cell[432095, 12580, 2764, 80, 70, "ProofInfo",
   CellTags->{"Proof|ID:30448165", "Proof|ID:30448165-1"},
   CellID->64482358]},
 "Proof|ID:30448165-1"->{
  Cell[432095, 12580, 2764, 80, 70, "ProofInfo",
   CellTags->{"Proof|ID:30448165", "Proof|ID:30448165-1"},
   CellID->64482358]},
 "Proof|ID:546802157"->{
  Cell[435378, 12679, 3940, 107, 70, "ProofInfo",
   CellTags->{"Proof|ID:546802157", "Proof|ID:546802157-1"},
   CellID->527891345]},
 "Proof|ID:546802157-1"->{
  Cell[435378, 12679, 3940, 107, 70, "ProofInfo",
   CellTags->{"Proof|ID:546802157", "Proof|ID:546802157-1"},
   CellID->527891345]},
 "Proof|ID:352201475"->{
  Cell[439838, 12805, 4003, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:352201475", "Proof|ID:352201475-1"},
   CellID->76510990]},
 "Proof|ID:352201475-1"->{
  Cell[439838, 12805, 4003, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:352201475", "Proof|ID:352201475-1"},
   CellID->76510990]},
 "Proof|ID:97128652"->{
  Cell[444361, 12934, 3995, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:97128652", "Proof|ID:97128652-1"},
   CellID->393536374]},
 "Proof|ID:97128652-1"->{
  Cell[444361, 12934, 3995, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:97128652", "Proof|ID:97128652-1"},
   CellID->393536374]},
 "Proof|ID:278931212"->{
  Cell[448875, 13063, 2960, 86, 70, "ProofInfo",
   CellTags->{"Proof|ID:278931212", "Proof|ID:278931212-1"},
   CellID->260422124]},
 "Proof|ID:278931212-1"->{
  Cell[448875, 13063, 2960, 86, 70, "ProofInfo",
   CellTags->{"Proof|ID:278931212", "Proof|ID:278931212-1"},
   CellID->260422124]},
 "Proof|ID:335211897"->{
  Cell[452355, 13168, 3950, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:335211897", "Proof|ID:335211897-1"},
   CellID->171667115]},
 "Proof|ID:335211897-1"->{
  Cell[452355, 13168, 3950, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:335211897", "Proof|ID:335211897-1"},
   CellID->171667115]},
 "Proof|ID:32863246"->{
  Cell[456825, 13295, 3315, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:32863246", "Proof|ID:32863246-1"},
   CellID->145909442]},
 "Proof|ID:32863246-1"->{
  Cell[456825, 13295, 3315, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:32863246", "Proof|ID:32863246-1"},
   CellID->145909442]},
 "ID:609763659"->{
  Cell[461486, 13436, 717, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:609763659", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->609763659]},
 "ID:19208428"->{
  Cell[462206, 13457, 816, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:19208428", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- 0 Nat", "IR"},
   CellID->19208428]},
 "- 0 Nat"->{
  Cell[462206, 13457, 816, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:19208428", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- 0 Nat", "IR"},
   CellID->19208428]},
 "ID:593421999"->{
  Cell[463025, 13482, 987, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:593421999", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- succ Nat", "IR"},
   CellID->593421999]},
 "- succ Nat"->{
  Cell[463025, 13482, 987, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:593421999", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- succ Nat", "IR"},
   CellID->593421999]},
 "ID:153357458"->{
  Cell[464015, 13512, 848, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:153357458", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- closed Nat", "IR"},
   CellID->153357458]},
 "- closed Nat"->{
  Cell[464015, 13512, 848, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:153357458", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- closed Nat", "IR"},
   CellID->153357458]},
 "ID:616125500"->{
  Cell[464866, 13538, 871, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:616125500", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- 1 Nat"},
   CellID->616125500]},
 "- 1 Nat"->{
  Cell[464866, 13538, 871, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:616125500", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- 1 Nat"},
   CellID->616125500]},
 "ID:491337343"->{
  Cell[465740, 13565, 845, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:491337343", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 - Nat", "IR"},
   CellID->491337343]},
 "0 - Nat"->{
  Cell[465740, 13565, 845, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:491337343", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 - Nat", "IR"},
   CellID->491337343]},
 "ID:621700658"->{
  Cell[466588, 13591, 1028, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:621700658", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ - succ Nat", "IR"},
   CellID->621700658]},
 "succ - succ Nat"->{
  Cell[466588, 13591, 1028, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:621700658", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "succ - succ Nat", "IR"},
   CellID->621700658]},
 "ID:236021775"->{
  Cell[467619, 13621, 869, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:236021775", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ - Nat"},
   CellID->236021775]},
 "+ - Nat"->{
  Cell[467619, 13621, 869, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:236021775", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ - Nat"},
   CellID->236021775]},
 "ID:3006634"->{
  Cell[468491, 13649, 965, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:3006634", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- + Nat"},
   CellID->3006634]},
 "- + Nat"->{
  Cell[468491, 13649, 965, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:3006634", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- + Nat"},
   CellID->3006634]},
 "ID:80465146"->{
  Cell[469459, 13680, 663, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:80465146", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->80465146]},
 "ID:404996142"->{
  Cell[470125, 13700, 996, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:404996142", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ - + Nat"},
   CellID->404996142]},
 "+ - + Nat"->{
  Cell[470125, 13700, 996, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:404996142", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ - + Nat"},
   CellID->404996142]},
 "ID:225131303"->{
  Cell[471124, 13732, 1106, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:225131303", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] distr - Nat"},
   CellID->225131303]},
 "\[RawStar] distr - Nat"->{
  Cell[471124, 13732, 1106, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:225131303", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] distr - Nat"},
   CellID->225131303]},
 "ID:142079924"->{
  Cell[472233, 13766, 1206, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:142079924", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "(-)+ Nat"},
   CellID->142079924]},
 "(-)+ Nat"->{
  Cell[472233, 13766, 1206, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:142079924", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "(-)+ Nat"},
   CellID->142079924]},
 "ID:367201929"->{
  Cell[473442, 13806, 2138, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:367201929", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- = - Nat"},
   CellID->367201929]},
 "- = - Nat"->{
  Cell[473442, 13806, 2138, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:367201929", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- = - Nat"},
   CellID->367201929]},
 "Proof|ID:19208428"->{
  Cell[475736, 13886, 2540, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:19208428", "Proof|ID:19208428-1"},
   CellID->94129044]},
 "Proof|ID:19208428-1"->{
  Cell[475736, 13886, 2540, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:19208428", "Proof|ID:19208428-1"},
   CellID->94129044]},
 "Proof|ID:593421999"->{
  Cell[478795, 13983, 3337, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:593421999", "Proof|ID:593421999-1"},
   CellID->286638854]},
 "Proof|ID:593421999-1"->{
  Cell[478795, 13983, 3337, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:593421999", "Proof|ID:593421999-1"},
   CellID->286638854]},
 "Proof|ID:153357458"->{
  Cell[482651, 14096, 2841, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:153357458", "Proof|ID:153357458-1"},
   CellID->454920045]},
 "Proof|ID:153357458-1"->{
  Cell[482651, 14096, 2841, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:153357458", "Proof|ID:153357458-1"},
   CellID->454920045]},
 "Proof|ID:616125500"->{
  Cell[486012, 14197, 2694, 80, 70, "ProofInfo",
   CellTags->{"Proof|ID:616125500", "Proof|ID:616125500-1"},
   CellID->14344385]},
 "Proof|ID:616125500-1"->{
  Cell[486012, 14197, 2694, 80, 70, "ProofInfo",
   CellTags->{"Proof|ID:616125500", "Proof|ID:616125500-1"},
   CellID->14344385]},
 "Proof|ID:491337343"->{
  Cell[489225, 14296, 2504, 76, 70, "ProofInfo",
   CellTags->{"Proof|ID:491337343", "Proof|ID:491337343-1"},
   CellID->458038256]},
 "Proof|ID:491337343-1"->{
  Cell[489225, 14296, 2504, 76, 70, "ProofInfo",
   CellTags->{"Proof|ID:491337343", "Proof|ID:491337343-1"},
   CellID->458038256]},
 "Proof|ID:621700658"->{
  Cell[492248, 14391, 3277, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:621700658", "Proof|ID:621700658-1"},
   CellID->31914115]},
 "Proof|ID:621700658-1"->{
  Cell[492248, 14391, 3277, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:621700658", "Proof|ID:621700658-1"},
   CellID->31914115]},
 "Proof|ID:236021775"->{
  Cell[496044, 14504, 3094, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:236021775", "Proof|ID:236021775-1"},
   CellID->322826090]},
 "Proof|ID:236021775-1"->{
  Cell[496044, 14504, 3094, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:236021775", "Proof|ID:236021775-1"},
   CellID->322826090]},
 "Proof|ID:3006634"->{
  Cell[499658, 14613, 3552, 100, 70, "ProofInfo",
   CellTags->{"Proof|ID:3006634", "Proof|ID:3006634-1"},
   CellID->539834930]},
 "Proof|ID:3006634-1"->{
  Cell[499658, 14613, 3552, 100, 70, "ProofInfo",
   CellTags->{"Proof|ID:3006634", "Proof|ID:3006634-1"},
   CellID->539834930]},
 "Proof|ID:404996142"->{
  Cell[503727, 14732, 3866, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:404996142", "Proof|ID:404996142-1"},
   CellID->457219728]},
 "Proof|ID:404996142-1"->{
  Cell[503727, 14732, 3866, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:404996142", "Proof|ID:404996142-1"},
   CellID->457219728]},
 "Proof|ID:225131303"->{
  Cell[508113, 14859, 4120, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:225131303", "Proof|ID:225131303-1"},
   CellID->442897560]},
 "Proof|ID:225131303-1"->{
  Cell[508113, 14859, 4120, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:225131303", "Proof|ID:225131303-1"},
   CellID->442897560]},
 "Proof|ID:142079924"->{
  Cell[512753, 14992, 4352, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:142079924", "Proof|ID:142079924-1"},
   CellID->230509363]},
 "Proof|ID:142079924-1"->{
  Cell[512753, 14992, 4352, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:142079924", "Proof|ID:142079924-1"},
   CellID->230509363]},
 "Proof|ID:367201929"->{
  Cell[517624, 15129, 5539, 144, 70, "ProofInfo",
   CellTags->{"Proof|ID:367201929", "Proof|ID:367201929-1"},
   CellID->109183652]},
 "Proof|ID:367201929-1"->{
  Cell[517624, 15129, 5539, 144, 70, "ProofInfo",
   CellTags->{"Proof|ID:367201929", "Proof|ID:367201929-1"},
   CellID->109183652]},
 "ID:513393113"->{
  Cell[524838, 15330, 649, 16, 30, "GlobalDeclaration",
   CellTags->{
    "ID:513393113", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->513393113]},
 "ID:623651727"->{
  Cell[525490, 15348, 637, 16, 31, "GlobalDeclaration",
   CellTags->{
    "ID:623651727", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->623651727]},
 "ID:100384801"->{
  Cell[526130, 15366, 1235, 41, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:100384801", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt \[Element]"},
   CellID->100384801]},
 "PreInt \[Element]"->{
  Cell[526130, 15366, 1235, 41, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:100384801", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt \[Element]"},
   CellID->100384801]},
 "ID:398024454"->{
  Cell[527368, 15409, 865, 25, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:398024454", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt 0"},
   CellID->398024454]},
 "PreInt 0"->{
  Cell[527368, 15409, 865, 25, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:398024454", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt 0"},
   CellID->398024454]},
 "ID:418726841"->{
  Cell[528236, 15436, 869, 25, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:418726841", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt 1"},
   CellID->418726841]},
 "PreInt 1"->{
  Cell[528236, 15436, 869, 25, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:418726841", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt 1"},
   CellID->418726841]},
 "ID:10580372"->{
  Cell[529108, 15463, 1072, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:10580372", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt +"},
   CellID->10580372]},
 "PreInt +"->{
  Cell[529108, 15463, 1072, 32, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:10580372", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt +"},
   CellID->10580372]},
 "ID:106096058"->{
  Cell[530183, 15497, 906, 26, 28, "FormalTextInputFormula",
   CellTags->{
    "ID:106096058", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt - unary"},
   CellID->106096058]},
 "PreInt - unary"->{
  Cell[530183, 15497, 906, 26, 28, "FormalTextInputFormula",
   CellTags->{
    "ID:106096058", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt - unary"},
   CellID->106096058]},
 "ID:12727554"->{
  Cell[531092, 15525, 1436, 44, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:12727554", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt \[RawStar]"},
   CellID->12727554]},
 "PreInt \[RawStar]"->{
  Cell[531092, 15525, 1436, 44, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:12727554", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt \[RawStar]"},
   CellID->12727554]},
 "ID:329154583"->{
  Cell[532531, 15571, 1463, 48, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:329154583", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt \[GreaterEqual]"},
   CellID->329154583]},
 "PreInt \[GreaterEqual]"->{
  Cell[532531, 15571, 1463, 48, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:329154583", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt \[GreaterEqual]"},
   CellID->329154583]},
 "ID:468243965"->{
  Cell[533997, 15621, 1332, 43, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:468243965", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt equiv"},
   CellID->468243965]},
 "PreInt equiv"->{
  Cell[533997, 15621, 1332, 43, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:468243965", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt equiv"},
   CellID->468243965]},
 "ID:168445606"->{
  Cell[536187, 15699, 770, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:168445606", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt isDomain", "IR"},
   CellID->168445606]},
 "PreInt isDomain"->{
  Cell[536187, 15699, 770, 20, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:168445606", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "PreInt isDomain", "IR"},
   CellID->168445606]},
 "ID:977301511"->{
  Cell[536960, 15721, 1640, 54, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:977301511", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[Element] PreInt alternative", "IR"},
   CellID->977301511]},
 "\[Element] PreInt alternative"->{
  Cell[536960, 15721, 1640, 54, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:977301511", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[Element] PreInt alternative", "IR"},
   CellID->977301511]},
 "ID:353341722"->{
  Cell[538603, 15777, 1529, 50, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:353341722", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Pair \[Element] PreInt", "IR"},
   CellID->353341722]},
 "Pair \[Element] PreInt"->{
  Cell[538603, 15777, 1529, 50, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:353341722", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Pair \[Element] PreInt", "IR"},
   CellID->353341722]},
 "ID:34141802"->{
  Cell[540135, 15829, 840, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:34141802", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 \[Element] PreInt", "IR"},
   CellID->34141802]},
 "0 \[Element] PreInt"->{
  Cell[540135, 15829, 840, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:34141802", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 \[Element] PreInt", "IR"},
   CellID->34141802]},
 "ID:553448562"->{
  Cell[540978, 15854, 847, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:553448562", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[Element] PreInt", "IR"},
   CellID->553448562]},
 "1 \[Element] PreInt"->{
  Cell[540978, 15854, 847, 23, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:553448562", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[Element] PreInt", "IR"},
   CellID->553448562]},
 "Proof|ID:168445606"->{
  Cell[541979, 15888, 1720, 54, 28, "ProofInfo",
   CellTags->{"Proof|ID:168445606", "Proof|ID:168445606-1"},
   CellID->142038033]},
 "Proof|ID:168445606-1"->{
  Cell[541979, 15888, 1720, 54, 28, "ProofInfo",
   CellTags->{"Proof|ID:168445606", "Proof|ID:168445606-1"},
   CellID->142038033]},
 "Proof|ID:977301511"->{
  Cell[544219, 15961, 3752, 102, 26, "ProofInfo",
   CellTags->{"Proof|ID:977301511", "Proof|ID:977301511-1"},
   CellID->32364991]},
 "Proof|ID:977301511-1"->{
  Cell[544219, 15961, 3752, 102, 26, "ProofInfo",
   CellTags->{"Proof|ID:977301511", "Proof|ID:977301511-1"},
   CellID->32364991]},
 "Proof|ID:353341722"->{
  Cell[548491, 16082, 3475, 96, 26, "ProofInfo",
   CellTags->{"Proof|ID:353341722", "Proof|ID:353341722-1"},
   CellID->352067017]},
 "Proof|ID:353341722-1"->{
  Cell[548491, 16082, 3475, 96, 26, "ProofInfo",
   CellTags->{"Proof|ID:353341722", "Proof|ID:353341722-1"},
   CellID->352067017]},
 "Proof|ID:34141802"->{
  Cell[552486, 16197, 1860, 58, 26, "ProofInfo",
   CellTags->{"Proof|ID:34141802", "Proof|ID:34141802-1"},
   CellID->53936406]},
 "Proof|ID:34141802-1"->{
  Cell[552486, 16197, 1860, 58, 26, "ProofInfo",
   CellTags->{"Proof|ID:34141802", "Proof|ID:34141802-1"},
   CellID->53936406]},
 "Proof|ID:553448562"->{
  Cell[554865, 16274, 1866, 58, 26, "ProofInfo",
   CellTags->{"Proof|ID:553448562", "Proof|ID:553448562-1"},
   CellID->265994209]},
 "Proof|ID:553448562-1"->{
  Cell[554865, 16274, 1866, 58, 26, "ProofInfo",
   CellTags->{"Proof|ID:553448562", "Proof|ID:553448562-1"},
   CellID->265994209]},
 "ID:220417861"->{
  Cell[557926, 16373, 1504, 47, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:220417861", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv Pair PreInt", "IR"},
   CellID->220417861]},
 "equiv Pair PreInt"->{
  Cell[557926, 16373, 1504, 47, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:220417861", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv Pair PreInt", "IR"},
   CellID->220417861]},
 "ID:52539739"->{
  Cell[559433, 16422, 1398, 44, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:52539739", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv 0 PreInt", "IR"},
   CellID->52539739]},
 "equiv 0 PreInt"->{
  Cell[559433, 16422, 1398, 44, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:52539739", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv 0 PreInt", "IR"},
   CellID->52539739]},
 "ID:94484613"->{
  Cell[560834, 16468, 1416, 43, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:94484613", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv Nat \[GreaterEqual] PreInt"},
   CellID->94484613]},
 "equiv Nat \[GreaterEqual] PreInt"->{
  Cell[560834, 16468, 1416, 43, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:94484613", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv Nat \[GreaterEqual] PreInt"},
   CellID->94484613]},
 "ID:321256465"->{
  Cell[562253, 16513, 1413, 43, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:321256465", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv Nat \[LessEqual] PreInt"},
   CellID->321256465]},
 "equiv Nat \[LessEqual] PreInt"->{
  Cell[562253, 16513, 1413, 43, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:321256465", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv Nat \[LessEqual] PreInt"},
   CellID->321256465]},
 "ID:129809121"->{
  Cell[563669, 16558, 911, 25, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:129809121", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[Not] 1 equiv 0 PreInt"},
   CellID->129809121]},
 "\[Not] 1 equiv 0 PreInt"->{
  Cell[563669, 16558, 911, 25, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:129809121", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[Not] 1 equiv 0 PreInt"},
   CellID->129809121]},
 "ID:413771523"->{
  Cell[564583, 16585, 720, 19, 41, "GlobalDeclaration",
   CellTags->{
    "ID:413771523", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->413771523]},
 "ID:637862621"->{
  Cell[565306, 16606, 1703, 51, 69, "FormalTextInputFormula",
   CellTags->{
    "ID:637862621", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv Nat PreInt"},
   CellID->637862621]},
 "equiv Nat PreInt"->{
  Cell[565306, 16606, 1703, 51, 69, "FormalTextInputFormula",
   CellTags->{
    "ID:637862621", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv Nat PreInt"},
   CellID->637862621]},
 "ID:12369324"->{
  Cell[567012, 16659, 817, 22, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:12369324", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv reflexive PreInt"},
   CellID->12369324]},
 "equiv reflexive PreInt"->{
  Cell[567012, 16659, 817, 22, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:12369324", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv reflexive PreInt"},
   CellID->12369324]},
 "ID:19276"->{
  Cell[567832, 16683, 930, 26, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:19276", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv symmetric PreInt"},
   CellID->19276]},
 "equiv symmetric PreInt"->{
  Cell[567832, 16683, 930, 26, 32, "FormalTextInputFormula",
   CellTags->{
    "ID:19276", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv symmetric PreInt"},
   CellID->19276]},
 "ID:164883544"->{
  Cell[568765, 16711, 1420, 45, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:164883544", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv transitive PreInt"},
   CellID->164883544]},
 "equiv transitive PreInt"->{
  Cell[568765, 16711, 1420, 45, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:164883544", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv transitive PreInt"},
   CellID->164883544]},
 "ID:357440757"->{
  Cell[570188, 16758, 849, 23, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:357440757", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv isEquivalence PreInt"},
   CellID->357440757]},
 "equiv isEquivalence PreInt"->{
  Cell[570188, 16758, 849, 23, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:357440757", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "equiv isEquivalence PreInt"},
   CellID->357440757]},
 "Proof|ID:220417861"->{
  Cell[571192, 16792, 4123, 112, 28, "ProofInfo",
   CellTags->{"Proof|ID:220417861", "Proof|ID:220417861-1"},
   CellID->227048585]},
 "Proof|ID:220417861-1"->{
  Cell[571192, 16792, 4123, 112, 28, "ProofInfo",
   CellTags->{"Proof|ID:220417861", "Proof|ID:220417861-1"},
   CellID->227048585]},
 "Proof|ID:52539739"->{
  Cell[575835, 16923, 3453, 98, 26, "ProofInfo",
   CellTags->{"Proof|ID:52539739", "Proof|ID:52539739-1"},
   CellID->142039401]},
 "Proof|ID:52539739-1"->{
  Cell[575835, 16923, 3453, 98, 26, "ProofInfo",
   CellTags->{"Proof|ID:52539739", "Proof|ID:52539739-1"},
   CellID->142039401]},
 "Proof|ID:94484613"->{
  Cell[579807, 17040, 4068, 109, 26, "ProofInfo",
   CellTags->{"Proof|ID:94484613", "Proof|ID:94484613-1"},
   CellID->325556718]},
 "Proof|ID:94484613-1"->{
  Cell[579807, 17040, 4068, 109, 26, "ProofInfo",
   CellTags->{"Proof|ID:94484613", "Proof|ID:94484613-1"},
   CellID->325556718]},
 "Proof|ID:321256465"->{
  Cell[584437, 17169, 4059, 108, 26, "ProofInfo",
   CellTags->{"Proof|ID:321256465", "Proof|ID:321256465-1"},
   CellID->438580499]},
 "Proof|ID:321256465-1"->{
  Cell[584437, 17169, 4059, 108, 26, "ProofInfo",
   CellTags->{"Proof|ID:321256465", "Proof|ID:321256465-1"},
   CellID->438580499]},
 "Proof|ID:129809121"->{
  Cell[589016, 17296, 2316, 70, 26, "ProofInfo",
   CellTags->{"Proof|ID:129809121", "Proof|ID:129809121-1"},
   CellID->242312613]},
 "Proof|ID:129809121-1"->{
  Cell[589016, 17296, 2316, 70, 26, "ProofInfo",
   CellTags->{"Proof|ID:129809121", "Proof|ID:129809121-1"},
   CellID->242312613]},
 "Proof|ID:637862621"->{
  Cell[591852, 17385, 4053, 110, 26, "ProofInfo",
   CellTags->{"Proof|ID:637862621", "Proof|ID:637862621-1"},
   CellID->752123440]},
 "Proof|ID:637862621-1"->{
  Cell[591852, 17385, 4053, 110, 26, "ProofInfo",
   CellTags->{"Proof|ID:637862621", "Proof|ID:637862621-1"},
   CellID->752123440]},
 "Proof|ID:12369324"->{
  Cell[596425, 17514, 2499, 74, 26, "ProofInfo",
   CellTags->{"Proof|ID:12369324", "Proof|ID:12369324-1"},
   CellID->28237667]},
 "Proof|ID:12369324-1"->{
  Cell[596425, 17514, 2499, 74, 26, "ProofInfo",
   CellTags->{"Proof|ID:12369324", "Proof|ID:12369324-1"},
   CellID->28237667]},
 "Proof|ID:19276"->{
  Cell[599443, 17607, 3299, 92, 26, "ProofInfo",
   CellTags->{"Proof|ID:19276", "Proof|ID:19276-1"},
   CellID->146831318]},
 "Proof|ID:19276-1"->{
  Cell[599443, 17607, 3299, 92, 26, "ProofInfo",
   CellTags->{"Proof|ID:19276", "Proof|ID:19276-1"},
   CellID->146831318]},
 "Proof|ID:164883544"->{
  Cell[603262, 17718, 4146, 110, 26, "ProofInfo",
   CellTags->{"Proof|ID:164883544", "Proof|ID:164883544-1"},
   CellID->62903797]},
 "Proof|ID:164883544-1"->{
  Cell[603262, 17718, 4146, 110, 26, "ProofInfo",
   CellTags->{"Proof|ID:164883544", "Proof|ID:164883544-1"},
   CellID->62903797]},
 "Proof|ID:357440757"->{
  Cell[607928, 17847, 1902, 58, 26, "ProofInfo",
   CellTags->{"Proof|ID:357440757", "Proof|ID:357440757-1"},
   CellID->81770248]},
 "Proof|ID:357440757-1"->{
  Cell[607928, 17847, 1902, 58, 26, "ProofInfo",
   CellTags->{"Proof|ID:357440757", "Proof|ID:357440757-1"},
   CellID->81770248]},
 "ID:125790650"->{
  Cell[611236, 17955, 1361, 41, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:125790650", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ Pair PreInt", "IR"},
   CellID->125790650]},
 "+ Pair PreInt"->{
  Cell[611236, 17955, 1361, 41, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:125790650", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ Pair PreInt", "IR"},
   CellID->125790650]},
 "ID:86057977"->{
  Cell[612600, 17998, 715, 19, 41, "GlobalDeclaration",
   CellTags->{
    "ID:86057977", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->86057977]},
 "ID:196108575"->{
  Cell[613318, 18019, 856, 24, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:196108575", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ closed PreInt"},
   CellID->196108575]},
 "+ closed PreInt"->{
  Cell[613318, 18019, 856, 24, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:196108575", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ closed PreInt"},
   CellID->196108575]},
 "ID:132949647"->{
  Cell[614177, 18045, 844, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:132949647", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 rneutral PreInt"},
   CellID->132949647]},
 "0 rneutral PreInt"->{
  Cell[614177, 18045, 844, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:132949647", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 rneutral PreInt"},
   CellID->132949647]},
 "ID:25630206"->{
  Cell[615024, 18071, 851, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:25630206", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ comm PreInt"},
   CellID->25630206]},
 "+ comm PreInt"->{
  Cell[615024, 18071, 851, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:25630206", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ comm PreInt"},
   CellID->25630206]},
 "ID:16604231"->{
  Cell[615878, 18097, 694, 18, 41, "GlobalDeclaration",
   CellTags->{
    "ID:16604231", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->16604231]},
 "ID:73792979"->{
  Cell[616575, 18117, 1018, 31, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:73792979", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ assoc PreInt"},
   CellID->73792979]},
 "+ assoc PreInt"->{
  Cell[616575, 18117, 1018, 31, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:73792979", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ assoc PreInt"},
   CellID->73792979]},
 "ID:243295365"->{
  Cell[617596, 18150, 1093, 31, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:243295365", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ rcanc PreInt"},
   CellID->243295365]},
 "+ rcanc PreInt"->{
  Cell[617596, 18150, 1093, 31, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:243295365", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ rcanc PreInt"},
   CellID->243295365]},
 "ID:47524840"->{
  Cell[618692, 18183, 1057, 31, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:47524840", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ invaiant equiv PreInt"},
   CellID->47524840]},
 "+ invaiant equiv PreInt"->{
  Cell[618692, 18183, 1057, 31, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:47524840", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ invaiant equiv PreInt"},
   CellID->47524840]},
 "ID:107686175"->{
  Cell[619752, 18216, 1030, 31, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:107686175", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ isInvariantFunction2 PreInt"},
   CellID->107686175]},
 "+ isInvariantFunction2 PreInt"->{
  Cell[619752, 18216, 1030, 31, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:107686175", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ isInvariantFunction2 PreInt"},
   CellID->107686175]},
 "Proof|ID:125790650"->{
  Cell[620937, 18258, 4037, 108, 28, "ProofInfo",
   CellTags->{"Proof|ID:125790650", "Proof|ID:125790650-1"},
   CellID->219935054]},
 "Proof|ID:125790650-1"->{
  Cell[620937, 18258, 4037, 108, 28, "ProofInfo",
   CellTags->{"Proof|ID:125790650", "Proof|ID:125790650-1"},
   CellID->219935054]},
 "Proof|ID:196108575"->{
  Cell[625494, 18385, 2871, 82, 26, "ProofInfo",
   CellTags->{"Proof|ID:196108575", "Proof|ID:196108575-1"},
   CellID->746037930]},
 "Proof|ID:196108575-1"->{
  Cell[625494, 18385, 2871, 82, 26, "ProofInfo",
   CellTags->{"Proof|ID:196108575", "Proof|ID:196108575-1"},
   CellID->746037930]},
 "Proof|ID:132949647"->{
  Cell[628885, 18486, 2583, 78, 26, "ProofInfo",
   CellTags->{"Proof|ID:132949647", "Proof|ID:132949647-1"},
   CellID->16458106]},
 "Proof|ID:132949647-1"->{
  Cell[628885, 18486, 2583, 78, 26, "ProofInfo",
   CellTags->{"Proof|ID:132949647", "Proof|ID:132949647-1"},
   CellID->16458106]},
 "Proof|ID:25630206"->{
  Cell[631987, 18583, 3117, 90, 26, "ProofInfo",
   CellTags->{"Proof|ID:25630206", "Proof|ID:25630206-1"},
   CellID->184131697]},
 "Proof|ID:25630206-1"->{
  Cell[631987, 18583, 3117, 90, 26, "ProofInfo",
   CellTags->{"Proof|ID:25630206", "Proof|ID:25630206-1"},
   CellID->184131697]},
 "Proof|ID:73792979"->{
  Cell[635622, 18692, 3913, 108, 26, "ProofInfo",
   CellTags->{"Proof|ID:73792979", "Proof|ID:73792979-1"},
   CellID->273331503]},
 "Proof|ID:73792979-1"->{
  Cell[635622, 18692, 3913, 108, 26, "ProofInfo",
   CellTags->{"Proof|ID:73792979", "Proof|ID:73792979-1"},
   CellID->273331503]},
 "Proof|ID:243295365"->{
  Cell[640054, 18819, 4121, 110, 26, "ProofInfo",
   CellTags->{"Proof|ID:243295365", "Proof|ID:243295365-1"},
   CellID->446127764]},
 "Proof|ID:243295365-1"->{
  Cell[640054, 18819, 4121, 110, 26, "ProofInfo",
   CellTags->{"Proof|ID:243295365", "Proof|ID:243295365-1"},
   CellID->446127764]},
 "Proof|ID:47524840"->{
  Cell[644695, 18948, 4134, 110, 26, "ProofInfo",
   CellTags->{"Proof|ID:47524840", "Proof|ID:47524840-1"},
   CellID->317988883]},
 "Proof|ID:47524840-1"->{
  Cell[644695, 18948, 4134, 110, 26, "ProofInfo",
   CellTags->{"Proof|ID:47524840", "Proof|ID:47524840-1"},
   CellID->317988883]},
 "Proof|ID:107686175"->{
  Cell[649347, 19077, 2019, 60, 26, "ProofInfo",
   CellTags->{"Proof|ID:107686175", "Proof|ID:107686175-1"},
   CellID->207023133]},
 "Proof|ID:107686175-1"->{
  Cell[649347, 19077, 2019, 60, 26, "ProofInfo",
   CellTags->{"Proof|ID:107686175", "Proof|ID:107686175-1"},
   CellID->207023133]},
 "ID:360943410"->{
  Cell[652392, 19175, 1112, 34, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:360943410", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- unary Pair PreInt", "IR"},
   CellID->360943410]},
 "- unary Pair PreInt"->{
  Cell[652392, 19175, 1112, 34, 34, "FormalTextInputFormula",
   CellTags->{
    "ID:360943410", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- unary Pair PreInt", "IR"},
   CellID->360943410]},
 "ID:474936069"->{
  Cell[653507, 19211, 695, 18, 41, "GlobalDeclaration",
   CellTags->{
    "ID:474936069", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->474936069]},
 "ID:5265224"->{
  Cell[654205, 19231, 863, 24, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:5265224", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- unary closed PreInt", "IR"},
   CellID->5265224]},
 "- unary closed PreInt"->{
  Cell[654205, 19231, 863, 24, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:5265224", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- unary closed PreInt", "IR"},
   CellID->5265224]},
 "ID:203475906"->{
  Cell[655071, 19257, 1076, 30, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:203475906", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- inverse PreInt"},
   CellID->203475906]},
 "- inverse PreInt"->{
  Cell[655071, 19257, 1076, 30, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:203475906", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- inverse PreInt"},
   CellID->203475906]},
 "ID:611704112"->{
  Cell[656150, 19289, 959, 25, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:611704112", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- isInvariantFunction1 PreInt"},
   CellID->611704112]},
 "- isInvariantFunction1 PreInt"->{
  Cell[656150, 19289, 959, 25, 35, "FormalTextInputFormula",
   CellTags->{
    "ID:611704112", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- isInvariantFunction1 PreInt"},
   CellID->611704112]},
 "Proof|ID:360943410"->{
  Cell[657265, 19325, 2987, 86, 28, "ProofInfo",
   CellTags->{"Proof|ID:360943410", "Proof|ID:360943410-1"},
   CellID->307773152]},
 "Proof|ID:360943410-1"->{
  Cell[657265, 19325, 2987, 86, 28, "ProofInfo",
   CellTags->{"Proof|ID:360943410", "Proof|ID:360943410-1"},
   CellID->307773152]},
 "Proof|ID:5265224"->{
  Cell[660772, 19430, 2509, 74, 26, "ProofInfo",
   CellTags->{"Proof|ID:5265224", "Proof|ID:5265224-1"},
   CellID->489898059]},
 "Proof|ID:5265224-1"->{
  Cell[660772, 19430, 2509, 74, 26, "ProofInfo",
   CellTags->{"Proof|ID:5265224", "Proof|ID:5265224-1"},
   CellID->489898059]},
 "Proof|ID:203475906"->{
  Cell[663798, 19523, 2915, 84, 26, "ProofInfo",
   CellTags->{"Proof|ID:203475906", "Proof|ID:203475906-1"},
   CellID->819919783]},
 "Proof|ID:203475906-1"->{
  Cell[663798, 19523, 2915, 84, 26, "ProofInfo",
   CellTags->{"Proof|ID:203475906", "Proof|ID:203475906-1"},
   CellID->819919783]},
 "Proof|ID:611704112"->{
  Cell[667233, 19626, 2027, 60, 26, "ProofInfo",
   CellTags->{"Proof|ID:611704112", "Proof|ID:611704112-1"},
   CellID->443769915]},
 "Proof|ID:611704112-1"->{
  Cell[667233, 19626, 2027, 60, 26, "ProofInfo",
   CellTags->{"Proof|ID:611704112", "Proof|ID:611704112-1"},
   CellID->443769915]},
 "ID:455358938"->{
  Cell[670688, 19739, 1603, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:455358938", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] Pair PreInt", "IR"},
   CellID->455358938]},
 "\[RawStar] Pair PreInt"->{
  Cell[670688, 19739, 1603, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:455358938", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] Pair PreInt", "IR"},
   CellID->455358938]},
 "ID:206670450"->{
  Cell[672294, 19790, 771, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:206670450", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->206670450]},
 "ID:80978087"->{
  Cell[673068, 19812, 908, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:80978087", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] closed PreInt", "IR"},
   CellID->80978087]},
 "\[RawStar] closed PreInt"->{
  Cell[673068, 19812, 908, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:80978087", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] closed PreInt", "IR"},
   CellID->80978087]},
 "ID:561254237"->{
  Cell[673979, 19839, 878, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:561254237", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 rannihilator PreInt"},
   CellID->561254237]},
 "0 rannihilator PreInt"->{
  Cell[673979, 19839, 878, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:561254237", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 rannihilator PreInt"},
   CellID->561254237]},
 "ID:201948360"->{
  Cell[674860, 19865, 897, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:201948360", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 rneutral PreInt"},
   CellID->201948360]},
 "1 rneutral PreInt"->{
  Cell[674860, 19865, 897, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:201948360", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 rneutral PreInt"},
   CellID->201948360]},
 "ID:152728157"->{
  Cell[675760, 19892, 853, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:152728157", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] comm PreInt"},
   CellID->152728157]},
 "\[RawStar] comm PreInt"->{
  Cell[675760, 19892, 853, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:152728157", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] comm PreInt"},
   CellID->152728157]},
 "ID:258603229"->{
  Cell[676616, 19918, 1344, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:258603229", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] zdf PreInt"},
   CellID->258603229]},
 "\[RawStar] zdf PreInt"->{
  Cell[676616, 19918, 1344, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:258603229", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] zdf PreInt"},
   CellID->258603229]},
 "ID:200297603"->{
  Cell[677963, 19961, 695, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:200297603", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->200297603]},
 "ID:142359696"->{
  Cell[678661, 19981, 1023, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:142359696", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] assoc PreInt"},
   CellID->142359696]},
 "\[RawStar] assoc PreInt"->{
  Cell[678661, 19981, 1023, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:142359696", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] assoc PreInt"},
   CellID->142359696]},
 "ID:262630475"->{
  Cell[679687, 20014, 1053, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:262630475", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] distr + PreInt"},
   CellID->262630475]},
 "\[RawStar] distr + PreInt"->{
  Cell[679687, 20014, 1053, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:262630475", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] distr + PreInt"},
   CellID->262630475]},
 "ID:146416852"->{
  Cell[680743, 20047, 1064, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:146416852", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] invariant equiv PreInt"},
   CellID->146416852]},
 "\[RawStar] invariant equiv PreInt"->{
  Cell[680743, 20047, 1064, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:146416852", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] invariant equiv PreInt"},
   CellID->146416852]},
 "ID:204021385"->{
  Cell[681810, 20080, 1030, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:204021385", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] isInvariantFunction2 PreInt"},
   CellID->204021385]},
 "\[RawStar] isInvariantFunction2 PreInt"->{
  Cell[681810, 20080, 1030, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:204021385", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] isInvariantFunction2 PreInt"},
   CellID->204021385]},
 "Proof|ID:455358938"->{
  Cell[682996, 20122, 5125, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:455358938", "Proof|ID:455358938-1"},
   CellID->297124414]},
 "Proof|ID:455358938-1"->{
  Cell[682996, 20122, 5125, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:455358938", "Proof|ID:455358938-1"},
   CellID->297124414]},
 "Proof|ID:80978087"->{
  Cell[688641, 20273, 2866, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:80978087", "Proof|ID:80978087-1"},
   CellID->295496011]},
 "Proof|ID:80978087-1"->{
  Cell[688641, 20273, 2866, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:80978087", "Proof|ID:80978087-1"},
   CellID->295496011]},
 "Proof|ID:561254237"->{
  Cell[692026, 20374, 2556, 76, 70, "ProofInfo",
   CellTags->{"Proof|ID:561254237", "Proof|ID:561254237-1"},
   CellID->690925667]},
 "Proof|ID:561254237-1"->{
  Cell[692026, 20374, 2556, 76, 70, "ProofInfo",
   CellTags->{"Proof|ID:561254237", "Proof|ID:561254237-1"},
   CellID->690925667]},
 "Proof|ID:201948360"->{
  Cell[695101, 20469, 2584, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:201948360", "Proof|ID:201948360-1"},
   CellID->558544782]},
 "Proof|ID:201948360-1"->{
  Cell[695101, 20469, 2584, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:201948360", "Proof|ID:201948360-1"},
   CellID->558544782]},
 "Proof|ID:152728157"->{
  Cell[698205, 20566, 3121, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:152728157", "Proof|ID:152728157-1"},
   CellID->24746466]},
 "Proof|ID:152728157-1"->{
  Cell[698205, 20566, 3121, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:152728157", "Proof|ID:152728157-1"},
   CellID->24746466]},
 "Proof|ID:258603229"->{
  Cell[701846, 20675, 4065, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:258603229", "Proof|ID:258603229-1"},
   CellID->149136758]},
 "Proof|ID:258603229-1"->{
  Cell[701846, 20675, 4065, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:258603229", "Proof|ID:258603229-1"},
   CellID->149136758]},
 "Proof|ID:142359696"->{
  Cell[706430, 20804, 3915, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:142359696", "Proof|ID:142359696-1"},
   CellID->796805]},
 "Proof|ID:142359696-1"->{
  Cell[706430, 20804, 3915, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:142359696", "Proof|ID:142359696-1"},
   CellID->796805]},
 "Proof|ID:262630475"->{
  Cell[710865, 20931, 4174, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:262630475", "Proof|ID:262630475-1"},
   CellID->457849091]},
 "Proof|ID:262630475-1"->{
  Cell[710865, 20931, 4174, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:262630475", "Proof|ID:262630475-1"},
   CellID->457849091]},
 "Proof|ID:146416852"->{
  Cell[715559, 21064, 4140, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:146416852", "Proof|ID:146416852-1"},
   CellID->83067894]},
 "Proof|ID:146416852-1"->{
  Cell[715559, 21064, 4140, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:146416852", "Proof|ID:146416852-1"},
   CellID->83067894]},
 "Proof|ID:204021385"->{
  Cell[720219, 21193, 2019, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:204021385", "Proof|ID:204021385-1"},
   CellID->716469443]},
 "Proof|ID:204021385-1"->{
  Cell[720219, 21193, 2019, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:204021385", "Proof|ID:204021385-1"},
   CellID->716469443]},
 "ID:129935970"->{
  Cell[723510, 21297, 1590, 50, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:129935970", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] Pair PreInt", "IR"},
   CellID->129935970]},
 "\[GreaterEqual] Pair PreInt"->{
  Cell[723510, 21297, 1590, 50, 38, "FormalTextInputFormula",
   CellTags->{
    "ID:129935970", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] Pair PreInt", "IR"},
   CellID->129935970]},
 "ID:95876614"->{
  Cell[725103, 21349, 1539, 49, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:95876614", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] 0 PreInt", "IR"},
   CellID->95876614]},
 "\[GreaterEqual] 0 PreInt"->{
  Cell[725103, 21349, 1539, 49, 48, "FormalTextInputFormula",
   CellTags->{
    "ID:95876614", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] 0 PreInt", "IR"},
   CellID->95876614]},
 "ID:221338547"->{
  Cell[726645, 21400, 919, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:221338547", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[GreaterEqual] 0 PreInt"},
   CellID->221338547]},
 "1 \[GreaterEqual] 0 PreInt"->{
  Cell[726645, 21400, 919, 24, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:221338547", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[GreaterEqual] 0 PreInt"},
   CellID->221338547]},
 "ID:13569836"->{
  Cell[727567, 21426, 1078, 32, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:13569836", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] isInvariantRelation2 PreInt"},
   CellID->13569836]},
 "\[GreaterEqual] isInvariantRelation2 PreInt"->{
  Cell[727567, 21426, 1078, 32, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:13569836", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] isInvariantRelation2 PreInt"},
   CellID->13569836]},
 "ID:337456597"->{
  Cell[728648, 21460, 722, 19, 41, "GlobalDeclaration",
   CellTags->{
    "ID:337456597", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->337456597]},
 "ID:83533617"->{
  Cell[729373, 21481, 822, 22, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:83533617", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] reflexive PreInt"},
   CellID->83533617]},
 "\[GreaterEqual] reflexive PreInt"->{
  Cell[729373, 21481, 822, 22, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:83533617", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] reflexive PreInt"},
   CellID->83533617]},
 "ID:281735911"->{
  Cell[730198, 21505, 1157, 34, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:281735911", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] antisymmetric PreInt"},
   CellID->281735911]},
 "\[GreaterEqual] antisymmetric PreInt"->{
  Cell[730198, 21505, 1157, 34, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:281735911", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] antisymmetric PreInt"},
   CellID->281735911]},
 "ID:115550209"->{
  Cell[731358, 21541, 913, 25, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:115550209", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] total PreInt"},
   CellID->115550209]},
 "\[GreaterEqual] total PreInt"->{
  Cell[731358, 21541, 913, 25, 33, "FormalTextInputFormula",
   CellTags->{
    "ID:115550209", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] total PreInt"},
   CellID->115550209]},
 "ID:215340491"->{
  Cell[732274, 21568, 1372, 42, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:215340491", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] discrete PreInt"},
   CellID->215340491]},
 "\[GreaterEqual] discrete PreInt"->{
  Cell[732274, 21568, 1372, 42, 39, "FormalTextInputFormula",
   CellTags->{
    "ID:215340491", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] discrete PreInt"},
   CellID->215340491]},
 "ID:92075332"->{
  Cell[733649, 21612, 1298, 38, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:92075332", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] \[RawStar] PreInt"},
   CellID->92075332]},
 "\[GreaterEqual] \[RawStar] PreInt"->{
  Cell[733649, 21612, 1298, 38, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:92075332", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] \[RawStar] PreInt"},
   CellID->92075332]},
 "ID:145078248"->{
  Cell[734950, 21652, 1299, 39, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:145078248", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat \[SucceedsTilde] PreInt+"},
   CellID->145078248]},
 "Nat \[SucceedsTilde] PreInt+"->{
  Cell[734950, 21652, 1299, 39, 50, "FormalTextInputFormula",
   CellTags->{
    "ID:145078248", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat \[SucceedsTilde] PreInt+"},
   CellID->145078248]},
 "ID:118070489"->{
  Cell[736252, 21693, 697, 18, 41, "GlobalDeclaration",
   CellTags->{
    "ID:118070489", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->118070489]},
 "ID:307661531"->{
  Cell[736952, 21713, 1125, 33, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:307661531", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] transitive PreInt"},
   CellID->307661531]},
 "\[GreaterEqual] transitive PreInt"->{
  Cell[736952, 21713, 1125, 33, 37, "FormalTextInputFormula",
   CellTags->{
    "ID:307661531", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] transitive PreInt"},
   CellID->307661531]},
 "ID:647932785"->{
  Cell[738080, 21748, 1012, 28, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:647932785", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] + PreInt"},
   CellID->647932785]},
 "\[GreaterEqual] + PreInt"->{
  Cell[738080, 21748, 1012, 28, 30, "FormalTextInputFormula",
   CellTags->{
    "ID:647932785", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] + PreInt"},
   CellID->647932785]},
 "Proof|ID:129935970"->{
  Cell[739247, 21787, 4139, 112, 28, "ProofInfo",
   CellTags->{"Proof|ID:129935970", "Proof|ID:129935970-1"},
   CellID->9772636]},
 "Proof|ID:129935970-1"->{
  Cell[739247, 21787, 4139, 112, 28, "ProofInfo",
   CellTags->{"Proof|ID:129935970", "Proof|ID:129935970-1"},
   CellID->9772636]},
 "Proof|ID:95876614"->{
  Cell[743905, 21918, 3458, 96, 26, "ProofInfo",
   CellTags->{"Proof|ID:95876614", "Proof|ID:95876614-1"},
   CellID->62135307]},
 "Proof|ID:95876614-1"->{
  Cell[743905, 21918, 3458, 96, 26, "ProofInfo",
   CellTags->{"Proof|ID:95876614", "Proof|ID:95876614-1"},
   CellID->62135307]},
 "Proof|ID:221338547"->{
  Cell[747882, 22033, 1951, 60, 26, "ProofInfo",
   CellTags->{"Proof|ID:221338547", "Proof|ID:221338547-1"},
   CellID->51061701]},
 "Proof|ID:221338547-1"->{
  Cell[747882, 22033, 1951, 60, 26, "ProofInfo",
   CellTags->{"Proof|ID:221338547", "Proof|ID:221338547-1"},
   CellID->51061701]},
 "Proof|ID:13569836"->{
  Cell[750352, 22112, 2086, 62, 26, "ProofInfo",
   CellTags->{"Proof|ID:13569836", "Proof|ID:13569836-1"},
   CellID->553106541]},
 "Proof|ID:13569836-1"->{
  Cell[750352, 22112, 2086, 62, 26, "ProofInfo",
   CellTags->{"Proof|ID:13569836", "Proof|ID:13569836-1"},
   CellID->553106541]},
 "Proof|ID:83533617"->{
  Cell[752957, 22193, 2477, 75, 26, "ProofInfo",
   CellTags->{"Proof|ID:83533617", "Proof|ID:83533617-1"},
   CellID->1756479]},
 "Proof|ID:83533617-1"->{
  Cell[752957, 22193, 2477, 75, 26, "ProofInfo",
   CellTags->{"Proof|ID:83533617", "Proof|ID:83533617-1"},
   CellID->1756479]},
 "Proof|ID:281735911"->{
  Cell[755953, 22287, 3758, 102, 26, "ProofInfo",
   CellTags->{"Proof|ID:281735911", "Proof|ID:281735911-1"},
   CellID->44053081]},
 "Proof|ID:281735911-1"->{
  Cell[755953, 22287, 3758, 102, 26, "ProofInfo",
   CellTags->{"Proof|ID:281735911", "Proof|ID:281735911-1"},
   CellID->44053081]},
 "Proof|ID:115550209"->{
  Cell[760231, 22408, 3202, 90, 26, "ProofInfo",
   CellTags->{"Proof|ID:115550209", "Proof|ID:115550209-1"},
   CellID->877842890]},
 "Proof|ID:115550209-1"->{
  Cell[760231, 22408, 3202, 90, 26, "ProofInfo",
   CellTags->{"Proof|ID:115550209", "Proof|ID:115550209-1"},
   CellID->877842890]},
 "Proof|ID:215340491"->{
  Cell[763952, 22517, 3760, 103, 26, "ProofInfo",
   CellTags->{"Proof|ID:215340491", "Proof|ID:215340491-1"},
   CellID->147476088]},
 "Proof|ID:215340491-1"->{
  Cell[763952, 22517, 3760, 103, 26, "ProofInfo",
   CellTags->{"Proof|ID:215340491", "Proof|ID:215340491-1"},
   CellID->147476088]},
 "Proof|ID:92075332"->{
  Cell[768232, 22639, 3848, 104, 26, "ProofInfo",
   CellTags->{"Proof|ID:92075332", "Proof|ID:92075332-1"},
   CellID->47485746]},
 "Proof|ID:92075332-1"->{
  Cell[768232, 22639, 3848, 104, 26, "ProofInfo",
   CellTags->{"Proof|ID:92075332", "Proof|ID:92075332-1"},
   CellID->47485746]},
 "Proof|ID:145078248"->{
  Cell[772599, 22762, 3743, 102, 26, "ProofInfo",
   CellTags->{"Proof|ID:145078248", "Proof|ID:145078248-1"},
   CellID->565621444]},
 "Proof|ID:145078248-1"->{
  Cell[772599, 22762, 3743, 102, 26, "ProofInfo",
   CellTags->{"Proof|ID:145078248", "Proof|ID:145078248-1"},
   CellID->565621444]},
 "Proof|ID:307661531"->{
  Cell[776862, 22883, 3981, 107, 26, "ProofInfo",
   CellTags->{"Proof|ID:307661531", "Proof|ID:307661531-1"},
   CellID->58099428]},
 "Proof|ID:307661531-1"->{
  Cell[776862, 22883, 3981, 107, 26, "ProofInfo",
   CellTags->{"Proof|ID:307661531", "Proof|ID:307661531-1"},
   CellID->58099428]},
 "Proof|ID:647932785"->{
  Cell[781363, 23009, 3996, 108, 26, "ProofInfo",
   CellTags->{"Proof|ID:647932785", "Proof|ID:647932785-1"},
   CellID->4451951]},
 "Proof|ID:647932785-1"->{
  Cell[781363, 23009, 3996, 108, 26, "ProofInfo",
   CellTags->{"Proof|ID:647932785", "Proof|ID:647932785-1"},
   CellID->4451951]},
 "ID:463204962"->{
  Cell[786089, 23143, 749, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:463204962", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->463204962]},
 "ID:32325634"->{
  Cell[787445, 23186, 645, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:32325634", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->32325634]},
 "ID:112681936"->{
  Cell[788093, 23204, 1067, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:112681936", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int \[Element]"},
   CellID->112681936]},
 "Int \[Element]"->{
  Cell[788093, 23204, 1067, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:112681936", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int \[Element]"},
   CellID->112681936]},
 "ID:52976508"->{
  Cell[789163, 23239, 849, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:52976508", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int 0"},
   CellID->52976508]},
 "Int 0"->{
  Cell[789163, 23239, 849, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:52976508", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int 0"},
   CellID->52976508]},
 "ID:712864778"->{
  Cell[790015, 23265, 852, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:712864778", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int 1"},
   CellID->712864778]},
 "Int 1"->{
  Cell[790015, 23265, 852, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:712864778", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int 1"},
   CellID->712864778]},
 "ID:19419340"->{
  Cell[790870, 23291, 1098, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:19419340", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int +"},
   CellID->19419340]},
 "Int +"->{
  Cell[790870, 23291, 1098, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:19419340", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int +"},
   CellID->19419340]},
 "ID:280798627"->{
  Cell[791971, 23331, 872, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:280798627", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int - unary"},
   CellID->280798627]},
 "Int - unary"->{
  Cell[791971, 23331, 872, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:280798627", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int - unary"},
   CellID->280798627]},
 "ID:328192039"->{
  Cell[792846, 23357, 1124, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:328192039", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int \[RawStar]"},
   CellID->328192039]},
 "Int \[RawStar]"->{
  Cell[792846, 23357, 1124, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:328192039", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int \[RawStar]"},
   CellID->328192039]},
 "ID:819512290"->{
  Cell[793973, 23398, 1184, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:819512290", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int \[GreaterEqual]"},
   CellID->819512290]},
 "Int \[GreaterEqual]"->{
  Cell[793973, 23398, 1184, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:819512290", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int \[GreaterEqual]"},
   CellID->819512290]},
 "ID:354647619"->{
  Cell[796056, 23473, 757, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:354647619", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int isDomain", "IR"},
   CellID->354647619]},
 "Int isDomain"->{
  Cell[796056, 23473, 757, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:354647619", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int isDomain", "IR"},
   CellID->354647619]},
 "ID:565027904"->{
  Cell[796816, 23495, 1619, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:565027904", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[Element] Int alternative", "IR"},
   CellID->565027904]},
 "\[Element] Int alternative"->{
  Cell[796816, 23495, 1619, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:565027904", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[Element] Int alternative", "IR"},
   CellID->565027904]},
 "ID:40757894"->{
  Cell[798438, 23550, 1163, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:40757894", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "EC \[Element] Int", "IR"},
   CellID->40757894]},
 "EC \[Element] Int"->{
  Cell[798438, 23550, 1163, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:40757894", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "EC \[Element] Int", "IR"},
   CellID->40757894]},
 "ID:350791961"->{
  Cell[799604, 23587, 833, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:350791961", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 \[Element] Int", "IR"},
   CellID->350791961]},
 "0 \[Element] Int"->{
  Cell[799604, 23587, 833, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:350791961", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 \[Element] Int", "IR"},
   CellID->350791961]},
 "ID:181240486"->{
  Cell[800440, 23612, 825, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:181240486", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[Element] Int"},
   CellID->181240486]},
 "1 \[Element] Int"->{
  Cell[800440, 23612, 825, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:181240486", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[Element] Int"},
   CellID->181240486]},
 "ID:364462666"->{
  Cell[801268, 23637, 815, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:364462666", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[NotEqual] 0 Int"},
   CellID->364462666]},
 "1 \[NotEqual] 0 Int"->{
  Cell[801268, 23637, 815, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:364462666", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[NotEqual] 0 Int"},
   CellID->364462666]},
 "Proof|ID:354647619"->{
  Cell[802239, 23670, 1708, 54, 70, "ProofInfo",
   CellTags->{"Proof|ID:354647619", "Proof|ID:354647619-1"},
   CellID->268377469]},
 "Proof|ID:354647619-1"->{
  Cell[802239, 23670, 1708, 54, 70, "ProofInfo",
   CellTags->{"Proof|ID:354647619", "Proof|ID:354647619-1"},
   CellID->268377469]},
 "Proof|ID:565027904"->{
  Cell[804465, 23743, 3621, 100, 70, "ProofInfo",
   CellTags->{"Proof|ID:565027904", "Proof|ID:565027904-1"},
   CellID->760024466]},
 "Proof|ID:565027904-1"->{
  Cell[804465, 23743, 3621, 100, 70, "ProofInfo",
   CellTags->{"Proof|ID:565027904", "Proof|ID:565027904-1"},
   CellID->760024466]},
 "Proof|ID:40757894"->{
  Cell[808606, 23862, 2755, 80, 70, "ProofInfo",
   CellTags->{"Proof|ID:40757894", "Proof|ID:40757894-1"},
   CellID->35500606]},
 "Proof|ID:40757894-1"->{
  Cell[808606, 23862, 2755, 80, 70, "ProofInfo",
   CellTags->{"Proof|ID:40757894", "Proof|ID:40757894-1"},
   CellID->35500606]},
 "Proof|ID:350791961"->{
  Cell[811880, 23961, 1847, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:350791961", "Proof|ID:350791961-1"},
   CellID->16051126]},
 "Proof|ID:350791961-1"->{
  Cell[811880, 23961, 1847, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:350791961", "Proof|ID:350791961-1"},
   CellID->16051126]},
 "Proof|ID:181240486"->{
  Cell[814247, 24038, 1848, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:181240486", "Proof|ID:181240486-1"},
   CellID->157610485]},
 "Proof|ID:181240486-1"->{
  Cell[814247, 24038, 1848, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:181240486", "Proof|ID:181240486-1"},
   CellID->157610485]},
 "Proof|ID:364462666"->{
  Cell[816614, 24115, 1846, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:364462666", "Proof|ID:364462666-1"},
   CellID->279790495]},
 "Proof|ID:364462666-1"->{
  Cell[816614, 24115, 1846, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:364462666", "Proof|ID:364462666-1"},
   CellID->279790495]},
 "ID:56650873"->{
  Cell[819770, 24222, 1339, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:56650873", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ EC Int", "IR"},
   CellID->56650873]},
 "+ EC Int"->{
  Cell[819770, 24222, 1339, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:56650873", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ EC Int", "IR"},
   CellID->56650873]},
 "ID:13236461"->{
  Cell[821112, 24266, 777, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:13236461", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->13236461]},
 "ID:562490241"->{
  Cell[821892, 24288, 848, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:562490241", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ closed Int", "IR"},
   CellID->562490241]},
 "+ closed Int"->{
  Cell[821892, 24288, 848, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:562490241", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ closed Int", "IR"},
   CellID->562490241]},
 "ID:624613835"->{
  Cell[822743, 24314, 827, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:624613835", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 rneutral Int"},
   CellID->624613835]},
 "0 rneutral Int"->{
  Cell[822743, 24314, 827, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:624613835", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 rneutral Int"},
   CellID->624613835]},
 "ID:115787764"->{
  Cell[823573, 24339, 844, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:115787764", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ comm Int"},
   CellID->115787764]},
 "+ comm Int"->{
  Cell[823573, 24339, 844, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:115787764", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ comm Int"},
   CellID->115787764]},
 "ID:11815767"->{
  Cell[824420, 24365, 1048, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:11815767", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ assoc Int"},
   CellID->11815767]},
 "+ assoc Int"->{
  Cell[824420, 24365, 1048, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:11815767", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ assoc Int"},
   CellID->11815767]},
 "ID:139539770"->{
  Cell[825471, 24398, 961, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:139539770", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ rcanc Int"},
   CellID->139539770]},
 "+ rcanc Int"->{
  Cell[825471, 24398, 961, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:139539770", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "+ rcanc Int"},
   CellID->139539770]},
 "Proof|ID:56650873"->{
  Cell[826588, 24436, 4406, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:56650873", "Proof|ID:56650873-1"},
   CellID->106415148]},
 "Proof|ID:56650873-1"->{
  Cell[826588, 24436, 4406, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:56650873", "Proof|ID:56650873-1"},
   CellID->106415148]},
 "Proof|ID:562490241"->{
  Cell[831512, 24573, 2841, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:562490241", "Proof|ID:562490241-1"},
   CellID->110816780]},
 "Proof|ID:562490241-1"->{
  Cell[831512, 24573, 2841, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:562490241", "Proof|ID:562490241-1"},
   CellID->110816780]},
 "Proof|ID:624613835"->{
  Cell[834873, 24674, 2559, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:624613835", "Proof|ID:624613835-1"},
   CellID->23051911]},
 "Proof|ID:624613835-1"->{
  Cell[834873, 24674, 2559, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:624613835", "Proof|ID:624613835-1"},
   CellID->23051911]},
 "Proof|ID:115787764"->{
  Cell[837951, 24771, 3092, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:115787764", "Proof|ID:115787764-1"},
   CellID->253640770]},
 "Proof|ID:115787764-1"->{
  Cell[837951, 24771, 3092, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:115787764", "Proof|ID:115787764-1"},
   CellID->253640770]},
 "Proof|ID:11815767"->{
  Cell[841563, 24880, 3864, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:11815767", "Proof|ID:11815767-1"},
   CellID->44906707]},
 "Proof|ID:11815767-1"->{
  Cell[841563, 24880, 3864, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:11815767", "Proof|ID:11815767-1"},
   CellID->44906707]},
 "Proof|ID:139539770"->{
  Cell[845945, 25007, 3753, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:139539770", "Proof|ID:139539770-1"},
   CellID->42190126]},
 "Proof|ID:139539770-1"->{
  Cell[845945, 25007, 3753, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:139539770", "Proof|ID:139539770-1"},
   CellID->42190126]},
 "ID:331988502"->{
  Cell[850720, 25149, 1242, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:331988502", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- unary EC Int", "IR"},
   CellID->331988502]},
 "- unary EC Int"->{
  Cell[850720, 25149, 1242, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:331988502", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- unary EC Int", "IR"},
   CellID->331988502]},
 "ID:138388913"->{
  Cell[851965, 25190, 694, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:138388913", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->138388913]},
 "ID:789456798"->{
  Cell[852662, 25210, 857, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:789456798", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- unary closed Int", "IR"},
   CellID->789456798]},
 "- unary closed Int"->{
  Cell[852662, 25210, 857, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:789456798", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- unary closed Int", "IR"},
   CellID->789456798]},
 "ID:561244613"->{
  Cell[853522, 25236, 895, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:561244613", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- inverse Int"},
   CellID->561244613]},
 "- inverse Int"->{
  Cell[853522, 25236, 895, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:561244613", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "- inverse Int"},
   CellID->561244613]},
 "Proof|ID:331988502"->{
  Cell[854573, 25273, 3539, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:331988502", "Proof|ID:331988502-1"},
   CellID->217485270]},
 "Proof|ID:331988502-1"->{
  Cell[854573, 25273, 3539, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:331988502", "Proof|ID:331988502-1"},
   CellID->217485270]},
 "Proof|ID:789456798"->{
  Cell[858632, 25390, 2495, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:789456798", "Proof|ID:789456798-1"},
   CellID->617130052]},
 "Proof|ID:789456798-1"->{
  Cell[858632, 25390, 2495, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:789456798", "Proof|ID:789456798-1"},
   CellID->617130052]},
 "Proof|ID:561244613"->{
  Cell[861647, 25483, 2702, 80, 70, "ProofInfo",
   CellTags->{"Proof|ID:561244613", "Proof|ID:561244613-1"},
   CellID->394996459]},
 "Proof|ID:561244613-1"->{
  Cell[861647, 25483, 2702, 80, 70, "ProofInfo",
   CellTags->{"Proof|ID:561244613", "Proof|ID:561244613-1"},
   CellID->394996459]},
 "ID:947607354"->{
  Cell[865668, 25614, 1340, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:947607354", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] EC Int", "IR"},
   CellID->947607354]},
 "\[RawStar] EC Int"->{
  Cell[865668, 25614, 1340, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:947607354", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] EC Int", "IR"},
   CellID->947607354]},
 "ID:105543058"->{
  Cell[867011, 25658, 701, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:105543058", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->105543058]},
 "ID:696766377"->{
  Cell[867715, 25679, 852, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:696766377", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] closed Int", "IR"},
   CellID->696766377]},
 "\[RawStar] closed Int"->{
  Cell[867715, 25679, 852, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:696766377", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] closed Int", "IR"},
   CellID->696766377]},
 "ID:202378771"->{
  Cell[868570, 25705, 861, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:202378771", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 rannihilator Int"},
   CellID->202378771]},
 "0 rannihilator Int"->{
  Cell[868570, 25705, 861, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:202378771", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "0 rannihilator Int"},
   CellID->202378771]},
 "ID:689609352"->{
  Cell[869434, 25731, 827, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:689609352", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 rneutral Int"},
   CellID->689609352]},
 "1 rneutral Int"->{
  Cell[869434, 25731, 827, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:689609352", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 rneutral Int"},
   CellID->689609352]},
 "ID:196343023"->{
  Cell[870264, 25756, 842, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:196343023", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] comm Int"},
   CellID->196343023]},
 "\[RawStar] comm Int"->{
  Cell[870264, 25756, 842, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:196343023", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] comm Int"},
   CellID->196343023]},
 "ID:207834126"->{
  Cell[871109, 25782, 1193, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:207834126", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] zdf Int"},
   CellID->207834126]},
 "\[RawStar] zdf Int"->{
  Cell[871109, 25782, 1193, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:207834126", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] zdf Int"},
   CellID->207834126]},
 "ID:10196611"->{
  Cell[872305, 25823, 999, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:10196611", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] assoc Int"},
   CellID->10196611]},
 "\[RawStar] assoc Int"->{
  Cell[872305, 25823, 999, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:10196611", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] assoc Int"},
   CellID->10196611]},
 "ID:777222529"->{
  Cell[873307, 25855, 1032, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:777222529", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] distr + Int"},
   CellID->777222529]},
 "\[RawStar] distr + Int"->{
  Cell[873307, 25855, 1032, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:777222529", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[RawStar] distr + Int"},
   CellID->777222529]},
 "Proof|ID:947607354"->{
  Cell[874494, 25897, 4411, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:947607354", "Proof|ID:947607354-1"},
   CellID->558776955]},
 "Proof|ID:947607354-1"->{
  Cell[874494, 25897, 4411, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:947607354", "Proof|ID:947607354-1"},
   CellID->558776955]},
 "Proof|ID:696766377"->{
  Cell[879425, 26034, 2840, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:696766377", "Proof|ID:696766377-1"},
   CellID->81854393]},
 "Proof|ID:696766377-1"->{
  Cell[879425, 26034, 2840, 82, 70, "ProofInfo",
   CellTags->{"Proof|ID:696766377", "Proof|ID:696766377-1"},
   CellID->81854393]},
 "Proof|ID:202378771"->{
  Cell[882785, 26135, 2526, 76, 70, "ProofInfo",
   CellTags->{"Proof|ID:202378771", "Proof|ID:202378771-1"},
   CellID->465404293]},
 "Proof|ID:202378771-1"->{
  Cell[882785, 26135, 2526, 76, 70, "ProofInfo",
   CellTags->{"Proof|ID:202378771", "Proof|ID:202378771-1"},
   CellID->465404293]},
 "Proof|ID:689609352"->{
  Cell[885831, 26230, 2560, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:689609352", "Proof|ID:689609352-1"},
   CellID->485464469]},
 "Proof|ID:689609352-1"->{
  Cell[885831, 26230, 2560, 78, 70, "ProofInfo",
   CellTags->{"Proof|ID:689609352", "Proof|ID:689609352-1"},
   CellID->485464469]},
 "Proof|ID:196343023"->{
  Cell[888911, 26327, 3091, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:196343023", "Proof|ID:196343023-1"},
   CellID->43558645]},
 "Proof|ID:196343023-1"->{
  Cell[888911, 26327, 3091, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:196343023", "Proof|ID:196343023-1"},
   CellID->43558645]},
 "Proof|ID:207834126"->{
  Cell[892522, 26436, 3504, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:207834126", "Proof|ID:207834126-1"},
   CellID->5339075]},
 "Proof|ID:207834126-1"->{
  Cell[892522, 26436, 3504, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:207834126", "Proof|ID:207834126-1"},
   CellID->5339075]},
 "Proof|ID:10196611"->{
  Cell[896546, 26553, 3865, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:10196611", "Proof|ID:10196611-1"},
   CellID->268794065]},
 "Proof|ID:10196611-1"->{
  Cell[896546, 26553, 3865, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:10196611", "Proof|ID:10196611-1"},
   CellID->268794065]},
 "Proof|ID:777222529"->{
  Cell[900930, 26680, 4120, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:777222529", "Proof|ID:777222529-1"},
   CellID->343265473]},
 "Proof|ID:777222529-1"->{
  Cell[900930, 26680, 4120, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:777222529", "Proof|ID:777222529-1"},
   CellID->343265473]},
 "ID:785793765"->{
  Cell[906257, 26835, 1550, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:785793765", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] EC Int", "IR"},
   CellID->785793765]},
 "\[GreaterEqual] EC Int"->{
  Cell[906257, 26835, 1550, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:785793765", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] EC Int", "IR"},
   CellID->785793765]},
 "ID:72386745"->{
  Cell[907810, 26886, 850, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:72386745", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[GreaterEqual] 0 Int"},
   CellID->72386745]},
 "1 \[GreaterEqual] 0 Int"->{
  Cell[907810, 26886, 850, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:72386745", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "1 \[GreaterEqual] 0 Int"},
   CellID->72386745]},
 "ID:100510373"->{
  Cell[908663, 26911, 726, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:100510373", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
   CellID->100510373]},
 "ID:267773274"->{
  Cell[909392, 26932, 815, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:267773274", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] reflexive Int"},
   CellID->267773274]},
 "\[GreaterEqual] reflexive Int"->{
  Cell[909392, 26932, 815, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:267773274", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] reflexive Int"},
   CellID->267773274]},
 "ID:69005952"->{
  Cell[910210, 26955, 1075, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:69005952", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] antisymmetric Int"},
   CellID->69005952]},
 "\[GreaterEqual] antisymmetric Int"->{
  Cell[910210, 26955, 1075, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:69005952", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] antisymmetric Int"},
   CellID->69005952]},
 "ID:275182510"->{
  Cell[911288, 26988, 1112, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:275182510", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] transitive Int"},
   CellID->275182510]},
 "\[GreaterEqual] transitive Int"->{
  Cell[911288, 26988, 1112, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:275182510", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] transitive Int"},
   CellID->275182510]},
 "ID:176048001"->{
  Cell[912403, 27023, 899, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:176048001", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] total Int"},
   CellID->176048001]},
 "\[GreaterEqual] total Int"->{
  Cell[912403, 27023, 899, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:176048001", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] total Int"},
   CellID->176048001]},
 "ID:566590035"->{
  Cell[913305, 27050, 1156, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:566590035", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] discrete Int"},
   CellID->566590035]},
 "\[GreaterEqual] discrete Int"->{
  Cell[913305, 27050, 1156, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:566590035", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] discrete Int"},
   CellID->566590035]},
 "ID:262876021"->{
  Cell[914464, 27086, 1000, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:262876021", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] + Int"},
   CellID->262876021]},
 "\[GreaterEqual] + Int"->{
  Cell[914464, 27086, 1000, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:262876021", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] + Int"},
   CellID->262876021]},
 "ID:34443687"->{
  Cell[915467, 27116, 1224, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:34443687", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] \[RawStar] Int"},
   CellID->34443687]},
 "\[GreaterEqual] \[RawStar] Int"->{
  Cell[915467, 27116, 1224, 37, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:34443687", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "\[GreaterEqual] \[RawStar] Int"},
   CellID->34443687]},
 "Proof|ID:785793765"->{
  Cell[916847, 27164, 4128, 112, 70, "ProofInfo",
   CellTags->{"Proof|ID:785793765", "Proof|ID:785793765-1"},
   CellID->518056285]},
 "Proof|ID:785793765-1"->{
  Cell[916847, 27164, 4128, 112, 70, "ProofInfo",
   CellTags->{"Proof|ID:785793765", "Proof|ID:785793765-1"},
   CellID->518056285]},
 "Proof|ID:72386745"->{
  Cell[921495, 27295, 1922, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:72386745", "Proof|ID:72386745-1"},
   CellID->72615069]},
 "Proof|ID:72386745-1"->{
  Cell[921495, 27295, 1922, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:72386745", "Proof|ID:72386745-1"},
   CellID->72615069]},
 "Proof|ID:267773274"->{
  Cell[923935, 27374, 2457, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:267773274", "Proof|ID:267773274-1"},
   CellID->50495462]},
 "Proof|ID:267773274-1"->{
  Cell[923935, 27374, 2457, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:267773274", "Proof|ID:267773274-1"},
   CellID->50495462]},
 "Proof|ID:69005952"->{
  Cell[926912, 27467, 3568, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:69005952", "Proof|ID:69005952-1"},
   CellID->49834528]},
 "Proof|ID:69005952-1"->{
  Cell[926912, 27467, 3568, 99, 70, "ProofInfo",
   CellTags->{"Proof|ID:69005952", "Proof|ID:69005952-1"},
   CellID->49834528]},
 "Proof|ID:275182510"->{
  Cell[930999, 27585, 3939, 107, 70, "ProofInfo",
   CellTags->{"Proof|ID:275182510", "Proof|ID:275182510-1"},
   CellID->94179691]},
 "Proof|ID:275182510-1"->{
  Cell[930999, 27585, 3939, 107, 70, "ProofInfo",
   CellTags->{"Proof|ID:275182510", "Proof|ID:275182510-1"},
   CellID->94179691]},
 "Proof|ID:176048001"->{
  Cell[935457, 27711, 3172, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:176048001", "Proof|ID:176048001-1"},
   CellID->320141850]},
 "Proof|ID:176048001-1"->{
  Cell[935457, 27711, 3172, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:176048001", "Proof|ID:176048001-1"},
   CellID->320141850]},
 "Proof|ID:566590035"->{
  Cell[939149, 27820, 3191, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:566590035", "Proof|ID:566590035-1"},
   CellID->88604489]},
 "Proof|ID:566590035-1"->{
  Cell[939149, 27820, 3191, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:566590035", "Proof|ID:566590035-1"},
   CellID->88604489]},
 "Proof|ID:262876021"->{
  Cell[942860, 27929, 3950, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:262876021", "Proof|ID:262876021-1"},
   CellID->490373417]},
 "Proof|ID:262876021-1"->{
  Cell[942860, 27929, 3950, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:262876021", "Proof|ID:262876021-1"},
   CellID->490373417]},
 "Proof|ID:34443687"->{
  Cell[947330, 28056, 3789, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:34443687", "Proof|ID:34443687-1"},
   CellID->161887690]},
 "Proof|ID:34443687-1"->{
  Cell[947330, 28056, 3789, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:34443687", "Proof|ID:34443687-1"},
   CellID->161887690]},
 "ID:134029858"->{
  Cell[952069, 28193, 1720, 52, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:134029858", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat \[SucceedsTilde] Int+"},
   CellID->134029858]},
 "Nat \[SucceedsTilde] Int+"->{
  Cell[952069, 28193, 1720, 52, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:134029858", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Nat \[SucceedsTilde] Int+"},
   CellID->134029858]},
 "ID:181242736"->{
  Cell[953792, 28247, 2965, 92, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:181242736", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int+ induction"},
   CellID->181242736]},
 "Int+ induction"->{
  Cell[953792, 28247, 2965, 92, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:181242736", 
     "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb",
      "Int+ induction"},
   CellID->181242736]},
 "Proof|ID:134029858"->{
  Cell[956912, 28350, 4013, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:134029858", "Proof|ID:134029858-1"},
   CellID->379729521]},
 "Proof|ID:134029858-1"->{
  Cell[956912, 28350, 4013, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:134029858", "Proof|ID:134029858-1"},
   CellID->379729521]},
 "Proof|ID:181242736"->{
  Cell[961445, 28477, 6055, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:181242736", "Proof|ID:181242736-1"},
   CellID->625873853]},
 "Proof|ID:181242736-1"->{
  Cell[961445, 28477, 6055, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:181242736", "Proof|ID:181242736-1"},
   CellID->625873853]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"ID:346267501", 1005624, 29412},
 {"Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", \
1005895, 29418},
 {"ID:143274494", 1055566, 30714},
 {"natRecSet", 1055801, 30721},
 {"ID:669214708", 1056039, 30728},
 {"natRec", 1056268, 30735},
 {"ID:97279932", 1056502, 30742},
 {"ID:504518491", 1056713, 30748},
 {"ID:79522091", 1056925, 30754},
 {"natRecSet induction", 1057184, 30761},
 {"IR", 1057426, 30768},
 {"ID:212626428", 1070915, 31117},
 {"natRecSet isSet", 1071167, 31124},
 {"ID:433946629", 1071416, 31131},
 {"base \[Element] natRecSet", 1071682, 31138},
 {"ID:198301498", 1071935, 31145},
 {"natRecSet inductive", 1072190, 31152},
 {"ID:972417912", 1072438, 31159},
 {"natRecSet shape", 1072685, 31166},
 {"ID:57153239", 1072928, 31173},
 {"natRecSet cases", 1073174, 31180},
 {"ID:240965330", 1073417, 31187},
 {"natRecSet induction'", 1073676, 31194},
 {"ID:207844390", 1073927, 31201},
 {"natRecSet functional", 1074185, 31208},
 {"ID:348247310", 1074435, 31215},
 {"natRecSet shape'", 1074685, 31222},
 {"ID:117940020", 1074931, 31229},
 {"natRecSet total", 1075178, 31236},
 {"Proof|ID:79522091", 1075427, 31243},
 {"Proof|ID:79522091-1", 1075586, 31247},
 {"Proof|ID:212626428", 1075744, 31251},
 {"Proof|ID:212626428-1", 1075905, 31255},
 {"Proof|ID:433946629", 1076064, 31259},
 {"Proof|ID:433946629-1", 1076223, 31263},
 {"Proof|ID:198301498", 1076380, 31267},
 {"Proof|ID:198301498-1", 1076542, 31271},
 {"Proof|ID:972417912", 1076702, 31275},
 {"Proof|ID:972417912-1", 1076862, 31279},
 {"Proof|ID:57153239", 1077019, 31283},
 {"Proof|ID:57153239-1", 1077177, 31287},
 {"Proof|ID:240965330", 1077334, 31291},
 {"Proof|ID:240965330-1", 1077495, 31295},
 {"Proof|ID:207844390", 1077654, 31299},
 {"Proof|ID:207844390-1", 1077815, 31303},
 {"Proof|ID:348247310", 1077974, 31307},
 {"Proof|ID:348247310-1", 1078134, 31311},
 {"Proof|ID:117940020", 1078292, 31315},
 {"Proof|ID:117940020-1", 1078452, 31319},
 {"ID:127781311", 1078604, 31323},
 {"natRec base", 1078849, 31330},
 {"ID:186992137", 1079095, 31337},
 {"natRec step", 1079341, 31344},
 {"Proof|ID:127781311", 1079594, 31351},
 {"Proof|ID:127781311-1", 1079752, 31355},
 {"Proof|ID:186992137", 1079908, 31359},
 {"Proof|ID:186992137-1", 1080066, 31363},
 {"ID:379686829", 1080216, 31367},
 {"ID:778005153", 1080430, 31373},
 {"ID:97787251", 1080643, 31379},
 {"Nat \[Element]", 1080887, 31386},
 {"ID:720795295", 1081129, 31393},
 {"Nat 0", 1081356, 31400},
 {"ID:3778917", 1081588, 31407},
 {"Nat 1", 1081811, 31414},
 {"ID:54060264", 1082040, 31421},
 {"Nat Pred", 1082271, 31428},
 {"ID:205601935", 1082506, 31435},
 {"Nat +", 1082734, 31442},
 {"ID:74238769", 1082968, 31449},
 {"Nat -", 1083194, 31456},
 {"ID:288612943", 1083427, 31463},
 {"Nat \[RawStar]", 1083673, 31470},
 {"ID:155532301", 1083917, 31477},
 {"Nat \[GreaterEqual]", 1084173, 31484},
 {"ID:15394254", 1084421, 31491},
 {"Nat isDomain", 1084660, 31498},
 {"ID:478904098", 1084899, 31505},
 {"Nat induction", 1085149, 31512},
 {"ID:60850406", 1085397, 31519},
 {"0 \[Element] Nat", 1085650, 31526},
 {"ID:114723517", 1085899, 31533},
 {"succ \[Element] Nat", 1086161, 31540},
 {"ID:275276526", 1086416, 31547},
 {"1 \[Element] Nat", 1086671, 31554},
 {"ID:65979872", 1086921, 31561},
 {"ID:572443783", 1087133, 31567},
 {"Nat shape", 1087368, 31574},
 {"ID:252333509", 1087606, 31581},
 {"Nat cases", 1087842, 31588},
 {"ID:77756457", 1088080, 31595},
 {"succ \[NotEqual] 0 Nat", 1088339, 31602},
 {"ID:52299395", 1088587, 31609},
 {"succ injective Nat", 1088839, 31616},
 {"ID:263281689", 1089085, 31623},
 {"succ no fixpoint Nat", 1089342, 31630},
 {"ID:377798869", 1089591, 31637},
 {"1 \[NotEqual] 0 Nat", 1089846, 31644},
 {"ID:33739458", 1090093, 31651},
 {"\[Element] closed Nat", 1090350, 31658},
 {"ID:654854303", 1090598, 31665},
 {"Nat transitive", 1090844, 31672},
 {"Proof|ID:15394254", 1091093, 31679},
 {"Proof|ID:15394254-1", 1091250, 31683},
 {"Proof|ID:478904098", 1091406, 31687},
 {"Proof|ID:478904098-1", 1091566, 31691},
 {"Proof|ID:60850406", 1091723, 31695},
 {"Proof|ID:60850406-1", 1091880, 31699},
 {"Proof|ID:114723517", 1092036, 31703},
 {"Proof|ID:114723517-1", 1092196, 31707},
 {"Proof|ID:275276526", 1092354, 31711},
 {"Proof|ID:275276526-1", 1092513, 31715},
 {"Proof|ID:572443783", 1092670, 31719},
 {"Proof|ID:572443783-1", 1092829, 31723},
 {"Proof|ID:252333509", 1092986, 31727},
 {"Proof|ID:252333509-1", 1093146, 31731},
 {"Proof|ID:77756457", 1093303, 31735},
 {"Proof|ID:77756457-1", 1093459, 31739},
 {"Proof|ID:52299395", 1093613, 31743},
 {"Proof|ID:52299395-1", 1093770, 31747},
 {"Proof|ID:263281689", 1093926, 31751},
 {"Proof|ID:263281689-1", 1094086, 31755},
 {"Proof|ID:377798869", 1094244, 31759},
 {"Proof|ID:377798869-1", 1094404, 31763},
 {"Proof|ID:33739458", 1094561, 31767},
 {"Proof|ID:33739458-1", 1094718, 31771},
 {"Proof|ID:654854303", 1094874, 31775},
 {"Proof|ID:654854303-1", 1095032, 31779},
 {"ID:240175329", 1095182, 31783},
 {"Pred 0 Nat", 1095425, 31790},
 {"ID:158284383", 1095670, 31797},
 {"ID:145236688", 1095884, 31803},
 {"Pred succ Nat", 1096133, 31810},
 {"ID:225607097", 1096381, 31817},
 {"Pred closed Nat", 1096634, 31824},
 {"Proof|ID:240175329", 1096890, 31831},
 {"Proof|ID:240175329-1", 1097050, 31835},
 {"Proof|ID:145236688", 1097208, 31839},
 {"Proof|ID:145236688-1", 1097368, 31843},
 {"Proof|ID:225607097", 1097526, 31847},
 {"Proof|ID:225607097-1", 1097686, 31851},
 {"ID:218406943", 1097838, 31855},
 {"ID:257436422", 1098052, 31861},
 {"+ 0 Nat", 1098289, 31868},
 {"ID:249353684", 1098531, 31875},
 {"+ succ Nat", 1098775, 31882},
 {"ID:279658160", 1099021, 31889},
 {"+ closed Nat", 1099268, 31896},
 {"ID:647623443", 1099515, 31903},
 {"+ 1 Nat", 1099746, 31910},
 {"ID:18327536", 1099981, 31917},
 {"0 lneutral Nat", 1100230, 31924},
 {"ID:1914645", 1100475, 31931},
 {"succ + Nat", 1100714, 31938},
 {"ID:343580194", 1100955, 31945},
 {"+ = 0 Nat", 1101191, 31952},
 {"ID:834229092", 1101430, 31959},
 {"+ comm Nat", 1101667, 31966},
 {"ID:303913216", 1101906, 31973},
 {"ID:73705609", 1102119, 31979},
 {"+ assoc Nat", 1102357, 31986},
 {"ID:202544617", 1102596, 31993},
 {"+ comm-assoc Nat", 1102846, 32000},
 {"ID:555281086", 1103092, 32007},
 {"+ + + Nat", 1103328, 32014},
 {"ID:5229733", 1103565, 32021},
 {"+ rcanc Nat", 1103801, 32028},
 {"Proof|ID:257436422", 1104044, 32035},
 {"Proof|ID:257436422-1", 1104204, 32039},
 {"Proof|ID:249353684", 1104362, 32043},
 {"Proof|ID:249353684-1", 1104522, 32047},
 {"Proof|ID:279658160", 1104680, 32051},
 {"Proof|ID:279658160-1", 1104840, 32055},
 {"Proof|ID:647623443", 1104998, 32059},
 {"Proof|ID:647623443-1", 1105157, 32063},
 {"Proof|ID:18327536", 1105313, 32067},
 {"Proof|ID:18327536-1", 1105470, 32071},
 {"Proof|ID:1914645", 1105624, 32075},
 {"Proof|ID:1914645-1", 1105778, 32079},
 {"Proof|ID:343580194", 1105932, 32083},
 {"Proof|ID:343580194-1", 1106092, 32087},
 {"Proof|ID:834229092", 1106250, 32091},
 {"Proof|ID:834229092-1", 1106409, 32095},
 {"Proof|ID:73705609", 1106565, 32099},
 {"Proof|ID:73705609-1", 1106723, 32103},
 {"Proof|ID:202544617", 1106880, 32107},
 {"Proof|ID:202544617-1", 1107038, 32111},
 {"Proof|ID:555281086", 1107194, 32115},
 {"Proof|ID:555281086-1", 1107354, 32119},
 {"Proof|ID:5229733", 1107510, 32123},
 {"Proof|ID:5229733-1", 1107665, 32127},
 {"ID:301992507", 1107814, 32131},
 {"ID:99163941", 1108027, 32137},
 {"\[RawStar] 0 Nat", 1108280, 32144},
 {"ID:371484270", 1108529, 32151},
 {"\[RawStar] succ Nat", 1108790, 32158},
 {"ID:142418160", 1109044, 32165},
 {"\[RawStar] closed Nat", 1109309, 32172},
 {"ID:21383846", 1109564, 32179},
 {"1 rneutral Nat", 1109813, 32186},
 {"ID:28953873", 1110059, 32193},
 {"0 lannihilator Nat", 1110316, 32200},
 {"ID:218775165", 1110567, 32207},
 {"succ \[RawStar] Nat", 1110828, 32214},
 {"ID:470439981", 1111082, 32221},
 {"\[RawStar] zdf Nat", 1111336, 32228},
 {"ID:218136211", 1111584, 32235},
 {"\[RawStar] comm Nat", 1111839, 32242},
 {"ID:666029340", 1112087, 32249},
 {"ID:43150914", 1112300, 32255},
 {"\[RawStar] distr + Nat", 1112560, 32262},
 {"ID:475421150", 1112810, 32269},
 {"\[RawStar] assoc Nat", 1113068, 32276},
 {"ID:98508460", 1113317, 32283},
 {"\[RawStar] comm-assoc Nat", 1113583, 32290},
 {"ID:138249572", 1113836, 32297},
 {"\[RawStar] rcanc Nat", 1114094, 32304},
 {"Proof|ID:99163941", 1114349, 32311},
 {"Proof|ID:99163941-1", 1114506, 32315},
 {"Proof|ID:371484270", 1114662, 32319},
 {"Proof|ID:371484270-1", 1114820, 32323},
 {"Proof|ID:142418160", 1114976, 32327},
 {"Proof|ID:142418160-1", 1115135, 32331},
 {"Proof|ID:21383846", 1115291, 32335},
 {"Proof|ID:21383846-1", 1115447, 32339},
 {"Proof|ID:28953873", 1115601, 32343},
 {"Proof|ID:28953873-1", 1115757, 32347},
 {"Proof|ID:218775165", 1115912, 32351},
 {"Proof|ID:218775165-1", 1116071, 32355},
 {"Proof|ID:470439981", 1116228, 32359},
 {"Proof|ID:470439981-1", 1116389, 32363},
 {"Proof|ID:218136211", 1116548, 32367},
 {"Proof|ID:218136211-1", 1116709, 32371},
 {"Proof|ID:43150914", 1116867, 32375},
 {"Proof|ID:43150914-1", 1117025, 32379},
 {"Proof|ID:475421150", 1117182, 32383},
 {"Proof|ID:475421150-1", 1117344, 32387},
 {"Proof|ID:98508460", 1117503, 32391},
 {"Proof|ID:98508460-1", 1117661, 32395},
 {"Proof|ID:138249572", 1117818, 32399},
 {"Proof|ID:138249572-1", 1117980, 32403},
 {"ID:621974620", 1118134, 32407},
 {"ID:380981542", 1118349, 32413},
 {"0 \[GreaterEqual] Nat", 1118616, 32420},
 {"ID:42977349", 1118873, 32427},
 {"succ \[GreaterEqual] Nat", 1119144, 32434},
 {"ID:428740350", 1119403, 32441},
 {"\[GreaterEqual] reflexive Nat", 1119679, 32448},
 {"ID:297686800", 1119938, 32455},
 {"\[GreaterEqual] 0 Nat", 1120204, 32462},
 {"ID:585961470", 1120461, 32469},
 {"\[GreaterEqual] succ Nat", 1120734, 32476},
 {"ID:894887193", 1120995, 32483},
 {"\[GreaterEqual] antisymmetric Nat", 1121280, 32490},
 {"ID:208533627", 1121544, 32497},
 {"succ increasing Nat", 1121800, 32504},
 {"ID:112605784", 1122049, 32511},
 {"succ \[GreaterEqual] succ Nat", 1122332, 32518},
 {"ID:304315076", 1122598, 32525},
 {"\[GreaterEqual] total Nat", 1122866, 32532},
 {"ID:446896782", 1123121, 32539},
 {"\[GreaterEqual] \[Exists] Nat", 1123398, 32546},
 {"ID:196203313", 1123658, 32553},
 {"1 \[GreaterEqual] 0 Nat", 1123922, 32560},
 {"ID:30448165", 1124174, 32567},
 {"\[GreaterEqual] 1 Nat", 1124439, 32574},
 {"ID:222994735", 1124695, 32581},
 {"ID:546802157", 1124910, 32587},
 {"\[GreaterEqual] transitive Nat", 1125189, 32594},
 {"ID:352201475", 1125450, 32601},
 {"\[GreaterEqual] + \[DoubleLeftArrow] Nat", 1125749, 32608},
 {"ID:97128652", 1126019, 32615},
 {"\[GreaterEqual] + Nat", 1126278, 32622},
 {"ID:278931212", 1126528, 32629},
 {"+ \[GreaterEqual] Nat", 1126788, 32636},
 {"ID:335211897", 1127039, 32643},
 {"\[GreaterEqual] \[RawStar] Nat", 1127318, 32650},
 {"ID:32863246", 1127578, 32657},
 {"\[RawStar] \[GreaterEqual] Nat", 1127854, 32664},
 {"Proof|ID:380981542", 1128118, 32671},
 {"Proof|ID:380981542-1", 1128279, 32675},
 {"Proof|ID:42977349", 1128437, 32679},
 {"Proof|ID:42977349-1", 1128596, 32683},
 {"Proof|ID:428740350", 1128754, 32687},
 {"Proof|ID:428740350-1", 1128914, 32691},
 {"Proof|ID:297686800", 1129072, 32695},
 {"Proof|ID:297686800-1", 1129232, 32699},
 {"Proof|ID:585961470", 1129390, 32703},
 {"Proof|ID:585961470-1", 1129552, 32707},
 {"Proof|ID:894887193", 1129712, 32711},
 {"Proof|ID:894887193-1", 1129873, 32715},
 {"Proof|ID:208533627", 1130032, 32719},
 {"Proof|ID:208533627-1", 1130192, 32723},
 {"Proof|ID:112605784", 1130350, 32727},
 {"Proof|ID:112605784-1", 1130511, 32731},
 {"Proof|ID:304315076", 1130670, 32735},
 {"Proof|ID:304315076-1", 1130831, 32739},
 {"Proof|ID:446896782", 1130990, 32743},
 {"Proof|ID:446896782-1", 1131152, 32747},
 {"Proof|ID:196203313", 1131312, 32751},
 {"Proof|ID:196203313-1", 1131472, 32755},
 {"Proof|ID:30448165", 1131629, 32759},
 {"Proof|ID:30448165-1", 1131786, 32763},
 {"Proof|ID:546802157", 1131942, 32767},
 {"Proof|ID:546802157-1", 1132104, 32771},
 {"Proof|ID:352201475", 1132264, 32775},
 {"Proof|ID:352201475-1", 1132425, 32779},
 {"Proof|ID:97128652", 1132583, 32783},
 {"Proof|ID:97128652-1", 1132742, 32787},
 {"Proof|ID:278931212", 1132900, 32791},
 {"Proof|ID:278931212-1", 1133061, 32795},
 {"Proof|ID:335211897", 1133220, 32799},
 {"Proof|ID:335211897-1", 1133382, 32803},
 {"Proof|ID:32863246", 1133541, 32807},
 {"Proof|ID:32863246-1", 1133699, 32811},
 {"ID:609763659", 1133850, 32815},
 {"ID:19208428", 1134064, 32821},
 {"- 0 Nat", 1134300, 32828},
 {"ID:593421999", 1134541, 32835},
 {"- succ Nat", 1134785, 32842},
 {"ID:153357458", 1135031, 32849},
 {"- closed Nat", 1135279, 32856},
 {"ID:616125500", 1135527, 32863},
 {"- 1 Nat", 1135759, 32870},
 {"ID:491337343", 1135996, 32877},
 {"0 - Nat", 1136234, 32884},
 {"ID:621700658", 1136477, 32891},
 {"succ - succ Nat", 1136732, 32898},
 {"ID:236021775", 1136984, 32905},
 {"+ - Nat", 1137216, 32912},
 {"ID:3006634", 1137451, 32919},
 {"- + Nat", 1137679, 32926},
 {"ID:80465146", 1137911, 32933},
 {"ID:404996142", 1138124, 32939},
 {"+ - + Nat", 1138360, 32946},
 {"ID:225131303", 1138599, 32953},
 {"\[RawStar] distr - Nat", 1138862, 32960},
 {"ID:142079924", 1139115, 32967},
 {"(-)+ Nat", 1139350, 32974},
 {"ID:367201929", 1139589, 32981},
 {"- = - Nat", 1139826, 32988},
 {"Proof|ID:19208428", 1140071, 32995},
 {"Proof|ID:19208428-1", 1140228, 32999},
 {"Proof|ID:593421999", 1140384, 33003},
 {"Proof|ID:593421999-1", 1140545, 33007},
 {"Proof|ID:153357458", 1140704, 33011},
 {"Proof|ID:153357458-1", 1140865, 33015},
 {"Proof|ID:616125500", 1141024, 33019},
 {"Proof|ID:616125500-1", 1141184, 33023},
 {"Proof|ID:491337343", 1141342, 33027},
 {"Proof|ID:491337343-1", 1141503, 33031},
 {"Proof|ID:621700658", 1141662, 33035},
 {"Proof|ID:621700658-1", 1141822, 33039},
 {"Proof|ID:236021775", 1141980, 33043},
 {"Proof|ID:236021775-1", 1142141, 33047},
 {"Proof|ID:3006634", 1142298, 33051},
 {"Proof|ID:3006634-1", 1142454, 33055},
 {"Proof|ID:404996142", 1142610, 33059},
 {"Proof|ID:404996142-1", 1142772, 33063},
 {"Proof|ID:225131303", 1142932, 33067},
 {"Proof|ID:225131303-1", 1143094, 33071},
 {"Proof|ID:142079924", 1143254, 33075},
 {"Proof|ID:142079924-1", 1143416, 33079},
 {"Proof|ID:367201929", 1143576, 33083},
 {"Proof|ID:367201929-1", 1143738, 33087},
 {"ID:513393113", 1143892, 33091},
 {"ID:623651727", 1144107, 33097},
 {"ID:100384801", 1144322, 33103},
 {"PreInt \[Element]", 1144575, 33110},
 {"ID:398024454", 1144823, 33117},
 {"PreInt 0", 1145057, 33124},
 {"ID:418726841", 1145295, 33131},
 {"PreInt 1", 1145529, 33138},
 {"ID:10580372", 1145766, 33145},
 {"PreInt +", 1145999, 33152},
 {"ID:106096058", 1146236, 33159},
 {"PreInt - unary", 1146482, 33166},
 {"ID:12727554", 1146725, 33173},
 {"PreInt \[RawStar]", 1146976, 33180},
 {"ID:329154583", 1147222, 33187},
 {"PreInt \[GreaterEqual]", 1147485, 33194},
 {"ID:468243965", 1147738, 33201},
 {"PreInt equiv", 1147981, 33208},
 {"ID:168445606", 1148224, 33215},
 {"PreInt isDomain", 1148478, 33222},
 {"ID:977301511", 1148729, 33229},
 {"\[Element] PreInt alternative", 1149012, 33236},
 {"ID:353341722", 1149278, 33243},
 {"Pair \[Element] PreInt", 1149547, 33250},
 {"ID:34141802", 1149805, 33257},
 {"0 \[Element] PreInt", 1150065, 33264},
 {"ID:553448562", 1150318, 33271},
 {"1 \[Element] PreInt", 1150580, 33278},
 {"Proof|ID:168445606", 1150841, 33285},
 {"Proof|ID:168445606-1", 1151002, 33289},
 {"Proof|ID:977301511", 1151161, 33293},
 {"Proof|ID:977301511-1", 1151322, 33297},
 {"Proof|ID:353341722", 1151481, 33301},
 {"Proof|ID:353341722-1", 1151642, 33305},
 {"Proof|ID:34141802", 1151800, 33309},
 {"Proof|ID:34141802-1", 1151957, 33313},
 {"Proof|ID:553448562", 1152113, 33317},
 {"Proof|ID:553448562-1", 1152274, 33321},
 {"ID:220417861", 1152427, 33325},
 {"equiv Pair PreInt", 1152686, 33332},
 {"ID:52539739", 1152939, 33339},
 {"equiv 0 PreInt", 1153190, 33346},
 {"ID:94484613", 1153438, 33353},
 {"equiv Nat \[GreaterEqual] PreInt", 1153719, 33360},
 {"ID:321256465", 1153980, 33367},
 {"equiv Nat \[LessEqual] PreInt", 1154257, 33374},
 {"ID:129809121", 1154517, 33381},
 {"\[Not] 1 equiv 0 PreInt", 1154781, 33388},
 {"ID:413771523", 1155034, 33395},
 {"ID:637862621", 1155249, 33401},
 {"equiv Nat PreInt", 1155500, 33408},
 {"ID:12369324", 1155746, 33415},
 {"equiv reflexive PreInt", 1156006, 33422},
 {"ID:19276", 1156252, 33429},
 {"equiv symmetric PreInt", 1156506, 33436},
 {"ID:164883544", 1156750, 33443},
 {"equiv transitive PreInt", 1157015, 33450},
 {"ID:357440757", 1157269, 33457},
 {"equiv isEquivalence PreInt", 1157539, 33464},
 {"Proof|ID:220417861", 1157801, 33471},
 {"Proof|ID:220417861-1", 1157963, 33475},
 {"Proof|ID:52539739", 1158122, 33479},
 {"Proof|ID:52539739-1", 1158280, 33483},
 {"Proof|ID:94484613", 1158436, 33487},
 {"Proof|ID:94484613-1", 1158595, 33491},
 {"Proof|ID:321256465", 1158753, 33495},
 {"Proof|ID:321256465-1", 1158915, 33499},
 {"Proof|ID:129809121", 1159075, 33503},
 {"Proof|ID:129809121-1", 1159236, 33507},
 {"Proof|ID:637862621", 1159395, 33511},
 {"Proof|ID:637862621-1", 1159557, 33515},
 {"Proof|ID:12369324", 1159716, 33519},
 {"Proof|ID:12369324-1", 1159873, 33523},
 {"Proof|ID:19276", 1160025, 33527},
 {"Proof|ID:19276-1", 1160174, 33531},
 {"Proof|ID:164883544", 1160325, 33535},
 {"Proof|ID:164883544-1", 1160486, 33539},
 {"Proof|ID:357440757", 1160645, 33543},
 {"Proof|ID:357440757-1", 1160805, 33547},
 {"ID:125790650", 1160957, 33551},
 {"+ Pair PreInt", 1161208, 33558},
 {"ID:86057977", 1161457, 33565},
 {"ID:196108575", 1161670, 33571},
 {"+ closed PreInt", 1161918, 33578},
 {"ID:132949647", 1162163, 33585},
 {"0 rneutral PreInt", 1162415, 33592},
 {"ID:25630206", 1162661, 33599},
 {"+ comm PreInt", 1162903, 33606},
 {"ID:16604231", 1163143, 33613},
 {"ID:73792979", 1163355, 33619},
 {"+ assoc PreInt", 1163600, 33626},
 {"ID:243295365", 1163843, 33633},
 {"+ rcanc PreInt", 1164090, 33640},
 {"ID:47524840", 1164334, 33647},
 {"+ invaiant equiv PreInt", 1164597, 33654},
 {"ID:107686175", 1164849, 33661},
 {"+ isInvariantFunction2 PreInt", 1165126, 33668},
 {"Proof|ID:125790650", 1165392, 33675},
 {"Proof|ID:125790650-1", 1165554, 33679},
 {"Proof|ID:196108575", 1165714, 33683},
 {"Proof|ID:196108575-1", 1165875, 33687},
 {"Proof|ID:132949647", 1166034, 33691},
 {"Proof|ID:132949647-1", 1166194, 33695},
 {"Proof|ID:25630206", 1166351, 33699},
 {"Proof|ID:25630206-1", 1166509, 33703},
 {"Proof|ID:73792979", 1166665, 33707},
 {"Proof|ID:73792979-1", 1166824, 33711},
 {"Proof|ID:243295365", 1166982, 33715},
 {"Proof|ID:243295365-1", 1167144, 33719},
 {"Proof|ID:47524840", 1167303, 33723},
 {"Proof|ID:47524840-1", 1167462, 33727},
 {"Proof|ID:107686175", 1167620, 33731},
 {"Proof|ID:107686175-1", 1167781, 33735},
 {"ID:360943410", 1167934, 33739},
 {"- unary Pair PreInt", 1168197, 33746},
 {"ID:474936069", 1168453, 33753},
 {"ID:5265224", 1168666, 33759},
 {"- unary closed PreInt", 1168928, 33766},
 {"ID:203475906", 1169181, 33773},
 {"- inverse PreInt", 1169432, 33780},
 {"ID:611704112", 1169679, 33787},
 {"- isInvariantFunction1 PreInt", 1169955, 33794},
 {"Proof|ID:360943410", 1170220, 33801},
 {"Proof|ID:360943410-1", 1170381, 33805},
 {"Proof|ID:5265224", 1170538, 33809},
 {"Proof|ID:5265224-1", 1170693, 33813},
 {"Proof|ID:203475906", 1170848, 33817},
 {"Proof|ID:203475906-1", 1171009, 33821},
 {"Proof|ID:611704112", 1171168, 33825},
 {"Proof|ID:611704112-1", 1171329, 33829},
 {"ID:455358938", 1171482, 33833},
 {"\[RawStar] Pair PreInt", 1171751, 33840},
 {"ID:206670450", 1172010, 33847},
 {"ID:80978087", 1172224, 33853},
 {"\[RawStar] closed PreInt", 1172494, 33860},
 {"ID:561254237", 1172752, 33867},
 {"0 rannihilator PreInt", 1173012, 33874},
 {"ID:201948360", 1173263, 33881},
 {"1 rneutral PreInt", 1173515, 33888},
 {"ID:152728157", 1173762, 33895},
 {"\[RawStar] comm PreInt", 1174024, 33902},
 {"ID:258603229", 1174276, 33909},
 {"\[RawStar] zdf PreInt", 1174537, 33916},
 {"ID:200297603", 1174789, 33923},
 {"ID:142359696", 1175004, 33929},
 {"\[RawStar] assoc PreInt", 1175269, 33936},
 {"ID:262630475", 1175523, 33943},
 {"\[RawStar] distr + PreInt", 1175792, 33950},
 {"ID:146416852", 1176048, 33957},
 {"\[RawStar] invariant equiv PreInt", 1176333, 33964},
 {"ID:204021385", 1176597, 33971},
 {"\[RawStar] isInvariantFunction2 PreInt", 1176892, 33978},
 {"Proof|ID:455358938", 1177167, 33985},
 {"Proof|ID:455358938-1", 1177329, 33989},
 {"Proof|ID:80978087", 1177488, 33993},
 {"Proof|ID:80978087-1", 1177646, 33997},
 {"Proof|ID:561254237", 1177803, 34001},
 {"Proof|ID:561254237-1", 1177964, 34005},
 {"Proof|ID:201948360", 1178123, 34009},
 {"Proof|ID:201948360-1", 1178284, 34013},
 {"Proof|ID:152728157", 1178443, 34017},
 {"Proof|ID:152728157-1", 1178603, 34021},
 {"Proof|ID:258603229", 1178761, 34025},
 {"Proof|ID:258603229-1", 1178923, 34029},
 {"Proof|ID:142359696", 1179083, 34033},
 {"Proof|ID:142359696-1", 1179242, 34037},
 {"Proof|ID:262630475", 1179399, 34041},
 {"Proof|ID:262630475-1", 1179561, 34045},
 {"Proof|ID:146416852", 1179721, 34049},
 {"Proof|ID:146416852-1", 1179882, 34053},
 {"Proof|ID:204021385", 1180041, 34057},
 {"Proof|ID:204021385-1", 1180202, 34061},
 {"ID:129935970", 1180355, 34065},
 {"\[GreaterEqual] Pair PreInt", 1180634, 34072},
 {"ID:95876614", 1180897, 34079},
 {"\[GreaterEqual] 0 PreInt", 1181168, 34086},
 {"ID:221338547", 1181427, 34093},
 {"1 \[GreaterEqual] 0 PreInt", 1181697, 34100},
 {"ID:13569836", 1181952, 34107},
 {"\[GreaterEqual] isInvariantRelation2 PreInt", 1182255, 34114},
 {"ID:337456597", 1182527, 34121},
 {"ID:83533617", 1182741, 34127},
 {"\[GreaterEqual] reflexive PreInt", 1183021, 34134},
 {"ID:281735911", 1183281, 34141},
 {"\[GreaterEqual] antisymmetric PreInt", 1183572, 34148},
 {"ID:115550209", 1183839, 34155},
 {"\[GreaterEqual] total PreInt", 1184113, 34162},
 {"ID:215340491", 1184371, 34169},
 {"\[GreaterEqual] discrete PreInt", 1184652, 34176},
 {"ID:92075332", 1184913, 34183},
 {"\[GreaterEqual] \[RawStar] PreInt", 1185196, 34190},
 {"ID:145078248", 1185458, 34197},
 {"Nat \[SucceedsTilde] PreInt+", 1185733, 34204},
 {"ID:118070489", 1185992, 34211},
 {"ID:307661531", 1186207, 34217},
 {"\[GreaterEqual] transitive PreInt", 1186492, 34224},
 {"ID:647932785", 1186756, 34231},
 {"\[GreaterEqual] + PreInt", 1187023, 34238},
 {"Proof|ID:129935970", 1187284, 34245},
 {"Proof|ID:129935970-1", 1187444, 34249},
 {"Proof|ID:95876614", 1187601, 34253},
 {"Proof|ID:95876614-1", 1187758, 34257},
 {"Proof|ID:221338547", 1187914, 34261},
 {"Proof|ID:221338547-1", 1188074, 34265},
 {"Proof|ID:13569836", 1188231, 34269},
 {"Proof|ID:13569836-1", 1188389, 34273},
 {"Proof|ID:83533617", 1188545, 34277},
 {"Proof|ID:83533617-1", 1188701, 34281},
 {"Proof|ID:281735911", 1188856, 34285},
 {"Proof|ID:281735911-1", 1189017, 34289},
 {"Proof|ID:115550209", 1189176, 34293},
 {"Proof|ID:115550209-1", 1189337, 34297},
 {"Proof|ID:215340491", 1189496, 34301},
 {"Proof|ID:215340491-1", 1189658, 34305},
 {"Proof|ID:92075332", 1189817, 34309},
 {"Proof|ID:92075332-1", 1189975, 34313},
 {"Proof|ID:145078248", 1190132, 34317},
 {"Proof|ID:145078248-1", 1190294, 34321},
 {"Proof|ID:307661531", 1190454, 34325},
 {"Proof|ID:307661531-1", 1190615, 34329},
 {"Proof|ID:647932785", 1190774, 34333},
 {"Proof|ID:647932785-1", 1190934, 34337},
 {"ID:463204962", 1191086, 34341},
 {"ID:32325634", 1191300, 34347},
 {"ID:112681936", 1191513, 34353},
 {"Int \[Element]", 1191760, 34360},
 {"ID:52976508", 1192004, 34367},
 {"Int 0", 1192230, 34374},
 {"ID:712864778", 1192463, 34381},
 {"Int 1", 1192691, 34388},
 {"ID:19419340", 1192925, 34395},
 {"Int +", 1193152, 34402},
 {"ID:280798627", 1193386, 34409},
 {"Int - unary", 1193626, 34416},
 {"ID:328192039", 1193867, 34423},
 {"Int \[RawStar]", 1194114, 34430},
 {"ID:819512290", 1194359, 34437},
 {"Int \[GreaterEqual]", 1194616, 34444},
 {"ID:354647619", 1194866, 34451},
 {"Int isDomain", 1195114, 34458},
 {"ID:565027904", 1195362, 34465},
 {"\[Element] Int alternative", 1195639, 34472},
 {"ID:40757894", 1195901, 34479},
 {"EC \[Element] Int", 1196158, 34486},
 {"ID:350791961", 1196410, 34493},
 {"0 \[Element] Int", 1196666, 34500},
 {"ID:181240486", 1196918, 34507},
 {"1 \[Element] Int", 1197168, 34514},
 {"ID:364462666", 1197414, 34521},
 {"1 \[NotEqual] 0 Int", 1197670, 34528},
 {"Proof|ID:354647619", 1197925, 34535},
 {"Proof|ID:354647619-1", 1198086, 34539},
 {"Proof|ID:565027904", 1198245, 34543},
 {"Proof|ID:565027904-1", 1198407, 34547},
 {"Proof|ID:40757894", 1198566, 34551},
 {"Proof|ID:40757894-1", 1198723, 34555},
 {"Proof|ID:350791961", 1198879, 34559},
 {"Proof|ID:350791961-1", 1199039, 34563},
 {"Proof|ID:181240486", 1199197, 34567},
 {"Proof|ID:181240486-1", 1199358, 34571},
 {"Proof|ID:364462666", 1199517, 34575},
 {"Proof|ID:364462666-1", 1199678, 34579},
 {"ID:56650873", 1199830, 34583},
 {"+ EC Int", 1200069, 34590},
 {"ID:13236461", 1200311, 34597},
 {"ID:562490241", 1200524, 34603},
 {"+ closed Int", 1200772, 34610},
 {"ID:624613835", 1201020, 34617},
 {"0 rneutral Int", 1201266, 34624},
 {"ID:115787764", 1201510, 34631},
 {"+ comm Int", 1201748, 34638},
 {"ID:11815767", 1201987, 34645},
 {"+ assoc Int", 1202226, 34652},
 {"ID:139539770", 1202466, 34659},
 {"+ rcanc Int", 1202706, 34666},
 {"Proof|ID:56650873", 1202952, 34673},
 {"Proof|ID:56650873-1", 1203111, 34677},
 {"Proof|ID:562490241", 1203269, 34681},
 {"Proof|ID:562490241-1", 1203430, 34685},
 {"Proof|ID:624613835", 1203589, 34689},
 {"Proof|ID:624613835-1", 1203749, 34693},
 {"Proof|ID:115787764", 1203907, 34697},
 {"Proof|ID:115787764-1", 1204068, 34701},
 {"Proof|ID:11815767", 1204226, 34705},
 {"Proof|ID:11815767-1", 1204384, 34709},
 {"Proof|ID:139539770", 1204541, 34713},
 {"Proof|ID:139539770-1", 1204702, 34717},
 {"ID:331988502", 1204855, 34721},
 {"- unary EC Int", 1205108, 34728},
 {"ID:138388913", 1205359, 34735},
 {"ID:789456798", 1205574, 34741},
 {"- unary closed Int", 1205834, 34748},
 {"ID:561244613", 1206088, 34755},
 {"- inverse Int", 1206332, 34762},
 {"Proof|ID:331988502", 1206581, 34769},
 {"Proof|ID:331988502-1", 1206742, 34773},
 {"Proof|ID:789456798", 1206901, 34777},
 {"Proof|ID:789456798-1", 1207062, 34781},
 {"Proof|ID:561244613", 1207221, 34785},
 {"Proof|ID:561244613-1", 1207382, 34789},
 {"ID:947607354", 1207535, 34793},
 {"\[RawStar] EC Int", 1207794, 34800},
 {"ID:105543058", 1208048, 34807},
 {"ID:696766377", 1208263, 34813},
 {"\[RawStar] closed Int", 1208529, 34820},
 {"ID:202378771", 1208786, 34827},
 {"0 rannihilator Int", 1209040, 34834},
 {"ID:689609352", 1209288, 34841},
 {"1 rneutral Int", 1209534, 34848},
 {"ID:196343023", 1209778, 34855},
 {"\[RawStar] comm Int", 1210034, 34862},
 {"ID:207834126", 1210283, 34869},
 {"\[RawStar] zdf Int", 1210538, 34876},
 {"ID:10196611", 1210786, 34883},
 {"\[RawStar] assoc Int", 1211042, 34890},
 {"ID:777222529", 1211290, 34897},
 {"\[RawStar] distr + Int", 1211553, 34904},
 {"Proof|ID:947607354", 1211812, 34911},
 {"Proof|ID:947607354-1", 1211974, 34915},
 {"Proof|ID:696766377", 1212134, 34919},
 {"Proof|ID:696766377-1", 1212294, 34923},
 {"Proof|ID:202378771", 1212452, 34927},
 {"Proof|ID:202378771-1", 1212613, 34931},
 {"Proof|ID:689609352", 1212772, 34935},
 {"Proof|ID:689609352-1", 1212933, 34939},
 {"Proof|ID:196343023", 1213092, 34943},
 {"Proof|ID:196343023-1", 1213252, 34947},
 {"Proof|ID:207834126", 1213410, 34951},
 {"Proof|ID:207834126-1", 1213569, 34955},
 {"Proof|ID:10196611", 1213725, 34959},
 {"Proof|ID:10196611-1", 1213884, 34963},
 {"Proof|ID:777222529", 1214042, 34967},
 {"Proof|ID:777222529-1", 1214204, 34971},
 {"ID:785793765", 1214358, 34975},
 {"\[GreaterEqual] EC Int", 1214627, 34982},
 {"ID:72386745", 1214885, 34989},
 {"1 \[GreaterEqual] 0 Int", 1215147, 34996},
 {"ID:100510373", 1215398, 35003},
 {"ID:267773274", 1215613, 35009},
 {"\[GreaterEqual] reflexive Int", 1215889, 35016},
 {"ID:69005952", 1216147, 35023},
 {"\[GreaterEqual] antisymmetric Int", 1216430, 35030},
 {"ID:275182510", 1216692, 35037},
 {"\[GreaterEqual] transitive Int", 1216971, 35044},
 {"ID:176048001", 1217232, 35051},
 {"\[GreaterEqual] total Int", 1217500, 35058},
 {"ID:566590035", 1217755, 35065},
 {"\[GreaterEqual] discrete Int", 1218030, 35072},
 {"ID:262876021", 1218289, 35079},
 {"\[GreaterEqual] + Int", 1218550, 35086},
 {"ID:34443687", 1218801, 35093},
 {"\[GreaterEqual] \[RawStar] Int", 1219078, 35100},
 {"Proof|ID:785793765", 1219343, 35107},
 {"Proof|ID:785793765-1", 1219505, 35111},
 {"Proof|ID:72386745", 1219664, 35115},
 {"Proof|ID:72386745-1", 1219821, 35119},
 {"Proof|ID:267773274", 1219977, 35123},
 {"Proof|ID:267773274-1", 1220137, 35127},
 {"Proof|ID:69005952", 1220294, 35131},
 {"Proof|ID:69005952-1", 1220451, 35135},
 {"Proof|ID:275182510", 1220607, 35139},
 {"Proof|ID:275182510-1", 1220768, 35143},
 {"Proof|ID:176048001", 1220927, 35147},
 {"Proof|ID:176048001-1", 1221088, 35151},
 {"Proof|ID:566590035", 1221247, 35155},
 {"Proof|ID:566590035-1", 1221407, 35159},
 {"Proof|ID:262876021", 1221565, 35163},
 {"Proof|ID:262876021-1", 1221727, 35167},
 {"Proof|ID:34443687", 1221886, 35171},
 {"Proof|ID:34443687-1", 1222045, 35175},
 {"ID:134029858", 1222197, 35179},
 {"Nat \[SucceedsTilde] Int+", 1222466, 35186},
 {"ID:181242736", 1222722, 35193},
 {"Int+ induction", 1222969, 35200},
 {"Proof|ID:134029858", 1223220, 35207},
 {"Proof|ID:134029858-1", 1223382, 35211},
 {"Proof|ID:181242736", 1223542, 35215},
 {"Proof|ID:181242736-1", 1223704, 35219}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[579, 22, 113, 2, 84, "Title",
 CellID->21108537],
Cell[695, 26, 185, 3, 52, "Subtitle",
 CellID->138585177],
Cell[CellGroupData[{
Cell[905, 33, 116, 2, 73, "Section",
 CellID->909929215],
Cell[CellGroupData[{
Cell[1046, 39, 170, 3, 38, "Subsection",
 CellID->731805898],
Cell[1219, 44, 499, 9, 62, "Text",
 CellID->581780680],
Cell[1721, 55, 2187, 57, 138, "Text",
 CellID->902968620],
Cell[3911, 114, 4546, 105, 327, "Text",
 CellID->438330771]
}, Closed]],
Cell[CellGroupData[{
Cell[8494, 224, 125, 2, 30, "Subsection",
 CellID->102004898],
Cell[8622, 228, 393, 10, 32, "Input",
 CellID->498486718],
Cell[9018, 240, 355, 8, 32, "Input",
 CellID->406742757],
Cell[9376, 250, 419, 9, 32, "Input",
 CellID->6977455],
Cell[9798, 261, 330, 8, 32, "Input",
 CellID->139193289],
Cell[10131, 271, 151, 3, 32, "Input",
 CellID->182306024],
Cell[10285, 276, 326, 8, 32, "Input",
 CellID->20362838],
Cell[CellGroupData[{
Cell[10636, 288, 740, 11, 32, "Input",
 CellID->20666029],
Cell[11379, 301, 2524, 35, 32, "Output",
 CellID->569800073]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[13952, 342, 124, 2, 30, "Subsection",
 CellID->1386333768],
Cell[CellGroupData[{
Cell[14101, 348, 978, 23, 99, "Input",
 CellID->239327779],
Cell[CellGroupData[{
Cell[15104, 375, 583, 13, 23, "Print",
 CellID->1996807139],
Cell[15690, 390, 625, 14, 23, "Print",
 CellID->1264118277],
Cell[16318, 406, 323, 7, 23, "Print",
 CellID->1741093491],
Cell[16644, 415, 323, 7, 23, "Print",
 CellID->1767534958],
Cell[16970, 424, 603, 14, 23, "Print",
 CellID->1570628282],
Cell[17576, 440, 320, 7, 23, "Print",
 CellID->1431522614],
Cell[17899, 449, 603, 14, 23, "Print",
 CellID->46013314],
Cell[18505, 465, 322, 7, 23, "Print",
 CellID->413353689],
Cell[18830, 474, 325, 7, 23, "Print",
 CellID->1417273774],
Cell[19158, 483, 327, 7, 23, "Print",
 CellID->1414858982],
Cell[19488, 492, 326, 7, 23, "Print",
 CellID->632179620],
Cell[19817, 501, 326, 7, 23, "Print",
 CellID->445177995],
Cell[20146, 510, 598, 14, 23, "Print",
 CellID->256427531],
Cell[20747, 526, 325, 7, 23, "Print",
 CellID->1702810782],
Cell[21075, 535, 326, 7, 23, "Print",
 CellID->610477982],
Cell[21404, 544, 326, 7, 23, "Print",
 CellID->437563033],
Cell[21733, 553, 326, 7, 23, "Print",
 CellID->881701306],
Cell[22062, 562, 326, 7, 23, "Print",
 CellID->1547923834],
Cell[22391, 571, 327, 7, 23, "Print",
 CellID->1241639991],
Cell[22721, 580, 617, 14, 23, "Print",
 CellID->1907575484],
Cell[23341, 596, 321, 7, 23, "Print",
 CellID->25744908],
Cell[23665, 605, 325, 7, 23, "Print",
 CellID->1066639784],
Cell[23993, 614, 327, 7, 23, "Print",
 CellID->1142307976],
Cell[24323, 623, 326, 7, 23, "Print",
 CellID->559336593],
Cell[24652, 632, 326, 7, 23, "Print",
 CellID->1038330147],
Cell[24981, 641, 602, 14, 23, "Print",
 CellID->818971569],
Cell[25586, 657, 323, 7, 23, "Print",
 CellID->1237330945],
Cell[25912, 666, 324, 7, 23, "Print",
 CellID->265312282],
Cell[26239, 675, 327, 7, 23, "Print",
 CellID->2034025804],
Cell[26569, 684, 325, 7, 23, "Print",
 CellID->299202341],
Cell[26897, 693, 606, 14, 23, "Print",
 CellID->225271960],
Cell[27506, 709, 319, 7, 23, "Print",
 CellID->26177497],
Cell[27828, 718, 598, 14, 23, "Print",
 CellID->995664070],
Cell[28429, 734, 325, 7, 23, "Print",
 CellID->1802485517],
Cell[28757, 743, 324, 7, 23, "Print",
 CellID->102721391],
Cell[29084, 752, 326, 7, 23, "Print",
 CellID->772960722],
Cell[29413, 761, 326, 7, 23, "Print",
 CellID->545971667],
Cell[29742, 770, 329, 7, 23, "Print",
 CellID->1544609497],
Cell[30074, 779, 327, 7, 23, "Print",
 CellID->1356748633],
Cell[30404, 788, 588, 13, 23, "Print",
 CellID->872782520],
Cell[30995, 803, 603, 13, 23, "Print",
 CellID->2040460043],
Cell[31601, 818, 320, 7, 23, "Print",
 CellID->1123372340],
Cell[31924, 827, 615, 13, 23, "Print",
 CellID->13998605],
Cell[32542, 842, 322, 7, 23, "Print",
 CellID->244795172],
Cell[32867, 851, 324, 7, 23, "Print",
 CellID->303869597],
Cell[33194, 860, 327, 7, 23, "Print",
 CellID->1680057154],
Cell[33524, 869, 325, 7, 23, "Print",
 CellID->415121188],
Cell[33852, 878, 628, 13, 23, "Print",
 CellID->290229500],
Cell[34483, 893, 322, 7, 23, "Print",
 CellID->917949173],
Cell[34808, 902, 324, 7, 23, "Print",
 CellID->861684230],
Cell[35135, 911, 607, 13, 23, "Print",
 CellID->1480145550],
Cell[35745, 926, 320, 7, 23, "Print",
 CellID->159503147],
Cell[36068, 935, 322, 7, 23, "Print",
 CellID->344990273],
Cell[36393, 944, 605, 13, 23, "Print",
 CellID->89052781],
Cell[37001, 959, 614, 13, 23, "Print",
 CellID->742993044],
Cell[37618, 974, 323, 7, 23, "Print",
 CellID->1964139600],
Cell[37944, 983, 326, 7, 23, "Print",
 CellID->437821055],
Cell[38273, 992, 327, 7, 23, "Print",
 CellID->1141074587],
Cell[38603, 1001, 327, 7, 23, "Print",
 CellID->1905329242],
Cell[38933, 1010, 326, 7, 23, "Print",
 CellID->694875228],
Cell[39262, 1019, 326, 7, 23, "Print",
 CellID->343196592],
Cell[39591, 1028, 327, 7, 23, "Print",
 CellID->1299769609],
Cell[39921, 1037, 613, 13, 23, "Print",
 CellID->1747622676],
Cell[40537, 1052, 325, 7, 23, "Print",
 CellID->1931994882],
Cell[40865, 1061, 325, 7, 23, "Print",
 CellID->1285042923],
Cell[41193, 1070, 326, 7, 23, "Print",
 CellID->656249480],
Cell[41522, 1079, 327, 7, 23, "Print",
 CellID->1610366942],
Cell[41852, 1088, 618, 13, 23, "Print",
 CellID->673784444],
Cell[42473, 1103, 325, 7, 23, "Print",
 CellID->1385058587],
Cell[42801, 1112, 325, 7, 23, "Print",
 CellID->1517845881],
Cell[43129, 1121, 327, 7, 23, "Print",
 CellID->2077601540],
Cell[43459, 1130, 329, 7, 23, "Print",
 CellID->2134222863],
Cell[43791, 1139, 326, 7, 23, "Print",
 CellID->196398467],
Cell[44120, 1148, 329, 7, 23, "Print",
 CellID->2072983810],
Cell[44452, 1157, 120, 2, 23, "Print",
 CellID->2123204866],
Cell[44575, 1161, 109, 2, 23, "Print",
 CellID->284217937],
Cell[44687, 1165, 310, 7, 23, "Print",
 CellID->399737596],
Cell[45000, 1174, 314, 7, 23, "Print",
 CellID->805060146],
Cell[45317, 1183, 314, 7, 23, "Print",
 CellID->263663299],
Cell[45634, 1192, 315, 7, 23, "Print",
 CellID->1733700269],
Cell[45952, 1201, 314, 7, 23, "Print",
 CellID->504235321],
Cell[46269, 1210, 317, 7, 23, "Print",
 CellID->1411520420],
Cell[46589, 1219, 314, 7, 23, "Print",
 CellID->257856182],
Cell[46906, 1228, 314, 7, 23, "Print",
 CellID->1540842051],
Cell[47223, 1237, 316, 7, 23, "Print",
 CellID->487865859],
Cell[47542, 1246, 314, 7, 23, "Print",
 CellID->602281323],
Cell[47859, 1255, 314, 7, 23, "Print",
 CellID->684893606],
Cell[48176, 1264, 315, 7, 23, "Print",
 CellID->1229499825],
Cell[48494, 1273, 314, 7, 23, "Print",
 CellID->736007362],
Cell[48811, 1282, 317, 7, 23, "Print",
 CellID->2047264904],
Cell[49131, 1291, 316, 7, 23, "Print",
 CellID->579248345],
Cell[49450, 1300, 314, 7, 23, "Print",
 CellID->301728890],
Cell[49767, 1309, 315, 7, 23, "Print",
 CellID->2091818475],
Cell[50085, 1318, 317, 7, 23, "Print",
 CellID->1034002157],
Cell[50405, 1327, 316, 7, 23, "Print",
 CellID->405159898],
Cell[50724, 1336, 314, 7, 23, "Print",
 CellID->880592609],
Cell[51041, 1345, 316, 7, 23, "Print",
 CellID->1428824071],
Cell[51360, 1354, 317, 7, 23, "Print",
 CellID->1545734938],
Cell[51680, 1363, 316, 7, 23, "Print",
 CellID->704434208],
Cell[51999, 1372, 316, 7, 23, "Print",
 CellID->904655799],
Cell[52318, 1381, 317, 7, 23, "Print",
 CellID->2006111296],
Cell[52638, 1390, 319, 7, 23, "Print",
 CellID->1851230075],
Cell[52960, 1399, 316, 7, 23, "Print",
 CellID->848859071],
Cell[53279, 1408, 317, 7, 23, "Print",
 CellID->1691606509],
Cell[53599, 1417, 316, 7, 23, "Print",
 CellID->293097688],
Cell[53918, 1426, 317, 7, 23, "Print",
 CellID->1366882988],
Cell[54238, 1435, 316, 7, 23, "Print",
 CellID->154429773],
Cell[54557, 1444, 316, 7, 23, "Print",
 CellID->744698092],
Cell[54876, 1453, 317, 7, 23, "Print",
 CellID->2124648104],
Cell[55196, 1462, 317, 7, 23, "Print",
 CellID->1405866756],
Cell[55516, 1471, 317, 7, 23, "Print",
 CellID->1678907350],
Cell[55836, 1480, 319, 7, 23, "Print",
 CellID->1176964354],
Cell[56158, 1489, 315, 7, 23, "Print",
 CellID->32915441],
Cell[56476, 1498, 317, 7, 23, "Print",
 CellID->1952766498],
Cell[56796, 1507, 317, 7, 23, "Print",
 CellID->79551376],
Cell[57116, 1516, 316, 7, 23, "Print",
 CellID->2039798323],
Cell[57435, 1525, 316, 7, 23, "Print",
 CellID->920011543],
Cell[57754, 1534, 319, 7, 23, "Print",
 CellID->1341994267],
Cell[58076, 1543, 317, 7, 23, "Print",
 CellID->1365657716],
Cell[58396, 1552, 318, 7, 23, "Print",
 CellID->750840837],
Cell[58717, 1561, 315, 7, 23, "Print",
 CellID->819852299],
Cell[59035, 1570, 315, 7, 23, "Print",
 CellID->90096710],
Cell[59353, 1579, 316, 7, 23, "Print",
 CellID->512166966],
Cell[59672, 1588, 317, 7, 23, "Print",
 CellID->1395279649],
Cell[59992, 1597, 316, 7, 23, "Print",
 CellID->907983991],
Cell[60311, 1606, 317, 7, 23, "Print",
 CellID->1271409132],
Cell[60631, 1615, 317, 7, 23, "Print",
 CellID->1007395009],
Cell[60951, 1624, 317, 7, 23, "Print",
 CellID->1008558891],
Cell[61271, 1633, 316, 7, 23, "Print",
 CellID->668606491],
Cell[61590, 1642, 316, 7, 23, "Print",
 CellID->346902053],
Cell[61909, 1651, 315, 7, 23, "Print",
 CellID->47434879],
Cell[62227, 1660, 316, 7, 23, "Print",
 CellID->339282318],
Cell[62546, 1669, 317, 7, 23, "Print",
 CellID->1946506496],
Cell[62866, 1678, 317, 7, 23, "Print",
 CellID->1917192120],
Cell[63186, 1687, 319, 7, 23, "Print",
 CellID->1728132324],
Cell[63508, 1696, 315, 7, 23, "Print",
 CellID->739649454],
Cell[63826, 1705, 317, 7, 23, "Print",
 CellID->1897894093],
Cell[64146, 1714, 319, 7, 23, "Print",
 CellID->1157191333],
Cell[64468, 1723, 316, 7, 23, "Print",
 CellID->448292855],
Cell[64787, 1732, 316, 7, 23, "Print",
 CellID->438328241],
Cell[65106, 1741, 319, 7, 23, "Print",
 CellID->2142336410],
Cell[65428, 1750, 122, 2, 23, "Print",
 CellID->1773987139],
Cell[65553, 1754, 120, 2, 23, "Print",
 CellID->434266741],
Cell[65676, 1758, 126, 2, 23, "Print",
 CellID->2009684850],
Cell[65805, 1762, 122, 2, 23, "Print",
 CellID->804196891],
Cell[65930, 1766, 108, 2, 23, "Print",
 CellID->689469934],
Cell[66041, 1770, 306, 7, 23, "Print",
 CellID->628793925],
Cell[66350, 1779, 101, 2, 23, "Print",
 CellID->1777775172]
}, Open  ]],
Cell[66466, 1784, 153, 3, 32, "Output",
 CellID->1546107737]
}, Closed]],
Cell[CellGroupData[{
Cell[66656, 1792, 278, 6, 26, "Input",
 CellID->1859739660],
Cell[CellGroupData[{
Cell[66959, 1802, 888, 18, 23, "Print",
 CellID->281141794],
Cell[67850, 1822, 613, 11, 23, "Print",
 CellID->1164344103],
Cell[68466, 1835, 615, 11, 23, "Print",
 CellID->1263200977],
Cell[69084, 1848, 615, 11, 23, "Print",
 CellID->61583932],
Cell[69702, 1861, 617, 11, 23, "Print",
 CellID->1482990827],
Cell[70322, 1874, 410, 6, 23, "Print",
 CellID->1725599731],
Cell[70735, 1882, 399, 6, 23, "Print",
 CellID->326369969],
Cell[71137, 1890, 598, 11, 23, "Print",
 CellID->406809379],
Cell[71738, 1903, 600, 11, 23, "Print",
 CellID->537433043],
Cell[72341, 1916, 602, 11, 23, "Print",
 CellID->241512502],
Cell[72946, 1929, 601, 11, 23, "Print",
 CellID->993359368],
Cell[73550, 1942, 603, 11, 23, "Print",
 CellID->2116037914],
Cell[74156, 1955, 411, 6, 23, "Print",
 CellID->222181995],
Cell[74570, 1963, 411, 6, 23, "Print",
 CellID->1898721665],
Cell[74984, 1971, 412, 6, 23, "Print",
 CellID->194167885],
Cell[75399, 1979, 412, 6, 23, "Print",
 CellID->578987836],
Cell[75814, 1987, 399, 6, 23, "Print",
 CellID->446001340],
Cell[76216, 1995, 594, 11, 23, "Print",
 CellID->1806064551],
Cell[76813, 2008, 391, 6, 23, "Print",
 CellID->1973519727]
}, Open  ]],
Cell[77219, 2017, 488, 10, 32, "Output",
 CellID->92225566]
}, Closed]],
Cell[77722, 2030, 1347, 33, 93, "Input",
 CellID->1039312110],
Cell[79072, 2065, 178, 4, 32, "Input",
 CellID->2077551358],
Cell[79253, 2071, 541, 12, 55, "Input",
 CellID->1909803754],
Cell[79797, 2085, 586, 15, 55, "Input",
 CellID->1780799392],
Cell[CellGroupData[{
Cell[80408, 2104, 496, 12, 32, "Input",
 CellID->63220416],
Cell[80907, 2118, 286, 5, 32, "Output",
 CellID->141359559]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[81242, 2129, 128, 2, 30, "Subsection",
 CellID->1454182661],
Cell[81373, 2133, 983, 25, 99, "Input",
 CellID->2074904605],
Cell[82359, 2160, 416, 9, 55, "Input",
 CellID->1980565193],
Cell[82778, 2171, 169, 3, 32, "Input",
 CellID->444074165],
Cell[82950, 2176, 327, 6, 32, "Input",
 CellID->584704106]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[83326, 2188, 351, 5, 43, "Section",
 CellID->252853686],
Cell[CellGroupData[{
Cell[83702, 2197, 366, 5, 38, "Subsection",
 CellID->394311767],
Cell[CellGroupData[{
Cell[84093, 2206, 135, 2, 28, "Subsubsection",
 CellID->250375803],
Cell[84231, 2210, 787, 18, 70, "Text",
 CellID->65046519],
Cell[85021, 2230, 1055, 20, 70, "Text",
 CellID->451096809]
}, Closed]],
Cell[CellGroupData[{
Cell[86113, 2255, 122, 2, 22, "Subsubsection",
 CellID->415653668],
Cell[86238, 2259, 47, 1, 12, "OpenEnvironment",
 CellID->562348793],
Cell[CellGroupData[{
Cell[86310, 2264, 689, 13, 43, "EnvironmentHeader",
 CellID->708005171],
Cell[87002, 2279, 635, 16, 31, "GlobalDeclaration",
 CellTags->{
  "ID:346267501", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->346267501],
Cell[87640, 2297, 1611, 43, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:143274494", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet"},
 CellID->143274494],
Cell[89254, 2342, 2201, 72, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:669214708", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRec"},
 CellID->669214708],
Cell[91458, 2416, 65, 1, 26, "EndEnvironmentMarker",
 CellID->382099679]
}, Open  ]],
Cell[91538, 2420, 47, 1, 16, "CloseEnvironment",
 CellID->47167836]
}, Closed]],
Cell[CellGroupData[{
Cell[91622, 2426, 312, 5, 22, "Subsubsection",
 CellID->217957457],
Cell[91937, 2433, 662, 17, 31, "GlobalDeclaration",
 CellTags->{
  "ID:97279932", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->97279932],
Cell[92602, 2452, 141, 3, 12, "OpenEnvironment",
 CellID->35112716],
Cell[CellGroupData[{
Cell[92768, 2459, 506, 11, 43, "EnvironmentHeader",
 CellID->85473914],
Cell[93277, 2472, 812, 20, 32, "GlobalDeclaration",
 CellTags->{
  "ID:504518491", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->504518491],
Cell[94092, 2494, 2480, 74, 46, "FormalTextInputFormula",
 CellTags->{
  "ID:79522091", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet induction", "IR"},
 CellID->79522091],
Cell[96575, 2570, 884, 22, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:212626428", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet isSet", "IR"},
 CellID->212626428],
Cell[97462, 2594, 998, 26, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:433946629", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "base \[Element] natRecSet"},
 CellID->433946629],
Cell[98463, 2622, 1248, 34, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:198301498", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet inductive"},
 CellID->198301498],
Cell[99714, 2658, 1281, 37, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:972417912", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet shape"},
 CellID->972417912],
Cell[100998, 2697, 2066, 59, 48, "FormalTextInputFormula",
 CellTags->{
  "ID:57153239", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet cases"},
 CellID->57153239],
Cell[103067, 2758, 3468, 104, 78, "FormalTextInputFormula",
 CellTags->{
  "ID:240965330", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet induction'"},
 CellID->240965330],
Cell[106538, 2864, 1558, 44, 47, "FormalTextInputFormula",
 CellTags->{
  "ID:207844390", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet functional"},
 CellID->207844390],
Cell[108099, 2910, 1303, 37, 32, "FormalTextInputFormula",
 CellTags->{
  "ID:348247310", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet shape'"},
 CellID->348247310],
Cell[109405, 2949, 985, 29, 32, "FormalTextInputFormula",
 CellTags->{
  "ID:117940020", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRecSet total"},
 CellID->117940020],
Cell[110393, 2980, 203, 4, 26, "EndEnvironmentMarker",
 CellID->893683]
}, Open  ]],
Cell[110611, 2987, 140, 3, 16, "CloseEnvironment",
 CellID->18123018],
Cell[CellGroupData[{
Cell[110776, 2994, 6385, 160, 28, "ProofInfo",
 CellTags->{"Proof|ID:79522091", "Proof|ID:79522091-1"},
 CellID->1623641860],
Cell[117164, 3156, 480, 12, 56, "ProofInfoBody",
 CellID->1522358468]
}, Closed]],
Cell[CellGroupData[{
Cell[117681, 3173, 2402, 72, 26, "ProofInfo",
 CellTags->{"Proof|ID:212626428", "Proof|ID:212626428-1"},
 CellID->2101062054],
Cell[120086, 3247, 480, 12, 56, "ProofInfoBody",
 CellID->699276041]
}, Closed]],
Cell[CellGroupData[{
Cell[120603, 3264, 2684, 78, 26, "ProofInfo",
 CellTags->{"Proof|ID:433946629", "Proof|ID:433946629-1"},
 CellID->61218476],
Cell[123290, 3344, 481, 12, 56, "ProofInfoBody",
 CellID->1894724465]
}, Closed]],
Cell[CellGroupData[{
Cell[123808, 3361, 4108, 110, 26, "ProofInfo",
 CellTags->{"Proof|ID:198301498", "Proof|ID:198301498-1"},
 CellID->1431447619],
Cell[127919, 3473, 480, 12, 56, "ProofInfoBody",
 CellID->867263181]
}, Closed]],
Cell[CellGroupData[{
Cell[128436, 3490, 3607, 98, 26, "ProofInfo",
 CellTags->{"Proof|ID:972417912", "Proof|ID:972417912-1"},
 CellID->196372021],
Cell[132046, 3590, 480, 12, 56, "ProofInfoBody",
 CellID->128215566]
}, Closed]],
Cell[CellGroupData[{
Cell[132563, 3607, 5226, 132, 26, "ProofInfo",
 CellTags->{"Proof|ID:57153239", "Proof|ID:57153239-1"},
 CellID->736236675],
Cell[137792, 3741, 479, 12, 58, "ProofInfoBody",
 CellID->197934379]
}, Closed]],
Cell[CellGroupData[{
Cell[138308, 3758, 9017, 218, 26, "ProofInfo",
 CellTags->{"Proof|ID:240965330", "Proof|ID:240965330-1"},
 CellID->221211846],
Cell[147328, 3978, 550, 13, 56, "ProofInfoBody",
 CellID->295739411]
}, Closed]],
Cell[CellGroupData[{
Cell[147915, 3996, 4960, 128, 26, "ProofInfo",
 CellTags->{"Proof|ID:207844390", "Proof|ID:207844390-1"},
 CellID->201549000],
Cell[152878, 4126, 480, 12, 56, "ProofInfoBody",
 CellID->386042070]
}, Closed]],
Cell[CellGroupData[{
Cell[153395, 4143, 3860, 106, 26, "ProofInfo",
 CellTags->{"Proof|ID:348247310", "Proof|ID:348247310-1"},
 CellID->22594877],
Cell[157258, 4251, 480, 12, 70, "ProofInfoBody",
 CellID->171131707]
}, Closed]],
Cell[CellGroupData[{
Cell[157775, 4268, 3333, 94, 26, "ProofInfo",
 CellTags->{"Proof|ID:117940020", "Proof|ID:117940020-1"},
 CellID->128642575],
Cell[161111, 4364, 479, 12, 70, "ProofInfoBody",
 CellID->27786247]
}, Closed]],
Cell[161605, 4379, 46, 1, 12, "OpenEnvironment",
 CellID->50539674],
Cell[CellGroupData[{
Cell[161676, 4384, 456, 12, 43, "EnvironmentHeader",
 CellID->155324481],
Cell[162135, 4398, 906, 25, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:127781311", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRec base", "IR"},
 CellID->127781311],
Cell[163044, 4425, 1267, 38, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:186992137", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "natRec step", "IR"},
 CellID->186992137],
Cell[164314, 4465, 64, 1, 26, "EndEnvironmentMarker",
 CellID->55183838]
}, Open  ]],
Cell[164393, 4469, 48, 1, 16, "CloseEnvironment",
 CellID->389335050],
Cell[CellGroupData[{
Cell[164466, 4474, 2654, 80, 28, "ProofInfo",
 CellTags->{"Proof|ID:127781311", "Proof|ID:127781311-1"},
 CellID->5572876],
Cell[167123, 4556, 480, 12, 70, "ProofInfoBody",
 CellID->462709909]
}, Closed]],
Cell[CellGroupData[{
Cell[167640, 4573, 3830, 104, 26, "ProofInfo",
 CellTags->{"Proof|ID:186992137", "Proof|ID:186992137-1"},
 CellID->546211],
Cell[171473, 4679, 480, 12, 70, "ProofInfoBody",
 CellID->116871600]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[172014, 4698, 389, 6, 30, "Subsection",
 CellID->748923816],
Cell[CellGroupData[{
Cell[172428, 4708, 126, 2, 28, "Subsubsection",
 CellID->424914336],
Cell[172557, 4712, 46, 1, 70, "OpenEnvironment",
 CellID->79952930],
Cell[CellGroupData[{
Cell[172628, 4717, 369, 9, 70, "EnvironmentHeader",
 CellID->635773539],
Cell[173000, 4728, 646, 16, 70, "GlobalDeclaration",
 CellTags->{
  "ID:379686829", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->379686829],
Cell[173649, 4746, 637, 16, 70, "GlobalDeclaration",
 CellTags->{
  "ID:778005153", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->778005153],
Cell[174289, 4764, 1189, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:97787251", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat \[Element]"},
 CellID->97787251],
Cell[175481, 4806, 819, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:720795295", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat 0"},
 CellID->720795295],
Cell[176303, 4831, 908, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:3778917", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat 1"},
 CellID->3778917],
Cell[177214, 4858, 970, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:54060264", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat Pred"},
 CellID->54060264],
Cell[178187, 4888, 1066, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:205601935", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat +"},
 CellID->205601935],
Cell[179256, 4922, 1011, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:74238769", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat -"},
 CellID->74238769],
Cell[180270, 4955, 1273, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:288612943", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat \[RawStar]"},
 CellID->288612943],
Cell[181546, 4997, 1659, 47, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:155532301", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat \[GreaterEqual]"},
 CellID->155532301],
Cell[183208, 5046, 65, 1, 70, "EndEnvironmentMarker",
 CellID->217986896]
}, Open  ]],
Cell[183288, 5050, 47, 1, 70, "CloseEnvironment",
 CellID->22540073]
}, Closed]],
Cell[CellGroupData[{
Cell[183372, 5056, 434, 8, 22, "Subsubsection",
 CellID->512561655],
Cell[183809, 5066, 47, 1, 12, "OpenEnvironment",
 CellID->399148770],
Cell[CellGroupData[{
Cell[183881, 5071, 426, 10, 43, "EnvironmentHeader",
 CellID->43728960],
Cell[184310, 5083, 748, 20, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:15394254", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat isDomain"},
 CellID->15394254],
Cell[185061, 5105, 1924, 62, 50, "FormalTextInputFormula",
 CellTags->{
  "ID:478904098", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat induction", "IR"},
 CellID->478904098],
Cell[186988, 5169, 832, 23, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:60850406", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 \[Element] Nat", "IR"},
 CellID->60850406],
Cell[187823, 5194, 1033, 29, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:114723517", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ \[Element] Nat", "IR"},
 CellID->114723517],
Cell[188859, 5225, 831, 23, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:275276526", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 \[Element] Nat", "IR"},
 CellID->275276526],
Cell[189693, 5250, 691, 18, 41, "GlobalDeclaration",
 CellTags->{
  "ID:65979872", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->65979872],
Cell[190387, 5270, 770, 21, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:572443783", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat shape"},
 CellID->572443783],
Cell[191160, 5293, 1442, 43, 66, "FormalTextInputFormula",
 CellTags->{
  "ID:252333509", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat cases"},
 CellID->252333509],
Cell[192605, 5338, 846, 23, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:77756457", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ \[NotEqual] 0 Nat"},
 CellID->77756457],
Cell[193454, 5363, 1116, 34, 41, "FormalTextInputFormula",
 CellTags->{
  "ID:52299395", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ injective Nat"},
 CellID->52299395],
Cell[194573, 5399, 796, 22, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:263281689", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ no fixpoint Nat"},
 CellID->263281689],
Cell[195372, 5423, 817, 22, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:377798869", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 \[NotEqual] 0 Nat"},
 CellID->377798869],
Cell[196192, 5447, 943, 26, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:33739458", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[Element] closed Nat"},
 CellID->33739458],
Cell[197138, 5475, 1138, 27, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:654854303", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat transitive"},
 CellID->654854303],
Cell[198279, 5504, 65, 1, 26, "EndEnvironmentMarker",
 CellID->241267144]
}, Open  ]],
Cell[198359, 5508, 48, 1, 16, "CloseEnvironment",
 CellID->142415995],
Cell[CellGroupData[{
Cell[198432, 5513, 1703, 54, 28, "ProofInfo",
 CellTags->{"Proof|ID:15394254", "Proof|ID:15394254-1"},
 CellID->198699355],
Cell[200138, 5569, 479, 12, 70, "ProofInfoBody",
 CellID->314318358]
}, Closed]],
Cell[CellGroupData[{
Cell[200654, 5586, 4374, 118, 26, "ProofInfo",
 CellTags->{"Proof|ID:478904098", "Proof|ID:478904098-1"},
 CellID->65480230],
Cell[205031, 5706, 478, 12, 70, "ProofInfoBody",
 CellID->5553732]
}, Closed]],
Cell[CellGroupData[{
Cell[205546, 5723, 1843, 58, 26, "ProofInfo",
 CellTags->{"Proof|ID:60850406", "Proof|ID:60850406-1"},
 CellID->291459920],
Cell[207392, 5783, 478, 12, 70, "ProofInfoBody",
 CellID->94542070]
}, Closed]],
Cell[CellGroupData[{
Cell[207907, 5800, 2421, 72, 26, "ProofInfo",
 CellTags->{"Proof|ID:114723517", "Proof|ID:114723517-1"},
 CellID->310507632],
Cell[210331, 5874, 480, 12, 70, "ProofInfoBody",
 CellID->740339103]
}, Closed]],
Cell[CellGroupData[{
Cell[210848, 5891, 1847, 58, 26, "ProofInfo",
 CellTags->{"Proof|ID:275276526", "Proof|ID:275276526-1"},
 CellID->58288735],
Cell[212698, 5951, 480, 12, 70, "ProofInfoBody",
 CellID->411515486]
}, Closed]],
Cell[CellGroupData[{
Cell[213215, 5968, 2236, 68, 26, "ProofInfo",
 CellTags->{"Proof|ID:572443783", "Proof|ID:572443783-1"},
 CellID->43340428],
Cell[215454, 6038, 479, 12, 70, "ProofInfoBody",
 CellID->73111381]
}, Closed]],
Cell[CellGroupData[{
Cell[215970, 6055, 3251, 92, 26, "ProofInfo",
 CellTags->{"Proof|ID:252333509", "Proof|ID:252333509-1"},
 CellID->179637041],
Cell[219224, 6149, 480, 12, 70, "ProofInfoBody",
 CellID->128846745]
}, Closed]],
Cell[CellGroupData[{
Cell[219741, 6166, 2402, 72, 26, "ProofInfo",
 CellTags->{"Proof|ID:77756457", "Proof|ID:77756457-1"},
 CellID->41859577],
Cell[222146, 6240, 479, 12, 70, "ProofInfoBody",
 CellID->154826070]
}, Closed]],
Cell[CellGroupData[{
Cell[222662, 6257, 3163, 90, 26, "ProofInfo",
 CellTags->{"Proof|ID:52299395", "Proof|ID:52299395-1"},
 CellID->590942562],
Cell[225828, 6349, 479, 12, 70, "ProofInfoBody",
 CellID->325144463]
}, Closed]],
Cell[CellGroupData[{
Cell[226344, 6366, 2446, 74, 26, "ProofInfo",
 CellTags->{"Proof|ID:263281689", "Proof|ID:263281689-1"},
 CellID->255034219],
Cell[228793, 6442, 480, 12, 70, "ProofInfoBody",
 CellID->341857592]
}, Closed]],
Cell[CellGroupData[{
Cell[229310, 6459, 1846, 58, 26, "ProofInfo",
 CellTags->{"Proof|ID:377798869", "Proof|ID:377798869-1"},
 CellID->103721996],
Cell[231159, 6519, 480, 12, 70, "ProofInfoBody",
 CellID->516295583]
}, Closed]],
Cell[CellGroupData[{
Cell[231676, 6536, 2628, 76, 26, "ProofInfo",
 CellTags->{"Proof|ID:33739458", "Proof|ID:33739458-1"},
 CellID->548186621],
Cell[234307, 6614, 479, 12, 70, "ProofInfoBody",
 CellID->520182775]
}, Closed]],
Cell[CellGroupData[{
Cell[234823, 6631, 2646, 78, 26, "ProofInfo",
 CellTags->{"Proof|ID:654854303", "Proof|ID:654854303-1"},
 CellID->8946125],
Cell[237472, 6711, 480, 12, 70, "ProofInfoBody",
 CellID->402341539]
}, Closed]],
Cell[237967, 6726, 47, 1, 12, "OpenEnvironment",
 CellID->385788823],
Cell[CellGroupData[{
Cell[238039, 6731, 380, 9, 43, "EnvironmentHeader",
 CellID->224309402],
Cell[238422, 6742, 861, 24, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:240175329", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Pred 0 Nat", "IR"},
 CellID->240175329],
Cell[239286, 6768, 666, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:158284383", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->158284383],
Cell[239955, 6788, 955, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:145236688", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Pred succ Nat", "IR"},
 CellID->145236688],
Cell[240913, 6816, 870, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:225607097", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Pred closed Nat", "IR"},
 CellID->225607097],
Cell[241786, 6843, 65, 1, 70, "EndEnvironmentMarker",
 CellID->562641991]
}, Open  ]],
Cell[241866, 6847, 46, 1, 70, "CloseEnvironment",
 CellID->3500975],
Cell[CellGroupData[{
Cell[241937, 6852, 1972, 62, 70, "ProofInfo",
 CellTags->{"Proof|ID:240175329", "Proof|ID:240175329-1"},
 CellID->415645142],
Cell[243912, 6916, 478, 12, 70, "ProofInfoBody",
 CellID->7870725]
}, Closed]],
Cell[CellGroupData[{
Cell[244427, 6933, 2588, 78, 70, "ProofInfo",
 CellTags->{"Proof|ID:145236688", "Proof|ID:145236688-1"},
 CellID->498357158],
Cell[247018, 7013, 480, 12, 70, "ProofInfoBody",
 CellID->228630812]
}, Closed]],
Cell[CellGroupData[{
Cell[247535, 7030, 2487, 74, 70, "ProofInfo",
 CellTags->{"Proof|ID:225607097", "Proof|ID:225607097-1"},
 CellID->787676860],
Cell[250025, 7106, 479, 12, 70, "ProofInfoBody",
 CellID->22707532]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[250553, 7124, 459, 13, 22, "Subsubsection",
 CellID->254557108],
Cell[251015, 7139, 47, 1, 70, "OpenEnvironment",
 CellID->427738811],
Cell[CellGroupData[{
Cell[251087, 7144, 386, 9, 70, "EnvironmentHeader",
 CellID->928488810],
Cell[251476, 7155, 766, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:218406943", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->218406943],
Cell[252245, 7177, 923, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:257436422", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ 0 Nat", "IR"},
 CellID->257436422],
Cell[253171, 7204, 1063, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:249353684", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ succ Nat", "IR"},
 CellID->249353684],
Cell[254237, 7234, 850, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:279658160", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ closed Nat", "IR"},
 CellID->279658160],
Cell[255090, 7260, 899, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:647623443", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ 1 Nat"},
 CellID->647623443],
Cell[255992, 7287, 939, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:18327536", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 lneutral Nat", "IR"},
 CellID->18327536],
Cell[256934, 7315, 956, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1914645", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ + Nat", "IR"},
 CellID->1914645],
Cell[257893, 7344, 1190, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:343580194", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ = 0 Nat"},
 CellID->343580194],
Cell[259086, 7385, 889, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:834229092", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ comm Nat"},
 CellID->834229092],
Cell[259978, 7412, 668, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:303913216", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->303913216],
Cell[260649, 7432, 1094, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:73705609", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ assoc Nat"},
 CellID->73705609],
Cell[261746, 7466, 1008, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:202544617", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ comm-assoc Nat"},
 CellID->202544617],
Cell[262757, 7498, 1407, 46, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:555281086", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ + + Nat"},
 CellID->555281086],
Cell[264167, 7546, 1147, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:5229733", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ rcanc Nat"},
 CellID->5229733],
Cell[265317, 7585, 64, 1, 70, "EndEnvironmentMarker",
 CellID->23399794]
}, Open  ]],
Cell[265396, 7589, 48, 1, 70, "CloseEnvironment",
 CellID->559100389],
Cell[CellGroupData[{
Cell[265469, 7594, 2546, 78, 70, "ProofInfo",
 CellTags->{"Proof|ID:257436422", "Proof|ID:257436422-1"},
 CellID->197432916],
Cell[268018, 7674, 480, 12, 70, "ProofInfoBody",
 CellID->565004225]
}, Closed]],
Cell[CellGroupData[{
Cell[268535, 7691, 3263, 92, 70, "ProofInfo",
 CellTags->{"Proof|ID:249353684", "Proof|ID:249353684-1"},
 CellID->376294367],
Cell[271801, 7785, 480, 12, 70, "ProofInfoBody",
 CellID->108601403]
}, Closed]],
Cell[CellGroupData[{
Cell[272318, 7802, 2841, 82, 70, "ProofInfo",
 CellTags->{"Proof|ID:279658160", "Proof|ID:279658160-1"},
 CellID->363010569],
Cell[275162, 7886, 480, 12, 70, "ProofInfoBody",
 CellID->585218458]
}, Closed]],
Cell[CellGroupData[{
Cell[275679, 7903, 2620, 78, 70, "ProofInfo",
 CellTags->{"Proof|ID:647623443", "Proof|ID:647623443-1"},
 CellID->35382149],
Cell[278302, 7983, 480, 12, 70, "ProofInfoBody",
 CellID->257589816]
}, Closed]],
Cell[CellGroupData[{
Cell[278819, 8000, 2555, 78, 70, "ProofInfo",
 CellTags->{"Proof|ID:18327536", "Proof|ID:18327536-1"},
 CellID->794865037],
Cell[281377, 8080, 478, 12, 70, "ProofInfoBody",
 CellID->71581738]
}, Closed]],
Cell[CellGroupData[{
Cell[281892, 8097, 3253, 92, 70, "ProofInfo",
 CellTags->{"Proof|ID:1914645", "Proof|ID:1914645-1"},
 CellID->218839119],
Cell[285148, 8191, 478, 12, 70, "ProofInfoBody",
 CellID->337393823]
}, Closed]],
Cell[CellGroupData[{
Cell[285663, 8208, 3508, 98, 70, "ProofInfo",
 CellTags->{"Proof|ID:343580194", "Proof|ID:343580194-1"},
 CellID->734595749],
Cell[289174, 8308, 480, 12, 70, "ProofInfoBody",
 CellID->407594587]
}, Closed]],
Cell[CellGroupData[{
Cell[289691, 8325, 3091, 90, 70, "ProofInfo",
 CellTags->{"Proof|ID:834229092", "Proof|ID:834229092-1"},
 CellID->12781957],
Cell[292785, 8417, 480, 12, 70, "ProofInfoBody",
 CellID->354412127]
}, Closed]],
Cell[CellGroupData[{
Cell[293302, 8434, 3865, 108, 70, "ProofInfo",
 CellTags->{"Proof|ID:73705609", "Proof|ID:73705609-1"},
 CellID->337185825],
Cell[297170, 8544, 478, 12, 70, "ProofInfoBody",
 CellID->62558116]
}, Closed]],
Cell[CellGroupData[{
Cell[297685, 8561, 3877, 108, 70, "ProofInfo",
 CellTags->{"Proof|ID:202544617", "Proof|ID:202544617-1"},
 CellID->299537],
Cell[301565, 8671, 480, 12, 70, "ProofInfoBody",
 CellID->173855211]
}, Closed]],
Cell[CellGroupData[{
Cell[302082, 8688, 4641, 126, 70, "ProofInfo",
 CellTags->{"Proof|ID:555281086", "Proof|ID:555281086-1"},
 CellID->94141421],
Cell[306726, 8816, 479, 12, 70, "ProofInfoBody",
 CellID->37231442]
}, Closed]],
Cell[CellGroupData[{
Cell[307242, 8833, 3744, 104, 70, "ProofInfo",
 CellTags->{"Proof|ID:5229733", "Proof|ID:5229733-1"},
 CellID->540060300],
Cell[310989, 8939, 478, 12, 70, "ProofInfoBody",
 CellID->235773962]
}, Closed]],
Cell[311482, 8954, 47, 1, 70, "OpenEnvironment",
 CellID->325165612],
Cell[CellGroupData[{
Cell[311554, 8959, 455, 14, 70, "EnvironmentHeader",
 CellID->901457357],
Cell[312012, 8975, 717, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:301992507", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->301992507],
Cell[312732, 8996, 895, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:99163941", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] 0 Nat", "IR"},
 CellID->99163941],
Cell[313630, 9023, 987, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:371484270", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] succ Nat", "IR"},
 CellID->371484270],
Cell[314620, 9053, 852, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:142418160", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] closed Nat", "IR"},
 CellID->142418160],
Cell[315475, 9079, 832, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:21383846", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 rneutral Nat", "IR"},
 CellID->21383846],
Cell[316310, 9104, 864, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:28953873", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 lannihilator Nat", "IR"},
 CellID->28953873],
Cell[317177, 9130, 981, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:218775165", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ \[RawStar] Nat", "IR"},
 CellID->218775165],
Cell[318161, 9160, 1191, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:470439981", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] zdf Nat"},
 CellID->470439981],
Cell[319355, 9201, 842, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:218136211", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] comm Nat"},
 CellID->218136211],
Cell[320200, 9227, 690, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:666029340", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->666029340],
Cell[320893, 9247, 1080, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:43150914", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] distr + Nat"},
 CellID->43150914],
Cell[321976, 9281, 1000, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:475421150", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] assoc Nat"},
 CellID->475421150],
Cell[322979, 9313, 1007, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:98508460", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] comm-assoc Nat"},
 CellID->98508460],
Cell[323989, 9345, 1194, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:138249572", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] rcanc Nat"},
 CellID->138249572],
Cell[325186, 9386, 63, 1, 70, "EndEnvironmentMarker",
 CellID->7112206]
}, Open  ]],
Cell[325264, 9390, 48, 1, 70, "CloseEnvironment",
 CellID->239279945],
Cell[CellGroupData[{
Cell[325337, 9395, 2499, 76, 70, "ProofInfo",
 CellTags->{"Proof|ID:99163941", "Proof|ID:99163941-1"},
 CellID->138225265],
Cell[327839, 9473, 479, 12, 70, "ProofInfoBody",
 CellID->392553738]
}, Closed]],
Cell[CellGroupData[{
Cell[328355, 9490, 3424, 98, 70, "ProofInfo",
 CellTags->{"Proof|ID:371484270", "Proof|ID:371484270-1"},
 CellID->2648213],
Cell[331782, 9590, 480, 12, 70, "ProofInfoBody",
 CellID->637195123]
}, Closed]],
Cell[CellGroupData[{
Cell[332299, 9607, 2840, 82, 70, "ProofInfo",
 CellTags->{"Proof|ID:142418160", "Proof|ID:142418160-1"},
 CellID->32519450],
Cell[335142, 9691, 480, 12, 70, "ProofInfoBody",
 CellID->748270549]
}, Closed]],
Cell[CellGroupData[{
Cell[335659, 9708, 2554, 78, 70, "ProofInfo",
 CellTags->{"Proof|ID:21383846", "Proof|ID:21383846-1"},
 CellID->16031556],
Cell[338216, 9788, 479, 12, 70, "ProofInfoBody",
 CellID->456731267]
}, Closed]],
Cell[CellGroupData[{
Cell[338732, 9805, 2520, 76, 70, "ProofInfo",
 CellTags->{"Proof|ID:28953873", "Proof|ID:28953873-1"},
 CellID->86840982],
Cell[341255, 9883, 478, 12, 70, "ProofInfoBody",
 CellID->14700220]
}, Closed]],
Cell[CellGroupData[{
Cell[341770, 9900, 3425, 98, 70, "ProofInfo",
 CellTags->{"Proof|ID:218775165", "Proof|ID:218775165-1"},
 CellID->82184287],
Cell[345198, 10000, 480, 12, 70, "ProofInfoBody",
 CellID->141785307]
}, Closed]],
Cell[CellGroupData[{
Cell[345715, 10017, 3506, 98, 70, "ProofInfo",
 CellTags->{"Proof|ID:470439981", "Proof|ID:470439981-1"},
 CellID->167271436],
Cell[349224, 10117, 479, 12, 70, "ProofInfoBody",
 CellID->20808447]
}, Closed]],
Cell[CellGroupData[{
Cell[349740, 10134, 3092, 90, 70, "ProofInfo",
 CellTags->{"Proof|ID:218136211", "Proof|ID:218136211-1"},
 CellID->116241402],
Cell[352835, 10226, 480, 12, 70, "ProofInfoBody",
 CellID->464340012]
}, Closed]],
Cell[CellGroupData[{
Cell[353352, 10243, 4114, 114, 70, "ProofInfo",
 CellTags->{"Proof|ID:43150914", "Proof|ID:43150914-1"},
 CellID->92944893],
Cell[357469, 10359, 479, 12, 70, "ProofInfoBody",
 CellID->422365210]
}, Closed]],
Cell[CellGroupData[{
Cell[357985, 10376, 3870, 108, 70, "ProofInfo",
 CellTags->{"Proof|ID:475421150", "Proof|ID:475421150-1"},
 CellID->574072770],
Cell[361858, 10486, 480, 12, 70, "ProofInfoBody",
 CellID->691990920]
}, Closed]],
Cell[CellGroupData[{
Cell[362375, 10503, 3874, 108, 70, "ProofInfo",
 CellTags->{"Proof|ID:98508460", "Proof|ID:98508460-1"},
 CellID->81244651],
Cell[366252, 10613, 479, 12, 70, "ProofInfoBody",
 CellID->176140155]
}, Closed]],
Cell[CellGroupData[{
Cell[366768, 10630, 4076, 110, 70, "ProofInfo",
 CellTags->{"Proof|ID:138249572", "Proof|ID:138249572-1"},
 CellID->165947681],
Cell[370847, 10742, 479, 12, 70, "ProofInfoBody",
 CellID->55566549]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[371375, 10760, 297, 4, 22, "Subsubsection",
 CellID->359761349],
Cell[371675, 10766, 46, 1, 70, "OpenEnvironment",
 CellID->54526147],
Cell[CellGroupData[{
Cell[371746, 10771, 396, 9, 70, "EnvironmentHeader",
 CellID->1339458],
Cell[372145, 10782, 715, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:621974620", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->621974620],
Cell[372863, 10803, 1126, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:380981542", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 \[GreaterEqual] Nat", "IR"},
 CellID->380981542],
Cell[373992, 10840, 1292, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:42977349", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ \[GreaterEqual] Nat", "IR"},
 CellID->42977349],
Cell[375287, 10881, 864, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:428740350", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] reflexive Nat"},
 CellID->428740350],
Cell[376154, 10905, 882, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:297686800", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] 0 Nat", "IR"},
 CellID->297686800],
Cell[377039, 10930, 1316, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:585961470", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] succ Nat", "IR"},
 CellID->585961470],
Cell[378358, 10971, 1080, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:894887193", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] antisymmetric Nat"},
 CellID->894887193],
Cell[379441, 11004, 875, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:208533627", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ increasing Nat"},
 CellID->208533627],
Cell[380319, 11029, 1239, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:112605784", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ \[GreaterEqual] succ Nat", "IR"},
 CellID->112605784],
Cell[381561, 11068, 901, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:304315076", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] total Nat"},
 CellID->304315076],
Cell[382465, 11095, 1167, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:446896782", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] \[Exists] Nat"},
 CellID->446896782],
Cell[383635, 11132, 853, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:196203313", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 \[GreaterEqual] 0 Nat"},
 CellID->196203313],
Cell[384491, 11157, 1179, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:30448165", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] 1 Nat", "IR"},
 CellID->30448165],
Cell[385673, 11195, 690, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:222994735", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->222994735],
Cell[386366, 11215, 1107, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:546802157", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] transitive Nat"},
 CellID->546802157],
Cell[387476, 11250, 1045, 29, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:352201475", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] + \[DoubleLeftArrow] Nat"},
 CellID->352201475],
Cell[388524, 11281, 1208, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:97128652", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] + Nat"},
 CellID->97128652],
Cell[389735, 11321, 849, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:278931212", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ \[GreaterEqual] Nat"},
 CellID->278931212],
Cell[390587, 11346, 1000, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:335211897", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] \[RawStar] Nat"},
 CellID->335211897],
Cell[391590, 11376, 943, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:32863246", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] \[GreaterEqual] Nat"},
 CellID->32863246],
Cell[392536, 11405, 65, 1, 70, "EndEnvironmentMarker",
 CellID->251470773]
}, Open  ]],
Cell[392616, 11409, 48, 1, 70, "CloseEnvironment",
 CellID->123859362],
Cell[CellGroupData[{
Cell[392689, 11414, 2764, 80, 70, "ProofInfo",
 CellTags->{"Proof|ID:380981542", "Proof|ID:380981542-1"},
 CellID->159585745],
Cell[395456, 11496, 479, 12, 70, "ProofInfoBody",
 CellID->36940991]
}, Closed]],
Cell[CellGroupData[{
Cell[395972, 11513, 3830, 102, 70, "ProofInfo",
 CellTags->{"Proof|ID:42977349", "Proof|ID:42977349-1"},
 CellID->255247870],
Cell[399805, 11617, 478, 12, 70, "ProofInfoBody",
 CellID->60792811]
}, Closed]],
Cell[CellGroupData[{
Cell[400320, 11634, 2457, 74, 70, "ProofInfo",
 CellTags->{"Proof|ID:428740350", "Proof|ID:428740350-1"},
 CellID->34057048],
Cell[402780, 11710, 480, 12, 70, "ProofInfoBody",
 CellID->812129610]
}, Closed]],
Cell[CellGroupData[{
Cell[403297, 11727, 2399, 72, 70, "ProofInfo",
 CellTags->{"Proof|ID:297686800", "Proof|ID:297686800-1"},
 CellID->45781616],
Cell[405699, 11801, 480, 12, 70, "ProofInfoBody",
 CellID->541714067]
}, Closed]],
Cell[CellGroupData[{
Cell[406216, 11818, 3771, 102, 70, "ProofInfo",
 CellTags->{"Proof|ID:585961470", "Proof|ID:585961470-1"},
 CellID->252534163],
Cell[409990, 11922, 479, 12, 70, "ProofInfoBody",
 CellID->60797310]
}, Closed]],
Cell[CellGroupData[{
Cell[410506, 11939, 3574, 99, 70, "ProofInfo",
 CellTags->{"Proof|ID:894887193", "Proof|ID:894887193-1"},
 CellID->246933219],
Cell[414083, 12040, 480, 12, 70, "ProofInfoBody",
 CellID->141573684]
}, Closed]],
Cell[CellGroupData[{
Cell[414600, 12057, 2521, 76, 70, "ProofInfo",
 CellTags->{"Proof|ID:208533627", "Proof|ID:208533627-1"},
 CellID->10870655],
Cell[417124, 12135, 479, 12, 70, "ProofInfoBody",
 CellID->71717834]
}, Closed]],
Cell[CellGroupData[{
Cell[417640, 12152, 3416, 96, 70, "ProofInfo",
 CellTags->{"Proof|ID:112605784", "Proof|ID:112605784-1"},
 CellID->362360428],
Cell[421059, 12250, 480, 12, 70, "ProofInfoBody",
 CellID->292064837]
}, Closed]],
Cell[CellGroupData[{
Cell[421576, 12267, 3172, 90, 70, "ProofInfo",
 CellTags->{"Proof|ID:304315076", "Proof|ID:304315076-1"},
 CellID->318982970],
Cell[424751, 12359, 480, 12, 70, "ProofInfoBody",
 CellID->873958818]
}, Closed]],
Cell[CellGroupData[{
Cell[425268, 12376, 3861, 106, 70, "ProofInfo",
 CellTags->{"Proof|ID:446896782", "Proof|ID:446896782-1"},
 CellID->239855888],
Cell[429132, 12484, 480, 12, 70, "ProofInfoBody",
 CellID->356600252]
}, Closed]],
Cell[CellGroupData[{
Cell[429649, 12501, 1927, 60, 70, "ProofInfo",
 CellTags->{"Proof|ID:196203313", "Proof|ID:196203313-1"},
 CellID->28380975],
Cell[431579, 12563, 479, 12, 70, "ProofInfoBody",
 CellID->13411257]
}, Closed]],
Cell[CellGroupData[{
Cell[432095, 12580, 2764, 80, 70, "ProofInfo",
 CellTags->{"Proof|ID:30448165", "Proof|ID:30448165-1"},
 CellID->64482358],
Cell[434862, 12662, 479, 12, 70, "ProofInfoBody",
 CellID->202694291]
}, Closed]],
Cell[CellGroupData[{
Cell[435378, 12679, 3940, 107, 70, "ProofInfo",
 CellTags->{"Proof|ID:546802157", "Proof|ID:546802157-1"},
 CellID->527891345],
Cell[439321, 12788, 480, 12, 70, "ProofInfoBody",
 CellID->823045757]
}, Closed]],
Cell[CellGroupData[{
Cell[439838, 12805, 4003, 110, 70, "ProofInfo",
 CellTags->{"Proof|ID:352201475", "Proof|ID:352201475-1"},
 CellID->76510990],
Cell[443844, 12917, 480, 12, 70, "ProofInfoBody",
 CellID->109258721]
}, Closed]],
Cell[CellGroupData[{
Cell[444361, 12934, 3995, 110, 70, "ProofInfo",
 CellTags->{"Proof|ID:97128652", "Proof|ID:97128652-1"},
 CellID->393536374],
Cell[448359, 13046, 479, 12, 70, "ProofInfoBody",
 CellID->644252183]
}, Closed]],
Cell[CellGroupData[{
Cell[448875, 13063, 2960, 86, 70, "ProofInfo",
 CellTags->{"Proof|ID:278931212", "Proof|ID:278931212-1"},
 CellID->260422124],
Cell[451838, 13151, 480, 12, 70, "ProofInfoBody",
 CellID->478276088]
}, Closed]],
Cell[CellGroupData[{
Cell[452355, 13168, 3950, 108, 70, "ProofInfo",
 CellTags->{"Proof|ID:335211897", "Proof|ID:335211897-1"},
 CellID->171667115],
Cell[456308, 13278, 480, 12, 70, "ProofInfoBody",
 CellID->548791013]
}, Closed]],
Cell[CellGroupData[{
Cell[456825, 13295, 3315, 94, 70, "ProofInfo",
 CellTags->{"Proof|ID:32863246", "Proof|ID:32863246-1"},
 CellID->145909442],
Cell[460143, 13391, 479, 12, 70, "ProofInfoBody",
 CellID->204709576]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[460671, 13409, 354, 9, 22, "Subsubsection",
 CellID->505759711],
Cell[461028, 13420, 46, 1, 70, "OpenEnvironment",
 CellID->56489176],
Cell[CellGroupData[{
Cell[461099, 13425, 384, 9, 70, "EnvironmentHeader",
 CellID->227256827],
Cell[461486, 13436, 717, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:609763659", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->609763659],
Cell[462206, 13457, 816, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:19208428", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- 0 Nat", "IR"},
 CellID->19208428],
Cell[463025, 13482, 987, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:593421999", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- succ Nat", "IR"},
 CellID->593421999],
Cell[464015, 13512, 848, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:153357458", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- closed Nat", "IR"},
 CellID->153357458],
Cell[464866, 13538, 871, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:616125500", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- 1 Nat"},
 CellID->616125500],
Cell[465740, 13565, 845, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:491337343", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 - Nat", "IR"},
 CellID->491337343],
Cell[466588, 13591, 1028, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:621700658", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "succ - succ Nat", "IR"},
 CellID->621700658],
Cell[467619, 13621, 869, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:236021775", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ - Nat"},
 CellID->236021775],
Cell[468491, 13649, 965, 29, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:3006634", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- + Nat"},
 CellID->3006634],
Cell[469459, 13680, 663, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:80465146", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->80465146],
Cell[470125, 13700, 996, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:404996142", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ - + Nat"},
 CellID->404996142],
Cell[471124, 13732, 1106, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:225131303", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] distr - Nat"},
 CellID->225131303],
Cell[472233, 13766, 1206, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:142079924", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "(-)+ Nat"},
 CellID->142079924],
Cell[473442, 13806, 2138, 69, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:367201929", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- = - Nat"},
 CellID->367201929],
Cell[475583, 13877, 65, 1, 70, "EndEnvironmentMarker",
 CellID->833478354]
}, Open  ]],
Cell[475663, 13881, 48, 1, 70, "CloseEnvironment",
 CellID->435769634],
Cell[CellGroupData[{
Cell[475736, 13886, 2540, 78, 70, "ProofInfo",
 CellTags->{"Proof|ID:19208428", "Proof|ID:19208428-1"},
 CellID->94129044],
Cell[478279, 13966, 479, 12, 70, "ProofInfoBody",
 CellID->264683537]
}, Closed]],
Cell[CellGroupData[{
Cell[478795, 13983, 3337, 94, 70, "ProofInfo",
 CellTags->{"Proof|ID:593421999", "Proof|ID:593421999-1"},
 CellID->286638854],
Cell[482135, 14079, 479, 12, 70, "ProofInfoBody",
 CellID->97602613]
}, Closed]],
Cell[CellGroupData[{
Cell[482651, 14096, 2841, 82, 70, "ProofInfo",
 CellTags->{"Proof|ID:153357458", "Proof|ID:153357458-1"},
 CellID->454920045],
Cell[485495, 14180, 480, 12, 70, "ProofInfoBody",
 CellID->935874446]
}, Closed]],
Cell[CellGroupData[{
Cell[486012, 14197, 2694, 80, 70, "ProofInfo",
 CellTags->{"Proof|ID:616125500", "Proof|ID:616125500-1"},
 CellID->14344385],
Cell[488709, 14279, 479, 12, 70, "ProofInfoBody",
 CellID->49030196]
}, Closed]],
Cell[CellGroupData[{
Cell[489225, 14296, 2504, 76, 70, "ProofInfo",
 CellTags->{"Proof|ID:491337343", "Proof|ID:491337343-1"},
 CellID->458038256],
Cell[491732, 14374, 479, 12, 70, "ProofInfoBody",
 CellID->21026969]
}, Closed]],
Cell[CellGroupData[{
Cell[492248, 14391, 3277, 94, 70, "ProofInfo",
 CellTags->{"Proof|ID:621700658", "Proof|ID:621700658-1"},
 CellID->31914115],
Cell[495528, 14487, 479, 12, 70, "ProofInfoBody",
 CellID->35212336]
}, Closed]],
Cell[CellGroupData[{
Cell[496044, 14504, 3094, 90, 70, "ProofInfo",
 CellTags->{"Proof|ID:236021775", "Proof|ID:236021775-1"},
 CellID->322826090],
Cell[499141, 14596, 480, 12, 70, "ProofInfoBody",
 CellID->105703489]
}, Closed]],
Cell[CellGroupData[{
Cell[499658, 14613, 3552, 100, 70, "ProofInfo",
 CellTags->{"Proof|ID:3006634", "Proof|ID:3006634-1"},
 CellID->539834930],
Cell[503213, 14715, 477, 12, 70, "ProofInfoBody",
 CellID->55369331]
}, Closed]],
Cell[CellGroupData[{
Cell[503727, 14732, 3866, 108, 70, "ProofInfo",
 CellTags->{"Proof|ID:404996142", "Proof|ID:404996142-1"},
 CellID->457219728],
Cell[507596, 14842, 480, 12, 70, "ProofInfoBody",
 CellID->822478605]
}, Closed]],
Cell[CellGroupData[{
Cell[508113, 14859, 4120, 114, 70, "ProofInfo",
 CellTags->{"Proof|ID:225131303", "Proof|ID:225131303-1"},
 CellID->442897560],
Cell[512236, 14975, 480, 12, 70, "ProofInfoBody",
 CellID->567742436]
}, Closed]],
Cell[CellGroupData[{
Cell[512753, 14992, 4352, 118, 70, "ProofInfo",
 CellTags->{"Proof|ID:142079924", "Proof|ID:142079924-1"},
 CellID->230509363],
Cell[517108, 15112, 479, 12, 70, "ProofInfoBody",
 CellID->21742626]
}, Closed]],
Cell[CellGroupData[{
Cell[517624, 15129, 5539, 144, 70, "ProofInfo",
 CellTags->{"Proof|ID:367201929", "Proof|ID:367201929-1"},
 CellID->109183652],
Cell[523166, 15275, 480, 12, 70, "ProofInfoBody",
 CellID->145493550]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[523719, 15295, 293, 4, 43, "Section",
 CellID->193970655],
Cell[CellGroupData[{
Cell[524037, 15303, 192, 3, 38, "Subsection",
 CellID->884540480],
Cell[CellGroupData[{
Cell[524254, 15310, 124, 2, 28, "Subsubsection",
 CellID->684862369],
Cell[524381, 15314, 47, 1, 12, "OpenEnvironment",
 CellID->475123522],
Cell[CellGroupData[{
Cell[524453, 15319, 382, 9, 43, "EnvironmentHeader",
 CellID->11831326],
Cell[524838, 15330, 649, 16, 30, "GlobalDeclaration",
 CellTags->{
  "ID:513393113", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->513393113],
Cell[525490, 15348, 637, 16, 31, "GlobalDeclaration",
 CellTags->{
  "ID:623651727", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->623651727],
Cell[526130, 15366, 1235, 41, 37, "FormalTextInputFormula",
 CellTags->{
  "ID:100384801", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "PreInt \[Element]"},
 CellID->100384801],
Cell[527368, 15409, 865, 25, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:398024454", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "PreInt 0"},
 CellID->398024454],
Cell[528236, 15436, 869, 25, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:418726841", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "PreInt 1"},
 CellID->418726841],
Cell[529108, 15463, 1072, 32, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:10580372", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "PreInt +"},
 CellID->10580372],
Cell[530183, 15497, 906, 26, 28, "FormalTextInputFormula",
 CellTags->{
  "ID:106096058", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "PreInt - unary"},
 CellID->106096058],
Cell[531092, 15525, 1436, 44, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:12727554", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "PreInt \[RawStar]"},
 CellID->12727554],
Cell[532531, 15571, 1463, 48, 37, "FormalTextInputFormula",
 CellTags->{
  "ID:329154583", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "PreInt \[GreaterEqual]"},
 CellID->329154583],
Cell[533997, 15621, 1332, 43, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:468243965", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "PreInt equiv"},
 CellID->468243965],
Cell[535332, 15666, 65, 1, 26, "EndEnvironmentMarker",
 CellID->386145701]
}, Open  ]],
Cell[535412, 15670, 48, 1, 16, "CloseEnvironment",
 CellID->138864013]
}, Closed]],
Cell[CellGroupData[{
Cell[535497, 15676, 223, 5, 22, "Subsubsection",
 CellID->100030134],
Cell[535723, 15683, 46, 1, 12, "OpenEnvironment",
 CellID->31288446],
Cell[CellGroupData[{
Cell[535794, 15688, 390, 9, 43, "EnvironmentHeader",
 CellID->207945970],
Cell[536187, 15699, 770, 20, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:168445606", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "PreInt isDomain", "IR"},
 CellID->168445606],
Cell[536960, 15721, 1640, 54, 50, "FormalTextInputFormula",
 CellTags->{
  "ID:977301511", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[Element] PreInt alternative", "IR"},
 CellID->977301511],
Cell[538603, 15777, 1529, 50, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:353341722", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Pair \[Element] PreInt", "IR"},
 CellID->353341722],
Cell[540135, 15829, 840, 23, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:34141802", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 \[Element] PreInt", "IR"},
 CellID->34141802],
Cell[540978, 15854, 847, 23, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:553448562", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 \[Element] PreInt", "IR"},
 CellID->553448562],
Cell[541828, 15879, 64, 1, 26, "EndEnvironmentMarker",
 CellID->18036666]
}, Open  ]],
Cell[541907, 15883, 47, 1, 16, "CloseEnvironment",
 CellID->54156207],
Cell[CellGroupData[{
Cell[541979, 15888, 1720, 54, 28, "ProofInfo",
 CellTags->{"Proof|ID:168445606", "Proof|ID:168445606-1"},
 CellID->142038033],
Cell[543702, 15944, 480, 12, 70, "ProofInfoBody",
 CellID->283190144]
}, Closed]],
Cell[CellGroupData[{
Cell[544219, 15961, 3752, 102, 26, "ProofInfo",
 CellTags->{"Proof|ID:977301511", "Proof|ID:977301511-1"},
 CellID->32364991],
Cell[547974, 16065, 480, 12, 70, "ProofInfoBody",
 CellID->351963332]
}, Closed]],
Cell[CellGroupData[{
Cell[548491, 16082, 3475, 96, 26, "ProofInfo",
 CellTags->{"Proof|ID:353341722", "Proof|ID:353341722-1"},
 CellID->352067017],
Cell[551969, 16180, 480, 12, 70, "ProofInfoBody",
 CellID->147706619]
}, Closed]],
Cell[CellGroupData[{
Cell[552486, 16197, 1860, 58, 26, "ProofInfo",
 CellTags->{"Proof|ID:34141802", "Proof|ID:34141802-1"},
 CellID->53936406],
Cell[554349, 16257, 479, 12, 70, "ProofInfoBody",
 CellID->312339613]
}, Closed]],
Cell[CellGroupData[{
Cell[554865, 16274, 1866, 58, 26, "ProofInfo",
 CellTags->{"Proof|ID:553448562", "Proof|ID:553448562-1"},
 CellID->265994209],
Cell[556734, 16334, 480, 12, 70, "ProofInfoBody",
 CellID->510886051]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[557263, 16352, 192, 3, 22, "Subsubsection",
 CellID->428038701],
Cell[557458, 16357, 47, 1, 12, "OpenEnvironment",
 CellID->234603150],
Cell[CellGroupData[{
Cell[557530, 16362, 393, 9, 43, "EnvironmentHeader",
 CellID->734012110],
Cell[557926, 16373, 1504, 47, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:220417861", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "equiv Pair PreInt", "IR"},
 CellID->220417861],
Cell[559433, 16422, 1398, 44, 48, "FormalTextInputFormula",
 CellTags->{
  "ID:52539739", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "equiv 0 PreInt", "IR"},
 CellID->52539739],
Cell[560834, 16468, 1416, 43, 48, "FormalTextInputFormula",
 CellTags->{
  "ID:94484613", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "equiv Nat \[GreaterEqual] PreInt"},
 CellID->94484613],
Cell[562253, 16513, 1413, 43, 48, "FormalTextInputFormula",
 CellTags->{
  "ID:321256465", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "equiv Nat \[LessEqual] PreInt"},
 CellID->321256465],
Cell[563669, 16558, 911, 25, 35, "FormalTextInputFormula",
 CellTags->{
  "ID:129809121", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[Not] 1 equiv 0 PreInt"},
 CellID->129809121],
Cell[564583, 16585, 720, 19, 41, "GlobalDeclaration",
 CellTags->{
  "ID:413771523", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->413771523],
Cell[565306, 16606, 1703, 51, 69, "FormalTextInputFormula",
 CellTags->{
  "ID:637862621", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "equiv Nat PreInt"},
 CellID->637862621],
Cell[567012, 16659, 817, 22, 32, "FormalTextInputFormula",
 CellTags->{
  "ID:12369324", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "equiv reflexive PreInt"},
 CellID->12369324],
Cell[567832, 16683, 930, 26, 32, "FormalTextInputFormula",
 CellTags->{
  "ID:19276", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "equiv symmetric PreInt"},
 CellID->19276],
Cell[568765, 16711, 1420, 45, 48, "FormalTextInputFormula",
 CellTags->{
  "ID:164883544", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "equiv transitive PreInt"},
 CellID->164883544],
Cell[570188, 16758, 849, 23, 35, "FormalTextInputFormula",
 CellTags->{
  "ID:357440757", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "equiv isEquivalence PreInt"},
 CellID->357440757],
Cell[571040, 16783, 64, 1, 26, "EndEnvironmentMarker",
 CellID->24574481]
}, Open  ]],
Cell[571119, 16787, 48, 1, 16, "CloseEnvironment",
 CellID->259129778],
Cell[CellGroupData[{
Cell[571192, 16792, 4123, 112, 28, "ProofInfo",
 CellTags->{"Proof|ID:220417861", "Proof|ID:220417861-1"},
 CellID->227048585],
Cell[575318, 16906, 480, 12, 70, "ProofInfoBody",
 CellID->124141065]
}, Closed]],
Cell[CellGroupData[{
Cell[575835, 16923, 3453, 98, 26, "ProofInfo",
 CellTags->{"Proof|ID:52539739", "Proof|ID:52539739-1"},
 CellID->142039401],
Cell[579291, 17023, 479, 12, 70, "ProofInfoBody",
 CellID->135864141]
}, Closed]],
Cell[CellGroupData[{
Cell[579807, 17040, 4068, 109, 26, "ProofInfo",
 CellTags->{"Proof|ID:94484613", "Proof|ID:94484613-1"},
 CellID->325556718],
Cell[583878, 17151, 522, 13, 70, "ProofInfoBody",
 CellID->55143000]
}, Closed]],
Cell[CellGroupData[{
Cell[584437, 17169, 4059, 108, 26, "ProofInfo",
 CellTags->{"Proof|ID:321256465", "Proof|ID:321256465-1"},
 CellID->438580499],
Cell[588499, 17279, 480, 12, 70, "ProofInfoBody",
 CellID->227406264]
}, Closed]],
Cell[CellGroupData[{
Cell[589016, 17296, 2316, 70, 26, "ProofInfo",
 CellTags->{"Proof|ID:129809121", "Proof|ID:129809121-1"},
 CellID->242312613],
Cell[591335, 17368, 480, 12, 70, "ProofInfoBody",
 CellID->768460961]
}, Closed]],
Cell[CellGroupData[{
Cell[591852, 17385, 4053, 110, 26, "ProofInfo",
 CellTags->{"Proof|ID:637862621", "Proof|ID:637862621-1"},
 CellID->752123440],
Cell[595908, 17497, 480, 12, 70, "ProofInfoBody",
 CellID->189898560]
}, Closed]],
Cell[CellGroupData[{
Cell[596425, 17514, 2499, 74, 26, "ProofInfo",
 CellTags->{"Proof|ID:12369324", "Proof|ID:12369324-1"},
 CellID->28237667],
Cell[598927, 17590, 479, 12, 70, "ProofInfoBody",
 CellID->609896313]
}, Closed]],
Cell[CellGroupData[{
Cell[599443, 17607, 3299, 92, 26, "ProofInfo",
 CellTags->{"Proof|ID:19276", "Proof|ID:19276-1"},
 CellID->146831318],
Cell[602745, 17701, 480, 12, 70, "ProofInfoBody",
 CellID->835535524]
}, Closed]],
Cell[CellGroupData[{
Cell[603262, 17718, 4146, 110, 26, "ProofInfo",
 CellTags->{"Proof|ID:164883544", "Proof|ID:164883544-1"},
 CellID->62903797],
Cell[607411, 17830, 480, 12, 70, "ProofInfoBody",
 CellID->325542796]
}, Closed]],
Cell[CellGroupData[{
Cell[607928, 17847, 1902, 58, 26, "ProofInfo",
 CellTags->{"Proof|ID:357440757", "Proof|ID:357440757-1"},
 CellID->81770248],
Cell[609833, 17907, 479, 12, 70, "ProofInfoBody",
 CellID->82676088]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[610361, 17925, 412, 12, 22, "Subsubsection",
 CellID->55494429],
Cell[610776, 17939, 47, 1, 12, "OpenEnvironment",
 CellID->135470470],
Cell[CellGroupData[{
Cell[610848, 17944, 385, 9, 43, "EnvironmentHeader",
 CellID->348809526],
Cell[611236, 17955, 1361, 41, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:125790650", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ Pair PreInt", "IR"},
 CellID->125790650],
Cell[612600, 17998, 715, 19, 41, "GlobalDeclaration",
 CellTags->{
  "ID:86057977", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->86057977],
Cell[613318, 18019, 856, 24, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:196108575", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ closed PreInt"},
 CellID->196108575],
Cell[614177, 18045, 844, 24, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:132949647", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 rneutral PreInt"},
 CellID->132949647],
Cell[615024, 18071, 851, 24, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:25630206", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ comm PreInt"},
 CellID->25630206],
Cell[615878, 18097, 694, 18, 41, "GlobalDeclaration",
 CellTags->{
  "ID:16604231", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->16604231],
Cell[616575, 18117, 1018, 31, 37, "FormalTextInputFormula",
 CellTags->{
  "ID:73792979", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ assoc PreInt"},
 CellID->73792979],
Cell[617596, 18150, 1093, 31, 35, "FormalTextInputFormula",
 CellTags->{
  "ID:243295365", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ rcanc PreInt"},
 CellID->243295365],
Cell[618692, 18183, 1057, 31, 35, "FormalTextInputFormula",
 CellTags->{
  "ID:47524840", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ invaiant equiv PreInt"},
 CellID->47524840],
Cell[619752, 18216, 1030, 31, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:107686175", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ isInvariantFunction2 PreInt"},
 CellID->107686175],
Cell[620785, 18249, 65, 1, 26, "EndEnvironmentMarker",
 CellID->125892535]
}, Open  ]],
Cell[620865, 18253, 47, 1, 16, "CloseEnvironment",
 CellID->92071022],
Cell[CellGroupData[{
Cell[620937, 18258, 4037, 108, 28, "ProofInfo",
 CellTags->{"Proof|ID:125790650", "Proof|ID:125790650-1"},
 CellID->219935054],
Cell[624977, 18368, 480, 12, 70, "ProofInfoBody",
 CellID->267994709]
}, Closed]],
Cell[CellGroupData[{
Cell[625494, 18385, 2871, 82, 26, "ProofInfo",
 CellTags->{"Proof|ID:196108575", "Proof|ID:196108575-1"},
 CellID->746037930],
Cell[628368, 18469, 480, 12, 70, "ProofInfoBody",
 CellID->198300422]
}, Closed]],
Cell[CellGroupData[{
Cell[628885, 18486, 2583, 78, 26, "ProofInfo",
 CellTags->{"Proof|ID:132949647", "Proof|ID:132949647-1"},
 CellID->16458106],
Cell[631471, 18566, 479, 12, 70, "ProofInfoBody",
 CellID->24014104]
}, Closed]],
Cell[CellGroupData[{
Cell[631987, 18583, 3117, 90, 26, "ProofInfo",
 CellTags->{"Proof|ID:25630206", "Proof|ID:25630206-1"},
 CellID->184131697],
Cell[635107, 18675, 478, 12, 70, "ProofInfoBody",
 CellID->15556707]
}, Closed]],
Cell[CellGroupData[{
Cell[635622, 18692, 3913, 108, 26, "ProofInfo",
 CellTags->{"Proof|ID:73792979", "Proof|ID:73792979-1"},
 CellID->273331503],
Cell[639538, 18802, 479, 12, 70, "ProofInfoBody",
 CellID->675199731]
}, Closed]],
Cell[CellGroupData[{
Cell[640054, 18819, 4121, 110, 26, "ProofInfo",
 CellTags->{"Proof|ID:243295365", "Proof|ID:243295365-1"},
 CellID->446127764],
Cell[644178, 18931, 480, 12, 70, "ProofInfoBody",
 CellID->136545031]
}, Closed]],
Cell[CellGroupData[{
Cell[644695, 18948, 4134, 110, 26, "ProofInfo",
 CellTags->{"Proof|ID:47524840", "Proof|ID:47524840-1"},
 CellID->317988883],
Cell[648832, 19060, 478, 12, 70, "ProofInfoBody",
 CellID->49485106]
}, Closed]],
Cell[CellGroupData[{
Cell[649347, 19077, 2019, 60, 26, "ProofInfo",
 CellTags->{"Proof|ID:107686175", "Proof|ID:107686175-1"},
 CellID->207023133],
Cell[651369, 19139, 479, 12, 70, "ProofInfoBody",
 CellID->11877910]
}, Closed]],
Cell[651863, 19154, 47, 1, 12, "OpenEnvironment",
 CellID->115221006],
Cell[CellGroupData[{
Cell[651935, 19159, 454, 14, 43, "EnvironmentHeader",
 CellID->286130886],
Cell[652392, 19175, 1112, 34, 34, "FormalTextInputFormula",
 CellTags->{
  "ID:360943410", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- unary Pair PreInt", "IR"},
 CellID->360943410],
Cell[653507, 19211, 695, 18, 41, "GlobalDeclaration",
 CellTags->{
  "ID:474936069", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->474936069],
Cell[654205, 19231, 863, 24, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:5265224", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- unary closed PreInt", "IR"},
 CellID->5265224],
Cell[655071, 19257, 1076, 30, 35, "FormalTextInputFormula",
 CellTags->{
  "ID:203475906", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- inverse PreInt"},
 CellID->203475906],
Cell[656150, 19289, 959, 25, 35, "FormalTextInputFormula",
 CellTags->{
  "ID:611704112", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- isInvariantFunction1 PreInt"},
 CellID->611704112],
Cell[657112, 19316, 65, 1, 26, "EndEnvironmentMarker",
 CellID->416398914]
}, Open  ]],
Cell[657192, 19320, 48, 1, 16, "CloseEnvironment",
 CellID->313664736],
Cell[CellGroupData[{
Cell[657265, 19325, 2987, 86, 28, "ProofInfo",
 CellTags->{"Proof|ID:360943410", "Proof|ID:360943410-1"},
 CellID->307773152],
Cell[660255, 19413, 480, 12, 70, "ProofInfoBody",
 CellID->129842997]
}, Closed]],
Cell[CellGroupData[{
Cell[660772, 19430, 2509, 74, 26, "ProofInfo",
 CellTags->{"Proof|ID:5265224", "Proof|ID:5265224-1"},
 CellID->489898059],
Cell[663284, 19506, 477, 12, 70, "ProofInfoBody",
 CellID->58997533]
}, Closed]],
Cell[CellGroupData[{
Cell[663798, 19523, 2915, 84, 26, "ProofInfo",
 CellTags->{"Proof|ID:203475906", "Proof|ID:203475906-1"},
 CellID->819919783],
Cell[666716, 19609, 480, 12, 70, "ProofInfoBody",
 CellID->420597009]
}, Closed]],
Cell[CellGroupData[{
Cell[667233, 19626, 2027, 60, 26, "ProofInfo",
 CellTags->{"Proof|ID:611704112", "Proof|ID:611704112-1"},
 CellID->443769915],
Cell[669263, 19688, 480, 12, 70, "ProofInfoBody",
 CellID->218910620]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[669792, 19706, 310, 9, 22, "Subsubsection",
 CellID->119023973],
Cell[670105, 19717, 47, 1, 70, "OpenEnvironment",
 CellID->170850154],
Cell[CellGroupData[{
Cell[670177, 19722, 508, 15, 70, "EnvironmentHeader",
 CellID->17732315],
Cell[670688, 19739, 1603, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:455358938", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] Pair PreInt", "IR"},
 CellID->455358938],
Cell[672294, 19790, 771, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:206670450", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->206670450],
Cell[673068, 19812, 908, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:80978087", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] closed PreInt", "IR"},
 CellID->80978087],
Cell[673979, 19839, 878, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:561254237", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 rannihilator PreInt"},
 CellID->561254237],
Cell[674860, 19865, 897, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:201948360", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 rneutral PreInt"},
 CellID->201948360],
Cell[675760, 19892, 853, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:152728157", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] comm PreInt"},
 CellID->152728157],
Cell[676616, 19918, 1344, 41, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:258603229", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] zdf PreInt"},
 CellID->258603229],
Cell[677963, 19961, 695, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:200297603", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->200297603],
Cell[678661, 19981, 1023, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:142359696", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] assoc PreInt"},
 CellID->142359696],
Cell[679687, 20014, 1053, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:262630475", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] distr + PreInt"},
 CellID->262630475],
Cell[680743, 20047, 1064, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:146416852", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] invariant equiv PreInt"},
 CellID->146416852],
Cell[681810, 20080, 1030, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:204021385", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] isInvariantFunction2 PreInt"},
 CellID->204021385],
Cell[682843, 20113, 65, 1, 70, "EndEnvironmentMarker",
 CellID->102737931]
}, Open  ]],
Cell[682923, 20117, 48, 1, 70, "CloseEnvironment",
 CellID->282267077],
Cell[CellGroupData[{
Cell[682996, 20122, 5125, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:455358938", "Proof|ID:455358938-1"},
 CellID->297124414],
Cell[688124, 20256, 480, 12, 70, "ProofInfoBody",
 CellID->122132350]
}, Closed]],
Cell[CellGroupData[{
Cell[688641, 20273, 2866, 82, 70, "ProofInfo",
 CellTags->{"Proof|ID:80978087", "Proof|ID:80978087-1"},
 CellID->295496011],
Cell[691510, 20357, 479, 12, 70, "ProofInfoBody",
 CellID->498778090]
}, Closed]],
Cell[CellGroupData[{
Cell[692026, 20374, 2556, 76, 70, "ProofInfo",
 CellTags->{"Proof|ID:561254237", "Proof|ID:561254237-1"},
 CellID->690925667],
Cell[694585, 20452, 479, 12, 70, "ProofInfoBody",
 CellID->25562229]
}, Closed]],
Cell[CellGroupData[{
Cell[695101, 20469, 2584, 78, 70, "ProofInfo",
 CellTags->{"Proof|ID:201948360", "Proof|ID:201948360-1"},
 CellID->558544782],
Cell[697688, 20549, 480, 12, 70, "ProofInfoBody",
 CellID->365680064]
}, Closed]],
Cell[CellGroupData[{
Cell[698205, 20566, 3121, 90, 70, "ProofInfo",
 CellTags->{"Proof|ID:152728157", "Proof|ID:152728157-1"},
 CellID->24746466],
Cell[701329, 20658, 480, 12, 70, "ProofInfoBody",
 CellID->497158307]
}, Closed]],
Cell[CellGroupData[{
Cell[701846, 20675, 4065, 110, 70, "ProofInfo",
 CellTags->{"Proof|ID:258603229", "Proof|ID:258603229-1"},
 CellID->149136758],
Cell[705914, 20787, 479, 12, 70, "ProofInfoBody",
 CellID->40095241]
}, Closed]],
Cell[CellGroupData[{
Cell[706430, 20804, 3915, 108, 70, "ProofInfo",
 CellTags->{"Proof|ID:142359696", "Proof|ID:142359696-1"},
 CellID->796805],
Cell[710348, 20914, 480, 12, 70, "ProofInfoBody",
 CellID->116032710]
}, Closed]],
Cell[CellGroupData[{
Cell[710865, 20931, 4174, 114, 70, "ProofInfo",
 CellTags->{"Proof|ID:262630475", "Proof|ID:262630475-1"},
 CellID->457849091],
Cell[715042, 21047, 480, 12, 70, "ProofInfoBody",
 CellID->105684652]
}, Closed]],
Cell[CellGroupData[{
Cell[715559, 21064, 4140, 110, 70, "ProofInfo",
 CellTags->{"Proof|ID:146416852", "Proof|ID:146416852-1"},
 CellID->83067894],
Cell[719702, 21176, 480, 12, 70, "ProofInfoBody",
 CellID->109470325]
}, Closed]],
Cell[CellGroupData[{
Cell[720219, 21193, 2019, 60, 70, "ProofInfo",
 CellTags->{"Proof|ID:204021385", "Proof|ID:204021385-1"},
 CellID->716469443],
Cell[722241, 21255, 480, 12, 70, "ProofInfoBody",
 CellID->145105386]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[722770, 21273, 253, 4, 22, "Subsubsection",
 CellID->293842696],
Cell[723026, 21279, 47, 1, 12, "OpenEnvironment",
 CellID->614975653],
Cell[CellGroupData[{
Cell[723098, 21284, 409, 11, 43, "EnvironmentHeader",
 CellID->100053836],
Cell[723510, 21297, 1590, 50, 38, "FormalTextInputFormula",
 CellTags->{
  "ID:129935970", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] Pair PreInt", "IR"},
 CellID->129935970],
Cell[725103, 21349, 1539, 49, 48, "FormalTextInputFormula",
 CellTags->{
  "ID:95876614", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] 0 PreInt", "IR"},
 CellID->95876614],
Cell[726645, 21400, 919, 24, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:221338547", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 \[GreaterEqual] 0 PreInt"},
 CellID->221338547],
Cell[727567, 21426, 1078, 32, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:13569836", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] isInvariantRelation2 PreInt"},
 CellID->13569836],
Cell[728648, 21460, 722, 19, 41, "GlobalDeclaration",
 CellTags->{
  "ID:337456597", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->337456597],
Cell[729373, 21481, 822, 22, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:83533617", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] reflexive PreInt"},
 CellID->83533617],
Cell[730198, 21505, 1157, 34, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:281735911", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] antisymmetric PreInt"},
 CellID->281735911],
Cell[731358, 21541, 913, 25, 33, "FormalTextInputFormula",
 CellTags->{
  "ID:115550209", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] total PreInt"},
 CellID->115550209],
Cell[732274, 21568, 1372, 42, 39, "FormalTextInputFormula",
 CellTags->{
  "ID:215340491", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] discrete PreInt"},
 CellID->215340491],
Cell[733649, 21612, 1298, 38, 37, "FormalTextInputFormula",
 CellTags->{
  "ID:92075332", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] \[RawStar] PreInt"},
 CellID->92075332],
Cell[734950, 21652, 1299, 39, 50, "FormalTextInputFormula",
 CellTags->{
  "ID:145078248", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat \[SucceedsTilde] PreInt+"},
 CellID->145078248],
Cell[736252, 21693, 697, 18, 41, "GlobalDeclaration",
 CellTags->{
  "ID:118070489", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->118070489],
Cell[736952, 21713, 1125, 33, 37, "FormalTextInputFormula",
 CellTags->{
  "ID:307661531", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] transitive PreInt"},
 CellID->307661531],
Cell[738080, 21748, 1012, 28, 30, "FormalTextInputFormula",
 CellTags->{
  "ID:647932785", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] + PreInt"},
 CellID->647932785],
Cell[739095, 21778, 64, 1, 26, "EndEnvironmentMarker",
 CellID->15173820]
}, Open  ]],
Cell[739174, 21782, 48, 1, 16, "CloseEnvironment",
 CellID->362378933],
Cell[CellGroupData[{
Cell[739247, 21787, 4139, 112, 28, "ProofInfo",
 CellTags->{"Proof|ID:129935970", "Proof|ID:129935970-1"},
 CellID->9772636],
Cell[743389, 21901, 479, 12, 70, "ProofInfoBody",
 CellID->22268996]
}, Closed]],
Cell[CellGroupData[{
Cell[743905, 21918, 3458, 96, 26, "ProofInfo",
 CellTags->{"Proof|ID:95876614", "Proof|ID:95876614-1"},
 CellID->62135307],
Cell[747366, 22016, 479, 12, 70, "ProofInfoBody",
 CellID->392128574]
}, Closed]],
Cell[CellGroupData[{
Cell[747882, 22033, 1951, 60, 26, "ProofInfo",
 CellTags->{"Proof|ID:221338547", "Proof|ID:221338547-1"},
 CellID->51061701],
Cell[749836, 22095, 479, 12, 70, "ProofInfoBody",
 CellID->92240380]
}, Closed]],
Cell[CellGroupData[{
Cell[750352, 22112, 2086, 62, 26, "ProofInfo",
 CellTags->{"Proof|ID:13569836", "Proof|ID:13569836-1"},
 CellID->553106541],
Cell[752441, 22176, 479, 12, 70, "ProofInfoBody",
 CellID->461282630]
}, Closed]],
Cell[CellGroupData[{
Cell[752957, 22193, 2477, 75, 26, "ProofInfo",
 CellTags->{"Proof|ID:83533617", "Proof|ID:83533617-1"},
 CellID->1756479],
Cell[755437, 22270, 479, 12, 70, "ProofInfoBody",
 CellID->169206792]
}, Closed]],
Cell[CellGroupData[{
Cell[755953, 22287, 3758, 102, 26, "ProofInfo",
 CellTags->{"Proof|ID:281735911", "Proof|ID:281735911-1"},
 CellID->44053081],
Cell[759714, 22391, 480, 12, 70, "ProofInfoBody",
 CellID->621527603]
}, Closed]],
Cell[CellGroupData[{
Cell[760231, 22408, 3202, 90, 26, "ProofInfo",
 CellTags->{"Proof|ID:115550209", "Proof|ID:115550209-1"},
 CellID->877842890],
Cell[763436, 22500, 479, 12, 70, "ProofInfoBody",
 CellID->46153360]
}, Closed]],
Cell[CellGroupData[{
Cell[763952, 22517, 3760, 103, 26, "ProofInfo",
 CellTags->{"Proof|ID:215340491", "Proof|ID:215340491-1"},
 CellID->147476088],
Cell[767715, 22622, 480, 12, 70, "ProofInfoBody",
 CellID->342294249]
}, Closed]],
Cell[CellGroupData[{
Cell[768232, 22639, 3848, 104, 26, "ProofInfo",
 CellTags->{"Proof|ID:92075332", "Proof|ID:92075332-1"},
 CellID->47485746],
Cell[772083, 22745, 479, 12, 70, "ProofInfoBody",
 CellID->108029984]
}, Closed]],
Cell[CellGroupData[{
Cell[772599, 22762, 3743, 102, 26, "ProofInfo",
 CellTags->{"Proof|ID:145078248", "Proof|ID:145078248-1"},
 CellID->565621444],
Cell[776345, 22866, 480, 12, 70, "ProofInfoBody",
 CellID->548997320]
}, Closed]],
Cell[CellGroupData[{
Cell[776862, 22883, 3981, 107, 26, "ProofInfo",
 CellTags->{"Proof|ID:307661531", "Proof|ID:307661531-1"},
 CellID->58099428],
Cell[780846, 22992, 480, 12, 70, "ProofInfoBody",
 CellID->476167925]
}, Closed]],
Cell[CellGroupData[{
Cell[781363, 23009, 3996, 108, 26, "ProofInfo",
 CellTags->{"Proof|ID:647932785", "Proof|ID:647932785-1"},
 CellID->4451951],
Cell[785362, 23119, 479, 12, 70, "ProofInfoBody",
 CellID->15683258]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[785902, 23138, 184, 3, 30, "Subsection",
 CellID->414452305],
Cell[786089, 23143, 749, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:463204962", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->463204962],
Cell[CellGroupData[{
Cell[786863, 23166, 126, 2, 70, "Subsubsection",
 CellID->269593602],
Cell[786992, 23170, 47, 1, 70, "OpenEnvironment",
 CellID->614819668],
Cell[CellGroupData[{
Cell[787064, 23175, 378, 9, 70, "EnvironmentHeader",
 CellID->201483006],
Cell[787445, 23186, 645, 16, 70, "GlobalDeclaration",
 CellTags->{
  "ID:32325634", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->32325634],
Cell[788093, 23204, 1067, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:112681936", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Int \[Element]"},
 CellID->112681936],
Cell[789163, 23239, 849, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:52976508", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Int 0"},
 CellID->52976508],
Cell[790015, 23265, 852, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:712864778", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Int 1"},
 CellID->712864778],
Cell[790870, 23291, 1098, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:19419340", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Int +"},
 CellID->19419340],
Cell[791971, 23331, 872, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:280798627", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Int - unary"},
 CellID->280798627],
Cell[792846, 23357, 1124, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:328192039", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Int \[RawStar]"},
 CellID->328192039],
Cell[793973, 23398, 1184, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:819512290", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Int \[GreaterEqual]"},
 CellID->819512290],
Cell[795160, 23439, 65, 1, 70, "EndEnvironmentMarker",
 CellID->395448942]
}, Open  ]],
Cell[795240, 23443, 48, 1, 70, "CloseEnvironment",
 CellID->281466815]
}, Closed]],
Cell[CellGroupData[{
Cell[795325, 23449, 267, 6, 70, "Subsubsection",
 CellID->95650746],
Cell[795595, 23457, 47, 1, 70, "OpenEnvironment",
 CellID->570814532],
Cell[CellGroupData[{
Cell[795667, 23462, 386, 9, 70, "EnvironmentHeader",
 CellID->30393050],
Cell[796056, 23473, 757, 20, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:354647619", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Int isDomain", "IR"},
 CellID->354647619],
Cell[796816, 23495, 1619, 53, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:565027904", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[Element] Int alternative", "IR"},
 CellID->565027904],
Cell[798438, 23550, 1163, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:40757894", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "EC \[Element] Int", "IR"},
 CellID->40757894],
Cell[799604, 23587, 833, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:350791961", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 \[Element] Int", "IR"},
 CellID->350791961],
Cell[800440, 23612, 825, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:181240486", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 \[Element] Int"},
 CellID->181240486],
Cell[801268, 23637, 815, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:364462666", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 \[NotEqual] 0 Int"},
 CellID->364462666],
Cell[802086, 23661, 65, 1, 70, "EndEnvironmentMarker",
 CellID->388453292]
}, Open  ]],
Cell[802166, 23665, 48, 1, 70, "CloseEnvironment",
 CellID->418936054],
Cell[CellGroupData[{
Cell[802239, 23670, 1708, 54, 70, "ProofInfo",
 CellTags->{"Proof|ID:354647619", "Proof|ID:354647619-1"},
 CellID->268377469],
Cell[803950, 23726, 478, 12, 70, "ProofInfoBody",
 CellID->8704553]
}, Closed]],
Cell[CellGroupData[{
Cell[804465, 23743, 3621, 100, 70, "ProofInfo",
 CellTags->{"Proof|ID:565027904", "Proof|ID:565027904-1"},
 CellID->760024466],
Cell[808089, 23845, 480, 12, 70, "ProofInfoBody",
 CellID->102735102]
}, Closed]],
Cell[CellGroupData[{
Cell[808606, 23862, 2755, 80, 70, "ProofInfo",
 CellTags->{"Proof|ID:40757894", "Proof|ID:40757894-1"},
 CellID->35500606],
Cell[811364, 23944, 479, 12, 70, "ProofInfoBody",
 CellID->168541484]
}, Closed]],
Cell[CellGroupData[{
Cell[811880, 23961, 1847, 58, 70, "ProofInfo",
 CellTags->{"Proof|ID:350791961", "Proof|ID:350791961-1"},
 CellID->16051126],
Cell[813730, 24021, 480, 12, 70, "ProofInfoBody",
 CellID->704723967]
}, Closed]],
Cell[CellGroupData[{
Cell[814247, 24038, 1848, 58, 70, "ProofInfo",
 CellTags->{"Proof|ID:181240486", "Proof|ID:181240486-1"},
 CellID->157610485],
Cell[816098, 24098, 479, 12, 70, "ProofInfoBody",
 CellID->95057389]
}, Closed]],
Cell[CellGroupData[{
Cell[816614, 24115, 1846, 58, 70, "ProofInfo",
 CellTags->{"Proof|ID:364462666", "Proof|ID:364462666-1"},
 CellID->279790495],
Cell[818463, 24175, 480, 12, 70, "ProofInfoBody",
 CellID->107741559]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[818992, 24193, 317, 11, 70, "Subsubsection",
 CellID->74984569],
Cell[819312, 24206, 47, 1, 70, "OpenEnvironment",
 CellID->234630122],
Cell[CellGroupData[{
Cell[819384, 24211, 383, 9, 70, "EnvironmentHeader",
 CellID->25494490],
Cell[819770, 24222, 1339, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:56650873", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ EC Int", "IR"},
 CellID->56650873],
Cell[821112, 24266, 777, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:13236461", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->13236461],
Cell[821892, 24288, 848, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:562490241", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ closed Int", "IR"},
 CellID->562490241],
Cell[822743, 24314, 827, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:624613835", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 rneutral Int"},
 CellID->624613835],
Cell[823573, 24339, 844, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:115787764", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ comm Int"},
 CellID->115787764],
Cell[824420, 24365, 1048, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:11815767", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ assoc Int"},
 CellID->11815767],
Cell[825471, 24398, 961, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:139539770", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "+ rcanc Int"},
 CellID->139539770],
Cell[826435, 24427, 65, 1, 70, "EndEnvironmentMarker",
 CellID->317123382]
}, Open  ]],
Cell[826515, 24431, 48, 1, 70, "CloseEnvironment",
 CellID->213410482],
Cell[CellGroupData[{
Cell[826588, 24436, 4406, 118, 70, "ProofInfo",
 CellTags->{"Proof|ID:56650873", "Proof|ID:56650873-1"},
 CellID->106415148],
Cell[830997, 24556, 478, 12, 70, "ProofInfoBody",
 CellID->80588368]
}, Closed]],
Cell[CellGroupData[{
Cell[831512, 24573, 2841, 82, 70, "ProofInfo",
 CellTags->{"Proof|ID:562490241", "Proof|ID:562490241-1"},
 CellID->110816780],
Cell[834356, 24657, 480, 12, 70, "ProofInfoBody",
 CellID->120666679]
}, Closed]],
Cell[CellGroupData[{
Cell[834873, 24674, 2559, 78, 70, "ProofInfo",
 CellTags->{"Proof|ID:624613835", "Proof|ID:624613835-1"},
 CellID->23051911],
Cell[837435, 24754, 479, 12, 70, "ProofInfoBody",
 CellID->26362014]
}, Closed]],
Cell[CellGroupData[{
Cell[837951, 24771, 3092, 90, 70, "ProofInfo",
 CellTags->{"Proof|ID:115787764", "Proof|ID:115787764-1"},
 CellID->253640770],
Cell[841046, 24863, 480, 12, 70, "ProofInfoBody",
 CellID->217136621]
}, Closed]],
Cell[CellGroupData[{
Cell[841563, 24880, 3864, 108, 70, "ProofInfo",
 CellTags->{"Proof|ID:11815767", "Proof|ID:11815767-1"},
 CellID->44906707],
Cell[845430, 24990, 478, 12, 70, "ProofInfoBody",
 CellID->31471341]
}, Closed]],
Cell[CellGroupData[{
Cell[845945, 25007, 3753, 104, 70, "ProofInfo",
 CellTags->{"Proof|ID:139539770", "Proof|ID:139539770-1"},
 CellID->42190126],
Cell[849701, 25113, 480, 12, 70, "ProofInfoBody",
 CellID->194323372]
}, Closed]],
Cell[850196, 25128, 46, 1, 70, "OpenEnvironment",
 CellID->13262632],
Cell[CellGroupData[{
Cell[850267, 25133, 450, 14, 70, "EnvironmentHeader",
 CellID->18951672],
Cell[850720, 25149, 1242, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:331988502", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- unary EC Int", "IR"},
 CellID->331988502],
Cell[851965, 25190, 694, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:138388913", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->138388913],
Cell[852662, 25210, 857, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:789456798", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- unary closed Int", "IR"},
 CellID->789456798],
Cell[853522, 25236, 895, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:561244613", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "- inverse Int"},
 CellID->561244613],
Cell[854420, 25264, 65, 1, 70, "EndEnvironmentMarker",
 CellID->183233282]
}, Open  ]],
Cell[854500, 25268, 48, 1, 70, "CloseEnvironment",
 CellID->412087319],
Cell[CellGroupData[{
Cell[854573, 25273, 3539, 98, 70, "ProofInfo",
 CellTags->{"Proof|ID:331988502", "Proof|ID:331988502-1"},
 CellID->217485270],
Cell[858115, 25373, 480, 12, 70, "ProofInfoBody",
 CellID->999928246]
}, Closed]],
Cell[CellGroupData[{
Cell[858632, 25390, 2495, 74, 70, "ProofInfo",
 CellTags->{"Proof|ID:789456798", "Proof|ID:789456798-1"},
 CellID->617130052],
Cell[861130, 25466, 480, 12, 70, "ProofInfoBody",
 CellID->194480606]
}, Closed]],
Cell[CellGroupData[{
Cell[861647, 25483, 2702, 80, 70, "ProofInfo",
 CellTags->{"Proof|ID:561244613", "Proof|ID:561244613-1"},
 CellID->394996459],
Cell[864352, 25565, 477, 12, 70, "ProofInfoBody",
 CellID->983259]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[864878, 25583, 259, 8, 70, "Subsubsection",
 CellID->716281922],
Cell[865140, 25593, 47, 1, 70, "OpenEnvironment",
 CellID->101122110],
Cell[CellGroupData[{
Cell[865212, 25598, 453, 14, 70, "EnvironmentHeader",
 CellID->317349246],
Cell[865668, 25614, 1340, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:947607354", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] EC Int", "IR"},
 CellID->947607354],
Cell[867011, 25658, 701, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:105543058", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->105543058],
Cell[867715, 25679, 852, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:696766377", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] closed Int", "IR"},
 CellID->696766377],
Cell[868570, 25705, 861, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:202378771", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "0 rannihilator Int"},
 CellID->202378771],
Cell[869434, 25731, 827, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:689609352", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 rneutral Int"},
 CellID->689609352],
Cell[870264, 25756, 842, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:196343023", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] comm Int"},
 CellID->196343023],
Cell[871109, 25782, 1193, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:207834126", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] zdf Int"},
 CellID->207834126],
Cell[872305, 25823, 999, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:10196611", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] assoc Int"},
 CellID->10196611],
Cell[873307, 25855, 1032, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:777222529", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[RawStar] distr + Int"},
 CellID->777222529],
Cell[874342, 25888, 64, 1, 70, "EndEnvironmentMarker",
 CellID->35189982]
}, Open  ]],
Cell[874421, 25892, 48, 1, 70, "CloseEnvironment",
 CellID->566077695],
Cell[CellGroupData[{
Cell[874494, 25897, 4411, 118, 70, "ProofInfo",
 CellTags->{"Proof|ID:947607354", "Proof|ID:947607354-1"},
 CellID->558776955],
Cell[878908, 26017, 480, 12, 70, "ProofInfoBody",
 CellID->522627315]
}, Closed]],
Cell[CellGroupData[{
Cell[879425, 26034, 2840, 82, 70, "ProofInfo",
 CellTags->{"Proof|ID:696766377", "Proof|ID:696766377-1"},
 CellID->81854393],
Cell[882268, 26118, 480, 12, 70, "ProofInfoBody",
 CellID->106710227]
}, Closed]],
Cell[CellGroupData[{
Cell[882785, 26135, 2526, 76, 70, "ProofInfo",
 CellTags->{"Proof|ID:202378771", "Proof|ID:202378771-1"},
 CellID->465404293],
Cell[885314, 26213, 480, 12, 70, "ProofInfoBody",
 CellID->561652051]
}, Closed]],
Cell[CellGroupData[{
Cell[885831, 26230, 2560, 78, 70, "ProofInfo",
 CellTags->{"Proof|ID:689609352", "Proof|ID:689609352-1"},
 CellID->485464469],
Cell[888394, 26310, 480, 12, 70, "ProofInfoBody",
 CellID->294236516]
}, Closed]],
Cell[CellGroupData[{
Cell[888911, 26327, 3091, 90, 70, "ProofInfo",
 CellTags->{"Proof|ID:196343023", "Proof|ID:196343023-1"},
 CellID->43558645],
Cell[892005, 26419, 480, 12, 70, "ProofInfoBody",
 CellID->308643022]
}, Closed]],
Cell[CellGroupData[{
Cell[892522, 26436, 3504, 98, 70, "ProofInfo",
 CellTags->{"Proof|ID:207834126", "Proof|ID:207834126-1"},
 CellID->5339075],
Cell[896029, 26536, 480, 12, 70, "ProofInfoBody",
 CellID->181354457]
}, Closed]],
Cell[CellGroupData[{
Cell[896546, 26553, 3865, 108, 70, "ProofInfo",
 CellTags->{"Proof|ID:10196611", "Proof|ID:10196611-1"},
 CellID->268794065],
Cell[900414, 26663, 479, 12, 70, "ProofInfoBody",
 CellID->283564566]
}, Closed]],
Cell[CellGroupData[{
Cell[900930, 26680, 4120, 114, 70, "ProofInfo",
 CellTags->{"Proof|ID:777222529", "Proof|ID:777222529-1"},
 CellID->343265473],
Cell[905053, 26796, 480, 12, 70, "ProofInfoBody",
 CellID->885039319]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[905582, 26814, 202, 3, 70, "Subsubsection",
 CellID->389529963],
Cell[905787, 26819, 46, 1, 70, "OpenEnvironment",
 CellID->77983804],
Cell[CellGroupData[{
Cell[905858, 26824, 396, 9, 70, "EnvironmentHeader",
 CellID->210469497],
Cell[906257, 26835, 1550, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:785793765", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] EC Int", "IR"},
 CellID->785793765],
Cell[907810, 26886, 850, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:72386745", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "1 \[GreaterEqual] 0 Int"},
 CellID->72386745],
Cell[908663, 26911, 726, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:100510373", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb"},
 CellID->100510373],
Cell[909392, 26932, 815, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:267773274", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] reflexive Int"},
 CellID->267773274],
Cell[910210, 26955, 1075, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:69005952", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] antisymmetric Int"},
 CellID->69005952],
Cell[911288, 26988, 1112, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:275182510", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] transitive Int"},
 CellID->275182510],
Cell[912403, 27023, 899, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:176048001", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] total Int"},
 CellID->176048001],
Cell[913305, 27050, 1156, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:566590035", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] discrete Int"},
 CellID->566590035],
Cell[914464, 27086, 1000, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:262876021", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] + Int"},
 CellID->262876021],
Cell[915467, 27116, 1224, 37, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:34443687", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "\[GreaterEqual] \[RawStar] Int"},
 CellID->34443687],
Cell[916694, 27155, 65, 1, 70, "EndEnvironmentMarker",
 CellID->341695448]
}, Open  ]],
Cell[916774, 27159, 48, 1, 70, "CloseEnvironment",
 CellID->127316464],
Cell[CellGroupData[{
Cell[916847, 27164, 4128, 112, 70, "ProofInfo",
 CellTags->{"Proof|ID:785793765", "Proof|ID:785793765-1"},
 CellID->518056285],
Cell[920978, 27278, 480, 12, 70, "ProofInfoBody",
 CellID->605637846]
}, Closed]],
Cell[CellGroupData[{
Cell[921495, 27295, 1922, 60, 70, "ProofInfo",
 CellTags->{"Proof|ID:72386745", "Proof|ID:72386745-1"},
 CellID->72615069],
Cell[923420, 27357, 478, 12, 70, "ProofInfoBody",
 CellID->22111013]
}, Closed]],
Cell[CellGroupData[{
Cell[923935, 27374, 2457, 74, 70, "ProofInfo",
 CellTags->{"Proof|ID:267773274", "Proof|ID:267773274-1"},
 CellID->50495462],
Cell[926395, 27450, 480, 12, 70, "ProofInfoBody",
 CellID->442279872]
}, Closed]],
Cell[CellGroupData[{
Cell[926912, 27467, 3568, 99, 70, "ProofInfo",
 CellTags->{"Proof|ID:69005952", "Proof|ID:69005952-1"},
 CellID->49834528],
Cell[930483, 27568, 479, 12, 70, "ProofInfoBody",
 CellID->128346525]
}, Closed]],
Cell[CellGroupData[{
Cell[930999, 27585, 3939, 107, 70, "ProofInfo",
 CellTags->{"Proof|ID:275182510", "Proof|ID:275182510-1"},
 CellID->94179691],
Cell[934941, 27694, 479, 12, 70, "ProofInfoBody",
 CellID->25059238]
}, Closed]],
Cell[CellGroupData[{
Cell[935457, 27711, 3172, 90, 70, "ProofInfo",
 CellTags->{"Proof|ID:176048001", "Proof|ID:176048001-1"},
 CellID->320141850],
Cell[938632, 27803, 480, 12, 70, "ProofInfoBody",
 CellID->123691091]
}, Closed]],
Cell[CellGroupData[{
Cell[939149, 27820, 3191, 90, 70, "ProofInfo",
 CellTags->{"Proof|ID:566590035", "Proof|ID:566590035-1"},
 CellID->88604489],
Cell[942343, 27912, 480, 12, 70, "ProofInfoBody",
 CellID->339162341]
}, Closed]],
Cell[CellGroupData[{
Cell[942860, 27929, 3950, 108, 70, "ProofInfo",
 CellTags->{"Proof|ID:262876021", "Proof|ID:262876021-1"},
 CellID->490373417],
Cell[946813, 28039, 480, 12, 70, "ProofInfoBody",
 CellID->399550044]
}, Closed]],
Cell[CellGroupData[{
Cell[947330, 28056, 3789, 104, 70, "ProofInfo",
 CellTags->{"Proof|ID:34443687", "Proof|ID:34443687-1"},
 CellID->161887690],
Cell[951122, 28162, 479, 12, 70, "ProofInfoBody",
 CellID->247568277]
}, Closed]],
Cell[951616, 28177, 47, 1, 70, "OpenEnvironment",
 CellID->541251984],
Cell[CellGroupData[{
Cell[951688, 28182, 378, 9, 70, "EnvironmentHeader",
 CellID->7001469],
Cell[952069, 28193, 1720, 52, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:134029858", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Nat \[SucceedsTilde] Int+"},
 CellID->134029858],
Cell[953792, 28247, 2965, 92, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:181242736", 
   "Source:/home/amaletzk/Theorema Notebooks/ElementaryTheories/Numbers.nb", 
   "Int+ induction"},
 CellID->181242736],
Cell[956760, 28341, 65, 1, 70, "EndEnvironmentMarker",
 CellID->331282669]
}, Open  ]],
Cell[956840, 28345, 47, 1, 70, "CloseEnvironment",
 CellID->16861641],
Cell[CellGroupData[{
Cell[956912, 28350, 4013, 108, 70, "ProofInfo",
 CellTags->{"Proof|ID:134029858", "Proof|ID:134029858-1"},
 CellID->379729521],
Cell[960928, 28460, 480, 12, 70, "ProofInfoBody",
 CellID->330204082]
}, Closed]],
Cell[CellGroupData[{
Cell[961445, 28477, 6055, 154, 70, "ProofInfo",
 CellTags->{"Proof|ID:181242736", "Proof|ID:181242736-1"},
 CellID->625873853],
Cell[967503, 28633, 480, 12, 70, "ProofInfoBody",
 CellID->165801038]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
