(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 9.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[   1887349,      52164]
NotebookOptionsPosition[   1423177,      39606]
NotebookOutlinePosition[   1748287,      48001]
CellTagsIndexPosition[   1714370,      47149]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["Reduction Polynomials Represented as Tuples of Monomials", "Title",
 CellChangeTimes->{{3.645759346310628*^9, 3.645759357126149*^9}},
 CellID->526162817],

Cell[CellGroupData[{

Cell["Preliminaries", "Section",
 CellChangeTimes->{{3.645759363233799*^9, 3.645759365981883*^9}},
 CellID->524135902],

Cell[CellGroupData[{

Cell["Description", "Subsection",
 CellChangeTimes->{{3.6457593699906588`*^9, 3.645759370840067*^9}, {
  3.6609066132255974`*^9, 3.6609066151288004`*^9}},
 CellID->1085100381],

Cell[TextData[{
 "This notebook contains the definition of functors ",
 Cell[BoxData[
  FormBox[
   RowBox[{"PPTuples", "[", "n", "]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"PolyTuples", "[", 
    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
   TraditionalForm]]],
 " that construct the domain of ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 "-variate power-poducts represented as tuples of natural numbers and the \
domain of reduction-polynomials over coefficient domain \[ScriptCapitalR] and \
power-product domain \[ScriptCapitalT], represented as ordered tuples of \
monomials, respectively. In addition, a functor for extending an arbitrary \
unordered power-product domain by a term order, called ",
 Cell[BoxData[
  FormBox[
   RowBox[{"OrderedPP", "[", 
    RowBox[{"\[ScriptCapitalT]", ",", "R"}], "]"}], TraditionalForm]]],
 ", as well as the three concrete term orders \[OpenCurlyDoubleQuote]Lex\
\[CloseCurlyDoubleQuote], \[OpenCurlyDoubleQuote]DegLex\
\[CloseCurlyDoubleQuote] and \[OpenCurlyDoubleQuote]DegRevLex\
\[CloseCurlyDoubleQuote], are also provided.\nThe core part of this notebook \
is the proof that, if \[ScriptCapitalR] is an algorithmic reduction ring and \
\[ScriptCapitalT] is an ordered power-product- and Dickson domain, the domain \
generated by ",
 Cell[BoxData[
  FormBox[
   RowBox[{"PPTuples", "[", 
    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
   TraditionalForm]]],
 " belongs to the class of algorithmic reduction polynomial domains (as \
defined in \[OpenCurlyDoubleQuote]Polynomials.nb\[CloseCurlyDoubleQuote]) and \
thus, according to the conservation theorem proved in \
\[OpenCurlyDoubleQuote]Polynomials.nb\[CloseCurlyDoubleQuote], also to the \
class of algorithmic reduction rings.\nOf course, the proofs of the claims \
that \[OpenCurlyDoubleQuote]Lex\[CloseCurlyDoubleQuote], \
\[OpenCurlyDoubleQuote]DegLex\[CloseCurlyDoubleQuote] and \
\[OpenCurlyDoubleQuote]DegRevLex\[CloseCurlyDoubleQuote], as defined here, \
are term orders, and that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"PPTuples", "[", "n", "]"}], TraditionalForm]]],
 " constructs power-product domains, can be found in this notebook, too."
}], "Text",
 CellChangeTimes->{{3.660906696155343*^9, 3.6609068437316017`*^9}, {
  3.6609069032301064`*^9, 3.660907122784892*^9}, {3.660907170162175*^9, 
  3.6609071743117824`*^9}, {3.6609072230306683`*^9, 3.6609072261194735`*^9}, {
  3.660907274713559*^9, 3.6609073507324924`*^9}},
 CellID->700836624]
}, Closed]],

Cell[CellGroupData[{

Cell["Theory Packages", "Subsection",
 CellChangeTimes->{{3.6554666129697094`*^9, 3.655466615543714*^9}},
 CellID->265926963],

Cell[BoxData[
 RowBox[{"$PackageDirectory", "=", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{
    "$RootDirectory", ",", "\"\<scratch\>\"", ",", "\"\<amaletzk\>\"", ",", 
     "\"\<Theorema Environment Data\>\""}], "}"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.666438783233775*^9, 3.666438821083324*^9}},
 CellID->498486718],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<Polynomials.m\>\""}], "}"}], "]"}],
   "]"}]], "Input",
 CellChangeTimes->{{3.6554666293029385`*^9, 3.655466662187796*^9}, 
   3.666440740699753*^9},
 CellID->36134333],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<ReductionRings.m\>\""}], "}"}], 
   "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.6554666293029385`*^9, 3.6554666705962105`*^9}, 
   3.666440741723823*^9},
 CellID->4041358],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<LogicSets.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554666293029385`*^9, 3.6554666783650246`*^9}, 
   3.666440742603622*^9},
 CellID->200738036],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<AlgebraicStructures.m\>\""}], 
    "}"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.6554666293029385`*^9, 3.6554666877406406`*^9}, 
   3.666440743675661*^9},
 CellID->582390848],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<NatInt.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554666293029385`*^9, 3.655466692186649*^9}, 
   3.666440744563591*^9},
 CellID->194899501],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<Tuples.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554666293029385`*^9, 3.655466662187796*^9}, {
   3.6554666950570536`*^9, 3.6554666973814583`*^9}, 3.66644074552351*^9},
 CellID->713552420],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<Sequences.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554666293029385`*^9, 3.655466662187796*^9}, {
   3.6554667016246653`*^9, 3.655466703059868*^9}, 3.666440746507523*^9},
 CellID->233536255],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", 
  RowBox[{"FileNameJoin", "[", 
   RowBox[{"{", 
    RowBox[{"$PackageDirectory", ",", "\"\<Functors.m\>\""}], "}"}], "]"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.6554666293029385`*^9, 3.655466662187796*^9}, {
   3.6554667058990726`*^9, 3.655466707084675*^9}, 3.666440747331586*^9},
 CellID->14602192],

Cell[BoxData[
 RowBox[{"LoadTheoryPackage", "[", "]"}]], "Input",
 CellChangeTimes->{{3.655466620613723*^9, 3.6554666250753307`*^9}},
 CellID->2803571],

Cell[BoxData[
 RowBox[{"MakeTheoryPackage", "[", "]"}]], "Input",
 CellChangeTimes->{{3.6554666293029385`*^9, 3.655466662187796*^9}, {
  3.655466711000282*^9, 3.6554667142762876`*^9}},
 CellID->31571396],

Cell[BoxData[
 RowBox[{"CompressProofData", "[", "]"}]], "Input",
 CellChangeTimes->{{3.6555272930250034`*^9, 3.655527297393011*^9}},
 CellID->794947268],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DeleteCases", "[", 
  RowBox[{
   RowBox[{"Map", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Quiet", "[", 
       RowBox[{"Check", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"With", "[", 
           RowBox[{
            RowBox[{"{", 
             RowBox[{"s", "=", 
              RowBox[{"\"\<VAR$\>\"", "<>", "#", "<>", "\"\<$TM\>\""}]}], 
             "}"}], ",", 
            RowBox[{"Context", "[", "s", "]"}]}], "]"}], ";", "#"}], ",", 
         "Null"}], "]"}], "]"}], "&"}], ",", 
     RowBox[{"{", 
      RowBox[{
      "\"\<i\>\"", ",", "\"\<j\>\"", ",", "\"\<n\>\"", ",", "\"\<m\>\"", ",", 
       "\"\<P\>\"", ",", "\"\<x\>\"", ",", "\"\<y\>\"", ",", "\"\<z\>\"", 
       ",", "\"\<a\>\"", ",", "\"\<A\>\"", ",", "\"\<b\>\"", ",", 
       "\"\<c\>\""}], "}"}]}], "]"}], ",", "Null"}], "]"}]], "Input",
 CellChangeTimes->{{3.6592618519471025`*^9, 3.6592618631011224`*^9}, {
   3.659681175379635*^9, 3.659681185706853*^9}, 3.659683710620489*^9, 
   3.65968441725893*^9, 3.6597695015379353`*^9, {3.6597698034056654`*^9, 
   3.659769808569275*^9}, {3.6597702854721127`*^9, 3.659770290573322*^9}, {
   3.659848272310069*^9, 3.6598483467534*^9}, {3.6598515848824887`*^9, 
   3.6598516763142495`*^9}, {3.6598555136659927`*^9, 3.659855522090008*^9}, {
   3.6598559051642804`*^9, 3.659855919516306*^9}, {3.6602862205309095`*^9, 
   3.6602862222781124`*^9}, {3.660479643354128*^9, 3.660479646380533*^9}},
 CellID->20666029],

Cell[BoxData[
 RowBox[{"{", "}"}]], "Output",
 CellChangeTimes->{3.661584318664313*^9, 3.662288533094919*^9, 
  3.666176997199605*^9},
 CellID->44505675]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Theory Analysis", "Subsection",
 CellChangeTimes->{{3.640668830437483*^9, 3.640668834318727*^9}},
 CellID->1386333768],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"AnalyzeProofs", "[", "]"}], "//", "Timing"}]], "Input",
 CellChangeTimes->{{3.6406688560104437`*^9, 3.640668867513785*^9}, 
   3.64067664060292*^9, 3.644222019629964*^9, {3.6569157648931513`*^9, 
   3.656915767966357*^9}},
 CellID->1039312110],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading directory \
\\\"C:\\\\Users\\\\Alex\\\\Documents\\\\Doktorat\\\\Theorema \
Notebooks\\\\Reduction Ring Theory\\\\PolyTuples\\\" (\"\>", 
   "\[InvisibleSpace]", "1", "\[InvisibleSpace]", "\<\" of \"\>", 
   "\[InvisibleSpace]", "\<\"1\"\>", "\[InvisibleSpace]", "\<\"):\"\>"}],
  SequenceForm[
  "Reading directory \"C:\\Users\\Alex\\Documents\\Doktorat\\Theorema \
Notebooks\\Reduction Ring Theory\\PolyTuples\" (", 1, " of ", "1", "):"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAOKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtkkVSGe2f+249/8pdWMgDQCP7pCe
  "],
 CellID->332409596],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"163\"\>"}],
  SequenceForm["Reading proof ", 1, " of ", "163"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAOKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtkkVSGe2f+3YUX9a3RhIAwCKzI/q
  "],
 CellID->149409841],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"163\"\>"}],
  SequenceForm["Reading proof ", 51, " of ", "163"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAOKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtkkVSGe2f+244YKthjGQBgCK9I/W
  "],
 CellID->172974039],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"163\"\>"}],
  SequenceForm["Reading proof ", 101, " of ", "163"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAOKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtkkVSGe2f+34TnWZhjGQBgCKrY/S
  "],
 CellID->49653511],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Reading proof \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "\<\"163\"\>"}],
  SequenceForm["Reading proof ", 151, " of ", "163"],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAOKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtkkVSGe2f+246buYpjGQBgCLMI/Z
  "],
 CellID->473922720],

Cell[BoxData["\<\"Creating forward rules:\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAOKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtkkVSGe2f+0okqenaQykAYMRjso=
  "],
 CellID->257520971],

Cell[BoxData["\<\"Preparing ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAOKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtkkVSGe2f+24Jldf0xhIAwCH0I9i
  "],
 CellID->159051417],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "449"}],
  SequenceForm["Creating rule ", 1, " of ", 449],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAOKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtkkVSGe2f+3I99ZA0xhIAwCGZo9F
  "],
 CellID->71026772],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "51", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "449"}],
  SequenceForm["Creating rule ", 51, " of ", 449],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAOKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtkkVSGe2f+24INtE0xhIAwCHgI9Z
  "],
 CellID->38953335],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "101", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "449"}],
  SequenceForm["Creating rule ", 101, " of ", 449],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAOKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtkkVSGe2f+1o9Mpc0xhIAwCHm49t
  "],
 CellID->545021598],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "151", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "449"}],
  SequenceForm["Creating rule ", 151, " of ", 449],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAOKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtkkVSGe2f+3Y98Ja0xhIAwCKhY/L
  "],
 CellID->206823069],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "201", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "449"}],
  SequenceForm["Creating rule ", 201, " of ", 449],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAOKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtkkVSGe2f+2okG6vaQykAYOmjuA=
  "],
 CellID->165393538],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "251", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "449"}],
  SequenceForm["Creating rule ", 251, " of ", 449],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAOKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtkkVSGe2B7r/qZOmMZAGAIuyj/I=
  "],
 CellID->4978992],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "301", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "449"}],
  SequenceForm["Creating rule ", 301, " of ", 449],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAOKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtkkVSGe2f+3I98RN0xhIAwCGq49S
  "],
 CellID->799672086],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "351", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "449"}],
  SequenceForm["Creating rule ", 351, " of ", 449],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAOKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtkkVSGe2f+24IMlL0xhIAwCHxY9m
  "],
 CellID->512697675],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Creating rule \"\>", "\[InvisibleSpace]", "401", 
   "\[InvisibleSpace]", "\<\" of \"\>", "\[InvisibleSpace]", "449"}],
  SequenceForm["Creating rule ", 401, " of ", 449],
  Editable->False]], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAOKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtkkVSGe2f+1o+NBX0xhIAwCH2I95
  "],
 CellID->109571920],

Cell[BoxData["\<\"Creating thinned data ...\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAOKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtkkVSGe2f+14KD5Q0xhIAwCI6o+M
  "],
 CellID->102929758],

Cell[BoxData["\<\"Done\"\>"], "Print",
 CellChangeTimes->CompressedData["
1:eJxTTMoPSmVkYGDQAOKTa2Uj2RReO15msVThAtKvylwegGjzgkD3XUC6TOLR
hWtAelVT050HQHraYe5+PsXXjr47zoubAOkDaR0nTYH0lM0LtM2AdHP2b5+b
IL5616s7QDpL8n/yP6XXjqc2uj1mUn7tuNxg9wRWIL1x2uViNiA9rcJlMzuQ
lvEXeKAKpJfpLpVVB9ITdFbO1wbSXywO9+sA6aajoTa6QDqNfc4WAyA9p2ey
nCGQ/hPx1K8RSNux7/zLovLacadz2yV1IH1Q+s0dEC1l2mCmqPrakSdt/g9P
IN0wMb4wEEjXFTscrQHSedOqi8+qvXac9T6bo0YdKB8oHq+o8doxSk+aUV7z
tWNYwo8uWyB9Y4G1pRuQtkkVSGe2f+3Ydz9Q0xhIAwCKyo/Y
  "],
 CellID->712415489]
}, Open  ]],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "4.64882979999999967901658237678930163383`6.687943559640405", ",", "Null"}],
   "}"}]], "Output",
 CellChangeTimes->{3.656915782677183*^9, 3.6570135785741396`*^9},
 CellID->143333797]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"ComputeFormulaPositions", "[", "]"}], ";"}]], "Input",
 CellChangeTimes->{{3.6406695270485573`*^9, 3.6406695415951242`*^9}, {
   3.64067664934828*^9, 3.640676663567005*^9}, 3.6412902845300865`*^9, 
   3.641290346822996*^9, {3.6442219079512787`*^9, 3.644221909273131*^9}},
 CellID->19428626],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CheckIntegrity", "[", 
  RowBox[{"\"\<CheckPositions\>\"", "\[Rule]", 
   RowBox[{"{", 
    RowBox[{"FileNameJoin", "[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"CurrentValue", "[", "\"\<NotebookDirectory\>\"", "]"}], ",", 
       RowBox[{"CurrentValue", "[", "\"\<NotebookFileName\>\"", "]"}]}], 
      "}"}], "]"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.64067003109768*^9, 3.6406700374311132`*^9}, {
   3.6406797903560762`*^9, 3.640679792950843*^9}, 3.6442220291527653`*^9},
 CellID->1909803754],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{
  3.640670040858255*^9, 3.640670099679185*^9, 3.640670397721546*^9, 
   3.640679793300535*^9, 3.6406798406436234`*^9, 3.640679891162312*^9, 
   3.640679971009297*^9, 3.64068009182163*^9, 3.640680173340745*^9, 
   3.641037395456129*^9, 3.641116403388618*^9, 3.641286881014108*^9, 
   3.642225466578094*^9, 3.642234672070362*^9, 3.642240151894621*^9, 
   3.6423063269273643`*^9, {3.642308832539488*^9, 3.642308857563932*^9}, 
   3.642321396697977*^9, 3.6423285030212173`*^9, 3.642493582768566*^9, 
   3.642763270724422*^9, 3.642834604308413*^9, 3.643428832901271*^9, 
   3.6434441292624083`*^9, 3.643532851309215*^9, 3.644918543299676*^9, 
   3.645518304732339*^9},
 CellID->636966437]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Auxiliary Functions", "Subsection",
 CellChangeTimes->{{3.640668579267561*^9, 3.640668583347287*^9}},
 CellID->1454182661],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "LoadPackage", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"LoadPackage", "[", "name_String", "]"}], ":=", 
   RowBox[{"LoadPackage", "[", 
    RowBox[{"{", "name", "}"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"LoadPackage", "[", 
    RowBox[{"l", ":", 
     RowBox[{"{", "___String", "}"}]}], "]"}], ":=", 
   RowBox[{"Scan", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Get", "[", 
       RowBox[{"FileNameJoin", "[", 
        RowBox[{"{", 
         RowBox[{"$HomeDirectory", ",", "\"\<Theorema Notebooks\>\"", ",", 
          RowBox[{"#", "<>", "\"\<.m\>\""}]}], "}"}], "]"}], "]"}], "&"}], 
     ",", "l"}], "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.640668603629292*^9, 3.640668727460284*^9}, 
   3.641286816086794*^9, 3.6434459961973133`*^9, {3.6555260761694636`*^9, 
   3.6555260852486796`*^9}, {3.66644075329471*^9, 3.666440759502613*^9}},
 CellID->2074904605],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "FindFormula", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"FindFormula", "[", "label_String", "]"}], ":=", 
   RowBox[{"NotebookLocate", "[", "label", "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.640677756847624*^9, 3.6406777896243896`*^9}, 
   3.640677822869519*^9, {3.640677854902379*^9, 3.6406778579321823`*^9}},
 CellID->1980565193],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "ActivateKnowledge", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"ActivateKnowledge", "[", 
    RowBox[{
     RowBox[{"labels", ":", 
      RowBox[{"{", "___String", "}"}]}], ",", "value_"}], "]"}], ":=", 
   "\[IndentingNewLine]", 
   RowBox[{"Scan", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"ActivateKnowledge", "[", 
       RowBox[{"#", ",", "value"}], "]"}], "&"}], ",", "labels"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"ActivateKnowledge", "[", 
    RowBox[{"label_String", ",", "value_"}], "]"}], ":=", 
   "\[IndentingNewLine]", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"key", "=", 
       RowBox[{"Cases", "[", 
        RowBox[{"Theorema`Common`$tmaEnv", ",", 
         RowBox[{
          RowBox[{"_", "[", 
           RowBox[{"k_", ",", "_", ",", "label", ",", "___"}], "]"}], 
          "\[RuleDelayed]", "k"}]}], "]"}]}], "}"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"Switch", "[", 
      RowBox[{"key", ",", "\[IndentingNewLine]", 
       RowBox[{"{", "}"}], ",", "\[IndentingNewLine]", 
       "\"\<Formula not found\>\"", ",", "\[IndentingNewLine]", 
       RowBox[{"{", "_", "}"}], ",", "\[IndentingNewLine]", 
       RowBox[{"With", "[", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{"k", "=", 
           RowBox[{"First", "[", "key", "]"}]}], "}"}], ",", 
         "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{
           RowBox[{"Theorema`Common`kbSelectProve", "[", "k", "]"}], "=", 
           "value"}], ";"}]}], "\[IndentingNewLine]", "]"}], ",", 
       "\[IndentingNewLine]", "_", ",", "\[IndentingNewLine]", 
       "\"\<Label not unique\>\""}], "\[IndentingNewLine]", "]"}]}], 
    "\[IndentingNewLine]", "]"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.6560489370706587`*^9, 3.65604906870469*^9}, {
   3.656049121666783*^9, 3.6560491231331854`*^9}, {3.656049157562446*^9, 
   3.6560491794804845`*^9}, {3.656049671349348*^9, 3.6560497147174244`*^9}, {
   3.656049781610342*^9, 3.656049811999195*^9}, 3.6560501616290092`*^9},
 CellID->658527806],

Cell[BoxData[
 RowBox[{"LoadPackage", "[", "\"\<ProofPalette\>\"", "]"}]], "Input",
 CellChangeTimes->{{3.640668742006607*^9, 3.640668751417616*^9}},
 CellID->444074165],

Cell[BoxData[
 RowBox[{"LoadPackage", "[", "\"\<TheoryTools/TheoryTools\>\"", 
  "]"}]], "Input",
 CellChangeTimes->{{3.6555260963870993`*^9, 3.655526104436713*^9}, {
   3.6622883954558773`*^9, 3.662288398809883*^9}, 3.6664407647948303`*^9},
 CellID->69584241]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Polynomial Parsing and Pretty-Printing", "Section",
 CellChangeTimes->{{3.631446380731266*^9, 3.631446389513492*^9}},
 CellID->1575793246],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "MonomialTupleIn", "]"}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"MonomialTupleIn", "[", 
    RowBox[{"p_", ",", "R_", ",", "T_"}], "]"}], ":=", "\[IndentingNewLine]", 
   
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"one", "=", 
        RowBox[{"Theorema`Computation`Language`DomainOperation$TM", "[", 
         RowBox[{"R", ",", "1"}], "]"}]}], ",", 
       RowBox[{"minus", "=", 
        RowBox[{"Theorema`Computation`Language`DomainOperation$TM", "[", 
         RowBox[{"R", ",", "Theorema`Computation`Language`Minus$TM"}], 
         "]"}]}], ",", "vars", ",", "tmp", ",", "pnew"}], "}"}], ",", 
     "\[IndentingNewLine]", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"Begin", "[", "\"\<Theorema`Computation`Knowledge`\>\"", "]"}], 
      ";", "\[IndentingNewLine]", 
      RowBox[{"vars", "=", 
       RowBox[{"Map", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"ToExpression", "[", "#", "]"}], "&"}], ",", 
         RowBox[{"{", 
          RowBox[{"\"\<x$TM\>\"", ",", "\"\<y$TM\>\"", ",", "\"\<z$TM\>\""}], 
          "}"}]}], "]"}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"End", "[", "]"}], ";", "\[IndentingNewLine]", 
      "\[IndentingNewLine]", 
      RowBox[{"tmp", "=", 
       RowBox[{"Cases", "[", 
        RowBox[{"p", ",", 
         RowBox[{
          RowBox[{"Theorema`Computation`Language`Subscript$TM", "[", 
           RowBox[{
            RowBox[{"First", "[", "vars", "]"}], ",", 
            RowBox[{"i_Integer", "?", "Positive"}]}], "]"}], "\[RuleDelayed]",
           "i"}], ",", "Infinity"}], "]"}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{"tmp", "=!=", 
         RowBox[{"{", "}"}]}], ",", "\[IndentingNewLine]", 
        RowBox[{"vars", "=", 
         RowBox[{"Table", "[", 
          RowBox[{
           RowBox[{"Theorema`Computation`Language`Subscript$TM", "[", 
            RowBox[{
             RowBox[{"First", "[", "vars", "]"}], ",", "i"}], "]"}], ",", 
           RowBox[{"{", 
            RowBox[{"i", ",", 
             RowBox[{"Max", "@@", "tmp"}]}], "}"}]}], "]"}]}], ",", 
        "\[IndentingNewLine]", "\[IndentingNewLine]", 
        RowBox[{"While", "[", 
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{"Length", "[", "vars", "]"}], ">", "1"}], "&&", 
           RowBox[{"FreeQ", "[", 
            RowBox[{"p", ",", 
             RowBox[{"Last", "[", "vars", "]"}]}], "]"}]}], ",", 
          "\[IndentingNewLine]", 
          RowBox[{"vars", "=", 
           RowBox[{"Most", "[", "vars", "]"}]}]}], "\[IndentingNewLine]", 
         "]"}]}], "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"pnew", "=", 
       RowBox[{"p", "/.", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{
          "Theorema`Computation`Language`Plus$TM", "\[Rule]", "Plus"}], ",", 
          RowBox[{
          "Theorema`Computation`Language`Subtract$TM", "\[Rule]", 
           "Subtract"}], ",", 
          RowBox[{
          "Theorema`Computation`Language`Minus$TM", "\[Rule]", "Minus"}], ",", 
          RowBox[{
          "Theorema`Computation`Language`Times$TM", "\[Rule]", "Times"}], ",", 
          RowBox[{
          "Theorema`Computation`Language`Power$TM", "\[Rule]", "Power"}]}], 
         "}"}]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"Apply", "[", 
       RowBox[{"Theorema`Computation`Language`Tuple$TM", ",", 
        RowBox[{
         RowBox[{"CoefficientRules", "[", 
          RowBox[{"pnew", ",", "vars"}], "]"}], "/.", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{
            RowBox[{"Rule", "[", 
             RowBox[{
              RowBox[{"{", "exp___", "}"}], ",", "1"}], "]"}], 
            "\[RuleDelayed]", 
            RowBox[{"Theorema`Computation`Language`Tuple$TM", "[", 
             RowBox[{"one", ",", 
              RowBox[{"ParseExpTuple", "[", 
               RowBox[{"exp", ",", "T"}], "]"}]}], "]"}]}], ",", 
           "\[IndentingNewLine]", 
           RowBox[{
            RowBox[{"Rule", "[", 
             RowBox[{
              RowBox[{"{", "exp___", "}"}], ",", 
              RowBox[{"-", "1"}]}], "]"}], "\[RuleDelayed]", 
            RowBox[{"Theorema`Computation`Language`Tuple$TM", "[", 
             RowBox[{
              RowBox[{"minus", "[", "one", "]"}], ",", 
              RowBox[{"ParseExpTuple", "[", 
               RowBox[{"exp", ",", "T"}], "]"}]}], "]"}]}], ",", 
           "\[IndentingNewLine]", 
           RowBox[{
            RowBox[{"Rule", "[", 
             RowBox[{
              RowBox[{"{", "exp___", "}"}], ",", 
              RowBox[{"coeff_", "?", "Negative"}]}], "]"}], "\[RuleDelayed]", 
            
            RowBox[{"Theorema`Computation`Language`Tuple$TM", "[", 
             RowBox[{
              RowBox[{"minus", "[", 
               RowBox[{"-", "coeff"}], "]"}], ",", 
              RowBox[{"ParseExpTuple", "[", 
               RowBox[{"exp", ",", "T"}], "]"}]}], "]"}]}], ",", 
           RowBox[{
            RowBox[{"Rule", "[", 
             RowBox[{
              RowBox[{"{", "exp___", "}"}], ",", "coeff_"}], "]"}], 
            "\[RuleDelayed]", 
            RowBox[{"Theorema`Computation`Language`Tuple$TM", "[", 
             RowBox[{"coeff", ",", 
              RowBox[{"ParseExpTuple", "[", 
               RowBox[{"exp", ",", "T"}], "]"}]}], "]"}]}]}], "}"}]}]}], 
       "]"}]}]}], "\[IndentingNewLine]", "]"}]}], ";"}]}], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.595826414247673*^9, 3.595826501118219*^9}, {
   3.595826601073849*^9, 3.595826634440883*^9}, {3.595826692680787*^9, 
   3.595826701566459*^9}, {3.608007216400512*^9, 3.608007217195429*^9}, {
   3.608007357914502*^9, 3.608007359743864*^9}, {3.608007391726764*^9, 
   3.608007411973763*^9}, {3.608007643685813*^9, 3.608007646221676*^9}, {
   3.608007746486024*^9, 3.608007775598164*^9}, {3.60800785399566*^9, 
   3.608007864991464*^9}, {3.608008048553611*^9, 3.608008252785018*^9}, 
   3.608008378797148*^9, {3.608021787058804*^9, 3.608021792928964*^9}, {
   3.608022057214825*^9, 3.608022107140704*^9}, {3.608022153080122*^9, 
   3.608022161558264*^9}, {3.608022281980045*^9, 3.608022285257821*^9}, {
   3.608022456554984*^9, 3.608022456816393*^9}, {3.635059940518152*^9, 
   3.635059970722624*^9}, {3.635754188353105*^9, 3.635754191678447*^9}, {
   3.635754223611846*^9, 3.635754240758956*^9}, {3.635754325141889*^9, 
   3.635754548471017*^9}, {3.635754794838705*^9, 3.635754813656164*^9}, 
   3.635755517060069*^9, {3.6357560128053913`*^9, 3.635756014075346*^9}, {
   3.635758155548319*^9, 3.635758174547298*^9}, {3.6357582291515427`*^9, 
   3.635758229564262*^9}, {3.6357583698562403`*^9, 3.635758370078277*^9}, {
   3.645764384004578*^9, 3.645764515592821*^9}, {3.645778509280991*^9, 
   3.645778555820409*^9}},
 CellID->1559908265],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "ParseExpTuple", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"ParseExpTuple", "[", 
    RowBox[{"exp___", ",", "T_"}], "]"}], ":=", 
   RowBox[{
    RowBox[{"Theorema`Computation`Language`DomainOperation$TM", "[", 
     RowBox[{"T", ",", "Theorema`Computation`Language`Get$TM"}], "]"}], "[", 
    RowBox[{"Theorema`Computation`Language`Tuple$TM", "[", "exp", "]"}], 
    "]"}]}], ";"}]}], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.635754685918868*^9, 3.63575477075211*^9}, 
   3.635755514278469*^9},
 CellID->423723764],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "SortMonomials", "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"SortMonomials", "[", 
    RowBox[{"p_", ",", "rel_"}], "]"}], ":=", 
   RowBox[{"Sort", "[", 
    RowBox[{"p", ",", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"rel", "[", 
         RowBox[{
          RowBox[{"Last", "[", "#1", "]"}], ",", 
          RowBox[{"Last", "[", "#2", "]"}]}], "]"}], "=!=", "False"}], ")"}], 
      "&"}]}], "]"}]}], ";"}]}], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.635754685918868*^9, 3.63575477075211*^9}, 
   3.635755514278469*^9, {3.635756035700601*^9, 3.635756053567562*^9}, {
   3.635756092943556*^9, 3.635756149157618*^9}, 3.635756193323399*^9},
 CellID->142555754],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "MonomialTupleOut", "]"}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"MonomialTupleOut", "[", 
    RowBox[{
     RowBox[{"p", ":", 
      RowBox[{
      "Theorema`Computation`Language`Tuple$TM", "[", 
       "___Theorema`Computation`Language`Tuple$TM", "]"}]}], ",", "R_", ",", 
     "T_"}], "]"}], ":=", "\[IndentingNewLine]", 
   RowBox[{"With", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"f", "=", 
       RowBox[{"Theorema`Computation`Language`DomainOperation$TM", "[", 
        RowBox[{"R", ",", "Theorema`Computation`Language`Put$TM"}], "]"}]}], 
      "}"}], ",", "\[IndentingNewLine]", 
     RowBox[{"Module", "[", 
      RowBox[{
       RowBox[{"{", "c0", "}"}], ",", "\[IndentingNewLine]", 
       RowBox[{"Replace", "[", 
        RowBox[{"p", ",", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"h_", "[", 
           RowBox[{"c_", ",", "e_"}], "]"}], "\[RuleDelayed]", 
          RowBox[{"(", "\[IndentingNewLine]", 
           RowBox[{
            RowBox[{"c0", "=", 
             RowBox[{"f", "[", "c", "]"}]}], ";", "\[IndentingNewLine]", 
            RowBox[{"If", "[", 
             RowBox[{
              RowBox[{
               RowBox[{"Head", "[", "c0", "]"}], "===", "f"}], ",", 
              RowBox[{"c0", "=", "c"}]}], "]"}], ";", "\[IndentingNewLine]", 
            RowBox[{"h", "[", 
             RowBox[{"c0", ",", 
              RowBox[{
               RowBox[{
               "Theorema`Computation`Language`DomainOperation$TM", "[", 
                RowBox[{"T", ",", "Theorema`Computation`Language`Put$TM"}], 
                "]"}], "[", "e", "]"}]}], "]"}]}], "\[IndentingNewLine]", 
           ")"}]}], ",", "\[IndentingNewLine]", 
         RowBox[{"{", "1", "}"}]}], "\[IndentingNewLine]", "]"}]}], 
      "\[IndentingNewLine]", "]"}]}], "\[IndentingNewLine]", "]"}]}], 
  ";"}]}], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.595826414247673*^9, 3.595826501118219*^9}, {
   3.595826601073849*^9, 3.595826634440883*^9}, {3.595826692680787*^9, 
   3.595826701566459*^9}, {3.608007216400512*^9, 3.608007217195429*^9}, {
   3.608007357914502*^9, 3.608007359743864*^9}, {3.608007391726764*^9, 
   3.608007411973763*^9}, {3.608007643685813*^9, 3.608007646221676*^9}, {
   3.608007746486024*^9, 3.608007775598164*^9}, {3.60800785399566*^9, 
   3.608007864991464*^9}, {3.608008048553611*^9, 3.608008252785018*^9}, 
   3.608008378797148*^9, {3.608021787058804*^9, 3.608021792928964*^9}, {
   3.608022057214825*^9, 3.608022107140704*^9}, {3.608022153080122*^9, 
   3.608022161558264*^9}, {3.608022281980045*^9, 3.608022285257821*^9}, {
   3.608022456554984*^9, 3.608022456816393*^9}, {3.635059940518152*^9, 
   3.635059970722624*^9}, {3.635754188353105*^9, 3.635754191678447*^9}, {
   3.635754223611846*^9, 3.635754240758956*^9}, {3.635754325141889*^9, 
   3.635754548471017*^9}, {3.635754794838705*^9, 3.635754813656164*^9}, {
   3.635754968131425*^9, 3.6357549718627653`*^9}, 3.63575551087724*^9, {
   3.635755588057486*^9, 3.635755713061338*^9}, {3.6357557479413643`*^9, 
   3.63575593326855*^9}, {3.63575601641434*^9, 3.635756018088565*^9}, {
   3.645764922395565*^9, 3.64576493632953*^9}, 3.6457650349122133`*^9, {
   3.645765072238769*^9, 3.6457650844759817`*^9}},
 CellID->751753410],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", "PrintPoly", "]"}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintPoly", "[", 
    RowBox[{"Theorema`Computation`Language`Tuple$TM", "[", "]"}], "]"}], ":=",
    "0"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintPoly", "[", 
    RowBox[{"p", ":", 
     RowBox[{"Theorema`Computation`Language`Tuple$TM", "[", 
      RowBox[{
       RowBox[{"Theorema`Computation`Language`Tuple$TM", "[", 
        RowBox[{"_", ",", "exp_Theorema`Computation`Language`Tuple$TM"}], 
        "]"}], ",", "___Theorema`Computation`Language`Tuple$TM"}], "]"}]}], 
    "]"}], ":=", "\[IndentingNewLine]", 
   RowBox[{"With", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"vars", "=", 
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"Length", "[", "exp", "]"}], "\[LessEqual]", "3"}], ",", 
         "\[IndentingNewLine]", 
         RowBox[{"Take", "[", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{
            "Theorema`Computation`Knowledge`x$TM", ",", 
             "Theorema`Computation`Knowledge`y$TM", ",", 
             "Theorema`Computation`Knowledge`z$TM"}], "}"}], ",", 
           RowBox[{"Length", "[", "exp", "]"}]}], "]"}], ",", 
         "\[IndentingNewLine]", 
         RowBox[{"Table", "[", 
          RowBox[{
           RowBox[{"Theorema`Computation`Language`Subscript$TM", "[", 
            RowBox[{"Theorema`Computation`Knowledge`x$TM", ",", "i"}], "]"}], 
           ",", 
           RowBox[{"{", 
            RowBox[{"i", ",", 
             RowBox[{"Length", "[", "exp", "]"}]}], "}"}]}], "]"}]}], 
        "\[IndentingNewLine]", "]"}]}], "}"}], ",", "\[IndentingNewLine]", 
     RowBox[{"FromCoefficientRules", "[", 
      RowBox[{
       RowBox[{"Replace", "[", 
        RowBox[{
         RowBox[{"List", "@@", "p"}], ",", 
         RowBox[{
          RowBox[{"_", "[", 
           RowBox[{"c_", ",", "e_"}], "]"}], "\[RuleDelayed]", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"List", "@@", "e"}], "\[Rule]", "c"}], ")"}]}], ",", 
         RowBox[{"{", "1", "}"}]}], "]"}], ",", "vars"}], "]"}]}], 
    "\[IndentingNewLine]", "]"}]}], ";"}]}], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.595826414247673*^9, 3.595826501118219*^9}, {
   3.595826601073849*^9, 3.595826634440883*^9}, {3.595826692680787*^9, 
   3.595826701566459*^9}, {3.608007216400512*^9, 3.608007217195429*^9}, {
   3.608007357914502*^9, 3.608007359743864*^9}, {3.608007391726764*^9, 
   3.608007411973763*^9}, {3.608007643685813*^9, 3.608007646221676*^9}, {
   3.608007746486024*^9, 3.608007775598164*^9}, {3.60800785399566*^9, 
   3.608007864991464*^9}, {3.608008048553611*^9, 3.608008252785018*^9}, 
   3.608008378797148*^9, {3.608021787058804*^9, 3.608021792928964*^9}, {
   3.608022057214825*^9, 3.608022107140704*^9}, {3.608022153080122*^9, 
   3.608022161558264*^9}, {3.608022281980045*^9, 3.608022285257821*^9}, {
   3.608022456554984*^9, 3.608022456816393*^9}, {3.635059940518152*^9, 
   3.635059970722624*^9}, {3.635754188353105*^9, 3.635754191678447*^9}, {
   3.635754223611846*^9, 3.635754240758956*^9}, {3.635754325141889*^9, 
   3.635754548471017*^9}, {3.635754794838705*^9, 3.635754813656164*^9}, {
   3.63575501224885*^9, 3.63575502267082*^9}, {3.635755070759357*^9, 
   3.635755098395145*^9}, {3.635755129683401*^9, 3.6357554733794403`*^9}, {
   3.635755509106204*^9, 3.6357555349088163`*^9}},
 CellID->1199342921]
}, Closed]],

Cell[CellGroupData[{

Cell["Abstract Power-Products (\[Checkmark])", "Section",
 CellChangeTimes->{{3.645759414649465*^9, 3.6457594165192013`*^9}, {
  3.6568442588393784`*^9, 3.6568442700401983`*^9}, {3.656929888913473*^9, 
  3.6569298931098804`*^9}},
 CellID->1600179867],

Cell[CellGroupData[{

Cell["\<\
Abstract Power-Products as Exponent Functions and Term Orders (\[Checkmark])\
\>", "Subsection",
 CellChangeTimes->{{3.656821141929216*^9, 3.656821157919244*^9}, {
   3.656821213581142*^9, 3.656821219914753*^9}, {3.6568212804272594`*^9, 
   3.6568212818156614`*^9}, 3.6568216852375703`*^9, {3.6568233337194676`*^9, 
   3.6568233733747373`*^9}, {3.656841930234888*^9, 3.6568419390333033`*^9}, {
   3.6568442045824833`*^9, 3.65684420888809*^9}, {3.6568442794314146`*^9, 
   3.6568442811630173`*^9}},
 CellID->579563147],

Cell[CellGroupData[{

Cell["Definitions", "Subsubsection",
 CellChangeTimes->{{3.656821168043662*^9, 3.656821169854265*^9}, 
   3.6568216852375703`*^9},
 CellID->79894345],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", "\[ScriptCapitalT]"]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", 303368290}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568212558416157`*^9, 3.6568212580880203`*^9}, 
   3.6568216852375703`*^9},
 CellTags->{
  "ID:303368290", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->303368290],

Cell["", "OpenEnvironment",
 CellChangeTimes->{3.6568216852375703`*^9},
 CellID->630672694],

Cell[CellGroupData[{

Cell["\<\
Definition: Class of (Unordered) Power-Products as Exponent Functions\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656821191865904*^9, 3.656821192021904*^9}, {
   3.6568212268411655`*^9, 3.6568212307723722`*^9}, 3.6568216852375703`*^9, {
   3.6568232893841896`*^9, 3.656823366432725*^9}},
 CellID->264955074],

Cell[BoxData[
 RowBox[{
  RowBox[{"isExpPPDomain", "[", "\[ScriptCapitalT]", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["dim", "\[ScriptCapitalT]"], "\[Element]", 
         SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]},
       {
        RowBox[{"isDomain", "[", "\[ScriptCapitalT]", "]"}]},
       {
        RowBox[{"hasDecisionPredicatePP", "[", "\[ScriptCapitalT]", "]"}]},
       {
        RowBox[{"hasExponentFunctionPP", "[", "\[ScriptCapitalT]", "]"}]},
       {
        RowBox[{"isExtensionalPP", "[", "\[ScriptCapitalT]", "]"}]},
       {
        RowBox[{"hasDegreeFunctionPP", "[", "\[ScriptCapitalT]", "]"}]},
       {
        RowBox[{"hasDivisibilityPP", "[", "\[ScriptCapitalT]", "]"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:603497554", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656821262721228*^9, 3.65682128640207*^9}, {
   3.6568213459171743`*^9, 3.6568214241055117`*^9}, {3.6568214594551735`*^9, 
   3.6568216311990757`*^9}, 3.6568216852375703`*^9, {3.6568218934833364`*^9, 
   3.656821907570161*^9}, {3.656822243969152*^9, 3.6568222488363605`*^9}, {
   3.6568233849343576`*^9, 3.6568234126244063`*^9}, {3.656823645409015*^9, 
   3.656823653318229*^9}, {3.6568239956608305`*^9, 3.6568240011676397`*^9}, {
   3.6568261125329494`*^9, 3.656826124513771*^9}},
 CellTags->{
  "ID:603497554", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain"},
 CellID->603497554],

Cell[BoxData[
 RowBox[{
  RowBox[{"hasDecisionPredicatePP", "[", "\[ScriptCapitalT]", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
         RowBox[{"[", 
          UnderscriptBox["1", "\[ScriptCapitalT]"], "]"}]}]},
       {
        RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
           RowBox[{"[", 
            RowBox[{"s", ",", "t"}], "]"}]}]], 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"\[And]", 
           RowBox[{"\[Piecewise]", GridBox[{
              {
               RowBox[{
                UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
                RowBox[{"[", 
                 RowBox[{"s", 
                  UnderscriptBox["*", "\[ScriptCapitalT]"], "t"}], "]"}]}]},
              {
               RowBox[{
                UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
                RowBox[{"[", 
                 RowBox[{
                  UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                  RowBox[{"s", ",", "t"}], "]"}], "]"}]}]},
              {
               RowBox[{
                RowBox[{"t", 
                 UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "s"}],
                 "\[Implies]", 
                RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
                 RowBox[{"[", 
                  RowBox[{"s", 
                   UnderscriptBox["/", "\[ScriptCapitalT]"], "t"}], 
                  "]"}]}]}]}
             },
             GridBoxAlignment->{
              "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
          TagBox[")",
           "AutoParentheses"]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(hasDecisionPredicatePP)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:49777475", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568217131928196`*^9, 3.6568217383868637`*^9}, {
  3.656821794095562*^9, 3.6568218623144817`*^9}, {3.656822958584609*^9, 
  3.6568229879282603`*^9}},
 CellTags->{
  "ID:49777475", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasDecisionPredicatePP"},
 CellID->49777475],

Cell[BoxData[
 RowBox[{
  RowBox[{"hasExponentFunctionPP", "[", "\[ScriptCapitalT]", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      UnderscriptBox["dim", "\[ScriptCapitalT]"]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           RowBox[{
            UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
            RowBox[{
             UnderscriptBox["1", "\[ScriptCapitalT]"], ",", "i"}], "]"}], 
           "\[Equal]", "0"}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
             RowBox[{"[", "t", "]"}]}]], 
           RowBox[{
            RowBox[{
             UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
             RowBox[{"t", ",", "i"}], "]"}], "\[Element]", 
            SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
             RowBox[{"[", 
              RowBox[{"s", ",", "t"}], "]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{"\[And]", 
             RowBox[{"\[Piecewise]", GridBox[{
                {
                 RowBox[{
                  RowBox[{
                   UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
                   RowBox[{
                    RowBox[{"s", 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], "t"}], ",", 
                    "i"}], "]"}], "\[Equal]", 
                  RowBox[{
                   RowBox[{
                    UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
                    RowBox[{"s", ",", "i"}], "]"}], "+", 
                   RowBox[{
                    UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
                    RowBox[{"t", ",", "i"}], "]"}]}]}]},
                {
                 RowBox[{
                  RowBox[{
                   UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{"s", ",", "t"}], "]"}], ",", "i"}], "]"}], 
                  "\[Equal]", 
                  RowBox[{
                   SubscriptBox["max", "\[LessEqual]"], "[", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
                    RowBox[{"s", ",", "i"}], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
                    RowBox[{"t", ",", "i"}], "]"}]}], 
                    "\[RightAngleBracket]"}], "]"}]}]},
                {
                 RowBox[{
                  RowBox[{"t", 
                   UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                   "s"}], "\[Implies]", 
                  RowBox[{
                   TagBox["(",
                    "AutoParentheses"], 
                   RowBox[{
                    RowBox[{
                    UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{"s", 
                    UnderscriptBox["/", "\[ScriptCapitalT]"], "t"}], ",", 
                    "i"}], "]"}], "\[Equal]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
                    RowBox[{"s", ",", "i"}], "]"}], "-", 
                    RowBox[{
                    UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
                    RowBox[{"t", ",", "i"}], "]"}]}]}], 
                   TagBox[")",
                    "AutoParentheses"]}]}]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(hasExponentFunctionPP)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:225295367", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568216972651916`*^9, 3.6568217046596045`*^9}, {
  3.656821875746105*^9, 3.656822083834871*^9}, {3.6568221655478144`*^9, 
  3.6568221656882143`*^9}, {3.6568230109071007`*^9, 3.656823070218405*^9}, {
  3.6568258975957727`*^9, 3.6568258994677753`*^9}},
 CellTags->{
  "ID:225295367", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasExponentFunctionPP"},
 CellID->225295367],

Cell[BoxData[
 RowBox[{
  RowBox[{"isExtensionalPP", "[", "\[ScriptCapitalT]", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", 
       RowBox[{"s", ",", "t"}], "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
          UnderscriptBox["dim", "\[ScriptCapitalT]"]}]], 
        RowBox[{
         RowBox[{
          UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
          RowBox[{"s", ",", "i"}], "]"}], "\[Equal]", 
         RowBox[{
          UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
          RowBox[{"t", ",", "i"}], "]"}]}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[Implies]", 
      RowBox[{"s", "\[Equal]", "t"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExtensionalPP)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:292978558", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568217341124563`*^9, 3.656821746499878*^9}, {
  3.65682209487969*^9, 3.6568221378577657`*^9}, {3.6568248015424466`*^9, 
  3.656824802182048*^9}},
 CellTags->{
  "ID:292978558", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExtensionalPP"},
 CellID->292978558],

Cell[BoxData[
 RowBox[{
  RowBox[{"hasDegreeFunctionPP", "[", "\[ScriptCapitalT]", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "t", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["deg", "\[ScriptCapitalT]"], "[", "t", "]"}], 
      "\[Equal]", 
      RowBox[{
       UnderscriptBox[
        SubscriptBox["\[Sum]", 
         RowBox[{"\[DoubleStruckCapitalZ]IntL", "[", "0", "]"}]], 
        RowBox[{
         RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
         UnderscriptBox["dim", "\[ScriptCapitalT]"]}]], 
       RowBox[{
        UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
        RowBox[{"t", ",", "i"}], "]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(hasDegreeFunctionPP)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:340781435", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656823563711672*^9, 3.6568236283581853`*^9}, {
  3.656915840787285*^9, 3.6569158462004943`*^9}, {3.661582006080447*^9, 
  3.6615820105108547`*^9}},
 CellTags->{
  "ID:340781435", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasDegreeFunctionPP"},
 CellID->340781435],

Cell[BoxData[
 RowBox[{
  RowBox[{"hasDivisibilityPP", "[", "\[ScriptCapitalT]", "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", 
       RowBox[{"s", ",", "t"}], "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"s", 
        UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "t"}], 
       TagBox[")",
        "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
          UnderscriptBox["dim", "\[ScriptCapitalT]"]}]], 
        RowBox[{
         RowBox[{
          UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
          RowBox[{"s", ",", "i"}], "]"}], "\[LessEqual]", 
         RowBox[{
          UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
          RowBox[{"t", ",", "i"}], "]"}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(hasDivisibilityPP)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:480617864", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568260137223763`*^9, 3.6568260896321096`*^9}},
 CellTags->{
  "ID:480617864", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasDivisibilityPP"},
 CellID->480617864],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellChangeTimes->{3.6568216852375703`*^9},
 CellID->380818943]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellChangeTimes->{3.6568216852375703`*^9},
 CellID->707551302],

Cell["", "OpenEnvironment",
 CellID->15645017],

Cell[CellGroupData[{

Cell["Definition: Term Order", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656841958548938*^9, 3.656841960904542*^9}},
 CellID->417009435],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "R"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"isTermOrder", "[", 
     RowBox[{"R", ",", "\[ScriptCapitalT]"}], "]"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"isTotalIrreflOrder", "[", 
           RowBox[{"R", ",", "\[ScriptCapitalT]"}], "]"}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
             RowBox[{"[", "t", "]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{"t", "\[NotEqual]", 
              UnderscriptBox["1", "\[ScriptCapitalT]"]}], "\[Implies]", 
             RowBox[{"R", "[", 
              RowBox[{"t", ",", 
               UnderscriptBox["1", "\[ScriptCapitalT]"]}], "]"}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]},
         {
          RowBox[{
           UnderscriptBox["\[ForAll]", 
            RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
             RowBox[{"[", 
              RowBox[{"s", ",", "t", ",", "u"}], "]"}]}]], 
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{"R", "[", 
              RowBox[{"s", ",", "t"}], "]"}], "\[Implies]", 
             RowBox[{"R", "[", 
              RowBox[{
               RowBox[{"s", 
                UnderscriptBox["*", "\[ScriptCapitalT]"], "u"}], ",", 
               RowBox[{"t", 
                UnderscriptBox["*", "\[ScriptCapitalT]"], "u"}]}], "]"}]}], 
            TagBox[")",
             "AutoParentheses"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isTermOrder)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:897055662", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656841967908954*^9, 3.6568420938479757`*^9}},
 CellTags->{
  "ID:897055662", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isTermOrder"},
 CellID->897055662],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->144066118]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->378822340]
}, Closed]],

Cell[CellGroupData[{

Cell["Immediate Consequences (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.656824210173608*^9, 3.6568242124044113`*^9}, {
  3.65682440483275*^9, 3.6568244098715587`*^9}, {3.6568245152185435`*^9, 
  3.6568245193681507`*^9}},
 CellID->44692600],

Cell["", "OpenEnvironment",
 CellID->473301257],

Cell[CellGroupData[{

Cell["Lemma: Consequences of Being a Power-Product Domain", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656824221390027*^9, 3.656824238893258*^9}},
 CellID->200657340],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "isExpPPDomain", "[", "\[ScriptCapitalT]", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 532395027}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656824243744867*^9, 3.6568242505308785`*^9}},
 CellTags->{
  "ID:532395027", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->532395027],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["dim", "\[ScriptCapitalT]"], "\[Element]", 
  SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain dim)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:160997957", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656824254664886*^9, 3.6568242639625025`*^9}},
 CellTags->{
  "ID:160997957", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain dim"},
 CellID->160997957],

Cell[BoxData[
 RowBox[{"hasDecisionPredicatePP", "[", "\[ScriptCapitalT]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain hasDecisionPredicate)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:83558965", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.656824322415805*^9},
 CellTags->{
  "ID:83558965", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasDecisionPredicate"},
 CellID->83558965],

Cell[BoxData[
 RowBox[{"hasExponentFunctionPP", "[", "\[ScriptCapitalT]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain hasExponentFunction)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:349898371", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.6568243267682123`*^9},
 CellTags->{
  "ID:349898371", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasExponentFunction"},
 CellID->349898371],

Cell[BoxData[
 RowBox[{"isExtensionalPP", "[", "\[ScriptCapitalT]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain isExtensional)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:200216737", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.65682433118402*^9},
 CellTags->{
  "ID:200216737", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isExtensional"},
 CellID->200216737],

Cell[BoxData[
 RowBox[{"hasDegreeFunctionPP", "[", "\[ScriptCapitalT]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain hasDegreeFunction)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:276506833", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.656824338500433*^9},
 CellTags->{
  "ID:276506833", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasDegreeFunction"},
 CellID->276506833],

Cell[BoxData[
 RowBox[{"hasDivisibilityPP", "[", "\[ScriptCapitalT]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain hasDivisibility)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:135986343", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656826137056193*^9, 3.656826142953003*^9}},
 CellTags->{
  "ID:135986343", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasDivisibility"},
 CellID->135986343],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    UnderscriptBox["dim", "\[ScriptCapitalT]"]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
     RowBox[{"[", "s", "]"}]}]], 
   RowBox[{
    RowBox[{
     UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
     RowBox[{"s", ",", "i"}], "]"}], "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalN]", 
     "0"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain membership exp)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:122009120", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656927950086667*^9, 3.656927983548726*^9}},
 CellTags->{
  "ID:122009120", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership exp"},
 CellID->122009120],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->711208495]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->141443493],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain dim)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain dim)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["dim", 
            StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
           "\[Element]", 
           SubscriptBox[
           "\[DoubleStruckCapitalN]", "0"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["dim", 
            StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
           "\[Element]", 
           SubscriptBox[
           "\[DoubleStruckCapitalN]", "0"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:160997957"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p160997957-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p160997957-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:160997957", "Proof|ID:160997957-1"},
 CellID->4369865],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p160997957-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->74490556]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain hasDecisionPredicate)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain hasDecisionPredicate)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"hasDecisionPredicatePP", "[", 
           StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"hasDecisionPredicatePP", "[", 
           StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:83558965"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p83558965-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p83558965-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:83558965", "Proof|ID:83558965-1"},
 CellID->777822562],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p83558965-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->348442790]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain hasExponentFunction)\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain hasExponentFunction)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"hasExponentFunctionPP", "[", 
           StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"hasExponentFunctionPP", "[", 
           StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:349898371"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p349898371-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p349898371-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:349898371", "Proof|ID:349898371-1"},
 CellID->51022891],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p349898371-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->494240957]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain isExtensional)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain isExtensional)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isExtensionalPP", "[", 
           StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isExtensionalPP", "[", 
           StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:200216737"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p200216737-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p200216737-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:200216737", "Proof|ID:200216737-1"},
 CellID->19036167],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p200216737-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->454301965]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain hasDegreeFunction)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain hasDegreeFunction)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"hasDegreeFunctionPP", "[", 
           StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"hasDegreeFunctionPP", "[", 
           StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:276506833"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p276506833-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p276506833-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:276506833", "Proof|ID:276506833-1"},
 CellID->447792752],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p276506833-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->204590069]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain hasDivisibility)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain hasDivisibility)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"hasDivisibilityPP", "[", 
           StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"hasDivisibilityPP", "[", 
           StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:135986343"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p135986343-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p135986343-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:135986343", "Proof|ID:135986343-1"},
 CellID->380432363],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p135986343-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellChangeTimes->{3.6568264580423565`*^9},
 CellID->100546045]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain membership exp)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain membership exp)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               UnderscriptBox["dim", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["exp", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], "]"}], "\[Element]", 
           SubscriptBox[
           "\[DoubleStruckCapitalN]", "0"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               UnderscriptBox["dim", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["exp", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["i", "ExpressionVariable"]}], "]"}], "\[Element]", 
           SubscriptBox[
           "\[DoubleStruckCapitalN]", "0"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:122009120"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p122009120-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p122009120-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:122009120", "Proof|ID:122009120-1"},
 CellID->131414947],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p122009120-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->34728721]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Cancellative Commutative Monoids (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.6568238726546144`*^9, 3.6568238889098425`*^9}, {
   3.656828523676586*^9, 3.656828527732593*^9}, 3.656828610990939*^9},
 CellID->180962323],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "isExpPPDomain", "[", "\[ScriptCapitalT]", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 185429719}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656823928814713*^9, 3.6568239365367265`*^9}},
 CellTags->{
  "ID:185429719", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->185429719],

Cell["", "OpenEnvironment",
 CellID->430279676],

Cell[CellGroupData[{

Cell["Lemma: ExpPPDomains have Properties of CCMonoids", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656823896787857*^9, 3.656823897006257*^9}, {
  3.6568239661611786`*^9, 3.6568239797956023`*^9}},
 CellID->158100854],

Cell[BoxData[
 RowBox[{"isDomain", "[", "\[ScriptCapitalT]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain isDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:81626705", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656824281450133*^9, 3.65682428538134*^9}},
 CellTags->{
  "ID:81626705", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isDomain"},
 CellID->81626705],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", 
     RowBox[{"s", ",", "t"}], "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", 
    RowBox[{"s", 
     UnderscriptBox["*", "\[ScriptCapitalT]"], "t"}], 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain membership *)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:154517895", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568249074042325`*^9, 3.65682492862027*^9}},
 CellTags->{
  "ID:154517895", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership *", "IR"},
 CellID->154517895],

Cell[BoxData[
 RowBox[{"isAssoc", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["*", "\[ScriptCapitalT]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalT]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain * isAssoc)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:226316196", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656824026938885*^9, 3.6568240453001175`*^9}},
 CellTags->{
  "ID:226316196", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain * isAssoc"},
 CellID->226316196],

Cell[BoxData[
 RowBox[{"isComm", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["*", "\[ScriptCapitalT]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalT]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain * isComm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:96938761", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.65682405250733*^9},
 CellTags->{
  "ID:96938761", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain * isComm"},
 CellID->96938761],

Cell[BoxData[
 RowBox[{"isNeutral", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["*", "\[ScriptCapitalT]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", 
   UnderscriptBox["1", "\[ScriptCapitalT]"], ",", "\[ScriptCapitalT]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain 1 isNeutral)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:499736780", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.656824060416544*^9},
 CellTags->{
  "ID:499736780", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain 1 isNeutral"},
 CellID->499736780],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", 
     RowBox[{"x", ",", "y", ",", "z"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"z", 
       UnderscriptBox["*", "\[ScriptCapitalT]"], "x"}], "\[Equal]", 
      RowBox[{"z", 
       UnderscriptBox["*", "\[ScriptCapitalT]"], "y"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"x", "\[Equal]", "y"}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain * lcanc)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:745587327", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.656824068684559*^9},
 CellTags->{
  "ID:745587327", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain * lcanc"},
 CellID->745587327],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->131282212]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->207922316],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain isDomain)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain isDomain)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isDomain", "[", 
           StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isDomain", "[", 
           StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:81626705"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p81626705-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p81626705-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:81626705", "Proof|ID:81626705-1"},
 CellID->281230526],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p81626705-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->215127808]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain membership *)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain membership *)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
           RowBox[{
             StyleBox["s", "ExpressionVariable"], 
             UnderscriptBox["*", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
             StyleBox["t", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
           RowBox[{
             StyleBox["s", "ExpressionVariable"], 
             UnderscriptBox["*", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
             StyleBox["t", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:154517895"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p154517895-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p154517895-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:154517895", "Proof|ID:154517895-1"},
 CellID->527357288],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p154517895-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->122633837]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain * isAssoc)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain * isAssoc)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isAssoc", "[", 
           RowBox[{
             UnderscriptBox["*", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], ",", 
             StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isAssoc", "[", 
           RowBox[{
             UnderscriptBox["*", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], ",", 
             StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:226316196"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p226316196-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p226316196-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:226316196", "Proof|ID:226316196-1"},
 CellID->31171248],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p226316196-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->166332303]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain * isComm)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain * isComm)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isComm", "[", 
           RowBox[{
             UnderscriptBox["*", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], ",", 
             StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isComm", "[", 
           RowBox[{
             UnderscriptBox["*", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], ",", 
             StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:96938761"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p96938761-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p96938761-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:96938761", "Proof|ID:96938761-1"},
 CellID->6993021],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p96938761-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->47945419]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain 1 isNeutral)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain 1 isNeutral)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isNeutral", "[", 
           RowBox[{
             UnderscriptBox["*", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], ",", 
             UnderscriptBox["1", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], ",", 
             StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isNeutral", "[", 
           RowBox[{
             UnderscriptBox["*", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], ",", 
             UnderscriptBox["1", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], ",", 
             StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:499736780"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p499736780-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p499736780-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:499736780", "Proof|ID:499736780-1"},
 CellID->247701442],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p499736780-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->34826465]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain * lcanc)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain * lcanc)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
               StyleBox["x", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["y", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["z", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
               StyleBox["x", "ExpressionVariable"]}], "\[Equal]", 
             RowBox[{
               StyleBox["z", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
               StyleBox["y", "ExpressionVariable"]}]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["x", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "y", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:745587327"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p745587327-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p745587327-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:745587327", "Proof|ID:745587327-1"},
 CellID->99539526],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p745587327-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->1175411]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->65103093],

Cell[CellGroupData[{

Cell["Theorem: ExpPPDomains are CCMonoids", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656824127165862*^9, 3.6568241341858745`*^9}},
 CellID->429762970],

Cell[BoxData[
 RowBox[{"isCCMonoid", "[", "\[ScriptCapitalT]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain isCCMonoid)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:333683969", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656824136447878*^9, 3.656824141861088*^9}},
 CellTags->{
  "ID:333683969", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isCCMonoid"},
 CellID->333683969],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->14565163]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->47457899],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain isCCMonoid)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain isCCMonoid)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isCCMonoid", "[", 
           StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isCCMonoid", "[", 
           StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:333683969"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p333683969-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p333683969-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:333683969", "Proof|ID:333683969-1"},
 CellID->190665795],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p333683969-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->331229762]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Divisibility and LCM (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6568286124105415`*^9, 3.656828617558551*^9}, {
  3.6568311415019846`*^9, 3.656831145557992*^9}},
 CellID->71800829],

Cell["", "OpenEnvironment",
 CellID->656590807],

Cell[CellGroupData[{

Cell["Lemma: Divisibility and LCM in ExpPPDomains", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6568286283069696`*^9, 3.656828642175394*^9}},
 CellID->290258020],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "isExpPPDomain", "[", "\[ScriptCapitalT]", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 274960855}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65682869663509*^9, 3.656828703015501*^9}},
 CellTags->{
  "ID:274960855", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->274960855],

Cell[BoxData[
 RowBox[{"isJoin", "[", 
  RowBox[{
   UnderscriptBox["lcm", "\[ScriptCapitalT]"], ",", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalT]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain lcm isJoin)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:52136714", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.656828672626648*^9},
 CellTags->{
  "ID:52136714", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain lcm isJoin"},
 CellID->52136714],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", 
    RowBox[{"s", ",", "t"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 35376325}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568286533918133`*^9, 3.6568286611918273`*^9}},
 CellTags->{
  "ID:35376325", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->35376325],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"s", 
    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "t"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"t", 
       UnderscriptBox["/", "\[ScriptCapitalT]"], "s"}], ")"}], 
     UnderscriptBox["*", "\[ScriptCapitalT]"], "s"}], "\[Equal]", "t"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain div.1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:138508454", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656828691783481*^9, 3.656828709739113*^9}, {
  3.6568297272393*^9, 3.656829735304514*^9}},
 CellTags->{
  "ID:138508454", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain div.1"},
 CellID->138508454],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"s", 
    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "t"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "u", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"t", "\[Equal]", 
      RowBox[{"u", 
       UnderscriptBox["*", "\[ScriptCapitalT]"], "s"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain div.2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:14785826", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.656828723155136*^9, {3.6568297390173206`*^9, 3.65682975001534*^9}},
 CellTags->{
  "ID:14785826", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain div.2"},
 CellID->14785826],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->147667495]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->416318934],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain lcm isJoin)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain lcm isJoin)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isJoin", "[", 
           RowBox[{
             UnderscriptBox["lcm", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], ",", 
             UnderscriptBox["\[VerticalBar]", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], ",", 
             StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isJoin", "[", 
           RowBox[{
             UnderscriptBox["lcm", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], ",", 
             UnderscriptBox["\[VerticalBar]", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], ",", 
             StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:52136714"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p52136714-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p52136714-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:52136714", "Proof|ID:52136714-1"},
 CellID->110160557],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p52136714-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->72948561]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain div.1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain div.1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["s", "ExpressionVariable"], 
             UnderscriptBox["\[VerticalBar]", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
             StyleBox["t", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["/", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["s", "ExpressionVariable"]}], 
               UnderscriptBox["*", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
               StyleBox["s", "ExpressionVariable"]}], "\[Equal]", 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["s", "ExpressionVariable"], 
             UnderscriptBox["\[VerticalBar]", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
             StyleBox["t", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["/", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["s", "ExpressionVariable"]}], 
               UnderscriptBox["*", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
               StyleBox["s", "ExpressionVariable"]}], "\[Equal]", 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:138508454"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p138508454-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p138508454-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:138508454", "Proof|ID:138508454-1"},
 CellID->245133977],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p138508454-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->598366505]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain div.2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain div.2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["s", "ExpressionVariable"], 
             UnderscriptBox["\[VerticalBar]", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
             StyleBox["t", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[Equal]", 
               RowBox[{
                 StyleBox["u", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox[
                 "s", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["s", "ExpressionVariable"], 
             UnderscriptBox["\[VerticalBar]", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
             StyleBox["t", "ExpressionVariable"]}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
                   StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[Equal]", 
               RowBox[{
                 StyleBox["u", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox[
                 "s", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:14785826"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p14785826-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p14785826-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:14785826", "Proof|ID:14785826-1"},
 CellID->2945667],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p14785826-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->9554763]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Degree Function (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.65691717358483*^9, 3.6569171764552345`*^9}, {
  3.6569294029258194`*^9, 3.6569294073874273`*^9}},
 CellID->194654141],

Cell["", "OpenEnvironment",
 CellID->21521330],

Cell[CellGroupData[{

Cell["Lemma: Simple Properties of the Degree Function", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656917186735653*^9, 3.6569171950660677`*^9}},
 CellID->129749829],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "isExpPPDomain", "[", "\[ScriptCapitalT]", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 181034073}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656917202741281*^9, 3.6569172104320946`*^9}},
 CellTags->{
  "ID:181034073", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->181034073],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", "s", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 943617998}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6569172216329145`*^9, 3.6569172292613277`*^9}},
 CellTags->{
  "ID:943617998", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->943617998],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"n", "=", "0"}], ",", "\[Ellipsis]", ",", 
    UnderscriptBox["dim", "\[ScriptCapitalT]"]}]], 
  RowBox[{
   UnderscriptBox["\[Element]", 
    RowBox[{"\[DoubleStruckCapitalZ]IntL", "[", "0", "]"}]], 
   RowBox[{"[", 
    RowBox[{
     UnderscriptBox[
      SubscriptBox["\[Sum]", 
       RowBox[{"\[DoubleStruckCapitalZ]IntL", "[", "0", "]"}]], 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]], 
     RowBox[{
      UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
      RowBox[{"s", ",", "i"}], "]"}]}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain membership deg ind)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:706945753", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656924760343779*^9, 3.656924819577083*^9}, {
  3.661582066344353*^9, 3.661582081133179*^9}},
 CellTags->{
  "ID:706945753", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership deg ind"},
 CellID->706945753],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"n", "=", "0"}], ",", "\[Ellipsis]", ",", 
    UnderscriptBox["dim", "\[ScriptCapitalT]"]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox[
         SubscriptBox["\[Sum]", 
          RowBox[{"\[DoubleStruckCapitalZ]IntL", "[", "0", "]"}]], 
         RowBox[{
          RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]], 
        RowBox[{
         UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
         RowBox[{"s", ",", "i"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[Equal]", "0"}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
         RowBox[{"s", ",", "i"}], "]"}], "\[Equal]", "0"}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain deg=0 ind)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:606418540", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6569248737871785`*^9, 3.656924938886093*^9}, 
   3.6615820767183714`*^9},
 CellTags->{
  "ID:606418540", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain deg=0 ind"},
 CellID->606418540],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["deg", "\[ScriptCapitalT]"], "[", "s", "]"}], "\[Element]", 
  
  SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain membership deg)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:47160", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6569172630821867`*^9, 3.656917276872611*^9}, {
  3.656917404356035*^9, 3.656917413965652*^9}},
 CellTags->{
  "ID:47160", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership deg", "IR"},
 CellID->47160],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["deg", "\[ScriptCapitalT]"], "[", "s", "]"}], "\[Equal]", 
    "0"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"s", "\[Equal]", 
    UnderscriptBox["1", "\[ScriptCapitalT]"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain deg=0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:58897330", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6569173782571893`*^9, 3.6569173964000216`*^9}},
 CellTags->{
  "ID:58897330", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain deg=0"},
 CellID->58897330],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["deg", "\[ScriptCapitalT]"], "[", 
     RowBox[{"s", 
      UnderscriptBox["*", "\[ScriptCapitalT]"], "t"}], "]"}], "\[Equal]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["deg", "\[ScriptCapitalT]"], "[", "s", "]"}], "+", 
     RowBox[{
      UnderscriptBox["deg", "\[ScriptCapitalT]"], "[", "t", "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain deg *)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:170894204", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656917434042887*^9, 3.6569174757417603`*^9}},
 CellTags->{
  "ID:170894204", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain deg *"},
 CellID->170894204],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->31663413]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->458193557],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain membership deg ind)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain membership deg ind)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               UnderscriptBox["dim", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            SubscriptBox["\[DoubleStruckCapitalN]", "0"]], "[", 
           RowBox[{
             UnderscriptBox[
              SubscriptBox["\[Sum]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]], 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   StyleBox["n", "ExpressionVariable"]}]}}]], 
             RowBox[{
               UnderscriptBox["exp", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"]}], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               UnderscriptBox["dim", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            SubscriptBox["\[DoubleStruckCapitalN]", "0"]], "[", 
           RowBox[{
             UnderscriptBox[
              SubscriptBox["\[Sum]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]], 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   StyleBox["n", "ExpressionVariable"]}]}}]], 
             RowBox[{
               UnderscriptBox["exp", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"]}], "]"}]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:706945753"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p706945753-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p706945753-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:706945753", "Proof|ID:706945753-1"},
 CellID->905512569],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p706945753-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->5970720]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain deg=0 ind)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain deg=0 ind)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               UnderscriptBox["dim", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox[
                SubscriptBox["\[Sum]", 
                 SubscriptBox["\[DoubleStruckCapitalN]", "0"]], 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]}}]], 
               RowBox[{
                 UnderscriptBox["exp", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["s", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"]}], "]"}]}], "\[Equal]",
              "0"}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   StyleBox["n", "ExpressionVariable"]}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["exp", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["s", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               "0"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               UnderscriptBox["dim", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox[
                SubscriptBox["\[Sum]", 
                 SubscriptBox["\[DoubleStruckCapitalN]", "0"]], 
                GridBox[{{
                   RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    StyleBox["n", "ExpressionVariable"]}]}}]], 
               RowBox[{
                 UnderscriptBox["exp", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["s", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"]}], "]"}]}], "\[Equal]",
              "0"}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   StyleBox["n", "ExpressionVariable"]}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["exp", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["s", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               "0"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:606418540"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p606418540-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p606418540-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:606418540", "Proof|ID:606418540-1"},
 CellID->408182534],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p606418540-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->193074823]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain membership deg)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain membership deg)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["deg", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
             StyleBox["s", "ExpressionVariable"], "]"}], "\[Element]", 
           SubscriptBox[
           "\[DoubleStruckCapitalN]", "0"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["deg", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
             StyleBox["s", "ExpressionVariable"], "]"}], "\[Element]", 
           SubscriptBox[
           "\[DoubleStruckCapitalN]", "0"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:47160"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p47160-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p47160-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:47160", "Proof|ID:47160-1"},
 CellID->168232727],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p47160-1-display.m\"}]"]],
      Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->170693603]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain deg=0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain deg=0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["deg", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}], "\[Equal]", "0"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["s", "ExpressionVariable"], "\[Equal]", 
             UnderscriptBox["1", 
              StyleBox[
              "\[ScriptCapitalT]", 
               "ExpressionVariable"]]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["deg", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}], "\[Equal]", "0"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             StyleBox["s", "ExpressionVariable"], "\[Equal]", 
             UnderscriptBox["1", 
              StyleBox[
              "\[ScriptCapitalT]", 
               "ExpressionVariable"]]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:58897330"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p58897330-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p58897330-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:58897330", "Proof|ID:58897330-1"},
 CellID->293749343],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p58897330-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->470353965]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain deg *)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain deg *)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["deg", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["deg", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}], "+", 
             RowBox[{
               UnderscriptBox["deg", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["deg", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["*", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["deg", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}], "+", 
             RowBox[{
               UnderscriptBox["deg", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:170894204"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p170894204-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p170894204-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:170894204", "Proof|ID:170894204-1"},
 CellID->366355002],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p170894204-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->6017937]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Dickson Domain (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6568311518136024`*^9, 3.656831154262807*^9}, {
  3.6568418268379064`*^9, 3.656841831018714*^9}},
 CellID->467568792],

Cell["", "OpenEnvironment",
 CellID->187255726],

Cell[CellGroupData[{

Cell["\<\
Lemma: Every Sequence of Natural Numbers Contains an Increasing Subsequence\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656831675587323*^9, 3.6568316934805546`*^9}, 
   3.656831723744608*^9},
 CellID->152834815],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSequences", "[", 
      RowBox[{"\[DoubleStruckCapitalZ]IntL", "[", "0", "]"}], "]"}]], 
    RowBox[{"[", "f", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{"isIncrMapping", "[", "g", "]"}]], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "\[Element]", "\[DoubleStruckCapitalN]"}]}]], 
     RowBox[{"j", ">", "i"}]], 
    RowBox[{
     RowBox[{
      RowBox[{"subSeq", "[", 
       RowBox[{"f", ",", "g"}], "]"}], "[", "i", "]"}], "\[LessEqual]", 
     RowBox[{
      RowBox[{"subSeq", "[", 
       RowBox[{"f", ",", "g"}], "]"}], "[", "j", 
      "]"}]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(increasing subsequence \[DoubleStruckCapitalN]0)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:49640627", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656742136612149*^9, 3.656742255783758*^9}, {
  3.6567424095844283`*^9, 3.6567424725305386`*^9}, {3.65674268960892*^9, 
  3.656742697284134*^9}, {3.6615828161052704`*^9, 3.661582820847679*^9}},
 CellTags->{
  "ID:49640627", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "increasing subsequence \[DoubleStruckCapitalN]0"},
 CellID->49640627],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->454910408]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->460125164],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(increasing subsequence \[DoubleStruckCapitalN]0)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(increasing subsequence \[DoubleStruckCapitalN]0)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSequences", "[", 
                  SubscriptBox["\[DoubleStruckCapitalN]", "0"], "]"}]], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{"isIncrMapping", "[", 
                 StyleBox["g", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}, {
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
              RowBox[{
                StyleBox["j", "ExpressionVariable"], ">", 
                StyleBox["i", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"subSeq", "[", 
                   RowBox[{
                    StyleBox["f", "ExpressionVariable"], ",", 
                    StyleBox["g", "ExpressionVariable"]}], "]"}], "[", 
                 StyleBox["i", "ExpressionVariable"], "]"}], "\[LessEqual]", 
               RowBox[{
                 RowBox[{"subSeq", "[", 
                   RowBox[{
                    StyleBox["f", "ExpressionVariable"], ",", 
                    StyleBox["g", "ExpressionVariable"]}], "]"}], "[", 
                 StyleBox["j", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSequences", "[", 
                  SubscriptBox["\[DoubleStruckCapitalN]", "0"], "]"}]], "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{"isIncrMapping", "[", 
                 StyleBox["g", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}, {
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
              RowBox[{
                StyleBox["j", "ExpressionVariable"], ">", 
                StyleBox["i", "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"subSeq", "[", 
                   RowBox[{
                    StyleBox["f", "ExpressionVariable"], ",", 
                    StyleBox["g", "ExpressionVariable"]}], "]"}], "[", 
                 StyleBox["i", "ExpressionVariable"], "]"}], "\[LessEqual]", 
               RowBox[{
                 RowBox[{"subSeq", "[", 
                   RowBox[{
                    StyleBox["f", "ExpressionVariable"], ",", 
                    StyleBox["g", "ExpressionVariable"]}], "]"}], "[", 
                 StyleBox["j", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:49640627"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p49640627-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p49640627-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:49640627", "Proof|ID:49640627-1"},
 CellID->620572316],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p49640627-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->563905649]
}, Closed]],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "isExpPPDomain", "[", "\[ScriptCapitalT]", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 256365628}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656831745725046*^9, 3.656831753899461*^9}},
 CellTags->{
  "ID:256365628", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->256365628],

Cell["", "OpenEnvironment",
 CellID->346122881],

Cell[CellGroupData[{

Cell["Definition: Divisibility of First Components", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6568317719174924`*^9, 3.656831784927915*^9}},
 CellID->216115192],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"n", ",", "s", ",", "t"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["divN", "\[ScriptCapitalT]"], "[", 
     RowBox[{"s", ",", "t", ",", "n"}], "]"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]], 
      RowBox[{
       RowBox[{
        UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
        RowBox[{"s", ",", "i"}], "]"}], "\[LessEqual]", 
       RowBox[{
        UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
        RowBox[{"t", ",", "i"}], "]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain divN)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:443096051", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568317931179295`*^9, 3.6568318489972277`*^9}, {
  3.6568321044320765`*^9, 3.656832124010111*^9}},
 CellTags->{
  "ID:443096051", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain divN"},
 CellID->443096051],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->306607494]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->45893119],

Cell["", "OpenEnvironment",
 CellID->21114114],

Cell[CellGroupData[{

Cell["Lemma: Preparations for Dickson Domain", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6568311884112673`*^9, 3.6568311962580805`*^9}},
 CellID->13494331],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"n", "\[Element]", 
    SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", 
       RowBox[{"s", ",", "t"}], "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["divN", "\[ScriptCapitalT]"], "[", 
       RowBox[{"s", ",", "t", ",", 
        RowBox[{"n", "+", "1"}]}], "]"}], "\[DoubleLongLeftRightArrow]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         RowBox[{
          UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
          RowBox[{"s", ",", 
           RowBox[{"n", "+", "1"}]}], "]"}], "\[LessEqual]", 
         RowBox[{
          UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
          RowBox[{"t", ",", 
           RowBox[{"n", "+", "1"}]}], "]"}]}], "\[And]", 
        RowBox[{
         UnderscriptBox["divN", "\[ScriptCapitalT]"], "[", 
         RowBox[{"s", ",", "t", ",", "n"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain divN +1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:27206868", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568314072670517`*^9, 3.656831473052367*^9}, {
  3.6568315583231173`*^9, 3.656831620238426*^9}, {3.6568318911017017`*^9, 
  3.6568319145173426`*^9}, {3.6568321277073174`*^9, 3.6568321303437223`*^9}},
 CellTags->{
  "ID:27206868", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain divN +1"},
 CellID->27206868],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", 
     RowBox[{"DomainSequences", "[", "\[ScriptCapitalT]", "]"}]], 
    RowBox[{"[", "f", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[Exists]", 
    RowBox[{"isIncrMapping", "[", "g", "]"}]], 
   RowBox[{
    UnderscriptBox[
     UnderscriptBox["\[ForAll]", 
      RowBox[{"i", ",", 
       RowBox[{"j", "\[Element]", "\[DoubleStruckCapitalN]"}]}]], 
     RowBox[{"j", ">", "i"}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["divN", "\[ScriptCapitalT]"], "[", 
      RowBox[{
       RowBox[{
        RowBox[{"subSeq", "[", 
         RowBox[{"f", ",", "g"}], "]"}], "[", "i", "]"}], ",", 
       RowBox[{
        RowBox[{"subSeq", "[", 
         RowBox[{"f", ",", "g"}], "]"}], "[", "j", "]"}], ",", "0"}], "]"}], 
     TagBox[")",
      "AutoParentheses"]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(increasing subsequence isExpPPDomain 0)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:184673495", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65683131528629*^9, 3.6568313410731354`*^9}, {
   3.656831529103266*^9, 3.6568315490567007`*^9}, {3.6568318579048433`*^9, 
   3.656831881648085*^9}, 3.656832136973734*^9},
 CellTags->{
  "ID:184673495", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "increasing subsequence isExpPPDomain 0"},
 CellID->184673495],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"n", "=", "0"}], ",", "\[Ellipsis]", ",", 
    UnderscriptBox["dim", "\[ScriptCapitalT]"]}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     UnderscriptBox["\[Element]", 
      RowBox[{"DomainSequences", "[", "\[ScriptCapitalT]", "]"}]], 
     RowBox[{"[", "f", "]"}]}]], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{"isIncrMapping", "[", "g", "]"}]], 
    RowBox[{
     UnderscriptBox[
      UnderscriptBox["\[ForAll]", 
       RowBox[{"i", ",", 
        RowBox[{"j", "\[Element]", "\[DoubleStruckCapitalN]"}]}]], 
      RowBox[{"j", ">", "i"}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{
       UnderscriptBox["divN", "\[ScriptCapitalT]"], "[", 
       RowBox[{
        RowBox[{
         RowBox[{"subSeq", "[", 
          RowBox[{"f", ",", "g"}], "]"}], "[", "i", "]"}], ",", 
        RowBox[{
         RowBox[{"subSeq", "[", 
          RowBox[{"f", ",", "g"}], "]"}], "[", "j", "]"}], ",", "n"}], "]"}], 
      
      TagBox[")",
       "AutoParentheses"]}]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(increasing subsequence isExpPPDomain)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:43287351", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656831970443441*^9, 3.656831986573869*^9}, 
   3.6568321397973385`*^9, {3.656832312692442*^9, 3.6568323204300556`*^9}, 
   3.656834891601375*^9},
 CellTags->{
  "ID:43287351", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "increasing subsequence isExpPPDomain"},
 CellID->43287351],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->506192652]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->19274976],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain divN +1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain divN +1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["divN", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], "+", "1"}]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["exp", 
                    StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["s", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "+", "1"}]}], "]"}], 
                 "\[LessEqual]", 
                 RowBox[{
                   UnderscriptBox["exp", 
                    StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "+", "1"}]}], 
                   "]"}]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["divN", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["s", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"], ",", 
                   StyleBox["n", "ExpressionVariable"]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["divN", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"], ",", 
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], "+", "1"}]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["exp", 
                    StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["s", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "+", "1"}]}], "]"}], 
                 "\[LessEqual]", 
                 RowBox[{
                   UnderscriptBox["exp", 
                    StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    RowBox[{
                    StyleBox["n", "ExpressionVariable"], "+", "1"}]}], 
                   "]"}]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["divN", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
                 RowBox[{
                   StyleBox["s", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"], ",", 
                   StyleBox["n", "ExpressionVariable"]}], "]"}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:27206868"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p27206868-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p27206868-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:27206868", "Proof|ID:27206868-1"},
 CellID->99745602],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p27206868-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->98056265]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(increasing subsequence isExpPPDomain 0)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(increasing subsequence isExpPPDomain 0)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSequences", "[", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{"isIncrMapping", "[", 
                 StyleBox["g", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}, {
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
              RowBox[{
                StyleBox["j", "ExpressionVariable"], ">", 
                StyleBox["i", "ExpressionVariable"]}]], 
             RowBox[{
               UnderscriptBox["divN", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   RowBox[{"subSeq", "[", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], ",", 
                    StyleBox["g", "ExpressionVariable"]}], "]"}], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   RowBox[{"subSeq", "[", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], ",", 
                    StyleBox["g", "ExpressionVariable"]}], "]"}], "[", 
                   StyleBox["j", "ExpressionVariable"], "]"}], ",", "0"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSequences", "[", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{"isIncrMapping", "[", 
                 StyleBox["g", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}, {
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
              RowBox[{
                StyleBox["j", "ExpressionVariable"], ">", 
                StyleBox["i", "ExpressionVariable"]}]], 
             RowBox[{
               UnderscriptBox["divN", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   RowBox[{"subSeq", "[", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], ",", 
                    StyleBox["g", "ExpressionVariable"]}], "]"}], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   RowBox[{"subSeq", "[", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], ",", 
                    StyleBox["g", "ExpressionVariable"]}], "]"}], "[", 
                   StyleBox["j", "ExpressionVariable"], "]"}], ",", "0"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:184673495"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p184673495-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p184673495-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:184673495", "Proof|ID:184673495-1"},
 CellID->104166294],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p184673495-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->267847356]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(increasing subsequence isExpPPDomain)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(increasing subsequence isExpPPDomain)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               UnderscriptBox["dim", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSequences", "[", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{"isIncrMapping", "[", 
                 StyleBox["g", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}, {
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
              RowBox[{
                StyleBox["j", "ExpressionVariable"], ">", 
                StyleBox["i", "ExpressionVariable"]}]], 
             RowBox[{
               UnderscriptBox["divN", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   RowBox[{"subSeq", "[", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], ",", 
                    StyleBox["g", "ExpressionVariable"]}], "]"}], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   RowBox[{"subSeq", "[", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], ",", 
                    StyleBox["g", "ExpressionVariable"]}], "]"}], "[", 
                   StyleBox["j", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["n", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["n", "ExpressionVariable"], "=", "0"}], ",", 
               "\[Ellipsis]", ",", 
               UnderscriptBox["dim", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"DomainSequences", "[", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]],
                "[", 
               StyleBox["f", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Exists]", 
            GridBox[{{
               RowBox[{"isIncrMapping", "[", 
                 StyleBox["g", "ExpressionVariable"], "]"}]}}]], 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[ForAll]", 
               GridBox[{{
                  RowBox[{
                    StyleBox["i", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}, {
                  RowBox[{
                    StyleBox["j", "ExpressionVariable"], "\[Element]", 
                    "\[DoubleStruckCapitalN]"}]}}]], 
              RowBox[{
                StyleBox["j", "ExpressionVariable"], ">", 
                StyleBox["i", "ExpressionVariable"]}]], 
             RowBox[{
               UnderscriptBox["divN", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               RowBox[{
                 RowBox[{
                   RowBox[{"subSeq", "[", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], ",", 
                    StyleBox["g", "ExpressionVariable"]}], "]"}], "[", 
                   StyleBox["i", "ExpressionVariable"], "]"}], ",", 
                 RowBox[{
                   RowBox[{"subSeq", "[", 
                    RowBox[{
                    StyleBox["f", "ExpressionVariable"], ",", 
                    StyleBox["g", "ExpressionVariable"]}], "]"}], "[", 
                   StyleBox["j", "ExpressionVariable"], "]"}], ",", 
                 StyleBox["n", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:43287351"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p43287351-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p43287351-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:43287351", "Proof|ID:43287351-1"},
 CellID->460075245],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p43287351-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {61., 67.}}]], "ProofInfoBody",
 CellID->392596903]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->503713153],

Cell[CellGroupData[{

Cell["Theorem: ExpPPDomains are Dickson Domains", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6568411773595657`*^9, 3.6568411960171986`*^9}},
 CellID->315610582],

Cell[BoxData[
 RowBox[{"isDicksonDomain", "[", 
  RowBox[{"\[ScriptCapitalT]", ",", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}]}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(isExpPPDomain isDicksonDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:30919390", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568411993712044`*^9, 3.6568412170460353`*^9}},
 CellTags->{
  "ID:30919390", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isDicksonDomain"},
 CellID->30919390],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->23678156]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->272481360],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isExpPPDomain isDicksonDomain)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isExpPPDomain isDicksonDomain)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isDicksonDomain", "[", 
           RowBox[{
             StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], ",", 
             UnderscriptBox["\[VerticalBar]", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isDicksonDomain", "[", 
           RowBox[{
             StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], ",", 
             UnderscriptBox["\[VerticalBar]", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:30919390"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p30919390-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p30919390-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:30919390", "Proof|ID:30919390-1"},
 CellID->278064001],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p30919390-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->89121114]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Properties of Term Orders (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.656842592970852*^9, 3.6568425960596576`*^9}, {
  3.6568441942708645`*^9, 3.6568441990288734`*^9}, {3.6569294285722647`*^9, 
  3.6569294305222683`*^9}},
 CellID->309098395],

Cell["", "OpenEnvironment",
 CellID->214207545],

Cell[CellGroupData[{

Cell["Lemma: Properties of Term Orders", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6568426120340853`*^9, 3.656842618305297*^9}},
 CellID->436328982],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"isExpPPDomain", "[", "\[ScriptCapitalT]", "]"}], ",", 
   "R"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 323333890}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568426220649033`*^9, 3.656842628195714*^9}, {
  3.6568428396696854`*^9, 3.6568428400596857`*^9}},
 CellTags->{
  "ID:323333890", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->323333890],

Cell[BoxData[
 RowBox[{
  RowBox[{"isTermOrder", "[", 
   RowBox[{"R", ",", "\[ScriptCapitalT]"}], "]"}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 235080407}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656842633608924*^9, 3.6568426427817397`*^9}, {
  3.6568428435540924`*^9, 3.656842846393297*^9}},
 CellTags->{
  "ID:235080407", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->235080407],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", 
     RowBox[{"s", ",", "t"}], "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"s", 
     UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "t"}], 
    "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"R", "[", 
       RowBox[{"t", ",", "s"}], "]"}], "\[Or]", 
      RowBox[{"t", "\[Equal]", "s"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(term order |)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:691531428", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568426585377674`*^9, 3.6568426915630255`*^9}},
 CellTags->{
  "ID:691531428", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "term order |"},
 CellID->691531428],

Cell[BoxData[
 RowBox[{"isOrderEmbedding", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    UnderscriptBox["*", "\[ScriptCapitalT]"], 
    TagBox[")",
     "AutoParentheses"]}], ",", "R", ",", "\[ScriptCapitalT]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(term order isOrderEmbedding)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:772895867", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6569129212377567`*^9, 3.6569129461042*^9}},
 CellTags->{
  "ID:772895867", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "term order isOrderEmbedding"},
 CellID->772895867],

Cell[BoxData[
 RowBox[{"isNoetherian", "[", 
  RowBox[{"R", ",", "\[ScriptCapitalT]"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(term order isNoetherian)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:83375799", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65684271121906*^9, 3.65684271697547*^9}},
 CellTags->{
  "ID:83375799", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "term order isNoetherian"},
 CellID->83375799],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->368030963]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->7746425],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(term order |)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(term order |)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             StyleBox["R", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"isTermOrder", "[", 
             RowBox[{
               StyleBox["R", "ExpressionVariable"], ",", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], "]"}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
                   StyleBox["s", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["\[VerticalBar]", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["t", "ExpressionVariable"]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   StyleBox["R", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Or]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                   StyleBox[
                   "s", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             StyleBox["R", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"isTermOrder", "[", 
             RowBox[{
               StyleBox["R", "ExpressionVariable"], ",", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], "]"}], 
           "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
                   StyleBox["s", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["\[VerticalBar]", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["t", "ExpressionVariable"]}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   StyleBox["R", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Or]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                   StyleBox[
                   "s", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:691531428"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p691531428-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p691531428-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:691531428", "Proof|ID:691531428-1"},
 CellID->26598154],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p691531428-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->174801662]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(term order isOrderEmbedding)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(term order isOrderEmbedding)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             StyleBox["R", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"isTermOrder", "[", 
             RowBox[{
               StyleBox["R", "ExpressionVariable"], ",", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], "]"}], 
           "\[Implies]", 
           RowBox[{"isOrderEmbedding", "[", 
             RowBox[{
               UnderscriptBox["*", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], ",", 
               StyleBox["R", "ExpressionVariable"], ",", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             StyleBox["R", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"isTermOrder", "[", 
             RowBox[{
               StyleBox["R", "ExpressionVariable"], ",", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], "]"}], 
           "\[Implies]", 
           RowBox[{"isOrderEmbedding", "[", 
             RowBox[{
               UnderscriptBox["*", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], ",", 
               StyleBox["R", "ExpressionVariable"], ",", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:772895867"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p772895867-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p772895867-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:772895867", "Proof|ID:772895867-1"},
 CellID->409263698],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p772895867-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->511868347]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(term order isNoetherian)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(term order isNoetherian)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             StyleBox["R", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"isTermOrder", "[", 
             RowBox[{
               StyleBox["R", "ExpressionVariable"], ",", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], "]"}], 
           "\[Implies]", 
           RowBox[{"isNoetherian", "[", 
             RowBox[{
               StyleBox["R", "ExpressionVariable"], ",", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             StyleBox["R", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"isTermOrder", "[", 
             RowBox[{
               StyleBox["R", "ExpressionVariable"], ",", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], "]"}], 
           "\[Implies]", 
           RowBox[{"isNoetherian", "[", 
             RowBox[{
               StyleBox["R", "ExpressionVariable"], ",", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:83375799"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p83375799-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p83375799-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:83375799", "Proof|ID:83375799-1"},
 CellID->549031787],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p83375799-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->459125148]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Ordered Power-Products (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6568233484458933`*^9, 3.656823353297502*^9}, {
  3.6569133083000364`*^9, 3.656913312371643*^9}},
 CellID->317472498],

Cell[CellGroupData[{

Cell["Definitions", "Subsubsection",
 CellChangeTimes->{{3.656823448177869*^9, 3.656823449940672*^9}},
 CellID->12853081],

Cell["", "OpenEnvironment",
 CellID->72677768],

Cell[CellGroupData[{

Cell["Definition: Monoid of Ordered Power-Products", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6568224502971144`*^9, 3.6568224753819585`*^9}},
 CellID->107281972],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"\[ScriptCapitalT]", ",", "R"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 69380281}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656822490404785*^9, 3.6568225054432116`*^9}},
 CellTags->{
  "ID:69380281", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->69380281],

Cell[BoxData[
 RowBox[{
  RowBox[{"OrderedPP", "[", 
   RowBox[{"\[ScriptCapitalT]", ",", "R"}], "]"}], ":=", 
  UnderscriptBox["\[CapitalDelta]", "N"]}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 70274751}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.6568231013872595`*^9},
 CellTags->{
  "ID:70274751", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->70274751],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"x", ",", "y"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 417719229}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568225362688656`*^9, 3.6568225373452673`*^9}},
 CellTags->{
  "ID:417719229", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->417719229],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", "N"], 
    RowBox[{"[", "x", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "x", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(OrderedPP \[Element])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:296713523", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568225673909197`*^9, 3.656822576657336*^9}},
 CellTags->{
  "ID:296713523", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP \[Element]"},
 CellID->296713523],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["dim", "N"], ":=", 
  UnderscriptBox["dim", "\[ScriptCapitalT]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(OrderedPP dim)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:275450356", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656822726673601*^9, 3.656822735238016*^9}},
 CellTags->{
  "ID:275450356", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP dim"},
 CellID->275450356],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "i"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["exp", "N"], "[", 
     RowBox[{"x", ",", "i"}], "]"}], ":=", 
    RowBox[{
     UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
     RowBox[{"x", ",", "i"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(OrderedPP exp)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:126765714", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656822713303378*^9, 3.6568227542388496`*^9}},
 CellTags->{
  "ID:126765714", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP exp"},
 CellID->126765714],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["1", "N"], ":=", 
  UnderscriptBox["1", "\[ScriptCapitalT]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(OrderedPP 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:300238463", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656822582148546*^9, 3.656822584972151*^9}},
 CellTags->{
  "ID:300238463", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP 1"},
 CellID->300238463],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["*", "N"], "y"}], ":=", 
  RowBox[{"x", 
   UnderscriptBox["*", "\[ScriptCapitalT]"], 
   "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(OrderedPP *)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:91738893", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568225909625616`*^9, 3.6568225940357666`*^9}},
 CellTags->{
  "ID:91738893", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP *"},
 CellID->91738893],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["/", "N"], "y"}], ":=", 
  RowBox[{"x", 
   UnderscriptBox["/", "\[ScriptCapitalT]"], 
   "y"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(OrderedPP /)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1017172830", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568225989809756`*^9, 3.656822602163381*^9}},
 CellTags->{
  "ID:1017172830", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP /"},
 CellID->1017172830],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lcm", "N"], "[", 
   RowBox[{"x", ",", "y"}], "]"}], ":=", 
  RowBox[{
   UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
   RowBox[{"x", ",", "y"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(OrderedPP lcm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:62103041", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656822622615017*^9, 3.6568226255010223`*^9}},
 CellTags->{
  "ID:62103041", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP lcm"},
 CellID->62103041],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", 
    UnderscriptBox["\[VerticalBar]", "N"], "y"}], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", 
    UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "y"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(OrderedPP |)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:229667832", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568226080601916`*^9, 3.6568226113985977`*^9}, {
  3.656822852378622*^9, 3.6568228809578724`*^9}, {3.6568285464224257`*^9, 
  3.6568285496516314`*^9}},
 CellTags->{
  "ID:229667832", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP |"},
 CellID->229667832],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["deg", "N"], "[", "x", "]"}], ":=", 
  RowBox[{
   UnderscriptBox["deg", "\[ScriptCapitalT]"], "[", "x", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(OrderedPP deg)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:687245437", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568227592776585`*^9, 3.6568227939097195`*^9}, {
  3.656828564003657*^9, 3.6568285698536673`*^9}},
 CellTags->{
  "ID:687245437", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP deg"},
 CellID->687245437],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"x", 
    UnderscriptBox["\[Succeeds]", "N"], "y"}], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{"R", "[", 
   RowBox[{"x", ",", "y"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(OrderedPP \[Succeeds])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:353388395", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656822633347836*^9, 3.6568226483862624`*^9}},
 CellTags->{
  "ID:353388395", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP \[Succeeds]"},
 CellID->353388395],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["<<", "N"], 
    RowBox[{"[", "x", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], ":=", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["<<", "\[ScriptCapitalT]"], 
    RowBox[{"[", "x", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(OrderedPP <<)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:144609801", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568228041433372`*^9, 3.6568228091977463`*^9}},
 CellTags->{
  "ID:144609801", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP <<"},
 CellID->144609801],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[">>", "N"], 
    RowBox[{"[", "x", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], ":=", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[">>", "\[ScriptCapitalT]"], 
    RowBox[{"[", "x", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(OrderedPP >>)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:189296520", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568228143301554`*^9, 3.6568228173409605`*^9}},
 CellTags->{
  "ID:189296520", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP >>"},
 CellID->189296520],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->170554027]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->887551028]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Commutative Power-Product Domains and Dickson Domains (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.6568421379180527`*^9, 3.6568421559516845`*^9}, {
  3.656910236318836*^9, 3.6569102404996433`*^9}, {3.656913297832418*^9, 
  3.6569133018572245`*^9}},
 CellID->75242152],

Cell["", "OpenEnvironment",
 CellID->67119962],

Cell[CellGroupData[{

Cell["\<\
Theorem: Ordered Power-Products are Commutative Power-Products\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656842165342901*^9, 3.6568421870269394`*^9}},
 CellID->40275436],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"isExpPPDomain", "[", "\[ScriptCapitalT]", "]"}], ",", 
   "R"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 29953053}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568423609984446`*^9, 3.656842365943653*^9}, {
  3.656842818874849*^9, 3.6568428192648497`*^9}},
 CellTags->{
  "ID:29953053", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->29953053],

Cell[BoxData[
 RowBox[{"isExpPPDomain", "[", 
  RowBox[{"OrderedPP", "[", 
   RowBox[{"\[ScriptCapitalT]", ",", "R"}], "]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(OrderedPP isExpPPDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:309264225", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568433920354557`*^9, 3.6568434005842705`*^9}},
 CellTags->{
  "ID:309264225", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isExpPPDomain"},
 CellID->309264225],

Cell[BoxData[
 RowBox[{"isDicksonDomain", "[", 
  RowBox[{
   RowBox[{"OrderedPP", "[", 
    RowBox[{"\[ScriptCapitalT]", ",", "R"}], "]"}], ",", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[VerticalBar]", 
      RowBox[{"OrderedPP", "[", 
       RowBox[{"\[ScriptCapitalT]", ",", "R"}], "]"}]],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}]}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(OrderedPP isDicksonDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:499712177", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656842387674492*^9, 3.656842396550907*^9}, {
  3.6569104058911343`*^9, 3.6569104215223613`*^9}},
 CellTags->{
  "ID:499712177", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isDicksonDomain"},
 CellID->499712177],

Cell[BoxData[
 RowBox[{
  RowBox[{"isTermOrder", "[", 
   RowBox[{"R", ",", "\[ScriptCapitalT]"}], "]"}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 275069792}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6569114094936967`*^9, 3.656911415265707*^9}},
 CellTags->{
  "ID:275069792", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->275069792],

Cell[BoxData[
 RowBox[{"isTermOrder", "[", 
  RowBox[{"R", ",", 
   RowBox[{"OrderedPP", "[", 
    RowBox[{"\[ScriptCapitalT]", ",", "R"}], "]"}]}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(OrderedPP isTermOrder)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:834856233", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6569114213029175`*^9, 3.656911436419344*^9}},
 CellTags->{
  "ID:834856233", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isTermOrder"},
 CellID->834856233],

Cell[BoxData[
 RowBox[{"isCommPPDomain", "[", 
  RowBox[{"OrderedPP", "[", 
   RowBox[{"\[ScriptCapitalT]", ",", "R"}], "]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(OrderedPP isCommPPDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:260893370", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6568422024241657`*^9, 3.656842243733039*^9}, {
  3.656842364430451*^9, 3.6568423812628803`*^9}, {3.6569106726214027`*^9, 
  3.65691068269902*^9}, {3.656911412286102*^9, 3.6569114133937035`*^9}},
 CellTags->{
  "ID:260893370", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isCommPPDomain"},
 CellID->260893370],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->163286488]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->981953859],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(OrderedPP isExpPPDomain)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(OrderedPP isExpPPDomain)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             StyleBox["R", "ExpressionVariable"]}}]], 
         RowBox[{"isExpPPDomain", "[", 
           RowBox[{"OrderedPP", "[", 
             RowBox[{
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], ",", 
               StyleBox["R", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             StyleBox["R", "ExpressionVariable"]}}]], 
         RowBox[{"isExpPPDomain", "[", 
           RowBox[{"OrderedPP", "[", 
             RowBox[{
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], ",", 
               StyleBox["R", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:309264225"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p309264225-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p309264225-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:309264225", "Proof|ID:309264225-1"},
 CellID->51410204],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p309264225-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->3099999]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(OrderedPP isDicksonDomain)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(OrderedPP isDicksonDomain)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             StyleBox["R", "ExpressionVariable"]}}]], 
         RowBox[{"isDicksonDomain", "[", 
           RowBox[{
             RowBox[{"OrderedPP", "[", 
               RowBox[{
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], ",", 
                 StyleBox["R", "ExpressionVariable"]}], "]"}], ",", 
             UnderscriptBox["\[VerticalBar]", 
              RowBox[{"OrderedPP", "[", 
                RowBox[{
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], ",", 
                  StyleBox["R", "ExpressionVariable"]}], "]"}]]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             StyleBox["R", "ExpressionVariable"]}}]], 
         RowBox[{"isDicksonDomain", "[", 
           RowBox[{
             RowBox[{"OrderedPP", "[", 
               RowBox[{
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], ",", 
                 StyleBox["R", "ExpressionVariable"]}], "]"}], ",", 
             UnderscriptBox["\[VerticalBar]", 
              RowBox[{"OrderedPP", "[", 
                RowBox[{
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], ",", 
                  StyleBox["R", "ExpressionVariable"]}], "]"}]]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:499712177"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p499712177-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p499712177-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:499712177", "Proof|ID:499712177-1"},
 CellID->50025897],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p499712177-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->631880945]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(OrderedPP isTermOrder)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(OrderedPP isTermOrder)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             StyleBox["R", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"isTermOrder", "[", 
             RowBox[{
               StyleBox["R", "ExpressionVariable"], ",", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], "]"}], 
           "\[Implies]", 
           RowBox[{"isTermOrder", "[", 
             RowBox[{
               StyleBox["R", "ExpressionVariable"], ",", 
               RowBox[{"OrderedPP", "[", 
                 RowBox[{
                   StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], ",", 
                   StyleBox["R", "ExpressionVariable"]}], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             StyleBox["R", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"isTermOrder", "[", 
             RowBox[{
               StyleBox["R", "ExpressionVariable"], ",", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], "]"}], 
           "\[Implies]", 
           RowBox[{"isTermOrder", "[", 
             RowBox[{
               StyleBox["R", "ExpressionVariable"], ",", 
               RowBox[{"OrderedPP", "[", 
                 RowBox[{
                   StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], ",", 
                   StyleBox["R", "ExpressionVariable"]}], "]"}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:834856233"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p834856233-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p834856233-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:834856233", "Proof|ID:834856233-1"},
 CellID->65135779],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p834856233-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->687624990]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(OrderedPP isCommPPDomain)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(OrderedPP isCommPPDomain)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             StyleBox["R", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"isTermOrder", "[", 
             RowBox[{
               StyleBox["R", "ExpressionVariable"], ",", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], "]"}], 
           "\[Implies]", 
           RowBox[{"isCommPPDomain", "[", 
             RowBox[{"OrderedPP", "[", 
               RowBox[{
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], ",", 
                 StyleBox["R", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             StyleBox["R", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"isTermOrder", "[", 
             RowBox[{
               StyleBox["R", "ExpressionVariable"], ",", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], "]"}], 
           "\[Implies]", 
           RowBox[{"isCommPPDomain", "[", 
             RowBox[{"OrderedPP", "[", 
               RowBox[{
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], ",", 
                 StyleBox["R", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:260893370"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p260893370-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p260893370-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:260893370", "Proof|ID:260893370-1"},
 CellID->97936963],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p260893370-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->264288389]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Concrete Term Orders (\[Checkmark])", "Section",
 CellChangeTimes->{{3.656929442643489*^9, 3.656929447120697*^9}, {
  3.657013536048465*^9, 3.6570135404788723`*^9}},
 CellID->46760114],

Cell[CellGroupData[{

Cell["Lexicographic (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.656929466573931*^9, 3.6569294692103357`*^9}, {
  3.657013235358136*^9, 3.657013239694944*^9}},
 CellID->19116656],

Cell[CellGroupData[{

Cell["Definition", "Subsubsection",
 CellChangeTimes->{{3.656929919583127*^9, 3.6569299234207335`*^9}, 
   3.656999336387774*^9},
 CellID->452220060],

Cell["", "OpenEnvironment",
 CellID->883840617],

Cell[CellGroupData[{

Cell["Definition: Lexicographic Term Order", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6569294918927755`*^9, 3.6569294980235863`*^9}},
 CellID->79250563],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"\[ScriptCapitalT]", ",", "x", ",", "y"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"Lex", "[", "\[ScriptCapitalT]", "]"}], "[", 
     RowBox[{"x", ",", "y"}], "]"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
        UnderscriptBox["dim", "\[ScriptCapitalT]"]}]], 
      RowBox[{"\[And]", 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{
            RowBox[{
             UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
             RowBox[{"x", ",", "i"}], "]"}], ">", 
            RowBox[{
             UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
             RowBox[{"y", ",", "i"}], "]"}]}]},
          {
           RowBox[{
            UnderscriptBox["\[ForAll]", 
             RowBox[{
              RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
              RowBox[{"i", "-", "1"}]}]], 
            RowBox[{
             RowBox[{
              UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
              RowBox[{"x", ",", "j"}], "]"}], "\[Equal]", 
             RowBox[{
              UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
              RowBox[{"y", ",", "j"}], "]"}]}]}]}
         }]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Lex)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:223902051", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656929503421196*^9, 3.656929567428108*^9}},
 CellTags->{
  "ID:223902051", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex"},
 CellID->223902051],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->86850538]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->119543112]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Lexicographic Order is Term Order (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.6569299313299475`*^9, 3.6569299420159664`*^9}, {
  3.6569977877642527`*^9, 3.656997788450654*^9}, {3.6569993430489855`*^9, 
  3.656999347494993*^9}, {3.657013227074522*^9, 3.6570132320197306`*^9}},
 CellID->585882636],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "isExpPPDomain", "[", "\[ScriptCapitalT]", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 376155678}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6569300980806403`*^9, 3.656930104883252*^9}},
 CellTags->{
  "ID:376155678", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->376155678],

Cell["", "OpenEnvironment",
 CellID->175815204],

Cell[CellGroupData[{

Cell["Lemma: Lex Satisifies All Properties of Term Orders", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656930077285804*^9, 3.6569300941650333`*^9}, {
   3.656997443033847*^9, 3.6569974436578484`*^9}, 3.657002306710392*^9},
 CellID->104761298],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", 
    RowBox[{"s", ",", "t", ",", "u"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 310375877}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6569301131668673`*^9, 3.6569301219808826`*^9}},
 CellTags->{
  "ID:310375877", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->310375877],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Lex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"s", ",", "t"}], "]"}], "\[Or]", 
  RowBox[{
   RowBox[{"Lex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"t", ",", "s"}], "]"}], "\[Or]", 
  RowBox[{"s", "\[Equal]", "t"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Lex total)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:463764498", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656930131340899*^9, 3.6569301561449423`*^9}},
 CellTags->{
  "ID:463764498", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex total"},
 CellID->463764498],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Lex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"s", ",", "t"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{
       RowBox[{"Lex", "[", "\[ScriptCapitalT]", "]"}], "[", 
       RowBox[{"t", ",", "s"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{"s", "\[NotEqual]", "t"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Lex asymmetric)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:6465370", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656930174069374*^9, 3.65693019445861*^9}},
 CellTags->{
  "ID:6465370", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex asymmetric"},
 CellID->6465370],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"Lex", "[", "\[ScriptCapitalT]", "]"}], "[", 
     RowBox[{"s", ",", "t"}], "]"}], "\[And]", 
    RowBox[{
     RowBox[{"Lex", "[", "\[ScriptCapitalT]", "]"}], "[", 
     RowBox[{"t", ",", "u"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"Lex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"s", ",", "u"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Lex transitive)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:78173747", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656930203132225*^9, 3.656930213537443*^9}},
 CellTags->{
  "ID:78173747", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex transitive"},
 CellID->78173747],

Cell[BoxData[
 RowBox[{
  RowBox[{"s", "\[NotEqual]", 
   UnderscriptBox["1", "\[ScriptCapitalT]"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"Lex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"s", ",", 
    UnderscriptBox["1", "\[ScriptCapitalT]"]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Lex 1 min)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:341999039", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656997452284663*^9, 3.6569974694914937`*^9}},
 CellTags->{
  "ID:341999039", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex 1 min"},
 CellID->341999039],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Lex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"s", ",", "t"}], "]"}], "\[Implies]", 
  RowBox[{
   RowBox[{"Lex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{
    RowBox[{"s", 
     UnderscriptBox["*", "\[ScriptCapitalT]"], "u"}], ",", 
    RowBox[{"t", 
     UnderscriptBox["*", "\[ScriptCapitalT]"], "u"}]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Lex *)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:34516336", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6569974773383074`*^9, 3.656997518288379*^9}},
 CellTags->{
  "ID:34516336", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex *"},
 CellID->34516336],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->51206845]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->212901590],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Lex total)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Lex total)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"Lex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Or]", 
           RowBox[{
             RowBox[{"Lex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], ",", 
               StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Or]", 
           RowBox[{
             StyleBox["s", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"Lex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Or]", 
           RowBox[{
             RowBox[{"Lex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], ",", 
               StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Or]", 
           RowBox[{
             StyleBox["s", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:463764498"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p463764498-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p463764498-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:463764498", "Proof|ID:463764498-1"},
 CellID->355435549],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p463764498-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.657002852503151*^9, 3.6570028544843545`*^9}},
 CellID->81470365]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Lex asymmetric)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Lex asymmetric)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"Lex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   RowBox[{"Lex", "[", 
                    StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
                    "]"}], "[", 
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox[
               "t", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"Lex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   RowBox[{"Lex", "[", 
                    StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
                    "]"}], "[", 
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox[
               "t", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:6465370"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p6465370-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p6465370-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:6465370", "Proof|ID:6465370-1"},
 CellID->258075395],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p6465370-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.6570034008787146`*^9, 3.6570034029067183`*^9}},
 CellID->146173952]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Lex transitive)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Lex transitive)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{"Lex", "[", 
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
               "[", 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               RowBox[{"Lex", "[", 
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
               "[", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], ",", 
                 StyleBox["u", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"Lex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["u", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{"Lex", "[", 
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
               "[", 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               RowBox[{"Lex", "[", 
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
               "[", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], ",", 
                 StyleBox["u", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"Lex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["u", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:78173747"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p78173747-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p78173747-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:78173747", "Proof|ID:78173747-1"},
 CellID->631231714],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p78173747-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.657004923558589*^9, 3.6570049254773927`*^9}},
 CellID->247058823]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Lex 1 min)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Lex 1 min)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["s", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["1", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{"Lex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               UnderscriptBox["1", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["s", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["1", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{"Lex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               UnderscriptBox["1", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:341999039"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p341999039-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p341999039-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:341999039", "Proof|ID:341999039-1"},
 CellID->366853708],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p341999039-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->235356693]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Lex *)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Lex *)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"Lex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"Lex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["u", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["u", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"Lex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"Lex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["u", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["u", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:34516336"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p34516336-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p34516336-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:34516336", "Proof|ID:34516336-1"},
 CellID->188021315],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p34516336-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.657006272582959*^9, 3.657006274486163*^9}},
 CellID->11866489]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->15806341],

Cell[CellGroupData[{

Cell["Theorem: Lex is Term Order", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656999501295664*^9, 3.6569995053360705`*^9}},
 CellID->12908456],

Cell[BoxData[
 RowBox[{"isTermOrder", "[", 
  RowBox[{
   RowBox[{"Lex", "[", "\[ScriptCapitalT]", "]"}], ",", "\[ScriptCapitalT]"}],
   "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(Lex isTermOrder)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:414199224", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6569995092048774`*^9, 3.6569995189704943`*^9}},
 CellTags->{
  "ID:414199224", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex isTermOrder"},
 CellID->414199224],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->118815614]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->11715767],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(Lex isTermOrder)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(Lex isTermOrder)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isTermOrder", "[", 
           RowBox[{
             RowBox[{"Lex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             ",", 
             StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isTermOrder", "[", 
           RowBox[{
             RowBox[{"Lex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             ",", 
             StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:414199224"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p414199224-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p414199224-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:414199224", "Proof|ID:414199224-1"},
 CellID->87459975],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p414199224-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->920517951]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Degree-Lexicographic (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6569993242509527`*^9, 3.6569993298201623`*^9}, {
  3.657013389518607*^9, 3.6570133936526146`*^9}},
 CellID->417561155],

Cell[CellGroupData[{

Cell["Definition", "Subsubsection",
 CellChangeTimes->{{3.656999355123407*^9, 3.6569993566522093`*^9}},
 CellID->46310993],

Cell["", "OpenEnvironment",
 CellID->327096997],

Cell[CellGroupData[{

Cell["Definition: Degree-Lexicographic Term Order", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656999388335865*^9, 3.65699939666628*^9}},
 CellID->210690636],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"\[ScriptCapitalT]", ",", "x", ",", "y"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"DegLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
     RowBox[{"x", ",", "y"}], "]"}], 
    TagBox[
     RowBox[{
     ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
      "\[DoubleLongLeftRightArrow]"}],
     Identity,
     SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["let", 
       RowBox[{
        RowBox[{"dx", "=", 
         RowBox[{
          UnderscriptBox["deg", "\[ScriptCapitalT]"], "[", "x", "]"}]}], ",", 
        
        RowBox[{"dy", "=", 
         RowBox[{
          UnderscriptBox["deg", "\[ScriptCapitalT]"], "[", "y", "]"}]}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[Or]", 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{"dx", ">", "dy"}]},
           {
            RowBox[{
             RowBox[{"dx", "\[Equal]", "dy"}], "\[And]", 
             RowBox[{
              RowBox[{"Lex", "[", "\[ScriptCapitalT]", "]"}], "[", 
              RowBox[{"x", ",", "y"}], "]"}]}]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DegLex)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:116320630", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6569994037486925`*^9, 3.6569994658212013`*^9}},
 CellTags->{
  "ID:116320630", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex"},
 CellID->116320630],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->208585242]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->49927192]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Degree-Lexicographic Order is Term Order (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.656999361691018*^9, 3.6569993709262342`*^9}, {
  3.6570133825297947`*^9, 3.6570133867574024`*^9}},
 CellID->122425200],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "isExpPPDomain", "[", "\[ScriptCapitalT]", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 106786369}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656999606736249*^9, 3.6569996136002607`*^9}},
 CellTags->{
  "ID:106786369", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->106786369],

Cell["", "OpenEnvironment",
 CellID->56827951],

Cell[CellGroupData[{

Cell["Lemma: DegLex Satisfies All Properties of Term Orders", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6569995554277587`*^9, 3.656999570996586*^9}, 
   3.6570023104231987`*^9},
 CellID->168509974],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", 
    RowBox[{"s", ",", "t", ",", "u"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 43019543}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6569995769869967`*^9, 3.65699958492741*^9}},
 CellTags->{
  "ID:43019543", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->43019543],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"DegLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"s", ",", "t"}], "]"}], "\[Or]", 
  RowBox[{
   RowBox[{"DegLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"t", ",", "s"}], "]"}], "\[Or]", 
  RowBox[{"s", "\[Equal]", "t"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DegLex total)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:75557233", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6569996215718746`*^9, 3.656999626813484*^9}},
 CellTags->{
  "ID:75557233", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex total"},
 CellID->75557233],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"DegLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"s", ",", "t"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{
       RowBox[{"DegLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
       RowBox[{"t", ",", "s"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{"s", "\[NotEqual]", "t"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DegLex asymmetric)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:49531328", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656999634051897*^9, 3.6569996396679068`*^9}},
 CellTags->{
  "ID:49531328", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex asymmetric"},
 CellID->49531328],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"DegLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
     RowBox[{"s", ",", "t"}], "]"}], "\[And]", 
    RowBox[{
     RowBox[{"DegLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
     RowBox[{"t", ",", "u"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"DegLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"s", ",", "u"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DegLex transitive)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:260398187", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656999647951521*^9, 3.6569996578419385`*^9}},
 CellTags->{
  "ID:260398187", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex transitive"},
 CellID->260398187],

Cell[BoxData[
 RowBox[{
  RowBox[{"s", "\[NotEqual]", 
   UnderscriptBox["1", "\[ScriptCapitalT]"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"DegLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"s", ",", 
    UnderscriptBox["1", "\[ScriptCapitalT]"]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DegLex 1 min)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:59133109", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656999663972749*^9, 3.6569996664219537`*^9}},
 CellTags->{
  "ID:59133109", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex 1 min"},
 CellID->59133109],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"DegLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"s", ",", "t"}], "]"}], "\[Implies]", 
  RowBox[{
   RowBox[{"DegLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{
    RowBox[{"s", 
     UnderscriptBox["*", "\[ScriptCapitalT]"], "u"}], ",", 
    RowBox[{"t", 
     UnderscriptBox["*", "\[ScriptCapitalT]"], "u"}]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DegLex *)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:572110615", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6569996717259626`*^9, 3.6569996768583717`*^9}},
 CellTags->{
  "ID:572110615", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex *"},
 CellID->572110615],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->467550442]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->213596470],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DegLex total)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DegLex total)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"DegLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Or]", 
           RowBox[{
             RowBox[{"DegLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], ",", 
               StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Or]", 
           RowBox[{
             StyleBox["s", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"DegLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Or]", 
           RowBox[{
             RowBox[{"DegLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], ",", 
               StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Or]", 
           RowBox[{
             StyleBox["s", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:75557233"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p75557233-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p75557233-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:75557233", "Proof|ID:75557233-1"},
 CellID->210478017],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p75557233-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{3.657006903745468*^9, 3.6570070521797285`*^9},
 CellID->35178206]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DegLex asymmetric)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DegLex asymmetric)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"DegLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   RowBox[{"DegLex", "[", 
                    StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
                    "]"}], "[", 
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox[
               "t", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"DegLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   RowBox[{"DegLex", "[", 
                    StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
                    "]"}], "[", 
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox[
               "t", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:49531328"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p49531328-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p49531328-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:49531328", "Proof|ID:49531328-1"},
 CellID->563764769],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p49531328-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.6570072453392677`*^9, 3.6570072473828716`*^9}},
 CellID->334172145]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DegLex transitive)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DegLex transitive)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{"DegLex", "[", 
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
               "[", 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               RowBox[{"DegLex", "[", 
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
               "[", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], ",", 
                 StyleBox["u", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"DegLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["u", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{"DegLex", "[", 
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
               "[", 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               RowBox[{"DegLex", "[", 
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
               "[", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], ",", 
                 StyleBox["u", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"DegLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["u", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:260398187"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p260398187-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p260398187-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:260398187", "Proof|ID:260398187-1"},
 CellID->100109944],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p260398187-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.657007773824396*^9, 3.6570077757587996`*^9}},
 CellID->450035939]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DegLex 1 min)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DegLex 1 min)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["s", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["1", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{"DegLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               UnderscriptBox["1", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["s", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["1", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{"DegLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               UnderscriptBox["1", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:59133109"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p59133109-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p59133109-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:59133109", "Proof|ID:59133109-1"},
 CellID->171486032],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p59133109-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{3.65701248029821*^9},
 CellID->51248171]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DegLex *)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DegLex *)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"DegLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"DegLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["u", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["u", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"DegLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"DegLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["u", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["u", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:572110615"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p572110615-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p572110615-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:572110615", "Proof|ID:572110615-1"},
 CellID->427513857],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p572110615-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.6570126841915684`*^9, 3.657012685860771*^9}},
 CellID->171210370]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->50768423],

Cell[CellGroupData[{

Cell["Theorem: DegLex is Term Order", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.657001933698137*^9, 3.6570019393453465`*^9}},
 CellID->584850280],

Cell[BoxData[
 RowBox[{"isTermOrder", "[", 
  RowBox[{
   RowBox[{"DegLex", "[", "\[ScriptCapitalT]", "]"}], ",", 
   "\[ScriptCapitalT]"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DegLex isTermOrder)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:463127740", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657001943120553*^9, 3.6570019530265703`*^9}},
 CellTags->{
  "ID:463127740", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex isTermOrder"},
 CellID->463127740],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->93774533]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->214012042],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DegLex isTermOrder)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DegLex isTermOrder)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isTermOrder", "[", 
           RowBox[{
             RowBox[{"DegLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             ",", 
             StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isTermOrder", "[", 
           RowBox[{
             RowBox[{"DegLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             ",", 
             StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:463127740"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p463127740-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p463127740-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:463127740", "Proof|ID:463127740-1"},
 CellID->102580793],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p463127740-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->73039304]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Degree-Reverse-Lexicographic (\[Checkmark])\
\>", "Subsection",
 CellChangeTimes->{{3.657001978407815*^9, 3.6570019853654275`*^9}, {
  3.6570135267976484`*^9, 3.6570135320704575`*^9}},
 CellID->187433264],

Cell[CellGroupData[{

Cell["Definitions", "Subsubsection",
 CellChangeTimes->{{3.6570019923698397`*^9, 3.6570019995458527`*^9}, 
   3.6570028778219957`*^9},
 CellID->290355525],

Cell["", "OpenEnvironment",
 CellID->337107021],

Cell[CellGroupData[{

Cell["\<\
Definition: RevLex and Degree-Reverse-Lexicographic Term Order\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6570020462679343`*^9, 3.6570020623515625`*^9}},
 CellID->513549404],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"\[ScriptCapitalT]", ",", "x", ",", "y"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 267616510}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6570020674683714`*^9, 3.657002071056378*^9}},
 CellTags->{
  "ID:267616510", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->267616510],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"x", ",", "y"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
      UnderscriptBox["dim", "\[ScriptCapitalT]"]}]], 
    RowBox[{"\[And]", 
     RowBox[{"\[Piecewise]", GridBox[{
        {
         RowBox[{
          RowBox[{
           UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
           RowBox[{"y", ",", "i"}], "]"}], ">", 
          RowBox[{
           UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
           RowBox[{"x", ",", "i"}], "]"}]}]},
        {
         RowBox[{
          UnderscriptBox["\[ForAll]", 
           RowBox[{
            RowBox[{"j", "=", 
             RowBox[{"i", "+", "1"}]}], ",", "\[Ellipsis]", ",", 
            UnderscriptBox["dim", "\[ScriptCapitalT]"]}]], 
          RowBox[{
           RowBox[{
            UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
            RowBox[{"x", ",", "j"}], "]"}], "\[Equal]", 
           RowBox[{
            UnderscriptBox["exp", "\[ScriptCapitalT]"], "[", 
            RowBox[{"y", ",", "j"}], "]"}]}]}]}
       }]}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RevLex)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1929370", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657002081274396*^9, 3.657002104237636*^9}},
 CellTags->{
  "ID:1929370", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex"},
 CellID->1929370],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"DegRevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"x", ",", "y"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[DoubleLongLeftRightArrow]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["let", 
     RowBox[{
      RowBox[{"dx", "=", 
       RowBox[{
        UnderscriptBox["deg", "\[ScriptCapitalT]"], "[", "x", "]"}]}], ",", 
      RowBox[{"dy", "=", 
       RowBox[{
        UnderscriptBox["deg", "\[ScriptCapitalT]"], "[", "y", "]"}]}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Or]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{"dx", ">", "dy"}]},
         {
          RowBox[{
           RowBox[{"dx", "\[Equal]", "dy"}], "\[And]", 
           RowBox[{
            RowBox[{"RevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
            RowBox[{"x", ",", "y"}], "]"}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DegRevLex)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:116917210", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6570021511157184`*^9, 3.6570021628001394`*^9}},
 CellTags->{
  "ID:116917210", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex"},
 CellID->116917210],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->494887762]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->50267427]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Degree-Reverse-Lexicographic Order is Term Order (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.6570020041478605`*^9, 3.6570020220098915`*^9}, {
  3.6570135183268337`*^9, 3.657013523178442*^9}},
 CellID->148960609],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "isExpPPDomain", "[", "\[ScriptCapitalT]", "]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 431887307}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657002319221614*^9, 3.657002326023226*^9}},
 CellTags->{
  "ID:431887307", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->431887307],

Cell["", "OpenEnvironment",
 CellID->181550980],

Cell[CellGroupData[{

Cell["Lemma: RevLex Satisfies Almost All Properties of Term Orders", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.657002278879943*^9, 3.6570023008915815`*^9}},
 CellID->417200827],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", 
    RowBox[{"s", ",", "t", ",", "u"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 204943512}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657002332528437*^9, 3.657002341841654*^9}},
 CellTags->{
  "ID:204943512", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->204943512],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"s", ",", "t"}], "]"}], "\[Or]", 
  RowBox[{
   RowBox[{"RevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"t", ",", "s"}], "]"}], "\[Or]", 
  RowBox[{"s", "\[Equal]", "t"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RevLex total)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:781115991", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6570023599688854`*^9, 3.657002386364132*^9}},
 CellTags->{
  "ID:781115991", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex total"},
 CellID->781115991],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"s", ",", "t"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{
       RowBox[{"RevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
       RowBox[{"t", ",", "s"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{"s", "\[NotEqual]", "t"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RevLex asymmetric)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2983958", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657002366645697*^9, 3.6570023918241415`*^9}},
 CellTags->{
  "ID:2983958", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex asymmetric"},
 CellID->2983958],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"RevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
     RowBox[{"s", ",", "t"}], "]"}], "\[And]", 
    RowBox[{
     RowBox[{"RevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
     RowBox[{"t", ",", "u"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"RevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"s", ",", "u"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RevLex transitive)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:240564679", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657002374383311*^9, 3.6570024009813576`*^9}},
 CellTags->{
  "ID:240564679", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex transitive"},
 CellID->240564679],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"s", ",", "t"}], "]"}], "\[Implies]", 
  RowBox[{
   RowBox[{"RevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{
    RowBox[{"s", 
     UnderscriptBox["*", "\[ScriptCapitalT]"], "u"}], ",", 
    RowBox[{"t", 
     UnderscriptBox["*", "\[ScriptCapitalT]"], "u"}]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(RevLex *)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:299677825", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657002380716922*^9, 3.657002406940568*^9}},
 CellTags->{
  "ID:299677825", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex *"},
 CellID->299677825],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->92970195]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->522695883],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(RevLex total)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(RevLex total)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"RevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Or]", 
           RowBox[{
             RowBox[{"RevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], ",", 
               StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Or]", 
           RowBox[{
             StyleBox["s", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"RevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Or]", 
           RowBox[{
             RowBox[{"RevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], ",", 
               StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Or]", 
           RowBox[{
             StyleBox["s", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:781115991"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p781115991-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p781115991-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:781115991", "Proof|ID:781115991-1"},
 CellID->124604986],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p781115991-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->126314747]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(RevLex asymmetric)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(RevLex asymmetric)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"RevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   RowBox[{"RevLex", "[", 
                    StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
                    "]"}], "[", 
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox[
               "t", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"RevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   RowBox[{"RevLex", "[", 
                    StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
                    "]"}], "[", 
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox[
               "t", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2983958"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2983958-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2983958-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2983958", "Proof|ID:2983958-1"},
 CellID->83920353],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2983958-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->206111400]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(RevLex transitive)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(RevLex transitive)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{"RevLex", "[", 
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
               "[", 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               RowBox[{"RevLex", "[", 
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
               "[", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], ",", 
                 StyleBox["u", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"RevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["u", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{"RevLex", "[", 
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
               "[", 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               RowBox[{"RevLex", "[", 
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
               "[", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], ",", 
                 StyleBox["u", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"RevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["u", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:240564679"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p240564679-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p240564679-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:240564679", "Proof|ID:240564679-1"},
 CellID->163719088],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p240564679-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->40115550]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(RevLex *)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(RevLex *)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"RevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"RevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["u", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["u", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"RevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"RevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["u", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["u", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:299677825"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p299677825-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p299677825-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:299677825", "Proof|ID:299677825-1"},
 CellID->464155749],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p299677825-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->805072015]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->97289005],

Cell[CellGroupData[{

Cell["Lemma: DegRevLex Satisfies All Properties of Term Orders", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6570067538292046`*^9, 3.657006769507232*^9}},
 CellID->39640643],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", 
    RowBox[{"s", ",", "t", ",", "u"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 449289304}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6570067737192397`*^9, 3.6570067827516556`*^9}},
 CellTags->{
  "ID:449289304", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->449289304],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"DegRevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"s", ",", "t"}], "]"}], "\[Or]", 
  RowBox[{
   RowBox[{"DegRevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"t", ",", "s"}], "]"}], "\[Or]", 
  RowBox[{"s", "\[Equal]", "t"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DegRevLex total)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:240154468", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.6570069182066936`*^9, {3.6570069621051702`*^9, 3.6570069661923776`*^9}},
 CellTags->{
  "ID:240154468", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex total"},
 CellID->240154468],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"DegRevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"s", ",", "t"}], "]"}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{
       RowBox[{"DegRevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
       RowBox[{"t", ",", "s"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{"s", "\[NotEqual]", "t"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DegRevLex asymmetric)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:808110452", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.6570069251487055`*^9, {3.6570069681891813`*^9, 3.6570069708411856`*^9}},
 CellTags->{
  "ID:808110452", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex asymmetric"},
 CellID->808110452],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{"DegRevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
     RowBox[{"s", ",", "t"}], "]"}], "\[And]", 
    RowBox[{
     RowBox[{"DegRevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
     RowBox[{"t", ",", "u"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"DegRevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"s", ",", "u"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DegRevLex transitive)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:224322863", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.657006932948719*^9, {3.6570069724635887`*^9, 3.6570069752247934`*^9}},
 CellTags->{
  "ID:224322863", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex transitive"},
 CellID->224322863],

Cell[BoxData[
 RowBox[{
  RowBox[{"s", "\[NotEqual]", 
   UnderscriptBox["1", "\[ScriptCapitalT]"]}], "\[Implies]", 
  RowBox[{
   RowBox[{"DegRevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"s", ",", 
    UnderscriptBox["1", "\[ScriptCapitalT]"]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DegRevLex 1 min)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:424804407", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6570069502335496`*^9, 3.6570069769719963`*^9}},
 CellTags->{
  "ID:424804407", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex 1 min"},
 CellID->424804407],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"DegRevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{"s", ",", "t"}], "]"}], "\[Implies]", 
  RowBox[{
   RowBox[{"DegRevLex", "[", "\[ScriptCapitalT]", "]"}], "[", 
   RowBox[{
    RowBox[{"s", 
     UnderscriptBox["*", "\[ScriptCapitalT]"], "u"}], ",", 
    RowBox[{"t", 
     UnderscriptBox["*", "\[ScriptCapitalT]"], "u"}]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DegRevLex *)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:308108885", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6570069570819616`*^9, 3.6570069807160034`*^9}},
 CellTags->{
  "ID:308108885", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex *"},
 CellID->308108885],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->951139]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->130964344],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DegRevLex total)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DegRevLex total)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"DegRevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Or]", 
           RowBox[{
             RowBox[{"DegRevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], ",", 
               StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Or]", 
           RowBox[{
             StyleBox["s", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"DegRevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Or]", 
           RowBox[{
             RowBox[{"DegRevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], ",", 
               StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Or]", 
           RowBox[{
             StyleBox["s", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:240154468"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p240154468-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p240154468-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:240154468", "Proof|ID:240154468-1"},
 CellID->162210883],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p240154468-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->17882270]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DegRevLex asymmetric)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DegRevLex asymmetric)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"DegRevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   RowBox[{"DegRevLex", "[", 
                    StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
                    "]"}], "[", 
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox[
               "t", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"DegRevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   RowBox[{"DegRevLex", "[", 
                    StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], 
                    "]"}], "[", 
                   RowBox[{
                    StyleBox["t", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 TagBox[")", "AutoParentheses"]}]}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], "\[NotEqual]", 
               StyleBox[
               "t", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:808110452"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p808110452-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p808110452-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:808110452", "Proof|ID:808110452-1"},
 CellID->14371058],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p808110452-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->223768150]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DegRevLex transitive)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DegRevLex transitive)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{"DegRevLex", "[", 
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
               "[", 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               RowBox[{"DegRevLex", "[", 
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
               "[", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], ",", 
                 StyleBox["u", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"DegRevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["u", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               RowBox[{"DegRevLex", "[", 
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
               "[", 
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[And]", 
             RowBox[{
               RowBox[{"DegRevLex", "[", 
                 StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
               "[", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], ",", 
                 StyleBox["u", "ExpressionVariable"]}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"DegRevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["u", "ExpressionVariable"]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:224322863"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p224322863-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p224322863-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:224322863", "Proof|ID:224322863-1"},
 CellID->483386676],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p224322863-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->255376339]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DegRevLex 1 min)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DegRevLex 1 min)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["s", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["1", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{"DegRevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               UnderscriptBox["1", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["s", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["1", 
              StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}], 
           "\[Implies]", 
           RowBox[{
             RowBox[{"DegRevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               UnderscriptBox["1", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:424804407"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p424804407-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p424804407-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:424804407", "Proof|ID:424804407-1"},
 CellID->138222520],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p424804407-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->421322267]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DegRevLex *)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DegRevLex *)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"DegRevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"DegRevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["u", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["u", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], "[", 
               StyleBox["u", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"DegRevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{"DegRevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             "[", 
             RowBox[{
               RowBox[{
                 StyleBox["s", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["u", "ExpressionVariable"]}], ",", 
               RowBox[{
                 StyleBox["t", "ExpressionVariable"], 
                 UnderscriptBox["*", 
                  StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]], 
                 StyleBox["u", "ExpressionVariable"]}]}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:308108885"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p308108885-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p308108885-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:308108885", "Proof|ID:308108885-1"},
 CellID->152180612],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p308108885-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->286784836]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->409591721],

Cell[CellGroupData[{

Cell["Theorem: DegRevLex is Term Order", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6570129791734867`*^9, 3.657012985787898*^9}},
 CellID->101138670],

Cell[BoxData[
 RowBox[{"isTermOrder", "[", 
  RowBox[{
   RowBox[{"DegRevLex", "[", "\[ScriptCapitalT]", "]"}], ",", 
   "\[ScriptCapitalT]"}], "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(DegRevLex isTermOrder)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:936963717", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.657012988627103*^9, 3.6570130026983275`*^9}},
 CellTags->{
  "ID:936963717", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex isTermOrder"},
 CellID->936963717],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->28325649]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->269251289],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(DegRevLex isTermOrder)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(DegRevLex isTermOrder)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isTermOrder", "[", 
           RowBox[{
             RowBox[{"DegRevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             ",", 
             StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isExpPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isTermOrder", "[", 
           RowBox[{
             RowBox[{"DegRevLex", "[", 
               StyleBox["\[ScriptCapitalT]", "ExpressionVariable"], "]"}], 
             ",", 
             StyleBox["\[ScriptCapitalT]", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:936963717"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p936963717-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p936963717-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:936963717", "Proof|ID:936963717-1"},
 CellID->319190197],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p936963717-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->379817477]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Definitions", "Section",
 CellChangeTimes->{{3.656844246218956*^9, 3.656844247903759*^9}},
 CellID->338638037],

Cell[CellGroupData[{

Cell["\<\
Power-Products Represented as Exponent-Tuples (\[Checkmark])\
\>", "Subsection",
 CellChangeTimes->{{3.645759502241544*^9, 3.645759505759798*^9}, {
  3.6457616135234213`*^9, 3.645761619417138*^9}, {3.645761808710765*^9, 
  3.6457618193733807`*^9}, {3.6569171582968025`*^9, 3.6569171624776106`*^9}},
 CellID->1801223056],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"n", "\[Element]", 
   SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 44994065}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656734787759634*^9, 3.656734792767243*^9}},
 CellTags->{
  "ID:44994065", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->44994065],

Cell[CellGroupData[{

Cell["Definitions", "Subsubsection",
 CellChangeTimes->{{3.656913934348736*^9, 3.656913936033539*^9}},
 CellID->494981958],

Cell["", "OpenEnvironment",
 CellID->1947010109],

Cell[CellGroupData[{

Cell["\<\
Definition: Monoid of Commutative Power-Products Represented as Exponent \
Tuples\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.593771603429861*^9, 3.593771617611169*^9}, {
   3.594446968548263*^9, 3.594446978341965*^9}, {3.645761182225607*^9, 
   3.6457611853575172`*^9}, 3.656822479406766*^9},
 CellID->1551875723],

Cell[BoxData[
 RowBox[{
  RowBox[{"PPTuples", "[", "n", "]"}], ":=", 
  UnderscriptBox["\[CapitalDelta]", "N"]}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 1997108144}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594447391269643*^9, 3.594447416311537*^9}, 
   3.594448041642974*^9, 3.645759847841826*^9},
 CellTags->{
  "ID:1997108144", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->1997108144],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"j", "\[Element]", "\[DoubleStruckCapitalN]"}], ",", "x", ",", 
   "y"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 608617422}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594447434298481*^9, 3.594447435349838*^9}, {
  3.594447525202013*^9, 3.594447535154495*^9}, {3.594447851812896*^9, 
  3.59444786744574*^9}, {3.594448310687039*^9, 3.594448326880621*^9}, {
  3.595057326902137*^9, 3.595057363317688*^9}},
 CellTags->{
  "ID:608617422", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->608617422],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Element]", "N"], 
   RowBox[{"[", "x", "]"}]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"\[And]", 
   RowBox[{"\[Piecewise]", GridBox[{
      {
       RowBox[{"isTuple", "[", "x", "]"}]},
      {
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "x", "\[RightBracketingBar]"}], 
        "\[Equal]", "n"}]},
      {
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]], 
        RowBox[{
         SubscriptBox["x", "i"], "\[Element]", 
         SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}]}
     }]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PPTuples \[Element])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:271011394", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594447444614265*^9, 3.59444750748188*^9}, 
   3.594447560223815*^9, {3.5944478742609*^9, 3.594447874816151*^9}, {
   3.595057340712281*^9, 3.595057345122045*^9}, {3.607338937311548*^9, 
   3.607338940497684*^9}},
 CellTags->{
  "ID:271011394", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples \[Element]"},
 CellID->271011394],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["1", "N"], ":=", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"0", 
    UnderscriptBox["|", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]]}], 
   "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PPTuples 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1820126795", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594447617596407*^9, 3.594447633084292*^9}},
 CellTags->{
  "ID:1820126795", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples 1"},
 CellID->1820126795],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["dim", "N"], ":=", "n"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PPTuples dim)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:281355476", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594461359983288*^9, 3.594461368066716*^9}},
 CellTags->{
  "ID:281355476", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples dim"},
 CellID->281355476],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["deg", "N"], "[", "x", "]"}], ":=", 
  RowBox[{
   UnderscriptBox[
    SubscriptBox["\[Sum]", 
     RowBox[{"\[DoubleStruckCapitalZ]IntL", "[", "0", "]"}]], 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]], 
   SubscriptBox["x", "i"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PPTuples deg)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2109519396", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.59444781924905*^9, {3.6569164100238886`*^9, 3.6569164132374945`*^9}, 
   3.666174183446455*^9},
 CellTags->{
  "ID:2109519396", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples deg"},
 CellID->2109519396],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["exp", "N"], "[", 
   RowBox[{"x", ",", "j"}], "]"}], ":=", 
  SubscriptBox["x", "j"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PPTuples exp)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1563833422", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594447828827002*^9, 3.594447847944791*^9}, {
  3.594447899163643*^9, 3.594447899574111*^9}},
 CellTags->{
  "ID:1563833422", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples exp"},
 CellID->1563833422],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["*", "N"], "y"}], ":=", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{
     SubscriptBox["x", "i"], "+", 
     SubscriptBox["y", "i"]}], 
    UnderscriptBox["|", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]]}], 
   "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PPTuples *)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1180963317", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594447647673286*^9, 3.594447668767288*^9}},
 CellTags->{
  "ID:1180963317", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples *"},
 CellID->1180963317],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["/", "N"], "y"}], ":=", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{
     SubscriptBox["x", "i"], "-", 
     SubscriptBox["y", "i"]}], 
    UnderscriptBox["|", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]]}], 
   "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PPTuples /)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1392702003", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.59444795789345*^9, 3.59444797158885*^9}},
 CellTags->{
  "ID:1392702003", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples /"},
 CellID->1392702003],

Cell[BoxData[
 RowBox[{
  RowBox[{"x", 
   UnderscriptBox["\[VerticalBar]", "N"], "y"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]], 
    RowBox[{
     SubscriptBox["x", "i"], "\[LessEqual]", 
     SubscriptBox["y", "i"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PPTuples \[VerticalBar])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1587875066", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594447675691046*^9, 3.594447705947123*^9}},
 CellTags->{
  "ID:1587875066", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples \[VerticalBar]"},
 CellID->1587875066],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lcm", "N"], "[", 
   RowBox[{"x", ",", "y"}], "]"}], ":=", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{
     SubscriptBox["max", "\[LessEqual]"], "[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       SubscriptBox["x", "i"], ",", 
       SubscriptBox["y", "i"]}], "\[RightAngleBracket]"}], "]"}], 
    UnderscriptBox["|", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", "n"}]]}], 
   "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PPTuples lcm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:803590963", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594450504568254*^9, 3.594450542588684*^9}, {
  3.607338963843925*^9, 3.607338966108112*^9}, {3.656747396326189*^9, 
  3.656747398073392*^9}, {3.656747492921559*^9, 3.6567475429040465`*^9}, {
  3.656747597862943*^9, 3.6567475996101465`*^9}},
 CellTags->{
  "ID:803590963", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples lcm"},
 CellID->803590963],

Cell["\<\
Power-product parsing and -pretty-printing is w.r.t. tuples of exponent \
vectors. Hence, since this functor stores power-products in exactly that way, \
there is no need to do anything.\
\>", "EnvironmentText",
 CellChangeTimes->{{3.635754858144409*^9, 3.6357549043177*^9}},
 CellID->425740884],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["<<", "N"], 
    RowBox[{"[", "x", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], ":=", "x"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PPTuples <<)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:106372713", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.635750548720985*^9, 3.635750557528364*^9}, {
   3.6357506090057096`*^9, 3.635750616099864*^9}, 3.635750649342054*^9, 
   3.635753633992157*^9, {3.63575367415648*^9, 3.635753702771008*^9}, 
   3.635754831350194*^9, {3.645763975508439*^9, 3.645763975993311*^9}},
 CellTags->{
  "ID:106372713", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples <<"},
 CellID->106372713],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[">>", "N"], 
    RowBox[{"[", "x", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], ":=", "x"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PPTuples >>)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1589980318", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.635750625193828*^9, 3.635750647078392*^9}, {
   3.635753587709537*^9, 3.635753595448114*^9}, 3.635754841092922*^9, {
   3.6457615660563307`*^9, 3.645761591553129*^9}},
 CellTags->{
  "ID:1589980318", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples >>"},
 CellID->1589980318],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->653212082]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->829024171]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
PPTuples Belong to ExpPPDomain (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.6569139444107533`*^9, 3.6569139825840206`*^9}, {
  3.656917135286762*^9, 3.65691713967037*^9}},
 CellID->89472363],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"\[ScriptCapitalT]", "=", 
   RowBox[{"PPTuples", "[", "n", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 61015536}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6567348750885873`*^9, 3.6567348850258045`*^9}},
 CellTags->{
  "ID:61015536", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->61015536],

Cell["", "OpenEnvironment",
 CellID->36911472],

Cell[CellGroupData[{

Cell["Lemma: PPTuples Satisfies all Properties of ExpPPDomains", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6569142359128656`*^9, 3.656914268641723*^9}},
 CellID->85270262],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "x"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "x", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[Element]", 
        RowBox[{"DomainTuples", "[", 
         SubscriptBox["\[DoubleStruckCapitalN]", "0"], "]"}]], 
       RowBox[{"[", "x", "]"}]}], "\[And]", 
      RowBox[{
       RowBox[{"\[LeftBracketingBar]", "x", "\[RightBracketingBar]"}], 
       "\[Equal]", "n"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] PPTuples alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:617800164", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6567351964659543`*^9, 3.656735228680011*^9}, {
  3.6570137225467925`*^9, 3.6570137266027994`*^9}},
 CellTags->{
  "ID:617800164", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PPTuples alternative", "IR"},
 CellID->617800164],

Cell[BoxData[
 RowBox[{"isDomain", "[", "\[ScriptCapitalT]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PPTuples isDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:41499157", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656735372372864*^9, 3.6567353770060716`*^9}},
 CellTags->{
  "ID:41499157", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples isDomain"},
 CellID->41499157],

Cell[BoxData[
 RowBox[{"hasDecisionPredicatePP", "[", "\[ScriptCapitalT]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PPTuples hasDecisionPredicate)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:142297858", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6569142988589764`*^9, 3.6569143098257957`*^9}},
 CellTags->{
  "ID:142297858", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasDecisionPredicate"},
 CellID->142297858],

Cell[BoxData[
 RowBox[{"hasExponentFunctionPP", "[", "\[ScriptCapitalT]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PPTuples hasExponentFunction)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:201916088", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656914326595825*^9, 3.656914334427039*^9}},
 CellTags->{
  "ID:201916088", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasExponentFunction"},
 CellID->201916088],

Cell[BoxData[
 RowBox[{"isExtensionalPP", "[", "\[ScriptCapitalT]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PPTuples isExtensional)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:102507049", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6569143371726437`*^9, 3.656914343116254*^9}},
 CellTags->{
  "ID:102507049", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples isExtensional"},
 CellID->102507049],

Cell[BoxData[
 RowBox[{"hasDegreeFunctionPP", "[", "\[ScriptCapitalT]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PPTuples hasDegreeFunction)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:93844102", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6569143472814617`*^9, 3.6569143606662846`*^9}},
 CellTags->{
  "ID:93844102", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasDegreeFunction"},
 CellID->93844102],

Cell[BoxData[
 RowBox[{"hasDivisibilityPP", "[", "\[ScriptCapitalT]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PPTuples hasDivisibility)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:34803991", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65691436369269*^9, 3.6569143707283025`*^9}, {
  3.6569167099052153`*^9, 3.656916710092416*^9}},
 CellTags->{
  "ID:34803991", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasDivisibility"},
 CellID->34803991],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->241614839]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->372933962],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] PPTuples alternative,[IR])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] PPTuples alternative,[IR])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             StyleBox["x", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PPTuples", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainTuples", "[", 
                    SubscriptBox["\[DoubleStruckCapitalN]", "0"], "]"}]], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["x", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], "\[Equal]", 
                 StyleBox["n", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}, {
             StyleBox["x", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PPTuples", "[", 
                  StyleBox["n", "ExpressionVariable"], "]"}]], "[", 
               StyleBox["x", "ExpressionVariable"], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"DomainTuples", "[", 
                    SubscriptBox["\[DoubleStruckCapitalN]", "0"], "]"}]], "[", 
                 StyleBox["x", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["x", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], "\[Equal]", 
                 StyleBox["n", "ExpressionVariable"]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:617800164"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p617800164-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p617800164-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:617800164", "Proof|ID:617800164-1"},
 CellID->427612524],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p617800164-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->712068464]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PPTuples isDomain)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PPTuples isDomain)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{"isDomain", "[", 
           RowBox[{"PPTuples", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{"isDomain", "[", 
           RowBox[{"PPTuples", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:41499157"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p41499157-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p41499157-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:41499157", "Proof|ID:41499157-1"},
 CellID->394527339],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p41499157-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->260522834]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PPTuples hasDecisionPredicate)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PPTuples hasDecisionPredicate)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{"hasDecisionPredicatePP", "[", 
           RowBox[{"PPTuples", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{"hasDecisionPredicatePP", "[", 
           RowBox[{"PPTuples", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:142297858"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p142297858-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p142297858-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:142297858", "Proof|ID:142297858-1"},
 CellID->706958866],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p142297858-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->471518847]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PPTuples hasExponentFunction)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PPTuples hasExponentFunction)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{"hasExponentFunctionPP", "[", 
           RowBox[{"PPTuples", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{"hasExponentFunctionPP", "[", 
           RowBox[{"PPTuples", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:201916088"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p201916088-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p201916088-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:201916088", "Proof|ID:201916088-1"},
 CellID->220074858],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p201916088-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->162708875]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PPTuples isExtensional)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PPTuples isExtensional)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{"isExtensionalPP", "[", 
           RowBox[{"PPTuples", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{"isExtensionalPP", "[", 
           RowBox[{"PPTuples", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:102507049"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p102507049-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p102507049-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:102507049", "Proof|ID:102507049-1"},
 CellID->628960870],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p102507049-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->439772086]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PPTuples hasDegreeFunction)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PPTuples hasDegreeFunction)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{"hasDegreeFunctionPP", "[", 
           RowBox[{"PPTuples", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{"hasDegreeFunctionPP", "[", 
           RowBox[{"PPTuples", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:93844102"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p93844102-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p93844102-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:93844102", "Proof|ID:93844102-1"},
 CellID->78679481],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p93844102-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->285701216]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PPTuples hasDivisibility)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PPTuples hasDivisibility)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{"hasDivisibilityPP", "[", 
           RowBox[{"PPTuples", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{"hasDivisibilityPP", "[", 
           RowBox[{"PPTuples", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:34803991"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p34803991-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p34803991-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:34803991", "Proof|ID:34803991-1"},
 CellID->42217056],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p34803991-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->294930096]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->79399544],

Cell[CellGroupData[{

Cell["Theorem: PPTuples is ExpPPDomain", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6569168369986386`*^9, 3.6569168452978535`*^9}, {
   3.656916890818733*^9, 3.6569169249515934`*^9}, 3.656917067021043*^9},
 CellID->12203207],

Cell[BoxData[
 RowBox[{"isExpPPDomain", "[", "\[ScriptCapitalT]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PPTuples isExpPPDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:338277855", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6569168483242583`*^9, 3.6569168613346815`*^9}},
 CellTags->{
  "ID:338277855", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples isExpPPDomain"},
 CellID->338277855],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->492675449]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->33968442],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PPTuples isExpPPDomain)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PPTuples isExpPPDomain)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{"isExpPPDomain", "[", 
           RowBox[{"PPTuples", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               StyleBox["n", "ExpressionVariable"], "\[Element]", 
               SubscriptBox["\[DoubleStruckCapitalN]", "0"]}]}}]], 
         RowBox[{"isExpPPDomain", "[", 
           RowBox[{"PPTuples", "[", 
             StyleBox["n", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:338277855"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p338277855-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p338277855-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:338277855", "Proof|ID:338277855-1"},
 CellID->630181205],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p338277855-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->497464270]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Polynomials Represented as Monomial-Tuples", "Subsection",
 CellChangeTimes->{{3.6457595170337048`*^9, 3.645759518549262*^9}, {
  3.64576182427752*^9, 3.645761831997134*^9}},
 CellID->1620536900],

Cell[CellGroupData[{

Cell["Informal Explanation", "Subsubsection",
 CellChangeTimes->{{3.592042340132761*^9, 3.592042344354285*^9}},
 CellID->1187960384],

Cell[TextData[{
 "The functor \[OpenCurlyDoubleQuote]PolynomialRing\[CloseCurlyDoubleQuote] \
takes a reduction ring ",
 Cell[BoxData[
  FormBox["R", TraditionalForm]]],
 " and a domain of power products ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 " and returns the domain of polynomials over ",
 Cell[BoxData[
  FormBox["R", TraditionalForm]]],
 ", where power products and operations thereof are represented by the \
elements and operations of ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 ". Polynomials are represented as tuples of monomials, which are in turn \
represented as pairs of non-zero coefficients and power-products, where the \
tuples are sorted descending w.r.t. the ordering relation \[Succeeds] defined \
on ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.592042505479131*^9, 3.592042626361019*^9}, {
  3.592042685781382*^9, 3.5920426957642*^9}, {3.592043025829199*^9, 
  3.592043053408197*^9}, {3.593771368263559*^9, 3.593771454643677*^9}, {
  3.645759627903304*^9, 3.645759682634451*^9}},
 CellID->1936291687],

Cell[TextData[{
 "The following is a list of operations that ",
 Cell[BoxData[
  FormBox["R", TraditionalForm]]],
 " has to provide (besides the usual ring operations):"
}], "Text",
 CellChangeTimes->{{3.592042848830953*^9, 3.592042894592203*^9}},
 CellID->1116953765],

Cell[CellGroupData[{

Cell["\<\
A noetherian irreflexive partial order relation (\[OpenCurlyDoubleQuote]\
\[Succeeds]\[CloseCurlyDoubleQuote]).\
\>", "Item",
 CellChangeTimes->{
  3.59204273196003*^9, {3.592042921714549*^9, 3.592042928306835*^9}},
 CellID->1456273069],

Cell[TextData[{
 "A function that, for all ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 ", returns ",
 StyleBox["all",
  FontWeight->"Bold"],
 " minimal non-trivial common reducibles of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " in a tuple (\[OpenCurlyDoubleQuote]lcrd\[CloseCurlyDoubleQuote]). Note \
that ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Equal]", "y"}], TraditionalForm]]],
 " is also possible!\nThis is different to [Buchberger2003]: There, only ",
 StyleBox["one",
  FontWeight->"Bold"],
 " minimal non-trivial common reducible is considered, and the case ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Equal]", "y"}], TraditionalForm]]],
 " is ruled out."
}], "Item",
 CellChangeTimes->{{3.592042742322896*^9, 3.592042750441021*^9}},
 CellID->1501844232],

Cell[TextData[{
 "A function that returns a valid reduction multiplier for reducing ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " modulo ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " (\[OpenCurlyDoubleQuote]rdm\[CloseCurlyDoubleQuote]). We follow the same \
paradigm as in [Buchberger2003] by setting this reduction multiplier to 0 iff \
no reduction is possible (modulo ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 ").\nOf course, reduction has to be compatible with the ordering, i.e. ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " is only reducible modulo ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " if ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[Succeeds]", 
    RowBox[{"x", "-", 
     RowBox[{
      RowBox[{"rdm", "[", 
       RowBox[{"x", ",", "y"}], "]"}], "*", "y"}]}]}], TraditionalForm]]],
 "."
}], "Item",
 CellChangeTimes->{{3.592042742322896*^9, 3.592042759074751*^9}},
 CellID->1006360754],

Cell[TextData[{
 "A function that returns the critical pair of ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " and minimial non-trivial common reducible ",
 Cell[BoxData[
  FormBox["z", TraditionalForm]]],
 " (\[OpenCurlyDoubleQuote]cp\[CloseCurlyDoubleQuote])."
}], "Item",
 CellChangeTimes->{{3.592042742322896*^9, 3.592042763764432*^9}, {
  3.645759699688065*^9, 3.645759707817293*^9}},
 CellID->2014536577]
}, Open  ]],

Cell[TextData[{
 "The following is a list of operations that ",
 Cell[BoxData[
  FormBox["T", TraditionalForm]]],
 " has to provide:"
}], "Text",
 CellChangeTimes->{{3.592042943680172*^9, 3.592042969457688*^9}},
 CellID->1075439607],

Cell[CellGroupData[{

Cell["\<\
An associative, commutative operation (\[OpenCurlyDoubleQuote]*\
\[CloseCurlyDoubleQuote]).\
\>", "Item",
 CellChangeTimes->{{3.59204307433266*^9, 3.592043108640753*^9}},
 CellID->643750130],

Cell["\<\
A neutral element w.r.t. \[OpenCurlyDoubleQuote]*\[CloseCurlyDoubleQuote] (\
\[OpenCurlyDoubleQuote]1\[CloseCurlyDoubleQuote]).\
\>", "Item",
 CellChangeTimes->{{3.59204307433266*^9, 3.592043129817933*^9}},
 CellID->1845811392],

Cell["A test for divisibility (\[OpenCurlyDoubleQuote]\[VerticalBar]\
\[CloseCurlyDoubleQuote]).", "Item",
 CellChangeTimes->{{3.59204307433266*^9, 3.592043177498439*^9}},
 CellID->1608954637],

Cell["\<\
A function that returns the least common multiple of two elements (\
\[OpenCurlyDoubleQuote]lcm\[CloseCurlyDoubleQuote]).\
\>", "Item",
 CellChangeTimes->{{3.59204307433266*^9, 3.592043217264247*^9}},
 CellID->1643257907]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Definition", "Subsubsection",
 CellChangeTimes->{{3.594449667860968*^9, 3.594449674571539*^9}, 
   3.645759912205167*^9},
 CellID->1331228005],

Cell["", "OpenEnvironment",
 CellID->1371979610],

Cell[CellGroupData[{

Cell["\<\
Definition: Reduction Ring of Polynomials Represented as Ordered Tuples of \
Monomials\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.592042368179434*^9, 3.592042393864496*^9}, {
  3.594446999906397*^9, 3.594447029676959*^9}},
 CellID->1149637019],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 611122547}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.592042401668746*^9, 3.592042405492165*^9}, {
   3.594450725564261*^9, 3.594450727161235*^9}, {3.594452087554145*^9, 
   3.59445209571691*^9}, {3.631601337799827*^9, 3.631601343033226*^9}, 
   3.6457623035765676`*^9},
 CellTags->{
  "ID:611122547", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->611122547],

Cell[BoxData[
 RowBox[{
  RowBox[{"PolyTuples", "[", 
   RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], ":=", 
  UnderscriptBox["\[CapitalDelta]", "N"]}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 1216459107}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.592042414902493*^9, 3.592042441369373*^9}, 
   3.59445073163715*^9, {3.59445210077302*^9, 3.594452107158702*^9}, {
   3.607339275536327*^9, 3.60733927697537*^9}, 3.6316013595709457`*^9, {
   3.634886270120873*^9, 3.63488627126783*^9}, {3.6457622957064*^9, 
   3.645762299419221*^9}},
 CellTags->{
  "ID:1216459107", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->1216459107],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
  "i", ",", "j", ",", "q", ",", "p", ",", "m", ",", "c", ",", "d", ",", "s", 
   ",", "t", ",", 
   RowBox[{"a", "..."}], ",", 
   RowBox[{"b", "..."}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 594519699}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.593494610094498*^9, 3.593494613120072*^9}, {
  3.593771469117906*^9, 3.593771491499526*^9}, {3.59377154770247*^9, 
  3.593771548185558*^9}, {3.594035843198792*^9, 3.59403584398054*^9}, {
  3.594451145720354*^9, 3.59445115103715*^9}, {3.594451936786313*^9, 
  3.594451946253508*^9}, {3.594452151680469*^9, 3.594452182371833*^9}, {
  3.594452370086054*^9, 3.594452371485838*^9}, {3.594455248663959*^9, 
  3.594455248849539*^9}, {3.59445533393801*^9, 3.594455337752419*^9}, {
  3.595057432199294*^9, 3.595057459269483*^9}, {3.6085498099461*^9, 
  3.60854984267891*^9}},
 CellTags->{
  "ID:594519699", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->594519699],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Element]", "N"], 
   RowBox[{"[", "p", "]"}]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"\[And]", 
   RowBox[{"\[Piecewise]", GridBox[{
      {
       RowBox[{"isTuple", "[", "p", "]"}]},
      {
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
          RowBox[{"\[LeftBracketingBar]", "p", "\[RightBracketingBar]"}]}]], 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"\[And]", 
          RowBox[{"\[Piecewise]", GridBox[{
             {
              RowBox[{"isTuple", "[", 
               SubscriptBox["p", "k"], "]"}]},
             {
              RowBox[{
               RowBox[{"\[LeftBracketingBar]", 
                SubscriptBox["p", "k"], "\[RightBracketingBar]"}], "\[Equal]",
                "2"}]},
             {
              RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
               RowBox[{"[", 
                SubscriptBox[
                 RowBox[{"(", 
                  SubscriptBox["p", "k"], ")"}], "1"], "]"}]}]},
             {
              RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
               RowBox[{"[", 
                SubscriptBox[
                 RowBox[{"(", 
                  SubscriptBox["p", "k"], ")"}], "2"], "]"}]}]},
             {
              RowBox[{
               SubscriptBox[
                RowBox[{"(", 
                 SubscriptBox["p", "k"], ")"}], "1"], "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalR]"]}]}
            }]}]}], 
         TagBox[")",
          "AutoParentheses"]}]}]},
      {
       RowBox[{
        UnderscriptBox["\[ForAll]", 
         RowBox[{
          RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
          RowBox[{
           RowBox[{"\[LeftBracketingBar]", "p", "\[RightBracketingBar]"}], 
           "-", "1"}]}]], 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          SubscriptBox[
           RowBox[{"(", 
            SubscriptBox["p", "k"], ")"}], "2"], 
          UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
          SubscriptBox[
           RowBox[{"(", 
            SubscriptBox["p", 
             RowBox[{"k", "+", "1"}]], ")"}], "2"]}], 
         TagBox[")",
          "AutoParentheses"]}]}]}
     }]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples \[Element])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:65019775", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.593494625629224*^9, 3.593494701540691*^9}, 
   3.59377115048849*^9, {3.593771200229455*^9, 3.593771290165935*^9}, {
   3.594035849263751*^9, 3.594035854600325*^9}, {3.59445075023562*^9, 
   3.59445076097245*^9}, {3.594451101486164*^9, 3.594451109082149*^9}, {
   3.594452191552515*^9, 3.59445232061715*^9}, {3.594455477126506*^9, 
   3.594455495779226*^9}, {3.595057462475311*^9, 3.595057474212757*^9}, {
   3.607339280024721*^9, 3.60733930328321*^9}, {3.608007717491008*^9, 
   3.608007719150078*^9}, {3.608549851757632*^9, 3.608549863712901*^9}, {
   3.6316013663152523`*^9, 3.631601367762216*^9}, 3.6316014601925793`*^9, {
   3.6457624135932493`*^9, 3.645762475166613*^9}},
 CellTags->{
  "ID:65019775", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Element]"},
 CellID->65019775],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["C", "N"], "[", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "s"}], 
   "]"}], ":=", 
  UnderscriptBox["0", "\[ScriptCapitalR]"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples C base)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:399496924", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.655538475935485*^9, 3.6555385155127544`*^9}},
 CellTags->{
  "ID:399496924", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C base"},
 CellID->399496924],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["C", "N"], "[", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"a", "..."}]}], "\[RightAngleBracket]"}], ",", "s"}], "]"}], ":=", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Piecewise]", GridBox[{
      {"c", "\[DoubleLeftArrow]", 
       RowBox[{"t", "\[Equal]", "s"}]},
      {
       RowBox[{
        UnderscriptBox["C", "N"], "[", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"a", "..."}], "\[RightAngleBracket]"}], ",", "s"}], "]"}], 
       "\[DoubleLeftArrow]", "otherwise"}
     }]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples C rec)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:983317445", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6555384977131233`*^9, 3.6555385497236147`*^9}, {
  3.655538670265026*^9, 3.655538671357028*^9}},
 CellTags->{
  "ID:983317445", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C rec"},
 CellID->983317445],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["0", "N"], ":=", 
  RowBox[{
  "\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1360683691", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.593771338010247*^9, {3.607339313279467*^9, 3.60733931418446*^9}},
 CellTags->{
  "ID:1360683691", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples 0"},
 CellID->1360683691],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["1", "N"], ":=", 
  RowBox[{"\[Piecewise]", GridBox[{
     {
      RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
      "\[DoubleLeftArrow]", 
      RowBox[{
       UnderscriptBox["1", "\[ScriptCapitalR]"], "\[Equal]", 
       UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
     {
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         UnderscriptBox["1", "\[ScriptCapitalR]"], ",", 
         UnderscriptBox["1", "\[ScriptCapitalT]"]}], "\[RightAngleBracket]"}],
        "\[RightAngleBracket]"}], "\[DoubleLeftArrow]", "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:13289303", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.593771323135162*^9, 3.593771326331638*^9}, {
   3.594451117899345*^9, 3.594451120044741*^9}, {3.594452328163875*^9, 
   3.594452341207251*^9}, {3.607339310695833*^9, 3.607339311692473*^9}, 
   3.631601373974741*^9, 3.6316014666278467`*^9, {3.6560458871090994`*^9, 
   3.656045920649158*^9}},
 CellTags->{
  "ID:13289303", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples 1"},
 CellID->13289303],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
   UnderscriptBox["+", "N"], "q"}], ":=", "q"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples + 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1777247200", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.593771349616116*^9, {3.593771509750953*^9, 3.593771511779852*^9}, {
   3.594465988489223*^9, 3.594465990377636*^9}, {3.607339314997389*^9, 
   3.6073393157365*^9}},
 CellTags->{
  "ID:1777247200", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + 1"},
 CellID->1777247200],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"a", "..."}]}], "\[RightAngleBracket]"}], 
   UnderscriptBox["+", "N"], 
   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], ":=", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"a", "..."}]}], 
   "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples + 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2122267989", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.593771502883749*^9, 3.593771516821926*^9}, {
  3.59446599262808*^9, 3.594465994109239*^9}, {3.607339316623276*^9, 
  3.60733931758368*^9}, {3.64576290307687*^9, 3.6457629241516953`*^9}},
 CellTags->{
  "ID:2122267989", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + 2"},
 CellID->2122267989],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"a", "..."}]}], "\[RightAngleBracket]"}], 
   UnderscriptBox["+", "N"], 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"d", ",", "t"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"b", "..."}]}], "\[RightAngleBracket]"}]}], ":=", 
  RowBox[{"\[Piecewise]", GridBox[{
     {
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], 
       TagBox["\:293b",
        Identity,
        SyntaxForm->"a*b"], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"a", "..."}], "\[RightAngleBracket]"}], 
         UnderscriptBox["+", "N"], 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"d", ",", "t"}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"b", "..."}]}], "\[RightAngleBracket]"}]}], ")"}]}], 
      "\[DoubleLongLeftArrow]", 
      RowBox[{"s", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]},
     {
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"d", ",", "t"}], "\[RightAngleBracket]"}], 
       TagBox["\:293b",
        Identity,
        SyntaxForm->"a*b"], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"a", "..."}]}], "\[RightAngleBracket]"}], 
         UnderscriptBox["+", "N"], 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"b", "..."}], "\[RightAngleBracket]"}]}], ")"}]}], 
      "\[DoubleLongLeftArrow]", 
      RowBox[{"t", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "s"}]},
     {
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         RowBox[{"c", 
          UnderscriptBox["+", "\[ScriptCapitalR]"], "d"}], ",", "s"}], 
        "\[RightAngleBracket]"}], 
       TagBox["\:293b",
        Identity,
        SyntaxForm->"a*b"], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"a", "..."}], "\[RightAngleBracket]"}], 
         UnderscriptBox["+", "N"], 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"b", "..."}], "\[RightAngleBracket]"}]}], ")"}]}], 
      "\[DoubleLongLeftArrow]", 
      RowBox[{
       RowBox[{"c", 
        UnderscriptBox["+", "\[ScriptCapitalR]"], "d"}], "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
     {
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"a", "..."}], "\[RightAngleBracket]"}], 
       UnderscriptBox["+", "N"], 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"b", "..."}], "\[RightAngleBracket]"}]}], 
      "\[DoubleLongLeftArrow]", "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples + 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:290578418", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.593771535298327*^9, {3.594451951784225*^9, 3.594452068262923*^9}, {
   3.594452376695929*^9, 3.59445243011773*^9}, {3.594452467396512*^9, 
   3.594452525584392*^9}, {3.607339319803906*^9, 3.607339360820973*^9}, {
   3.608616776204025*^9, 3.608616797757431*^9}, {3.6316013789759483`*^9, 
   3.631601381571271*^9}, {3.6457623243128023`*^9, 3.6457623398207703`*^9}},
 CellTags->{
  "ID:290578418", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + 3"},
 CellID->290578418],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["-", "N"], 
   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], ":=", 
  RowBox[{
  "\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples - unary 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1495569651", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594452624928613*^9, 3.59445264858098*^9}, {
  3.594452844350174*^9, 3.594452850337731*^9}},
 CellTags->{
  "ID:1495569651", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - unary 1"},
 CellID->1495569651],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["-", "N"], 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"a", "..."}]}], "\[RightAngleBracket]"}]}], ":=", 
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["-", "\[ScriptCapitalR]"], "c"}], ",", "s"}], 
    "\[RightAngleBracket]"}], 
   TagBox["\:293b",
    Identity,
    SyntaxForm->"a*b"], 
   RowBox[{"(", 
    RowBox[{
     UnderscriptBox["-", "N"], 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"a", "..."}], "\[RightAngleBracket]"}]}], 
    ")"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples - unary 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1285234307", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594452655840393*^9, 3.594452715467961*^9}, {
   3.607339371118301*^9, 3.607339376856483*^9}, {3.608616802549688*^9, 
   3.608616806779823*^9}, 3.6316013855223017`*^9},
 CellTags->{
  "ID:1285234307", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - unary 2"},
 CellID->1285234307],

Cell[BoxData[
 RowBox[{
  RowBox[{"p", 
   UnderscriptBox["-", "N"], "q"}], ":=", 
  RowBox[{"p", 
   UnderscriptBox["+", "N"], 
   RowBox[{"(", 
    RowBox[{
     UnderscriptBox["-", "N"], "q"}], ")"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples -)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:551740538", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594452728093301*^9, 3.594452743728123*^9}},
 CellTags->{
  "ID:551740538", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples -"},
 CellID->551740538],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
   UnderscriptBox["*", "N"], "q"}], ":=", 
  RowBox[{
  "\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples * 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:419928300", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.594452773429185*^9, {3.594452821668337*^9, 3.594452858056516*^9}},
 CellTags->{
  "ID:419928300", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples * 1"},
 CellID->419928300],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"a", "..."}]}], "\[RightAngleBracket]"}], 
   UnderscriptBox["*", "N"], 
   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], ":=", 
  RowBox[{
  "\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples * 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:209271439", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.594452779708953*^9, {3.594452829405809*^9, 3.594452865116559*^9}, 
   3.6457629402019587`*^9},
 CellTags->{
  "ID:209271439", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples * 2"},
 CellID->209271439],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"a", "..."}]}], "\[RightAngleBracket]"}], 
   UnderscriptBox["*", "N"], 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"d", ",", "t"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"b", "..."}]}], "\[RightAngleBracket]"}]}], ":=", 
  RowBox[{"\[Piecewise]", GridBox[{
     {
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{
            RowBox[{"c", 
             UnderscriptBox["*", "\[ScriptCapitalR]"], "d"}], ",", 
            RowBox[{"s", 
             UnderscriptBox["*", "\[ScriptCapitalT]"], "t"}]}], 
           "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], 
         UnderscriptBox["+", "N"], 
         RowBox[{
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], 
           "\[RightAngleBracket]"}], 
          UnderscriptBox["*", "N"], 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"b", "..."}], "\[RightAngleBracket]"}]}]}], ")"}], 
       UnderscriptBox["+", "N"], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"a", "..."}], "\[RightAngleBracket]"}], 
         UnderscriptBox["*", "N"], 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"d", ",", "t"}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"b", "..."}]}], "\[RightAngleBracket]"}]}], ")"}]}], 
      "\[DoubleLeftArrow]", 
      RowBox[{
       RowBox[{"c", 
        UnderscriptBox["*", "\[ScriptCapitalR]"], "d"}], "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
     {
      RowBox[{
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], 
          "\[RightAngleBracket]"}], 
         UnderscriptBox["*", "N"], 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"b", "..."}], "\[RightAngleBracket]"}]}], ")"}], 
       UnderscriptBox["+", "N"], 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"a", "..."}], "\[RightAngleBracket]"}], 
         UnderscriptBox["*", "N"], 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"d", ",", "t"}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"b", "..."}]}], "\[RightAngleBracket]"}]}], ")"}]}], 
      "\[DoubleLeftArrow]", "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples * 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1411844363", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.594452787008683*^9, {3.594452868912309*^9, 3.59445289405804*^9}, 
   3.594465942620106*^9, {3.607767252726596*^9, 3.607767254750163*^9}, {
   3.607769878440762*^9, 3.607769893829508*^9}, {3.607769931019342*^9, 
   3.607769934230881*^9}, {3.60777020788632*^9, 3.607770221626101*^9}, {
   3.631601390527039*^9, 3.631601394603779*^9}, 3.631601478293705*^9},
 CellTags->{
  "ID:1411844363", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples * 3"},
 CellID->1411844363],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
    UnderscriptBox["\[Succeeds]", "N"], "p"}], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], "False"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples \[Succeeds] 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1128062400", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.594452795398282*^9, {3.594452911005001*^9, 3.594452914902691*^9}, 
   3.64576235900149*^9},
 CellTags->{
  "ID:1128062400", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Succeeds] 1"},
 CellID->1128062400],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"a", "..."}]}], "\[RightAngleBracket]"}], 
    UnderscriptBox["\[Succeeds]", "N"], 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], "True"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples \[Succeeds] 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1280515048", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.594452803505665*^9, {3.594452899621533*^9, 3.594452918628251*^9}, {
   3.645762363317094*^9, 3.645762366984333*^9}},
 CellTags->{
  "ID:1280515048", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Succeeds] 2"},
 CellID->1280515048],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"a", "..."}]}], "\[RightAngleBracket]"}], 
    UnderscriptBox["\[Succeeds]", "N"], 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"d", ",", "t"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"b", "..."}]}], "\[RightAngleBracket]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"\[Or]", 
   RowBox[{"\[Piecewise]", GridBox[{
      {
       RowBox[{"s", 
        UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]},
      {
       RowBox[{
        RowBox[{"s", "\[Equal]", "t"}], "\[And]", 
        RowBox[{"c", 
         UnderscriptBox["\[Succeeds]", "\[ScriptCapitalR]"], "d"}]}]},
      {
       RowBox[{"\[And]", 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{"s", "\[Equal]", "t"}]},
           {
            RowBox[{"c", "\[Equal]", "d"}]},
           {
            RowBox[{
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"a", "..."}], "\[RightAngleBracket]"}], 
             UnderscriptBox["\[Succeeds]", "N"], 
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"b", "..."}], "\[RightAngleBracket]"}]}]}
          }]}]}]}
     }]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples \[Succeeds] 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2095908261", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.594452813491317*^9, {3.594452905443574*^9, 3.594452970210094*^9}, {
   3.607339394527795*^9, 3.607339404488731*^9}, 3.631601400636804*^9, {
   3.6457623703866053`*^9, 3.645762387317819*^9}},
 CellTags->{
  "ID:2095908261", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Succeeds] 3"},
 CellID->2095908261],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["multN", "N"], "[", 
   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], "]"}], ":=", 
  "0"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples multN 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1668822406", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594455300573168*^9, 3.594455310143608*^9}, {
  3.631855547457464*^9, 3.631855550559929*^9}, {3.631856582449052*^9, 
  3.631856584975881*^9}},
 CellTags->{
  "ID:1668822406", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples multN 1"},
 CellID->1668822406],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["multN", "N"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"a", "..."}]}], "\[RightAngleBracket]"}], "]"}], ":=", 
  RowBox[{
   UnderscriptBox["multN", "\[ScriptCapitalR]"], "[", "c", 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples multN 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:329725901", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594453230597199*^9, 3.594453266772131*^9}, {
   3.594455187214362*^9, 3.594455210711886*^9}, 3.6316014102918777`*^9, {
   3.631855553241474*^9, 3.6318555558328114`*^9}, {3.63185658675159*^9, 
   3.6318565893408337`*^9}},
 CellTags->{
  "ID:329725901", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples multN 2"},
 CellID->329725901],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "N"], "[", 
   RowBox[{"m", ",", "p"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
      RowBox[{
       UnderscriptBox["multN", "N"], "[", "p", "]"}]}]], 
    RowBox[{
     UnderscriptBox["mult", "N"], "[", 
     RowBox[{"m", ",", "p", ",", "k"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples mult)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:822409829", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594455218905547*^9, 3.594455296165126*^9}, {
   3.607339407770714*^9, 3.607339409026057*^9}, {3.608549880241295*^9, 
   3.608549882306764*^9}, {3.631855566808803*^9, 3.631855570731358*^9}, {
   3.6318565932337427`*^9, 3.631856598865741*^9}, 3.645762397773674*^9},
 CellTags->{
  "ID:822409829", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples mult"},
 CellID->822409829],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["mult", "N"], "[", 
   RowBox[{"m", ",", "p", ",", "i"}], "]"}], 
  TagBox[
   RowBox[{
   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
    "\[DoubleLongLeftRightArrow]"}],
   Identity,
   SyntaxForm->"a\[Implies]b"], 
  RowBox[{"\[And]", 
   RowBox[{"\[Piecewise]", GridBox[{
      {
       RowBox[{
        RowBox[{"\[LeftBracketingBar]", "m", "\[RightBracketingBar]"}], 
        "\[Equal]", "1"}]},
      {
       RowBox[{
        UnderscriptBox["mult", "\[ScriptCapitalR]"], "[", 
        RowBox[{
         SubscriptBox[
          RowBox[{"(", 
           SubscriptBox["m", "1"], ")"}], "1"], ",", 
         SubscriptBox[
          RowBox[{"(", 
           SubscriptBox["p", "1"], ")"}], "1"], ",", "i"}], "]"}]}
     }]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples mult i)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:922404041", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594455317700557*^9, 3.59445532286657*^9}, {
   3.594455363118537*^9, 3.594455418662535*^9}, {3.594455462671837*^9, 
   3.594455467662085*^9}, {3.607339413785221*^9, 3.607339414708534*^9}, {
   3.607767513787626*^9, 3.607767514000927*^9}, {3.608549887839101*^9, 
   3.608549893994661*^9}, 3.63160141482133*^9, {3.6457624953873863`*^9, 
   3.6457625074838552`*^9}},
 CellTags->{
  "ID:922404041", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples mult i"},
 CellID->922404041],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["rdm", "N"], "[", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "q"}], 
   "]"}], ":=", 
  UnderscriptBox["0", "N"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples rdm 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2033419264", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.59445567881711*^9, 3.594455698340902*^9}},
 CellTags->{
  "ID:2033419264", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples rdm 1"},
 CellID->2033419264],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["rdm", "N"], "[", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"a", "..."}]}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], "]"}], ":=", 
  UnderscriptBox["0", "N"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples rdm 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1308197671", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594455702818093*^9, 3.594455713110278*^9}, 
   3.645762955384719*^9},
 CellTags->{
  "ID:1308197671", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples rdm 2"},
 CellID->1308197671],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["rdm", "N"], "[", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"a", "..."}]}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"d", ",", "t"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"b", "..."}]}], "\[RightAngleBracket]"}]}], "]"}], ":=", 
  RowBox[{
   UnderscriptBox["let", 
    RowBox[{"rr", "=", 
     RowBox[{
      UnderscriptBox["rdm", "\[ScriptCapitalR]"], "[", 
      RowBox[{"c", ",", "d"}], "]"}]}]], 
   RowBox[{"\[Piecewise]", GridBox[{
      {
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"rr", ",", 
          RowBox[{"s", 
           UnderscriptBox["/", "\[ScriptCapitalT]"], "t"}]}], 
         "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], 
       "\[DoubleLeftArrow]", 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"\[Not]", 
          RowBox[{"rr", "\[Equal]", 
           UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
         TagBox[")",
          "AutoParentheses"]}], "\[And]", 
        RowBox[{"t", 
         UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "s"}]}]},
      {
       RowBox[{
        UnderscriptBox["rdm", "N"], "[", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"a", "..."}], "\[RightAngleBracket]"}], ",", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"d", ",", "t"}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"b", "..."}]}], "\[RightAngleBracket]"}]}], "]"}], 
       "\[DoubleLeftArrow]", "otherwise"}
     }]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples rdm 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1157498137", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594455606126402*^9, 3.594455662576593*^9}, {
   3.594455728873069*^9, 3.594455920392444*^9}, {3.594455975215451*^9, 
   3.594455975550145*^9}, 3.594468586862225*^9, {3.594469124445245*^9, 
   3.594469128309598*^9}, {3.607339436306461*^9, 3.607339446500299*^9}, {
   3.631601427814893*^9, 3.6316014300074253`*^9}, {3.6316014910234613`*^9, 
   3.631601492211052*^9}, {3.642997992615514*^9, 3.642997993552527*^9}, {
   3.6563977170795116`*^9, 3.6563977225707216`*^9}},
 CellTags->{
  "ID:1157498137", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples rdm 3"},
 CellID->1157498137],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lcrd", "N"], "[", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "q", ",", 
    "i", ",", "j"}], "]"}], ":=", 
  RowBox[{
  "\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples lcrd 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:103388547", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6563089176724243`*^9, 3.6563089319464493`*^9}},
 CellTags->{
  "ID:103388547", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd 1"},
 CellID->103388547],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lcrd", "N"], "[", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"a", "..."}]}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "i", ",", 
    "j"}], "]"}], ":=", 
  RowBox[{
  "\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples lcrd 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:166800573", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656308935830856*^9, 3.6563089406200643`*^9}, 
   3.656397381162122*^9},
 CellTags->{
  "ID:166800573", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd 2"},
 CellID->166800573],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lcrd", "N"], "[", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"a", "..."}]}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"d", ",", "t"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"b", "..."}]}], "\[RightAngleBracket]"}], ",", "i", ",", "j"}], 
   "]"}], ":=", 
  RowBox[{"\[Piecewise]", GridBox[{
     {
      RowBox[{
       UnderscriptBox["let", 
        RowBox[{"lr", "=", 
         RowBox[{
          UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
          RowBox[{"d", ",", "d", ",", "i", ",", "j"}], "]"}]}]], 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{
            SubscriptBox["lr", "i"], ",", "t"}], "\[RightAngleBracket]"}], 
          "\[RightAngleBracket]"}], 
         UnderscriptBox["|", 
          RowBox[{
           RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{
           "\[LeftBracketingBar]", "lr", "\[RightBracketingBar]"}]}]]}], 
        "\[RightAngleBracket]"}]}], "\[DoubleLeftArrow]", 
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], ",", 
         RowBox[{"a", "..."}]}], "\[RightAngleBracket]"}], "\[Equal]", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"d", ",", "t"}], "\[RightAngleBracket]"}], ",", 
         RowBox[{"b", "..."}]}], "\[RightAngleBracket]"}]}]},
     {
      RowBox[{
       UnderscriptBox["let", 
        RowBox[{"lp", "=", 
         RowBox[{
          UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
          RowBox[{"s", ",", "t"}], "]"}]}]], 
       RowBox[{"\[Piecewise]", GridBox[{
          {
           RowBox[{
            UnderscriptBox["let", 
             RowBox[{"lr", "=", 
              RowBox[{
               UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", "d", 
               "]"}]}]], 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
              RowBox[{"\[LeftAngleBracket]", 
               RowBox[{"\[LeftAngleBracket]", 
                RowBox[{
                 SubscriptBox["lr", "i"], ",", "lp"}], 
                "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], 
              UnderscriptBox["|", 
               RowBox[{
                RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
                RowBox[{
                "\[LeftBracketingBar]", "lr", "\[RightBracketingBar]"}]}]]}], 
             "\[RightAngleBracket]"}]}], "\[DoubleLeftArrow]", 
           RowBox[{"c", "\[Equal]", "d"}]},
          {
           RowBox[{
            UnderscriptBox["let", 
             RowBox[{"lr", "=", 
              RowBox[{
               UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
               RowBox[{"c", ",", "d", ",", "i", ",", "j"}], "]"}]}]], 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
              RowBox[{"\[LeftAngleBracket]", 
               RowBox[{"\[LeftAngleBracket]", 
                RowBox[{
                 SubscriptBox["lr", "i"], ",", "lp"}], 
                "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], 
              UnderscriptBox["|", 
               RowBox[{
                RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
                RowBox[{
                "\[LeftBracketingBar]", "lr", "\[RightBracketingBar]"}]}]]}], 
             "\[RightAngleBracket]"}]}], "\[DoubleLeftArrow]", "otherwise"}
         }]}]}], "\[DoubleLongLeftArrow]", "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples lcrd 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:319850711", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.59445822831031*^9, 3.594458248361637*^9}, {
   3.608549922126339*^9, 3.608549923051816*^9}, {3.610097740389538*^9, 
   3.610097826366921*^9}, {3.61009786192786*^9, 3.610097931272047*^9}, {
   3.610097972837281*^9, 3.610097999577977*^9}, {3.631601435162877*^9, 
   3.631601443530683*^9}, 3.631601497437171*^9, {3.645762533360882*^9, 
   3.645762592793336*^9}, {3.6563089447696714`*^9, 3.6563090486814537`*^9}},
 CellTags->{
  "ID:319850711", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd 3"},
 CellID->319850711],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lcrd", "N"], "[", 
   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], "]"}], ":=", 
  RowBox[{
  "\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples lcrd unary 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:112650434", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594466770198428*^9, 3.594466802307595*^9}},
 CellTags->{
  "ID:112650434", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd unary 1"},
 CellID->112650434],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lcrd", "N"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"a", "..."}]}], "\[RightAngleBracket]"}], "]"}], ":=", 
  RowBox[{
   UnderscriptBox["let", 
    RowBox[{"lr", "=", 
     RowBox[{
      UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", "c", "]"}]}]], 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        SubscriptBox["lr", "i"], ",", "s"}], "\[RightAngleBracket]"}], 
      "\[RightAngleBracket]"}], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "lr", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples lcrd unary 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1729493745", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.59446680615005*^9, 3.594466909001557*^9}, {
   3.607339482779186*^9, 3.607339484034992*^9}, 3.631601448960155*^9},
 CellTags->{
  "ID:1729493745", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd unary 2"},
 CellID->1729493745],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["cpm", "N"], "[", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "q", ",", 
    "m", ",", "i", ",", "j"}], "]"}], ":=", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    UnderscriptBox["0", "N"], ",", 
    UnderscriptBox["0", "N"]}], 
   "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples cpm 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:414290689", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6563091660560603`*^9, 3.6563091823580885`*^9}},
 CellTags->{
  "ID:414290689", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples cpm 1"},
 CellID->414290689],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["cpm", "N"], "[", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"a", "..."}]}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], ",", "m", ",", 
    "i", ",", "j"}], "]"}], ":=", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    UnderscriptBox["0", "N"], ",", 
    UnderscriptBox["0", "N"]}], 
   "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples cpm 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:27385995", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656309185400094*^9, 3.6563091905637035`*^9}, 
   3.656397390241338*^9},
 CellTags->{
  "ID:27385995", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples cpm 2"},
 CellID->27385995],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["cpm", "N"], "[", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"a", "..."}]}], "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"d", ",", "t"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"b", "..."}]}], "\[RightAngleBracket]"}], ",", "m", ",", "i", 
    ",", "j"}], "]"}], ":=", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Piecewise]", GridBox[{
      {
       RowBox[{
        UnderscriptBox["let", 
         RowBox[{"cm", "=", 
          RowBox[{
           UnderscriptBox["cpm", "\[ScriptCapitalR]"], "[", 
           RowBox[{"c", ",", "d", ",", 
            SubscriptBox[
             RowBox[{"(", 
              SubscriptBox["m", "1"], ")"}], "1"], ",", "i", ",", "j"}], 
           "]"}]}]], 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{
             SubscriptBox["cm", "1"], ",", 
             RowBox[{
              SubscriptBox[
               RowBox[{"(", 
                SubscriptBox["m", "1"], ")"}], "2"], 
              UnderscriptBox["/", "\[ScriptCapitalT]"], "s"}]}], 
            "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], ",", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{
             SubscriptBox["cm", "2"], ",", 
             RowBox[{
              SubscriptBox[
               RowBox[{"(", 
                SubscriptBox["m", "1"], ")"}], "2"], 
              UnderscriptBox["/", "\[ScriptCapitalT]"], "t"}]}], 
            "\[RightAngleBracket]"}], "\[RightAngleBracket]"}]}], 
         "\[RightAngleBracket]"}]}], "\[DoubleLeftArrow]", 
       RowBox[{
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"c", ",", "s"}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"a", "..."}]}], "\[RightAngleBracket]"}], "\[Equal]", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"d", ",", "t"}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"b", "..."}]}], "\[RightAngleBracket]"}]}], "\[Or]", 
        RowBox[{"c", "\[NotEqual]", "d"}]}]},
      {
       RowBox[{
        UnderscriptBox["let", 
         RowBox[{"r", "=", 
          RowBox[{
           UnderscriptBox["rdm", "\[ScriptCapitalR]"], "[", 
           RowBox[{
            SubscriptBox[
             RowBox[{"(", 
              SubscriptBox["m", "1"], ")"}], "1"], ",", "c"}], "]"}]}]], 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"r", ",", 
             RowBox[{
              SubscriptBox[
               RowBox[{"(", 
                SubscriptBox["m", "1"], ")"}], "2"], 
              UnderscriptBox["/", "\[ScriptCapitalT]"], "s"}]}], 
            "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], ",", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"r", ",", 
             RowBox[{
              SubscriptBox[
               RowBox[{"(", 
                SubscriptBox["m", "1"], ")"}], "2"], 
              UnderscriptBox["/", "\[ScriptCapitalT]"], "t"}]}], 
            "\[RightAngleBracket]"}], "\[RightAngleBracket]"}]}], 
         "\[RightAngleBracket]"}]}], "\[DoubleLeftArrow]", "otherwise"}
     }]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples cpm 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1444186163", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594458543232637*^9, 3.594458574911888*^9}, {
   3.595057543892115*^9, 3.595057574715891*^9}, {3.607339487800038*^9, 
   3.607339488907539*^9}, {3.610098046958749*^9, 3.610098099738793*^9}, 
   3.63160145196136*^9, {3.631601504167103*^9, 3.631601505614423*^9}, {
   3.6457629699539623`*^9, 3.645763018538412*^9}, {3.645765749408908*^9, 
   3.645765854908393*^9}, {3.645765885130582*^9, 3.645765946866288*^9}, {
   3.645766346569352*^9, 3.645766353065627*^9}, {3.6563091992217183`*^9, 
   3.656309280796262*^9}},
 CellTags->{
  "ID:1444186163", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples cpm 3"},
 CellID->1444186163],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["<<", "N"], 
    RowBox[{"[", "p", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], ":=", 
  RowBox[{"\[Wolf]SortMonomials", "[", 
   RowBox[{
    RowBox[{"\[Wolf]MonomialTupleIn", "[", 
     RowBox[{"p", ",", "\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}],
     ",", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     TagBox[
      UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"],
      Identity], 
     TagBox[")",
      "AutoParentheses"]}]}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples <<)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2018930893", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.63307995105399*^9, 3.6330800081856823`*^9}, {
  3.6357482955308533`*^9, 3.6357483021986713`*^9}, {3.63574835193331*^9, 
  3.63574835524253*^9}, {3.63574945517863*^9, 3.6357494667202187`*^9}, {
  3.6357495597802677`*^9, 3.635749563959304*^9}, {3.635749826926386*^9, 
  3.635749856448213*^9}, {3.635749946439636*^9, 3.635749963547842*^9}, {
  3.6357500168851843`*^9, 3.635750018367378*^9}, {3.635750224798779*^9, 
  3.635750225067432*^9}, {3.635755973487458*^9, 3.635756002324532*^9}, {
  3.645762603163488*^9, 3.6457626091533537`*^9}, {3.6457639172698727`*^9, 
  3.645763917950746*^9}},
 CellTags->{
  "ID:2018930893", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples <<"},
 CellID->2018930893],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox[">>", "N"], 
    RowBox[{"[", "p", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], ":=", 
  RowBox[{"\[Wolf]PrintPoly", "[", 
   RowBox[{"\[Wolf]MonomialTupleOut", "[", 
    RowBox[{"p", ",", "\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples >>)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1442847663", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6330800950307093`*^9, 3.633080100268716*^9}, {
   3.633080141071067*^9, 3.633080166793787*^9}, {3.6357483089335413`*^9, 
   3.635748319684381*^9}, {3.6357496056725607`*^9, 3.635749612146576*^9}, {
   3.63574986257734*^9, 3.635749875574164*^9}, {3.635750074021977*^9, 
   3.635750075389398*^9}, {3.635750174292903*^9, 3.635750176576387*^9}, {
   3.635750242111219*^9, 3.635750244758223*^9}, 3.635753894888608*^9, {
   3.635755976470223*^9, 3.6357559771811733`*^9}},
 CellTags->{
  "ID:1442847663", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples >>"},
 CellID->1442847663],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1513890663]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1465096073]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Tests", "Section",
 CellChangeTimes->{{3.645759533075899*^9, 3.645759533820724*^9}},
 CellID->1812227459],

Cell[CellGroupData[{

Cell["Power Products", "Subsection",
 CellChangeTimes->{{3.645759885122527*^9, 3.645759887663216*^9}},
 CellID->1790574279],

Cell["", "OpenEnvironment",
 CellID->2085486573],

Cell[CellGroupData[{

Cell["\<\
Definition: 2-Variate and 3-Variate Power Products with \
Degree-Reverse-Lexicographic Order\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.594460609631138*^9, 3.594460620214307*^9}, {
  3.596192419821697*^9, 3.596192425340894*^9}, {3.645763067920206*^9, 
  3.645763077109971*^9}},
 CellID->748671929],

Cell[BoxData[
 RowBox[{"\[ScriptCapitalT]2", ":=", 
  RowBox[{"OrderedPP", "[", 
   RowBox[{
    RowBox[{"PPTuples", "[", "2", "]"}], ",", 
    RowBox[{"DegRevLex", "[", 
     RowBox[{"PPTuples", "[", "2", "]"}], "]"}]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalT]2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1486930793", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.596192432181783*^9, 3.596192439037655*^9}, {
   3.645763082667824*^9, 3.645763098659034*^9}, 3.645763234786858*^9, {
   3.6569176419444523`*^9, 3.6569176540968738`*^9}},
 CellTags->{
  "ID:1486930793", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalT]2"},
 CellID->1486930793],

Cell[BoxData[
 RowBox[{"\[ScriptCapitalT]3", ":=", 
  RowBox[{"OrderedPP", "[", 
   RowBox[{
    RowBox[{"PPTuples", "[", "3", "]"}], ",", 
    RowBox[{"DegRevLex", "[", 
     RowBox[{"PPTuples", "[", "3", "]"}], "]"}]}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalT]3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:933000257", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594460624908743*^9, 3.594460647909824*^9}, {
   3.645763089593129*^9, 3.645763101579585*^9}, 3.6457632358554287`*^9, {
   3.656917658808082*^9, 3.6569176623960886`*^9}},
 CellTags->{
  "ID:933000257", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalT]3"},
 CellID->933000257],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->1614174784]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->1930771981],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalT]3"], 
  RowBox[{"[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"0", ",", "7", ",", "3"}], "\[RightAngleBracket]"}], 
   "]"}]}]], "Computation",
 CellChangeTimes->{{3.594460662764123*^9, 3.594460678192673*^9}, {
  3.594460716424183*^9, 3.594460729775287*^9}, {3.645763177181121*^9, 
  3.64576318211237*^9}},
 CellID->300068762],

Cell[BoxData[
 FormBox["True", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{{3.607339204896455*^9, 3.60733921292517*^9}, 
   3.610097554574191*^9, 3.61571399624681*^9, 3.615797918718175*^9, 
   3.615801319778202*^9, 3.63144747366044*^9, 3.635758031280484*^9, {
   3.645763226095035*^9, 3.645763242482129*^9}, {3.6570138945234947`*^9, 
   3.6570139055059137`*^9}},
 CellID->42587073]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c300068762"]]]]}, {None, None}},
 CellChangeTimes->{3.657013906114315*^9},
 CellID->30747916],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c300068762-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6570139063171153`*^9},
 CellID->96247615]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalT]2"], 
  RowBox[{"[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"0", ",", "7", ",", "3"}], "\[RightAngleBracket]"}], 
   "]"}]}]], "Computation",
 CellChangeTimes->{3.6457632636529007`*^9},
 CellID->1717056269],

Cell[BoxData[
 FormBox["False", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645763264956834*^9, 3.657013908407519*^9},
 CellID->258122591]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1717056269"]]]]}, {None, None}},
 CellChangeTimes->{3.6570139088443193`*^9},
 CellID->92227752],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1717056269-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.65701390904712*^9},
 CellID->642042499]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"2", ",", "1"}], "\[RightAngleBracket]"}], 
  UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]2"], 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"2", ",", "0"}], "\[RightAngleBracket]"}]}]], "Computation",
 CellChangeTimes->{{3.645763282752521*^9, 3.64576331508995*^9}},
 CellID->1221343315],

Cell[BoxData[
 FormBox["False", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{{3.645763305411647*^9, 3.645763316019462*^9}, 
   3.6570139116211243`*^9},
 CellID->369293207]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1221343315"]]]]}, {None, None}},
 CellChangeTimes->{3.657013912120325*^9},
 CellID->623331724],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1221343315-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6570139123387256`*^9},
 CellID->330571709]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"0", ",", "0", ",", "0"}], "\[RightAngleBracket]"}], 
  UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]3"], 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"1", ",", "2", ",", "3"}], 
   "\[RightAngleBracket]"}]}]], "Computation",
 CellChangeTimes->{{3.594467723945379*^9, 3.594467748049016*^9}, {
  3.645763184284588*^9, 3.64576318839495*^9}},
 CellID->361486860],

Cell[BoxData[
 FormBox["True", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.607339219350796*^9, 3.610097558819793*^9, 
  3.615714001631713*^9, 3.615797923066301*^9, 3.615801323849364*^9, 
  3.6314474804250813`*^9, 3.635758036435273*^9, 3.645763245545463*^9, 
  3.6570139141327286`*^9},
 CellID->324244333]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c361486860"]]]]}, {None, None}},
 CellChangeTimes->{3.6570139145695295`*^9},
 CellID->69411863],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c361486860-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6570139147723303`*^9},
 CellID->556797800]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"7", ",", "1"}], "\[RightAngleBracket]"}], 
  UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]2"], 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"6", ",", "2"}], "\[RightAngleBracket]"}]}]], "Computation",
 CellChangeTimes->{{3.645763320281314*^9, 3.645763335586051*^9}},
 CellID->1508699062],

Cell[BoxData[
 FormBox["True", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645763339724745*^9, 3.657013916644333*^9},
 CellID->19039667]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1508699062"]]]]}, {None, None}},
 CellChangeTimes->{3.657013917096734*^9},
 CellID->19048599],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1508699062-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6570139172995343`*^9},
 CellID->584992054]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"7", ",", "0", ",", "1"}], "\[RightAngleBracket]"}], 
  UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]3"], 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"6", ",", "2", ",", "0"}], 
   "\[RightAngleBracket]"}]}]], "Computation",
 CellChangeTimes->{{3.594460778031136*^9, 3.594460800695582*^9}, {
  3.594461526041698*^9, 3.594461529181095*^9}, {3.594461878620723*^9, 
  3.594461899141757*^9}, {3.594462079942355*^9, 3.594462112940048*^9}, {
  3.607339239457304*^9, 3.607339242412022*^9}, {3.645763170138084*^9, 
  3.645763174752243*^9}},
 CellID->286294197],

Cell[BoxData[
 FormBox["False", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{{3.607339221600148*^9, 3.607339242845807*^9}, 
   3.610097561165618*^9, 3.615714004474963*^9, 3.61579792537529*^9, 
   3.61580132614625*^9, 3.63144748358661*^9, 3.63496721210321*^9, 
   3.635758039078506*^9, 3.645763248273024*^9, 3.6570139190311375`*^9},
 CellID->71897093]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c286294197"]]]]}, {None, None}},
 CellChangeTimes->{3.6570139195615387`*^9},
 CellID->125781715],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c286294197-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6570139197799387`*^9},
 CellID->541156049]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Polynomials", "Subsection",
 CellChangeTimes->{{3.645759892554407*^9, 3.645759894215169*^9}},
 CellID->1580386408],

Cell["", "OpenEnvironment",
 CellID->299725121],

Cell[CellGroupData[{

Cell["\<\
Definition: 3-Variate Polynomials over \[DoubleStruckCapitalQ] with \
Degree-Reverse-Lexicographic Ordering\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.594462652517175*^9, 3.594462691403917*^9}, {
  3.594462733801779*^9, 3.594462746009465*^9}},
 CellID->769281711],

Cell[BoxData[
 RowBox[{"\[ScriptCapitalQ]xyz", ":=", 
  RowBox[{"PolyTuples", "[", 
   RowBox[{"\[ScriptCapitalQ]", ",", "\[ScriptCapitalT]3"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalQ]xyz)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1375788385", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.594462696055421*^9, 3.594462765978983*^9}, {
  3.645763366165936*^9, 3.645763378185211*^9}, {3.645763465760006*^9, 
  3.645763469795065*^9}, {3.6457799703563423`*^9, 3.645779980966342*^9}},
 CellTags->{
  "ID:1375788385", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalQ]xyz"},
 CellID->1375788385],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->284898274]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->446422152],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"poly1", "=", 
  RowBox[{
   UnderscriptBox["<<", "\[ScriptCapitalQ]xyz"], 
   RowBox[{"[", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"7", "x", " ", "y", " ", "z"}], "-", 
      RowBox[{
       FractionBox["2", "3"], 
       SuperscriptBox["x", "2"]}], "+", 
      RowBox[{
       FractionBox["3", "8"], 
       SuperscriptBox["z", "2"]}], "-", "y", "-", "4"}], 
     TagBox[")",
      "AutoParentheses"]}], "]"}]}]}]], "Computation",
 CellChangeTimes->{{3.64576398516992*^9, 3.645764004047679*^9}, {
  3.6457800065349894`*^9, 3.645780006823175*^9}},
 CellID->342550773],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"7", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"1", ",", "1", ",", "1"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"-", 
       FractionBox["2", "3"]}], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"2", ",", "0", ",", "0"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      FractionBox["3", "8"], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"0", ",", "0", ",", "2"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"0", ",", "1", ",", "0"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"-", "4"}], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"0", ",", "0", ",", "0"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}]}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.6457635260454187`*^9, 3.645763589252672*^9, 
  3.6457640088338823`*^9, 3.645764210139819*^9, 3.645764529555179*^9, 
  3.6457648228543367`*^9, 3.6457661788126097`*^9},
 CellID->2126912284]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c342550773"]]]]}, {None, None}},
 CellChangeTimes->{3.6457661793983088`*^9},
 CellID->1970788451],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c342550773-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645766179741029*^9},
 CellID->353659982]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"poly2", "=", 
  RowBox[{
   UnderscriptBox["<<", "\[ScriptCapitalQ]xyz"], 
   RowBox[{"[", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"3", 
       SuperscriptBox["x", "5"], 
       SuperscriptBox["z", "3"]}], "-", 
      RowBox[{"x", " ", 
       SuperscriptBox["y", "3"], 
       SuperscriptBox["z", "3"]}], "+", 
      RowBox[{
       FractionBox["4", "5"], 
       SuperscriptBox["x", "3"], "y", " ", 
       SuperscriptBox["z", "2"]}], "-", 
      RowBox[{"3", "x", " ", "y", " ", "z"}], "+", 
      RowBox[{
       FractionBox["1", "2"], 
       SuperscriptBox["z", "2"]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "]"}]}]}]], "Computation",
 CellChangeTimes->{{3.645764629886777*^9, 3.645764636482317*^9}, 
   3.645765426347533*^9, 3.645766391563362*^9, {3.6457800104674253`*^9, 
   3.645780010764881*^9}},
 CellID->1172786959],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"5", ",", "0", ",", "3"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"1", ",", "3", ",", "3"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      FractionBox["4", "5"], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"3", ",", "1", ",", "2"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"-", "3"}], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"1", ",", "1", ",", "1"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      FractionBox["1", "2"], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"0", ",", "0", ",", "2"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}]}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645764638093253*^9, 3.645765427621984*^9, 
  3.6457661818856163`*^9, 3.645766392840357*^9},
 CellID->1664446879]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1172786959"]]]]}, {None, None}},
 CellChangeTimes->{3.6457663934305677`*^9},
 CellID->116485574],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1172786959-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645766393768714*^9},
 CellID->1171363384]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"poly3", "=", 
  RowBox[{
   UnderscriptBox["<<", "\[ScriptCapitalQ]xyz"], 
   RowBox[{"[", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"3", 
       SuperscriptBox["x", "2"], "z"}], "-", "y"}], 
     TagBox[")",
      "AutoParentheses"]}], "]"}]}]}]], "Computation",
 CellChangeTimes->{{3.645764665612184*^9, 3.645764671747218*^9}, 
   3.6457663868506517`*^9, {3.645780015291027*^9, 3.64578001559838*^9}},
 CellID->1934266891],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"3", ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"2", ",", "0", ",", "1"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}], ",", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"-", "1"}], ",", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"0", ",", "1", ",", "0"}], "\[RightAngleBracket]"}]}], 
     "\[RightAngleBracket]"}]}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645764672720992*^9, 3.645766184654084*^9, 
  3.645766395186729*^9},
 CellID->1218180393]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1934266891"]]]]}, {None, None}},
 CellChangeTimes->{3.645766395766296*^9},
 CellID->1620691346],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1934266891-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645766396108508*^9},
 CellID->511072915]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 UnderscriptBox["0", "\[ScriptCapitalQ]xyz"]], "Computation",
 CellChangeTimes->{{3.594462829433548*^9, 3.594462833648411*^9}, {
  3.608017401934998*^9, 3.608017426555695*^9}, {3.645764687918786*^9, 
  3.645764706509848*^9}, {3.645780017755313*^9, 3.645780018089044*^9}},
 CellID->174499053],

Cell[BoxData[
 FormBox[
  RowBox[{
  "\[LeftAngleBracket]", 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{
  3.607339549171228*^9, {3.607339584893338*^9, 3.607339612365939*^9}, {
   3.607339677198781*^9, 3.607339702155318*^9}, {3.608017371699668*^9, 
   3.608017379353537*^9}, {3.608017419247571*^9, 3.608017427475157*^9}, 
   3.608022677856492*^9, 3.608023189532117*^9, 3.610098197795411*^9, 
   3.615714207883848*^9, {3.615715010408205*^9, 3.61571502510691*^9}, 
   3.615801387218752*^9, 3.6314487507488613`*^9, {3.6457646982848997`*^9, 
   3.645764707410831*^9}, 3.6457661869907417`*^9},
 CellID->622829772]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c174499053"]]]]}, {None, None}},
 CellChangeTimes->{3.645766187546783*^9},
 CellID->1258356781],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c174499053-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645766187880597*^9},
 CellID->835786078]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 UnderscriptBox["1", "\[ScriptCapitalQ]xyz"]], "Computation",
 CellChangeTimes->{{3.594462861022269*^9, 3.594462864874531*^9}, {
  3.645780022836759*^9, 3.645780023101982*^9}},
 CellID->715169566],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"1", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"0", ",", "0", ",", "0"}], "\[RightAngleBracket]"}]}], 
    "\[RightAngleBracket]"}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{{3.60733967910075*^9, 3.607339703817065*^9}, 
   3.608017376845251*^9, 3.608017429393859*^9, 3.608022679640739*^9, 
   3.608023191344335*^9, 3.610098206033902*^9, 3.615714210920287*^9, {
   3.615715012627264*^9, 3.615715027032525*^9}, 3.615801389396498*^9, 
   3.631448753699223*^9, 3.645764709823059*^9, 3.6457661894613047`*^9},
 CellID->886905339]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c715169566"]]]]}, {None, None}},
 CellChangeTimes->{3.6457661900365543`*^9},
 CellID->1108652826],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c715169566-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645766190379096*^9},
 CellID->1600102270]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", "\[ScriptCapitalQ]xyz"], 
  RowBox[{"[", 
   RowBox[{"poly2", 
    UnderscriptBox["+", "\[ScriptCapitalQ]xyz"], "poly1"}], 
   "]"}]}]], "Computation",
 CellChangeTimes->{{3.645765350118043*^9, 3.645765358447914*^9}, {
  3.6457800263047953`*^9, 3.645780030140196*^9}},
 CellID->853820995],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"-", "4"}], "-", 
   FractionBox[
    RowBox[{"2", " ", 
     SuperscriptBox["x", "2"]}], "3"], "-", "y", "+", 
   RowBox[{"4", " ", "x", " ", "y", " ", "z"}], "+", 
   FractionBox[
    RowBox[{"7", " ", 
     SuperscriptBox["z", "2"]}], "8"], "+", 
   RowBox[{
    FractionBox["4", "5"], " ", 
    SuperscriptBox["x", "3"], " ", "y", " ", 
    SuperscriptBox["z", "2"]}], "+", 
   RowBox[{"3", " ", 
    SuperscriptBox["x", "5"], " ", 
    SuperscriptBox["z", "3"]}], "-", 
   RowBox[{"x", " ", 
    SuperscriptBox["y", "3"], " ", 
    SuperscriptBox["z", 
     "3"]}]}], $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645765360059493*^9, 3.645765438218364*^9, 
  3.645766192311069*^9, 3.645766398908619*^9},
 CellID->676131250]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c853820995"]]]]}, {None, None}},
 CellChangeTimes->{3.645766399475238*^9},
 CellID->1821544516],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c853820995-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645766399826702*^9},
 CellID->1704096014]
}, Closed]],

Cell[CellGroupData[{

Cell["rdm", "Subsubsection",
 CellChangeTimes->{{3.645765455538384*^9, 3.645765455872617*^9}},
 CellID->1216185475],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", "\[ScriptCapitalQ]xyz"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["rdm", "\[ScriptCapitalQ]xyz"], "[", 
    RowBox[{"poly1", ",", "poly2"}], "]"}], "]"}]}]], "Computation",
 CellChangeTimes->{{3.645765549586059*^9, 3.645765550544517*^9}, {
  3.645780035151434*^9, 3.645780039716502*^9}},
 CellID->1933070300],

Cell[BoxData[
 FormBox["0", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.6457655514092484`*^9, 3.6457661964965267`*^9, 
  3.645766411523917*^9},
 CellID->1659834227]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1933070300"]]]]}, {None, None}},
 CellChangeTimes->{3.645766412120782*^9},
 CellID->445837838],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1933070300-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645766412452237*^9},
 CellID->1713495286]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", "\[ScriptCapitalQ]xyz"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["rdm", "\[ScriptCapitalQ]xyz"], "[", 
    RowBox[{"poly2", ",", "poly1"}], "]"}], "]"}]}]], "Computation",
 CellChangeTimes->{
  3.645765378635017*^9, {3.6457800410444307`*^9, 3.645780041749384*^9}},
 CellID->1181004978],

Cell[BoxData[
 FormBox[
  RowBox[{
   RowBox[{"-", 
    FractionBox["1", "7"]}], " ", 
   SuperscriptBox["y", "2"], " ", 
   SuperscriptBox["z", 
    "2"]}], $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645765380162537*^9, 3.645765504997562*^9, 
  3.645766199473411*^9, 3.645766413956168*^9},
 CellID->611302185]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1181004978"]]]]}, {None, None}},
 CellChangeTimes->{3.645766414513838*^9},
 CellID->442383335],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1181004978-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645766414848254*^9},
 CellID->1889001502]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", "\[ScriptCapitalQ]xyz"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["rdm", "\[ScriptCapitalQ]xyz"], "[", 
    RowBox[{"poly1", ",", "poly3"}], "]"}], "]"}]}]], "Computation",
 CellChangeTimes->{{3.645765498423123*^9, 3.64576549989614*^9}, {
  3.645780042857295*^9, 3.645780044294766*^9}},
 CellID->829389969],

Cell[BoxData[
 FormBox["0", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645765501475936*^9, 3.645766201760437*^9, 
  3.645766416433666*^9},
 CellID->1237458846]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c829389969"]]]]}, {None, None}},
 CellChangeTimes->{3.64576641699306*^9},
 CellID->301477109],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c829389969-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645766417322541*^9},
 CellID->2087589975]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", "\[ScriptCapitalQ]xyz"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["rdm", "\[ScriptCapitalQ]xyz"], "[", 
    RowBox[{"poly3", ",", "poly1"}], "]"}], "]"}]}]], "Computation",
 CellChangeTimes->{{3.645765517243834*^9, 3.645765518606854*^9}, {
  3.645780045512334*^9, 3.645780046472439*^9}},
 CellID->806902398],

Cell[BoxData[
 FormBox["0", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645765519489339*^9, 3.64576620403376*^9, 
  3.6457664184854603`*^9},
 CellID->1613258704]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c806902398"]]]]}, {None, None}},
 CellChangeTimes->{3.645766419038682*^9},
 CellID->1603337453],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c806902398-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6457664193705463`*^9},
 CellID->236846226]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", "\[ScriptCapitalQ]xyz"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["rdm", "\[ScriptCapitalQ]xyz"], "[", 
    RowBox[{"poly2", ",", "poly3"}], "]"}], "]"}]}]], "Computation",
 CellChangeTimes->{
  3.6457655592084417`*^9, {3.64578004835179*^9, 3.645780049422667*^9}},
 CellID->1918925811],

Cell[BoxData[
 FormBox[
  RowBox[{
   SuperscriptBox["x", "3"], " ", 
   SuperscriptBox["z", 
    "2"]}], $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645765560035543*^9, 3.64576620659756*^9, 
  3.645766421250575*^9},
 CellID->993729665]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1918925811"]]]]}, {None, None}},
 CellChangeTimes->{3.645766421828837*^9},
 CellID->938624865],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1918925811-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6457664221678677`*^9},
 CellID->1603628063]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", "\[ScriptCapitalQ]xyz"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["rdm", "\[ScriptCapitalQ]xyz"], "[", 
    RowBox[{"poly3", ",", "poly2"}], "]"}], "]"}]}]], "Computation",
 CellChangeTimes->{{3.6457655659867268`*^9, 3.6457655673135023`*^9}, {
  3.645780050565607*^9, 3.64578005134164*^9}},
 CellID->1741569015],

Cell[BoxData[
 FormBox["0", $CellContext`TheoremaForm]], "ComputationResult",
 CellChangeTimes->{3.645765568147441*^9, 3.6457662088795424`*^9, 
  3.6457664238220263`*^9},
 CellID->1732022027]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1741569015"]]]]}, {None, None}},
 CellChangeTimes->{3.6457664245407677`*^9},
 CellID->1921971350],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1741569015-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645766424876966*^9},
 CellID->1105863375]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["lcrd", "Subsubsection",
 CellChangeTimes->{{3.6457654615785847`*^9, 3.645765462104848*^9}},
 CellID->190084383],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalQ]xyz", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["lcrd", "\[ScriptCapitalQ]xyz"], "[", 
    RowBox[{"poly1", ",", "poly2", ",", "1", ",", "1"}], "]"}], 
   "]"}]}]], "Computation",
 CellChangeTimes->{{3.645764724438426*^9, 3.645764734513777*^9}, {
  3.645764829190384*^9, 3.645764835295643*^9}, {3.64578005767222*^9, 
  3.645780058815464*^9}},
 CellID->1777111242],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    SuperscriptBox["x", "5"], " ", "y", " ", 
    SuperscriptBox["z", "3"]}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645764735303334*^9, 3.645764836161963*^9, 
  3.645764941499309*^9, 3.645765131638033*^9, 3.645765217311171*^9, 
  3.645765266273521*^9, 3.645765299285797*^9, 3.645765431157659*^9, 
  3.64576621355851*^9, 3.64576644005368*^9},
 CellID->266105935]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1777111242"]]]]}, {None, None}},
 CellChangeTimes->{3.645766440665159*^9},
 CellID->2144087656],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1777111242-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645766441010068*^9},
 CellID->1031259517]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalQ]xyz", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["lcrd", "\[ScriptCapitalQ]xyz"], "[", 
    RowBox[{"poly1", ",", "poly3", ",", "1", ",", "1"}], "]"}], 
   "]"}]}]], "Computation",
 CellChangeTimes->{
  3.645765584040159*^9, {3.645780060584959*^9, 3.645780061552779*^9}},
 CellID->999756278],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    SuperscriptBox["x", "2"], " ", "y", " ", "z"}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645765584885375*^9, 3.6457662158439817`*^9, 
  3.645766442235869*^9},
 CellID->333730610]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c999756278"]]]]}, {None, None}},
 CellChangeTimes->{3.645766442799624*^9},
 CellID->262548371],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c999756278-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6457664431387453`*^9},
 CellID->839150767]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalQ]xyz", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["lcrd", "\[ScriptCapitalQ]xyz"], "[", 
    RowBox[{"poly2", ",", "poly3", ",", "1", ",", "1"}], "]"}], 
   "]"}]}]], "Computation",
 CellChangeTimes->{{3.645765589972931*^9, 3.645765591409313*^9}, {
  3.6457800627049913`*^9, 3.645780063590839*^9}},
 CellID->1213404407],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    SuperscriptBox["x", "5"], " ", 
    SuperscriptBox["z", "3"]}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645765592267527*^9, 3.645766218237338*^9, 
  3.645766444475341*^9},
 CellID->240280514]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1213404407"]]]]}, {None, None}},
 CellChangeTimes->{3.6457664450655413`*^9},
 CellID->838793558],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1213404407-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.6457664453995047`*^9},
 CellID->1634384220]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["cpm", "Subsubsection",
 CellChangeTimes->{{3.645765466228437*^9, 3.645765466599585*^9}},
 CellID->800084030],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalQ]xyz", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["cpm", "\[ScriptCapitalQ]xyz"], "[", 
    RowBox[{"poly1", ",", "poly2", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"1", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"5", ",", "1", ",", "3"}], "\[RightAngleBracket]"}]}], 
       "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], ",", "1", ",", 
     "1"}], "]"}], "]"}]}]], "Computation",
 CellChangeTimes->{{3.6457653271772957`*^9, 3.6457653335195293`*^9}, {
  3.645780067199642*^9, 3.645780068380158*^9}},
 CellID->1088310899],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    FractionBox[
     RowBox[{
      SuperscriptBox["x", "4"], " ", 
      SuperscriptBox["z", "2"]}], "7"], ",", 
    FractionBox["y", "3"]}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.64576533509509*^9, 3.645765434466372*^9, 
  3.645766229689699*^9, 3.645766448272992*^9},
 CellID->1642074452]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1088310899"]]]]}, {None, None}},
 CellChangeTimes->{3.645766448868815*^9},
 CellID->2048206854],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1088310899-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645766449201729*^9},
 CellID->2139465790]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalQ]xyz", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["cpm", "\[ScriptCapitalQ]xyz"], "[", 
    RowBox[{"poly1", ",", "poly3", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"1", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"2", ",", "1", ",", "1"}], "\[RightAngleBracket]"}]}], 
       "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], ",", "1", ",", 
     "1"}], "]"}], "]"}]}]], "Computation",
 CellChangeTimes->{{3.645765602164707*^9, 3.6457656093862753`*^9}, {
  3.645780069812755*^9, 3.6457800708083467`*^9}},
 CellID->1164918771],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    FractionBox["x", "7"], ",", 
    FractionBox["y", "3"]}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645765610471281*^9, 3.645766232819556*^9, 
  3.6457662667156467`*^9, 3.645766450997334*^9},
 CellID->1954185803]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c1164918771"]]]]}, {None, None}},
 CellChangeTimes->{3.645766451592238*^9},
 CellID->1294693027],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c1164918771-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645766451928073*^9},
 CellID->475108056]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  UnderscriptBox[">>", 
   RowBox[{"DomainTuples", "[", "\[ScriptCapitalQ]xyz", "]"}]], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["cpm", "\[ScriptCapitalQ]xyz"], "[", 
    RowBox[{"poly2", ",", "poly3", ",", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"1", ",", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"5", ",", "0", ",", "3"}], "\[RightAngleBracket]"}]}], 
       "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], ",", "1", ",", 
     "1"}], "]"}], "]"}]}]], "Computation",
 CellChangeTimes->{{3.6457656162063913`*^9, 3.645765621435318*^9}, {
  3.645780072431006*^9, 3.645780073423128*^9}},
 CellID->697197138],

Cell[BoxData[
 FormBox[
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{
    FractionBox["1", "3"], ",", 
    FractionBox[
     RowBox[{
      SuperscriptBox["x", "3"], " ", 
      SuperscriptBox["z", "2"]}], "3"]}], 
   "\[RightAngleBracket]"}], $CellContext`TheoremaForm]], "ComputationResult",\

 CellChangeTimes->{3.645765622483877*^9, 3.645766257927593*^9, 
  3.64576635940349*^9, 3.645766454165551*^9},
 CellID->126279016]
}, Open  ]],

Cell[CellGroupData[{

Cell["", "ComputationInfo",
 CellAutoOverwrite->True,
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "c697197138"]]]]}, {None, None}},
 CellChangeTimes->{3.645766454704507*^9},
 CellID->1368887684],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"c697197138-display.m\"}]"]], {
     Theorema`Interface`GUI`Private`FORM -> 
      Theorema`Interface`GUI`Private`displayFormulaFromKey, 
      Theorema`Interface`GUI`Private`RESULT -> 
      Theorema`Common`theoremaDisplay}], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{56., {2., 9.}}]], "ComputationInfoBody",
 CellAutoOverwrite->True,
 CellChangeTimes->{3.645766455044561*^9},
 CellID->1052908597]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Theorem: PolyTuples Belong to Category of Reduction Polynomials (\[Checkmark])\
\>", "Section",
 CellChangeTimes->{{3.6457595406233997`*^9, 3.645759569365563*^9}, {
  3.6564132852141275`*^9, 3.6564132886149335`*^9}},
 CellID->444952074],

Cell[TextData[{
 "For proving that the reduction polynomials in tuple representation belong \
to the category of polynomial reduction rings, we have to show this not for \
all ",
 Cell[BoxData[
  FormBox["\[ScriptCapitalR]", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["\[ScriptCapitalT]", TraditionalForm]]],
 ", but only for the \[OpenCurlyDoubleQuote]arbitrary but fixed\
\[CloseCurlyDoubleQuote] constants \[ScriptCapitalR] and \[ScriptCapitalT] \
from theory \[OpenCurlyDoubleQuote]Polynomials.nb\[CloseCurlyDoubleQuote]."
}], "Text",
 CellChangeTimes->{{3.631601138786726*^9, 3.6316012370861053`*^9}, {
  3.634886337193138*^9, 3.6348863889125233`*^9}, {3.64575996780709*^9, 
  3.645759979171678*^9}, {3.6661769918175955`*^9, 3.6661769920047956`*^9}},
 CellID->707250474],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"\[ScriptCapitalP]", "=", 
   RowBox[{"PolyTuples", "[", 
    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
    "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 395821466}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.655526427244481*^9, 3.6555264381956997`*^9}},
 CellTags->{
  "ID:395821466", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->395821466],

Cell[CellGroupData[{

Cell["Auxiliary Results", "Subsection",
 CellChangeTimes->{{3.6555263389589252`*^9, 3.6555263434517336`*^9}},
 CellID->62492604],

Cell[CellGroupData[{

Cell["Membership in PolyTuples (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.655526348896143*^9, 3.6555263543249526`*^9}, {
  3.655537309884431*^9, 3.6555373143304386`*^9}, {3.6555491480859165`*^9, 
  3.655549156463132*^9}},
 CellID->15574753],

Cell["", "OpenEnvironment",
 CellID->88808795],

Cell[CellGroupData[{

Cell["\<\
Lemma: Alternative Characterizations of \[Element] of PolyTuples\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.655526374682988*^9, 3.6555263906916165`*^9}},
 CellID->318992154],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "p", "]"}]}]], 
  RowBox[{"isTuple", "[", "p", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples isTuple)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:689486298", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6562331911179433`*^9, 3.656233204689967*^9}},
 CellTags->{
  "ID:689486298", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isTuple", "IR"},
 CellID->689486298],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "p"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "p", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", 
        RowBox[{"ProdDomain", "[", 
         RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
        "]"}]], 
      RowBox[{"[", "p", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] PolyTuples DomainTuples)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:302371753", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6555264112846527`*^9, 3.6555264845911818`*^9}},
 CellTags->{
  "ID:302371753", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples DomainTuples"},
 CellID->302371753],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
  RowBox[{"[", 
   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] PolyTuples empty)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:619631752", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.655526512562031*^9, 3.6555265218440475`*^9}},
 CellTags->{
  "ID:619631752", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples empty", "IR"},
 CellID->619631752],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"isTuple", "[", "p", "]"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "p", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"p", "\[NotEqual]", 
       RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
      "\[Implies]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[And]", 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"ProdDomain", "[", 
               RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
               "]"}]], 
             RowBox[{"[", 
              SubscriptBox["p", "1"], "]"}]}]},
           {
            RowBox[{
             SubscriptBox[
              RowBox[{"(", 
               SubscriptBox["p", "1"], ")"}], "1"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
           {
            RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
             RowBox[{"[", 
              RowBox[{"Rest", "[", "p", "]"}], "]"}]}]},
           {
            RowBox[{
             RowBox[{
              RowBox[{"Rest", "[", "p", "]"}], "\[NotEqual]", 
              RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
             "\[Implies]", 
             RowBox[{
              SubscriptBox[
               RowBox[{"(", 
                SubscriptBox["p", "1"], ")"}], "2"], 
              UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
              SubscriptBox[
               RowBox[{"(", 
                SubscriptBox[
                 RowBox[{"Rest", "[", "p", "]"}], "1"], ")"}], "2"]}]}]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] PolyTuples rec)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:93154383", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.655527475845725*^9, 3.6555274984969645`*^9}, {
  3.6555275861701183`*^9, 3.655527732849577*^9}, {3.655534009259226*^9, 
  3.6555340123792315`*^9}},
 CellTags->{
  "ID:93154383", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples rec", "IR"},
 CellID->93154383],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"c", ",", "t", ",", 
    RowBox[{"p", "..."}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], ",", 
         RowBox[{"p", "..."}]}], "\[RightAngleBracket]"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
           RowBox[{"[", "c", "]"}]}]},
         {
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
           RowBox[{"[", "t", "]"}]}]},
         {
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
           RowBox[{"[", 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"p", "..."}], "\[RightAngleBracket]"}], "]"}]}]},
         {
          RowBox[{"c", "\[NotEqual]", 
           UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
         {
          RowBox[{
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"p", "..."}], "\[RightAngleBracket]"}], "\[NotEqual]", 
             RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
            TagBox[")",
             "AutoParentheses"]}], "\[Implies]", 
           RowBox[{"t", 
            UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
            SubscriptBox[
             RowBox[{"(", 
              SubscriptBox[
               RowBox[{"\[LeftAngleBracket]", 
                RowBox[{"p", "..."}], "\[RightAngleBracket]"}], "1"], ")"}], 
             "2"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] PolyTuples rec II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:110840922", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6555265288796597`*^9, 3.655526700365161*^9}, {
  3.6555267473066435`*^9, 3.655526789130317*^9}, {3.655528101157025*^9, 
  3.6555281271466703`*^9}},
 CellTags->{
  "ID:110840922", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples rec II", "IR"},
 CellID->110840922],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"h", ",", 
    RowBox[{"r", "..."}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"h", ",", 
         RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
         RowBox[{"[", "c", "]"}]}], ",", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
         RowBox[{"[", "t", "]"}]}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"\[And]", 
        RowBox[{"\[Piecewise]", GridBox[{
           {
            RowBox[{"c", "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
           {
            RowBox[{
             RowBox[{
              TagBox["(",
               "AutoParentheses"], 
              RowBox[{
               RowBox[{"\[LeftAngleBracket]", 
                RowBox[{"r", "..."}], "\[RightAngleBracket]"}], "\[NotEqual]", 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
              TagBox[")",
               "AutoParentheses"]}], "\[Implies]", 
             RowBox[{"t", 
              UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
              SubscriptBox[
               RowBox[{"(", 
                SubscriptBox[
                 RowBox[{"\[LeftAngleBracket]", 
                  RowBox[{"r", "..."}], "\[RightAngleBracket]"}], "1"], ")"}],
                "2"]}]}]},
           {
            RowBox[{"h", "\[Equal]", 
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}]}]},
           {
            RowBox[{
             UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
             RowBox[{"[", 
              RowBox[{"\[LeftAngleBracket]", 
               RowBox[{"r", "..."}], "\[RightAngleBracket]"}], "]"}]}]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] PolyTuples rec III)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:45787564", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6556447279126015`*^9, 3.655644832402585*^9}, {
  3.6556449057851143`*^9, 3.655644915285531*^9}, {3.655645033954939*^9, 
  3.6556450467937617`*^9}},
 CellTags->{
  "ID:45787564", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples rec III", "IR"},
 CellID->45787564],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", "p", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 34480440}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.655647928109227*^9},
 CellTags->{
  "ID:34480440", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->34480440],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"c", ",", "t"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], 
        TagBox["\:293b",
         Identity,
         SyntaxForm->"a*b"], "p"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
           RowBox[{"[", "c", "]"}]}]},
         {
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
           RowBox[{"[", "t", "]"}]}]},
         {
          RowBox[{"c", "\[NotEqual]", 
           UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
         {
          RowBox[{
           RowBox[{
            TagBox["(",
             "AutoParentheses"], 
            RowBox[{"p", "\[NotEqual]", 
             RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
            TagBox[")",
             "AutoParentheses"]}], "\[Implies]", 
           RowBox[{"t", 
            UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
            SubscriptBox[
             RowBox[{"(", 
              SubscriptBox["p", "1"], ")"}], "2"]}]}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] PolyTuples prepend)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:134755829", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656046351116314*^9, 3.6560463887903805`*^9}, {
  3.6560464582271023`*^9, 3.6560464834991465`*^9}, {3.656046528084025*^9, 
  3.656046533310034*^9}, {3.656046979034017*^9, 3.6560469815768213`*^9}},
 CellTags->{
  "ID:134755829", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples prepend", "IR"},
 CellID->134755829],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "p", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", 
    SubscriptBox[
     RowBox[{"(", 
      SubscriptBox["p", "i"], ")"}], "1"], 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples membership first)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:218126712", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6556479430168533`*^9, 3.6556479699737005`*^9}},
 CellTags->{
  "ID:218126712", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership first", "IR"},
 CellID->218126712],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "p", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", 
    SubscriptBox[
     RowBox[{"(", 
      SubscriptBox["p", "i"], ")"}], "2"], 
    "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples membership second)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:36095185", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.655647974047308*^9, 3.6556479790705166`*^9}},
 CellTags->{
  "ID:36095185", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership second", "IR"},
 CellID->36095185],

Cell[BoxData[
 RowBox[{"\[Or]", 
  RowBox[{"\[Piecewise]", GridBox[{
     {
      RowBox[{"p", "\[Equal]", 
       RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]},
     {
      RowBox[{
       UnderscriptBox["\[Exists]", 
        RowBox[{
         RowBox[{
          UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
          RowBox[{"[", "c", "]"}]}], ",", 
         RowBox[{
          UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
          RowBox[{"[", "t", "]"}]}], ",", 
         RowBox[{"r", "..."}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{"\[And]", 
         RowBox[{"\[Piecewise]", GridBox[{
            {
             RowBox[{"c", "\[NotEqual]", 
              UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
            {
             RowBox[{
              UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
              RowBox[{"[", 
               RowBox[{"\[LeftAngleBracket]", 
                RowBox[{"r", "..."}], "\[RightAngleBracket]"}], "]"}]}]},
            {
             RowBox[{
              RowBox[{
               TagBox["(",
                "AutoParentheses"], 
               RowBox[{
                RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{"r", "..."}], "\[RightAngleBracket]"}], 
                "\[NotEqual]", 
                RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               TagBox[")",
                "AutoParentheses"]}], "\[Implies]", 
              RowBox[{"t", 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               SubscriptBox[
                RowBox[{"(", 
                 SubscriptBox[
                  RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{"r", "..."}], "\[RightAngleBracket]"}], "1"], 
                 ")"}], "2"]}]}]},
            {
             RowBox[{"p", "\[Equal]", 
              RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], ",", 
                RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}]}]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[Element] PolyTuples shape)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:266402037", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6555387080170927`*^9, 3.655538779528618*^9}, {
  3.6555388189508877`*^9, 3.655538865348369*^9}, {3.655647926939225*^9, 
  3.6556479341808376`*^9}},
 CellTags->{
  "ID:266402037", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples shape"},
 CellID->266402037],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"\[LeftBracketingBar]", "p", "\[RightBracketingBar]"}], 
    "\[Equal]", "1"}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", "c", "]"}]}], ",", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
       RowBox[{"[", "t", "]"}]}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{"c", "\[NotEqual]", 
       UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"p", "\[Equal]", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], 
         "\[RightAngleBracket]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples length=1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:314594392", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656304601110423*^9, 3.6563046466957035`*^9}},
 CellTags->{
  "ID:314594392", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples length=1"},
 CellID->314594392],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->91535020]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->558368068],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples isTuple)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples isTuple)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isTuple", "[", 
           StyleBox["p", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isTuple", "[", 
           StyleBox["p", "ExpressionVariable"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:689486298"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p689486298-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p689486298-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:689486298", "Proof|ID:689486298-1"},
 CellID->851460394],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p689486298-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->16049767]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] PolyTuples DomainTuples)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] PolyTuples DomainTuples)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["p", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
                RowBox[{"ProdDomain", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}], "]"}]], "[", 
             StyleBox["p", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["p", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"DomainTuples", "[", 
                RowBox[{"ProdDomain", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}], "]"}]], "[", 
             StyleBox["p", "ExpressionVariable"], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:302371753"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p302371753-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p302371753-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:302371753", "Proof|ID:302371753-1"},
 CellID->426132571],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p302371753-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->241646450]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] PolyTuples empty)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] PolyTuples empty)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", 
          RowBox[{"PolyTuples", "[", 
            RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}]], 
         "[", 
         RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", 
          RowBox[{"PolyTuples", "[", 
            RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}]], 
         "[", 
         RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:619631752"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p619631752-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p619631752-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:619631752", "Proof|ID:619631752-1"},
 CellID->236569764],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p619631752-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->229281091]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] PolyTuples rec)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] PolyTuples rec)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTuple", "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[Implies]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"ProdDomain", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   SubscriptBox[
                    StyleBox["p", "ExpressionVariable"], "1"], "]"}], 
                 "\[And]", 
                 RowBox[{
                   SubscriptBox[
                    SubscriptBox[
                    StyleBox["p", "ExpressionVariable"], "1"], "1"], 
                   "\[NotEqual]", 
                   UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{"Rest", "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}], 
                 "\[And]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"Rest", "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}],
                    "\[Implies]", 
                   RowBox[{
                    SubscriptBox[
                    SubscriptBox[
                    StyleBox["p", "ExpressionVariable"], "1"], "2"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    SubscriptBox[
                    SubscriptBox[
                    RowBox[{"Rest", "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "1"], 
                    "2"]}]}]}]}], ")"}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isTuple", "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[Implies]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"ProdDomain", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   SubscriptBox[
                    StyleBox["p", "ExpressionVariable"], "1"], "]"}], 
                 "\[And]", 
                 RowBox[{
                   SubscriptBox[
                    SubscriptBox[
                    StyleBox["p", "ExpressionVariable"], "1"], "1"], 
                   "\[NotEqual]", 
                   UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{"Rest", "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}], 
                 "\[And]", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"Rest", "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}],
                    "\[Implies]", 
                   RowBox[{
                    SubscriptBox[
                    SubscriptBox[
                    StyleBox["p", "ExpressionVariable"], "1"], "2"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    SubscriptBox[
                    SubscriptBox[
                    RowBox[{"Rest", "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "1"], 
                    "2"]}]}]}]}], ")"}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:93154383"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p93154383-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p93154383-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:93154383", "Proof|ID:93154383-1"},
 CellID->290123111],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p93154383-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->117687721]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] PolyTuples rec II)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] PolyTuples rec II)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"p", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"p", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                 StyleBox["t", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox[
                    RowBox[{"p", "..."}], "ExpressionVariable"], 
                   "\[RightAngleBracket]"}], "]"}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"p", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "\[NotEqual]", 
                   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                   SubscriptBox[
                    SubscriptBox[
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"p", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "1"], "2"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"p", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"p", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                 StyleBox["t", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox[
                    RowBox[{"p", "..."}], "ExpressionVariable"], 
                   "\[RightAngleBracket]"}], "]"}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"p", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "\[NotEqual]", 
                   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                   SubscriptBox[
                    SubscriptBox[
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"p", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "1"], "2"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:110840922"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p110840922-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p110840922-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:110840922", "Proof|ID:110840922-1"},
 CellID->99121591],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p110840922-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->642964469]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] PolyTuples rec III)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] PolyTuples rec III)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["h", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["h", "ExpressionVariable"], ",", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "\[NotEqual]", 
                   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                   SubscriptBox[
                    SubscriptBox[
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "1"], "2"]}]}], "\[And]", 
               RowBox[{
                 StyleBox["h", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["h", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["h", "ExpressionVariable"], ",", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "\[NotEqual]", 
                   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                   SubscriptBox[
                    SubscriptBox[
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "1"], "2"]}]}], "\[And]", 
               RowBox[{
                 StyleBox["h", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                   "\[RightAngleBracket]"}], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:45787564"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p45787564-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p45787564-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:45787564", "Proof|ID:45787564-1"},
 CellID->483935740],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p45787564-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->13219130]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] PolyTuples prepend)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] PolyTuples prepend)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             StyleBox["t", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "\:293b", 
                 StyleBox["p", "ExpressionVariable"]}], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                 StyleBox["t", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                   SubscriptBox[
                    SubscriptBox[
                    StyleBox["p", "ExpressionVariable"], "1"], "2"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             StyleBox["t", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "\:293b", 
                 StyleBox["p", "ExpressionVariable"]}], "]"}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"(", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                 StyleBox["c", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                 StyleBox["t", "ExpressionVariable"], "]"}], "\[And]", 
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
                   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                   SubscriptBox[
                    SubscriptBox[
                    StyleBox["p", "ExpressionVariable"], "1"], "2"]}]}]}], 
             ")"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:134755829"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p134755829-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p134755829-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:134755829", "Proof|ID:134755829-1"},
 CellID->955362936],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p134755829-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->11634208]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples membership first)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples membership first)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
           SubscriptBox[
            SubscriptBox[
             StyleBox["p", "ExpressionVariable"], 
             StyleBox["i", "ExpressionVariable"]], "1"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
           SubscriptBox[
            SubscriptBox[
             StyleBox["p", "ExpressionVariable"], 
             StyleBox["i", "ExpressionVariable"]], "1"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:218126712"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p218126712-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p218126712-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:218126712", "Proof|ID:218126712-1"},
 CellID->69811972],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p218126712-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->436964383]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples membership second)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples membership second)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
           SubscriptBox[
            SubscriptBox[
             StyleBox["p", "ExpressionVariable"], 
             StyleBox["i", "ExpressionVariable"]], "2"], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
           SubscriptBox[
            SubscriptBox[
             StyleBox["p", "ExpressionVariable"], 
             StyleBox["i", "ExpressionVariable"]], "2"], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:36095185"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p36095185-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p36095185-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:36095185", "Proof|ID:36095185-1"},
 CellID->66892996],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p36095185-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->319554292]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[Element] PolyTuples shape)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[Element] PolyTuples shape)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[Equal]", 
             RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
           "\[Or]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}, {
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                   "\[RightAngleBracket]"}], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "\[NotEqual]", 
                   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                   SubscriptBox[
                    SubscriptBox[
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "1"], "2"]}]}], "\[And]", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[Equal]", 
             RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
           "\[Or]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}, {
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                   "\[RightAngleBracket]"}], "]"}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "\[NotEqual]", 
                   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
                 "\[Implies]", 
                 RowBox[{
                   StyleBox["t", "ExpressionVariable"], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                   SubscriptBox[
                    SubscriptBox[
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "1"], "2"]}]}], "\[And]", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:266402037"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p266402037-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p266402037-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:266402037", "Proof|ID:266402037-1"},
 CellID->167245154],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p266402037-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->5730164]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples length=1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples length=1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["p", "ExpressionVariable"], "\[RightBracketingBar]"}],
              "\[Equal]", "1"}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], 
                   "\[RightAngleBracket]"}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{"\[LeftBracketingBar]", 
               StyleBox["p", "ExpressionVariable"], "\[RightBracketingBar]"}],
              "\[Equal]", "1"}], "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                   StyleBox["c", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], 
                   
                   "\[RightAngleBracket]"}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:314594392"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p314594392-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p314594392-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:314594392", "Proof|ID:314594392-1"},
 CellID->424017657],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p314594392-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->350735652]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Induction Rules (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.655549967970157*^9, 3.6555499720417643`*^9}, {
  3.6555626017280073`*^9, 3.655562605690414*^9}, {3.656052494942322*^9, 
  3.656052495925123*^9}},
 CellID->48602793],

Cell["", "OpenEnvironment",
 CellID->66481260],

Cell[CellGroupData[{

Cell["Lemma: Induction Rules for PolyTuples", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6555499902781963`*^9, 3.655549998859212*^9}, {
  3.656052497968727*^9, 3.6560524989515285`*^9}},
 CellID->56424937],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", "P"]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 441087600}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.655550001887617*^9, 3.655550002901619*^9}},
 CellTags->{
  "ID:441087600", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->441087600],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "p", "]"}]}]], 
    RowBox[{"P", "[", "p", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", 
       RowBox[{"DomainTuples", "[", 
        RowBox[{"ProdDomain", "[", 
         RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
        "]"}]], 
      RowBox[{"[", "p", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
        RowBox[{"[", "p", "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}], "\[Implies]", 
      RowBox[{"P", "[", "p", "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples \[ForAll] alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:678316554", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6555503787902794`*^9, 3.655550421083954*^9}, {
  3.6555510147065973`*^9, 3.6555510358456345`*^9}},
 CellTags->{
  "ID:678316554", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[ForAll] alternative"},
 CellID->678316554],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"P", "[", 
     RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], "]"}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
         RowBox[{"[", "c", "]"}]}], ",", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
         RowBox[{"[", "t", "]"}]}], ",", 
        RowBox[{"r", "..."}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"\[And]", 
          RowBox[{"\[Piecewise]", GridBox[{
             {
              RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
               RowBox[{"[", 
                RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{"r", "..."}], "\[RightAngleBracket]"}], "]"}]}]},
             {
              RowBox[{"P", "[", 
               RowBox[{"\[LeftAngleBracket]", 
                RowBox[{"r", "..."}], "\[RightAngleBracket]"}], "]"}]},
             {
              RowBox[{"c", "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
             {
              RowBox[{
               RowBox[{
                RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{"r", "..."}], "\[RightAngleBracket]"}], 
                "\[NotEqual]", 
                RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[Implies]", 
               RowBox[{"t", 
                UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                SubscriptBox[
                 RowBox[{"(", 
                  SubscriptBox[
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{"r", "..."}], "\[RightAngleBracket]"}], "1"], 
                  ")"}], "2"]}]}]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
         TagBox[")",
          "AutoParentheses"]}], "\[Implies]", 
        RowBox[{"P", "[", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "p", "]"}]}]], 
    RowBox[{"P", "[", "p", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples induction)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:329976481", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6555500438526907`*^9, 3.65555027083809*^9}},
 CellTags->{
  "ID:329976481", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples induction", "IR"},
 CellID->329976481],

Cell[TextData[{
 "The only difference between the two induction rules is that in the one \
below ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], ",", 
       RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], "]"}]}], 
   TraditionalForm]]],
 " may be assumed as well."
}], "EnvironmentText",
 CellChangeTimes->{{3.6560517269605722`*^9, 3.656051783838272*^9}},
 CellID->120515761],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"P", "[", 
     RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], "]"}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
         RowBox[{"[", "c", "]"}]}], ",", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
         RowBox[{"[", "t", "]"}]}], ",", 
        RowBox[{"r", "..."}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"\[And]", 
          RowBox[{"\[Piecewise]", GridBox[{
             {
              RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
               RowBox[{"[", 
                RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{"r", "..."}], "\[RightAngleBracket]"}], "]"}]}]},
             {
              RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
               RowBox[{"[", 
                RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                  RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], ",", 
                  RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], "]"}]}]},
             {
              RowBox[{"P", "[", 
               RowBox[{"\[LeftAngleBracket]", 
                RowBox[{"r", "..."}], "\[RightAngleBracket]"}], "]"}]},
             {
              RowBox[{"c", "\[NotEqual]", 
               UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
             {
              RowBox[{
               RowBox[{
                RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{"r", "..."}], "\[RightAngleBracket]"}], 
                "\[NotEqual]", 
                RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
               "\[Implies]", 
               RowBox[{"t", 
                UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                SubscriptBox[
                 RowBox[{"(", 
                  SubscriptBox[
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{"r", "..."}], "\[RightAngleBracket]"}], "1"], 
                  ")"}], "2"]}]}]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
         TagBox[")",
          "AutoParentheses"]}], "\[Implies]", 
        RowBox[{"P", "[", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], "]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", "p", "]"}]}]], 
    RowBox[{"P", "[", "p", "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples induction')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:379661071", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.656051664217262*^9, {3.6560516990053225`*^9, 3.6560517180373564`*^9}},
 CellTags->{
  "ID:379661071", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples induction'", "IR"},
 CellID->379661071],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->122025085]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->70749121],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples \[ForAll] alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples \[ForAll] alternative)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 StyleBox["P", "ExpressionVariable"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}], "]"}]], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[Implies]", 
               RowBox[{
                 StyleBox["P", "ExpressionVariable"], "[", 
                 StyleBox["p", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 StyleBox["P", "ExpressionVariable"], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"DomainTuples", "[", 
                    RowBox[{"ProdDomain", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}], "]"}]], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], "\[Implies]", 
               RowBox[{
                 StyleBox["P", "ExpressionVariable"], "[", 
                 StyleBox["p", "ExpressionVariable"], 
                 "]"}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:678316554"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p678316554-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p678316554-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:678316554", "Proof|ID:678316554-1"},
 CellID->716699503],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p678316554-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->496685325]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples induction)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples induction)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], "]"}],
              "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}, {
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}], "\[And]", 
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}], "\[And]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "\[NotEqual]", 
                    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}],
                     "\[Implies]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    SubscriptBox[
                    SubscriptBox[
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "1"], "2"]}]}]}], "\[Implies]", 
                 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "]"}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               StyleBox["p", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], "]"}],
              "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}, {
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}], "\[And]", 
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}], "\[And]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "\[NotEqual]", 
                    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}],
                     "\[Implies]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    SubscriptBox[
                    SubscriptBox[
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "1"], "2"]}]}]}], "\[Implies]", 
                 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "]"}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               StyleBox["p", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:329976481"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p329976481-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p329976481-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:329976481", "Proof|ID:329976481-1"},
 CellID->3050184],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p329976481-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->87959892]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples induction')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples induction')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], "]"}],
              "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}, {
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "]"}], "\[And]", 
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}], "\[And]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "\[NotEqual]", 
                    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}],
                     "\[Implies]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    SubscriptBox[
                    SubscriptBox[
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "1"], "2"]}]}]}], "\[Implies]", 
                 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "]"}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               StyleBox["p", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], "]"}],
              "\[And]", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}, {
                   RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}, {
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}}]], 
               RowBox[{
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}], "\[And]", 
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "]"}], "\[And]", 
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "]"}], "\[And]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "\[NotEqual]", 
                    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}],
                     "\[Implies]", 
                    RowBox[{
                    StyleBox["t", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    SubscriptBox[
                    SubscriptBox[
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "1"], "2"]}]}]}], "\[Implies]", 
                 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "]"}]}]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               StyleBox["P", "ExpressionVariable"], "[", 
               StyleBox["p", "ExpressionVariable"], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:379661071"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p379661071-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p379661071-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:379661071", "Proof|ID:379661071-1"},
 CellID->321989072],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p379661071-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->625844841]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Power-Products are Strictly Decreasing (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.655549107929446*^9, 3.6555491188046656`*^9}, {
  3.655648386797833*^9, 3.6556483952842484`*^9}, {3.6556983312936525`*^9, 
  3.6556983375024633`*^9}},
 CellID->29457529],

Cell["", "OpenEnvironment",
 CellID->1102219],

Cell[CellGroupData[{

Cell["\<\
Lemma: Power Products in Polynomials are Strictly Decreasing I\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6555495431096106`*^9, 3.6555495534992294`*^9}, {
  3.6555625809955707`*^9, 3.655562584068776*^9}, {3.6556481694240513`*^9, 
  3.6556481697048516`*^9}},
 CellID->45823306],

Cell[BoxData[
 RowBox[{"isTotalIrreflOrder", "[", 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    TagBox[
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"],
     Identity], 
    TagBox[")",
     "AutoParentheses"]}], ",", "\[ScriptCapitalT]"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(isTotalIrreflOrder \[Succeeds] \[ScriptCapitalT])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:485298517", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.655553727958969*^9, 3.65555375085981*^9}},
 CellTags->{
  "ID:485298517", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isTotalIrreflOrder \[Succeeds] \[ScriptCapitalT]"},
 CellID->485298517],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", "p", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 561716382}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.655549562250845*^9, 3.655549569426857*^9}},
 CellTags->{
  "ID:561716382", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->561716382],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{
     RowBox[{"\[LeftBracketingBar]", "p", "\[RightBracketingBar]"}], "-", 
     "1"}]}]], 
  RowBox[{
   UnderscriptBox[
    UnderscriptBox["\[ForAll]", 
     RowBox[{"j", "\[Element]", "\[DoubleStruckCapitalN]"}]], 
    RowBox[{
     RowBox[{"i", "+", "j"}], "\[Element]", 
     SubscriptBox["\[DoubleStruckCapitalN]", 
      RowBox[{"1", ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "p", "\[RightBracketingBar]"}]}]]}]], 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     SubscriptBox[
      RowBox[{"(", 
       SubscriptBox["p", "i"], ")"}], "2"], 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
     SubscriptBox[
      RowBox[{"(", 
       SubscriptBox["p", 
        RowBox[{"i", "+", "j"}]], ")"}], "2"]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples pp decreasing ind)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:107805279", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6555535277030168`*^9, 3.6555535746620994`*^9}, {
  3.6555536085191593`*^9, 3.65555360878436*^9}},
 CellTags->{
  "ID:107805279", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp decreasing ind"},
 CellID->107805279],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", 
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "p", "\[RightBracketingBar]"}]}]], 
   RowBox[{"i", ">", "j"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    SubscriptBox[
     RowBox[{"(", 
      SubscriptBox["p", "j"], ")"}], "2"], 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
    SubscriptBox[
     RowBox[{"(", 
      SubscriptBox["p", "i"], ")"}], "2"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples pp decreasing)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:220534381", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6555495741692657`*^9, 3.6555496167885404`*^9}},
 CellTags->{
  "ID:220534381", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp decreasing"},
 CellID->220534381],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "p", "\[RightBracketingBar]"}]}]], 
   RowBox[{"i", "\[NotEqual]", "1"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    SubscriptBox[
     RowBox[{"(", 
      SubscriptBox["p", "1"], ")"}], "2"], 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
    SubscriptBox[
     RowBox[{"(", 
      SubscriptBox["p", "i"], ")"}], "2"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples first pp max)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:5709061", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6555496330281687`*^9, 3.6555496809876537`*^9}},
 CellTags->{
  "ID:5709061", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max"},
 CellID->5709061],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "p", "\[RightBracketingBar]"}]}]], 
  RowBox[{"\[Not]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     SubscriptBox[
      RowBox[{"(", 
       SubscriptBox["p", "i"], ")"}], "2"], 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
     SubscriptBox[
      RowBox[{"(", 
       SubscriptBox["p", "1"], ")"}], "2"]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples first pp max')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:133904640", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6556223050407863`*^9, 3.6556223155396047`*^9}},
 CellTags->{
  "ID:133904640", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max'"},
 CellID->133904640],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", 
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", "p", "\[RightBracketingBar]"}]}]], 
   RowBox[{"i", "\[NotEqual]", "j"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    SubscriptBox[
     RowBox[{"(", 
      SubscriptBox["p", "i"], ")"}], "2"], "\[NotEqual]", 
    SubscriptBox[
     RowBox[{"(", 
      SubscriptBox["p", "j"], ")"}], "2"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples pp distinct)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:12736634", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.655550561858601*^9, 3.6555505737146215`*^9}},
 CellTags->{
  "ID:12736634", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp distinct"},
 CellID->12736634],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->99753670]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->145531536],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(isTotalIrreflOrder \[Succeeds] \[ScriptCapitalT])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(isTotalIrreflOrder \[Succeeds] \[ScriptCapitalT])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isTotalIrreflOrder", "[", 
         RowBox[{
           UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], ",", 
           "\[ScriptCapitalT]"}], "]"}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isTotalIrreflOrder", "[", 
         RowBox[{
           UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], ",", 
           "\[ScriptCapitalT]"}], "]"}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:485298517"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p485298517-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p485298517-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:485298517", "Proof|ID:485298517-1"},
 CellID->638746182],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p485298517-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->34385987]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples pp decreasing ind)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples pp decreasing ind)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["p", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], "-", "1"}]}]}, {
             RowBox[{
               StyleBox["j", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "+", 
               StyleBox["j", "ExpressionVariable"]}], "\[Element]", 
             SubscriptBox["\[DoubleStruckCapitalN]", 
              RowBox[{"1", ",", "\[Ellipsis]", ",", 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["p", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}]}]]}], "\[Implies]", 
           RowBox[{
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], 
               StyleBox["i", "ExpressionVariable"]], "2"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], 
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "+", 
                 StyleBox["j", "ExpressionVariable"]}]], 
              "2"]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 RowBox[{"\[LeftBracketingBar]", 
                   StyleBox["p", "ExpressionVariable"], 
                   "\[RightBracketingBar]"}], "-", "1"}]}]}, {
             RowBox[{
               StyleBox["j", "ExpressionVariable"], "\[Element]", 
               "\[DoubleStruckCapitalN]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "+", 
               StyleBox["j", "ExpressionVariable"]}], "\[Element]", 
             SubscriptBox["\[DoubleStruckCapitalN]", 
              RowBox[{"1", ",", "\[Ellipsis]", ",", 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["p", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}]}]]}], "\[Implies]", 
           RowBox[{
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], 
               StyleBox["i", "ExpressionVariable"]], "2"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], 
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "+", 
                 StyleBox["j", "ExpressionVariable"]}]], 
              "2"]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:107805279"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p107805279-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p107805279-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:107805279", "Proof|ID:107805279-1"},
 CellID->218458895],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p107805279-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->263931670]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples pp decreasing)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples pp decreasing)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["i", "ExpressionVariable"], ">", 
             StyleBox["j", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], 
               StyleBox["j", "ExpressionVariable"]], "2"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], 
               StyleBox["i", "ExpressionVariable"]], 
              "2"]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["i", "ExpressionVariable"], ">", 
             StyleBox["j", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], 
               StyleBox["j", "ExpressionVariable"]], "2"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], 
               StyleBox["i", "ExpressionVariable"]], 
              "2"]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:220534381"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p220534381-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p220534381-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:220534381", "Proof|ID:220534381-1"},
 CellID->457923117],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p220534381-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->110795177]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples first pp max)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples first pp max)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["i", "ExpressionVariable"], "\[NotEqual]", "1"}], 
           "\[Implies]", 
           RowBox[{
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], "1"], "2"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], 
               StyleBox["i", "ExpressionVariable"]], 
              "2"]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["i", "ExpressionVariable"], "\[NotEqual]", "1"}], 
           "\[Implies]", 
           RowBox[{
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], "1"], "2"], 
             UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], 
               StyleBox["i", "ExpressionVariable"]], 
              "2"]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:5709061"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p5709061-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p5709061-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:5709061", "Proof|ID:5709061-1"},
 CellID->199136965],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p5709061-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->55914907]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples first pp max')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples first pp max')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               SubscriptBox[
                SubscriptBox[
                 StyleBox["p", "ExpressionVariable"], 
                 StyleBox["i", "ExpressionVariable"]], "2"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               SubscriptBox[
                SubscriptBox[
                 StyleBox["p", "ExpressionVariable"], "1"], "2"]}], 
             TagBox[")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{"\[Not]", 
           RowBox[{
             TagBox["(", "AutoParentheses"], 
             RowBox[{
               SubscriptBox[
                SubscriptBox[
                 StyleBox["p", "ExpressionVariable"], 
                 StyleBox["i", "ExpressionVariable"]], "2"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               SubscriptBox[
                SubscriptBox[
                 StyleBox["p", "ExpressionVariable"], "1"], "2"]}], 
             TagBox[
             ")", "AutoParentheses"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:133904640"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p133904640-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p133904640-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:133904640", "Proof|ID:133904640-1"},
 CellID->185035893],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p133904640-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->169582083]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples pp distinct)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples pp distinct)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox["j", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], 
               StyleBox["i", "ExpressionVariable"]], "2"], "\[NotEqual]", 
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], 
               StyleBox["j", "ExpressionVariable"]], 
              "2"]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["i", "ExpressionVariable"], "\[NotEqual]", 
             StyleBox["j", "ExpressionVariable"]}], "\[Implies]", 
           RowBox[{
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], 
               StyleBox["i", "ExpressionVariable"]], "2"], "\[NotEqual]", 
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], 
               StyleBox["j", "ExpressionVariable"]], 
              "2"]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:12736634"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p12736634-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p12736634-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:12736634", "Proof|ID:12736634-1"},
 CellID->225795913],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p12736634-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->1112935]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->297754641],

Cell[CellGroupData[{

Cell["\<\
Lemma: Power Products in Polynomials are Strictly Decreasing II\
\>", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.655648181202072*^9, 3.6556481821068735`*^9}},
 CellID->480387543],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"c", ",", "t", ",", 
   RowBox[{"r", "..."}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 723780914}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6556481856948795`*^9, 3.6556481901096873`*^9}},
 CellTags->{
  "ID:723780914", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->723780914],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], "]"}]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 284026551}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.655648227987554*^9, 3.655648231809561*^9}},
 CellTags->{
  "ID:284026551", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->284026551],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
   RowBox[{"\[LeftBracketingBar]", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"r", "..."}], "\[RightAngleBracket]"}], 
    "\[RightBracketingBar]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 548327311}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6556482557400026`*^9, 3.655648260810012*^9}},
 CellTags->{
  "ID:548327311", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->548327311],

Cell[BoxData[
 RowBox[{"t", 
  UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
  SubscriptBox[
   RowBox[{"(", 
    SubscriptBox[
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"r", "..."}], "\[RightAngleBracket]"}], "i"], ")"}], 
   "2"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples first pp max seq)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:387361496", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6556470045736046`*^9, 3.655647083681343*^9}, {
  3.6556482031211104`*^9, 3.65564826558362*^9}},
 CellTags->{
  "ID:387361496", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max seq"},
 CellID->387361496],

Cell[BoxData[
 RowBox[{"\[Not]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    SubscriptBox[
     RowBox[{"(", 
      SubscriptBox[
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"r", "..."}], "\[RightAngleBracket]"}], "i"], ")"}], "2"], 
    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples first pp max seq')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:70991441", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6556482721044316`*^9, 3.6556482881412597`*^9}},
 CellTags->{
  "ID:70991441", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max seq'"},
 CellID->70991441],

Cell[BoxData[
 RowBox[{"t", "\[NotEqual]", 
  SubscriptBox[
   RowBox[{"(", 
    SubscriptBox[
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"r", "..."}], "\[RightAngleBracket]"}], "i"], ")"}], 
   "2"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples pp distinct seq)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:365249767", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6556482988750787`*^9, 3.655648309873098*^9}},
 CellTags->{
  "ID:365249767", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp distinct seq"},
 CellID->365249767],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->16959157]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->47716413],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples first pp max seq)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples first pp max seq)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               SubscriptBox[
                SubscriptBox[
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                   "\[RightAngleBracket]"}], 
                 StyleBox["i", "ExpressionVariable"]], 
                "2"]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               SubscriptBox[
                SubscriptBox[
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                   "\[RightAngleBracket]"}], 
                 StyleBox["i", "ExpressionVariable"]], 
                "2"]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:387361496"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p387361496-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p387361496-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:387361496", "Proof|ID:387361496-1"},
 CellID->597599025],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p387361496-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->278306075]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples first pp max seq')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples first pp max seq')\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   SubscriptBox[
                    SubscriptBox[
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                    StyleBox["i", "ExpressionVariable"]], "2"], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                   StyleBox["t", "ExpressionVariable"]}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{"\[Not]", 
               RowBox[{
                 TagBox["(", "AutoParentheses"], 
                 RowBox[{
                   SubscriptBox[
                    SubscriptBox[
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                    StyleBox["i", "ExpressionVariable"]], "2"], 
                   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                   StyleBox["t", "ExpressionVariable"]}], 
                 TagBox[
                 ")", "AutoParentheses"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:70991441"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p70991441-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p70991441-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:70991441", "Proof|ID:70991441-1"},
 CellID->346547304],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p70991441-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->635659261]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples pp distinct seq)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples pp distinct seq)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
               SubscriptBox[
                SubscriptBox[
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                   "\[RightAngleBracket]"}], 
                 StyleBox["i", "ExpressionVariable"]], 
                "2"]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
               SubscriptBox[
                SubscriptBox[
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                   "\[RightAngleBracket]"}], 
                 StyleBox["i", "ExpressionVariable"]], 
                "2"]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:365249767"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p365249767-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p365249767-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:365249767", "Proof|ID:365249767-1"},
 CellID->408497668],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p365249767-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->459278606]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Properties of the Coefficient Function (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.655648371416206*^9, 3.6556483782958183`*^9}, {
  3.6557019315907784`*^9, 3.6557019357715855`*^9}},
 CellID->92723618],

Cell["", "OpenEnvironment",
 CellID->269082022],

Cell[CellGroupData[{

Cell["Lemma: Properties of the Coefficient Function", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.655549124389475*^9, 3.6555491354030943`*^9}, {
  3.6556469664939375`*^9, 3.6556469767587557`*^9}},
 CellID->90259041],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", "p", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 16132582}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6555491642943454`*^9, 3.6555491844183807`*^9}, {
  3.6556164983917866`*^9, 3.655616498563387*^9}},
 CellTags->{
  "ID:16132582", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->16132582],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "p", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", 
      SubscriptBox[
       RowBox[{"(", 
        SubscriptBox["p", "i"], ")"}], "2"]}], "]"}], "\[Equal]", 
    SubscriptBox[
     RowBox[{"(", 
      SubscriptBox["p", "i"], ")"}], "1"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples C 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:742018993", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6556165065350013`*^9, 3.655616536487054*^9}},
 CellTags->{
  "ID:742018993", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C 1"},
 CellID->742018993],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     RowBox[{
      UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
      RowBox[{"p", ",", "t"}], "]"}], "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Exists]", 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "p", "\[RightBracketingBar]"}]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"t", "\[Equal]", 
        SubscriptBox[
         RowBox[{"(", 
          SubscriptBox["p", "i"], ")"}], "2"]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples C \[NotEqual] 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:281673814", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6556287577386274`*^9, 3.655628802354706*^9}, {
  3.655628898156474*^9, 3.655628909138893*^9}},
 CellTags->{
  "ID:281673814", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C \[NotEqual] 0"},
 CellID->281673814],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]], 
  RowBox[{
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     RowBox[{
      UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
      RowBox[{"p", ",", "t"}], "]"}], "\[Equal]", 
     UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
    TagBox[")",
     "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
   RowBox[{
    TagBox["(",
     "AutoParentheses"], 
    RowBox[{
     UnderscriptBox["\[ForAll]", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "p", "\[RightBracketingBar]"}]}]], 
     RowBox[{
      TagBox["(",
       "AutoParentheses"], 
      RowBox[{"t", "\[NotEqual]", 
       SubscriptBox[
        RowBox[{"(", 
         SubscriptBox["p", "i"], ")"}], "2"]}], 
      TagBox[")",
       "AutoParentheses"]}]}], 
    TagBox[")",
     "AutoParentheses"]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples C=0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:122376381", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6555492024364123`*^9, 3.6555492538073025`*^9}, {
  3.6556164874405675`*^9, 3.65561649485058*^9}, {3.6556167135317645`*^9, 
  3.655616716136969*^9}},
 CellTags->{
  "ID:122376381", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C=0"},
 CellID->122376381],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
     RowBox[{"[", "c", "]"}]}], ",", 
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
     RowBox[{"[", 
      RowBox[{"s", ",", "t"}], "]"}]}], ",", 
    RowBox[{"r", "..."}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
       RowBox[{"[", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], ",", 
          RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], "]"}]}], "\[And]", 
      RowBox[{"s", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], ",", 
          RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], ",", "s"}], "]"}],
       "\[Equal]", 
      UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples C=0 seq)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:256177412", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6556468194012794`*^9, 3.655646906262232*^9}},
 CellTags->{
  "ID:256177412", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C=0 seq"},
 CellID->256177412],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"c", ",", "t", ",", 
    RowBox[{"r", "..."}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
      RowBox[{"[", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], ",", 
         RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"r", "..."}], "\[RightAngleBracket]"}], ",", "t"}], "]"}], 
      "\[Equal]", 
      UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples C=0 seq')", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:293354245", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656136672682327*^9, 3.656136700746776*^9}},
 CellTags->{
  "ID:293354245", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C=0 seq'"},
 CellID->293354245],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"c", ",", "t", ",", "s"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], 
       TagBox["\:293b",
        Identity,
        SyntaxForm->"a*b"], "p"}], ",", "s"}], "]"}], "\[Equal]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Piecewise]", GridBox[{
        {"c", "\[DoubleLeftArrow]", 
         RowBox[{"t", "\[Equal]", "s"}]},
        {
         RowBox[{
          UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
          RowBox[{"p", ",", "s"}], "]"}], "\[DoubleLeftArrow]", "otherwise"}
       }]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples C prepend)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:129332919", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656134268120303*^9, 3.6561343386636267`*^9}},
 CellTags->{
  "ID:129332919", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C prepend"},
 CellID->129332919],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->208838147]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->929166910],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples C 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples C 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               SubscriptBox[
                SubscriptBox[
                 StyleBox["p", "ExpressionVariable"], 
                 StyleBox["i", "ExpressionVariable"]], "2"]}], "]"}], 
           "\[Equal]", 
           SubscriptBox[
            SubscriptBox[
             StyleBox["p", "ExpressionVariable"], 
             StyleBox["i", "ExpressionVariable"]], 
            "1"]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               SubscriptBox[
                SubscriptBox[
                 StyleBox["p", "ExpressionVariable"], 
                 StyleBox["i", "ExpressionVariable"]], "2"]}], "]"}], 
           "\[Equal]", 
           SubscriptBox[
            SubscriptBox[
             StyleBox["p", "ExpressionVariable"], 
             StyleBox["i", "ExpressionVariable"]], 
            "1"]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:742018993"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p742018993-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p742018993-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:742018993", "Proof|ID:742018993-1"},
 CellID->161841610],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p742018993-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.6556243726714344`*^9, 3.655624375354639*^9}},
 CellID->117326975]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples C \[NotEqual] 0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples C \[NotEqual] 0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["C", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[NotEqual]", 
             
             UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["p", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[Equal]", 
               SubscriptBox[
                SubscriptBox[
                 StyleBox["p", "ExpressionVariable"], 
                 StyleBox["i", "ExpressionVariable"]], 
                "2"]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["C", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[NotEqual]", 
             
             UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["p", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[Equal]", 
               SubscriptBox[
                SubscriptBox[
                 StyleBox["p", "ExpressionVariable"], 
                 StyleBox["i", "ExpressionVariable"]], 
                "2"]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:281673814"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p281673814-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p281673814-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:281673814", "Proof|ID:281673814-1"},
 CellID->294457098],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p281673814-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->162988210]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples C=0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples C=0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["C", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["p", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
               SubscriptBox[
                SubscriptBox[
                 StyleBox["p", "ExpressionVariable"], 
                 StyleBox["i", "ExpressionVariable"]], 
                "2"]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["C", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["p", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[NotEqual]", 
               SubscriptBox[
                SubscriptBox[
                 StyleBox["p", "ExpressionVariable"], 
                 StyleBox["i", "ExpressionVariable"]], 
                "2"]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:122376381"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p122376381-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p122376381-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:122376381", "Proof|ID:122376381-1"},
 CellID->6731680],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p122376381-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->2400693]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples C=0 seq)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples C=0 seq)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             StyleBox[
              RowBox[{"r", "..."}], "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             StyleBox[
              RowBox[{"r", "..."}], "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[And]", 
             RowBox[{
               StyleBox["s", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               StyleBox["t", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:256177412"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p256177412-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p256177412-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:256177412", "Proof|ID:256177412-1"},
 CellID->606101095],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p256177412-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->21528983]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples C=0 seq')\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples C=0 seq')\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             UnderscriptBox[
             "0", "\[ScriptCapitalR]"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:293354245"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p293354245-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p293354245-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:293354245", "Proof|ID:293354245-1"},
 CellID->371170725],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p293354245-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->14209107]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples C prepend)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples C prepend)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             StyleBox["t", "ExpressionVariable"]}, {
             StyleBox["s", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "\:293b", 
                 StyleBox["p", "ExpressionVariable"]}], ",", 
               StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{"\[Piecewise]", 
             GridBox[{{
                StyleBox["c", "ExpressionVariable"], "\[DoubleLeftArrow]", 
                RowBox[{
                  StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                  StyleBox["s", "ExpressionVariable"]}]}, {
                RowBox[{
                  UnderscriptBox["C", 
                   RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                  RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                "\[DoubleLeftArrow]", 
                "True"}}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             StyleBox["c", "ExpressionVariable"]}, {
             StyleBox["t", "ExpressionVariable"]}, {
             StyleBox["s", "ExpressionVariable"]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "\:293b", 
                 StyleBox["p", "ExpressionVariable"]}], ",", 
               StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{"\[Piecewise]", 
             GridBox[{{
                StyleBox["c", "ExpressionVariable"], "\[DoubleLeftArrow]", 
                RowBox[{
                  StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                  StyleBox["s", "ExpressionVariable"]}]}, {
                RowBox[{
                  UnderscriptBox["C", 
                   RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                  RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                "\[DoubleLeftArrow]", 
                "True"}}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:129332919"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p129332919-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p129332919-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:129332919", "Proof|ID:129332919-1"},
 CellID->88799612],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p129332919-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->277564924]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Properties of Addition (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.656047421392394*^9, 3.6560474255732017`*^9}, {
  3.656225324358184*^9, 3.6562253287729917`*^9}},
 CellID->11122401],

Cell["", "OpenEnvironment",
 CellID->53346088],

Cell[CellGroupData[{

Cell["Lemma: Properties of Addition in PolyTuples I", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656047433014415*^9, 3.656047440221627*^9}, {
  3.656054090781325*^9, 3.6560540910933256`*^9}},
 CellID->610570600],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"p", ",", "q"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 136746483}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656047445198036*^9, 3.656047455353654*^9}},
 CellTags->{
  "ID:136746483", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->136746483],

Cell[BoxData[
 UnderscriptBox["let", 
  RowBox[{"r", "=", 
   RowBox[{"p", 
    UnderscriptBox["+", "\[ScriptCapitalP]"], "q"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 110691644}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6560474717336826`*^9, 3.6560474801264973`*^9}},
 CellTags->{
  "ID:110691644", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->110691644],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", 
   RowBox[{"DomainTuples", "[", 
    RowBox[{"ProdDomain", "[", 
     RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], "]"}]], 
  RowBox[{"[", "r", "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples + ProdDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:371581865", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6560480671603303`*^9, 3.656048069937135*^9}, {
  3.6560481149432144`*^9, 3.6560481393884573`*^9}},
 CellTags->{
  "ID:371581865", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + ProdDomain"},
 CellID->371581865],

Cell[BoxData[
 RowBox[{"isTuple", "[", "r", "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples + isTuple)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:694735421", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6560495979824195`*^9, 3.6560496007436247`*^9}},
 CellTags->{
  "ID:694735421", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + isTuple"},
 CellID->694735421],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "r", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    SubscriptBox[
     RowBox[{"(", 
      SubscriptBox["r", "i"], ")"}], "1"], "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples + \[NotEqual] 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:442662347", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656049853354868*^9, 3.6560498778313107`*^9}},
 CellTags->{
  "ID:442662347", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[NotEqual] 0"},
 CellID->442662347],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "r", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[Or]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["\[Exists]", 
          RowBox[{
           RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{"\[LeftBracketingBar]", "p", "\[RightBracketingBar]"}]}]], 
         
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{
           SubscriptBox[
            RowBox[{"(", 
             SubscriptBox["r", "i"], ")"}], "2"], "\[Equal]", 
           SubscriptBox[
            RowBox[{"(", 
             SubscriptBox["p", "j"], ")"}], "2"]}], 
          TagBox[")",
           "AutoParentheses"]}]}]},
       {
        RowBox[{
         UnderscriptBox["\[Exists]", 
          RowBox[{
           RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
           RowBox[{"\[LeftBracketingBar]", "q", "\[RightBracketingBar]"}]}]], 
         
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{
           SubscriptBox[
            RowBox[{"(", 
             SubscriptBox["r", "i"], ")"}], "2"], "\[Equal]", 
           SubscriptBox[
            RowBox[{"(", 
             SubscriptBox["q", "j"], ")"}], "2"]}], 
          TagBox[")",
           "AutoParentheses"]}]}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples + pp \[SubsetEqual])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:66376992", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656047466273673*^9, 3.6560475268173795`*^9}},
 CellTags->{
  "ID:66376992", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + pp \[SubsetEqual]"},
 CellID->66376992],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->580166704]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->469006967],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples + ProdDomain)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples + ProdDomain)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", 
              RowBox[{"ProdDomain", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}], "]"}]], "[", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["+", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], 
             StyleBox["q", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"DomainTuples", "[", 
              RowBox[{"ProdDomain", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}], "]"}]], "[", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["+", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], 
             StyleBox["q", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:371581865"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p371581865-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p371581865-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:371581865", "Proof|ID:371581865-1"},
 CellID->414829965],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p371581865-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellChangeTimes->{3.656049939170619*^9},
 CellID->847751266]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples + isTuple)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples + isTuple)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isTuple", "[", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["+", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], 
             StyleBox["q", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isTuple", "[", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["+", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], 
             StyleBox["q", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:694735421"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p694735421-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p694735421-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:694735421", "Proof|ID:694735421-1"},
 CellID->511476329],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p694735421-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->251211459]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples + \[NotEqual] 0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples + \[NotEqual] 0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   StyleBox["q", "ExpressionVariable"]}], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           SubscriptBox[
            SubscriptBox[
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["+", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], 
               StyleBox["q", "ExpressionVariable"]}], 
             StyleBox["i", "ExpressionVariable"]], "1"], "\[NotEqual]", 
           UnderscriptBox[
           "0", "\[ScriptCapitalR]"]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   StyleBox["q", "ExpressionVariable"]}], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           SubscriptBox[
            SubscriptBox[
             RowBox[{
               StyleBox["p", "ExpressionVariable"], 
               UnderscriptBox["+", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], 
               StyleBox["q", "ExpressionVariable"]}], 
             StyleBox["i", "ExpressionVariable"]], "1"], "\[NotEqual]", 
           UnderscriptBox[
           "0", "\[ScriptCapitalR]"]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:442662347"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p442662347-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p442662347-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:442662347", "Proof|ID:442662347-1"},
 CellID->874127013],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p442662347-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->341494949]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples + pp \[SubsetEqual])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples + pp \[SubsetEqual])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   StyleBox["q", "ExpressionVariable"]}], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["p", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               SubscriptBox[
                SubscriptBox[
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   StyleBox["q", "ExpressionVariable"]}], 
                 StyleBox["i", "ExpressionVariable"]], "2"], "\[Equal]", 
               SubscriptBox[
                SubscriptBox[
                 StyleBox["p", "ExpressionVariable"], 
                 StyleBox["j", "ExpressionVariable"]], "2"]}]}], "\[Or]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["q", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               SubscriptBox[
                SubscriptBox[
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   StyleBox["q", "ExpressionVariable"]}], 
                 StyleBox["i", "ExpressionVariable"]], "2"], "\[Equal]", 
               SubscriptBox[
                SubscriptBox[
                 StyleBox["q", "ExpressionVariable"], 
                 StyleBox["j", "ExpressionVariable"]], 
                "2"]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   StyleBox["q", "ExpressionVariable"]}], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["p", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               SubscriptBox[
                SubscriptBox[
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   StyleBox["q", "ExpressionVariable"]}], 
                 StyleBox["i", "ExpressionVariable"]], "2"], "\[Equal]", 
               SubscriptBox[
                SubscriptBox[
                 StyleBox["p", "ExpressionVariable"], 
                 StyleBox["j", "ExpressionVariable"]], "2"]}]}], "\[Or]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   RowBox[{
                    StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
                   "\[Ellipsis]", ",", 
                   RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["q", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]], 
             RowBox[{
               SubscriptBox[
                SubscriptBox[
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   StyleBox["q", "ExpressionVariable"]}], 
                 StyleBox["i", "ExpressionVariable"]], "2"], "\[Equal]", 
               SubscriptBox[
                SubscriptBox[
                 StyleBox["q", "ExpressionVariable"], 
                 StyleBox["j", "ExpressionVariable"]], 
                "2"]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:66376992"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p66376992-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p66376992-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:66376992", "Proof|ID:66376992-1"},
 CellID->490152990],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p66376992-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->654010454]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->144928206],

Cell[CellGroupData[{

Cell["Lemma: Properties of Addition in PolyTuples II", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656054098799739*^9, 3.6560540998137407`*^9}},
 CellID->736686],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"c1", ",", "c2", ",", "t1", ",", "t2", ",", 
   RowBox[{"r1", "..."}], ",", 
   RowBox[{"r2", "..."}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 480657044}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65605410510215*^9, 3.656054133119799*^9}, {
  3.6560541813706837`*^9, 3.6560542267667637`*^9}, {3.6560543817686357`*^9, 
  3.656054388523448*^9}},
 CellTags->{
  "ID:480657044", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->480657044],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
        RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], "]"}]}], "\[And]", 
    
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"c2", ",", "t2"}], "\[RightAngleBracket]"}], ",", 
        RowBox[{"r2", "..."}]}], "\[RightAngleBracket]"}], "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 907803476}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6560541367078056`*^9, 3.6560541716362667`*^9}},
 CellTags->{
  "ID:907803476", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->907803476],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"t1", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t2"}], "\[And]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"r1", "..."}], "\[RightAngleBracket]"}], 
       UnderscriptBox["+", "\[ScriptCapitalP]"], 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"c2", ",", "t2"}], "\[RightAngleBracket]"}], ",", 
         RowBox[{"r2", "..."}]}], "\[RightAngleBracket]"}]}], ")"}], 
     "\[NotEqual]", 
     RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"t1", 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
   SubscriptBox[
    RowBox[{"(", 
     SubscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"r1", "..."}], "\[RightAngleBracket]"}], 
        UnderscriptBox["+", "\[ScriptCapitalP]"], 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"c2", ",", "t2"}], "\[RightAngleBracket]"}], ",", 
          RowBox[{"r2", "..."}]}], "\[RightAngleBracket]"}]}], ")"}], "1"], 
     ")"}], "2"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples + \[Succeeds] 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:442268007", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656054257561218*^9, 3.6560542758912497`*^9}, {
  3.65605432153693*^9, 3.6560543616290007`*^9}, {3.656054464355181*^9, 
  3.6560544996424427`*^9}, {3.6560546923963814`*^9, 3.656054693878384*^9}},
 CellTags->{
  "ID:442268007", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[Succeeds] 1"},
 CellID->442268007],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"t2", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t1"}], "\[And]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
         RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], 
       UnderscriptBox["+", "\[ScriptCapitalP]"], 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"r2", "..."}], "\[RightAngleBracket]"}]}], ")"}], 
     "\[NotEqual]", 
     RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"t2", 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
   SubscriptBox[
    RowBox[{"(", 
     SubscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
          RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], 
        UnderscriptBox["+", "\[ScriptCapitalP]"], 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"r2", "..."}], "\[RightAngleBracket]"}]}], ")"}], "1"], 
     ")"}], "2"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples + \[Succeeds] 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:26802188", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656054574132574*^9, 3.6560546185302515`*^9}, {
  3.656054696093588*^9, 3.6560546973883905`*^9}},
 CellTags->{
  "ID:26802188", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[Succeeds] 2"},
 CellID->26802188],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{"t1", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t2"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{"t2", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t1"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"r1", "..."}], "\[RightAngleBracket]"}], 
       UnderscriptBox["+", "\[ScriptCapitalP]"], 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"r2", "..."}], "\[RightAngleBracket]"}]}], ")"}], 
     "\[NotEqual]", 
     RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{"t1", 
   UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
   SubscriptBox[
    RowBox[{"(", 
     SubscriptBox[
      RowBox[{"(", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"r1", "..."}], "\[RightAngleBracket]"}], 
        UnderscriptBox["+", "\[ScriptCapitalP]"], 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"r2", "..."}], "\[RightAngleBracket]"}]}], ")"}], "1"], 
     ")"}], "2"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples + \[Succeeds] 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:4890345", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6560546711959443`*^9, 3.6560547597729*^9}, {
  3.656057127450859*^9, 3.6560571276224594`*^9}},
 CellTags->{
  "ID:4890345", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[Succeeds] 3"},
 CellID->4890345],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->29004364]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->199903433],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples + \[Succeeds] 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples + \[Succeeds] 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["c2", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox["t2", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r2", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["t1", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 StyleBox["t2", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]}], 
             "\[Implies]", 
             RowBox[{
               StyleBox["t1", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               SubscriptBox[
                SubscriptBox[
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "1"], 
                "2"]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["c2", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox["t2", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r2", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["t1", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 StyleBox["t2", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]}], 
             "\[Implies]", 
             RowBox[{
               StyleBox["t1", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               SubscriptBox[
                SubscriptBox[
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "1"], 
                "2"]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:442268007"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p442268007-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p442268007-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:442268007", "Proof|ID:442268007-1"},
 CellID->23969624],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p442268007-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->196804738]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples + \[Succeeds] 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples + \[Succeeds] 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["c2", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox["t2", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r2", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["t2", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 StyleBox["t1", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]}], 
             "\[Implies]", 
             RowBox[{
               StyleBox["t2", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               SubscriptBox[
                SubscriptBox[
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "1"], 
                "2"]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["c2", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox["t2", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r2", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["t2", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 StyleBox["t1", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]}], 
             "\[Implies]", 
             RowBox[{
               StyleBox["t2", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               SubscriptBox[
                SubscriptBox[
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "1"], 
                "2"]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:26802188"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p26802188-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p26802188-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:26802188", "Proof|ID:26802188-1"},
 CellID->77790849],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p26802188-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->96981217]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples + \[Succeeds] 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples + \[Succeeds] 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["c2", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox["t2", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r2", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    StyleBox["t1", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t2", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    StyleBox["t2", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t1", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]}], 
             "\[Implies]", 
             RowBox[{
               StyleBox["t1", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               SubscriptBox[
                SubscriptBox[
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "1"], 
                "2"]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["c2", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox["t2", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r2", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    StyleBox["t1", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t2", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    StyleBox["t2", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t1", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "\[NotEqual]", 
                 RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]}], 
             "\[Implies]", 
             RowBox[{
               StyleBox["t1", "ExpressionVariable"], 
               UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
               SubscriptBox[
                SubscriptBox[
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                   UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "1"], 
                "2"]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:4890345"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p4890345-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p4890345-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:4890345", "Proof|ID:4890345-1"},
 CellID->650383762],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p4890345-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->139699914]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->39778867],

Cell[CellGroupData[{

Cell["Lemma: Properties of Negaton", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.65606266910218*^9, 3.6560626762313924`*^9}},
 CellID->9504892],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", "p", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 212784844}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6560631942781024`*^9, 3.6560632022965164`*^9}},
 CellTags->{
  "ID:212784844", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->212784844],

Cell[BoxData[
 RowBox[{"isTuple", "[", 
  RowBox[{
   UnderscriptBox["-", "\[ScriptCapitalP]"], "p"}], 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples - isTuple)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:106856858", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6560632146673384`*^9, 3.656063222997753*^9}},
 CellTags->{
  "ID:106856858", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - isTuple"},
 CellID->106856858],

Cell[BoxData[
 RowBox[{
  RowBox[{"\[LeftBracketingBar]", 
   RowBox[{
    UnderscriptBox["-", "\[ScriptCapitalP]"], "p"}], 
   "\[RightBracketingBar]"}], "\[Equal]", 
  RowBox[{
  "\[LeftBracketingBar]", "p", 
   "\[RightBracketingBar]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples - length)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:200895043", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656130376056265*^9, 3.656130392186693*^9}},
 CellTags->{
  "ID:200895043", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - length"},
 CellID->200895043],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[LeftBracketingBar]", "p", "\[RightBracketingBar]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    SubscriptBox[
     RowBox[{"(", 
      SubscriptBox["p", "i"], ")"}], "2"], "\[Equal]", 
    SubscriptBox[
     RowBox[{"(", 
      SubscriptBox[
       RowBox[{"(", 
        RowBox[{
         UnderscriptBox["-", "\[ScriptCapitalP]"], "p"}], ")"}], "i"], ")"}], 
     "2"]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples - pp)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:19952680", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656062705653044*^9, 3.6560627130942574`*^9}, {
   3.6560627438575115`*^9, 3.656062784869983*^9}, 3.656063205182522*^9},
 CellTags->{
  "ID:19952680", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - pp"},
 CellID->19952680],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->697897580]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->8829566],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples - isTuple)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples - isTuple)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isTuple", "[", 
           RowBox[{
             UnderscriptBox["Minus", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{"isTuple", "[", 
           RowBox[{
             UnderscriptBox["Minus", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:106856858"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p106856858-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p106856858-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:106856858", "Proof|ID:106856858-1"},
 CellID->555324734],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p106856858-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->2967280]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples - length)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples - length)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"\[LeftBracketingBar]", 
             RowBox[{
               UnderscriptBox["Minus", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             "\[RightBracketingBar]"}], "\[Equal]", 
           RowBox[{"\[LeftBracketingBar]", 
             StyleBox["p", "ExpressionVariable"], 
             "\[RightBracketingBar]"}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{"\[LeftBracketingBar]", 
             RowBox[{
               UnderscriptBox["Minus", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             "\[RightBracketingBar]"}], "\[Equal]", 
           RowBox[{"\[LeftBracketingBar]", 
             StyleBox["p", "ExpressionVariable"], 
             "\[RightBracketingBar]"}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:200895043"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p200895043-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p200895043-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:200895043", "Proof|ID:200895043-1"},
 CellID->593511758],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p200895043-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->5689008]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples - pp)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples - pp)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           SubscriptBox[
            SubscriptBox[
             StyleBox["p", "ExpressionVariable"], 
             StyleBox["i", "ExpressionVariable"]], "2"], "\[Equal]", 
           SubscriptBox[
            SubscriptBox[
             RowBox[{
               UnderscriptBox["Minus", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             StyleBox["i", "ExpressionVariable"]], 
            "2"]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 StyleBox["p", "ExpressionVariable"], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           SubscriptBox[
            SubscriptBox[
             StyleBox["p", "ExpressionVariable"], 
             StyleBox["i", "ExpressionVariable"]], "2"], "\[Equal]", 
           SubscriptBox[
            SubscriptBox[
             RowBox[{
               UnderscriptBox["Minus", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], 
             StyleBox["i", "ExpressionVariable"]], 
            "2"]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:19952680"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p19952680-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p19952680-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:19952680", "Proof|ID:19952680-1"},
 CellID->287758024],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p19952680-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->19013154]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Translations of Auxiliary Objects (\[Checkmark])\
\>", "Subsubsection",
 CellChangeTimes->{{3.656131692704578*^9, 3.6561317088506064`*^9}, {
  3.656241751670148*^9, 3.6562417560381556`*^9}},
 CellID->27460276],

Cell["", "OpenEnvironment",
 CellID->7836954],

Cell[CellGroupData[{

Cell["Lemma: Monomials in PolyTuples", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.65622716663982*^9, 3.6562271720062294`*^9}},
 CellID->639878451],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "m", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", "m", "]"}], 
    "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox[
       UnderscriptBox["\[Exists]", 
        RowBox[{
         RowBox[{
          UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
          RowBox[{"[", "c", "]"}]}], ",", 
         RowBox[{
          UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
          RowBox[{"[", "t", "]"}]}]}]], 
       RowBox[{"c", "\[NotEqual]", 
        UnderscriptBox["0", "\[ScriptCapitalR]"]}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{"m", "\[Equal]", 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], 
         "\[RightAngleBracket]"}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples isMonomial I)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:118174250", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656227525565251*^9, 3.6562275876689596`*^9}},
 CellTags->{
  "ID:118174250", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isMonomial I"},
 CellID->118174250],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
    RowBox[{"[", "c", "]"}]}], ",", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 78163703}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.656227380640996*^9},
 CellTags->{
  "ID:78163703", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->78163703],

Cell[BoxData[
 RowBox[{
  RowBox[{"c", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], 
       "\[RightAngleBracket]"}], "]"}]}], "\[And]", 
    RowBox[{
     UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], 
      "\[RightAngleBracket]"}], "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples isMonomial II)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:748742057", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656227631489437*^9, 3.656227654749078*^9}, {
  3.656227717039987*^9, 3.6562277194423914`*^9}, {3.6562277655414724`*^9, 
  3.6562277743242874`*^9}, {3.6562328419893303`*^9, 3.6562328458113365`*^9}, {
  3.6562328945146227`*^9, 3.6562329354178944`*^9}},
 CellTags->{
  "ID:748742057", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isMonomial II"},
 CellID->748742057],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "m", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isMonomial", "\[ScriptCapitalP]"], "[", 
     RowBox[{"m", ",", "c", ",", "t"}], "]"}], "\[DoubleLongLeftRightArrow]", 
    
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Piecewise]", GridBox[{
        {
         RowBox[{"m", "\[Equal]", 
          RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
         "\[DoubleLeftArrow]", 
         RowBox[{"c", "\[Equal]", 
          UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
        {
         RowBox[{"m", "\[Equal]", 
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], 
           "\[RightAngleBracket]"}]}], "\[DoubleLeftArrow]", "otherwise"}
       }]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples isMonomial III)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:104425812", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6562273928714175`*^9, 3.656227455989128*^9}},
 CellTags->{
  "ID:104425812", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isMonomial III"},
 CellID->104425812],

Cell[BoxData[
 RowBox[{
  RowBox[{"c", 
   UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "t"}], "\[Equal]", 
  RowBox[{"\[Piecewise]", GridBox[{
     {
      RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
      "\[DoubleLeftArrow]", 
      RowBox[{"c", "\[Equal]", 
       UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
     {
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], 
       "\[RightAngleBracket]"}], "\[DoubleLeftArrow]", "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples \[CenterDot])", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:191443971", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6562271761246367`*^9, 3.6562272248903227`*^9}, {
  3.656227379096594*^9, 3.6562273838390017`*^9}},
 CellTags->{
  "ID:191443971", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CenterDot]"},
 CellID->191443971],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->291201157]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->439301014],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples isMonomial I)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples isMonomial I)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", "\[ScriptCapitalR]"]}]], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], 
                 "\[RightAngleBracket]"}]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             StyleBox["m", "ExpressionVariable"], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox[
              UnderscriptBox["\[Exists]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                    StyleBox["c", "ExpressionVariable"], "]"}]}, {
                  RowBox[{
                    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                    StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
              RowBox[{
                StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
                UnderscriptBox["0", "\[ScriptCapitalR]"]}]], 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], "\[Equal]", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], 
                 "\[RightAngleBracket]"}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:118174250"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p118174250-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p118174250-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:118174250", "Proof|ID:118174250-1"},
 CellID->45600071],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p118174250-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->33001478]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples isMonomial II)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples isMonomial II)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], "]"}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], "]"}], 
             "\[And]", 
             RowBox[{
               UnderscriptBox["isMonomial", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:748742057"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p748742057-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p748742057-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:748742057", "Proof|ID:748742057-1"},
 CellID->104846486],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p748742057-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->561651541]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples isMonomial III)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples isMonomial III)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"\[Piecewise]", 
             GridBox[{{
                RowBox[{
                  StyleBox["m", "ExpressionVariable"], "\[Equal]", 
                  RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
                "\[DoubleLeftArrow]", 
                RowBox[{
                  StyleBox["c", "ExpressionVariable"], "\[Equal]", 
                  UnderscriptBox["0", "\[ScriptCapitalR]"]}]}, {
                RowBox[{
                  StyleBox["m", "ExpressionVariable"], "\[Equal]", 
                  RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\[RightAngleBracket]"}]}], 
                "\[DoubleLeftArrow]", 
                "True"}}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["m", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["isMonomial", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{
               StyleBox["m", "ExpressionVariable"], ",", 
               StyleBox["c", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{"\[Piecewise]", 
             GridBox[{{
                RowBox[{
                  StyleBox["m", "ExpressionVariable"], "\[Equal]", 
                  RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
                "\[DoubleLeftArrow]", 
                RowBox[{
                  StyleBox["c", "ExpressionVariable"], "\[Equal]", 
                  UnderscriptBox["0", "\[ScriptCapitalR]"]}]}, {
                RowBox[{
                  StyleBox["m", "ExpressionVariable"], "\[Equal]", 
                  RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\[RightAngleBracket]"}]}], 
                "\[DoubleLeftArrow]", 
                "True"}}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:104425812"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p104425812-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p104425812-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:104425812", "Proof|ID:104425812-1"},
 CellID->41476004],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p104425812-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->112588225]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples \[CenterDot])\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples \[CenterDot])\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], 
             UnderscriptBox["\[CenterDot]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], 
             StyleBox["t", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{"\[Piecewise]", 
             GridBox[{{
                RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
                "\[DoubleLeftArrow]", 
                RowBox[{
                  StyleBox["c", "ExpressionVariable"], "\[Equal]", 
                  UnderscriptBox["0", "\[ScriptCapitalR]"]}]}, {
                RowBox[{"\[LeftAngleBracket]", 
                  RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], 
                "\[DoubleLeftArrow]", 
                "True"}}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["c", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["c", "ExpressionVariable"], 
             UnderscriptBox["\[CenterDot]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], 
             StyleBox["t", "ExpressionVariable"]}], "\[Equal]", 
           RowBox[{"\[Piecewise]", 
             GridBox[{{
                RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}], 
                "\[DoubleLeftArrow]", 
                RowBox[{
                  StyleBox["c", "ExpressionVariable"], "\[Equal]", 
                  UnderscriptBox["0", "\[ScriptCapitalR]"]}]}, {
                RowBox[{"\[LeftAngleBracket]", 
                  RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], 
                "\[DoubleLeftArrow]", 
                "True"}}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:191443971"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p191443971-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p191443971-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:191443971", "Proof|ID:191443971-1"},
 CellID->309334880],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p191443971-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->255214453]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->41488312],

Cell[CellGroupData[{

Cell["Lemma: lp, lc and R in PolyTuples I", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6562242057132187`*^9, 3.656224213575632*^9}, {
  3.6562242591433125`*^9, 3.6562242597361135`*^9}},
 CellID->568608957],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", "p", "]"}]}]], 
  RowBox[{"p", "\[NotEqual]", 
   UnderscriptBox["0", "\[ScriptCapitalP]"]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 956509664}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656224217023238*^9, 3.656224232638866*^9}},
 CellTags->{
  "ID:956509664", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->956509664],

Cell[BoxData[
 RowBox[{"1", "\[Element]", 
  SubscriptBox["\[DoubleStruckCapitalN]", 
   RowBox[{"1", ",", "\[Ellipsis]", ",", 
    RowBox[{
    "\[LeftBracketingBar]", "p", 
     "\[RightBracketingBar]"}]}]]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples first index)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:52346492", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656225866428936*^9, 3.6562258764753537`*^9}},
 CellTags->{
  "ID:52346492", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first index", "IR"},
 CellID->52346492],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["isLP", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "t"}], "]"}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"t", "\[Equal]", 
      SubscriptBox[
       RowBox[{"(", 
        SubscriptBox["p", "1"], ")"}], "2"]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples isLP)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:3252425", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6562247353673487`*^9, 3.656224743307763*^9}, {
  3.656224788128642*^9, 3.65622479277745*^9}, {3.6562256554687653`*^9, 
  3.656225685514418*^9}},
 CellTags->{
  "ID:3252425", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isLP"},
 CellID->3252425],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], "\[Equal]", 
  SubscriptBox[
   RowBox[{"(", 
    SubscriptBox["p", "1"], ")"}], "2"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples lp)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:301012536", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6562242392688775`*^9, 3.656224252544501*^9}},
 CellTags->{
  "ID:301012536", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lp"},
 CellID->301012536],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], "\[Equal]", 
  SubscriptBox[
   RowBox[{"(", 
    SubscriptBox["p", "1"], ")"}], "1"]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples lc)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:499482967", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6562242670057263`*^9, 3.656224280998951*^9}},
 CellTags->{
  "ID:499482967", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lc"},
 CellID->499482967],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", "p", "]"}], "\[Equal]", 
  RowBox[{"\[LeftAngleBracket]", 
   SubscriptBox["p", "1"], 
   "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples lm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:77776384", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656240051862356*^9, 3.6562400684139853`*^9}},
 CellTags->{
  "ID:77776384", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lm"},
 CellID->77776384],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->677673986]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->96813974],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples first index)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples first index)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]]}], "\[Implies]", 
           RowBox[{"1", "\[Element]", 
             SubscriptBox["\[DoubleStruckCapitalN]", 
              RowBox[{"1", ",", "\[Ellipsis]", ",", 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["p", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}]}]]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]]}], "\[Implies]", 
           RowBox[{"1", "\[Element]", 
             SubscriptBox["\[DoubleStruckCapitalN]", 
              RowBox[{"1", ",", "\[Ellipsis]", ",", 
                RowBox[{"\[LeftBracketingBar]", 
                  StyleBox["p", "ExpressionVariable"], 
                  "\[RightBracketingBar]"}]}]]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:52346492"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p52346492-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p52346492-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:52346492", "Proof|ID:52346492-1"},
 CellID->156386449],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p52346492-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->286974475]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples isLP)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples isLP)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isLP", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[Equal]", 
               SubscriptBox[
                SubscriptBox[
                 StyleBox["p", "ExpressionVariable"], "1"], 
                "2"]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["isLP", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             "\[DoubleLeftRightArrow]", 
             RowBox[{
               StyleBox["t", "ExpressionVariable"], "\[Equal]", 
               SubscriptBox[
                SubscriptBox[
                 StyleBox["p", "ExpressionVariable"], "1"], 
                "2"]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:3252425"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p3252425-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p3252425-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:3252425", "Proof|ID:3252425-1"},
 CellID->201371222],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p3252425-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->543101995]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples lp)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples lp)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], "1"], 
              "2"]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], "1"], 
              "2"]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:301012536"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p301012536-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p301012536-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:301012536", "Proof|ID:301012536-1"},
 CellID->132883546],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p301012536-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->95709161]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples lc)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples lc)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lc", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], "1"], 
              "1"]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lc", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
             SubscriptBox[
              SubscriptBox[
               StyleBox["p", "ExpressionVariable"], "1"], 
              "1"]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:499482967"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p499482967-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p499482967-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:499482967", "Proof|ID:499482967-1"},
 CellID->358731136],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p499482967-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->606474251]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples lm)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples lm)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lm", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{"\[LeftAngleBracket]", 
               SubscriptBox[
                StyleBox["p", "ExpressionVariable"], "1"], 
               "\[RightAngleBracket]"}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lm", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{"\[LeftAngleBracket]", 
               SubscriptBox[
                StyleBox["p", "ExpressionVariable"], "1"], 
               "\[RightAngleBracket]"}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:77776384"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p77776384-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p77776384-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:77776384", "Proof|ID:77776384-1"},
 CellID->379573710],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p77776384-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->99872566]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->95739086],

Cell[CellGroupData[{

Cell["Lemma: lp, lc, lm and R in PolyTuples II", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6562406760608587`*^9, 3.6562406911460853`*^9}},
 CellID->41342827],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"c", ",", "t", ",", 
   RowBox[{"r", "..."}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 1015977913}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6562406967776947`*^9, 3.6562407056229105`*^9}},
 CellTags->{
  "ID:1015977913", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->1015977913],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], "]"}]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 731677923}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656240709257717*^9, 3.6562407259809465`*^9}},
 CellTags->{
  "ID:731677923", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->731677923],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
  "t"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples lp seq)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:23863949", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6562407321897573`*^9, 3.6562407449817796`*^9}},
 CellTags->{
  "ID:23863949", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lp seq"},
 CellID->23863949],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
  "c"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples lc seq)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:721799663", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6562407575086017`*^9, 3.6562407630154114`*^9}},
 CellTags->{
  "ID:721799663", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lc seq"},
 CellID->721799663],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lm", "\[ScriptCapitalP]"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], 
   "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples lm seq)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2939774", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6562407584290032`*^9, 3.6562407719542274`*^9}},
 CellTags->{
  "ID:2939774", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lm seq"},
 CellID->2939774],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["R", "\[ScriptCapitalP]"], "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"c", ",", "t"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"r", "..."}]}], "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
  RowBox[{"\[LeftAngleBracket]", 
   RowBox[{"r", "..."}], "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples R seq)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:172576768", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6562407591934047`*^9, 3.65624077931744*^9}},
 CellTags->{
  "ID:172576768", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples R seq"},
 CellID->172576768],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
      RowBox[{"[", "d", "]"}]}], ",", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "s", "]"}]}]}]], 
   RowBox[{"d", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalR]"]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"\[And]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         RowBox[{
          UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", 
          RowBox[{"d", 
           UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "s"}], "]"}], 
         "\[Equal]", "d"}]},
       {
        RowBox[{
         RowBox[{
          UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", 
          RowBox[{"d", 
           UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], "s"}], "]"}], 
         "\[Equal]", "s"}]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}}]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples \[CenterDot] lp lc)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:3206806", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656399034660826*^9, 3.6563991540010357`*^9}},
 CellTags->{
  "ID:3206806", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CenterDot] lp lc"},
 CellID->3206806],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->331618674]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->834387043],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples lp seq)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples lp seq)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lp", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
             StyleBox[
             "t", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:23863949"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p23863949-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p23863949-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:23863949", "Proof|ID:23863949-1"},
 CellID->2176190],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p23863949-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->315083762]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples lc seq)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples lc seq)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lc", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lc", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
             StyleBox[
             "c", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:721799663"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p721799663-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p721799663-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:721799663", "Proof|ID:721799663-1"},
 CellID->307273605],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p721799663-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->320776169]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples lm seq)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples lm seq)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lm", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], 
               "\[RightAngleBracket]"}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lm", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   StyleBox["c", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], 
               "\[RightAngleBracket]"}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2939774"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2939774-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2939774-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2939774", "Proof|ID:2939774-1"},
 CellID->75418897],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2939774-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->142228474]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples R seq)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples R seq)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["R", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
             RowBox[{"\[LeftAngleBracket]", 
               StyleBox[
                RowBox[{"r", "..."}], "ExpressionVariable"], 
               "\[RightAngleBracket]"}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c", "ExpressionVariable"], ",", 
            StyleBox["t", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["R", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c", "ExpressionVariable"], ",", 
                    StyleBox["t", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[Equal]", 
             RowBox[{"\[LeftAngleBracket]", 
               StyleBox[
                RowBox[{"r", "..."}], "ExpressionVariable"], 
               "\[RightAngleBracket]"}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:172576768"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p172576768-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p172576768-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:172576768", "Proof|ID:172576768-1"},
 CellID->126187940],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p172576768-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->454221170]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples \[CenterDot] lp lc)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples \[CenterDot] lp lc)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["d", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["d", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               StyleBox["d", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["d", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               StyleBox[
               "s", "ExpressionVariable"]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
               StyleBox["d", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["d", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 UnderscriptBox["lc", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["d", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               StyleBox["d", "ExpressionVariable"]}], "\[And]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["lp", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["d", "ExpressionVariable"], 
                   UnderscriptBox["\[CenterDot]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               StyleBox[
               "s", "ExpressionVariable"]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:3206806"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p3206806-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p3206806-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:3206806", "Proof|ID:3206806-1"},
 CellID->37996798],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p3206806-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->471222850]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Reduction Polynomials (\[Checkmark])", "Subsection",
 CellChangeTimes->{{3.6555261920202675`*^9, 3.655526195561474*^9}, {
  3.6563079290596867`*^9, 3.656307933833295*^9}},
 CellID->335068244],

Cell[CellGroupData[{

Cell["isDomain (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.655526220756518*^9, 3.655526235124143*^9}, {
  3.6555383870591288`*^9, 3.655538391286736*^9}, {3.6555488948300714`*^9, 
  3.655548894939272*^9}},
 CellID->420471911],

Cell["", "OpenEnvironment",
 CellID->657209226],

Cell[CellGroupData[{

Cell["Lemma: PolyTuples form Domain", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.655526250848971*^9, 3.655526255653779*^9}},
 CellID->711892999],

Cell[BoxData[
 RowBox[{"isDomain", "[", "\[ScriptCapitalR]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalR] isDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:31351138", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.6555376275773897`*^9, {3.655537675813675*^9, 3.655537677092877*^9}},
 CellTags->{
  "ID:31351138", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalR] isDomain"},
 CellID->31351138],

Cell[BoxData[
 RowBox[{"isDomain", "[", "\[ScriptCapitalT]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(\[ScriptCapitalT] isDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:632462358", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{
  3.655537628731792*^9, {3.6555376791832805`*^9, 3.6555376805560827`*^9}},
 CellTags->{
  "ID:632462358", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalT] isDomain"},
 CellID->632462358],

Cell[BoxData[
 RowBox[{"isDomain", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples isDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:404915746", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.655526258976585*^9, 3.6555262808946238`*^9}, 
   3.65552691005853*^9, {3.6555373423052883`*^9, 3.6555373436780906`*^9}},
 CellTags->{
  "ID:404915746", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isDomain"},
 CellID->404915746],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->654702049]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->432439441],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[ScriptCapitalR] isDomain)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[ScriptCapitalR] isDomain)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "isDomain", "[", "\[ScriptCapitalR]", 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "isDomain", "[", "\[ScriptCapitalR]", 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:31351138"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p31351138-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p31351138-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:31351138", "Proof|ID:31351138-1"},
 CellID->124332373],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p31351138-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->498504441]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(\[ScriptCapitalT] isDomain)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(\[ScriptCapitalT] isDomain)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
        "isDomain", "[", "\[ScriptCapitalT]", 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
        "isDomain", "[", "\[ScriptCapitalT]", 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:632462358"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p632462358-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p632462358-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:632462358", "Proof|ID:632462358-1"},
 CellID->403093521],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p632462358-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->85073108]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples isDomain)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples isDomain)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isDomain", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isDomain", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:404915746"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p404915746-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p404915746-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:404915746", "Proof|ID:404915746-1"},
 CellID->277496693],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p404915746-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->80377566]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Decision Predicate (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6557009567086654`*^9, 3.655700959984671*^9}, {
  3.6562238958322744`*^9, 3.656223900356282*^9}},
 CellID->110666071],

Cell["", "OpenEnvironment",
 CellID->498410766],

Cell[CellGroupData[{

Cell["Lemma: Decision Predicate", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6557015078618336`*^9, 3.6557015303258734`*^9}},
 CellID->351279416],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"p", ",", "q"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 3279840}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.655701535302282*^9, 3.655701545021099*^9}},
 CellTags->{
  "ID:3279840", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->3279840],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
    RowBox[{"[", "t", "]"}]}]], 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
   RowBox[{"[", 
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "t"}], "]"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples membership C)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:141435198", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.655701550184708*^9, 3.655701574801551*^9}},
 CellTags->{
  "ID:141435198", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership C"},
 CellID->141435198],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
  RowBox[{"[", 
   UnderscriptBox["0", "\[ScriptCapitalP]"], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples membership 0)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:45852914", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6557017316910267`*^9, 3.655701742096245*^9}},
 CellTags->{
  "ID:45852914", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership 0"},
 CellID->45852914],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
  RowBox[{"[", 
   UnderscriptBox["1", "\[ScriptCapitalP]"], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples membership 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:320592129", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.655701748211456*^9, 3.6557017497870584`*^9}},
 CellTags->{
  "ID:320592129", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership 1"},
 CellID->320592129],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
  RowBox[{"[", 
   RowBox[{"p", 
    UnderscriptBox["+", "\[ScriptCapitalP]"], "q"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples membership +)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:93322392", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6557016137860193`*^9, 3.6557016266872425`*^9}},
 CellTags->{
  "ID:93322392", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership +"},
 CellID->93322392],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
  RowBox[{"[", 
   RowBox[{
    UnderscriptBox["-", "\[ScriptCapitalP]"], "p"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples membership -)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:6751189", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.655701700522172*^9, 3.6557017121285925`*^9}},
 CellTags->{
  "ID:6751189", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership -"},
 CellID->6751189],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["supp", "\[ScriptCapitalP]"], "[", "p", "]"}], "\[Equal]", 
  RowBox[{"TupleToSet", "[", 
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     SubscriptBox[
      RowBox[{"(", 
       SubscriptBox["p", "i"], ")"}], "2"], 
     UnderscriptBox["|", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
       RowBox[{"\[LeftBracketingBar]", "p", "\[RightBracketingBar]"}]}]]}], 
    "\[RightAngleBracket]"}], "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples supp)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:179493910", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6557015799963603`*^9, 3.655701592429582*^9}, {
  3.6561313068993*^9, 3.656131337459754*^9}},
 CellTags->{
  "ID:179493910", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples supp"},
 CellID->179493910],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->154048488]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->51765692],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples membership C)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples membership C)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
           RowBox[{
             UnderscriptBox["C", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
           RowBox[{
             UnderscriptBox["C", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{
               StyleBox["p", "ExpressionVariable"], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:141435198"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p141435198-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p141435198-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:141435198", "Proof|ID:141435198-1"},
 CellID->225697908],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p141435198-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->89331975]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples membership 0)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples membership 0)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", 
          RowBox[{"PolyTuples", "[", 
            RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}]], 
         "[", 
         UnderscriptBox["0", 
          RowBox[{"PolyTuples", "[", 
            RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}]], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", 
          RowBox[{"PolyTuples", "[", 
            RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}]], 
         "[", 
         UnderscriptBox["0", 
          RowBox[{"PolyTuples", "[", 
            RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}]], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:45852914"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p45852914-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p45852914-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:45852914", "Proof|ID:45852914-1"},
 CellID->91077423],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p45852914-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->169634120]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples membership 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples membership 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", 
          RowBox[{"PolyTuples", "[", 
            RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}]], 
         "[", 
         UnderscriptBox["1", 
          RowBox[{"PolyTuples", "[", 
            RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}]], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[Element]", 
          RowBox[{"PolyTuples", "[", 
            RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}]], 
         "[", 
         UnderscriptBox["1", 
          RowBox[{"PolyTuples", "[", 
            RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}]], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:320592129"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p320592129-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p320592129-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:320592129", "Proof|ID:320592129-1"},
 CellID->94174531],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p320592129-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->339585414]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples membership +)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples membership +)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"PolyTuples", "[", 
              RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}]],
            "[", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["+", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], 
             StyleBox["q", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"PolyTuples", "[", 
              RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}]],
            "[", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], 
             UnderscriptBox["+", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], 
             StyleBox["q", "ExpressionVariable"]}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:93322392"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p93322392-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p93322392-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:93322392", "Proof|ID:93322392-1"},
 CellID->411309806],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p93322392-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->207218404]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples membership -)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples membership -)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"PolyTuples", "[", 
              RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}]],
            "[", 
           RowBox[{
             UnderscriptBox["Minus", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           UnderscriptBox["\[Element]", 
            RowBox[{"PolyTuples", "[", 
              RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}]],
            "[", 
           RowBox[{
             UnderscriptBox["Minus", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], 
           "]"}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:6751189"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p6751189-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p6751189-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:6751189", "Proof|ID:6751189-1"},
 CellID->16346637],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p6751189-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->78211768]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples supp)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples supp)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["supp", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
           RowBox[{"TupleToSet", "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 SubscriptBox[
                  SubscriptBox[
                   StyleBox["p", "ExpressionVariable"], 
                   StyleBox["i", "ExpressionVariable"]], "2"], 
                 UnderscriptBox["|", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["p", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
               "\[RightAngleBracket]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["supp", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
           RowBox[{"TupleToSet", "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 SubscriptBox[
                  SubscriptBox[
                   StyleBox["p", "ExpressionVariable"], 
                   StyleBox["i", "ExpressionVariable"]], "2"], 
                 UnderscriptBox["|", 
                  GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["p", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
               "\[RightAngleBracket]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:179493910"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p179493910-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p179493910-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:179493910", "Proof|ID:179493910-1"},
 CellID->162772222],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p179493910-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->62815219]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->212237021],

Cell[CellGroupData[{

Cell["Theorem: PolyTuples has Decision Predicate", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6557009899055233`*^9, 3.6557009967227354`*^9}},
 CellID->439058519],

Cell[BoxData[
 RowBox[{"hasDecisionPredicate", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples hasDecisionPredicate)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:644982519", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6557010021515455`*^9, 3.655701009389958*^9}},
 CellTags->{
  "ID:644982519", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasDecisionPredicate"},
 CellID->644982519],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->686721649]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->6665501],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples hasDecisionPredicate)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples hasDecisionPredicate)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"hasDecisionPredicate", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"hasDecisionPredicate", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:644982519"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p644982519-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p644982519-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:644982519", "Proof|ID:644982519-1"},
 CellID->323902770],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p644982519-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->440058230]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Monomials (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6561399805181417`*^9, 3.6561399816881437`*^9}, {
  3.6561412015276866`*^9, 3.6561412058644943`*^9}},
 CellID->704433491],

Cell["", "OpenEnvironment",
 CellID->360614755],

Cell[CellGroupData[{

Cell["Theorem: PolyTuples contains Monomials", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6561399943709655`*^9, 3.656140000673377*^9}},
 CellID->332255835],

Cell[BoxData[
 RowBox[{"containsMonomials", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples containsMonomials)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:114843091", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656140003169381*^9, 3.656140011078595*^9}},
 CellTags->{
  "ID:114843091", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples containsMonomials"},
 CellID->114843091],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->80976911]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->35940],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples containsMonomials)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples containsMonomials)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"containsMonomials", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"containsMonomials", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:114843091"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p114843091-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p114843091-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:114843091", "Proof|ID:114843091-1"},
 CellID->255004292],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p114843091-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellChangeTimes->{3.6561417822705226`*^9},
 CellID->81650813]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Coefficient Function (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.656043621808221*^9, 3.656043625848628*^9}, {
  3.6561399150604267`*^9, 3.656139919490834*^9}},
 CellID->42388596],

Cell["", "OpenEnvironment",
 CellID->154907406],

Cell[CellGroupData[{

Cell["Lemma: Coefficient Function and Addition", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656134891481798*^9, 3.6561348980182095`*^9}},
 CellID->358230070],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"s", ",", "c1", ",", "c2", ",", "t1", ",", "t2", ",", 
   RowBox[{"r1", "..."}], ",", 
   RowBox[{"r2", "..."}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 78312379}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.65613492690946*^9, 3.656134941994687*^9}},
 CellTags->{
  "ID:78312379", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->78312379],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
        RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], "]"}]}], "\[And]", 
    
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"c2", ",", "t2"}], "\[RightAngleBracket]"}], ",", 
        RowBox[{"r2", "..."}]}], "\[RightAngleBracket]"}], "]"}]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 210280178}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{3.65613493243187*^9},
 CellTags->{
  "ID:210280178", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->210280178],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"t1", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t2"}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"r1", "..."}], "\[RightAngleBracket]"}], 
         UnderscriptBox["+", "\[ScriptCapitalP]"], 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"c2", ",", "t2"}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"r2", "..."}]}], "\[RightAngleBracket]"}]}], ",", "s"}], 
       "]"}], "\[Equal]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"r1", "..."}], "\[RightAngleBracket]"}], ",", "s"}], "]"}], 
       
       UnderscriptBox["+", "\[ScriptCapitalR]"], 
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"c2", ",", "t2"}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"r2", "..."}]}], "\[RightAngleBracket]"}], ",", "s"}], 
        "]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
    RowBox[{
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], 
      TagBox["\:293b",
       Identity,
       SyntaxForm->"a*b"], 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"r1", "..."}], "\[RightAngleBracket]"}], 
        UnderscriptBox["+", "\[ScriptCapitalP]"], 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"c2", ",", "t2"}], "\[RightAngleBracket]"}], ",", 
          RowBox[{"r2", "..."}]}], "\[RightAngleBracket]"}]}], ")"}]}], ",", 
     "s"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
        RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], ",", "s"}], "]"}], 
    
    UnderscriptBox["+", "\[ScriptCapitalR]"], 
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"c2", ",", "t2"}], "\[RightAngleBracket]"}], ",", 
        RowBox[{"r2", "..."}]}], "\[RightAngleBracket]"}], ",", "s"}], 
     "]"}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples + C 1)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:261724624", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6561339558213544`*^9, 3.656133965634772*^9}, {
  3.656134012512854*^9, 3.656134106409419*^9}, {3.656134143771485*^9, 
  3.6561342005087843`*^9}, {3.6561346331297445`*^9, 3.656134634923747*^9}, {
  3.656134948827499*^9, 3.656134956955113*^9}, {3.656135284369488*^9, 
  3.6561352863662915`*^9}},
 CellTags->{
  "ID:261724624", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 1"},
 CellID->261724624],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"t2", 
     UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t1"}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], 
         UnderscriptBox["+", "\[ScriptCapitalP]"], 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"r2", "..."}], "\[RightAngleBracket]"}]}], ",", "s"}], 
       "]"}], "\[Equal]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
           RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], ",", "s"}], 
        "]"}], 
       UnderscriptBox["+", "\[ScriptCapitalR]"], 
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"r2", "..."}], "\[RightAngleBracket]"}], ",", "s"}], 
        "]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
    RowBox[{
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"c2", ",", "t2"}], "\[RightAngleBracket]"}], 
      TagBox["\:293b",
       Identity,
       SyntaxForm->"a*b"], 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{
          RowBox[{"\[LeftAngleBracket]", 
           RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
          RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], 
        UnderscriptBox["+", "\[ScriptCapitalP]"], 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"r2", "..."}], "\[RightAngleBracket]"}]}], ")"}]}], ",", 
     "s"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
        RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], ",", "s"}], "]"}], 
    
    UnderscriptBox["+", "\[ScriptCapitalR]"], 
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"c2", ",", "t2"}], "\[RightAngleBracket]"}], ",", 
        RowBox[{"r2", "..."}]}], "\[RightAngleBracket]"}], ",", "s"}], 
     "]"}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples + C 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:115965410", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6561356302864957`*^9, 3.6561356700509653`*^9}},
 CellTags->{
  "ID:115965410", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 2"},
 CellID->115965410],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{"t1", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t2"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{"t2", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t1"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     RowBox[{"c1", 
      UnderscriptBox["+", "\[ScriptCapitalR]"], "c2"}], "\[NotEqual]", 
     UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"r1", "..."}], "\[RightAngleBracket]"}], 
         UnderscriptBox["+", "\[ScriptCapitalP]"], 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"r2", "..."}], "\[RightAngleBracket]"}]}], ",", "s"}], 
       "]"}], "\[Equal]", 
      RowBox[{
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"r1", "..."}], "\[RightAngleBracket]"}], ",", "s"}], "]"}], 
       
       UnderscriptBox["+", "\[ScriptCapitalR]"], 
       RowBox[{
        UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"r2", "..."}], "\[RightAngleBracket]"}], ",", "s"}], 
        "]"}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
    RowBox[{
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"c1", 
         UnderscriptBox["+", "\[ScriptCapitalR]"], "c2"}], ",", "t1"}], 
       "\[RightAngleBracket]"}], 
      TagBox["\:293b",
       Identity,
       SyntaxForm->"a*b"], 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"r1", "..."}], "\[RightAngleBracket]"}], 
        UnderscriptBox["+", "\[ScriptCapitalP]"], 
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"r2", "..."}], "\[RightAngleBracket]"}]}], ")"}]}], ",", 
     "s"}], "]"}], "\[Equal]", 
   RowBox[{
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
        RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], ",", "s"}], "]"}], 
    
    UnderscriptBox["+", "\[ScriptCapitalR]"], 
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"c2", ",", "t2"}], "\[RightAngleBracket]"}], ",", 
        RowBox[{"r2", "..."}]}], "\[RightAngleBracket]"}], ",", "s"}], 
     "]"}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples + C 3)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:384214099", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6561363670309896`*^9, 3.656136444609926*^9}, {
  3.656136483141994*^9, 3.656136512906846*^9}},
 CellTags->{
  "ID:384214099", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 3"},
 CellID->384214099],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{"t1", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t2"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[Not]", 
      RowBox[{"t2", 
       UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], "t1"}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[And]", 
    RowBox[{
     RowBox[{"c1", 
      UnderscriptBox["+", "\[ScriptCapitalR]"], "c2"}], "\[Equal]", 
     UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   RowBox[{
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"r1", "..."}], "\[RightAngleBracket]"}], ",", "s"}], "]"}], 
    UnderscriptBox["+", "\[ScriptCapitalR]"], 
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"r2", "..."}], "\[RightAngleBracket]"}], ",", "s"}], "]"}]}], 
   "\[Equal]", 
   RowBox[{
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
        RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], ",", "s"}], "]"}], 
    
    UnderscriptBox["+", "\[ScriptCapitalR]"], 
    RowBox[{
     UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"c2", ",", "t2"}], "\[RightAngleBracket]"}], ",", 
        RowBox[{"r2", "..."}]}], "\[RightAngleBracket]"}], ",", "s"}], "]"}]}]
    }]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples + C 4)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:56134460", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6561360680554647`*^9, 3.656136112141142*^9}, {
  3.6561362518403873`*^9, 3.656136307236085*^9}},
 CellTags->{
  "ID:56134460", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 4"},
 CellID->56134460],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->106716937]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->658773040],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples + C 1)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples + C 1)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["s", "ExpressionVariable"], ",", 
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["c2", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox["t2", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r2", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["t1", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 StyleBox["t2", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                    UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["+", "\[ScriptCapitalR]"], 
                   RowBox[{
                    UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}]}]}]}], 
             "\[Implies]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\:293b", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                    UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}]}], ",", 
                   StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["+", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["s", "ExpressionVariable"], ",", 
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["c2", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox["t2", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r2", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["t1", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 StyleBox["t2", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                    UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["+", "\[ScriptCapitalR]"], 
                   RowBox[{
                    UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}]}]}]}], 
             "\[Implies]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\:293b", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                    UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}]}], ",", 
                   StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["+", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:261724624"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p261724624-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p261724624-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:261724624", "Proof|ID:261724624-1"},
 CellID->305685053],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p261724624-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->219011818]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples + C 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples + C 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["s", "ExpressionVariable"], ",", 
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["c2", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox["t2", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r2", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["t2", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 StyleBox["t1", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], 
                    UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["+", "\[ScriptCapitalR]"], 
                   RowBox[{
                    UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}]}]}]}], 
             "\[Implies]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\:293b", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], 
                    UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}]}], ",", 
                   StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["+", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["s", "ExpressionVariable"], ",", 
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["c2", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox["t2", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r2", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{
                 StyleBox["t2", "ExpressionVariable"], 
                 UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                 StyleBox["t1", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], 
                    UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["+", "\[ScriptCapitalR]"], 
                   RowBox[{
                    UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}]}]}]}], 
             "\[Implies]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\:293b", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], 
                    UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}]}], ",", 
                   StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["+", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:115965410"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p115965410-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p115965410-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:115965410", "Proof|ID:115965410-1"},
 CellID->175933698],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p115965410-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{3.656137099001876*^9},
 CellID->283308081]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples + C 3)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples + C 3)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["s", "ExpressionVariable"], ",", 
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["c2", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox["t2", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r2", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    StyleBox["t1", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t2", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    StyleBox["t2", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t1", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], 
                   UnderscriptBox["+", "\[ScriptCapitalR]"], 
                   StyleBox["c2", "ExpressionVariable"]}], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                    UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["+", "\[ScriptCapitalR]"], 
                   RowBox[{
                    UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}]}]}]}], 
             "\[Implies]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], 
                    UnderscriptBox["+", "\[ScriptCapitalR]"], 
                    StyleBox["c2", "ExpressionVariable"]}], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\:293b", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                    UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}]}], ",", 
                   StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["+", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["s", "ExpressionVariable"], ",", 
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["c2", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox["t2", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r2", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    StyleBox["t1", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t2", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    StyleBox["t2", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t1", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], 
                   UnderscriptBox["+", "\[ScriptCapitalR]"], 
                   StyleBox["c2", "ExpressionVariable"]}], "\[NotEqual]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                    UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
                 
                 RowBox[{
                   RowBox[{
                    UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                   UnderscriptBox["+", "\[ScriptCapitalR]"], 
                   RowBox[{
                    UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}]}]}]}], 
             "\[Implies]", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], 
                    UnderscriptBox["+", "\[ScriptCapitalR]"], 
                    StyleBox["c2", "ExpressionVariable"]}], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\:293b", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                    UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}]}], ",", 
                   StyleBox["s", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["+", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:384214099"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p384214099-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p384214099-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:384214099", "Proof|ID:384214099-1"},
 CellID->84967562],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p384214099-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->467820545]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples + C 4)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples + C 4)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["s", "ExpressionVariable"], ",", 
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["c2", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox["t2", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r2", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    StyleBox["t1", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t2", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    StyleBox["t2", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t1", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], 
                   UnderscriptBox["+", "\[ScriptCapitalR]"], 
                   StyleBox["c2", "ExpressionVariable"]}], "\[Equal]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], "\[Implies]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["+", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}]}], 
               "\[Equal]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["+", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["s", "ExpressionVariable"], ",", 
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["c2", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox["t2", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r2", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}], "\[And]", 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], "]"}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    StyleBox["t1", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t2", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{"\[Not]", 
                 RowBox[{
                   TagBox["(", "AutoParentheses"], 
                   RowBox[{
                    StyleBox["t2", "ExpressionVariable"], 
                    UnderscriptBox["\[Succeeds]", "\[ScriptCapitalT]"], 
                    StyleBox["t1", "ExpressionVariable"]}], 
                   TagBox[")", "AutoParentheses"]}]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   StyleBox["c1", "ExpressionVariable"], 
                   UnderscriptBox["+", "\[ScriptCapitalR]"], 
                   StyleBox["c2", "ExpressionVariable"]}], "\[Equal]", 
                 UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], "\[Implies]", 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["+", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}]}], 
               "\[Equal]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], "]"}], 
                 UnderscriptBox["+", "\[ScriptCapitalR]"], 
                 RowBox[{
                   UnderscriptBox["C", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox["s", "ExpressionVariable"]}], 
                   "]"}]}]}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:56134460"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p56134460-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p56134460-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:56134460", "Proof|ID:56134460-1"},
 CellID->170413464],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p56134460-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->121669868]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->67737875],

Cell[CellGroupData[{

Cell["Lemma: Coefficient Function", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656043635505045*^9, 3.656043640793454*^9}},
 CellID->299731454],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
   RowBox[{"[", "t", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 235178571}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6560436806354065`*^9, 3.6560436885758204`*^9}},
 CellTags->{
  "ID:235178571", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->235178571],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    UnderscriptBox["1", "\[ScriptCapitalP]"], ",", "t"}], "]"}], "\[Equal]", 
  RowBox[{"\[Piecewise]", GridBox[{
     {
      UnderscriptBox["1", "\[ScriptCapitalR]"], "\[DoubleLeftArrow]", 
      RowBox[{"t", "\[Equal]", 
       UnderscriptBox["1", "\[ScriptCapitalT]"]}]},
     {
      UnderscriptBox["0", "\[ScriptCapitalR]"], "\[DoubleLeftArrow]", 
      "otherwise"}
    }]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples C 2)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:474434374", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656044108266358*^9, 3.65604414381882*^9}},
 CellTags->{
  "ID:474434374", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C 2"},
 CellID->474434374],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"p", ",", "q"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 108633346}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6560436523998747`*^9, 3.6560436621186914`*^9}},
 CellTags->{
  "ID:108633346", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->108633346],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["+", "\[ScriptCapitalP]"], "q"}], ",", "t"}], "]"}], 
  "\[Equal]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
    RowBox[{"p", ",", "t"}], "]"}], 
   UnderscriptBox["+", "\[ScriptCapitalR]"], 
   RowBox[{
    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
    RowBox[{"q", ",", "t"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples C +)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:611741232", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6560437369203053`*^9, 3.656043774360371*^9}, {
  3.6560438132980394`*^9, 3.656043814655242*^9}},
 CellTags->{
  "ID:611741232", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C +"},
 CellID->611741232],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{
     UnderscriptBox["-", "\[ScriptCapitalP]"], "p"}], ",", "t"}], "]"}], 
  "\[Equal]", 
  RowBox[{
   UnderscriptBox["-", "\[ScriptCapitalR]"], 
   RowBox[{
    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
    RowBox[{"p", ",", "t"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples C - unary)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:438277340", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656043801613619*^9, 3.656043831160071*^9}},
 CellTags->{
  "ID:438277340", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C - unary"},
 CellID->438277340],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
   RowBox[{
    RowBox[{"p", 
     UnderscriptBox["-", "\[ScriptCapitalP]"], "q"}], ",", "t"}], "]"}], 
  "\[Equal]", 
  RowBox[{
   RowBox[{
    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
    RowBox[{"p", ",", "t"}], "]"}], 
   UnderscriptBox["-", "\[ScriptCapitalR]"], 
   RowBox[{
    UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
    RowBox[{"q", ",", "t"}], "]"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples C -)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:412638768", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656043852251308*^9, 3.6560438696765385`*^9}},
 CellTags->{
  "ID:412638768", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C -"},
 CellID->412638768],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->166044720]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->203713653],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples C 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples C 2)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{
               UnderscriptBox["1", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{"\[Piecewise]", 
             GridBox[{{
                UnderscriptBox["1", "\[ScriptCapitalR]"], 
                "\[DoubleLeftArrow]", 
                RowBox[{
                  StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                  UnderscriptBox["1", "\[ScriptCapitalT]"]}]}, {
                UnderscriptBox["0", "\[ScriptCapitalR]"], 
                "\[DoubleLeftArrow]", 
                "True"}}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{
               UnderscriptBox["1", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{"\[Piecewise]", 
             GridBox[{{
                UnderscriptBox["1", "\[ScriptCapitalR]"], 
                "\[DoubleLeftArrow]", 
                RowBox[{
                  StyleBox["t", "ExpressionVariable"], "\[Equal]", 
                  UnderscriptBox["1", "\[ScriptCapitalT]"]}]}, {
                UnderscriptBox["0", "\[ScriptCapitalR]"], 
                "\[DoubleLeftArrow]", 
                "True"}}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:474434374"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p474434374-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p474434374-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:474434374", "Proof|ID:474434374-1"},
 CellID->85632222],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p474434374-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->518387433]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples C +)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples C +)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                 StyleBox["q", "ExpressionVariable"]}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["+", "\[ScriptCapitalR]"], 
             RowBox[{
               UnderscriptBox["C", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["+", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                 StyleBox["q", "ExpressionVariable"]}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["+", "\[ScriptCapitalR]"], 
             RowBox[{
               UnderscriptBox["C", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:611741232"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p611741232-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p611741232-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:611741232", "Proof|ID:611741232-1"},
 CellID->352968606],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p611741232-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->576517268]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples C - unary)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples C - unary)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["Minus", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["Minus", "\[ScriptCapitalR]"], "[", 
             RowBox[{
               UnderscriptBox["C", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{
               RowBox[{
                 UnderscriptBox["Minus", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             UnderscriptBox["Minus", "\[ScriptCapitalR]"], "[", 
             RowBox[{
               UnderscriptBox["C", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:438277340"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p438277340-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p438277340-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:438277340", "Proof|ID:438277340-1"},
 CellID->232831160],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p438277340-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->159284763]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples C -)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples C -)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                 StyleBox["q", "ExpressionVariable"]}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["-", "\[ScriptCapitalR]"], 
             RowBox[{
               UnderscriptBox["C", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
               StyleBox["t", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["C", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], 
                 UnderscriptBox["-", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                 StyleBox["q", "ExpressionVariable"]}], ",", 
               StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["C", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], "]"}], 
             UnderscriptBox["-", "\[ScriptCapitalR]"], 
             RowBox[{
               UnderscriptBox["C", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["t", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:412638768"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p412638768-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p412638768-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:412638768", "Proof|ID:412638768-1"},
 CellID->382215156],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p412638768-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->149522829]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->107107139],

Cell[CellGroupData[{

Cell["Theorem: PolyTuples has Coefficient Function", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6561396437271495`*^9, 3.656139671323598*^9}},
 CellID->105654882],

Cell[BoxData[
 RowBox[{"hasCoefficientFunction", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples hasCoefficientFunction)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:392022323", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6561396741004033`*^9, 3.6561396817132163`*^9}},
 CellTags->{
  "ID:392022323", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasCoefficientFunction"},
 CellID->392022323],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->154812201]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->77149556],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples hasCoefficientFunction)\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples hasCoefficientFunction)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"hasCoefficientFunction", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"hasCoefficientFunction", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:392022323"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p392022323-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p392022323-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:392022323", "Proof|ID:392022323-1"},
 CellID->43251408],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p392022323-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->19455043]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Multiplication (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6561409296494093`*^9, 3.6561409321454134`*^9}, {
  3.6562491856396065`*^9, 3.656249189820414*^9}},
 CellID->196031232],

Cell["", "OpenEnvironment",
 CellID->557047419],

Cell[CellGroupData[{

Cell["Lemma: Monomial Multiplication", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6562417979710293`*^9, 3.6562418047882414`*^9}},
 CellID->240207582],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{"c1", ",", "t1", ",", 
   RowBox[{"r1", "..."}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 23554005}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6562418999328084`*^9, 3.6562419255792537`*^9}, {
  3.656242069801507*^9, 3.6562420725471115`*^9}},
 CellTags->{
  "ID:23554005", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->23554005],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], "]"}]}], 
  "\[Implies]"}]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 392464903}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6562419322872653`*^9, 3.656241949369295*^9}, {
  3.6562420737951136`*^9, 3.656242075573517*^9}},
 CellTags->{
  "ID:392464903", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->392464903],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], 
   UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
   RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], "\[Equal]", 
  RowBox[{
  "\[LeftAngleBracket]", "\[RightAngleBracket]"}]}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples \[CircleTimes] empty)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:161672314", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656241956545308*^9, 3.6562419893833656`*^9}, {
  3.656242076634319*^9, 3.656242077991521*^9}},
 CellTags->{
  "ID:161672314", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] empty"},
 CellID->161672314],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{"c2", ",", "t2", ",", 
    RowBox[{"r2", "..."}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
     RowBox[{"[", 
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{
        RowBox[{"\[LeftAngleBracket]", 
         RowBox[{"c2", ",", "t2"}], "\[RightAngleBracket]"}], ",", 
        RowBox[{"r2", "..."}]}], "\[RightAngleBracket]"}], "]"}]}], 
    "\[Implies]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
         RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], 
       UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
       RowBox[{"\[LeftAngleBracket]", 
        RowBox[{
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{"c2", ",", "t2"}], "\[RightAngleBracket]"}], ",", 
         RowBox[{"r2", "..."}]}], "\[RightAngleBracket]"}]}], "\[Equal]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
              RowBox[{"c1", 
               UnderscriptBox["*", "\[ScriptCapitalR]"], "c2"}], ",", 
              RowBox[{"t1", 
               UnderscriptBox["*", "\[ScriptCapitalT]"], "t2"}]}], 
             "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], 
           UnderscriptBox["+", "\[ScriptCapitalP]"], 
           RowBox[{
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{
              RowBox[{"\[LeftAngleBracket]", 
               RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
              RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], 
            UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"r2", "..."}], "\[RightAngleBracket]"}]}]}], 
          "\[DoubleLeftArrow]", 
          RowBox[{
           RowBox[{"c1", 
            UnderscriptBox["*", "\[ScriptCapitalR]"], "c2"}], "\[NotEqual]", 
           UnderscriptBox["0", "\[ScriptCapitalR]"]}]},
         {
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
             RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], 
           UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"r2", "..."}], "\[RightAngleBracket]"}]}], 
          "\[DoubleLeftArrow]", "otherwise"}
        }]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples \[CircleTimes] prepend)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:71703914", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656242052735077*^9, 3.6562421910137196`*^9}, {
  3.6562422365345993`*^9, 3.6562422578598375`*^9}, {3.6562425499235497`*^9, 
  3.6562425701879854`*^9}, {3.656243185890867*^9, 3.6562432138843164`*^9}, {
  3.6562450349721155`*^9, 3.656245042460129*^9}},
 CellTags->{
  "ID:71703914", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] prepend"},
 CellID->71703914],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", "q", "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 127729286}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6562457384895515`*^9, 3.6562457457435637`*^9}},
 CellTags->{
  "ID:127729286", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->127729286],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], 
   UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], "\[Equal]", 
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], 
    "\[RightAngleBracket]"}], 
   UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], 
   "q"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples \[CircleTimes] lm)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:2718947", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656244173988003*^9, 3.65624420064845*^9}, {
  3.656245748613969*^9, 3.6562457508447733`*^9}},
 CellTags->{
  "ID:2718947", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] lm"},
 CellID->2718947],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
  RowBox[{"[", 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], 
    UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], 
   "]"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples membership \[CircleTimes])", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:204263407", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656245768972005*^9, 3.656245781140026*^9}, {
  3.656245841231332*^9, 3.656245850700548*^9}},
 CellTags->{
  "ID:204263407", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership \[CircleTimes]"},
 CellID->204263407],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], 
   UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], "\[Equal]", 
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], 
    "\[RightAngleBracket]"}], 
   UnderscriptBox["*", "\[ScriptCapitalP]"], 
   "q"}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples \[CircleTimes] *)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:134536220", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6562426760029716`*^9, 3.656242747341897*^9}, {
  3.656242781318756*^9, 3.6562427916771746`*^9}, {3.6562428735159187`*^9, 
  3.656242877181925*^9}, {3.656244415383827*^9, 3.656244419439834*^9}, {
  3.656245752217575*^9, 3.656245758753987*^9}, {3.6562458435245357`*^9, 
  3.6562458545225554`*^9}},
 CellTags->{
  "ID:134536220", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] *"},
 CellID->134536220],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
     RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], 
   UnderscriptBox["*", "\[ScriptCapitalP]"], "q"}], "\[Equal]", 
  RowBox[{
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"\[LeftAngleBracket]", 
       RowBox[{"c1", ",", "t1"}], "\[RightAngleBracket]"}], ",", 
      RowBox[{"r1", "..."}]}], "\[RightAngleBracket]"}], 
    UnderscriptBox["\[CircleTimes]", "\[ScriptCapitalP]"], "q"}], 
   UnderscriptBox["+", "\[ScriptCapitalP]"], 
   RowBox[{
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{"r1", "..."}], "\[RightAngleBracket]"}], 
    UnderscriptBox["*", "\[ScriptCapitalP]"], 
    "q"}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples * alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:387729163", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6562481372855654`*^9, 3.6562481482991843`*^9}, {
  3.656248205098884*^9, 3.656248252819368*^9}},
 CellTags->{
  "ID:387729163", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples * alternative"},
 CellID->387729163],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->90621649]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->483115236],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples \[CircleTimes] empty)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples \[CircleTimes] empty)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], 
               UnderscriptBox["\[CircleTimes]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
             "\[Equal]", 
             
             RowBox[{"\[LeftAngleBracket]", 
               "\[RightAngleBracket]"}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             RowBox[{
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                   StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                 "\[RightAngleBracket]"}], 
               UnderscriptBox["\[CircleTimes]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], 
               RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
             "\[Equal]", 
             
             RowBox[{"\[LeftAngleBracket]", 
               "\[RightAngleBracket]"}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:161672314"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p161672314-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p161672314-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:161672314", "Proof|ID:161672314-1"},
 CellID->221366480],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p161672314-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->341310380]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples \[CircleTimes] prepend)\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples \[CircleTimes] prepend)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              RowBox[{
                StyleBox["c2", "ExpressionVariable"], ",", 
                StyleBox["t2", "ExpressionVariable"], ",", 
                StyleBox[
                 RowBox[{"r2", "..."}], "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], 
                   UnderscriptBox["\[CircleTimes]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[Equal]", 
                 RowBox[{"\[Piecewise]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["c2", "ExpressionVariable"]}], ",", 
                    RowBox[{
                    StyleBox["t1", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    StyleBox["t2", "ExpressionVariable"]}]}], 
                    "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], 
                    UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], 
                    UnderscriptBox["\[CircleTimes]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}]}], "\[DoubleLeftArrow]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["c2", "ExpressionVariable"]}], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalR]"]}]}, {
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], 
                    UnderscriptBox["\[CircleTimes]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "\[DoubleLeftArrow]", 
                    "True"}}]}]}]}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              RowBox[{
                StyleBox["c2", "ExpressionVariable"], ",", 
                StyleBox["t2", "ExpressionVariable"], ",", 
                StyleBox[
                 RowBox[{"r2", "..."}], "ExpressionVariable"]}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], 
                   UnderscriptBox["\[CircleTimes]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c2", "ExpressionVariable"], ",", 
                    StyleBox["t2", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}]}], "\[Equal]", 
                 RowBox[{"\[Piecewise]", 
                   GridBox[{{
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["c2", "ExpressionVariable"]}], ",", 
                    RowBox[{
                    StyleBox["t1", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalT]"], 
                    StyleBox["t2", "ExpressionVariable"]}]}], 
                    "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], 
                    UnderscriptBox["+", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], 
                    UnderscriptBox["\[CircleTimes]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}]}], "\[DoubleLeftArrow]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], 
                    UnderscriptBox["*", "\[ScriptCapitalR]"], 
                    StyleBox["c2", "ExpressionVariable"]}], "\[NotEqual]", 
                    UnderscriptBox["0", "\[ScriptCapitalR]"]}]}, {
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], 
                    UnderscriptBox["\[CircleTimes]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r2", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}]}], "\[DoubleLeftArrow]", 
                    "True"}}]}]}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:71703914"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p71703914-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p71703914-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:71703914", "Proof|ID:71703914-1"},
 CellID->82366827],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p71703914-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->220781038]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples \[CircleTimes] lm)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples \[CircleTimes] lm)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], 
                 UnderscriptBox["\[CircleTimes]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                 StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], 
                 UnderscriptBox["\[CircleTimes]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                 StyleBox[
                 "q", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], 
                 UnderscriptBox["\[CircleTimes]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                 StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], 
                 UnderscriptBox["\[CircleTimes]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                 StyleBox[
                 "q", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:2718947"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p2718947-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p2718947-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:2718947", "Proof|ID:2718947-1"},
 CellID->181253020],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p2718947-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellChangeTimes->{3.6562459827411804`*^9},
 CellID->302282372]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples membership \[CircleTimes])\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples membership \[CircleTimes])\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], 
                 UnderscriptBox["\[CircleTimes]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                 StyleBox["q", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], 
                 UnderscriptBox["\[CircleTimes]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                 StyleBox["q", "ExpressionVariable"]}], 
               "]"}]}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:204263407"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p204263407-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p204263407-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:204263407", "Proof|ID:204263407-1"},
 CellID->426944387],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p204263407-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->74746441]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples \[CircleTimes] *)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples \[CircleTimes] *)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], 
                 UnderscriptBox["\[CircleTimes]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                 StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], 
                 UnderscriptBox["*", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                 StyleBox[
                 "q", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], 
                 UnderscriptBox["\[CircleTimes]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                 StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], 
                 UnderscriptBox["*", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                 StyleBox[
                 "q", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:134536220"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p134536220-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p134536220-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:134536220", "Proof|ID:134536220-1"},
 CellID->170925754],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p134536220-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellID->103658595]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples * alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples * alternative)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], 
                 UnderscriptBox["*", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                 StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], 
                   UnderscriptBox["\[CircleTimes]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   StyleBox["q", "ExpressionVariable"]}], 
                 UnderscriptBox["+", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                   UnderscriptBox["*", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   StyleBox[
                   "q", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["c1", "ExpressionVariable"], ",", 
            StyleBox["t1", "ExpressionVariable"], ",", 
            StyleBox[
             RowBox[{"r1", "..."}], "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[Element]", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]], "[", 
             RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], ",", 
                 StyleBox[
                  RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
               "\[RightAngleBracket]"}], "]"}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                   "\[RightAngleBracket]"}], 
                 UnderscriptBox["*", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                 StyleBox["q", "ExpressionVariable"]}], "\[Equal]", 
               RowBox[{
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["c1", "ExpressionVariable"], ",", 
                    StyleBox["t1", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], ",", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], 
                   UnderscriptBox["\[CircleTimes]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   StyleBox["q", "ExpressionVariable"]}], 
                 UnderscriptBox["+", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                 RowBox[{
                   RowBox[{"\[LeftAngleBracket]", 
                    StyleBox[
                    RowBox[{"r1", "..."}], "ExpressionVariable"], 
                    "\[RightAngleBracket]"}], 
                   UnderscriptBox["*", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                   StyleBox[
                   "q", "ExpressionVariable"]}]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:387729163"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p387729163-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p387729163-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:387729163", "Proof|ID:387729163-1"},
 CellID->722597358],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p387729163-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->89283540]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->8782086],

Cell[CellGroupData[{

Cell["Theorem: PolyTuples has Multiplication", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6562424521737785`*^9, 3.6562424584605894`*^9}},
 CellID->208907366],

Cell[BoxData[
 RowBox[{"hasMultiplication", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples hasMultiplication)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:65280633", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656242461096994*^9, 3.656242467446205*^9}},
 CellTags->{
  "ID:65280633", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasMultiplication"},
 CellID->65280633],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->606425772]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->6209804],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples hasMultiplication)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples hasMultiplication)\"\>", "FormReferenceHover",
       
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"hasMultiplication", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"hasMultiplication", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:65280633"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p65280633-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p65280633-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:65280633", "Proof|ID:65280633-1"},
 CellID->187277372],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p65280633-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->355668730]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Extensionality (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.655548901056483*^9, 3.6555489036460876`*^9}, {
  3.655700964617879*^9, 3.6557009686426864`*^9}},
 CellID->40538141],

Cell["", "OpenEnvironment",
 CellID->87454831],

Cell[CellGroupData[{

Cell["Lemma: Extensionality", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.655548974548212*^9, 3.6555489788694196`*^9}},
 CellID->164838296],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", "P"], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{
        UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
        RowBox[{"[", "p", "]"}]}]], 
      RowBox[{
       UnderscriptBox["\[ForAll]", 
        RowBox[{
         UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
         RowBox[{"[", "q", "]"}]}]], 
       RowBox[{
        TagBox["(",
         "AutoParentheses"], 
        RowBox[{
         RowBox[{"P", "[", 
          RowBox[{"p", ",", "q"}], "]"}], "\[Implies]", 
         RowBox[{
          TagBox["(",
           "AutoParentheses"], 
          RowBox[{"p", "\[Equal]", "q"}], 
          TagBox[")",
           "AutoParentheses"]}]}], 
        TagBox[")",
         "AutoParentheses"]}]}]}], 
     TagBox[")",
      "AutoParentheses"]}], "\[DoubleLongLeftRightArrow]", 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      UnderscriptBox["\[ForAll]", 
       RowBox[{"isTuple", "[", 
        RowBox[{"p", ",", "q"}], "]"}]], 
      RowBox[{
       TagBox["(",
        "AutoParentheses"], 
       RowBox[{
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
           RowBox[{"[", "p", "]"}]}], "\[And]", 
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
           RowBox[{"[", "q", "]"}]}], "\[And]", 
          RowBox[{"P", "[", 
           RowBox[{"p", ",", "q"}], "]"}]}], 
         TagBox[")",
          "AutoParentheses"]}], "\[Implies]", 
        RowBox[{
         TagBox["(",
          "AutoParentheses"], 
         RowBox[{"p", "\[Equal]", "q"}], 
         TagBox[")",
          "AutoParentheses"]}]}], 
       TagBox[")",
        "AutoParentheses"]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples \[ForAll] alternative 2)", "FrameLabel"], "  ", 
        
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:48571309", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.655642046967084*^9, 3.6556421639360895`*^9}},
 CellTags->{
  "ID:48571309", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[ForAll] alternative 2"},
 CellID->48571309],

Cell[BoxData[
 UnderscriptBox["\[ForAll]", 
  RowBox[{
   UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
   RowBox[{"[", 
    RowBox[{"p", ",", "q"}], "]"}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 617696571}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6555494025701637`*^9, 3.6555494112125793`*^9}},
 CellTags->{
  "ID:617696571", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->617696571],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "t", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{"p", ",", "t"}], "]"}], "\[Equal]", 
      UnderscriptBox["0", "\[ScriptCapitalR]"]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"p", "\[Equal]", 
    RowBox[{"\[LeftAngleBracket]", "\[RightAngleBracket]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples=empty sufficient)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:776548", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.655643240066986*^9, 3.655643258584219*^9}},
 CellTags->{
  "ID:776548", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples=empty sufficient"},
 CellID->776548],

Cell[BoxData[
 RowBox[{
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["\[ForAll]", 
     RowBox[{
      UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
      RowBox[{"[", "t", "]"}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{"p", ",", "t"}], "]"}], "\[Equal]", 
      RowBox[{
       UnderscriptBox["C", "\[ScriptCapitalP]"], "[", 
       RowBox[{"q", ",", "t"}], "]"}]}], 
     TagBox[")",
      "AutoParentheses"]}]}], 
   TagBox[")",
    "AutoParentheses"]}], "\[Implies]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{"p", "\[Equal]", "q"}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples extensional)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:304028020", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.655548918809314*^9, 3.655548965624996*^9}, {
  3.655549413380983*^9, 3.655549418028591*^9}},
 CellTags->{
  "ID:304028020", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples extensional"},
 CellID->304028020],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->121515229]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->85801361],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples \[ForAll] alternative 2)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples \[ForAll] alternative 2)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"]}], "]"}], 
                   "\[Implies]", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["q", "ExpressionVariable"]}]}]}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isTuple", "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isTuple", "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}], "\[And]", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"]}], "]"}]}], 
               "\[Implies]", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                 StyleBox[
                 "q", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          RowBox[{
            StyleBox["P", "ExpressionVariable"]}]], 
         RowBox[{
           RowBox[{"(", 
             RowBox[{
               UnderscriptBox["\[ForAll]", 
                GridBox[{{
                   RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", 
                  GridBox[{{
                    RowBox[{
                    UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
                 RowBox[{
                   RowBox[{
                    StyleBox["P", "ExpressionVariable"], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"]}], "]"}], 
                   "\[Implies]", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["q", "ExpressionVariable"]}]}]}]}], ")"}], 
           "\[DoubleLeftRightArrow]", 
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{"isTuple", "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}]}, {
                 RowBox[{"isTuple", "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], "\[And]", 
                 RowBox[{
                   UnderscriptBox["\[Element]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}], "\[And]", 
                 RowBox[{
                   StyleBox["P", "ExpressionVariable"], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"]}], "]"}]}], 
               "\[Implies]", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                 StyleBox[
                 "q", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:48571309"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p48571309-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p48571309-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:48571309", "Proof|ID:48571309-1"},
 CellID->399659369],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p48571309-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {25., 31.}}]], "ProofInfoBody",
 CellID->164181676]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples=empty sufficient)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples=empty sufficient)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[Equal]", 
             
             RowBox[{"\[LeftAngleBracket]", 
               "\[RightAngleBracket]"}]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               UnderscriptBox["0", "\[ScriptCapitalR]"]}]}], "\[Implies]", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[Equal]", 
             
             RowBox[{"\[LeftAngleBracket]", 
               "\[RightAngleBracket]"}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:776548"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p776548-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p776548-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:776548", "Proof|ID:776548-1"},
 CellID->379176191],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p776548-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->227491426]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples extensional)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples extensional)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["C", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "q", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm], 
      DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             UnderscriptBox["\[ForAll]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["t", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["C", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}], "\[Equal]", 
               RowBox[{
                 UnderscriptBox["C", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["t", "ExpressionVariable"]}], "]"}]}]}], 
           "\[Implies]", 
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[Equal]", 
             StyleBox[
             "q", "ExpressionVariable"]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:304028020"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p304028020-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p304028020-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:304028020", "Proof|ID:304028020-1"},
 CellID->86924546],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p304028020-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->131380879]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->186415220],

Cell[CellGroupData[{

Cell["Theorem: PolyTuples is Extensional", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.655700613474862*^9, 3.6557006225072784`*^9}, {
  3.6557009867699184`*^9, 3.6557009868791184`*^9}},
 CellID->193487632],

Cell[BoxData[
 RowBox[{"isExtensional", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples isExtensional)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:589891585", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6557006261108847`*^9, 3.6557006311496935`*^9}},
 CellTags->{
  "ID:589891585", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isExtensional"},
 CellID->589891585],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->54820822]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->232514858],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples isExtensional)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples isExtensional)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isExtensional", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isExtensional", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:589891585"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p589891585-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p589891585-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:589891585", "Proof|ID:589891585-1"},
 CellID->116448114],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p589891585-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->332438606]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Order Relation (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6561412183913164`*^9, 3.656141221604922*^9}, {
  3.6562518348094606`*^9, 3.656251838959068*^9}},
 CellID->69393044],

Cell["", "OpenEnvironment",
 CellID->263183869],

Cell[CellGroupData[{

Cell["Theorem: PolyTuples is Ordered", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6562492170580616`*^9, 3.656249222518071*^9}},
 CellID->39919617],

Cell[BoxData[
 RowBox[{"isOrderedRec", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples isOrderedRec)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:160815443", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656249263858144*^9, 3.6562492660421476`*^9}},
 CellTags->{
  "ID:160815443", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isOrderedRec"},
 CellID->160815443],

Cell[BoxData[
 RowBox[{"isOrdered", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples isOrdered)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:397107848", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656249224780075*^9, 3.656249229881284*^9}},
 CellTags->{
  "ID:397107848", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isOrdered"},
 CellID->397107848],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->46796859]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->17008929],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples isOrderedRec)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples isOrderedRec)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isOrderedRec", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isOrderedRec", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:160815443"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p160815443-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p160815443-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:160815443", "Proof|ID:160815443-1"},
 CellID->9775189],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p160815443-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->144892605]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples isOrdered)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples isOrdered)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isOrdered", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isOrdered", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:397107848"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p397107848-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p397107848-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:397107848", "Proof|ID:397107848-1"},
 CellID->3787871],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p397107848-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->521252860]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Multipliers (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.65614122642533*^9, 3.6561412282505336`*^9}, {
  3.6563075249803724`*^9, 3.6563075295199804`*^9}},
 CellID->11061911],

Cell["", "OpenEnvironment",
 CellID->55195651],

Cell[CellGroupData[{

Cell["Theorem: PolyTuples has Multipliers", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656251890251958*^9, 3.6562518968819695`*^9}},
 CellID->186389390],

Cell[BoxData[
 RowBox[{"hasMultipliers", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples hasMultipliers)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:156782366", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6562518989567733`*^9, 3.656251904541583*^9}},
 CellTags->{
  "ID:156782366", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasMultipliers"},
 CellID->156782366],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->56161024]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->666198591],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples hasMultipliers)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples hasMultipliers)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"hasMultipliers", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"hasMultipliers", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:156782366"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p156782366-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p156782366-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:156782366", "Proof|ID:156782366-1"},
 CellID->558592865],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p156782366-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->82438810]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Main Result (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6563075196295633`*^9, 3.6563075228431687`*^9}, {
  3.6563078801335964`*^9, 3.656307884860405*^9}},
 CellID->78809743],

Cell["", "OpenEnvironment",
 CellID->149984650],

Cell[CellGroupData[{

Cell["Theorem: PolyTuples forms Reduction Polynomials", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656307553013622*^9, 3.6563075767568636`*^9}},
 CellID->157312756],

Cell[BoxData[
 RowBox[{"isReductionPolynomialDomain", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples isReductionPolynomialDomain)", "FrameLabel"], 
        "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:667417187", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6563075798300686`*^9, 3.656307586865681*^9}},
 CellTags->{
  "ID:667417187", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isReductionPolynomialDomain"},
 CellID->667417187],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->117448790]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->82630341],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples isReductionPolynomialDomain)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples isReductionPolynomialDomain)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isReductionPolynomialDomain", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isReductionPolynomialDomain", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:667417187"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p667417187-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p667417187-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:667417187", "Proof|ID:667417187-1"},
 CellID->486349173],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p667417187-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->78869945]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Algorithmic Reduction Polynomials (\[Checkmark])\
\>", "Subsection",
 CellChangeTimes->{{3.6555262007406826`*^9, 3.6555262064668927`*^9}, {
  3.656413277133313*^9, 3.656413280549719*^9}},
 CellID->9669308],

Cell[CellGroupData[{

Cell["LCRD (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.656307950026123*^9, 3.6563079507905245`*^9}, {
  3.6563964897923326`*^9, 3.65639649384834*^9}},
 CellID->375801328],

Cell["", "OpenEnvironment",
 CellID->325998897],

Cell[CellGroupData[{

Cell["Lemma: Properties of lcrd in PolyTuples", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656308291377124*^9, 3.656308299879139*^9}},
 CellID->90125590],

Cell[BoxData[
 UnderscriptBox[
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
    RowBox[{"[", 
     RowBox[{"p", ",", "q"}], "]"}]}]], 
  RowBox[{
   RowBox[{"p", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}], "\[And]", 
   RowBox[{"q", "\[NotEqual]", 
    UnderscriptBox["0", "\[ScriptCapitalP]"]}]}]]], "GlobalDeclaration",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Language`Session`Private`removeGlobal[{
         "C:\\Users\\Alex\\Documents\\Doktorat\\Theorema Notebooks\\Reduction \
Ring Theory\\PolyTuples.nb", 714775050}]]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656308374104069*^9, 3.656308397769311*^9}},
 CellTags->{
  "ID:714775050", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->714775050],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", 
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{
      UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
   RowBox[{"j", "\[GreaterEqual]", "i"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{
     UnderscriptBox["lcrd", "\[ScriptCapitalP]"], "[", 
     RowBox[{"p", ",", "q", ",", "i", ",", "j"}], "]"}], "\[Equal]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{
         UnderscriptBox["let", 
          RowBox[{"l", "=", 
           RowBox[{
            UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
            RowBox[{
             RowBox[{
              UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
             RowBox[{
              UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], ",",
              "i", ",", "j"}], "]"}]}]], 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           RowBox[{
            SubscriptBox["l", "i"], 
            UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
            RowBox[{
             UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
             RowBox[{
              RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
              ",", 
              RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], 
             "]"}]}], 
           UnderscriptBox["|", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
             "\[LeftBracketingBar]", "l", "\[RightBracketingBar]"}]}]]}], 
          "\[RightAngleBracket]"}]}], "\[DoubleLeftArrow]", 
        RowBox[{
         RowBox[{"p", "\[Equal]", "q"}], "\[Or]", 
         RowBox[{
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
          "\[NotEqual]", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}]}]}]},
       {
        RowBox[{
         UnderscriptBox["let", 
          RowBox[{"l", "=", 
           RowBox[{
            UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
            RowBox[{
             UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
            "]"}]}]], 
         RowBox[{"\[LeftAngleBracket]", 
          RowBox[{
           RowBox[{
            SubscriptBox["l", "i"], 
            UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
            RowBox[{
             UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
             RowBox[{
              RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
              ",", 
              RowBox[{
               UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}]}], 
             "]"}]}], 
           UnderscriptBox["|", 
            RowBox[{
             RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
             RowBox[{
             "\[LeftBracketingBar]", "l", "\[RightBracketingBar]"}]}]]}], 
          "\[RightAngleBracket]"}]}], "\[DoubleLeftArrow]", "otherwise"}
      }]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples lcrd alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:15424584", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656308511931311*^9, 3.656308546376172*^9}, {
  3.6563085865462427`*^9, 3.656308708132856*^9}, {3.6563921590447206`*^9, 
  3.6563921612443247`*^9}, {3.656393708645447*^9, 3.6563937098622494`*^9}},
 CellTags->{
  "ID:15424584", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd alternative"},
 CellID->15424584],

Cell[BoxData[
 RowBox[{
  RowBox[{
   UnderscriptBox["lcrd", "\[ScriptCapitalP]"], "[", "p", "]"}], "\[Equal]", 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    UnderscriptBox["let", 
     RowBox[{"l", "=", 
      RowBox[{
       UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
       RowBox[{
        UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], "]"}]}]], 
    
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{
       SubscriptBox["l", "i"], 
       UnderscriptBox["\[CenterDot]", "\[ScriptCapitalP]"], 
       RowBox[{
        UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}]}], 
      UnderscriptBox["|", 
       RowBox[{
        RowBox[{"i", "=", "1"}], ",", "\[Ellipsis]", ",", 
        RowBox[{"\[LeftBracketingBar]", "l", "\[RightBracketingBar]"}]}]]}], 
     "\[RightAngleBracket]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples lcrd unary alternative)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:168826371", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6563084022777185`*^9, 3.6563084620726233`*^9}},
 CellTags->{
  "ID:168826371", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd unary alternative"},
 CellID->168826371],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", 
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{
      UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
   RowBox[{"i", "<", "j"}]], 
  RowBox[{
   UnderscriptBox["\[ForAll]", 
    RowBox[{
     RowBox[{"k", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{"\[LeftBracketingBar]", 
      RowBox[{
       UnderscriptBox["lcrd", "\[ScriptCapitalP]"], "[", 
       RowBox[{"p", ",", "q", ",", "i", ",", "j"}], "]"}], 
      "\[RightBracketingBar]"}]}]], 
   RowBox[{
    UnderscriptBox["\[Exists]", 
     RowBox[{
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], 
       RowBox[{"[", "m", "]"}]}], ",", 
      RowBox[{
       UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], 
       RowBox[{"[", "s", "]"}]}]}]], 
    RowBox[{
     TagBox["(",
      "AutoParentheses"], 
     RowBox[{"\[And]", 
      RowBox[{"\[Piecewise]", GridBox[{
         {
          RowBox[{
           UnderscriptBox["\[Element]", "\[ScriptCapitalP]"], 
           RowBox[{"[", 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"m", ",", "s"}], "\[RightAngleBracket]"}], 
             "\[RightAngleBracket]"}], "]"}]}]},
         {
          RowBox[{"m", "\[Equal]", 
           RowBox[{"\[Piecewise]", GridBox[{
              {
               SubscriptBox[
                RowBox[{
                 UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                  RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}],
                   ",", 
                  RowBox[{
                   UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}],
                   ",", "i", ",", "j"}], "]"}], "k"], "\[DoubleLeftArrow]", 
               RowBox[{
                RowBox[{"p", "\[Equal]", "q"}], "\[Or]", 
                RowBox[{
                 RowBox[{
                  UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
                 "\[NotEqual]", 
                 RowBox[{
                  UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", 
                  "]"}]}]}]},
              {
               SubscriptBox[
                RowBox[{
                 UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                 RowBox[{
                  UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
                 "]"}], "k"], "\[DoubleLeftArrow]", "otherwise"}
             }]}]}]},
         {
          RowBox[{
           SubscriptBox[
            RowBox[{
             UnderscriptBox["lcrd", "\[ScriptCapitalP]"], "[", 
             RowBox[{"p", ",", "q", ",", "i", ",", "j"}], "]"}], "k"], 
           "\[Equal]", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"m", ",", "s"}], "\[RightAngleBracket]"}], 
            "\[RightAngleBracket]"}]}]},
         {
          RowBox[{
           RowBox[{
            UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "p", "]"}], 
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "s"}]},
         {
          RowBox[{
           RowBox[{
            UnderscriptBox["lp", "\[ScriptCapitalP]"], "[", "q", "]"}], 
           UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], "s"}]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}]}]}], 
     TagBox[")",
      "AutoParentheses"]}]}]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples lcrd lc lp)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:228703376", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656398345498416*^9, 3.6563984796898513`*^9}, {
   3.6563986465165443`*^9, 3.656398668309783*^9}, {3.65639879744681*^9, 
   3.6563988443716917`*^9}, {3.6563989226994295`*^9, 3.656398946505071*^9}, 
   3.656399360609799*^9, {3.6564020244652877`*^9, 3.656402066149561*^9}, {
   3.6564023874789257`*^9, 3.6564024336082067`*^9}, {3.6564024751510797`*^9, 
   3.656402495992716*^9}, {3.656402650744988*^9, 3.656402671290224*^9}, {
   3.6564029089098415`*^9, 3.656402934930687*^9}},
 CellTags->{
  "ID:228703376", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd lc lp"},
 CellID->228703376],

Cell[BoxData[
 RowBox[{
  UnderscriptBox[
   UnderscriptBox["\[ForAll]", 
    RowBox[{"i", ",", 
     RowBox[{"j", "=", "1"}], ",", "\[Ellipsis]", ",", 
     RowBox[{
      UnderscriptBox["multN", "\[ScriptCapitalP]"], "[", "p", "]"}]}]], 
   RowBox[{"i", "<", "j"}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"\[LeftBracketingBar]", 
     RowBox[{
      UnderscriptBox["lcrd", "\[ScriptCapitalP]"], "[", 
      RowBox[{"p", ",", "q", ",", "i", ",", "j"}], "]"}], 
     "\[RightBracketingBar]"}], "\[Equal]", 
    RowBox[{"\[Piecewise]", GridBox[{
       {
        RowBox[{"\[LeftBracketingBar]", 
         RowBox[{
          UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
          RowBox[{
           RowBox[{
            UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], ",", 
           RowBox[{
            UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}], ",", 
           "i", ",", "j"}], "]"}], "\[RightBracketingBar]"}], 
        "\[DoubleLeftArrow]", 
        RowBox[{
         RowBox[{"p", "\[Equal]", "q"}], "\[Or]", 
         RowBox[{
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], 
          "\[NotEqual]", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "q", "]"}]}]}]},
       {
        RowBox[{"\[LeftBracketingBar]", 
         RowBox[{
          UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
          RowBox[{
           UnderscriptBox["lc", "\[ScriptCapitalP]"], "[", "p", "]"}], "]"}], 
         "\[RightBracketingBar]"}], "\[DoubleLeftArrow]", "otherwise"}
      }]}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples lcrd length)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:354399952", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6564108217357965`*^9, 3.656410879143897*^9}},
 CellTags->{
  "ID:354399952", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd length"},
 CellID->354399952],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->13100681]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->640374478],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples lcrd alternative)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples lcrd alternative)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
               StyleBox["i", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lcrd", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{"\[Piecewise]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["let", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["l", "ExpressionVariable"], "=", 
                    RowBox[{
                    UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["l", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["l", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}]}], "\[DoubleLeftArrow]", 
                  RowBox[{
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["q", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}]}, {
                  RowBox[{
                    UnderscriptBox["let", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["l", "ExpressionVariable"], "=", 
                    RowBox[{
                    UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}]}]}}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["l", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["l", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}]}], "\[DoubleLeftArrow]", 
                  "True"}}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["j", "ExpressionVariable"], "\[GreaterEqual]", 
               StyleBox["i", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lcrd", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               RowBox[{
                 StyleBox["p", "ExpressionVariable"], ",", 
                 StyleBox["q", "ExpressionVariable"], ",", 
                 StyleBox["i", "ExpressionVariable"], ",", 
                 StyleBox["j", "ExpressionVariable"]}], "]"}], "\[Equal]", 
             RowBox[{"\[Piecewise]", 
               GridBox[{{
                  RowBox[{
                    UnderscriptBox["let", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["l", "ExpressionVariable"], "=", 
                    RowBox[{
                    UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}]}]}}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["l", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["l", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}]}], "\[DoubleLeftArrow]", 
                  RowBox[{
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["q", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}]}, {
                  RowBox[{
                    UnderscriptBox["let", 
                    GridBox[{{
                    RowBox[{
                    StyleBox["l", "ExpressionVariable"], "=", 
                    RowBox[{
                    UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}]}]}}]], 
                    RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    RowBox[{
                    SubscriptBox[
                    StyleBox["l", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["lcm", "\[ScriptCapitalT]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lp", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lp", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}], "]"}]}], 
                    UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["l", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                    "\[RightAngleBracket]"}]}], "\[DoubleLeftArrow]", 
                  "True"}}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:15424584"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p15424584-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p15424584-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:15424584", "Proof|ID:15424584-1"},
 CellID->190628320],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p15424584-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.65639596991842*^9, 3.656395972336424*^9}},
 CellID->124361579]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples lcrd unary alternative)\"\>", "FormReference",
       
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples lcrd unary alternative)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lcrd", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["let", 
                GridBox[{{
                   RowBox[{
                    StyleBox["l", "ExpressionVariable"], "=", 
                    RowBox[{
                    UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}]}]}}]], 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    SubscriptBox[
                    StyleBox["l", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["lp", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["l", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                 "\[RightAngleBracket]"}]}]}]}]}], $CellContext`TheoremaForm],
       DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}}]], 
         RowBox[{
           RowBox[{
             StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
             UnderscriptBox["0", 
              RowBox[{"PolyTuples", "[", 
                RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                "]"}]]}], "\[Implies]", 
           RowBox[{
             RowBox[{
               UnderscriptBox["lcrd", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}], "\[Equal]", 
             RowBox[{
               UnderscriptBox["let", 
                GridBox[{{
                   RowBox[{
                    StyleBox["l", "ExpressionVariable"], "=", 
                    RowBox[{
                    UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}]}]}}]], 
               RowBox[{"\[LeftAngleBracket]", 
                 RowBox[{
                   RowBox[{
                    SubscriptBox[
                    StyleBox["l", "ExpressionVariable"], 
                    StyleBox["i", "ExpressionVariable"]], 
                    UnderscriptBox["\[CenterDot]", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], 
                    RowBox[{
                    UnderscriptBox["lp", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}]}], 
                   UnderscriptBox["|", 
                    GridBox[{{
                    RowBox[{
                    RowBox[{
                    StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
                    "\[Ellipsis]", ",", 
                    RowBox[{"\[LeftBracketingBar]", 
                    StyleBox["l", "ExpressionVariable"], 
                    "\[RightBracketingBar]"}]}]}}]]}], 
                 "\[RightAngleBracket]"}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:168826371"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p168826371-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p168826371-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:168826371", "Proof|ID:168826371-1"},
 CellID->493554176],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p168826371-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->11398291]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples lcrd lc lp)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples lcrd lc lp)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 RowBox[{
                   UnderscriptBox["lcrd", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "<", 
               StyleBox["j", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], "]"}],
                "\[And]", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{"\[Piecewise]", 
                   GridBox[{{
                    SubscriptBox[
                    RowBox[{
                    UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    "\[DoubleLeftArrow]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["q", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}]}, {
                    SubscriptBox[
                    RowBox[{
                    UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    "\[DoubleLeftArrow]", "True"}}]}]}], "\[And]", 
               RowBox[{
                 SubscriptBox[
                  RowBox[{
                    UnderscriptBox["lcrd", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                  StyleBox["k", "ExpressionVariable"]], "\[Equal]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                 StyleBox["s", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                 StyleBox[
                 "s", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["k", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{"\[LeftBracketingBar]", 
                 RowBox[{
                   UnderscriptBox["lcrd", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                 "\[RightBracketingBar]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "<", 
               StyleBox["j", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             UnderscriptBox["\[Exists]", 
              GridBox[{{
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalR]"], "[", 
                   StyleBox["m", "ExpressionVariable"], "]"}]}, {
                 RowBox[{
                   UnderscriptBox["\[Element]", "\[ScriptCapitalT]"], "[", 
                   StyleBox["s", "ExpressionVariable"], "]"}]}}]], 
             RowBox[{
               RowBox[{
                 UnderscriptBox["\[Element]", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\[RightAngleBracket]"}], "]"}],
                "\[And]", 
               RowBox[{
                 StyleBox["m", "ExpressionVariable"], "\[Equal]", 
                 RowBox[{"\[Piecewise]", 
                   GridBox[{{
                    SubscriptBox[
                    RowBox[{
                    UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    "\[DoubleLeftArrow]", 
                    RowBox[{
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["q", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}]}, {
                    SubscriptBox[
                    RowBox[{
                    UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}], 
                    StyleBox["k", "ExpressionVariable"]], 
                    "\[DoubleLeftArrow]", "True"}}]}]}], "\[And]", 
               RowBox[{
                 SubscriptBox[
                  RowBox[{
                    UnderscriptBox["lcrd", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], ",", 
                    StyleBox["q", "ExpressionVariable"], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                  StyleBox["k", "ExpressionVariable"]], "\[Equal]", 
                 RowBox[{"\[LeftAngleBracket]", 
                   RowBox[{"\[LeftAngleBracket]", 
                    RowBox[{
                    StyleBox["m", "ExpressionVariable"], ",", 
                    StyleBox["s", "ExpressionVariable"]}], 
                    "\[RightAngleBracket]"}], "\[RightAngleBracket]"}]}], 
               "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["p", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                 StyleBox["s", "ExpressionVariable"]}], "\[And]", 
               RowBox[{
                 RowBox[{
                   UnderscriptBox["lp", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                   StyleBox["q", "ExpressionVariable"], "]"}], 
                 UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]"], 
                 StyleBox[
                 "s", "ExpressionVariable"]}]}]}]}]}], \
$CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:228703376"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p228703376-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p228703376-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:228703376", "Proof|ID:228703376-1"},
 CellID->48977005],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p228703376-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {53., 60.}}]], "ProofInfoBody",
 CellChangeTimes->{{3.6564085562012124`*^9, 3.6564085574180145`*^9}},
 CellID->1819865]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples lcrd length)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples lcrd length)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "<", 
               StyleBox["j", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[LeftBracketingBar]", 
               RowBox[{
                 UnderscriptBox["lcrd", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], "]"}], 
               "\[RightBracketingBar]"}], "\[Equal]", 
             RowBox[{"\[Piecewise]", 
               GridBox[{{
                  RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    "\[RightBracketingBar]"}], "\[DoubleLeftArrow]", 
                  RowBox[{
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["q", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}]}, {
                  RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}], 
                    "\[RightBracketingBar]"}], "\[DoubleLeftArrow]", 
                  "True"}}]}]}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["p", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               UnderscriptBox["\[Element]", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]], "[", 
               StyleBox["q", "ExpressionVariable"], "]"}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["i", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}, {
             RowBox[{
               RowBox[{
                 StyleBox["j", "ExpressionVariable"], "=", "1"}], ",", 
               "\[Ellipsis]", ",", 
               RowBox[{
                 UnderscriptBox["multN", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 StyleBox["p", "ExpressionVariable"], "]"}]}]}}]], 
         RowBox[{
           RowBox[{
             RowBox[{
               StyleBox["p", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["q", "ExpressionVariable"], "\[NotEqual]", 
               UnderscriptBox["0", 
                RowBox[{"PolyTuples", "[", 
                  RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                  "]"}]]}], "\[And]", 
             RowBox[{
               StyleBox["i", "ExpressionVariable"], "<", 
               StyleBox["j", "ExpressionVariable"]}]}], "\[Implies]", 
           RowBox[{
             RowBox[{"\[LeftBracketingBar]", 
               RowBox[{
                 UnderscriptBox["lcrd", 
                  RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                 RowBox[{
                   StyleBox["p", "ExpressionVariable"], ",", 
                   StyleBox["q", "ExpressionVariable"], ",", 
                   StyleBox["i", "ExpressionVariable"], ",", 
                   StyleBox["j", "ExpressionVariable"]}], "]"}], 
               "\[RightBracketingBar]"}], "\[Equal]", 
             RowBox[{"\[Piecewise]", 
               GridBox[{{
                  RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], ",", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}], ",", 
                    StyleBox["i", "ExpressionVariable"], ",", 
                    StyleBox["j", "ExpressionVariable"]}], "]"}], 
                    "\[RightBracketingBar]"}], "\[DoubleLeftArrow]", 
                  RowBox[{
                    RowBox[{
                    StyleBox["p", "ExpressionVariable"], "\[Equal]", 
                    StyleBox["q", "ExpressionVariable"]}], "\[Or]", 
                    RowBox[{
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "\[NotEqual]", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["q", "ExpressionVariable"], "]"}]}]}]}, {
                  RowBox[{"\[LeftBracketingBar]", 
                    RowBox[{
                    UnderscriptBox["lcrd", "\[ScriptCapitalR]"], "[", 
                    RowBox[{
                    UnderscriptBox["lc", 
                    RowBox[{"PolyTuples", "[", 
                    RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], 
                    "]"}]], "[", 
                    StyleBox["p", "ExpressionVariable"], "]"}], "]"}], 
                    "\[RightBracketingBar]"}], "\[DoubleLeftArrow]", 
                  "True"}}]}]}]}]}], $CellContext`TheoremaForm]], 
     "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:354399952"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p354399952-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p354399952-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:354399952", "Proof|ID:354399952-1"},
 CellID->182853122],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p354399952-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {38., 45.}}]], "ProofInfoBody",
 CellID->237714126]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->187246246],

Cell[CellGroupData[{

Cell["Theorem: PolyTuples has LCRD", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6563962939807887`*^9, 3.656396298535997*^9}},
 CellID->332568537],

Cell[BoxData[
 RowBox[{"hasLCRD", "[", "\[ScriptCapitalP]", "]"}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples hasLCRD)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:479598778", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656396301203602*^9, 3.6563963058056097`*^9}},
 CellTags->{
  "ID:479598778", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasLCRD"},
 CellID->479598778],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->324782272]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->144574756],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples hasLCRD)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples hasLCRD)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"hasLCRD", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"hasLCRD", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:479598778"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p479598778-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p479598778-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:479598778", "Proof|ID:479598778-1"},
 CellID->690537389],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p479598778-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {24., 30.}}]], "ProofInfoBody",
 CellID->458247862]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["RDM (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.656307955751333*^9, 3.6563079564845343`*^9}, {
  3.6563981408104563`*^9, 3.656398145334464*^9}},
 CellID->31285283],

Cell["", "OpenEnvironment",
 CellID->169886156],

Cell[CellGroupData[{

Cell["Theorem: PolyTuples has RDM", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6563965073579636`*^9, 3.656396511991172*^9}},
 CellID->922306034],

Cell[BoxData[
 RowBox[{"hasRDM", "[", "\[ScriptCapitalP]", "]"}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples hasRDM)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:463361785", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656396514487176*^9, 3.6563965203683863`*^9}},
 CellTags->{
  "ID:463361785", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasRDM"},
 CellID->463361785],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->133661739]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->121927638],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples hasRDM)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples hasRDM)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"hasRDM", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"hasRDM", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:463361785"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p463361785-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p463361785-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:463361785", "Proof|ID:463361785-1"},
 CellID->260093492],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p463361785-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->130872278]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["CPM (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.6563079611333427`*^9, 3.6563079619133444`*^9}, {
  3.656411838938385*^9, 3.656411843197192*^9}},
 CellID->5046155],

Cell["", "OpenEnvironment",
 CellID->8219039],

Cell[CellGroupData[{

Cell["Theorem: PolyTuples has CPM", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6563981654740996`*^9, 3.6563981700137076`*^9}},
 CellID->180390280],

Cell[BoxData[
 RowBox[{"hasCPM", "[", "\[ScriptCapitalP]", "]"}]], "FormalTextInputFormula",\

 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples hasCPM)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:7345776", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6563981723693113`*^9, 3.6563981767841196`*^9}},
 CellTags->{
  "ID:7345776", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasCPM"},
 CellID->7345776],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->593022431]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->276041390],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples hasCPM)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples hasCPM)\"\>", "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"hasCPM", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"hasCPM", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:7345776"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p7345776-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p7345776-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:7345776", "Proof|ID:7345776-1"},
 CellID->11260575],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p7345776-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {46., 52.}}]], "ProofInfoBody",
 CellID->419962729]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Main Result (\[Checkmark])", "Subsubsection",
 CellChangeTimes->{{3.656307966983353*^9, 3.6563079701969585`*^9}, {
  3.6564132697857003`*^9, 3.6564132733581066`*^9}},
 CellID->2896086],

Cell["", "OpenEnvironment",
 CellID->124935155],

Cell[CellGroupData[{

Cell["Theorem: PolyTuples forms Algorithmic Reduction Polynomials", \
"EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.656307980024976*^9, 3.656307994860602*^9}},
 CellID->572686989],

Cell[BoxData[
 RowBox[{"isAlgoReductionPolynomialDomain", "[", "\[ScriptCapitalP]", 
  "]"}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox[
        "(PolyTuples isAlgoReductionPolynomialDomain)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:1014777887", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.656307998869809*^9, 3.6563080071378236`*^9}},
 CellTags->{
  "ID:1014777887", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isAlgoReductionPolynomialDomain"},
 CellID->1014777887],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->253163528]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->268518370],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples isAlgoReductionPolynomialDomain)\"\>", 
       "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples isAlgoReductionPolynomialDomain)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{"isAlgoReductionPolynomialDomain", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{"isAlgoReductionPolynomialDomain", "[", 
         RowBox[{"PolyTuples", "[", 
           RowBox[{"\[ScriptCapitalR]", ",", "\[ScriptCapitalT]"}], "]"}], 
         "]"}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:1014777887"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p1014777887-1-po.m\"}]"]],
   
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p1014777887-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:1014777887", "Proof|ID:1014777887-1"},
 CellID->68078823],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p1014777887-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->423181797]
}, Closed]],

Cell["", "OpenEnvironment",
 CellID->48219384],

Cell[CellGroupData[{

Cell["Theorem: PolyTuples is Algorithmic Reduction Ring", "EnvironmentHeader",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Common`removeEnvironment[
         ButtonNotebook[]]]]]}, {None, None}},
 CellChangeTimes->{{3.6563080362006745`*^9, 3.656308045529491*^9}},
 CellID->190868484],

Cell[BoxData[
 RowBox[{
  UnderscriptBox["\[ForAll]", 
   RowBox[{
    RowBox[{"isAlgoReductionRing", "[", "\[ScriptCapitalR]1", "]"}], ",", 
    RowBox[{"isCommPPDomain", "[", "\[ScriptCapitalT]1", "]"}]}]], 
  RowBox[{
   TagBox["(",
    "AutoParentheses"], 
   RowBox[{
    RowBox[{"isDicksonDomain", "[", 
     RowBox[{"\[ScriptCapitalT]1", ",", 
      RowBox[{"(", 
       UnderscriptBox["\[VerticalBar]", "\[ScriptCapitalT]1"], ")"}]}], "]"}],
     "\[Implies]", 
    RowBox[{"isAlgoReductionRing", "[", 
     RowBox[{"PolyTuples", "[", 
      RowBox[{"\[ScriptCapitalR]1", ",", "\[ScriptCapitalT]1"}], "]"}], 
     "]"}]}], 
   TagBox[")",
    "AutoParentheses"]}]}]], "FormalTextInputFormula",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      RowBox[{
        StyleBox["(PolyTuples isAlgoReductionRing)", "FrameLabel"], "  ", 
        ButtonBox[
        "\[Times]", Evaluator -> Automatic, Appearance -> None, 
         ButtonFunction :> 
         Theorema`Language`Session`Private`removeFormula[{
           "ID:67407018", 
            "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"}]]}]]]}, {None, None}},
 ShowCellTags->False,
 CellChangeTimes->{{3.6563080524715033`*^9, 3.656308094669577*^9}, {
   3.656413005143835*^9, 3.656413016812656*^9}, {3.6564131069506145`*^9, 
   3.6564131135182257`*^9}, 3.65641318985116*^9},
 CellTags->{
  "ID:67407018", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isAlgoReductionRing"},
 CellID->67407018],

Cell["\[GraySquare]", "EndEnvironmentMarker",
 CellID->59995559]
}, Open  ]],

Cell["", "CloseEnvironment",
 CellID->628697588],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  StyleBox["\<\"\[CheckmarkedBox]\"\>",
   StripOnInput->False,
   ShowStringCharacters->False]]],
 " Proof of ",
 Cell[BoxData[
  ButtonBox[
   TagBox[
    TooltipBox[
     PaneSelectorBox[{False->
      StyleBox["\<\"(PolyTuples isAlgoReductionRing)\"\>", "FormReference",
       StripOnInput->False], True->
      StyleBox["\<\"(PolyTuples isAlgoReductionRing)\"\>", 
       "FormReferenceHover",
       StripOnInput->False]}, Dynamic[
       CurrentValue["MouseOver"]],
      FrameMargins->0,
      ImageSize->Automatic],
     TagBox[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]1", "ExpressionVariable"], "]"}]}, {
            
             RowBox[{"isCommPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]1", "ExpressionVariable"], 
               "]"}]}}]], 
         RowBox[{
           RowBox[{"isDicksonDomain", "[", 
             RowBox[{
               StyleBox["\[ScriptCapitalT]1", "ExpressionVariable"], ",", 
               UnderscriptBox["\[VerticalBar]", 
                StyleBox["\[ScriptCapitalT]1", "ExpressionVariable"]]}], 
             "]"}], "\[Implies]", 
           RowBox[{"isAlgoReductionRing", "[", 
             RowBox[{"PolyTuples", "[", 
               RowBox[{
                 StyleBox["\[ScriptCapitalR]1", "ExpressionVariable"], ",", 
                 StyleBox["\[ScriptCapitalT]1", "ExpressionVariable"]}], 
               "]"}], "]"}]}]}], $CellContext`TheoremaForm], DisplayForm]],
    Annotation[#, 
     DisplayForm[
      FormBox[
       RowBox[{
         UnderscriptBox["\[ForAll]", 
          GridBox[{{
             RowBox[{"isAlgoReductionRing", "[", 
               StyleBox["\[ScriptCapitalR]1", "ExpressionVariable"], "]"}]}, {
            
             RowBox[{"isCommPPDomain", "[", 
               StyleBox["\[ScriptCapitalT]1", "ExpressionVariable"], 
               "]"}]}}]], 
         RowBox[{
           RowBox[{"isDicksonDomain", "[", 
             RowBox[{
               StyleBox["\[ScriptCapitalT]1", "ExpressionVariable"], ",", 
               UnderscriptBox["\[VerticalBar]", 
                StyleBox["\[ScriptCapitalT]1", "ExpressionVariable"]]}], 
             "]"}], "\[Implies]", 
           RowBox[{"isAlgoReductionRing", "[", 
             RowBox[{"PolyTuples", "[", 
               RowBox[{
                 StyleBox["\[ScriptCapitalR]1", "ExpressionVariable"], ",", 
                 StyleBox["\[ScriptCapitalT]1", "ExpressionVariable"]}], 
               "]"}], "]"}]}]}], $CellContext`TheoremaForm]], "Tooltip"]& ],
   Appearance->None,
   ButtonFunction:>NotebookLocate["ID:67407018"],
   Evaluator->Automatic,
   Method->"Preemptive"]]],
 " #1:   ",
 Cell[BoxData[
  ButtonBox[
   StyleBox["\<\"Show proof\"\>",
    StripOnInput->False,
    FontVariations->{"Underline"->True}],
   Appearance->None,
   ButtonFunction:>Theorema`Common`displayProof[
     ToExpression[
     "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \"p67407018-1-po.m\"}]"]],
   Evaluator->Automatic,
   ImageSize->Automatic,
   Method->"Queued"]]]
}], "ProofInfo",
 CellFrameLabels->{{None, 
    Cell[
     BoxData[
      ButtonBox[
      "\[Times]", Evaluator -> Automatic, Appearance -> None, ButtonFunction :> 
       Theorema`Interface`GUI`Private`removeGroup[
         ButtonNotebook[], "p67407018-1"]]]]}, {None, None}},
 CellTags->{"Proof|ID:67407018", "Proof|ID:67407018-1"},
 CellID->163016488],

Cell[BoxData[
 DynamicBox[ToBoxes[
   Refresh[
    ReplaceAll[
     Get[
      ToExpression[
      "FileNameJoin[{CurrentValue[\"NotebookDirectory\"], \
FileBaseName[CurrentValue[\"NotebookFileName\"]], \
\"p67407018-1-display.m\"}]"]], Theorema`Interface`GUI`Private`FORM -> 
     Theorema`Interface`GUI`Private`displayFormulaFromKey], 
    TrackedSymbols :> {Theorema`Common`$tmaEnv}], StandardForm],
  ImageSizeCache->{516., {31., 37.}}]], "ProofInfoBody",
 CellID->310436971]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]]
}, Open  ]]
},
WindowSize->{1107, 757},
WindowMargins->{{1, Automatic}, {Automatic, 0}},
ShowSelection->True,
Magnification->1.,
FrontEndVersion->"9.0 for Linux x86 (64-bit) (February 7, 2013)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Theorema"}, "TheoremaCore.nb"]]], 
   
   Cell["Modify the definitions below to change the default appearance of all \
cells in a given style. \nMake modifications to any definition using commands \
in the Format menu.", "Text"], 
   Cell[
    CellGroupData[{
      Cell["Notebook Options Settings", "Section"], 
      Cell[
      "The options defined for the style below will be used at the Notebook \
level.", "Text"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Notebook"], InputAliases -> {"and" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[And]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "or" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", "\[Or]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "not" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Not]", "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "impl" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[Implies]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[DoubleLongLeftRightArrow]", 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               
               RowBox[{
                "\[SelectionPlaceholder]", "\[Equal]", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":equiv" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox[
                  
                  RowBox[{
                   ":", "\[NegativeThickSpace]\[NegativeThinSpace]", 
                    "\[DoubleLongLeftRightArrow]"}], Identity, SyntaxForm -> 
                  "a:=b"], "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], ":eq" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", ":=", "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "andn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[And]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "orn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Or]", "\[Piecewise]", 
                 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}, 
                  ColumnAlignments -> Left]}], 
               TagBox[")", "AutoParentheses"]}], "equivn" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[DoubleLeftRightArrow]", "\[Piecewise]", 
                 GridBox[{{"\[SelectionPlaceholder]"}, {"\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "cdist" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[Piecewise]", 
                 
                 GridBox[{{
                   "\[SelectionPlaceholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}, {
                   "\[Placeholder]", "\[DoubleLeftArrow]", 
                    "\[Placeholder]"}}]}], 
               TagBox[")", "AutoParentheses"]}], "far" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exr" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "farc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[ForAll]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "exrc" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox[
                  UnderscriptBox["\[Exists]", "\[Placeholder]"], 
                  "\[Placeholder]"], "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "let" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 UnderscriptBox["let", 
                  RowBox[{"\[Placeholder]", "=", "\[Placeholder]"}]], 
                 "\[SelectionPlaceholder]"}], 
               TagBox[")", "AutoParentheses"]}], "app" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293a", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "prep" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:293b", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "join" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22c8", Identity, SyntaxForm -> "a*b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "elemT" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{"\[SelectionPlaceholder]", 
                 TagBox["\:22ff", Identity, SyntaxForm -> "a+b"], 
                 "\[Placeholder]"}], 
               TagBox[")", "AutoParentheses"]}], "subop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "supop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SuperscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "subsupop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubsuperscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "oop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "uoop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               UnderoverscriptBox[
               "\[SelectionPlaceholder]", "\[Placeholder]", "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "osop" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               SubscriptBox[
                OverscriptBox["\[SelectionPlaceholder]", "\[Placeholder]"], 
                "\[Placeholder]"], 
               TagBox[")", "AutoParentheses"]}], "op" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               TagBox["\[SelectionPlaceholder]", Identity], 
               TagBox[")", "AutoParentheses"]}], "[]" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:e114", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:e115", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "()" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:fd3e", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:fd3f", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "<>" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27e8", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27e9", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "{}" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:e117", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:e118", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "[[]]" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27e6", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27e7", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "(())" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:2e28", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:2e29", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "<<>>" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:27ea", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:27eb", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "{{}}" -> RowBox[{
               TagBox["(", "AutoParentheses"], 
               RowBox[{
                 TagBox["\:2983", Identity, SyntaxForm -> "("], 
                 "\[SelectionPlaceholder]", 
                 TagBox["\:2984", Identity, SyntaxForm -> ")"]}], 
               TagBox[")", "AutoParentheses"]}], "(" -> 
            TagBox["(", "AutoParentheses"], ")" -> 
            TagBox[")", "AutoParentheses"]}, Background -> 
          RGBColor[0.957, 0.96, 0.97]], 
         Cell[
          StyleData["Notebook", "Printout"], Background -> None]}, Open]]}, 
     Open]], 
   Cell[
    CellGroupData[{
      Cell["Style Environment Names", "Section"], 
      Cell[
       StyleData[All, "Working"], DockedCells -> {
         Cell[
         "Theorema 2.0", "SBO", FontTracking -> 8, 
          CellMargins -> {{0, 0}, {0, 0}}, 
          CellFrameMargins -> {{0, 50}, {0, 0}}, FontSize -> 28, FontSlant -> 
          "Plain", FontColor -> RGBColor[0.29, 0.46, 0.6], TextAlignment -> 
          Right, Background -> RGBColor[0.36, 0.54, 0.69]]}], 
      Cell[
       StyleData[
       All, "Layout", StyleDefinitions -> StyleData[All, "Working"]], 
       MenuSortingValue -> 1200]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Styles for Title and Section Cells", "Section"], 
      Cell[
       CellGroupData[{
         Cell["Title", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], CellMargins -> {{27, 27}, {10, 30}}, 
             TextAlignment -> Center, MenuSortingValue -> 1100, 
             MenuCommandKey -> "1", FontFamily -> "Bitstream Charter", 
             FontSize -> 36, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Title", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{6, 10}, {5, 30}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], CellMargins -> {{27, 27}, {20, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1150, FontFamily -> 
             "Bitstream Charter", FontSize -> 24, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{6, 10}, {5, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], CellMargins -> {{27, 27}, {8, 2}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Author"], CellMargins -> {{27, 27}, {18, 20}}, 
             TextAlignment -> Center, MenuSortingValue -> 1200, FontFamily -> 
             "Bitstream Charter", FontSize -> 16, FontWeight -> "Plain", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.23, 0.25, 0.34]], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{6, 10}, {0, 0}}]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Sectioning", "Subsection"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], ShowGroupOpener -> True, MenuSortingValue -> 
             1250, MenuCommandKey -> "4", FontFamily -> "Bitstream Charter", 
             FontSize -> 20, FontWeight -> "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Section", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{2, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {8, 12}}, 
             MenuSortingValue -> 1300, MenuCommandKey -> "5", FontFamily -> 
             "Bitstream Charter", FontSize -> 14, FontWeight -> "Bold", 
             FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.3, 0.3, 0.3]], 
            Cell[
             StyleData["Subsection", "Printout"], ShowGroupOpener -> False, 
             CellMargins -> {{9, 0}, {7, 22}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> None, ShowGroupOpener -> 
             True, CellMargins -> {{27, Inherited}, {2, 10}}, 
             MenuSortingValue -> 1350, MenuCommandKey -> "6", FontFamily -> 
             "Bitstream Charter", FontSize -> 12, FontWeight -> "Bold", 
             FontSlant -> "Italic", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsection", "Printout"], ShowGroupOpener -> False,
              CellMargins -> {{35, 0}, {7, 14}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsubsection"], CellDingbat -> None, 
             ShowGroupOpener -> True, 
             CellMargins -> {{27, Inherited}, {Inherited, Inherited}}, 
             FontFamily -> "Bitstream Charter", FontSize -> 11, FontWeight -> 
             "Bold", FontSlant -> "Plain", 
             FontVariations -> {
              "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
             RGBColor[0.4, 0.4, 0.4]], 
            Cell[
             StyleData["Subsubsubsection", "Printout"], 
             CellMargins -> {{50, 0}, {7, 14}}]}, Closed]]}, Open]]}, Open]], 
   
   Cell[
    CellGroupData[{
      Cell["Styles for Publication\[Hyphen]ready Documents", "Section"], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Author"], CellMargins -> {{27, 30}, {20, 5}}, 
          MenuSortingValue -> 1900, FontFamily -> "Helvetica", FontSize -> 14,
           FontSlant -> "Italic"], 
         Cell[
          StyleData["Author", "Printout"], CellMargins -> {{6, 10}, {0, 10}}, 
          FontSize -> 10]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Abstract"], CellFrame -> {{0, 0}, {0.5, 0.5}}, 
          CellMargins -> {{27, 140}, {20, 20}}, LineSpacing -> {0.9, 0}, 
          MenuSortingValue -> 2050, FontFamily -> "Times", FontSize -> 12], 
         Cell[
          StyleData["Abstract", "Printout"], Hyphenation -> True, 
          LineSpacing -> {1, 2}, FontSize -> 10]}, Closed]]}, Closed]], 
   Cell[
    CellGroupData[{
      Cell["Text", "Section"], 
      Cell[
       StyleData["Text"], CellMargins -> {{27, 27}, {Inherited, Inherited}}, 
       TextJustification -> 1., ParagraphIndent -> 15, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["EnvironmentText"], ShowCellBracket -> Automatic, 
       CellMargins -> {{82, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextJustification -> 1., ParagraphSpacing -> {0.3, 0.}, 
       MenuSortingValue -> 1480, MenuCommandKey -> "T", FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Item"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "Item", 
          CounterAssignments -> {{"Subitem", 0}, {"Subsubitem", 0}}, 
          MenuSortingValue -> 1600, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Item", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemParagraph"], CellMargins -> {{81, 10}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          DefaultNewCellStyle -> "Item", MenuSortingValue -> 1600, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemParagraph", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "Subitem", 
          CounterAssignments -> {{"Subsubitem", 0}}, MenuSortingValue -> 1610,
           FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subitem", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemParagraph"], CellMargins -> {{105, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subitem", 
          MenuSortingValue -> 1610, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemParagraph", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Subsubitem"], CellDingbat -> 
          Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1620, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["Subsubitem", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemParagraph"], 
          CellMargins -> {{129, 12}, {4, 1}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
          CounterIncrements -> "Subsubitem", MenuSortingValue -> 1625, 
          FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemParagraph", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 2}}, FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["ItemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{81, 10}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
          CellFrameLabelMargins -> 4, CounterIncrements -> "ItemNumbered", 
          CounterAssignments -> {{"SubitemNumbered", 0}, {
            "SubsubitemNumbered", 0}}, MenuSortingValue -> 1630, FontFamily -> 
          "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", 
          FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["ItemNumbered", "Presentation"], 
          CellMargins -> {{140, 10}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{105, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", 
          CounterIncrements -> "SubitemNumbered", 
          CounterAssignments -> {{"SubsubitemNumbered", 0}}, MenuSortingValue -> 
          1640, FontFamily -> "Bitstream Charter", FontSize -> 12, FontWeight -> 
          "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubitemNumbered", "Presentation"], 
          CellMargins -> {{176, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["SubsubitemNumbered"], CellDingbat -> Cell[
            TextData[{
              CounterBox["ItemNumbered"], ".", 
              CounterBox["SubitemNumbered"], ".", 
              CounterBox["SubsubitemNumbered"], "."}], FontWeight -> "Bold"], 
          CellMargins -> {{129, 12}, {4, 4}}, 
          CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
          CellFrameLabelMargins -> 4, DefaultNewCellStyle -> 
          "SubitemNumbered", CounterIncrements -> "SubsubitemNumbered", 
          MenuSortingValue -> 1645, FontFamily -> "Bitstream Charter", 
          FontSize -> 12, FontWeight -> "Plain", FontSlant -> "Plain", 
          FontVariations -> {
           "StrikeThrough" -> False, "Underline" -> 
            False}, $CellContext`ReturnCreatesNewCell -> True], 
         Cell[
          StyleData["SubsubitemNumbered", "Presentation"], 
          CellMargins -> {{212, 12}, {7, 7}}, CellFrameLabelMargins -> 6, 
          FontSize -> 24]}, Open]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell["Theorema", "Section"], 
      Cell[
       StyleData["GlobalDeclaration"], CellFrame -> {{0, 2}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 3}, {2, 2}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       InitializationCell -> True, CellFrameMargins -> 10, CellFrameColor -> 
       RGBColor[0.73, 0.65, 0.61], CellFrameLabelMargins -> {{33, 7}, {1, 1}},
        ShowAutoStyles -> False, MenuSortingValue -> 1441, MenuCommandKey -> 
       "G", FontSize -> 14], 
      Cell[
       StyleData["OpenEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {-4, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       StyleData["EnvironmentHeader"], 
       CellFrame -> {{False, False}, {False, False}}, 
       CellMargins -> {{76, Inherited}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 6, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       ShowAutoStyles -> False, MenuSortingValue -> 1450, FontSize -> 14, 
       FontWeight -> "Bold", FontVariations -> {"CapsType" -> "SmallCaps"}, 
       Background -> RGBColor[0.89, 0.8, 0.69]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FormalTextInputFormula"], ShowCellBracket -> Automatic, 
          CellMargins -> {{96, 10}, {2, 2}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
          InitializationCell -> True, CellFrameLabelMargins -> 53, 
          ShowAutoStyles -> False, LimitsPositioningTokens -> {}, 
          MenuSortingValue -> 1470, MenuCommandKey -> "F", FontWeight -> 
          "Bold", FontColor -> RGBColor[0.25, 0.19, 0.2]], 
         Cell[
          StyleData["FormalTextInputFormula", "Layout"], Editable -> False], 
         Cell[
          StyleData["FormalTextInputFormula", "Presentation"], Editable -> 
          False]}, Open]], 
      Cell[
       StyleData["EndEnvironmentMarker"], 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       TextAlignment -> Right, MenuSortingValue -> 1480], 
      Cell[
       StyleData["CloseEnvironment"], Editable -> False, Selectable -> False, 
       CellFrame -> {{False, False}, {False, True}}, ShowCellBracket -> True, 
       CellMargins -> {{76, 6}, {Inherited, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 2, CellFrameColor -> RGBColor[0.4, 0.4, 0.4], 
       MenuSortingValue -> 1485], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Computation"], CellFrame -> {{1.5, 0}, {0, 0}}, 
          ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 7}}, 
          CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
          CellFrameMargins -> {{8, 8}, {3, 4}}, CellFrameColor -> 
          RGBColor[0.486, 0.33, 0.32], ShowAutoStyles -> False, 
          MenuSortingValue -> 1510, FontWeight -> "Bold", FontColor -> 
          RGBColor[0.23, 0.25, 0.34], Background -> 
          RGBColor[0.88, 0.88, 0.92]], 
         Cell[
          StyleData["Computation", "Layout"], Editable -> False], 
         Cell[
          StyleData["Computation", "Presentation"], Editable -> False]}, 
        Open]], 
      Cell[
       StyleData["ComputationResult"], CellFrame -> {{1.5, 0}, {0, 0}}, 
       ShowCellBracket -> Automatic, CellMargins -> {{76, 10}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.23, 0.25, 0.34]}, 
       CellFrameMargins -> {{8, 8}, {4, 3}}, CellFrameColor -> 
       RGBColor[0.486, 0.33, 0.32], StyleMenuListing -> None, FontColor -> 
       RGBColor[0.23, 0.25, 0.34], Background -> RGBColor[0.88, 0.88, 0.92]], 
      
      Cell[
       StyleData["ComputationInfo"], ShowCellBracket -> False, 
       CellMargins -> {{120, 10}, {10, 2}}, StyleMenuListing -> None, 
       FontFamily -> $CellContext`Arial], 
      Cell[
       StyleData["ComputationInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{144, 10}, {10, 2}}, StyleMenuListing -> None], 
      Cell[
       StyleData["ProofInfo"], ShowCellBracket -> False, 
       CellMargins -> {{76, 10}, {6, 6}}, CellFrameLabelMargins -> 15, 
       StyleMenuListing -> None, FontFamily -> "Arial"], 
      Cell[
       StyleData["ProofInfoBody"], ShowCellBracket -> False, 
       CellMargins -> {{100, 10}, {0, 0}}, StyleMenuListing -> None], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["FrameLabel"], LanguageCategory -> None, StyleMenuListing -> 
          None, FontFamily -> "Times New Roman", FontSize -> 10, FontWeight -> 
          "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
          FontVariations -> {
           "Masked" -> False, "Outline" -> False, "Shadow" -> False, 
            "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
          RGBColor[0.55, 0.44, 0.42], Background -> None], 
         Cell[
          StyleData["FrameLabel", "Printout"]]}, Open]], 
      Cell[
       StyleData["OpenArchive"], CellFrame -> {{False, False}, {False, True}},
        ShowCellBracket -> Automatic, CellMargins -> {{70, 70}, {0, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 20}, CellFrameMargins -> 3, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], CellFrameLabelMargins -> 1, 
       DefaultFormatType -> DefaultInputFormatType, ShowAutoStyles -> False, 
       MenuSortingValue -> 1210, FontWeight -> Bold, Background -> 
       RGBColor[0.73, 0.74, 0.84]], 
      Cell[
       StyleData["ArchiveInfo"], ShowCellBracket -> Automatic, 
       CellMargins -> {{70, 70}, {0, 0}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameMargins -> 5, CellFrameLabelMargins -> 3, ShowCellTags -> 
       False, ShowAutoStyles -> False, MenuSortingValue -> 1220, FontWeight -> 
       "Bold", Background -> RGBColor[0.88, 0.88, 0.92]], 
      Cell[
       StyleData["CloseArchive"], 
       CellFrame -> {{False, False}, {True, False}}, ShowCellBracket -> 
       Automatic, CellMargins -> {{0, 0}, {5, 10}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellGroupingRules -> {"SectionGrouping", 25}, CellFrameMargins -> 2, 
       CellFrameColor -> RGBColor[0.4, 0.4, 0.4], TextAlignment -> Center, 
       MenuSortingValue -> 1230, FontFamily -> "Arial Black", FontSize -> 24, 
       FontWeight -> "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       Background -> RGBColor[0.73, 0.65, 0.61]], 
      Cell[
       StyleData["IncludeArchive"], CellFrame -> {{4, 0}, {False, False}}, 
       ShowCellBracket -> Automatic, 
       CellMargins -> {{81, 100}, {Inherited, Inherited}}, 
       CellBracketOptions -> {"Color" -> RGBColor[0.73, 0.74, 0.84]}, 
       CellFrameColor -> RGBColor[0.55, 0.44, 0.42], 
       CellFrameLabels -> {{"\[RightSkeleton]", None}, {None, None}}, 
       CellFrameLabelMargins -> 25, MenuSortingValue -> 1445], 
      Cell[
       StyleData["ExpressionVariable"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionABF"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       StyleData["ExpressionMeta"], SingleLetterItalics -> True, 
       MultiLetterItalics -> True, FontColor -> RGBColor[0.29, 0.46, 0.6]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["AutoParentheses"], Editable -> False, SpanMaxSize -> 3, 
          FontOpacity -> 0.15], 
         Cell[
          StyleData["AutoParentheses", "Layout"], FontOpacity -> 0], 
         Cell[
          StyleData["AutoParentheses", "SlideShow"], FontOpacity -> 0]}, 
        Open]], 
      Cell[
       CellGroupData[{
         Cell[
          StyleData["Emphasized"], MenuCommandKey -> "E", FontSlant -> 
          "Italic"], 
         Cell[
          StyleData["Emphasized", "SlideShow"], FontSlant -> "Plain", 
          FontColor -> RGBColor[0.36, 0.54, 0.69]]}, Open]], 
      Cell[
       StyleData["GenericButton"], FontFamily -> "Bitstream Charter", 
       FontWeight -> "Plain", FontSlant -> "Italic", FontTracking -> "Plain", 
       FontVariations -> {
        "Masked" -> False, "Outline" -> False, "Shadow" -> False, 
         "StrikeThrough" -> False, "Underline" -> False}, FontColor -> 
       RGBColor[0.23, 0.25, 0.34]], 
      Cell[
       StyleData["TheoremaForm"], "TwoByteSyntaxCharacterAutoReplacement" -> 
       True, LineSpacing -> {1.25, 0, 2}, SingleLetterItalics -> True, 
       StyleMenuListing -> None, FontFamily -> "Courier"], 
      Cell[
       StyleData["ComponentActive"], FontFamily -> "Times New Roman"], 
      Cell[
       StyleData["ComponentInactive"], FontFamily -> "Times New Roman", 
       FontOpacity -> 0.4], 
      Cell[
       StyleData["ComponentEmpty"], FontFamily -> "Times New Roman", 
       FontVariations -> {"StrikeThrough" -> True}], 
      Cell[
       StyleData["TabView"], FontFamily -> "Times New Roman", FontSize -> 10], 
      Cell[
       StyleData["TabViewLabel"], FontFamily -> "Times New Roman", FontSize -> 
       10, FontWeight -> "Bold"], 
      Cell[
       StyleData["FormReference", StyleDefinitions -> StyleData["Text"]], 
       FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> "Plain",
        FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}], 
      Cell[
       StyleData["FormReferenceHover", StyleDefinitions -> StyleData["Text"]],
        FontFamily -> "Times New Roman", FontSize -> 12, FontWeight -> 
       "Plain", FontSlant -> "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> True}, 
       FontColor -> RGBColor[0.404, 0.43, 0.545]], 
      Cell[
       StyleData["DisplayFormula"], CellMargins -> {{36, 6}, {7, 7}}, 
       ConversionRules :> {
        "HTML" -> {{"", ""}, {"<p class=\"Equation\">", "</p>"}}}, 
       CellHorizontalScrolling -> True, DefaultFormatType -> TraditionalForm, 
       InputAutoReplacements -> {}, TextAlignment -> Center, FontFamily -> 
       "Bitstream Charter", FontSize -> 12, FontWeight -> "Plain", FontSlant -> 
       "Plain", 
       FontVariations -> {"StrikeThrough" -> False, "Underline" -> False}, 
       ButtonBoxOptions -> {Appearance -> {Automatic, None}}]}, Open]]}, 
  Visible -> False, FrontEndVersion -> 
  "9.0 for Linux x86 (64-bit) (February 7, 2013)", StyleDefinitions -> 
  "StylesheetFormatting.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "ID:303368290"->{
  Cell[44486, 1105, 665, 17, 31, "GlobalDeclaration",
   CellTags->{
    "ID:303368290", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->303368290]},
 "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"->{
  Cell[44486, 1105, 665, 17, 31, "GlobalDeclaration",
   CellTags->{
    "ID:303368290", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->303368290],
  Cell[45833, 1145, 2342, 60, 137, "FormalTextInputFormula",
   CellTags->{
    "ID:603497554", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain"},
   CellID->603497554],
  Cell[48178, 1207, 3150, 87, 135, "FormalTextInputFormula",
   CellTags->{
    "ID:49777475", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasDecisionPredicatePP"},
   CellID->49777475],
  Cell[51331, 1296, 5696, 147, 224, "FormalTextInputFormula",
   CellTags->{
    "ID:225295367", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasExponentFunctionPP"},
   CellID->225295367],
  Cell[57030, 1445, 2105, 64, 52, "FormalTextInputFormula",
   CellTags->{
    "ID:292978558", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExtensionalPP"},
   CellID->292978558],
  Cell[59138, 1511, 1950, 58, 63, "FormalTextInputFormula",
   CellTags->{
    "ID:340781435", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasDegreeFunctionPP"},
   CellID->340781435],
  Cell[61091, 1571, 2199, 68, 52, "FormalTextInputFormula",
   CellTags->{
    "ID:480617864", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasDivisibilityPP"},
   CellID->480617864],
  Cell[63961, 1666, 2958, 87, 115, "FormalTextInputFormula",
   CellTags->{
    "ID:897055662", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isTermOrder"},
   CellID->897055662],
  Cell[67823, 1788, 703, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:532395027", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->532395027],
  Cell[68529, 1808, 854, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:160997957", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain dim"},
   CellID->160997957],
  Cell[69386, 1832, 818, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:83558965", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasDecisionPredicate"},
   CellID->83558965],
  Cell[70207, 1856, 811, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:349898371", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasExponentFunction"},
   CellID->349898371],
  Cell[71021, 1879, 790, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:200216737", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isExtensional"},
   CellID->200216737],
  Cell[71814, 1902, 803, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:276506833", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasDegreeFunction"},
   CellID->276506833],
  Cell[72620, 1925, 821, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:135986343", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasDivisibility"},
   CellID->135986343],
  Cell[73444, 1948, 1254, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:122009120", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership exp"},
   CellID->122009120],
  Cell[96528, 2645, 703, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:185429719", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->185429719],
  Cell[97763, 2682, 794, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:81626705", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isDomain"},
   CellID->81626705],
  Cell[98560, 2705, 1095, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:154517895", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership \[RawStar]", "IR"},
   CellID->154517895],
  Cell[99658, 2738, 993, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:226316196", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain \[RawStar] isAssoc"},
   CellID->226316196],
  Cell[100654, 2770, 960, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:96938761", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain \[RawStar] isComm"},
   CellID->96938761],
  Cell[101617, 2802, 1024, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:499736780", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain 1 isNeutral"},
   CellID->499736780],
  Cell[102644, 2835, 1458, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:745587327", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain \[RawStar] lcanc"},
   CellID->745587327],
  Cell[126159, 3521, 804, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:333683969", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isCCMonoid"},
   CellID->333683969],
  Cell[130592, 3660, 700, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:274960855", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->274960855],
  Cell[131295, 3680, 1031, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:52136714", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain lcm isJoin"},
   CellID->52136714],
  Cell[132329, 3713, 760, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:35376325", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->35376325],
  Cell[133092, 3735, 1279, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:138508454", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain div.1"},
   CellID->138508454],
  Cell[134374, 3776, 1437, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:14785826", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain div.2"},
   CellID->14785826],
  Cell[150172, 4222, 703, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:181034073", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->181034073],
  Cell[150878, 4242, 738, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:943617998", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->943617998],
  Cell[151619, 4263, 1424, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:706945753", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership deg ind"},
   CellID->706945753],
  Cell[153046, 4303, 2172, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:606418540", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain deg=0 ind"},
   CellID->606418540],
  Cell[155221, 4374, 951, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:47160", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership deg", "IR"},
   CellID->47160],
  Cell[156175, 4401, 1158, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:58897330", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain deg=0"},
   CellID->58897330],
  Cell[157336, 4439, 1333, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:170894204", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain deg \[RawStar]"},
   CellID->170894204],
  Cell[183800, 5175, 1710, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:49640627", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "increasing subsequence \[DoubleStruckCapitalN]0"},
   CellID->49640627],
  Cell[191307, 5382, 701, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:256365628", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->256365628],
  Cell[192485, 5418, 1782, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:443096051", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain divN"},
   CellID->443096051],
  Cell[194868, 5497, 2207, 64, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:27206868", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain divN +1"},
   CellID->27206868],
  Cell[197078, 5563, 1809, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:184673495", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "increasing subsequence isExpPPDomain 0"},
   CellID->184673495],
  Cell[198890, 5616, 1975, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:43287351", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "increasing subsequence isExpPPDomain"},
   CellID->43287351],
  Cell[221550, 6213, 1022, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:30919390", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isDicksonDomain"},
   CellID->30919390],
  Cell[226555, 6367, 781, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:323333890", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->323333890],
  Cell[227339, 6389, 771, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:235080407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->235080407],
  Cell[228113, 6411, 1338, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:691531428", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "term order |"},
   CellID->691531428],
  Cell[229454, 6456, 989, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:772895867", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "term order isOrderEmbedding"},
   CellID->772895867],
  Cell[230446, 6486, 819, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:83375799", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "term order isNoetherian"},
   CellID->83375799],
  Cell[245321, 6906, 680, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:69380281", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->69380281],
  Cell[246004, 6925, 723, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:70274751", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->70274751],
  Cell[246730, 6946, 669, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:417719229", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->417719229],
  Cell[247402, 6965, 1279, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:296713523", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP \[Element]"},
   CellID->296713523],
  Cell[248684, 7009, 818, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:275450356", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP dim"},
   CellID->275450356],
  Cell[249505, 7033, 1082, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:126765714", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP exp"},
   CellID->126765714],
  Cell[250590, 7068, 810, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:300238463", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP 1"},
   CellID->300238463],
  Cell[251403, 7092, 863, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:91738893", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP \[RawStar]"},
   CellID->91738893],
  Cell[252269, 7119, 867, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1017172830", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP /"},
   CellID->1017172830],
  Cell[253139, 7146, 923, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:62103041", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP lcm"},
   CellID->62103041],
  Cell[254065, 7174, 1335, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:229667832", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP |"},
   CellID->229667832],
  Cell[255403, 7218, 935, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:687245437", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP deg"},
   CellID->687245437],
  Cell[256341, 7246, 1143, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:353388395", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP \[Succeeds]"},
   CellID->353388395],
  Cell[257487, 7283, 1096, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:144609801", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP <<"},
   CellID->144609801],
  Cell[258586, 7321, 1096, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:189296520", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP >>"},
   CellID->189296520],
  Cell[260643, 7395, 776, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:29953053", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->29953053],
  Cell[261422, 7417, 869, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:309264225", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isExpPPDomain"},
   CellID->309264225],
  Cell[262294, 7442, 1196, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:499712177", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isDicksonDomain"},
   CellID->499712177],
  Cell[263493, 7478, 720, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:275069792", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->275069792],
  Cell[264216, 7499, 886, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:834856233", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isTermOrder"},
   CellID->834856233],
  Cell[265105, 7525, 1019, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:260893370", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isCommPPDomain"},
   CellID->260893370],
  Cell[281789, 8011, 2379, 71, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:223902051", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex"},
   CellID->223902051],
  Cell[284659, 8102, 703, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:376155678", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->376155678],
  Cell[285922, 8140, 773, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:310375877", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->310375877],
  Cell[286698, 8162, 989, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:463764498", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex total"},
   CellID->463764498],
  Cell[287690, 8191, 1249, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:6465370", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex asymmetric"},
   CellID->6465370],
  Cell[288942, 8232, 1188, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:78173747", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex transitive"},
   CellID->78173747],
  Cell[290133, 8269, 971, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:341999039", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex 1 min"},
   CellID->341999039],
  Cell[291107, 8298, 1079, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:34516336", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex \[RawStar]"},
   CellID->34516336],
  Cell[318896, 9057, 861, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:414199224", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex isTermOrder"},
   CellID->414199224],
  Cell[323832, 9213, 2280, 72, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:116320630", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex"},
   CellID->116320630],
  Cell[326511, 9304, 703, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:106786369", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->106786369],
  Cell[327729, 9342, 767, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:43019543", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->43019543],
  Cell[328499, 9364, 998, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:75557233", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex total"},
   CellID->75557233],
  Cell[329500, 9393, 1267, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:49531328", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex asymmetric"},
   CellID->49531328],
  Cell[330770, 9434, 1208, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:260398187", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex transitive"},
   CellID->260398187],
  Cell[331981, 9471, 977, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:59133109", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex 1 min"},
   CellID->59133109],
  Cell[332961, 9500, 1096, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:572110615", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex \[RawStar]"},
   CellID->572110615],
  Cell[360909, 10260, 869, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:463127740", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex isTermOrder"},
   CellID->463127740],
  Cell[365945, 10421, 693, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:267616510", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->267616510],
  Cell[366641, 10440, 2155, 64, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1929370", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex"},
   CellID->1929370],
  Cell[368799, 10506, 1994, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:116917210", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex"},
   CellID->116917210],
  Cell[371200, 10587, 701, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:431887307", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->431887307],
  Cell[372396, 10624, 769, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:204943512", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->204943512],
  Cell[373168, 10646, 1001, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:781115991", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex total"},
   CellID->781115991],
  Cell[374172, 10675, 1264, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2983958", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex asymmetric"},
   CellID->2983958],
  Cell[375439, 10716, 1208, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:240564679", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex transitive"},
   CellID->240564679],
  Cell[376650, 10753, 1092, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:299677825", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex \[RawStar]"},
   CellID->299677825],
  Cell[400183, 11392, 773, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:449289304", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->449289304],
  Cell[400959, 11414, 1042, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:240154468", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex total"},
   CellID->240154468],
  Cell[402004, 11444, 1311, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:808110452", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex asymmetric"},
   CellID->808110452],
  Cell[403318, 11486, 1250, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:224322863", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex transitive"},
   CellID->224322863],
  Cell[404571, 11524, 991, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:424804407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex 1 min"},
   CellID->424804407],
  Cell[405565, 11553, 1108, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:308108885", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex \[RawStar]"},
   CellID->308108885],
  Cell[433321, 12308, 878, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:936963717", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex isTermOrder"},
   CellID->936963717],
  Cell[437966, 12452, 716, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:44994065", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->44994065],
  Cell[439493, 12499, 758, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1997108144", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->1997108144],
  Cell[440254, 12520, 926, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:608617422", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->608617422],
  Cell[441183, 12544, 1676, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:271011394", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples \[Element]"},
   CellID->271011394],
  Cell[442862, 12595, 953, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1820126795", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples 1"},
   CellID->1820126795],
  Cell[443818, 12624, 774, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:281355476", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples dim"},
   CellID->281355476],
  Cell[444595, 12647, 1083, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2109519396", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples deg"},
   CellID->2109519396],
  Cell[445681, 12680, 901, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1563833422", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples exp"},
   CellID->1563833422],
  Cell[446585, 12707, 1052, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1180963317", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples \[RawStar]"},
   CellID->1180963317],
  Cell[447640, 12740, 1050, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1392702003", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples /"},
   CellID->1392702003],
  Cell[448693, 12773, 1292, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1587875066", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples \[VerticalBar]"},
   CellID->1587875066],
  Cell[449988, 12815, 1418, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:803590963", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples lcm"},
   CellID->803590963],
  Cell[451717, 12863, 1127, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:106372713", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples <<"},
   CellID->106372713],
  Cell[452847, 12896, 1037, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1589980318", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples >>"},
   CellID->1589980318],
  Cell[454270, 12945, 712, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:61015536", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->61015536],
  Cell[455471, 12982, 1647, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:617800164", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PPTuples alternative", "IR"},
   CellID->617800164],
  Cell[457121, 13034, 787, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:41499157", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples isDomain"},
   CellID->41499157],
  Cell[457911, 13057, 830, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:142297858", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasDecisionPredicate"},
   CellID->142297858],
  Cell[458744, 13080, 823, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:201916088", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasExponentFunction"},
   CellID->201916088],
  Cell[459570, 13103, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:102507049", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples isExtensional"},
   CellID->102507049],
  Cell[460380, 13126, 818, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:93844102", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasDegreeFunction"},
   CellID->93844102],
  Cell[461201, 13149, 860, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:34803991", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasDivisibility"},
   CellID->34803991],
  Cell[484890, 13859, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:338277855", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples isExpPPDomain"},
   CellID->338277855],
  Cell[495075, 14188, 873, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:611122547", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->611122547],
  Cell[495951, 14211, 1008, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1216459107", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->1216459107],
  Cell[496962, 14236, 1327, 29, 70, "GlobalDeclaration",
   CellTags->{
    "ID:594519699", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->594519699],
  Cell[498292, 14267, 3917, 107, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:65019775", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Element]"},
   CellID->65019775],
  Cell[502212, 14376, 938, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:399496924", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C base"},
   CellID->399496924],
  Cell[503153, 14404, 1544, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:983317445", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C rec"},
   CellID->983317445],
  Cell[504700, 14451, 859, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1360683691", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples 0"},
   CellID->1360683691],
  Cell[505562, 14478, 1594, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:13289303", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples 1"},
   CellID->13289303],
  Cell[507159, 14520, 982, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1777247200", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + 1"},
   CellID->1777247200],
  Cell[508144, 14548, 1399, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2122267989", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + 2"},
   CellID->2122267989],
  Cell[509546, 14586, 4077, 109, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:290578418", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + 3"},
   CellID->290578418],
  Cell[513626, 14697, 974, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1495569651", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - unary 1"},
   CellID->1495569651],
  Cell[514603, 14726, 1529, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1285234307", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - unary 2"},
   CellID->1285234307],
  Cell[516135, 14772, 915, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:551740538", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples -"},
   CellID->551740538],
  Cell[517053, 14801, 941, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:419928300", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[RawStar] 1"},
   CellID->419928300],
  Cell[517997, 14830, 1168, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:209271439", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[RawStar] 2"},
   CellID->209271439],
  Cell[519168, 14865, 3884, 99, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1411844363", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[RawStar] 3"},
   CellID->1411844363],
  Cell[523055, 14966, 1192, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1128062400", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Succeeds] 1"},
   CellID->1128062400],
  Cell[524250, 15004, 1420, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1280515048", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Succeeds] 2"},
   CellID->1280515048],
  Cell[525673, 15047, 2480, 72, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2095908261", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Succeeds] 3"},
   CellID->2095908261],
  Cell[528156, 15121, 977, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1668822406", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples multN 1"},
   CellID->1668822406],
  Cell[529136, 15149, 1270, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:329725901", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples multN 2"},
   CellID->329725901],
  Cell[530409, 15184, 1587, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:822409829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples mult"},
   CellID->822409829],
  Cell[531999, 15232, 1828, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:922404041", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples mult i"},
   CellID->922404041],
  Cell[533830, 15285, 922, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2033419264", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples rdm 1"},
   CellID->2033419264],
  Cell[534755, 15313, 1149, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1308197671", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples rdm 2"},
   CellID->1308197671],
  Cell[535907, 15346, 3023, 78, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1157498137", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples rdm 3"},
   CellID->1157498137],
  Cell[538933, 15426, 984, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:103388547", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd 1"},
   CellID->103388547],
  Cell[539920, 15456, 1212, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:166800573", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd 2"},
   CellID->166800573],
  Cell[541135, 15492, 4959, 122, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:319850711", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd 3"},
   CellID->319850711],
  Cell[546097, 15616, 942, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:112650434", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd unary 1"},
   CellID->112650434],
  Cell[547042, 15644, 1695, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1729493745", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd unary 2"},
   CellID->1729493745],
  Cell[548740, 15691, 1071, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:414290689", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples cpm 1"},
   CellID->414290689],
  Cell[549814, 15723, 1296, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:27385995", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples cpm 2"},
   CellID->27385995],
  Cell[551113, 15761, 4968, 126, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1444186163", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples cpm 3"},
   CellID->1444186163],
  Cell[556084, 15889, 1841, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2018930893", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples <<"},
   CellID->2018930893],
  Cell[557928, 15940, 1503, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1442847663", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples >>"},
   CellID->1442847663],
  Cell[560532, 16023, 1083, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1486930793", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalT]2"},
   CellID->1486930793],
  Cell[561618, 16053, 1080, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:933000257", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalT]3"},
   CellID->933000257],
  Cell[574360, 16449, 1030, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1375788385", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalQ]xyz"},
   CellID->1375788385],
  Cell[614870, 17740, 777, 20, 32, "GlobalDeclaration",
   CellTags->{
    "ID:395821466", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->395821466],
  Cell[616582, 17794, 923, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:689486298", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isTuple", "IR"},
   CellID->689486298],
  Cell[617508, 17821, 1507, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:302371753", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples DomainTuples"},
   CellID->302371753],
  Cell[619018, 17871, 912, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:619631752", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples empty", "IR"},
   CellID->619631752],
  Cell[619933, 17897, 3154, 93, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:93154383", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples rec", "IR"},
   CellID->93154383],
  Cell[623090, 17992, 3179, 91, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:110840922", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples rec II", "IR"},
   CellID->110840922],
  Cell[626272, 18085, 3467, 100, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:45787564", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples rec III", "IR"},
   CellID->45787564],
  Cell[629742, 18187, 707, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:34480440", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->34480440],
  Cell[630452, 18208, 2714, 80, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:134755829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples prepend", "IR"},
   CellID->134755829],
  Cell[633169, 18290, 1120, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:218126712", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership first", "IR"},
   CellID->218126712],
  Cell[634292, 18323, 1117, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:36095185", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership second", "IR"},
   CellID->36095185],
  Cell[635412, 18356, 3357, 89, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:266402037", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples shape"},
   CellID->266402037],
  Cell[638772, 18447, 1900, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:314594392", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples length=1"},
   CellID->314594392],
  Cell[702890, 20161, 642, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:441087600", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->441087600],
  Cell[703535, 20179, 1924, 61, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:678316554", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[ForAll] alternative"},
   CellID->678316554],
  Cell[705462, 20242, 3708, 107, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:329976481", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples induction", "IR"},
   CellID->329976481],
  Cell[709772, 20370, 4150, 117, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:379661071", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples induction'", "IR"},
   CellID->379661071],
  Cell[741770, 21181, 1079, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:485298517", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isTotalIrreflOrder \[Succeeds] \[ScriptCapitalT]"},
   CellID->485298517],
  Cell[742852, 21215, 734, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:561716382", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->561716382],
  Cell[743589, 21236, 1743, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:107805279", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp decreasing ind"},
   CellID->107805279],
  Cell[745335, 21288, 1318, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:220534381", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp decreasing"},
   CellID->220534381],
  Cell[746656, 21329, 1310, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:5709061", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max"},
   CellID->5709061],
  Cell[747969, 21370, 1292, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:133904640", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max'"},
   CellID->133904640],
  Cell[749264, 21410, 1277, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:12736634", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp distinct"},
   CellID->12736634],
  Cell[779169, 22259, 700, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:723780914", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->723780914],
  Cell[779872, 22279, 936, 25, 70, "GlobalDeclaration",
   CellTags->{
    "ID:284026551", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->284026551],
  Cell[780811, 22306, 864, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:548327311", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->548327311],
  Cell[781678, 22330, 1040, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:387361496", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max seq"},
   CellID->387361496],
  Cell[782721, 22360, 1116, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:70991441", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max seq'"},
   CellID->70991441],
  Cell[783840, 22394, 948, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:365249767", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp distinct seq"},
   CellID->365249767],
  Cell[804351, 22948, 786, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:16132582", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->16132582],
  Cell[805140, 22970, 1294, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:742018993", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C 1"},
   CellID->742018993],
  Cell[806437, 23011, 1795, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:281673814", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C \[NotEqual] 0"},
   CellID->281673814],
  Cell[808235, 23068, 1827, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:122376381", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C=0"},
   CellID->122376381],
  Cell[810065, 23126, 2253, 68, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:256177412", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C=0 seq"},
   CellID->256177412],
  Cell[812321, 23196, 1778, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:293354245", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C=0 seq'"},
   CellID->293354245],
  Cell[814102, 23253, 1629, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:129332919", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C prepend"},
   CellID->129332919],
  Cell[851032, 24258, 759, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:136746483", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->136746483],
  Cell[851794, 24280, 729, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:110691644", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->110691644],
  Cell[852526, 24301, 1012, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:371581865", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + ProdDomain"},
   CellID->371581865],
  Cell[853541, 24329, 778, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:694735421", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + isTuple"},
   CellID->694735421],
  Cell[854322, 24351, 1182, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:442662347", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[NotEqual] 0"},
   CellID->442662347],
  Cell[855507, 24387, 2440, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:66376992", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + pp \[SubsetEqual]"},
   CellID->66376992],
  Cell[882031, 25113, 856, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:480657044", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->480657044],
  Cell[882890, 25136, 1398, 39, 70, "GlobalDeclaration",
   CellTags->{
    "ID:907803476", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->907803476],
  Cell[884291, 25177, 2224, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:442268007", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[Succeeds] 1"},
   CellID->442268007],
  Cell[886518, 25238, 2123, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:26802188", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[Succeeds] 2"},
   CellID->26802188],
  Cell[888644, 25298, 2216, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:4890345", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[Succeeds] 3"},
   CellID->4890345],
  Cell[924151, 26172, 738, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:212784844", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->212784844],
  Cell[924892, 26193, 838, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:106856858", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - isTuple"},
   CellID->106856858],
  Cell[925733, 26218, 964, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:200895043", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - length"},
   CellID->200895043],
  Cell[926700, 26247, 1367, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:19952680", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - pp"},
   CellID->19952680],
  Cell[940945, 26670, 1973, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:118174250", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isMonomial I"},
   CellID->118174250],
  Cell[942921, 26732, 828, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:78163703", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->78163703],
  Cell[943752, 26757, 1691, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:748742057", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isMonomial II"},
   CellID->748742057],
  Cell[945446, 26803, 1819, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:104425812", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isMonomial III"},
   CellID->104425812],
  Cell[947268, 26858, 1344, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:191443971", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CenterDot]"},
   CellID->191443971],
  Cell[971779, 27505, 835, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:956509664", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->956509664],
  Cell[972617, 27529, 944, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:52346492", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first index", "IR"},
   CellID->52346492],
  Cell[973564, 27556, 1433, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:3252425", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isLP"},
   CellID->3252425],
  Cell[975000, 27602, 898, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:301012536", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lp"},
   CellID->301012536],
  Cell[975901, 27629, 898, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:499482967", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lc"},
   CellID->499482967],
  Cell[976802, 27656, 912, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:77776384", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lm"},
   CellID->77776384],
  Cell[1000465, 28331, 703, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1015977913", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->1015977913],
  Cell[1001171, 28351, 938, 25, 70, "GlobalDeclaration",
   CellTags->{
    "ID:731677923", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->731677923],
  Cell[1002112, 28378, 1037, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:23863949", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lp seq"},
   CellID->23863949],
  Cell[1003152, 28408, 1040, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:721799663", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lc seq"},
   CellID->721799663],
  Cell[1004195, 28438, 1181, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2939774", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lm seq"},
   CellID->2939774],
  Cell[1005379, 28471, 1114, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:172576768", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples R seq"},
   CellID->172576768],
  Cell[1006496, 28503, 1899, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:3206806", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CenterDot] lp lc"},
   CellID->3206806],
  Cell[1038907, 29384, 830, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:31351138", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalR] isDomain"},
   CellID->31351138],
  Cell[1039740, 29408, 835, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:632462358", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalT] isDomain"},
   CellID->632462358],
  Cell[1040578, 29432, 869, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:404915746", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isDomain"},
   CellID->404915746],
  Cell[1049343, 29714, 753, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:3279840", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->3279840],
  Cell[1050099, 29736, 1082, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:141435198", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership C"},
   CellID->141435198],
  Cell[1051184, 29768, 880, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:45852914", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership 0"},
   CellID->45852914],
  Cell[1052067, 29794, 883, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:320592129", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership 1"},
   CellID->320592129],
  Cell[1052953, 29820, 907, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:93322392", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership +"},
   CellID->93322392],
  Cell[1053863, 29847, 897, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:6751189", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership -"},
   CellID->6751189],
  Cell[1054763, 29874, 1252, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:179493910", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples supp"},
   CellID->179493910],
  Cell[1079572, 30595, 830, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:644982519", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasDecisionPredicate"},
   CellID->644982519],
  Cell[1083748, 30727, 819, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:114843091", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples containsMonomials"},
   CellID->114843091],
  Cell[1087947, 30860, 759, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:78312379", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->78312379],
  Cell[1088709, 30881, 1369, 39, 70, "GlobalDeclaration",
   CellTags->{
    "ID:210280178", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->210280178],
  Cell[1090081, 30922, 3948, 109, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:261724624", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 1"},
   CellID->261724624],
  Cell[1094032, 31033, 3703, 105, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:115965410", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 2"},
   CellID->115965410],
  Cell[1097738, 31140, 3955, 118, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:384214099", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 3"},
   CellID->384214099],
  Cell[1101696, 31260, 2710, 81, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:56134460", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 4"},
   CellID->56134460],
  Cell[1175749, 32989, 738, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:235178571", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->235178571],
  Cell[1176490, 33010, 1193, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:474434374", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C 2"},
   CellID->474434374],
  Cell[1177686, 33045, 763, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:108633346", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->108633346],
  Cell[1178452, 33067, 1231, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:611741232", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C +"},
   CellID->611741232],
  Cell[1179686, 33104, 1084, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:438277340", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C - unary"},
   CellID->438277340],
  Cell[1180773, 33137, 1180, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:412638768", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C -"},
   CellID->412638768],
  Cell[1205632, 33810, 838, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:392022323", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasCoefficientFunction"},
   CellID->392022323],
  Cell[1209828, 33943, 751, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:23554005", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->23554005],
  Cell[1210582, 33964, 992, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:392464903", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->392464903],
  Cell[1211577, 33992, 1238, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:161672314", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] empty"},
   CellID->161672314],
  Cell[1212818, 34026, 3855, 97, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:71703914", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] prepend"},
   CellID->71703914],
  Cell[1216676, 34125, 738, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:127729286", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->127729286],
  Cell[1217417, 34146, 1346, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2718947", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] lm"},
   CellID->2718947],
  Cell[1218766, 34183, 1210, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:204263407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership \[CircleTimes]"},
   CellID->204263407],
  Cell[1219979, 34217, 1539, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:134536220", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] \[RawStar]"},
   CellID->134536220],
  Cell[1221521, 34257, 1604, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:387729163", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[RawStar] alternative"},
   CellID->387729163],
  Cell[1276715, 35633, 816, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:65280633", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasMultiplication"},
   CellID->65280633],
  Cell[1280841, 35765, 2804, 92, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:48571309", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[ForAll] alternative 2"},
   CellID->48571309],
  Cell[1283648, 35859, 763, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:617696571", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->617696571],
  Cell[1284414, 35881, 1490, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:776548", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples=empty sufficient"},
   CellID->776548],
  Cell[1285907, 35930, 1542, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:304028020", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples extensional"},
   CellID->304028020],
  Cell[1306777, 36501, 811, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:589891585", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isExtensional"},
   CellID->589891585],
  Cell[1310890, 36632, 806, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:160815443", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isOrderedRec"},
   CellID->160815443],
  Cell[1311699, 36655, 795, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:397107848", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isOrdered"},
   CellID->397107848],
  Cell[1318255, 36863, 812, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:156782366", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasMultipliers"},
   CellID->156782366],
  Cell[1322389, 36994, 860, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:667417187", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isReductionPolynomialDomain"},
   CellID->667417187],
  Cell[1326871, 37138, 967, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:714775050", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->714775050],
  Cell[1327841, 37166, 4116, 109, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:15424584", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd alternative"},
   CellID->15424584],
  Cell[1331960, 37277, 1655, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:168826371", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd unary alternative"},
   CellID->168826371],
  Cell[1333618, 37326, 4778, 122, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:228703376", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd lc lp"},
   CellID->228703376],
  Cell[1338399, 37450, 2404, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:354399952", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd length"},
   CellID->354399952],
  Cell[1401104, 38917, 790, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:479598778", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasLCRD"},
   CellID->479598778],
  Cell[1405160, 39048, 787, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:463361785", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasRDM"},
   CellID->463361785],
  Cell[1409208, 39179, 783, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:7345776", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasCPM"},
   CellID->7345776],
  Cell[1413281, 39311, 875, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1014777887", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isAlgoReductionPolynomialDomain"},
   CellID->1014777887],
  Cell[1417346, 39438, 1542, 41, 78, "FormalTextInputFormula",
   CellTags->{
    "ID:67407018", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isAlgoReductionRing"},
   CellID->67407018]},
 "ID:603497554"->{
  Cell[45833, 1145, 2342, 60, 137, "FormalTextInputFormula",
   CellTags->{
    "ID:603497554", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain"},
   CellID->603497554]},
 "isExpPPDomain"->{
  Cell[45833, 1145, 2342, 60, 137, "FormalTextInputFormula",
   CellTags->{
    "ID:603497554", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain"},
   CellID->603497554]},
 "ID:49777475"->{
  Cell[48178, 1207, 3150, 87, 135, "FormalTextInputFormula",
   CellTags->{
    "ID:49777475", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasDecisionPredicatePP"},
   CellID->49777475]},
 "hasDecisionPredicatePP"->{
  Cell[48178, 1207, 3150, 87, 135, "FormalTextInputFormula",
   CellTags->{
    "ID:49777475", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasDecisionPredicatePP"},
   CellID->49777475]},
 "ID:225295367"->{
  Cell[51331, 1296, 5696, 147, 224, "FormalTextInputFormula",
   CellTags->{
    "ID:225295367", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasExponentFunctionPP"},
   CellID->225295367]},
 "hasExponentFunctionPP"->{
  Cell[51331, 1296, 5696, 147, 224, "FormalTextInputFormula",
   CellTags->{
    "ID:225295367", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasExponentFunctionPP"},
   CellID->225295367]},
 "ID:292978558"->{
  Cell[57030, 1445, 2105, 64, 52, "FormalTextInputFormula",
   CellTags->{
    "ID:292978558", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExtensionalPP"},
   CellID->292978558]},
 "isExtensionalPP"->{
  Cell[57030, 1445, 2105, 64, 52, "FormalTextInputFormula",
   CellTags->{
    "ID:292978558", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExtensionalPP"},
   CellID->292978558]},
 "ID:340781435"->{
  Cell[59138, 1511, 1950, 58, 63, "FormalTextInputFormula",
   CellTags->{
    "ID:340781435", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasDegreeFunctionPP"},
   CellID->340781435]},
 "hasDegreeFunctionPP"->{
  Cell[59138, 1511, 1950, 58, 63, "FormalTextInputFormula",
   CellTags->{
    "ID:340781435", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasDegreeFunctionPP"},
   CellID->340781435]},
 "ID:480617864"->{
  Cell[61091, 1571, 2199, 68, 52, "FormalTextInputFormula",
   CellTags->{
    "ID:480617864", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasDivisibilityPP"},
   CellID->480617864]},
 "hasDivisibilityPP"->{
  Cell[61091, 1571, 2199, 68, 52, "FormalTextInputFormula",
   CellTags->{
    "ID:480617864", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasDivisibilityPP"},
   CellID->480617864]},
 "ID:897055662"->{
  Cell[63961, 1666, 2958, 87, 115, "FormalTextInputFormula",
   CellTags->{
    "ID:897055662", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isTermOrder"},
   CellID->897055662]},
 "isTermOrder"->{
  Cell[63961, 1666, 2958, 87, 115, "FormalTextInputFormula",
   CellTags->{
    "ID:897055662", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isTermOrder"},
   CellID->897055662]},
 "ID:532395027"->{
  Cell[67823, 1788, 703, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:532395027", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->532395027]},
 "ID:160997957"->{
  Cell[68529, 1808, 854, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:160997957", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain dim"},
   CellID->160997957]},
 "isExpPPDomain dim"->{
  Cell[68529, 1808, 854, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:160997957", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain dim"},
   CellID->160997957]},
 "ID:83558965"->{
  Cell[69386, 1832, 818, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:83558965", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasDecisionPredicate"},
   CellID->83558965]},
 "isExpPPDomain hasDecisionPredicate"->{
  Cell[69386, 1832, 818, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:83558965", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasDecisionPredicate"},
   CellID->83558965]},
 "ID:349898371"->{
  Cell[70207, 1856, 811, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:349898371", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasExponentFunction"},
   CellID->349898371]},
 "isExpPPDomain hasExponentFunction"->{
  Cell[70207, 1856, 811, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:349898371", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasExponentFunction"},
   CellID->349898371]},
 "ID:200216737"->{
  Cell[71021, 1879, 790, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:200216737", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isExtensional"},
   CellID->200216737]},
 "isExpPPDomain isExtensional"->{
  Cell[71021, 1879, 790, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:200216737", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isExtensional"},
   CellID->200216737]},
 "ID:276506833"->{
  Cell[71814, 1902, 803, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:276506833", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasDegreeFunction"},
   CellID->276506833]},
 "isExpPPDomain hasDegreeFunction"->{
  Cell[71814, 1902, 803, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:276506833", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasDegreeFunction"},
   CellID->276506833]},
 "ID:135986343"->{
  Cell[72620, 1925, 821, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:135986343", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasDivisibility"},
   CellID->135986343]},
 "isExpPPDomain hasDivisibility"->{
  Cell[72620, 1925, 821, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:135986343", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasDivisibility"},
   CellID->135986343]},
 "ID:122009120"->{
  Cell[73444, 1948, 1254, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:122009120", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership exp"},
   CellID->122009120]},
 "isExpPPDomain membership exp"->{
  Cell[73444, 1948, 1254, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:122009120", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership exp"},
   CellID->122009120]},
 "Proof|ID:160997957"->{
  Cell[74854, 1994, 2458, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:160997957", "Proof|ID:160997957-1"},
   CellID->4369865]},
 "Proof|ID:160997957-1"->{
  Cell[74854, 1994, 2458, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:160997957", "Proof|ID:160997957-1"},
   CellID->4369865]},
 "Proof|ID:83558965"->{
  Cell[77831, 2087, 2316, 67, 70, "ProofInfo",
   CellTags->{"Proof|ID:83558965", "Proof|ID:83558965-1"},
   CellID->777822562]},
 "Proof|ID:83558965-1"->{
  Cell[77831, 2087, 2316, 67, 70, "ProofInfo",
   CellTags->{"Proof|ID:83558965", "Proof|ID:83558965-1"},
   CellID->777822562]},
 "Proof|ID:349898371"->{
  Cell[80666, 2173, 2324, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:349898371", "Proof|ID:349898371-1"},
   CellID->51022891]},
 "Proof|ID:349898371-1"->{
  Cell[80666, 2173, 2324, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:349898371", "Proof|ID:349898371-1"},
   CellID->51022891]},
 "Proof|ID:200216737"->{
  Cell[83510, 2260, 2284, 66, 70, "ProofInfo",
   CellTags->{"Proof|ID:200216737", "Proof|ID:200216737-1"},
   CellID->19036167]},
 "Proof|ID:200216737-1"->{
  Cell[83510, 2260, 2284, 66, 70, "ProofInfo",
   CellTags->{"Proof|ID:200216737", "Proof|ID:200216737-1"},
   CellID->19036167]},
 "Proof|ID:276506833"->{
  Cell[86314, 2345, 2309, 67, 70, "ProofInfo",
   CellTags->{"Proof|ID:276506833", "Proof|ID:276506833-1"},
   CellID->447792752]},
 "Proof|ID:276506833-1"->{
  Cell[86314, 2345, 2309, 67, 70, "ProofInfo",
   CellTags->{"Proof|ID:276506833", "Proof|ID:276506833-1"},
   CellID->447792752]},
 "Proof|ID:135986343"->{
  Cell[89143, 2431, 2293, 66, 70, "ProofInfo",
   CellTags->{"Proof|ID:135986343", "Proof|ID:135986343-1"},
   CellID->380432363]},
 "Proof|ID:135986343-1"->{
  Cell[89143, 2431, 2293, 66, 70, "ProofInfo",
   CellTags->{"Proof|ID:135986343", "Proof|ID:135986343-1"},
   CellID->380432363]},
 "Proof|ID:122009120"->{
  Cell[92000, 2517, 3752, 101, 70, "ProofInfo",
   CellTags->{"Proof|ID:122009120", "Proof|ID:122009120-1"},
   CellID->131414947]},
 "Proof|ID:122009120-1"->{
  Cell[92000, 2517, 3752, 101, 70, "ProofInfo",
   CellTags->{"Proof|ID:122009120", "Proof|ID:122009120-1"},
   CellID->131414947]},
 "ID:185429719"->{
  Cell[96528, 2645, 703, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:185429719", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->185429719]},
 "ID:81626705"->{
  Cell[97763, 2682, 794, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:81626705", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isDomain"},
   CellID->81626705]},
 "isExpPPDomain isDomain"->{
  Cell[97763, 2682, 794, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:81626705", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isDomain"},
   CellID->81626705]},
 "ID:154517895"->{
  Cell[98560, 2705, 1095, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:154517895", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership \[RawStar]", "IR"},
   CellID->154517895]},
 "isExpPPDomain membership \[RawStar]"->{
  Cell[98560, 2705, 1095, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:154517895", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership \[RawStar]", "IR"},
   CellID->154517895]},
 "IR"->{
  Cell[98560, 2705, 1095, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:154517895", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership \[RawStar]", "IR"},
   CellID->154517895],
  Cell[155221, 4374, 951, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:47160", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership deg", "IR"},
   CellID->47160],
  Cell[455471, 12982, 1647, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:617800164", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PPTuples alternative", "IR"},
   CellID->617800164],
  Cell[616582, 17794, 923, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:689486298", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isTuple", "IR"},
   CellID->689486298],
  Cell[619018, 17871, 912, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:619631752", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples empty", "IR"},
   CellID->619631752],
  Cell[619933, 17897, 3154, 93, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:93154383", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples rec", "IR"},
   CellID->93154383],
  Cell[623090, 17992, 3179, 91, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:110840922", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples rec II", "IR"},
   CellID->110840922],
  Cell[626272, 18085, 3467, 100, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:45787564", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples rec III", "IR"},
   CellID->45787564],
  Cell[630452, 18208, 2714, 80, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:134755829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples prepend", "IR"},
   CellID->134755829],
  Cell[633169, 18290, 1120, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:218126712", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership first", "IR"},
   CellID->218126712],
  Cell[634292, 18323, 1117, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:36095185", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership second", "IR"},
   CellID->36095185],
  Cell[705462, 20242, 3708, 107, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:329976481", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples induction", "IR"},
   CellID->329976481],
  Cell[709772, 20370, 4150, 117, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:379661071", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples induction'", "IR"},
   CellID->379661071],
  Cell[972617, 27529, 944, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:52346492", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first index", "IR"},
   CellID->52346492]},
 "ID:226316196"->{
  Cell[99658, 2738, 993, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:226316196", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain \[RawStar] isAssoc"},
   CellID->226316196]},
 "isExpPPDomain \[RawStar] isAssoc"->{
  Cell[99658, 2738, 993, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:226316196", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain \[RawStar] isAssoc"},
   CellID->226316196]},
 "ID:96938761"->{
  Cell[100654, 2770, 960, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:96938761", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain \[RawStar] isComm"},
   CellID->96938761]},
 "isExpPPDomain \[RawStar] isComm"->{
  Cell[100654, 2770, 960, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:96938761", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain \[RawStar] isComm"},
   CellID->96938761]},
 "ID:499736780"->{
  Cell[101617, 2802, 1024, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:499736780", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain 1 isNeutral"},
   CellID->499736780]},
 "isExpPPDomain 1 isNeutral"->{
  Cell[101617, 2802, 1024, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:499736780", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain 1 isNeutral"},
   CellID->499736780]},
 "ID:745587327"->{
  Cell[102644, 2835, 1458, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:745587327", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain \[RawStar] lcanc"},
   CellID->745587327]},
 "isExpPPDomain \[RawStar] lcanc"->{
  Cell[102644, 2835, 1458, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:745587327", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain \[RawStar] lcanc"},
   CellID->745587327]},
 "Proof|ID:81626705"->{
  Cell[104258, 2893, 2256, 66, 70, "ProofInfo",
   CellTags->{"Proof|ID:81626705", "Proof|ID:81626705-1"},
   CellID->281230526]},
 "Proof|ID:81626705-1"->{
  Cell[104258, 2893, 2256, 66, 70, "ProofInfo",
   CellTags->{"Proof|ID:81626705", "Proof|ID:81626705-1"},
   CellID->281230526]},
 "Proof|ID:154517895"->{
  Cell[107033, 2978, 3605, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:154517895", "Proof|ID:154517895-1"},
   CellID->527357288]},
 "Proof|ID:154517895-1"->{
  Cell[107033, 2978, 3605, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:154517895", "Proof|ID:154517895-1"},
   CellID->527357288]},
 "Proof|ID:226316196"->{
  Cell[111158, 3091, 2524, 72, 70, "ProofInfo",
   CellTags->{"Proof|ID:226316196", "Proof|ID:226316196-1"},
   CellID->31171248]},
 "Proof|ID:226316196-1"->{
  Cell[111158, 3091, 2524, 72, 70, "ProofInfo",
   CellTags->{"Proof|ID:226316196", "Proof|ID:226316196-1"},
   CellID->31171248]},
 "Proof|ID:96938761"->{
  Cell[114202, 3182, 2514, 72, 70, "ProofInfo",
   CellTags->{"Proof|ID:96938761", "Proof|ID:96938761-1"},
   CellID->6993021]},
 "Proof|ID:96938761-1"->{
  Cell[114202, 3182, 2514, 72, 70, "ProofInfo",
   CellTags->{"Proof|ID:96938761", "Proof|ID:96938761-1"},
   CellID->6993021]},
 "Proof|ID:499736780"->{
  Cell[117234, 3273, 2749, 76, 70, "ProofInfo",
   CellTags->{"Proof|ID:499736780", "Proof|ID:499736780-1"},
   CellID->247701442]},
 "Proof|ID:499736780-1"->{
  Cell[117234, 3273, 2749, 76, 70, "ProofInfo",
   CellTags->{"Proof|ID:499736780", "Proof|ID:499736780-1"},
   CellID->247701442]},
 "Proof|ID:745587327"->{
  Cell[120502, 3368, 4697, 120, 70, "ProofInfo",
   CellTags->{"Proof|ID:745587327", "Proof|ID:745587327-1"},
   CellID->99539526]},
 "Proof|ID:745587327-1"->{
  Cell[120502, 3368, 4697, 120, 70, "ProofInfo",
   CellTags->{"Proof|ID:745587327", "Proof|ID:745587327-1"},
   CellID->99539526]},
 "ID:333683969"->{
  Cell[126159, 3521, 804, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:333683969", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isCCMonoid"},
   CellID->333683969]},
 "isExpPPDomain isCCMonoid"->{
  Cell[126159, 3521, 804, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:333683969", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isCCMonoid"},
   CellID->333683969]},
 "Proof|ID:333683969"->{
  Cell[127117, 3553, 2269, 66, 70, "ProofInfo",
   CellTags->{"Proof|ID:333683969", "Proof|ID:333683969-1"},
   CellID->190665795]},
 "Proof|ID:333683969-1"->{
  Cell[127117, 3553, 2269, 66, 70, "ProofInfo",
   CellTags->{"Proof|ID:333683969", "Proof|ID:333683969-1"},
   CellID->190665795]},
 "ID:274960855"->{
  Cell[130592, 3660, 700, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:274960855", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->274960855]},
 "ID:52136714"->{
  Cell[131295, 3680, 1031, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:52136714", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain lcm isJoin"},
   CellID->52136714]},
 "isExpPPDomain lcm isJoin"->{
  Cell[131295, 3680, 1031, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:52136714", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain lcm isJoin"},
   CellID->52136714]},
 "ID:35376325"->{
  Cell[132329, 3713, 760, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:35376325", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->35376325]},
 "ID:138508454"->{
  Cell[133092, 3735, 1279, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:138508454", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain div.1"},
   CellID->138508454]},
 "isExpPPDomain div.1"->{
  Cell[133092, 3735, 1279, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:138508454", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain div.1"},
   CellID->138508454]},
 "ID:14785826"->{
  Cell[134374, 3776, 1437, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:14785826", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain div.2"},
   CellID->14785826]},
 "isExpPPDomain div.2"->{
  Cell[134374, 3776, 1437, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:14785826", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain div.2"},
   CellID->14785826]},
 "Proof|ID:52136714"->{
  Cell[135967, 3833, 2766, 76, 70, "ProofInfo",
   CellTags->{"Proof|ID:52136714", "Proof|ID:52136714-1"},
   CellID->110160557]},
 "Proof|ID:52136714-1"->{
  Cell[135967, 3833, 2766, 76, 70, "ProofInfo",
   CellTags->{"Proof|ID:52136714", "Proof|ID:52136714-1"},
   CellID->110160557]},
 "Proof|ID:138508454"->{
  Cell[139251, 3928, 4458, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:138508454", "Proof|ID:138508454-1"},
   CellID->245133977]},
 "Proof|ID:138508454-1"->{
  Cell[139251, 3928, 4458, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:138508454", "Proof|ID:138508454-1"},
   CellID->245133977]},
 "Proof|ID:14785826"->{
  Cell[144229, 4061, 4740, 120, 70, "ProofInfo",
   CellTags->{"Proof|ID:14785826", "Proof|ID:14785826-1"},
   CellID->2945667]},
 "Proof|ID:14785826-1"->{
  Cell[144229, 4061, 4740, 120, 70, "ProofInfo",
   CellTags->{"Proof|ID:14785826", "Proof|ID:14785826-1"},
   CellID->2945667]},
 "ID:181034073"->{
  Cell[150172, 4222, 703, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:181034073", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->181034073]},
 "ID:943617998"->{
  Cell[150878, 4242, 738, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:943617998", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->943617998]},
 "ID:706945753"->{
  Cell[151619, 4263, 1424, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:706945753", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership deg ind"},
   CellID->706945753]},
 "isExpPPDomain membership deg ind"->{
  Cell[151619, 4263, 1424, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:706945753", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership deg ind"},
   CellID->706945753]},
 "ID:606418540"->{
  Cell[153046, 4303, 2172, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:606418540", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain deg=0 ind"},
   CellID->606418540]},
 "isExpPPDomain deg=0 ind"->{
  Cell[153046, 4303, 2172, 69, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:606418540", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain deg=0 ind"},
   CellID->606418540]},
 "ID:47160"->{
  Cell[155221, 4374, 951, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:47160", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership deg", "IR"},
   CellID->47160]},
 "isExpPPDomain membership deg"->{
  Cell[155221, 4374, 951, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:47160", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership deg", "IR"},
   CellID->47160]},
 "ID:58897330"->{
  Cell[156175, 4401, 1158, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:58897330", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain deg=0"},
   CellID->58897330]},
 "isExpPPDomain deg=0"->{
  Cell[156175, 4401, 1158, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:58897330", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain deg=0"},
   CellID->58897330]},
 "ID:170894204"->{
  Cell[157336, 4439, 1333, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:170894204", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain deg \[RawStar]"},
   CellID->170894204]},
 "isExpPPDomain deg \[RawStar]"->{
  Cell[157336, 4439, 1333, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:170894204", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain deg \[RawStar]"},
   CellID->170894204]},
 "Proof|ID:706945753"->{
  Cell[158824, 4489, 4667, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:706945753", "Proof|ID:706945753-1"},
   CellID->905512569]},
 "Proof|ID:706945753-1"->{
  Cell[158824, 4489, 4667, 121, 70, "ProofInfo",
   CellTags->{"Proof|ID:706945753", "Proof|ID:706945753-1"},
   CellID->905512569]},
 "Proof|ID:606418540"->{
  Cell[164009, 4629, 5957, 150, 70, "ProofInfo",
   CellTags->{"Proof|ID:606418540", "Proof|ID:606418540-1"},
   CellID->408182534]},
 "Proof|ID:606418540-1"->{
  Cell[164009, 4629, 5957, 150, 70, "ProofInfo",
   CellTags->{"Proof|ID:606418540", "Proof|ID:606418540-1"},
   CellID->408182534]},
 "Proof|ID:47160"->{
  Cell[170486, 4798, 3030, 85, 70, "ProofInfo",
   CellTags->{"Proof|ID:47160", "Proof|ID:47160-1"},
   CellID->168232727]},
 "Proof|ID:47160-1"->{
  Cell[170486, 4798, 3030, 85, 70, "ProofInfo",
   CellTags->{"Proof|ID:47160", "Proof|ID:47160-1"},
   CellID->168232727]},
 "Proof|ID:58897330"->{
  Cell[174036, 4902, 3461, 96, 70, "ProofInfo",
   CellTags->{"Proof|ID:58897330", "Proof|ID:58897330-1"},
   CellID->293749343]},
 "Proof|ID:58897330-1"->{
  Cell[174036, 4902, 3461, 96, 70, "ProofInfo",
   CellTags->{"Proof|ID:58897330", "Proof|ID:58897330-1"},
   CellID->293749343]},
 "Proof|ID:170894204"->{
  Cell[178016, 5017, 4519, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:170894204", "Proof|ID:170894204-1"},
   CellID->366355002]},
 "Proof|ID:170894204-1"->{
  Cell[178016, 5017, 4519, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:170894204", "Proof|ID:170894204-1"},
   CellID->366355002]},
 "ID:49640627"->{
  Cell[183800, 5175, 1710, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:49640627", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "increasing subsequence \[DoubleStruckCapitalN]0"},
   CellID->49640627]},
 "increasing subsequence \[DoubleStruckCapitalN]0"->{
  Cell[183800, 5175, 1710, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:49640627", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "increasing subsequence \[DoubleStruckCapitalN]0"},
   CellID->49640627]},
 "Proof|ID:49640627"->{
  Cell[185666, 5233, 5144, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:49640627", "Proof|ID:49640627-1"},
   CellID->620572316]},
 "Proof|ID:49640627-1"->{
  Cell[185666, 5233, 5144, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:49640627", "Proof|ID:49640627-1"},
   CellID->620572316]},
 "ID:256365628"->{
  Cell[191307, 5382, 701, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:256365628", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->256365628]},
 "ID:443096051"->{
  Cell[192485, 5418, 1782, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:443096051", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain divN"},
   CellID->443096051]},
 "isExpPPDomain divN"->{
  Cell[192485, 5418, 1782, 54, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:443096051", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain divN"},
   CellID->443096051]},
 "ID:27206868"->{
  Cell[194868, 5497, 2207, 64, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:27206868", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain divN +1"},
   CellID->27206868]},
 "isExpPPDomain divN +1"->{
  Cell[194868, 5497, 2207, 64, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:27206868", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain divN +1"},
   CellID->27206868]},
 "ID:184673495"->{
  Cell[197078, 5563, 1809, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:184673495", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "increasing subsequence isExpPPDomain 0"},
   CellID->184673495]},
 "increasing subsequence isExpPPDomain 0"->{
  Cell[197078, 5563, 1809, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:184673495", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "increasing subsequence isExpPPDomain 0"},
   CellID->184673495]},
 "ID:43287351"->{
  Cell[198890, 5616, 1975, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:43287351", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "increasing subsequence isExpPPDomain"},
   CellID->43287351]},
 "increasing subsequence isExpPPDomain"->{
  Cell[198890, 5616, 1975, 57, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:43287351", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "increasing subsequence isExpPPDomain"},
   CellID->43287351]},
 "Proof|ID:27206868"->{
  Cell[201020, 5684, 6427, 156, 70, "ProofInfo",
   CellTags->{"Proof|ID:27206868", "Proof|ID:27206868-1"},
   CellID->99745602]},
 "Proof|ID:27206868-1"->{
  Cell[201020, 5684, 6427, 156, 70, "ProofInfo",
   CellTags->{"Proof|ID:27206868", "Proof|ID:27206868-1"},
   CellID->99745602]},
 "Proof|ID:184673495"->{
  Cell[207965, 5859, 5735, 144, 70, "ProofInfo",
   CellTags->{"Proof|ID:184673495", "Proof|ID:184673495-1"},
   CellID->104166294]},
 "Proof|ID:184673495-1"->{
  Cell[207965, 5859, 5735, 144, 70, "ProofInfo",
   CellTags->{"Proof|ID:184673495", "Proof|ID:184673495-1"},
   CellID->104166294]},
 "Proof|ID:43287351"->{
  Cell[214220, 6022, 6360, 158, 70, "ProofInfo",
   CellTags->{"Proof|ID:43287351", "Proof|ID:43287351-1"},
   CellID->460075245]},
 "Proof|ID:43287351-1"->{
  Cell[214220, 6022, 6360, 158, 70, "ProofInfo",
   CellTags->{"Proof|ID:43287351", "Proof|ID:43287351-1"},
   CellID->460075245]},
 "ID:30919390"->{
  Cell[221550, 6213, 1022, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:30919390", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isDicksonDomain"},
   CellID->30919390]},
 "isExpPPDomain isDicksonDomain"->{
  Cell[221550, 6213, 1022, 29, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:30919390", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isDicksonDomain"},
   CellID->30919390]},
 "Proof|ID:30919390"->{
  Cell[222727, 6253, 2574, 72, 70, "ProofInfo",
   CellTags->{"Proof|ID:30919390", "Proof|ID:30919390-1"},
   CellID->278064001]},
 "Proof|ID:30919390-1"->{
  Cell[222727, 6253, 2574, 72, 70, "ProofInfo",
   CellTags->{"Proof|ID:30919390", "Proof|ID:30919390-1"},
   CellID->278064001]},
 "ID:323333890"->{
  Cell[226555, 6367, 781, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:323333890", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->323333890]},
 "ID:235080407"->{
  Cell[227339, 6389, 771, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:235080407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->235080407]},
 "ID:691531428"->{
  Cell[228113, 6411, 1338, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:691531428", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "term order |"},
   CellID->691531428]},
 "term order |"->{
  Cell[228113, 6411, 1338, 43, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:691531428", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "term order |"},
   CellID->691531428]},
 "ID:772895867"->{
  Cell[229454, 6456, 989, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:772895867", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "term order isOrderEmbedding"},
   CellID->772895867]},
 "term order isOrderEmbedding"->{
  Cell[229454, 6456, 989, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:772895867", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "term order isOrderEmbedding"},
   CellID->772895867]},
 "ID:83375799"->{
  Cell[230446, 6486, 819, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:83375799", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "term order isNoetherian"},
   CellID->83375799]},
 "term order isNoetherian"->{
  Cell[230446, 6486, 819, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:83375799", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "term order isNoetherian"},
   CellID->83375799]},
 "Proof|ID:691531428"->{
  Cell[231419, 6518, 5172, 130, 70, "ProofInfo",
   CellTags->{"Proof|ID:691531428", "Proof|ID:691531428-1"},
   CellID->26598154]},
 "Proof|ID:691531428-1"->{
  Cell[231419, 6518, 5172, 130, 70, "ProofInfo",
   CellTags->{"Proof|ID:691531428", "Proof|ID:691531428-1"},
   CellID->26598154]},
 "Proof|ID:772895867"->{
  Cell[237111, 6667, 3285, 88, 70, "ProofInfo",
   CellTags->{"Proof|ID:772895867", "Proof|ID:772895867-1"},
   CellID->409263698]},
 "Proof|ID:772895867-1"->{
  Cell[237111, 6667, 3285, 88, 70, "ProofInfo",
   CellTags->{"Proof|ID:772895867", "Proof|ID:772895867-1"},
   CellID->409263698]},
 "Proof|ID:83375799"->{
  Cell[240916, 6774, 3040, 84, 70, "ProofInfo",
   CellTags->{"Proof|ID:83375799", "Proof|ID:83375799-1"},
   CellID->549031787]},
 "Proof|ID:83375799-1"->{
  Cell[240916, 6774, 3040, 84, 70, "ProofInfo",
   CellTags->{"Proof|ID:83375799", "Proof|ID:83375799-1"},
   CellID->549031787]},
 "ID:69380281"->{
  Cell[245321, 6906, 680, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:69380281", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->69380281]},
 "ID:70274751"->{
  Cell[246004, 6925, 723, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:70274751", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->70274751]},
 "ID:417719229"->{
  Cell[246730, 6946, 669, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:417719229", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->417719229]},
 "ID:296713523"->{
  Cell[247402, 6965, 1279, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:296713523", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP \[Element]"},
   CellID->296713523]},
 "OrderedPP \[Element]"->{
  Cell[247402, 6965, 1279, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:296713523", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP \[Element]"},
   CellID->296713523]},
 "ID:275450356"->{
  Cell[248684, 7009, 818, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:275450356", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP dim"},
   CellID->275450356]},
 "OrderedPP dim"->{
  Cell[248684, 7009, 818, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:275450356", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP dim"},
   CellID->275450356]},
 "ID:126765714"->{
  Cell[249505, 7033, 1082, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:126765714", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP exp"},
   CellID->126765714]},
 "OrderedPP exp"->{
  Cell[249505, 7033, 1082, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:126765714", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP exp"},
   CellID->126765714]},
 "ID:300238463"->{
  Cell[250590, 7068, 810, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:300238463", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP 1"},
   CellID->300238463]},
 "OrderedPP 1"->{
  Cell[250590, 7068, 810, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:300238463", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP 1"},
   CellID->300238463]},
 "ID:91738893"->{
  Cell[251403, 7092, 863, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:91738893", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP \[RawStar]"},
   CellID->91738893]},
 "OrderedPP \[RawStar]"->{
  Cell[251403, 7092, 863, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:91738893", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP \[RawStar]"},
   CellID->91738893]},
 "ID:1017172830"->{
  Cell[252269, 7119, 867, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1017172830", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP /"},
   CellID->1017172830]},
 "OrderedPP /"->{
  Cell[252269, 7119, 867, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1017172830", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP /"},
   CellID->1017172830]},
 "ID:62103041"->{
  Cell[253139, 7146, 923, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:62103041", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP lcm"},
   CellID->62103041]},
 "OrderedPP lcm"->{
  Cell[253139, 7146, 923, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:62103041", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP lcm"},
   CellID->62103041]},
 "ID:229667832"->{
  Cell[254065, 7174, 1335, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:229667832", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP |"},
   CellID->229667832]},
 "OrderedPP |"->{
  Cell[254065, 7174, 1335, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:229667832", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP |"},
   CellID->229667832]},
 "ID:687245437"->{
  Cell[255403, 7218, 935, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:687245437", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP deg"},
   CellID->687245437]},
 "OrderedPP deg"->{
  Cell[255403, 7218, 935, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:687245437", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP deg"},
   CellID->687245437]},
 "ID:353388395"->{
  Cell[256341, 7246, 1143, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:353388395", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP \[Succeeds]"},
   CellID->353388395]},
 "OrderedPP \[Succeeds]"->{
  Cell[256341, 7246, 1143, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:353388395", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP \[Succeeds]"},
   CellID->353388395]},
 "ID:144609801"->{
  Cell[257487, 7283, 1096, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:144609801", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP <<"},
   CellID->144609801]},
 "OrderedPP <<"->{
  Cell[257487, 7283, 1096, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:144609801", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP <<"},
   CellID->144609801]},
 "ID:189296520"->{
  Cell[258586, 7321, 1096, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:189296520", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP >>"},
   CellID->189296520]},
 "OrderedPP >>"->{
  Cell[258586, 7321, 1096, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:189296520", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP >>"},
   CellID->189296520]},
 "ID:29953053"->{
  Cell[260643, 7395, 776, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:29953053", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->29953053]},
 "ID:309264225"->{
  Cell[261422, 7417, 869, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:309264225", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isExpPPDomain"},
   CellID->309264225]},
 "OrderedPP isExpPPDomain"->{
  Cell[261422, 7417, 869, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:309264225", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isExpPPDomain"},
   CellID->309264225]},
 "ID:499712177"->{
  Cell[262294, 7442, 1196, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:499712177", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isDicksonDomain"},
   CellID->499712177]},
 "OrderedPP isDicksonDomain"->{
  Cell[262294, 7442, 1196, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:499712177", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isDicksonDomain"},
   CellID->499712177]},
 "ID:275069792"->{
  Cell[263493, 7478, 720, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:275069792", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->275069792]},
 "ID:834856233"->{
  Cell[264216, 7499, 886, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:834856233", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isTermOrder"},
   CellID->834856233]},
 "OrderedPP isTermOrder"->{
  Cell[264216, 7499, 886, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:834856233", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isTermOrder"},
   CellID->834856233]},
 "ID:260893370"->{
  Cell[265105, 7525, 1019, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:260893370", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isCommPPDomain"},
   CellID->260893370]},
 "OrderedPP isCommPPDomain"->{
  Cell[265105, 7525, 1019, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:260893370", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isCommPPDomain"},
   CellID->260893370]},
 "Proof|ID:309264225"->{
  Cell[266280, 7561, 2640, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:309264225", "Proof|ID:309264225-1"},
   CellID->51410204]},
 "Proof|ID:309264225-1"->{
  Cell[266280, 7561, 2640, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:309264225", "Proof|ID:309264225-1"},
   CellID->51410204]},
 "Proof|ID:499712177"->{
  Cell[269438, 7654, 3230, 86, 70, "ProofInfo",
   CellTags->{"Proof|ID:499712177", "Proof|ID:499712177-1"},
   CellID->50025897]},
 "Proof|ID:499712177-1"->{
  Cell[269438, 7654, 3230, 86, 70, "ProofInfo",
   CellTags->{"Proof|ID:499712177", "Proof|ID:499712177-1"},
   CellID->50025897]},
 "Proof|ID:834856233"->{
  Cell[273188, 7759, 3324, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:834856233", "Proof|ID:834856233-1"},
   CellID->65135779]},
 "Proof|ID:834856233-1"->{
  Cell[273188, 7759, 3324, 90, 70, "ProofInfo",
   CellTags->{"Proof|ID:834856233", "Proof|ID:834856233-1"},
   CellID->65135779]},
 "Proof|ID:260893370"->{
  Cell[277032, 7868, 3156, 86, 70, "ProofInfo",
   CellTags->{"Proof|ID:260893370", "Proof|ID:260893370-1"},
   CellID->97936963]},
 "Proof|ID:260893370-1"->{
  Cell[277032, 7868, 3156, 86, 70, "ProofInfo",
   CellTags->{"Proof|ID:260893370", "Proof|ID:260893370-1"},
   CellID->97936963]},
 "ID:223902051"->{
  Cell[281789, 8011, 2379, 71, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:223902051", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex"},
   CellID->223902051]},
 "Lex"->{
  Cell[281789, 8011, 2379, 71, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:223902051", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex"},
   CellID->223902051]},
 "ID:376155678"->{
  Cell[284659, 8102, 703, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:376155678", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->376155678]},
 "ID:310375877"->{
  Cell[285922, 8140, 773, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:310375877", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->310375877]},
 "ID:463764498"->{
  Cell[286698, 8162, 989, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:463764498", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex total"},
   CellID->463764498]},
 "Lex total"->{
  Cell[286698, 8162, 989, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:463764498", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex total"},
   CellID->463764498]},
 "ID:6465370"->{
  Cell[287690, 8191, 1249, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:6465370", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex asymmetric"},
   CellID->6465370]},
 "Lex asymmetric"->{
  Cell[287690, 8191, 1249, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:6465370", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex asymmetric"},
   CellID->6465370]},
 "ID:78173747"->{
  Cell[288942, 8232, 1188, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:78173747", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex transitive"},
   CellID->78173747]},
 "Lex transitive"->{
  Cell[288942, 8232, 1188, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:78173747", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex transitive"},
   CellID->78173747]},
 "ID:341999039"->{
  Cell[290133, 8269, 971, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:341999039", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex 1 min"},
   CellID->341999039]},
 "Lex 1 min"->{
  Cell[290133, 8269, 971, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:341999039", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex 1 min"},
   CellID->341999039]},
 "ID:34516336"->{
  Cell[291107, 8298, 1079, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:34516336", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex \[RawStar]"},
   CellID->34516336]},
 "Lex \[RawStar]"->{
  Cell[291107, 8298, 1079, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:34516336", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex \[RawStar]"},
   CellID->34516336]},
 "Proof|ID:463764498"->{
  Cell[292341, 8340, 4370, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:463764498", "Proof|ID:463764498-1"},
   CellID->355435549]},
 "Proof|ID:463764498-1"->{
  Cell[292341, 8340, 4370, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:463764498", "Proof|ID:463764498-1"},
   CellID->355435549]},
 "Proof|ID:6465370"->{
  Cell[297298, 8476, 4844, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:6465370", "Proof|ID:6465370-1"},
   CellID->258075395]},
 "Proof|ID:6465370-1"->{
  Cell[297298, 8476, 4844, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:6465370", "Proof|ID:6465370-1"},
   CellID->258075395]},
 "Proof|ID:78173747"->{
  Cell[302730, 8622, 5194, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:78173747", "Proof|ID:78173747-1"},
   CellID->631231714]},
 "Proof|ID:78173747-1"->{
  Cell[302730, 8622, 5194, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:78173747", "Proof|ID:78173747-1"},
   CellID->631231714]},
 "Proof|ID:341999039"->{
  Cell[308511, 8774, 3599, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:341999039", "Proof|ID:341999039-1"},
   CellID->366853708]},
 "Proof|ID:341999039-1"->{
  Cell[308511, 8774, 3599, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:341999039", "Proof|ID:341999039-1"},
   CellID->366853708]},
 "Proof|ID:34516336"->{
  Cell[312630, 8891, 5250, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:34516336", "Proof|ID:34516336-1"},
   CellID->188021315]},
 "Proof|ID:34516336-1"->{
  Cell[312630, 8891, 5250, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:34516336", "Proof|ID:34516336-1"},
   CellID->188021315]},
 "ID:414199224"->{
  Cell[318896, 9057, 861, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:414199224", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex isTermOrder"},
   CellID->414199224]},
 "Lex isTermOrder"->{
  Cell[318896, 9057, 861, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:414199224", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex isTermOrder"},
   CellID->414199224]},
 "Proof|ID:414199224"->{
  Cell[319912, 9091, 2558, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:414199224", "Proof|ID:414199224-1"},
   CellID->87459975]},
 "Proof|ID:414199224-1"->{
  Cell[319912, 9091, 2558, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:414199224", "Proof|ID:414199224-1"},
   CellID->87459975]},
 "ID:116320630"->{
  Cell[323832, 9213, 2280, 72, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:116320630", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex"},
   CellID->116320630]},
 "DegLex"->{
  Cell[323832, 9213, 2280, 72, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:116320630", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex"},
   CellID->116320630]},
 "ID:106786369"->{
  Cell[326511, 9304, 703, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:106786369", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->106786369]},
 "ID:43019543"->{
  Cell[327729, 9342, 767, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:43019543", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->43019543]},
 "ID:75557233"->{
  Cell[328499, 9364, 998, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:75557233", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex total"},
   CellID->75557233]},
 "DegLex total"->{
  Cell[328499, 9364, 998, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:75557233", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex total"},
   CellID->75557233]},
 "ID:49531328"->{
  Cell[329500, 9393, 1267, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:49531328", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex asymmetric"},
   CellID->49531328]},
 "DegLex asymmetric"->{
  Cell[329500, 9393, 1267, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:49531328", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex asymmetric"},
   CellID->49531328]},
 "ID:260398187"->{
  Cell[330770, 9434, 1208, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:260398187", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex transitive"},
   CellID->260398187]},
 "DegLex transitive"->{
  Cell[330770, 9434, 1208, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:260398187", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex transitive"},
   CellID->260398187]},
 "ID:59133109"->{
  Cell[331981, 9471, 977, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:59133109", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex 1 min"},
   CellID->59133109]},
 "DegLex 1 min"->{
  Cell[331981, 9471, 977, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:59133109", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex 1 min"},
   CellID->59133109]},
 "ID:572110615"->{
  Cell[332961, 9500, 1096, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:572110615", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex \[RawStar]"},
   CellID->572110615]},
 "DegLex \[RawStar]"->{
  Cell[332961, 9500, 1096, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:572110615", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex \[RawStar]"},
   CellID->572110615]},
 "Proof|ID:75557233"->{
  Cell[334213, 9542, 4383, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:75557233", "Proof|ID:75557233-1"},
   CellID->210478017]},
 "Proof|ID:75557233-1"->{
  Cell[334213, 9542, 4383, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:75557233", "Proof|ID:75557233-1"},
   CellID->210478017]},
 "Proof|ID:49531328"->{
  Cell[339180, 9678, 4867, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:49531328", "Proof|ID:49531328-1"},
   CellID->563764769]},
 "Proof|ID:49531328-1"->{
  Cell[339180, 9678, 4867, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:49531328", "Proof|ID:49531328-1"},
   CellID->563764769]},
 "Proof|ID:260398187"->{
  Cell[344636, 9824, 5223, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:260398187", "Proof|ID:260398187-1"},
   CellID->100109944]},
 "Proof|ID:260398187-1"->{
  Cell[344636, 9824, 5223, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:260398187", "Proof|ID:260398187-1"},
   CellID->100109944]},
 "Proof|ID:59133109"->{
  Cell[350447, 9976, 3606, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:59133109", "Proof|ID:59133109-1"},
   CellID->171486032]},
 "Proof|ID:59133109-1"->{
  Cell[350447, 9976, 3606, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:59133109", "Proof|ID:59133109-1"},
   CellID->171486032]},
 "Proof|ID:572110615"->{
  Cell[354612, 10094, 5273, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:572110615", "Proof|ID:572110615-1"},
   CellID->427513857]},
 "Proof|ID:572110615-1"->{
  Cell[354612, 10094, 5273, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:572110615", "Proof|ID:572110615-1"},
   CellID->427513857]},
 "ID:463127740"->{
  Cell[360909, 10260, 869, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:463127740", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex isTermOrder"},
   CellID->463127740]},
 "DegLex isTermOrder"->{
  Cell[360909, 10260, 869, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:463127740", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex isTermOrder"},
   CellID->463127740]},
 "Proof|ID:463127740"->{
  Cell[361933, 10294, 2571, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:463127740", "Proof|ID:463127740-1"},
   CellID->102580793]},
 "Proof|ID:463127740-1"->{
  Cell[361933, 10294, 2571, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:463127740", "Proof|ID:463127740-1"},
   CellID->102580793]},
 "ID:267616510"->{
  Cell[365945, 10421, 693, 17, 70, "GlobalDeclaration",
   CellTags->{
    "ID:267616510", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->267616510]},
 "ID:1929370"->{
  Cell[366641, 10440, 2155, 64, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1929370", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex"},
   CellID->1929370]},
 "RevLex"->{
  Cell[366641, 10440, 2155, 64, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1929370", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex"},
   CellID->1929370]},
 "ID:116917210"->{
  Cell[368799, 10506, 1994, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:116917210", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex"},
   CellID->116917210]},
 "DegRevLex"->{
  Cell[368799, 10506, 1994, 62, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:116917210", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex"},
   CellID->116917210]},
 "ID:431887307"->{
  Cell[371200, 10587, 701, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:431887307", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->431887307]},
 "ID:204943512"->{
  Cell[372396, 10624, 769, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:204943512", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->204943512]},
 "ID:781115991"->{
  Cell[373168, 10646, 1001, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:781115991", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex total"},
   CellID->781115991]},
 "RevLex total"->{
  Cell[373168, 10646, 1001, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:781115991", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex total"},
   CellID->781115991]},
 "ID:2983958"->{
  Cell[374172, 10675, 1264, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2983958", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex asymmetric"},
   CellID->2983958]},
 "RevLex asymmetric"->{
  Cell[374172, 10675, 1264, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2983958", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex asymmetric"},
   CellID->2983958]},
 "ID:240564679"->{
  Cell[375439, 10716, 1208, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:240564679", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex transitive"},
   CellID->240564679]},
 "RevLex transitive"->{
  Cell[375439, 10716, 1208, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:240564679", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex transitive"},
   CellID->240564679]},
 "ID:299677825"->{
  Cell[376650, 10753, 1092, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:299677825", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex \[RawStar]"},
   CellID->299677825]},
 "RevLex \[RawStar]"->{
  Cell[376650, 10753, 1092, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:299677825", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex \[RawStar]"},
   CellID->299677825]},
 "Proof|ID:781115991"->{
  Cell[377897, 10795, 4388, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:781115991", "Proof|ID:781115991-1"},
   CellID->124604986]},
 "Proof|ID:781115991-1"->{
  Cell[377897, 10795, 4388, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:781115991", "Proof|ID:781115991-1"},
   CellID->124604986]},
 "Proof|ID:2983958"->{
  Cell[382805, 10930, 4861, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:2983958", "Proof|ID:2983958-1"},
   CellID->83920353]},
 "Proof|ID:2983958-1"->{
  Cell[382805, 10930, 4861, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:2983958", "Proof|ID:2983958-1"},
   CellID->83920353]},
 "Proof|ID:240564679"->{
  Cell[388184, 11075, 5223, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:240564679", "Proof|ID:240564679-1"},
   CellID->163719088]},
 "Proof|ID:240564679-1"->{
  Cell[388184, 11075, 5223, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:240564679", "Proof|ID:240564679-1"},
   CellID->163719088]},
 "Proof|ID:299677825"->{
  Cell[393926, 11226, 5273, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:299677825", "Proof|ID:299677825-1"},
   CellID->464155749]},
 "Proof|ID:299677825-1"->{
  Cell[393926, 11226, 5273, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:299677825", "Proof|ID:299677825-1"},
   CellID->464155749]},
 "ID:449289304"->{
  Cell[400183, 11392, 773, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:449289304", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->449289304]},
 "ID:240154468"->{
  Cell[400959, 11414, 1042, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:240154468", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex total"},
   CellID->240154468]},
 "DegRevLex total"->{
  Cell[400959, 11414, 1042, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:240154468", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex total"},
   CellID->240154468]},
 "ID:808110452"->{
  Cell[402004, 11444, 1311, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:808110452", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex asymmetric"},
   CellID->808110452]},
 "DegRevLex asymmetric"->{
  Cell[402004, 11444, 1311, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:808110452", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex asymmetric"},
   CellID->808110452]},
 "ID:224322863"->{
  Cell[403318, 11486, 1250, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:224322863", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex transitive"},
   CellID->224322863]},
 "DegRevLex transitive"->{
  Cell[403318, 11486, 1250, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:224322863", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex transitive"},
   CellID->224322863]},
 "ID:424804407"->{
  Cell[404571, 11524, 991, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:424804407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex 1 min"},
   CellID->424804407]},
 "DegRevLex 1 min"->{
  Cell[404571, 11524, 991, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:424804407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex 1 min"},
   CellID->424804407]},
 "ID:308108885"->{
  Cell[405565, 11553, 1108, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:308108885", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex \[RawStar]"},
   CellID->308108885]},
 "DegRevLex \[RawStar]"->{
  Cell[405565, 11553, 1108, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:308108885", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex \[RawStar]"},
   CellID->308108885]},
 "Proof|ID:240154468"->{
  Cell[406826, 11595, 4406, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:240154468", "Proof|ID:240154468-1"},
   CellID->162210883]},
 "Proof|ID:240154468-1"->{
  Cell[406826, 11595, 4406, 116, 70, "ProofInfo",
   CellTags->{"Proof|ID:240154468", "Proof|ID:240154468-1"},
   CellID->162210883]},
 "Proof|ID:808110452"->{
  Cell[411751, 11730, 4889, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:808110452", "Proof|ID:808110452-1"},
   CellID->14371058]},
 "Proof|ID:808110452-1"->{
  Cell[411751, 11730, 4889, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:808110452", "Proof|ID:808110452-1"},
   CellID->14371058]},
 "Proof|ID:224322863"->{
  Cell[417160, 11875, 5247, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:224322863", "Proof|ID:224322863-1"},
   CellID->483386676]},
 "Proof|ID:224322863-1"->{
  Cell[417160, 11875, 5247, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:224322863", "Proof|ID:224322863-1"},
   CellID->483386676]},
 "Proof|ID:424804407"->{
  Cell[422927, 12026, 3623, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:424804407", "Proof|ID:424804407-1"},
   CellID->138222520]},
 "Proof|ID:424804407-1"->{
  Cell[422927, 12026, 3623, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:424804407", "Proof|ID:424804407-1"},
   CellID->138222520]},
 "Proof|ID:308108885"->{
  Cell[427070, 12143, 5291, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:308108885", "Proof|ID:308108885-1"},
   CellID->152180612]},
 "Proof|ID:308108885-1"->{
  Cell[427070, 12143, 5291, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:308108885", "Proof|ID:308108885-1"},
   CellID->152180612]},
 "ID:936963717"->{
  Cell[433321, 12308, 878, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:936963717", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex isTermOrder"},
   CellID->936963717]},
 "DegRevLex isTermOrder"->{
  Cell[433321, 12308, 878, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:936963717", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex isTermOrder"},
   CellID->936963717]},
 "Proof|ID:936963717"->{
  Cell[434354, 12342, 2583, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:936963717", "Proof|ID:936963717-1"},
   CellID->319190197]},
 "Proof|ID:936963717-1"->{
  Cell[434354, 12342, 2583, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:936963717", "Proof|ID:936963717-1"},
   CellID->319190197]},
 "ID:44994065"->{
  Cell[437966, 12452, 716, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:44994065", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->44994065]},
 "ID:1997108144"->{
  Cell[439493, 12499, 758, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1997108144", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->1997108144]},
 "ID:608617422"->{
  Cell[440254, 12520, 926, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:608617422", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->608617422]},
 "ID:271011394"->{
  Cell[441183, 12544, 1676, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:271011394", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples \[Element]"},
   CellID->271011394]},
 "PPTuples \[Element]"->{
  Cell[441183, 12544, 1676, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:271011394", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples \[Element]"},
   CellID->271011394]},
 "ID:1820126795"->{
  Cell[442862, 12595, 953, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1820126795", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples 1"},
   CellID->1820126795]},
 "PPTuples 1"->{
  Cell[442862, 12595, 953, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1820126795", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples 1"},
   CellID->1820126795]},
 "ID:281355476"->{
  Cell[443818, 12624, 774, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:281355476", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples dim"},
   CellID->281355476]},
 "PPTuples dim"->{
  Cell[443818, 12624, 774, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:281355476", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples dim"},
   CellID->281355476]},
 "ID:2109519396"->{
  Cell[444595, 12647, 1083, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2109519396", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples deg"},
   CellID->2109519396]},
 "PPTuples deg"->{
  Cell[444595, 12647, 1083, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2109519396", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples deg"},
   CellID->2109519396]},
 "ID:1563833422"->{
  Cell[445681, 12680, 901, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1563833422", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples exp"},
   CellID->1563833422]},
 "PPTuples exp"->{
  Cell[445681, 12680, 901, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1563833422", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples exp"},
   CellID->1563833422]},
 "ID:1180963317"->{
  Cell[446585, 12707, 1052, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1180963317", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples \[RawStar]"},
   CellID->1180963317]},
 "PPTuples \[RawStar]"->{
  Cell[446585, 12707, 1052, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1180963317", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples \[RawStar]"},
   CellID->1180963317]},
 "ID:1392702003"->{
  Cell[447640, 12740, 1050, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1392702003", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples /"},
   CellID->1392702003]},
 "PPTuples /"->{
  Cell[447640, 12740, 1050, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1392702003", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples /"},
   CellID->1392702003]},
 "ID:1587875066"->{
  Cell[448693, 12773, 1292, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1587875066", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples \[VerticalBar]"},
   CellID->1587875066]},
 "PPTuples \[VerticalBar]"->{
  Cell[448693, 12773, 1292, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1587875066", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples \[VerticalBar]"},
   CellID->1587875066]},
 "ID:803590963"->{
  Cell[449988, 12815, 1418, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:803590963", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples lcm"},
   CellID->803590963]},
 "PPTuples lcm"->{
  Cell[449988, 12815, 1418, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:803590963", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples lcm"},
   CellID->803590963]},
 "ID:106372713"->{
  Cell[451717, 12863, 1127, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:106372713", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples <<"},
   CellID->106372713]},
 "PPTuples <<"->{
  Cell[451717, 12863, 1127, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:106372713", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples <<"},
   CellID->106372713]},
 "ID:1589980318"->{
  Cell[452847, 12896, 1037, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1589980318", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples >>"},
   CellID->1589980318]},
 "PPTuples >>"->{
  Cell[452847, 12896, 1037, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1589980318", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples >>"},
   CellID->1589980318]},
 "ID:61015536"->{
  Cell[454270, 12945, 712, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:61015536", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->61015536]},
 "ID:617800164"->{
  Cell[455471, 12982, 1647, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:617800164", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PPTuples alternative", "IR"},
   CellID->617800164]},
 "\[Element] PPTuples alternative"->{
  Cell[455471, 12982, 1647, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:617800164", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PPTuples alternative", "IR"},
   CellID->617800164]},
 "ID:41499157"->{
  Cell[457121, 13034, 787, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:41499157", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples isDomain"},
   CellID->41499157]},
 "PPTuples isDomain"->{
  Cell[457121, 13034, 787, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:41499157", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples isDomain"},
   CellID->41499157]},
 "ID:142297858"->{
  Cell[457911, 13057, 830, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:142297858", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasDecisionPredicate"},
   CellID->142297858]},
 "PPTuples hasDecisionPredicate"->{
  Cell[457911, 13057, 830, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:142297858", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasDecisionPredicate"},
   CellID->142297858]},
 "ID:201916088"->{
  Cell[458744, 13080, 823, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:201916088", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasExponentFunction"},
   CellID->201916088]},
 "PPTuples hasExponentFunction"->{
  Cell[458744, 13080, 823, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:201916088", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasExponentFunction"},
   CellID->201916088]},
 "ID:102507049"->{
  Cell[459570, 13103, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:102507049", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples isExtensional"},
   CellID->102507049]},
 "PPTuples isExtensional"->{
  Cell[459570, 13103, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:102507049", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples isExtensional"},
   CellID->102507049]},
 "ID:93844102"->{
  Cell[460380, 13126, 818, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:93844102", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasDegreeFunction"},
   CellID->93844102]},
 "PPTuples hasDegreeFunction"->{
  Cell[460380, 13126, 818, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:93844102", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasDegreeFunction"},
   CellID->93844102]},
 "ID:34803991"->{
  Cell[461201, 13149, 860, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:34803991", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasDivisibility"},
   CellID->34803991]},
 "PPTuples hasDivisibility"->{
  Cell[461201, 13149, 860, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:34803991", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasDivisibility"},
   CellID->34803991]},
 "Proof|ID:617800164"->{
  Cell[462217, 13182, 4073, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:617800164", "Proof|ID:617800164-1"},
   CellID->427612524]},
 "Proof|ID:617800164-1"->{
  Cell[462217, 13182, 4073, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:617800164", "Proof|ID:617800164-1"},
   CellID->427612524]},
 "Proof|ID:41499157"->{
  Cell[466810, 13309, 2372, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:41499157", "Proof|ID:41499157-1"},
   CellID->394527339]},
 "Proof|ID:41499157-1"->{
  Cell[466810, 13309, 2372, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:41499157", "Proof|ID:41499157-1"},
   CellID->394527339]},
 "Proof|ID:142297858"->{
  Cell[469701, 13398, 2429, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:142297858", "Proof|ID:142297858-1"},
   CellID->706958866]},
 "Proof|ID:142297858-1"->{
  Cell[469701, 13398, 2429, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:142297858", "Proof|ID:142297858-1"},
   CellID->706958866]},
 "Proof|ID:201916088"->{
  Cell[472650, 13487, 2433, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:201916088", "Proof|ID:201916088-1"},
   CellID->220074858]},
 "Proof|ID:201916088-1"->{
  Cell[472650, 13487, 2433, 71, 70, "ProofInfo",
   CellTags->{"Proof|ID:201916088", "Proof|ID:201916088-1"},
   CellID->220074858]},
 "Proof|ID:102507049"->{
  Cell[475603, 13577, 2401, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:102507049", "Proof|ID:102507049-1"},
   CellID->628960870]},
 "Proof|ID:102507049-1"->{
  Cell[475603, 13577, 2401, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:102507049", "Proof|ID:102507049-1"},
   CellID->628960870]},
 "Proof|ID:93844102"->{
  Cell[478524, 13666, 2411, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:93844102", "Proof|ID:93844102-1"},
   CellID->78679481]},
 "Proof|ID:93844102-1"->{
  Cell[478524, 13666, 2411, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:93844102", "Proof|ID:93844102-1"},
   CellID->78679481]},
 "Proof|ID:34803991"->{
  Cell[481454, 13755, 2403, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:34803991", "Proof|ID:34803991-1"},
   CellID->42217056]},
 "Proof|ID:34803991-1"->{
  Cell[481454, 13755, 2403, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:34803991", "Proof|ID:34803991-1"},
   CellID->42217056]},
 "ID:338277855"->{
  Cell[484890, 13859, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:338277855", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples isExpPPDomain"},
   CellID->338277855]},
 "PPTuples isExpPPDomain"->{
  Cell[484890, 13859, 807, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:338277855", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples isExpPPDomain"},
   CellID->338277855]},
 "Proof|ID:338277855"->{
  Cell[485852, 13891, 2397, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:338277855", "Proof|ID:338277855-1"},
   CellID->630181205]},
 "Proof|ID:338277855-1"->{
  Cell[485852, 13891, 2397, 70, 70, "ProofInfo",
   CellTags->{"Proof|ID:338277855", "Proof|ID:338277855-1"},
   CellID->630181205]},
 "ID:611122547"->{
  Cell[495075, 14188, 873, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:611122547", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->611122547]},
 "ID:1216459107"->{
  Cell[495951, 14211, 1008, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1216459107", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->1216459107]},
 "ID:594519699"->{
  Cell[496962, 14236, 1327, 29, 70, "GlobalDeclaration",
   CellTags->{
    "ID:594519699", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->594519699]},
 "ID:65019775"->{
  Cell[498292, 14267, 3917, 107, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:65019775", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Element]"},
   CellID->65019775]},
 "PolyTuples \[Element]"->{
  Cell[498292, 14267, 3917, 107, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:65019775", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Element]"},
   CellID->65019775]},
 "ID:399496924"->{
  Cell[502212, 14376, 938, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:399496924", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C base"},
   CellID->399496924]},
 "PolyTuples C base"->{
  Cell[502212, 14376, 938, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:399496924", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C base"},
   CellID->399496924]},
 "ID:983317445"->{
  Cell[503153, 14404, 1544, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:983317445", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C rec"},
   CellID->983317445]},
 "PolyTuples C rec"->{
  Cell[503153, 14404, 1544, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:983317445", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C rec"},
   CellID->983317445]},
 "ID:1360683691"->{
  Cell[504700, 14451, 859, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1360683691", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples 0"},
   CellID->1360683691]},
 "PolyTuples 0"->{
  Cell[504700, 14451, 859, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1360683691", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples 0"},
   CellID->1360683691]},
 "ID:13289303"->{
  Cell[505562, 14478, 1594, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:13289303", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples 1"},
   CellID->13289303]},
 "PolyTuples 1"->{
  Cell[505562, 14478, 1594, 40, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:13289303", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples 1"},
   CellID->13289303]},
 "ID:1777247200"->{
  Cell[507159, 14520, 982, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1777247200", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + 1"},
   CellID->1777247200]},
 "PolyTuples + 1"->{
  Cell[507159, 14520, 982, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1777247200", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + 1"},
   CellID->1777247200]},
 "ID:2122267989"->{
  Cell[508144, 14548, 1399, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2122267989", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + 2"},
   CellID->2122267989]},
 "PolyTuples + 2"->{
  Cell[508144, 14548, 1399, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2122267989", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + 2"},
   CellID->2122267989]},
 "ID:290578418"->{
  Cell[509546, 14586, 4077, 109, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:290578418", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + 3"},
   CellID->290578418]},
 "PolyTuples + 3"->{
  Cell[509546, 14586, 4077, 109, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:290578418", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + 3"},
   CellID->290578418]},
 "ID:1495569651"->{
  Cell[513626, 14697, 974, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1495569651", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - unary 1"},
   CellID->1495569651]},
 "PolyTuples - unary 1"->{
  Cell[513626, 14697, 974, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1495569651", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - unary 1"},
   CellID->1495569651]},
 "ID:1285234307"->{
  Cell[514603, 14726, 1529, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1285234307", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - unary 2"},
   CellID->1285234307]},
 "PolyTuples - unary 2"->{
  Cell[514603, 14726, 1529, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1285234307", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - unary 2"},
   CellID->1285234307]},
 "ID:551740538"->{
  Cell[516135, 14772, 915, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:551740538", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples -"},
   CellID->551740538]},
 "PolyTuples -"->{
  Cell[516135, 14772, 915, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:551740538", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples -"},
   CellID->551740538]},
 "ID:419928300"->{
  Cell[517053, 14801, 941, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:419928300", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[RawStar] 1"},
   CellID->419928300]},
 "PolyTuples \[RawStar] 1"->{
  Cell[517053, 14801, 941, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:419928300", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[RawStar] 1"},
   CellID->419928300]},
 "ID:209271439"->{
  Cell[517997, 14830, 1168, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:209271439", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[RawStar] 2"},
   CellID->209271439]},
 "PolyTuples \[RawStar] 2"->{
  Cell[517997, 14830, 1168, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:209271439", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[RawStar] 2"},
   CellID->209271439]},
 "ID:1411844363"->{
  Cell[519168, 14865, 3884, 99, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1411844363", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[RawStar] 3"},
   CellID->1411844363]},
 "PolyTuples \[RawStar] 3"->{
  Cell[519168, 14865, 3884, 99, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1411844363", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[RawStar] 3"},
   CellID->1411844363]},
 "ID:1128062400"->{
  Cell[523055, 14966, 1192, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1128062400", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Succeeds] 1"},
   CellID->1128062400]},
 "PolyTuples \[Succeeds] 1"->{
  Cell[523055, 14966, 1192, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1128062400", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Succeeds] 1"},
   CellID->1128062400]},
 "ID:1280515048"->{
  Cell[524250, 15004, 1420, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1280515048", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Succeeds] 2"},
   CellID->1280515048]},
 "PolyTuples \[Succeeds] 2"->{
  Cell[524250, 15004, 1420, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1280515048", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Succeeds] 2"},
   CellID->1280515048]},
 "ID:2095908261"->{
  Cell[525673, 15047, 2480, 72, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2095908261", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Succeeds] 3"},
   CellID->2095908261]},
 "PolyTuples \[Succeeds] 3"->{
  Cell[525673, 15047, 2480, 72, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2095908261", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Succeeds] 3"},
   CellID->2095908261]},
 "ID:1668822406"->{
  Cell[528156, 15121, 977, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1668822406", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples multN 1"},
   CellID->1668822406]},
 "PolyTuples multN 1"->{
  Cell[528156, 15121, 977, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1668822406", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples multN 1"},
   CellID->1668822406]},
 "ID:329725901"->{
  Cell[529136, 15149, 1270, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:329725901", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples multN 2"},
   CellID->329725901]},
 "PolyTuples multN 2"->{
  Cell[529136, 15149, 1270, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:329725901", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples multN 2"},
   CellID->329725901]},
 "ID:822409829"->{
  Cell[530409, 15184, 1587, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:822409829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples mult"},
   CellID->822409829]},
 "PolyTuples mult"->{
  Cell[530409, 15184, 1587, 46, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:822409829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples mult"},
   CellID->822409829]},
 "ID:922404041"->{
  Cell[531999, 15232, 1828, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:922404041", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples mult i"},
   CellID->922404041]},
 "PolyTuples mult i"->{
  Cell[531999, 15232, 1828, 51, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:922404041", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples mult i"},
   CellID->922404041]},
 "ID:2033419264"->{
  Cell[533830, 15285, 922, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2033419264", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples rdm 1"},
   CellID->2033419264]},
 "PolyTuples rdm 1"->{
  Cell[533830, 15285, 922, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2033419264", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples rdm 1"},
   CellID->2033419264]},
 "ID:1308197671"->{
  Cell[534755, 15313, 1149, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1308197671", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples rdm 2"},
   CellID->1308197671]},
 "PolyTuples rdm 2"->{
  Cell[534755, 15313, 1149, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1308197671", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples rdm 2"},
   CellID->1308197671]},
 "ID:1157498137"->{
  Cell[535907, 15346, 3023, 78, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1157498137", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples rdm 3"},
   CellID->1157498137]},
 "PolyTuples rdm 3"->{
  Cell[535907, 15346, 3023, 78, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1157498137", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples rdm 3"},
   CellID->1157498137]},
 "ID:103388547"->{
  Cell[538933, 15426, 984, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:103388547", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd 1"},
   CellID->103388547]},
 "PolyTuples lcrd 1"->{
  Cell[538933, 15426, 984, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:103388547", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd 1"},
   CellID->103388547]},
 "ID:166800573"->{
  Cell[539920, 15456, 1212, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:166800573", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd 2"},
   CellID->166800573]},
 "PolyTuples lcrd 2"->{
  Cell[539920, 15456, 1212, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:166800573", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd 2"},
   CellID->166800573]},
 "ID:319850711"->{
  Cell[541135, 15492, 4959, 122, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:319850711", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd 3"},
   CellID->319850711]},
 "PolyTuples lcrd 3"->{
  Cell[541135, 15492, 4959, 122, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:319850711", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd 3"},
   CellID->319850711]},
 "ID:112650434"->{
  Cell[546097, 15616, 942, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:112650434", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd unary 1"},
   CellID->112650434]},
 "PolyTuples lcrd unary 1"->{
  Cell[546097, 15616, 942, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:112650434", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd unary 1"},
   CellID->112650434]},
 "ID:1729493745"->{
  Cell[547042, 15644, 1695, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1729493745", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd unary 2"},
   CellID->1729493745]},
 "PolyTuples lcrd unary 2"->{
  Cell[547042, 15644, 1695, 45, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1729493745", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd unary 2"},
   CellID->1729493745]},
 "ID:414290689"->{
  Cell[548740, 15691, 1071, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:414290689", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples cpm 1"},
   CellID->414290689]},
 "PolyTuples cpm 1"->{
  Cell[548740, 15691, 1071, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:414290689", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples cpm 1"},
   CellID->414290689]},
 "ID:27385995"->{
  Cell[549814, 15723, 1296, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:27385995", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples cpm 2"},
   CellID->27385995]},
 "PolyTuples cpm 2"->{
  Cell[549814, 15723, 1296, 36, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:27385995", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples cpm 2"},
   CellID->27385995]},
 "ID:1444186163"->{
  Cell[551113, 15761, 4968, 126, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1444186163", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples cpm 3"},
   CellID->1444186163]},
 "PolyTuples cpm 3"->{
  Cell[551113, 15761, 4968, 126, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1444186163", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples cpm 3"},
   CellID->1444186163]},
 "ID:2018930893"->{
  Cell[556084, 15889, 1841, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2018930893", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples <<"},
   CellID->2018930893]},
 "PolyTuples <<"->{
  Cell[556084, 15889, 1841, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2018930893", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples <<"},
   CellID->2018930893]},
 "ID:1442847663"->{
  Cell[557928, 15940, 1503, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1442847663", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples >>"},
   CellID->1442847663]},
 "PolyTuples >>"->{
  Cell[557928, 15940, 1503, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1442847663", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples >>"},
   CellID->1442847663]},
 "ID:1486930793"->{
  Cell[560532, 16023, 1083, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1486930793", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalT]2"},
   CellID->1486930793]},
 "\[ScriptCapitalT]2"->{
  Cell[560532, 16023, 1083, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1486930793", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalT]2"},
   CellID->1486930793]},
 "ID:933000257"->{
  Cell[561618, 16053, 1080, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:933000257", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalT]3"},
   CellID->933000257]},
 "\[ScriptCapitalT]3"->{
  Cell[561618, 16053, 1080, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:933000257", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalT]3"},
   CellID->933000257]},
 "ID:1375788385"->{
  Cell[574360, 16449, 1030, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1375788385", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalQ]xyz"},
   CellID->1375788385]},
 "\[ScriptCapitalQ]xyz"->{
  Cell[574360, 16449, 1030, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:1375788385", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalQ]xyz"},
   CellID->1375788385]},
 "ID:395821466"->{
  Cell[614870, 17740, 777, 20, 32, "GlobalDeclaration",
   CellTags->{
    "ID:395821466", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->395821466]},
 "ID:689486298"->{
  Cell[616582, 17794, 923, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:689486298", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isTuple", "IR"},
   CellID->689486298]},
 "PolyTuples isTuple"->{
  Cell[616582, 17794, 923, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:689486298", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isTuple", "IR"},
   CellID->689486298]},
 "ID:302371753"->{
  Cell[617508, 17821, 1507, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:302371753", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples DomainTuples"},
   CellID->302371753]},
 "\[Element] PolyTuples DomainTuples"->{
  Cell[617508, 17821, 1507, 48, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:302371753", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples DomainTuples"},
   CellID->302371753]},
 "ID:619631752"->{
  Cell[619018, 17871, 912, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:619631752", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples empty", "IR"},
   CellID->619631752]},
 "\[Element] PolyTuples empty"->{
  Cell[619018, 17871, 912, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:619631752", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples empty", "IR"},
   CellID->619631752]},
 "ID:93154383"->{
  Cell[619933, 17897, 3154, 93, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:93154383", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples rec", "IR"},
   CellID->93154383]},
 "\[Element] PolyTuples rec"->{
  Cell[619933, 17897, 3154, 93, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:93154383", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples rec", "IR"},
   CellID->93154383]},
 "ID:110840922"->{
  Cell[623090, 17992, 3179, 91, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:110840922", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples rec II", "IR"},
   CellID->110840922]},
 "\[Element] PolyTuples rec II"->{
  Cell[623090, 17992, 3179, 91, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:110840922", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples rec II", "IR"},
   CellID->110840922]},
 "ID:45787564"->{
  Cell[626272, 18085, 3467, 100, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:45787564", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples rec III", "IR"},
   CellID->45787564]},
 "\[Element] PolyTuples rec III"->{
  Cell[626272, 18085, 3467, 100, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:45787564", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples rec III", "IR"},
   CellID->45787564]},
 "ID:34480440"->{
  Cell[629742, 18187, 707, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:34480440", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->34480440]},
 "ID:134755829"->{
  Cell[630452, 18208, 2714, 80, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:134755829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples prepend", "IR"},
   CellID->134755829]},
 "\[Element] PolyTuples prepend"->{
  Cell[630452, 18208, 2714, 80, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:134755829", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples prepend", "IR"},
   CellID->134755829]},
 "ID:218126712"->{
  Cell[633169, 18290, 1120, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:218126712", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership first", "IR"},
   CellID->218126712]},
 "PolyTuples membership first"->{
  Cell[633169, 18290, 1120, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:218126712", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership first", "IR"},
   CellID->218126712]},
 "ID:36095185"->{
  Cell[634292, 18323, 1117, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:36095185", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership second", "IR"},
   CellID->36095185]},
 "PolyTuples membership second"->{
  Cell[634292, 18323, 1117, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:36095185", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership second", "IR"},
   CellID->36095185]},
 "ID:266402037"->{
  Cell[635412, 18356, 3357, 89, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:266402037", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples shape"},
   CellID->266402037]},
 "\[Element] PolyTuples shape"->{
  Cell[635412, 18356, 3357, 89, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:266402037", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples shape"},
   CellID->266402037]},
 "ID:314594392"->{
  Cell[638772, 18447, 1900, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:314594392", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples length=1"},
   CellID->314594392]},
 "PolyTuples length=1"->{
  Cell[638772, 18447, 1900, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:314594392", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples length=1"},
   CellID->314594392]},
 "Proof|ID:689486298"->{
  Cell[640827, 18518, 2537, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:689486298", "Proof|ID:689486298-1"},
   CellID->851460394]},
 "Proof|ID:689486298-1"->{
  Cell[640827, 18518, 2537, 74, 70, "ProofInfo",
   CellTags->{"Proof|ID:689486298", "Proof|ID:689486298-1"},
   CellID->851460394]},
 "Proof|ID:302371753"->{
  Cell[643883, 18611, 3187, 89, 70, "ProofInfo",
   CellTags->{"Proof|ID:302371753", "Proof|ID:302371753-1"},
   CellID->426132571]},
 "Proof|ID:302371753-1"->{
  Cell[643883, 18611, 3187, 89, 70, "ProofInfo",
   CellTags->{"Proof|ID:302371753", "Proof|ID:302371753-1"},
   CellID->426132571]},
 "Proof|ID:619631752"->{
  Cell[647590, 18719, 2171, 64, 70, "ProofInfo",
   CellTags->{"Proof|ID:619631752", "Proof|ID:619631752-1"},
   CellID->236569764]},
 "Proof|ID:619631752-1"->{
  Cell[647590, 18719, 2171, 64, 70, "ProofInfo",
   CellTags->{"Proof|ID:619631752", "Proof|ID:619631752-1"},
   CellID->236569764]},
 "Proof|ID:93154383"->{
  Cell[650281, 18802, 6811, 170, 70, "ProofInfo",
   CellTags->{"Proof|ID:93154383", "Proof|ID:93154383-1"},
   CellID->290123111]},
 "Proof|ID:93154383-1"->{
  Cell[650281, 18802, 6811, 170, 70, "ProofInfo",
   CellTags->{"Proof|ID:93154383", "Proof|ID:93154383-1"},
   CellID->290123111]},
 "Proof|ID:110840922"->{
  Cell[657611, 18991, 7380, 177, 70, "ProofInfo",
   CellTags->{"Proof|ID:110840922", "Proof|ID:110840922-1"},
   CellID->99121591]},
 "Proof|ID:110840922-1"->{
  Cell[657611, 18991, 7380, 177, 70, "ProofInfo",
   CellTags->{"Proof|ID:110840922", "Proof|ID:110840922-1"},
   CellID->99121591]},
 "Proof|ID:45787564"->{
  Cell[665511, 19187, 7690, 184, 70, "ProofInfo",
   CellTags->{"Proof|ID:45787564", "Proof|ID:45787564-1"},
   CellID->483935740]},
 "Proof|ID:45787564-1"->{
  Cell[665511, 19187, 7690, 184, 70, "ProofInfo",
   CellTags->{"Proof|ID:45787564", "Proof|ID:45787564-1"},
   CellID->483935740]},
 "Proof|ID:134755829"->{
  Cell[673719, 19390, 6085, 148, 70, "ProofInfo",
   CellTags->{"Proof|ID:134755829", "Proof|ID:134755829-1"},
   CellID->955362936]},
 "Proof|ID:134755829-1"->{
  Cell[673719, 19390, 6085, 148, 70, "ProofInfo",
   CellTags->{"Proof|ID:134755829", "Proof|ID:134755829-1"},
   CellID->955362936]},
 "Proof|ID:218126712"->{
  Cell[680323, 19557, 3492, 96, 70, "ProofInfo",
   CellTags->{"Proof|ID:218126712", "Proof|ID:218126712-1"},
   CellID->69811972]},
 "Proof|ID:218126712-1"->{
  Cell[680323, 19557, 3492, 96, 70, "ProofInfo",
   CellTags->{"Proof|ID:218126712", "Proof|ID:218126712-1"},
   CellID->69811972]},
 "Proof|ID:36095185"->{
  Cell[684335, 19672, 3497, 97, 70, "ProofInfo",
   CellTags->{"Proof|ID:36095185", "Proof|ID:36095185-1"},
   CellID->66892996]},
 "Proof|ID:36095185-1"->{
  Cell[684335, 19672, 3497, 97, 70, "ProofInfo",
   CellTags->{"Proof|ID:36095185", "Proof|ID:36095185-1"},
   CellID->66892996]},
 "Proof|ID:266402037"->{
  Cell[688351, 19788, 7807, 186, 70, "ProofInfo",
   CellTags->{"Proof|ID:266402037", "Proof|ID:266402037-1"},
   CellID->167245154]},
 "Proof|ID:266402037-1"->{
  Cell[688351, 19788, 7807, 186, 70, "ProofInfo",
   CellTags->{"Proof|ID:266402037", "Proof|ID:266402037-1"},
   CellID->167245154]},
 "Proof|ID:314594392"->{
  Cell[696676, 19993, 4921, 125, 70, "ProofInfo",
   CellTags->{"Proof|ID:314594392", "Proof|ID:314594392-1"},
   CellID->424017657]},
 "Proof|ID:314594392-1"->{
  Cell[696676, 19993, 4921, 125, 70, "ProofInfo",
   CellTags->{"Proof|ID:314594392", "Proof|ID:314594392-1"},
   CellID->424017657]},
 "ID:441087600"->{
  Cell[702890, 20161, 642, 16, 70, "GlobalDeclaration",
   CellTags->{
    "ID:441087600", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->441087600]},
 "ID:678316554"->{
  Cell[703535, 20179, 1924, 61, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:678316554", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[ForAll] alternative"},
   CellID->678316554]},
 "PolyTuples \[ForAll] alternative"->{
  Cell[703535, 20179, 1924, 61, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:678316554", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[ForAll] alternative"},
   CellID->678316554]},
 "ID:329976481"->{
  Cell[705462, 20242, 3708, 107, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:329976481", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples induction", "IR"},
   CellID->329976481]},
 "PolyTuples induction"->{
  Cell[705462, 20242, 3708, 107, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:329976481", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples induction", "IR"},
   CellID->329976481]},
 "ID:379661071"->{
  Cell[709772, 20370, 4150, 117, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:379661071", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples induction'", "IR"},
   CellID->379661071]},
 "PolyTuples induction'"->{
  Cell[709772, 20370, 4150, 117, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:379661071", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples induction'", "IR"},
   CellID->379661071]},
 "Proof|ID:678316554"->{
  Cell[714077, 20498, 5067, 131, 70, "ProofInfo",
   CellTags->{"Proof|ID:678316554", "Proof|ID:678316554-1"},
   CellID->716699503]},
 "Proof|ID:678316554-1"->{
  Cell[714077, 20498, 5067, 131, 70, "ProofInfo",
   CellTags->{"Proof|ID:678316554", "Proof|ID:678316554-1"},
   CellID->716699503]},
 "Proof|ID:329976481"->{
  Cell[719664, 20648, 9347, 218, 70, "ProofInfo",
   CellTags->{"Proof|ID:329976481", "Proof|ID:329976481-1"},
   CellID->3050184]},
 "Proof|ID:329976481-1"->{
  Cell[719664, 20648, 9347, 218, 70, "ProofInfo",
   CellTags->{"Proof|ID:329976481", "Proof|ID:329976481-1"},
   CellID->3050184]},
 "Proof|ID:379661071"->{
  Cell[729530, 20885, 10825, 248, 70, "ProofInfo",
   CellTags->{"Proof|ID:379661071", "Proof|ID:379661071-1"},
   CellID->321989072]},
 "Proof|ID:379661071-1"->{
  Cell[729530, 20885, 10825, 248, 70, "ProofInfo",
   CellTags->{"Proof|ID:379661071", "Proof|ID:379661071-1"},
   CellID->321989072]},
 "ID:485298517"->{
  Cell[741770, 21181, 1079, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:485298517", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isTotalIrreflOrder \[Succeeds] \[ScriptCapitalT]"},
   CellID->485298517]},
 "isTotalIrreflOrder \[Succeeds] \[ScriptCapitalT]"->{
  Cell[741770, 21181, 1079, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:485298517", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isTotalIrreflOrder \[Succeeds] \[ScriptCapitalT]"},
   CellID->485298517]},
 "ID:561716382"->{
  Cell[742852, 21215, 734, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:561716382", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->561716382]},
 "ID:107805279"->{
  Cell[743589, 21236, 1743, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:107805279", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp decreasing ind"},
   CellID->107805279]},
 "PolyTuples pp decreasing ind"->{
  Cell[743589, 21236, 1743, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:107805279", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp decreasing ind"},
   CellID->107805279]},
 "ID:220534381"->{
  Cell[745335, 21288, 1318, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:220534381", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp decreasing"},
   CellID->220534381]},
 "PolyTuples pp decreasing"->{
  Cell[745335, 21288, 1318, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:220534381", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp decreasing"},
   CellID->220534381]},
 "ID:5709061"->{
  Cell[746656, 21329, 1310, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:5709061", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max"},
   CellID->5709061]},
 "PolyTuples first pp max"->{
  Cell[746656, 21329, 1310, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:5709061", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max"},
   CellID->5709061]},
 "ID:133904640"->{
  Cell[747969, 21370, 1292, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:133904640", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max'"},
   CellID->133904640]},
 "PolyTuples first pp max'"->{
  Cell[747969, 21370, 1292, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:133904640", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max'"},
   CellID->133904640]},
 "ID:12736634"->{
  Cell[749264, 21410, 1277, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:12736634", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp distinct"},
   CellID->12736634]},
 "PolyTuples pp distinct"->{
  Cell[749264, 21410, 1277, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:12736634", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp distinct"},
   CellID->12736634]},
 "Proof|ID:485298517"->{
  Cell[750696, 21459, 2046, 62, 70, "ProofInfo",
   CellTags->{"Proof|ID:485298517", "Proof|ID:485298517-1"},
   CellID->638746182]},
 "Proof|ID:485298517-1"->{
  Cell[750696, 21459, 2046, 62, 70, "ProofInfo",
   CellTags->{"Proof|ID:485298517", "Proof|ID:485298517-1"},
   CellID->638746182]},
 "Proof|ID:107805279"->{
  Cell[753261, 21540, 5315, 137, 70, "ProofInfo",
   CellTags->{"Proof|ID:107805279", "Proof|ID:107805279-1"},
   CellID->218458895]},
 "Proof|ID:107805279-1"->{
  Cell[753261, 21540, 5315, 137, 70, "ProofInfo",
   CellTags->{"Proof|ID:107805279", "Proof|ID:107805279-1"},
   CellID->218458895]},
 "Proof|ID:220534381"->{
  Cell[759096, 21696, 4773, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:220534381", "Proof|ID:220534381-1"},
   CellID->457923117]},
 "Proof|ID:220534381-1"->{
  Cell[759096, 21696, 4773, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:220534381", "Proof|ID:220534381-1"},
   CellID->457923117]},
 "Proof|ID:5709061"->{
  Cell[764389, 21841, 4005, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:5709061", "Proof|ID:5709061-1"},
   CellID->199136965]},
 "Proof|ID:5709061-1"->{
  Cell[764389, 21841, 4005, 110, 70, "ProofInfo",
   CellTags->{"Proof|ID:5709061", "Proof|ID:5709061-1"},
   CellID->199136965]},
 "Proof|ID:133904640"->{
  Cell[768911, 21970, 4060, 111, 70, "ProofInfo",
   CellTags->{"Proof|ID:133904640", "Proof|ID:133904640-1"},
   CellID->185035893]},
 "Proof|ID:133904640-1"->{
  Cell[768911, 21970, 4060, 111, 70, "ProofInfo",
   CellTags->{"Proof|ID:133904640", "Proof|ID:133904640-1"},
   CellID->185035893]},
 "Proof|ID:12736634"->{
  Cell[773491, 22100, 4682, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:12736634", "Proof|ID:12736634-1"},
   CellID->225795913]},
 "Proof|ID:12736634-1"->{
  Cell[773491, 22100, 4682, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:12736634", "Proof|ID:12736634-1"},
   CellID->225795913]},
 "ID:723780914"->{
  Cell[779169, 22259, 700, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:723780914", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->723780914]},
 "ID:284026551"->{
  Cell[779872, 22279, 936, 25, 70, "GlobalDeclaration",
   CellTags->{
    "ID:284026551", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->284026551]},
 "ID:548327311"->{
  Cell[780811, 22306, 864, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:548327311", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->548327311]},
 "ID:387361496"->{
  Cell[781678, 22330, 1040, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:387361496", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max seq"},
   CellID->387361496]},
 "PolyTuples first pp max seq"->{
  Cell[781678, 22330, 1040, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:387361496", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max seq"},
   CellID->387361496]},
 "ID:70991441"->{
  Cell[782721, 22360, 1116, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:70991441", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max seq'"},
   CellID->70991441]},
 "PolyTuples first pp max seq'"->{
  Cell[782721, 22360, 1116, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:70991441", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max seq'"},
   CellID->70991441]},
 "ID:365249767"->{
  Cell[783840, 22394, 948, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:365249767", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp distinct seq"},
   CellID->365249767]},
 "PolyTuples pp distinct seq"->{
  Cell[783840, 22394, 948, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:365249767", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp distinct seq"},
   CellID->365249767]},
 "Proof|ID:387361496"->{
  Cell[784942, 22431, 5601, 142, 70, "ProofInfo",
   CellTags->{"Proof|ID:387361496", "Proof|ID:387361496-1"},
   CellID->597599025]},
 "Proof|ID:387361496-1"->{
  Cell[784942, 22431, 5601, 142, 70, "ProofInfo",
   CellTags->{"Proof|ID:387361496", "Proof|ID:387361496-1"},
   CellID->597599025]},
 "Proof|ID:70991441"->{
  Cell[791063, 22592, 5988, 153, 70, "ProofInfo",
   CellTags->{"Proof|ID:70991441", "Proof|ID:70991441-1"},
   CellID->346547304]},
 "Proof|ID:70991441-1"->{
  Cell[791063, 22592, 5988, 153, 70, "ProofInfo",
   CellTags->{"Proof|ID:70991441", "Proof|ID:70991441-1"},
   CellID->346547304]},
 "Proof|ID:365249767"->{
  Cell[797570, 22764, 5493, 140, 70, "ProofInfo",
   CellTags->{"Proof|ID:365249767", "Proof|ID:365249767-1"},
   CellID->408497668]},
 "Proof|ID:365249767-1"->{
  Cell[797570, 22764, 5493, 140, 70, "ProofInfo",
   CellTags->{"Proof|ID:365249767", "Proof|ID:365249767-1"},
   CellID->408497668]},
 "ID:16132582"->{
  Cell[804351, 22948, 786, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:16132582", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->16132582]},
 "ID:742018993"->{
  Cell[805140, 22970, 1294, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:742018993", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C 1"},
   CellID->742018993]},
 "PolyTuples C 1"->{
  Cell[805140, 22970, 1294, 39, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:742018993", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C 1"},
   CellID->742018993]},
 "ID:281673814"->{
  Cell[806437, 23011, 1795, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:281673814", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C \[NotEqual] 0"},
   CellID->281673814]},
 "PolyTuples C \[NotEqual] 0"->{
  Cell[806437, 23011, 1795, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:281673814", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C \[NotEqual] 0"},
   CellID->281673814]},
 "ID:122376381"->{
  Cell[808235, 23068, 1827, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:122376381", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C=0"},
   CellID->122376381]},
 "PolyTuples C=0"->{
  Cell[808235, 23068, 1827, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:122376381", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C=0"},
   CellID->122376381]},
 "ID:256177412"->{
  Cell[810065, 23126, 2253, 68, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:256177412", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C=0 seq"},
   CellID->256177412]},
 "PolyTuples C=0 seq"->{
  Cell[810065, 23126, 2253, 68, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:256177412", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C=0 seq"},
   CellID->256177412]},
 "ID:293354245"->{
  Cell[812321, 23196, 1778, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:293354245", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C=0 seq'"},
   CellID->293354245]},
 "PolyTuples C=0 seq'"->{
  Cell[812321, 23196, 1778, 55, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:293354245", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C=0 seq'"},
   CellID->293354245]},
 "ID:129332919"->{
  Cell[814102, 23253, 1629, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:129332919", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C prepend"},
   CellID->129332919]},
 "PolyTuples C prepend"->{
  Cell[814102, 23253, 1629, 50, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:129332919", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C prepend"},
   CellID->129332919]},
 "Proof|ID:742018993"->{
  Cell[815887, 23314, 4301, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:742018993", "Proof|ID:742018993-1"},
   CellID->161841610]},
 "Proof|ID:742018993-1"->{
  Cell[815887, 23314, 4301, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:742018993", "Proof|ID:742018993-1"},
   CellID->161841610]},
 "Proof|ID:281673814"->{
  Cell[820776, 23452, 5067, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:281673814", "Proof|ID:281673814-1"},
   CellID->294457098]},
 "Proof|ID:281673814-1"->{
  Cell[820776, 23452, 5067, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:281673814", "Proof|ID:281673814-1"},
   CellID->294457098]},
 "Proof|ID:122376381"->{
  Cell[826363, 23603, 5063, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:122376381", "Proof|ID:122376381-1"},
   CellID->6731680]},
 "Proof|ID:122376381-1"->{
  Cell[826363, 23603, 5063, 132, 70, "ProofInfo",
   CellTags->{"Proof|ID:122376381", "Proof|ID:122376381-1"},
   CellID->6731680]},
 "Proof|ID:256177412"->{
  Cell[831944, 23754, 6576, 162, 70, "ProofInfo",
   CellTags->{"Proof|ID:256177412", "Proof|ID:256177412-1"},
   CellID->606101095]},
 "Proof|ID:256177412-1"->{
  Cell[831944, 23754, 6576, 162, 70, "ProofInfo",
   CellTags->{"Proof|ID:256177412", "Proof|ID:256177412-1"},
   CellID->606101095]},
 "Proof|ID:293354245"->{
  Cell[839039, 23935, 4770, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:293354245", "Proof|ID:293354245-1"},
   CellID->371170725]},
 "Proof|ID:293354245-1"->{
  Cell[839039, 23935, 4770, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:293354245", "Proof|ID:293354245-1"},
   CellID->371170725]},
 "Proof|ID:129332919"->{
  Cell[844328, 24080, 5446, 136, 70, "ProofInfo",
   CellTags->{"Proof|ID:129332919", "Proof|ID:129332919-1"},
   CellID->88799612]},
 "Proof|ID:129332919-1"->{
  Cell[844328, 24080, 5446, 136, 70, "ProofInfo",
   CellTags->{"Proof|ID:129332919", "Proof|ID:129332919-1"},
   CellID->88799612]},
 "ID:136746483"->{
  Cell[851032, 24258, 759, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:136746483", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->136746483]},
 "ID:110691644"->{
  Cell[851794, 24280, 729, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:110691644", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->110691644]},
 "ID:371581865"->{
  Cell[852526, 24301, 1012, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:371581865", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + ProdDomain"},
   CellID->371581865]},
 "PolyTuples + ProdDomain"->{
  Cell[852526, 24301, 1012, 26, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:371581865", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + ProdDomain"},
   CellID->371581865]},
 "ID:694735421"->{
  Cell[853541, 24329, 778, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:694735421", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + isTuple"},
   CellID->694735421]},
 "PolyTuples + isTuple"->{
  Cell[853541, 24329, 778, 20, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:694735421", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + isTuple"},
   CellID->694735421]},
 "ID:442662347"->{
  Cell[854322, 24351, 1182, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:442662347", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[NotEqual] 0"},
   CellID->442662347]},
 "PolyTuples + \[NotEqual] 0"->{
  Cell[854322, 24351, 1182, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:442662347", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[NotEqual] 0"},
   CellID->442662347]},
 "ID:66376992"->{
  Cell[855507, 24387, 2440, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:66376992", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + pp \[SubsetEqual]"},
   CellID->66376992]},
 "PolyTuples + pp \[SubsetEqual]"->{
  Cell[855507, 24387, 2440, 73, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:66376992", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + pp \[SubsetEqual]"},
   CellID->66376992]},
 "Proof|ID:371581865"->{
  Cell[858103, 24471, 4049, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:371581865", "Proof|ID:371581865-1"},
   CellID->414829965]},
 "Proof|ID:371581865-1"->{
  Cell[858103, 24471, 4049, 108, 70, "ProofInfo",
   CellTags->{"Proof|ID:371581865", "Proof|ID:371581865-1"},
   CellID->414829965]},
 "Proof|ID:694735421"->{
  Cell[862714, 24599, 3603, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:694735421", "Proof|ID:694735421-1"},
   CellID->511476329]},
 "Proof|ID:694735421-1"->{
  Cell[862714, 24599, 3603, 98, 70, "ProofInfo",
   CellTags->{"Proof|ID:694735421", "Proof|ID:694735421-1"},
   CellID->511476329]},
 "Proof|ID:442662347"->{
  Cell[866837, 24716, 5146, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:442662347", "Proof|ID:442662347-1"},
   CellID->874127013]},
 "Proof|ID:442662347-1"->{
  Cell[866837, 24716, 5146, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:442662347", "Proof|ID:442662347-1"},
   CellID->874127013]},
 "Proof|ID:66376992"->{
  Cell[872503, 24869, 8558, 211, 70, "ProofInfo",
   CellTags->{"Proof|ID:66376992", "Proof|ID:66376992-1"},
   CellID->490152990]},
 "Proof|ID:66376992-1"->{
  Cell[872503, 24869, 8558, 211, 70, "ProofInfo",
   CellTags->{"Proof|ID:66376992", "Proof|ID:66376992-1"},
   CellID->490152990]},
 "ID:480657044"->{
  Cell[882031, 25113, 856, 21, 70, "GlobalDeclaration",
   CellTags->{
    "ID:480657044", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->480657044]},
 "ID:907803476"->{
  Cell[882890, 25136, 1398, 39, 70, "GlobalDeclaration",
   CellTags->{
    "ID:907803476", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->907803476]},
 "ID:442268007"->{
  Cell[884291, 25177, 2224, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:442268007", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[Succeeds] 1"},
   CellID->442268007]},
 "PolyTuples + \[Succeeds] 1"->{
  Cell[884291, 25177, 2224, 59, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:442268007", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[Succeeds] 1"},
   CellID->442268007]},
 "ID:26802188"->{
  Cell[886518, 25238, 2123, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:26802188", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[Succeeds] 2"},
   CellID->26802188]},
 "PolyTuples + \[Succeeds] 2"->{
  Cell[886518, 25238, 2123, 58, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:26802188", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[Succeeds] 2"},
   CellID->26802188]},
 "ID:4890345"->{
  Cell[888644, 25298, 2216, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:4890345", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[Succeeds] 3"},
   CellID->4890345]},
 "PolyTuples + \[Succeeds] 3"->{
  Cell[888644, 25298, 2216, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:4890345", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[Succeeds] 3"},
   CellID->4890345]},
 "Proof|ID:442268007"->{
  Cell[891015, 25375, 10386, 242, 70, "ProofInfo",
   CellTags->{"Proof|ID:442268007", "Proof|ID:442268007-1"},
   CellID->23969624]},
 "Proof|ID:442268007-1"->{
  Cell[891015, 25375, 10386, 242, 70, "ProofInfo",
   CellTags->{"Proof|ID:442268007", "Proof|ID:442268007-1"},
   CellID->23969624]},
 "Proof|ID:26802188"->{
  Cell[901921, 25636, 10381, 242, 70, "ProofInfo",
   CellTags->{"Proof|ID:26802188", "Proof|ID:26802188-1"},
   CellID->77790849]},
 "Proof|ID:26802188-1"->{
  Cell[901921, 25636, 10381, 242, 70, "ProofInfo",
   CellTags->{"Proof|ID:26802188", "Proof|ID:26802188-1"},
   CellID->77790849]},
 "Proof|ID:4890345"->{
  Cell[912820, 25897, 10381, 242, 70, "ProofInfo",
   CellTags->{"Proof|ID:4890345", "Proof|ID:4890345-1"},
   CellID->650383762]},
 "Proof|ID:4890345-1"->{
  Cell[912820, 25897, 10381, 242, 70, "ProofInfo",
   CellTags->{"Proof|ID:4890345", "Proof|ID:4890345-1"},
   CellID->650383762]},
 "ID:212784844"->{
  Cell[924151, 26172, 738, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:212784844", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->212784844]},
 "ID:106856858"->{
  Cell[924892, 26193, 838, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:106856858", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - isTuple"},
   CellID->106856858]},
 "PolyTuples - isTuple"->{
  Cell[924892, 26193, 838, 23, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:106856858", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - isTuple"},
   CellID->106856858]},
 "ID:200895043"->{
  Cell[925733, 26218, 964, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:200895043", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - length"},
   CellID->200895043]},
 "PolyTuples - length"->{
  Cell[925733, 26218, 964, 27, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:200895043", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - length"},
   CellID->200895043]},
 "ID:19952680"->{
  Cell[926700, 26247, 1367, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:19952680", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - pp"},
   CellID->19952680]},
 "PolyTuples - pp"->{
  Cell[926700, 26247, 1367, 41, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:19952680", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - pp"},
   CellID->19952680]},
 "Proof|ID:106856858"->{
  Cell[928221, 26299, 2967, 84, 70, "ProofInfo",
   CellTags->{"Proof|ID:106856858", "Proof|ID:106856858-1"},
   CellID->555324734]},
 "Proof|ID:106856858-1"->{
  Cell[928221, 26299, 2967, 84, 70, "ProofInfo",
   CellTags->{"Proof|ID:106856858", "Proof|ID:106856858-1"},
   CellID->555324734]},
 "Proof|ID:200895043"->{
  Cell[931706, 26402, 3402, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:200895043", "Proof|ID:200895043-1"},
   CellID->593511758]},
 "Proof|ID:200895043-1"->{
  Cell[931706, 26402, 3402, 94, 70, "ProofInfo",
   CellTags->{"Proof|ID:200895043", "Proof|ID:200895043-1"},
   CellID->593511758]},
 "Proof|ID:19952680"->{
  Cell[935626, 26515, 4110, 112, 70, "ProofInfo",
   CellTags->{"Proof|ID:19952680", "Proof|ID:19952680-1"},
   CellID->287758024]},
 "Proof|ID:19952680-1"->{
  Cell[935626, 26515, 4110, 112, 70, "ProofInfo",
   CellTags->{"Proof|ID:19952680", "Proof|ID:19952680-1"},
   CellID->287758024]},
 "ID:118174250"->{
  Cell[940945, 26670, 1973, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:118174250", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isMonomial I"},
   CellID->118174250]},
 "PolyTuples isMonomial I"->{
  Cell[940945, 26670, 1973, 60, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:118174250", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isMonomial I"},
   CellID->118174250]},
 "ID:78163703"->{
  Cell[942921, 26732, 828, 23, 70, "GlobalDeclaration",
   CellTags->{
    "ID:78163703", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->78163703]},
 "ID:748742057"->{
  Cell[943752, 26757, 1691, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:748742057", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isMonomial II"},
   CellID->748742057]},
 "PolyTuples isMonomial II"->{
  Cell[943752, 26757, 1691, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:748742057", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isMonomial II"},
   CellID->748742057]},
 "ID:104425812"->{
  Cell[945446, 26803, 1819, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:104425812", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isMonomial III"},
   CellID->104425812]},
 "PolyTuples isMonomial III"->{
  Cell[945446, 26803, 1819, 53, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:104425812", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isMonomial III"},
   CellID->104425812]},
 "ID:191443971"->{
  Cell[947268, 26858, 1344, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:191443971", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CenterDot]"},
   CellID->191443971]},
 "PolyTuples \[CenterDot]"->{
  Cell[947268, 26858, 1344, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:191443971", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CenterDot]"},
   CellID->191443971]},
 "Proof|ID:118174250"->{
  Cell[948768, 26904, 5091, 130, 70, "ProofInfo",
   CellTags->{"Proof|ID:118174250", "Proof|ID:118174250-1"},
   CellID->45600071]},
 "Proof|ID:118174250-1"->{
  Cell[948768, 26904, 5091, 130, 70, "ProofInfo",
   CellTags->{"Proof|ID:118174250", "Proof|ID:118174250-1"},
   CellID->45600071]},
 "Proof|ID:748742057"->{
  Cell[954378, 27053, 5067, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:748742057", "Proof|ID:748742057-1"},
   CellID->104846486]},
 "Proof|ID:748742057-1"->{
  Cell[954378, 27053, 5067, 126, 70, "ProofInfo",
   CellTags->{"Proof|ID:748742057", "Proof|ID:748742057-1"},
   CellID->104846486]},
 "Proof|ID:104425812"->{
  Cell[959965, 27198, 5744, 140, 70, "ProofInfo",
   CellTags->{"Proof|ID:104425812", "Proof|ID:104425812-1"},
   CellID->41476004]},
 "Proof|ID:104425812-1"->{
  Cell[959965, 27198, 5744, 140, 70, "ProofInfo",
   CellTags->{"Proof|ID:104425812", "Proof|ID:104425812-1"},
   CellID->41476004]},
 "Proof|ID:191443971"->{
  Cell[966229, 27357, 4535, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:191443971", "Proof|ID:191443971-1"},
   CellID->309334880]},
 "Proof|ID:191443971-1"->{
  Cell[966229, 27357, 4535, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:191443971", "Proof|ID:191443971-1"},
   CellID->309334880]},
 "ID:956509664"->{
  Cell[971779, 27505, 835, 22, 70, "GlobalDeclaration",
   CellTags->{
    "ID:956509664", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->956509664]},
 "ID:52346492"->{
  Cell[972617, 27529, 944, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:52346492", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first index", "IR"},
   CellID->52346492]},
 "PolyTuples first index"->{
  Cell[972617, 27529, 944, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:52346492", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first index", "IR"},
   CellID->52346492]},
 "ID:3252425"->{
  Cell[973564, 27556, 1433, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:3252425", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isLP"},
   CellID->3252425]},
 "PolyTuples isLP"->{
  Cell[973564, 27556, 1433, 44, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:3252425", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isLP"},
   CellID->3252425]},
 "ID:301012536"->{
  Cell[975000, 27602, 898, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:301012536", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lp"},
   CellID->301012536]},
 "PolyTuples lp"->{
  Cell[975000, 27602, 898, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:301012536", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lp"},
   CellID->301012536]},
 "ID:499482967"->{
  Cell[975901, 27629, 898, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:499482967", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lc"},
   CellID->499482967]},
 "PolyTuples lc"->{
  Cell[975901, 27629, 898, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:499482967", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lc"},
   CellID->499482967]},
 "ID:77776384"->{
  Cell[976802, 27656, 912, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:77776384", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lm"},
   CellID->77776384]},
 "PolyTuples lm"->{
  Cell[976802, 27656, 912, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:77776384", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lm"},
   CellID->77776384]},
 "Proof|ID:52346492"->{
  Cell[977869, 27692, 3538, 96, 70, "ProofInfo",
   CellTags->{"Proof|ID:52346492", "Proof|ID:52346492-1"},
   CellID->156386449]},
 "Proof|ID:52346492-1"->{
  Cell[977869, 27692, 3538, 96, 70, "ProofInfo",
   CellTags->{"Proof|ID:52346492", "Proof|ID:52346492-1"},
   CellID->156386449]},
 "Proof|ID:3252425"->{
  Cell[981926, 27807, 4535, 120, 70, "ProofInfo",
   CellTags->{"Proof|ID:3252425", "Proof|ID:3252425-1"},
   CellID->201371222]},
 "Proof|ID:3252425-1"->{
  Cell[981926, 27807, 4535, 120, 70, "ProofInfo",
   CellTags->{"Proof|ID:3252425", "Proof|ID:3252425-1"},
   CellID->201371222]},
 "Proof|ID:301012536"->{
  Cell[986979, 27946, 3797, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:301012536", "Proof|ID:301012536-1"},
   CellID->132883546]},
 "Proof|ID:301012536-1"->{
  Cell[986979, 27946, 3797, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:301012536", "Proof|ID:301012536-1"},
   CellID->132883546]},
 "Proof|ID:499482967"->{
  Cell[991295, 28069, 3797, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:499482967", "Proof|ID:499482967-1"},
   CellID->358731136]},
 "Proof|ID:499482967-1"->{
  Cell[991295, 28069, 3797, 104, 70, "ProofInfo",
   CellTags->{"Proof|ID:499482967", "Proof|ID:499482967-1"},
   CellID->358731136]},
 "Proof|ID:77776384"->{
  Cell[995612, 28192, 3887, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:77776384", "Proof|ID:77776384-1"},
   CellID->379573710]},
 "Proof|ID:77776384-1"->{
  Cell[995612, 28192, 3887, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:77776384", "Proof|ID:77776384-1"},
   CellID->379573710]},
 "ID:1015977913"->{
  Cell[1000465, 28331, 703, 18, 70, "GlobalDeclaration",
   CellTags->{
    "ID:1015977913", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->1015977913]},
 "ID:731677923"->{
  Cell[1001171, 28351, 938, 25, 70, "GlobalDeclaration",
   CellTags->{
    "ID:731677923", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->731677923]},
 "ID:23863949"->{
  Cell[1002112, 28378, 1037, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:23863949", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lp seq"},
   CellID->23863949]},
 "PolyTuples lp seq"->{
  Cell[1002112, 28378, 1037, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:23863949", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lp seq"},
   CellID->23863949]},
 "ID:721799663"->{
  Cell[1003152, 28408, 1040, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:721799663", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lc seq"},
   CellID->721799663]},
 "PolyTuples lc seq"->{
  Cell[1003152, 28408, 1040, 28, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:721799663", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lc seq"},
   CellID->721799663]},
 "ID:2939774"->{
  Cell[1004195, 28438, 1181, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2939774", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lm seq"},
   CellID->2939774]},
 "PolyTuples lm seq"->{
  Cell[1004195, 28438, 1181, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2939774", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lm seq"},
   CellID->2939774]},
 "ID:172576768"->{
  Cell[1005379, 28471, 1114, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:172576768", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples R seq"},
   CellID->172576768]},
 "PolyTuples R seq"->{
  Cell[1005379, 28471, 1114, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:172576768", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples R seq"},
   CellID->172576768]},
 "ID:3206806"->{
  Cell[1006496, 28503, 1899, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:3206806", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CenterDot] lp lc"},
   CellID->3206806]},
 "PolyTuples \[CenterDot] lp lc"->{
  Cell[1006496, 28503, 1899, 56, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:3206806", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CenterDot] lp lc"},
   CellID->3206806]},
 "Proof|ID:23863949"->{
  Cell[1008551, 28570, 5137, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:23863949", "Proof|ID:23863949-1"},
   CellID->2176190]},
 "Proof|ID:23863949-1"->{
  Cell[1008551, 28570, 5137, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:23863949", "Proof|ID:23863949-1"},
   CellID->2176190]},
 "Proof|ID:721799663"->{
  Cell[1014207, 28723, 5144, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:721799663", "Proof|ID:721799663-1"},
   CellID->307273605]},
 "Proof|ID:721799663-1"->{
  Cell[1014207, 28723, 5144, 134, 70, "ProofInfo",
   CellTags->{"Proof|ID:721799663", "Proof|ID:721799663-1"},
   CellID->307273605]},
 "Proof|ID:2939774"->{
  Cell[1019871, 28876, 5653, 144, 70, "ProofInfo",
   CellTags->{"Proof|ID:2939774", "Proof|ID:2939774-1"},
   CellID->75418897]},
 "Proof|ID:2939774-1"->{
  Cell[1019871, 28876, 5653, 144, 70, "ProofInfo",
   CellTags->{"Proof|ID:2939774", "Proof|ID:2939774-1"},
   CellID->75418897]},
 "Proof|ID:172576768"->{
  Cell[1026042, 29039, 5358, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:172576768", "Proof|ID:172576768-1"},
   CellID->126187940]},
 "Proof|ID:172576768-1"->{
  Cell[1026042, 29039, 5358, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:172576768", "Proof|ID:172576768-1"},
   CellID->126187940]},
 "Proof|ID:3206806"->{
  Cell[1031920, 29196, 5527, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:3206806", "Proof|ID:3206806-1"},
   CellID->37996798]},
 "Proof|ID:3206806-1"->{
  Cell[1031920, 29196, 5527, 138, 70, "ProofInfo",
   CellTags->{"Proof|ID:3206806", "Proof|ID:3206806-1"},
   CellID->37996798]},
 "ID:31351138"->{
  Cell[1038907, 29384, 830, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:31351138", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalR] isDomain"},
   CellID->31351138]},
 "\[ScriptCapitalR] isDomain"->{
  Cell[1038907, 29384, 830, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:31351138", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalR] isDomain"},
   CellID->31351138]},
 "ID:632462358"->{
  Cell[1039740, 29408, 835, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:632462358", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalT] isDomain"},
   CellID->632462358]},
 "\[ScriptCapitalT] isDomain"->{
  Cell[1039740, 29408, 835, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:632462358", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalT] isDomain"},
   CellID->632462358]},
 "ID:404915746"->{
  Cell[1040578, 29432, 869, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:404915746", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isDomain"},
   CellID->404915746]},
 "PolyTuples isDomain"->{
  Cell[1040578, 29432, 869, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:404915746", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isDomain"},
   CellID->404915746]},
 "Proof|ID:31351138"->{
  Cell[1041603, 29465, 1784, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:31351138", "Proof|ID:31351138-1"},
   CellID->124332373]},
 "Proof|ID:31351138-1"->{
  Cell[1041603, 29465, 1784, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:31351138", "Proof|ID:31351138-1"},
   CellID->124332373]},
 "Proof|ID:632462358"->{
  Cell[1043906, 29540, 1789, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:632462358", "Proof|ID:632462358-1"},
   CellID->403093521]},
 "Proof|ID:632462358-1"->{
  Cell[1043906, 29540, 1789, 56, 70, "ProofInfo",
   CellTags->{"Proof|ID:632462358", "Proof|ID:632462358-1"},
   CellID->403093521]},
 "Proof|ID:404915746"->{
  Cell[1046214, 29615, 1941, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:404915746", "Proof|ID:404915746-1"},
   CellID->277496693]},
 "Proof|ID:404915746-1"->{
  Cell[1046214, 29615, 1941, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:404915746", "Proof|ID:404915746-1"},
   CellID->277496693]},
 "ID:3279840"->{
  Cell[1049343, 29714, 753, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:3279840", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->3279840]},
 "ID:141435198"->{
  Cell[1050099, 29736, 1082, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:141435198", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership C"},
   CellID->141435198]},
 "PolyTuples membership C"->{
  Cell[1050099, 29736, 1082, 30, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:141435198", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership C"},
   CellID->141435198]},
 "ID:45852914"->{
  Cell[1051184, 29768, 880, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:45852914", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership 0"},
   CellID->45852914]},
 "PolyTuples membership 0"->{
  Cell[1051184, 29768, 880, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:45852914", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership 0"},
   CellID->45852914]},
 "ID:320592129"->{
  Cell[1052067, 29794, 883, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:320592129", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership 1"},
   CellID->320592129]},
 "PolyTuples membership 1"->{
  Cell[1052067, 29794, 883, 24, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:320592129", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership 1"},
   CellID->320592129]},
 "ID:93322392"->{
  Cell[1052953, 29820, 907, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:93322392", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership +"},
   CellID->93322392]},
 "PolyTuples membership +"->{
  Cell[1052953, 29820, 907, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:93322392", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership +"},
   CellID->93322392]},
 "ID:6751189"->{
  Cell[1053863, 29847, 897, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:6751189", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership -"},
   CellID->6751189]},
 "PolyTuples membership -"->{
  Cell[1053863, 29847, 897, 25, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:6751189", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership -"},
   CellID->6751189]},
 "ID:179493910"->{
  Cell[1054763, 29874, 1252, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:179493910", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples supp"},
   CellID->179493910]},
 "PolyTuples supp"->{
  Cell[1054763, 29874, 1252, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:179493910", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples supp"},
   CellID->179493910]},
 "Proof|ID:141435198"->{
  Cell[1056170, 29919, 3549, 96, 70, "ProofInfo",
   CellTags->{"Proof|ID:141435198", "Proof|ID:141435198-1"},
   CellID->225697908]},
 "Proof|ID:141435198-1"->{
  Cell[1056170, 29919, 3549, 96, 70, "ProofInfo",
   CellTags->{"Proof|ID:141435198", "Proof|ID:141435198-1"},
   CellID->225697908]},
 "Proof|ID:45852914"->{
  Cell[1060238, 30034, 2315, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:45852914", "Proof|ID:45852914-1"},
   CellID->91077423]},
 "Proof|ID:45852914-1"->{
  Cell[1060238, 30034, 2315, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:45852914", "Proof|ID:45852914-1"},
   CellID->91077423]},
 "Proof|ID:320592129"->{
  Cell[1063072, 30121, 2320, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:320592129", "Proof|ID:320592129-1"},
   CellID->94174531]},
 "Proof|ID:320592129-1"->{
  Cell[1063072, 30121, 2320, 68, 70, "ProofInfo",
   CellTags->{"Proof|ID:320592129", "Proof|ID:320592129-1"},
   CellID->94174531]},
 "Proof|ID:93322392"->{
  Cell[1065912, 30208, 3928, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:93322392", "Proof|ID:93322392-1"},
   CellID->411309806]},
 "Proof|ID:93322392-1"->{
  Cell[1065912, 30208, 3928, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:93322392", "Proof|ID:93322392-1"},
   CellID->411309806]},
 "Proof|ID:6751189"->{
  Cell[1070359, 30333, 3286, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:6751189", "Proof|ID:6751189-1"},
   CellID->16346637]},
 "Proof|ID:6751189-1"->{
  Cell[1070359, 30333, 3286, 92, 70, "ProofInfo",
   CellTags->{"Proof|ID:6751189", "Proof|ID:6751189-1"},
   CellID->16346637]},
 "Proof|ID:179493910"->{
  Cell[1074162, 30444, 4439, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:179493910", "Proof|ID:179493910-1"},
   CellID->162772222]},
 "Proof|ID:179493910-1"->{
  Cell[1074162, 30444, 4439, 118, 70, "ProofInfo",
   CellTags->{"Proof|ID:179493910", "Proof|ID:179493910-1"},
   CellID->162772222]},
 "ID:644982519"->{
  Cell[1079572, 30595, 830, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:644982519", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasDecisionPredicate"},
   CellID->644982519]},
 "PolyTuples hasDecisionPredicate"->{
  Cell[1079572, 30595, 830, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:644982519", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasDecisionPredicate"},
   CellID->644982519]},
 "Proof|ID:644982519"->{
  Cell[1080556, 30627, 1997, 59, 70, "ProofInfo",
   CellTags->{"Proof|ID:644982519", "Proof|ID:644982519-1"},
   CellID->323902770]},
 "Proof|ID:644982519-1"->{
  Cell[1080556, 30627, 1997, 59, 70, "ProofInfo",
   CellTags->{"Proof|ID:644982519", "Proof|ID:644982519-1"},
   CellID->323902770]},
 "ID:114843091"->{
  Cell[1083748, 30727, 819, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:114843091", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples containsMonomials"},
   CellID->114843091]},
 "PolyTuples containsMonomials"->{
  Cell[1083748, 30727, 819, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:114843091", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples containsMonomials"},
   CellID->114843091]},
 "Proof|ID:114843091"->{
  Cell[1084718, 30759, 1985, 59, 70, "ProofInfo",
   CellTags->{"Proof|ID:114843091", "Proof|ID:114843091-1"},
   CellID->255004292]},
 "Proof|ID:114843091-1"->{
  Cell[1084718, 30759, 1985, 59, 70, "ProofInfo",
   CellTags->{"Proof|ID:114843091", "Proof|ID:114843091-1"},
   CellID->255004292]},
 "ID:78312379"->{
  Cell[1087947, 30860, 759, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:78312379", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->78312379]},
 "ID:210280178"->{
  Cell[1088709, 30881, 1369, 39, 70, "GlobalDeclaration",
   CellTags->{
    "ID:210280178", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->210280178]},
 "ID:261724624"->{
  Cell[1090081, 30922, 3948, 109, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:261724624", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 1"},
   CellID->261724624]},
 "PolyTuples + C 1"->{
  Cell[1090081, 30922, 3948, 109, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:261724624", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 1"},
   CellID->261724624]},
 "ID:115965410"->{
  Cell[1094032, 31033, 3703, 105, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:115965410", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 2"},
   CellID->115965410]},
 "PolyTuples + C 2"->{
  Cell[1094032, 31033, 3703, 105, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:115965410", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 2"},
   CellID->115965410]},
 "ID:384214099"->{
  Cell[1097738, 31140, 3955, 118, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:384214099", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 3"},
   CellID->384214099]},
 "PolyTuples + C 3"->{
  Cell[1097738, 31140, 3955, 118, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:384214099", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 3"},
   CellID->384214099]},
 "ID:56134460"->{
  Cell[1101696, 31260, 2710, 81, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:56134460", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 4"},
   CellID->56134460]},
 "PolyTuples + C 4"->{
  Cell[1101696, 31260, 2710, 81, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:56134460", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 4"},
   CellID->56134460]},
 "Proof|ID:261724624"->{
  Cell[1104562, 31352, 18180, 409, 70, "ProofInfo",
   CellTags->{"Proof|ID:261724624", "Proof|ID:261724624-1"},
   CellID->305685053]},
 "Proof|ID:261724624-1"->{
  Cell[1104562, 31352, 18180, 409, 70, "ProofInfo",
   CellTags->{"Proof|ID:261724624", "Proof|ID:261724624-1"},
   CellID->305685053]},
 "Proof|ID:115965410"->{
  Cell[1123262, 31780, 18180, 409, 70, "ProofInfo",
   CellTags->{"Proof|ID:115965410", "Proof|ID:115965410-1"},
   CellID->175933698]},
 "Proof|ID:115965410-1"->{
  Cell[1123262, 31780, 18180, 409, 70, "ProofInfo",
   CellTags->{"Proof|ID:115965410", "Proof|ID:115965410-1"},
   CellID->175933698]},
 "Proof|ID:384214099"->{
  Cell[1142004, 32209, 18545, 415, 70, "ProofInfo",
   CellTags->{"Proof|ID:384214099", "Proof|ID:384214099-1"},
   CellID->84967562]},
 "Proof|ID:384214099-1"->{
  Cell[1142004, 32209, 18545, 415, 70, "ProofInfo",
   CellTags->{"Proof|ID:384214099", "Proof|ID:384214099-1"},
   CellID->84967562]},
 "Proof|ID:56134460"->{
  Cell[1161069, 32643, 13729, 313, 70, "ProofInfo",
   CellTags->{"Proof|ID:56134460", "Proof|ID:56134460-1"},
   CellID->170413464]},
 "Proof|ID:56134460-1"->{
  Cell[1161069, 32643, 13729, 313, 70, "ProofInfo",
   CellTags->{"Proof|ID:56134460", "Proof|ID:56134460-1"},
   CellID->170413464]},
 "ID:235178571"->{
  Cell[1175749, 32989, 738, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:235178571", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->235178571]},
 "ID:474434374"->{
  Cell[1176490, 33010, 1193, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:474434374", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C 2"},
   CellID->474434374]},
 "PolyTuples C 2"->{
  Cell[1176490, 33010, 1193, 33, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:474434374", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C 2"},
   CellID->474434374]},
 "ID:108633346"->{
  Cell[1177686, 33045, 763, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:108633346", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->108633346]},
 "ID:611741232"->{
  Cell[1178452, 33067, 1231, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:611741232", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C +"},
   CellID->611741232]},
 "PolyTuples C +"->{
  Cell[1178452, 33067, 1231, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:611741232", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C +"},
   CellID->611741232]},
 "ID:438277340"->{
  Cell[1179686, 33104, 1084, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:438277340", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C - unary"},
   CellID->438277340]},
 "PolyTuples C - unary"->{
  Cell[1179686, 33104, 1084, 31, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:438277340", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C - unary"},
   CellID->438277340]},
 "ID:412638768"->{
  Cell[1180773, 33137, 1180, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:412638768", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C -"},
   CellID->412638768]},
 "PolyTuples C -"->{
  Cell[1180773, 33137, 1180, 34, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:412638768", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C -"},
   CellID->412638768]},
 "Proof|ID:474434374"->{
  Cell[1182109, 33182, 3970, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:474434374", "Proof|ID:474434374-1"},
   CellID->85632222]},
 "Proof|ID:474434374-1"->{
  Cell[1182109, 33182, 3970, 106, 70, "ProofInfo",
   CellTags->{"Proof|ID:474434374", "Proof|ID:474434374-1"},
   CellID->85632222]},
 "Proof|ID:611741232"->{
  Cell[1186599, 33307, 6115, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:611741232", "Proof|ID:611741232-1"},
   CellID->352968606]},
 "Proof|ID:611741232-1"->{
  Cell[1186599, 33307, 6115, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:611741232", "Proof|ID:611741232-1"},
   CellID->352968606]},
 "Proof|ID:438277340"->{
  Cell[1193234, 33480, 4791, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:438277340", "Proof|ID:438277340-1"},
   CellID->232831160]},
 "Proof|ID:438277340-1"->{
  Cell[1193234, 33480, 4791, 124, 70, "ProofInfo",
   CellTags->{"Proof|ID:438277340", "Proof|ID:438277340-1"},
   CellID->232831160]},
 "Proof|ID:412638768"->{
  Cell[1198545, 33623, 6115, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:412638768", "Proof|ID:412638768-1"},
   CellID->382215156]},
 "Proof|ID:412638768-1"->{
  Cell[1198545, 33623, 6115, 154, 70, "ProofInfo",
   CellTags->{"Proof|ID:412638768", "Proof|ID:412638768-1"},
   CellID->382215156]},
 "ID:392022323"->{
  Cell[1205632, 33810, 838, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:392022323", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasCoefficientFunction"},
   CellID->392022323]},
 "PolyTuples hasCoefficientFunction"->{
  Cell[1205632, 33810, 838, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:392022323", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasCoefficientFunction"},
   CellID->392022323]},
 "Proof|ID:392022323"->{
  Cell[1206625, 33842, 2012, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:392022323", "Proof|ID:392022323-1"},
   CellID->43251408]},
 "Proof|ID:392022323-1"->{
  Cell[1206625, 33842, 2012, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:392022323", "Proof|ID:392022323-1"},
   CellID->43251408]},
 "ID:23554005"->{
  Cell[1209828, 33943, 751, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:23554005", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->23554005]},
 "ID:392464903"->{
  Cell[1210582, 33964, 992, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:392464903", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->392464903]},
 "ID:161672314"->{
  Cell[1211577, 33992, 1238, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:161672314", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] empty"},
   CellID->161672314]},
 "PolyTuples \[CircleTimes] empty"->{
  Cell[1211577, 33992, 1238, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:161672314", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] empty"},
   CellID->161672314]},
 "ID:71703914"->{
  Cell[1212818, 34026, 3855, 97, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:71703914", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] prepend"},
   CellID->71703914]},
 "PolyTuples \[CircleTimes] prepend"->{
  Cell[1212818, 34026, 3855, 97, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:71703914", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] prepend"},
   CellID->71703914]},
 "ID:127729286"->{
  Cell[1216676, 34125, 738, 19, 70, "GlobalDeclaration",
   CellTags->{
    "ID:127729286", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->127729286]},
 "ID:2718947"->{
  Cell[1217417, 34146, 1346, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2718947", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] lm"},
   CellID->2718947]},
 "PolyTuples \[CircleTimes] lm"->{
  Cell[1217417, 34146, 1346, 35, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:2718947", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] lm"},
   CellID->2718947]},
 "ID:204263407"->{
  Cell[1218766, 34183, 1210, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:204263407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership \[CircleTimes]"},
   CellID->204263407]},
 "PolyTuples membership \[CircleTimes]"->{
  Cell[1218766, 34183, 1210, 32, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:204263407", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership \[CircleTimes]"},
   CellID->204263407]},
 "ID:134536220"->{
  Cell[1219979, 34217, 1539, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:134536220", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] \[RawStar]"},
   CellID->134536220]},
 "PolyTuples \[CircleTimes] \[RawStar]"->{
  Cell[1219979, 34217, 1539, 38, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:134536220", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] \[RawStar]"},
   CellID->134536220]},
 "ID:387729163"->{
  Cell[1221521, 34257, 1604, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:387729163", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[RawStar] alternative"},
   CellID->387729163]},
 "PolyTuples \[RawStar] alternative"->{
  Cell[1221521, 34257, 1604, 42, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:387729163", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[RawStar] alternative"},
   CellID->387729163]},
 "Proof|ID:161672314"->{
  Cell[1223280, 34310, 5448, 141, 70, "ProofInfo",
   CellTags->{"Proof|ID:161672314", "Proof|ID:161672314-1"},
   CellID->221366480]},
 "Proof|ID:161672314-1"->{
  Cell[1223280, 34310, 5448, 141, 70, "ProofInfo",
   CellTags->{"Proof|ID:161672314", "Proof|ID:161672314-1"},
   CellID->221366480]},
 "Proof|ID:71703914"->{
  Cell[1229248, 34470, 14576, 326, 70, "ProofInfo",
   CellTags->{"Proof|ID:71703914", "Proof|ID:71703914-1"},
   CellID->82366827]},
 "Proof|ID:71703914-1"->{
  Cell[1229248, 34470, 14576, 326, 70, "ProofInfo",
   CellTags->{"Proof|ID:71703914", "Proof|ID:71703914-1"},
   CellID->82366827]},
 "Proof|ID:2718947"->{
  Cell[1244343, 34815, 7241, 177, 70, "ProofInfo",
   CellTags->{"Proof|ID:2718947", "Proof|ID:2718947-1"},
   CellID->181253020]},
 "Proof|ID:2718947-1"->{
  Cell[1244343, 34815, 7241, 177, 70, "ProofInfo",
   CellTags->{"Proof|ID:2718947", "Proof|ID:2718947-1"},
   CellID->181253020]},
 "Proof|ID:204263407"->{
  Cell[1252146, 35012, 6423, 160, 70, "ProofInfo",
   CellTags->{"Proof|ID:204263407", "Proof|ID:204263407-1"},
   CellID->426944387]},
 "Proof|ID:204263407-1"->{
  Cell[1252146, 35012, 6423, 160, 70, "ProofInfo",
   CellTags->{"Proof|ID:204263407", "Proof|ID:204263407-1"},
   CellID->426944387]},
 "Proof|ID:134536220"->{
  Cell[1259088, 35191, 7215, 176, 70, "ProofInfo",
   CellTags->{"Proof|ID:134536220", "Proof|ID:134536220-1"},
   CellID->170925754]},
 "Proof|ID:134536220-1"->{
  Cell[1259088, 35191, 7215, 176, 70, "ProofInfo",
   CellTags->{"Proof|ID:134536220", "Proof|ID:134536220-1"},
   CellID->170925754]},
 "Proof|ID:387729163"->{
  Cell[1266823, 35386, 8927, 214, 70, "ProofInfo",
   CellTags->{"Proof|ID:387729163", "Proof|ID:387729163-1"},
   CellID->722597358]},
 "Proof|ID:387729163-1"->{
  Cell[1266823, 35386, 8927, 214, 70, "ProofInfo",
   CellTags->{"Proof|ID:387729163", "Proof|ID:387729163-1"},
   CellID->722597358]},
 "ID:65280633"->{
  Cell[1276715, 35633, 816, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:65280633", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasMultiplication"},
   CellID->65280633]},
 "PolyTuples hasMultiplication"->{
  Cell[1276715, 35633, 816, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:65280633", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasMultiplication"},
   CellID->65280633]},
 "Proof|ID:65280633"->{
  Cell[1277685, 35665, 1980, 59, 70, "ProofInfo",
   CellTags->{"Proof|ID:65280633", "Proof|ID:65280633-1"},
   CellID->187277372]},
 "Proof|ID:65280633-1"->{
  Cell[1277685, 35665, 1980, 59, 70, "ProofInfo",
   CellTags->{"Proof|ID:65280633", "Proof|ID:65280633-1"},
   CellID->187277372]},
 "ID:48571309"->{
  Cell[1280841, 35765, 2804, 92, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:48571309", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[ForAll] alternative 2"},
   CellID->48571309]},
 "PolyTuples \[ForAll] alternative 2"->{
  Cell[1280841, 35765, 2804, 92, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:48571309", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[ForAll] alternative 2"},
   CellID->48571309]},
 "ID:617696571"->{
  Cell[1283648, 35859, 763, 20, 70, "GlobalDeclaration",
   CellTags->{
    "ID:617696571", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->617696571]},
 "ID:776548"->{
  Cell[1284414, 35881, 1490, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:776548", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples=empty sufficient"},
   CellID->776548]},
 "PolyTuples=empty sufficient"->{
  Cell[1284414, 35881, 1490, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:776548", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples=empty sufficient"},
   CellID->776548]},
 "ID:304028020"->{
  Cell[1285907, 35930, 1542, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:304028020", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples extensional"},
   CellID->304028020]},
 "PolyTuples extensional"->{
  Cell[1285907, 35930, 1542, 49, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:304028020", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples extensional"},
   CellID->304028020]},
 "Proof|ID:48571309"->{
  Cell[1287604, 35990, 7486, 185, 70, "ProofInfo",
   CellTags->{"Proof|ID:48571309", "Proof|ID:48571309-1"},
   CellID->399659369]},
 "Proof|ID:48571309-1"->{
  Cell[1287604, 35990, 7486, 185, 70, "ProofInfo",
   CellTags->{"Proof|ID:48571309", "Proof|ID:48571309-1"},
   CellID->399659369]},
 "Proof|ID:776548"->{
  Cell[1295609, 36194, 4243, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:776548", "Proof|ID:776548-1"},
   CellID->379176191]},
 "Proof|ID:776548-1"->{
  Cell[1295609, 36194, 4243, 114, 70, "ProofInfo",
   CellTags->{"Proof|ID:776548", "Proof|ID:776548-1"},
   CellID->379176191]},
 "Proof|ID:304028020"->{
  Cell[1300369, 36327, 5393, 140, 70, "ProofInfo",
   CellTags->{"Proof|ID:304028020", "Proof|ID:304028020-1"},
   CellID->86924546]},
 "Proof|ID:304028020-1"->{
  Cell[1300369, 36327, 5393, 140, 70, "ProofInfo",
   CellTags->{"Proof|ID:304028020", "Proof|ID:304028020-1"},
   CellID->86924546]},
 "ID:589891585"->{
  Cell[1306777, 36501, 811, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:589891585", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isExtensional"},
   CellID->589891585]},
 "PolyTuples isExtensional"->{
  Cell[1306777, 36501, 811, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:589891585", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isExtensional"},
   CellID->589891585]},
 "Proof|ID:589891585"->{
  Cell[1307743, 36533, 1961, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:589891585", "Proof|ID:589891585-1"},
   CellID->116448114]},
 "Proof|ID:589891585-1"->{
  Cell[1307743, 36533, 1961, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:589891585", "Proof|ID:589891585-1"},
   CellID->116448114]},
 "ID:160815443"->{
  Cell[1310890, 36632, 806, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:160815443", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isOrderedRec"},
   CellID->160815443]},
 "PolyTuples isOrderedRec"->{
  Cell[1310890, 36632, 806, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:160815443", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isOrderedRec"},
   CellID->160815443]},
 "ID:397107848"->{
  Cell[1311699, 36655, 795, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:397107848", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isOrdered"},
   CellID->397107848]},
 "PolyTuples isOrdered"->{
  Cell[1311699, 36655, 795, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:397107848", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isOrdered"},
   CellID->397107848]},
 "Proof|ID:160815443"->{
  Cell[1312648, 36687, 1955, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:160815443", "Proof|ID:160815443-1"},
   CellID->9775189]},
 "Proof|ID:160815443-1"->{
  Cell[1312648, 36687, 1955, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:160815443", "Proof|ID:160815443-1"},
   CellID->9775189]},
 "Proof|ID:397107848"->{
  Cell[1315123, 36764, 1943, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:397107848", "Proof|ID:397107848-1"},
   CellID->3787871]},
 "Proof|ID:397107848-1"->{
  Cell[1315123, 36764, 1943, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:397107848", "Proof|ID:397107848-1"},
   CellID->3787871]},
 "ID:156782366"->{
  Cell[1318255, 36863, 812, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:156782366", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasMultipliers"},
   CellID->156782366]},
 "PolyTuples hasMultipliers"->{
  Cell[1318255, 36863, 812, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:156782366", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasMultipliers"},
   CellID->156782366]},
 "Proof|ID:156782366"->{
  Cell[1319222, 36895, 1965, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:156782366", "Proof|ID:156782366-1"},
   CellID->558592865]},
 "Proof|ID:156782366-1"->{
  Cell[1319222, 36895, 1965, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:156782366", "Proof|ID:156782366-1"},
   CellID->558592865]},
 "ID:667417187"->{
  Cell[1322389, 36994, 860, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:667417187", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isReductionPolynomialDomain"},
   CellID->667417187]},
 "PolyTuples isReductionPolynomialDomain"->{
  Cell[1322389, 36994, 860, 22, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:667417187", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isReductionPolynomialDomain"},
   CellID->667417187]},
 "Proof|ID:667417187"->{
  Cell[1323404, 37027, 2033, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:667417187", "Proof|ID:667417187-1"},
   CellID->486349173]},
 "Proof|ID:667417187-1"->{
  Cell[1323404, 37027, 2033, 60, 70, "ProofInfo",
   CellTags->{"Proof|ID:667417187", "Proof|ID:667417187-1"},
   CellID->486349173]},
 "ID:714775050"->{
  Cell[1326871, 37138, 967, 26, 70, "GlobalDeclaration",
   CellTags->{
    "ID:714775050", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
   CellID->714775050]},
 "ID:15424584"->{
  Cell[1327841, 37166, 4116, 109, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:15424584", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd alternative"},
   CellID->15424584]},
 "PolyTuples lcrd alternative"->{
  Cell[1327841, 37166, 4116, 109, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:15424584", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd alternative"},
   CellID->15424584]},
 "ID:168826371"->{
  Cell[1331960, 37277, 1655, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:168826371", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd unary alternative"},
   CellID->168826371]},
 "PolyTuples lcrd unary alternative"->{
  Cell[1331960, 37277, 1655, 47, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:168826371", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd unary alternative"},
   CellID->168826371]},
 "ID:228703376"->{
  Cell[1333618, 37326, 4778, 122, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:228703376", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd lc lp"},
   CellID->228703376]},
 "PolyTuples lcrd lc lp"->{
  Cell[1333618, 37326, 4778, 122, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:228703376", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd lc lp"},
   CellID->228703376]},
 "ID:354399952"->{
  Cell[1338399, 37450, 2404, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:354399952", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd length"},
   CellID->354399952]},
 "PolyTuples lcrd length"->{
  Cell[1338399, 37450, 2404, 66, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:354399952", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd length"},
   CellID->354399952]},
 "Proof|ID:15424584"->{
  Cell[1340958, 37527, 19212, 427, 70, "ProofInfo",
   CellTags->{"Proof|ID:15424584", "Proof|ID:15424584-1"},
   CellID->190628320]},
 "Proof|ID:15424584-1"->{
  Cell[1340958, 37527, 19212, 427, 70, "ProofInfo",
   CellTags->{"Proof|ID:15424584", "Proof|ID:15424584-1"},
   CellID->190628320]},
 "Proof|ID:168826371"->{
  Cell[1360754, 37974, 7200, 178, 70, "ProofInfo",
   CellTags->{"Proof|ID:168826371", "Proof|ID:168826371-1"},
   CellID->493554176]},
 "Proof|ID:168826371-1"->{
  Cell[1360754, 37974, 7200, 178, 70, "ProofInfo",
   CellTags->{"Proof|ID:168826371", "Proof|ID:168826371-1"},
   CellID->493554176]},
 "Proof|ID:228703376"->{
  Cell[1368473, 38171, 18754, 414, 70, "ProofInfo",
   CellTags->{"Proof|ID:228703376", "Proof|ID:228703376-1"},
   CellID->48977005]},
 "Proof|ID:228703376-1"->{
  Cell[1368473, 38171, 18754, 414, 70, "ProofInfo",
   CellTags->{"Proof|ID:228703376", "Proof|ID:228703376-1"},
   CellID->48977005]},
 "Proof|ID:354399952"->{
  Cell[1387815, 38605, 12333, 279, 70, "ProofInfo",
   CellTags->{"Proof|ID:354399952", "Proof|ID:354399952-1"},
   CellID->182853122]},
 "Proof|ID:354399952-1"->{
  Cell[1387815, 38605, 12333, 279, 70, "ProofInfo",
   CellTags->{"Proof|ID:354399952", "Proof|ID:354399952-1"},
   CellID->182853122]},
 "ID:479598778"->{
  Cell[1401104, 38917, 790, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:479598778", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasLCRD"},
   CellID->479598778]},
 "PolyTuples hasLCRD"->{
  Cell[1401104, 38917, 790, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:479598778", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasLCRD"},
   CellID->479598778]},
 "Proof|ID:479598778"->{
  Cell[1402050, 38949, 1937, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:479598778", "Proof|ID:479598778-1"},
   CellID->690537389]},
 "Proof|ID:479598778-1"->{
  Cell[1402050, 38949, 1937, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:479598778", "Proof|ID:479598778-1"},
   CellID->690537389]},
 "ID:463361785"->{
  Cell[1405160, 39048, 787, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:463361785", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasRDM"},
   CellID->463361785]},
 "PolyTuples hasRDM"->{
  Cell[1405160, 39048, 787, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:463361785", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasRDM"},
   CellID->463361785]},
 "Proof|ID:463361785"->{
  Cell[1406103, 39080, 1933, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:463361785", "Proof|ID:463361785-1"},
   CellID->260093492]},
 "Proof|ID:463361785-1"->{
  Cell[1406103, 39080, 1933, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:463361785", "Proof|ID:463361785-1"},
   CellID->260093492]},
 "ID:7345776"->{
  Cell[1409208, 39179, 783, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:7345776", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasCPM"},
   CellID->7345776]},
 "PolyTuples hasCPM"->{
  Cell[1409208, 39179, 783, 21, 70, "FormalTextInputFormula",
   CellTags->{
    "ID:7345776", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasCPM"},
   CellID->7345776]},
 "Proof|ID:7345776"->{
  Cell[1410147, 39211, 1922, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:7345776", "Proof|ID:7345776-1"},
   CellID->11260575]},
 "Proof|ID:7345776-1"->{
  Cell[1410147, 39211, 1922, 58, 70, "ProofInfo",
   CellTags->{"Proof|ID:7345776", "Proof|ID:7345776-1"},
   CellID->11260575]},
 "ID:1014777887"->{
  Cell[1413281, 39311, 875, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1014777887", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isAlgoReductionPolynomialDomain"},
   CellID->1014777887]},
 "PolyTuples isAlgoReductionPolynomialDomain"->{
  Cell[1413281, 39311, 875, 22, 21, "FormalTextInputFormula",
   CellTags->{
    "ID:1014777887", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isAlgoReductionPolynomialDomain"},
   CellID->1014777887]},
 "Proof|ID:1014777887"->{
  Cell[1414312, 39344, 2057, 61, 28, "ProofInfo",
   CellTags->{"Proof|ID:1014777887", "Proof|ID:1014777887-1"},
   CellID->68078823]},
 "Proof|ID:1014777887-1"->{
  Cell[1414312, 39344, 2057, 61, 28, "ProofInfo",
   CellTags->{"Proof|ID:1014777887", "Proof|ID:1014777887-1"},
   CellID->68078823]},
 "ID:67407018"->{
  Cell[1417346, 39438, 1542, 41, 78, "FormalTextInputFormula",
   CellTags->{
    "ID:67407018", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isAlgoReductionRing"},
   CellID->67407018]},
 "PolyTuples isAlgoReductionRing"->{
  Cell[1417346, 39438, 1542, 41, 78, "FormalTextInputFormula",
   CellTags->{
    "ID:67407018", 
     "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isAlgoReductionRing"},
   CellID->67407018]},
 "Proof|ID:67407018"->{
  Cell[1419043, 39490, 3588, 95, 28, "ProofInfo",
   CellTags->{"Proof|ID:67407018", "Proof|ID:67407018-1"},
   CellID->163016488]},
 "Proof|ID:67407018-1"->{
  Cell[1419043, 39490, 3588, 95, 28, "ProofInfo",
   CellTags->{"Proof|ID:67407018", "Proof|ID:67407018-1"},
   CellID->163016488]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"ID:303368290", 1460751, 40370},
 {"Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", 1461038, 40378},
 {"ID:603497554", 1527026, 42119},
 {"isExpPPDomain", 1527272, 42126},
 {"ID:49777475", 1527516, 42133},
 {"hasDecisionPredicatePP", 1527778, 42140},
 {"ID:225295367", 1528030, 42147},
 {"hasExponentFunctionPP", 1528293, 42154},
 {"ID:292978558", 1528547, 42161},
 {"isExtensionalPP", 1528796, 42168},
 {"ID:340781435", 1529042, 42175},
 {"hasDegreeFunctionPP", 1529299, 42182},
 {"ID:480617864", 1529549, 42189},
 {"hasDivisibilityPP", 1529802, 42196},
 {"ID:897055662", 1530050, 42203},
 {"isTermOrder", 1530292, 42210},
 {"ID:532395027", 1530535, 42217},
 {"ID:160997957", 1530756, 42224},
 {"isExpPPDomain dim", 1531008, 42231},
 {"ID:83558965", 1531254, 42238},
 {"isExpPPDomain hasDecisionPredicate", 1531538, 42245},
 {"ID:349898371", 1531800, 42252},
 {"isExpPPDomain hasExponentFunction", 1532084, 42259},
 {"ID:200216737", 1532347, 42266},
 {"isExpPPDomain isExtensional", 1532619, 42273},
 {"ID:276506833", 1532876, 42280},
 {"isExpPPDomain hasDegreeFunction", 1533156, 42287},
 {"ID:135986343", 1533417, 42294},
 {"isExpPPDomain hasDivisibility", 1533693, 42301},
 {"ID:122009120", 1533952, 42308},
 {"isExpPPDomain membership exp", 1534227, 42315},
 {"Proof|ID:160997957", 1534492, 42322},
 {"Proof|ID:160997957-1", 1534649, 42326},
 {"Proof|ID:83558965", 1534803, 42330},
 {"Proof|ID:83558965-1", 1534959, 42334},
 {"Proof|ID:349898371", 1535114, 42338},
 {"Proof|ID:349898371-1", 1535272, 42342},
 {"Proof|ID:200216737", 1535428, 42346},
 {"Proof|ID:200216737-1", 1535586, 42350},
 {"Proof|ID:276506833", 1535742, 42354},
 {"Proof|ID:276506833-1", 1535901, 42358},
 {"Proof|ID:135986343", 1536058, 42362},
 {"Proof|ID:135986343-1", 1536217, 42366},
 {"Proof|ID:122009120", 1536374, 42370},
 {"Proof|ID:122009120-1", 1536534, 42374},
 {"ID:185429719", 1536686, 42378},
 {"ID:81626705", 1536906, 42385},
 {"isExpPPDomain isDomain", 1537166, 42392},
 {"ID:154517895", 1537416, 42399},
 {"isExpPPDomain membership \[RawStar]", 1537711, 42406},
 {"IR", 1537973, 42413},
 {"ID:226316196", 1541407, 42498},
 {"isExpPPDomain \[RawStar] isAssoc", 1541689, 42505},
 {"ID:96938761", 1541950, 42512},
 {"isExpPPDomain \[RawStar] isComm", 1542229, 42519},
 {"ID:499736780", 1542489, 42526},
 {"isExpPPDomain 1 isNeutral", 1542759, 42533},
 {"ID:745587327", 1543016, 42540},
 {"isExpPPDomain \[RawStar] lcanc", 1543296, 42547},
 {"Proof|ID:81626705", 1543563, 42554},
 {"Proof|ID:81626705-1", 1543720, 42558},
 {"Proof|ID:154517895", 1543876, 42562},
 {"Proof|ID:154517895-1", 1544036, 42566},
 {"Proof|ID:226316196", 1544194, 42570},
 {"Proof|ID:226316196-1", 1544353, 42574},
 {"Proof|ID:96938761", 1544509, 42578},
 {"Proof|ID:96938761-1", 1544664, 42582},
 {"Proof|ID:499736780", 1544818, 42586},
 {"Proof|ID:499736780-1", 1544978, 42590},
 {"Proof|ID:745587327", 1545136, 42594},
 {"Proof|ID:745587327-1", 1545296, 42598},
 {"ID:333683969", 1545448, 42602},
 {"isExpPPDomain isCCMonoid", 1545715, 42609},
 {"Proof|ID:333683969", 1545976, 42616},
 {"Proof|ID:333683969-1", 1546136, 42620},
 {"ID:274960855", 1546288, 42624},
 {"ID:52136714", 1546509, 42631},
 {"isExpPPDomain lcm isJoin", 1546775, 42638},
 {"ID:35376325", 1547028, 42645},
 {"ID:138508454", 1547248, 42652},
 {"isExpPPDomain div.1", 1547506, 42659},
 {"ID:14785826", 1547756, 42666},
 {"isExpPPDomain div.2", 1548012, 42673},
 {"Proof|ID:52136714", 1548266, 42680},
 {"Proof|ID:52136714-1", 1548423, 42684},
 {"Proof|ID:138508454", 1548579, 42688},
 {"Proof|ID:138508454-1", 1548740, 42692},
 {"Proof|ID:14785826", 1548898, 42696},
 {"Proof|ID:14785826-1", 1549054, 42700},
 {"ID:181034073", 1549203, 42704},
 {"ID:943617998", 1549425, 42711},
 {"ID:706945753", 1549647, 42718},
 {"isExpPPDomain membership deg ind", 1549931, 42725},
 {"ID:606418540", 1550195, 42732},
 {"isExpPPDomain deg=0 ind", 1550461, 42739},
 {"ID:47160", 1550712, 42746},
 {"isExpPPDomain membership deg", 1550985, 42753},
 {"ID:58897330", 1551241, 42760},
 {"isExpPPDomain deg=0", 1551497, 42767},
 {"ID:170894204", 1551746, 42774},
 {"isExpPPDomain deg \[RawStar]", 1552022, 42781},
 {"Proof|ID:706945753", 1552288, 42788},
 {"Proof|ID:706945753-1", 1552449, 42792},
 {"Proof|ID:606418540", 1552608, 42796},
 {"Proof|ID:606418540-1", 1552769, 42800},
 {"Proof|ID:47160", 1552924, 42804},
 {"Proof|ID:47160-1", 1553072, 42808},
 {"Proof|ID:58897330", 1553221, 42812},
 {"Proof|ID:58897330-1", 1553378, 42816},
 {"Proof|ID:170894204", 1553534, 42820},
 {"Proof|ID:170894204-1", 1553695, 42824},
 {"ID:49640627", 1553847, 42828},
 {"increasing subsequence \[DoubleStruckCapitalN]0", 1554159, 42835},
 {"Proof|ID:49640627", 1554441, 42842},
 {"Proof|ID:49640627-1", 1554599, 42846},
 {"ID:256365628", 1554750, 42850},
 {"ID:443096051", 1554972, 42857},
 {"isExpPPDomain divN", 1555228, 42864},
 {"ID:27206868", 1555477, 42871},
 {"isExpPPDomain divN +1", 1555737, 42878},
 {"ID:184673495", 1555988, 42885},
 {"increasing subsequence isExpPPDomain 0", 1556284, 42892},
 {"ID:43287351", 1556553, 42899},
 {"increasing subsequence isExpPPDomain", 1556843, 42906},
 {"Proof|ID:27206868", 1557114, 42913},
 {"Proof|ID:27206868-1", 1557271, 42917},
 {"Proof|ID:184673495", 1557427, 42921},
 {"Proof|ID:184673495-1", 1557588, 42925},
 {"Proof|ID:43287351", 1557746, 42929},
 {"Proof|ID:43287351-1", 1557904, 42933},
 {"ID:30919390", 1558054, 42937},
 {"isExpPPDomain isDicksonDomain", 1558330, 42944},
 {"Proof|ID:30919390", 1558594, 42951},
 {"Proof|ID:30919390-1", 1558751, 42955},
 {"ID:323333890", 1558901, 42959},
 {"ID:235080407", 1559123, 42966},
 {"ID:691531428", 1559345, 42973},
 {"term order |", 1559589, 42980},
 {"ID:772895867", 1559833, 42987},
 {"term order isOrderEmbedding", 1560106, 42994},
 {"ID:83375799", 1560363, 43001},
 {"term order isNoetherian", 1560626, 43008},
 {"Proof|ID:691531428", 1560884, 43015},
 {"Proof|ID:691531428-1", 1561044, 43019},
 {"Proof|ID:772895867", 1561202, 43023},
 {"Proof|ID:772895867-1", 1561362, 43027},
 {"Proof|ID:83375799", 1561519, 43031},
 {"Proof|ID:83375799-1", 1561676, 43035},
 {"ID:69380281", 1561825, 43039},
 {"ID:70274751", 1562044, 43046},
 {"ID:417719229", 1562264, 43053},
 {"ID:296713523", 1562486, 43060},
 {"OrderedPP \[Element]", 1562746, 43067},
 {"ID:275450356", 1562998, 43074},
 {"OrderedPP dim", 1563243, 43081},
 {"ID:126765714", 1563487, 43088},
 {"OrderedPP exp", 1563733, 43095},
 {"ID:300238463", 1563978, 43102},
 {"OrderedPP 1", 1564219, 43109},
 {"ID:91738893", 1564460, 43116},
 {"OrderedPP \[RawStar]", 1564717, 43123},
 {"ID:1017172830", 1564967, 43130},
 {"OrderedPP /", 1565210, 43137},
 {"ID:62103041", 1565453, 43144},
 {"OrderedPP lcm", 1565696, 43151},
 {"ID:229667832", 1565938, 43158},
 {"OrderedPP |", 1566180, 43165},
 {"ID:687245437", 1566423, 43172},
 {"OrderedPP deg", 1566668, 43179},
 {"ID:353388395", 1566912, 43186},
 {"OrderedPP \[Succeeds]", 1567174, 43193},
 {"ID:144609801", 1567427, 43200},
 {"OrderedPP <<", 1567671, 43207},
 {"ID:189296520", 1567915, 43214},
 {"OrderedPP >>", 1568159, 43221},
 {"ID:29953053", 1568402, 43228},
 {"ID:309264225", 1568622, 43235},
 {"OrderedPP isExpPPDomain", 1568887, 43242},
 {"ID:499712177", 1569141, 43249},
 {"OrderedPP isDicksonDomain", 1569411, 43256},
 {"ID:275069792", 1569668, 43263},
 {"ID:834856233", 1569890, 43270},
 {"OrderedPP isTermOrder", 1570151, 43277},
 {"ID:260893370", 1570403, 43284},
 {"OrderedPP isCommPPDomain", 1570671, 43291},
 {"Proof|ID:309264225", 1570933, 43298},
 {"Proof|ID:309264225-1", 1571092, 43302},
 {"Proof|ID:499712177", 1571249, 43306},
 {"Proof|ID:499712177-1", 1571408, 43310},
 {"Proof|ID:834856233", 1571565, 43314},
 {"Proof|ID:834856233-1", 1571724, 43318},
 {"Proof|ID:260893370", 1571881, 43322},
 {"Proof|ID:260893370-1", 1572040, 43326},
 {"ID:223902051", 1572191, 43330},
 {"Lex", 1572417, 43337},
 {"ID:376155678", 1572652, 43344},
 {"ID:310375877", 1572874, 43351},
 {"ID:463764498", 1573096, 43358},
 {"Lex total", 1573333, 43365},
 {"ID:6465370", 1573571, 43372},
 {"Lex asymmetric", 1573815, 43379},
 {"ID:78173747", 1574056, 43386},
 {"Lex transitive", 1574302, 43393},
 {"ID:341999039", 1574546, 43400},
 {"Lex 1 min", 1574783, 43407},
 {"ID:34516336", 1575022, 43414},
 {"Lex \[RawStar]", 1575268, 43421},
 {"Proof|ID:463764498", 1575518, 43428},
 {"Proof|ID:463764498-1", 1575679, 43432},
 {"Proof|ID:6465370", 1575836, 43436},
 {"Proof|ID:6465370-1", 1575991, 43440},
 {"Proof|ID:78173747", 1576145, 43444},
 {"Proof|ID:78173747-1", 1576303, 43448},
 {"Proof|ID:341999039", 1576460, 43452},
 {"Proof|ID:341999039-1", 1576620, 43456},
 {"Proof|ID:34516336", 1576777, 43460},
 {"Proof|ID:34516336-1", 1576935, 43464},
 {"ID:414199224", 1577086, 43468},
 {"Lex isTermOrder", 1577335, 43475},
 {"Proof|ID:414199224", 1577587, 43482},
 {"Proof|ID:414199224-1", 1577746, 43486},
 {"ID:116320630", 1577897, 43490},
 {"DegLex", 1578129, 43497},
 {"ID:106786369", 1578367, 43504},
 {"ID:43019543", 1578588, 43511},
 {"ID:75557233", 1578807, 43518},
 {"DegLex total", 1579048, 43525},
 {"ID:49531328", 1579288, 43532},
 {"DegLex asymmetric", 1579540, 43539},
 {"ID:260398187", 1579787, 43546},
 {"DegLex transitive", 1580041, 43553},
 {"ID:59133109", 1580289, 43560},
 {"DegLex 1 min", 1580530, 43567},
 {"ID:572110615", 1580771, 43574},
 {"DegLex \[RawStar]", 1581025, 43581},
 {"Proof|ID:75557233", 1581279, 43588},
 {"Proof|ID:75557233-1", 1581437, 43592},
 {"Proof|ID:49531328", 1581593, 43596},
 {"Proof|ID:49531328-1", 1581751, 43600},
 {"Proof|ID:260398187", 1581908, 43604},
 {"Proof|ID:260398187-1", 1582069, 43608},
 {"Proof|ID:59133109", 1582227, 43612},
 {"Proof|ID:59133109-1", 1582384, 43616},
 {"Proof|ID:572110615", 1582540, 43620},
 {"Proof|ID:572110615-1", 1582702, 43624},
 {"ID:463127740", 1582856, 43628},
 {"DegLex isTermOrder", 1583112, 43635},
 {"Proof|ID:463127740", 1583368, 43642},
 {"Proof|ID:463127740-1", 1583529, 43646},
 {"ID:267616510", 1583682, 43650},
 {"ID:1929370", 1583903, 43657},
 {"RevLex", 1584132, 43664},
 {"ID:116917210", 1584367, 43671},
 {"DegRevLex", 1584606, 43678},
 {"ID:431887307", 1584848, 43685},
 {"ID:204943512", 1585071, 43692},
 {"ID:781115991", 1585294, 43699},
 {"RevLex total", 1585539, 43706},
 {"ID:2983958", 1585782, 43713},
 {"RevLex asymmetric", 1586033, 43720},
 {"ID:240564679", 1586279, 43727},
 {"RevLex transitive", 1586534, 43734},
 {"ID:299677825", 1586784, 43741},
 {"RevLex \[RawStar]", 1587039, 43748},
 {"Proof|ID:781115991", 1587295, 43755},
 {"Proof|ID:781115991-1", 1587457, 43759},
 {"Proof|ID:2983958", 1587615, 43763},
 {"Proof|ID:2983958-1", 1587770, 43767},
 {"Proof|ID:240564679", 1587925, 43771},
 {"Proof|ID:240564679-1", 1588087, 43775},
 {"Proof|ID:299677825", 1588247, 43779},
 {"Proof|ID:299677825-1", 1588409, 43783},
 {"ID:449289304", 1588563, 43787},
 {"ID:240154468", 1588786, 43794},
 {"DegRevLex total", 1589037, 43801},
 {"ID:808110452", 1589285, 43808},
 {"DegRevLex asymmetric", 1589546, 43815},
 {"ID:224322863", 1589799, 43822},
 {"DegRevLex transitive", 1590060, 43829},
 {"ID:424804407", 1590313, 43836},
 {"DegRevLex 1 min", 1590563, 43843},
 {"ID:308108885", 1590810, 43850},
 {"DegRevLex \[RawStar]", 1591071, 43857},
 {"Proof|ID:240154468", 1591330, 43864},
 {"Proof|ID:240154468-1", 1591492, 43868},
 {"Proof|ID:808110452", 1591652, 43872},
 {"Proof|ID:808110452-1", 1591813, 43876},
 {"Proof|ID:224322863", 1591972, 43880},
 {"Proof|ID:224322863-1", 1592134, 43884},
 {"Proof|ID:424804407", 1592294, 43888},
 {"Proof|ID:424804407-1", 1592455, 43892},
 {"Proof|ID:308108885", 1592614, 43896},
 {"Proof|ID:308108885-1", 1592776, 43900},
 {"ID:936963717", 1592930, 43904},
 {"DegRevLex isTermOrder", 1593192, 43911},
 {"Proof|ID:936963717", 1593451, 43918},
 {"Proof|ID:936963717-1", 1593612, 43922},
 {"ID:44994065", 1593764, 43926},
 {"ID:1997108144", 1593986, 43933},
 {"ID:608617422", 1594211, 43940},
 {"ID:271011394", 1594434, 43947},
 {"PPTuples \[Element]", 1594693, 43954},
 {"ID:1820126795", 1594946, 43961},
 {"PPTuples 1", 1595188, 43968},
 {"ID:281355476", 1595432, 43975},
 {"PPTuples dim", 1595676, 43982},
 {"ID:2109519396", 1595921, 43989},
 {"PPTuples deg", 1596168, 43996},
 {"ID:1563833422", 1596416, 44003},
 {"PPTuples exp", 1596662, 44010},
 {"ID:1180963317", 1596909, 44017},
 {"PPTuples \[RawStar]", 1597170, 44024},
 {"ID:1392702003", 1597425, 44031},
 {"PPTuples /", 1597668, 44038},
 {"ID:1587875066", 1597914, 44045},
 {"PPTuples \[VerticalBar]", 1598183, 44052},
 {"ID:803590963", 1598441, 44059},
 {"PPTuples lcm", 1598686, 44066},
 {"ID:106372713", 1598931, 44073},
 {"PPTuples <<", 1599174, 44080},
 {"ID:1589980318", 1599419, 44087},
 {"PPTuples >>", 1599664, 44094},
 {"ID:61015536", 1599909, 44101},
 {"ID:617800164", 1600130, 44108},
 {"\[Element] PPTuples alternative", 1600419, 44115},
 {"ID:41499157", 1600688, 44122},
 {"PPTuples isDomain", 1600940, 44129},
 {"ID:142297858", 1601187, 44136},
 {"PPTuples hasDecisionPredicate", 1601465, 44143},
 {"ID:201916088", 1601726, 44150},
 {"PPTuples hasExponentFunction", 1602002, 44157},
 {"ID:102507049", 1602262, 44164},
 {"PPTuples isExtensional", 1602526, 44171},
 {"ID:93844102", 1602779, 44178},
 {"PPTuples hasDegreeFunction", 1603049, 44185},
 {"ID:34803991", 1603304, 44192},
 {"PPTuples hasDivisibility", 1603570, 44199},
 {"Proof|ID:617800164", 1603830, 44206},
 {"Proof|ID:617800164-1", 1603992, 44210},
 {"Proof|ID:41499157", 1604151, 44214},
 {"Proof|ID:41499157-1", 1604309, 44218},
 {"Proof|ID:142297858", 1604466, 44222},
 {"Proof|ID:142297858-1", 1604627, 44226},
 {"Proof|ID:201916088", 1604786, 44230},
 {"Proof|ID:201916088-1", 1604947, 44234},
 {"Proof|ID:102507049", 1605106, 44238},
 {"Proof|ID:102507049-1", 1605267, 44242},
 {"Proof|ID:93844102", 1605425, 44246},
 {"Proof|ID:93844102-1", 1605582, 44250},
 {"Proof|ID:34803991", 1605737, 44254},
 {"Proof|ID:34803991-1", 1605894, 44258},
 {"ID:338277855", 1606044, 44262},
 {"PPTuples isExpPPDomain", 1606308, 44269},
 {"Proof|ID:338277855", 1606568, 44276},
 {"Proof|ID:338277855-1", 1606729, 44280},
 {"ID:611122547", 1606882, 44284},
 {"ID:1216459107", 1607106, 44291},
 {"ID:594519699", 1607332, 44298},
 {"ID:65019775", 1607555, 44305},
 {"PolyTuples \[Element]", 1607817, 44312},
 {"ID:399496924", 1608070, 44319},
 {"PolyTuples C base", 1608324, 44326},
 {"ID:983317445", 1608573, 44333},
 {"PolyTuples C rec", 1608826, 44340},
 {"ID:1360683691", 1609076, 44347},
 {"PolyTuples 0", 1609322, 44354},
 {"ID:13289303", 1609567, 44361},
 {"PolyTuples 1", 1609810, 44368},
 {"ID:1777247200", 1610054, 44375},
 {"PolyTuples + 1", 1610304, 44382},
 {"ID:2122267989", 1610553, 44389},
 {"PolyTuples + 2", 1610804, 44396},
 {"ID:290578418", 1611053, 44403},
 {"PolyTuples + 3", 1611303, 44410},
 {"ID:1495569651", 1611552, 44417},
 {"PolyTuples - unary 1", 1611814, 44424},
 {"ID:1285234307", 1612069, 44431},
 {"PolyTuples - unary 2", 1612332, 44438},
 {"ID:551740538", 1612587, 44445},
 {"PolyTuples -", 1612831, 44452},
 {"ID:419928300", 1613075, 44459},
 {"PolyTuples \[RawStar] 1", 1613341, 44466},
 {"ID:209271439", 1613596, 44473},
 {"PolyTuples \[RawStar] 2", 1613863, 44480},
 {"ID:1411844363", 1614120, 44487},
 {"PolyTuples \[RawStar] 3", 1614389, 44494},
 {"ID:1128062400", 1614648, 44501},
 {"PolyTuples \[Succeeds] 1", 1614919, 44508},
 {"ID:1280515048", 1615179, 44515},
 {"PolyTuples \[Succeeds] 2", 1615450, 44522},
 {"ID:2095908261", 1615710, 44529},
 {"PolyTuples \[Succeeds] 3", 1615981, 44536},
 {"ID:1668822406", 1616241, 44543},
 {"PolyTuples multN 1", 1616499, 44550},
 {"ID:329725901", 1616751, 44557},
 {"PolyTuples multN 2", 1617008, 44564},
 {"ID:822409829", 1617259, 44571},
 {"PolyTuples mult", 1617510, 44578},
 {"ID:922404041", 1617758, 44585},
 {"PolyTuples mult i", 1618013, 44592},
 {"ID:2033419264", 1618264, 44599},
 {"PolyTuples rdm 1", 1618518, 44606},
 {"ID:1308197671", 1618769, 44613},
 {"PolyTuples rdm 2", 1619024, 44620},
 {"ID:1157498137", 1619276, 44627},
 {"PolyTuples rdm 3", 1619531, 44634},
 {"ID:103388547", 1619782, 44641},
 {"PolyTuples lcrd 1", 1620036, 44648},
 {"ID:166800573", 1620285, 44655},
 {"PolyTuples lcrd 2", 1620540, 44662},
 {"ID:319850711", 1620790, 44669},
 {"PolyTuples lcrd 3", 1621046, 44676},
 {"ID:112650434", 1621297, 44683},
 {"PolyTuples lcrd unary 1", 1621563, 44690},
 {"ID:1729493745", 1621819, 44697},
 {"PolyTuples lcrd unary 2", 1622088, 44704},
 {"ID:414290689", 1622346, 44711},
 {"PolyTuples cpm 1", 1622599, 44718},
 {"ID:27385995", 1622847, 44725},
 {"PolyTuples cpm 2", 1623098, 44732},
 {"ID:1444186163", 1623346, 44739},
 {"PolyTuples cpm 3", 1623602, 44746},
 {"ID:2018930893", 1623855, 44753},
 {"PolyTuples <<", 1624104, 44760},
 {"ID:1442847663", 1624353, 44767},
 {"PolyTuples >>", 1624602, 44774},
 {"ID:1486930793", 1624851, 44781},
 {"\[ScriptCapitalT]2", 1625110, 44788},
 {"ID:933000257", 1625363, 44795},
 {"\[ScriptCapitalT]3", 1625620, 44802},
 {"ID:1375788385", 1625872, 44809},
 {"\[ScriptCapitalQ]xyz", 1626135, 44816},
 {"ID:395821466", 1626390, 44823},
 {"ID:689486298", 1626613, 44830},
 {"PolyTuples isTuple", 1626875, 44837},
 {"ID:302371753", 1627131, 44844},
 {"\[Element] PolyTuples DomainTuples", 1627420, 44851},
 {"ID:619631752", 1627687, 44858},
 {"\[Element] PolyTuples empty", 1627967, 44865},
 {"ID:93154383", 1628231, 44872},
 {"\[Element] PolyTuples rec", 1628506, 44879},
 {"ID:110840922", 1628768, 44886},
 {"\[Element] PolyTuples rec II", 1629051, 44893},
 {"ID:45787564", 1629317, 44900},
 {"\[Element] PolyTuples rec III", 1629601, 44907},
 {"ID:34480440", 1629867, 44914},
 {"ID:134755829", 1630088, 44921},
 {"\[Element] PolyTuples prepend", 1630373, 44928},
 {"ID:218126712", 1630641, 44935},
 {"PolyTuples membership first", 1630922, 44942},
 {"ID:36095185", 1631187, 44949},
 {"PolyTuples membership second", 1631468, 44956},
 {"ID:266402037", 1631733, 44963},
 {"\[Element] PolyTuples shape", 1632008, 44970},
 {"ID:314594392", 1632268, 44977},
 {"PolyTuples length=1", 1632527, 44984},
 {"Proof|ID:689486298", 1632785, 44991},
 {"Proof|ID:689486298-1", 1632946, 44995},
 {"Proof|ID:302371753", 1633105, 44999},
 {"Proof|ID:302371753-1", 1633266, 45003},
 {"Proof|ID:619631752", 1633425, 45007},
 {"Proof|ID:619631752-1", 1633586, 45011},
 {"Proof|ID:93154383", 1633744, 45015},
 {"Proof|ID:93154383-1", 1633903, 45019},
 {"Proof|ID:110840922", 1634061, 45023},
 {"Proof|ID:110840922-1", 1634222, 45027},
 {"Proof|ID:45787564", 1634380, 45031},
 {"Proof|ID:45787564-1", 1634539, 45035},
 {"Proof|ID:134755829", 1634697, 45039},
 {"Proof|ID:134755829-1", 1634859, 45043},
 {"Proof|ID:218126712", 1635019, 45047},
 {"Proof|ID:218126712-1", 1635179, 45051},
 {"Proof|ID:36095185", 1635336, 45055},
 {"Proof|ID:36095185-1", 1635493, 45059},
 {"Proof|ID:266402037", 1635649, 45063},
 {"Proof|ID:266402037-1", 1635811, 45067},
 {"Proof|ID:314594392", 1635971, 45071},
 {"Proof|ID:314594392-1", 1636133, 45075},
 {"ID:441087600", 1636287, 45079},
 {"ID:678316554", 1636510, 45086},
 {"PolyTuples \[ForAll] alternative", 1636795, 45093},
 {"ID:329976481", 1637060, 45100},
 {"PolyTuples induction", 1637328, 45107},
 {"ID:379661071", 1637588, 45114},
 {"PolyTuples induction'", 1637858, 45121},
 {"Proof|ID:678316554", 1638125, 45128},
 {"Proof|ID:678316554-1", 1638287, 45132},
 {"Proof|ID:329976481", 1638447, 45136},
 {"Proof|ID:329976481-1", 1638607, 45140},
 {"Proof|ID:379661071", 1638765, 45144},
 {"Proof|ID:379661071-1", 1638928, 45148},
 {"ID:485298517", 1639083, 45152},
 {"isTotalIrreflOrder \[Succeeds] \[ScriptCapitalT]", 1639400, 45159},
 {"ID:561716382", 1639681, 45166},
 {"ID:107805279", 1639904, 45173},
 {"PolyTuples pp decreasing ind", 1640181, 45180},
 {"ID:220534381", 1640442, 45187},
 {"PolyTuples pp decreasing", 1640711, 45194},
 {"ID:5709061", 1640966, 45201},
 {"PolyTuples first pp max", 1641229, 45208},
 {"ID:133904640", 1641481, 45215},
 {"PolyTuples first pp max'", 1641750, 45222},
 {"ID:12736634", 1642006, 45229},
 {"PolyTuples pp distinct", 1642269, 45236},
 {"Proof|ID:485298517", 1642528, 45243},
 {"Proof|ID:485298517-1", 1642689, 45247},
 {"Proof|ID:107805279", 1642848, 45251},
 {"Proof|ID:107805279-1", 1643010, 45255},
 {"Proof|ID:220534381", 1643170, 45259},
 {"Proof|ID:220534381-1", 1643332, 45263},
 {"Proof|ID:5709061", 1643490, 45267},
 {"Proof|ID:5709061-1", 1643646, 45271},
 {"Proof|ID:133904640", 1643802, 45275},
 {"Proof|ID:133904640-1", 1643964, 45279},
 {"Proof|ID:12736634", 1644123, 45283},
 {"Proof|ID:12736634-1", 1644282, 45287},
 {"ID:723780914", 1644434, 45291},
 {"ID:284026551", 1644657, 45298},
 {"ID:548327311", 1644880, 45305},
 {"ID:387361496", 1645103, 45312},
 {"PolyTuples first pp max seq", 1645378, 45319},
 {"ID:70991441", 1645637, 45326},
 {"PolyTuples first pp max seq'", 1645912, 45333},
 {"ID:365249767", 1646171, 45340},
 {"PolyTuples pp distinct seq", 1646443, 45347},
 {"Proof|ID:387361496", 1646707, 45354},
 {"Proof|ID:387361496-1", 1646869, 45358},
 {"Proof|ID:70991441", 1647028, 45362},
 {"Proof|ID:70991441-1", 1647187, 45366},
 {"Proof|ID:365249767", 1647345, 45370},
 {"Proof|ID:365249767-1", 1647507, 45374},
 {"ID:16132582", 1647660, 45378},
 {"ID:742018993", 1647881, 45385},
 {"PolyTuples C 1", 1648130, 45392},
 {"ID:281673814", 1648377, 45399},
 {"PolyTuples C \[NotEqual] 0", 1648650, 45406},
 {"ID:122376381", 1648909, 45413},
 {"PolyTuples C=0", 1649158, 45420},
 {"ID:256177412", 1649405, 45427},
 {"PolyTuples C=0 seq", 1649662, 45434},
 {"ID:293354245", 1649913, 45441},
 {"PolyTuples C=0 seq'", 1650172, 45448},
 {"ID:129332919", 1650424, 45455},
 {"PolyTuples C prepend", 1650685, 45462},
 {"Proof|ID:742018993", 1650944, 45469},
 {"Proof|ID:742018993-1", 1651106, 45473},
 {"Proof|ID:281673814", 1651266, 45477},
 {"Proof|ID:281673814-1", 1651428, 45481},
 {"Proof|ID:122376381", 1651588, 45485},
 {"Proof|ID:122376381-1", 1651748, 45489},
 {"Proof|ID:256177412", 1651906, 45493},
 {"Proof|ID:256177412-1", 1652068, 45497},
 {"Proof|ID:293354245", 1652228, 45501},
 {"Proof|ID:293354245-1", 1652390, 45505},
 {"Proof|ID:129332919", 1652550, 45509},
 {"Proof|ID:129332919-1", 1652711, 45513},
 {"ID:136746483", 1652864, 45517},
 {"ID:110691644", 1653087, 45524},
 {"ID:371581865", 1653310, 45531},
 {"PolyTuples + ProdDomain", 1653577, 45538},
 {"ID:694735421", 1653833, 45545},
 {"PolyTuples + isTuple", 1654093, 45552},
 {"ID:442662347", 1654345, 45559},
 {"PolyTuples + \[NotEqual] 0", 1654618, 45566},
 {"ID:66376992", 1654876, 45573},
 {"PolyTuples + pp \[SubsetEqual]", 1655155, 45580},
 {"Proof|ID:371581865", 1655422, 45587},
 {"Proof|ID:371581865-1", 1655584, 45591},
 {"Proof|ID:694735421", 1655744, 45595},
 {"Proof|ID:694735421-1", 1655905, 45599},
 {"Proof|ID:442662347", 1656064, 45603},
 {"Proof|ID:442662347-1", 1656226, 45607},
 {"Proof|ID:66376992", 1656385, 45611},
 {"Proof|ID:66376992-1", 1656544, 45615},
 {"ID:480657044", 1656696, 45619},
 {"ID:907803476", 1656919, 45626},
 {"ID:442268007", 1657143, 45633},
 {"PolyTuples + \[Succeeds] 1", 1657416, 45640},
 {"ID:26802188", 1657674, 45647},
 {"PolyTuples + \[Succeeds] 2", 1657945, 45654},
 {"ID:4890345", 1658200, 45661},
 {"PolyTuples + \[Succeeds] 3", 1658469, 45668},
 {"Proof|ID:442268007", 1658730, 45675},
 {"Proof|ID:442268007-1", 1658892, 45679},
 {"Proof|ID:26802188", 1659051, 45683},
 {"Proof|ID:26802188-1", 1659210, 45687},
 {"Proof|ID:4890345", 1659366, 45691},
 {"Proof|ID:4890345-1", 1659523, 45695},
 {"ID:212784844", 1659674, 45699},
 {"ID:106856858", 1659897, 45706},
 {"PolyTuples - isTuple", 1660157, 45713},
 {"ID:200895043", 1660409, 45720},
 {"PolyTuples - length", 1660667, 45727},
 {"ID:19952680", 1660917, 45734},
 {"PolyTuples - pp", 1661166, 45741},
 {"Proof|ID:106856858", 1661418, 45748},
 {"Proof|ID:106856858-1", 1661579, 45752},
 {"Proof|ID:200895043", 1661738, 45756},
 {"Proof|ID:200895043-1", 1661899, 45760},
 {"Proof|ID:19952680", 1662057, 45764},
 {"Proof|ID:19952680-1", 1662216, 45768},
 {"ID:118174250", 1662368, 45772},
 {"PolyTuples isMonomial I", 1662635, 45779},
 {"ID:78163703", 1662890, 45786},
 {"ID:748742057", 1663111, 45793},
 {"PolyTuples isMonomial II", 1663380, 45800},
 {"ID:104425812", 1663637, 45807},
 {"PolyTuples isMonomial III", 1663908, 45814},
 {"ID:191443971", 1664166, 45821},
 {"PolyTuples \[CenterDot]", 1664433, 45828},
 {"Proof|ID:118174250", 1664695, 45835},
 {"Proof|ID:118174250-1", 1664856, 45839},
 {"Proof|ID:748742057", 1665015, 45843},
 {"Proof|ID:748742057-1", 1665177, 45847},
 {"Proof|ID:104425812", 1665337, 45851},
 {"Proof|ID:104425812-1", 1665498, 45855},
 {"Proof|ID:191443971", 1665657, 45859},
 {"Proof|ID:191443971-1", 1665819, 45863},
 {"ID:956509664", 1665973, 45867},
 {"ID:52346492", 1666195, 45874},
 {"PolyTuples first index", 1666463, 45881},
 {"ID:3252425", 1666719, 45888},
 {"PolyTuples isLP", 1666966, 45895},
 {"ID:301012536", 1667210, 45902},
 {"PolyTuples lp", 1667456, 45909},
 {"ID:499482967", 1667701, 45916},
 {"PolyTuples lc", 1667947, 45923},
 {"ID:77776384", 1668191, 45930},
 {"PolyTuples lm", 1668435, 45937},
 {"Proof|ID:52346492", 1668683, 45944},
 {"Proof|ID:52346492-1", 1668841, 45948},
 {"Proof|ID:3252425", 1668996, 45952},
 {"Proof|ID:3252425-1", 1669152, 45956},
 {"Proof|ID:301012536", 1669308, 45960},
 {"Proof|ID:301012536-1", 1669470, 45964},
 {"Proof|ID:499482967", 1669630, 45968},
 {"Proof|ID:499482967-1", 1669792, 45972},
 {"Proof|ID:77776384", 1669951, 45976},
 {"Proof|ID:77776384-1", 1670110, 45980},
 {"ID:1015977913", 1670263, 45984},
 {"ID:731677923", 1670489, 45991},
 {"ID:23863949", 1670712, 45998},
 {"PolyTuples lp seq", 1670966, 46005},
 {"ID:721799663", 1671215, 46012},
 {"PolyTuples lc seq", 1671471, 46019},
 {"ID:2939774", 1671720, 46026},
 {"PolyTuples lm seq", 1671972, 46033},
 {"ID:172576768", 1672219, 46040},
 {"PolyTuples R seq", 1672473, 46047},
 {"ID:3206806", 1672721, 46054},
 {"PolyTuples \[CenterDot] lp lc", 1672997, 46061},
 {"Proof|ID:23863949", 1673261, 46068},
 {"Proof|ID:23863949-1", 1673419, 46072},
 {"Proof|ID:721799663", 1673576, 46076},
 {"Proof|ID:721799663-1", 1673739, 46080},
 {"Proof|ID:2939774", 1673898, 46084},
 {"Proof|ID:2939774-1", 1674054, 46088},
 {"Proof|ID:172576768", 1674210, 46092},
 {"Proof|ID:172576768-1", 1674373, 46096},
 {"Proof|ID:3206806", 1674532, 46100},
 {"Proof|ID:3206806-1", 1674688, 46104},
 {"ID:31351138", 1674837, 46108},
 {"\[ScriptCapitalR] isDomain", 1675108, 46115},
 {"ID:632462358", 1675365, 46122},
 {"\[ScriptCapitalT] isDomain", 1675638, 46129},
 {"ID:404915746", 1675897, 46136},
 {"PolyTuples isDomain", 1676156, 46143},
 {"Proof|ID:31351138", 1676413, 46150},
 {"Proof|ID:31351138-1", 1676572, 46154},
 {"Proof|ID:632462358", 1676730, 46158},
 {"Proof|ID:632462358-1", 1676892, 46162},
 {"Proof|ID:404915746", 1677052, 46166},
 {"Proof|ID:404915746-1", 1677214, 46170},
 {"ID:3279840", 1677366, 46174},
 {"ID:141435198", 1677586, 46181},
 {"PolyTuples membership C", 1677854, 46188},
 {"ID:45852914", 1678110, 46195},
 {"PolyTuples membership 0", 1678375, 46202},
 {"ID:320592129", 1678629, 46209},
 {"PolyTuples membership 1", 1678896, 46216},
 {"ID:93322392", 1679151, 46223},
 {"PolyTuples membership +", 1679416, 46230},
 {"ID:6751189", 1679668, 46237},
 {"PolyTuples membership -", 1679931, 46244},
 {"ID:179493910", 1680183, 46251},
 {"PolyTuples supp", 1680435, 46258},
 {"Proof|ID:141435198", 1680690, 46265},
 {"Proof|ID:141435198-1", 1680852, 46269},
 {"Proof|ID:45852914", 1681011, 46273},
 {"Proof|ID:45852914-1", 1681169, 46277},
 {"Proof|ID:320592129", 1681326, 46281},
 {"Proof|ID:320592129-1", 1681487, 46285},
 {"Proof|ID:93322392", 1681645, 46289},
 {"Proof|ID:93322392-1", 1681805, 46293},
 {"Proof|ID:6751189", 1681962, 46297},
 {"Proof|ID:6751189-1", 1682117, 46301},
 {"Proof|ID:179493910", 1682272, 46305},
 {"Proof|ID:179493910-1", 1682435, 46309},
 {"ID:644982519", 1682590, 46313},
 {"PolyTuples hasDecisionPredicate", 1682873, 46320},
 {"Proof|ID:644982519", 1683143, 46327},
 {"Proof|ID:644982519-1", 1683305, 46331},
 {"ID:114843091", 1683459, 46335},
 {"PolyTuples containsMonomials", 1683736, 46342},
 {"Proof|ID:114843091", 1684003, 46349},
 {"Proof|ID:114843091-1", 1684165, 46353},
 {"ID:78312379", 1684318, 46357},
 {"ID:210280178", 1684540, 46364},
 {"ID:261724624", 1684765, 46371},
 {"PolyTuples + C 1", 1685020, 46378},
 {"ID:115965410", 1685271, 46385},
 {"PolyTuples + C 2", 1685526, 46392},
 {"ID:384214099", 1685777, 46399},
 {"PolyTuples + C 3", 1686032, 46406},
 {"ID:56134460", 1686282, 46413},
 {"PolyTuples + C 4", 1686534, 46420},
 {"Proof|ID:261724624", 1686788, 46427},
 {"Proof|ID:261724624-1", 1686952, 46431},
 {"Proof|ID:115965410", 1687114, 46435},
 {"Proof|ID:115965410-1", 1687278, 46439},
 {"Proof|ID:384214099", 1687440, 46443},
 {"Proof|ID:384214099-1", 1687603, 46447},
 {"Proof|ID:56134460", 1687763, 46451},
 {"Proof|ID:56134460-1", 1687924, 46455},
 {"ID:235178571", 1688078, 46459},
 {"ID:474434374", 1688302, 46466},
 {"PolyTuples C 2", 1688552, 46473},
 {"ID:108633346", 1688800, 46480},
 {"ID:611741232", 1689024, 46487},
 {"PolyTuples C +", 1689274, 46494},
 {"ID:438277340", 1689522, 46501},
 {"PolyTuples C - unary", 1689784, 46508},
 {"ID:412638768", 1690038, 46515},
 {"PolyTuples C -", 1690288, 46522},
 {"Proof|ID:474434374", 1690542, 46529},
 {"Proof|ID:474434374-1", 1690704, 46533},
 {"Proof|ID:611741232", 1690864, 46537},
 {"Proof|ID:611741232-1", 1691027, 46541},
 {"Proof|ID:438277340", 1691188, 46545},
 {"Proof|ID:438277340-1", 1691351, 46549},
 {"Proof|ID:412638768", 1691512, 46553},
 {"Proof|ID:412638768-1", 1691675, 46557},
 {"ID:392022323", 1691830, 46561},
 {"PolyTuples hasCoefficientFunction", 1692117, 46568},
 {"Proof|ID:392022323", 1692389, 46575},
 {"Proof|ID:392022323-1", 1692550, 46579},
 {"ID:23554005", 1692702, 46583},
 {"ID:392464903", 1692924, 46590},
 {"ID:161672314", 1693148, 46597},
 {"PolyTuples \[CircleTimes] empty", 1693432, 46604},
 {"ID:71703914", 1693696, 46611},
 {"PolyTuples \[CircleTimes] prepend", 1693982, 46618},
 {"ID:127729286", 1694247, 46625},
 {"ID:2718947", 1694469, 46632},
 {"PolyTuples \[CircleTimes] lm", 1694743, 46639},
 {"ID:204263407", 1695001, 46646},
 {"PolyTuples membership \[CircleTimes]", 1695295, 46653},
 {"ID:134536220", 1695565, 46660},
 {"PolyTuples \[CircleTimes] \[RawStar]", 1695859, 46667},
 {"ID:387729163", 1696129, 46674},
 {"PolyTuples \[RawStar] alternative", 1696417, 46681},
 {"Proof|ID:161672314", 1696690, 46688},
 {"Proof|ID:161672314-1", 1696853, 46692},
 {"Proof|ID:71703914", 1697013, 46696},
 {"Proof|ID:71703914-1", 1697173, 46700},
 {"Proof|ID:2718947", 1697330, 46704},
 {"Proof|ID:2718947-1", 1697487, 46708},
 {"Proof|ID:204263407", 1697644, 46712},
 {"Proof|ID:204263407-1", 1697807, 46716},
 {"Proof|ID:134536220", 1697968, 46720},
 {"Proof|ID:134536220-1", 1698131, 46724},
 {"Proof|ID:387729163", 1698292, 46728},
 {"Proof|ID:387729163-1", 1698455, 46732},
 {"ID:65280633", 1698609, 46736},
 {"PolyTuples hasMultiplication", 1698884, 46743},
 {"Proof|ID:65280633", 1699148, 46750},
 {"Proof|ID:65280633-1", 1699307, 46754},
 {"ID:48571309", 1699458, 46758},
 {"PolyTuples \[ForAll] alternative 2", 1699746, 46765},
 {"ID:617696571", 1700012, 46772},
 {"ID:776548", 1700233, 46779},
 {"PolyTuples=empty sufficient", 1700503, 46786},
 {"ID:304028020", 1700758, 46793},
 {"PolyTuples extensional", 1701024, 46800},
 {"Proof|ID:48571309", 1701285, 46807},
 {"Proof|ID:48571309-1", 1701445, 46811},
 {"Proof|ID:776548", 1701601, 46815},
 {"Proof|ID:776548-1", 1701755, 46819},
 {"Proof|ID:304028020", 1701910, 46823},
 {"Proof|ID:304028020-1", 1702072, 46827},
 {"ID:589891585", 1702226, 46831},
 {"PolyTuples isExtensional", 1702495, 46838},
 {"Proof|ID:589891585", 1702758, 46845},
 {"Proof|ID:589891585-1", 1702920, 46849},
 {"ID:160815443", 1703074, 46853},
 {"PolyTuples isOrderedRec", 1703341, 46860},
 {"ID:397107848", 1703597, 46867},
 {"PolyTuples isOrdered", 1703858, 46874},
 {"Proof|ID:160815443", 1704117, 46881},
 {"Proof|ID:160815443-1", 1704277, 46885},
 {"Proof|ID:397107848", 1704435, 46889},
 {"Proof|ID:397107848-1", 1704595, 46893},
 {"ID:156782366", 1704747, 46897},
 {"PolyTuples hasMultipliers", 1705018, 46904},
 {"Proof|ID:156782366", 1705282, 46911},
 {"Proof|ID:156782366-1", 1705444, 46915},
 {"ID:667417187", 1705598, 46919},
 {"PolyTuples isReductionPolynomialDomain", 1705895, 46926},
 {"Proof|ID:667417187", 1706172, 46933},
 {"Proof|ID:667417187-1", 1706334, 46937},
 {"ID:714775050", 1706488, 46941},
 {"ID:15424584", 1706711, 46948},
 {"PolyTuples lcrd alternative", 1706986, 46955},
 {"ID:168826371", 1707246, 46962},
 {"PolyTuples lcrd unary alternative", 1707534, 46969},
 {"ID:228703376", 1707801, 46976},
 {"PolyTuples lcrd lc lp", 1708066, 46983},
 {"ID:354399952", 1708322, 46990},
 {"PolyTuples lcrd length", 1708588, 46997},
 {"Proof|ID:15424584", 1708849, 47004},
 {"Proof|ID:15424584-1", 1709010, 47008},
 {"Proof|ID:168826371", 1709170, 47012},
 {"Proof|ID:168826371-1", 1709333, 47016},
 {"Proof|ID:228703376", 1709494, 47020},
 {"Proof|ID:228703376-1", 1709657, 47024},
 {"Proof|ID:354399952", 1709818, 47028},
 {"Proof|ID:354399952-1", 1709982, 47032},
 {"ID:479598778", 1710138, 47036},
 {"PolyTuples hasLCRD", 1710395, 47043},
 {"Proof|ID:479598778", 1710652, 47050},
 {"Proof|ID:479598778-1", 1710814, 47054},
 {"ID:463361785", 1710968, 47058},
 {"PolyTuples hasRDM", 1711223, 47065},
 {"Proof|ID:463361785", 1711479, 47072},
 {"Proof|ID:463361785-1", 1711641, 47076},
 {"ID:7345776", 1711793, 47080},
 {"PolyTuples hasCPM", 1712044, 47087},
 {"Proof|ID:7345776", 1712294, 47094},
 {"Proof|ID:7345776-1", 1712449, 47098},
 {"ID:1014777887", 1712599, 47102},
 {"PolyTuples isAlgoReductionPolynomialDomain", 1712906, 47109},
 {"Proof|ID:1014777887", 1713190, 47116},
 {"Proof|ID:1014777887-1", 1713354, 47120},
 {"ID:67407018", 1713508, 47124},
 {"PolyTuples isAlgoReductionRing", 1713788, 47131},
 {"Proof|ID:67407018", 1714055, 47138},
 {"Proof|ID:67407018-1", 1714214, 47142}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[579, 22, 159, 2, 132, "Title",
 CellID->526162817],
Cell[CellGroupData[{
Cell[763, 28, 118, 2, 73, "Section",
 CellID->524135902],
Cell[CellGroupData[{
Cell[906, 34, 175, 3, 70, "Subsection",
 CellID->1085100381],
Cell[1084, 39, 2541, 53, 70, "Text",
 CellID->700836624]
}, Closed]],
Cell[CellGroupData[{
Cell[3662, 97, 125, 2, 70, "Subsection",
 CellID->265926963],
Cell[3790, 101, 344, 8, 70, "Input",
 CellID->498486718],
Cell[4137, 111, 310, 8, 70, "Input",
 CellID->36134333],
Cell[4450, 121, 315, 8, 70, "Input",
 CellID->4041358],
Cell[4768, 131, 311, 8, 70, "Input",
 CellID->200738036],
Cell[5082, 141, 323, 8, 70, "Input",
 CellID->582390848],
Cell[5408, 151, 306, 8, 70, "Input",
 CellID->194899501],
Cell[5717, 161, 355, 8, 70, "Input",
 CellID->713552420],
Cell[6075, 171, 357, 8, 70, "Input",
 CellID->233536255],
Cell[6435, 181, 355, 8, 70, "Input",
 CellID->14602192],
Cell[6793, 191, 151, 3, 70, "Input",
 CellID->2803571],
Cell[6947, 196, 203, 4, 70, "Input",
 CellID->31571396],
Cell[7153, 202, 153, 3, 70, "Input",
 CellID->794947268],
Cell[CellGroupData[{
Cell[7331, 209, 1468, 32, 70, "Input",
 CellID->20666029],
Cell[8802, 243, 153, 4, 70, "Output",
 CellID->44505675]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[9004, 253, 124, 2, 70, "Subsection",
 CellID->1386333768],
Cell[CellGroupData[{
Cell[9153, 259, 278, 6, 70, "Input",
 CellID->1039312110],
Cell[CellGroupData[{
Cell[9456, 269, 1004, 20, 70, "Print",
 CellID->332409596],
Cell[10463, 291, 734, 15, 70, "Print",
 CellID->149409841],
Cell[11200, 308, 736, 15, 70, "Print",
 CellID->172974039],
Cell[11939, 325, 737, 15, 70, "Print",
 CellID->49653511],
Cell[12679, 342, 738, 15, 70, "Print",
 CellID->473922720],
Cell[13420, 359, 535, 10, 70, "Print",
 CellID->257520971],
Cell[13958, 371, 525, 10, 70, "Print",
 CellID->159051417],
Cell[14486, 383, 723, 15, 70, "Print",
 CellID->71026772],
Cell[15212, 400, 725, 15, 70, "Print",
 CellID->38953335],
Cell[15940, 417, 728, 15, 70, "Print",
 CellID->545021598],
Cell[16671, 434, 728, 15, 70, "Print",
 CellID->206823069],
Cell[17402, 451, 728, 15, 70, "Print",
 CellID->165393538],
Cell[18133, 468, 726, 15, 70, "Print",
 CellID->4978992],
Cell[18862, 485, 728, 15, 70, "Print",
 CellID->799672086],
Cell[19593, 502, 728, 15, 70, "Print",
 CellID->512697675],
Cell[20324, 519, 728, 15, 70, "Print",
 CellID->109571920],
Cell[21055, 536, 537, 10, 70, "Print",
 CellID->102929758],
Cell[21595, 548, 516, 10, 70, "Print",
 CellID->712415489]
}, Open  ]],
Cell[22126, 561, 225, 6, 70, "Output",
 CellID->143333797]
}, Open  ]],
Cell[22366, 570, 325, 6, 70, "Input",
 CellID->19428626],
Cell[CellGroupData[{
Cell[22716, 580, 541, 12, 70, "Input",
 CellID->1909803754],
Cell[23260, 594, 735, 12, 70, "Output",
 CellID->636966437]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[24044, 612, 128, 2, 70, "Subsection",
 CellID->1454182661],
Cell[24175, 616, 986, 25, 70, "Input",
 CellID->2074904605],
Cell[25164, 643, 416, 9, 70, "Input",
 CellID->1980565193],
Cell[25583, 654, 2163, 55, 70, "Input",
 CellID->658527806],
Cell[27749, 711, 169, 3, 70, "Input",
 CellID->444074165],
Cell[27921, 716, 260, 5, 70, "Input",
 CellID->69584241]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[28230, 727, 144, 2, 43, "Section",
 CellID->1575793246],
Cell[28377, 731, 6914, 154, 70, "Input",
 InitializationCell->True,
 CellID->1559908265],
Cell[35294, 887, 613, 16, 70, "Input",
 InitializationCell->True,
 CellID->423723764],
Cell[35910, 905, 780, 22, 70, "Input",
 InitializationCell->True,
 CellID->142555754],
Cell[36693, 929, 3302, 67, 70, "Input",
 InitializationCell->True,
 CellID->751753410],
Cell[39998, 998, 3472, 76, 70, "Input",
 InitializationCell->True,
 CellID->1199342921]
}, Closed]],
Cell[CellGroupData[{
Cell[43507, 1079, 250, 4, 43, "Section",
 CellID->1600179867],
Cell[CellGroupData[{
Cell[43782, 1087, 527, 9, 38, "Subsection",
 CellID->579563147],
Cell[CellGroupData[{
Cell[44334, 1100, 149, 3, 28, "Subsubsection",
 CellID->79894345],
Cell[44486, 1105, 665, 17, 31, "GlobalDeclaration",
 CellTags->{
  "ID:303368290", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->303368290],
Cell[45154, 1124, 91, 2, 12, "OpenEnvironment",
 CellID->630672694],
Cell[CellGroupData[{
Cell[45270, 1130, 560, 13, 43, "EnvironmentHeader",
 CellID->264955074],
Cell[45833, 1145, 2342, 60, 137, "FormalTextInputFormula",
 CellTags->{
  "ID:603497554", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain"},
 CellID->603497554],
Cell[48178, 1207, 3150, 87, 135, "FormalTextInputFormula",
 CellTags->{
  "ID:49777475", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasDecisionPredicatePP"},
 CellID->49777475],
Cell[51331, 1296, 5696, 147, 224, "FormalTextInputFormula",
 CellTags->{
  "ID:225295367", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasExponentFunctionPP"},
 CellID->225295367],
Cell[57030, 1445, 2105, 64, 52, "FormalTextInputFormula",
 CellTags->{
  "ID:292978558", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExtensionalPP"},
 CellID->292978558],
Cell[59138, 1511, 1950, 58, 63, "FormalTextInputFormula",
 CellTags->{
  "ID:340781435", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasDegreeFunctionPP"},
 CellID->340781435],
Cell[61091, 1571, 2199, 68, 52, "FormalTextInputFormula",
 CellTags->{
  "ID:480617864", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "hasDivisibilityPP"},
 CellID->480617864],
Cell[63293, 1641, 109, 2, 26, "EndEnvironmentMarker",
 CellID->380818943]
}, Open  ]],
Cell[63417, 1646, 92, 2, 16, "CloseEnvironment",
 CellID->707551302],
Cell[63512, 1650, 46, 1, 12, "OpenEnvironment",
 CellID->15645017],
Cell[CellGroupData[{
Cell[63583, 1655, 375, 9, 43, "EnvironmentHeader",
 CellID->417009435],
Cell[63961, 1666, 2958, 87, 115, "FormalTextInputFormula",
 CellTags->{
  "ID:897055662", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isTermOrder"},
 CellID->897055662],
Cell[66922, 1755, 65, 1, 26, "EndEnvironmentMarker",
 CellID->144066118]
}, Open  ]],
Cell[67002, 1759, 48, 1, 16, "CloseEnvironment",
 CellID->378822340]
}, Closed]],
Cell[CellGroupData[{
Cell[67087, 1765, 252, 4, 22, "Subsubsection",
 CellID->44692600],
Cell[67342, 1771, 47, 1, 70, "OpenEnvironment",
 CellID->473301257],
Cell[CellGroupData[{
Cell[67414, 1776, 406, 10, 70, "EnvironmentHeader",
 CellID->200657340],
Cell[67823, 1788, 703, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:532395027", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->532395027],
Cell[68529, 1808, 854, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:160997957", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain dim"},
 CellID->160997957],
Cell[69386, 1832, 818, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:83558965", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasDecisionPredicate"},
 CellID->83558965],
Cell[70207, 1856, 811, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:349898371", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasExponentFunction"},
 CellID->349898371],
Cell[71021, 1879, 790, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:200216737", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isExtensional"},
 CellID->200216737],
Cell[71814, 1902, 803, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:276506833", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasDegreeFunction"},
 CellID->276506833],
Cell[72620, 1925, 821, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:135986343", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain hasDivisibility"},
 CellID->135986343],
Cell[73444, 1948, 1254, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:122009120", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership exp"},
 CellID->122009120],
Cell[74701, 1985, 65, 1, 70, "EndEnvironmentMarker",
 CellID->711208495]
}, Open  ]],
Cell[74781, 1989, 48, 1, 70, "CloseEnvironment",
 CellID->141443493],
Cell[CellGroupData[{
Cell[74854, 1994, 2458, 74, 70, "ProofInfo",
 CellTags->{"Proof|ID:160997957", "Proof|ID:160997957-1"},
 CellID->4369865],
Cell[77315, 2070, 479, 12, 70, "ProofInfoBody",
 CellID->74490556]
}, Closed]],
Cell[CellGroupData[{
Cell[77831, 2087, 2316, 67, 70, "ProofInfo",
 CellTags->{"Proof|ID:83558965", "Proof|ID:83558965-1"},
 CellID->777822562],
Cell[80150, 2156, 479, 12, 70, "ProofInfoBody",
 CellID->348442790]
}, Closed]],
Cell[CellGroupData[{
Cell[80666, 2173, 2324, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:349898371", "Proof|ID:349898371-1"},
 CellID->51022891],
Cell[82993, 2243, 480, 12, 70, "ProofInfoBody",
 CellID->494240957]
}, Closed]],
Cell[CellGroupData[{
Cell[83510, 2260, 2284, 66, 70, "ProofInfo",
 CellTags->{"Proof|ID:200216737", "Proof|ID:200216737-1"},
 CellID->19036167],
Cell[85797, 2328, 480, 12, 70, "ProofInfoBody",
 CellID->454301965]
}, Closed]],
Cell[CellGroupData[{
Cell[86314, 2345, 2309, 67, 70, "ProofInfo",
 CellTags->{"Proof|ID:276506833", "Proof|ID:276506833-1"},
 CellID->447792752],
Cell[88626, 2414, 480, 12, 70, "ProofInfoBody",
 CellID->204590069]
}, Closed]],
Cell[CellGroupData[{
Cell[89143, 2431, 2293, 66, 70, "ProofInfo",
 CellTags->{"Proof|ID:135986343", "Proof|ID:135986343-1"},
 CellID->380432363],
Cell[91439, 2499, 524, 13, 70, "ProofInfoBody",
 CellID->100546045]
}, Closed]],
Cell[CellGroupData[{
Cell[92000, 2517, 3752, 101, 70, "ProofInfo",
 CellTags->{"Proof|ID:122009120", "Proof|ID:122009120-1"},
 CellID->131414947],
Cell[95755, 2620, 479, 12, 70, "ProofInfoBody",
 CellID->34728721]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[96283, 2638, 242, 5, 22, "Subsubsection",
 CellID->180962323],
Cell[96528, 2645, 703, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:185429719", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->185429719],
Cell[97234, 2665, 47, 1, 70, "OpenEnvironment",
 CellID->430279676],
Cell[CellGroupData[{
Cell[97306, 2670, 454, 10, 70, "EnvironmentHeader",
 CellID->158100854],
Cell[97763, 2682, 794, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:81626705", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isDomain"},
 CellID->81626705],
Cell[98560, 2705, 1095, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:154517895", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership \[RawStar]", "IR"},
 CellID->154517895],
Cell[99658, 2738, 993, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:226316196", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain \[RawStar] isAssoc"},
 CellID->226316196],
Cell[100654, 2770, 960, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:96938761", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain \[RawStar] isComm"},
 CellID->96938761],
Cell[101617, 2802, 1024, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:499736780", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain 1 isNeutral"},
 CellID->499736780],
Cell[102644, 2835, 1458, 47, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:745587327", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain \[RawStar] lcanc"},
 CellID->745587327],
Cell[104105, 2884, 65, 1, 70, "EndEnvironmentMarker",
 CellID->131282212]
}, Open  ]],
Cell[104185, 2888, 48, 1, 70, "CloseEnvironment",
 CellID->207922316],
Cell[CellGroupData[{
Cell[104258, 2893, 2256, 66, 70, "ProofInfo",
 CellTags->{"Proof|ID:81626705", "Proof|ID:81626705-1"},
 CellID->281230526],
Cell[106517, 2961, 479, 12, 70, "ProofInfoBody",
 CellID->215127808]
}, Closed]],
Cell[CellGroupData[{
Cell[107033, 2978, 3605, 94, 70, "ProofInfo",
 CellTags->{"Proof|ID:154517895", "Proof|ID:154517895-1"},
 CellID->527357288],
Cell[110641, 3074, 480, 12, 70, "ProofInfoBody",
 CellID->122633837]
}, Closed]],
Cell[CellGroupData[{
Cell[111158, 3091, 2524, 72, 70, "ProofInfo",
 CellTags->{"Proof|ID:226316196", "Proof|ID:226316196-1"},
 CellID->31171248],
Cell[113685, 3165, 480, 12, 70, "ProofInfoBody",
 CellID->166332303]
}, Closed]],
Cell[CellGroupData[{
Cell[114202, 3182, 2514, 72, 70, "ProofInfo",
 CellTags->{"Proof|ID:96938761", "Proof|ID:96938761-1"},
 CellID->6993021],
Cell[116719, 3256, 478, 12, 70, "ProofInfoBody",
 CellID->47945419]
}, Closed]],
Cell[CellGroupData[{
Cell[117234, 3273, 2749, 76, 70, "ProofInfo",
 CellTags->{"Proof|ID:499736780", "Proof|ID:499736780-1"},
 CellID->247701442],
Cell[119986, 3351, 479, 12, 70, "ProofInfoBody",
 CellID->34826465]
}, Closed]],
Cell[CellGroupData[{
Cell[120502, 3368, 4697, 120, 70, "ProofInfo",
 CellTags->{"Proof|ID:745587327", "Proof|ID:745587327-1"},
 CellID->99539526],
Cell[125202, 3490, 478, 12, 70, "ProofInfoBody",
 CellID->1175411]
}, Closed]],
Cell[125695, 3505, 46, 1, 70, "OpenEnvironment",
 CellID->65103093],
Cell[CellGroupData[{
Cell[125766, 3510, 390, 9, 70, "EnvironmentHeader",
 CellID->429762970],
Cell[126159, 3521, 804, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:333683969", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isCCMonoid"},
 CellID->333683969],
Cell[126966, 3544, 64, 1, 70, "EndEnvironmentMarker",
 CellID->14565163]
}, Open  ]],
Cell[127045, 3548, 47, 1, 70, "CloseEnvironment",
 CellID->47457899],
Cell[CellGroupData[{
Cell[127117, 3553, 2269, 66, 70, "ProofInfo",
 CellTags->{"Proof|ID:333683969", "Proof|ID:333683969-1"},
 CellID->190665795],
Cell[129389, 3621, 480, 12, 70, "ProofInfoBody",
 CellID->331229762]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[129918, 3639, 198, 3, 22, "Subsubsection",
 CellID->71800829],
Cell[130119, 3644, 47, 1, 70, "OpenEnvironment",
 CellID->656590807],
Cell[CellGroupData[{
Cell[130191, 3649, 398, 9, 70, "EnvironmentHeader",
 CellID->290258020],
Cell[130592, 3660, 700, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:274960855", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->274960855],
Cell[131295, 3680, 1031, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:52136714", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain lcm isJoin"},
 CellID->52136714],
Cell[132329, 3713, 760, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:35376325", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->35376325],
Cell[133092, 3735, 1279, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:138508454", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain div.1"},
 CellID->138508454],
Cell[134374, 3776, 1437, 46, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:14785826", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain div.2"},
 CellID->14785826],
Cell[135814, 3824, 65, 1, 70, "EndEnvironmentMarker",
 CellID->147667495]
}, Open  ]],
Cell[135894, 3828, 48, 1, 70, "CloseEnvironment",
 CellID->416318934],
Cell[CellGroupData[{
Cell[135967, 3833, 2766, 76, 70, "ProofInfo",
 CellTags->{"Proof|ID:52136714", "Proof|ID:52136714-1"},
 CellID->110160557],
Cell[138736, 3911, 478, 12, 70, "ProofInfoBody",
 CellID->72948561]
}, Closed]],
Cell[CellGroupData[{
Cell[139251, 3928, 4458, 114, 70, "ProofInfo",
 CellTags->{"Proof|ID:138508454", "Proof|ID:138508454-1"},
 CellID->245133977],
Cell[143712, 4044, 480, 12, 70, "ProofInfoBody",
 CellID->598366505]
}, Closed]],
Cell[CellGroupData[{
Cell[144229, 4061, 4740, 120, 70, "ProofInfo",
 CellTags->{"Proof|ID:14785826", "Proof|ID:14785826-1"},
 CellID->2945667],
Cell[148972, 4183, 477, 12, 70, "ProofInfoBody",
 CellID->9554763]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[149498, 4201, 195, 3, 22, "Subsubsection",
 CellID->194654141],
Cell[149696, 4206, 46, 1, 70, "OpenEnvironment",
 CellID->21521330],
Cell[CellGroupData[{
Cell[149767, 4211, 402, 9, 70, "EnvironmentHeader",
 CellID->129749829],
Cell[150172, 4222, 703, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:181034073", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->181034073],
Cell[150878, 4242, 738, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:943617998", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->943617998],
Cell[151619, 4263, 1424, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:706945753", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership deg ind"},
 CellID->706945753],
Cell[153046, 4303, 2172, 69, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:606418540", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain deg=0 ind"},
 CellID->606418540],
Cell[155221, 4374, 951, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:47160", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain membership deg", "IR"},
 CellID->47160],
Cell[156175, 4401, 1158, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:58897330", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain deg=0"},
 CellID->58897330],
Cell[157336, 4439, 1333, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:170894204", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain deg \[RawStar]"},
 CellID->170894204],
Cell[158672, 4480, 64, 1, 70, "EndEnvironmentMarker",
 CellID->31663413]
}, Open  ]],
Cell[158751, 4484, 48, 1, 70, "CloseEnvironment",
 CellID->458193557],
Cell[CellGroupData[{
Cell[158824, 4489, 4667, 121, 70, "ProofInfo",
 CellTags->{"Proof|ID:706945753", "Proof|ID:706945753-1"},
 CellID->905512569],
Cell[163494, 4612, 478, 12, 70, "ProofInfoBody",
 CellID->5970720]
}, Closed]],
Cell[CellGroupData[{
Cell[164009, 4629, 5957, 150, 70, "ProofInfo",
 CellTags->{"Proof|ID:606418540", "Proof|ID:606418540-1"},
 CellID->408182534],
Cell[169969, 4781, 480, 12, 70, "ProofInfoBody",
 CellID->193074823]
}, Closed]],
Cell[CellGroupData[{
Cell[170486, 4798, 3030, 85, 70, "ProofInfo",
 CellTags->{"Proof|ID:47160", "Proof|ID:47160-1"},
 CellID->168232727],
Cell[173519, 4885, 480, 12, 70, "ProofInfoBody",
 CellID->170693603]
}, Closed]],
Cell[CellGroupData[{
Cell[174036, 4902, 3461, 96, 70, "ProofInfo",
 CellTags->{"Proof|ID:58897330", "Proof|ID:58897330-1"},
 CellID->293749343],
Cell[177500, 5000, 479, 12, 70, "ProofInfoBody",
 CellID->470353965]
}, Closed]],
Cell[CellGroupData[{
Cell[178016, 5017, 4519, 114, 70, "ProofInfo",
 CellTags->{"Proof|ID:170894204", "Proof|ID:170894204-1"},
 CellID->366355002],
Cell[182538, 5133, 478, 12, 70, "ProofInfoBody",
 CellID->6017937]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[183065, 5151, 193, 3, 22, "Subsubsection",
 CellID->467568792],
Cell[183261, 5156, 47, 1, 70, "OpenEnvironment",
 CellID->187255726],
Cell[CellGroupData[{
Cell[183333, 5161, 464, 12, 70, "EnvironmentHeader",
 CellID->152834815],
Cell[183800, 5175, 1710, 47, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:49640627", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "increasing subsequence \[DoubleStruckCapitalN]0"},
 CellID->49640627],
Cell[185513, 5224, 65, 1, 70, "EndEnvironmentMarker",
 CellID->454910408]
}, Open  ]],
Cell[185593, 5228, 48, 1, 70, "CloseEnvironment",
 CellID->460125164],
Cell[CellGroupData[{
Cell[185666, 5233, 5144, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:49640627", "Proof|ID:49640627-1"},
 CellID->620572316],
Cell[190813, 5367, 479, 12, 70, "ProofInfoBody",
 CellID->563905649]
}, Closed]],
Cell[191307, 5382, 701, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:256365628", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->256365628],
Cell[192011, 5402, 47, 1, 70, "OpenEnvironment",
 CellID->346122881],
Cell[CellGroupData[{
Cell[192083, 5407, 399, 9, 70, "EnvironmentHeader",
 CellID->216115192],
Cell[192485, 5418, 1782, 54, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:443096051", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain divN"},
 CellID->443096051],
Cell[194270, 5474, 65, 1, 70, "EndEnvironmentMarker",
 CellID->306607494]
}, Open  ]],
Cell[194350, 5478, 47, 1, 70, "CloseEnvironment",
 CellID->45893119],
Cell[194400, 5481, 46, 1, 70, "OpenEnvironment",
 CellID->21114114],
Cell[CellGroupData[{
Cell[194471, 5486, 394, 9, 70, "EnvironmentHeader",
 CellID->13494331],
Cell[194868, 5497, 2207, 64, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:27206868", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain divN +1"},
 CellID->27206868],
Cell[197078, 5563, 1809, 51, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:184673495", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "increasing subsequence isExpPPDomain 0"},
 CellID->184673495],
Cell[198890, 5616, 1975, 57, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:43287351", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "increasing subsequence isExpPPDomain"},
 CellID->43287351],
Cell[200868, 5675, 65, 1, 70, "EndEnvironmentMarker",
 CellID->506192652]
}, Open  ]],
Cell[200948, 5679, 47, 1, 70, "CloseEnvironment",
 CellID->19274976],
Cell[CellGroupData[{
Cell[201020, 5684, 6427, 156, 70, "ProofInfo",
 CellTags->{"Proof|ID:27206868", "Proof|ID:27206868-1"},
 CellID->99745602],
Cell[207450, 5842, 478, 12, 70, "ProofInfoBody",
 CellID->98056265]
}, Closed]],
Cell[CellGroupData[{
Cell[207965, 5859, 5735, 144, 70, "ProofInfo",
 CellTags->{"Proof|ID:184673495", "Proof|ID:184673495-1"},
 CellID->104166294],
Cell[213703, 6005, 480, 12, 70, "ProofInfoBody",
 CellID->267847356]
}, Closed]],
Cell[CellGroupData[{
Cell[214220, 6022, 6360, 158, 70, "ProofInfo",
 CellTags->{"Proof|ID:43287351", "Proof|ID:43287351-1"},
 CellID->460075245],
Cell[220583, 6182, 479, 12, 70, "ProofInfoBody",
 CellID->392596903]
}, Closed]],
Cell[221077, 6197, 47, 1, 70, "OpenEnvironment",
 CellID->503713153],
Cell[CellGroupData[{
Cell[221149, 6202, 398, 9, 70, "EnvironmentHeader",
 CellID->315610582],
Cell[221550, 6213, 1022, 29, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:30919390", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isExpPPDomain isDicksonDomain"},
 CellID->30919390],
Cell[222575, 6244, 64, 1, 70, "EndEnvironmentMarker",
 CellID->23678156]
}, Open  ]],
Cell[222654, 6248, 48, 1, 70, "CloseEnvironment",
 CellID->272481360],
Cell[CellGroupData[{
Cell[222727, 6253, 2574, 72, 70, "ProofInfo",
 CellTags->{"Proof|ID:30919390", "Proof|ID:30919390-1"},
 CellID->278064001],
Cell[225304, 6327, 478, 12, 70, "ProofInfoBody",
 CellID->89121114]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[225831, 6345, 259, 4, 22, "Subsubsection",
 CellID->309098395],
Cell[226093, 6351, 47, 1, 70, "OpenEnvironment",
 CellID->214207545],
Cell[CellGroupData[{
Cell[226165, 6356, 387, 9, 70, "EnvironmentHeader",
 CellID->436328982],
Cell[226555, 6367, 781, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:323333890", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->323333890],
Cell[227339, 6389, 771, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:235080407", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->235080407],
Cell[228113, 6411, 1338, 43, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:691531428", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "term order |"},
 CellID->691531428],
Cell[229454, 6456, 989, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:772895867", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "term order isOrderEmbedding"},
 CellID->772895867],
Cell[230446, 6486, 819, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:83375799", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "term order isNoetherian"},
 CellID->83375799],
Cell[231268, 6509, 65, 1, 70, "EndEnvironmentMarker",
 CellID->368030963]
}, Open  ]],
Cell[231348, 6513, 46, 1, 70, "CloseEnvironment",
 CellID->7746425],
Cell[CellGroupData[{
Cell[231419, 6518, 5172, 130, 70, "ProofInfo",
 CellTags->{"Proof|ID:691531428", "Proof|ID:691531428-1"},
 CellID->26598154],
Cell[236594, 6650, 480, 12, 70, "ProofInfoBody",
 CellID->174801662]
}, Closed]],
Cell[CellGroupData[{
Cell[237111, 6667, 3285, 88, 70, "ProofInfo",
 CellTags->{"Proof|ID:772895867", "Proof|ID:772895867-1"},
 CellID->409263698],
Cell[240399, 6757, 480, 12, 70, "ProofInfoBody",
 CellID->511868347]
}, Closed]],
Cell[CellGroupData[{
Cell[240916, 6774, 3040, 84, 70, "ProofInfo",
 CellTags->{"Proof|ID:83375799", "Proof|ID:83375799-1"},
 CellID->549031787],
Cell[243959, 6860, 479, 12, 70, "ProofInfoBody",
 CellID->459125148]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[244499, 6879, 198, 3, 30, "Subsection",
 CellID->317472498],
Cell[CellGroupData[{
Cell[244722, 6886, 121, 2, 70, "Subsubsection",
 CellID->12853081],
Cell[244846, 6890, 46, 1, 70, "OpenEnvironment",
 CellID->72677768],
Cell[CellGroupData[{
Cell[244917, 6895, 401, 9, 70, "EnvironmentHeader",
 CellID->107281972],
Cell[245321, 6906, 680, 17, 70, "GlobalDeclaration",
 CellTags->{
  "ID:69380281", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->69380281],
Cell[246004, 6925, 723, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:70274751", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->70274751],
Cell[246730, 6946, 669, 17, 70, "GlobalDeclaration",
 CellTags->{
  "ID:417719229", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->417719229],
Cell[247402, 6965, 1279, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:296713523", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP \[Element]"},
 CellID->296713523],
Cell[248684, 7009, 818, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:275450356", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP dim"},
 CellID->275450356],
Cell[249505, 7033, 1082, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:126765714", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP exp"},
 CellID->126765714],
Cell[250590, 7068, 810, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:300238463", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP 1"},
 CellID->300238463],
Cell[251403, 7092, 863, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:91738893", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP \[RawStar]"},
 CellID->91738893],
Cell[252269, 7119, 867, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1017172830", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP /"},
 CellID->1017172830],
Cell[253139, 7146, 923, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:62103041", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP lcm"},
 CellID->62103041],
Cell[254065, 7174, 1335, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:229667832", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP |"},
 CellID->229667832],
Cell[255403, 7218, 935, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:687245437", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP deg"},
 CellID->687245437],
Cell[256341, 7246, 1143, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:353388395", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP \[Succeeds]"},
 CellID->353388395],
Cell[257487, 7283, 1096, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:144609801", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP <<"},
 CellID->144609801],
Cell[258586, 7321, 1096, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:189296520", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP >>"},
 CellID->189296520],
Cell[259685, 7359, 65, 1, 70, "EndEnvironmentMarker",
 CellID->170554027]
}, Open  ]],
Cell[259765, 7363, 48, 1, 70, "CloseEnvironment",
 CellID->887551028]
}, Closed]],
Cell[CellGroupData[{
Cell[259850, 7369, 292, 6, 70, "Subsubsection",
 CellID->75242152],
Cell[260145, 7377, 46, 1, 70, "OpenEnvironment",
 CellID->67119962],
Cell[CellGroupData[{
Cell[260216, 7382, 424, 11, 70, "EnvironmentHeader",
 CellID->40275436],
Cell[260643, 7395, 776, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:29953053", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->29953053],
Cell[261422, 7417, 869, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:309264225", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isExpPPDomain"},
 CellID->309264225],
Cell[262294, 7442, 1196, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:499712177", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isDicksonDomain"},
 CellID->499712177],
Cell[263493, 7478, 720, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:275069792", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->275069792],
Cell[264216, 7499, 886, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:834856233", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isTermOrder"},
 CellID->834856233],
Cell[265105, 7525, 1019, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:260893370", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "OrderedPP isCommPPDomain"},
 CellID->260893370],
Cell[266127, 7552, 65, 1, 70, "EndEnvironmentMarker",
 CellID->163286488]
}, Open  ]],
Cell[266207, 7556, 48, 1, 70, "CloseEnvironment",
 CellID->981953859],
Cell[CellGroupData[{
Cell[266280, 7561, 2640, 74, 70, "ProofInfo",
 CellTags->{"Proof|ID:309264225", "Proof|ID:309264225-1"},
 CellID->51410204],
Cell[268923, 7637, 478, 12, 70, "ProofInfoBody",
 CellID->3099999]
}, Closed]],
Cell[CellGroupData[{
Cell[269438, 7654, 3230, 86, 70, "ProofInfo",
 CellTags->{"Proof|ID:499712177", "Proof|ID:499712177-1"},
 CellID->50025897],
Cell[272671, 7742, 480, 12, 70, "ProofInfoBody",
 CellID->631880945]
}, Closed]],
Cell[CellGroupData[{
Cell[273188, 7759, 3324, 90, 70, "ProofInfo",
 CellTags->{"Proof|ID:834856233", "Proof|ID:834856233-1"},
 CellID->65135779],
Cell[276515, 7851, 480, 12, 70, "ProofInfoBody",
 CellID->687624990]
}, Closed]],
Cell[CellGroupData[{
Cell[277032, 7868, 3156, 86, 70, "ProofInfo",
 CellTags->{"Proof|ID:260893370", "Proof|ID:260893370-1"},
 CellID->97936963],
Cell[280191, 7956, 480, 12, 70, "ProofInfoBody",
 CellID->264288389]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[280744, 7976, 190, 3, 43, "Section",
 CellID->46760114],
Cell[CellGroupData[{
Cell[280959, 7983, 186, 3, 70, "Subsection",
 CellID->19116656],
Cell[CellGroupData[{
Cell[281170, 7990, 149, 3, 70, "Subsubsection",
 CellID->452220060],
Cell[281322, 7995, 47, 1, 70, "OpenEnvironment",
 CellID->883840617],
Cell[CellGroupData[{
Cell[281394, 8000, 392, 9, 70, "EnvironmentHeader",
 CellID->79250563],
Cell[281789, 8011, 2379, 71, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:223902051", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex"},
 CellID->223902051],
Cell[284171, 8084, 64, 1, 70, "EndEnvironmentMarker",
 CellID->86850538]
}, Open  ]],
Cell[284250, 8088, 48, 1, 70, "CloseEnvironment",
 CellID->119543112]
}, Closed]],
Cell[CellGroupData[{
Cell[284335, 8094, 321, 6, 70, "Subsubsection",
 CellID->585882636],
Cell[284659, 8102, 703, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:376155678", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->376155678],
Cell[285365, 8122, 47, 1, 70, "OpenEnvironment",
 CellID->175815204],
Cell[CellGroupData[{
Cell[285437, 8127, 482, 11, 70, "EnvironmentHeader",
 CellID->104761298],
Cell[285922, 8140, 773, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:310375877", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->310375877],
Cell[286698, 8162, 989, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:463764498", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex total"},
 CellID->463764498],
Cell[287690, 8191, 1249, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:6465370", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex asymmetric"},
 CellID->6465370],
Cell[288942, 8232, 1188, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:78173747", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex transitive"},
 CellID->78173747],
Cell[290133, 8269, 971, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:341999039", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex 1 min"},
 CellID->341999039],
Cell[291107, 8298, 1079, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:34516336", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex \[RawStar]"},
 CellID->34516336],
Cell[292189, 8331, 64, 1, 70, "EndEnvironmentMarker",
 CellID->51206845]
}, Open  ]],
Cell[292268, 8335, 48, 1, 70, "CloseEnvironment",
 CellID->212901590],
Cell[CellGroupData[{
Cell[292341, 8340, 4370, 116, 70, "ProofInfo",
 CellTags->{"Proof|ID:463764498", "Proof|ID:463764498-1"},
 CellID->355435549],
Cell[296714, 8458, 547, 13, 70, "ProofInfoBody",
 CellID->81470365]
}, Closed]],
Cell[CellGroupData[{
Cell[297298, 8476, 4844, 126, 70, "ProofInfo",
 CellTags->{"Proof|ID:6465370", "Proof|ID:6465370-1"},
 CellID->258075395],
Cell[302145, 8604, 548, 13, 70, "ProofInfoBody",
 CellID->146173952]
}, Closed]],
Cell[CellGroupData[{
Cell[302730, 8622, 5194, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:78173747", "Proof|ID:78173747-1"},
 CellID->631231714],
Cell[307927, 8756, 547, 13, 70, "ProofInfoBody",
 CellID->247058823]
}, Closed]],
Cell[CellGroupData[{
Cell[308511, 8774, 3599, 98, 70, "ProofInfo",
 CellTags->{"Proof|ID:341999039", "Proof|ID:341999039-1"},
 CellID->366853708],
Cell[312113, 8874, 480, 12, 70, "ProofInfoBody",
 CellID->235356693]
}, Closed]],
Cell[CellGroupData[{
Cell[312630, 8891, 5250, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:34516336", "Proof|ID:34516336-1"},
 CellID->188021315],
Cell[317883, 9025, 544, 13, 70, "ProofInfoBody",
 CellID->11866489]
}, Closed]],
Cell[318442, 9041, 46, 1, 70, "OpenEnvironment",
 CellID->15806341],
Cell[CellGroupData[{
Cell[318513, 9046, 380, 9, 70, "EnvironmentHeader",
 CellID->12908456],
Cell[318896, 9057, 861, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:414199224", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "Lex isTermOrder"},
 CellID->414199224],
Cell[319760, 9082, 65, 1, 70, "EndEnvironmentMarker",
 CellID->118815614]
}, Open  ]],
Cell[319840, 9086, 47, 1, 70, "CloseEnvironment",
 CellID->11715767],
Cell[CellGroupData[{
Cell[319912, 9091, 2558, 74, 70, "ProofInfo",
 CellTags->{"Proof|ID:414199224", "Proof|ID:414199224-1"},
 CellID->87459975],
Cell[322473, 9167, 480, 12, 70, "ProofInfoBody",
 CellID->920517951]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[323014, 9186, 198, 3, 70, "Subsection",
 CellID->417561155],
Cell[CellGroupData[{
Cell[323237, 9193, 122, 2, 70, "Subsubsection",
 CellID->46310993],
Cell[323362, 9197, 47, 1, 70, "OpenEnvironment",
 CellID->327096997],
Cell[CellGroupData[{
Cell[323434, 9202, 395, 9, 70, "EnvironmentHeader",
 CellID->210690636],
Cell[323832, 9213, 2280, 72, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:116320630", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex"},
 CellID->116320630],
Cell[326115, 9287, 65, 1, 70, "EndEnvironmentMarker",
 CellID->208585242]
}, Open  ]],
Cell[326195, 9291, 47, 1, 70, "CloseEnvironment",
 CellID->49927192]
}, Closed]],
Cell[CellGroupData[{
Cell[326279, 9297, 229, 5, 70, "Subsubsection",
 CellID->122425200],
Cell[326511, 9304, 703, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:106786369", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->106786369],
Cell[327217, 9324, 46, 1, 70, "OpenEnvironment",
 CellID->56827951],
Cell[CellGroupData[{
Cell[327288, 9329, 438, 11, 70, "EnvironmentHeader",
 CellID->168509974],
Cell[327729, 9342, 767, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:43019543", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->43019543],
Cell[328499, 9364, 998, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:75557233", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex total"},
 CellID->75557233],
Cell[329500, 9393, 1267, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:49531328", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex asymmetric"},
 CellID->49531328],
Cell[330770, 9434, 1208, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:260398187", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex transitive"},
 CellID->260398187],
Cell[331981, 9471, 977, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:59133109", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex 1 min"},
 CellID->59133109],
Cell[332961, 9500, 1096, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:572110615", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex \[RawStar]"},
 CellID->572110615],
Cell[334060, 9533, 65, 1, 70, "EndEnvironmentMarker",
 CellID->467550442]
}, Open  ]],
Cell[334140, 9537, 48, 1, 70, "CloseEnvironment",
 CellID->213596470],
Cell[CellGroupData[{
Cell[334213, 9542, 4383, 116, 70, "ProofInfo",
 CellTags->{"Proof|ID:75557233", "Proof|ID:75557233-1"},
 CellID->210478017],
Cell[338599, 9660, 544, 13, 70, "ProofInfoBody",
 CellID->35178206]
}, Closed]],
Cell[CellGroupData[{
Cell[339180, 9678, 4867, 126, 70, "ProofInfo",
 CellTags->{"Proof|ID:49531328", "Proof|ID:49531328-1"},
 CellID->563764769],
Cell[344050, 9806, 549, 13, 70, "ProofInfoBody",
 CellID->334172145]
}, Closed]],
Cell[CellGroupData[{
Cell[344636, 9824, 5223, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:260398187", "Proof|ID:260398187-1"},
 CellID->100109944],
Cell[349862, 9958, 548, 13, 70, "ProofInfoBody",
 CellID->450035939]
}, Closed]],
Cell[CellGroupData[{
Cell[350447, 9976, 3606, 98, 70, "ProofInfo",
 CellTags->{"Proof|ID:59133109", "Proof|ID:59133109-1"},
 CellID->171486032],
Cell[354056, 10076, 519, 13, 70, "ProofInfoBody",
 CellID->51248171]
}, Closed]],
Cell[CellGroupData[{
Cell[354612, 10094, 5273, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:572110615", "Proof|ID:572110615-1"},
 CellID->427513857],
Cell[359888, 10228, 548, 13, 70, "ProofInfoBody",
 CellID->171210370]
}, Closed]],
Cell[360451, 10244, 46, 1, 70, "OpenEnvironment",
 CellID->50768423],
Cell[CellGroupData[{
Cell[360522, 10249, 384, 9, 70, "EnvironmentHeader",
 CellID->584850280],
Cell[360909, 10260, 869, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:463127740", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegLex isTermOrder"},
 CellID->463127740],
Cell[361781, 10285, 64, 1, 70, "EndEnvironmentMarker",
 CellID->93774533]
}, Open  ]],
Cell[361860, 10289, 48, 1, 70, "CloseEnvironment",
 CellID->214012042],
Cell[CellGroupData[{
Cell[361933, 10294, 2571, 74, 70, "ProofInfo",
 CellTags->{"Proof|ID:463127740", "Proof|ID:463127740-1"},
 CellID->102580793],
Cell[364507, 10370, 479, 12, 70, "ProofInfoBody",
 CellID->73039304]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[365047, 10389, 214, 5, 70, "Subsection",
 CellID->187433264],
Cell[CellGroupData[{
Cell[365286, 10398, 154, 3, 70, "Subsubsection",
 CellID->290355525],
Cell[365443, 10403, 47, 1, 70, "OpenEnvironment",
 CellID->337107021],
Cell[CellGroupData[{
Cell[365515, 10408, 427, 11, 70, "EnvironmentHeader",
 CellID->513549404],
Cell[365945, 10421, 693, 17, 70, "GlobalDeclaration",
 CellTags->{
  "ID:267616510", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->267616510],
Cell[366641, 10440, 2155, 64, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1929370", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex"},
 CellID->1929370],
Cell[368799, 10506, 1994, 62, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:116917210", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex"},
 CellID->116917210],
Cell[370796, 10570, 65, 1, 70, "EndEnvironmentMarker",
 CellID->494887762]
}, Open  ]],
Cell[370876, 10574, 47, 1, 70, "CloseEnvironment",
 CellID->50267427]
}, Closed]],
Cell[CellGroupData[{
Cell[370960, 10580, 237, 5, 70, "Subsubsection",
 CellID->148960609],
Cell[371200, 10587, 701, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:431887307", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->431887307],
Cell[371904, 10607, 47, 1, 70, "OpenEnvironment",
 CellID->181550980],
Cell[CellGroupData[{
Cell[371976, 10612, 417, 10, 70, "EnvironmentHeader",
 CellID->417200827],
Cell[372396, 10624, 769, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:204943512", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->204943512],
Cell[373168, 10646, 1001, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:781115991", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex total"},
 CellID->781115991],
Cell[374172, 10675, 1264, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2983958", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex asymmetric"},
 CellID->2983958],
Cell[375439, 10716, 1208, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:240564679", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex transitive"},
 CellID->240564679],
Cell[376650, 10753, 1092, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:299677825", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "RevLex \[RawStar]"},
 CellID->299677825],
Cell[377745, 10786, 64, 1, 70, "EndEnvironmentMarker",
 CellID->92970195]
}, Open  ]],
Cell[377824, 10790, 48, 1, 70, "CloseEnvironment",
 CellID->522695883],
Cell[CellGroupData[{
Cell[377897, 10795, 4388, 116, 70, "ProofInfo",
 CellTags->{"Proof|ID:781115991", "Proof|ID:781115991-1"},
 CellID->124604986],
Cell[382288, 10913, 480, 12, 70, "ProofInfoBody",
 CellID->126314747]
}, Closed]],
Cell[CellGroupData[{
Cell[382805, 10930, 4861, 126, 70, "ProofInfo",
 CellTags->{"Proof|ID:2983958", "Proof|ID:2983958-1"},
 CellID->83920353],
Cell[387669, 11058, 478, 12, 70, "ProofInfoBody",
 CellID->206111400]
}, Closed]],
Cell[CellGroupData[{
Cell[388184, 11075, 5223, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:240564679", "Proof|ID:240564679-1"},
 CellID->163719088],
Cell[393410, 11209, 479, 12, 70, "ProofInfoBody",
 CellID->40115550]
}, Closed]],
Cell[CellGroupData[{
Cell[393926, 11226, 5273, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:299677825", "Proof|ID:299677825-1"},
 CellID->464155749],
Cell[399202, 11360, 480, 12, 70, "ProofInfoBody",
 CellID->805072015]
}, Closed]],
Cell[399697, 11375, 46, 1, 70, "OpenEnvironment",
 CellID->97289005],
Cell[CellGroupData[{
Cell[399768, 11380, 412, 10, 70, "EnvironmentHeader",
 CellID->39640643],
Cell[400183, 11392, 773, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:449289304", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->449289304],
Cell[400959, 11414, 1042, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:240154468", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex total"},
 CellID->240154468],
Cell[402004, 11444, 1311, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:808110452", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex asymmetric"},
 CellID->808110452],
Cell[403318, 11486, 1250, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:224322863", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex transitive"},
 CellID->224322863],
Cell[404571, 11524, 991, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:424804407", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex 1 min"},
 CellID->424804407],
Cell[405565, 11553, 1108, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:308108885", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex \[RawStar]"},
 CellID->308108885],
Cell[406676, 11586, 62, 1, 70, "EndEnvironmentMarker",
 CellID->951139]
}, Open  ]],
Cell[406753, 11590, 48, 1, 70, "CloseEnvironment",
 CellID->130964344],
Cell[CellGroupData[{
Cell[406826, 11595, 4406, 116, 70, "ProofInfo",
 CellTags->{"Proof|ID:240154468", "Proof|ID:240154468-1"},
 CellID->162210883],
Cell[411235, 11713, 479, 12, 70, "ProofInfoBody",
 CellID->17882270]
}, Closed]],
Cell[CellGroupData[{
Cell[411751, 11730, 4889, 126, 70, "ProofInfo",
 CellTags->{"Proof|ID:808110452", "Proof|ID:808110452-1"},
 CellID->14371058],
Cell[416643, 11858, 480, 12, 70, "ProofInfoBody",
 CellID->223768150]
}, Closed]],
Cell[CellGroupData[{
Cell[417160, 11875, 5247, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:224322863", "Proof|ID:224322863-1"},
 CellID->483386676],
Cell[422410, 12009, 480, 12, 70, "ProofInfoBody",
 CellID->255376339]
}, Closed]],
Cell[CellGroupData[{
Cell[422927, 12026, 3623, 98, 70, "ProofInfo",
 CellTags->{"Proof|ID:424804407", "Proof|ID:424804407-1"},
 CellID->138222520],
Cell[426553, 12126, 480, 12, 70, "ProofInfoBody",
 CellID->421322267]
}, Closed]],
Cell[CellGroupData[{
Cell[427070, 12143, 5291, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:308108885", "Proof|ID:308108885-1"},
 CellID->152180612],
Cell[432364, 12277, 480, 12, 70, "ProofInfoBody",
 CellID->286784836]
}, Closed]],
Cell[432859, 12292, 47, 1, 70, "OpenEnvironment",
 CellID->409591721],
Cell[CellGroupData[{
Cell[432931, 12297, 387, 9, 70, "EnvironmentHeader",
 CellID->101138670],
Cell[433321, 12308, 878, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:936963717", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "DegRevLex isTermOrder"},
 CellID->936963717],
Cell[434202, 12333, 64, 1, 70, "EndEnvironmentMarker",
 CellID->28325649]
}, Open  ]],
Cell[434281, 12337, 48, 1, 70, "CloseEnvironment",
 CellID->269251289],
Cell[CellGroupData[{
Cell[434354, 12342, 2583, 74, 70, "ProofInfo",
 CellTags->{"Proof|ID:936963717", "Proof|ID:936963717-1"},
 CellID->319190197],
Cell[436940, 12418, 480, 12, 70, "ProofInfoBody",
 CellID->379817477]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[437493, 12438, 116, 2, 43, "Section",
 CellID->338638037],
Cell[CellGroupData[{
Cell[437634, 12444, 329, 6, 70, "Subsection",
 CellID->1801223056],
Cell[437966, 12452, 716, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:44994065", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->44994065],
Cell[CellGroupData[{
Cell[438707, 12474, 122, 2, 70, "Subsubsection",
 CellID->494981958],
Cell[438832, 12478, 48, 1, 70, "OpenEnvironment",
 CellID->1947010109],
Cell[CellGroupData[{
Cell[438905, 12483, 585, 14, 70, "EnvironmentHeader",
 CellID->1551875723],
Cell[439493, 12499, 758, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1997108144", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->1997108144],
Cell[440254, 12520, 926, 22, 70, "GlobalDeclaration",
 CellTags->{
  "ID:608617422", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->608617422],
Cell[441183, 12544, 1676, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:271011394", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples \[Element]"},
 CellID->271011394],
Cell[442862, 12595, 953, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1820126795", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples 1"},
 CellID->1820126795],
Cell[443818, 12624, 774, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:281355476", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples dim"},
 CellID->281355476],
Cell[444595, 12647, 1083, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2109519396", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples deg"},
 CellID->2109519396],
Cell[445681, 12680, 901, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1563833422", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples exp"},
 CellID->1563833422],
Cell[446585, 12707, 1052, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1180963317", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples \[RawStar]"},
 CellID->1180963317],
Cell[447640, 12740, 1050, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1392702003", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples /"},
 CellID->1392702003],
Cell[448693, 12773, 1292, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1587875066", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples \[VerticalBar]"},
 CellID->1587875066],
Cell[449988, 12815, 1418, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:803590963", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples lcm"},
 CellID->803590963],
Cell[451409, 12855, 305, 6, 70, "EnvironmentText",
 CellID->425740884],
Cell[451717, 12863, 1127, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:106372713", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples <<"},
 CellID->106372713],
Cell[452847, 12896, 1037, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1589980318", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples >>"},
 CellID->1589980318],
Cell[453887, 12928, 65, 1, 70, "EndEnvironmentMarker",
 CellID->653212082]
}, Open  ]],
Cell[453967, 12932, 48, 1, 70, "CloseEnvironment",
 CellID->829024171]
}, Closed]],
Cell[CellGroupData[{
Cell[454052, 12938, 215, 5, 70, "Subsubsection",
 CellID->89472363],
Cell[454270, 12945, 712, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:61015536", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->61015536],
Cell[454985, 12965, 46, 1, 70, "OpenEnvironment",
 CellID->36911472],
Cell[CellGroupData[{
Cell[455056, 12970, 412, 10, 70, "EnvironmentHeader",
 CellID->85270262],
Cell[455471, 12982, 1647, 50, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:617800164", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PPTuples alternative", "IR"},
 CellID->617800164],
Cell[457121, 13034, 787, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:41499157", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples isDomain"},
 CellID->41499157],
Cell[457911, 13057, 830, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:142297858", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasDecisionPredicate"},
 CellID->142297858],
Cell[458744, 13080, 823, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:201916088", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasExponentFunction"},
 CellID->201916088],
Cell[459570, 13103, 807, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:102507049", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples isExtensional"},
 CellID->102507049],
Cell[460380, 13126, 818, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:93844102", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasDegreeFunction"},
 CellID->93844102],
Cell[461201, 13149, 860, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:34803991", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples hasDivisibility"},
 CellID->34803991],
Cell[462064, 13173, 65, 1, 70, "EndEnvironmentMarker",
 CellID->241614839]
}, Open  ]],
Cell[462144, 13177, 48, 1, 70, "CloseEnvironment",
 CellID->372933962],
Cell[CellGroupData[{
Cell[462217, 13182, 4073, 108, 70, "ProofInfo",
 CellTags->{"Proof|ID:617800164", "Proof|ID:617800164-1"},
 CellID->427612524],
Cell[466293, 13292, 480, 12, 70, "ProofInfoBody",
 CellID->712068464]
}, Closed]],
Cell[CellGroupData[{
Cell[466810, 13309, 2372, 70, 70, "ProofInfo",
 CellTags->{"Proof|ID:41499157", "Proof|ID:41499157-1"},
 CellID->394527339],
Cell[469185, 13381, 479, 12, 70, "ProofInfoBody",
 CellID->260522834]
}, Closed]],
Cell[CellGroupData[{
Cell[469701, 13398, 2429, 70, 70, "ProofInfo",
 CellTags->{"Proof|ID:142297858", "Proof|ID:142297858-1"},
 CellID->706958866],
Cell[472133, 13470, 480, 12, 70, "ProofInfoBody",
 CellID->471518847]
}, Closed]],
Cell[CellGroupData[{
Cell[472650, 13487, 2433, 71, 70, "ProofInfo",
 CellTags->{"Proof|ID:201916088", "Proof|ID:201916088-1"},
 CellID->220074858],
Cell[475086, 13560, 480, 12, 70, "ProofInfoBody",
 CellID->162708875]
}, Closed]],
Cell[CellGroupData[{
Cell[475603, 13577, 2401, 70, 70, "ProofInfo",
 CellTags->{"Proof|ID:102507049", "Proof|ID:102507049-1"},
 CellID->628960870],
Cell[478007, 13649, 480, 12, 70, "ProofInfoBody",
 CellID->439772086]
}, Closed]],
Cell[CellGroupData[{
Cell[478524, 13666, 2411, 70, 70, "ProofInfo",
 CellTags->{"Proof|ID:93844102", "Proof|ID:93844102-1"},
 CellID->78679481],
Cell[480938, 13738, 479, 12, 70, "ProofInfoBody",
 CellID->285701216]
}, Closed]],
Cell[CellGroupData[{
Cell[481454, 13755, 2403, 70, 70, "ProofInfo",
 CellTags->{"Proof|ID:34803991", "Proof|ID:34803991-1"},
 CellID->42217056],
Cell[483860, 13827, 479, 12, 70, "ProofInfoBody",
 CellID->294930096]
}, Closed]],
Cell[484354, 13842, 46, 1, 70, "OpenEnvironment",
 CellID->79399544],
Cell[CellGroupData[{
Cell[484425, 13847, 462, 10, 70, "EnvironmentHeader",
 CellID->12203207],
Cell[484890, 13859, 807, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:338277855", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PPTuples isExpPPDomain"},
 CellID->338277855],
Cell[485700, 13882, 65, 1, 70, "EndEnvironmentMarker",
 CellID->492675449]
}, Open  ]],
Cell[485780, 13886, 47, 1, 70, "CloseEnvironment",
 CellID->33968442],
Cell[CellGroupData[{
Cell[485852, 13891, 2397, 70, 70, "ProofInfo",
 CellTags->{"Proof|ID:338277855", "Proof|ID:338277855-1"},
 CellID->630181205],
Cell[488252, 13963, 480, 12, 70, "ProofInfoBody",
 CellID->497464270]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[488793, 13982, 201, 3, 70, "Subsection",
 CellID->1620536900],
Cell[CellGroupData[{
Cell[489019, 13989, 132, 2, 70, "Subsubsection",
 CellID->1187960384],
Cell[489154, 13993, 1093, 27, 70, "Text",
 CellID->1936291687],
Cell[490250, 14022, 268, 7, 70, "Text",
 CellID->1116953765],
Cell[CellGroupData[{
Cell[490543, 14033, 246, 6, 70, "Item",
 CellID->1456273069],
Cell[490792, 14041, 911, 31, 70, "Item",
 CellID->1501844232],
Cell[491706, 14074, 959, 29, 70, "Item",
 CellID->1006360754],
Cell[492668, 14105, 475, 14, 70, "Item",
 CellID->2014536577]
}, Open  ]],
Cell[493158, 14122, 232, 7, 70, "Text",
 CellID->1075439607],
Cell[CellGroupData[{
Cell[493415, 14133, 200, 5, 70, "Item",
 CellID->643750130],
Cell[493618, 14140, 237, 5, 70, "Item",
 CellID->1845811392],
Cell[493858, 14147, 192, 3, 70, "Item",
 CellID->1608954637],
Cell[494053, 14152, 231, 5, 70, "Item",
 CellID->1643257907]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[494333, 14163, 148, 3, 70, "Subsubsection",
 CellID->1331228005],
Cell[494484, 14168, 48, 1, 70, "OpenEnvironment",
 CellID->1371979610],
Cell[CellGroupData[{
Cell[494557, 14173, 515, 13, 70, "EnvironmentHeader",
 CellID->1149637019],
Cell[495075, 14188, 873, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:611122547", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->611122547],
Cell[495951, 14211, 1008, 23, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1216459107", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->1216459107],
Cell[496962, 14236, 1327, 29, 70, "GlobalDeclaration",
 CellTags->{
  "ID:594519699", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->594519699],
Cell[498292, 14267, 3917, 107, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:65019775", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Element]"},
 CellID->65019775],
Cell[502212, 14376, 938, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:399496924", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C base"},
 CellID->399496924],
Cell[503153, 14404, 1544, 45, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:983317445", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C rec"},
 CellID->983317445],
Cell[504700, 14451, 859, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1360683691", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples 0"},
 CellID->1360683691],
Cell[505562, 14478, 1594, 40, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:13289303", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples 1"},
 CellID->13289303],
Cell[507159, 14520, 982, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1777247200", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + 1"},
 CellID->1777247200],
Cell[508144, 14548, 1399, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2122267989", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + 2"},
 CellID->2122267989],
Cell[509546, 14586, 4077, 109, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:290578418", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + 3"},
 CellID->290578418],
Cell[513626, 14697, 974, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1495569651", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - unary 1"},
 CellID->1495569651],
Cell[514603, 14726, 1529, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1285234307", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - unary 2"},
 CellID->1285234307],
Cell[516135, 14772, 915, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:551740538", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples -"},
 CellID->551740538],
Cell[517053, 14801, 941, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:419928300", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[RawStar] 1"},
 CellID->419928300],
Cell[517997, 14830, 1168, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:209271439", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[RawStar] 2"},
 CellID->209271439],
Cell[519168, 14865, 3884, 99, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1411844363", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[RawStar] 3"},
 CellID->1411844363],
Cell[523055, 14966, 1192, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1128062400", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Succeeds] 1"},
 CellID->1128062400],
Cell[524250, 15004, 1420, 41, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1280515048", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Succeeds] 2"},
 CellID->1280515048],
Cell[525673, 15047, 2480, 72, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2095908261", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[Succeeds] 3"},
 CellID->2095908261],
Cell[528156, 15121, 977, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1668822406", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples multN 1"},
 CellID->1668822406],
Cell[529136, 15149, 1270, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:329725901", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples multN 2"},
 CellID->329725901],
Cell[530409, 15184, 1587, 46, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:822409829", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples mult"},
 CellID->822409829],
Cell[531999, 15232, 1828, 51, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:922404041", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples mult i"},
 CellID->922404041],
Cell[533830, 15285, 922, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2033419264", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples rdm 1"},
 CellID->2033419264],
Cell[534755, 15313, 1149, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1308197671", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples rdm 2"},
 CellID->1308197671],
Cell[535907, 15346, 3023, 78, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1157498137", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples rdm 3"},
 CellID->1157498137],
Cell[538933, 15426, 984, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:103388547", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd 1"},
 CellID->103388547],
Cell[539920, 15456, 1212, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:166800573", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd 2"},
 CellID->166800573],
Cell[541135, 15492, 4959, 122, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:319850711", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd 3"},
 CellID->319850711],
Cell[546097, 15616, 942, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:112650434", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd unary 1"},
 CellID->112650434],
Cell[547042, 15644, 1695, 45, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1729493745", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd unary 2"},
 CellID->1729493745],
Cell[548740, 15691, 1071, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:414290689", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples cpm 1"},
 CellID->414290689],
Cell[549814, 15723, 1296, 36, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:27385995", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples cpm 2"},
 CellID->27385995],
Cell[551113, 15761, 4968, 126, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1444186163", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples cpm 3"},
 CellID->1444186163],
Cell[556084, 15889, 1841, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2018930893", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples <<"},
 CellID->2018930893],
Cell[557928, 15940, 1503, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1442847663", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples >>"},
 CellID->1442847663],
Cell[559434, 15980, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1513890663]
}, Open  ]],
Cell[559515, 15984, 49, 1, 70, "CloseEnvironment",
 CellID->1465096073]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[559625, 15992, 111, 2, 43, "Section",
 CellID->1812227459],
Cell[CellGroupData[{
Cell[559761, 15998, 123, 2, 70, "Subsection",
 CellID->1790574279],
Cell[559887, 16002, 48, 1, 70, "OpenEnvironment",
 CellID->2085486573],
Cell[CellGroupData[{
Cell[559960, 16007, 569, 14, 70, "EnvironmentHeader",
 CellID->748671929],
Cell[560532, 16023, 1083, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1486930793", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalT]2"},
 CellID->1486930793],
Cell[561618, 16053, 1080, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:933000257", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalT]3"},
 CellID->933000257],
Cell[562701, 16083, 66, 1, 70, "EndEnvironmentMarker",
 CellID->1614174784]
}, Open  ]],
Cell[562782, 16087, 49, 1, 70, "CloseEnvironment",
 CellID->1930771981],
Cell[CellGroupData[{
Cell[562856, 16092, 406, 10, 70, "Computation",
 CellID->300068762],
Cell[563265, 16104, 406, 7, 70, "ComputationResult",
 CellID->42587073]
}, Open  ]],
Cell[CellGroupData[{
Cell[563708, 16116, 375, 10, 70, "ComputationInfo",
 CellID->30747916],
Cell[564086, 16128, 650, 17, 70, "ComputationInfoBody",
 CellID->96247615]
}, Closed]],
Cell[CellGroupData[{
Cell[564773, 16150, 288, 8, 70, "Computation",
 CellID->1717056269],
Cell[565064, 16160, 165, 3, 70, "ComputationResult",
 CellID->258122591]
}, Open  ]],
Cell[CellGroupData[{
Cell[565266, 16168, 378, 10, 70, "ComputationInfo",
 CellID->92227752],
Cell[565647, 16180, 649, 17, 70, "ComputationInfoBody",
 CellID->642042499]
}, Closed]],
Cell[CellGroupData[{
Cell[566333, 16202, 363, 8, 70, "Computation",
 CellID->1221343315],
Cell[566699, 16212, 195, 4, 70, "ComputationResult",
 CellID->369293207]
}, Open  ]],
Cell[CellGroupData[{
Cell[566931, 16221, 377, 10, 70, "ComputationInfo",
 CellID->623331724],
Cell[567311, 16233, 652, 17, 70, "ComputationInfoBody",
 CellID->330571709]
}, Closed]],
Cell[CellGroupData[{
Cell[568000, 16255, 435, 10, 70, "Computation",
 CellID->361486860],
Cell[568438, 16267, 331, 6, 70, "ComputationResult",
 CellID->324244333]
}, Open  ]],
Cell[CellGroupData[{
Cell[568806, 16278, 377, 10, 70, "ComputationInfo",
 CellID->69411863],
Cell[569186, 16290, 651, 17, 70, "ComputationInfoBody",
 CellID->556797800]
}, Closed]],
Cell[CellGroupData[{
Cell[569874, 16312, 361, 8, 70, "Computation",
 CellID->1508699062],
Cell[570238, 16322, 163, 3, 70, "ComputationResult",
 CellID->19039667]
}, Open  ]],
Cell[CellGroupData[{
Cell[570438, 16330, 376, 10, 70, "ComputationInfo",
 CellID->19048599],
Cell[570817, 16342, 652, 17, 70, "ComputationInfoBody",
 CellID->584992054]
}, Closed]],
Cell[CellGroupData[{
Cell[571506, 16364, 626, 13, 70, "Computation",
 CellID->286294197],
Cell[572135, 16379, 374, 6, 70, "ComputationResult",
 CellID->71897093]
}, Open  ]],
Cell[CellGroupData[{
Cell[572546, 16390, 378, 10, 70, "ComputationInfo",
 CellID->125781715],
Cell[572927, 16402, 651, 17, 70, "ComputationInfoBody",
 CellID->541156049]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[573627, 16425, 120, 2, 70, "Subsection",
 CellID->1580386408],
Cell[573750, 16429, 47, 1, 70, "OpenEnvironment",
 CellID->299725121],
Cell[CellGroupData[{
Cell[573822, 16434, 535, 13, 70, "EnvironmentHeader",
 CellID->769281711],
Cell[574360, 16449, 1030, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:1375788385", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalQ]xyz"},
 CellID->1375788385],
Cell[575393, 16476, 65, 1, 70, "EndEnvironmentMarker",
 CellID->284898274]
}, Open  ]],
Cell[575473, 16480, 48, 1, 70, "CloseEnvironment",
 CellID->446422152],
Cell[CellGroupData[{
Cell[575546, 16485, 637, 20, 70, "Computation",
 CellID->342550773],
Cell[576186, 16507, 1503, 39, 70, "ComputationResult",
 CellID->2126912284]
}, Open  ]],
Cell[CellGroupData[{
Cell[577726, 16551, 379, 10, 70, "ComputationInfo",
 CellID->1970788451],
Cell[578108, 16563, 646, 17, 70, "ComputationInfoBody",
 CellID->353659982]
}, Closed]],
Cell[CellGroupData[{
Cell[578791, 16585, 907, 28, 70, "Computation",
 CellID->1172786959],
Cell[579701, 16615, 1405, 37, 70, "ComputationResult",
 CellID->1664446879]
}, Open  ]],
Cell[CellGroupData[{
Cell[581143, 16657, 379, 10, 70, "ComputationInfo",
 CellID->116485574],
Cell[581525, 16669, 648, 17, 70, "ComputationInfoBody",
 CellID->1171363384]
}, Closed]],
Cell[CellGroupData[{
Cell[582210, 16691, 491, 15, 70, "Computation",
 CellID->1934266891],
Cell[582704, 16708, 691, 19, 70, "ComputationResult",
 CellID->1218180393]
}, Open  ]],
Cell[CellGroupData[{
Cell[583432, 16732, 378, 10, 70, "ComputationInfo",
 CellID->1620691346],
Cell[583813, 16744, 647, 17, 70, "ComputationInfoBody",
 CellID->511072915]
}, Closed]],
Cell[CellGroupData[{
Cell[584497, 16766, 305, 5, 70, "Computation",
 CellID->174499053],
Cell[584805, 16773, 665, 14, 70, "ComputationResult",
 CellID->622829772]
}, Open  ]],
Cell[CellGroupData[{
Cell[585507, 16792, 377, 10, 70, "ComputationInfo",
 CellID->1258356781],
Cell[585887, 16804, 646, 17, 70, "ComputationInfoBody",
 CellID->835786078]
}, Closed]],
Cell[CellGroupData[{
Cell[586570, 16826, 210, 4, 70, "Computation",
 CellID->715169566],
Cell[586783, 16832, 703, 15, 70, "ComputationResult",
 CellID->886905339]
}, Open  ]],
Cell[CellGroupData[{
Cell[587523, 16852, 379, 10, 70, "ComputationInfo",
 CellID->1108652826],
Cell[587905, 16864, 647, 17, 70, "ComputationInfoBody",
 CellID->1600102270]
}, Closed]],
Cell[CellGroupData[{
Cell[588589, 16886, 335, 9, 70, "Computation",
 CellID->853820995],
Cell[588927, 16897, 799, 24, 70, "ComputationResult",
 CellID->676131250]
}, Open  ]],
Cell[CellGroupData[{
Cell[589763, 16926, 377, 10, 70, "ComputationInfo",
 CellID->1821544516],
Cell[590143, 16938, 647, 17, 70, "ComputationInfoBody",
 CellID->1704096014]
}, Closed]],
Cell[CellGroupData[{
Cell[590827, 16960, 115, 2, 70, "Subsubsection",
 CellID->1216185475],
Cell[CellGroupData[{
Cell[590967, 16966, 362, 9, 70, "Computation",
 CellID->1933070300],
Cell[591332, 16977, 191, 4, 70, "ComputationResult",
 CellID->1659834227]
}, Open  ]],
Cell[CellGroupData[{
Cell[591560, 16986, 377, 10, 70, "ComputationInfo",
 CellID->445837838],
Cell[591940, 16998, 648, 17, 70, "ComputationInfoBody",
 CellID->1713495286]
}, Closed]],
Cell[CellGroupData[{
Cell[592625, 17020, 340, 9, 70, "Computation",
 CellID->1181004978],
Cell[592968, 17031, 338, 10, 70, "ComputationResult",
 CellID->611302185]
}, Open  ]],
Cell[CellGroupData[{
Cell[593343, 17046, 377, 10, 70, "ComputationInfo",
 CellID->442383335],
Cell[593723, 17058, 648, 17, 70, "ComputationInfoBody",
 CellID->1889001502]
}, Closed]],
Cell[CellGroupData[{
Cell[594408, 17080, 360, 9, 70, "Computation",
 CellID->829389969],
Cell[594771, 17091, 187, 4, 70, "ComputationResult",
 CellID->1237458846]
}, Open  ]],
Cell[CellGroupData[{
Cell[594995, 17100, 375, 10, 70, "ComputationInfo",
 CellID->301477109],
Cell[595373, 17112, 647, 17, 70, "ComputationInfoBody",
 CellID->2087589975]
}, Closed]],
Cell[CellGroupData[{
Cell[596057, 17134, 361, 9, 70, "Computation",
 CellID->806902398],
Cell[596421, 17145, 188, 4, 70, "ComputationResult",
 CellID->1613258704]
}, Open  ]],
Cell[CellGroupData[{
Cell[596646, 17154, 377, 10, 70, "ComputationInfo",
 CellID->1603337453],
Cell[597026, 17166, 648, 17, 70, "ComputationInfoBody",
 CellID->236846226]
}, Closed]],
Cell[CellGroupData[{
Cell[597711, 17188, 339, 9, 70, "Computation",
 CellID->1918925811],
Cell[598053, 17199, 263, 8, 70, "ComputationResult",
 CellID->993729665]
}, Open  ]],
Cell[CellGroupData[{
Cell[598353, 17212, 377, 10, 70, "ComputationInfo",
 CellID->938624865],
Cell[598733, 17224, 650, 17, 70, "ComputationInfoBody",
 CellID->1603628063]
}, Closed]],
Cell[CellGroupData[{
Cell[599420, 17246, 365, 9, 70, "Computation",
 CellID->1741569015],
Cell[599788, 17257, 191, 4, 70, "ComputationResult",
 CellID->1732022027]
}, Open  ]],
Cell[CellGroupData[{
Cell[600016, 17266, 380, 10, 70, "ComputationInfo",
 CellID->1921971350],
Cell[600399, 17278, 648, 17, 70, "ComputationInfoBody",
 CellID->1105863375]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[601096, 17301, 117, 2, 70, "Subsubsection",
 CellID->190084383],
Cell[CellGroupData[{
Cell[601238, 17307, 475, 12, 70, "Computation",
 CellID->1777111242],
Cell[601716, 17321, 496, 12, 70, "ComputationResult",
 CellID->266105935]
}, Open  ]],
Cell[CellGroupData[{
Cell[602249, 17338, 378, 10, 70, "ComputationInfo",
 CellID->2144087656],
Cell[602630, 17350, 648, 17, 70, "ComputationInfoBody",
 CellID->1031259517]
}, Closed]],
Cell[CellGroupData[{
Cell[603315, 17372, 402, 11, 70, "Computation",
 CellID->999756278],
Cell[603720, 17385, 314, 9, 70, "ComputationResult",
 CellID->333730610]
}, Open  ]],
Cell[CellGroupData[{
Cell[604071, 17399, 376, 10, 70, "ComputationInfo",
 CellID->262548371],
Cell[604450, 17411, 648, 17, 70, "ComputationInfoBody",
 CellID->839150767]
}, Closed]],
Cell[CellGroupData[{
Cell[605135, 17433, 429, 11, 70, "Computation",
 CellID->1213404407],
Cell[605567, 17446, 328, 10, 70, "ComputationResult",
 CellID->240280514]
}, Open  ]],
Cell[CellGroupData[{
Cell[605932, 17461, 379, 10, 70, "ComputationInfo",
 CellID->838793558],
Cell[606314, 17473, 650, 17, 70, "ComputationInfoBody",
 CellID->1634384220]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[607013, 17496, 114, 2, 70, "Subsubsection",
 CellID->800084030],
Cell[CellGroupData[{
Cell[607152, 17502, 711, 17, 70, "Computation",
 CellID->1088310899],
Cell[607866, 17521, 426, 13, 70, "ComputationResult",
 CellID->1642074452]
}, Open  ]],
Cell[CellGroupData[{
Cell[608329, 17539, 378, 10, 70, "ComputationInfo",
 CellID->2048206854],
Cell[608710, 17551, 648, 17, 70, "ComputationInfoBody",
 CellID->2139465790]
}, Closed]],
Cell[CellGroupData[{
Cell[609395, 17573, 711, 17, 70, "Computation",
 CellID->1164918771],
Cell[610109, 17592, 347, 10, 70, "ComputationResult",
 CellID->1954185803]
}, Open  ]],
Cell[CellGroupData[{
Cell[610493, 17607, 378, 10, 70, "ComputationInfo",
 CellID->1294693027],
Cell[610874, 17619, 647, 17, 70, "ComputationInfoBody",
 CellID->475108056]
}, Closed]],
Cell[CellGroupData[{
Cell[611558, 17641, 708, 17, 70, "Computation",
 CellID->697197138],
Cell[612269, 17660, 425, 13, 70, "ComputationResult",
 CellID->126279016]
}, Open  ]],
Cell[CellGroupData[{
Cell[612731, 17678, 377, 10, 70, "ComputationInfo",
 CellID->1368887684],
Cell[613111, 17690, 647, 17, 70, "ComputationInfoBody",
 CellID->1052908597]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[613831, 17715, 246, 5, 43, "Section",
 CellID->444952074],
Cell[614080, 17722, 787, 16, 43, "Text",
 CellID->707250474],
Cell[614870, 17740, 777, 20, 32, "GlobalDeclaration",
 CellTags->{
  "ID:395821466", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->395821466],
Cell[CellGroupData[{
Cell[615672, 17764, 128, 2, 38, "Subsection",
 CellID->62492604],
Cell[CellGroupData[{
Cell[615825, 17770, 253, 4, 70, "Subsubsection",
 CellID->15574753],
Cell[616081, 17776, 46, 1, 70, "OpenEnvironment",
 CellID->88808795],
Cell[CellGroupData[{
Cell[616152, 17781, 427, 11, 70, "EnvironmentHeader",
 CellID->318992154],
Cell[616582, 17794, 923, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:689486298", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isTuple", "IR"},
 CellID->689486298],
Cell[617508, 17821, 1507, 48, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:302371753", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples DomainTuples"},
 CellID->302371753],
Cell[619018, 17871, 912, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:619631752", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples empty", "IR"},
 CellID->619631752],
Cell[619933, 17897, 3154, 93, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:93154383", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples rec", "IR"},
 CellID->93154383],
Cell[623090, 17992, 3179, 91, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:110840922", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples rec II", "IR"},
 CellID->110840922],
Cell[626272, 18085, 3467, 100, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:45787564", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples rec III", "IR"},
 CellID->45787564],
Cell[629742, 18187, 707, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:34480440", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->34480440],
Cell[630452, 18208, 2714, 80, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:134755829", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples prepend", "IR"},
 CellID->134755829],
Cell[633169, 18290, 1120, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:218126712", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership first", "IR"},
 CellID->218126712],
Cell[634292, 18323, 1117, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:36095185", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership second", "IR"},
 CellID->36095185],
Cell[635412, 18356, 3357, 89, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:266402037", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[Element] PolyTuples shape"},
 CellID->266402037],
Cell[638772, 18447, 1900, 60, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:314594392", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples length=1"},
 CellID->314594392],
Cell[640675, 18509, 64, 1, 70, "EndEnvironmentMarker",
 CellID->91535020]
}, Open  ]],
Cell[640754, 18513, 48, 1, 70, "CloseEnvironment",
 CellID->558368068],
Cell[CellGroupData[{
Cell[640827, 18518, 2537, 74, 70, "ProofInfo",
 CellTags->{"Proof|ID:689486298", "Proof|ID:689486298-1"},
 CellID->851460394],
Cell[643367, 18594, 479, 12, 70, "ProofInfoBody",
 CellID->16049767]
}, Closed]],
Cell[CellGroupData[{
Cell[643883, 18611, 3187, 89, 70, "ProofInfo",
 CellTags->{"Proof|ID:302371753", "Proof|ID:302371753-1"},
 CellID->426132571],
Cell[647073, 18702, 480, 12, 70, "ProofInfoBody",
 CellID->241646450]
}, Closed]],
Cell[CellGroupData[{
Cell[647590, 18719, 2171, 64, 70, "ProofInfo",
 CellTags->{"Proof|ID:619631752", "Proof|ID:619631752-1"},
 CellID->236569764],
Cell[649764, 18785, 480, 12, 70, "ProofInfoBody",
 CellID->229281091]
}, Closed]],
Cell[CellGroupData[{
Cell[650281, 18802, 6811, 170, 70, "ProofInfo",
 CellTags->{"Proof|ID:93154383", "Proof|ID:93154383-1"},
 CellID->290123111],
Cell[657095, 18974, 479, 12, 70, "ProofInfoBody",
 CellID->117687721]
}, Closed]],
Cell[CellGroupData[{
Cell[657611, 18991, 7380, 177, 70, "ProofInfo",
 CellTags->{"Proof|ID:110840922", "Proof|ID:110840922-1"},
 CellID->99121591],
Cell[664994, 19170, 480, 12, 70, "ProofInfoBody",
 CellID->642964469]
}, Closed]],
Cell[CellGroupData[{
Cell[665511, 19187, 7690, 184, 70, "ProofInfo",
 CellTags->{"Proof|ID:45787564", "Proof|ID:45787564-1"},
 CellID->483935740],
Cell[673204, 19373, 478, 12, 70, "ProofInfoBody",
 CellID->13219130]
}, Closed]],
Cell[CellGroupData[{
Cell[673719, 19390, 6085, 148, 70, "ProofInfo",
 CellTags->{"Proof|ID:134755829", "Proof|ID:134755829-1"},
 CellID->955362936],
Cell[679807, 19540, 479, 12, 70, "ProofInfoBody",
 CellID->11634208]
}, Closed]],
Cell[CellGroupData[{
Cell[680323, 19557, 3492, 96, 70, "ProofInfo",
 CellTags->{"Proof|ID:218126712", "Proof|ID:218126712-1"},
 CellID->69811972],
Cell[683818, 19655, 480, 12, 70, "ProofInfoBody",
 CellID->436964383]
}, Closed]],
Cell[CellGroupData[{
Cell[684335, 19672, 3497, 97, 70, "ProofInfo",
 CellTags->{"Proof|ID:36095185", "Proof|ID:36095185-1"},
 CellID->66892996],
Cell[687835, 19771, 479, 12, 70, "ProofInfoBody",
 CellID->319554292]
}, Closed]],
Cell[CellGroupData[{
Cell[688351, 19788, 7807, 186, 70, "ProofInfo",
 CellTags->{"Proof|ID:266402037", "Proof|ID:266402037-1"},
 CellID->167245154],
Cell[696161, 19976, 478, 12, 70, "ProofInfoBody",
 CellID->5730164]
}, Closed]],
Cell[CellGroupData[{
Cell[696676, 19993, 4921, 125, 70, "ProofInfo",
 CellTags->{"Proof|ID:314594392", "Proof|ID:314594392-1"},
 CellID->424017657],
Cell[701600, 20120, 480, 12, 70, "ProofInfoBody",
 CellID->350735652]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[702129, 20138, 242, 4, 70, "Subsubsection",
 CellID->48602793],
Cell[702374, 20144, 46, 1, 70, "OpenEnvironment",
 CellID->66481260],
Cell[CellGroupData[{
Cell[702445, 20149, 442, 10, 70, "EnvironmentHeader",
 CellID->56424937],
Cell[702890, 20161, 642, 16, 70, "GlobalDeclaration",
 CellTags->{
  "ID:441087600", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->441087600],
Cell[703535, 20179, 1924, 61, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:678316554", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[ForAll] alternative"},
 CellID->678316554],
Cell[705462, 20242, 3708, 107, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:329976481", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples induction", "IR"},
 CellID->329976481],
Cell[709173, 20351, 596, 17, 70, "EnvironmentText",
 CellID->120515761],
Cell[709772, 20370, 4150, 117, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:379661071", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples induction'", "IR"},
 CellID->379661071],
Cell[713925, 20489, 65, 1, 70, "EndEnvironmentMarker",
 CellID->122025085]
}, Open  ]],
Cell[714005, 20493, 47, 1, 70, "CloseEnvironment",
 CellID->70749121],
Cell[CellGroupData[{
Cell[714077, 20498, 5067, 131, 70, "ProofInfo",
 CellTags->{"Proof|ID:678316554", "Proof|ID:678316554-1"},
 CellID->716699503],
Cell[719147, 20631, 480, 12, 70, "ProofInfoBody",
 CellID->496685325]
}, Closed]],
Cell[CellGroupData[{
Cell[719664, 20648, 9347, 218, 70, "ProofInfo",
 CellTags->{"Proof|ID:329976481", "Proof|ID:329976481-1"},
 CellID->3050184],
Cell[729014, 20868, 479, 12, 70, "ProofInfoBody",
 CellID->87959892]
}, Closed]],
Cell[CellGroupData[{
Cell[729530, 20885, 10825, 248, 70, "ProofInfo",
 CellTags->{"Proof|ID:379661071", "Proof|ID:379661071-1"},
 CellID->321989072],
Cell[740358, 21135, 480, 12, 70, "ProofInfoBody",
 CellID->625844841]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[740887, 21153, 277, 6, 70, "Subsubsection",
 CellID->29457529],
Cell[741167, 21161, 45, 1, 70, "OpenEnvironment",
 CellID->1102219],
Cell[CellGroupData[{
Cell[741237, 21166, 530, 13, 70, "EnvironmentHeader",
 CellID->45823306],
Cell[741770, 21181, 1079, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:485298517", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "isTotalIrreflOrder \[Succeeds] \[ScriptCapitalT]"},
 CellID->485298517],
Cell[742852, 21215, 734, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:561716382", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->561716382],
Cell[743589, 21236, 1743, 50, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:107805279", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp decreasing ind"},
 CellID->107805279],
Cell[745335, 21288, 1318, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:220534381", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp decreasing"},
 CellID->220534381],
Cell[746656, 21329, 1310, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:5709061", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max"},
 CellID->5709061],
Cell[747969, 21370, 1292, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:133904640", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max'"},
 CellID->133904640],
Cell[749264, 21410, 1277, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:12736634", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp distinct"},
 CellID->12736634],
Cell[750544, 21450, 64, 1, 70, "EndEnvironmentMarker",
 CellID->99753670]
}, Open  ]],
Cell[750623, 21454, 48, 1, 70, "CloseEnvironment",
 CellID->145531536],
Cell[CellGroupData[{
Cell[750696, 21459, 2046, 62, 70, "ProofInfo",
 CellTags->{"Proof|ID:485298517", "Proof|ID:485298517-1"},
 CellID->638746182],
Cell[752745, 21523, 479, 12, 70, "ProofInfoBody",
 CellID->34385987]
}, Closed]],
Cell[CellGroupData[{
Cell[753261, 21540, 5315, 137, 70, "ProofInfo",
 CellTags->{"Proof|ID:107805279", "Proof|ID:107805279-1"},
 CellID->218458895],
Cell[758579, 21679, 480, 12, 70, "ProofInfoBody",
 CellID->263931670]
}, Closed]],
Cell[CellGroupData[{
Cell[759096, 21696, 4773, 126, 70, "ProofInfo",
 CellTags->{"Proof|ID:220534381", "Proof|ID:220534381-1"},
 CellID->457923117],
Cell[763872, 21824, 480, 12, 70, "ProofInfoBody",
 CellID->110795177]
}, Closed]],
Cell[CellGroupData[{
Cell[764389, 21841, 4005, 110, 70, "ProofInfo",
 CellTags->{"Proof|ID:5709061", "Proof|ID:5709061-1"},
 CellID->199136965],
Cell[768397, 21953, 477, 12, 70, "ProofInfoBody",
 CellID->55914907]
}, Closed]],
Cell[CellGroupData[{
Cell[768911, 21970, 4060, 111, 70, "ProofInfo",
 CellTags->{"Proof|ID:133904640", "Proof|ID:133904640-1"},
 CellID->185035893],
Cell[772974, 22083, 480, 12, 70, "ProofInfoBody",
 CellID->169582083]
}, Closed]],
Cell[CellGroupData[{
Cell[773491, 22100, 4682, 124, 70, "ProofInfo",
 CellTags->{"Proof|ID:12736634", "Proof|ID:12736634-1"},
 CellID->225795913],
Cell[778176, 22226, 477, 12, 70, "ProofInfoBody",
 CellID->1112935]
}, Closed]],
Cell[778668, 22241, 47, 1, 70, "OpenEnvironment",
 CellID->297754641],
Cell[CellGroupData[{
Cell[778740, 22246, 426, 11, 70, "EnvironmentHeader",
 CellID->480387543],
Cell[779169, 22259, 700, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:723780914", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->723780914],
Cell[779872, 22279, 936, 25, 70, "GlobalDeclaration",
 CellTags->{
  "ID:284026551", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->284026551],
Cell[780811, 22306, 864, 22, 70, "GlobalDeclaration",
 CellTags->{
  "ID:548327311", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->548327311],
Cell[781678, 22330, 1040, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:387361496", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max seq"},
 CellID->387361496],
Cell[782721, 22360, 1116, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:70991441", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first pp max seq'"},
 CellID->70991441],
Cell[783840, 22394, 948, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:365249767", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples pp distinct seq"},
 CellID->365249767],
Cell[784791, 22422, 64, 1, 70, "EndEnvironmentMarker",
 CellID->16959157]
}, Open  ]],
Cell[784870, 22426, 47, 1, 70, "CloseEnvironment",
 CellID->47716413],
Cell[CellGroupData[{
Cell[784942, 22431, 5601, 142, 70, "ProofInfo",
 CellTags->{"Proof|ID:387361496", "Proof|ID:387361496-1"},
 CellID->597599025],
Cell[790546, 22575, 480, 12, 70, "ProofInfoBody",
 CellID->278306075]
}, Closed]],
Cell[CellGroupData[{
Cell[791063, 22592, 5988, 153, 70, "ProofInfo",
 CellTags->{"Proof|ID:70991441", "Proof|ID:70991441-1"},
 CellID->346547304],
Cell[797054, 22747, 479, 12, 70, "ProofInfoBody",
 CellID->635659261]
}, Closed]],
Cell[CellGroupData[{
Cell[797570, 22764, 5493, 140, 70, "ProofInfo",
 CellTags->{"Proof|ID:365249767", "Proof|ID:365249767-1"},
 CellID->408497668],
Cell[803066, 22906, 480, 12, 70, "ProofInfoBody",
 CellID->459278606]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[803595, 22924, 226, 5, 70, "Subsubsection",
 CellID->92723618],
Cell[803824, 22931, 47, 1, 70, "OpenEnvironment",
 CellID->269082022],
Cell[CellGroupData[{
Cell[803896, 22936, 452, 10, 70, "EnvironmentHeader",
 CellID->90259041],
Cell[804351, 22948, 786, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:16132582", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->16132582],
Cell[805140, 22970, 1294, 39, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:742018993", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C 1"},
 CellID->742018993],
Cell[806437, 23011, 1795, 55, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:281673814", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C \[NotEqual] 0"},
 CellID->281673814],
Cell[808235, 23068, 1827, 56, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:122376381", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C=0"},
 CellID->122376381],
Cell[810065, 23126, 2253, 68, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:256177412", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C=0 seq"},
 CellID->256177412],
Cell[812321, 23196, 1778, 55, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:293354245", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C=0 seq'"},
 CellID->293354245],
Cell[814102, 23253, 1629, 50, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:129332919", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C prepend"},
 CellID->129332919],
Cell[815734, 23305, 65, 1, 70, "EndEnvironmentMarker",
 CellID->208838147]
}, Open  ]],
Cell[815814, 23309, 48, 1, 70, "CloseEnvironment",
 CellID->929166910],
Cell[CellGroupData[{
Cell[815887, 23314, 4301, 118, 70, "ProofInfo",
 CellTags->{"Proof|ID:742018993", "Proof|ID:742018993-1"},
 CellID->161841610],
Cell[820191, 23434, 548, 13, 70, "ProofInfoBody",
 CellID->117326975]
}, Closed]],
Cell[CellGroupData[{
Cell[820776, 23452, 5067, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:281673814", "Proof|ID:281673814-1"},
 CellID->294457098],
Cell[825846, 23586, 480, 12, 70, "ProofInfoBody",
 CellID->162988210]
}, Closed]],
Cell[CellGroupData[{
Cell[826363, 23603, 5063, 132, 70, "ProofInfo",
 CellTags->{"Proof|ID:122376381", "Proof|ID:122376381-1"},
 CellID->6731680],
Cell[831429, 23737, 478, 12, 70, "ProofInfoBody",
 CellID->2400693]
}, Closed]],
Cell[CellGroupData[{
Cell[831944, 23754, 6576, 162, 70, "ProofInfo",
 CellTags->{"Proof|ID:256177412", "Proof|ID:256177412-1"},
 CellID->606101095],
Cell[838523, 23918, 479, 12, 70, "ProofInfoBody",
 CellID->21528983]
}, Closed]],
Cell[CellGroupData[{
Cell[839039, 23935, 4770, 126, 70, "ProofInfo",
 CellTags->{"Proof|ID:293354245", "Proof|ID:293354245-1"},
 CellID->371170725],
Cell[843812, 24063, 479, 12, 70, "ProofInfoBody",
 CellID->14209107]
}, Closed]],
Cell[CellGroupData[{
Cell[844328, 24080, 5446, 136, 70, "ProofInfo",
 CellTags->{"Proof|ID:129332919", "Proof|ID:129332919-1"},
 CellID->88799612],
Cell[849777, 24218, 480, 12, 70, "ProofInfoBody",
 CellID->277564924]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[850306, 24236, 200, 3, 70, "Subsubsection",
 CellID->11122401],
Cell[850509, 24241, 46, 1, 70, "OpenEnvironment",
 CellID->53346088],
Cell[CellGroupData[{
Cell[850580, 24246, 449, 10, 70, "EnvironmentHeader",
 CellID->610570600],
Cell[851032, 24258, 759, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:136746483", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->136746483],
Cell[851794, 24280, 729, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:110691644", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->110691644],
Cell[852526, 24301, 1012, 26, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:371581865", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + ProdDomain"},
 CellID->371581865],
Cell[853541, 24329, 778, 20, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:694735421", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + isTuple"},
 CellID->694735421],
Cell[854322, 24351, 1182, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:442662347", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[NotEqual] 0"},
 CellID->442662347],
Cell[855507, 24387, 2440, 73, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:66376992", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + pp \[SubsetEqual]"},
 CellID->66376992],
Cell[857950, 24462, 65, 1, 70, "EndEnvironmentMarker",
 CellID->580166704]
}, Open  ]],
Cell[858030, 24466, 48, 1, 70, "CloseEnvironment",
 CellID->469006967],
Cell[CellGroupData[{
Cell[858103, 24471, 4049, 108, 70, "ProofInfo",
 CellTags->{"Proof|ID:371581865", "Proof|ID:371581865-1"},
 CellID->414829965],
Cell[862155, 24581, 522, 13, 70, "ProofInfoBody",
 CellID->847751266]
}, Closed]],
Cell[CellGroupData[{
Cell[862714, 24599, 3603, 98, 70, "ProofInfo",
 CellTags->{"Proof|ID:694735421", "Proof|ID:694735421-1"},
 CellID->511476329],
Cell[866320, 24699, 480, 12, 70, "ProofInfoBody",
 CellID->251211459]
}, Closed]],
Cell[CellGroupData[{
Cell[866837, 24716, 5146, 134, 70, "ProofInfo",
 CellTags->{"Proof|ID:442662347", "Proof|ID:442662347-1"},
 CellID->874127013],
Cell[871986, 24852, 480, 12, 70, "ProofInfoBody",
 CellID->341494949]
}, Closed]],
Cell[CellGroupData[{
Cell[872503, 24869, 8558, 211, 70, "ProofInfo",
 CellTags->{"Proof|ID:66376992", "Proof|ID:66376992-1"},
 CellID->490152990],
Cell[881064, 25082, 479, 12, 70, "ProofInfoBody",
 CellID->654010454]
}, Closed]],
Cell[881558, 25097, 47, 1, 70, "OpenEnvironment",
 CellID->144928206],
Cell[CellGroupData[{
Cell[881630, 25102, 398, 9, 70, "EnvironmentHeader",
 CellID->736686],
Cell[882031, 25113, 856, 21, 70, "GlobalDeclaration",
 CellTags->{
  "ID:480657044", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->480657044],
Cell[882890, 25136, 1398, 39, 70, "GlobalDeclaration",
 CellTags->{
  "ID:907803476", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->907803476],
Cell[884291, 25177, 2224, 59, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:442268007", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[Succeeds] 1"},
 CellID->442268007],
Cell[886518, 25238, 2123, 58, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:26802188", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[Succeeds] 2"},
 CellID->26802188],
Cell[888644, 25298, 2216, 66, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:4890345", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + \[Succeeds] 3"},
 CellID->4890345],
Cell[890863, 25366, 64, 1, 70, "EndEnvironmentMarker",
 CellID->29004364]
}, Open  ]],
Cell[890942, 25370, 48, 1, 70, "CloseEnvironment",
 CellID->199903433],
Cell[CellGroupData[{
Cell[891015, 25375, 10386, 242, 70, "ProofInfo",
 CellTags->{"Proof|ID:442268007", "Proof|ID:442268007-1"},
 CellID->23969624],
Cell[901404, 25619, 480, 12, 70, "ProofInfoBody",
 CellID->196804738]
}, Closed]],
Cell[CellGroupData[{
Cell[901921, 25636, 10381, 242, 70, "ProofInfo",
 CellTags->{"Proof|ID:26802188", "Proof|ID:26802188-1"},
 CellID->77790849],
Cell[912305, 25880, 478, 12, 70, "ProofInfoBody",
 CellID->96981217]
}, Closed]],
Cell[CellGroupData[{
Cell[912820, 25897, 10381, 242, 70, "ProofInfo",
 CellTags->{"Proof|ID:4890345", "Proof|ID:4890345-1"},
 CellID->650383762],
Cell[923204, 26141, 478, 12, 70, "ProofInfoBody",
 CellID->139699914]
}, Closed]],
Cell[923697, 26156, 46, 1, 70, "OpenEnvironment",
 CellID->39778867],
Cell[CellGroupData[{
Cell[923768, 26161, 380, 9, 70, "EnvironmentHeader",
 CellID->9504892],
Cell[924151, 26172, 738, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:212784844", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->212784844],
Cell[924892, 26193, 838, 23, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:106856858", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - isTuple"},
 CellID->106856858],
Cell[925733, 26218, 964, 27, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:200895043", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - length"},
 CellID->200895043],
Cell[926700, 26247, 1367, 41, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:19952680", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples - pp"},
 CellID->19952680],
Cell[928070, 26290, 65, 1, 70, "EndEnvironmentMarker",
 CellID->697897580]
}, Open  ]],
Cell[928150, 26294, 46, 1, 70, "CloseEnvironment",
 CellID->8829566],
Cell[CellGroupData[{
Cell[928221, 26299, 2967, 84, 70, "ProofInfo",
 CellTags->{"Proof|ID:106856858", "Proof|ID:106856858-1"},
 CellID->555324734],
Cell[931191, 26385, 478, 12, 70, "ProofInfoBody",
 CellID->2967280]
}, Closed]],
Cell[CellGroupData[{
Cell[931706, 26402, 3402, 94, 70, "ProofInfo",
 CellTags->{"Proof|ID:200895043", "Proof|ID:200895043-1"},
 CellID->593511758],
Cell[935111, 26498, 478, 12, 70, "ProofInfoBody",
 CellID->5689008]
}, Closed]],
Cell[CellGroupData[{
Cell[935626, 26515, 4110, 112, 70, "ProofInfo",
 CellTags->{"Proof|ID:19952680", "Proof|ID:19952680-1"},
 CellID->287758024],
Cell[939739, 26629, 478, 12, 70, "ProofInfoBody",
 CellID->19013154]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[940266, 26647, 219, 5, 70, "Subsubsection",
 CellID->27460276],
Cell[940488, 26654, 45, 1, 70, "OpenEnvironment",
 CellID->7836954],
Cell[CellGroupData[{
Cell[940558, 26659, 384, 9, 70, "EnvironmentHeader",
 CellID->639878451],
Cell[940945, 26670, 1973, 60, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:118174250", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isMonomial I"},
 CellID->118174250],
Cell[942921, 26732, 828, 23, 70, "GlobalDeclaration",
 CellTags->{
  "ID:78163703", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->78163703],
Cell[943752, 26757, 1691, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:748742057", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isMonomial II"},
 CellID->748742057],
Cell[945446, 26803, 1819, 53, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:104425812", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isMonomial III"},
 CellID->104425812],
Cell[947268, 26858, 1344, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:191443971", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CenterDot]"},
 CellID->191443971],
Cell[948615, 26895, 65, 1, 70, "EndEnvironmentMarker",
 CellID->291201157]
}, Open  ]],
Cell[948695, 26899, 48, 1, 70, "CloseEnvironment",
 CellID->439301014],
Cell[CellGroupData[{
Cell[948768, 26904, 5091, 130, 70, "ProofInfo",
 CellTags->{"Proof|ID:118174250", "Proof|ID:118174250-1"},
 CellID->45600071],
Cell[953862, 27036, 479, 12, 70, "ProofInfoBody",
 CellID->33001478]
}, Closed]],
Cell[CellGroupData[{
Cell[954378, 27053, 5067, 126, 70, "ProofInfo",
 CellTags->{"Proof|ID:748742057", "Proof|ID:748742057-1"},
 CellID->104846486],
Cell[959448, 27181, 480, 12, 70, "ProofInfoBody",
 CellID->561651541]
}, Closed]],
Cell[CellGroupData[{
Cell[959965, 27198, 5744, 140, 70, "ProofInfo",
 CellTags->{"Proof|ID:104425812", "Proof|ID:104425812-1"},
 CellID->41476004],
Cell[965712, 27340, 480, 12, 70, "ProofInfoBody",
 CellID->112588225]
}, Closed]],
Cell[CellGroupData[{
Cell[966229, 27357, 4535, 114, 70, "ProofInfo",
 CellTags->{"Proof|ID:191443971", "Proof|ID:191443971-1"},
 CellID->309334880],
Cell[970767, 27473, 480, 12, 70, "ProofInfoBody",
 CellID->255214453]
}, Closed]],
Cell[971262, 27488, 46, 1, 70, "OpenEnvironment",
 CellID->41488312],
Cell[CellGroupData[{
Cell[971333, 27493, 443, 10, 70, "EnvironmentHeader",
 CellID->568608957],
Cell[971779, 27505, 835, 22, 70, "GlobalDeclaration",
 CellTags->{
  "ID:956509664", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->956509664],
Cell[972617, 27529, 944, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:52346492", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples first index", "IR"},
 CellID->52346492],
Cell[973564, 27556, 1433, 44, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:3252425", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isLP"},
 CellID->3252425],
Cell[975000, 27602, 898, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:301012536", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lp"},
 CellID->301012536],
Cell[975901, 27629, 898, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:499482967", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lc"},
 CellID->499482967],
Cell[976802, 27656, 912, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:77776384", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lm"},
 CellID->77776384],
Cell[977717, 27683, 65, 1, 70, "EndEnvironmentMarker",
 CellID->677673986]
}, Open  ]],
Cell[977797, 27687, 47, 1, 70, "CloseEnvironment",
 CellID->96813974],
Cell[CellGroupData[{
Cell[977869, 27692, 3538, 96, 70, "ProofInfo",
 CellTags->{"Proof|ID:52346492", "Proof|ID:52346492-1"},
 CellID->156386449],
Cell[981410, 27790, 479, 12, 70, "ProofInfoBody",
 CellID->286974475]
}, Closed]],
Cell[CellGroupData[{
Cell[981926, 27807, 4535, 120, 70, "ProofInfo",
 CellTags->{"Proof|ID:3252425", "Proof|ID:3252425-1"},
 CellID->201371222],
Cell[986464, 27929, 478, 12, 70, "ProofInfoBody",
 CellID->543101995]
}, Closed]],
Cell[CellGroupData[{
Cell[986979, 27946, 3797, 104, 70, "ProofInfo",
 CellTags->{"Proof|ID:301012536", "Proof|ID:301012536-1"},
 CellID->132883546],
Cell[990779, 28052, 479, 12, 70, "ProofInfoBody",
 CellID->95709161]
}, Closed]],
Cell[CellGroupData[{
Cell[991295, 28069, 3797, 104, 70, "ProofInfo",
 CellTags->{"Proof|ID:499482967", "Proof|ID:499482967-1"},
 CellID->358731136],
Cell[995095, 28175, 480, 12, 70, "ProofInfoBody",
 CellID->606474251]
}, Closed]],
Cell[CellGroupData[{
Cell[995612, 28192, 3887, 106, 70, "ProofInfo",
 CellTags->{"Proof|ID:77776384", "Proof|ID:77776384-1"},
 CellID->379573710],
Cell[999502, 28300, 478, 12, 70, "ProofInfoBody",
 CellID->99872566]
}, Closed]],
Cell[999995, 28315, 46, 1, 70, "OpenEnvironment",
 CellID->95739086],
Cell[CellGroupData[{
Cell[1000066, 28320, 396, 9, 70, "EnvironmentHeader",
 CellID->41342827],
Cell[1000465, 28331, 703, 18, 70, "GlobalDeclaration",
 CellTags->{
  "ID:1015977913", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->1015977913],
Cell[1001171, 28351, 938, 25, 70, "GlobalDeclaration",
 CellTags->{
  "ID:731677923", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->731677923],
Cell[1002112, 28378, 1037, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:23863949", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lp seq"},
 CellID->23863949],
Cell[1003152, 28408, 1040, 28, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:721799663", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lc seq"},
 CellID->721799663],
Cell[1004195, 28438, 1181, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2939774", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lm seq"},
 CellID->2939774],
Cell[1005379, 28471, 1114, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:172576768", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples R seq"},
 CellID->172576768],
Cell[1006496, 28503, 1899, 56, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:3206806", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CenterDot] lp lc"},
 CellID->3206806],
Cell[1008398, 28561, 65, 1, 70, "EndEnvironmentMarker",
 CellID->331618674]
}, Open  ]],
Cell[1008478, 28565, 48, 1, 70, "CloseEnvironment",
 CellID->834387043],
Cell[CellGroupData[{
Cell[1008551, 28570, 5137, 134, 70, "ProofInfo",
 CellTags->{"Proof|ID:23863949", "Proof|ID:23863949-1"},
 CellID->2176190],
Cell[1013691, 28706, 479, 12, 70, "ProofInfoBody",
 CellID->315083762]
}, Closed]],
Cell[CellGroupData[{
Cell[1014207, 28723, 5144, 134, 70, "ProofInfo",
 CellTags->{"Proof|ID:721799663", "Proof|ID:721799663-1"},
 CellID->307273605],
Cell[1019354, 28859, 480, 12, 70, "ProofInfoBody",
 CellID->320776169]
}, Closed]],
Cell[CellGroupData[{
Cell[1019871, 28876, 5653, 144, 70, "ProofInfo",
 CellTags->{"Proof|ID:2939774", "Proof|ID:2939774-1"},
 CellID->75418897],
Cell[1025527, 29022, 478, 12, 70, "ProofInfoBody",
 CellID->142228474]
}, Closed]],
Cell[CellGroupData[{
Cell[1026042, 29039, 5358, 138, 70, "ProofInfo",
 CellTags->{"Proof|ID:172576768", "Proof|ID:172576768-1"},
 CellID->126187940],
Cell[1031403, 29179, 480, 12, 70, "ProofInfoBody",
 CellID->454221170]
}, Closed]],
Cell[CellGroupData[{
Cell[1031920, 29196, 5527, 138, 70, "ProofInfo",
 CellTags->{"Proof|ID:3206806", "Proof|ID:3206806-1"},
 CellID->37996798],
Cell[1037450, 29336, 478, 12, 70, "ProofInfoBody",
 CellID->471222850]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1037989, 29355, 197, 3, 30, "Subsection",
 CellID->335068244],
Cell[CellGroupData[{
Cell[1038211, 29362, 236, 4, 70, "Subsubsection",
 CellID->420471911],
Cell[1038450, 29368, 47, 1, 70, "OpenEnvironment",
 CellID->657209226],
Cell[CellGroupData[{
Cell[1038522, 29373, 382, 9, 70, "EnvironmentHeader",
 CellID->711892999],
Cell[1038907, 29384, 830, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:31351138", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalR] isDomain"},
 CellID->31351138],
Cell[1039740, 29408, 835, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:632462358", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "\[ScriptCapitalT] isDomain"},
 CellID->632462358],
Cell[1040578, 29432, 869, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:404915746", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isDomain"},
 CellID->404915746],
Cell[1041450, 29456, 65, 1, 70, "EndEnvironmentMarker",
 CellID->654702049]
}, Open  ]],
Cell[1041530, 29460, 48, 1, 70, "CloseEnvironment",
 CellID->432439441],
Cell[CellGroupData[{
Cell[1041603, 29465, 1784, 56, 70, "ProofInfo",
 CellTags->{"Proof|ID:31351138", "Proof|ID:31351138-1"},
 CellID->124332373],
Cell[1043390, 29523, 479, 12, 70, "ProofInfoBody",
 CellID->498504441]
}, Closed]],
Cell[CellGroupData[{
Cell[1043906, 29540, 1789, 56, 70, "ProofInfo",
 CellTags->{"Proof|ID:632462358", "Proof|ID:632462358-1"},
 CellID->403093521],
Cell[1045698, 29598, 479, 12, 70, "ProofInfoBody",
 CellID->85073108]
}, Closed]],
Cell[CellGroupData[{
Cell[1046214, 29615, 1941, 58, 70, "ProofInfo",
 CellTags->{"Proof|ID:404915746", "Proof|ID:404915746-1"},
 CellID->277496693],
Cell[1048158, 29675, 479, 12, 70, "ProofInfoBody",
 CellID->80377566]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1048686, 29693, 197, 3, 70, "Subsubsection",
 CellID->110666071],
Cell[1048886, 29698, 47, 1, 70, "OpenEnvironment",
 CellID->498410766],
Cell[CellGroupData[{
Cell[1048958, 29703, 382, 9, 70, "EnvironmentHeader",
 CellID->351279416],
Cell[1049343, 29714, 753, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:3279840", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->3279840],
Cell[1050099, 29736, 1082, 30, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:141435198", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership C"},
 CellID->141435198],
Cell[1051184, 29768, 880, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:45852914", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership 0"},
 CellID->45852914],
Cell[1052067, 29794, 883, 24, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:320592129", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership 1"},
 CellID->320592129],
Cell[1052953, 29820, 907, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:93322392", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership +"},
 CellID->93322392],
Cell[1053863, 29847, 897, 25, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:6751189", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership -"},
 CellID->6751189],
Cell[1054763, 29874, 1252, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:179493910", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples supp"},
 CellID->179493910],
Cell[1056018, 29910, 65, 1, 70, "EndEnvironmentMarker",
 CellID->154048488]
}, Open  ]],
Cell[1056098, 29914, 47, 1, 70, "CloseEnvironment",
 CellID->51765692],
Cell[CellGroupData[{
Cell[1056170, 29919, 3549, 96, 70, "ProofInfo",
 CellTags->{"Proof|ID:141435198", "Proof|ID:141435198-1"},
 CellID->225697908],
Cell[1059722, 30017, 479, 12, 70, "ProofInfoBody",
 CellID->89331975]
}, Closed]],
Cell[CellGroupData[{
Cell[1060238, 30034, 2315, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:45852914", "Proof|ID:45852914-1"},
 CellID->91077423],
Cell[1062556, 30104, 479, 12, 70, "ProofInfoBody",
 CellID->169634120]
}, Closed]],
Cell[CellGroupData[{
Cell[1063072, 30121, 2320, 68, 70, "ProofInfo",
 CellTags->{"Proof|ID:320592129", "Proof|ID:320592129-1"},
 CellID->94174531],
Cell[1065395, 30191, 480, 12, 70, "ProofInfoBody",
 CellID->339585414]
}, Closed]],
Cell[CellGroupData[{
Cell[1065912, 30208, 3928, 106, 70, "ProofInfo",
 CellTags->{"Proof|ID:93322392", "Proof|ID:93322392-1"},
 CellID->411309806],
Cell[1069843, 30316, 479, 12, 70, "ProofInfoBody",
 CellID->207218404]
}, Closed]],
Cell[CellGroupData[{
Cell[1070359, 30333, 3286, 92, 70, "ProofInfo",
 CellTags->{"Proof|ID:6751189", "Proof|ID:6751189-1"},
 CellID->16346637],
Cell[1073648, 30427, 477, 12, 70, "ProofInfoBody",
 CellID->78211768]
}, Closed]],
Cell[CellGroupData[{
Cell[1074162, 30444, 4439, 118, 70, "ProofInfo",
 CellTags->{"Proof|ID:179493910", "Proof|ID:179493910-1"},
 CellID->162772222],
Cell[1078604, 30564, 479, 12, 70, "ProofInfoBody",
 CellID->62815219]
}, Closed]],
Cell[1079098, 30579, 47, 1, 70, "OpenEnvironment",
 CellID->212237021],
Cell[CellGroupData[{
Cell[1079170, 30584, 399, 9, 70, "EnvironmentHeader",
 CellID->439058519],
Cell[1079572, 30595, 830, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:644982519", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasDecisionPredicate"},
 CellID->644982519],
Cell[1080405, 30618, 65, 1, 70, "EndEnvironmentMarker",
 CellID->686721649]
}, Open  ]],
Cell[1080485, 30622, 46, 1, 70, "CloseEnvironment",
 CellID->6665501],
Cell[CellGroupData[{
Cell[1080556, 30627, 1997, 59, 70, "ProofInfo",
 CellTags->{"Proof|ID:644982519", "Proof|ID:644982519-1"},
 CellID->323902770],
Cell[1082556, 30688, 480, 12, 70, "ProofInfoBody",
 CellID->440058230]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1083085, 30706, 192, 3, 70, "Subsubsection",
 CellID->704433491],
Cell[1083280, 30711, 47, 1, 70, "OpenEnvironment",
 CellID->360614755],
Cell[CellGroupData[{
Cell[1083352, 30716, 393, 9, 70, "EnvironmentHeader",
 CellID->332255835],
Cell[1083748, 30727, 819, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:114843091", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples containsMonomials"},
 CellID->114843091],
Cell[1084570, 30750, 64, 1, 70, "EndEnvironmentMarker",
 CellID->80976911]
}, Open  ]],
Cell[1084649, 30754, 44, 1, 70, "CloseEnvironment",
 CellID->35940],
Cell[CellGroupData[{
Cell[1084718, 30759, 1985, 59, 70, "ProofInfo",
 CellTags->{"Proof|ID:114843091", "Proof|ID:114843091-1"},
 CellID->255004292],
Cell[1086706, 30820, 523, 13, 70, "ProofInfoBody",
 CellID->81650813]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1087278, 30839, 196, 3, 70, "Subsubsection",
 CellID->42388596],
Cell[1087477, 30844, 47, 1, 70, "OpenEnvironment",
 CellID->154907406],
Cell[CellGroupData[{
Cell[1087549, 30849, 395, 9, 70, "EnvironmentHeader",
 CellID->358230070],
Cell[1087947, 30860, 759, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:78312379", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->78312379],
Cell[1088709, 30881, 1369, 39, 70, "GlobalDeclaration",
 CellTags->{
  "ID:210280178", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->210280178],
Cell[1090081, 30922, 3948, 109, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:261724624", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 1"},
 CellID->261724624],
Cell[1094032, 31033, 3703, 105, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:115965410", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 2"},
 CellID->115965410],
Cell[1097738, 31140, 3955, 118, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:384214099", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 3"},
 CellID->384214099],
Cell[1101696, 31260, 2710, 81, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:56134460", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples + C 4"},
 CellID->56134460],
Cell[1104409, 31343, 65, 1, 70, "EndEnvironmentMarker",
 CellID->106716937]
}, Open  ]],
Cell[1104489, 31347, 48, 1, 70, "CloseEnvironment",
 CellID->658773040],
Cell[CellGroupData[{
Cell[1104562, 31352, 18180, 409, 70, "ProofInfo",
 CellTags->{"Proof|ID:261724624", "Proof|ID:261724624-1"},
 CellID->305685053],
Cell[1122745, 31763, 480, 12, 70, "ProofInfoBody",
 CellID->219011818]
}, Closed]],
Cell[CellGroupData[{
Cell[1123262, 31780, 18180, 409, 70, "ProofInfo",
 CellTags->{"Proof|ID:115965410", "Proof|ID:115965410-1"},
 CellID->175933698],
Cell[1141445, 32191, 522, 13, 70, "ProofInfoBody",
 CellID->283308081]
}, Closed]],
Cell[CellGroupData[{
Cell[1142004, 32209, 18545, 415, 70, "ProofInfo",
 CellTags->{"Proof|ID:384214099", "Proof|ID:384214099-1"},
 CellID->84967562],
Cell[1160552, 32626, 480, 12, 70, "ProofInfoBody",
 CellID->467820545]
}, Closed]],
Cell[CellGroupData[{
Cell[1161069, 32643, 13729, 313, 70, "ProofInfo",
 CellTags->{"Proof|ID:56134460", "Proof|ID:56134460-1"},
 CellID->170413464],
Cell[1174801, 32958, 479, 12, 70, "ProofInfoBody",
 CellID->121669868]
}, Closed]],
Cell[1175295, 32973, 46, 1, 70, "OpenEnvironment",
 CellID->67737875],
Cell[CellGroupData[{
Cell[1175366, 32978, 380, 9, 70, "EnvironmentHeader",
 CellID->299731454],
Cell[1175749, 32989, 738, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:235178571", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->235178571],
Cell[1176490, 33010, 1193, 33, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:474434374", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C 2"},
 CellID->474434374],
Cell[1177686, 33045, 763, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:108633346", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->108633346],
Cell[1178452, 33067, 1231, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:611741232", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C +"},
 CellID->611741232],
Cell[1179686, 33104, 1084, 31, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:438277340", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C - unary"},
 CellID->438277340],
Cell[1180773, 33137, 1180, 34, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:412638768", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples C -"},
 CellID->412638768],
Cell[1181956, 33173, 65, 1, 70, "EndEnvironmentMarker",
 CellID->166044720]
}, Open  ]],
Cell[1182036, 33177, 48, 1, 70, "CloseEnvironment",
 CellID->203713653],
Cell[CellGroupData[{
Cell[1182109, 33182, 3970, 106, 70, "ProofInfo",
 CellTags->{"Proof|ID:474434374", "Proof|ID:474434374-1"},
 CellID->85632222],
Cell[1186082, 33290, 480, 12, 70, "ProofInfoBody",
 CellID->518387433]
}, Closed]],
Cell[CellGroupData[{
Cell[1186599, 33307, 6115, 154, 70, "ProofInfo",
 CellTags->{"Proof|ID:611741232", "Proof|ID:611741232-1"},
 CellID->352968606],
Cell[1192717, 33463, 480, 12, 70, "ProofInfoBody",
 CellID->576517268]
}, Closed]],
Cell[CellGroupData[{
Cell[1193234, 33480, 4791, 124, 70, "ProofInfo",
 CellTags->{"Proof|ID:438277340", "Proof|ID:438277340-1"},
 CellID->232831160],
Cell[1198028, 33606, 480, 12, 70, "ProofInfoBody",
 CellID->159284763]
}, Closed]],
Cell[CellGroupData[{
Cell[1198545, 33623, 6115, 154, 70, "ProofInfo",
 CellTags->{"Proof|ID:412638768", "Proof|ID:412638768-1"},
 CellID->382215156],
Cell[1204663, 33779, 480, 12, 70, "ProofInfoBody",
 CellID->149522829]
}, Closed]],
Cell[1205158, 33794, 47, 1, 70, "OpenEnvironment",
 CellID->107107139],
Cell[CellGroupData[{
Cell[1205230, 33799, 399, 9, 70, "EnvironmentHeader",
 CellID->105654882],
Cell[1205632, 33810, 838, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:392022323", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasCoefficientFunction"},
 CellID->392022323],
Cell[1206473, 33833, 65, 1, 70, "EndEnvironmentMarker",
 CellID->154812201]
}, Open  ]],
Cell[1206553, 33837, 47, 1, 70, "CloseEnvironment",
 CellID->77149556],
Cell[CellGroupData[{
Cell[1206625, 33842, 2012, 60, 70, "ProofInfo",
 CellTags->{"Proof|ID:392022323", "Proof|ID:392022323-1"},
 CellID->43251408],
Cell[1208640, 33904, 479, 12, 70, "ProofInfoBody",
 CellID->19455043]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1209168, 33922, 195, 3, 70, "Subsubsection",
 CellID->196031232],
Cell[1209366, 33927, 47, 1, 70, "OpenEnvironment",
 CellID->557047419],
Cell[CellGroupData[{
Cell[1209438, 33932, 387, 9, 70, "EnvironmentHeader",
 CellID->240207582],
Cell[1209828, 33943, 751, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:23554005", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->23554005],
Cell[1210582, 33964, 992, 26, 70, "GlobalDeclaration",
 CellTags->{
  "ID:392464903", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->392464903],
Cell[1211577, 33992, 1238, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:161672314", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] empty"},
 CellID->161672314],
Cell[1212818, 34026, 3855, 97, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:71703914", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] prepend"},
 CellID->71703914],
Cell[1216676, 34125, 738, 19, 70, "GlobalDeclaration",
 CellTags->{
  "ID:127729286", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->127729286],
Cell[1217417, 34146, 1346, 35, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:2718947", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] lm"},
 CellID->2718947],
Cell[1218766, 34183, 1210, 32, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:204263407", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples membership \[CircleTimes]"},
 CellID->204263407],
Cell[1219979, 34217, 1539, 38, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:134536220", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[CircleTimes] \[RawStar]"},
 CellID->134536220],
Cell[1221521, 34257, 1604, 42, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:387729163", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[RawStar] alternative"},
 CellID->387729163],
Cell[1223128, 34301, 64, 1, 70, "EndEnvironmentMarker",
 CellID->90621649]
}, Open  ]],
Cell[1223207, 34305, 48, 1, 70, "CloseEnvironment",
 CellID->483115236],
Cell[CellGroupData[{
Cell[1223280, 34310, 5448, 141, 70, "ProofInfo",
 CellTags->{"Proof|ID:161672314", "Proof|ID:161672314-1"},
 CellID->221366480],
Cell[1228731, 34453, 480, 12, 70, "ProofInfoBody",
 CellID->341310380]
}, Closed]],
Cell[CellGroupData[{
Cell[1229248, 34470, 14576, 326, 70, "ProofInfo",
 CellTags->{"Proof|ID:71703914", "Proof|ID:71703914-1"},
 CellID->82366827],
Cell[1243827, 34798, 479, 12, 70, "ProofInfoBody",
 CellID->220781038]
}, Closed]],
Cell[CellGroupData[{
Cell[1244343, 34815, 7241, 177, 70, "ProofInfo",
 CellTags->{"Proof|ID:2718947", "Proof|ID:2718947-1"},
 CellID->181253020],
Cell[1251587, 34994, 522, 13, 70, "ProofInfoBody",
 CellID->302282372]
}, Closed]],
Cell[CellGroupData[{
Cell[1252146, 35012, 6423, 160, 70, "ProofInfo",
 CellTags->{"Proof|ID:204263407", "Proof|ID:204263407-1"},
 CellID->426944387],
Cell[1258572, 35174, 479, 12, 70, "ProofInfoBody",
 CellID->74746441]
}, Closed]],
Cell[CellGroupData[{
Cell[1259088, 35191, 7215, 176, 70, "ProofInfo",
 CellTags->{"Proof|ID:134536220", "Proof|ID:134536220-1"},
 CellID->170925754],
Cell[1266306, 35369, 480, 12, 70, "ProofInfoBody",
 CellID->103658595]
}, Closed]],
Cell[CellGroupData[{
Cell[1266823, 35386, 8927, 214, 70, "ProofInfo",
 CellTags->{"Proof|ID:387729163", "Proof|ID:387729163-1"},
 CellID->722597358],
Cell[1275753, 35602, 479, 12, 70, "ProofInfoBody",
 CellID->89283540]
}, Closed]],
Cell[1276247, 35617, 45, 1, 70, "OpenEnvironment",
 CellID->8782086],
Cell[CellGroupData[{
Cell[1276317, 35622, 395, 9, 70, "EnvironmentHeader",
 CellID->208907366],
Cell[1276715, 35633, 816, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:65280633", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasMultiplication"},
 CellID->65280633],
Cell[1277534, 35656, 65, 1, 70, "EndEnvironmentMarker",
 CellID->606425772]
}, Open  ]],
Cell[1277614, 35660, 46, 1, 70, "CloseEnvironment",
 CellID->6209804],
Cell[CellGroupData[{
Cell[1277685, 35665, 1980, 59, 70, "ProofInfo",
 CellTags->{"Proof|ID:65280633", "Proof|ID:65280633-1"},
 CellID->187277372],
Cell[1279668, 35726, 479, 12, 70, "ProofInfoBody",
 CellID->355668730]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1280196, 35744, 192, 3, 70, "Subsubsection",
 CellID->40538141],
Cell[1280391, 35749, 46, 1, 70, "OpenEnvironment",
 CellID->87454831],
Cell[CellGroupData[{
Cell[1280462, 35754, 376, 9, 70, "EnvironmentHeader",
 CellID->164838296],
Cell[1280841, 35765, 2804, 92, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:48571309", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples \[ForAll] alternative 2"},
 CellID->48571309],
Cell[1283648, 35859, 763, 20, 70, "GlobalDeclaration",
 CellTags->{
  "ID:617696571", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->617696571],
Cell[1284414, 35881, 1490, 47, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:776548", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples=empty sufficient"},
 CellID->776548],
Cell[1285907, 35930, 1542, 49, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:304028020", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples extensional"},
 CellID->304028020],
Cell[1287452, 35981, 65, 1, 70, "EndEnvironmentMarker",
 CellID->121515229]
}, Open  ]],
Cell[1287532, 35985, 47, 1, 70, "CloseEnvironment",
 CellID->85801361],
Cell[CellGroupData[{
Cell[1287604, 35990, 7486, 185, 70, "ProofInfo",
 CellTags->{"Proof|ID:48571309", "Proof|ID:48571309-1"},
 CellID->399659369],
Cell[1295093, 36177, 479, 12, 70, "ProofInfoBody",
 CellID->164181676]
}, Closed]],
Cell[CellGroupData[{
Cell[1295609, 36194, 4243, 114, 70, "ProofInfo",
 CellTags->{"Proof|ID:776548", "Proof|ID:776548-1"},
 CellID->379176191],
Cell[1299855, 36310, 477, 12, 70, "ProofInfoBody",
 CellID->227491426]
}, Closed]],
Cell[CellGroupData[{
Cell[1300369, 36327, 5393, 140, 70, "ProofInfo",
 CellTags->{"Proof|ID:304028020", "Proof|ID:304028020-1"},
 CellID->86924546],
Cell[1305765, 36469, 480, 12, 70, "ProofInfoBody",
 CellID->131380879]
}, Closed]],
Cell[1306260, 36484, 47, 1, 70, "OpenEnvironment",
 CellID->186415220],
Cell[CellGroupData[{
Cell[1306332, 36489, 442, 10, 70, "EnvironmentHeader",
 CellID->193487632],
Cell[1306777, 36501, 811, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:589891585", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isExtensional"},
 CellID->589891585],
Cell[1307591, 36524, 64, 1, 70, "EndEnvironmentMarker",
 CellID->54820822]
}, Open  ]],
Cell[1307670, 36528, 48, 1, 70, "CloseEnvironment",
 CellID->232514858],
Cell[CellGroupData[{
Cell[1307743, 36533, 1961, 58, 70, "ProofInfo",
 CellTags->{"Proof|ID:589891585", "Proof|ID:589891585-1"},
 CellID->116448114],
Cell[1309707, 36593, 480, 12, 70, "ProofInfoBody",
 CellID->332438606]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1310236, 36611, 192, 3, 70, "Subsubsection",
 CellID->69393044],
Cell[1310431, 36616, 47, 1, 70, "OpenEnvironment",
 CellID->263183869],
Cell[CellGroupData[{
Cell[1310503, 36621, 384, 9, 70, "EnvironmentHeader",
 CellID->39919617],
Cell[1310890, 36632, 806, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:160815443", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isOrderedRec"},
 CellID->160815443],
Cell[1311699, 36655, 795, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:397107848", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isOrdered"},
 CellID->397107848],
Cell[1312497, 36678, 64, 1, 70, "EndEnvironmentMarker",
 CellID->46796859]
}, Open  ]],
Cell[1312576, 36682, 47, 1, 70, "CloseEnvironment",
 CellID->17008929],
Cell[CellGroupData[{
Cell[1312648, 36687, 1955, 58, 70, "ProofInfo",
 CellTags->{"Proof|ID:160815443", "Proof|ID:160815443-1"},
 CellID->9775189],
Cell[1314606, 36747, 480, 12, 70, "ProofInfoBody",
 CellID->144892605]
}, Closed]],
Cell[CellGroupData[{
Cell[1315123, 36764, 1943, 58, 70, "ProofInfo",
 CellTags->{"Proof|ID:397107848", "Proof|ID:397107848-1"},
 CellID->3787871],
Cell[1317069, 36824, 480, 12, 70, "ProofInfoBody",
 CellID->521252860]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1317598, 36842, 190, 3, 70, "Subsubsection",
 CellID->11061911],
Cell[1317791, 36847, 46, 1, 70, "OpenEnvironment",
 CellID->55195651],
Cell[CellGroupData[{
Cell[1317862, 36852, 390, 9, 70, "EnvironmentHeader",
 CellID->186389390],
Cell[1318255, 36863, 812, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:156782366", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasMultipliers"},
 CellID->156782366],
Cell[1319070, 36886, 64, 1, 70, "EndEnvironmentMarker",
 CellID->56161024]
}, Open  ]],
Cell[1319149, 36890, 48, 1, 70, "CloseEnvironment",
 CellID->666198591],
Cell[CellGroupData[{
Cell[1319222, 36895, 1965, 58, 70, "ProofInfo",
 CellTags->{"Proof|ID:156782366", "Proof|ID:156782366-1"},
 CellID->558592865],
Cell[1321190, 36955, 479, 12, 70, "ProofInfoBody",
 CellID->82438810]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1321718, 36973, 191, 3, 70, "Subsubsection",
 CellID->78809743],
Cell[1321912, 36978, 47, 1, 70, "OpenEnvironment",
 CellID->149984650],
Cell[CellGroupData[{
Cell[1321984, 36983, 402, 9, 70, "EnvironmentHeader",
 CellID->157312756],
Cell[1322389, 36994, 860, 22, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:667417187", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isReductionPolynomialDomain"},
 CellID->667417187],
Cell[1323252, 37018, 65, 1, 70, "EndEnvironmentMarker",
 CellID->117448790]
}, Open  ]],
Cell[1323332, 37022, 47, 1, 70, "CloseEnvironment",
 CellID->82630341],
Cell[CellGroupData[{
Cell[1323404, 37027, 2033, 60, 70, "ProofInfo",
 CellTags->{"Proof|ID:667417187", "Proof|ID:667417187-1"},
 CellID->486349173],
Cell[1325440, 37089, 479, 12, 70, "ProofInfoBody",
 CellID->78869945]
}, Closed]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1325980, 37108, 215, 5, 30, "Subsection",
 CellID->9669308],
Cell[CellGroupData[{
Cell[1326220, 37117, 182, 3, 28, "Subsubsection",
 CellID->375801328],
Cell[1326405, 37122, 47, 1, 70, "OpenEnvironment",
 CellID->325998897],
Cell[CellGroupData[{
Cell[1326477, 37127, 391, 9, 70, "EnvironmentHeader",
 CellID->90125590],
Cell[1326871, 37138, 967, 26, 70, "GlobalDeclaration",
 CellTags->{
  "ID:714775050", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb"},
 CellID->714775050],
Cell[1327841, 37166, 4116, 109, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:15424584", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd alternative"},
 CellID->15424584],
Cell[1331960, 37277, 1655, 47, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:168826371", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd unary alternative"},
 CellID->168826371],
Cell[1333618, 37326, 4778, 122, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:228703376", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd lc lp"},
 CellID->228703376],
Cell[1338399, 37450, 2404, 66, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:354399952", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples lcrd length"},
 CellID->354399952],
Cell[1340806, 37518, 64, 1, 70, "EndEnvironmentMarker",
 CellID->13100681]
}, Open  ]],
Cell[1340885, 37522, 48, 1, 70, "CloseEnvironment",
 CellID->640374478],
Cell[CellGroupData[{
Cell[1340958, 37527, 19212, 427, 70, "ProofInfo",
 CellTags->{"Proof|ID:15424584", "Proof|ID:15424584-1"},
 CellID->190628320],
Cell[1360173, 37956, 544, 13, 70, "ProofInfoBody",
 CellID->124361579]
}, Closed]],
Cell[CellGroupData[{
Cell[1360754, 37974, 7200, 178, 70, "ProofInfo",
 CellTags->{"Proof|ID:168826371", "Proof|ID:168826371-1"},
 CellID->493554176],
Cell[1367957, 38154, 479, 12, 70, "ProofInfoBody",
 CellID->11398291]
}, Closed]],
Cell[CellGroupData[{
Cell[1368473, 38171, 18754, 414, 70, "ProofInfo",
 CellTags->{"Proof|ID:228703376", "Proof|ID:228703376-1"},
 CellID->48977005],
Cell[1387230, 38587, 548, 13, 70, "ProofInfoBody",
 CellID->1819865]
}, Closed]],
Cell[CellGroupData[{
Cell[1387815, 38605, 12333, 279, 70, "ProofInfo",
 CellTags->{"Proof|ID:354399952", "Proof|ID:354399952-1"},
 CellID->182853122],
Cell[1400151, 38886, 480, 12, 70, "ProofInfoBody",
 CellID->237714126]
}, Closed]],
Cell[1400646, 38901, 47, 1, 70, "OpenEnvironment",
 CellID->187246246],
Cell[CellGroupData[{
Cell[1400718, 38906, 383, 9, 70, "EnvironmentHeader",
 CellID->332568537],
Cell[1401104, 38917, 790, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:479598778", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasLCRD"},
 CellID->479598778],
Cell[1401897, 38940, 65, 1, 70, "EndEnvironmentMarker",
 CellID->324782272]
}, Open  ]],
Cell[1401977, 38944, 48, 1, 70, "CloseEnvironment",
 CellID->144574756],
Cell[CellGroupData[{
Cell[1402050, 38949, 1937, 58, 70, "ProofInfo",
 CellTags->{"Proof|ID:479598778", "Proof|ID:479598778-1"},
 CellID->690537389],
Cell[1403990, 39009, 480, 12, 70, "ProofInfoBody",
 CellID->458247862]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1404519, 39027, 181, 3, 22, "Subsubsection",
 CellID->31285283],
Cell[1404703, 39032, 47, 1, 70, "OpenEnvironment",
 CellID->169886156],
Cell[CellGroupData[{
Cell[1404775, 39037, 382, 9, 70, "EnvironmentHeader",
 CellID->922306034],
Cell[1405160, 39048, 787, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:463361785", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasRDM"},
 CellID->463361785],
Cell[1405950, 39071, 65, 1, 70, "EndEnvironmentMarker",
 CellID->133661739]
}, Open  ]],
Cell[1406030, 39075, 48, 1, 70, "CloseEnvironment",
 CellID->121927638],
Cell[CellGroupData[{
Cell[1406103, 39080, 1933, 58, 70, "ProofInfo",
 CellTags->{"Proof|ID:463361785", "Proof|ID:463361785-1"},
 CellID->260093492],
Cell[1408039, 39140, 480, 12, 70, "ProofInfoBody",
 CellID->130872278]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1408568, 39158, 180, 3, 22, "Subsubsection",
 CellID->5046155],
Cell[1408751, 39163, 45, 1, 70, "OpenEnvironment",
 CellID->8219039],
Cell[CellGroupData[{
Cell[1408821, 39168, 384, 9, 70, "EnvironmentHeader",
 CellID->180390280],
Cell[1409208, 39179, 783, 21, 70, "FormalTextInputFormula",
 CellTags->{
  "ID:7345776", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples hasCPM"},
 CellID->7345776],
Cell[1409994, 39202, 65, 1, 70, "EndEnvironmentMarker",
 CellID->593022431]
}, Open  ]],
Cell[1410074, 39206, 48, 1, 70, "CloseEnvironment",
 CellID->276041390],
Cell[CellGroupData[{
Cell[1410147, 39211, 1922, 58, 70, "ProofInfo",
 CellTags->{"Proof|ID:7345776", "Proof|ID:7345776-1"},
 CellID->11260575],
Cell[1412072, 39271, 478, 12, 70, "ProofInfoBody",
 CellID->419962729]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[1412599, 39289, 190, 3, 22, "Subsubsection",
 CellID->2896086],
Cell[1412792, 39294, 47, 1, 12, "OpenEnvironment",
 CellID->124935155],
Cell[CellGroupData[{
Cell[1412864, 39299, 414, 10, 43, "EnvironmentHeader",
 CellID->572686989],
Cell[1413281, 39311, 875, 22, 21, "FormalTextInputFormula",
 CellTags->{
  "ID:1014777887", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isAlgoReductionPolynomialDomain"},
 CellID->1014777887],
Cell[1414159, 39335, 65, 1, 26, "EndEnvironmentMarker",
 CellID->253163528]
}, Open  ]],
Cell[1414239, 39339, 48, 1, 16, "CloseEnvironment",
 CellID->268518370],
Cell[CellGroupData[{
Cell[1414312, 39344, 2057, 61, 28, "ProofInfo",
 CellTags->{"Proof|ID:1014777887", "Proof|ID:1014777887-1"},
 CellID->68078823],
Cell[1416372, 39407, 481, 12, 70, "ProofInfoBody",
 CellID->423181797]
}, Closed]],
Cell[1416868, 39422, 46, 1, 12, "OpenEnvironment",
 CellID->48219384],
Cell[CellGroupData[{
Cell[1416939, 39427, 404, 9, 43, "EnvironmentHeader",
 CellID->190868484],
Cell[1417346, 39438, 1542, 41, 78, "FormalTextInputFormula",
 CellTags->{
  "ID:67407018", 
   "Source:/home/amaletzk/Theorema Notebooks/Reduction Ring \
Theory/PolyTuples.nb", "PolyTuples isAlgoReductionRing"},
 CellID->67407018],
Cell[1418891, 39481, 64, 1, 26, "EndEnvironmentMarker",
 CellID->59995559]
}, Open  ]],
Cell[1418970, 39485, 48, 1, 16, "CloseEnvironment",
 CellID->628697588],
Cell[CellGroupData[{
Cell[1419043, 39490, 3588, 95, 28, "ProofInfo",
 CellTags->{"Proof|ID:67407018", "Proof|ID:67407018-1"},
 CellID->163016488],
Cell[1422634, 39587, 479, 12, 70, "ProofInfoBody",
 CellID->310436971]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
